/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 3.0.8
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */


#ifndef SWIGPYTHON
#define SWIGPYTHON
#endif

#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#  ifndef GCC_HASCLASSVISIBILITY
#    define GCC_HASCLASSVISIBILITY
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif


#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */
# undef _DEBUG
# include <Python.h>
# define _DEBUG
#else
# include <Python.h>
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = ((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = ((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define PyString_AS_STRING(x) PyUnicode_AS_STRING(x)
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Warning: This function will allocate a new string in Python 3,
 * so please call SWIG_Python_str_DelForPy3(x) to free the space.
 */
SWIGINTERN char*
SWIG_Python_str_AsChar(PyObject *str)
{
#if PY_VERSION_HEX >= 0x03000000
  char *cstr;
  char *newstr;
  Py_ssize_t len;
  str = PyUnicode_AsUTF8String(str);
  PyBytes_AsStringAndSize(str, &cstr, &len);
  newstr = (char *) malloc(len+1);
  memcpy(newstr, cstr, len+1);
  Py_XDECREF(str);
  return newstr;
#else
  return PyString_AsString(str);
#endif
}

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_DelForPy3(x) free( (void*) (x) )
#else
#  define SWIG_Python_str_DelForPy3(x) 
#endif


SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

/* Add PyOS_snprintf for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# if defined(_MSC_VER) || defined(__BORLANDC__) || defined(_WATCOM)
#  define PyOS_snprintf _snprintf
# else
#  define PyOS_snprintf snprintf
# endif
#endif

/* A crude PyString_FromFormat implementation for old Pythons */
#if PY_VERSION_HEX < 0x02020000

#ifndef SWIG_PYBUFFER_SIZE
# define SWIG_PYBUFFER_SIZE 1024
#endif

static PyObject *
PyString_FromFormat(const char *fmt, ...) {
  va_list ap;
  char buf[SWIG_PYBUFFER_SIZE * 2];
  int res;
  va_start(ap, fmt);
  res = vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);
  return (res < 0 || res >= (int)sizeof(buf)) ? 0 : PyString_FromString(buf);
}
#endif

/* Add PyObject_Del for old Pythons */
#if PY_VERSION_HEX < 0x01060000
# define PyObject_Del(op) PyMem_DEL((op))
#endif
#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

/* A crude PyExc_StopIteration exception for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# ifndef PyExc_StopIteration
#  define PyExc_StopIteration PyExc_RuntimeError
# endif
# ifndef PyObject_GenericGetAttr
#  define PyObject_GenericGetAttr 0
# endif
#endif

/* Py_NotImplemented is defined in 2.1 and up. */
#if PY_VERSION_HEX < 0x02010000
# ifndef Py_NotImplemented
#  define Py_NotImplemented PyExc_RuntimeError
# endif
#endif

/* A crude PyString_AsStringAndSize implementation for old Pythons */
#if PY_VERSION_HEX < 0x02010000
# ifndef PyString_AsStringAndSize
#  define PyString_AsStringAndSize(obj, s, len) {*s = PyString_AsString(obj); *len = *s ? strlen(*s) : 0;}
# endif
#endif

/* PySequence_Size for old Pythons */
#if PY_VERSION_HEX < 0x02000000
# ifndef PySequence_Size
#  define PySequence_Size PySequence_Length
# endif
#endif

/* PyBool_FromLong for old Pythons */
#if PY_VERSION_HEX < 0x02030000
static
PyObject *PyBool_FromLong(long ok)
{
  PyObject *result = ok ? Py_True : Py_False;
  Py_INCREF(result);
  return result;
}
#endif

/* Py_ssize_t for old Pythons */
/* This code is as recommended by: */
/* http://www.python.org/dev/peps/pep-0353/#conversion-guidelines */
#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
typedef int Py_ssize_t;
# define PY_SSIZE_T_MAX INT_MAX
# define PY_SSIZE_T_MIN INT_MIN
typedef inquiry lenfunc;
typedef intargfunc ssizeargfunc;
typedef intintargfunc ssizessizeargfunc;
typedef intobjargproc ssizeobjargproc;
typedef intintobjargproc ssizessizeobjargproc;
typedef getreadbufferproc readbufferproc;
typedef getwritebufferproc writebufferproc;
typedef getsegcountproc segcountproc;
typedef getcharbufferproc charbufferproc;
static long PyNumber_AsSsize_t (PyObject *x, void *SWIGUNUSEDPARM(exc))
{
  long result = 0;
  PyObject *i = PyNumber_Int(x);
  if (i) {
    result = PyInt_AsLong(i);
    Py_DECREF(i);
  }
  return result;
}
#endif

#if PY_VERSION_HEX < 0x02050000
#define PyInt_FromSize_t(x) PyInt_FromLong((long)x)
#endif

#if PY_VERSION_HEX < 0x02040000
#define Py_VISIT(op)				\
  do { 						\
    if (op) {					\
      int vret = visit((op), arg);		\
      if (vret)					\
        return vret;				\
    }						\
  } while (0)
#endif

#if PY_VERSION_HEX < 0x02030000
typedef struct {
  PyTypeObject type;
  PyNumberMethods as_number;
  PyMappingMethods as_mapping;
  PySequenceMethods as_sequence;
  PyBufferProcs as_buffer;
  PyObject *name, *slots;
} PyHeapTypeObject;
#endif

#if PY_VERSION_HEX < 0x02030000
typedef destructor freefunc;
#endif

#if ((PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION > 6) || \
     (PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION > 0) || \
     (PY_MAJOR_VERSION > 3))
# define SWIGPY_USE_CAPSULE
# define SWIGPY_CAPSULE_NAME ((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION ".type_pointer_capsule" SWIG_TYPE_TABLE_NAME)
#endif

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred()) PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    char *tmp;
    PyObject *old_str = PyObject_Str(value);
    PyErr_Clear();
    Py_XINCREF(type);

    PyErr_Format(type, "%s %s", tmp = SWIG_Python_str_AsChar(old_str), mesg);
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    if (PY_VERSION_HEX >= 0x02030000) /* For 2.3 or later, use the PyGILState calls */
#      define SWIG_PYTHON_USE_GIL
#    endif
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;


/* -----------------------------------------------------------------------------
 * Wrapper of PyInstanceMethod_New() used in Python 3
 * It is exported to the generated module, used for -fastproxy
 * ----------------------------------------------------------------------------- */
#if PY_VERSION_HEX >= 0x03000000
SWIGRUNTIME PyObject* SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func)
{
  return PyInstanceMethod_New(func);
}
#else
SWIGRUNTIME PyObject* SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *SWIGUNUSEDPARM(func))
{
  return NULL;
}
#endif

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
#if PY_VERSION_HEX < 0x02030000
  PyDict_SetItemString(d, (char *)name, obj);
#else
  PyDict_SetItemString(d, name, obj);
#endif
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
#if PY_VERSION_HEX < 0x02030000
  PyDict_SetItemString(d, (char *)name, obj);
#else
  PyDict_SetItemString(d, name, obj);
#endif
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
#if !defined(SWIG_PYTHON_OUTPUT_TUPLE)
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
#else
  PyObject*   o2;
  PyObject*   o3;
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyTuple_Check(result)) {
      o2 = result;
      result = PyTuple_New(1);
      PyTuple_SET_ITEM(result, 0, o2);
    }
    o3 = PyTuple_New(1);
    PyTuple_SET_ITEM(o3, 0, obj);
    o2 = result;
    result = PySequence_Concat(o2, o3);
    Py_DECREF(o2);
    Py_DECREF(o3);
  }
  return result;
#endif
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

/* A functor is a function object with one single object argument */
#if PY_VERSION_HEX >= 0x02020000
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);
#else
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunction(functor, "O", obj);
#endif

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/*  How to access Py_None */
#if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#  ifndef SWIG_PYTHON_NO_BUILD_NONE
#    ifndef SWIG_PYTHON_BUILD_NONE
#      define SWIG_PYTHON_BUILD_NONE
#    endif
#  endif
#endif

#ifdef SWIG_PYTHON_BUILD_NONE
#  ifdef Py_None
#   undef Py_None
#   define Py_None SWIG_Py_None()
#  endif
SWIGRUNTIMEINLINE PyObject * 
_SWIG_Py_None(void)
{
  PyObject *none = Py_BuildValue((char*)"");
  Py_DECREF(none);
  return none;
}
SWIGRUNTIME PyObject * 
SWIG_Py_None(void)
{
  static PyObject *SWIG_STATIC_POINTER(none) = _SWIG_Py_None();
  return none;
}
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  return data ? data->implicitconv : 0;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
#if (PY_VERSION_HEX < 0x02020000)
      data->newraw = 0;
#else
      data->newraw = PyObject_GetAttrString(data->klass, (char *)"__new__");
#endif
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, (char *)"__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
#ifdef METH_O
      data->delargs = !(flags & (METH_O));
#else
      data->delargs = 0;
#endif
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data) {
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->dict)
    sobj->dict = PyDict_New();

  Py_INCREF(sobj->dict);
  return sobj->dict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, SwigPyObject_long(v)) == 0) {
      PyObject *ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
	res = PyUnicode_Format(ofmt,args);
#else
	res = PyString_Format(ofmt,args);
#endif
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
#ifdef METH_NOARGS
SwigPyObject_repr(SwigPyObject *v)
#else
SwigPyObject_repr(SwigPyObject *v, PyObject *args)
#endif
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (v->next) {
# ifdef METH_NOARGS
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
# else
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next, args);
# endif
# if PY_VERSION_HEX >= 0x03000000
    PyObject *joined = PyUnicode_Concat(repr, nrep);
    Py_DecRef(repr);
    Py_DecRef(nrep);
    repr = joined;
# else
    PyString_ConcatAndDel(&repr,nrep);
# endif
  }
  return repr;  
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res;
  if( op != Py_EQ && op != Py_NE ) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
#ifdef SWIGPYTHON_BUILTIN
  PyTypeObject *target_tp = SwigPyObject_type();
  if (PyType_IsSubtype(op->ob_type, target_tp))
    return 1;
  return (strcmp(op->ob_type->tp_name, "SwigPyObject") == 0);
#else
  return (Py_TYPE(op) == SwigPyObject_type())
    || (strcmp(Py_TYPE(op)->tp_name,"SwigPyObject") == 0);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *val = NULL, *type = NULL, *tb = NULL;
      PyErr_Fetch(&val, &type, &tb);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        res = SWIG_Python_CallFunctor(destroy, tmp);
        Py_DECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(val, type, tb);

      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
#ifndef METH_O
  PyObject *tmp = 0;
  if (!PyArg_ParseTuple(next,(char *)"O:append", &tmp)) return NULL;
  next = tmp;
#endif
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
#ifdef METH_NOARGS
SwigPyObject_next(PyObject* v)
#else
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
SwigPyObject_disown(PyObject *v)
#else
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
SwigPyObject_acquire(PyObject *v)
#else
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
#if (PY_VERSION_HEX < 0x02020000)
  if (!PyArg_ParseTuple(args,(char *)"|O:own",&val))
#elif (PY_VERSION_HEX < 0x02050000)
  if (!PyArg_UnpackTuple(args, (char *)"own", 0, 1, &val)) 
#else
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) 
#endif
    {
      return NULL;
    } 
  else
    {
      SwigPyObject *sobj = (SwigPyObject *)v;
      PyObject *obj = PyBool_FromLong(sobj->own);
      if (val) {
#ifdef METH_NOARGS
	if (PyObject_IsTrue(val)) {
	  SwigPyObject_acquire(v);
	} else {
	  SwigPyObject_disown(v);
	}
#else
	if (PyObject_IsTrue(val)) {
	  SwigPyObject_acquire(v,args);
	} else {
	  SwigPyObject_disown(v,args);
	}
#endif
      } 
      return obj;
    }
}

#ifdef METH_O
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)SwigPyObject_disown,  METH_NOARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)SwigPyObject_acquire, METH_NOARGS,  (char *)"acquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)SwigPyObject_own,     METH_VARARGS, (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)SwigPyObject_append,  METH_O,       (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)SwigPyObject_next,    METH_NOARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)SwigPyObject_repr,    METH_NOARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#else
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)SwigPyObject_disown,  METH_VARARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)SwigPyObject_acquire, METH_VARARGS,  (char *)"acquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)SwigPyObject_own,     METH_VARARGS,  (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)SwigPyObject_append,  METH_VARARGS,  (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)SwigPyObject_next,    METH_VARARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)SwigPyObject_repr,   METH_VARARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#endif

#if PY_VERSION_HEX < 0x02020000
SWIGINTERN PyObject *
SwigPyObject_getattr(SwigPyObject *sobj,char *name)
{
  return Py_FindMethod(swigobject_methods, (PyObject *)sobj, name);
}
#endif

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";

  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#elif PY_VERSION_HEX >= 0x02050000 /* 2.5.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#elif PY_VERSION_HEX >= 0x02020000 /* 2.2.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_true_divide */
#elif PY_VERSION_HEX >= 0x02000000 /* 2.0.0 */
    0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_or */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
      /* PyObject header changed in Python 3 */
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      (char *)"SwigPyObject",               /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
      0,				    /* tp_print */
#if PY_VERSION_HEX < 0x02020000
      (getattrfunc)SwigPyObject_getattr,    /* tp_getattr */
#else
      (getattrfunc)0,                       /* tp_getattr */
#endif
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,				    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
      0,                                    /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
      0,                                    /* tp_version_tag */
#endif
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
#if PY_VERSION_HEX >= 0x02050000
      0,                                    /* tp_prev */
#endif
      0                                     /* tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
#if PY_VERSION_HEX < 0x02020000
    swigpyobject_type.ob_type = &PyType_Type;
#else
    if (PyType_Ready(&swigpyobject_type) < 0)
      return NULL;
#endif
  }
  return &swigpyobject_type;
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_NEW(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME int
SwigPyPacked_print(SwigPyPacked *v, FILE *fp, int SWIGUNUSEDPARM(flags))
{
  char result[SWIG_BUFFER_SIZE];
  fputs("<Swig Packed ", fp); 
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    fputs("at ", fp); 
    fputs(result, fp); 
  }
  fputs(v->ty->name,fp); 
  fputs(">", fp);
  return 0; 
}
  
SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((char *)v->pack, (char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  return ((op)->ob_type == SwigPyPacked_TypeOnce()) 
    || (strcmp((op)->ob_type->tp_name,"SwigPyPacked") == 0);
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
      /* PyObject header changed in Python 3 */
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      (char *)"SwigPyPacked",               /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
      (printfunc)SwigPyPacked_print,        /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
      0,                                    /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
      0,                                    /* tp_version_tag */
#endif
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
#if PY_VERSION_HEX >= 0x02050000
      0,                                    /* tp_prev */
#endif
      0                                     /* tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
#if PY_VERSION_HEX < 0x02020000
    swigpypacked_type.ob_type = &PyType_Type;
#else
    if (PyType_Ready(&swigpypacked_type) < 0)
      return NULL;
#endif
  }
  return &swigpypacked_type;
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_NEW(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIMEINLINE PyObject *
_SWIG_This(void)
{
    return SWIG_Python_str_FromChar("this");
}

static PyObject *swig_this = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (swig_this == NULL)
    swig_this = _SWIG_This();
  return swig_this;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if (!defined(SWIG_PYTHON_SLOW_GETSET_THIS) && (PY_VERSION_HEX >= 0x02030000))
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (own)
      *own = *own | sobj->own;
    if (flags & SWIG_POINTER_DISOWN) {
      sobj->own = 0;
    }
    res = SWIG_OK;
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
    }
    if (!SWIG_IsOK(res) && obj == Py_None) {
      if (ptr)
        *ptr = 0;
      if (PyErr_Occurred())
        PyErr_Clear();
      res = SWIG_OK;
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    
    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
    if (!desc) 
      return SWIG_ERROR;
    if (ty) {
      swig_cast_info *tc = SWIG_TypeCheck(desc,ty);
      if (tc) {
        int newmemory = 0;
        *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
        assert(!newmemory); /* newmemory handling not yet implemented */
      } else {
        return SWIG_ERROR;
      }
    } else {
      *ptr = vptr;
    }
    return SWIG_OK;
  }
}

/* Convert a packed value value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
#if (PY_VERSION_HEX >= 0x02020000)
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      PyObject *key = SWIG_This();
      PyObject_SetAttr(inst, key, swig_this);
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    inst = ((PyTypeObject*) data->newargs)->tp_new((PyTypeObject*) data->newargs, Py_None, Py_None);
    if (inst) {
      PyObject_SetAttr(inst, SWIG_This(), swig_this);
      Py_TYPE(inst)->tp_flags &= ~Py_TPFLAGS_VALID_VERSION_TAG;
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
#else
#if (PY_VERSION_HEX >= 0x02010000)
  PyObject *inst = 0;
  PyObject *dict = PyDict_New();
  if (dict) {
    PyDict_SetItem(dict, SWIG_This(), swig_this);
    inst = PyInstance_NewRaw(data->newargs, dict);
    Py_DECREF(dict);
  }
  return (PyObject *) inst;
#else
  PyInstanceObject *inst = PyObject_NEW(PyInstanceObject, &PyInstance_Type);
  if (inst == NULL) {
    return NULL;
  }
  inst->in_class = (PyClassObject *)data->newargs;
  Py_INCREF(inst->in_class);
  inst->in_dict = PyDict_New();
  if (inst->in_dict == NULL) {
    Py_DECREF(inst);
    return NULL;
  }
#ifdef Py_TPFLAGS_HAVE_WEAKREFS
  inst->in_weakreflist = NULL;
#endif
#ifdef Py_TPFLAGS_GC
  PyObject_GC_Init(inst);
#endif
  PyDict_SetItem(inst->in_dict, SWIG_This(), swig_this);
  return (PyObject *) inst;
#endif
#endif
}

SWIGRUNTIME void
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
 PyObject *dict;
#if (PY_VERSION_HEX >= 0x02020000) && !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
 PyObject **dictptr = _PyObject_GetDictPtr(inst);
 if (dictptr != NULL) {
   dict = *dictptr;
   if (dict == NULL) {
     dict = PyDict_New();
     *dictptr = dict;
   }
   PyDict_SetItem(dict, SWIG_This(), swig_this);
   return;
 }
#endif
 dict = PyObject_GetAttrString(inst, (char*)"__dict__");
 PyDict_SetItem(dict, SWIG_This(), swig_this);
 Py_DECREF(dict);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SwigPyObject_append((PyObject*) sthis, obj[1]);
    } else {
      SWIG_Python_SetSwigThis(obj[0], obj[1]);
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
        PyObject *next_self = clientdata->pytype->tp_alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
#ifdef SWIGPYTHON_BUILTIN
        newobj->dict = 0;
#endif
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
#ifdef SWIGPYTHON_BUILTIN
      newobj->dict = 0;
#endif
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
# ifdef SWIGPY_USE_CAPSULE
    type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
# else
    type_pointer = PyCObject_Import((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION,
				    (char*)"type_pointer" SWIG_TYPE_TABLE_NAME);
# endif
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

#if PY_MAJOR_VERSION < 2
/* PyModule_AddObject function was introduced in Python 2.0.  The following function
   is copied out of Python/modsupport.c in python version 2.3.4 */
SWIGINTERN int
PyModule_AddObject(PyObject *m, char *name, PyObject *o)
{
  PyObject *dict;
  if (!PyModule_Check(m)) {
    PyErr_SetString(PyExc_TypeError, "PyModule_AddObject() needs module as first arg");
    return SWIG_ERROR;
  }
  if (!o) {
    PyErr_SetString(PyExc_TypeError, "PyModule_AddObject() needs non-NULL value");
    return SWIG_ERROR;
  }
  
  dict = PyModule_GetDict(m);
  if (dict == NULL) {
    /* Internal error -- modules must have a dict! */
    PyErr_Format(PyExc_SystemError, "module '%s' has no __dict__",
		 PyModule_GetName(m));
    return SWIG_ERROR;
  }
  if (PyDict_SetItemString(dict, name, o))
    return SWIG_ERROR;
  Py_DECREF(o);
  return SWIG_OK;
}
#endif

SWIGRUNTIME void
#ifdef SWIGPY_USE_CAPSULE
SWIG_Python_DestroyModule(PyObject *obj)
#else
SWIG_Python_DestroyModule(void *vptr)
#endif
{
#ifdef SWIGPY_USE_CAPSULE
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
#else
  swig_module_info *swig_module = (swig_module_info *) vptr;
#endif
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  swig_this = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
#ifdef SWIGPY_USE_CAPSULE
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer_capsule" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
#else
  PyObject *pointer = PyCObject_FromVoidPtr((void *) swig_module, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
#endif
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
#ifdef SWIGPY_USE_CAPSULE
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
#else
    descriptor = (swig_type_info *) PyCObject_AsVoidPtr(obj);
#endif
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
#ifdef SWIGPY_USE_CAPSULE
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
#else
      obj = PyCObject_FromVoidPtr(descriptor, NULL);
#endif
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      char *tmp;
      PyObject *old_str = PyObject_Str(value);
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, tmp = SWIG_Python_str_AsChar(old_str));
      } else {
	PyErr_Format(type, "%s %s", tmp = SWIG_Python_str_AsChar(old_str), mesg);
      }
      SWIG_Python_str_DelForPy3(tmp);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? SWIG_Python_str_AsChar(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
          SWIG_Python_str_DelForPy3(cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
#if SWIG_POINTER_EXCEPTION
    if (flags) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
#endif
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) < 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_ByteArray swig_types[0]
#define SWIGTYPE_p_ESYS_CONTEXT swig_types[1]
#define SWIGTYPE_p_ESYS_TR_PTR swig_types[2]
#define SWIGTYPE_p_TPM2B_ATTEST swig_types[3]
#define SWIGTYPE_p_TPM2B_ATTEST_ARRAY swig_types[4]
#define SWIGTYPE_p_TPM2B_CONTEXT_DATA swig_types[5]
#define SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY swig_types[6]
#define SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE swig_types[7]
#define SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY swig_types[8]
#define SWIGTYPE_p_TPM2B_CREATION_DATA swig_types[9]
#define SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY swig_types[10]
#define SWIGTYPE_p_TPM2B_DATA swig_types[11]
#define SWIGTYPE_p_TPM2B_DATA_ARRAY swig_types[12]
#define SWIGTYPE_p_TPM2B_DIGEST swig_types[13]
#define SWIGTYPE_p_TPM2B_DIGEST_ARRAY swig_types[14]
#define SWIGTYPE_p_TPM2B_ECC_PARAMETER swig_types[15]
#define SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY swig_types[16]
#define SWIGTYPE_p_TPM2B_ECC_POINT swig_types[17]
#define SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY swig_types[18]
#define SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET swig_types[19]
#define SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY swig_types[20]
#define SWIGTYPE_p_TPM2B_EVENT swig_types[21]
#define SWIGTYPE_p_TPM2B_EVENT_ARRAY swig_types[22]
#define SWIGTYPE_p_TPM2B_ID_OBJECT swig_types[23]
#define SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY swig_types[24]
#define SWIGTYPE_p_TPM2B_IV swig_types[25]
#define SWIGTYPE_p_TPM2B_IV_ARRAY swig_types[26]
#define SWIGTYPE_p_TPM2B_MAX_BUFFER swig_types[27]
#define SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY swig_types[28]
#define SWIGTYPE_p_TPM2B_MAX_NV_BUFFER swig_types[29]
#define SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY swig_types[30]
#define SWIGTYPE_p_TPM2B_NAME swig_types[31]
#define SWIGTYPE_p_TPM2B_NAME_ARRAY swig_types[32]
#define SWIGTYPE_p_TPM2B_NV_PUBLIC swig_types[33]
#define SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY swig_types[34]
#define SWIGTYPE_p_TPM2B_PRIVATE swig_types[35]
#define SWIGTYPE_p_TPM2B_PRIVATE_ARRAY swig_types[36]
#define SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA swig_types[37]
#define SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY swig_types[38]
#define SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC swig_types[39]
#define SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY swig_types[40]
#define SWIGTYPE_p_TPM2B_PUBLIC swig_types[41]
#define SWIGTYPE_p_TPM2B_PUBLIC_ARRAY swig_types[42]
#define SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA swig_types[43]
#define SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY swig_types[44]
#define SWIGTYPE_p_TPM2B_SENSITIVE swig_types[45]
#define SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY swig_types[46]
#define SWIGTYPE_p_TPM2B_SENSITIVE_CREATE swig_types[47]
#define SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY swig_types[48]
#define SWIGTYPE_p_TPM2B_SENSITIVE_DATA swig_types[49]
#define SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY swig_types[50]
#define SWIGTYPE_p_TPM2B_SYM_KEY swig_types[51]
#define SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY swig_types[52]
#define SWIGTYPE_p_TPM2B_TEMPLATE swig_types[53]
#define SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY swig_types[54]
#define SWIGTYPE_p_TPML_AC_CAPABILITIES swig_types[55]
#define SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY swig_types[56]
#define SWIGTYPE_p_TPML_ALG swig_types[57]
#define SWIGTYPE_p_TPML_ALG_ARRAY swig_types[58]
#define SWIGTYPE_p_TPML_ALG_PROPERTY swig_types[59]
#define SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY swig_types[60]
#define SWIGTYPE_p_TPML_CC swig_types[61]
#define SWIGTYPE_p_TPML_CCA swig_types[62]
#define SWIGTYPE_p_TPML_CCA_ARRAY swig_types[63]
#define SWIGTYPE_p_TPML_CC_ARRAY swig_types[64]
#define SWIGTYPE_p_TPML_DIGEST swig_types[65]
#define SWIGTYPE_p_TPML_DIGEST_ARRAY swig_types[66]
#define SWIGTYPE_p_TPML_DIGEST_VALUES swig_types[67]
#define SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY swig_types[68]
#define SWIGTYPE_p_TPML_ECC_CURVE swig_types[69]
#define SWIGTYPE_p_TPML_ECC_CURVE_ARRAY swig_types[70]
#define SWIGTYPE_p_TPML_HANDLE swig_types[71]
#define SWIGTYPE_p_TPML_HANDLE_ARRAY swig_types[72]
#define SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY swig_types[73]
#define SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY swig_types[74]
#define SWIGTYPE_p_TPML_PCR_SELECTION swig_types[75]
#define SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY swig_types[76]
#define SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY swig_types[77]
#define SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY swig_types[78]
#define SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY swig_types[79]
#define SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY swig_types[80]
#define SWIGTYPE_p_TPMS_AC_OUTPUT swig_types[81]
#define SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY swig_types[82]
#define SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION swig_types[83]
#define SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY swig_types[84]
#define SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC swig_types[85]
#define SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY swig_types[86]
#define SWIGTYPE_p_TPMS_ALG_PROPERTY swig_types[87]
#define SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY swig_types[88]
#define SWIGTYPE_p_TPMS_ASYM_PARMS swig_types[89]
#define SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY swig_types[90]
#define SWIGTYPE_p_TPMS_ATTEST swig_types[91]
#define SWIGTYPE_p_TPMS_ATTEST_ARRAY swig_types[92]
#define SWIGTYPE_p_TPMS_AUTH_COMMAND swig_types[93]
#define SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY swig_types[94]
#define SWIGTYPE_p_TPMS_AUTH_RESPONSE swig_types[95]
#define SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY swig_types[96]
#define SWIGTYPE_p_TPMS_CAPABILITY_DATA swig_types[97]
#define SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY swig_types[98]
#define SWIGTYPE_p_TPMS_CERTIFY_INFO swig_types[99]
#define SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY swig_types[100]
#define SWIGTYPE_p_TPMS_CLOCK_INFO swig_types[101]
#define SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY swig_types[102]
#define SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO swig_types[103]
#define SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY swig_types[104]
#define SWIGTYPE_p_TPMS_CONTEXT swig_types[105]
#define SWIGTYPE_p_TPMS_CONTEXT_ARRAY swig_types[106]
#define SWIGTYPE_p_TPMS_CONTEXT_DATA swig_types[107]
#define SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY swig_types[108]
#define SWIGTYPE_p_TPMS_CREATION_DATA swig_types[109]
#define SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY swig_types[110]
#define SWIGTYPE_p_TPMS_CREATION_INFO swig_types[111]
#define SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY swig_types[112]
#define SWIGTYPE_p_TPMS_ECC_PARMS swig_types[113]
#define SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY swig_types[114]
#define SWIGTYPE_p_TPMS_ECC_POINT swig_types[115]
#define SWIGTYPE_p_TPMS_ECC_POINT_ARRAY swig_types[116]
#define SWIGTYPE_p_TPMS_EMPTY swig_types[117]
#define SWIGTYPE_p_TPMS_EMPTY_ARRAY swig_types[118]
#define SWIGTYPE_p_TPMS_ID_OBJECT swig_types[119]
#define SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY swig_types[120]
#define SWIGTYPE_p_TPMS_KEYEDHASH_PARMS swig_types[121]
#define SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY swig_types[122]
#define SWIGTYPE_p_TPMS_NV_CERTIFY_INFO swig_types[123]
#define SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY swig_types[124]
#define SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS swig_types[125]
#define SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY swig_types[126]
#define SWIGTYPE_p_TPMS_NV_PUBLIC swig_types[127]
#define SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY swig_types[128]
#define SWIGTYPE_p_TPMS_PCR_SELECT swig_types[129]
#define SWIGTYPE_p_TPMS_PCR_SELECTION swig_types[130]
#define SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY swig_types[131]
#define SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY swig_types[132]
#define SWIGTYPE_p_TPMS_QUOTE_INFO swig_types[133]
#define SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY swig_types[134]
#define SWIGTYPE_p_TPMS_RSA_PARMS swig_types[135]
#define SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY swig_types[136]
#define SWIGTYPE_p_TPMS_SCHEME_ECDAA swig_types[137]
#define SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY swig_types[138]
#define SWIGTYPE_p_TPMS_SCHEME_HASH swig_types[139]
#define SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY swig_types[140]
#define SWIGTYPE_p_TPMS_SCHEME_XOR swig_types[141]
#define SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY swig_types[142]
#define SWIGTYPE_p_TPMS_SENSITIVE_CREATE swig_types[143]
#define SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY swig_types[144]
#define SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO swig_types[145]
#define SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY swig_types[146]
#define SWIGTYPE_p_TPMS_SIGNATURE_ECC swig_types[147]
#define SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY swig_types[148]
#define SWIGTYPE_p_TPMS_SIGNATURE_RSA swig_types[149]
#define SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY swig_types[150]
#define SWIGTYPE_p_TPMS_SYMCIPHER_PARMS swig_types[151]
#define SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY swig_types[152]
#define SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT swig_types[153]
#define SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY swig_types[154]
#define SWIGTYPE_p_TPMS_TAGGED_POLICY swig_types[155]
#define SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY swig_types[156]
#define SWIGTYPE_p_TPMS_TAGGED_PROPERTY swig_types[157]
#define SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY swig_types[158]
#define SWIGTYPE_p_TPMS_TIME_ATTEST_INFO swig_types[159]
#define SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY swig_types[160]
#define SWIGTYPE_p_TPMS_TIME_INFO swig_types[161]
#define SWIGTYPE_p_TPMS_TIME_INFO_ARRAY swig_types[162]
#define SWIGTYPE_p_TPMT_ASYM_SCHEME swig_types[163]
#define SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY swig_types[164]
#define SWIGTYPE_p_TPMT_ECC_SCHEME swig_types[165]
#define SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY swig_types[166]
#define SWIGTYPE_p_TPMT_HA swig_types[167]
#define SWIGTYPE_p_TPMT_HA_ARRAY swig_types[168]
#define SWIGTYPE_p_TPMT_KDF_SCHEME swig_types[169]
#define SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY swig_types[170]
#define SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME swig_types[171]
#define SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY swig_types[172]
#define SWIGTYPE_p_TPMT_PUBLIC swig_types[173]
#define SWIGTYPE_p_TPMT_PUBLIC_ARRAY swig_types[174]
#define SWIGTYPE_p_TPMT_PUBLIC_PARMS swig_types[175]
#define SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY swig_types[176]
#define SWIGTYPE_p_TPMT_RSA_DECRYPT swig_types[177]
#define SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY swig_types[178]
#define SWIGTYPE_p_TPMT_RSA_SCHEME swig_types[179]
#define SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY swig_types[180]
#define SWIGTYPE_p_TPMT_SENSITIVE swig_types[181]
#define SWIGTYPE_p_TPMT_SENSITIVE_ARRAY swig_types[182]
#define SWIGTYPE_p_TPMT_SIGNATURE swig_types[183]
#define SWIGTYPE_p_TPMT_SIGNATURE_ARRAY swig_types[184]
#define SWIGTYPE_p_TPMT_SIG_SCHEME swig_types[185]
#define SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY swig_types[186]
#define SWIGTYPE_p_TPMT_SYM_DEF swig_types[187]
#define SWIGTYPE_p_TPMT_SYM_DEF_ARRAY swig_types[188]
#define SWIGTYPE_p_TPMT_SYM_DEF_OBJECT swig_types[189]
#define SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY swig_types[190]
#define SWIGTYPE_p_TPMT_TK_AUTH swig_types[191]
#define SWIGTYPE_p_TPMT_TK_AUTH_ARRAY swig_types[192]
#define SWIGTYPE_p_TPMT_TK_CREATION swig_types[193]
#define SWIGTYPE_p_TPMT_TK_CREATION_ARRAY swig_types[194]
#define SWIGTYPE_p_TPMT_TK_HASHCHECK swig_types[195]
#define SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY swig_types[196]
#define SWIGTYPE_p_TPMT_TK_VERIFIED swig_types[197]
#define SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY swig_types[198]
#define SWIGTYPE_p_TPMU_ASYM_SCHEME swig_types[199]
#define SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY swig_types[200]
#define SWIGTYPE_p_TPMU_ATTEST swig_types[201]
#define SWIGTYPE_p_TPMU_ATTEST_ARRAY swig_types[202]
#define SWIGTYPE_p_TPMU_CAPABILITIES swig_types[203]
#define SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY swig_types[204]
#define SWIGTYPE_p_TPMU_ENCRYPTED_SECRET swig_types[205]
#define SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY swig_types[206]
#define SWIGTYPE_p_TPMU_HA swig_types[207]
#define SWIGTYPE_p_TPMU_HA_ARRAY swig_types[208]
#define SWIGTYPE_p_TPMU_KDF_SCHEME swig_types[209]
#define SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY swig_types[210]
#define SWIGTYPE_p_TPMU_NAME swig_types[211]
#define SWIGTYPE_p_TPMU_NAME_ARRAY swig_types[212]
#define SWIGTYPE_p_TPMU_PUBLIC_ID swig_types[213]
#define SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY swig_types[214]
#define SWIGTYPE_p_TPMU_PUBLIC_PARMS swig_types[215]
#define SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY swig_types[216]
#define SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH swig_types[217]
#define SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY swig_types[218]
#define SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE swig_types[219]
#define SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY swig_types[220]
#define SWIGTYPE_p_TPMU_SIGNATURE swig_types[221]
#define SWIGTYPE_p_TPMU_SIGNATURE_ARRAY swig_types[222]
#define SWIGTYPE_p_TPMU_SIG_SCHEME swig_types[223]
#define SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY swig_types[224]
#define SWIGTYPE_p_TPMU_SYM_KEY_BITS swig_types[225]
#define SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY swig_types[226]
#define SWIGTYPE_p_TPMU_SYM_MODE swig_types[227]
#define SWIGTYPE_p_TPMU_SYM_MODE_ARRAY swig_types[228]
#define SWIGTYPE_p_TSS2_ABI_VERSION swig_types[229]
#define SWIGTYPE_p_TSS2_TCTI_CONTEXT swig_types[230]
#define SWIGTYPE_p_UINT16_PTR swig_types[231]
#define SWIGTYPE_p_UINT32_PTR swig_types[232]
#define SWIGTYPE_p__PRIVATE swig_types[233]
#define SWIGTYPE_p__PRIVATE_ARRAY swig_types[234]
#define SWIGTYPE_p_char swig_types[235]
#define SWIGTYPE_p_int swig_types[236]
#define SWIGTYPE_p_long_long swig_types[237]
#define SWIGTYPE_p_p_ESYS_CONTEXT swig_types[238]
#define SWIGTYPE_p_p_ESYS_TR_PTR swig_types[239]
#define SWIGTYPE_p_p_TPM2B_ATTEST swig_types[240]
#define SWIGTYPE_p_p_TPM2B_ATTEST_ARRAY swig_types[241]
#define SWIGTYPE_p_p_TPM2B_CONTEXT_DATA swig_types[242]
#define SWIGTYPE_p_p_TPM2B_CONTEXT_DATA_ARRAY swig_types[243]
#define SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE swig_types[244]
#define SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY swig_types[245]
#define SWIGTYPE_p_p_TPM2B_CREATION_DATA swig_types[246]
#define SWIGTYPE_p_p_TPM2B_CREATION_DATA_ARRAY swig_types[247]
#define SWIGTYPE_p_p_TPM2B_DATA swig_types[248]
#define SWIGTYPE_p_p_TPM2B_DATA_ARRAY swig_types[249]
#define SWIGTYPE_p_p_TPM2B_DIGEST swig_types[250]
#define SWIGTYPE_p_p_TPM2B_DIGEST_ARRAY swig_types[251]
#define SWIGTYPE_p_p_TPM2B_ECC_PARAMETER swig_types[252]
#define SWIGTYPE_p_p_TPM2B_ECC_PARAMETER_ARRAY swig_types[253]
#define SWIGTYPE_p_p_TPM2B_ECC_POINT swig_types[254]
#define SWIGTYPE_p_p_TPM2B_ECC_POINT_ARRAY swig_types[255]
#define SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET swig_types[256]
#define SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET_ARRAY swig_types[257]
#define SWIGTYPE_p_p_TPM2B_EVENT swig_types[258]
#define SWIGTYPE_p_p_TPM2B_EVENT_ARRAY swig_types[259]
#define SWIGTYPE_p_p_TPM2B_ID_OBJECT swig_types[260]
#define SWIGTYPE_p_p_TPM2B_ID_OBJECT_ARRAY swig_types[261]
#define SWIGTYPE_p_p_TPM2B_IV swig_types[262]
#define SWIGTYPE_p_p_TPM2B_IV_ARRAY swig_types[263]
#define SWIGTYPE_p_p_TPM2B_MAX_BUFFER swig_types[264]
#define SWIGTYPE_p_p_TPM2B_MAX_BUFFER_ARRAY swig_types[265]
#define SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER swig_types[266]
#define SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER_ARRAY swig_types[267]
#define SWIGTYPE_p_p_TPM2B_NAME swig_types[268]
#define SWIGTYPE_p_p_TPM2B_NAME_ARRAY swig_types[269]
#define SWIGTYPE_p_p_TPM2B_NV_PUBLIC swig_types[270]
#define SWIGTYPE_p_p_TPM2B_NV_PUBLIC_ARRAY swig_types[271]
#define SWIGTYPE_p_p_TPM2B_PRIVATE swig_types[272]
#define SWIGTYPE_p_p_TPM2B_PRIVATE_ARRAY swig_types[273]
#define SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA swig_types[274]
#define SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY swig_types[275]
#define SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC swig_types[276]
#define SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY swig_types[277]
#define SWIGTYPE_p_p_TPM2B_PUBLIC swig_types[278]
#define SWIGTYPE_p_p_TPM2B_PUBLIC_ARRAY swig_types[279]
#define SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA swig_types[280]
#define SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY swig_types[281]
#define SWIGTYPE_p_p_TPM2B_SENSITIVE swig_types[282]
#define SWIGTYPE_p_p_TPM2B_SENSITIVE_ARRAY swig_types[283]
#define SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE swig_types[284]
#define SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE_ARRAY swig_types[285]
#define SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA swig_types[286]
#define SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA_ARRAY swig_types[287]
#define SWIGTYPE_p_p_TPM2B_SYM_KEY swig_types[288]
#define SWIGTYPE_p_p_TPM2B_SYM_KEY_ARRAY swig_types[289]
#define SWIGTYPE_p_p_TPM2B_TEMPLATE swig_types[290]
#define SWIGTYPE_p_p_TPM2B_TEMPLATE_ARRAY swig_types[291]
#define SWIGTYPE_p_p_TPML_AC_CAPABILITIES swig_types[292]
#define SWIGTYPE_p_p_TPML_AC_CAPABILITIES_ARRAY swig_types[293]
#define SWIGTYPE_p_p_TPML_ALG swig_types[294]
#define SWIGTYPE_p_p_TPML_ALG_ARRAY swig_types[295]
#define SWIGTYPE_p_p_TPML_ALG_PROPERTY swig_types[296]
#define SWIGTYPE_p_p_TPML_ALG_PROPERTY_ARRAY swig_types[297]
#define SWIGTYPE_p_p_TPML_CC swig_types[298]
#define SWIGTYPE_p_p_TPML_CCA swig_types[299]
#define SWIGTYPE_p_p_TPML_CCA_ARRAY swig_types[300]
#define SWIGTYPE_p_p_TPML_CC_ARRAY swig_types[301]
#define SWIGTYPE_p_p_TPML_DIGEST swig_types[302]
#define SWIGTYPE_p_p_TPML_DIGEST_ARRAY swig_types[303]
#define SWIGTYPE_p_p_TPML_DIGEST_VALUES swig_types[304]
#define SWIGTYPE_p_p_TPML_DIGEST_VALUES_ARRAY swig_types[305]
#define SWIGTYPE_p_p_TPML_ECC_CURVE swig_types[306]
#define SWIGTYPE_p_p_TPML_ECC_CURVE_ARRAY swig_types[307]
#define SWIGTYPE_p_p_TPML_HANDLE swig_types[308]
#define SWIGTYPE_p_p_TPML_HANDLE_ARRAY swig_types[309]
#define SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY swig_types[310]
#define SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY_ARRAY swig_types[311]
#define SWIGTYPE_p_p_TPML_PCR_SELECTION swig_types[312]
#define SWIGTYPE_p_p_TPML_PCR_SELECTION_ARRAY swig_types[313]
#define SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY swig_types[314]
#define SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY swig_types[315]
#define SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY swig_types[316]
#define SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY swig_types[317]
#define SWIGTYPE_p_p_TPMS_AC_OUTPUT swig_types[318]
#define SWIGTYPE_p_p_TPMS_AC_OUTPUT_ARRAY swig_types[319]
#define SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION swig_types[320]
#define SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY swig_types[321]
#define SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC swig_types[322]
#define SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY swig_types[323]
#define SWIGTYPE_p_p_TPMS_ALG_PROPERTY swig_types[324]
#define SWIGTYPE_p_p_TPMS_ALG_PROPERTY_ARRAY swig_types[325]
#define SWIGTYPE_p_p_TPMS_ASYM_PARMS swig_types[326]
#define SWIGTYPE_p_p_TPMS_ASYM_PARMS_ARRAY swig_types[327]
#define SWIGTYPE_p_p_TPMS_ATTEST swig_types[328]
#define SWIGTYPE_p_p_TPMS_ATTEST_ARRAY swig_types[329]
#define SWIGTYPE_p_p_TPMS_AUTH_COMMAND swig_types[330]
#define SWIGTYPE_p_p_TPMS_AUTH_COMMAND_ARRAY swig_types[331]
#define SWIGTYPE_p_p_TPMS_AUTH_RESPONSE swig_types[332]
#define SWIGTYPE_p_p_TPMS_AUTH_RESPONSE_ARRAY swig_types[333]
#define SWIGTYPE_p_p_TPMS_CAPABILITY_DATA swig_types[334]
#define SWIGTYPE_p_p_TPMS_CAPABILITY_DATA_ARRAY swig_types[335]
#define SWIGTYPE_p_p_TPMS_CERTIFY_INFO swig_types[336]
#define SWIGTYPE_p_p_TPMS_CERTIFY_INFO_ARRAY swig_types[337]
#define SWIGTYPE_p_p_TPMS_CLOCK_INFO swig_types[338]
#define SWIGTYPE_p_p_TPMS_CLOCK_INFO_ARRAY swig_types[339]
#define SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO swig_types[340]
#define SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY swig_types[341]
#define SWIGTYPE_p_p_TPMS_CONTEXT swig_types[342]
#define SWIGTYPE_p_p_TPMS_CONTEXT_ARRAY swig_types[343]
#define SWIGTYPE_p_p_TPMS_CONTEXT_DATA swig_types[344]
#define SWIGTYPE_p_p_TPMS_CONTEXT_DATA_ARRAY swig_types[345]
#define SWIGTYPE_p_p_TPMS_CREATION_DATA swig_types[346]
#define SWIGTYPE_p_p_TPMS_CREATION_DATA_ARRAY swig_types[347]
#define SWIGTYPE_p_p_TPMS_CREATION_INFO swig_types[348]
#define SWIGTYPE_p_p_TPMS_CREATION_INFO_ARRAY swig_types[349]
#define SWIGTYPE_p_p_TPMS_ECC_PARMS swig_types[350]
#define SWIGTYPE_p_p_TPMS_ECC_PARMS_ARRAY swig_types[351]
#define SWIGTYPE_p_p_TPMS_ECC_POINT swig_types[352]
#define SWIGTYPE_p_p_TPMS_ECC_POINT_ARRAY swig_types[353]
#define SWIGTYPE_p_p_TPMS_EMPTY swig_types[354]
#define SWIGTYPE_p_p_TPMS_EMPTY_ARRAY swig_types[355]
#define SWIGTYPE_p_p_TPMS_ID_OBJECT swig_types[356]
#define SWIGTYPE_p_p_TPMS_ID_OBJECT_ARRAY swig_types[357]
#define SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS swig_types[358]
#define SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS_ARRAY swig_types[359]
#define SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO swig_types[360]
#define SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO_ARRAY swig_types[361]
#define SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS swig_types[362]
#define SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY swig_types[363]
#define SWIGTYPE_p_p_TPMS_NV_PUBLIC swig_types[364]
#define SWIGTYPE_p_p_TPMS_NV_PUBLIC_ARRAY swig_types[365]
#define SWIGTYPE_p_p_TPMS_PCR_SELECT swig_types[366]
#define SWIGTYPE_p_p_TPMS_PCR_SELECTION swig_types[367]
#define SWIGTYPE_p_p_TPMS_PCR_SELECTION_ARRAY swig_types[368]
#define SWIGTYPE_p_p_TPMS_PCR_SELECT_ARRAY swig_types[369]
#define SWIGTYPE_p_p_TPMS_QUOTE_INFO swig_types[370]
#define SWIGTYPE_p_p_TPMS_QUOTE_INFO_ARRAY swig_types[371]
#define SWIGTYPE_p_p_TPMS_RSA_PARMS swig_types[372]
#define SWIGTYPE_p_p_TPMS_RSA_PARMS_ARRAY swig_types[373]
#define SWIGTYPE_p_p_TPMS_SCHEME_ECDAA swig_types[374]
#define SWIGTYPE_p_p_TPMS_SCHEME_ECDAA_ARRAY swig_types[375]
#define SWIGTYPE_p_p_TPMS_SCHEME_HASH swig_types[376]
#define SWIGTYPE_p_p_TPMS_SCHEME_HASH_ARRAY swig_types[377]
#define SWIGTYPE_p_p_TPMS_SCHEME_XOR swig_types[378]
#define SWIGTYPE_p_p_TPMS_SCHEME_XOR_ARRAY swig_types[379]
#define SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE swig_types[380]
#define SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE_ARRAY swig_types[381]
#define SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO swig_types[382]
#define SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO_ARRAY swig_types[383]
#define SWIGTYPE_p_p_TPMS_SIGNATURE_ECC swig_types[384]
#define SWIGTYPE_p_p_TPMS_SIGNATURE_ECC_ARRAY swig_types[385]
#define SWIGTYPE_p_p_TPMS_SIGNATURE_RSA swig_types[386]
#define SWIGTYPE_p_p_TPMS_SIGNATURE_RSA_ARRAY swig_types[387]
#define SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS swig_types[388]
#define SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS_ARRAY swig_types[389]
#define SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT swig_types[390]
#define SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT_ARRAY swig_types[391]
#define SWIGTYPE_p_p_TPMS_TAGGED_POLICY swig_types[392]
#define SWIGTYPE_p_p_TPMS_TAGGED_POLICY_ARRAY swig_types[393]
#define SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY swig_types[394]
#define SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY_ARRAY swig_types[395]
#define SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO swig_types[396]
#define SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO_ARRAY swig_types[397]
#define SWIGTYPE_p_p_TPMS_TIME_INFO swig_types[398]
#define SWIGTYPE_p_p_TPMS_TIME_INFO_ARRAY swig_types[399]
#define SWIGTYPE_p_p_TPMT_ASYM_SCHEME swig_types[400]
#define SWIGTYPE_p_p_TPMT_ASYM_SCHEME_ARRAY swig_types[401]
#define SWIGTYPE_p_p_TPMT_ECC_SCHEME swig_types[402]
#define SWIGTYPE_p_p_TPMT_ECC_SCHEME_ARRAY swig_types[403]
#define SWIGTYPE_p_p_TPMT_HA swig_types[404]
#define SWIGTYPE_p_p_TPMT_HA_ARRAY swig_types[405]
#define SWIGTYPE_p_p_TPMT_KDF_SCHEME swig_types[406]
#define SWIGTYPE_p_p_TPMT_KDF_SCHEME_ARRAY swig_types[407]
#define SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME swig_types[408]
#define SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME_ARRAY swig_types[409]
#define SWIGTYPE_p_p_TPMT_PUBLIC swig_types[410]
#define SWIGTYPE_p_p_TPMT_PUBLIC_ARRAY swig_types[411]
#define SWIGTYPE_p_p_TPMT_PUBLIC_PARMS swig_types[412]
#define SWIGTYPE_p_p_TPMT_PUBLIC_PARMS_ARRAY swig_types[413]
#define SWIGTYPE_p_p_TPMT_RSA_DECRYPT swig_types[414]
#define SWIGTYPE_p_p_TPMT_RSA_DECRYPT_ARRAY swig_types[415]
#define SWIGTYPE_p_p_TPMT_RSA_SCHEME swig_types[416]
#define SWIGTYPE_p_p_TPMT_RSA_SCHEME_ARRAY swig_types[417]
#define SWIGTYPE_p_p_TPMT_SENSITIVE swig_types[418]
#define SWIGTYPE_p_p_TPMT_SENSITIVE_ARRAY swig_types[419]
#define SWIGTYPE_p_p_TPMT_SIGNATURE swig_types[420]
#define SWIGTYPE_p_p_TPMT_SIGNATURE_ARRAY swig_types[421]
#define SWIGTYPE_p_p_TPMT_SIG_SCHEME swig_types[422]
#define SWIGTYPE_p_p_TPMT_SIG_SCHEME_ARRAY swig_types[423]
#define SWIGTYPE_p_p_TPMT_SYM_DEF swig_types[424]
#define SWIGTYPE_p_p_TPMT_SYM_DEF_ARRAY swig_types[425]
#define SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT swig_types[426]
#define SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT_ARRAY swig_types[427]
#define SWIGTYPE_p_p_TPMT_TK_AUTH swig_types[428]
#define SWIGTYPE_p_p_TPMT_TK_AUTH_ARRAY swig_types[429]
#define SWIGTYPE_p_p_TPMT_TK_CREATION swig_types[430]
#define SWIGTYPE_p_p_TPMT_TK_CREATION_ARRAY swig_types[431]
#define SWIGTYPE_p_p_TPMT_TK_HASHCHECK swig_types[432]
#define SWIGTYPE_p_p_TPMT_TK_HASHCHECK_ARRAY swig_types[433]
#define SWIGTYPE_p_p_TPMT_TK_VERIFIED swig_types[434]
#define SWIGTYPE_p_p_TPMT_TK_VERIFIED_ARRAY swig_types[435]
#define SWIGTYPE_p_p_TPMU_ASYM_SCHEME swig_types[436]
#define SWIGTYPE_p_p_TPMU_ASYM_SCHEME_ARRAY swig_types[437]
#define SWIGTYPE_p_p_TPMU_ATTEST swig_types[438]
#define SWIGTYPE_p_p_TPMU_ATTEST_ARRAY swig_types[439]
#define SWIGTYPE_p_p_TPMU_CAPABILITIES swig_types[440]
#define SWIGTYPE_p_p_TPMU_CAPABILITIES_ARRAY swig_types[441]
#define SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET swig_types[442]
#define SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET_ARRAY swig_types[443]
#define SWIGTYPE_p_p_TPMU_HA swig_types[444]
#define SWIGTYPE_p_p_TPMU_HA_ARRAY swig_types[445]
#define SWIGTYPE_p_p_TPMU_KDF_SCHEME swig_types[446]
#define SWIGTYPE_p_p_TPMU_KDF_SCHEME_ARRAY swig_types[447]
#define SWIGTYPE_p_p_TPMU_NAME swig_types[448]
#define SWIGTYPE_p_p_TPMU_NAME_ARRAY swig_types[449]
#define SWIGTYPE_p_p_TPMU_PUBLIC_ID swig_types[450]
#define SWIGTYPE_p_p_TPMU_PUBLIC_ID_ARRAY swig_types[451]
#define SWIGTYPE_p_p_TPMU_PUBLIC_PARMS swig_types[452]
#define SWIGTYPE_p_p_TPMU_PUBLIC_PARMS_ARRAY swig_types[453]
#define SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH swig_types[454]
#define SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH_ARRAY swig_types[455]
#define SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE swig_types[456]
#define SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY swig_types[457]
#define SWIGTYPE_p_p_TPMU_SIGNATURE swig_types[458]
#define SWIGTYPE_p_p_TPMU_SIGNATURE_ARRAY swig_types[459]
#define SWIGTYPE_p_p_TPMU_SIG_SCHEME swig_types[460]
#define SWIGTYPE_p_p_TPMU_SIG_SCHEME_ARRAY swig_types[461]
#define SWIGTYPE_p_p_TPMU_SYM_KEY_BITS swig_types[462]
#define SWIGTYPE_p_p_TPMU_SYM_KEY_BITS_ARRAY swig_types[463]
#define SWIGTYPE_p_p_TPMU_SYM_MODE swig_types[464]
#define SWIGTYPE_p_p_TPMU_SYM_MODE_ARRAY swig_types[465]
#define SWIGTYPE_p_p_TSS2_TCTI_CONTEXT swig_types[466]
#define SWIGTYPE_p_p_TSS2_TCTI_POLL_HANDLE swig_types[467]
#define SWIGTYPE_p_p__PRIVATE swig_types[468]
#define SWIGTYPE_p_p__PRIVATE_ARRAY swig_types[469]
#define SWIGTYPE_p_p_char swig_types[470]
#define SWIGTYPE_p_p_size_t swig_types[471]
#define SWIGTYPE_p_p_unsigned_char swig_types[472]
#define SWIGTYPE_p_p_unsigned_int swig_types[473]
#define SWIGTYPE_p_p_void swig_types[474]
#define SWIGTYPE_p_short swig_types[475]
#define SWIGTYPE_p_signed_char swig_types[476]
#define SWIGTYPE_p_size_t swig_types[477]
#define SWIGTYPE_p_uint16_t_PTR swig_types[478]
#define SWIGTYPE_p_uint32_t_PTR swig_types[479]
#define SWIGTYPE_p_unsigned_char swig_types[480]
#define SWIGTYPE_p_unsigned_int swig_types[481]
#define SWIGTYPE_p_unsigned_long_long swig_types[482]
#define SWIGTYPE_p_unsigned_short swig_types[483]
#define SWIGTYPE_p_void swig_types[484]
static swig_type_info *swig_types[486];
static swig_module_info swig_module = {swig_types, 485, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#if (PY_VERSION_HEX <= 0x02000000)
# if !defined(SWIG_PYTHON_CLASSIC)
#  error "This python version requires swig to be run with the '-classic' option"
# endif
#endif

/*-----------------------------------------------
              @(target):= _esys_binding.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__esys_binding

#else
#  define SWIG_init    init_esys_binding

#endif
#define SWIG_name    "_esys_binding"

#define SWIGVERSION 0x030008 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#include <tss2/tss2_esys.h>
#include <tss2/tss2_tctildr.h>



#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_short (PyObject * obj, unsigned short *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > USHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned short)(v);
    }
  }  
  return res;
}


  #define SWIG_From_long   PyLong_FromLong 


SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyLong_FromLong((long)(value)); 
}


SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_short  (unsigned short value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_char (PyObject * obj, unsigned char *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UCHAR_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned char)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_char  (unsigned char value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (PyObject *obj, unsigned long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    unsigned long long v = PyLong_AsUnsignedLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
      if (val) *val = (unsigned long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}


SWIGINTERNINLINE PyObject* 
SWIG_From_long_SS_long  (long long value)
{
  return ((value < LONG_MIN) || (value > LONG_MAX)) ?
    PyLong_FromLongLong(value) : PyLong_FromLong((long)(value)); 
}


SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLongLong(value) : PyLong_FromLong((long)(value)); 
}


#include <tss2/tss2_rc.h>


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj((char *)(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if PY_VERSION_HEX >= 0x03010000
      return PyUnicode_DecodeUTF8(carray, (Py_ssize_t)(size), "surrogateescape");
#else
      return PyUnicode_FromStringAndSize(carray, (Py_ssize_t)(size));
#endif
#else
      return PyString_FromStringAndSize(carray, (Py_ssize_t)(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


#include <stdlib.h>


  static TPMI_YES_NO *new_TPMI_YES_NO_PTR() { 
    return (TPMI_YES_NO *)malloc(sizeof(TPMI_YES_NO));
  }
  
  static TPMI_YES_NO *copy_TPMI_YES_NO_PTR(TPMI_YES_NO value) { 
    return (TPMI_YES_NO *)memcpy((TPMI_YES_NO *)malloc(sizeof(TPMI_YES_NO)),&value,sizeof(TPMI_YES_NO));
  }

  static void delete_TPMI_YES_NO_PTR(TPMI_YES_NO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMI_YES_NO_PTR_assign(TPMI_YES_NO *obj, TPMI_YES_NO value) {
    *obj = value;
  }

  static TPMI_YES_NO TPMI_YES_NO_PTR_value(TPMI_YES_NO *obj) {
    return *obj;
  }


SWIGINTERNINLINE PyObject *
SWIG_From_size_t  (size_t value)
{    
  return SWIG_From_unsigned_SS_long  ((unsigned long)(value));
}


typedef TPMS_EMPTY TPMS_EMPTY_ARRAY;


SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
  if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  return res;
}

SWIGINTERN TPMS_EMPTY_ARRAY *new_TPMS_EMPTY_ARRAY(size_t nelements){
    return (TPMS_EMPTY *)malloc((nelements)*sizeof(TPMS_EMPTY));
  }
SWIGINTERN void delete_TPMS_EMPTY_ARRAY(TPMS_EMPTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_EMPTY TPMS_EMPTY_ARRAY___getitem__(TPMS_EMPTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_EMPTY_ARRAY___setitem__(TPMS_EMPTY_ARRAY *self,size_t index,TPMS_EMPTY value){
    self[index] = value;
  }
SWIGINTERN TPMS_EMPTY *TPMS_EMPTY_ARRAY_cast(TPMS_EMPTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_EMPTY_ARRAY *TPMS_EMPTY_ARRAY_frompointer(TPMS_EMPTY *t){
    return (TPMS_EMPTY_ARRAY *)(t);
  }

  static TPMS_EMPTY *new_TPMS_EMPTY_PTR() { 
    return (TPMS_EMPTY *)malloc(sizeof(TPMS_EMPTY));
  }
  
  static TPMS_EMPTY *copy_TPMS_EMPTY_PTR(TPMS_EMPTY value) { 
    return (TPMS_EMPTY *)memcpy((TPMS_EMPTY *)malloc(sizeof(TPMS_EMPTY)),&value,sizeof(TPMS_EMPTY));
  }

  static void delete_TPMS_EMPTY_PTR(TPMS_EMPTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_EMPTY_PTR_assign(TPMS_EMPTY *obj, TPMS_EMPTY value) {
    *obj = value;
  }

  static TPMS_EMPTY TPMS_EMPTY_PTR_value(TPMS_EMPTY *obj) {
    return *obj;
  }


  static TPMS_EMPTY * *new_TPMS_EMPTY_PTR_PTR() { 
    return (TPMS_EMPTY * *)malloc(sizeof(TPMS_EMPTY *));
  }
  
  static TPMS_EMPTY * *copy_TPMS_EMPTY_PTR_PTR(TPMS_EMPTY * value) { 
    return (TPMS_EMPTY * *)memcpy((TPMS_EMPTY * *)malloc(sizeof(TPMS_EMPTY *)),&value,sizeof(TPMS_EMPTY *));
  }

  static void delete_TPMS_EMPTY_PTR_PTR(TPMS_EMPTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_EMPTY_PTR_PTR_assign(TPMS_EMPTY * *obj, TPMS_EMPTY * value) {
    *obj = value;
  }

  static TPMS_EMPTY * TPMS_EMPTY_PTR_PTR_value(TPMS_EMPTY * *obj) {
    return *obj;
  }


typedef TPMS_ALGORITHM_DESCRIPTION TPMS_ALGORITHM_DESCRIPTION_ARRAY;

SWIGINTERN TPMS_ALGORITHM_DESCRIPTION_ARRAY *new_TPMS_ALGORITHM_DESCRIPTION_ARRAY(size_t nelements){
    return (TPMS_ALGORITHM_DESCRIPTION *)malloc((nelements)*sizeof(TPMS_ALGORITHM_DESCRIPTION));
  }
SWIGINTERN void delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY(TPMS_ALGORITHM_DESCRIPTION_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ALGORITHM_DESCRIPTION TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__(TPMS_ALGORITHM_DESCRIPTION_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__(TPMS_ALGORITHM_DESCRIPTION_ARRAY *self,size_t index,TPMS_ALGORITHM_DESCRIPTION value){
    self[index] = value;
  }
SWIGINTERN TPMS_ALGORITHM_DESCRIPTION *TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast(TPMS_ALGORITHM_DESCRIPTION_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ALGORITHM_DESCRIPTION_ARRAY *TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer(TPMS_ALGORITHM_DESCRIPTION *t){
    return (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)(t);
  }

  static TPMS_ALGORITHM_DESCRIPTION *new_TPMS_ALGORITHM_DESCRIPTION_PTR() { 
    return (TPMS_ALGORITHM_DESCRIPTION *)malloc(sizeof(TPMS_ALGORITHM_DESCRIPTION));
  }
  
  static TPMS_ALGORITHM_DESCRIPTION *copy_TPMS_ALGORITHM_DESCRIPTION_PTR(TPMS_ALGORITHM_DESCRIPTION value) { 
    return (TPMS_ALGORITHM_DESCRIPTION *)memcpy((TPMS_ALGORITHM_DESCRIPTION *)malloc(sizeof(TPMS_ALGORITHM_DESCRIPTION)),&value,sizeof(TPMS_ALGORITHM_DESCRIPTION));
  }

  static void delete_TPMS_ALGORITHM_DESCRIPTION_PTR(TPMS_ALGORITHM_DESCRIPTION *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ALGORITHM_DESCRIPTION_PTR_assign(TPMS_ALGORITHM_DESCRIPTION *obj, TPMS_ALGORITHM_DESCRIPTION value) {
    *obj = value;
  }

  static TPMS_ALGORITHM_DESCRIPTION TPMS_ALGORITHM_DESCRIPTION_PTR_value(TPMS_ALGORITHM_DESCRIPTION *obj) {
    return *obj;
  }


  static TPMS_ALGORITHM_DESCRIPTION * *new_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR() { 
    return (TPMS_ALGORITHM_DESCRIPTION * *)malloc(sizeof(TPMS_ALGORITHM_DESCRIPTION *));
  }
  
  static TPMS_ALGORITHM_DESCRIPTION * *copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(TPMS_ALGORITHM_DESCRIPTION * value) { 
    return (TPMS_ALGORITHM_DESCRIPTION * *)memcpy((TPMS_ALGORITHM_DESCRIPTION * *)malloc(sizeof(TPMS_ALGORITHM_DESCRIPTION *)),&value,sizeof(TPMS_ALGORITHM_DESCRIPTION *));
  }

  static void delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(TPMS_ALGORITHM_DESCRIPTION * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign(TPMS_ALGORITHM_DESCRIPTION * *obj, TPMS_ALGORITHM_DESCRIPTION * value) {
    *obj = value;
  }

  static TPMS_ALGORITHM_DESCRIPTION * TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value(TPMS_ALGORITHM_DESCRIPTION * *obj) {
    return *obj;
  }


typedef TPMU_HA TPMU_HA_ARRAY;

SWIGINTERN TPMU_HA_ARRAY *new_TPMU_HA_ARRAY(size_t nelements){
    return (TPMU_HA *)malloc((nelements)*sizeof(TPMU_HA));
  }
SWIGINTERN void delete_TPMU_HA_ARRAY(TPMU_HA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_HA TPMU_HA_ARRAY___getitem__(TPMU_HA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_HA_ARRAY___setitem__(TPMU_HA_ARRAY *self,size_t index,TPMU_HA value){
    self[index] = value;
  }
SWIGINTERN TPMU_HA *TPMU_HA_ARRAY_cast(TPMU_HA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_HA_ARRAY *TPMU_HA_ARRAY_frompointer(TPMU_HA *t){
    return (TPMU_HA_ARRAY *)(t);
  }

  static TPMU_HA *new_TPMU_HA_PTR() { 
    return (TPMU_HA *)malloc(sizeof(TPMU_HA));
  }
  
  static TPMU_HA *copy_TPMU_HA_PTR(TPMU_HA value) { 
    return (TPMU_HA *)memcpy((TPMU_HA *)malloc(sizeof(TPMU_HA)),&value,sizeof(TPMU_HA));
  }

  static void delete_TPMU_HA_PTR(TPMU_HA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_HA_PTR_assign(TPMU_HA *obj, TPMU_HA value) {
    *obj = value;
  }

  static TPMU_HA TPMU_HA_PTR_value(TPMU_HA *obj) {
    return *obj;
  }


  static TPMU_HA * *new_TPMU_HA_PTR_PTR() { 
    return (TPMU_HA * *)malloc(sizeof(TPMU_HA *));
  }
  
  static TPMU_HA * *copy_TPMU_HA_PTR_PTR(TPMU_HA * value) { 
    return (TPMU_HA * *)memcpy((TPMU_HA * *)malloc(sizeof(TPMU_HA *)),&value,sizeof(TPMU_HA *));
  }

  static void delete_TPMU_HA_PTR_PTR(TPMU_HA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_HA_PTR_PTR_assign(TPMU_HA * *obj, TPMU_HA * value) {
    *obj = value;
  }

  static TPMU_HA * TPMU_HA_PTR_PTR_value(TPMU_HA * *obj) {
    return *obj;
  }


typedef TPMT_HA TPMT_HA_ARRAY;

SWIGINTERN TPMT_HA_ARRAY *new_TPMT_HA_ARRAY(size_t nelements){
    return (TPMT_HA *)malloc((nelements)*sizeof(TPMT_HA));
  }
SWIGINTERN void delete_TPMT_HA_ARRAY(TPMT_HA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_HA TPMT_HA_ARRAY___getitem__(TPMT_HA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_HA_ARRAY___setitem__(TPMT_HA_ARRAY *self,size_t index,TPMT_HA value){
    self[index] = value;
  }
SWIGINTERN TPMT_HA *TPMT_HA_ARRAY_cast(TPMT_HA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_HA_ARRAY *TPMT_HA_ARRAY_frompointer(TPMT_HA *t){
    return (TPMT_HA_ARRAY *)(t);
  }

  static TPMT_HA *new_TPMT_HA_PTR() { 
    return (TPMT_HA *)malloc(sizeof(TPMT_HA));
  }
  
  static TPMT_HA *copy_TPMT_HA_PTR(TPMT_HA value) { 
    return (TPMT_HA *)memcpy((TPMT_HA *)malloc(sizeof(TPMT_HA)),&value,sizeof(TPMT_HA));
  }

  static void delete_TPMT_HA_PTR(TPMT_HA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_HA_PTR_assign(TPMT_HA *obj, TPMT_HA value) {
    *obj = value;
  }

  static TPMT_HA TPMT_HA_PTR_value(TPMT_HA *obj) {
    return *obj;
  }


  static TPMT_HA * *new_TPMT_HA_PTR_PTR() { 
    return (TPMT_HA * *)malloc(sizeof(TPMT_HA *));
  }
  
  static TPMT_HA * *copy_TPMT_HA_PTR_PTR(TPMT_HA * value) { 
    return (TPMT_HA * *)memcpy((TPMT_HA * *)malloc(sizeof(TPMT_HA *)),&value,sizeof(TPMT_HA *));
  }

  static void delete_TPMT_HA_PTR_PTR(TPMT_HA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_HA_PTR_PTR_assign(TPMT_HA * *obj, TPMT_HA * value) {
    *obj = value;
  }

  static TPMT_HA * TPMT_HA_PTR_PTR_value(TPMT_HA * *obj) {
    return *obj;
  }


typedef TPM2B_DIGEST TPM2B_DIGEST_ARRAY;

SWIGINTERN TPM2B_DIGEST_ARRAY *new_TPM2B_DIGEST_ARRAY(size_t nelements){
    return (TPM2B_DIGEST *)malloc((nelements)*sizeof(TPM2B_DIGEST));
  }
SWIGINTERN void delete_TPM2B_DIGEST_ARRAY(TPM2B_DIGEST_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_DIGEST TPM2B_DIGEST_ARRAY___getitem__(TPM2B_DIGEST_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_DIGEST_ARRAY___setitem__(TPM2B_DIGEST_ARRAY *self,size_t index,TPM2B_DIGEST value){
    self[index] = value;
  }
SWIGINTERN TPM2B_DIGEST *TPM2B_DIGEST_ARRAY_cast(TPM2B_DIGEST_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_DIGEST_ARRAY *TPM2B_DIGEST_ARRAY_frompointer(TPM2B_DIGEST *t){
    return (TPM2B_DIGEST_ARRAY *)(t);
  }

  static TPM2B_DIGEST *new_TPM2B_DIGEST_PTR() { 
    return (TPM2B_DIGEST *)malloc(sizeof(TPM2B_DIGEST));
  }
  
  static TPM2B_DIGEST *copy_TPM2B_DIGEST_PTR(TPM2B_DIGEST value) { 
    return (TPM2B_DIGEST *)memcpy((TPM2B_DIGEST *)malloc(sizeof(TPM2B_DIGEST)),&value,sizeof(TPM2B_DIGEST));
  }

  static void delete_TPM2B_DIGEST_PTR(TPM2B_DIGEST *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_DIGEST_PTR_assign(TPM2B_DIGEST *obj, TPM2B_DIGEST value) {
    *obj = value;
  }

  static TPM2B_DIGEST TPM2B_DIGEST_PTR_value(TPM2B_DIGEST *obj) {
    return *obj;
  }


  static TPM2B_DIGEST * *new_TPM2B_DIGEST_PTR_PTR() { 
    return (TPM2B_DIGEST * *)malloc(sizeof(TPM2B_DIGEST *));
  }
  
  static TPM2B_DIGEST * *copy_TPM2B_DIGEST_PTR_PTR(TPM2B_DIGEST * value) { 
    return (TPM2B_DIGEST * *)memcpy((TPM2B_DIGEST * *)malloc(sizeof(TPM2B_DIGEST *)),&value,sizeof(TPM2B_DIGEST *));
  }

  static void delete_TPM2B_DIGEST_PTR_PTR(TPM2B_DIGEST * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_DIGEST_PTR_PTR_assign(TPM2B_DIGEST * *obj, TPM2B_DIGEST * value) {
    *obj = value;
  }

  static TPM2B_DIGEST * TPM2B_DIGEST_PTR_PTR_value(TPM2B_DIGEST * *obj) {
    return *obj;
  }


typedef TPM2B_DATA TPM2B_DATA_ARRAY;

SWIGINTERN TPM2B_DATA_ARRAY *new_TPM2B_DATA_ARRAY(size_t nelements){
    return (TPM2B_DATA *)malloc((nelements)*sizeof(TPM2B_DATA));
  }
SWIGINTERN void delete_TPM2B_DATA_ARRAY(TPM2B_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_DATA TPM2B_DATA_ARRAY___getitem__(TPM2B_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_DATA_ARRAY___setitem__(TPM2B_DATA_ARRAY *self,size_t index,TPM2B_DATA value){
    self[index] = value;
  }
SWIGINTERN TPM2B_DATA *TPM2B_DATA_ARRAY_cast(TPM2B_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_DATA_ARRAY *TPM2B_DATA_ARRAY_frompointer(TPM2B_DATA *t){
    return (TPM2B_DATA_ARRAY *)(t);
  }

  static TPM2B_DATA *new_TPM2B_DATA_PTR() { 
    return (TPM2B_DATA *)malloc(sizeof(TPM2B_DATA));
  }
  
  static TPM2B_DATA *copy_TPM2B_DATA_PTR(TPM2B_DATA value) { 
    return (TPM2B_DATA *)memcpy((TPM2B_DATA *)malloc(sizeof(TPM2B_DATA)),&value,sizeof(TPM2B_DATA));
  }

  static void delete_TPM2B_DATA_PTR(TPM2B_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_DATA_PTR_assign(TPM2B_DATA *obj, TPM2B_DATA value) {
    *obj = value;
  }

  static TPM2B_DATA TPM2B_DATA_PTR_value(TPM2B_DATA *obj) {
    return *obj;
  }


  static TPM2B_DATA * *new_TPM2B_DATA_PTR_PTR() { 
    return (TPM2B_DATA * *)malloc(sizeof(TPM2B_DATA *));
  }
  
  static TPM2B_DATA * *copy_TPM2B_DATA_PTR_PTR(TPM2B_DATA * value) { 
    return (TPM2B_DATA * *)memcpy((TPM2B_DATA * *)malloc(sizeof(TPM2B_DATA *)),&value,sizeof(TPM2B_DATA *));
  }

  static void delete_TPM2B_DATA_PTR_PTR(TPM2B_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_DATA_PTR_PTR_assign(TPM2B_DATA * *obj, TPM2B_DATA * value) {
    *obj = value;
  }

  static TPM2B_DATA * TPM2B_DATA_PTR_PTR_value(TPM2B_DATA * *obj) {
    return *obj;
  }


typedef TPM2B_EVENT TPM2B_EVENT_ARRAY;

SWIGINTERN TPM2B_EVENT_ARRAY *new_TPM2B_EVENT_ARRAY(size_t nelements){
    return (TPM2B_EVENT *)malloc((nelements)*sizeof(TPM2B_EVENT));
  }
SWIGINTERN void delete_TPM2B_EVENT_ARRAY(TPM2B_EVENT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_EVENT TPM2B_EVENT_ARRAY___getitem__(TPM2B_EVENT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_EVENT_ARRAY___setitem__(TPM2B_EVENT_ARRAY *self,size_t index,TPM2B_EVENT value){
    self[index] = value;
  }
SWIGINTERN TPM2B_EVENT *TPM2B_EVENT_ARRAY_cast(TPM2B_EVENT_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_EVENT_ARRAY *TPM2B_EVENT_ARRAY_frompointer(TPM2B_EVENT *t){
    return (TPM2B_EVENT_ARRAY *)(t);
  }

  static TPM2B_EVENT *new_TPM2B_EVENT_PTR() { 
    return (TPM2B_EVENT *)malloc(sizeof(TPM2B_EVENT));
  }
  
  static TPM2B_EVENT *copy_TPM2B_EVENT_PTR(TPM2B_EVENT value) { 
    return (TPM2B_EVENT *)memcpy((TPM2B_EVENT *)malloc(sizeof(TPM2B_EVENT)),&value,sizeof(TPM2B_EVENT));
  }

  static void delete_TPM2B_EVENT_PTR(TPM2B_EVENT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_EVENT_PTR_assign(TPM2B_EVENT *obj, TPM2B_EVENT value) {
    *obj = value;
  }

  static TPM2B_EVENT TPM2B_EVENT_PTR_value(TPM2B_EVENT *obj) {
    return *obj;
  }


  static TPM2B_EVENT * *new_TPM2B_EVENT_PTR_PTR() { 
    return (TPM2B_EVENT * *)malloc(sizeof(TPM2B_EVENT *));
  }
  
  static TPM2B_EVENT * *copy_TPM2B_EVENT_PTR_PTR(TPM2B_EVENT * value) { 
    return (TPM2B_EVENT * *)memcpy((TPM2B_EVENT * *)malloc(sizeof(TPM2B_EVENT *)),&value,sizeof(TPM2B_EVENT *));
  }

  static void delete_TPM2B_EVENT_PTR_PTR(TPM2B_EVENT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_EVENT_PTR_PTR_assign(TPM2B_EVENT * *obj, TPM2B_EVENT * value) {
    *obj = value;
  }

  static TPM2B_EVENT * TPM2B_EVENT_PTR_PTR_value(TPM2B_EVENT * *obj) {
    return *obj;
  }


typedef TPM2B_MAX_BUFFER TPM2B_MAX_BUFFER_ARRAY;

SWIGINTERN TPM2B_MAX_BUFFER_ARRAY *new_TPM2B_MAX_BUFFER_ARRAY(size_t nelements){
    return (TPM2B_MAX_BUFFER *)malloc((nelements)*sizeof(TPM2B_MAX_BUFFER));
  }
SWIGINTERN void delete_TPM2B_MAX_BUFFER_ARRAY(TPM2B_MAX_BUFFER_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_MAX_BUFFER TPM2B_MAX_BUFFER_ARRAY___getitem__(TPM2B_MAX_BUFFER_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_MAX_BUFFER_ARRAY___setitem__(TPM2B_MAX_BUFFER_ARRAY *self,size_t index,TPM2B_MAX_BUFFER value){
    self[index] = value;
  }
SWIGINTERN TPM2B_MAX_BUFFER *TPM2B_MAX_BUFFER_ARRAY_cast(TPM2B_MAX_BUFFER_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_MAX_BUFFER_ARRAY *TPM2B_MAX_BUFFER_ARRAY_frompointer(TPM2B_MAX_BUFFER *t){
    return (TPM2B_MAX_BUFFER_ARRAY *)(t);
  }

  static TPM2B_MAX_BUFFER *new_TPM2B_MAX_BUFFER_PTR() { 
    return (TPM2B_MAX_BUFFER *)malloc(sizeof(TPM2B_MAX_BUFFER));
  }
  
  static TPM2B_MAX_BUFFER *copy_TPM2B_MAX_BUFFER_PTR(TPM2B_MAX_BUFFER value) { 
    return (TPM2B_MAX_BUFFER *)memcpy((TPM2B_MAX_BUFFER *)malloc(sizeof(TPM2B_MAX_BUFFER)),&value,sizeof(TPM2B_MAX_BUFFER));
  }

  static void delete_TPM2B_MAX_BUFFER_PTR(TPM2B_MAX_BUFFER *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_MAX_BUFFER_PTR_assign(TPM2B_MAX_BUFFER *obj, TPM2B_MAX_BUFFER value) {
    *obj = value;
  }

  static TPM2B_MAX_BUFFER TPM2B_MAX_BUFFER_PTR_value(TPM2B_MAX_BUFFER *obj) {
    return *obj;
  }


  static TPM2B_MAX_BUFFER * *new_TPM2B_MAX_BUFFER_PTR_PTR() { 
    return (TPM2B_MAX_BUFFER * *)malloc(sizeof(TPM2B_MAX_BUFFER *));
  }
  
  static TPM2B_MAX_BUFFER * *copy_TPM2B_MAX_BUFFER_PTR_PTR(TPM2B_MAX_BUFFER * value) { 
    return (TPM2B_MAX_BUFFER * *)memcpy((TPM2B_MAX_BUFFER * *)malloc(sizeof(TPM2B_MAX_BUFFER *)),&value,sizeof(TPM2B_MAX_BUFFER *));
  }

  static void delete_TPM2B_MAX_BUFFER_PTR_PTR(TPM2B_MAX_BUFFER * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_MAX_BUFFER_PTR_PTR_assign(TPM2B_MAX_BUFFER * *obj, TPM2B_MAX_BUFFER * value) {
    *obj = value;
  }

  static TPM2B_MAX_BUFFER * TPM2B_MAX_BUFFER_PTR_PTR_value(TPM2B_MAX_BUFFER * *obj) {
    return *obj;
  }


typedef TPM2B_MAX_NV_BUFFER TPM2B_MAX_NV_BUFFER_ARRAY;

SWIGINTERN TPM2B_MAX_NV_BUFFER_ARRAY *new_TPM2B_MAX_NV_BUFFER_ARRAY(size_t nelements){
    return (TPM2B_MAX_NV_BUFFER *)malloc((nelements)*sizeof(TPM2B_MAX_NV_BUFFER));
  }
SWIGINTERN void delete_TPM2B_MAX_NV_BUFFER_ARRAY(TPM2B_MAX_NV_BUFFER_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_MAX_NV_BUFFER TPM2B_MAX_NV_BUFFER_ARRAY___getitem__(TPM2B_MAX_NV_BUFFER_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_MAX_NV_BUFFER_ARRAY___setitem__(TPM2B_MAX_NV_BUFFER_ARRAY *self,size_t index,TPM2B_MAX_NV_BUFFER value){
    self[index] = value;
  }
SWIGINTERN TPM2B_MAX_NV_BUFFER *TPM2B_MAX_NV_BUFFER_ARRAY_cast(TPM2B_MAX_NV_BUFFER_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_MAX_NV_BUFFER_ARRAY *TPM2B_MAX_NV_BUFFER_ARRAY_frompointer(TPM2B_MAX_NV_BUFFER *t){
    return (TPM2B_MAX_NV_BUFFER_ARRAY *)(t);
  }

  static TPM2B_MAX_NV_BUFFER *new_TPM2B_MAX_NV_BUFFER_PTR() { 
    return (TPM2B_MAX_NV_BUFFER *)malloc(sizeof(TPM2B_MAX_NV_BUFFER));
  }
  
  static TPM2B_MAX_NV_BUFFER *copy_TPM2B_MAX_NV_BUFFER_PTR(TPM2B_MAX_NV_BUFFER value) { 
    return (TPM2B_MAX_NV_BUFFER *)memcpy((TPM2B_MAX_NV_BUFFER *)malloc(sizeof(TPM2B_MAX_NV_BUFFER)),&value,sizeof(TPM2B_MAX_NV_BUFFER));
  }

  static void delete_TPM2B_MAX_NV_BUFFER_PTR(TPM2B_MAX_NV_BUFFER *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_MAX_NV_BUFFER_PTR_assign(TPM2B_MAX_NV_BUFFER *obj, TPM2B_MAX_NV_BUFFER value) {
    *obj = value;
  }

  static TPM2B_MAX_NV_BUFFER TPM2B_MAX_NV_BUFFER_PTR_value(TPM2B_MAX_NV_BUFFER *obj) {
    return *obj;
  }


  static TPM2B_MAX_NV_BUFFER * *new_TPM2B_MAX_NV_BUFFER_PTR_PTR() { 
    return (TPM2B_MAX_NV_BUFFER * *)malloc(sizeof(TPM2B_MAX_NV_BUFFER *));
  }
  
  static TPM2B_MAX_NV_BUFFER * *copy_TPM2B_MAX_NV_BUFFER_PTR_PTR(TPM2B_MAX_NV_BUFFER * value) { 
    return (TPM2B_MAX_NV_BUFFER * *)memcpy((TPM2B_MAX_NV_BUFFER * *)malloc(sizeof(TPM2B_MAX_NV_BUFFER *)),&value,sizeof(TPM2B_MAX_NV_BUFFER *));
  }

  static void delete_TPM2B_MAX_NV_BUFFER_PTR_PTR(TPM2B_MAX_NV_BUFFER * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_MAX_NV_BUFFER_PTR_PTR_assign(TPM2B_MAX_NV_BUFFER * *obj, TPM2B_MAX_NV_BUFFER * value) {
    *obj = value;
  }

  static TPM2B_MAX_NV_BUFFER * TPM2B_MAX_NV_BUFFER_PTR_PTR_value(TPM2B_MAX_NV_BUFFER * *obj) {
    return *obj;
  }


typedef TPM2B_IV TPM2B_IV_ARRAY;

SWIGINTERN TPM2B_IV_ARRAY *new_TPM2B_IV_ARRAY(size_t nelements){
    return (TPM2B_IV *)malloc((nelements)*sizeof(TPM2B_IV));
  }
SWIGINTERN void delete_TPM2B_IV_ARRAY(TPM2B_IV_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_IV TPM2B_IV_ARRAY___getitem__(TPM2B_IV_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_IV_ARRAY___setitem__(TPM2B_IV_ARRAY *self,size_t index,TPM2B_IV value){
    self[index] = value;
  }
SWIGINTERN TPM2B_IV *TPM2B_IV_ARRAY_cast(TPM2B_IV_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_IV_ARRAY *TPM2B_IV_ARRAY_frompointer(TPM2B_IV *t){
    return (TPM2B_IV_ARRAY *)(t);
  }

  static TPM2B_IV *new_TPM2B_IV_PTR() { 
    return (TPM2B_IV *)malloc(sizeof(TPM2B_IV));
  }
  
  static TPM2B_IV *copy_TPM2B_IV_PTR(TPM2B_IV value) { 
    return (TPM2B_IV *)memcpy((TPM2B_IV *)malloc(sizeof(TPM2B_IV)),&value,sizeof(TPM2B_IV));
  }

  static void delete_TPM2B_IV_PTR(TPM2B_IV *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_IV_PTR_assign(TPM2B_IV *obj, TPM2B_IV value) {
    *obj = value;
  }

  static TPM2B_IV TPM2B_IV_PTR_value(TPM2B_IV *obj) {
    return *obj;
  }


  static TPM2B_IV * *new_TPM2B_IV_PTR_PTR() { 
    return (TPM2B_IV * *)malloc(sizeof(TPM2B_IV *));
  }
  
  static TPM2B_IV * *copy_TPM2B_IV_PTR_PTR(TPM2B_IV * value) { 
    return (TPM2B_IV * *)memcpy((TPM2B_IV * *)malloc(sizeof(TPM2B_IV *)),&value,sizeof(TPM2B_IV *));
  }

  static void delete_TPM2B_IV_PTR_PTR(TPM2B_IV * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_IV_PTR_PTR_assign(TPM2B_IV * *obj, TPM2B_IV * value) {
    *obj = value;
  }

  static TPM2B_IV * TPM2B_IV_PTR_PTR_value(TPM2B_IV * *obj) {
    return *obj;
  }


typedef TPMU_NAME TPMU_NAME_ARRAY;

SWIGINTERN TPMU_NAME_ARRAY *new_TPMU_NAME_ARRAY(size_t nelements){
    return (TPMU_NAME *)malloc((nelements)*sizeof(TPMU_NAME));
  }
SWIGINTERN void delete_TPMU_NAME_ARRAY(TPMU_NAME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_NAME TPMU_NAME_ARRAY___getitem__(TPMU_NAME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_NAME_ARRAY___setitem__(TPMU_NAME_ARRAY *self,size_t index,TPMU_NAME value){
    self[index] = value;
  }
SWIGINTERN TPMU_NAME *TPMU_NAME_ARRAY_cast(TPMU_NAME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_NAME_ARRAY *TPMU_NAME_ARRAY_frompointer(TPMU_NAME *t){
    return (TPMU_NAME_ARRAY *)(t);
  }

  static TPMU_NAME *new_TPMU_NAME_PTR() { 
    return (TPMU_NAME *)malloc(sizeof(TPMU_NAME));
  }
  
  static TPMU_NAME *copy_TPMU_NAME_PTR(TPMU_NAME value) { 
    return (TPMU_NAME *)memcpy((TPMU_NAME *)malloc(sizeof(TPMU_NAME)),&value,sizeof(TPMU_NAME));
  }

  static void delete_TPMU_NAME_PTR(TPMU_NAME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_NAME_PTR_assign(TPMU_NAME *obj, TPMU_NAME value) {
    *obj = value;
  }

  static TPMU_NAME TPMU_NAME_PTR_value(TPMU_NAME *obj) {
    return *obj;
  }


  static TPMU_NAME * *new_TPMU_NAME_PTR_PTR() { 
    return (TPMU_NAME * *)malloc(sizeof(TPMU_NAME *));
  }
  
  static TPMU_NAME * *copy_TPMU_NAME_PTR_PTR(TPMU_NAME * value) { 
    return (TPMU_NAME * *)memcpy((TPMU_NAME * *)malloc(sizeof(TPMU_NAME *)),&value,sizeof(TPMU_NAME *));
  }

  static void delete_TPMU_NAME_PTR_PTR(TPMU_NAME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_NAME_PTR_PTR_assign(TPMU_NAME * *obj, TPMU_NAME * value) {
    *obj = value;
  }

  static TPMU_NAME * TPMU_NAME_PTR_PTR_value(TPMU_NAME * *obj) {
    return *obj;
  }


typedef TPM2B_NAME TPM2B_NAME_ARRAY;

SWIGINTERN TPM2B_NAME_ARRAY *new_TPM2B_NAME_ARRAY(size_t nelements){
    return (TPM2B_NAME *)malloc((nelements)*sizeof(TPM2B_NAME));
  }
SWIGINTERN void delete_TPM2B_NAME_ARRAY(TPM2B_NAME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_NAME TPM2B_NAME_ARRAY___getitem__(TPM2B_NAME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_NAME_ARRAY___setitem__(TPM2B_NAME_ARRAY *self,size_t index,TPM2B_NAME value){
    self[index] = value;
  }
SWIGINTERN TPM2B_NAME *TPM2B_NAME_ARRAY_cast(TPM2B_NAME_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_NAME_ARRAY *TPM2B_NAME_ARRAY_frompointer(TPM2B_NAME *t){
    return (TPM2B_NAME_ARRAY *)(t);
  }

  static TPM2B_NAME *new_TPM2B_NAME_PTR() { 
    return (TPM2B_NAME *)malloc(sizeof(TPM2B_NAME));
  }
  
  static TPM2B_NAME *copy_TPM2B_NAME_PTR(TPM2B_NAME value) { 
    return (TPM2B_NAME *)memcpy((TPM2B_NAME *)malloc(sizeof(TPM2B_NAME)),&value,sizeof(TPM2B_NAME));
  }

  static void delete_TPM2B_NAME_PTR(TPM2B_NAME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_NAME_PTR_assign(TPM2B_NAME *obj, TPM2B_NAME value) {
    *obj = value;
  }

  static TPM2B_NAME TPM2B_NAME_PTR_value(TPM2B_NAME *obj) {
    return *obj;
  }


  static TPM2B_NAME * *new_TPM2B_NAME_PTR_PTR() { 
    return (TPM2B_NAME * *)malloc(sizeof(TPM2B_NAME *));
  }
  
  static TPM2B_NAME * *copy_TPM2B_NAME_PTR_PTR(TPM2B_NAME * value) { 
    return (TPM2B_NAME * *)memcpy((TPM2B_NAME * *)malloc(sizeof(TPM2B_NAME *)),&value,sizeof(TPM2B_NAME *));
  }

  static void delete_TPM2B_NAME_PTR_PTR(TPM2B_NAME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_NAME_PTR_PTR_assign(TPM2B_NAME * *obj, TPM2B_NAME * value) {
    *obj = value;
  }

  static TPM2B_NAME * TPM2B_NAME_PTR_PTR_value(TPM2B_NAME * *obj) {
    return *obj;
  }


typedef TPMS_PCR_SELECT TPMS_PCR_SELECT_ARRAY;

SWIGINTERN TPMS_PCR_SELECT_ARRAY *new_TPMS_PCR_SELECT_ARRAY(size_t nelements){
    return (TPMS_PCR_SELECT *)malloc((nelements)*sizeof(TPMS_PCR_SELECT));
  }
SWIGINTERN void delete_TPMS_PCR_SELECT_ARRAY(TPMS_PCR_SELECT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_PCR_SELECT TPMS_PCR_SELECT_ARRAY___getitem__(TPMS_PCR_SELECT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_PCR_SELECT_ARRAY___setitem__(TPMS_PCR_SELECT_ARRAY *self,size_t index,TPMS_PCR_SELECT value){
    self[index] = value;
  }
SWIGINTERN TPMS_PCR_SELECT *TPMS_PCR_SELECT_ARRAY_cast(TPMS_PCR_SELECT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_PCR_SELECT_ARRAY *TPMS_PCR_SELECT_ARRAY_frompointer(TPMS_PCR_SELECT *t){
    return (TPMS_PCR_SELECT_ARRAY *)(t);
  }

  static TPMS_PCR_SELECT *new_TPMS_PCR_SELECT_PTR() { 
    return (TPMS_PCR_SELECT *)malloc(sizeof(TPMS_PCR_SELECT));
  }
  
  static TPMS_PCR_SELECT *copy_TPMS_PCR_SELECT_PTR(TPMS_PCR_SELECT value) { 
    return (TPMS_PCR_SELECT *)memcpy((TPMS_PCR_SELECT *)malloc(sizeof(TPMS_PCR_SELECT)),&value,sizeof(TPMS_PCR_SELECT));
  }

  static void delete_TPMS_PCR_SELECT_PTR(TPMS_PCR_SELECT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_PCR_SELECT_PTR_assign(TPMS_PCR_SELECT *obj, TPMS_PCR_SELECT value) {
    *obj = value;
  }

  static TPMS_PCR_SELECT TPMS_PCR_SELECT_PTR_value(TPMS_PCR_SELECT *obj) {
    return *obj;
  }


  static TPMS_PCR_SELECT * *new_TPMS_PCR_SELECT_PTR_PTR() { 
    return (TPMS_PCR_SELECT * *)malloc(sizeof(TPMS_PCR_SELECT *));
  }
  
  static TPMS_PCR_SELECT * *copy_TPMS_PCR_SELECT_PTR_PTR(TPMS_PCR_SELECT * value) { 
    return (TPMS_PCR_SELECT * *)memcpy((TPMS_PCR_SELECT * *)malloc(sizeof(TPMS_PCR_SELECT *)),&value,sizeof(TPMS_PCR_SELECT *));
  }

  static void delete_TPMS_PCR_SELECT_PTR_PTR(TPMS_PCR_SELECT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_PCR_SELECT_PTR_PTR_assign(TPMS_PCR_SELECT * *obj, TPMS_PCR_SELECT * value) {
    *obj = value;
  }

  static TPMS_PCR_SELECT * TPMS_PCR_SELECT_PTR_PTR_value(TPMS_PCR_SELECT * *obj) {
    return *obj;
  }


typedef TPMS_PCR_SELECTION TPMS_PCR_SELECTION_ARRAY;

SWIGINTERN TPMS_PCR_SELECTION_ARRAY *new_TPMS_PCR_SELECTION_ARRAY(size_t nelements){
    return (TPMS_PCR_SELECTION *)malloc((nelements)*sizeof(TPMS_PCR_SELECTION));
  }
SWIGINTERN void delete_TPMS_PCR_SELECTION_ARRAY(TPMS_PCR_SELECTION_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_PCR_SELECTION TPMS_PCR_SELECTION_ARRAY___getitem__(TPMS_PCR_SELECTION_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_PCR_SELECTION_ARRAY___setitem__(TPMS_PCR_SELECTION_ARRAY *self,size_t index,TPMS_PCR_SELECTION value){
    self[index] = value;
  }
SWIGINTERN TPMS_PCR_SELECTION *TPMS_PCR_SELECTION_ARRAY_cast(TPMS_PCR_SELECTION_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_PCR_SELECTION_ARRAY *TPMS_PCR_SELECTION_ARRAY_frompointer(TPMS_PCR_SELECTION *t){
    return (TPMS_PCR_SELECTION_ARRAY *)(t);
  }

  static TPMS_PCR_SELECTION *new_TPMS_PCR_SELECTION_PTR() { 
    return (TPMS_PCR_SELECTION *)malloc(sizeof(TPMS_PCR_SELECTION));
  }
  
  static TPMS_PCR_SELECTION *copy_TPMS_PCR_SELECTION_PTR(TPMS_PCR_SELECTION value) { 
    return (TPMS_PCR_SELECTION *)memcpy((TPMS_PCR_SELECTION *)malloc(sizeof(TPMS_PCR_SELECTION)),&value,sizeof(TPMS_PCR_SELECTION));
  }

  static void delete_TPMS_PCR_SELECTION_PTR(TPMS_PCR_SELECTION *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_PCR_SELECTION_PTR_assign(TPMS_PCR_SELECTION *obj, TPMS_PCR_SELECTION value) {
    *obj = value;
  }

  static TPMS_PCR_SELECTION TPMS_PCR_SELECTION_PTR_value(TPMS_PCR_SELECTION *obj) {
    return *obj;
  }


  static TPMS_PCR_SELECTION * *new_TPMS_PCR_SELECTION_PTR_PTR() { 
    return (TPMS_PCR_SELECTION * *)malloc(sizeof(TPMS_PCR_SELECTION *));
  }
  
  static TPMS_PCR_SELECTION * *copy_TPMS_PCR_SELECTION_PTR_PTR(TPMS_PCR_SELECTION * value) { 
    return (TPMS_PCR_SELECTION * *)memcpy((TPMS_PCR_SELECTION * *)malloc(sizeof(TPMS_PCR_SELECTION *)),&value,sizeof(TPMS_PCR_SELECTION *));
  }

  static void delete_TPMS_PCR_SELECTION_PTR_PTR(TPMS_PCR_SELECTION * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_PCR_SELECTION_PTR_PTR_assign(TPMS_PCR_SELECTION * *obj, TPMS_PCR_SELECTION * value) {
    *obj = value;
  }

  static TPMS_PCR_SELECTION * TPMS_PCR_SELECTION_PTR_PTR_value(TPMS_PCR_SELECTION * *obj) {
    return *obj;
  }


typedef TPMT_TK_CREATION TPMT_TK_CREATION_ARRAY;

SWIGINTERN TPMT_TK_CREATION_ARRAY *new_TPMT_TK_CREATION_ARRAY(size_t nelements){
    return (TPMT_TK_CREATION *)malloc((nelements)*sizeof(TPMT_TK_CREATION));
  }
SWIGINTERN void delete_TPMT_TK_CREATION_ARRAY(TPMT_TK_CREATION_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_TK_CREATION TPMT_TK_CREATION_ARRAY___getitem__(TPMT_TK_CREATION_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_TK_CREATION_ARRAY___setitem__(TPMT_TK_CREATION_ARRAY *self,size_t index,TPMT_TK_CREATION value){
    self[index] = value;
  }
SWIGINTERN TPMT_TK_CREATION *TPMT_TK_CREATION_ARRAY_cast(TPMT_TK_CREATION_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_TK_CREATION_ARRAY *TPMT_TK_CREATION_ARRAY_frompointer(TPMT_TK_CREATION *t){
    return (TPMT_TK_CREATION_ARRAY *)(t);
  }

  static TPMT_TK_CREATION *new_TPMT_TK_CREATION_PTR() { 
    return (TPMT_TK_CREATION *)malloc(sizeof(TPMT_TK_CREATION));
  }
  
  static TPMT_TK_CREATION *copy_TPMT_TK_CREATION_PTR(TPMT_TK_CREATION value) { 
    return (TPMT_TK_CREATION *)memcpy((TPMT_TK_CREATION *)malloc(sizeof(TPMT_TK_CREATION)),&value,sizeof(TPMT_TK_CREATION));
  }

  static void delete_TPMT_TK_CREATION_PTR(TPMT_TK_CREATION *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_CREATION_PTR_assign(TPMT_TK_CREATION *obj, TPMT_TK_CREATION value) {
    *obj = value;
  }

  static TPMT_TK_CREATION TPMT_TK_CREATION_PTR_value(TPMT_TK_CREATION *obj) {
    return *obj;
  }


  static TPMT_TK_CREATION * *new_TPMT_TK_CREATION_PTR_PTR() { 
    return (TPMT_TK_CREATION * *)malloc(sizeof(TPMT_TK_CREATION *));
  }
  
  static TPMT_TK_CREATION * *copy_TPMT_TK_CREATION_PTR_PTR(TPMT_TK_CREATION * value) { 
    return (TPMT_TK_CREATION * *)memcpy((TPMT_TK_CREATION * *)malloc(sizeof(TPMT_TK_CREATION *)),&value,sizeof(TPMT_TK_CREATION *));
  }

  static void delete_TPMT_TK_CREATION_PTR_PTR(TPMT_TK_CREATION * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_CREATION_PTR_PTR_assign(TPMT_TK_CREATION * *obj, TPMT_TK_CREATION * value) {
    *obj = value;
  }

  static TPMT_TK_CREATION * TPMT_TK_CREATION_PTR_PTR_value(TPMT_TK_CREATION * *obj) {
    return *obj;
  }


typedef TPMT_TK_VERIFIED TPMT_TK_VERIFIED_ARRAY;

SWIGINTERN TPMT_TK_VERIFIED_ARRAY *new_TPMT_TK_VERIFIED_ARRAY(size_t nelements){
    return (TPMT_TK_VERIFIED *)malloc((nelements)*sizeof(TPMT_TK_VERIFIED));
  }
SWIGINTERN void delete_TPMT_TK_VERIFIED_ARRAY(TPMT_TK_VERIFIED_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_TK_VERIFIED TPMT_TK_VERIFIED_ARRAY___getitem__(TPMT_TK_VERIFIED_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_TK_VERIFIED_ARRAY___setitem__(TPMT_TK_VERIFIED_ARRAY *self,size_t index,TPMT_TK_VERIFIED value){
    self[index] = value;
  }
SWIGINTERN TPMT_TK_VERIFIED *TPMT_TK_VERIFIED_ARRAY_cast(TPMT_TK_VERIFIED_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_TK_VERIFIED_ARRAY *TPMT_TK_VERIFIED_ARRAY_frompointer(TPMT_TK_VERIFIED *t){
    return (TPMT_TK_VERIFIED_ARRAY *)(t);
  }

  static TPMT_TK_VERIFIED *new_TPMT_TK_VERIFIED_PTR() { 
    return (TPMT_TK_VERIFIED *)malloc(sizeof(TPMT_TK_VERIFIED));
  }
  
  static TPMT_TK_VERIFIED *copy_TPMT_TK_VERIFIED_PTR(TPMT_TK_VERIFIED value) { 
    return (TPMT_TK_VERIFIED *)memcpy((TPMT_TK_VERIFIED *)malloc(sizeof(TPMT_TK_VERIFIED)),&value,sizeof(TPMT_TK_VERIFIED));
  }

  static void delete_TPMT_TK_VERIFIED_PTR(TPMT_TK_VERIFIED *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_VERIFIED_PTR_assign(TPMT_TK_VERIFIED *obj, TPMT_TK_VERIFIED value) {
    *obj = value;
  }

  static TPMT_TK_VERIFIED TPMT_TK_VERIFIED_PTR_value(TPMT_TK_VERIFIED *obj) {
    return *obj;
  }


  static TPMT_TK_VERIFIED * *new_TPMT_TK_VERIFIED_PTR_PTR() { 
    return (TPMT_TK_VERIFIED * *)malloc(sizeof(TPMT_TK_VERIFIED *));
  }
  
  static TPMT_TK_VERIFIED * *copy_TPMT_TK_VERIFIED_PTR_PTR(TPMT_TK_VERIFIED * value) { 
    return (TPMT_TK_VERIFIED * *)memcpy((TPMT_TK_VERIFIED * *)malloc(sizeof(TPMT_TK_VERIFIED *)),&value,sizeof(TPMT_TK_VERIFIED *));
  }

  static void delete_TPMT_TK_VERIFIED_PTR_PTR(TPMT_TK_VERIFIED * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_VERIFIED_PTR_PTR_assign(TPMT_TK_VERIFIED * *obj, TPMT_TK_VERIFIED * value) {
    *obj = value;
  }

  static TPMT_TK_VERIFIED * TPMT_TK_VERIFIED_PTR_PTR_value(TPMT_TK_VERIFIED * *obj) {
    return *obj;
  }


typedef TPMT_TK_AUTH TPMT_TK_AUTH_ARRAY;

SWIGINTERN TPMT_TK_AUTH_ARRAY *new_TPMT_TK_AUTH_ARRAY(size_t nelements){
    return (TPMT_TK_AUTH *)malloc((nelements)*sizeof(TPMT_TK_AUTH));
  }
SWIGINTERN void delete_TPMT_TK_AUTH_ARRAY(TPMT_TK_AUTH_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_TK_AUTH TPMT_TK_AUTH_ARRAY___getitem__(TPMT_TK_AUTH_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_TK_AUTH_ARRAY___setitem__(TPMT_TK_AUTH_ARRAY *self,size_t index,TPMT_TK_AUTH value){
    self[index] = value;
  }
SWIGINTERN TPMT_TK_AUTH *TPMT_TK_AUTH_ARRAY_cast(TPMT_TK_AUTH_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_TK_AUTH_ARRAY *TPMT_TK_AUTH_ARRAY_frompointer(TPMT_TK_AUTH *t){
    return (TPMT_TK_AUTH_ARRAY *)(t);
  }

  static TPMT_TK_AUTH *new_TPMT_TK_AUTH_PTR() { 
    return (TPMT_TK_AUTH *)malloc(sizeof(TPMT_TK_AUTH));
  }
  
  static TPMT_TK_AUTH *copy_TPMT_TK_AUTH_PTR(TPMT_TK_AUTH value) { 
    return (TPMT_TK_AUTH *)memcpy((TPMT_TK_AUTH *)malloc(sizeof(TPMT_TK_AUTH)),&value,sizeof(TPMT_TK_AUTH));
  }

  static void delete_TPMT_TK_AUTH_PTR(TPMT_TK_AUTH *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_AUTH_PTR_assign(TPMT_TK_AUTH *obj, TPMT_TK_AUTH value) {
    *obj = value;
  }

  static TPMT_TK_AUTH TPMT_TK_AUTH_PTR_value(TPMT_TK_AUTH *obj) {
    return *obj;
  }


  static TPMT_TK_AUTH * *new_TPMT_TK_AUTH_PTR_PTR() { 
    return (TPMT_TK_AUTH * *)malloc(sizeof(TPMT_TK_AUTH *));
  }
  
  static TPMT_TK_AUTH * *copy_TPMT_TK_AUTH_PTR_PTR(TPMT_TK_AUTH * value) { 
    return (TPMT_TK_AUTH * *)memcpy((TPMT_TK_AUTH * *)malloc(sizeof(TPMT_TK_AUTH *)),&value,sizeof(TPMT_TK_AUTH *));
  }

  static void delete_TPMT_TK_AUTH_PTR_PTR(TPMT_TK_AUTH * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_AUTH_PTR_PTR_assign(TPMT_TK_AUTH * *obj, TPMT_TK_AUTH * value) {
    *obj = value;
  }

  static TPMT_TK_AUTH * TPMT_TK_AUTH_PTR_PTR_value(TPMT_TK_AUTH * *obj) {
    return *obj;
  }


typedef TPMT_TK_HASHCHECK TPMT_TK_HASHCHECK_ARRAY;

SWIGINTERN TPMT_TK_HASHCHECK_ARRAY *new_TPMT_TK_HASHCHECK_ARRAY(size_t nelements){
    return (TPMT_TK_HASHCHECK *)malloc((nelements)*sizeof(TPMT_TK_HASHCHECK));
  }
SWIGINTERN void delete_TPMT_TK_HASHCHECK_ARRAY(TPMT_TK_HASHCHECK_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_TK_HASHCHECK TPMT_TK_HASHCHECK_ARRAY___getitem__(TPMT_TK_HASHCHECK_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_TK_HASHCHECK_ARRAY___setitem__(TPMT_TK_HASHCHECK_ARRAY *self,size_t index,TPMT_TK_HASHCHECK value){
    self[index] = value;
  }
SWIGINTERN TPMT_TK_HASHCHECK *TPMT_TK_HASHCHECK_ARRAY_cast(TPMT_TK_HASHCHECK_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_TK_HASHCHECK_ARRAY *TPMT_TK_HASHCHECK_ARRAY_frompointer(TPMT_TK_HASHCHECK *t){
    return (TPMT_TK_HASHCHECK_ARRAY *)(t);
  }

  static TPMT_TK_HASHCHECK *new_TPMT_TK_HASHCHECK_PTR() { 
    return (TPMT_TK_HASHCHECK *)malloc(sizeof(TPMT_TK_HASHCHECK));
  }
  
  static TPMT_TK_HASHCHECK *copy_TPMT_TK_HASHCHECK_PTR(TPMT_TK_HASHCHECK value) { 
    return (TPMT_TK_HASHCHECK *)memcpy((TPMT_TK_HASHCHECK *)malloc(sizeof(TPMT_TK_HASHCHECK)),&value,sizeof(TPMT_TK_HASHCHECK));
  }

  static void delete_TPMT_TK_HASHCHECK_PTR(TPMT_TK_HASHCHECK *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_HASHCHECK_PTR_assign(TPMT_TK_HASHCHECK *obj, TPMT_TK_HASHCHECK value) {
    *obj = value;
  }

  static TPMT_TK_HASHCHECK TPMT_TK_HASHCHECK_PTR_value(TPMT_TK_HASHCHECK *obj) {
    return *obj;
  }


  static TPMT_TK_HASHCHECK * *new_TPMT_TK_HASHCHECK_PTR_PTR() { 
    return (TPMT_TK_HASHCHECK * *)malloc(sizeof(TPMT_TK_HASHCHECK *));
  }
  
  static TPMT_TK_HASHCHECK * *copy_TPMT_TK_HASHCHECK_PTR_PTR(TPMT_TK_HASHCHECK * value) { 
    return (TPMT_TK_HASHCHECK * *)memcpy((TPMT_TK_HASHCHECK * *)malloc(sizeof(TPMT_TK_HASHCHECK *)),&value,sizeof(TPMT_TK_HASHCHECK *));
  }

  static void delete_TPMT_TK_HASHCHECK_PTR_PTR(TPMT_TK_HASHCHECK * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_TK_HASHCHECK_PTR_PTR_assign(TPMT_TK_HASHCHECK * *obj, TPMT_TK_HASHCHECK * value) {
    *obj = value;
  }

  static TPMT_TK_HASHCHECK * TPMT_TK_HASHCHECK_PTR_PTR_value(TPMT_TK_HASHCHECK * *obj) {
    return *obj;
  }


typedef TPMS_ALG_PROPERTY TPMS_ALG_PROPERTY_ARRAY;

SWIGINTERN TPMS_ALG_PROPERTY_ARRAY *new_TPMS_ALG_PROPERTY_ARRAY(size_t nelements){
    return (TPMS_ALG_PROPERTY *)malloc((nelements)*sizeof(TPMS_ALG_PROPERTY));
  }
SWIGINTERN void delete_TPMS_ALG_PROPERTY_ARRAY(TPMS_ALG_PROPERTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ALG_PROPERTY TPMS_ALG_PROPERTY_ARRAY___getitem__(TPMS_ALG_PROPERTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ALG_PROPERTY_ARRAY___setitem__(TPMS_ALG_PROPERTY_ARRAY *self,size_t index,TPMS_ALG_PROPERTY value){
    self[index] = value;
  }
SWIGINTERN TPMS_ALG_PROPERTY *TPMS_ALG_PROPERTY_ARRAY_cast(TPMS_ALG_PROPERTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ALG_PROPERTY_ARRAY *TPMS_ALG_PROPERTY_ARRAY_frompointer(TPMS_ALG_PROPERTY *t){
    return (TPMS_ALG_PROPERTY_ARRAY *)(t);
  }

  static TPMS_ALG_PROPERTY *new_TPMS_ALG_PROPERTY_PTR() { 
    return (TPMS_ALG_PROPERTY *)malloc(sizeof(TPMS_ALG_PROPERTY));
  }
  
  static TPMS_ALG_PROPERTY *copy_TPMS_ALG_PROPERTY_PTR(TPMS_ALG_PROPERTY value) { 
    return (TPMS_ALG_PROPERTY *)memcpy((TPMS_ALG_PROPERTY *)malloc(sizeof(TPMS_ALG_PROPERTY)),&value,sizeof(TPMS_ALG_PROPERTY));
  }

  static void delete_TPMS_ALG_PROPERTY_PTR(TPMS_ALG_PROPERTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ALG_PROPERTY_PTR_assign(TPMS_ALG_PROPERTY *obj, TPMS_ALG_PROPERTY value) {
    *obj = value;
  }

  static TPMS_ALG_PROPERTY TPMS_ALG_PROPERTY_PTR_value(TPMS_ALG_PROPERTY *obj) {
    return *obj;
  }


  static TPMS_ALG_PROPERTY * *new_TPMS_ALG_PROPERTY_PTR_PTR() { 
    return (TPMS_ALG_PROPERTY * *)malloc(sizeof(TPMS_ALG_PROPERTY *));
  }
  
  static TPMS_ALG_PROPERTY * *copy_TPMS_ALG_PROPERTY_PTR_PTR(TPMS_ALG_PROPERTY * value) { 
    return (TPMS_ALG_PROPERTY * *)memcpy((TPMS_ALG_PROPERTY * *)malloc(sizeof(TPMS_ALG_PROPERTY *)),&value,sizeof(TPMS_ALG_PROPERTY *));
  }

  static void delete_TPMS_ALG_PROPERTY_PTR_PTR(TPMS_ALG_PROPERTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ALG_PROPERTY_PTR_PTR_assign(TPMS_ALG_PROPERTY * *obj, TPMS_ALG_PROPERTY * value) {
    *obj = value;
  }

  static TPMS_ALG_PROPERTY * TPMS_ALG_PROPERTY_PTR_PTR_value(TPMS_ALG_PROPERTY * *obj) {
    return *obj;
  }


typedef TPMS_TAGGED_PROPERTY TPMS_TAGGED_PROPERTY_ARRAY;

SWIGINTERN TPMS_TAGGED_PROPERTY_ARRAY *new_TPMS_TAGGED_PROPERTY_ARRAY(size_t nelements){
    return (TPMS_TAGGED_PROPERTY *)malloc((nelements)*sizeof(TPMS_TAGGED_PROPERTY));
  }
SWIGINTERN void delete_TPMS_TAGGED_PROPERTY_ARRAY(TPMS_TAGGED_PROPERTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_TAGGED_PROPERTY TPMS_TAGGED_PROPERTY_ARRAY___getitem__(TPMS_TAGGED_PROPERTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_TAGGED_PROPERTY_ARRAY___setitem__(TPMS_TAGGED_PROPERTY_ARRAY *self,size_t index,TPMS_TAGGED_PROPERTY value){
    self[index] = value;
  }
SWIGINTERN TPMS_TAGGED_PROPERTY *TPMS_TAGGED_PROPERTY_ARRAY_cast(TPMS_TAGGED_PROPERTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_TAGGED_PROPERTY_ARRAY *TPMS_TAGGED_PROPERTY_ARRAY_frompointer(TPMS_TAGGED_PROPERTY *t){
    return (TPMS_TAGGED_PROPERTY_ARRAY *)(t);
  }

  static TPMS_TAGGED_PROPERTY *new_TPMS_TAGGED_PROPERTY_PTR() { 
    return (TPMS_TAGGED_PROPERTY *)malloc(sizeof(TPMS_TAGGED_PROPERTY));
  }
  
  static TPMS_TAGGED_PROPERTY *copy_TPMS_TAGGED_PROPERTY_PTR(TPMS_TAGGED_PROPERTY value) { 
    return (TPMS_TAGGED_PROPERTY *)memcpy((TPMS_TAGGED_PROPERTY *)malloc(sizeof(TPMS_TAGGED_PROPERTY)),&value,sizeof(TPMS_TAGGED_PROPERTY));
  }

  static void delete_TPMS_TAGGED_PROPERTY_PTR(TPMS_TAGGED_PROPERTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TAGGED_PROPERTY_PTR_assign(TPMS_TAGGED_PROPERTY *obj, TPMS_TAGGED_PROPERTY value) {
    *obj = value;
  }

  static TPMS_TAGGED_PROPERTY TPMS_TAGGED_PROPERTY_PTR_value(TPMS_TAGGED_PROPERTY *obj) {
    return *obj;
  }


  static TPMS_TAGGED_PROPERTY * *new_TPMS_TAGGED_PROPERTY_PTR_PTR() { 
    return (TPMS_TAGGED_PROPERTY * *)malloc(sizeof(TPMS_TAGGED_PROPERTY *));
  }
  
  static TPMS_TAGGED_PROPERTY * *copy_TPMS_TAGGED_PROPERTY_PTR_PTR(TPMS_TAGGED_PROPERTY * value) { 
    return (TPMS_TAGGED_PROPERTY * *)memcpy((TPMS_TAGGED_PROPERTY * *)malloc(sizeof(TPMS_TAGGED_PROPERTY *)),&value,sizeof(TPMS_TAGGED_PROPERTY *));
  }

  static void delete_TPMS_TAGGED_PROPERTY_PTR_PTR(TPMS_TAGGED_PROPERTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TAGGED_PROPERTY_PTR_PTR_assign(TPMS_TAGGED_PROPERTY * *obj, TPMS_TAGGED_PROPERTY * value) {
    *obj = value;
  }

  static TPMS_TAGGED_PROPERTY * TPMS_TAGGED_PROPERTY_PTR_PTR_value(TPMS_TAGGED_PROPERTY * *obj) {
    return *obj;
  }


typedef TPMS_TAGGED_PCR_SELECT TPMS_TAGGED_PCR_SELECT_ARRAY;

SWIGINTERN TPMS_TAGGED_PCR_SELECT_ARRAY *new_TPMS_TAGGED_PCR_SELECT_ARRAY(size_t nelements){
    return (TPMS_TAGGED_PCR_SELECT *)malloc((nelements)*sizeof(TPMS_TAGGED_PCR_SELECT));
  }
SWIGINTERN void delete_TPMS_TAGGED_PCR_SELECT_ARRAY(TPMS_TAGGED_PCR_SELECT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_TAGGED_PCR_SELECT TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__(TPMS_TAGGED_PCR_SELECT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__(TPMS_TAGGED_PCR_SELECT_ARRAY *self,size_t index,TPMS_TAGGED_PCR_SELECT value){
    self[index] = value;
  }
SWIGINTERN TPMS_TAGGED_PCR_SELECT *TPMS_TAGGED_PCR_SELECT_ARRAY_cast(TPMS_TAGGED_PCR_SELECT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_TAGGED_PCR_SELECT_ARRAY *TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer(TPMS_TAGGED_PCR_SELECT *t){
    return (TPMS_TAGGED_PCR_SELECT_ARRAY *)(t);
  }

  static TPMS_TAGGED_PCR_SELECT *new_TPMS_TAGGED_PCR_SELECT_PTR() { 
    return (TPMS_TAGGED_PCR_SELECT *)malloc(sizeof(TPMS_TAGGED_PCR_SELECT));
  }
  
  static TPMS_TAGGED_PCR_SELECT *copy_TPMS_TAGGED_PCR_SELECT_PTR(TPMS_TAGGED_PCR_SELECT value) { 
    return (TPMS_TAGGED_PCR_SELECT *)memcpy((TPMS_TAGGED_PCR_SELECT *)malloc(sizeof(TPMS_TAGGED_PCR_SELECT)),&value,sizeof(TPMS_TAGGED_PCR_SELECT));
  }

  static void delete_TPMS_TAGGED_PCR_SELECT_PTR(TPMS_TAGGED_PCR_SELECT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TAGGED_PCR_SELECT_PTR_assign(TPMS_TAGGED_PCR_SELECT *obj, TPMS_TAGGED_PCR_SELECT value) {
    *obj = value;
  }

  static TPMS_TAGGED_PCR_SELECT TPMS_TAGGED_PCR_SELECT_PTR_value(TPMS_TAGGED_PCR_SELECT *obj) {
    return *obj;
  }


  static TPMS_TAGGED_PCR_SELECT * *new_TPMS_TAGGED_PCR_SELECT_PTR_PTR() { 
    return (TPMS_TAGGED_PCR_SELECT * *)malloc(sizeof(TPMS_TAGGED_PCR_SELECT *));
  }
  
  static TPMS_TAGGED_PCR_SELECT * *copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR(TPMS_TAGGED_PCR_SELECT * value) { 
    return (TPMS_TAGGED_PCR_SELECT * *)memcpy((TPMS_TAGGED_PCR_SELECT * *)malloc(sizeof(TPMS_TAGGED_PCR_SELECT *)),&value,sizeof(TPMS_TAGGED_PCR_SELECT *));
  }

  static void delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR(TPMS_TAGGED_PCR_SELECT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign(TPMS_TAGGED_PCR_SELECT * *obj, TPMS_TAGGED_PCR_SELECT * value) {
    *obj = value;
  }

  static TPMS_TAGGED_PCR_SELECT * TPMS_TAGGED_PCR_SELECT_PTR_PTR_value(TPMS_TAGGED_PCR_SELECT * *obj) {
    return *obj;
  }


typedef TPMS_TAGGED_POLICY TPMS_TAGGED_POLICY_ARRAY;

SWIGINTERN TPMS_TAGGED_POLICY_ARRAY *new_TPMS_TAGGED_POLICY_ARRAY(size_t nelements){
    return (TPMS_TAGGED_POLICY *)malloc((nelements)*sizeof(TPMS_TAGGED_POLICY));
  }
SWIGINTERN void delete_TPMS_TAGGED_POLICY_ARRAY(TPMS_TAGGED_POLICY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_TAGGED_POLICY TPMS_TAGGED_POLICY_ARRAY___getitem__(TPMS_TAGGED_POLICY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_TAGGED_POLICY_ARRAY___setitem__(TPMS_TAGGED_POLICY_ARRAY *self,size_t index,TPMS_TAGGED_POLICY value){
    self[index] = value;
  }
SWIGINTERN TPMS_TAGGED_POLICY *TPMS_TAGGED_POLICY_ARRAY_cast(TPMS_TAGGED_POLICY_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_TAGGED_POLICY_ARRAY *TPMS_TAGGED_POLICY_ARRAY_frompointer(TPMS_TAGGED_POLICY *t){
    return (TPMS_TAGGED_POLICY_ARRAY *)(t);
  }

  static TPMS_TAGGED_POLICY *new_TPMS_TAGGED_POLICY_PTR() { 
    return (TPMS_TAGGED_POLICY *)malloc(sizeof(TPMS_TAGGED_POLICY));
  }
  
  static TPMS_TAGGED_POLICY *copy_TPMS_TAGGED_POLICY_PTR(TPMS_TAGGED_POLICY value) { 
    return (TPMS_TAGGED_POLICY *)memcpy((TPMS_TAGGED_POLICY *)malloc(sizeof(TPMS_TAGGED_POLICY)),&value,sizeof(TPMS_TAGGED_POLICY));
  }

  static void delete_TPMS_TAGGED_POLICY_PTR(TPMS_TAGGED_POLICY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TAGGED_POLICY_PTR_assign(TPMS_TAGGED_POLICY *obj, TPMS_TAGGED_POLICY value) {
    *obj = value;
  }

  static TPMS_TAGGED_POLICY TPMS_TAGGED_POLICY_PTR_value(TPMS_TAGGED_POLICY *obj) {
    return *obj;
  }


  static TPMS_TAGGED_POLICY * *new_TPMS_TAGGED_POLICY_PTR_PTR() { 
    return (TPMS_TAGGED_POLICY * *)malloc(sizeof(TPMS_TAGGED_POLICY *));
  }
  
  static TPMS_TAGGED_POLICY * *copy_TPMS_TAGGED_POLICY_PTR_PTR(TPMS_TAGGED_POLICY * value) { 
    return (TPMS_TAGGED_POLICY * *)memcpy((TPMS_TAGGED_POLICY * *)malloc(sizeof(TPMS_TAGGED_POLICY *)),&value,sizeof(TPMS_TAGGED_POLICY *));
  }

  static void delete_TPMS_TAGGED_POLICY_PTR_PTR(TPMS_TAGGED_POLICY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TAGGED_POLICY_PTR_PTR_assign(TPMS_TAGGED_POLICY * *obj, TPMS_TAGGED_POLICY * value) {
    *obj = value;
  }

  static TPMS_TAGGED_POLICY * TPMS_TAGGED_POLICY_PTR_PTR_value(TPMS_TAGGED_POLICY * *obj) {
    return *obj;
  }


typedef TPML_CC TPML_CC_ARRAY;

SWIGINTERN TPML_CC_ARRAY *new_TPML_CC_ARRAY(size_t nelements){
    return (TPML_CC *)malloc((nelements)*sizeof(TPML_CC));
  }
SWIGINTERN void delete_TPML_CC_ARRAY(TPML_CC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_CC TPML_CC_ARRAY___getitem__(TPML_CC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_CC_ARRAY___setitem__(TPML_CC_ARRAY *self,size_t index,TPML_CC value){
    self[index] = value;
  }
SWIGINTERN TPML_CC *TPML_CC_ARRAY_cast(TPML_CC_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_CC_ARRAY *TPML_CC_ARRAY_frompointer(TPML_CC *t){
    return (TPML_CC_ARRAY *)(t);
  }

  static TPML_CC *new_TPML_CC_PTR() { 
    return (TPML_CC *)malloc(sizeof(TPML_CC));
  }
  
  static TPML_CC *copy_TPML_CC_PTR(TPML_CC value) { 
    return (TPML_CC *)memcpy((TPML_CC *)malloc(sizeof(TPML_CC)),&value,sizeof(TPML_CC));
  }

  static void delete_TPML_CC_PTR(TPML_CC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_CC_PTR_assign(TPML_CC *obj, TPML_CC value) {
    *obj = value;
  }

  static TPML_CC TPML_CC_PTR_value(TPML_CC *obj) {
    return *obj;
  }


  static TPML_CC * *new_TPML_CC_PTR_PTR() { 
    return (TPML_CC * *)malloc(sizeof(TPML_CC *));
  }
  
  static TPML_CC * *copy_TPML_CC_PTR_PTR(TPML_CC * value) { 
    return (TPML_CC * *)memcpy((TPML_CC * *)malloc(sizeof(TPML_CC *)),&value,sizeof(TPML_CC *));
  }

  static void delete_TPML_CC_PTR_PTR(TPML_CC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_CC_PTR_PTR_assign(TPML_CC * *obj, TPML_CC * value) {
    *obj = value;
  }

  static TPML_CC * TPML_CC_PTR_PTR_value(TPML_CC * *obj) {
    return *obj;
  }


typedef TPML_CCA TPML_CCA_ARRAY;

SWIGINTERN TPML_CCA_ARRAY *new_TPML_CCA_ARRAY(size_t nelements){
    return (TPML_CCA *)malloc((nelements)*sizeof(TPML_CCA));
  }
SWIGINTERN void delete_TPML_CCA_ARRAY(TPML_CCA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_CCA TPML_CCA_ARRAY___getitem__(TPML_CCA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_CCA_ARRAY___setitem__(TPML_CCA_ARRAY *self,size_t index,TPML_CCA value){
    self[index] = value;
  }
SWIGINTERN TPML_CCA *TPML_CCA_ARRAY_cast(TPML_CCA_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_CCA_ARRAY *TPML_CCA_ARRAY_frompointer(TPML_CCA *t){
    return (TPML_CCA_ARRAY *)(t);
  }

  static TPML_CCA *new_TPML_CCA_PTR() { 
    return (TPML_CCA *)malloc(sizeof(TPML_CCA));
  }
  
  static TPML_CCA *copy_TPML_CCA_PTR(TPML_CCA value) { 
    return (TPML_CCA *)memcpy((TPML_CCA *)malloc(sizeof(TPML_CCA)),&value,sizeof(TPML_CCA));
  }

  static void delete_TPML_CCA_PTR(TPML_CCA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_CCA_PTR_assign(TPML_CCA *obj, TPML_CCA value) {
    *obj = value;
  }

  static TPML_CCA TPML_CCA_PTR_value(TPML_CCA *obj) {
    return *obj;
  }


  static TPML_CCA * *new_TPML_CCA_PTR_PTR() { 
    return (TPML_CCA * *)malloc(sizeof(TPML_CCA *));
  }
  
  static TPML_CCA * *copy_TPML_CCA_PTR_PTR(TPML_CCA * value) { 
    return (TPML_CCA * *)memcpy((TPML_CCA * *)malloc(sizeof(TPML_CCA *)),&value,sizeof(TPML_CCA *));
  }

  static void delete_TPML_CCA_PTR_PTR(TPML_CCA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_CCA_PTR_PTR_assign(TPML_CCA * *obj, TPML_CCA * value) {
    *obj = value;
  }

  static TPML_CCA * TPML_CCA_PTR_PTR_value(TPML_CCA * *obj) {
    return *obj;
  }


typedef TPML_ALG TPML_ALG_ARRAY;

SWIGINTERN TPML_ALG_ARRAY *new_TPML_ALG_ARRAY(size_t nelements){
    return (TPML_ALG *)malloc((nelements)*sizeof(TPML_ALG));
  }
SWIGINTERN void delete_TPML_ALG_ARRAY(TPML_ALG_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_ALG TPML_ALG_ARRAY___getitem__(TPML_ALG_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_ALG_ARRAY___setitem__(TPML_ALG_ARRAY *self,size_t index,TPML_ALG value){
    self[index] = value;
  }
SWIGINTERN TPML_ALG *TPML_ALG_ARRAY_cast(TPML_ALG_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_ALG_ARRAY *TPML_ALG_ARRAY_frompointer(TPML_ALG *t){
    return (TPML_ALG_ARRAY *)(t);
  }

  static TPML_ALG *new_TPML_ALG_PTR() { 
    return (TPML_ALG *)malloc(sizeof(TPML_ALG));
  }
  
  static TPML_ALG *copy_TPML_ALG_PTR(TPML_ALG value) { 
    return (TPML_ALG *)memcpy((TPML_ALG *)malloc(sizeof(TPML_ALG)),&value,sizeof(TPML_ALG));
  }

  static void delete_TPML_ALG_PTR(TPML_ALG *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_ALG_PTR_assign(TPML_ALG *obj, TPML_ALG value) {
    *obj = value;
  }

  static TPML_ALG TPML_ALG_PTR_value(TPML_ALG *obj) {
    return *obj;
  }


  static TPML_ALG * *new_TPML_ALG_PTR_PTR() { 
    return (TPML_ALG * *)malloc(sizeof(TPML_ALG *));
  }
  
  static TPML_ALG * *copy_TPML_ALG_PTR_PTR(TPML_ALG * value) { 
    return (TPML_ALG * *)memcpy((TPML_ALG * *)malloc(sizeof(TPML_ALG *)),&value,sizeof(TPML_ALG *));
  }

  static void delete_TPML_ALG_PTR_PTR(TPML_ALG * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_ALG_PTR_PTR_assign(TPML_ALG * *obj, TPML_ALG * value) {
    *obj = value;
  }

  static TPML_ALG * TPML_ALG_PTR_PTR_value(TPML_ALG * *obj) {
    return *obj;
  }


typedef TPML_HANDLE TPML_HANDLE_ARRAY;

SWIGINTERN TPML_HANDLE_ARRAY *new_TPML_HANDLE_ARRAY(size_t nelements){
    return (TPML_HANDLE *)malloc((nelements)*sizeof(TPML_HANDLE));
  }
SWIGINTERN void delete_TPML_HANDLE_ARRAY(TPML_HANDLE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_HANDLE TPML_HANDLE_ARRAY___getitem__(TPML_HANDLE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_HANDLE_ARRAY___setitem__(TPML_HANDLE_ARRAY *self,size_t index,TPML_HANDLE value){
    self[index] = value;
  }
SWIGINTERN TPML_HANDLE *TPML_HANDLE_ARRAY_cast(TPML_HANDLE_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_HANDLE_ARRAY *TPML_HANDLE_ARRAY_frompointer(TPML_HANDLE *t){
    return (TPML_HANDLE_ARRAY *)(t);
  }

  static TPML_HANDLE *new_TPML_HANDLE_PTR() { 
    return (TPML_HANDLE *)malloc(sizeof(TPML_HANDLE));
  }
  
  static TPML_HANDLE *copy_TPML_HANDLE_PTR(TPML_HANDLE value) { 
    return (TPML_HANDLE *)memcpy((TPML_HANDLE *)malloc(sizeof(TPML_HANDLE)),&value,sizeof(TPML_HANDLE));
  }

  static void delete_TPML_HANDLE_PTR(TPML_HANDLE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_HANDLE_PTR_assign(TPML_HANDLE *obj, TPML_HANDLE value) {
    *obj = value;
  }

  static TPML_HANDLE TPML_HANDLE_PTR_value(TPML_HANDLE *obj) {
    return *obj;
  }


  static TPML_HANDLE * *new_TPML_HANDLE_PTR_PTR() { 
    return (TPML_HANDLE * *)malloc(sizeof(TPML_HANDLE *));
  }
  
  static TPML_HANDLE * *copy_TPML_HANDLE_PTR_PTR(TPML_HANDLE * value) { 
    return (TPML_HANDLE * *)memcpy((TPML_HANDLE * *)malloc(sizeof(TPML_HANDLE *)),&value,sizeof(TPML_HANDLE *));
  }

  static void delete_TPML_HANDLE_PTR_PTR(TPML_HANDLE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_HANDLE_PTR_PTR_assign(TPML_HANDLE * *obj, TPML_HANDLE * value) {
    *obj = value;
  }

  static TPML_HANDLE * TPML_HANDLE_PTR_PTR_value(TPML_HANDLE * *obj) {
    return *obj;
  }


typedef TPML_DIGEST TPML_DIGEST_ARRAY;

SWIGINTERN TPML_DIGEST_ARRAY *new_TPML_DIGEST_ARRAY(size_t nelements){
    return (TPML_DIGEST *)malloc((nelements)*sizeof(TPML_DIGEST));
  }
SWIGINTERN void delete_TPML_DIGEST_ARRAY(TPML_DIGEST_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_DIGEST TPML_DIGEST_ARRAY___getitem__(TPML_DIGEST_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_DIGEST_ARRAY___setitem__(TPML_DIGEST_ARRAY *self,size_t index,TPML_DIGEST value){
    self[index] = value;
  }
SWIGINTERN TPML_DIGEST *TPML_DIGEST_ARRAY_cast(TPML_DIGEST_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_DIGEST_ARRAY *TPML_DIGEST_ARRAY_frompointer(TPML_DIGEST *t){
    return (TPML_DIGEST_ARRAY *)(t);
  }

  static TPML_DIGEST *new_TPML_DIGEST_PTR() { 
    return (TPML_DIGEST *)malloc(sizeof(TPML_DIGEST));
  }
  
  static TPML_DIGEST *copy_TPML_DIGEST_PTR(TPML_DIGEST value) { 
    return (TPML_DIGEST *)memcpy((TPML_DIGEST *)malloc(sizeof(TPML_DIGEST)),&value,sizeof(TPML_DIGEST));
  }

  static void delete_TPML_DIGEST_PTR(TPML_DIGEST *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_DIGEST_PTR_assign(TPML_DIGEST *obj, TPML_DIGEST value) {
    *obj = value;
  }

  static TPML_DIGEST TPML_DIGEST_PTR_value(TPML_DIGEST *obj) {
    return *obj;
  }


  static TPML_DIGEST * *new_TPML_DIGEST_PTR_PTR() { 
    return (TPML_DIGEST * *)malloc(sizeof(TPML_DIGEST *));
  }
  
  static TPML_DIGEST * *copy_TPML_DIGEST_PTR_PTR(TPML_DIGEST * value) { 
    return (TPML_DIGEST * *)memcpy((TPML_DIGEST * *)malloc(sizeof(TPML_DIGEST *)),&value,sizeof(TPML_DIGEST *));
  }

  static void delete_TPML_DIGEST_PTR_PTR(TPML_DIGEST * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_DIGEST_PTR_PTR_assign(TPML_DIGEST * *obj, TPML_DIGEST * value) {
    *obj = value;
  }

  static TPML_DIGEST * TPML_DIGEST_PTR_PTR_value(TPML_DIGEST * *obj) {
    return *obj;
  }


typedef TPML_DIGEST_VALUES TPML_DIGEST_VALUES_ARRAY;

SWIGINTERN TPML_DIGEST_VALUES_ARRAY *new_TPML_DIGEST_VALUES_ARRAY(size_t nelements){
    return (TPML_DIGEST_VALUES *)malloc((nelements)*sizeof(TPML_DIGEST_VALUES));
  }
SWIGINTERN void delete_TPML_DIGEST_VALUES_ARRAY(TPML_DIGEST_VALUES_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_DIGEST_VALUES TPML_DIGEST_VALUES_ARRAY___getitem__(TPML_DIGEST_VALUES_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_DIGEST_VALUES_ARRAY___setitem__(TPML_DIGEST_VALUES_ARRAY *self,size_t index,TPML_DIGEST_VALUES value){
    self[index] = value;
  }
SWIGINTERN TPML_DIGEST_VALUES *TPML_DIGEST_VALUES_ARRAY_cast(TPML_DIGEST_VALUES_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_DIGEST_VALUES_ARRAY *TPML_DIGEST_VALUES_ARRAY_frompointer(TPML_DIGEST_VALUES *t){
    return (TPML_DIGEST_VALUES_ARRAY *)(t);
  }

  static TPML_DIGEST_VALUES *new_TPML_DIGEST_VALUES_PTR() { 
    return (TPML_DIGEST_VALUES *)malloc(sizeof(TPML_DIGEST_VALUES));
  }
  
  static TPML_DIGEST_VALUES *copy_TPML_DIGEST_VALUES_PTR(TPML_DIGEST_VALUES value) { 
    return (TPML_DIGEST_VALUES *)memcpy((TPML_DIGEST_VALUES *)malloc(sizeof(TPML_DIGEST_VALUES)),&value,sizeof(TPML_DIGEST_VALUES));
  }

  static void delete_TPML_DIGEST_VALUES_PTR(TPML_DIGEST_VALUES *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_DIGEST_VALUES_PTR_assign(TPML_DIGEST_VALUES *obj, TPML_DIGEST_VALUES value) {
    *obj = value;
  }

  static TPML_DIGEST_VALUES TPML_DIGEST_VALUES_PTR_value(TPML_DIGEST_VALUES *obj) {
    return *obj;
  }


  static TPML_DIGEST_VALUES * *new_TPML_DIGEST_VALUES_PTR_PTR() { 
    return (TPML_DIGEST_VALUES * *)malloc(sizeof(TPML_DIGEST_VALUES *));
  }
  
  static TPML_DIGEST_VALUES * *copy_TPML_DIGEST_VALUES_PTR_PTR(TPML_DIGEST_VALUES * value) { 
    return (TPML_DIGEST_VALUES * *)memcpy((TPML_DIGEST_VALUES * *)malloc(sizeof(TPML_DIGEST_VALUES *)),&value,sizeof(TPML_DIGEST_VALUES *));
  }

  static void delete_TPML_DIGEST_VALUES_PTR_PTR(TPML_DIGEST_VALUES * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_DIGEST_VALUES_PTR_PTR_assign(TPML_DIGEST_VALUES * *obj, TPML_DIGEST_VALUES * value) {
    *obj = value;
  }

  static TPML_DIGEST_VALUES * TPML_DIGEST_VALUES_PTR_PTR_value(TPML_DIGEST_VALUES * *obj) {
    return *obj;
  }


typedef TPML_PCR_SELECTION TPML_PCR_SELECTION_ARRAY;

SWIGINTERN TPML_PCR_SELECTION_ARRAY *new_TPML_PCR_SELECTION_ARRAY(size_t nelements){
    return (TPML_PCR_SELECTION *)malloc((nelements)*sizeof(TPML_PCR_SELECTION));
  }
SWIGINTERN void delete_TPML_PCR_SELECTION_ARRAY(TPML_PCR_SELECTION_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_PCR_SELECTION TPML_PCR_SELECTION_ARRAY___getitem__(TPML_PCR_SELECTION_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_PCR_SELECTION_ARRAY___setitem__(TPML_PCR_SELECTION_ARRAY *self,size_t index,TPML_PCR_SELECTION value){
    self[index] = value;
  }
SWIGINTERN TPML_PCR_SELECTION *TPML_PCR_SELECTION_ARRAY_cast(TPML_PCR_SELECTION_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_PCR_SELECTION_ARRAY *TPML_PCR_SELECTION_ARRAY_frompointer(TPML_PCR_SELECTION *t){
    return (TPML_PCR_SELECTION_ARRAY *)(t);
  }

  static TPML_PCR_SELECTION *new_TPML_PCR_SELECTION_PTR() { 
    return (TPML_PCR_SELECTION *)malloc(sizeof(TPML_PCR_SELECTION));
  }
  
  static TPML_PCR_SELECTION *copy_TPML_PCR_SELECTION_PTR(TPML_PCR_SELECTION value) { 
    return (TPML_PCR_SELECTION *)memcpy((TPML_PCR_SELECTION *)malloc(sizeof(TPML_PCR_SELECTION)),&value,sizeof(TPML_PCR_SELECTION));
  }

  static void delete_TPML_PCR_SELECTION_PTR(TPML_PCR_SELECTION *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_PCR_SELECTION_PTR_assign(TPML_PCR_SELECTION *obj, TPML_PCR_SELECTION value) {
    *obj = value;
  }

  static TPML_PCR_SELECTION TPML_PCR_SELECTION_PTR_value(TPML_PCR_SELECTION *obj) {
    return *obj;
  }


  static TPML_PCR_SELECTION * *new_TPML_PCR_SELECTION_PTR_PTR() { 
    return (TPML_PCR_SELECTION * *)malloc(sizeof(TPML_PCR_SELECTION *));
  }
  
  static TPML_PCR_SELECTION * *copy_TPML_PCR_SELECTION_PTR_PTR(TPML_PCR_SELECTION * value) { 
    return (TPML_PCR_SELECTION * *)memcpy((TPML_PCR_SELECTION * *)malloc(sizeof(TPML_PCR_SELECTION *)),&value,sizeof(TPML_PCR_SELECTION *));
  }

  static void delete_TPML_PCR_SELECTION_PTR_PTR(TPML_PCR_SELECTION * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_PCR_SELECTION_PTR_PTR_assign(TPML_PCR_SELECTION * *obj, TPML_PCR_SELECTION * value) {
    *obj = value;
  }

  static TPML_PCR_SELECTION * TPML_PCR_SELECTION_PTR_PTR_value(TPML_PCR_SELECTION * *obj) {
    return *obj;
  }


typedef TPML_ALG_PROPERTY TPML_ALG_PROPERTY_ARRAY;

SWIGINTERN TPML_ALG_PROPERTY_ARRAY *new_TPML_ALG_PROPERTY_ARRAY(size_t nelements){
    return (TPML_ALG_PROPERTY *)malloc((nelements)*sizeof(TPML_ALG_PROPERTY));
  }
SWIGINTERN void delete_TPML_ALG_PROPERTY_ARRAY(TPML_ALG_PROPERTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_ALG_PROPERTY TPML_ALG_PROPERTY_ARRAY___getitem__(TPML_ALG_PROPERTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_ALG_PROPERTY_ARRAY___setitem__(TPML_ALG_PROPERTY_ARRAY *self,size_t index,TPML_ALG_PROPERTY value){
    self[index] = value;
  }
SWIGINTERN TPML_ALG_PROPERTY *TPML_ALG_PROPERTY_ARRAY_cast(TPML_ALG_PROPERTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_ALG_PROPERTY_ARRAY *TPML_ALG_PROPERTY_ARRAY_frompointer(TPML_ALG_PROPERTY *t){
    return (TPML_ALG_PROPERTY_ARRAY *)(t);
  }

  static TPML_ALG_PROPERTY *new_TPML_ALG_PROPERTY_PTR() { 
    return (TPML_ALG_PROPERTY *)malloc(sizeof(TPML_ALG_PROPERTY));
  }
  
  static TPML_ALG_PROPERTY *copy_TPML_ALG_PROPERTY_PTR(TPML_ALG_PROPERTY value) { 
    return (TPML_ALG_PROPERTY *)memcpy((TPML_ALG_PROPERTY *)malloc(sizeof(TPML_ALG_PROPERTY)),&value,sizeof(TPML_ALG_PROPERTY));
  }

  static void delete_TPML_ALG_PROPERTY_PTR(TPML_ALG_PROPERTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_ALG_PROPERTY_PTR_assign(TPML_ALG_PROPERTY *obj, TPML_ALG_PROPERTY value) {
    *obj = value;
  }

  static TPML_ALG_PROPERTY TPML_ALG_PROPERTY_PTR_value(TPML_ALG_PROPERTY *obj) {
    return *obj;
  }


  static TPML_ALG_PROPERTY * *new_TPML_ALG_PROPERTY_PTR_PTR() { 
    return (TPML_ALG_PROPERTY * *)malloc(sizeof(TPML_ALG_PROPERTY *));
  }
  
  static TPML_ALG_PROPERTY * *copy_TPML_ALG_PROPERTY_PTR_PTR(TPML_ALG_PROPERTY * value) { 
    return (TPML_ALG_PROPERTY * *)memcpy((TPML_ALG_PROPERTY * *)malloc(sizeof(TPML_ALG_PROPERTY *)),&value,sizeof(TPML_ALG_PROPERTY *));
  }

  static void delete_TPML_ALG_PROPERTY_PTR_PTR(TPML_ALG_PROPERTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_ALG_PROPERTY_PTR_PTR_assign(TPML_ALG_PROPERTY * *obj, TPML_ALG_PROPERTY * value) {
    *obj = value;
  }

  static TPML_ALG_PROPERTY * TPML_ALG_PROPERTY_PTR_PTR_value(TPML_ALG_PROPERTY * *obj) {
    return *obj;
  }


typedef TPML_TAGGED_TPM_PROPERTY TPML_TAGGED_TPM_PROPERTY_ARRAY;

SWIGINTERN TPML_TAGGED_TPM_PROPERTY_ARRAY *new_TPML_TAGGED_TPM_PROPERTY_ARRAY(size_t nelements){
    return (TPML_TAGGED_TPM_PROPERTY *)malloc((nelements)*sizeof(TPML_TAGGED_TPM_PROPERTY));
  }
SWIGINTERN void delete_TPML_TAGGED_TPM_PROPERTY_ARRAY(TPML_TAGGED_TPM_PROPERTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_TAGGED_TPM_PROPERTY TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__(TPML_TAGGED_TPM_PROPERTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__(TPML_TAGGED_TPM_PROPERTY_ARRAY *self,size_t index,TPML_TAGGED_TPM_PROPERTY value){
    self[index] = value;
  }
SWIGINTERN TPML_TAGGED_TPM_PROPERTY *TPML_TAGGED_TPM_PROPERTY_ARRAY_cast(TPML_TAGGED_TPM_PROPERTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_TAGGED_TPM_PROPERTY_ARRAY *TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer(TPML_TAGGED_TPM_PROPERTY *t){
    return (TPML_TAGGED_TPM_PROPERTY_ARRAY *)(t);
  }

  static TPML_TAGGED_TPM_PROPERTY *new_TPML_TAGGED_TPM_PROPERTY_PTR() { 
    return (TPML_TAGGED_TPM_PROPERTY *)malloc(sizeof(TPML_TAGGED_TPM_PROPERTY));
  }
  
  static TPML_TAGGED_TPM_PROPERTY *copy_TPML_TAGGED_TPM_PROPERTY_PTR(TPML_TAGGED_TPM_PROPERTY value) { 
    return (TPML_TAGGED_TPM_PROPERTY *)memcpy((TPML_TAGGED_TPM_PROPERTY *)malloc(sizeof(TPML_TAGGED_TPM_PROPERTY)),&value,sizeof(TPML_TAGGED_TPM_PROPERTY));
  }

  static void delete_TPML_TAGGED_TPM_PROPERTY_PTR(TPML_TAGGED_TPM_PROPERTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_TAGGED_TPM_PROPERTY_PTR_assign(TPML_TAGGED_TPM_PROPERTY *obj, TPML_TAGGED_TPM_PROPERTY value) {
    *obj = value;
  }

  static TPML_TAGGED_TPM_PROPERTY TPML_TAGGED_TPM_PROPERTY_PTR_value(TPML_TAGGED_TPM_PROPERTY *obj) {
    return *obj;
  }


  static TPML_TAGGED_TPM_PROPERTY * *new_TPML_TAGGED_TPM_PROPERTY_PTR_PTR() { 
    return (TPML_TAGGED_TPM_PROPERTY * *)malloc(sizeof(TPML_TAGGED_TPM_PROPERTY *));
  }
  
  static TPML_TAGGED_TPM_PROPERTY * *copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(TPML_TAGGED_TPM_PROPERTY * value) { 
    return (TPML_TAGGED_TPM_PROPERTY * *)memcpy((TPML_TAGGED_TPM_PROPERTY * *)malloc(sizeof(TPML_TAGGED_TPM_PROPERTY *)),&value,sizeof(TPML_TAGGED_TPM_PROPERTY *));
  }

  static void delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(TPML_TAGGED_TPM_PROPERTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign(TPML_TAGGED_TPM_PROPERTY * *obj, TPML_TAGGED_TPM_PROPERTY * value) {
    *obj = value;
  }

  static TPML_TAGGED_TPM_PROPERTY * TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value(TPML_TAGGED_TPM_PROPERTY * *obj) {
    return *obj;
  }


typedef TPML_TAGGED_PCR_PROPERTY TPML_TAGGED_PCR_PROPERTY_ARRAY;

SWIGINTERN TPML_TAGGED_PCR_PROPERTY_ARRAY *new_TPML_TAGGED_PCR_PROPERTY_ARRAY(size_t nelements){
    return (TPML_TAGGED_PCR_PROPERTY *)malloc((nelements)*sizeof(TPML_TAGGED_PCR_PROPERTY));
  }
SWIGINTERN void delete_TPML_TAGGED_PCR_PROPERTY_ARRAY(TPML_TAGGED_PCR_PROPERTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_TAGGED_PCR_PROPERTY TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__(TPML_TAGGED_PCR_PROPERTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__(TPML_TAGGED_PCR_PROPERTY_ARRAY *self,size_t index,TPML_TAGGED_PCR_PROPERTY value){
    self[index] = value;
  }
SWIGINTERN TPML_TAGGED_PCR_PROPERTY *TPML_TAGGED_PCR_PROPERTY_ARRAY_cast(TPML_TAGGED_PCR_PROPERTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_TAGGED_PCR_PROPERTY_ARRAY *TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer(TPML_TAGGED_PCR_PROPERTY *t){
    return (TPML_TAGGED_PCR_PROPERTY_ARRAY *)(t);
  }

  static TPML_TAGGED_PCR_PROPERTY *new_TPML_TAGGED_PCR_PROPERTY_PTR() { 
    return (TPML_TAGGED_PCR_PROPERTY *)malloc(sizeof(TPML_TAGGED_PCR_PROPERTY));
  }
  
  static TPML_TAGGED_PCR_PROPERTY *copy_TPML_TAGGED_PCR_PROPERTY_PTR(TPML_TAGGED_PCR_PROPERTY value) { 
    return (TPML_TAGGED_PCR_PROPERTY *)memcpy((TPML_TAGGED_PCR_PROPERTY *)malloc(sizeof(TPML_TAGGED_PCR_PROPERTY)),&value,sizeof(TPML_TAGGED_PCR_PROPERTY));
  }

  static void delete_TPML_TAGGED_PCR_PROPERTY_PTR(TPML_TAGGED_PCR_PROPERTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_TAGGED_PCR_PROPERTY_PTR_assign(TPML_TAGGED_PCR_PROPERTY *obj, TPML_TAGGED_PCR_PROPERTY value) {
    *obj = value;
  }

  static TPML_TAGGED_PCR_PROPERTY TPML_TAGGED_PCR_PROPERTY_PTR_value(TPML_TAGGED_PCR_PROPERTY *obj) {
    return *obj;
  }


  static TPML_TAGGED_PCR_PROPERTY * *new_TPML_TAGGED_PCR_PROPERTY_PTR_PTR() { 
    return (TPML_TAGGED_PCR_PROPERTY * *)malloc(sizeof(TPML_TAGGED_PCR_PROPERTY *));
  }
  
  static TPML_TAGGED_PCR_PROPERTY * *copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(TPML_TAGGED_PCR_PROPERTY * value) { 
    return (TPML_TAGGED_PCR_PROPERTY * *)memcpy((TPML_TAGGED_PCR_PROPERTY * *)malloc(sizeof(TPML_TAGGED_PCR_PROPERTY *)),&value,sizeof(TPML_TAGGED_PCR_PROPERTY *));
  }

  static void delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(TPML_TAGGED_PCR_PROPERTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign(TPML_TAGGED_PCR_PROPERTY * *obj, TPML_TAGGED_PCR_PROPERTY * value) {
    *obj = value;
  }

  static TPML_TAGGED_PCR_PROPERTY * TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value(TPML_TAGGED_PCR_PROPERTY * *obj) {
    return *obj;
  }


typedef TPML_ECC_CURVE TPML_ECC_CURVE_ARRAY;

SWIGINTERN TPML_ECC_CURVE_ARRAY *new_TPML_ECC_CURVE_ARRAY(size_t nelements){
    return (TPML_ECC_CURVE *)malloc((nelements)*sizeof(TPML_ECC_CURVE));
  }
SWIGINTERN void delete_TPML_ECC_CURVE_ARRAY(TPML_ECC_CURVE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_ECC_CURVE TPML_ECC_CURVE_ARRAY___getitem__(TPML_ECC_CURVE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_ECC_CURVE_ARRAY___setitem__(TPML_ECC_CURVE_ARRAY *self,size_t index,TPML_ECC_CURVE value){
    self[index] = value;
  }
SWIGINTERN TPML_ECC_CURVE *TPML_ECC_CURVE_ARRAY_cast(TPML_ECC_CURVE_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_ECC_CURVE_ARRAY *TPML_ECC_CURVE_ARRAY_frompointer(TPML_ECC_CURVE *t){
    return (TPML_ECC_CURVE_ARRAY *)(t);
  }

  static TPML_ECC_CURVE *new_TPML_ECC_CURVE_PTR() { 
    return (TPML_ECC_CURVE *)malloc(sizeof(TPML_ECC_CURVE));
  }
  
  static TPML_ECC_CURVE *copy_TPML_ECC_CURVE_PTR(TPML_ECC_CURVE value) { 
    return (TPML_ECC_CURVE *)memcpy((TPML_ECC_CURVE *)malloc(sizeof(TPML_ECC_CURVE)),&value,sizeof(TPML_ECC_CURVE));
  }

  static void delete_TPML_ECC_CURVE_PTR(TPML_ECC_CURVE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_ECC_CURVE_PTR_assign(TPML_ECC_CURVE *obj, TPML_ECC_CURVE value) {
    *obj = value;
  }

  static TPML_ECC_CURVE TPML_ECC_CURVE_PTR_value(TPML_ECC_CURVE *obj) {
    return *obj;
  }


  static TPML_ECC_CURVE * *new_TPML_ECC_CURVE_PTR_PTR() { 
    return (TPML_ECC_CURVE * *)malloc(sizeof(TPML_ECC_CURVE *));
  }
  
  static TPML_ECC_CURVE * *copy_TPML_ECC_CURVE_PTR_PTR(TPML_ECC_CURVE * value) { 
    return (TPML_ECC_CURVE * *)memcpy((TPML_ECC_CURVE * *)malloc(sizeof(TPML_ECC_CURVE *)),&value,sizeof(TPML_ECC_CURVE *));
  }

  static void delete_TPML_ECC_CURVE_PTR_PTR(TPML_ECC_CURVE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_ECC_CURVE_PTR_PTR_assign(TPML_ECC_CURVE * *obj, TPML_ECC_CURVE * value) {
    *obj = value;
  }

  static TPML_ECC_CURVE * TPML_ECC_CURVE_PTR_PTR_value(TPML_ECC_CURVE * *obj) {
    return *obj;
  }


typedef TPML_INTEL_PTT_PROPERTY TPML_INTEL_PTT_PROPERTY_ARRAY;

SWIGINTERN TPML_INTEL_PTT_PROPERTY_ARRAY *new_TPML_INTEL_PTT_PROPERTY_ARRAY(size_t nelements){
    return (TPML_INTEL_PTT_PROPERTY *)malloc((nelements)*sizeof(TPML_INTEL_PTT_PROPERTY));
  }
SWIGINTERN void delete_TPML_INTEL_PTT_PROPERTY_ARRAY(TPML_INTEL_PTT_PROPERTY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_INTEL_PTT_PROPERTY TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__(TPML_INTEL_PTT_PROPERTY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__(TPML_INTEL_PTT_PROPERTY_ARRAY *self,size_t index,TPML_INTEL_PTT_PROPERTY value){
    self[index] = value;
  }
SWIGINTERN TPML_INTEL_PTT_PROPERTY *TPML_INTEL_PTT_PROPERTY_ARRAY_cast(TPML_INTEL_PTT_PROPERTY_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_INTEL_PTT_PROPERTY_ARRAY *TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer(TPML_INTEL_PTT_PROPERTY *t){
    return (TPML_INTEL_PTT_PROPERTY_ARRAY *)(t);
  }

  static TPML_INTEL_PTT_PROPERTY *new_TPML_INTEL_PTT_PROPERTY_PTR() { 
    return (TPML_INTEL_PTT_PROPERTY *)malloc(sizeof(TPML_INTEL_PTT_PROPERTY));
  }
  
  static TPML_INTEL_PTT_PROPERTY *copy_TPML_INTEL_PTT_PROPERTY_PTR(TPML_INTEL_PTT_PROPERTY value) { 
    return (TPML_INTEL_PTT_PROPERTY *)memcpy((TPML_INTEL_PTT_PROPERTY *)malloc(sizeof(TPML_INTEL_PTT_PROPERTY)),&value,sizeof(TPML_INTEL_PTT_PROPERTY));
  }

  static void delete_TPML_INTEL_PTT_PROPERTY_PTR(TPML_INTEL_PTT_PROPERTY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_INTEL_PTT_PROPERTY_PTR_assign(TPML_INTEL_PTT_PROPERTY *obj, TPML_INTEL_PTT_PROPERTY value) {
    *obj = value;
  }

  static TPML_INTEL_PTT_PROPERTY TPML_INTEL_PTT_PROPERTY_PTR_value(TPML_INTEL_PTT_PROPERTY *obj) {
    return *obj;
  }


  static TPML_INTEL_PTT_PROPERTY * *new_TPML_INTEL_PTT_PROPERTY_PTR_PTR() { 
    return (TPML_INTEL_PTT_PROPERTY * *)malloc(sizeof(TPML_INTEL_PTT_PROPERTY *));
  }
  
  static TPML_INTEL_PTT_PROPERTY * *copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR(TPML_INTEL_PTT_PROPERTY * value) { 
    return (TPML_INTEL_PTT_PROPERTY * *)memcpy((TPML_INTEL_PTT_PROPERTY * *)malloc(sizeof(TPML_INTEL_PTT_PROPERTY *)),&value,sizeof(TPML_INTEL_PTT_PROPERTY *));
  }

  static void delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR(TPML_INTEL_PTT_PROPERTY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign(TPML_INTEL_PTT_PROPERTY * *obj, TPML_INTEL_PTT_PROPERTY * value) {
    *obj = value;
  }

  static TPML_INTEL_PTT_PROPERTY * TPML_INTEL_PTT_PROPERTY_PTR_PTR_value(TPML_INTEL_PTT_PROPERTY * *obj) {
    return *obj;
  }


typedef TPMU_CAPABILITIES TPMU_CAPABILITIES_ARRAY;

SWIGINTERN TPMU_CAPABILITIES_ARRAY *new_TPMU_CAPABILITIES_ARRAY(size_t nelements){
    return (TPMU_CAPABILITIES *)malloc((nelements)*sizeof(TPMU_CAPABILITIES));
  }
SWIGINTERN void delete_TPMU_CAPABILITIES_ARRAY(TPMU_CAPABILITIES_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_CAPABILITIES TPMU_CAPABILITIES_ARRAY___getitem__(TPMU_CAPABILITIES_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_CAPABILITIES_ARRAY___setitem__(TPMU_CAPABILITIES_ARRAY *self,size_t index,TPMU_CAPABILITIES value){
    self[index] = value;
  }
SWIGINTERN TPMU_CAPABILITIES *TPMU_CAPABILITIES_ARRAY_cast(TPMU_CAPABILITIES_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_CAPABILITIES_ARRAY *TPMU_CAPABILITIES_ARRAY_frompointer(TPMU_CAPABILITIES *t){
    return (TPMU_CAPABILITIES_ARRAY *)(t);
  }

  static TPMU_CAPABILITIES *new_TPMU_CAPABILITIES_PTR() { 
    return (TPMU_CAPABILITIES *)malloc(sizeof(TPMU_CAPABILITIES));
  }
  
  static TPMU_CAPABILITIES *copy_TPMU_CAPABILITIES_PTR(TPMU_CAPABILITIES value) { 
    return (TPMU_CAPABILITIES *)memcpy((TPMU_CAPABILITIES *)malloc(sizeof(TPMU_CAPABILITIES)),&value,sizeof(TPMU_CAPABILITIES));
  }

  static void delete_TPMU_CAPABILITIES_PTR(TPMU_CAPABILITIES *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_CAPABILITIES_PTR_assign(TPMU_CAPABILITIES *obj, TPMU_CAPABILITIES value) {
    *obj = value;
  }

  static TPMU_CAPABILITIES TPMU_CAPABILITIES_PTR_value(TPMU_CAPABILITIES *obj) {
    return *obj;
  }


  static TPMU_CAPABILITIES * *new_TPMU_CAPABILITIES_PTR_PTR() { 
    return (TPMU_CAPABILITIES * *)malloc(sizeof(TPMU_CAPABILITIES *));
  }
  
  static TPMU_CAPABILITIES * *copy_TPMU_CAPABILITIES_PTR_PTR(TPMU_CAPABILITIES * value) { 
    return (TPMU_CAPABILITIES * *)memcpy((TPMU_CAPABILITIES * *)malloc(sizeof(TPMU_CAPABILITIES *)),&value,sizeof(TPMU_CAPABILITIES *));
  }

  static void delete_TPMU_CAPABILITIES_PTR_PTR(TPMU_CAPABILITIES * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_CAPABILITIES_PTR_PTR_assign(TPMU_CAPABILITIES * *obj, TPMU_CAPABILITIES * value) {
    *obj = value;
  }

  static TPMU_CAPABILITIES * TPMU_CAPABILITIES_PTR_PTR_value(TPMU_CAPABILITIES * *obj) {
    return *obj;
  }


typedef TPMS_CAPABILITY_DATA TPMS_CAPABILITY_DATA_ARRAY;

SWIGINTERN TPMS_CAPABILITY_DATA_ARRAY *new_TPMS_CAPABILITY_DATA_ARRAY(size_t nelements){
    return (TPMS_CAPABILITY_DATA *)malloc((nelements)*sizeof(TPMS_CAPABILITY_DATA));
  }
SWIGINTERN void delete_TPMS_CAPABILITY_DATA_ARRAY(TPMS_CAPABILITY_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CAPABILITY_DATA TPMS_CAPABILITY_DATA_ARRAY___getitem__(TPMS_CAPABILITY_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CAPABILITY_DATA_ARRAY___setitem__(TPMS_CAPABILITY_DATA_ARRAY *self,size_t index,TPMS_CAPABILITY_DATA value){
    self[index] = value;
  }
SWIGINTERN TPMS_CAPABILITY_DATA *TPMS_CAPABILITY_DATA_ARRAY_cast(TPMS_CAPABILITY_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CAPABILITY_DATA_ARRAY *TPMS_CAPABILITY_DATA_ARRAY_frompointer(TPMS_CAPABILITY_DATA *t){
    return (TPMS_CAPABILITY_DATA_ARRAY *)(t);
  }

  static TPMS_CAPABILITY_DATA *new_TPMS_CAPABILITY_DATA_PTR() { 
    return (TPMS_CAPABILITY_DATA *)malloc(sizeof(TPMS_CAPABILITY_DATA));
  }
  
  static TPMS_CAPABILITY_DATA *copy_TPMS_CAPABILITY_DATA_PTR(TPMS_CAPABILITY_DATA value) { 
    return (TPMS_CAPABILITY_DATA *)memcpy((TPMS_CAPABILITY_DATA *)malloc(sizeof(TPMS_CAPABILITY_DATA)),&value,sizeof(TPMS_CAPABILITY_DATA));
  }

  static void delete_TPMS_CAPABILITY_DATA_PTR(TPMS_CAPABILITY_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CAPABILITY_DATA_PTR_assign(TPMS_CAPABILITY_DATA *obj, TPMS_CAPABILITY_DATA value) {
    *obj = value;
  }

  static TPMS_CAPABILITY_DATA TPMS_CAPABILITY_DATA_PTR_value(TPMS_CAPABILITY_DATA *obj) {
    return *obj;
  }


  static TPMS_CAPABILITY_DATA * *new_TPMS_CAPABILITY_DATA_PTR_PTR() { 
    return (TPMS_CAPABILITY_DATA * *)malloc(sizeof(TPMS_CAPABILITY_DATA *));
  }
  
  static TPMS_CAPABILITY_DATA * *copy_TPMS_CAPABILITY_DATA_PTR_PTR(TPMS_CAPABILITY_DATA * value) { 
    return (TPMS_CAPABILITY_DATA * *)memcpy((TPMS_CAPABILITY_DATA * *)malloc(sizeof(TPMS_CAPABILITY_DATA *)),&value,sizeof(TPMS_CAPABILITY_DATA *));
  }

  static void delete_TPMS_CAPABILITY_DATA_PTR_PTR(TPMS_CAPABILITY_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CAPABILITY_DATA_PTR_PTR_assign(TPMS_CAPABILITY_DATA * *obj, TPMS_CAPABILITY_DATA * value) {
    *obj = value;
  }

  static TPMS_CAPABILITY_DATA * TPMS_CAPABILITY_DATA_PTR_PTR_value(TPMS_CAPABILITY_DATA * *obj) {
    return *obj;
  }


typedef TPMS_CLOCK_INFO TPMS_CLOCK_INFO_ARRAY;

SWIGINTERN TPMS_CLOCK_INFO_ARRAY *new_TPMS_CLOCK_INFO_ARRAY(size_t nelements){
    return (TPMS_CLOCK_INFO *)malloc((nelements)*sizeof(TPMS_CLOCK_INFO));
  }
SWIGINTERN void delete_TPMS_CLOCK_INFO_ARRAY(TPMS_CLOCK_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CLOCK_INFO TPMS_CLOCK_INFO_ARRAY___getitem__(TPMS_CLOCK_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CLOCK_INFO_ARRAY___setitem__(TPMS_CLOCK_INFO_ARRAY *self,size_t index,TPMS_CLOCK_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_CLOCK_INFO *TPMS_CLOCK_INFO_ARRAY_cast(TPMS_CLOCK_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CLOCK_INFO_ARRAY *TPMS_CLOCK_INFO_ARRAY_frompointer(TPMS_CLOCK_INFO *t){
    return (TPMS_CLOCK_INFO_ARRAY *)(t);
  }

  static TPMS_CLOCK_INFO *new_TPMS_CLOCK_INFO_PTR() { 
    return (TPMS_CLOCK_INFO *)malloc(sizeof(TPMS_CLOCK_INFO));
  }
  
  static TPMS_CLOCK_INFO *copy_TPMS_CLOCK_INFO_PTR(TPMS_CLOCK_INFO value) { 
    return (TPMS_CLOCK_INFO *)memcpy((TPMS_CLOCK_INFO *)malloc(sizeof(TPMS_CLOCK_INFO)),&value,sizeof(TPMS_CLOCK_INFO));
  }

  static void delete_TPMS_CLOCK_INFO_PTR(TPMS_CLOCK_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CLOCK_INFO_PTR_assign(TPMS_CLOCK_INFO *obj, TPMS_CLOCK_INFO value) {
    *obj = value;
  }

  static TPMS_CLOCK_INFO TPMS_CLOCK_INFO_PTR_value(TPMS_CLOCK_INFO *obj) {
    return *obj;
  }


  static TPMS_CLOCK_INFO * *new_TPMS_CLOCK_INFO_PTR_PTR() { 
    return (TPMS_CLOCK_INFO * *)malloc(sizeof(TPMS_CLOCK_INFO *));
  }
  
  static TPMS_CLOCK_INFO * *copy_TPMS_CLOCK_INFO_PTR_PTR(TPMS_CLOCK_INFO * value) { 
    return (TPMS_CLOCK_INFO * *)memcpy((TPMS_CLOCK_INFO * *)malloc(sizeof(TPMS_CLOCK_INFO *)),&value,sizeof(TPMS_CLOCK_INFO *));
  }

  static void delete_TPMS_CLOCK_INFO_PTR_PTR(TPMS_CLOCK_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CLOCK_INFO_PTR_PTR_assign(TPMS_CLOCK_INFO * *obj, TPMS_CLOCK_INFO * value) {
    *obj = value;
  }

  static TPMS_CLOCK_INFO * TPMS_CLOCK_INFO_PTR_PTR_value(TPMS_CLOCK_INFO * *obj) {
    return *obj;
  }


typedef TPMS_TIME_INFO TPMS_TIME_INFO_ARRAY;

SWIGINTERN TPMS_TIME_INFO_ARRAY *new_TPMS_TIME_INFO_ARRAY(size_t nelements){
    return (TPMS_TIME_INFO *)malloc((nelements)*sizeof(TPMS_TIME_INFO));
  }
SWIGINTERN void delete_TPMS_TIME_INFO_ARRAY(TPMS_TIME_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_TIME_INFO TPMS_TIME_INFO_ARRAY___getitem__(TPMS_TIME_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_TIME_INFO_ARRAY___setitem__(TPMS_TIME_INFO_ARRAY *self,size_t index,TPMS_TIME_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_TIME_INFO *TPMS_TIME_INFO_ARRAY_cast(TPMS_TIME_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_TIME_INFO_ARRAY *TPMS_TIME_INFO_ARRAY_frompointer(TPMS_TIME_INFO *t){
    return (TPMS_TIME_INFO_ARRAY *)(t);
  }

  static TPMS_TIME_INFO *new_TPMS_TIME_INFO_PTR() { 
    return (TPMS_TIME_INFO *)malloc(sizeof(TPMS_TIME_INFO));
  }
  
  static TPMS_TIME_INFO *copy_TPMS_TIME_INFO_PTR(TPMS_TIME_INFO value) { 
    return (TPMS_TIME_INFO *)memcpy((TPMS_TIME_INFO *)malloc(sizeof(TPMS_TIME_INFO)),&value,sizeof(TPMS_TIME_INFO));
  }

  static void delete_TPMS_TIME_INFO_PTR(TPMS_TIME_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TIME_INFO_PTR_assign(TPMS_TIME_INFO *obj, TPMS_TIME_INFO value) {
    *obj = value;
  }

  static TPMS_TIME_INFO TPMS_TIME_INFO_PTR_value(TPMS_TIME_INFO *obj) {
    return *obj;
  }


  static TPMS_TIME_INFO * *new_TPMS_TIME_INFO_PTR_PTR() { 
    return (TPMS_TIME_INFO * *)malloc(sizeof(TPMS_TIME_INFO *));
  }
  
  static TPMS_TIME_INFO * *copy_TPMS_TIME_INFO_PTR_PTR(TPMS_TIME_INFO * value) { 
    return (TPMS_TIME_INFO * *)memcpy((TPMS_TIME_INFO * *)malloc(sizeof(TPMS_TIME_INFO *)),&value,sizeof(TPMS_TIME_INFO *));
  }

  static void delete_TPMS_TIME_INFO_PTR_PTR(TPMS_TIME_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TIME_INFO_PTR_PTR_assign(TPMS_TIME_INFO * *obj, TPMS_TIME_INFO * value) {
    *obj = value;
  }

  static TPMS_TIME_INFO * TPMS_TIME_INFO_PTR_PTR_value(TPMS_TIME_INFO * *obj) {
    return *obj;
  }


typedef TPMS_TIME_ATTEST_INFO TPMS_TIME_ATTEST_INFO_ARRAY;

SWIGINTERN TPMS_TIME_ATTEST_INFO_ARRAY *new_TPMS_TIME_ATTEST_INFO_ARRAY(size_t nelements){
    return (TPMS_TIME_ATTEST_INFO *)malloc((nelements)*sizeof(TPMS_TIME_ATTEST_INFO));
  }
SWIGINTERN void delete_TPMS_TIME_ATTEST_INFO_ARRAY(TPMS_TIME_ATTEST_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_TIME_ATTEST_INFO TPMS_TIME_ATTEST_INFO_ARRAY___getitem__(TPMS_TIME_ATTEST_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_TIME_ATTEST_INFO_ARRAY___setitem__(TPMS_TIME_ATTEST_INFO_ARRAY *self,size_t index,TPMS_TIME_ATTEST_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_TIME_ATTEST_INFO *TPMS_TIME_ATTEST_INFO_ARRAY_cast(TPMS_TIME_ATTEST_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_TIME_ATTEST_INFO_ARRAY *TPMS_TIME_ATTEST_INFO_ARRAY_frompointer(TPMS_TIME_ATTEST_INFO *t){
    return (TPMS_TIME_ATTEST_INFO_ARRAY *)(t);
  }

  static TPMS_TIME_ATTEST_INFO *new_TPMS_TIME_ATTEST_INFO_PTR() { 
    return (TPMS_TIME_ATTEST_INFO *)malloc(sizeof(TPMS_TIME_ATTEST_INFO));
  }
  
  static TPMS_TIME_ATTEST_INFO *copy_TPMS_TIME_ATTEST_INFO_PTR(TPMS_TIME_ATTEST_INFO value) { 
    return (TPMS_TIME_ATTEST_INFO *)memcpy((TPMS_TIME_ATTEST_INFO *)malloc(sizeof(TPMS_TIME_ATTEST_INFO)),&value,sizeof(TPMS_TIME_ATTEST_INFO));
  }

  static void delete_TPMS_TIME_ATTEST_INFO_PTR(TPMS_TIME_ATTEST_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TIME_ATTEST_INFO_PTR_assign(TPMS_TIME_ATTEST_INFO *obj, TPMS_TIME_ATTEST_INFO value) {
    *obj = value;
  }

  static TPMS_TIME_ATTEST_INFO TPMS_TIME_ATTEST_INFO_PTR_value(TPMS_TIME_ATTEST_INFO *obj) {
    return *obj;
  }


  static TPMS_TIME_ATTEST_INFO * *new_TPMS_TIME_ATTEST_INFO_PTR_PTR() { 
    return (TPMS_TIME_ATTEST_INFO * *)malloc(sizeof(TPMS_TIME_ATTEST_INFO *));
  }
  
  static TPMS_TIME_ATTEST_INFO * *copy_TPMS_TIME_ATTEST_INFO_PTR_PTR(TPMS_TIME_ATTEST_INFO * value) { 
    return (TPMS_TIME_ATTEST_INFO * *)memcpy((TPMS_TIME_ATTEST_INFO * *)malloc(sizeof(TPMS_TIME_ATTEST_INFO *)),&value,sizeof(TPMS_TIME_ATTEST_INFO *));
  }

  static void delete_TPMS_TIME_ATTEST_INFO_PTR_PTR(TPMS_TIME_ATTEST_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_TIME_ATTEST_INFO_PTR_PTR_assign(TPMS_TIME_ATTEST_INFO * *obj, TPMS_TIME_ATTEST_INFO * value) {
    *obj = value;
  }

  static TPMS_TIME_ATTEST_INFO * TPMS_TIME_ATTEST_INFO_PTR_PTR_value(TPMS_TIME_ATTEST_INFO * *obj) {
    return *obj;
  }


typedef TPMS_CERTIFY_INFO TPMS_CERTIFY_INFO_ARRAY;

SWIGINTERN TPMS_CERTIFY_INFO_ARRAY *new_TPMS_CERTIFY_INFO_ARRAY(size_t nelements){
    return (TPMS_CERTIFY_INFO *)malloc((nelements)*sizeof(TPMS_CERTIFY_INFO));
  }
SWIGINTERN void delete_TPMS_CERTIFY_INFO_ARRAY(TPMS_CERTIFY_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CERTIFY_INFO TPMS_CERTIFY_INFO_ARRAY___getitem__(TPMS_CERTIFY_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CERTIFY_INFO_ARRAY___setitem__(TPMS_CERTIFY_INFO_ARRAY *self,size_t index,TPMS_CERTIFY_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_CERTIFY_INFO *TPMS_CERTIFY_INFO_ARRAY_cast(TPMS_CERTIFY_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CERTIFY_INFO_ARRAY *TPMS_CERTIFY_INFO_ARRAY_frompointer(TPMS_CERTIFY_INFO *t){
    return (TPMS_CERTIFY_INFO_ARRAY *)(t);
  }

  static TPMS_CERTIFY_INFO *new_TPMS_CERTIFY_INFO_PTR() { 
    return (TPMS_CERTIFY_INFO *)malloc(sizeof(TPMS_CERTIFY_INFO));
  }
  
  static TPMS_CERTIFY_INFO *copy_TPMS_CERTIFY_INFO_PTR(TPMS_CERTIFY_INFO value) { 
    return (TPMS_CERTIFY_INFO *)memcpy((TPMS_CERTIFY_INFO *)malloc(sizeof(TPMS_CERTIFY_INFO)),&value,sizeof(TPMS_CERTIFY_INFO));
  }

  static void delete_TPMS_CERTIFY_INFO_PTR(TPMS_CERTIFY_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CERTIFY_INFO_PTR_assign(TPMS_CERTIFY_INFO *obj, TPMS_CERTIFY_INFO value) {
    *obj = value;
  }

  static TPMS_CERTIFY_INFO TPMS_CERTIFY_INFO_PTR_value(TPMS_CERTIFY_INFO *obj) {
    return *obj;
  }


  static TPMS_CERTIFY_INFO * *new_TPMS_CERTIFY_INFO_PTR_PTR() { 
    return (TPMS_CERTIFY_INFO * *)malloc(sizeof(TPMS_CERTIFY_INFO *));
  }
  
  static TPMS_CERTIFY_INFO * *copy_TPMS_CERTIFY_INFO_PTR_PTR(TPMS_CERTIFY_INFO * value) { 
    return (TPMS_CERTIFY_INFO * *)memcpy((TPMS_CERTIFY_INFO * *)malloc(sizeof(TPMS_CERTIFY_INFO *)),&value,sizeof(TPMS_CERTIFY_INFO *));
  }

  static void delete_TPMS_CERTIFY_INFO_PTR_PTR(TPMS_CERTIFY_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CERTIFY_INFO_PTR_PTR_assign(TPMS_CERTIFY_INFO * *obj, TPMS_CERTIFY_INFO * value) {
    *obj = value;
  }

  static TPMS_CERTIFY_INFO * TPMS_CERTIFY_INFO_PTR_PTR_value(TPMS_CERTIFY_INFO * *obj) {
    return *obj;
  }


typedef TPMS_QUOTE_INFO TPMS_QUOTE_INFO_ARRAY;

SWIGINTERN TPMS_QUOTE_INFO_ARRAY *new_TPMS_QUOTE_INFO_ARRAY(size_t nelements){
    return (TPMS_QUOTE_INFO *)malloc((nelements)*sizeof(TPMS_QUOTE_INFO));
  }
SWIGINTERN void delete_TPMS_QUOTE_INFO_ARRAY(TPMS_QUOTE_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_QUOTE_INFO TPMS_QUOTE_INFO_ARRAY___getitem__(TPMS_QUOTE_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_QUOTE_INFO_ARRAY___setitem__(TPMS_QUOTE_INFO_ARRAY *self,size_t index,TPMS_QUOTE_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_QUOTE_INFO *TPMS_QUOTE_INFO_ARRAY_cast(TPMS_QUOTE_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_QUOTE_INFO_ARRAY *TPMS_QUOTE_INFO_ARRAY_frompointer(TPMS_QUOTE_INFO *t){
    return (TPMS_QUOTE_INFO_ARRAY *)(t);
  }

  static TPMS_QUOTE_INFO *new_TPMS_QUOTE_INFO_PTR() { 
    return (TPMS_QUOTE_INFO *)malloc(sizeof(TPMS_QUOTE_INFO));
  }
  
  static TPMS_QUOTE_INFO *copy_TPMS_QUOTE_INFO_PTR(TPMS_QUOTE_INFO value) { 
    return (TPMS_QUOTE_INFO *)memcpy((TPMS_QUOTE_INFO *)malloc(sizeof(TPMS_QUOTE_INFO)),&value,sizeof(TPMS_QUOTE_INFO));
  }

  static void delete_TPMS_QUOTE_INFO_PTR(TPMS_QUOTE_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_QUOTE_INFO_PTR_assign(TPMS_QUOTE_INFO *obj, TPMS_QUOTE_INFO value) {
    *obj = value;
  }

  static TPMS_QUOTE_INFO TPMS_QUOTE_INFO_PTR_value(TPMS_QUOTE_INFO *obj) {
    return *obj;
  }


  static TPMS_QUOTE_INFO * *new_TPMS_QUOTE_INFO_PTR_PTR() { 
    return (TPMS_QUOTE_INFO * *)malloc(sizeof(TPMS_QUOTE_INFO *));
  }
  
  static TPMS_QUOTE_INFO * *copy_TPMS_QUOTE_INFO_PTR_PTR(TPMS_QUOTE_INFO * value) { 
    return (TPMS_QUOTE_INFO * *)memcpy((TPMS_QUOTE_INFO * *)malloc(sizeof(TPMS_QUOTE_INFO *)),&value,sizeof(TPMS_QUOTE_INFO *));
  }

  static void delete_TPMS_QUOTE_INFO_PTR_PTR(TPMS_QUOTE_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_QUOTE_INFO_PTR_PTR_assign(TPMS_QUOTE_INFO * *obj, TPMS_QUOTE_INFO * value) {
    *obj = value;
  }

  static TPMS_QUOTE_INFO * TPMS_QUOTE_INFO_PTR_PTR_value(TPMS_QUOTE_INFO * *obj) {
    return *obj;
  }


typedef TPMS_COMMAND_AUDIT_INFO TPMS_COMMAND_AUDIT_INFO_ARRAY;

SWIGINTERN TPMS_COMMAND_AUDIT_INFO_ARRAY *new_TPMS_COMMAND_AUDIT_INFO_ARRAY(size_t nelements){
    return (TPMS_COMMAND_AUDIT_INFO *)malloc((nelements)*sizeof(TPMS_COMMAND_AUDIT_INFO));
  }
SWIGINTERN void delete_TPMS_COMMAND_AUDIT_INFO_ARRAY(TPMS_COMMAND_AUDIT_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_COMMAND_AUDIT_INFO TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__(TPMS_COMMAND_AUDIT_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__(TPMS_COMMAND_AUDIT_INFO_ARRAY *self,size_t index,TPMS_COMMAND_AUDIT_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_COMMAND_AUDIT_INFO *TPMS_COMMAND_AUDIT_INFO_ARRAY_cast(TPMS_COMMAND_AUDIT_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_COMMAND_AUDIT_INFO_ARRAY *TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer(TPMS_COMMAND_AUDIT_INFO *t){
    return (TPMS_COMMAND_AUDIT_INFO_ARRAY *)(t);
  }

  static TPMS_COMMAND_AUDIT_INFO *new_TPMS_COMMAND_AUDIT_INFO_PTR() { 
    return (TPMS_COMMAND_AUDIT_INFO *)malloc(sizeof(TPMS_COMMAND_AUDIT_INFO));
  }
  
  static TPMS_COMMAND_AUDIT_INFO *copy_TPMS_COMMAND_AUDIT_INFO_PTR(TPMS_COMMAND_AUDIT_INFO value) { 
    return (TPMS_COMMAND_AUDIT_INFO *)memcpy((TPMS_COMMAND_AUDIT_INFO *)malloc(sizeof(TPMS_COMMAND_AUDIT_INFO)),&value,sizeof(TPMS_COMMAND_AUDIT_INFO));
  }

  static void delete_TPMS_COMMAND_AUDIT_INFO_PTR(TPMS_COMMAND_AUDIT_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_COMMAND_AUDIT_INFO_PTR_assign(TPMS_COMMAND_AUDIT_INFO *obj, TPMS_COMMAND_AUDIT_INFO value) {
    *obj = value;
  }

  static TPMS_COMMAND_AUDIT_INFO TPMS_COMMAND_AUDIT_INFO_PTR_value(TPMS_COMMAND_AUDIT_INFO *obj) {
    return *obj;
  }


  static TPMS_COMMAND_AUDIT_INFO * *new_TPMS_COMMAND_AUDIT_INFO_PTR_PTR() { 
    return (TPMS_COMMAND_AUDIT_INFO * *)malloc(sizeof(TPMS_COMMAND_AUDIT_INFO *));
  }
  
  static TPMS_COMMAND_AUDIT_INFO * *copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(TPMS_COMMAND_AUDIT_INFO * value) { 
    return (TPMS_COMMAND_AUDIT_INFO * *)memcpy((TPMS_COMMAND_AUDIT_INFO * *)malloc(sizeof(TPMS_COMMAND_AUDIT_INFO *)),&value,sizeof(TPMS_COMMAND_AUDIT_INFO *));
  }

  static void delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(TPMS_COMMAND_AUDIT_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign(TPMS_COMMAND_AUDIT_INFO * *obj, TPMS_COMMAND_AUDIT_INFO * value) {
    *obj = value;
  }

  static TPMS_COMMAND_AUDIT_INFO * TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value(TPMS_COMMAND_AUDIT_INFO * *obj) {
    return *obj;
  }


typedef TPMS_SESSION_AUDIT_INFO TPMS_SESSION_AUDIT_INFO_ARRAY;

SWIGINTERN TPMS_SESSION_AUDIT_INFO_ARRAY *new_TPMS_SESSION_AUDIT_INFO_ARRAY(size_t nelements){
    return (TPMS_SESSION_AUDIT_INFO *)malloc((nelements)*sizeof(TPMS_SESSION_AUDIT_INFO));
  }
SWIGINTERN void delete_TPMS_SESSION_AUDIT_INFO_ARRAY(TPMS_SESSION_AUDIT_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SESSION_AUDIT_INFO TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__(TPMS_SESSION_AUDIT_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__(TPMS_SESSION_AUDIT_INFO_ARRAY *self,size_t index,TPMS_SESSION_AUDIT_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_SESSION_AUDIT_INFO *TPMS_SESSION_AUDIT_INFO_ARRAY_cast(TPMS_SESSION_AUDIT_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SESSION_AUDIT_INFO_ARRAY *TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer(TPMS_SESSION_AUDIT_INFO *t){
    return (TPMS_SESSION_AUDIT_INFO_ARRAY *)(t);
  }

  static TPMS_SESSION_AUDIT_INFO *new_TPMS_SESSION_AUDIT_INFO_PTR() { 
    return (TPMS_SESSION_AUDIT_INFO *)malloc(sizeof(TPMS_SESSION_AUDIT_INFO));
  }
  
  static TPMS_SESSION_AUDIT_INFO *copy_TPMS_SESSION_AUDIT_INFO_PTR(TPMS_SESSION_AUDIT_INFO value) { 
    return (TPMS_SESSION_AUDIT_INFO *)memcpy((TPMS_SESSION_AUDIT_INFO *)malloc(sizeof(TPMS_SESSION_AUDIT_INFO)),&value,sizeof(TPMS_SESSION_AUDIT_INFO));
  }

  static void delete_TPMS_SESSION_AUDIT_INFO_PTR(TPMS_SESSION_AUDIT_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SESSION_AUDIT_INFO_PTR_assign(TPMS_SESSION_AUDIT_INFO *obj, TPMS_SESSION_AUDIT_INFO value) {
    *obj = value;
  }

  static TPMS_SESSION_AUDIT_INFO TPMS_SESSION_AUDIT_INFO_PTR_value(TPMS_SESSION_AUDIT_INFO *obj) {
    return *obj;
  }


  static TPMS_SESSION_AUDIT_INFO * *new_TPMS_SESSION_AUDIT_INFO_PTR_PTR() { 
    return (TPMS_SESSION_AUDIT_INFO * *)malloc(sizeof(TPMS_SESSION_AUDIT_INFO *));
  }
  
  static TPMS_SESSION_AUDIT_INFO * *copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR(TPMS_SESSION_AUDIT_INFO * value) { 
    return (TPMS_SESSION_AUDIT_INFO * *)memcpy((TPMS_SESSION_AUDIT_INFO * *)malloc(sizeof(TPMS_SESSION_AUDIT_INFO *)),&value,sizeof(TPMS_SESSION_AUDIT_INFO *));
  }

  static void delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR(TPMS_SESSION_AUDIT_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign(TPMS_SESSION_AUDIT_INFO * *obj, TPMS_SESSION_AUDIT_INFO * value) {
    *obj = value;
  }

  static TPMS_SESSION_AUDIT_INFO * TPMS_SESSION_AUDIT_INFO_PTR_PTR_value(TPMS_SESSION_AUDIT_INFO * *obj) {
    return *obj;
  }


typedef TPMS_CREATION_INFO TPMS_CREATION_INFO_ARRAY;

SWIGINTERN TPMS_CREATION_INFO_ARRAY *new_TPMS_CREATION_INFO_ARRAY(size_t nelements){
    return (TPMS_CREATION_INFO *)malloc((nelements)*sizeof(TPMS_CREATION_INFO));
  }
SWIGINTERN void delete_TPMS_CREATION_INFO_ARRAY(TPMS_CREATION_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CREATION_INFO TPMS_CREATION_INFO_ARRAY___getitem__(TPMS_CREATION_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CREATION_INFO_ARRAY___setitem__(TPMS_CREATION_INFO_ARRAY *self,size_t index,TPMS_CREATION_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_CREATION_INFO *TPMS_CREATION_INFO_ARRAY_cast(TPMS_CREATION_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CREATION_INFO_ARRAY *TPMS_CREATION_INFO_ARRAY_frompointer(TPMS_CREATION_INFO *t){
    return (TPMS_CREATION_INFO_ARRAY *)(t);
  }

  static TPMS_CREATION_INFO *new_TPMS_CREATION_INFO_PTR() { 
    return (TPMS_CREATION_INFO *)malloc(sizeof(TPMS_CREATION_INFO));
  }
  
  static TPMS_CREATION_INFO *copy_TPMS_CREATION_INFO_PTR(TPMS_CREATION_INFO value) { 
    return (TPMS_CREATION_INFO *)memcpy((TPMS_CREATION_INFO *)malloc(sizeof(TPMS_CREATION_INFO)),&value,sizeof(TPMS_CREATION_INFO));
  }

  static void delete_TPMS_CREATION_INFO_PTR(TPMS_CREATION_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CREATION_INFO_PTR_assign(TPMS_CREATION_INFO *obj, TPMS_CREATION_INFO value) {
    *obj = value;
  }

  static TPMS_CREATION_INFO TPMS_CREATION_INFO_PTR_value(TPMS_CREATION_INFO *obj) {
    return *obj;
  }


  static TPMS_CREATION_INFO * *new_TPMS_CREATION_INFO_PTR_PTR() { 
    return (TPMS_CREATION_INFO * *)malloc(sizeof(TPMS_CREATION_INFO *));
  }
  
  static TPMS_CREATION_INFO * *copy_TPMS_CREATION_INFO_PTR_PTR(TPMS_CREATION_INFO * value) { 
    return (TPMS_CREATION_INFO * *)memcpy((TPMS_CREATION_INFO * *)malloc(sizeof(TPMS_CREATION_INFO *)),&value,sizeof(TPMS_CREATION_INFO *));
  }

  static void delete_TPMS_CREATION_INFO_PTR_PTR(TPMS_CREATION_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CREATION_INFO_PTR_PTR_assign(TPMS_CREATION_INFO * *obj, TPMS_CREATION_INFO * value) {
    *obj = value;
  }

  static TPMS_CREATION_INFO * TPMS_CREATION_INFO_PTR_PTR_value(TPMS_CREATION_INFO * *obj) {
    return *obj;
  }


typedef TPMS_NV_CERTIFY_INFO TPMS_NV_CERTIFY_INFO_ARRAY;

SWIGINTERN TPMS_NV_CERTIFY_INFO_ARRAY *new_TPMS_NV_CERTIFY_INFO_ARRAY(size_t nelements){
    return (TPMS_NV_CERTIFY_INFO *)malloc((nelements)*sizeof(TPMS_NV_CERTIFY_INFO));
  }
SWIGINTERN void delete_TPMS_NV_CERTIFY_INFO_ARRAY(TPMS_NV_CERTIFY_INFO_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_NV_CERTIFY_INFO TPMS_NV_CERTIFY_INFO_ARRAY___getitem__(TPMS_NV_CERTIFY_INFO_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_NV_CERTIFY_INFO_ARRAY___setitem__(TPMS_NV_CERTIFY_INFO_ARRAY *self,size_t index,TPMS_NV_CERTIFY_INFO value){
    self[index] = value;
  }
SWIGINTERN TPMS_NV_CERTIFY_INFO *TPMS_NV_CERTIFY_INFO_ARRAY_cast(TPMS_NV_CERTIFY_INFO_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_NV_CERTIFY_INFO_ARRAY *TPMS_NV_CERTIFY_INFO_ARRAY_frompointer(TPMS_NV_CERTIFY_INFO *t){
    return (TPMS_NV_CERTIFY_INFO_ARRAY *)(t);
  }

  static TPMS_NV_CERTIFY_INFO *new_TPMS_NV_CERTIFY_INFO_PTR() { 
    return (TPMS_NV_CERTIFY_INFO *)malloc(sizeof(TPMS_NV_CERTIFY_INFO));
  }
  
  static TPMS_NV_CERTIFY_INFO *copy_TPMS_NV_CERTIFY_INFO_PTR(TPMS_NV_CERTIFY_INFO value) { 
    return (TPMS_NV_CERTIFY_INFO *)memcpy((TPMS_NV_CERTIFY_INFO *)malloc(sizeof(TPMS_NV_CERTIFY_INFO)),&value,sizeof(TPMS_NV_CERTIFY_INFO));
  }

  static void delete_TPMS_NV_CERTIFY_INFO_PTR(TPMS_NV_CERTIFY_INFO *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_NV_CERTIFY_INFO_PTR_assign(TPMS_NV_CERTIFY_INFO *obj, TPMS_NV_CERTIFY_INFO value) {
    *obj = value;
  }

  static TPMS_NV_CERTIFY_INFO TPMS_NV_CERTIFY_INFO_PTR_value(TPMS_NV_CERTIFY_INFO *obj) {
    return *obj;
  }


  static TPMS_NV_CERTIFY_INFO * *new_TPMS_NV_CERTIFY_INFO_PTR_PTR() { 
    return (TPMS_NV_CERTIFY_INFO * *)malloc(sizeof(TPMS_NV_CERTIFY_INFO *));
  }
  
  static TPMS_NV_CERTIFY_INFO * *copy_TPMS_NV_CERTIFY_INFO_PTR_PTR(TPMS_NV_CERTIFY_INFO * value) { 
    return (TPMS_NV_CERTIFY_INFO * *)memcpy((TPMS_NV_CERTIFY_INFO * *)malloc(sizeof(TPMS_NV_CERTIFY_INFO *)),&value,sizeof(TPMS_NV_CERTIFY_INFO *));
  }

  static void delete_TPMS_NV_CERTIFY_INFO_PTR_PTR(TPMS_NV_CERTIFY_INFO * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_NV_CERTIFY_INFO_PTR_PTR_assign(TPMS_NV_CERTIFY_INFO * *obj, TPMS_NV_CERTIFY_INFO * value) {
    *obj = value;
  }

  static TPMS_NV_CERTIFY_INFO * TPMS_NV_CERTIFY_INFO_PTR_PTR_value(TPMS_NV_CERTIFY_INFO * *obj) {
    return *obj;
  }


typedef TPMU_ATTEST TPMU_ATTEST_ARRAY;

SWIGINTERN TPMU_ATTEST_ARRAY *new_TPMU_ATTEST_ARRAY(size_t nelements){
    return (TPMU_ATTEST *)malloc((nelements)*sizeof(TPMU_ATTEST));
  }
SWIGINTERN void delete_TPMU_ATTEST_ARRAY(TPMU_ATTEST_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_ATTEST TPMU_ATTEST_ARRAY___getitem__(TPMU_ATTEST_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_ATTEST_ARRAY___setitem__(TPMU_ATTEST_ARRAY *self,size_t index,TPMU_ATTEST value){
    self[index] = value;
  }
SWIGINTERN TPMU_ATTEST *TPMU_ATTEST_ARRAY_cast(TPMU_ATTEST_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_ATTEST_ARRAY *TPMU_ATTEST_ARRAY_frompointer(TPMU_ATTEST *t){
    return (TPMU_ATTEST_ARRAY *)(t);
  }

  static TPMU_ATTEST *new_TPMU_ATTEST_PTR() { 
    return (TPMU_ATTEST *)malloc(sizeof(TPMU_ATTEST));
  }
  
  static TPMU_ATTEST *copy_TPMU_ATTEST_PTR(TPMU_ATTEST value) { 
    return (TPMU_ATTEST *)memcpy((TPMU_ATTEST *)malloc(sizeof(TPMU_ATTEST)),&value,sizeof(TPMU_ATTEST));
  }

  static void delete_TPMU_ATTEST_PTR(TPMU_ATTEST *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_ATTEST_PTR_assign(TPMU_ATTEST *obj, TPMU_ATTEST value) {
    *obj = value;
  }

  static TPMU_ATTEST TPMU_ATTEST_PTR_value(TPMU_ATTEST *obj) {
    return *obj;
  }


  static TPMU_ATTEST * *new_TPMU_ATTEST_PTR_PTR() { 
    return (TPMU_ATTEST * *)malloc(sizeof(TPMU_ATTEST *));
  }
  
  static TPMU_ATTEST * *copy_TPMU_ATTEST_PTR_PTR(TPMU_ATTEST * value) { 
    return (TPMU_ATTEST * *)memcpy((TPMU_ATTEST * *)malloc(sizeof(TPMU_ATTEST *)),&value,sizeof(TPMU_ATTEST *));
  }

  static void delete_TPMU_ATTEST_PTR_PTR(TPMU_ATTEST * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_ATTEST_PTR_PTR_assign(TPMU_ATTEST * *obj, TPMU_ATTEST * value) {
    *obj = value;
  }

  static TPMU_ATTEST * TPMU_ATTEST_PTR_PTR_value(TPMU_ATTEST * *obj) {
    return *obj;
  }


typedef TPMS_ATTEST TPMS_ATTEST_ARRAY;

SWIGINTERN TPMS_ATTEST_ARRAY *new_TPMS_ATTEST_ARRAY(size_t nelements){
    return (TPMS_ATTEST *)malloc((nelements)*sizeof(TPMS_ATTEST));
  }
SWIGINTERN void delete_TPMS_ATTEST_ARRAY(TPMS_ATTEST_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ATTEST TPMS_ATTEST_ARRAY___getitem__(TPMS_ATTEST_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ATTEST_ARRAY___setitem__(TPMS_ATTEST_ARRAY *self,size_t index,TPMS_ATTEST value){
    self[index] = value;
  }
SWIGINTERN TPMS_ATTEST *TPMS_ATTEST_ARRAY_cast(TPMS_ATTEST_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ATTEST_ARRAY *TPMS_ATTEST_ARRAY_frompointer(TPMS_ATTEST *t){
    return (TPMS_ATTEST_ARRAY *)(t);
  }

  static TPMS_ATTEST *new_TPMS_ATTEST_PTR() { 
    return (TPMS_ATTEST *)malloc(sizeof(TPMS_ATTEST));
  }
  
  static TPMS_ATTEST *copy_TPMS_ATTEST_PTR(TPMS_ATTEST value) { 
    return (TPMS_ATTEST *)memcpy((TPMS_ATTEST *)malloc(sizeof(TPMS_ATTEST)),&value,sizeof(TPMS_ATTEST));
  }

  static void delete_TPMS_ATTEST_PTR(TPMS_ATTEST *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ATTEST_PTR_assign(TPMS_ATTEST *obj, TPMS_ATTEST value) {
    *obj = value;
  }

  static TPMS_ATTEST TPMS_ATTEST_PTR_value(TPMS_ATTEST *obj) {
    return *obj;
  }


  static TPMS_ATTEST * *new_TPMS_ATTEST_PTR_PTR() { 
    return (TPMS_ATTEST * *)malloc(sizeof(TPMS_ATTEST *));
  }
  
  static TPMS_ATTEST * *copy_TPMS_ATTEST_PTR_PTR(TPMS_ATTEST * value) { 
    return (TPMS_ATTEST * *)memcpy((TPMS_ATTEST * *)malloc(sizeof(TPMS_ATTEST *)),&value,sizeof(TPMS_ATTEST *));
  }

  static void delete_TPMS_ATTEST_PTR_PTR(TPMS_ATTEST * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ATTEST_PTR_PTR_assign(TPMS_ATTEST * *obj, TPMS_ATTEST * value) {
    *obj = value;
  }

  static TPMS_ATTEST * TPMS_ATTEST_PTR_PTR_value(TPMS_ATTEST * *obj) {
    return *obj;
  }


typedef TPM2B_ATTEST TPM2B_ATTEST_ARRAY;

SWIGINTERN TPM2B_ATTEST_ARRAY *new_TPM2B_ATTEST_ARRAY(size_t nelements){
    return (TPM2B_ATTEST *)malloc((nelements)*sizeof(TPM2B_ATTEST));
  }
SWIGINTERN void delete_TPM2B_ATTEST_ARRAY(TPM2B_ATTEST_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_ATTEST TPM2B_ATTEST_ARRAY___getitem__(TPM2B_ATTEST_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_ATTEST_ARRAY___setitem__(TPM2B_ATTEST_ARRAY *self,size_t index,TPM2B_ATTEST value){
    self[index] = value;
  }
SWIGINTERN TPM2B_ATTEST *TPM2B_ATTEST_ARRAY_cast(TPM2B_ATTEST_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_ATTEST_ARRAY *TPM2B_ATTEST_ARRAY_frompointer(TPM2B_ATTEST *t){
    return (TPM2B_ATTEST_ARRAY *)(t);
  }

  static TPM2B_ATTEST *new_TPM2B_ATTEST_PTR() { 
    return (TPM2B_ATTEST *)malloc(sizeof(TPM2B_ATTEST));
  }
  
  static TPM2B_ATTEST *copy_TPM2B_ATTEST_PTR(TPM2B_ATTEST value) { 
    return (TPM2B_ATTEST *)memcpy((TPM2B_ATTEST *)malloc(sizeof(TPM2B_ATTEST)),&value,sizeof(TPM2B_ATTEST));
  }

  static void delete_TPM2B_ATTEST_PTR(TPM2B_ATTEST *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ATTEST_PTR_assign(TPM2B_ATTEST *obj, TPM2B_ATTEST value) {
    *obj = value;
  }

  static TPM2B_ATTEST TPM2B_ATTEST_PTR_value(TPM2B_ATTEST *obj) {
    return *obj;
  }


  static TPM2B_ATTEST * *new_TPM2B_ATTEST_PTR_PTR() { 
    return (TPM2B_ATTEST * *)malloc(sizeof(TPM2B_ATTEST *));
  }
  
  static TPM2B_ATTEST * *copy_TPM2B_ATTEST_PTR_PTR(TPM2B_ATTEST * value) { 
    return (TPM2B_ATTEST * *)memcpy((TPM2B_ATTEST * *)malloc(sizeof(TPM2B_ATTEST *)),&value,sizeof(TPM2B_ATTEST *));
  }

  static void delete_TPM2B_ATTEST_PTR_PTR(TPM2B_ATTEST * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ATTEST_PTR_PTR_assign(TPM2B_ATTEST * *obj, TPM2B_ATTEST * value) {
    *obj = value;
  }

  static TPM2B_ATTEST * TPM2B_ATTEST_PTR_PTR_value(TPM2B_ATTEST * *obj) {
    return *obj;
  }


typedef TPMS_AUTH_COMMAND TPMS_AUTH_COMMAND_ARRAY;

SWIGINTERN TPMS_AUTH_COMMAND_ARRAY *new_TPMS_AUTH_COMMAND_ARRAY(size_t nelements){
    return (TPMS_AUTH_COMMAND *)malloc((nelements)*sizeof(TPMS_AUTH_COMMAND));
  }
SWIGINTERN void delete_TPMS_AUTH_COMMAND_ARRAY(TPMS_AUTH_COMMAND_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_AUTH_COMMAND TPMS_AUTH_COMMAND_ARRAY___getitem__(TPMS_AUTH_COMMAND_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_AUTH_COMMAND_ARRAY___setitem__(TPMS_AUTH_COMMAND_ARRAY *self,size_t index,TPMS_AUTH_COMMAND value){
    self[index] = value;
  }
SWIGINTERN TPMS_AUTH_COMMAND *TPMS_AUTH_COMMAND_ARRAY_cast(TPMS_AUTH_COMMAND_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_AUTH_COMMAND_ARRAY *TPMS_AUTH_COMMAND_ARRAY_frompointer(TPMS_AUTH_COMMAND *t){
    return (TPMS_AUTH_COMMAND_ARRAY *)(t);
  }

  static TPMS_AUTH_COMMAND *new_TPMS_AUTH_COMMAND_PTR() { 
    return (TPMS_AUTH_COMMAND *)malloc(sizeof(TPMS_AUTH_COMMAND));
  }
  
  static TPMS_AUTH_COMMAND *copy_TPMS_AUTH_COMMAND_PTR(TPMS_AUTH_COMMAND value) { 
    return (TPMS_AUTH_COMMAND *)memcpy((TPMS_AUTH_COMMAND *)malloc(sizeof(TPMS_AUTH_COMMAND)),&value,sizeof(TPMS_AUTH_COMMAND));
  }

  static void delete_TPMS_AUTH_COMMAND_PTR(TPMS_AUTH_COMMAND *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_AUTH_COMMAND_PTR_assign(TPMS_AUTH_COMMAND *obj, TPMS_AUTH_COMMAND value) {
    *obj = value;
  }

  static TPMS_AUTH_COMMAND TPMS_AUTH_COMMAND_PTR_value(TPMS_AUTH_COMMAND *obj) {
    return *obj;
  }


  static TPMS_AUTH_COMMAND * *new_TPMS_AUTH_COMMAND_PTR_PTR() { 
    return (TPMS_AUTH_COMMAND * *)malloc(sizeof(TPMS_AUTH_COMMAND *));
  }
  
  static TPMS_AUTH_COMMAND * *copy_TPMS_AUTH_COMMAND_PTR_PTR(TPMS_AUTH_COMMAND * value) { 
    return (TPMS_AUTH_COMMAND * *)memcpy((TPMS_AUTH_COMMAND * *)malloc(sizeof(TPMS_AUTH_COMMAND *)),&value,sizeof(TPMS_AUTH_COMMAND *));
  }

  static void delete_TPMS_AUTH_COMMAND_PTR_PTR(TPMS_AUTH_COMMAND * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_AUTH_COMMAND_PTR_PTR_assign(TPMS_AUTH_COMMAND * *obj, TPMS_AUTH_COMMAND * value) {
    *obj = value;
  }

  static TPMS_AUTH_COMMAND * TPMS_AUTH_COMMAND_PTR_PTR_value(TPMS_AUTH_COMMAND * *obj) {
    return *obj;
  }


typedef TPMS_AUTH_RESPONSE TPMS_AUTH_RESPONSE_ARRAY;

SWIGINTERN TPMS_AUTH_RESPONSE_ARRAY *new_TPMS_AUTH_RESPONSE_ARRAY(size_t nelements){
    return (TPMS_AUTH_RESPONSE *)malloc((nelements)*sizeof(TPMS_AUTH_RESPONSE));
  }
SWIGINTERN void delete_TPMS_AUTH_RESPONSE_ARRAY(TPMS_AUTH_RESPONSE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_AUTH_RESPONSE TPMS_AUTH_RESPONSE_ARRAY___getitem__(TPMS_AUTH_RESPONSE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_AUTH_RESPONSE_ARRAY___setitem__(TPMS_AUTH_RESPONSE_ARRAY *self,size_t index,TPMS_AUTH_RESPONSE value){
    self[index] = value;
  }
SWIGINTERN TPMS_AUTH_RESPONSE *TPMS_AUTH_RESPONSE_ARRAY_cast(TPMS_AUTH_RESPONSE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_AUTH_RESPONSE_ARRAY *TPMS_AUTH_RESPONSE_ARRAY_frompointer(TPMS_AUTH_RESPONSE *t){
    return (TPMS_AUTH_RESPONSE_ARRAY *)(t);
  }

  static TPMS_AUTH_RESPONSE *new_TPMS_AUTH_RESPONSE_PTR() { 
    return (TPMS_AUTH_RESPONSE *)malloc(sizeof(TPMS_AUTH_RESPONSE));
  }
  
  static TPMS_AUTH_RESPONSE *copy_TPMS_AUTH_RESPONSE_PTR(TPMS_AUTH_RESPONSE value) { 
    return (TPMS_AUTH_RESPONSE *)memcpy((TPMS_AUTH_RESPONSE *)malloc(sizeof(TPMS_AUTH_RESPONSE)),&value,sizeof(TPMS_AUTH_RESPONSE));
  }

  static void delete_TPMS_AUTH_RESPONSE_PTR(TPMS_AUTH_RESPONSE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_AUTH_RESPONSE_PTR_assign(TPMS_AUTH_RESPONSE *obj, TPMS_AUTH_RESPONSE value) {
    *obj = value;
  }

  static TPMS_AUTH_RESPONSE TPMS_AUTH_RESPONSE_PTR_value(TPMS_AUTH_RESPONSE *obj) {
    return *obj;
  }


  static TPMS_AUTH_RESPONSE * *new_TPMS_AUTH_RESPONSE_PTR_PTR() { 
    return (TPMS_AUTH_RESPONSE * *)malloc(sizeof(TPMS_AUTH_RESPONSE *));
  }
  
  static TPMS_AUTH_RESPONSE * *copy_TPMS_AUTH_RESPONSE_PTR_PTR(TPMS_AUTH_RESPONSE * value) { 
    return (TPMS_AUTH_RESPONSE * *)memcpy((TPMS_AUTH_RESPONSE * *)malloc(sizeof(TPMS_AUTH_RESPONSE *)),&value,sizeof(TPMS_AUTH_RESPONSE *));
  }

  static void delete_TPMS_AUTH_RESPONSE_PTR_PTR(TPMS_AUTH_RESPONSE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_AUTH_RESPONSE_PTR_PTR_assign(TPMS_AUTH_RESPONSE * *obj, TPMS_AUTH_RESPONSE * value) {
    *obj = value;
  }

  static TPMS_AUTH_RESPONSE * TPMS_AUTH_RESPONSE_PTR_PTR_value(TPMS_AUTH_RESPONSE * *obj) {
    return *obj;
  }


typedef TPMU_SYM_KEY_BITS TPMU_SYM_KEY_BITS_ARRAY;

SWIGINTERN TPMU_SYM_KEY_BITS_ARRAY *new_TPMU_SYM_KEY_BITS_ARRAY(size_t nelements){
    return (TPMU_SYM_KEY_BITS *)malloc((nelements)*sizeof(TPMU_SYM_KEY_BITS));
  }
SWIGINTERN void delete_TPMU_SYM_KEY_BITS_ARRAY(TPMU_SYM_KEY_BITS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_SYM_KEY_BITS TPMU_SYM_KEY_BITS_ARRAY___getitem__(TPMU_SYM_KEY_BITS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_SYM_KEY_BITS_ARRAY___setitem__(TPMU_SYM_KEY_BITS_ARRAY *self,size_t index,TPMU_SYM_KEY_BITS value){
    self[index] = value;
  }
SWIGINTERN TPMU_SYM_KEY_BITS *TPMU_SYM_KEY_BITS_ARRAY_cast(TPMU_SYM_KEY_BITS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_SYM_KEY_BITS_ARRAY *TPMU_SYM_KEY_BITS_ARRAY_frompointer(TPMU_SYM_KEY_BITS *t){
    return (TPMU_SYM_KEY_BITS_ARRAY *)(t);
  }

  static TPMU_SYM_KEY_BITS *new_TPMU_SYM_KEY_BITS_PTR() { 
    return (TPMU_SYM_KEY_BITS *)malloc(sizeof(TPMU_SYM_KEY_BITS));
  }
  
  static TPMU_SYM_KEY_BITS *copy_TPMU_SYM_KEY_BITS_PTR(TPMU_SYM_KEY_BITS value) { 
    return (TPMU_SYM_KEY_BITS *)memcpy((TPMU_SYM_KEY_BITS *)malloc(sizeof(TPMU_SYM_KEY_BITS)),&value,sizeof(TPMU_SYM_KEY_BITS));
  }

  static void delete_TPMU_SYM_KEY_BITS_PTR(TPMU_SYM_KEY_BITS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SYM_KEY_BITS_PTR_assign(TPMU_SYM_KEY_BITS *obj, TPMU_SYM_KEY_BITS value) {
    *obj = value;
  }

  static TPMU_SYM_KEY_BITS TPMU_SYM_KEY_BITS_PTR_value(TPMU_SYM_KEY_BITS *obj) {
    return *obj;
  }


  static TPMU_SYM_KEY_BITS * *new_TPMU_SYM_KEY_BITS_PTR_PTR() { 
    return (TPMU_SYM_KEY_BITS * *)malloc(sizeof(TPMU_SYM_KEY_BITS *));
  }
  
  static TPMU_SYM_KEY_BITS * *copy_TPMU_SYM_KEY_BITS_PTR_PTR(TPMU_SYM_KEY_BITS * value) { 
    return (TPMU_SYM_KEY_BITS * *)memcpy((TPMU_SYM_KEY_BITS * *)malloc(sizeof(TPMU_SYM_KEY_BITS *)),&value,sizeof(TPMU_SYM_KEY_BITS *));
  }

  static void delete_TPMU_SYM_KEY_BITS_PTR_PTR(TPMU_SYM_KEY_BITS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SYM_KEY_BITS_PTR_PTR_assign(TPMU_SYM_KEY_BITS * *obj, TPMU_SYM_KEY_BITS * value) {
    *obj = value;
  }

  static TPMU_SYM_KEY_BITS * TPMU_SYM_KEY_BITS_PTR_PTR_value(TPMU_SYM_KEY_BITS * *obj) {
    return *obj;
  }


typedef TPMU_SYM_MODE TPMU_SYM_MODE_ARRAY;

SWIGINTERN TPMU_SYM_MODE_ARRAY *new_TPMU_SYM_MODE_ARRAY(size_t nelements){
    return (TPMU_SYM_MODE *)malloc((nelements)*sizeof(TPMU_SYM_MODE));
  }
SWIGINTERN void delete_TPMU_SYM_MODE_ARRAY(TPMU_SYM_MODE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_SYM_MODE TPMU_SYM_MODE_ARRAY___getitem__(TPMU_SYM_MODE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_SYM_MODE_ARRAY___setitem__(TPMU_SYM_MODE_ARRAY *self,size_t index,TPMU_SYM_MODE value){
    self[index] = value;
  }
SWIGINTERN TPMU_SYM_MODE *TPMU_SYM_MODE_ARRAY_cast(TPMU_SYM_MODE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_SYM_MODE_ARRAY *TPMU_SYM_MODE_ARRAY_frompointer(TPMU_SYM_MODE *t){
    return (TPMU_SYM_MODE_ARRAY *)(t);
  }

  static TPMU_SYM_MODE *new_TPMU_SYM_MODE_PTR() { 
    return (TPMU_SYM_MODE *)malloc(sizeof(TPMU_SYM_MODE));
  }
  
  static TPMU_SYM_MODE *copy_TPMU_SYM_MODE_PTR(TPMU_SYM_MODE value) { 
    return (TPMU_SYM_MODE *)memcpy((TPMU_SYM_MODE *)malloc(sizeof(TPMU_SYM_MODE)),&value,sizeof(TPMU_SYM_MODE));
  }

  static void delete_TPMU_SYM_MODE_PTR(TPMU_SYM_MODE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SYM_MODE_PTR_assign(TPMU_SYM_MODE *obj, TPMU_SYM_MODE value) {
    *obj = value;
  }

  static TPMU_SYM_MODE TPMU_SYM_MODE_PTR_value(TPMU_SYM_MODE *obj) {
    return *obj;
  }


  static TPMU_SYM_MODE * *new_TPMU_SYM_MODE_PTR_PTR() { 
    return (TPMU_SYM_MODE * *)malloc(sizeof(TPMU_SYM_MODE *));
  }
  
  static TPMU_SYM_MODE * *copy_TPMU_SYM_MODE_PTR_PTR(TPMU_SYM_MODE * value) { 
    return (TPMU_SYM_MODE * *)memcpy((TPMU_SYM_MODE * *)malloc(sizeof(TPMU_SYM_MODE *)),&value,sizeof(TPMU_SYM_MODE *));
  }

  static void delete_TPMU_SYM_MODE_PTR_PTR(TPMU_SYM_MODE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SYM_MODE_PTR_PTR_assign(TPMU_SYM_MODE * *obj, TPMU_SYM_MODE * value) {
    *obj = value;
  }

  static TPMU_SYM_MODE * TPMU_SYM_MODE_PTR_PTR_value(TPMU_SYM_MODE * *obj) {
    return *obj;
  }


typedef TPMT_SYM_DEF TPMT_SYM_DEF_ARRAY;

SWIGINTERN TPMT_SYM_DEF_ARRAY *new_TPMT_SYM_DEF_ARRAY(size_t nelements){
    return (TPMT_SYM_DEF *)malloc((nelements)*sizeof(TPMT_SYM_DEF));
  }
SWIGINTERN void delete_TPMT_SYM_DEF_ARRAY(TPMT_SYM_DEF_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_SYM_DEF TPMT_SYM_DEF_ARRAY___getitem__(TPMT_SYM_DEF_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_SYM_DEF_ARRAY___setitem__(TPMT_SYM_DEF_ARRAY *self,size_t index,TPMT_SYM_DEF value){
    self[index] = value;
  }
SWIGINTERN TPMT_SYM_DEF *TPMT_SYM_DEF_ARRAY_cast(TPMT_SYM_DEF_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_SYM_DEF_ARRAY *TPMT_SYM_DEF_ARRAY_frompointer(TPMT_SYM_DEF *t){
    return (TPMT_SYM_DEF_ARRAY *)(t);
  }

  static TPMT_SYM_DEF *new_TPMT_SYM_DEF_PTR() { 
    return (TPMT_SYM_DEF *)malloc(sizeof(TPMT_SYM_DEF));
  }
  
  static TPMT_SYM_DEF *copy_TPMT_SYM_DEF_PTR(TPMT_SYM_DEF value) { 
    return (TPMT_SYM_DEF *)memcpy((TPMT_SYM_DEF *)malloc(sizeof(TPMT_SYM_DEF)),&value,sizeof(TPMT_SYM_DEF));
  }

  static void delete_TPMT_SYM_DEF_PTR(TPMT_SYM_DEF *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SYM_DEF_PTR_assign(TPMT_SYM_DEF *obj, TPMT_SYM_DEF value) {
    *obj = value;
  }

  static TPMT_SYM_DEF TPMT_SYM_DEF_PTR_value(TPMT_SYM_DEF *obj) {
    return *obj;
  }


  static TPMT_SYM_DEF * *new_TPMT_SYM_DEF_PTR_PTR() { 
    return (TPMT_SYM_DEF * *)malloc(sizeof(TPMT_SYM_DEF *));
  }
  
  static TPMT_SYM_DEF * *copy_TPMT_SYM_DEF_PTR_PTR(TPMT_SYM_DEF * value) { 
    return (TPMT_SYM_DEF * *)memcpy((TPMT_SYM_DEF * *)malloc(sizeof(TPMT_SYM_DEF *)),&value,sizeof(TPMT_SYM_DEF *));
  }

  static void delete_TPMT_SYM_DEF_PTR_PTR(TPMT_SYM_DEF * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SYM_DEF_PTR_PTR_assign(TPMT_SYM_DEF * *obj, TPMT_SYM_DEF * value) {
    *obj = value;
  }

  static TPMT_SYM_DEF * TPMT_SYM_DEF_PTR_PTR_value(TPMT_SYM_DEF * *obj) {
    return *obj;
  }


typedef TPMT_SYM_DEF_OBJECT TPMT_SYM_DEF_OBJECT_ARRAY;

SWIGINTERN TPMT_SYM_DEF_OBJECT_ARRAY *new_TPMT_SYM_DEF_OBJECT_ARRAY(size_t nelements){
    return (TPMT_SYM_DEF_OBJECT *)malloc((nelements)*sizeof(TPMT_SYM_DEF_OBJECT));
  }
SWIGINTERN void delete_TPMT_SYM_DEF_OBJECT_ARRAY(TPMT_SYM_DEF_OBJECT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_SYM_DEF_OBJECT TPMT_SYM_DEF_OBJECT_ARRAY___getitem__(TPMT_SYM_DEF_OBJECT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_SYM_DEF_OBJECT_ARRAY___setitem__(TPMT_SYM_DEF_OBJECT_ARRAY *self,size_t index,TPMT_SYM_DEF_OBJECT value){
    self[index] = value;
  }
SWIGINTERN TPMT_SYM_DEF_OBJECT *TPMT_SYM_DEF_OBJECT_ARRAY_cast(TPMT_SYM_DEF_OBJECT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_SYM_DEF_OBJECT_ARRAY *TPMT_SYM_DEF_OBJECT_ARRAY_frompointer(TPMT_SYM_DEF_OBJECT *t){
    return (TPMT_SYM_DEF_OBJECT_ARRAY *)(t);
  }

  static TPMT_SYM_DEF_OBJECT *new_TPMT_SYM_DEF_OBJECT_PTR() { 
    return (TPMT_SYM_DEF_OBJECT *)malloc(sizeof(TPMT_SYM_DEF_OBJECT));
  }
  
  static TPMT_SYM_DEF_OBJECT *copy_TPMT_SYM_DEF_OBJECT_PTR(TPMT_SYM_DEF_OBJECT value) { 
    return (TPMT_SYM_DEF_OBJECT *)memcpy((TPMT_SYM_DEF_OBJECT *)malloc(sizeof(TPMT_SYM_DEF_OBJECT)),&value,sizeof(TPMT_SYM_DEF_OBJECT));
  }

  static void delete_TPMT_SYM_DEF_OBJECT_PTR(TPMT_SYM_DEF_OBJECT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SYM_DEF_OBJECT_PTR_assign(TPMT_SYM_DEF_OBJECT *obj, TPMT_SYM_DEF_OBJECT value) {
    *obj = value;
  }

  static TPMT_SYM_DEF_OBJECT TPMT_SYM_DEF_OBJECT_PTR_value(TPMT_SYM_DEF_OBJECT *obj) {
    return *obj;
  }


  static TPMT_SYM_DEF_OBJECT * *new_TPMT_SYM_DEF_OBJECT_PTR_PTR() { 
    return (TPMT_SYM_DEF_OBJECT * *)malloc(sizeof(TPMT_SYM_DEF_OBJECT *));
  }
  
  static TPMT_SYM_DEF_OBJECT * *copy_TPMT_SYM_DEF_OBJECT_PTR_PTR(TPMT_SYM_DEF_OBJECT * value) { 
    return (TPMT_SYM_DEF_OBJECT * *)memcpy((TPMT_SYM_DEF_OBJECT * *)malloc(sizeof(TPMT_SYM_DEF_OBJECT *)),&value,sizeof(TPMT_SYM_DEF_OBJECT *));
  }

  static void delete_TPMT_SYM_DEF_OBJECT_PTR_PTR(TPMT_SYM_DEF_OBJECT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SYM_DEF_OBJECT_PTR_PTR_assign(TPMT_SYM_DEF_OBJECT * *obj, TPMT_SYM_DEF_OBJECT * value) {
    *obj = value;
  }

  static TPMT_SYM_DEF_OBJECT * TPMT_SYM_DEF_OBJECT_PTR_PTR_value(TPMT_SYM_DEF_OBJECT * *obj) {
    return *obj;
  }


typedef TPM2B_SYM_KEY TPM2B_SYM_KEY_ARRAY;

SWIGINTERN TPM2B_SYM_KEY_ARRAY *new_TPM2B_SYM_KEY_ARRAY(size_t nelements){
    return (TPM2B_SYM_KEY *)malloc((nelements)*sizeof(TPM2B_SYM_KEY));
  }
SWIGINTERN void delete_TPM2B_SYM_KEY_ARRAY(TPM2B_SYM_KEY_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_SYM_KEY TPM2B_SYM_KEY_ARRAY___getitem__(TPM2B_SYM_KEY_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_SYM_KEY_ARRAY___setitem__(TPM2B_SYM_KEY_ARRAY *self,size_t index,TPM2B_SYM_KEY value){
    self[index] = value;
  }
SWIGINTERN TPM2B_SYM_KEY *TPM2B_SYM_KEY_ARRAY_cast(TPM2B_SYM_KEY_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_SYM_KEY_ARRAY *TPM2B_SYM_KEY_ARRAY_frompointer(TPM2B_SYM_KEY *t){
    return (TPM2B_SYM_KEY_ARRAY *)(t);
  }

  static TPM2B_SYM_KEY *new_TPM2B_SYM_KEY_PTR() { 
    return (TPM2B_SYM_KEY *)malloc(sizeof(TPM2B_SYM_KEY));
  }
  
  static TPM2B_SYM_KEY *copy_TPM2B_SYM_KEY_PTR(TPM2B_SYM_KEY value) { 
    return (TPM2B_SYM_KEY *)memcpy((TPM2B_SYM_KEY *)malloc(sizeof(TPM2B_SYM_KEY)),&value,sizeof(TPM2B_SYM_KEY));
  }

  static void delete_TPM2B_SYM_KEY_PTR(TPM2B_SYM_KEY *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SYM_KEY_PTR_assign(TPM2B_SYM_KEY *obj, TPM2B_SYM_KEY value) {
    *obj = value;
  }

  static TPM2B_SYM_KEY TPM2B_SYM_KEY_PTR_value(TPM2B_SYM_KEY *obj) {
    return *obj;
  }


  static TPM2B_SYM_KEY * *new_TPM2B_SYM_KEY_PTR_PTR() { 
    return (TPM2B_SYM_KEY * *)malloc(sizeof(TPM2B_SYM_KEY *));
  }
  
  static TPM2B_SYM_KEY * *copy_TPM2B_SYM_KEY_PTR_PTR(TPM2B_SYM_KEY * value) { 
    return (TPM2B_SYM_KEY * *)memcpy((TPM2B_SYM_KEY * *)malloc(sizeof(TPM2B_SYM_KEY *)),&value,sizeof(TPM2B_SYM_KEY *));
  }

  static void delete_TPM2B_SYM_KEY_PTR_PTR(TPM2B_SYM_KEY * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SYM_KEY_PTR_PTR_assign(TPM2B_SYM_KEY * *obj, TPM2B_SYM_KEY * value) {
    *obj = value;
  }

  static TPM2B_SYM_KEY * TPM2B_SYM_KEY_PTR_PTR_value(TPM2B_SYM_KEY * *obj) {
    return *obj;
  }


typedef TPMS_SYMCIPHER_PARMS TPMS_SYMCIPHER_PARMS_ARRAY;

SWIGINTERN TPMS_SYMCIPHER_PARMS_ARRAY *new_TPMS_SYMCIPHER_PARMS_ARRAY(size_t nelements){
    return (TPMS_SYMCIPHER_PARMS *)malloc((nelements)*sizeof(TPMS_SYMCIPHER_PARMS));
  }
SWIGINTERN void delete_TPMS_SYMCIPHER_PARMS_ARRAY(TPMS_SYMCIPHER_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SYMCIPHER_PARMS TPMS_SYMCIPHER_PARMS_ARRAY___getitem__(TPMS_SYMCIPHER_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SYMCIPHER_PARMS_ARRAY___setitem__(TPMS_SYMCIPHER_PARMS_ARRAY *self,size_t index,TPMS_SYMCIPHER_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMS_SYMCIPHER_PARMS *TPMS_SYMCIPHER_PARMS_ARRAY_cast(TPMS_SYMCIPHER_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SYMCIPHER_PARMS_ARRAY *TPMS_SYMCIPHER_PARMS_ARRAY_frompointer(TPMS_SYMCIPHER_PARMS *t){
    return (TPMS_SYMCIPHER_PARMS_ARRAY *)(t);
  }

  static TPMS_SYMCIPHER_PARMS *new_TPMS_SYMCIPHER_PARMS_PTR() { 
    return (TPMS_SYMCIPHER_PARMS *)malloc(sizeof(TPMS_SYMCIPHER_PARMS));
  }
  
  static TPMS_SYMCIPHER_PARMS *copy_TPMS_SYMCIPHER_PARMS_PTR(TPMS_SYMCIPHER_PARMS value) { 
    return (TPMS_SYMCIPHER_PARMS *)memcpy((TPMS_SYMCIPHER_PARMS *)malloc(sizeof(TPMS_SYMCIPHER_PARMS)),&value,sizeof(TPMS_SYMCIPHER_PARMS));
  }

  static void delete_TPMS_SYMCIPHER_PARMS_PTR(TPMS_SYMCIPHER_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SYMCIPHER_PARMS_PTR_assign(TPMS_SYMCIPHER_PARMS *obj, TPMS_SYMCIPHER_PARMS value) {
    *obj = value;
  }

  static TPMS_SYMCIPHER_PARMS TPMS_SYMCIPHER_PARMS_PTR_value(TPMS_SYMCIPHER_PARMS *obj) {
    return *obj;
  }


  static TPMS_SYMCIPHER_PARMS * *new_TPMS_SYMCIPHER_PARMS_PTR_PTR() { 
    return (TPMS_SYMCIPHER_PARMS * *)malloc(sizeof(TPMS_SYMCIPHER_PARMS *));
  }
  
  static TPMS_SYMCIPHER_PARMS * *copy_TPMS_SYMCIPHER_PARMS_PTR_PTR(TPMS_SYMCIPHER_PARMS * value) { 
    return (TPMS_SYMCIPHER_PARMS * *)memcpy((TPMS_SYMCIPHER_PARMS * *)malloc(sizeof(TPMS_SYMCIPHER_PARMS *)),&value,sizeof(TPMS_SYMCIPHER_PARMS *));
  }

  static void delete_TPMS_SYMCIPHER_PARMS_PTR_PTR(TPMS_SYMCIPHER_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SYMCIPHER_PARMS_PTR_PTR_assign(TPMS_SYMCIPHER_PARMS * *obj, TPMS_SYMCIPHER_PARMS * value) {
    *obj = value;
  }

  static TPMS_SYMCIPHER_PARMS * TPMS_SYMCIPHER_PARMS_PTR_PTR_value(TPMS_SYMCIPHER_PARMS * *obj) {
    return *obj;
  }


typedef TPM2B_SENSITIVE_DATA TPM2B_SENSITIVE_DATA_ARRAY;

SWIGINTERN TPM2B_SENSITIVE_DATA_ARRAY *new_TPM2B_SENSITIVE_DATA_ARRAY(size_t nelements){
    return (TPM2B_SENSITIVE_DATA *)malloc((nelements)*sizeof(TPM2B_SENSITIVE_DATA));
  }
SWIGINTERN void delete_TPM2B_SENSITIVE_DATA_ARRAY(TPM2B_SENSITIVE_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_SENSITIVE_DATA TPM2B_SENSITIVE_DATA_ARRAY___getitem__(TPM2B_SENSITIVE_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_SENSITIVE_DATA_ARRAY___setitem__(TPM2B_SENSITIVE_DATA_ARRAY *self,size_t index,TPM2B_SENSITIVE_DATA value){
    self[index] = value;
  }
SWIGINTERN TPM2B_SENSITIVE_DATA *TPM2B_SENSITIVE_DATA_ARRAY_cast(TPM2B_SENSITIVE_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_SENSITIVE_DATA_ARRAY *TPM2B_SENSITIVE_DATA_ARRAY_frompointer(TPM2B_SENSITIVE_DATA *t){
    return (TPM2B_SENSITIVE_DATA_ARRAY *)(t);
  }

  static TPM2B_SENSITIVE_DATA *new_TPM2B_SENSITIVE_DATA_PTR() { 
    return (TPM2B_SENSITIVE_DATA *)malloc(sizeof(TPM2B_SENSITIVE_DATA));
  }
  
  static TPM2B_SENSITIVE_DATA *copy_TPM2B_SENSITIVE_DATA_PTR(TPM2B_SENSITIVE_DATA value) { 
    return (TPM2B_SENSITIVE_DATA *)memcpy((TPM2B_SENSITIVE_DATA *)malloc(sizeof(TPM2B_SENSITIVE_DATA)),&value,sizeof(TPM2B_SENSITIVE_DATA));
  }

  static void delete_TPM2B_SENSITIVE_DATA_PTR(TPM2B_SENSITIVE_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SENSITIVE_DATA_PTR_assign(TPM2B_SENSITIVE_DATA *obj, TPM2B_SENSITIVE_DATA value) {
    *obj = value;
  }

  static TPM2B_SENSITIVE_DATA TPM2B_SENSITIVE_DATA_PTR_value(TPM2B_SENSITIVE_DATA *obj) {
    return *obj;
  }


  static TPM2B_SENSITIVE_DATA * *new_TPM2B_SENSITIVE_DATA_PTR_PTR() { 
    return (TPM2B_SENSITIVE_DATA * *)malloc(sizeof(TPM2B_SENSITIVE_DATA *));
  }
  
  static TPM2B_SENSITIVE_DATA * *copy_TPM2B_SENSITIVE_DATA_PTR_PTR(TPM2B_SENSITIVE_DATA * value) { 
    return (TPM2B_SENSITIVE_DATA * *)memcpy((TPM2B_SENSITIVE_DATA * *)malloc(sizeof(TPM2B_SENSITIVE_DATA *)),&value,sizeof(TPM2B_SENSITIVE_DATA *));
  }

  static void delete_TPM2B_SENSITIVE_DATA_PTR_PTR(TPM2B_SENSITIVE_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SENSITIVE_DATA_PTR_PTR_assign(TPM2B_SENSITIVE_DATA * *obj, TPM2B_SENSITIVE_DATA * value) {
    *obj = value;
  }

  static TPM2B_SENSITIVE_DATA * TPM2B_SENSITIVE_DATA_PTR_PTR_value(TPM2B_SENSITIVE_DATA * *obj) {
    return *obj;
  }


typedef TPMS_SENSITIVE_CREATE TPMS_SENSITIVE_CREATE_ARRAY;

SWIGINTERN TPMS_SENSITIVE_CREATE_ARRAY *new_TPMS_SENSITIVE_CREATE_ARRAY(size_t nelements){
    return (TPMS_SENSITIVE_CREATE *)malloc((nelements)*sizeof(TPMS_SENSITIVE_CREATE));
  }
SWIGINTERN void delete_TPMS_SENSITIVE_CREATE_ARRAY(TPMS_SENSITIVE_CREATE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SENSITIVE_CREATE TPMS_SENSITIVE_CREATE_ARRAY___getitem__(TPMS_SENSITIVE_CREATE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SENSITIVE_CREATE_ARRAY___setitem__(TPMS_SENSITIVE_CREATE_ARRAY *self,size_t index,TPMS_SENSITIVE_CREATE value){
    self[index] = value;
  }
SWIGINTERN TPMS_SENSITIVE_CREATE *TPMS_SENSITIVE_CREATE_ARRAY_cast(TPMS_SENSITIVE_CREATE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SENSITIVE_CREATE_ARRAY *TPMS_SENSITIVE_CREATE_ARRAY_frompointer(TPMS_SENSITIVE_CREATE *t){
    return (TPMS_SENSITIVE_CREATE_ARRAY *)(t);
  }

  static TPMS_SENSITIVE_CREATE *new_TPMS_SENSITIVE_CREATE_PTR() { 
    return (TPMS_SENSITIVE_CREATE *)malloc(sizeof(TPMS_SENSITIVE_CREATE));
  }
  
  static TPMS_SENSITIVE_CREATE *copy_TPMS_SENSITIVE_CREATE_PTR(TPMS_SENSITIVE_CREATE value) { 
    return (TPMS_SENSITIVE_CREATE *)memcpy((TPMS_SENSITIVE_CREATE *)malloc(sizeof(TPMS_SENSITIVE_CREATE)),&value,sizeof(TPMS_SENSITIVE_CREATE));
  }

  static void delete_TPMS_SENSITIVE_CREATE_PTR(TPMS_SENSITIVE_CREATE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SENSITIVE_CREATE_PTR_assign(TPMS_SENSITIVE_CREATE *obj, TPMS_SENSITIVE_CREATE value) {
    *obj = value;
  }

  static TPMS_SENSITIVE_CREATE TPMS_SENSITIVE_CREATE_PTR_value(TPMS_SENSITIVE_CREATE *obj) {
    return *obj;
  }


  static TPMS_SENSITIVE_CREATE * *new_TPMS_SENSITIVE_CREATE_PTR_PTR() { 
    return (TPMS_SENSITIVE_CREATE * *)malloc(sizeof(TPMS_SENSITIVE_CREATE *));
  }
  
  static TPMS_SENSITIVE_CREATE * *copy_TPMS_SENSITIVE_CREATE_PTR_PTR(TPMS_SENSITIVE_CREATE * value) { 
    return (TPMS_SENSITIVE_CREATE * *)memcpy((TPMS_SENSITIVE_CREATE * *)malloc(sizeof(TPMS_SENSITIVE_CREATE *)),&value,sizeof(TPMS_SENSITIVE_CREATE *));
  }

  static void delete_TPMS_SENSITIVE_CREATE_PTR_PTR(TPMS_SENSITIVE_CREATE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SENSITIVE_CREATE_PTR_PTR_assign(TPMS_SENSITIVE_CREATE * *obj, TPMS_SENSITIVE_CREATE * value) {
    *obj = value;
  }

  static TPMS_SENSITIVE_CREATE * TPMS_SENSITIVE_CREATE_PTR_PTR_value(TPMS_SENSITIVE_CREATE * *obj) {
    return *obj;
  }


typedef TPM2B_SENSITIVE_CREATE TPM2B_SENSITIVE_CREATE_ARRAY;

SWIGINTERN TPM2B_SENSITIVE_CREATE_ARRAY *new_TPM2B_SENSITIVE_CREATE_ARRAY(size_t nelements){
    return (TPM2B_SENSITIVE_CREATE *)malloc((nelements)*sizeof(TPM2B_SENSITIVE_CREATE));
  }
SWIGINTERN void delete_TPM2B_SENSITIVE_CREATE_ARRAY(TPM2B_SENSITIVE_CREATE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_SENSITIVE_CREATE TPM2B_SENSITIVE_CREATE_ARRAY___getitem__(TPM2B_SENSITIVE_CREATE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_SENSITIVE_CREATE_ARRAY___setitem__(TPM2B_SENSITIVE_CREATE_ARRAY *self,size_t index,TPM2B_SENSITIVE_CREATE value){
    self[index] = value;
  }
SWIGINTERN TPM2B_SENSITIVE_CREATE *TPM2B_SENSITIVE_CREATE_ARRAY_cast(TPM2B_SENSITIVE_CREATE_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_SENSITIVE_CREATE_ARRAY *TPM2B_SENSITIVE_CREATE_ARRAY_frompointer(TPM2B_SENSITIVE_CREATE *t){
    return (TPM2B_SENSITIVE_CREATE_ARRAY *)(t);
  }

  static TPM2B_SENSITIVE_CREATE *new_TPM2B_SENSITIVE_CREATE_PTR() { 
    return (TPM2B_SENSITIVE_CREATE *)malloc(sizeof(TPM2B_SENSITIVE_CREATE));
  }
  
  static TPM2B_SENSITIVE_CREATE *copy_TPM2B_SENSITIVE_CREATE_PTR(TPM2B_SENSITIVE_CREATE value) { 
    return (TPM2B_SENSITIVE_CREATE *)memcpy((TPM2B_SENSITIVE_CREATE *)malloc(sizeof(TPM2B_SENSITIVE_CREATE)),&value,sizeof(TPM2B_SENSITIVE_CREATE));
  }

  static void delete_TPM2B_SENSITIVE_CREATE_PTR(TPM2B_SENSITIVE_CREATE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SENSITIVE_CREATE_PTR_assign(TPM2B_SENSITIVE_CREATE *obj, TPM2B_SENSITIVE_CREATE value) {
    *obj = value;
  }

  static TPM2B_SENSITIVE_CREATE TPM2B_SENSITIVE_CREATE_PTR_value(TPM2B_SENSITIVE_CREATE *obj) {
    return *obj;
  }


  static TPM2B_SENSITIVE_CREATE * *new_TPM2B_SENSITIVE_CREATE_PTR_PTR() { 
    return (TPM2B_SENSITIVE_CREATE * *)malloc(sizeof(TPM2B_SENSITIVE_CREATE *));
  }
  
  static TPM2B_SENSITIVE_CREATE * *copy_TPM2B_SENSITIVE_CREATE_PTR_PTR(TPM2B_SENSITIVE_CREATE * value) { 
    return (TPM2B_SENSITIVE_CREATE * *)memcpy((TPM2B_SENSITIVE_CREATE * *)malloc(sizeof(TPM2B_SENSITIVE_CREATE *)),&value,sizeof(TPM2B_SENSITIVE_CREATE *));
  }

  static void delete_TPM2B_SENSITIVE_CREATE_PTR_PTR(TPM2B_SENSITIVE_CREATE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SENSITIVE_CREATE_PTR_PTR_assign(TPM2B_SENSITIVE_CREATE * *obj, TPM2B_SENSITIVE_CREATE * value) {
    *obj = value;
  }

  static TPM2B_SENSITIVE_CREATE * TPM2B_SENSITIVE_CREATE_PTR_PTR_value(TPM2B_SENSITIVE_CREATE * *obj) {
    return *obj;
  }


typedef TPMS_SCHEME_HASH TPMS_SCHEME_HASH_ARRAY;

SWIGINTERN TPMS_SCHEME_HASH_ARRAY *new_TPMS_SCHEME_HASH_ARRAY(size_t nelements){
    return (TPMS_SCHEME_HASH *)malloc((nelements)*sizeof(TPMS_SCHEME_HASH));
  }
SWIGINTERN void delete_TPMS_SCHEME_HASH_ARRAY(TPMS_SCHEME_HASH_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SCHEME_HASH TPMS_SCHEME_HASH_ARRAY___getitem__(TPMS_SCHEME_HASH_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SCHEME_HASH_ARRAY___setitem__(TPMS_SCHEME_HASH_ARRAY *self,size_t index,TPMS_SCHEME_HASH value){
    self[index] = value;
  }
SWIGINTERN TPMS_SCHEME_HASH *TPMS_SCHEME_HASH_ARRAY_cast(TPMS_SCHEME_HASH_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SCHEME_HASH_ARRAY *TPMS_SCHEME_HASH_ARRAY_frompointer(TPMS_SCHEME_HASH *t){
    return (TPMS_SCHEME_HASH_ARRAY *)(t);
  }

  static TPMS_SCHEME_HASH *new_TPMS_SCHEME_HASH_PTR() { 
    return (TPMS_SCHEME_HASH *)malloc(sizeof(TPMS_SCHEME_HASH));
  }
  
  static TPMS_SCHEME_HASH *copy_TPMS_SCHEME_HASH_PTR(TPMS_SCHEME_HASH value) { 
    return (TPMS_SCHEME_HASH *)memcpy((TPMS_SCHEME_HASH *)malloc(sizeof(TPMS_SCHEME_HASH)),&value,sizeof(TPMS_SCHEME_HASH));
  }

  static void delete_TPMS_SCHEME_HASH_PTR(TPMS_SCHEME_HASH *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SCHEME_HASH_PTR_assign(TPMS_SCHEME_HASH *obj, TPMS_SCHEME_HASH value) {
    *obj = value;
  }

  static TPMS_SCHEME_HASH TPMS_SCHEME_HASH_PTR_value(TPMS_SCHEME_HASH *obj) {
    return *obj;
  }


  static TPMS_SCHEME_HASH * *new_TPMS_SCHEME_HASH_PTR_PTR() { 
    return (TPMS_SCHEME_HASH * *)malloc(sizeof(TPMS_SCHEME_HASH *));
  }
  
  static TPMS_SCHEME_HASH * *copy_TPMS_SCHEME_HASH_PTR_PTR(TPMS_SCHEME_HASH * value) { 
    return (TPMS_SCHEME_HASH * *)memcpy((TPMS_SCHEME_HASH * *)malloc(sizeof(TPMS_SCHEME_HASH *)),&value,sizeof(TPMS_SCHEME_HASH *));
  }

  static void delete_TPMS_SCHEME_HASH_PTR_PTR(TPMS_SCHEME_HASH * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SCHEME_HASH_PTR_PTR_assign(TPMS_SCHEME_HASH * *obj, TPMS_SCHEME_HASH * value) {
    *obj = value;
  }

  static TPMS_SCHEME_HASH * TPMS_SCHEME_HASH_PTR_PTR_value(TPMS_SCHEME_HASH * *obj) {
    return *obj;
  }


typedef TPMS_SCHEME_ECDAA TPMS_SCHEME_ECDAA_ARRAY;

SWIGINTERN TPMS_SCHEME_ECDAA_ARRAY *new_TPMS_SCHEME_ECDAA_ARRAY(size_t nelements){
    return (TPMS_SCHEME_ECDAA *)malloc((nelements)*sizeof(TPMS_SCHEME_ECDAA));
  }
SWIGINTERN void delete_TPMS_SCHEME_ECDAA_ARRAY(TPMS_SCHEME_ECDAA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SCHEME_ECDAA TPMS_SCHEME_ECDAA_ARRAY___getitem__(TPMS_SCHEME_ECDAA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SCHEME_ECDAA_ARRAY___setitem__(TPMS_SCHEME_ECDAA_ARRAY *self,size_t index,TPMS_SCHEME_ECDAA value){
    self[index] = value;
  }
SWIGINTERN TPMS_SCHEME_ECDAA *TPMS_SCHEME_ECDAA_ARRAY_cast(TPMS_SCHEME_ECDAA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SCHEME_ECDAA_ARRAY *TPMS_SCHEME_ECDAA_ARRAY_frompointer(TPMS_SCHEME_ECDAA *t){
    return (TPMS_SCHEME_ECDAA_ARRAY *)(t);
  }

  static TPMS_SCHEME_ECDAA *new_TPMS_SCHEME_ECDAA_PTR() { 
    return (TPMS_SCHEME_ECDAA *)malloc(sizeof(TPMS_SCHEME_ECDAA));
  }
  
  static TPMS_SCHEME_ECDAA *copy_TPMS_SCHEME_ECDAA_PTR(TPMS_SCHEME_ECDAA value) { 
    return (TPMS_SCHEME_ECDAA *)memcpy((TPMS_SCHEME_ECDAA *)malloc(sizeof(TPMS_SCHEME_ECDAA)),&value,sizeof(TPMS_SCHEME_ECDAA));
  }

  static void delete_TPMS_SCHEME_ECDAA_PTR(TPMS_SCHEME_ECDAA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SCHEME_ECDAA_PTR_assign(TPMS_SCHEME_ECDAA *obj, TPMS_SCHEME_ECDAA value) {
    *obj = value;
  }

  static TPMS_SCHEME_ECDAA TPMS_SCHEME_ECDAA_PTR_value(TPMS_SCHEME_ECDAA *obj) {
    return *obj;
  }


  static TPMS_SCHEME_ECDAA * *new_TPMS_SCHEME_ECDAA_PTR_PTR() { 
    return (TPMS_SCHEME_ECDAA * *)malloc(sizeof(TPMS_SCHEME_ECDAA *));
  }
  
  static TPMS_SCHEME_ECDAA * *copy_TPMS_SCHEME_ECDAA_PTR_PTR(TPMS_SCHEME_ECDAA * value) { 
    return (TPMS_SCHEME_ECDAA * *)memcpy((TPMS_SCHEME_ECDAA * *)malloc(sizeof(TPMS_SCHEME_ECDAA *)),&value,sizeof(TPMS_SCHEME_ECDAA *));
  }

  static void delete_TPMS_SCHEME_ECDAA_PTR_PTR(TPMS_SCHEME_ECDAA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SCHEME_ECDAA_PTR_PTR_assign(TPMS_SCHEME_ECDAA * *obj, TPMS_SCHEME_ECDAA * value) {
    *obj = value;
  }

  static TPMS_SCHEME_ECDAA * TPMS_SCHEME_ECDAA_PTR_PTR_value(TPMS_SCHEME_ECDAA * *obj) {
    return *obj;
  }


typedef TPMS_SCHEME_XOR TPMS_SCHEME_XOR_ARRAY;

SWIGINTERN TPMS_SCHEME_XOR_ARRAY *new_TPMS_SCHEME_XOR_ARRAY(size_t nelements){
    return (TPMS_SCHEME_XOR *)malloc((nelements)*sizeof(TPMS_SCHEME_XOR));
  }
SWIGINTERN void delete_TPMS_SCHEME_XOR_ARRAY(TPMS_SCHEME_XOR_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SCHEME_XOR TPMS_SCHEME_XOR_ARRAY___getitem__(TPMS_SCHEME_XOR_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SCHEME_XOR_ARRAY___setitem__(TPMS_SCHEME_XOR_ARRAY *self,size_t index,TPMS_SCHEME_XOR value){
    self[index] = value;
  }
SWIGINTERN TPMS_SCHEME_XOR *TPMS_SCHEME_XOR_ARRAY_cast(TPMS_SCHEME_XOR_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SCHEME_XOR_ARRAY *TPMS_SCHEME_XOR_ARRAY_frompointer(TPMS_SCHEME_XOR *t){
    return (TPMS_SCHEME_XOR_ARRAY *)(t);
  }

  static TPMS_SCHEME_XOR *new_TPMS_SCHEME_XOR_PTR() { 
    return (TPMS_SCHEME_XOR *)malloc(sizeof(TPMS_SCHEME_XOR));
  }
  
  static TPMS_SCHEME_XOR *copy_TPMS_SCHEME_XOR_PTR(TPMS_SCHEME_XOR value) { 
    return (TPMS_SCHEME_XOR *)memcpy((TPMS_SCHEME_XOR *)malloc(sizeof(TPMS_SCHEME_XOR)),&value,sizeof(TPMS_SCHEME_XOR));
  }

  static void delete_TPMS_SCHEME_XOR_PTR(TPMS_SCHEME_XOR *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SCHEME_XOR_PTR_assign(TPMS_SCHEME_XOR *obj, TPMS_SCHEME_XOR value) {
    *obj = value;
  }

  static TPMS_SCHEME_XOR TPMS_SCHEME_XOR_PTR_value(TPMS_SCHEME_XOR *obj) {
    return *obj;
  }


  static TPMS_SCHEME_XOR * *new_TPMS_SCHEME_XOR_PTR_PTR() { 
    return (TPMS_SCHEME_XOR * *)malloc(sizeof(TPMS_SCHEME_XOR *));
  }
  
  static TPMS_SCHEME_XOR * *copy_TPMS_SCHEME_XOR_PTR_PTR(TPMS_SCHEME_XOR * value) { 
    return (TPMS_SCHEME_XOR * *)memcpy((TPMS_SCHEME_XOR * *)malloc(sizeof(TPMS_SCHEME_XOR *)),&value,sizeof(TPMS_SCHEME_XOR *));
  }

  static void delete_TPMS_SCHEME_XOR_PTR_PTR(TPMS_SCHEME_XOR * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SCHEME_XOR_PTR_PTR_assign(TPMS_SCHEME_XOR * *obj, TPMS_SCHEME_XOR * value) {
    *obj = value;
  }

  static TPMS_SCHEME_XOR * TPMS_SCHEME_XOR_PTR_PTR_value(TPMS_SCHEME_XOR * *obj) {
    return *obj;
  }


typedef TPMU_SCHEME_KEYEDHASH TPMU_SCHEME_KEYEDHASH_ARRAY;

SWIGINTERN TPMU_SCHEME_KEYEDHASH_ARRAY *new_TPMU_SCHEME_KEYEDHASH_ARRAY(size_t nelements){
    return (TPMU_SCHEME_KEYEDHASH *)malloc((nelements)*sizeof(TPMU_SCHEME_KEYEDHASH));
  }
SWIGINTERN void delete_TPMU_SCHEME_KEYEDHASH_ARRAY(TPMU_SCHEME_KEYEDHASH_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_SCHEME_KEYEDHASH TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__(TPMU_SCHEME_KEYEDHASH_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__(TPMU_SCHEME_KEYEDHASH_ARRAY *self,size_t index,TPMU_SCHEME_KEYEDHASH value){
    self[index] = value;
  }
SWIGINTERN TPMU_SCHEME_KEYEDHASH *TPMU_SCHEME_KEYEDHASH_ARRAY_cast(TPMU_SCHEME_KEYEDHASH_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_SCHEME_KEYEDHASH_ARRAY *TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer(TPMU_SCHEME_KEYEDHASH *t){
    return (TPMU_SCHEME_KEYEDHASH_ARRAY *)(t);
  }

  static TPMU_SCHEME_KEYEDHASH *new_TPMU_SCHEME_KEYEDHASH_PTR() { 
    return (TPMU_SCHEME_KEYEDHASH *)malloc(sizeof(TPMU_SCHEME_KEYEDHASH));
  }
  
  static TPMU_SCHEME_KEYEDHASH *copy_TPMU_SCHEME_KEYEDHASH_PTR(TPMU_SCHEME_KEYEDHASH value) { 
    return (TPMU_SCHEME_KEYEDHASH *)memcpy((TPMU_SCHEME_KEYEDHASH *)malloc(sizeof(TPMU_SCHEME_KEYEDHASH)),&value,sizeof(TPMU_SCHEME_KEYEDHASH));
  }

  static void delete_TPMU_SCHEME_KEYEDHASH_PTR(TPMU_SCHEME_KEYEDHASH *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SCHEME_KEYEDHASH_PTR_assign(TPMU_SCHEME_KEYEDHASH *obj, TPMU_SCHEME_KEYEDHASH value) {
    *obj = value;
  }

  static TPMU_SCHEME_KEYEDHASH TPMU_SCHEME_KEYEDHASH_PTR_value(TPMU_SCHEME_KEYEDHASH *obj) {
    return *obj;
  }


  static TPMU_SCHEME_KEYEDHASH * *new_TPMU_SCHEME_KEYEDHASH_PTR_PTR() { 
    return (TPMU_SCHEME_KEYEDHASH * *)malloc(sizeof(TPMU_SCHEME_KEYEDHASH *));
  }
  
  static TPMU_SCHEME_KEYEDHASH * *copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR(TPMU_SCHEME_KEYEDHASH * value) { 
    return (TPMU_SCHEME_KEYEDHASH * *)memcpy((TPMU_SCHEME_KEYEDHASH * *)malloc(sizeof(TPMU_SCHEME_KEYEDHASH *)),&value,sizeof(TPMU_SCHEME_KEYEDHASH *));
  }

  static void delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR(TPMU_SCHEME_KEYEDHASH * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign(TPMU_SCHEME_KEYEDHASH * *obj, TPMU_SCHEME_KEYEDHASH * value) {
    *obj = value;
  }

  static TPMU_SCHEME_KEYEDHASH * TPMU_SCHEME_KEYEDHASH_PTR_PTR_value(TPMU_SCHEME_KEYEDHASH * *obj) {
    return *obj;
  }


typedef TPMT_KEYEDHASH_SCHEME TPMT_KEYEDHASH_SCHEME_ARRAY;

SWIGINTERN TPMT_KEYEDHASH_SCHEME_ARRAY *new_TPMT_KEYEDHASH_SCHEME_ARRAY(size_t nelements){
    return (TPMT_KEYEDHASH_SCHEME *)malloc((nelements)*sizeof(TPMT_KEYEDHASH_SCHEME));
  }
SWIGINTERN void delete_TPMT_KEYEDHASH_SCHEME_ARRAY(TPMT_KEYEDHASH_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_KEYEDHASH_SCHEME TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__(TPMT_KEYEDHASH_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__(TPMT_KEYEDHASH_SCHEME_ARRAY *self,size_t index,TPMT_KEYEDHASH_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMT_KEYEDHASH_SCHEME *TPMT_KEYEDHASH_SCHEME_ARRAY_cast(TPMT_KEYEDHASH_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_KEYEDHASH_SCHEME_ARRAY *TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer(TPMT_KEYEDHASH_SCHEME *t){
    return (TPMT_KEYEDHASH_SCHEME_ARRAY *)(t);
  }

  static TPMT_KEYEDHASH_SCHEME *new_TPMT_KEYEDHASH_SCHEME_PTR() { 
    return (TPMT_KEYEDHASH_SCHEME *)malloc(sizeof(TPMT_KEYEDHASH_SCHEME));
  }
  
  static TPMT_KEYEDHASH_SCHEME *copy_TPMT_KEYEDHASH_SCHEME_PTR(TPMT_KEYEDHASH_SCHEME value) { 
    return (TPMT_KEYEDHASH_SCHEME *)memcpy((TPMT_KEYEDHASH_SCHEME *)malloc(sizeof(TPMT_KEYEDHASH_SCHEME)),&value,sizeof(TPMT_KEYEDHASH_SCHEME));
  }

  static void delete_TPMT_KEYEDHASH_SCHEME_PTR(TPMT_KEYEDHASH_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_KEYEDHASH_SCHEME_PTR_assign(TPMT_KEYEDHASH_SCHEME *obj, TPMT_KEYEDHASH_SCHEME value) {
    *obj = value;
  }

  static TPMT_KEYEDHASH_SCHEME TPMT_KEYEDHASH_SCHEME_PTR_value(TPMT_KEYEDHASH_SCHEME *obj) {
    return *obj;
  }


  static TPMT_KEYEDHASH_SCHEME * *new_TPMT_KEYEDHASH_SCHEME_PTR_PTR() { 
    return (TPMT_KEYEDHASH_SCHEME * *)malloc(sizeof(TPMT_KEYEDHASH_SCHEME *));
  }
  
  static TPMT_KEYEDHASH_SCHEME * *copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR(TPMT_KEYEDHASH_SCHEME * value) { 
    return (TPMT_KEYEDHASH_SCHEME * *)memcpy((TPMT_KEYEDHASH_SCHEME * *)malloc(sizeof(TPMT_KEYEDHASH_SCHEME *)),&value,sizeof(TPMT_KEYEDHASH_SCHEME *));
  }

  static void delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR(TPMT_KEYEDHASH_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign(TPMT_KEYEDHASH_SCHEME * *obj, TPMT_KEYEDHASH_SCHEME * value) {
    *obj = value;
  }

  static TPMT_KEYEDHASH_SCHEME * TPMT_KEYEDHASH_SCHEME_PTR_PTR_value(TPMT_KEYEDHASH_SCHEME * *obj) {
    return *obj;
  }


typedef TPMU_SIG_SCHEME TPMU_SIG_SCHEME_ARRAY;

SWIGINTERN TPMU_SIG_SCHEME_ARRAY *new_TPMU_SIG_SCHEME_ARRAY(size_t nelements){
    return (TPMU_SIG_SCHEME *)malloc((nelements)*sizeof(TPMU_SIG_SCHEME));
  }
SWIGINTERN void delete_TPMU_SIG_SCHEME_ARRAY(TPMU_SIG_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_SIG_SCHEME TPMU_SIG_SCHEME_ARRAY___getitem__(TPMU_SIG_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_SIG_SCHEME_ARRAY___setitem__(TPMU_SIG_SCHEME_ARRAY *self,size_t index,TPMU_SIG_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMU_SIG_SCHEME *TPMU_SIG_SCHEME_ARRAY_cast(TPMU_SIG_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_SIG_SCHEME_ARRAY *TPMU_SIG_SCHEME_ARRAY_frompointer(TPMU_SIG_SCHEME *t){
    return (TPMU_SIG_SCHEME_ARRAY *)(t);
  }

  static TPMU_SIG_SCHEME *new_TPMU_SIG_SCHEME_PTR() { 
    return (TPMU_SIG_SCHEME *)malloc(sizeof(TPMU_SIG_SCHEME));
  }
  
  static TPMU_SIG_SCHEME *copy_TPMU_SIG_SCHEME_PTR(TPMU_SIG_SCHEME value) { 
    return (TPMU_SIG_SCHEME *)memcpy((TPMU_SIG_SCHEME *)malloc(sizeof(TPMU_SIG_SCHEME)),&value,sizeof(TPMU_SIG_SCHEME));
  }

  static void delete_TPMU_SIG_SCHEME_PTR(TPMU_SIG_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SIG_SCHEME_PTR_assign(TPMU_SIG_SCHEME *obj, TPMU_SIG_SCHEME value) {
    *obj = value;
  }

  static TPMU_SIG_SCHEME TPMU_SIG_SCHEME_PTR_value(TPMU_SIG_SCHEME *obj) {
    return *obj;
  }


  static TPMU_SIG_SCHEME * *new_TPMU_SIG_SCHEME_PTR_PTR() { 
    return (TPMU_SIG_SCHEME * *)malloc(sizeof(TPMU_SIG_SCHEME *));
  }
  
  static TPMU_SIG_SCHEME * *copy_TPMU_SIG_SCHEME_PTR_PTR(TPMU_SIG_SCHEME * value) { 
    return (TPMU_SIG_SCHEME * *)memcpy((TPMU_SIG_SCHEME * *)malloc(sizeof(TPMU_SIG_SCHEME *)),&value,sizeof(TPMU_SIG_SCHEME *));
  }

  static void delete_TPMU_SIG_SCHEME_PTR_PTR(TPMU_SIG_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SIG_SCHEME_PTR_PTR_assign(TPMU_SIG_SCHEME * *obj, TPMU_SIG_SCHEME * value) {
    *obj = value;
  }

  static TPMU_SIG_SCHEME * TPMU_SIG_SCHEME_PTR_PTR_value(TPMU_SIG_SCHEME * *obj) {
    return *obj;
  }


typedef TPMT_SIG_SCHEME TPMT_SIG_SCHEME_ARRAY;

SWIGINTERN TPMT_SIG_SCHEME_ARRAY *new_TPMT_SIG_SCHEME_ARRAY(size_t nelements){
    return (TPMT_SIG_SCHEME *)malloc((nelements)*sizeof(TPMT_SIG_SCHEME));
  }
SWIGINTERN void delete_TPMT_SIG_SCHEME_ARRAY(TPMT_SIG_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_SIG_SCHEME TPMT_SIG_SCHEME_ARRAY___getitem__(TPMT_SIG_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_SIG_SCHEME_ARRAY___setitem__(TPMT_SIG_SCHEME_ARRAY *self,size_t index,TPMT_SIG_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMT_SIG_SCHEME *TPMT_SIG_SCHEME_ARRAY_cast(TPMT_SIG_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_SIG_SCHEME_ARRAY *TPMT_SIG_SCHEME_ARRAY_frompointer(TPMT_SIG_SCHEME *t){
    return (TPMT_SIG_SCHEME_ARRAY *)(t);
  }

  static TPMT_SIG_SCHEME *new_TPMT_SIG_SCHEME_PTR() { 
    return (TPMT_SIG_SCHEME *)malloc(sizeof(TPMT_SIG_SCHEME));
  }
  
  static TPMT_SIG_SCHEME *copy_TPMT_SIG_SCHEME_PTR(TPMT_SIG_SCHEME value) { 
    return (TPMT_SIG_SCHEME *)memcpy((TPMT_SIG_SCHEME *)malloc(sizeof(TPMT_SIG_SCHEME)),&value,sizeof(TPMT_SIG_SCHEME));
  }

  static void delete_TPMT_SIG_SCHEME_PTR(TPMT_SIG_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SIG_SCHEME_PTR_assign(TPMT_SIG_SCHEME *obj, TPMT_SIG_SCHEME value) {
    *obj = value;
  }

  static TPMT_SIG_SCHEME TPMT_SIG_SCHEME_PTR_value(TPMT_SIG_SCHEME *obj) {
    return *obj;
  }


  static TPMT_SIG_SCHEME * *new_TPMT_SIG_SCHEME_PTR_PTR() { 
    return (TPMT_SIG_SCHEME * *)malloc(sizeof(TPMT_SIG_SCHEME *));
  }
  
  static TPMT_SIG_SCHEME * *copy_TPMT_SIG_SCHEME_PTR_PTR(TPMT_SIG_SCHEME * value) { 
    return (TPMT_SIG_SCHEME * *)memcpy((TPMT_SIG_SCHEME * *)malloc(sizeof(TPMT_SIG_SCHEME *)),&value,sizeof(TPMT_SIG_SCHEME *));
  }

  static void delete_TPMT_SIG_SCHEME_PTR_PTR(TPMT_SIG_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SIG_SCHEME_PTR_PTR_assign(TPMT_SIG_SCHEME * *obj, TPMT_SIG_SCHEME * value) {
    *obj = value;
  }

  static TPMT_SIG_SCHEME * TPMT_SIG_SCHEME_PTR_PTR_value(TPMT_SIG_SCHEME * *obj) {
    return *obj;
  }


typedef TPMU_KDF_SCHEME TPMU_KDF_SCHEME_ARRAY;

SWIGINTERN TPMU_KDF_SCHEME_ARRAY *new_TPMU_KDF_SCHEME_ARRAY(size_t nelements){
    return (TPMU_KDF_SCHEME *)malloc((nelements)*sizeof(TPMU_KDF_SCHEME));
  }
SWIGINTERN void delete_TPMU_KDF_SCHEME_ARRAY(TPMU_KDF_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_KDF_SCHEME TPMU_KDF_SCHEME_ARRAY___getitem__(TPMU_KDF_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_KDF_SCHEME_ARRAY___setitem__(TPMU_KDF_SCHEME_ARRAY *self,size_t index,TPMU_KDF_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMU_KDF_SCHEME *TPMU_KDF_SCHEME_ARRAY_cast(TPMU_KDF_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_KDF_SCHEME_ARRAY *TPMU_KDF_SCHEME_ARRAY_frompointer(TPMU_KDF_SCHEME *t){
    return (TPMU_KDF_SCHEME_ARRAY *)(t);
  }

  static TPMU_KDF_SCHEME *new_TPMU_KDF_SCHEME_PTR() { 
    return (TPMU_KDF_SCHEME *)malloc(sizeof(TPMU_KDF_SCHEME));
  }
  
  static TPMU_KDF_SCHEME *copy_TPMU_KDF_SCHEME_PTR(TPMU_KDF_SCHEME value) { 
    return (TPMU_KDF_SCHEME *)memcpy((TPMU_KDF_SCHEME *)malloc(sizeof(TPMU_KDF_SCHEME)),&value,sizeof(TPMU_KDF_SCHEME));
  }

  static void delete_TPMU_KDF_SCHEME_PTR(TPMU_KDF_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_KDF_SCHEME_PTR_assign(TPMU_KDF_SCHEME *obj, TPMU_KDF_SCHEME value) {
    *obj = value;
  }

  static TPMU_KDF_SCHEME TPMU_KDF_SCHEME_PTR_value(TPMU_KDF_SCHEME *obj) {
    return *obj;
  }


  static TPMU_KDF_SCHEME * *new_TPMU_KDF_SCHEME_PTR_PTR() { 
    return (TPMU_KDF_SCHEME * *)malloc(sizeof(TPMU_KDF_SCHEME *));
  }
  
  static TPMU_KDF_SCHEME * *copy_TPMU_KDF_SCHEME_PTR_PTR(TPMU_KDF_SCHEME * value) { 
    return (TPMU_KDF_SCHEME * *)memcpy((TPMU_KDF_SCHEME * *)malloc(sizeof(TPMU_KDF_SCHEME *)),&value,sizeof(TPMU_KDF_SCHEME *));
  }

  static void delete_TPMU_KDF_SCHEME_PTR_PTR(TPMU_KDF_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_KDF_SCHEME_PTR_PTR_assign(TPMU_KDF_SCHEME * *obj, TPMU_KDF_SCHEME * value) {
    *obj = value;
  }

  static TPMU_KDF_SCHEME * TPMU_KDF_SCHEME_PTR_PTR_value(TPMU_KDF_SCHEME * *obj) {
    return *obj;
  }


typedef TPMT_KDF_SCHEME TPMT_KDF_SCHEME_ARRAY;

SWIGINTERN TPMT_KDF_SCHEME_ARRAY *new_TPMT_KDF_SCHEME_ARRAY(size_t nelements){
    return (TPMT_KDF_SCHEME *)malloc((nelements)*sizeof(TPMT_KDF_SCHEME));
  }
SWIGINTERN void delete_TPMT_KDF_SCHEME_ARRAY(TPMT_KDF_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_KDF_SCHEME TPMT_KDF_SCHEME_ARRAY___getitem__(TPMT_KDF_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_KDF_SCHEME_ARRAY___setitem__(TPMT_KDF_SCHEME_ARRAY *self,size_t index,TPMT_KDF_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMT_KDF_SCHEME *TPMT_KDF_SCHEME_ARRAY_cast(TPMT_KDF_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_KDF_SCHEME_ARRAY *TPMT_KDF_SCHEME_ARRAY_frompointer(TPMT_KDF_SCHEME *t){
    return (TPMT_KDF_SCHEME_ARRAY *)(t);
  }

  static TPMT_KDF_SCHEME *new_TPMT_KDF_SCHEME_PTR() { 
    return (TPMT_KDF_SCHEME *)malloc(sizeof(TPMT_KDF_SCHEME));
  }
  
  static TPMT_KDF_SCHEME *copy_TPMT_KDF_SCHEME_PTR(TPMT_KDF_SCHEME value) { 
    return (TPMT_KDF_SCHEME *)memcpy((TPMT_KDF_SCHEME *)malloc(sizeof(TPMT_KDF_SCHEME)),&value,sizeof(TPMT_KDF_SCHEME));
  }

  static void delete_TPMT_KDF_SCHEME_PTR(TPMT_KDF_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_KDF_SCHEME_PTR_assign(TPMT_KDF_SCHEME *obj, TPMT_KDF_SCHEME value) {
    *obj = value;
  }

  static TPMT_KDF_SCHEME TPMT_KDF_SCHEME_PTR_value(TPMT_KDF_SCHEME *obj) {
    return *obj;
  }


  static TPMT_KDF_SCHEME * *new_TPMT_KDF_SCHEME_PTR_PTR() { 
    return (TPMT_KDF_SCHEME * *)malloc(sizeof(TPMT_KDF_SCHEME *));
  }
  
  static TPMT_KDF_SCHEME * *copy_TPMT_KDF_SCHEME_PTR_PTR(TPMT_KDF_SCHEME * value) { 
    return (TPMT_KDF_SCHEME * *)memcpy((TPMT_KDF_SCHEME * *)malloc(sizeof(TPMT_KDF_SCHEME *)),&value,sizeof(TPMT_KDF_SCHEME *));
  }

  static void delete_TPMT_KDF_SCHEME_PTR_PTR(TPMT_KDF_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_KDF_SCHEME_PTR_PTR_assign(TPMT_KDF_SCHEME * *obj, TPMT_KDF_SCHEME * value) {
    *obj = value;
  }

  static TPMT_KDF_SCHEME * TPMT_KDF_SCHEME_PTR_PTR_value(TPMT_KDF_SCHEME * *obj) {
    return *obj;
  }


typedef TPMU_ASYM_SCHEME TPMU_ASYM_SCHEME_ARRAY;

SWIGINTERN TPMU_ASYM_SCHEME_ARRAY *new_TPMU_ASYM_SCHEME_ARRAY(size_t nelements){
    return (TPMU_ASYM_SCHEME *)malloc((nelements)*sizeof(TPMU_ASYM_SCHEME));
  }
SWIGINTERN void delete_TPMU_ASYM_SCHEME_ARRAY(TPMU_ASYM_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_ASYM_SCHEME TPMU_ASYM_SCHEME_ARRAY___getitem__(TPMU_ASYM_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_ASYM_SCHEME_ARRAY___setitem__(TPMU_ASYM_SCHEME_ARRAY *self,size_t index,TPMU_ASYM_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMU_ASYM_SCHEME *TPMU_ASYM_SCHEME_ARRAY_cast(TPMU_ASYM_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_ASYM_SCHEME_ARRAY *TPMU_ASYM_SCHEME_ARRAY_frompointer(TPMU_ASYM_SCHEME *t){
    return (TPMU_ASYM_SCHEME_ARRAY *)(t);
  }

  static TPMU_ASYM_SCHEME *new_TPMU_ASYM_SCHEME_PTR() { 
    return (TPMU_ASYM_SCHEME *)malloc(sizeof(TPMU_ASYM_SCHEME));
  }
  
  static TPMU_ASYM_SCHEME *copy_TPMU_ASYM_SCHEME_PTR(TPMU_ASYM_SCHEME value) { 
    return (TPMU_ASYM_SCHEME *)memcpy((TPMU_ASYM_SCHEME *)malloc(sizeof(TPMU_ASYM_SCHEME)),&value,sizeof(TPMU_ASYM_SCHEME));
  }

  static void delete_TPMU_ASYM_SCHEME_PTR(TPMU_ASYM_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_ASYM_SCHEME_PTR_assign(TPMU_ASYM_SCHEME *obj, TPMU_ASYM_SCHEME value) {
    *obj = value;
  }

  static TPMU_ASYM_SCHEME TPMU_ASYM_SCHEME_PTR_value(TPMU_ASYM_SCHEME *obj) {
    return *obj;
  }


  static TPMU_ASYM_SCHEME * *new_TPMU_ASYM_SCHEME_PTR_PTR() { 
    return (TPMU_ASYM_SCHEME * *)malloc(sizeof(TPMU_ASYM_SCHEME *));
  }
  
  static TPMU_ASYM_SCHEME * *copy_TPMU_ASYM_SCHEME_PTR_PTR(TPMU_ASYM_SCHEME * value) { 
    return (TPMU_ASYM_SCHEME * *)memcpy((TPMU_ASYM_SCHEME * *)malloc(sizeof(TPMU_ASYM_SCHEME *)),&value,sizeof(TPMU_ASYM_SCHEME *));
  }

  static void delete_TPMU_ASYM_SCHEME_PTR_PTR(TPMU_ASYM_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_ASYM_SCHEME_PTR_PTR_assign(TPMU_ASYM_SCHEME * *obj, TPMU_ASYM_SCHEME * value) {
    *obj = value;
  }

  static TPMU_ASYM_SCHEME * TPMU_ASYM_SCHEME_PTR_PTR_value(TPMU_ASYM_SCHEME * *obj) {
    return *obj;
  }


typedef TPMT_ASYM_SCHEME TPMT_ASYM_SCHEME_ARRAY;

SWIGINTERN TPMT_ASYM_SCHEME_ARRAY *new_TPMT_ASYM_SCHEME_ARRAY(size_t nelements){
    return (TPMT_ASYM_SCHEME *)malloc((nelements)*sizeof(TPMT_ASYM_SCHEME));
  }
SWIGINTERN void delete_TPMT_ASYM_SCHEME_ARRAY(TPMT_ASYM_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_ASYM_SCHEME TPMT_ASYM_SCHEME_ARRAY___getitem__(TPMT_ASYM_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_ASYM_SCHEME_ARRAY___setitem__(TPMT_ASYM_SCHEME_ARRAY *self,size_t index,TPMT_ASYM_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMT_ASYM_SCHEME *TPMT_ASYM_SCHEME_ARRAY_cast(TPMT_ASYM_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_ASYM_SCHEME_ARRAY *TPMT_ASYM_SCHEME_ARRAY_frompointer(TPMT_ASYM_SCHEME *t){
    return (TPMT_ASYM_SCHEME_ARRAY *)(t);
  }

  static TPMT_ASYM_SCHEME *new_TPMT_ASYM_SCHEME_PTR() { 
    return (TPMT_ASYM_SCHEME *)malloc(sizeof(TPMT_ASYM_SCHEME));
  }
  
  static TPMT_ASYM_SCHEME *copy_TPMT_ASYM_SCHEME_PTR(TPMT_ASYM_SCHEME value) { 
    return (TPMT_ASYM_SCHEME *)memcpy((TPMT_ASYM_SCHEME *)malloc(sizeof(TPMT_ASYM_SCHEME)),&value,sizeof(TPMT_ASYM_SCHEME));
  }

  static void delete_TPMT_ASYM_SCHEME_PTR(TPMT_ASYM_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_ASYM_SCHEME_PTR_assign(TPMT_ASYM_SCHEME *obj, TPMT_ASYM_SCHEME value) {
    *obj = value;
  }

  static TPMT_ASYM_SCHEME TPMT_ASYM_SCHEME_PTR_value(TPMT_ASYM_SCHEME *obj) {
    return *obj;
  }


  static TPMT_ASYM_SCHEME * *new_TPMT_ASYM_SCHEME_PTR_PTR() { 
    return (TPMT_ASYM_SCHEME * *)malloc(sizeof(TPMT_ASYM_SCHEME *));
  }
  
  static TPMT_ASYM_SCHEME * *copy_TPMT_ASYM_SCHEME_PTR_PTR(TPMT_ASYM_SCHEME * value) { 
    return (TPMT_ASYM_SCHEME * *)memcpy((TPMT_ASYM_SCHEME * *)malloc(sizeof(TPMT_ASYM_SCHEME *)),&value,sizeof(TPMT_ASYM_SCHEME *));
  }

  static void delete_TPMT_ASYM_SCHEME_PTR_PTR(TPMT_ASYM_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_ASYM_SCHEME_PTR_PTR_assign(TPMT_ASYM_SCHEME * *obj, TPMT_ASYM_SCHEME * value) {
    *obj = value;
  }

  static TPMT_ASYM_SCHEME * TPMT_ASYM_SCHEME_PTR_PTR_value(TPMT_ASYM_SCHEME * *obj) {
    return *obj;
  }


typedef TPMT_RSA_SCHEME TPMT_RSA_SCHEME_ARRAY;

SWIGINTERN TPMT_RSA_SCHEME_ARRAY *new_TPMT_RSA_SCHEME_ARRAY(size_t nelements){
    return (TPMT_RSA_SCHEME *)malloc((nelements)*sizeof(TPMT_RSA_SCHEME));
  }
SWIGINTERN void delete_TPMT_RSA_SCHEME_ARRAY(TPMT_RSA_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_RSA_SCHEME TPMT_RSA_SCHEME_ARRAY___getitem__(TPMT_RSA_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_RSA_SCHEME_ARRAY___setitem__(TPMT_RSA_SCHEME_ARRAY *self,size_t index,TPMT_RSA_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMT_RSA_SCHEME *TPMT_RSA_SCHEME_ARRAY_cast(TPMT_RSA_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_RSA_SCHEME_ARRAY *TPMT_RSA_SCHEME_ARRAY_frompointer(TPMT_RSA_SCHEME *t){
    return (TPMT_RSA_SCHEME_ARRAY *)(t);
  }

  static TPMT_RSA_SCHEME *new_TPMT_RSA_SCHEME_PTR() { 
    return (TPMT_RSA_SCHEME *)malloc(sizeof(TPMT_RSA_SCHEME));
  }
  
  static TPMT_RSA_SCHEME *copy_TPMT_RSA_SCHEME_PTR(TPMT_RSA_SCHEME value) { 
    return (TPMT_RSA_SCHEME *)memcpy((TPMT_RSA_SCHEME *)malloc(sizeof(TPMT_RSA_SCHEME)),&value,sizeof(TPMT_RSA_SCHEME));
  }

  static void delete_TPMT_RSA_SCHEME_PTR(TPMT_RSA_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_RSA_SCHEME_PTR_assign(TPMT_RSA_SCHEME *obj, TPMT_RSA_SCHEME value) {
    *obj = value;
  }

  static TPMT_RSA_SCHEME TPMT_RSA_SCHEME_PTR_value(TPMT_RSA_SCHEME *obj) {
    return *obj;
  }


  static TPMT_RSA_SCHEME * *new_TPMT_RSA_SCHEME_PTR_PTR() { 
    return (TPMT_RSA_SCHEME * *)malloc(sizeof(TPMT_RSA_SCHEME *));
  }
  
  static TPMT_RSA_SCHEME * *copy_TPMT_RSA_SCHEME_PTR_PTR(TPMT_RSA_SCHEME * value) { 
    return (TPMT_RSA_SCHEME * *)memcpy((TPMT_RSA_SCHEME * *)malloc(sizeof(TPMT_RSA_SCHEME *)),&value,sizeof(TPMT_RSA_SCHEME *));
  }

  static void delete_TPMT_RSA_SCHEME_PTR_PTR(TPMT_RSA_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_RSA_SCHEME_PTR_PTR_assign(TPMT_RSA_SCHEME * *obj, TPMT_RSA_SCHEME * value) {
    *obj = value;
  }

  static TPMT_RSA_SCHEME * TPMT_RSA_SCHEME_PTR_PTR_value(TPMT_RSA_SCHEME * *obj) {
    return *obj;
  }


typedef TPMT_RSA_DECRYPT TPMT_RSA_DECRYPT_ARRAY;

SWIGINTERN TPMT_RSA_DECRYPT_ARRAY *new_TPMT_RSA_DECRYPT_ARRAY(size_t nelements){
    return (TPMT_RSA_DECRYPT *)malloc((nelements)*sizeof(TPMT_RSA_DECRYPT));
  }
SWIGINTERN void delete_TPMT_RSA_DECRYPT_ARRAY(TPMT_RSA_DECRYPT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_RSA_DECRYPT TPMT_RSA_DECRYPT_ARRAY___getitem__(TPMT_RSA_DECRYPT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_RSA_DECRYPT_ARRAY___setitem__(TPMT_RSA_DECRYPT_ARRAY *self,size_t index,TPMT_RSA_DECRYPT value){
    self[index] = value;
  }
SWIGINTERN TPMT_RSA_DECRYPT *TPMT_RSA_DECRYPT_ARRAY_cast(TPMT_RSA_DECRYPT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_RSA_DECRYPT_ARRAY *TPMT_RSA_DECRYPT_ARRAY_frompointer(TPMT_RSA_DECRYPT *t){
    return (TPMT_RSA_DECRYPT_ARRAY *)(t);
  }

  static TPMT_RSA_DECRYPT *new_TPMT_RSA_DECRYPT_PTR() { 
    return (TPMT_RSA_DECRYPT *)malloc(sizeof(TPMT_RSA_DECRYPT));
  }
  
  static TPMT_RSA_DECRYPT *copy_TPMT_RSA_DECRYPT_PTR(TPMT_RSA_DECRYPT value) { 
    return (TPMT_RSA_DECRYPT *)memcpy((TPMT_RSA_DECRYPT *)malloc(sizeof(TPMT_RSA_DECRYPT)),&value,sizeof(TPMT_RSA_DECRYPT));
  }

  static void delete_TPMT_RSA_DECRYPT_PTR(TPMT_RSA_DECRYPT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_RSA_DECRYPT_PTR_assign(TPMT_RSA_DECRYPT *obj, TPMT_RSA_DECRYPT value) {
    *obj = value;
  }

  static TPMT_RSA_DECRYPT TPMT_RSA_DECRYPT_PTR_value(TPMT_RSA_DECRYPT *obj) {
    return *obj;
  }


  static TPMT_RSA_DECRYPT * *new_TPMT_RSA_DECRYPT_PTR_PTR() { 
    return (TPMT_RSA_DECRYPT * *)malloc(sizeof(TPMT_RSA_DECRYPT *));
  }
  
  static TPMT_RSA_DECRYPT * *copy_TPMT_RSA_DECRYPT_PTR_PTR(TPMT_RSA_DECRYPT * value) { 
    return (TPMT_RSA_DECRYPT * *)memcpy((TPMT_RSA_DECRYPT * *)malloc(sizeof(TPMT_RSA_DECRYPT *)),&value,sizeof(TPMT_RSA_DECRYPT *));
  }

  static void delete_TPMT_RSA_DECRYPT_PTR_PTR(TPMT_RSA_DECRYPT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_RSA_DECRYPT_PTR_PTR_assign(TPMT_RSA_DECRYPT * *obj, TPMT_RSA_DECRYPT * value) {
    *obj = value;
  }

  static TPMT_RSA_DECRYPT * TPMT_RSA_DECRYPT_PTR_PTR_value(TPMT_RSA_DECRYPT * *obj) {
    return *obj;
  }


typedef TPM2B_PUBLIC_KEY_RSA TPM2B_PUBLIC_KEY_RSA_ARRAY;

SWIGINTERN TPM2B_PUBLIC_KEY_RSA_ARRAY *new_TPM2B_PUBLIC_KEY_RSA_ARRAY(size_t nelements){
    return (TPM2B_PUBLIC_KEY_RSA *)malloc((nelements)*sizeof(TPM2B_PUBLIC_KEY_RSA));
  }
SWIGINTERN void delete_TPM2B_PUBLIC_KEY_RSA_ARRAY(TPM2B_PUBLIC_KEY_RSA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_PUBLIC_KEY_RSA TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__(TPM2B_PUBLIC_KEY_RSA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__(TPM2B_PUBLIC_KEY_RSA_ARRAY *self,size_t index,TPM2B_PUBLIC_KEY_RSA value){
    self[index] = value;
  }
SWIGINTERN TPM2B_PUBLIC_KEY_RSA *TPM2B_PUBLIC_KEY_RSA_ARRAY_cast(TPM2B_PUBLIC_KEY_RSA_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_PUBLIC_KEY_RSA_ARRAY *TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer(TPM2B_PUBLIC_KEY_RSA *t){
    return (TPM2B_PUBLIC_KEY_RSA_ARRAY *)(t);
  }

  static TPM2B_PUBLIC_KEY_RSA *new_TPM2B_PUBLIC_KEY_RSA_PTR() { 
    return (TPM2B_PUBLIC_KEY_RSA *)malloc(sizeof(TPM2B_PUBLIC_KEY_RSA));
  }
  
  static TPM2B_PUBLIC_KEY_RSA *copy_TPM2B_PUBLIC_KEY_RSA_PTR(TPM2B_PUBLIC_KEY_RSA value) { 
    return (TPM2B_PUBLIC_KEY_RSA *)memcpy((TPM2B_PUBLIC_KEY_RSA *)malloc(sizeof(TPM2B_PUBLIC_KEY_RSA)),&value,sizeof(TPM2B_PUBLIC_KEY_RSA));
  }

  static void delete_TPM2B_PUBLIC_KEY_RSA_PTR(TPM2B_PUBLIC_KEY_RSA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PUBLIC_KEY_RSA_PTR_assign(TPM2B_PUBLIC_KEY_RSA *obj, TPM2B_PUBLIC_KEY_RSA value) {
    *obj = value;
  }

  static TPM2B_PUBLIC_KEY_RSA TPM2B_PUBLIC_KEY_RSA_PTR_value(TPM2B_PUBLIC_KEY_RSA *obj) {
    return *obj;
  }


  static TPM2B_PUBLIC_KEY_RSA * *new_TPM2B_PUBLIC_KEY_RSA_PTR_PTR() { 
    return (TPM2B_PUBLIC_KEY_RSA * *)malloc(sizeof(TPM2B_PUBLIC_KEY_RSA *));
  }
  
  static TPM2B_PUBLIC_KEY_RSA * *copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(TPM2B_PUBLIC_KEY_RSA * value) { 
    return (TPM2B_PUBLIC_KEY_RSA * *)memcpy((TPM2B_PUBLIC_KEY_RSA * *)malloc(sizeof(TPM2B_PUBLIC_KEY_RSA *)),&value,sizeof(TPM2B_PUBLIC_KEY_RSA *));
  }

  static void delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(TPM2B_PUBLIC_KEY_RSA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign(TPM2B_PUBLIC_KEY_RSA * *obj, TPM2B_PUBLIC_KEY_RSA * value) {
    *obj = value;
  }

  static TPM2B_PUBLIC_KEY_RSA * TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value(TPM2B_PUBLIC_KEY_RSA * *obj) {
    return *obj;
  }


typedef TPM2B_PRIVATE_KEY_RSA TPM2B_PRIVATE_KEY_RSA_ARRAY;

SWIGINTERN TPM2B_PRIVATE_KEY_RSA_ARRAY *new_TPM2B_PRIVATE_KEY_RSA_ARRAY(size_t nelements){
    return (TPM2B_PRIVATE_KEY_RSA *)malloc((nelements)*sizeof(TPM2B_PRIVATE_KEY_RSA));
  }
SWIGINTERN void delete_TPM2B_PRIVATE_KEY_RSA_ARRAY(TPM2B_PRIVATE_KEY_RSA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_PRIVATE_KEY_RSA TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__(TPM2B_PRIVATE_KEY_RSA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__(TPM2B_PRIVATE_KEY_RSA_ARRAY *self,size_t index,TPM2B_PRIVATE_KEY_RSA value){
    self[index] = value;
  }
SWIGINTERN TPM2B_PRIVATE_KEY_RSA *TPM2B_PRIVATE_KEY_RSA_ARRAY_cast(TPM2B_PRIVATE_KEY_RSA_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_PRIVATE_KEY_RSA_ARRAY *TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer(TPM2B_PRIVATE_KEY_RSA *t){
    return (TPM2B_PRIVATE_KEY_RSA_ARRAY *)(t);
  }

  static TPM2B_PRIVATE_KEY_RSA *new_TPM2B_PRIVATE_KEY_RSA_PTR() { 
    return (TPM2B_PRIVATE_KEY_RSA *)malloc(sizeof(TPM2B_PRIVATE_KEY_RSA));
  }
  
  static TPM2B_PRIVATE_KEY_RSA *copy_TPM2B_PRIVATE_KEY_RSA_PTR(TPM2B_PRIVATE_KEY_RSA value) { 
    return (TPM2B_PRIVATE_KEY_RSA *)memcpy((TPM2B_PRIVATE_KEY_RSA *)malloc(sizeof(TPM2B_PRIVATE_KEY_RSA)),&value,sizeof(TPM2B_PRIVATE_KEY_RSA));
  }

  static void delete_TPM2B_PRIVATE_KEY_RSA_PTR(TPM2B_PRIVATE_KEY_RSA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PRIVATE_KEY_RSA_PTR_assign(TPM2B_PRIVATE_KEY_RSA *obj, TPM2B_PRIVATE_KEY_RSA value) {
    *obj = value;
  }

  static TPM2B_PRIVATE_KEY_RSA TPM2B_PRIVATE_KEY_RSA_PTR_value(TPM2B_PRIVATE_KEY_RSA *obj) {
    return *obj;
  }


  static TPM2B_PRIVATE_KEY_RSA * *new_TPM2B_PRIVATE_KEY_RSA_PTR_PTR() { 
    return (TPM2B_PRIVATE_KEY_RSA * *)malloc(sizeof(TPM2B_PRIVATE_KEY_RSA *));
  }
  
  static TPM2B_PRIVATE_KEY_RSA * *copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(TPM2B_PRIVATE_KEY_RSA * value) { 
    return (TPM2B_PRIVATE_KEY_RSA * *)memcpy((TPM2B_PRIVATE_KEY_RSA * *)malloc(sizeof(TPM2B_PRIVATE_KEY_RSA *)),&value,sizeof(TPM2B_PRIVATE_KEY_RSA *));
  }

  static void delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(TPM2B_PRIVATE_KEY_RSA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign(TPM2B_PRIVATE_KEY_RSA * *obj, TPM2B_PRIVATE_KEY_RSA * value) {
    *obj = value;
  }

  static TPM2B_PRIVATE_KEY_RSA * TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value(TPM2B_PRIVATE_KEY_RSA * *obj) {
    return *obj;
  }


typedef TPM2B_ECC_PARAMETER TPM2B_ECC_PARAMETER_ARRAY;

SWIGINTERN TPM2B_ECC_PARAMETER_ARRAY *new_TPM2B_ECC_PARAMETER_ARRAY(size_t nelements){
    return (TPM2B_ECC_PARAMETER *)malloc((nelements)*sizeof(TPM2B_ECC_PARAMETER));
  }
SWIGINTERN void delete_TPM2B_ECC_PARAMETER_ARRAY(TPM2B_ECC_PARAMETER_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_ECC_PARAMETER TPM2B_ECC_PARAMETER_ARRAY___getitem__(TPM2B_ECC_PARAMETER_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_ECC_PARAMETER_ARRAY___setitem__(TPM2B_ECC_PARAMETER_ARRAY *self,size_t index,TPM2B_ECC_PARAMETER value){
    self[index] = value;
  }
SWIGINTERN TPM2B_ECC_PARAMETER *TPM2B_ECC_PARAMETER_ARRAY_cast(TPM2B_ECC_PARAMETER_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_ECC_PARAMETER_ARRAY *TPM2B_ECC_PARAMETER_ARRAY_frompointer(TPM2B_ECC_PARAMETER *t){
    return (TPM2B_ECC_PARAMETER_ARRAY *)(t);
  }

  static TPM2B_ECC_PARAMETER *new_TPM2B_ECC_PARAMETER_PTR() { 
    return (TPM2B_ECC_PARAMETER *)malloc(sizeof(TPM2B_ECC_PARAMETER));
  }
  
  static TPM2B_ECC_PARAMETER *copy_TPM2B_ECC_PARAMETER_PTR(TPM2B_ECC_PARAMETER value) { 
    return (TPM2B_ECC_PARAMETER *)memcpy((TPM2B_ECC_PARAMETER *)malloc(sizeof(TPM2B_ECC_PARAMETER)),&value,sizeof(TPM2B_ECC_PARAMETER));
  }

  static void delete_TPM2B_ECC_PARAMETER_PTR(TPM2B_ECC_PARAMETER *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ECC_PARAMETER_PTR_assign(TPM2B_ECC_PARAMETER *obj, TPM2B_ECC_PARAMETER value) {
    *obj = value;
  }

  static TPM2B_ECC_PARAMETER TPM2B_ECC_PARAMETER_PTR_value(TPM2B_ECC_PARAMETER *obj) {
    return *obj;
  }


  static TPM2B_ECC_PARAMETER * *new_TPM2B_ECC_PARAMETER_PTR_PTR() { 
    return (TPM2B_ECC_PARAMETER * *)malloc(sizeof(TPM2B_ECC_PARAMETER *));
  }
  
  static TPM2B_ECC_PARAMETER * *copy_TPM2B_ECC_PARAMETER_PTR_PTR(TPM2B_ECC_PARAMETER * value) { 
    return (TPM2B_ECC_PARAMETER * *)memcpy((TPM2B_ECC_PARAMETER * *)malloc(sizeof(TPM2B_ECC_PARAMETER *)),&value,sizeof(TPM2B_ECC_PARAMETER *));
  }

  static void delete_TPM2B_ECC_PARAMETER_PTR_PTR(TPM2B_ECC_PARAMETER * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ECC_PARAMETER_PTR_PTR_assign(TPM2B_ECC_PARAMETER * *obj, TPM2B_ECC_PARAMETER * value) {
    *obj = value;
  }

  static TPM2B_ECC_PARAMETER * TPM2B_ECC_PARAMETER_PTR_PTR_value(TPM2B_ECC_PARAMETER * *obj) {
    return *obj;
  }


typedef TPMS_ECC_POINT TPMS_ECC_POINT_ARRAY;

SWIGINTERN TPMS_ECC_POINT_ARRAY *new_TPMS_ECC_POINT_ARRAY(size_t nelements){
    return (TPMS_ECC_POINT *)malloc((nelements)*sizeof(TPMS_ECC_POINT));
  }
SWIGINTERN void delete_TPMS_ECC_POINT_ARRAY(TPMS_ECC_POINT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ECC_POINT TPMS_ECC_POINT_ARRAY___getitem__(TPMS_ECC_POINT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ECC_POINT_ARRAY___setitem__(TPMS_ECC_POINT_ARRAY *self,size_t index,TPMS_ECC_POINT value){
    self[index] = value;
  }
SWIGINTERN TPMS_ECC_POINT *TPMS_ECC_POINT_ARRAY_cast(TPMS_ECC_POINT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ECC_POINT_ARRAY *TPMS_ECC_POINT_ARRAY_frompointer(TPMS_ECC_POINT *t){
    return (TPMS_ECC_POINT_ARRAY *)(t);
  }

  static TPMS_ECC_POINT *new_TPMS_ECC_POINT_PTR() { 
    return (TPMS_ECC_POINT *)malloc(sizeof(TPMS_ECC_POINT));
  }
  
  static TPMS_ECC_POINT *copy_TPMS_ECC_POINT_PTR(TPMS_ECC_POINT value) { 
    return (TPMS_ECC_POINT *)memcpy((TPMS_ECC_POINT *)malloc(sizeof(TPMS_ECC_POINT)),&value,sizeof(TPMS_ECC_POINT));
  }

  static void delete_TPMS_ECC_POINT_PTR(TPMS_ECC_POINT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ECC_POINT_PTR_assign(TPMS_ECC_POINT *obj, TPMS_ECC_POINT value) {
    *obj = value;
  }

  static TPMS_ECC_POINT TPMS_ECC_POINT_PTR_value(TPMS_ECC_POINT *obj) {
    return *obj;
  }


  static TPMS_ECC_POINT * *new_TPMS_ECC_POINT_PTR_PTR() { 
    return (TPMS_ECC_POINT * *)malloc(sizeof(TPMS_ECC_POINT *));
  }
  
  static TPMS_ECC_POINT * *copy_TPMS_ECC_POINT_PTR_PTR(TPMS_ECC_POINT * value) { 
    return (TPMS_ECC_POINT * *)memcpy((TPMS_ECC_POINT * *)malloc(sizeof(TPMS_ECC_POINT *)),&value,sizeof(TPMS_ECC_POINT *));
  }

  static void delete_TPMS_ECC_POINT_PTR_PTR(TPMS_ECC_POINT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ECC_POINT_PTR_PTR_assign(TPMS_ECC_POINT * *obj, TPMS_ECC_POINT * value) {
    *obj = value;
  }

  static TPMS_ECC_POINT * TPMS_ECC_POINT_PTR_PTR_value(TPMS_ECC_POINT * *obj) {
    return *obj;
  }


typedef TPM2B_ECC_POINT TPM2B_ECC_POINT_ARRAY;

SWIGINTERN TPM2B_ECC_POINT_ARRAY *new_TPM2B_ECC_POINT_ARRAY(size_t nelements){
    return (TPM2B_ECC_POINT *)malloc((nelements)*sizeof(TPM2B_ECC_POINT));
  }
SWIGINTERN void delete_TPM2B_ECC_POINT_ARRAY(TPM2B_ECC_POINT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_ECC_POINT TPM2B_ECC_POINT_ARRAY___getitem__(TPM2B_ECC_POINT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_ECC_POINT_ARRAY___setitem__(TPM2B_ECC_POINT_ARRAY *self,size_t index,TPM2B_ECC_POINT value){
    self[index] = value;
  }
SWIGINTERN TPM2B_ECC_POINT *TPM2B_ECC_POINT_ARRAY_cast(TPM2B_ECC_POINT_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_ECC_POINT_ARRAY *TPM2B_ECC_POINT_ARRAY_frompointer(TPM2B_ECC_POINT *t){
    return (TPM2B_ECC_POINT_ARRAY *)(t);
  }

  static TPM2B_ECC_POINT *new_TPM2B_ECC_POINT_PTR() { 
    return (TPM2B_ECC_POINT *)malloc(sizeof(TPM2B_ECC_POINT));
  }
  
  static TPM2B_ECC_POINT *copy_TPM2B_ECC_POINT_PTR(TPM2B_ECC_POINT value) { 
    return (TPM2B_ECC_POINT *)memcpy((TPM2B_ECC_POINT *)malloc(sizeof(TPM2B_ECC_POINT)),&value,sizeof(TPM2B_ECC_POINT));
  }

  static void delete_TPM2B_ECC_POINT_PTR(TPM2B_ECC_POINT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ECC_POINT_PTR_assign(TPM2B_ECC_POINT *obj, TPM2B_ECC_POINT value) {
    *obj = value;
  }

  static TPM2B_ECC_POINT TPM2B_ECC_POINT_PTR_value(TPM2B_ECC_POINT *obj) {
    return *obj;
  }


  static TPM2B_ECC_POINT * *new_TPM2B_ECC_POINT_PTR_PTR() { 
    return (TPM2B_ECC_POINT * *)malloc(sizeof(TPM2B_ECC_POINT *));
  }
  
  static TPM2B_ECC_POINT * *copy_TPM2B_ECC_POINT_PTR_PTR(TPM2B_ECC_POINT * value) { 
    return (TPM2B_ECC_POINT * *)memcpy((TPM2B_ECC_POINT * *)malloc(sizeof(TPM2B_ECC_POINT *)),&value,sizeof(TPM2B_ECC_POINT *));
  }

  static void delete_TPM2B_ECC_POINT_PTR_PTR(TPM2B_ECC_POINT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ECC_POINT_PTR_PTR_assign(TPM2B_ECC_POINT * *obj, TPM2B_ECC_POINT * value) {
    *obj = value;
  }

  static TPM2B_ECC_POINT * TPM2B_ECC_POINT_PTR_PTR_value(TPM2B_ECC_POINT * *obj) {
    return *obj;
  }


typedef TPMT_ECC_SCHEME TPMT_ECC_SCHEME_ARRAY;

SWIGINTERN TPMT_ECC_SCHEME_ARRAY *new_TPMT_ECC_SCHEME_ARRAY(size_t nelements){
    return (TPMT_ECC_SCHEME *)malloc((nelements)*sizeof(TPMT_ECC_SCHEME));
  }
SWIGINTERN void delete_TPMT_ECC_SCHEME_ARRAY(TPMT_ECC_SCHEME_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_ECC_SCHEME TPMT_ECC_SCHEME_ARRAY___getitem__(TPMT_ECC_SCHEME_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_ECC_SCHEME_ARRAY___setitem__(TPMT_ECC_SCHEME_ARRAY *self,size_t index,TPMT_ECC_SCHEME value){
    self[index] = value;
  }
SWIGINTERN TPMT_ECC_SCHEME *TPMT_ECC_SCHEME_ARRAY_cast(TPMT_ECC_SCHEME_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_ECC_SCHEME_ARRAY *TPMT_ECC_SCHEME_ARRAY_frompointer(TPMT_ECC_SCHEME *t){
    return (TPMT_ECC_SCHEME_ARRAY *)(t);
  }

  static TPMT_ECC_SCHEME *new_TPMT_ECC_SCHEME_PTR() { 
    return (TPMT_ECC_SCHEME *)malloc(sizeof(TPMT_ECC_SCHEME));
  }
  
  static TPMT_ECC_SCHEME *copy_TPMT_ECC_SCHEME_PTR(TPMT_ECC_SCHEME value) { 
    return (TPMT_ECC_SCHEME *)memcpy((TPMT_ECC_SCHEME *)malloc(sizeof(TPMT_ECC_SCHEME)),&value,sizeof(TPMT_ECC_SCHEME));
  }

  static void delete_TPMT_ECC_SCHEME_PTR(TPMT_ECC_SCHEME *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_ECC_SCHEME_PTR_assign(TPMT_ECC_SCHEME *obj, TPMT_ECC_SCHEME value) {
    *obj = value;
  }

  static TPMT_ECC_SCHEME TPMT_ECC_SCHEME_PTR_value(TPMT_ECC_SCHEME *obj) {
    return *obj;
  }


  static TPMT_ECC_SCHEME * *new_TPMT_ECC_SCHEME_PTR_PTR() { 
    return (TPMT_ECC_SCHEME * *)malloc(sizeof(TPMT_ECC_SCHEME *));
  }
  
  static TPMT_ECC_SCHEME * *copy_TPMT_ECC_SCHEME_PTR_PTR(TPMT_ECC_SCHEME * value) { 
    return (TPMT_ECC_SCHEME * *)memcpy((TPMT_ECC_SCHEME * *)malloc(sizeof(TPMT_ECC_SCHEME *)),&value,sizeof(TPMT_ECC_SCHEME *));
  }

  static void delete_TPMT_ECC_SCHEME_PTR_PTR(TPMT_ECC_SCHEME * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_ECC_SCHEME_PTR_PTR_assign(TPMT_ECC_SCHEME * *obj, TPMT_ECC_SCHEME * value) {
    *obj = value;
  }

  static TPMT_ECC_SCHEME * TPMT_ECC_SCHEME_PTR_PTR_value(TPMT_ECC_SCHEME * *obj) {
    return *obj;
  }


typedef TPMS_ALGORITHM_DETAIL_ECC TPMS_ALGORITHM_DETAIL_ECC_ARRAY;

SWIGINTERN TPMS_ALGORITHM_DETAIL_ECC_ARRAY *new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY(size_t nelements){
    return (TPMS_ALGORITHM_DETAIL_ECC *)malloc((nelements)*sizeof(TPMS_ALGORITHM_DETAIL_ECC));
  }
SWIGINTERN void delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY(TPMS_ALGORITHM_DETAIL_ECC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ALGORITHM_DETAIL_ECC TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__(TPMS_ALGORITHM_DETAIL_ECC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__(TPMS_ALGORITHM_DETAIL_ECC_ARRAY *self,size_t index,TPMS_ALGORITHM_DETAIL_ECC value){
    self[index] = value;
  }
SWIGINTERN TPMS_ALGORITHM_DETAIL_ECC *TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast(TPMS_ALGORITHM_DETAIL_ECC_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ALGORITHM_DETAIL_ECC_ARRAY *TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer(TPMS_ALGORITHM_DETAIL_ECC *t){
    return (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)(t);
  }

  static TPMS_ALGORITHM_DETAIL_ECC *new_TPMS_ALGORITHM_DETAIL_ECC_PTR() { 
    return (TPMS_ALGORITHM_DETAIL_ECC *)malloc(sizeof(TPMS_ALGORITHM_DETAIL_ECC));
  }
  
  static TPMS_ALGORITHM_DETAIL_ECC *copy_TPMS_ALGORITHM_DETAIL_ECC_PTR(TPMS_ALGORITHM_DETAIL_ECC value) { 
    return (TPMS_ALGORITHM_DETAIL_ECC *)memcpy((TPMS_ALGORITHM_DETAIL_ECC *)malloc(sizeof(TPMS_ALGORITHM_DETAIL_ECC)),&value,sizeof(TPMS_ALGORITHM_DETAIL_ECC));
  }

  static void delete_TPMS_ALGORITHM_DETAIL_ECC_PTR(TPMS_ALGORITHM_DETAIL_ECC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ALGORITHM_DETAIL_ECC_PTR_assign(TPMS_ALGORITHM_DETAIL_ECC *obj, TPMS_ALGORITHM_DETAIL_ECC value) {
    *obj = value;
  }

  static TPMS_ALGORITHM_DETAIL_ECC TPMS_ALGORITHM_DETAIL_ECC_PTR_value(TPMS_ALGORITHM_DETAIL_ECC *obj) {
    return *obj;
  }


  static TPMS_ALGORITHM_DETAIL_ECC * *new_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR() { 
    return (TPMS_ALGORITHM_DETAIL_ECC * *)malloc(sizeof(TPMS_ALGORITHM_DETAIL_ECC *));
  }
  
  static TPMS_ALGORITHM_DETAIL_ECC * *copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(TPMS_ALGORITHM_DETAIL_ECC * value) { 
    return (TPMS_ALGORITHM_DETAIL_ECC * *)memcpy((TPMS_ALGORITHM_DETAIL_ECC * *)malloc(sizeof(TPMS_ALGORITHM_DETAIL_ECC *)),&value,sizeof(TPMS_ALGORITHM_DETAIL_ECC *));
  }

  static void delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(TPMS_ALGORITHM_DETAIL_ECC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign(TPMS_ALGORITHM_DETAIL_ECC * *obj, TPMS_ALGORITHM_DETAIL_ECC * value) {
    *obj = value;
  }

  static TPMS_ALGORITHM_DETAIL_ECC * TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value(TPMS_ALGORITHM_DETAIL_ECC * *obj) {
    return *obj;
  }


typedef TPMS_SIGNATURE_RSA TPMS_SIGNATURE_RSA_ARRAY;

SWIGINTERN TPMS_SIGNATURE_RSA_ARRAY *new_TPMS_SIGNATURE_RSA_ARRAY(size_t nelements){
    return (TPMS_SIGNATURE_RSA *)malloc((nelements)*sizeof(TPMS_SIGNATURE_RSA));
  }
SWIGINTERN void delete_TPMS_SIGNATURE_RSA_ARRAY(TPMS_SIGNATURE_RSA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SIGNATURE_RSA TPMS_SIGNATURE_RSA_ARRAY___getitem__(TPMS_SIGNATURE_RSA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SIGNATURE_RSA_ARRAY___setitem__(TPMS_SIGNATURE_RSA_ARRAY *self,size_t index,TPMS_SIGNATURE_RSA value){
    self[index] = value;
  }
SWIGINTERN TPMS_SIGNATURE_RSA *TPMS_SIGNATURE_RSA_ARRAY_cast(TPMS_SIGNATURE_RSA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SIGNATURE_RSA_ARRAY *TPMS_SIGNATURE_RSA_ARRAY_frompointer(TPMS_SIGNATURE_RSA *t){
    return (TPMS_SIGNATURE_RSA_ARRAY *)(t);
  }

  static TPMS_SIGNATURE_RSA *new_TPMS_SIGNATURE_RSA_PTR() { 
    return (TPMS_SIGNATURE_RSA *)malloc(sizeof(TPMS_SIGNATURE_RSA));
  }
  
  static TPMS_SIGNATURE_RSA *copy_TPMS_SIGNATURE_RSA_PTR(TPMS_SIGNATURE_RSA value) { 
    return (TPMS_SIGNATURE_RSA *)memcpy((TPMS_SIGNATURE_RSA *)malloc(sizeof(TPMS_SIGNATURE_RSA)),&value,sizeof(TPMS_SIGNATURE_RSA));
  }

  static void delete_TPMS_SIGNATURE_RSA_PTR(TPMS_SIGNATURE_RSA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SIGNATURE_RSA_PTR_assign(TPMS_SIGNATURE_RSA *obj, TPMS_SIGNATURE_RSA value) {
    *obj = value;
  }

  static TPMS_SIGNATURE_RSA TPMS_SIGNATURE_RSA_PTR_value(TPMS_SIGNATURE_RSA *obj) {
    return *obj;
  }


  static TPMS_SIGNATURE_RSA * *new_TPMS_SIGNATURE_RSA_PTR_PTR() { 
    return (TPMS_SIGNATURE_RSA * *)malloc(sizeof(TPMS_SIGNATURE_RSA *));
  }
  
  static TPMS_SIGNATURE_RSA * *copy_TPMS_SIGNATURE_RSA_PTR_PTR(TPMS_SIGNATURE_RSA * value) { 
    return (TPMS_SIGNATURE_RSA * *)memcpy((TPMS_SIGNATURE_RSA * *)malloc(sizeof(TPMS_SIGNATURE_RSA *)),&value,sizeof(TPMS_SIGNATURE_RSA *));
  }

  static void delete_TPMS_SIGNATURE_RSA_PTR_PTR(TPMS_SIGNATURE_RSA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SIGNATURE_RSA_PTR_PTR_assign(TPMS_SIGNATURE_RSA * *obj, TPMS_SIGNATURE_RSA * value) {
    *obj = value;
  }

  static TPMS_SIGNATURE_RSA * TPMS_SIGNATURE_RSA_PTR_PTR_value(TPMS_SIGNATURE_RSA * *obj) {
    return *obj;
  }


typedef TPMS_SIGNATURE_ECC TPMS_SIGNATURE_ECC_ARRAY;

SWIGINTERN TPMS_SIGNATURE_ECC_ARRAY *new_TPMS_SIGNATURE_ECC_ARRAY(size_t nelements){
    return (TPMS_SIGNATURE_ECC *)malloc((nelements)*sizeof(TPMS_SIGNATURE_ECC));
  }
SWIGINTERN void delete_TPMS_SIGNATURE_ECC_ARRAY(TPMS_SIGNATURE_ECC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_SIGNATURE_ECC TPMS_SIGNATURE_ECC_ARRAY___getitem__(TPMS_SIGNATURE_ECC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_SIGNATURE_ECC_ARRAY___setitem__(TPMS_SIGNATURE_ECC_ARRAY *self,size_t index,TPMS_SIGNATURE_ECC value){
    self[index] = value;
  }
SWIGINTERN TPMS_SIGNATURE_ECC *TPMS_SIGNATURE_ECC_ARRAY_cast(TPMS_SIGNATURE_ECC_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_SIGNATURE_ECC_ARRAY *TPMS_SIGNATURE_ECC_ARRAY_frompointer(TPMS_SIGNATURE_ECC *t){
    return (TPMS_SIGNATURE_ECC_ARRAY *)(t);
  }

  static TPMS_SIGNATURE_ECC *new_TPMS_SIGNATURE_ECC_PTR() { 
    return (TPMS_SIGNATURE_ECC *)malloc(sizeof(TPMS_SIGNATURE_ECC));
  }
  
  static TPMS_SIGNATURE_ECC *copy_TPMS_SIGNATURE_ECC_PTR(TPMS_SIGNATURE_ECC value) { 
    return (TPMS_SIGNATURE_ECC *)memcpy((TPMS_SIGNATURE_ECC *)malloc(sizeof(TPMS_SIGNATURE_ECC)),&value,sizeof(TPMS_SIGNATURE_ECC));
  }

  static void delete_TPMS_SIGNATURE_ECC_PTR(TPMS_SIGNATURE_ECC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SIGNATURE_ECC_PTR_assign(TPMS_SIGNATURE_ECC *obj, TPMS_SIGNATURE_ECC value) {
    *obj = value;
  }

  static TPMS_SIGNATURE_ECC TPMS_SIGNATURE_ECC_PTR_value(TPMS_SIGNATURE_ECC *obj) {
    return *obj;
  }


  static TPMS_SIGNATURE_ECC * *new_TPMS_SIGNATURE_ECC_PTR_PTR() { 
    return (TPMS_SIGNATURE_ECC * *)malloc(sizeof(TPMS_SIGNATURE_ECC *));
  }
  
  static TPMS_SIGNATURE_ECC * *copy_TPMS_SIGNATURE_ECC_PTR_PTR(TPMS_SIGNATURE_ECC * value) { 
    return (TPMS_SIGNATURE_ECC * *)memcpy((TPMS_SIGNATURE_ECC * *)malloc(sizeof(TPMS_SIGNATURE_ECC *)),&value,sizeof(TPMS_SIGNATURE_ECC *));
  }

  static void delete_TPMS_SIGNATURE_ECC_PTR_PTR(TPMS_SIGNATURE_ECC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_SIGNATURE_ECC_PTR_PTR_assign(TPMS_SIGNATURE_ECC * *obj, TPMS_SIGNATURE_ECC * value) {
    *obj = value;
  }

  static TPMS_SIGNATURE_ECC * TPMS_SIGNATURE_ECC_PTR_PTR_value(TPMS_SIGNATURE_ECC * *obj) {
    return *obj;
  }


typedef TPMU_SIGNATURE TPMU_SIGNATURE_ARRAY;

SWIGINTERN TPMU_SIGNATURE_ARRAY *new_TPMU_SIGNATURE_ARRAY(size_t nelements){
    return (TPMU_SIGNATURE *)malloc((nelements)*sizeof(TPMU_SIGNATURE));
  }
SWIGINTERN void delete_TPMU_SIGNATURE_ARRAY(TPMU_SIGNATURE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_SIGNATURE TPMU_SIGNATURE_ARRAY___getitem__(TPMU_SIGNATURE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_SIGNATURE_ARRAY___setitem__(TPMU_SIGNATURE_ARRAY *self,size_t index,TPMU_SIGNATURE value){
    self[index] = value;
  }
SWIGINTERN TPMU_SIGNATURE *TPMU_SIGNATURE_ARRAY_cast(TPMU_SIGNATURE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_SIGNATURE_ARRAY *TPMU_SIGNATURE_ARRAY_frompointer(TPMU_SIGNATURE *t){
    return (TPMU_SIGNATURE_ARRAY *)(t);
  }

  static TPMU_SIGNATURE *new_TPMU_SIGNATURE_PTR() { 
    return (TPMU_SIGNATURE *)malloc(sizeof(TPMU_SIGNATURE));
  }
  
  static TPMU_SIGNATURE *copy_TPMU_SIGNATURE_PTR(TPMU_SIGNATURE value) { 
    return (TPMU_SIGNATURE *)memcpy((TPMU_SIGNATURE *)malloc(sizeof(TPMU_SIGNATURE)),&value,sizeof(TPMU_SIGNATURE));
  }

  static void delete_TPMU_SIGNATURE_PTR(TPMU_SIGNATURE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SIGNATURE_PTR_assign(TPMU_SIGNATURE *obj, TPMU_SIGNATURE value) {
    *obj = value;
  }

  static TPMU_SIGNATURE TPMU_SIGNATURE_PTR_value(TPMU_SIGNATURE *obj) {
    return *obj;
  }


  static TPMU_SIGNATURE * *new_TPMU_SIGNATURE_PTR_PTR() { 
    return (TPMU_SIGNATURE * *)malloc(sizeof(TPMU_SIGNATURE *));
  }
  
  static TPMU_SIGNATURE * *copy_TPMU_SIGNATURE_PTR_PTR(TPMU_SIGNATURE * value) { 
    return (TPMU_SIGNATURE * *)memcpy((TPMU_SIGNATURE * *)malloc(sizeof(TPMU_SIGNATURE *)),&value,sizeof(TPMU_SIGNATURE *));
  }

  static void delete_TPMU_SIGNATURE_PTR_PTR(TPMU_SIGNATURE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SIGNATURE_PTR_PTR_assign(TPMU_SIGNATURE * *obj, TPMU_SIGNATURE * value) {
    *obj = value;
  }

  static TPMU_SIGNATURE * TPMU_SIGNATURE_PTR_PTR_value(TPMU_SIGNATURE * *obj) {
    return *obj;
  }


typedef TPMT_SIGNATURE TPMT_SIGNATURE_ARRAY;

SWIGINTERN TPMT_SIGNATURE_ARRAY *new_TPMT_SIGNATURE_ARRAY(size_t nelements){
    return (TPMT_SIGNATURE *)malloc((nelements)*sizeof(TPMT_SIGNATURE));
  }
SWIGINTERN void delete_TPMT_SIGNATURE_ARRAY(TPMT_SIGNATURE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_SIGNATURE TPMT_SIGNATURE_ARRAY___getitem__(TPMT_SIGNATURE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_SIGNATURE_ARRAY___setitem__(TPMT_SIGNATURE_ARRAY *self,size_t index,TPMT_SIGNATURE value){
    self[index] = value;
  }
SWIGINTERN TPMT_SIGNATURE *TPMT_SIGNATURE_ARRAY_cast(TPMT_SIGNATURE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_SIGNATURE_ARRAY *TPMT_SIGNATURE_ARRAY_frompointer(TPMT_SIGNATURE *t){
    return (TPMT_SIGNATURE_ARRAY *)(t);
  }

  static TPMT_SIGNATURE *new_TPMT_SIGNATURE_PTR() { 
    return (TPMT_SIGNATURE *)malloc(sizeof(TPMT_SIGNATURE));
  }
  
  static TPMT_SIGNATURE *copy_TPMT_SIGNATURE_PTR(TPMT_SIGNATURE value) { 
    return (TPMT_SIGNATURE *)memcpy((TPMT_SIGNATURE *)malloc(sizeof(TPMT_SIGNATURE)),&value,sizeof(TPMT_SIGNATURE));
  }

  static void delete_TPMT_SIGNATURE_PTR(TPMT_SIGNATURE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SIGNATURE_PTR_assign(TPMT_SIGNATURE *obj, TPMT_SIGNATURE value) {
    *obj = value;
  }

  static TPMT_SIGNATURE TPMT_SIGNATURE_PTR_value(TPMT_SIGNATURE *obj) {
    return *obj;
  }


  static TPMT_SIGNATURE * *new_TPMT_SIGNATURE_PTR_PTR() { 
    return (TPMT_SIGNATURE * *)malloc(sizeof(TPMT_SIGNATURE *));
  }
  
  static TPMT_SIGNATURE * *copy_TPMT_SIGNATURE_PTR_PTR(TPMT_SIGNATURE * value) { 
    return (TPMT_SIGNATURE * *)memcpy((TPMT_SIGNATURE * *)malloc(sizeof(TPMT_SIGNATURE *)),&value,sizeof(TPMT_SIGNATURE *));
  }

  static void delete_TPMT_SIGNATURE_PTR_PTR(TPMT_SIGNATURE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SIGNATURE_PTR_PTR_assign(TPMT_SIGNATURE * *obj, TPMT_SIGNATURE * value) {
    *obj = value;
  }

  static TPMT_SIGNATURE * TPMT_SIGNATURE_PTR_PTR_value(TPMT_SIGNATURE * *obj) {
    return *obj;
  }


typedef TPMU_ENCRYPTED_SECRET TPMU_ENCRYPTED_SECRET_ARRAY;

SWIGINTERN TPMU_ENCRYPTED_SECRET_ARRAY *new_TPMU_ENCRYPTED_SECRET_ARRAY(size_t nelements){
    return (TPMU_ENCRYPTED_SECRET *)malloc((nelements)*sizeof(TPMU_ENCRYPTED_SECRET));
  }
SWIGINTERN void delete_TPMU_ENCRYPTED_SECRET_ARRAY(TPMU_ENCRYPTED_SECRET_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_ENCRYPTED_SECRET TPMU_ENCRYPTED_SECRET_ARRAY___getitem__(TPMU_ENCRYPTED_SECRET_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_ENCRYPTED_SECRET_ARRAY___setitem__(TPMU_ENCRYPTED_SECRET_ARRAY *self,size_t index,TPMU_ENCRYPTED_SECRET value){
    self[index] = value;
  }
SWIGINTERN TPMU_ENCRYPTED_SECRET *TPMU_ENCRYPTED_SECRET_ARRAY_cast(TPMU_ENCRYPTED_SECRET_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_ENCRYPTED_SECRET_ARRAY *TPMU_ENCRYPTED_SECRET_ARRAY_frompointer(TPMU_ENCRYPTED_SECRET *t){
    return (TPMU_ENCRYPTED_SECRET_ARRAY *)(t);
  }

  static TPMU_ENCRYPTED_SECRET *new_TPMU_ENCRYPTED_SECRET_PTR() { 
    return (TPMU_ENCRYPTED_SECRET *)malloc(sizeof(TPMU_ENCRYPTED_SECRET));
  }
  
  static TPMU_ENCRYPTED_SECRET *copy_TPMU_ENCRYPTED_SECRET_PTR(TPMU_ENCRYPTED_SECRET value) { 
    return (TPMU_ENCRYPTED_SECRET *)memcpy((TPMU_ENCRYPTED_SECRET *)malloc(sizeof(TPMU_ENCRYPTED_SECRET)),&value,sizeof(TPMU_ENCRYPTED_SECRET));
  }

  static void delete_TPMU_ENCRYPTED_SECRET_PTR(TPMU_ENCRYPTED_SECRET *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_ENCRYPTED_SECRET_PTR_assign(TPMU_ENCRYPTED_SECRET *obj, TPMU_ENCRYPTED_SECRET value) {
    *obj = value;
  }

  static TPMU_ENCRYPTED_SECRET TPMU_ENCRYPTED_SECRET_PTR_value(TPMU_ENCRYPTED_SECRET *obj) {
    return *obj;
  }


  static TPMU_ENCRYPTED_SECRET * *new_TPMU_ENCRYPTED_SECRET_PTR_PTR() { 
    return (TPMU_ENCRYPTED_SECRET * *)malloc(sizeof(TPMU_ENCRYPTED_SECRET *));
  }
  
  static TPMU_ENCRYPTED_SECRET * *copy_TPMU_ENCRYPTED_SECRET_PTR_PTR(TPMU_ENCRYPTED_SECRET * value) { 
    return (TPMU_ENCRYPTED_SECRET * *)memcpy((TPMU_ENCRYPTED_SECRET * *)malloc(sizeof(TPMU_ENCRYPTED_SECRET *)),&value,sizeof(TPMU_ENCRYPTED_SECRET *));
  }

  static void delete_TPMU_ENCRYPTED_SECRET_PTR_PTR(TPMU_ENCRYPTED_SECRET * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_ENCRYPTED_SECRET_PTR_PTR_assign(TPMU_ENCRYPTED_SECRET * *obj, TPMU_ENCRYPTED_SECRET * value) {
    *obj = value;
  }

  static TPMU_ENCRYPTED_SECRET * TPMU_ENCRYPTED_SECRET_PTR_PTR_value(TPMU_ENCRYPTED_SECRET * *obj) {
    return *obj;
  }


typedef TPM2B_ENCRYPTED_SECRET TPM2B_ENCRYPTED_SECRET_ARRAY;

SWIGINTERN TPM2B_ENCRYPTED_SECRET_ARRAY *new_TPM2B_ENCRYPTED_SECRET_ARRAY(size_t nelements){
    return (TPM2B_ENCRYPTED_SECRET *)malloc((nelements)*sizeof(TPM2B_ENCRYPTED_SECRET));
  }
SWIGINTERN void delete_TPM2B_ENCRYPTED_SECRET_ARRAY(TPM2B_ENCRYPTED_SECRET_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_ENCRYPTED_SECRET TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__(TPM2B_ENCRYPTED_SECRET_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__(TPM2B_ENCRYPTED_SECRET_ARRAY *self,size_t index,TPM2B_ENCRYPTED_SECRET value){
    self[index] = value;
  }
SWIGINTERN TPM2B_ENCRYPTED_SECRET *TPM2B_ENCRYPTED_SECRET_ARRAY_cast(TPM2B_ENCRYPTED_SECRET_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_ENCRYPTED_SECRET_ARRAY *TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer(TPM2B_ENCRYPTED_SECRET *t){
    return (TPM2B_ENCRYPTED_SECRET_ARRAY *)(t);
  }

  static TPM2B_ENCRYPTED_SECRET *new_TPM2B_ENCRYPTED_SECRET_PTR() { 
    return (TPM2B_ENCRYPTED_SECRET *)malloc(sizeof(TPM2B_ENCRYPTED_SECRET));
  }
  
  static TPM2B_ENCRYPTED_SECRET *copy_TPM2B_ENCRYPTED_SECRET_PTR(TPM2B_ENCRYPTED_SECRET value) { 
    return (TPM2B_ENCRYPTED_SECRET *)memcpy((TPM2B_ENCRYPTED_SECRET *)malloc(sizeof(TPM2B_ENCRYPTED_SECRET)),&value,sizeof(TPM2B_ENCRYPTED_SECRET));
  }

  static void delete_TPM2B_ENCRYPTED_SECRET_PTR(TPM2B_ENCRYPTED_SECRET *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ENCRYPTED_SECRET_PTR_assign(TPM2B_ENCRYPTED_SECRET *obj, TPM2B_ENCRYPTED_SECRET value) {
    *obj = value;
  }

  static TPM2B_ENCRYPTED_SECRET TPM2B_ENCRYPTED_SECRET_PTR_value(TPM2B_ENCRYPTED_SECRET *obj) {
    return *obj;
  }


  static TPM2B_ENCRYPTED_SECRET * *new_TPM2B_ENCRYPTED_SECRET_PTR_PTR() { 
    return (TPM2B_ENCRYPTED_SECRET * *)malloc(sizeof(TPM2B_ENCRYPTED_SECRET *));
  }
  
  static TPM2B_ENCRYPTED_SECRET * *copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR(TPM2B_ENCRYPTED_SECRET * value) { 
    return (TPM2B_ENCRYPTED_SECRET * *)memcpy((TPM2B_ENCRYPTED_SECRET * *)malloc(sizeof(TPM2B_ENCRYPTED_SECRET *)),&value,sizeof(TPM2B_ENCRYPTED_SECRET *));
  }

  static void delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR(TPM2B_ENCRYPTED_SECRET * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign(TPM2B_ENCRYPTED_SECRET * *obj, TPM2B_ENCRYPTED_SECRET * value) {
    *obj = value;
  }

  static TPM2B_ENCRYPTED_SECRET * TPM2B_ENCRYPTED_SECRET_PTR_PTR_value(TPM2B_ENCRYPTED_SECRET * *obj) {
    return *obj;
  }


typedef TPMU_PUBLIC_ID TPMU_PUBLIC_ID_ARRAY;

SWIGINTERN TPMU_PUBLIC_ID_ARRAY *new_TPMU_PUBLIC_ID_ARRAY(size_t nelements){
    return (TPMU_PUBLIC_ID *)malloc((nelements)*sizeof(TPMU_PUBLIC_ID));
  }
SWIGINTERN void delete_TPMU_PUBLIC_ID_ARRAY(TPMU_PUBLIC_ID_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_PUBLIC_ID TPMU_PUBLIC_ID_ARRAY___getitem__(TPMU_PUBLIC_ID_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_PUBLIC_ID_ARRAY___setitem__(TPMU_PUBLIC_ID_ARRAY *self,size_t index,TPMU_PUBLIC_ID value){
    self[index] = value;
  }
SWIGINTERN TPMU_PUBLIC_ID *TPMU_PUBLIC_ID_ARRAY_cast(TPMU_PUBLIC_ID_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_PUBLIC_ID_ARRAY *TPMU_PUBLIC_ID_ARRAY_frompointer(TPMU_PUBLIC_ID *t){
    return (TPMU_PUBLIC_ID_ARRAY *)(t);
  }

  static TPMU_PUBLIC_ID *new_TPMU_PUBLIC_ID_PTR() { 
    return (TPMU_PUBLIC_ID *)malloc(sizeof(TPMU_PUBLIC_ID));
  }
  
  static TPMU_PUBLIC_ID *copy_TPMU_PUBLIC_ID_PTR(TPMU_PUBLIC_ID value) { 
    return (TPMU_PUBLIC_ID *)memcpy((TPMU_PUBLIC_ID *)malloc(sizeof(TPMU_PUBLIC_ID)),&value,sizeof(TPMU_PUBLIC_ID));
  }

  static void delete_TPMU_PUBLIC_ID_PTR(TPMU_PUBLIC_ID *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_PUBLIC_ID_PTR_assign(TPMU_PUBLIC_ID *obj, TPMU_PUBLIC_ID value) {
    *obj = value;
  }

  static TPMU_PUBLIC_ID TPMU_PUBLIC_ID_PTR_value(TPMU_PUBLIC_ID *obj) {
    return *obj;
  }


  static TPMU_PUBLIC_ID * *new_TPMU_PUBLIC_ID_PTR_PTR() { 
    return (TPMU_PUBLIC_ID * *)malloc(sizeof(TPMU_PUBLIC_ID *));
  }
  
  static TPMU_PUBLIC_ID * *copy_TPMU_PUBLIC_ID_PTR_PTR(TPMU_PUBLIC_ID * value) { 
    return (TPMU_PUBLIC_ID * *)memcpy((TPMU_PUBLIC_ID * *)malloc(sizeof(TPMU_PUBLIC_ID *)),&value,sizeof(TPMU_PUBLIC_ID *));
  }

  static void delete_TPMU_PUBLIC_ID_PTR_PTR(TPMU_PUBLIC_ID * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_PUBLIC_ID_PTR_PTR_assign(TPMU_PUBLIC_ID * *obj, TPMU_PUBLIC_ID * value) {
    *obj = value;
  }

  static TPMU_PUBLIC_ID * TPMU_PUBLIC_ID_PTR_PTR_value(TPMU_PUBLIC_ID * *obj) {
    return *obj;
  }


typedef TPMS_KEYEDHASH_PARMS TPMS_KEYEDHASH_PARMS_ARRAY;

SWIGINTERN TPMS_KEYEDHASH_PARMS_ARRAY *new_TPMS_KEYEDHASH_PARMS_ARRAY(size_t nelements){
    return (TPMS_KEYEDHASH_PARMS *)malloc((nelements)*sizeof(TPMS_KEYEDHASH_PARMS));
  }
SWIGINTERN void delete_TPMS_KEYEDHASH_PARMS_ARRAY(TPMS_KEYEDHASH_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_KEYEDHASH_PARMS TPMS_KEYEDHASH_PARMS_ARRAY___getitem__(TPMS_KEYEDHASH_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_KEYEDHASH_PARMS_ARRAY___setitem__(TPMS_KEYEDHASH_PARMS_ARRAY *self,size_t index,TPMS_KEYEDHASH_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMS_KEYEDHASH_PARMS *TPMS_KEYEDHASH_PARMS_ARRAY_cast(TPMS_KEYEDHASH_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_KEYEDHASH_PARMS_ARRAY *TPMS_KEYEDHASH_PARMS_ARRAY_frompointer(TPMS_KEYEDHASH_PARMS *t){
    return (TPMS_KEYEDHASH_PARMS_ARRAY *)(t);
  }

  static TPMS_KEYEDHASH_PARMS *new_TPMS_KEYEDHASH_PARMS_PTR() { 
    return (TPMS_KEYEDHASH_PARMS *)malloc(sizeof(TPMS_KEYEDHASH_PARMS));
  }
  
  static TPMS_KEYEDHASH_PARMS *copy_TPMS_KEYEDHASH_PARMS_PTR(TPMS_KEYEDHASH_PARMS value) { 
    return (TPMS_KEYEDHASH_PARMS *)memcpy((TPMS_KEYEDHASH_PARMS *)malloc(sizeof(TPMS_KEYEDHASH_PARMS)),&value,sizeof(TPMS_KEYEDHASH_PARMS));
  }

  static void delete_TPMS_KEYEDHASH_PARMS_PTR(TPMS_KEYEDHASH_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_KEYEDHASH_PARMS_PTR_assign(TPMS_KEYEDHASH_PARMS *obj, TPMS_KEYEDHASH_PARMS value) {
    *obj = value;
  }

  static TPMS_KEYEDHASH_PARMS TPMS_KEYEDHASH_PARMS_PTR_value(TPMS_KEYEDHASH_PARMS *obj) {
    return *obj;
  }


  static TPMS_KEYEDHASH_PARMS * *new_TPMS_KEYEDHASH_PARMS_PTR_PTR() { 
    return (TPMS_KEYEDHASH_PARMS * *)malloc(sizeof(TPMS_KEYEDHASH_PARMS *));
  }
  
  static TPMS_KEYEDHASH_PARMS * *copy_TPMS_KEYEDHASH_PARMS_PTR_PTR(TPMS_KEYEDHASH_PARMS * value) { 
    return (TPMS_KEYEDHASH_PARMS * *)memcpy((TPMS_KEYEDHASH_PARMS * *)malloc(sizeof(TPMS_KEYEDHASH_PARMS *)),&value,sizeof(TPMS_KEYEDHASH_PARMS *));
  }

  static void delete_TPMS_KEYEDHASH_PARMS_PTR_PTR(TPMS_KEYEDHASH_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_KEYEDHASH_PARMS_PTR_PTR_assign(TPMS_KEYEDHASH_PARMS * *obj, TPMS_KEYEDHASH_PARMS * value) {
    *obj = value;
  }

  static TPMS_KEYEDHASH_PARMS * TPMS_KEYEDHASH_PARMS_PTR_PTR_value(TPMS_KEYEDHASH_PARMS * *obj) {
    return *obj;
  }


typedef TPMS_ASYM_PARMS TPMS_ASYM_PARMS_ARRAY;

SWIGINTERN TPMS_ASYM_PARMS_ARRAY *new_TPMS_ASYM_PARMS_ARRAY(size_t nelements){
    return (TPMS_ASYM_PARMS *)malloc((nelements)*sizeof(TPMS_ASYM_PARMS));
  }
SWIGINTERN void delete_TPMS_ASYM_PARMS_ARRAY(TPMS_ASYM_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ASYM_PARMS TPMS_ASYM_PARMS_ARRAY___getitem__(TPMS_ASYM_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ASYM_PARMS_ARRAY___setitem__(TPMS_ASYM_PARMS_ARRAY *self,size_t index,TPMS_ASYM_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMS_ASYM_PARMS *TPMS_ASYM_PARMS_ARRAY_cast(TPMS_ASYM_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ASYM_PARMS_ARRAY *TPMS_ASYM_PARMS_ARRAY_frompointer(TPMS_ASYM_PARMS *t){
    return (TPMS_ASYM_PARMS_ARRAY *)(t);
  }

  static TPMS_ASYM_PARMS *new_TPMS_ASYM_PARMS_PTR() { 
    return (TPMS_ASYM_PARMS *)malloc(sizeof(TPMS_ASYM_PARMS));
  }
  
  static TPMS_ASYM_PARMS *copy_TPMS_ASYM_PARMS_PTR(TPMS_ASYM_PARMS value) { 
    return (TPMS_ASYM_PARMS *)memcpy((TPMS_ASYM_PARMS *)malloc(sizeof(TPMS_ASYM_PARMS)),&value,sizeof(TPMS_ASYM_PARMS));
  }

  static void delete_TPMS_ASYM_PARMS_PTR(TPMS_ASYM_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ASYM_PARMS_PTR_assign(TPMS_ASYM_PARMS *obj, TPMS_ASYM_PARMS value) {
    *obj = value;
  }

  static TPMS_ASYM_PARMS TPMS_ASYM_PARMS_PTR_value(TPMS_ASYM_PARMS *obj) {
    return *obj;
  }


  static TPMS_ASYM_PARMS * *new_TPMS_ASYM_PARMS_PTR_PTR() { 
    return (TPMS_ASYM_PARMS * *)malloc(sizeof(TPMS_ASYM_PARMS *));
  }
  
  static TPMS_ASYM_PARMS * *copy_TPMS_ASYM_PARMS_PTR_PTR(TPMS_ASYM_PARMS * value) { 
    return (TPMS_ASYM_PARMS * *)memcpy((TPMS_ASYM_PARMS * *)malloc(sizeof(TPMS_ASYM_PARMS *)),&value,sizeof(TPMS_ASYM_PARMS *));
  }

  static void delete_TPMS_ASYM_PARMS_PTR_PTR(TPMS_ASYM_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ASYM_PARMS_PTR_PTR_assign(TPMS_ASYM_PARMS * *obj, TPMS_ASYM_PARMS * value) {
    *obj = value;
  }

  static TPMS_ASYM_PARMS * TPMS_ASYM_PARMS_PTR_PTR_value(TPMS_ASYM_PARMS * *obj) {
    return *obj;
  }


typedef TPMS_RSA_PARMS TPMS_RSA_PARMS_ARRAY;

SWIGINTERN TPMS_RSA_PARMS_ARRAY *new_TPMS_RSA_PARMS_ARRAY(size_t nelements){
    return (TPMS_RSA_PARMS *)malloc((nelements)*sizeof(TPMS_RSA_PARMS));
  }
SWIGINTERN void delete_TPMS_RSA_PARMS_ARRAY(TPMS_RSA_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_RSA_PARMS TPMS_RSA_PARMS_ARRAY___getitem__(TPMS_RSA_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_RSA_PARMS_ARRAY___setitem__(TPMS_RSA_PARMS_ARRAY *self,size_t index,TPMS_RSA_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMS_RSA_PARMS *TPMS_RSA_PARMS_ARRAY_cast(TPMS_RSA_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_RSA_PARMS_ARRAY *TPMS_RSA_PARMS_ARRAY_frompointer(TPMS_RSA_PARMS *t){
    return (TPMS_RSA_PARMS_ARRAY *)(t);
  }

  static TPMS_RSA_PARMS *new_TPMS_RSA_PARMS_PTR() { 
    return (TPMS_RSA_PARMS *)malloc(sizeof(TPMS_RSA_PARMS));
  }
  
  static TPMS_RSA_PARMS *copy_TPMS_RSA_PARMS_PTR(TPMS_RSA_PARMS value) { 
    return (TPMS_RSA_PARMS *)memcpy((TPMS_RSA_PARMS *)malloc(sizeof(TPMS_RSA_PARMS)),&value,sizeof(TPMS_RSA_PARMS));
  }

  static void delete_TPMS_RSA_PARMS_PTR(TPMS_RSA_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_RSA_PARMS_PTR_assign(TPMS_RSA_PARMS *obj, TPMS_RSA_PARMS value) {
    *obj = value;
  }

  static TPMS_RSA_PARMS TPMS_RSA_PARMS_PTR_value(TPMS_RSA_PARMS *obj) {
    return *obj;
  }


  static TPMS_RSA_PARMS * *new_TPMS_RSA_PARMS_PTR_PTR() { 
    return (TPMS_RSA_PARMS * *)malloc(sizeof(TPMS_RSA_PARMS *));
  }
  
  static TPMS_RSA_PARMS * *copy_TPMS_RSA_PARMS_PTR_PTR(TPMS_RSA_PARMS * value) { 
    return (TPMS_RSA_PARMS * *)memcpy((TPMS_RSA_PARMS * *)malloc(sizeof(TPMS_RSA_PARMS *)),&value,sizeof(TPMS_RSA_PARMS *));
  }

  static void delete_TPMS_RSA_PARMS_PTR_PTR(TPMS_RSA_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_RSA_PARMS_PTR_PTR_assign(TPMS_RSA_PARMS * *obj, TPMS_RSA_PARMS * value) {
    *obj = value;
  }

  static TPMS_RSA_PARMS * TPMS_RSA_PARMS_PTR_PTR_value(TPMS_RSA_PARMS * *obj) {
    return *obj;
  }


typedef TPMS_ECC_PARMS TPMS_ECC_PARMS_ARRAY;

SWIGINTERN TPMS_ECC_PARMS_ARRAY *new_TPMS_ECC_PARMS_ARRAY(size_t nelements){
    return (TPMS_ECC_PARMS *)malloc((nelements)*sizeof(TPMS_ECC_PARMS));
  }
SWIGINTERN void delete_TPMS_ECC_PARMS_ARRAY(TPMS_ECC_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ECC_PARMS TPMS_ECC_PARMS_ARRAY___getitem__(TPMS_ECC_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ECC_PARMS_ARRAY___setitem__(TPMS_ECC_PARMS_ARRAY *self,size_t index,TPMS_ECC_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMS_ECC_PARMS *TPMS_ECC_PARMS_ARRAY_cast(TPMS_ECC_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ECC_PARMS_ARRAY *TPMS_ECC_PARMS_ARRAY_frompointer(TPMS_ECC_PARMS *t){
    return (TPMS_ECC_PARMS_ARRAY *)(t);
  }

  static TPMS_ECC_PARMS *new_TPMS_ECC_PARMS_PTR() { 
    return (TPMS_ECC_PARMS *)malloc(sizeof(TPMS_ECC_PARMS));
  }
  
  static TPMS_ECC_PARMS *copy_TPMS_ECC_PARMS_PTR(TPMS_ECC_PARMS value) { 
    return (TPMS_ECC_PARMS *)memcpy((TPMS_ECC_PARMS *)malloc(sizeof(TPMS_ECC_PARMS)),&value,sizeof(TPMS_ECC_PARMS));
  }

  static void delete_TPMS_ECC_PARMS_PTR(TPMS_ECC_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ECC_PARMS_PTR_assign(TPMS_ECC_PARMS *obj, TPMS_ECC_PARMS value) {
    *obj = value;
  }

  static TPMS_ECC_PARMS TPMS_ECC_PARMS_PTR_value(TPMS_ECC_PARMS *obj) {
    return *obj;
  }


  static TPMS_ECC_PARMS * *new_TPMS_ECC_PARMS_PTR_PTR() { 
    return (TPMS_ECC_PARMS * *)malloc(sizeof(TPMS_ECC_PARMS *));
  }
  
  static TPMS_ECC_PARMS * *copy_TPMS_ECC_PARMS_PTR_PTR(TPMS_ECC_PARMS * value) { 
    return (TPMS_ECC_PARMS * *)memcpy((TPMS_ECC_PARMS * *)malloc(sizeof(TPMS_ECC_PARMS *)),&value,sizeof(TPMS_ECC_PARMS *));
  }

  static void delete_TPMS_ECC_PARMS_PTR_PTR(TPMS_ECC_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ECC_PARMS_PTR_PTR_assign(TPMS_ECC_PARMS * *obj, TPMS_ECC_PARMS * value) {
    *obj = value;
  }

  static TPMS_ECC_PARMS * TPMS_ECC_PARMS_PTR_PTR_value(TPMS_ECC_PARMS * *obj) {
    return *obj;
  }


typedef TPMU_PUBLIC_PARMS TPMU_PUBLIC_PARMS_ARRAY;

SWIGINTERN TPMU_PUBLIC_PARMS_ARRAY *new_TPMU_PUBLIC_PARMS_ARRAY(size_t nelements){
    return (TPMU_PUBLIC_PARMS *)malloc((nelements)*sizeof(TPMU_PUBLIC_PARMS));
  }
SWIGINTERN void delete_TPMU_PUBLIC_PARMS_ARRAY(TPMU_PUBLIC_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_PUBLIC_PARMS TPMU_PUBLIC_PARMS_ARRAY___getitem__(TPMU_PUBLIC_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_PUBLIC_PARMS_ARRAY___setitem__(TPMU_PUBLIC_PARMS_ARRAY *self,size_t index,TPMU_PUBLIC_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMU_PUBLIC_PARMS *TPMU_PUBLIC_PARMS_ARRAY_cast(TPMU_PUBLIC_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_PUBLIC_PARMS_ARRAY *TPMU_PUBLIC_PARMS_ARRAY_frompointer(TPMU_PUBLIC_PARMS *t){
    return (TPMU_PUBLIC_PARMS_ARRAY *)(t);
  }

  static TPMU_PUBLIC_PARMS *new_TPMU_PUBLIC_PARMS_PTR() { 
    return (TPMU_PUBLIC_PARMS *)malloc(sizeof(TPMU_PUBLIC_PARMS));
  }
  
  static TPMU_PUBLIC_PARMS *copy_TPMU_PUBLIC_PARMS_PTR(TPMU_PUBLIC_PARMS value) { 
    return (TPMU_PUBLIC_PARMS *)memcpy((TPMU_PUBLIC_PARMS *)malloc(sizeof(TPMU_PUBLIC_PARMS)),&value,sizeof(TPMU_PUBLIC_PARMS));
  }

  static void delete_TPMU_PUBLIC_PARMS_PTR(TPMU_PUBLIC_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_PUBLIC_PARMS_PTR_assign(TPMU_PUBLIC_PARMS *obj, TPMU_PUBLIC_PARMS value) {
    *obj = value;
  }

  static TPMU_PUBLIC_PARMS TPMU_PUBLIC_PARMS_PTR_value(TPMU_PUBLIC_PARMS *obj) {
    return *obj;
  }


  static TPMU_PUBLIC_PARMS * *new_TPMU_PUBLIC_PARMS_PTR_PTR() { 
    return (TPMU_PUBLIC_PARMS * *)malloc(sizeof(TPMU_PUBLIC_PARMS *));
  }
  
  static TPMU_PUBLIC_PARMS * *copy_TPMU_PUBLIC_PARMS_PTR_PTR(TPMU_PUBLIC_PARMS * value) { 
    return (TPMU_PUBLIC_PARMS * *)memcpy((TPMU_PUBLIC_PARMS * *)malloc(sizeof(TPMU_PUBLIC_PARMS *)),&value,sizeof(TPMU_PUBLIC_PARMS *));
  }

  static void delete_TPMU_PUBLIC_PARMS_PTR_PTR(TPMU_PUBLIC_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_PUBLIC_PARMS_PTR_PTR_assign(TPMU_PUBLIC_PARMS * *obj, TPMU_PUBLIC_PARMS * value) {
    *obj = value;
  }

  static TPMU_PUBLIC_PARMS * TPMU_PUBLIC_PARMS_PTR_PTR_value(TPMU_PUBLIC_PARMS * *obj) {
    return *obj;
  }


typedef TPMT_PUBLIC_PARMS TPMT_PUBLIC_PARMS_ARRAY;

SWIGINTERN TPMT_PUBLIC_PARMS_ARRAY *new_TPMT_PUBLIC_PARMS_ARRAY(size_t nelements){
    return (TPMT_PUBLIC_PARMS *)malloc((nelements)*sizeof(TPMT_PUBLIC_PARMS));
  }
SWIGINTERN void delete_TPMT_PUBLIC_PARMS_ARRAY(TPMT_PUBLIC_PARMS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_PUBLIC_PARMS TPMT_PUBLIC_PARMS_ARRAY___getitem__(TPMT_PUBLIC_PARMS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_PUBLIC_PARMS_ARRAY___setitem__(TPMT_PUBLIC_PARMS_ARRAY *self,size_t index,TPMT_PUBLIC_PARMS value){
    self[index] = value;
  }
SWIGINTERN TPMT_PUBLIC_PARMS *TPMT_PUBLIC_PARMS_ARRAY_cast(TPMT_PUBLIC_PARMS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_PUBLIC_PARMS_ARRAY *TPMT_PUBLIC_PARMS_ARRAY_frompointer(TPMT_PUBLIC_PARMS *t){
    return (TPMT_PUBLIC_PARMS_ARRAY *)(t);
  }

  static TPMT_PUBLIC_PARMS *new_TPMT_PUBLIC_PARMS_PTR() { 
    return (TPMT_PUBLIC_PARMS *)malloc(sizeof(TPMT_PUBLIC_PARMS));
  }
  
  static TPMT_PUBLIC_PARMS *copy_TPMT_PUBLIC_PARMS_PTR(TPMT_PUBLIC_PARMS value) { 
    return (TPMT_PUBLIC_PARMS *)memcpy((TPMT_PUBLIC_PARMS *)malloc(sizeof(TPMT_PUBLIC_PARMS)),&value,sizeof(TPMT_PUBLIC_PARMS));
  }

  static void delete_TPMT_PUBLIC_PARMS_PTR(TPMT_PUBLIC_PARMS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_PUBLIC_PARMS_PTR_assign(TPMT_PUBLIC_PARMS *obj, TPMT_PUBLIC_PARMS value) {
    *obj = value;
  }

  static TPMT_PUBLIC_PARMS TPMT_PUBLIC_PARMS_PTR_value(TPMT_PUBLIC_PARMS *obj) {
    return *obj;
  }


  static TPMT_PUBLIC_PARMS * *new_TPMT_PUBLIC_PARMS_PTR_PTR() { 
    return (TPMT_PUBLIC_PARMS * *)malloc(sizeof(TPMT_PUBLIC_PARMS *));
  }
  
  static TPMT_PUBLIC_PARMS * *copy_TPMT_PUBLIC_PARMS_PTR_PTR(TPMT_PUBLIC_PARMS * value) { 
    return (TPMT_PUBLIC_PARMS * *)memcpy((TPMT_PUBLIC_PARMS * *)malloc(sizeof(TPMT_PUBLIC_PARMS *)),&value,sizeof(TPMT_PUBLIC_PARMS *));
  }

  static void delete_TPMT_PUBLIC_PARMS_PTR_PTR(TPMT_PUBLIC_PARMS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_PUBLIC_PARMS_PTR_PTR_assign(TPMT_PUBLIC_PARMS * *obj, TPMT_PUBLIC_PARMS * value) {
    *obj = value;
  }

  static TPMT_PUBLIC_PARMS * TPMT_PUBLIC_PARMS_PTR_PTR_value(TPMT_PUBLIC_PARMS * *obj) {
    return *obj;
  }


typedef TPMT_PUBLIC TPMT_PUBLIC_ARRAY;

SWIGINTERN TPMT_PUBLIC_ARRAY *new_TPMT_PUBLIC_ARRAY(size_t nelements){
    return (TPMT_PUBLIC *)malloc((nelements)*sizeof(TPMT_PUBLIC));
  }
SWIGINTERN void delete_TPMT_PUBLIC_ARRAY(TPMT_PUBLIC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_PUBLIC TPMT_PUBLIC_ARRAY___getitem__(TPMT_PUBLIC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_PUBLIC_ARRAY___setitem__(TPMT_PUBLIC_ARRAY *self,size_t index,TPMT_PUBLIC value){
    self[index] = value;
  }
SWIGINTERN TPMT_PUBLIC *TPMT_PUBLIC_ARRAY_cast(TPMT_PUBLIC_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_PUBLIC_ARRAY *TPMT_PUBLIC_ARRAY_frompointer(TPMT_PUBLIC *t){
    return (TPMT_PUBLIC_ARRAY *)(t);
  }

  static TPMT_PUBLIC *new_TPMT_PUBLIC_PTR() { 
    return (TPMT_PUBLIC *)malloc(sizeof(TPMT_PUBLIC));
  }
  
  static TPMT_PUBLIC *copy_TPMT_PUBLIC_PTR(TPMT_PUBLIC value) { 
    return (TPMT_PUBLIC *)memcpy((TPMT_PUBLIC *)malloc(sizeof(TPMT_PUBLIC)),&value,sizeof(TPMT_PUBLIC));
  }

  static void delete_TPMT_PUBLIC_PTR(TPMT_PUBLIC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_PUBLIC_PTR_assign(TPMT_PUBLIC *obj, TPMT_PUBLIC value) {
    *obj = value;
  }

  static TPMT_PUBLIC TPMT_PUBLIC_PTR_value(TPMT_PUBLIC *obj) {
    return *obj;
  }


  static TPMT_PUBLIC * *new_TPMT_PUBLIC_PTR_PTR() { 
    return (TPMT_PUBLIC * *)malloc(sizeof(TPMT_PUBLIC *));
  }
  
  static TPMT_PUBLIC * *copy_TPMT_PUBLIC_PTR_PTR(TPMT_PUBLIC * value) { 
    return (TPMT_PUBLIC * *)memcpy((TPMT_PUBLIC * *)malloc(sizeof(TPMT_PUBLIC *)),&value,sizeof(TPMT_PUBLIC *));
  }

  static void delete_TPMT_PUBLIC_PTR_PTR(TPMT_PUBLIC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_PUBLIC_PTR_PTR_assign(TPMT_PUBLIC * *obj, TPMT_PUBLIC * value) {
    *obj = value;
  }

  static TPMT_PUBLIC * TPMT_PUBLIC_PTR_PTR_value(TPMT_PUBLIC * *obj) {
    return *obj;
  }


typedef TPM2B_PUBLIC TPM2B_PUBLIC_ARRAY;

SWIGINTERN TPM2B_PUBLIC_ARRAY *new_TPM2B_PUBLIC_ARRAY(size_t nelements){
    return (TPM2B_PUBLIC *)malloc((nelements)*sizeof(TPM2B_PUBLIC));
  }
SWIGINTERN void delete_TPM2B_PUBLIC_ARRAY(TPM2B_PUBLIC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_PUBLIC TPM2B_PUBLIC_ARRAY___getitem__(TPM2B_PUBLIC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_PUBLIC_ARRAY___setitem__(TPM2B_PUBLIC_ARRAY *self,size_t index,TPM2B_PUBLIC value){
    self[index] = value;
  }
SWIGINTERN TPM2B_PUBLIC *TPM2B_PUBLIC_ARRAY_cast(TPM2B_PUBLIC_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_PUBLIC_ARRAY *TPM2B_PUBLIC_ARRAY_frompointer(TPM2B_PUBLIC *t){
    return (TPM2B_PUBLIC_ARRAY *)(t);
  }

  static TPM2B_PUBLIC *new_TPM2B_PUBLIC_PTR() { 
    return (TPM2B_PUBLIC *)malloc(sizeof(TPM2B_PUBLIC));
  }
  
  static TPM2B_PUBLIC *copy_TPM2B_PUBLIC_PTR(TPM2B_PUBLIC value) { 
    return (TPM2B_PUBLIC *)memcpy((TPM2B_PUBLIC *)malloc(sizeof(TPM2B_PUBLIC)),&value,sizeof(TPM2B_PUBLIC));
  }

  static void delete_TPM2B_PUBLIC_PTR(TPM2B_PUBLIC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PUBLIC_PTR_assign(TPM2B_PUBLIC *obj, TPM2B_PUBLIC value) {
    *obj = value;
  }

  static TPM2B_PUBLIC TPM2B_PUBLIC_PTR_value(TPM2B_PUBLIC *obj) {
    return *obj;
  }


  static TPM2B_PUBLIC * *new_TPM2B_PUBLIC_PTR_PTR() { 
    return (TPM2B_PUBLIC * *)malloc(sizeof(TPM2B_PUBLIC *));
  }
  
  static TPM2B_PUBLIC * *copy_TPM2B_PUBLIC_PTR_PTR(TPM2B_PUBLIC * value) { 
    return (TPM2B_PUBLIC * *)memcpy((TPM2B_PUBLIC * *)malloc(sizeof(TPM2B_PUBLIC *)),&value,sizeof(TPM2B_PUBLIC *));
  }

  static void delete_TPM2B_PUBLIC_PTR_PTR(TPM2B_PUBLIC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PUBLIC_PTR_PTR_assign(TPM2B_PUBLIC * *obj, TPM2B_PUBLIC * value) {
    *obj = value;
  }

  static TPM2B_PUBLIC * TPM2B_PUBLIC_PTR_PTR_value(TPM2B_PUBLIC * *obj) {
    return *obj;
  }


typedef TPM2B_TEMPLATE TPM2B_TEMPLATE_ARRAY;

SWIGINTERN TPM2B_TEMPLATE_ARRAY *new_TPM2B_TEMPLATE_ARRAY(size_t nelements){
    return (TPM2B_TEMPLATE *)malloc((nelements)*sizeof(TPM2B_TEMPLATE));
  }
SWIGINTERN void delete_TPM2B_TEMPLATE_ARRAY(TPM2B_TEMPLATE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_TEMPLATE TPM2B_TEMPLATE_ARRAY___getitem__(TPM2B_TEMPLATE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_TEMPLATE_ARRAY___setitem__(TPM2B_TEMPLATE_ARRAY *self,size_t index,TPM2B_TEMPLATE value){
    self[index] = value;
  }
SWIGINTERN TPM2B_TEMPLATE *TPM2B_TEMPLATE_ARRAY_cast(TPM2B_TEMPLATE_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_TEMPLATE_ARRAY *TPM2B_TEMPLATE_ARRAY_frompointer(TPM2B_TEMPLATE *t){
    return (TPM2B_TEMPLATE_ARRAY *)(t);
  }

  static TPM2B_TEMPLATE *new_TPM2B_TEMPLATE_PTR() { 
    return (TPM2B_TEMPLATE *)malloc(sizeof(TPM2B_TEMPLATE));
  }
  
  static TPM2B_TEMPLATE *copy_TPM2B_TEMPLATE_PTR(TPM2B_TEMPLATE value) { 
    return (TPM2B_TEMPLATE *)memcpy((TPM2B_TEMPLATE *)malloc(sizeof(TPM2B_TEMPLATE)),&value,sizeof(TPM2B_TEMPLATE));
  }

  static void delete_TPM2B_TEMPLATE_PTR(TPM2B_TEMPLATE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_TEMPLATE_PTR_assign(TPM2B_TEMPLATE *obj, TPM2B_TEMPLATE value) {
    *obj = value;
  }

  static TPM2B_TEMPLATE TPM2B_TEMPLATE_PTR_value(TPM2B_TEMPLATE *obj) {
    return *obj;
  }


  static TPM2B_TEMPLATE * *new_TPM2B_TEMPLATE_PTR_PTR() { 
    return (TPM2B_TEMPLATE * *)malloc(sizeof(TPM2B_TEMPLATE *));
  }
  
  static TPM2B_TEMPLATE * *copy_TPM2B_TEMPLATE_PTR_PTR(TPM2B_TEMPLATE * value) { 
    return (TPM2B_TEMPLATE * *)memcpy((TPM2B_TEMPLATE * *)malloc(sizeof(TPM2B_TEMPLATE *)),&value,sizeof(TPM2B_TEMPLATE *));
  }

  static void delete_TPM2B_TEMPLATE_PTR_PTR(TPM2B_TEMPLATE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_TEMPLATE_PTR_PTR_assign(TPM2B_TEMPLATE * *obj, TPM2B_TEMPLATE * value) {
    *obj = value;
  }

  static TPM2B_TEMPLATE * TPM2B_TEMPLATE_PTR_PTR_value(TPM2B_TEMPLATE * *obj) {
    return *obj;
  }


typedef TPM2B_PRIVATE_VENDOR_SPECIFIC TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY;

SWIGINTERN TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY(size_t nelements){
    return (TPM2B_PRIVATE_VENDOR_SPECIFIC *)malloc((nelements)*sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC));
  }
SWIGINTERN void delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY(TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_PRIVATE_VENDOR_SPECIFIC TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__(TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__(TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *self,size_t index,TPM2B_PRIVATE_VENDOR_SPECIFIC value){
    self[index] = value;
  }
SWIGINTERN TPM2B_PRIVATE_VENDOR_SPECIFIC *TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast(TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer(TPM2B_PRIVATE_VENDOR_SPECIFIC *t){
    return (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)(t);
  }

  static TPM2B_PRIVATE_VENDOR_SPECIFIC *new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR() { 
    return (TPM2B_PRIVATE_VENDOR_SPECIFIC *)malloc(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC));
  }
  
  static TPM2B_PRIVATE_VENDOR_SPECIFIC *copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(TPM2B_PRIVATE_VENDOR_SPECIFIC value) { 
    return (TPM2B_PRIVATE_VENDOR_SPECIFIC *)memcpy((TPM2B_PRIVATE_VENDOR_SPECIFIC *)malloc(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC)),&value,sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC));
  }

  static void delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(TPM2B_PRIVATE_VENDOR_SPECIFIC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign(TPM2B_PRIVATE_VENDOR_SPECIFIC *obj, TPM2B_PRIVATE_VENDOR_SPECIFIC value) {
    *obj = value;
  }

  static TPM2B_PRIVATE_VENDOR_SPECIFIC TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value(TPM2B_PRIVATE_VENDOR_SPECIFIC *obj) {
    return *obj;
  }


  static TPM2B_PRIVATE_VENDOR_SPECIFIC * *new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR() { 
    return (TPM2B_PRIVATE_VENDOR_SPECIFIC * *)malloc(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC *));
  }
  
  static TPM2B_PRIVATE_VENDOR_SPECIFIC * *copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(TPM2B_PRIVATE_VENDOR_SPECIFIC * value) { 
    return (TPM2B_PRIVATE_VENDOR_SPECIFIC * *)memcpy((TPM2B_PRIVATE_VENDOR_SPECIFIC * *)malloc(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC *)),&value,sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC *));
  }

  static void delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(TPM2B_PRIVATE_VENDOR_SPECIFIC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign(TPM2B_PRIVATE_VENDOR_SPECIFIC * *obj, TPM2B_PRIVATE_VENDOR_SPECIFIC * value) {
    *obj = value;
  }

  static TPM2B_PRIVATE_VENDOR_SPECIFIC * TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value(TPM2B_PRIVATE_VENDOR_SPECIFIC * *obj) {
    return *obj;
  }


typedef TPMU_SENSITIVE_COMPOSITE TPMU_SENSITIVE_COMPOSITE_ARRAY;

SWIGINTERN TPMU_SENSITIVE_COMPOSITE_ARRAY *new_TPMU_SENSITIVE_COMPOSITE_ARRAY(size_t nelements){
    return (TPMU_SENSITIVE_COMPOSITE *)malloc((nelements)*sizeof(TPMU_SENSITIVE_COMPOSITE));
  }
SWIGINTERN void delete_TPMU_SENSITIVE_COMPOSITE_ARRAY(TPMU_SENSITIVE_COMPOSITE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMU_SENSITIVE_COMPOSITE TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__(TPMU_SENSITIVE_COMPOSITE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__(TPMU_SENSITIVE_COMPOSITE_ARRAY *self,size_t index,TPMU_SENSITIVE_COMPOSITE value){
    self[index] = value;
  }
SWIGINTERN TPMU_SENSITIVE_COMPOSITE *TPMU_SENSITIVE_COMPOSITE_ARRAY_cast(TPMU_SENSITIVE_COMPOSITE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMU_SENSITIVE_COMPOSITE_ARRAY *TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer(TPMU_SENSITIVE_COMPOSITE *t){
    return (TPMU_SENSITIVE_COMPOSITE_ARRAY *)(t);
  }

  static TPMU_SENSITIVE_COMPOSITE *new_TPMU_SENSITIVE_COMPOSITE_PTR() { 
    return (TPMU_SENSITIVE_COMPOSITE *)malloc(sizeof(TPMU_SENSITIVE_COMPOSITE));
  }
  
  static TPMU_SENSITIVE_COMPOSITE *copy_TPMU_SENSITIVE_COMPOSITE_PTR(TPMU_SENSITIVE_COMPOSITE value) { 
    return (TPMU_SENSITIVE_COMPOSITE *)memcpy((TPMU_SENSITIVE_COMPOSITE *)malloc(sizeof(TPMU_SENSITIVE_COMPOSITE)),&value,sizeof(TPMU_SENSITIVE_COMPOSITE));
  }

  static void delete_TPMU_SENSITIVE_COMPOSITE_PTR(TPMU_SENSITIVE_COMPOSITE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SENSITIVE_COMPOSITE_PTR_assign(TPMU_SENSITIVE_COMPOSITE *obj, TPMU_SENSITIVE_COMPOSITE value) {
    *obj = value;
  }

  static TPMU_SENSITIVE_COMPOSITE TPMU_SENSITIVE_COMPOSITE_PTR_value(TPMU_SENSITIVE_COMPOSITE *obj) {
    return *obj;
  }


  static TPMU_SENSITIVE_COMPOSITE * *new_TPMU_SENSITIVE_COMPOSITE_PTR_PTR() { 
    return (TPMU_SENSITIVE_COMPOSITE * *)malloc(sizeof(TPMU_SENSITIVE_COMPOSITE *));
  }
  
  static TPMU_SENSITIVE_COMPOSITE * *copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(TPMU_SENSITIVE_COMPOSITE * value) { 
    return (TPMU_SENSITIVE_COMPOSITE * *)memcpy((TPMU_SENSITIVE_COMPOSITE * *)malloc(sizeof(TPMU_SENSITIVE_COMPOSITE *)),&value,sizeof(TPMU_SENSITIVE_COMPOSITE *));
  }

  static void delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(TPMU_SENSITIVE_COMPOSITE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign(TPMU_SENSITIVE_COMPOSITE * *obj, TPMU_SENSITIVE_COMPOSITE * value) {
    *obj = value;
  }

  static TPMU_SENSITIVE_COMPOSITE * TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value(TPMU_SENSITIVE_COMPOSITE * *obj) {
    return *obj;
  }


typedef TPMT_SENSITIVE TPMT_SENSITIVE_ARRAY;

SWIGINTERN TPMT_SENSITIVE_ARRAY *new_TPMT_SENSITIVE_ARRAY(size_t nelements){
    return (TPMT_SENSITIVE *)malloc((nelements)*sizeof(TPMT_SENSITIVE));
  }
SWIGINTERN void delete_TPMT_SENSITIVE_ARRAY(TPMT_SENSITIVE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMT_SENSITIVE TPMT_SENSITIVE_ARRAY___getitem__(TPMT_SENSITIVE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMT_SENSITIVE_ARRAY___setitem__(TPMT_SENSITIVE_ARRAY *self,size_t index,TPMT_SENSITIVE value){
    self[index] = value;
  }
SWIGINTERN TPMT_SENSITIVE *TPMT_SENSITIVE_ARRAY_cast(TPMT_SENSITIVE_ARRAY *self){
    return self;
  }
SWIGINTERN TPMT_SENSITIVE_ARRAY *TPMT_SENSITIVE_ARRAY_frompointer(TPMT_SENSITIVE *t){
    return (TPMT_SENSITIVE_ARRAY *)(t);
  }

  static TPMT_SENSITIVE *new_TPMT_SENSITIVE_PTR() { 
    return (TPMT_SENSITIVE *)malloc(sizeof(TPMT_SENSITIVE));
  }
  
  static TPMT_SENSITIVE *copy_TPMT_SENSITIVE_PTR(TPMT_SENSITIVE value) { 
    return (TPMT_SENSITIVE *)memcpy((TPMT_SENSITIVE *)malloc(sizeof(TPMT_SENSITIVE)),&value,sizeof(TPMT_SENSITIVE));
  }

  static void delete_TPMT_SENSITIVE_PTR(TPMT_SENSITIVE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SENSITIVE_PTR_assign(TPMT_SENSITIVE *obj, TPMT_SENSITIVE value) {
    *obj = value;
  }

  static TPMT_SENSITIVE TPMT_SENSITIVE_PTR_value(TPMT_SENSITIVE *obj) {
    return *obj;
  }


  static TPMT_SENSITIVE * *new_TPMT_SENSITIVE_PTR_PTR() { 
    return (TPMT_SENSITIVE * *)malloc(sizeof(TPMT_SENSITIVE *));
  }
  
  static TPMT_SENSITIVE * *copy_TPMT_SENSITIVE_PTR_PTR(TPMT_SENSITIVE * value) { 
    return (TPMT_SENSITIVE * *)memcpy((TPMT_SENSITIVE * *)malloc(sizeof(TPMT_SENSITIVE *)),&value,sizeof(TPMT_SENSITIVE *));
  }

  static void delete_TPMT_SENSITIVE_PTR_PTR(TPMT_SENSITIVE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMT_SENSITIVE_PTR_PTR_assign(TPMT_SENSITIVE * *obj, TPMT_SENSITIVE * value) {
    *obj = value;
  }

  static TPMT_SENSITIVE * TPMT_SENSITIVE_PTR_PTR_value(TPMT_SENSITIVE * *obj) {
    return *obj;
  }


typedef TPM2B_SENSITIVE TPM2B_SENSITIVE_ARRAY;

SWIGINTERN TPM2B_SENSITIVE_ARRAY *new_TPM2B_SENSITIVE_ARRAY(size_t nelements){
    return (TPM2B_SENSITIVE *)malloc((nelements)*sizeof(TPM2B_SENSITIVE));
  }
SWIGINTERN void delete_TPM2B_SENSITIVE_ARRAY(TPM2B_SENSITIVE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_SENSITIVE TPM2B_SENSITIVE_ARRAY___getitem__(TPM2B_SENSITIVE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_SENSITIVE_ARRAY___setitem__(TPM2B_SENSITIVE_ARRAY *self,size_t index,TPM2B_SENSITIVE value){
    self[index] = value;
  }
SWIGINTERN TPM2B_SENSITIVE *TPM2B_SENSITIVE_ARRAY_cast(TPM2B_SENSITIVE_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_SENSITIVE_ARRAY *TPM2B_SENSITIVE_ARRAY_frompointer(TPM2B_SENSITIVE *t){
    return (TPM2B_SENSITIVE_ARRAY *)(t);
  }

  static TPM2B_SENSITIVE *new_TPM2B_SENSITIVE_PTR() { 
    return (TPM2B_SENSITIVE *)malloc(sizeof(TPM2B_SENSITIVE));
  }
  
  static TPM2B_SENSITIVE *copy_TPM2B_SENSITIVE_PTR(TPM2B_SENSITIVE value) { 
    return (TPM2B_SENSITIVE *)memcpy((TPM2B_SENSITIVE *)malloc(sizeof(TPM2B_SENSITIVE)),&value,sizeof(TPM2B_SENSITIVE));
  }

  static void delete_TPM2B_SENSITIVE_PTR(TPM2B_SENSITIVE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SENSITIVE_PTR_assign(TPM2B_SENSITIVE *obj, TPM2B_SENSITIVE value) {
    *obj = value;
  }

  static TPM2B_SENSITIVE TPM2B_SENSITIVE_PTR_value(TPM2B_SENSITIVE *obj) {
    return *obj;
  }


  static TPM2B_SENSITIVE * *new_TPM2B_SENSITIVE_PTR_PTR() { 
    return (TPM2B_SENSITIVE * *)malloc(sizeof(TPM2B_SENSITIVE *));
  }
  
  static TPM2B_SENSITIVE * *copy_TPM2B_SENSITIVE_PTR_PTR(TPM2B_SENSITIVE * value) { 
    return (TPM2B_SENSITIVE * *)memcpy((TPM2B_SENSITIVE * *)malloc(sizeof(TPM2B_SENSITIVE *)),&value,sizeof(TPM2B_SENSITIVE *));
  }

  static void delete_TPM2B_SENSITIVE_PTR_PTR(TPM2B_SENSITIVE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_SENSITIVE_PTR_PTR_assign(TPM2B_SENSITIVE * *obj, TPM2B_SENSITIVE * value) {
    *obj = value;
  }

  static TPM2B_SENSITIVE * TPM2B_SENSITIVE_PTR_PTR_value(TPM2B_SENSITIVE * *obj) {
    return *obj;
  }


typedef _PRIVATE _PRIVATE_ARRAY;

SWIGINTERN _PRIVATE_ARRAY *new__PRIVATE_ARRAY(size_t nelements){
    return (_PRIVATE *)malloc((nelements)*sizeof(_PRIVATE));
  }
SWIGINTERN void delete__PRIVATE_ARRAY(_PRIVATE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN _PRIVATE _PRIVATE_ARRAY___getitem__(_PRIVATE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void _PRIVATE_ARRAY___setitem__(_PRIVATE_ARRAY *self,size_t index,_PRIVATE value){
    self[index] = value;
  }
SWIGINTERN _PRIVATE *_PRIVATE_ARRAY_cast(_PRIVATE_ARRAY *self){
    return self;
  }
SWIGINTERN _PRIVATE_ARRAY *_PRIVATE_ARRAY_frompointer(_PRIVATE *t){
    return (_PRIVATE_ARRAY *)(t);
  }

  static _PRIVATE *new__PRIVATE_PTR() { 
    return (_PRIVATE *)malloc(sizeof(_PRIVATE));
  }
  
  static _PRIVATE *copy__PRIVATE_PTR(_PRIVATE value) { 
    return (_PRIVATE *)memcpy((_PRIVATE *)malloc(sizeof(_PRIVATE)),&value,sizeof(_PRIVATE));
  }

  static void delete__PRIVATE_PTR(_PRIVATE *obj) { 
    if (obj) free((char*)obj);
  }

  static void _PRIVATE_PTR_assign(_PRIVATE *obj, _PRIVATE value) {
    *obj = value;
  }

  static _PRIVATE _PRIVATE_PTR_value(_PRIVATE *obj) {
    return *obj;
  }


  static _PRIVATE * *new__PRIVATE_PTR_PTR() { 
    return (_PRIVATE * *)malloc(sizeof(_PRIVATE *));
  }
  
  static _PRIVATE * *copy__PRIVATE_PTR_PTR(_PRIVATE * value) { 
    return (_PRIVATE * *)memcpy((_PRIVATE * *)malloc(sizeof(_PRIVATE *)),&value,sizeof(_PRIVATE *));
  }

  static void delete__PRIVATE_PTR_PTR(_PRIVATE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void _PRIVATE_PTR_PTR_assign(_PRIVATE * *obj, _PRIVATE * value) {
    *obj = value;
  }

  static _PRIVATE * _PRIVATE_PTR_PTR_value(_PRIVATE * *obj) {
    return *obj;
  }


typedef TPM2B_PRIVATE TPM2B_PRIVATE_ARRAY;

SWIGINTERN TPM2B_PRIVATE_ARRAY *new_TPM2B_PRIVATE_ARRAY(size_t nelements){
    return (TPM2B_PRIVATE *)malloc((nelements)*sizeof(TPM2B_PRIVATE));
  }
SWIGINTERN void delete_TPM2B_PRIVATE_ARRAY(TPM2B_PRIVATE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_PRIVATE TPM2B_PRIVATE_ARRAY___getitem__(TPM2B_PRIVATE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_PRIVATE_ARRAY___setitem__(TPM2B_PRIVATE_ARRAY *self,size_t index,TPM2B_PRIVATE value){
    self[index] = value;
  }
SWIGINTERN TPM2B_PRIVATE *TPM2B_PRIVATE_ARRAY_cast(TPM2B_PRIVATE_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_PRIVATE_ARRAY *TPM2B_PRIVATE_ARRAY_frompointer(TPM2B_PRIVATE *t){
    return (TPM2B_PRIVATE_ARRAY *)(t);
  }

  static TPM2B_PRIVATE *new_TPM2B_PRIVATE_PTR() { 
    return (TPM2B_PRIVATE *)malloc(sizeof(TPM2B_PRIVATE));
  }
  
  static TPM2B_PRIVATE *copy_TPM2B_PRIVATE_PTR(TPM2B_PRIVATE value) { 
    return (TPM2B_PRIVATE *)memcpy((TPM2B_PRIVATE *)malloc(sizeof(TPM2B_PRIVATE)),&value,sizeof(TPM2B_PRIVATE));
  }

  static void delete_TPM2B_PRIVATE_PTR(TPM2B_PRIVATE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PRIVATE_PTR_assign(TPM2B_PRIVATE *obj, TPM2B_PRIVATE value) {
    *obj = value;
  }

  static TPM2B_PRIVATE TPM2B_PRIVATE_PTR_value(TPM2B_PRIVATE *obj) {
    return *obj;
  }


  static TPM2B_PRIVATE * *new_TPM2B_PRIVATE_PTR_PTR() { 
    return (TPM2B_PRIVATE * *)malloc(sizeof(TPM2B_PRIVATE *));
  }
  
  static TPM2B_PRIVATE * *copy_TPM2B_PRIVATE_PTR_PTR(TPM2B_PRIVATE * value) { 
    return (TPM2B_PRIVATE * *)memcpy((TPM2B_PRIVATE * *)malloc(sizeof(TPM2B_PRIVATE *)),&value,sizeof(TPM2B_PRIVATE *));
  }

  static void delete_TPM2B_PRIVATE_PTR_PTR(TPM2B_PRIVATE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_PRIVATE_PTR_PTR_assign(TPM2B_PRIVATE * *obj, TPM2B_PRIVATE * value) {
    *obj = value;
  }

  static TPM2B_PRIVATE * TPM2B_PRIVATE_PTR_PTR_value(TPM2B_PRIVATE * *obj) {
    return *obj;
  }


typedef TPMS_ID_OBJECT TPMS_ID_OBJECT_ARRAY;

SWIGINTERN TPMS_ID_OBJECT_ARRAY *new_TPMS_ID_OBJECT_ARRAY(size_t nelements){
    return (TPMS_ID_OBJECT *)malloc((nelements)*sizeof(TPMS_ID_OBJECT));
  }
SWIGINTERN void delete_TPMS_ID_OBJECT_ARRAY(TPMS_ID_OBJECT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_ID_OBJECT TPMS_ID_OBJECT_ARRAY___getitem__(TPMS_ID_OBJECT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_ID_OBJECT_ARRAY___setitem__(TPMS_ID_OBJECT_ARRAY *self,size_t index,TPMS_ID_OBJECT value){
    self[index] = value;
  }
SWIGINTERN TPMS_ID_OBJECT *TPMS_ID_OBJECT_ARRAY_cast(TPMS_ID_OBJECT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_ID_OBJECT_ARRAY *TPMS_ID_OBJECT_ARRAY_frompointer(TPMS_ID_OBJECT *t){
    return (TPMS_ID_OBJECT_ARRAY *)(t);
  }

  static TPMS_ID_OBJECT *new_TPMS_ID_OBJECT_PTR() { 
    return (TPMS_ID_OBJECT *)malloc(sizeof(TPMS_ID_OBJECT));
  }
  
  static TPMS_ID_OBJECT *copy_TPMS_ID_OBJECT_PTR(TPMS_ID_OBJECT value) { 
    return (TPMS_ID_OBJECT *)memcpy((TPMS_ID_OBJECT *)malloc(sizeof(TPMS_ID_OBJECT)),&value,sizeof(TPMS_ID_OBJECT));
  }

  static void delete_TPMS_ID_OBJECT_PTR(TPMS_ID_OBJECT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ID_OBJECT_PTR_assign(TPMS_ID_OBJECT *obj, TPMS_ID_OBJECT value) {
    *obj = value;
  }

  static TPMS_ID_OBJECT TPMS_ID_OBJECT_PTR_value(TPMS_ID_OBJECT *obj) {
    return *obj;
  }


  static TPMS_ID_OBJECT * *new_TPMS_ID_OBJECT_PTR_PTR() { 
    return (TPMS_ID_OBJECT * *)malloc(sizeof(TPMS_ID_OBJECT *));
  }
  
  static TPMS_ID_OBJECT * *copy_TPMS_ID_OBJECT_PTR_PTR(TPMS_ID_OBJECT * value) { 
    return (TPMS_ID_OBJECT * *)memcpy((TPMS_ID_OBJECT * *)malloc(sizeof(TPMS_ID_OBJECT *)),&value,sizeof(TPMS_ID_OBJECT *));
  }

  static void delete_TPMS_ID_OBJECT_PTR_PTR(TPMS_ID_OBJECT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_ID_OBJECT_PTR_PTR_assign(TPMS_ID_OBJECT * *obj, TPMS_ID_OBJECT * value) {
    *obj = value;
  }

  static TPMS_ID_OBJECT * TPMS_ID_OBJECT_PTR_PTR_value(TPMS_ID_OBJECT * *obj) {
    return *obj;
  }


typedef TPM2B_ID_OBJECT TPM2B_ID_OBJECT_ARRAY;

SWIGINTERN TPM2B_ID_OBJECT_ARRAY *new_TPM2B_ID_OBJECT_ARRAY(size_t nelements){
    return (TPM2B_ID_OBJECT *)malloc((nelements)*sizeof(TPM2B_ID_OBJECT));
  }
SWIGINTERN void delete_TPM2B_ID_OBJECT_ARRAY(TPM2B_ID_OBJECT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_ID_OBJECT TPM2B_ID_OBJECT_ARRAY___getitem__(TPM2B_ID_OBJECT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_ID_OBJECT_ARRAY___setitem__(TPM2B_ID_OBJECT_ARRAY *self,size_t index,TPM2B_ID_OBJECT value){
    self[index] = value;
  }
SWIGINTERN TPM2B_ID_OBJECT *TPM2B_ID_OBJECT_ARRAY_cast(TPM2B_ID_OBJECT_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_ID_OBJECT_ARRAY *TPM2B_ID_OBJECT_ARRAY_frompointer(TPM2B_ID_OBJECT *t){
    return (TPM2B_ID_OBJECT_ARRAY *)(t);
  }

  static TPM2B_ID_OBJECT *new_TPM2B_ID_OBJECT_PTR() { 
    return (TPM2B_ID_OBJECT *)malloc(sizeof(TPM2B_ID_OBJECT));
  }
  
  static TPM2B_ID_OBJECT *copy_TPM2B_ID_OBJECT_PTR(TPM2B_ID_OBJECT value) { 
    return (TPM2B_ID_OBJECT *)memcpy((TPM2B_ID_OBJECT *)malloc(sizeof(TPM2B_ID_OBJECT)),&value,sizeof(TPM2B_ID_OBJECT));
  }

  static void delete_TPM2B_ID_OBJECT_PTR(TPM2B_ID_OBJECT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ID_OBJECT_PTR_assign(TPM2B_ID_OBJECT *obj, TPM2B_ID_OBJECT value) {
    *obj = value;
  }

  static TPM2B_ID_OBJECT TPM2B_ID_OBJECT_PTR_value(TPM2B_ID_OBJECT *obj) {
    return *obj;
  }


  static TPM2B_ID_OBJECT * *new_TPM2B_ID_OBJECT_PTR_PTR() { 
    return (TPM2B_ID_OBJECT * *)malloc(sizeof(TPM2B_ID_OBJECT *));
  }
  
  static TPM2B_ID_OBJECT * *copy_TPM2B_ID_OBJECT_PTR_PTR(TPM2B_ID_OBJECT * value) { 
    return (TPM2B_ID_OBJECT * *)memcpy((TPM2B_ID_OBJECT * *)malloc(sizeof(TPM2B_ID_OBJECT *)),&value,sizeof(TPM2B_ID_OBJECT *));
  }

  static void delete_TPM2B_ID_OBJECT_PTR_PTR(TPM2B_ID_OBJECT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_ID_OBJECT_PTR_PTR_assign(TPM2B_ID_OBJECT * *obj, TPM2B_ID_OBJECT * value) {
    *obj = value;
  }

  static TPM2B_ID_OBJECT * TPM2B_ID_OBJECT_PTR_PTR_value(TPM2B_ID_OBJECT * *obj) {
    return *obj;
  }


typedef TPMS_NV_PIN_COUNTER_PARAMETERS TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY;

SWIGINTERN TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY(size_t nelements){
    return (TPMS_NV_PIN_COUNTER_PARAMETERS *)malloc((nelements)*sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS));
  }
SWIGINTERN void delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY(TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_NV_PIN_COUNTER_PARAMETERS TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__(TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__(TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *self,size_t index,TPMS_NV_PIN_COUNTER_PARAMETERS value){
    self[index] = value;
  }
SWIGINTERN TPMS_NV_PIN_COUNTER_PARAMETERS *TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast(TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer(TPMS_NV_PIN_COUNTER_PARAMETERS *t){
    return (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)(t);
  }

  static TPMS_NV_PIN_COUNTER_PARAMETERS *new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR() { 
    return (TPMS_NV_PIN_COUNTER_PARAMETERS *)malloc(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS));
  }
  
  static TPMS_NV_PIN_COUNTER_PARAMETERS *copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(TPMS_NV_PIN_COUNTER_PARAMETERS value) { 
    return (TPMS_NV_PIN_COUNTER_PARAMETERS *)memcpy((TPMS_NV_PIN_COUNTER_PARAMETERS *)malloc(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS)),&value,sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS));
  }

  static void delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(TPMS_NV_PIN_COUNTER_PARAMETERS *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign(TPMS_NV_PIN_COUNTER_PARAMETERS *obj, TPMS_NV_PIN_COUNTER_PARAMETERS value) {
    *obj = value;
  }

  static TPMS_NV_PIN_COUNTER_PARAMETERS TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value(TPMS_NV_PIN_COUNTER_PARAMETERS *obj) {
    return *obj;
  }


  static TPMS_NV_PIN_COUNTER_PARAMETERS * *new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR() { 
    return (TPMS_NV_PIN_COUNTER_PARAMETERS * *)malloc(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS *));
  }
  
  static TPMS_NV_PIN_COUNTER_PARAMETERS * *copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(TPMS_NV_PIN_COUNTER_PARAMETERS * value) { 
    return (TPMS_NV_PIN_COUNTER_PARAMETERS * *)memcpy((TPMS_NV_PIN_COUNTER_PARAMETERS * *)malloc(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS *)),&value,sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS *));
  }

  static void delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(TPMS_NV_PIN_COUNTER_PARAMETERS * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign(TPMS_NV_PIN_COUNTER_PARAMETERS * *obj, TPMS_NV_PIN_COUNTER_PARAMETERS * value) {
    *obj = value;
  }

  static TPMS_NV_PIN_COUNTER_PARAMETERS * TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value(TPMS_NV_PIN_COUNTER_PARAMETERS * *obj) {
    return *obj;
  }


typedef TPMS_NV_PUBLIC TPMS_NV_PUBLIC_ARRAY;

SWIGINTERN TPMS_NV_PUBLIC_ARRAY *new_TPMS_NV_PUBLIC_ARRAY(size_t nelements){
    return (TPMS_NV_PUBLIC *)malloc((nelements)*sizeof(TPMS_NV_PUBLIC));
  }
SWIGINTERN void delete_TPMS_NV_PUBLIC_ARRAY(TPMS_NV_PUBLIC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_NV_PUBLIC TPMS_NV_PUBLIC_ARRAY___getitem__(TPMS_NV_PUBLIC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_NV_PUBLIC_ARRAY___setitem__(TPMS_NV_PUBLIC_ARRAY *self,size_t index,TPMS_NV_PUBLIC value){
    self[index] = value;
  }
SWIGINTERN TPMS_NV_PUBLIC *TPMS_NV_PUBLIC_ARRAY_cast(TPMS_NV_PUBLIC_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_NV_PUBLIC_ARRAY *TPMS_NV_PUBLIC_ARRAY_frompointer(TPMS_NV_PUBLIC *t){
    return (TPMS_NV_PUBLIC_ARRAY *)(t);
  }

  static TPMS_NV_PUBLIC *new_TPMS_NV_PUBLIC_PTR() { 
    return (TPMS_NV_PUBLIC *)malloc(sizeof(TPMS_NV_PUBLIC));
  }
  
  static TPMS_NV_PUBLIC *copy_TPMS_NV_PUBLIC_PTR(TPMS_NV_PUBLIC value) { 
    return (TPMS_NV_PUBLIC *)memcpy((TPMS_NV_PUBLIC *)malloc(sizeof(TPMS_NV_PUBLIC)),&value,sizeof(TPMS_NV_PUBLIC));
  }

  static void delete_TPMS_NV_PUBLIC_PTR(TPMS_NV_PUBLIC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_NV_PUBLIC_PTR_assign(TPMS_NV_PUBLIC *obj, TPMS_NV_PUBLIC value) {
    *obj = value;
  }

  static TPMS_NV_PUBLIC TPMS_NV_PUBLIC_PTR_value(TPMS_NV_PUBLIC *obj) {
    return *obj;
  }


  static TPMS_NV_PUBLIC * *new_TPMS_NV_PUBLIC_PTR_PTR() { 
    return (TPMS_NV_PUBLIC * *)malloc(sizeof(TPMS_NV_PUBLIC *));
  }
  
  static TPMS_NV_PUBLIC * *copy_TPMS_NV_PUBLIC_PTR_PTR(TPMS_NV_PUBLIC * value) { 
    return (TPMS_NV_PUBLIC * *)memcpy((TPMS_NV_PUBLIC * *)malloc(sizeof(TPMS_NV_PUBLIC *)),&value,sizeof(TPMS_NV_PUBLIC *));
  }

  static void delete_TPMS_NV_PUBLIC_PTR_PTR(TPMS_NV_PUBLIC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_NV_PUBLIC_PTR_PTR_assign(TPMS_NV_PUBLIC * *obj, TPMS_NV_PUBLIC * value) {
    *obj = value;
  }

  static TPMS_NV_PUBLIC * TPMS_NV_PUBLIC_PTR_PTR_value(TPMS_NV_PUBLIC * *obj) {
    return *obj;
  }


typedef TPM2B_NV_PUBLIC TPM2B_NV_PUBLIC_ARRAY;

SWIGINTERN TPM2B_NV_PUBLIC_ARRAY *new_TPM2B_NV_PUBLIC_ARRAY(size_t nelements){
    return (TPM2B_NV_PUBLIC *)malloc((nelements)*sizeof(TPM2B_NV_PUBLIC));
  }
SWIGINTERN void delete_TPM2B_NV_PUBLIC_ARRAY(TPM2B_NV_PUBLIC_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_NV_PUBLIC TPM2B_NV_PUBLIC_ARRAY___getitem__(TPM2B_NV_PUBLIC_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_NV_PUBLIC_ARRAY___setitem__(TPM2B_NV_PUBLIC_ARRAY *self,size_t index,TPM2B_NV_PUBLIC value){
    self[index] = value;
  }
SWIGINTERN TPM2B_NV_PUBLIC *TPM2B_NV_PUBLIC_ARRAY_cast(TPM2B_NV_PUBLIC_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_NV_PUBLIC_ARRAY *TPM2B_NV_PUBLIC_ARRAY_frompointer(TPM2B_NV_PUBLIC *t){
    return (TPM2B_NV_PUBLIC_ARRAY *)(t);
  }

  static TPM2B_NV_PUBLIC *new_TPM2B_NV_PUBLIC_PTR() { 
    return (TPM2B_NV_PUBLIC *)malloc(sizeof(TPM2B_NV_PUBLIC));
  }
  
  static TPM2B_NV_PUBLIC *copy_TPM2B_NV_PUBLIC_PTR(TPM2B_NV_PUBLIC value) { 
    return (TPM2B_NV_PUBLIC *)memcpy((TPM2B_NV_PUBLIC *)malloc(sizeof(TPM2B_NV_PUBLIC)),&value,sizeof(TPM2B_NV_PUBLIC));
  }

  static void delete_TPM2B_NV_PUBLIC_PTR(TPM2B_NV_PUBLIC *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_NV_PUBLIC_PTR_assign(TPM2B_NV_PUBLIC *obj, TPM2B_NV_PUBLIC value) {
    *obj = value;
  }

  static TPM2B_NV_PUBLIC TPM2B_NV_PUBLIC_PTR_value(TPM2B_NV_PUBLIC *obj) {
    return *obj;
  }


  static TPM2B_NV_PUBLIC * *new_TPM2B_NV_PUBLIC_PTR_PTR() { 
    return (TPM2B_NV_PUBLIC * *)malloc(sizeof(TPM2B_NV_PUBLIC *));
  }
  
  static TPM2B_NV_PUBLIC * *copy_TPM2B_NV_PUBLIC_PTR_PTR(TPM2B_NV_PUBLIC * value) { 
    return (TPM2B_NV_PUBLIC * *)memcpy((TPM2B_NV_PUBLIC * *)malloc(sizeof(TPM2B_NV_PUBLIC *)),&value,sizeof(TPM2B_NV_PUBLIC *));
  }

  static void delete_TPM2B_NV_PUBLIC_PTR_PTR(TPM2B_NV_PUBLIC * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_NV_PUBLIC_PTR_PTR_assign(TPM2B_NV_PUBLIC * *obj, TPM2B_NV_PUBLIC * value) {
    *obj = value;
  }

  static TPM2B_NV_PUBLIC * TPM2B_NV_PUBLIC_PTR_PTR_value(TPM2B_NV_PUBLIC * *obj) {
    return *obj;
  }


typedef TPM2B_CONTEXT_SENSITIVE TPM2B_CONTEXT_SENSITIVE_ARRAY;

SWIGINTERN TPM2B_CONTEXT_SENSITIVE_ARRAY *new_TPM2B_CONTEXT_SENSITIVE_ARRAY(size_t nelements){
    return (TPM2B_CONTEXT_SENSITIVE *)malloc((nelements)*sizeof(TPM2B_CONTEXT_SENSITIVE));
  }
SWIGINTERN void delete_TPM2B_CONTEXT_SENSITIVE_ARRAY(TPM2B_CONTEXT_SENSITIVE_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_CONTEXT_SENSITIVE TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__(TPM2B_CONTEXT_SENSITIVE_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__(TPM2B_CONTEXT_SENSITIVE_ARRAY *self,size_t index,TPM2B_CONTEXT_SENSITIVE value){
    self[index] = value;
  }
SWIGINTERN TPM2B_CONTEXT_SENSITIVE *TPM2B_CONTEXT_SENSITIVE_ARRAY_cast(TPM2B_CONTEXT_SENSITIVE_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_CONTEXT_SENSITIVE_ARRAY *TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer(TPM2B_CONTEXT_SENSITIVE *t){
    return (TPM2B_CONTEXT_SENSITIVE_ARRAY *)(t);
  }

  static TPM2B_CONTEXT_SENSITIVE *new_TPM2B_CONTEXT_SENSITIVE_PTR() { 
    return (TPM2B_CONTEXT_SENSITIVE *)malloc(sizeof(TPM2B_CONTEXT_SENSITIVE));
  }
  
  static TPM2B_CONTEXT_SENSITIVE *copy_TPM2B_CONTEXT_SENSITIVE_PTR(TPM2B_CONTEXT_SENSITIVE value) { 
    return (TPM2B_CONTEXT_SENSITIVE *)memcpy((TPM2B_CONTEXT_SENSITIVE *)malloc(sizeof(TPM2B_CONTEXT_SENSITIVE)),&value,sizeof(TPM2B_CONTEXT_SENSITIVE));
  }

  static void delete_TPM2B_CONTEXT_SENSITIVE_PTR(TPM2B_CONTEXT_SENSITIVE *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_CONTEXT_SENSITIVE_PTR_assign(TPM2B_CONTEXT_SENSITIVE *obj, TPM2B_CONTEXT_SENSITIVE value) {
    *obj = value;
  }

  static TPM2B_CONTEXT_SENSITIVE TPM2B_CONTEXT_SENSITIVE_PTR_value(TPM2B_CONTEXT_SENSITIVE *obj) {
    return *obj;
  }


  static TPM2B_CONTEXT_SENSITIVE * *new_TPM2B_CONTEXT_SENSITIVE_PTR_PTR() { 
    return (TPM2B_CONTEXT_SENSITIVE * *)malloc(sizeof(TPM2B_CONTEXT_SENSITIVE *));
  }
  
  static TPM2B_CONTEXT_SENSITIVE * *copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(TPM2B_CONTEXT_SENSITIVE * value) { 
    return (TPM2B_CONTEXT_SENSITIVE * *)memcpy((TPM2B_CONTEXT_SENSITIVE * *)malloc(sizeof(TPM2B_CONTEXT_SENSITIVE *)),&value,sizeof(TPM2B_CONTEXT_SENSITIVE *));
  }

  static void delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(TPM2B_CONTEXT_SENSITIVE * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign(TPM2B_CONTEXT_SENSITIVE * *obj, TPM2B_CONTEXT_SENSITIVE * value) {
    *obj = value;
  }

  static TPM2B_CONTEXT_SENSITIVE * TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value(TPM2B_CONTEXT_SENSITIVE * *obj) {
    return *obj;
  }


typedef TPMS_CONTEXT_DATA TPMS_CONTEXT_DATA_ARRAY;

SWIGINTERN TPMS_CONTEXT_DATA_ARRAY *new_TPMS_CONTEXT_DATA_ARRAY(size_t nelements){
    return (TPMS_CONTEXT_DATA *)malloc((nelements)*sizeof(TPMS_CONTEXT_DATA));
  }
SWIGINTERN void delete_TPMS_CONTEXT_DATA_ARRAY(TPMS_CONTEXT_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CONTEXT_DATA TPMS_CONTEXT_DATA_ARRAY___getitem__(TPMS_CONTEXT_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CONTEXT_DATA_ARRAY___setitem__(TPMS_CONTEXT_DATA_ARRAY *self,size_t index,TPMS_CONTEXT_DATA value){
    self[index] = value;
  }
SWIGINTERN TPMS_CONTEXT_DATA *TPMS_CONTEXT_DATA_ARRAY_cast(TPMS_CONTEXT_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CONTEXT_DATA_ARRAY *TPMS_CONTEXT_DATA_ARRAY_frompointer(TPMS_CONTEXT_DATA *t){
    return (TPMS_CONTEXT_DATA_ARRAY *)(t);
  }

  static TPMS_CONTEXT_DATA *new_TPMS_CONTEXT_DATA_PTR() { 
    return (TPMS_CONTEXT_DATA *)malloc(sizeof(TPMS_CONTEXT_DATA));
  }
  
  static TPMS_CONTEXT_DATA *copy_TPMS_CONTEXT_DATA_PTR(TPMS_CONTEXT_DATA value) { 
    return (TPMS_CONTEXT_DATA *)memcpy((TPMS_CONTEXT_DATA *)malloc(sizeof(TPMS_CONTEXT_DATA)),&value,sizeof(TPMS_CONTEXT_DATA));
  }

  static void delete_TPMS_CONTEXT_DATA_PTR(TPMS_CONTEXT_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CONTEXT_DATA_PTR_assign(TPMS_CONTEXT_DATA *obj, TPMS_CONTEXT_DATA value) {
    *obj = value;
  }

  static TPMS_CONTEXT_DATA TPMS_CONTEXT_DATA_PTR_value(TPMS_CONTEXT_DATA *obj) {
    return *obj;
  }


  static TPMS_CONTEXT_DATA * *new_TPMS_CONTEXT_DATA_PTR_PTR() { 
    return (TPMS_CONTEXT_DATA * *)malloc(sizeof(TPMS_CONTEXT_DATA *));
  }
  
  static TPMS_CONTEXT_DATA * *copy_TPMS_CONTEXT_DATA_PTR_PTR(TPMS_CONTEXT_DATA * value) { 
    return (TPMS_CONTEXT_DATA * *)memcpy((TPMS_CONTEXT_DATA * *)malloc(sizeof(TPMS_CONTEXT_DATA *)),&value,sizeof(TPMS_CONTEXT_DATA *));
  }

  static void delete_TPMS_CONTEXT_DATA_PTR_PTR(TPMS_CONTEXT_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CONTEXT_DATA_PTR_PTR_assign(TPMS_CONTEXT_DATA * *obj, TPMS_CONTEXT_DATA * value) {
    *obj = value;
  }

  static TPMS_CONTEXT_DATA * TPMS_CONTEXT_DATA_PTR_PTR_value(TPMS_CONTEXT_DATA * *obj) {
    return *obj;
  }


typedef TPM2B_CONTEXT_DATA TPM2B_CONTEXT_DATA_ARRAY;

SWIGINTERN TPM2B_CONTEXT_DATA_ARRAY *new_TPM2B_CONTEXT_DATA_ARRAY(size_t nelements){
    return (TPM2B_CONTEXT_DATA *)malloc((nelements)*sizeof(TPM2B_CONTEXT_DATA));
  }
SWIGINTERN void delete_TPM2B_CONTEXT_DATA_ARRAY(TPM2B_CONTEXT_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_CONTEXT_DATA TPM2B_CONTEXT_DATA_ARRAY___getitem__(TPM2B_CONTEXT_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_CONTEXT_DATA_ARRAY___setitem__(TPM2B_CONTEXT_DATA_ARRAY *self,size_t index,TPM2B_CONTEXT_DATA value){
    self[index] = value;
  }
SWIGINTERN TPM2B_CONTEXT_DATA *TPM2B_CONTEXT_DATA_ARRAY_cast(TPM2B_CONTEXT_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_CONTEXT_DATA_ARRAY *TPM2B_CONTEXT_DATA_ARRAY_frompointer(TPM2B_CONTEXT_DATA *t){
    return (TPM2B_CONTEXT_DATA_ARRAY *)(t);
  }

  static TPM2B_CONTEXT_DATA *new_TPM2B_CONTEXT_DATA_PTR() { 
    return (TPM2B_CONTEXT_DATA *)malloc(sizeof(TPM2B_CONTEXT_DATA));
  }
  
  static TPM2B_CONTEXT_DATA *copy_TPM2B_CONTEXT_DATA_PTR(TPM2B_CONTEXT_DATA value) { 
    return (TPM2B_CONTEXT_DATA *)memcpy((TPM2B_CONTEXT_DATA *)malloc(sizeof(TPM2B_CONTEXT_DATA)),&value,sizeof(TPM2B_CONTEXT_DATA));
  }

  static void delete_TPM2B_CONTEXT_DATA_PTR(TPM2B_CONTEXT_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_CONTEXT_DATA_PTR_assign(TPM2B_CONTEXT_DATA *obj, TPM2B_CONTEXT_DATA value) {
    *obj = value;
  }

  static TPM2B_CONTEXT_DATA TPM2B_CONTEXT_DATA_PTR_value(TPM2B_CONTEXT_DATA *obj) {
    return *obj;
  }


  static TPM2B_CONTEXT_DATA * *new_TPM2B_CONTEXT_DATA_PTR_PTR() { 
    return (TPM2B_CONTEXT_DATA * *)malloc(sizeof(TPM2B_CONTEXT_DATA *));
  }
  
  static TPM2B_CONTEXT_DATA * *copy_TPM2B_CONTEXT_DATA_PTR_PTR(TPM2B_CONTEXT_DATA * value) { 
    return (TPM2B_CONTEXT_DATA * *)memcpy((TPM2B_CONTEXT_DATA * *)malloc(sizeof(TPM2B_CONTEXT_DATA *)),&value,sizeof(TPM2B_CONTEXT_DATA *));
  }

  static void delete_TPM2B_CONTEXT_DATA_PTR_PTR(TPM2B_CONTEXT_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_CONTEXT_DATA_PTR_PTR_assign(TPM2B_CONTEXT_DATA * *obj, TPM2B_CONTEXT_DATA * value) {
    *obj = value;
  }

  static TPM2B_CONTEXT_DATA * TPM2B_CONTEXT_DATA_PTR_PTR_value(TPM2B_CONTEXT_DATA * *obj) {
    return *obj;
  }


typedef TPMS_CONTEXT TPMS_CONTEXT_ARRAY;

SWIGINTERN TPMS_CONTEXT_ARRAY *new_TPMS_CONTEXT_ARRAY(size_t nelements){
    return (TPMS_CONTEXT *)malloc((nelements)*sizeof(TPMS_CONTEXT));
  }
SWIGINTERN void delete_TPMS_CONTEXT_ARRAY(TPMS_CONTEXT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CONTEXT TPMS_CONTEXT_ARRAY___getitem__(TPMS_CONTEXT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CONTEXT_ARRAY___setitem__(TPMS_CONTEXT_ARRAY *self,size_t index,TPMS_CONTEXT value){
    self[index] = value;
  }
SWIGINTERN TPMS_CONTEXT *TPMS_CONTEXT_ARRAY_cast(TPMS_CONTEXT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CONTEXT_ARRAY *TPMS_CONTEXT_ARRAY_frompointer(TPMS_CONTEXT *t){
    return (TPMS_CONTEXT_ARRAY *)(t);
  }

  static TPMS_CONTEXT *new_TPMS_CONTEXT_PTR() { 
    return (TPMS_CONTEXT *)malloc(sizeof(TPMS_CONTEXT));
  }
  
  static TPMS_CONTEXT *copy_TPMS_CONTEXT_PTR(TPMS_CONTEXT value) { 
    return (TPMS_CONTEXT *)memcpy((TPMS_CONTEXT *)malloc(sizeof(TPMS_CONTEXT)),&value,sizeof(TPMS_CONTEXT));
  }

  static void delete_TPMS_CONTEXT_PTR(TPMS_CONTEXT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CONTEXT_PTR_assign(TPMS_CONTEXT *obj, TPMS_CONTEXT value) {
    *obj = value;
  }

  static TPMS_CONTEXT TPMS_CONTEXT_PTR_value(TPMS_CONTEXT *obj) {
    return *obj;
  }


  static TPMS_CONTEXT * *new_TPMS_CONTEXT_PTR_PTR() { 
    return (TPMS_CONTEXT * *)malloc(sizeof(TPMS_CONTEXT *));
  }
  
  static TPMS_CONTEXT * *copy_TPMS_CONTEXT_PTR_PTR(TPMS_CONTEXT * value) { 
    return (TPMS_CONTEXT * *)memcpy((TPMS_CONTEXT * *)malloc(sizeof(TPMS_CONTEXT *)),&value,sizeof(TPMS_CONTEXT *));
  }

  static void delete_TPMS_CONTEXT_PTR_PTR(TPMS_CONTEXT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CONTEXT_PTR_PTR_assign(TPMS_CONTEXT * *obj, TPMS_CONTEXT * value) {
    *obj = value;
  }

  static TPMS_CONTEXT * TPMS_CONTEXT_PTR_PTR_value(TPMS_CONTEXT * *obj) {
    return *obj;
  }


typedef TPMS_CREATION_DATA TPMS_CREATION_DATA_ARRAY;

SWIGINTERN TPMS_CREATION_DATA_ARRAY *new_TPMS_CREATION_DATA_ARRAY(size_t nelements){
    return (TPMS_CREATION_DATA *)malloc((nelements)*sizeof(TPMS_CREATION_DATA));
  }
SWIGINTERN void delete_TPMS_CREATION_DATA_ARRAY(TPMS_CREATION_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_CREATION_DATA TPMS_CREATION_DATA_ARRAY___getitem__(TPMS_CREATION_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_CREATION_DATA_ARRAY___setitem__(TPMS_CREATION_DATA_ARRAY *self,size_t index,TPMS_CREATION_DATA value){
    self[index] = value;
  }
SWIGINTERN TPMS_CREATION_DATA *TPMS_CREATION_DATA_ARRAY_cast(TPMS_CREATION_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_CREATION_DATA_ARRAY *TPMS_CREATION_DATA_ARRAY_frompointer(TPMS_CREATION_DATA *t){
    return (TPMS_CREATION_DATA_ARRAY *)(t);
  }

  static TPMS_CREATION_DATA *new_TPMS_CREATION_DATA_PTR() { 
    return (TPMS_CREATION_DATA *)malloc(sizeof(TPMS_CREATION_DATA));
  }
  
  static TPMS_CREATION_DATA *copy_TPMS_CREATION_DATA_PTR(TPMS_CREATION_DATA value) { 
    return (TPMS_CREATION_DATA *)memcpy((TPMS_CREATION_DATA *)malloc(sizeof(TPMS_CREATION_DATA)),&value,sizeof(TPMS_CREATION_DATA));
  }

  static void delete_TPMS_CREATION_DATA_PTR(TPMS_CREATION_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CREATION_DATA_PTR_assign(TPMS_CREATION_DATA *obj, TPMS_CREATION_DATA value) {
    *obj = value;
  }

  static TPMS_CREATION_DATA TPMS_CREATION_DATA_PTR_value(TPMS_CREATION_DATA *obj) {
    return *obj;
  }


  static TPMS_CREATION_DATA * *new_TPMS_CREATION_DATA_PTR_PTR() { 
    return (TPMS_CREATION_DATA * *)malloc(sizeof(TPMS_CREATION_DATA *));
  }
  
  static TPMS_CREATION_DATA * *copy_TPMS_CREATION_DATA_PTR_PTR(TPMS_CREATION_DATA * value) { 
    return (TPMS_CREATION_DATA * *)memcpy((TPMS_CREATION_DATA * *)malloc(sizeof(TPMS_CREATION_DATA *)),&value,sizeof(TPMS_CREATION_DATA *));
  }

  static void delete_TPMS_CREATION_DATA_PTR_PTR(TPMS_CREATION_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_CREATION_DATA_PTR_PTR_assign(TPMS_CREATION_DATA * *obj, TPMS_CREATION_DATA * value) {
    *obj = value;
  }

  static TPMS_CREATION_DATA * TPMS_CREATION_DATA_PTR_PTR_value(TPMS_CREATION_DATA * *obj) {
    return *obj;
  }


typedef TPM2B_CREATION_DATA TPM2B_CREATION_DATA_ARRAY;

SWIGINTERN TPM2B_CREATION_DATA_ARRAY *new_TPM2B_CREATION_DATA_ARRAY(size_t nelements){
    return (TPM2B_CREATION_DATA *)malloc((nelements)*sizeof(TPM2B_CREATION_DATA));
  }
SWIGINTERN void delete_TPM2B_CREATION_DATA_ARRAY(TPM2B_CREATION_DATA_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPM2B_CREATION_DATA TPM2B_CREATION_DATA_ARRAY___getitem__(TPM2B_CREATION_DATA_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPM2B_CREATION_DATA_ARRAY___setitem__(TPM2B_CREATION_DATA_ARRAY *self,size_t index,TPM2B_CREATION_DATA value){
    self[index] = value;
  }
SWIGINTERN TPM2B_CREATION_DATA *TPM2B_CREATION_DATA_ARRAY_cast(TPM2B_CREATION_DATA_ARRAY *self){
    return self;
  }
SWIGINTERN TPM2B_CREATION_DATA_ARRAY *TPM2B_CREATION_DATA_ARRAY_frompointer(TPM2B_CREATION_DATA *t){
    return (TPM2B_CREATION_DATA_ARRAY *)(t);
  }

  static TPM2B_CREATION_DATA *new_TPM2B_CREATION_DATA_PTR() { 
    return (TPM2B_CREATION_DATA *)malloc(sizeof(TPM2B_CREATION_DATA));
  }
  
  static TPM2B_CREATION_DATA *copy_TPM2B_CREATION_DATA_PTR(TPM2B_CREATION_DATA value) { 
    return (TPM2B_CREATION_DATA *)memcpy((TPM2B_CREATION_DATA *)malloc(sizeof(TPM2B_CREATION_DATA)),&value,sizeof(TPM2B_CREATION_DATA));
  }

  static void delete_TPM2B_CREATION_DATA_PTR(TPM2B_CREATION_DATA *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_CREATION_DATA_PTR_assign(TPM2B_CREATION_DATA *obj, TPM2B_CREATION_DATA value) {
    *obj = value;
  }

  static TPM2B_CREATION_DATA TPM2B_CREATION_DATA_PTR_value(TPM2B_CREATION_DATA *obj) {
    return *obj;
  }


  static TPM2B_CREATION_DATA * *new_TPM2B_CREATION_DATA_PTR_PTR() { 
    return (TPM2B_CREATION_DATA * *)malloc(sizeof(TPM2B_CREATION_DATA *));
  }
  
  static TPM2B_CREATION_DATA * *copy_TPM2B_CREATION_DATA_PTR_PTR(TPM2B_CREATION_DATA * value) { 
    return (TPM2B_CREATION_DATA * *)memcpy((TPM2B_CREATION_DATA * *)malloc(sizeof(TPM2B_CREATION_DATA *)),&value,sizeof(TPM2B_CREATION_DATA *));
  }

  static void delete_TPM2B_CREATION_DATA_PTR_PTR(TPM2B_CREATION_DATA * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPM2B_CREATION_DATA_PTR_PTR_assign(TPM2B_CREATION_DATA * *obj, TPM2B_CREATION_DATA * value) {
    *obj = value;
  }

  static TPM2B_CREATION_DATA * TPM2B_CREATION_DATA_PTR_PTR_value(TPM2B_CREATION_DATA * *obj) {
    return *obj;
  }


typedef TPMS_AC_OUTPUT TPMS_AC_OUTPUT_ARRAY;

SWIGINTERN TPMS_AC_OUTPUT_ARRAY *new_TPMS_AC_OUTPUT_ARRAY(size_t nelements){
    return (TPMS_AC_OUTPUT *)malloc((nelements)*sizeof(TPMS_AC_OUTPUT));
  }
SWIGINTERN void delete_TPMS_AC_OUTPUT_ARRAY(TPMS_AC_OUTPUT_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPMS_AC_OUTPUT TPMS_AC_OUTPUT_ARRAY___getitem__(TPMS_AC_OUTPUT_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPMS_AC_OUTPUT_ARRAY___setitem__(TPMS_AC_OUTPUT_ARRAY *self,size_t index,TPMS_AC_OUTPUT value){
    self[index] = value;
  }
SWIGINTERN TPMS_AC_OUTPUT *TPMS_AC_OUTPUT_ARRAY_cast(TPMS_AC_OUTPUT_ARRAY *self){
    return self;
  }
SWIGINTERN TPMS_AC_OUTPUT_ARRAY *TPMS_AC_OUTPUT_ARRAY_frompointer(TPMS_AC_OUTPUT *t){
    return (TPMS_AC_OUTPUT_ARRAY *)(t);
  }

  static TPMS_AC_OUTPUT *new_TPMS_AC_OUTPUT_PTR() { 
    return (TPMS_AC_OUTPUT *)malloc(sizeof(TPMS_AC_OUTPUT));
  }
  
  static TPMS_AC_OUTPUT *copy_TPMS_AC_OUTPUT_PTR(TPMS_AC_OUTPUT value) { 
    return (TPMS_AC_OUTPUT *)memcpy((TPMS_AC_OUTPUT *)malloc(sizeof(TPMS_AC_OUTPUT)),&value,sizeof(TPMS_AC_OUTPUT));
  }

  static void delete_TPMS_AC_OUTPUT_PTR(TPMS_AC_OUTPUT *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_AC_OUTPUT_PTR_assign(TPMS_AC_OUTPUT *obj, TPMS_AC_OUTPUT value) {
    *obj = value;
  }

  static TPMS_AC_OUTPUT TPMS_AC_OUTPUT_PTR_value(TPMS_AC_OUTPUT *obj) {
    return *obj;
  }


  static TPMS_AC_OUTPUT * *new_TPMS_AC_OUTPUT_PTR_PTR() { 
    return (TPMS_AC_OUTPUT * *)malloc(sizeof(TPMS_AC_OUTPUT *));
  }
  
  static TPMS_AC_OUTPUT * *copy_TPMS_AC_OUTPUT_PTR_PTR(TPMS_AC_OUTPUT * value) { 
    return (TPMS_AC_OUTPUT * *)memcpy((TPMS_AC_OUTPUT * *)malloc(sizeof(TPMS_AC_OUTPUT *)),&value,sizeof(TPMS_AC_OUTPUT *));
  }

  static void delete_TPMS_AC_OUTPUT_PTR_PTR(TPMS_AC_OUTPUT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPMS_AC_OUTPUT_PTR_PTR_assign(TPMS_AC_OUTPUT * *obj, TPMS_AC_OUTPUT * value) {
    *obj = value;
  }

  static TPMS_AC_OUTPUT * TPMS_AC_OUTPUT_PTR_PTR_value(TPMS_AC_OUTPUT * *obj) {
    return *obj;
  }


typedef TPML_AC_CAPABILITIES TPML_AC_CAPABILITIES_ARRAY;

SWIGINTERN TPML_AC_CAPABILITIES_ARRAY *new_TPML_AC_CAPABILITIES_ARRAY(size_t nelements){
    return (TPML_AC_CAPABILITIES *)malloc((nelements)*sizeof(TPML_AC_CAPABILITIES));
  }
SWIGINTERN void delete_TPML_AC_CAPABILITIES_ARRAY(TPML_AC_CAPABILITIES_ARRAY *self){
    free((char*)self);
  }
SWIGINTERN TPML_AC_CAPABILITIES TPML_AC_CAPABILITIES_ARRAY___getitem__(TPML_AC_CAPABILITIES_ARRAY *self,size_t index){
    return self[index];
  }
SWIGINTERN void TPML_AC_CAPABILITIES_ARRAY___setitem__(TPML_AC_CAPABILITIES_ARRAY *self,size_t index,TPML_AC_CAPABILITIES value){
    self[index] = value;
  }
SWIGINTERN TPML_AC_CAPABILITIES *TPML_AC_CAPABILITIES_ARRAY_cast(TPML_AC_CAPABILITIES_ARRAY *self){
    return self;
  }
SWIGINTERN TPML_AC_CAPABILITIES_ARRAY *TPML_AC_CAPABILITIES_ARRAY_frompointer(TPML_AC_CAPABILITIES *t){
    return (TPML_AC_CAPABILITIES_ARRAY *)(t);
  }

  static TPML_AC_CAPABILITIES *new_TPML_AC_CAPABILITIES_PTR() { 
    return (TPML_AC_CAPABILITIES *)malloc(sizeof(TPML_AC_CAPABILITIES));
  }
  
  static TPML_AC_CAPABILITIES *copy_TPML_AC_CAPABILITIES_PTR(TPML_AC_CAPABILITIES value) { 
    return (TPML_AC_CAPABILITIES *)memcpy((TPML_AC_CAPABILITIES *)malloc(sizeof(TPML_AC_CAPABILITIES)),&value,sizeof(TPML_AC_CAPABILITIES));
  }

  static void delete_TPML_AC_CAPABILITIES_PTR(TPML_AC_CAPABILITIES *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_AC_CAPABILITIES_PTR_assign(TPML_AC_CAPABILITIES *obj, TPML_AC_CAPABILITIES value) {
    *obj = value;
  }

  static TPML_AC_CAPABILITIES TPML_AC_CAPABILITIES_PTR_value(TPML_AC_CAPABILITIES *obj) {
    return *obj;
  }


  static TPML_AC_CAPABILITIES * *new_TPML_AC_CAPABILITIES_PTR_PTR() { 
    return (TPML_AC_CAPABILITIES * *)malloc(sizeof(TPML_AC_CAPABILITIES *));
  }
  
  static TPML_AC_CAPABILITIES * *copy_TPML_AC_CAPABILITIES_PTR_PTR(TPML_AC_CAPABILITIES * value) { 
    return (TPML_AC_CAPABILITIES * *)memcpy((TPML_AC_CAPABILITIES * *)malloc(sizeof(TPML_AC_CAPABILITIES *)),&value,sizeof(TPML_AC_CAPABILITIES *));
  }

  static void delete_TPML_AC_CAPABILITIES_PTR_PTR(TPML_AC_CAPABILITIES * *obj) { 
    if (obj) free((char*)obj);
  }

  static void TPML_AC_CAPABILITIES_PTR_PTR_assign(TPML_AC_CAPABILITIES * *obj, TPML_AC_CAPABILITIES * value) {
    *obj = value;
  }

  static TPML_AC_CAPABILITIES * TPML_AC_CAPABILITIES_PTR_PTR_value(TPML_AC_CAPABILITIES * *obj) {
    return *obj;
  }


#include <stdint.h>		// Use the C99 official header


  static void * *new_VOID_PTR_PTR() { 
    return (void * *)malloc(sizeof(void *));
  }
  
  static void * *copy_VOID_PTR_PTR(void * value) { 
    return (void * *)memcpy((void * *)malloc(sizeof(void *)),&value,sizeof(void *));
  }

  static void delete_VOID_PTR_PTR(void * *obj) { 
    if (obj) free((char*)obj);
  }

  static void VOID_PTR_PTR_assign(void * *obj, void * value) {
    *obj = value;
  }

  static void * VOID_PTR_PTR_value(void * *obj) {
    return *obj;
  }


  static size_t *new_SIZE_T_PTR() { 
    return (size_t *)malloc(sizeof(size_t));
  }
  
  static size_t *copy_SIZE_T_PTR(size_t value) { 
    return (size_t *)memcpy((size_t *)malloc(sizeof(size_t)),&value,sizeof(size_t));
  }

  static void delete_SIZE_T_PTR(size_t *obj) { 
    if (obj) free((char*)obj);
  }

  static void SIZE_T_PTR_assign(size_t *obj, size_t value) {
    *obj = value;
  }

  static size_t SIZE_T_PTR_value(size_t *obj) {
    return *obj;
  }


  static size_t * *new_SIZE_T_PTR_PTR() { 
    return (size_t * *)malloc(sizeof(size_t *));
  }
  
  static size_t * *copy_SIZE_T_PTR_PTR(size_t * value) { 
    return (size_t * *)memcpy((size_t * *)malloc(sizeof(size_t *)),&value,sizeof(size_t *));
  }

  static void delete_SIZE_T_PTR_PTR(size_t * *obj) { 
    if (obj) free((char*)obj);
  }

  static void SIZE_T_PTR_PTR_assign(size_t * *obj, size_t * value) {
    *obj = value;
  }

  static size_t * SIZE_T_PTR_PTR_value(size_t * *obj) {
    return *obj;
  }


  static char *new_CHAR_PTR() { 
    return (char *)malloc(sizeof(char));
  }
  
  static char *copy_CHAR_PTR(char value) { 
    return (char *)memcpy((char *)malloc(sizeof(char)),&value,sizeof(char));
  }

  static void delete_CHAR_PTR(char *obj) { 
    if (obj) free((char*)obj);
  }

  static void CHAR_PTR_assign(char *obj, char value) {
    *obj = value;
  }

  static char CHAR_PTR_value(char *obj) {
    return *obj;
  }


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
  if (PyUnicode_Check(obj))
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
#if PY_VERSION_HEX>=0x03000000
    if (!alloc && cptr) {
        /* We can't allow converting without allocation, since the internal
           representation of string in Python 3 is UCS-2/UCS-4 but we require
           a UTF-8 representation.
           TODO(bhy) More detailed explanation */
        return SWIG_RuntimeError;
    }
    obj = PyUnicode_AsUTF8String(obj);
    PyBytes_AsStringAndSize(obj, &cstr, &len);
    if(alloc) *alloc = SWIG_NEWOBJ;
#else
    PyString_AsStringAndSize(obj, &cstr, &len);
#endif
    if (cptr) {
      if (alloc) {
	/* 
	   In python the user should not be able to modify the inner
	   string representation. To warranty that, if you define
	   SWIG_PYTHON_SAFE_CSTRINGS, a new/copy of the python string
	   buffer is always returned.

	   The default behavior is just to return the pointer value,
	   so, be careful.
	*/ 
#if defined(SWIG_PYTHON_SAFE_CSTRINGS)
	if (*alloc != SWIG_OLDOBJ) 
#else
	if (*alloc == SWIG_NEWOBJ) 
#endif
	{
	  *cptr = (char *)memcpy((char *)malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
	  *alloc = SWIG_NEWOBJ;
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
	#if PY_VERSION_HEX>=0x03000000
	assert(0); /* Should never reach here in Python 3 */
	#endif
	*cptr = SWIG_Python_str_AsChar(obj);
      }
    }
    if (psize) *psize = len + 1;
#if PY_VERSION_HEX>=0x03000000
    Py_XDECREF(obj);
#endif
    return SWIG_OK;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if PY_VERSION_HEX<0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = (char *)memcpy((char *)malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        }
        if (psize) *psize = len + 1;

        Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsCharArray(PyObject * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	free((char*)cptr);
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) free((char*)cptr);
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_char (PyObject * obj, char *val)
{    
  int res = SWIG_AsCharArray(obj, val, 1);
  if (!SWIG_IsOK(res)) {
    long v;
    res = SWIG_AddCast(SWIG_AsVal_long (obj, &v));
    if (SWIG_IsOK(res)) {
      if ((CHAR_MIN <= v) && (v <= CHAR_MAX)) {
	if (val) *val = (char)(v);
      } else {
	res = SWIG_OverflowError;
      }
    }
  }
  return res;
}





SWIGINTERNINLINE PyObject *
SWIG_From_char  (char c) 
{ 
  return SWIG_FromCharPtrAndSize(&c,1);
}


  static char * *new_CHAR_PTR_PTR() { 
    return (char * *)malloc(sizeof(char *));
  }
  
  static char * *copy_CHAR_PTR_PTR(char * value) { 
    return (char * *)memcpy((char * *)malloc(sizeof(char *)),&value,sizeof(char *));
  }

  static void delete_CHAR_PTR_PTR(char * *obj) { 
    if (obj) free((char*)obj);
  }

  static void CHAR_PTR_PTR_assign(char * *obj, char * value) {
    *obj = value;
  }

  static char * CHAR_PTR_PTR_value(char * *obj) {
    return *obj;
  }


  static UINT8 *new_UINT8_PTR() { 
    return (UINT8 *)malloc(sizeof(UINT8));
  }
  
  static UINT8 *copy_UINT8_PTR(UINT8 value) { 
    return (UINT8 *)memcpy((UINT8 *)malloc(sizeof(UINT8)),&value,sizeof(UINT8));
  }

  static void delete_UINT8_PTR(UINT8 *obj) { 
    if (obj) free((char*)obj);
  }

  static void UINT8_PTR_assign(UINT8 *obj, UINT8 value) {
    *obj = value;
  }

  static UINT8 UINT8_PTR_value(UINT8 *obj) {
    return *obj;
  }


  static UINT8 * *new_UINT8_PTR_PTR() { 
    return (UINT8 * *)malloc(sizeof(UINT8 *));
  }
  
  static UINT8 * *copy_UINT8_PTR_PTR(UINT8 * value) { 
    return (UINT8 * *)memcpy((UINT8 * *)malloc(sizeof(UINT8 *)),&value,sizeof(UINT8 *));
  }

  static void delete_UINT8_PTR_PTR(UINT8 * *obj) { 
    if (obj) free((char*)obj);
  }

  static void UINT8_PTR_PTR_assign(UINT8 * *obj, UINT8 * value) {
    *obj = value;
  }

  static UINT8 * UINT8_PTR_PTR_value(UINT8 * *obj) {
    return *obj;
  }


typedef uint32_t uint32_t_PTR;

SWIGINTERN uint32_t_PTR *new_uint32_t_PTR(void){
    return (uint32_t *)malloc(sizeof(uint32_t));
  }
SWIGINTERN void delete_uint32_t_PTR(uint32_t_PTR *self){
    if (self) free((char*)self);
  }
SWIGINTERN void uint32_t_PTR_assign(uint32_t_PTR *self,uint32_t value){
    *self = value;
  }
SWIGINTERN uint32_t uint32_t_PTR_value(uint32_t_PTR *self){
    return *self;
  }
SWIGINTERN uint32_t *uint32_t_PTR_cast(uint32_t_PTR *self){
    return self;
  }
SWIGINTERN uint32_t_PTR *uint32_t_PTR_frompointer(uint32_t *t){
    return (uint32_t_PTR *) t;
  }

  static uint32_t_PTR *new_uint32_t_PTR_PTR() { 
    return (uint32_t_PTR *)malloc(sizeof(uint32_t_PTR));
  }
  
  static uint32_t_PTR *copy_uint32_t_PTR_PTR(uint32_t_PTR value) { 
    return (uint32_t_PTR *)memcpy((uint32_t_PTR *)malloc(sizeof(uint32_t_PTR)),&value,sizeof(uint32_t_PTR));
  }

  static void delete_uint32_t_PTR_PTR(uint32_t_PTR *obj) { 
    if (obj) free((char*)obj);
  }

  static void uint32_t_PTR_PTR_assign(uint32_t_PTR *obj, uint32_t_PTR value) {
    *obj = value;
  }

  static uint32_t_PTR uint32_t_PTR_PTR_value(uint32_t_PTR *obj) {
    return *obj;
  }


typedef UINT32 UINT32_PTR;

SWIGINTERN UINT32_PTR *new_UINT32_PTR(void){
    return (UINT32 *)malloc(sizeof(UINT32));
  }
SWIGINTERN void delete_UINT32_PTR(UINT32_PTR *self){
    if (self) free((char*)self);
  }
SWIGINTERN void UINT32_PTR_assign(UINT32_PTR *self,UINT32 value){
    *self = value;
  }
SWIGINTERN UINT32 UINT32_PTR_value(UINT32_PTR *self){
    return *self;
  }
SWIGINTERN UINT32 *UINT32_PTR_cast(UINT32_PTR *self){
    return self;
  }
SWIGINTERN UINT32_PTR *UINT32_PTR_frompointer(UINT32 *t){
    return (UINT32_PTR *) t;
  }

  static UINT32_PTR *new_UINT32_PTR_PTR() { 
    return (UINT32_PTR *)malloc(sizeof(UINT32_PTR));
  }
  
  static UINT32_PTR *copy_UINT32_PTR_PTR(UINT32_PTR value) { 
    return (UINT32_PTR *)memcpy((UINT32_PTR *)malloc(sizeof(UINT32_PTR)),&value,sizeof(UINT32_PTR));
  }

  static void delete_UINT32_PTR_PTR(UINT32_PTR *obj) { 
    if (obj) free((char*)obj);
  }

  static void UINT32_PTR_PTR_assign(UINT32_PTR *obj, UINT32_PTR value) {
    *obj = value;
  }

  static UINT32_PTR UINT32_PTR_PTR_value(UINT32_PTR *obj) {
    return *obj;
  }


typedef uint16_t uint16_t_PTR;

SWIGINTERN uint16_t_PTR *new_uint16_t_PTR(void){
    return (uint16_t *)malloc(sizeof(uint16_t));
  }
SWIGINTERN void delete_uint16_t_PTR(uint16_t_PTR *self){
    if (self) free((char*)self);
  }
SWIGINTERN void uint16_t_PTR_assign(uint16_t_PTR *self,uint16_t value){
    *self = value;
  }
SWIGINTERN uint16_t uint16_t_PTR_value(uint16_t_PTR *self){
    return *self;
  }
SWIGINTERN uint16_t *uint16_t_PTR_cast(uint16_t_PTR *self){
    return self;
  }
SWIGINTERN uint16_t_PTR *uint16_t_PTR_frompointer(uint16_t *t){
    return (uint16_t_PTR *) t;
  }

  static uint16_t_PTR *new_uint16_t_PTR_PTR() { 
    return (uint16_t_PTR *)malloc(sizeof(uint16_t_PTR));
  }
  
  static uint16_t_PTR *copy_uint16_t_PTR_PTR(uint16_t_PTR value) { 
    return (uint16_t_PTR *)memcpy((uint16_t_PTR *)malloc(sizeof(uint16_t_PTR)),&value,sizeof(uint16_t_PTR));
  }

  static void delete_uint16_t_PTR_PTR(uint16_t_PTR *obj) { 
    if (obj) free((char*)obj);
  }

  static void uint16_t_PTR_PTR_assign(uint16_t_PTR *obj, uint16_t_PTR value) {
    *obj = value;
  }

  static uint16_t_PTR uint16_t_PTR_PTR_value(uint16_t_PTR *obj) {
    return *obj;
  }


typedef UINT16 UINT16_PTR;

SWIGINTERN UINT16_PTR *new_UINT16_PTR(void){
    return (UINT16 *)malloc(sizeof(UINT16));
  }
SWIGINTERN void delete_UINT16_PTR(UINT16_PTR *self){
    if (self) free((char*)self);
  }
SWIGINTERN void UINT16_PTR_assign(UINT16_PTR *self,UINT16 value){
    *self = value;
  }
SWIGINTERN UINT16 UINT16_PTR_value(UINT16_PTR *self){
    return *self;
  }
SWIGINTERN UINT16 *UINT16_PTR_cast(UINT16_PTR *self){
    return self;
  }
SWIGINTERN UINT16_PTR *UINT16_PTR_frompointer(UINT16 *t){
    return (UINT16_PTR *) t;
  }

  static UINT16_PTR *new_UINT16_PTR_PTR() { 
    return (UINT16_PTR *)malloc(sizeof(UINT16_PTR));
  }
  
  static UINT16_PTR *copy_UINT16_PTR_PTR(UINT16_PTR value) { 
    return (UINT16_PTR *)memcpy((UINT16_PTR *)malloc(sizeof(UINT16_PTR)),&value,sizeof(UINT16_PTR));
  }

  static void delete_UINT16_PTR_PTR(UINT16_PTR *obj) { 
    if (obj) free((char*)obj);
  }

  static void UINT16_PTR_PTR_assign(UINT16_PTR *obj, UINT16_PTR value) {
    *obj = value;
  }

  static UINT16_PTR UINT16_PTR_PTR_value(UINT16_PTR *obj) {
    return *obj;
  }


typedef BYTE ByteArray;

SWIGINTERN ByteArray *new_ByteArray(size_t nelements){
    return (BYTE *)malloc((nelements)*sizeof(BYTE));
  }
SWIGINTERN void delete_ByteArray(ByteArray *self){
    free((char*)self);
  }
SWIGINTERN BYTE ByteArray___getitem__(ByteArray *self,size_t index){
    return self[index];
  }
SWIGINTERN void ByteArray___setitem__(ByteArray *self,size_t index,BYTE value){
    self[index] = value;
  }
SWIGINTERN BYTE *ByteArray_cast(ByteArray *self){
    return self;
  }
SWIGINTERN ByteArray *ByteArray_frompointer(BYTE *t){
    return (ByteArray *)(t);
  }

typedef ByteArray BYTE_ARRAY;
typedef TPM2B_DIGEST TPM2B_NONCE;
typedef TPM2B_DIGEST TPM2B_AUTH;
typedef TPM2B_DIGEST TPM2B_OPERAND;
typedef TPM2B_DIGEST TPM2B_TIMEOUT;
typedef UINT16 TPM2_ALG_ID;
typedef UINT16 TPM2_ECC_CURVE;
typedef UINT32 TPM2_CC;
typedef UINT32 TPM2_ALGORITHM_ID;
typedef UINT32 TPM2_MODIFIER_INDICATOR;
typedef UINT32 TPM2_AUTHORIZATION_SIZE;
typedef UINT32 TPM2_PARAMETER_SIZE;
typedef UINT16 TPM2_KEY_SIZE;
typedef UINT16 TPM2_KEY_BITS;
typedef UINT32 TPM2_SPEC;
typedef UINT32 TPM2_GENERATED;
typedef UINT32 TPM2_RC;
typedef INT8 TPM2_CLOCK_ADJUST;
typedef UINT16 TPM2_EO;
typedef UINT16 TPM2_ST;
typedef UINT16 TPM2_SU;
typedef UINT8 TPM2_SE;
typedef UINT32 TPM2_CAP;
typedef UINT32 TPM2_PT;
typedef UINT32 TPM2_PT_PCR;
typedef UINT32 TPM2_PS;
typedef UINT32 TPM2_HANDLE;
typedef UINT8 TPM2_HT;
typedef TPM2_HANDLE TPM2_RH;
typedef TPM2_HANDLE TPM2_HC;
typedef uint32_t TPMA_ALGORITHM;
typedef uint32_t TPMA_OBJECT;
typedef UINT8 TPMA_SESSION;
typedef UINT8 TPMA_LOCALITY;
typedef uint32_t TPMA_PERMANENT;
typedef uint32_t TPMA_STARTUP_CLEAR;
typedef uint32_t TPMA_MEMORY;
typedef uint32_t TPMA_CC;
typedef uint32_t TPMA_MODES;
typedef BYTE TPMI_YES_NO;
typedef TPM2_HANDLE TPMI_DH_OBJECT;
typedef TPM2_HANDLE TPMI_DH_PERSISTENT;
typedef TPM2_HANDLE TPMI_DH_ENTITY;
typedef TPM2_HANDLE TPMI_DH_PCR;
typedef TPM2_HANDLE TPMI_SH_AUTH_SESSION;
typedef TPM2_HANDLE TPMI_SH_HMAC;
typedef TPM2_HANDLE TPMI_SH_POLICY;
typedef TPM2_HANDLE TPMI_DH_CONTEXT;
typedef TPM2_HANDLE TPMI_RH_HIERARCHY;
typedef TPM2_HANDLE TPMI_RH_ENABLES;
typedef TPM2_HANDLE TPMI_RH_HIERARCHY_AUTH;
typedef TPM2_HANDLE TPMI_RH_PLATFORM;
typedef TPM2_HANDLE TPMI_RH_OWNER;
typedef TPM2_HANDLE TPMI_RH_ENDORSEMENT;
typedef TPM2_HANDLE TPMI_RH_PROVISION;
typedef TPM2_HANDLE TPMI_RH_CLEAR;
typedef TPM2_HANDLE TPMI_RH_NV_AUTH;
typedef TPM2_HANDLE TPMI_RH_LOCKOUT;
typedef TPM2_HANDLE TPMI_RH_NV_INDEX;
typedef TPM2_ALG_ID TPMI_ALG_HASH;
typedef TPM2_ALG_ID TPMI_ALG_ASYM;
typedef TPM2_ALG_ID TPMI_ALG_SYM;
typedef TPM2_ALG_ID TPMI_ALG_SYM_OBJECT;
typedef TPM2_ALG_ID TPMI_ALG_SYM_MODE;
typedef TPM2_ALG_ID TPMI_ALG_KDF;
typedef TPM2_ALG_ID TPMI_ALG_SIG_SCHEME;
typedef TPM2_ALG_ID TPMI_ECC_KEY_EXCHANGE;
typedef TPM2_ST TPMI_ST_COMMAND_TAG;
typedef TPM2_ST TPMI_ST_ATTEST;
typedef TPM2_KEY_BITS TPMI_AES_KEY_BITS;
typedef TPM2_KEY_BITS TPMI_SM4_KEY_BITS;
typedef TPM2_KEY_BITS TPMI_TPM2_CAMELLIA_KEY_BITS;
typedef TPM2_ALG_ID TPMI_ALG_KEYEDHASH_SCHEME;
typedef TPMS_SCHEME_HASH TPMS_SCHEME_HMAC;
typedef TPMS_SCHEME_HASH TPMS_SIG_SCHEME_RSASSA;
typedef TPMS_SCHEME_HASH TPMS_SIG_SCHEME_RSAPSS;
typedef TPMS_SCHEME_HASH  TPMS_SIG_SCHEME_ECDSA;
typedef TPMS_SCHEME_HASH  TPMS_SIG_SCHEME_SM2;
typedef TPMS_SCHEME_HASH  TPMS_SIG_SCHEME_ECSCHNORR;
typedef TPMS_SCHEME_ECDAA TPMS_SIG_SCHEME_ECDAA;
typedef TPMS_SCHEME_HASH TPMS_ENC_SCHEME_OAEP;
typedef TPMS_EMPTY TPMS_ENC_SCHEME_RSAES;
typedef TPMS_SCHEME_HASH TPMS_KEY_SCHEME_ECDH;
typedef TPMS_SCHEME_HASH TPMS_KEY_SCHEME_ECMQV;
typedef TPMS_SCHEME_HASH TPMS_SCHEME_MGF1;
typedef TPMS_SCHEME_HASH TPMS_SCHEME_KDF1_SP800_56A;
typedef TPMS_SCHEME_HASH TPMS_SCHEME_KDF2;
typedef TPMS_SCHEME_HASH TPMS_SCHEME_KDF1_SP800_108;
typedef TPM2_ALG_ID TPMI_ALG_ASYM_SCHEME;
typedef TPM2_ALG_ID TPMI_ALG_RSA_SCHEME;
typedef TPM2_ALG_ID TPMI_ALG_RSA_DECRYPT;
typedef TPM2_KEY_BITS TPMI_RSA_KEY_BITS;
typedef TPM2_ALG_ID TPMI_ALG_ECC_SCHEME;
typedef TPM2_ECC_CURVE TPMI_ECC_CURVE;
typedef TPMS_SIGNATURE_RSA TPMS_SIGNATURE_RSASSA;
typedef TPMS_SIGNATURE_RSA TPMS_SIGNATURE_RSAPSS;
typedef TPMS_SIGNATURE_ECC TPMS_SIGNATURE_ECDSA;
typedef TPMS_SIGNATURE_ECC TPMS_SIGNATURE_ECDAA;
typedef TPMS_SIGNATURE_ECC TPMS_SIGNATURE_SM2;
typedef TPMS_SIGNATURE_ECC TPMS_SIGNATURE_ECSCHNORR;
typedef TPM2_ALG_ID TPMI_ALG_PUBLIC;
typedef uint32_t TPM2_NV_INDEX;
typedef UINT8 TPM2_NT;
typedef uint32_t TPMA_NV;
typedef UINT32 TPM_AT;
typedef UINT32 TPM_EA;
typedef TPM2_HANDLE TPMI_RH_AC;


  #include <tss2/tss2_mu.h>



SWIGINTERN int
SWIG_AsVal_signed_SS_char (PyObject * obj, signed char *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < SCHAR_MIN || v > SCHAR_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (signed char)(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsVal_short (PyObject * obj, short *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < SHRT_MIN || v > SHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (short)(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsVal_long_SS_long (PyObject *obj, long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    long long v = PyLong_AsLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    long v;
    res = SWIG_AsVal_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    const double mant_min = -mant_max;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, mant_min, mant_max)) {
      if (val) *val = (long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}


typedef unsigned int ESYS_TR_PTR;

SWIGINTERN ESYS_TR_PTR *new_ESYS_TR_PTR(void){
    return (unsigned int *)malloc(sizeof(unsigned int));
  }
SWIGINTERN void delete_ESYS_TR_PTR(ESYS_TR_PTR *self){
    if (self) free((char*)self);
  }
SWIGINTERN void ESYS_TR_PTR_assign(ESYS_TR_PTR *self,unsigned int value){
    *self = value;
  }
SWIGINTERN unsigned int ESYS_TR_PTR_value(ESYS_TR_PTR *self){
    return *self;
  }
SWIGINTERN unsigned int *ESYS_TR_PTR_cast(ESYS_TR_PTR *self){
    return self;
  }
SWIGINTERN ESYS_TR_PTR *ESYS_TR_PTR_frompointer(unsigned int *t){
    return (ESYS_TR_PTR *) t;
  }

  static unsigned int * *new_ESYS_TR_PTR_PTR() { 
    return (unsigned int * *)malloc(sizeof(unsigned int *));
  }
  
  static unsigned int * *copy_ESYS_TR_PTR_PTR(unsigned int * value) { 
    return (unsigned int * *)memcpy((unsigned int * *)malloc(sizeof(unsigned int *)),&value,sizeof(unsigned int *));
  }

  static void delete_ESYS_TR_PTR_PTR(unsigned int * *obj) { 
    if (obj) free((char*)obj);
  }

  static void ESYS_TR_PTR_PTR_assign(unsigned int * *obj, unsigned int * value) {
    *obj = value;
  }

  static unsigned int * ESYS_TR_PTR_PTR_value(unsigned int * *obj) {
    return *obj;
  }


  static struct ESYS_CONTEXT * *new_ctx_ptr() { 
    return (struct ESYS_CONTEXT * *)malloc(sizeof(struct ESYS_CONTEXT *));
  }
  
  static struct ESYS_CONTEXT * *copy_ctx_ptr(struct ESYS_CONTEXT * value) { 
    return (struct ESYS_CONTEXT * *)memcpy((struct ESYS_CONTEXT * *)malloc(sizeof(struct ESYS_CONTEXT *)),&value,sizeof(struct ESYS_CONTEXT *));
  }

  static void delete_ctx_ptr(struct ESYS_CONTEXT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void ctx_ptr_assign(struct ESYS_CONTEXT * *obj, struct ESYS_CONTEXT * value) {
    *obj = value;
  }

  static struct ESYS_CONTEXT * ctx_ptr_value(struct ESYS_CONTEXT * *obj) {
    return *obj;
  }


  static struct TSS2_TCTI_CONTEXT * *new_tcti_ctx_ptr_ptr() { 
    return (struct TSS2_TCTI_CONTEXT * *)malloc(sizeof(struct TSS2_TCTI_CONTEXT *));
  }
  
  static struct TSS2_TCTI_CONTEXT * *copy_tcti_ctx_ptr_ptr(struct TSS2_TCTI_CONTEXT * value) { 
    return (struct TSS2_TCTI_CONTEXT * *)memcpy((struct TSS2_TCTI_CONTEXT * *)malloc(sizeof(struct TSS2_TCTI_CONTEXT *)),&value,sizeof(struct TSS2_TCTI_CONTEXT *));
  }

  static void delete_tcti_ctx_ptr_ptr(struct TSS2_TCTI_CONTEXT * *obj) { 
    if (obj) free((char*)obj);
  }

  static void tcti_ctx_ptr_ptr_assign(struct TSS2_TCTI_CONTEXT * *obj, struct TSS2_TCTI_CONTEXT * value) {
    *obj = value;
  }

  static struct TSS2_TCTI_CONTEXT * tcti_ctx_ptr_ptr_value(struct TSS2_TCTI_CONTEXT * *obj) {
    return *obj;
  }

#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssCreator_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TSS2_ABI_VERSION_tssCreator_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssCreator_set" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TSS2_ABI_VERSION_tssCreator_set" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  if (arg1) (arg1)->tssCreator = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssCreator_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TSS2_ABI_VERSION_tssCreator_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssCreator_get" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  result = (uint32_t) ((arg1)->tssCreator);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssFamily_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TSS2_ABI_VERSION_tssFamily_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssFamily_set" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TSS2_ABI_VERSION_tssFamily_set" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  if (arg1) (arg1)->tssFamily = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssFamily_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TSS2_ABI_VERSION_tssFamily_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssFamily_get" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  result = (uint32_t) ((arg1)->tssFamily);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssLevel_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TSS2_ABI_VERSION_tssLevel_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssLevel_set" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TSS2_ABI_VERSION_tssLevel_set" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  if (arg1) (arg1)->tssLevel = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssLevel_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TSS2_ABI_VERSION_tssLevel_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssLevel_get" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  result = (uint32_t) ((arg1)->tssLevel);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssVersion_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TSS2_ABI_VERSION_tssVersion_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssVersion_set" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TSS2_ABI_VERSION_tssVersion_set" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  if (arg1) (arg1)->tssVersion = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TSS2_ABI_VERSION_tssVersion_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TSS2_ABI_VERSION_tssVersion_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TSS2_ABI_VERSION_tssVersion_get" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  result = (uint32_t) ((arg1)->tssVersion);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TSS2_ABI_VERSION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TSS2_ABI_VERSION")) SWIG_fail;
  result = (TSS2_ABI_VERSION *)calloc(1, sizeof(TSS2_ABI_VERSION));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TSS2_ABI_VERSION, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TSS2_ABI_VERSION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_ABI_VERSION *arg1 = (TSS2_ABI_VERSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TSS2_ABI_VERSION",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_ABI_VERSION, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TSS2_ABI_VERSION" "', argument " "1"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg1 = (TSS2_ABI_VERSION *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TSS2_ABI_VERSION_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TSS2_ABI_VERSION, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *TSS2_RC_LAYER_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_RC_LAYER_SHIFT",SWIG_From_int((int)((16))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_GENERAL_FAILURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_GENERAL_FAILURE",SWIG_From_unsigned_SS_int((unsigned int)(1U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NOT_IMPLEMENTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NOT_IMPLEMENTED",SWIG_From_unsigned_SS_int((unsigned int)(2U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_CONTEXT",SWIG_From_unsigned_SS_int((unsigned int)(3U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_ABI_MISMATCH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_ABI_MISMATCH",SWIG_From_unsigned_SS_int((unsigned int)(4U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_REFERENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_REFERENCE",SWIG_From_unsigned_SS_int((unsigned int)(5U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_INSUFFICIENT_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_INSUFFICIENT_BUFFER",SWIG_From_unsigned_SS_int((unsigned int)(6U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_SEQUENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_SEQUENCE",SWIG_From_unsigned_SS_int((unsigned int)(7U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_CONNECTION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_CONNECTION",SWIG_From_unsigned_SS_int((unsigned int)(8U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_TRY_AGAIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_TRY_AGAIN",SWIG_From_unsigned_SS_int((unsigned int)(9U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_IO_ERROR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_IO_ERROR",SWIG_From_unsigned_SS_int((unsigned int)(10U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_VALUE",SWIG_From_unsigned_SS_int((unsigned int)(11U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NOT_PERMITTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NOT_PERMITTED",SWIG_From_unsigned_SS_int((unsigned int)(12U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_INVALID_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_INVALID_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(13U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_DECRYPT_PARAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_DECRYPT_PARAM",SWIG_From_unsigned_SS_int((unsigned int)(14U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_ENCRYPT_PARAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_ENCRYPT_PARAM",SWIG_From_unsigned_SS_int((unsigned int)(15U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_SIZE",SWIG_From_unsigned_SS_int((unsigned int)(16U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_MALFORMED_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_MALFORMED_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(17U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_INSUFFICIENT_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_INSUFFICIENT_CONTEXT",SWIG_From_unsigned_SS_int((unsigned int)(18U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_INSUFFICIENT_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_INSUFFICIENT_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(19U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_INCOMPATIBLE_TCTI_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_INCOMPATIBLE_TCTI",SWIG_From_unsigned_SS_int((unsigned int)(20U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NOT_SUPPORTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NOT_SUPPORTED",SWIG_From_unsigned_SS_int((unsigned int)(21U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_TCTI_STRUCTURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_TCTI_STRUCTURE",SWIG_From_unsigned_SS_int((unsigned int)(22U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_MEMORY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_MEMORY",SWIG_From_unsigned_SS_int((unsigned int)(23U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_TR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_TR",SWIG_From_unsigned_SS_int((unsigned int)(24U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_MULTIPLE_DECRYPT_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_MULTIPLE_DECRYPT_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(25U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_MULTIPLE_ENCRYPT_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_MULTIPLE_ENCRYPT_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(26U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_RSP_AUTH_FAILED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_RSP_AUTH_FAILED",SWIG_From_unsigned_SS_int((unsigned int)(27U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_CONFIG_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_CONFIG",SWIG_From_unsigned_SS_int((unsigned int)(28U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_PATH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_PATH",SWIG_From_unsigned_SS_int((unsigned int)(29U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NOT_DELETABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NOT_DELETABLE",SWIG_From_unsigned_SS_int((unsigned int)(30U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_PATH_ALREADY_EXISTS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_PATH_ALREADY_EXISTS",SWIG_From_unsigned_SS_int((unsigned int)(31U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_KEY_NOT_FOUND_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_KEY_NOT_FOUND",SWIG_From_unsigned_SS_int((unsigned int)(32U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_SIGNATURE_VERIFICATION_FAILED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_SIGNATURE_VERIFICATION_FAILED",SWIG_From_unsigned_SS_int((unsigned int)(33U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_HASH_MISMATCH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_HASH_MISMATCH",SWIG_From_unsigned_SS_int((unsigned int)(34U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_KEY_NOT_DUPLICABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_KEY_NOT_DUPLICABLE",SWIG_From_unsigned_SS_int((unsigned int)(35U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_PATH_NOT_FOUND_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_PATH_NOT_FOUND",SWIG_From_unsigned_SS_int((unsigned int)(36U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_CERT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_CERT",SWIG_From_unsigned_SS_int((unsigned int)(37U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_PCR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_PCR",SWIG_From_unsigned_SS_int((unsigned int)(38U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_PCR_NOT_RESETTABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_PCR_NOT_RESETTABLE",SWIG_From_unsigned_SS_int((unsigned int)(39U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_TEMPLATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_TEMPLATE",SWIG_From_unsigned_SS_int((unsigned int)(40U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_AUTHORIZATION_FAILED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_AUTHORIZATION_FAILED",SWIG_From_unsigned_SS_int((unsigned int)(41U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_AUTHORIZATION_UNKNOWN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_AUTHORIZATION_UNKNOWN",SWIG_From_unsigned_SS_int((unsigned int)(42U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NV_NOT_READABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NV_NOT_READABLE",SWIG_From_unsigned_SS_int((unsigned int)(43U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NV_TOO_SMALL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NV_TOO_SMALL",SWIG_From_unsigned_SS_int((unsigned int)(44U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NV_NOT_WRITEABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NV_NOT_WRITEABLE",SWIG_From_unsigned_SS_int((unsigned int)(45U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_POLICY_UNKNOWN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_POLICY_UNKNOWN",SWIG_From_unsigned_SS_int((unsigned int)(46U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NV_WRONG_TYPE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NV_WRONG_TYPE",SWIG_From_unsigned_SS_int((unsigned int)(47U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NAME_ALREADY_EXISTS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NAME_ALREADY_EXISTS",SWIG_From_unsigned_SS_int((unsigned int)(48U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_TPM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_TPM",SWIG_From_unsigned_SS_int((unsigned int)(49U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_BAD_KEY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_BAD_KEY",SWIG_From_unsigned_SS_int((unsigned int)(50U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_BASE_RC_NO_HANDLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_BASE_RC_NO_HANDLE",SWIG_From_unsigned_SS_int((unsigned int)(51U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_LAYER_IMPLEMENTATION_SPECIFIC_OFFSET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_LAYER_IMPLEMENTATION_SPECIFIC_OFFSET",SWIG_From_int((int)(0xf800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_LEVEL_IMPLEMENTATION_SPECIFIC_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_LEVEL_IMPLEMENTATION_SPECIFIC_SHIFT",SWIG_From_int((int)(11)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_COMMAND_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_COMMAND_SIZE",SWIG_From_int((int)(4096)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_RESPONSE_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_RESPONSE_SIZE",SWIG_From_int((int)(4096)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NUM_PCR_BANKS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NUM_PCR_BANKS",SWIG_From_int((int)(16)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_DIGEST_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_DIGEST_BUFFER",SWIG_From_int((int)(1024)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_NV_BUFFER_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_NV_BUFFER_SIZE",SWIG_From_int((int)(2048)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_PCRS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_PCRS",SWIG_From_int((int)(32)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_ALG_LIST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_ALG_LIST_SIZE",SWIG_From_int((int)(128)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_CAP_CC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_CAP_CC",SWIG_From_int((int)(256)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_CAP_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_CAP_BUFFER",SWIG_From_int((int)(1024)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_CONTEXT_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_CONTEXT_SIZE",SWIG_From_int((int)(5120)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SHA_DIGEST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SHA_DIGEST_SIZE",SWIG_From_int((int)(20)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SHA1_DIGEST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SHA1_DIGEST_SIZE",SWIG_From_int((int)(20)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SHA256_DIGEST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SHA256_DIGEST_SIZE",SWIG_From_int((int)(32)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SHA384_DIGEST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SHA384_DIGEST_SIZE",SWIG_From_int((int)(48)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SHA512_DIGEST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SHA512_DIGEST_SIZE",SWIG_From_int((int)(64)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SM3_256_DIGEST_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SM3_256_DIGEST_SIZE",SWIG_From_int((int)(32)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_SYM_BLOCK_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_SYM_BLOCK_SIZE",SWIG_From_int((int)(16)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_SYM_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_SYM_DATA",SWIG_From_int((int)(256)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_ECC_KEY_BYTES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_ECC_KEY_BYTES",SWIG_From_int((int)(128)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_SYM_KEY_BYTES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_SYM_KEY_BYTES",SWIG_From_int((int)(32)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_RSA_KEY_BYTES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_RSA_KEY_BYTES",SWIG_From_int((int)(512)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_LABEL_MAX_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_LABEL_MAX_BUFFER",SWIG_From_int((int)(32)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PCR_SELECT_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PCR_SELECT_MAX",SWIG_From_int((int)(((32+7)/8))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PRIVATE_VENDOR_SPECIFIC_BYTES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PRIVATE_VENDOR_SPECIFIC_BYTES",SWIG_From_int((int)(((512/2) *(3+2)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_EXTENDED_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_EXTENDED_SHIFT",SWIG_From_int((int)((5))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_COMMANDINDEX_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_COMMANDINDEX_SHIFT",SWIG_From_int((int)((0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_CHANDLES_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_CHANDLES_SHIFT",SWIG_From_int((int)((25))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_RES_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_RES_SHIFT",SWIG_From_int((int)((30))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NO",SWIG_From_int((int)(0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_YES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_YES",SWIG_From_int((int)(1)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_empty_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_EMPTY_empty_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_empty_set" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_EMPTY_empty_set" "', argument " "2"" of type '" "BYTE [1]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)1; ++ii) *(BYTE *)&arg1->empty[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""empty""' of type '""BYTE [1]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_empty_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_EMPTY_empty_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_empty_get" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->empty);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_EMPTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_EMPTY")) SWIG_fail;
  result = (TPMS_EMPTY *)calloc(1, sizeof(TPMS_EMPTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_EMPTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_EMPTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_EMPTY" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_EMPTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_EMPTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_alg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  TPM2_ALG_ID arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DESCRIPTION_alg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_alg_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DESCRIPTION_alg_set" "', argument " "2"" of type '" "TPM2_ALG_ID""'");
  } 
  arg2 = (TPM2_ALG_ID)(val2);
  if (arg1) (arg1)->alg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_alg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ALG_ID result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DESCRIPTION_alg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_alg_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  result = (TPM2_ALG_ID) ((arg1)->alg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_attributes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  TPMA_ALGORITHM arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DESCRIPTION_attributes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_attributes_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DESCRIPTION_attributes_set" "', argument " "2"" of type '" "TPMA_ALGORITHM""'");
  } 
  arg2 = (TPMA_ALGORITHM)(val2);
  if (arg1) (arg1)->attributes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_attributes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_ALGORITHM result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DESCRIPTION_attributes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_attributes_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  result = (TPMA_ALGORITHM) ((arg1)->attributes);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DESCRIPTION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALGORITHM_DESCRIPTION")) SWIG_fail;
  result = (TPMS_ALGORITHM_DESCRIPTION *)calloc(1, sizeof(TPMS_ALGORITHM_DESCRIPTION));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DESCRIPTION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DESCRIPTION",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DESCRIPTION" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ALGORITHM_DESCRIPTION_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_HA_sha1_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_sha1_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha1_set" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_sha1_set" "', argument " "2"" of type '" "BYTE [20]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)20; ++ii) *(BYTE *)&arg1->sha1[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""sha1""' of type '""BYTE [20]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha1_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_sha1_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha1_get" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->sha1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha256_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_sha256_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha256_set" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_sha256_set" "', argument " "2"" of type '" "BYTE [32]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)32; ++ii) *(BYTE *)&arg1->sha256[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""sha256""' of type '""BYTE [32]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha256_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_sha256_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha256_get" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->sha256);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha384_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_sha384_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha384_set" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_sha384_set" "', argument " "2"" of type '" "BYTE [48]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)48; ++ii) *(BYTE *)&arg1->sha384[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""sha384""' of type '""BYTE [48]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha384_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_sha384_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha384_get" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->sha384);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha512_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_sha512_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha512_set" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_sha512_set" "', argument " "2"" of type '" "BYTE [64]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)64; ++ii) *(BYTE *)&arg1->sha512[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""sha512""' of type '""BYTE [64]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sha512_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_sha512_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sha512_get" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->sha512);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sm3_256_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_sm3_256_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sm3_256_set" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_sm3_256_set" "', argument " "2"" of type '" "BYTE [32]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)32; ++ii) *(BYTE *)&arg1->sm3_256[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""sm3_256""' of type '""BYTE [32]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_sm3_256_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_sm3_256_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_sm3_256_get" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->sm3_256);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_HA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_HA")) SWIG_fail;
  result = (TPMU_HA *)calloc(1, sizeof(TPMU_HA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_HA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_HA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_HA" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_HA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_HA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_HA_hashAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_HA_hashAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_hashAlg_set" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_HA_hashAlg_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hashAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_hashAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_HA_hashAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_hashAlg_get" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hashAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_digest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  TPMU_HA *arg2 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_HA_digest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_digest_set" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_HA_digest_set" "', argument " "2"" of type '" "TPMU_HA *""'"); 
  }
  arg2 = (TPMU_HA *)(argp2);
  if (arg1) (arg1)->digest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_digest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_HA_digest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_digest_get" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  result = (TPMU_HA *)& ((arg1)->digest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_HA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_HA")) SWIG_fail;
  result = (TPMT_HA *)calloc(1, sizeof(TPMT_HA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_HA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_HA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_HA" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_HA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_HA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DIGEST_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_size_set" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_DIGEST_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DIGEST_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_size_get" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DIGEST_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_buffer_set" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_DIGEST_buffer_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMU_HA)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMU_HA); ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [sizeof(TPMU_HA)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DIGEST_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_buffer_get" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_DIGEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_DIGEST")) SWIG_fail;
  result = (TPM2B_DIGEST *)calloc(1, sizeof(TPM2B_DIGEST));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DIGEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DIGEST",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DIGEST" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_DIGEST_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_DIGEST, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_DATA_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DATA_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_size_set" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_DATA_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DATA_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_size_get" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DATA_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_buffer_set" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_DATA_buffer_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMU_HA)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMU_HA); ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [sizeof(TPMU_HA)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DATA_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_buffer_get" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_DATA")) SWIG_fail;
  result = (TPM2B_DATA *)calloc(1, sizeof(TPM2B_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DATA" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_EVENT_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_EVENT_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_size_set" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_EVENT_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_EVENT_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_size_get" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_EVENT_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_buffer_set" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_EVENT_buffer_set" "', argument " "2"" of type '" "BYTE [1024]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)1024; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [1024]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_EVENT_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_buffer_get" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_EVENT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_EVENT")) SWIG_fail;
  result = (TPM2B_EVENT *)calloc(1, sizeof(TPM2B_EVENT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_EVENT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_EVENT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_EVENT" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_EVENT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_EVENT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_BUFFER_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_size_set" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_MAX_BUFFER_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_BUFFER_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_size_get" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_BUFFER_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_buffer_set" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_MAX_BUFFER_buffer_set" "', argument " "2"" of type '" "BYTE [1024]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)1024; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [1024]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_BUFFER_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_buffer_get" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_BUFFER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_MAX_BUFFER")) SWIG_fail;
  result = (TPM2B_MAX_BUFFER *)calloc(1, sizeof(TPM2B_MAX_BUFFER));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_BUFFER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_BUFFER",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_BUFFER" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_MAX_BUFFER_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_MAX_BUFFER, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_NV_BUFFER_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_size_set" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_MAX_NV_BUFFER_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_NV_BUFFER_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_size_get" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_NV_BUFFER_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_buffer_set" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_MAX_NV_BUFFER_buffer_set" "', argument " "2"" of type '" "BYTE [2048]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)2048; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [2048]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_NV_BUFFER_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_buffer_get" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_NV_BUFFER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_MAX_NV_BUFFER")) SWIG_fail;
  result = (TPM2B_MAX_NV_BUFFER *)calloc(1, sizeof(TPM2B_MAX_NV_BUFFER));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_NV_BUFFER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_NV_BUFFER",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_NV_BUFFER" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_MAX_NV_BUFFER_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_IV_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_IV_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_size_set" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_IV_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_IV_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_size_get" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_IV_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_buffer_set" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_IV_buffer_set" "', argument " "2"" of type '" "BYTE [16]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [16]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_IV_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_buffer_get" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_IV(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_IV")) SWIG_fail;
  result = (TPM2B_IV *)calloc(1, sizeof(TPM2B_IV));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_IV(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_IV",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_IV" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_IV_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_IV, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_NAME_digest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  TPMT_HA *arg2 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_NAME_digest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_digest_set" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_NAME_digest_set" "', argument " "2"" of type '" "TPMT_HA *""'"); 
  }
  arg2 = (TPMT_HA *)(argp2);
  if (arg1) (arg1)->digest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_digest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_NAME_digest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_digest_get" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  result = (TPMT_HA *)& ((arg1)->digest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_handle_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  TPM2_HANDLE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_NAME_handle_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_handle_set" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_NAME_handle_set" "', argument " "2"" of type '" "TPM2_HANDLE""'");
  } 
  arg2 = (TPM2_HANDLE)(val2);
  if (arg1) (arg1)->handle = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_handle_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_HANDLE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_NAME_handle_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_handle_get" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  result = (TPM2_HANDLE) ((arg1)->handle);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_NAME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_NAME")) SWIG_fail;
  result = (TPMU_NAME *)calloc(1, sizeof(TPMU_NAME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_NAME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_NAME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_NAME" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_NAME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_NAME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_NAME_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NAME_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_size_set" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_NAME_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NAME_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_size_get" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_name_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NAME_name_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_name_set" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_NAME_name_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMU_NAME)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMU_NAME); ++ii) *(BYTE *)&arg1->name[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""name""' of type '""BYTE [sizeof(TPMU_NAME)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_name_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NAME_name_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_name_get" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->name);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_NAME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_NAME")) SWIG_fail;
  result = (TPM2B_NAME *)calloc(1, sizeof(TPM2B_NAME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NAME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NAME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NAME" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_NAME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_NAME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_sizeofSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  UINT8 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECT_sizeofSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_sizeofSelect_set" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECT_sizeofSelect_set" "', argument " "2"" of type '" "UINT8""'");
  } 
  arg2 = (UINT8)(val2);
  if (arg1) (arg1)->sizeofSelect = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_sizeofSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECT_sizeofSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_sizeofSelect_get" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  result = (UINT8) ((arg1)->sizeofSelect);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_pcrSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECT_pcrSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_pcrSelect_set" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_PCR_SELECT_pcrSelect_set" "', argument " "2"" of type '" "BYTE [((32+7)/8)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((32+7)/8); ++ii) *(BYTE *)&arg1->pcrSelect[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""pcrSelect""' of type '""BYTE [((32+7)/8)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_pcrSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECT_pcrSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_pcrSelect_get" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->pcrSelect);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_PCR_SELECT")) SWIG_fail;
  result = (TPMS_PCR_SELECT *)calloc(1, sizeof(TPMS_PCR_SELECT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECT" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_PCR_SELECT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_PCR_SELECT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_hash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECTION_hash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_hash_set" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECTION_hash_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hash = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_hash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_hash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_hash_get" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hash);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_sizeofSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  UINT8 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECTION_sizeofSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_sizeofSelect_set" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECTION_sizeofSelect_set" "', argument " "2"" of type '" "UINT8""'");
  } 
  arg2 = (UINT8)(val2);
  if (arg1) (arg1)->sizeofSelect = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_sizeofSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_sizeofSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_sizeofSelect_get" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  result = (UINT8) ((arg1)->sizeofSelect);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_pcrSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECTION_pcrSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_pcrSelect_set" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_PCR_SELECTION_pcrSelect_set" "', argument " "2"" of type '" "BYTE [((32+7)/8)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((32+7)/8); ++ii) *(BYTE *)&arg1->pcrSelect[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""pcrSelect""' of type '""BYTE [((32+7)/8)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_pcrSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_pcrSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_pcrSelect_get" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->pcrSelect);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECTION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_PCR_SELECTION")) SWIG_fail;
  result = (TPMS_PCR_SELECTION *)calloc(1, sizeof(TPMS_PCR_SELECTION));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECTION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECTION",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECTION" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_PCR_SELECTION_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_PCR_SELECTION, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_tag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  TPM2_ST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_CREATION_tag_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_tag_set" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_CREATION_tag_set" "', argument " "2"" of type '" "TPM2_ST""'");
  } 
  arg2 = (TPM2_ST)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_tag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_tag_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_tag_get" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  result = (TPM2_ST) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_hierarchy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  TPMI_RH_HIERARCHY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_CREATION_hierarchy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_hierarchy_set" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_CREATION_hierarchy_set" "', argument " "2"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg2 = (TPMI_RH_HIERARCHY)(val2);
  if (arg1) (arg1)->hierarchy = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_hierarchy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RH_HIERARCHY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_hierarchy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_hierarchy_get" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  result = (TPMI_RH_HIERARCHY) ((arg1)->hierarchy);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_digest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_CREATION_digest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_digest_set" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_CREATION_digest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->digest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_digest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_digest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_digest_get" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->digest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_CREATION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_CREATION")) SWIG_fail;
  result = (TPMT_TK_CREATION *)calloc(1, sizeof(TPMT_TK_CREATION));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_CREATION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_CREATION",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_CREATION" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_CREATION_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_CREATION, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_tag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  TPM2_ST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_VERIFIED_tag_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_tag_set" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_VERIFIED_tag_set" "', argument " "2"" of type '" "TPM2_ST""'");
  } 
  arg2 = (TPM2_ST)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_tag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_tag_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_tag_get" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  result = (TPM2_ST) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_hierarchy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  TPMI_RH_HIERARCHY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_VERIFIED_hierarchy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_hierarchy_set" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_VERIFIED_hierarchy_set" "', argument " "2"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg2 = (TPMI_RH_HIERARCHY)(val2);
  if (arg1) (arg1)->hierarchy = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_hierarchy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RH_HIERARCHY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_hierarchy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_hierarchy_get" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  result = (TPMI_RH_HIERARCHY) ((arg1)->hierarchy);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_digest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_VERIFIED_digest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_digest_set" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_VERIFIED_digest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->digest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_digest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_digest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_digest_get" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->digest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_VERIFIED(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_VERIFIED")) SWIG_fail;
  result = (TPMT_TK_VERIFIED *)calloc(1, sizeof(TPMT_TK_VERIFIED));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_VERIFIED(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_VERIFIED",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_VERIFIED" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_VERIFIED_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_VERIFIED, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_tag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  TPM2_ST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_AUTH_tag_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_tag_set" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_AUTH_tag_set" "', argument " "2"" of type '" "TPM2_ST""'");
  } 
  arg2 = (TPM2_ST)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_tag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_tag_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_tag_get" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  result = (TPM2_ST) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_hierarchy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  TPMI_RH_HIERARCHY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_AUTH_hierarchy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_hierarchy_set" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_AUTH_hierarchy_set" "', argument " "2"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg2 = (TPMI_RH_HIERARCHY)(val2);
  if (arg1) (arg1)->hierarchy = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_hierarchy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RH_HIERARCHY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_hierarchy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_hierarchy_get" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  result = (TPMI_RH_HIERARCHY) ((arg1)->hierarchy);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_digest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_AUTH_digest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_digest_set" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_AUTH_digest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->digest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_digest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_digest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_digest_get" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->digest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_AUTH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_AUTH")) SWIG_fail;
  result = (TPMT_TK_AUTH *)calloc(1, sizeof(TPMT_TK_AUTH));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_AUTH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_AUTH",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_AUTH" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_AUTH_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_AUTH, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_tag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  TPM2_ST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_HASHCHECK_tag_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_tag_set" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_HASHCHECK_tag_set" "', argument " "2"" of type '" "TPM2_ST""'");
  } 
  arg2 = (TPM2_ST)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_tag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_tag_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_tag_get" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  result = (TPM2_ST) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_hierarchy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  TPMI_RH_HIERARCHY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_HASHCHECK_hierarchy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_hierarchy_set" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_HASHCHECK_hierarchy_set" "', argument " "2"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg2 = (TPMI_RH_HIERARCHY)(val2);
  if (arg1) (arg1)->hierarchy = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_hierarchy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RH_HIERARCHY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_hierarchy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_hierarchy_get" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  result = (TPMI_RH_HIERARCHY) ((arg1)->hierarchy);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_digest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_HASHCHECK_digest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_digest_set" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_HASHCHECK_digest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->digest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_digest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_digest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_digest_get" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->digest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_HASHCHECK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_HASHCHECK")) SWIG_fail;
  result = (TPMT_TK_HASHCHECK *)calloc(1, sizeof(TPMT_TK_HASHCHECK));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_HASHCHECK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_HASHCHECK",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_HASHCHECK" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_HASHCHECK_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_HASHCHECK, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_alg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  TPM2_ALG_ID arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALG_PROPERTY_alg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_alg_set" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALG_PROPERTY_alg_set" "', argument " "2"" of type '" "TPM2_ALG_ID""'");
  } 
  arg2 = (TPM2_ALG_ID)(val2);
  if (arg1) (arg1)->alg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_alg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ALG_ID result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALG_PROPERTY_alg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_alg_get" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  result = (TPM2_ALG_ID) ((arg1)->alg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_algProperties_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  TPMA_ALGORITHM arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALG_PROPERTY_algProperties_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_algProperties_set" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALG_PROPERTY_algProperties_set" "', argument " "2"" of type '" "TPMA_ALGORITHM""'");
  } 
  arg2 = (TPMA_ALGORITHM)(val2);
  if (arg1) (arg1)->algProperties = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_algProperties_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_ALGORITHM result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALG_PROPERTY_algProperties_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_algProperties_get" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  result = (TPMA_ALGORITHM) ((arg1)->algProperties);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALG_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALG_PROPERTY")) SWIG_fail;
  result = (TPMS_ALG_PROPERTY *)calloc(1, sizeof(TPMS_ALG_PROPERTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALG_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALG_PROPERTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALG_PROPERTY" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ALG_PROPERTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ALG_PROPERTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_property_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  TPM2_PT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PROPERTY_property_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_property_set" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PROPERTY_property_set" "', argument " "2"" of type '" "TPM2_PT""'");
  } 
  arg2 = (TPM2_PT)(val2);
  if (arg1) (arg1)->property = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_property_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_PT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PROPERTY_property_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_property_get" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  result = (TPM2_PT) ((arg1)->property);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_value_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PROPERTY_value_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_value_set" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PROPERTY_value_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->value = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_value_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PROPERTY_value_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_value_get" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  result = (UINT32) ((arg1)->value);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_PROPERTY")) SWIG_fail;
  result = (TPMS_TAGGED_PROPERTY *)calloc(1, sizeof(TPMS_TAGGED_PROPERTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PROPERTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PROPERTY" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TAGGED_PROPERTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TAGGED_PROPERTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_tag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  TPM2_PT_PCR arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PCR_SELECT_tag_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_tag_set" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PCR_SELECT_tag_set" "', argument " "2"" of type '" "TPM2_PT_PCR""'");
  } 
  arg2 = (TPM2_PT_PCR)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_tag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_PT_PCR result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_tag_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_tag_get" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  result = (TPM2_PT_PCR) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_sizeofSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  UINT8 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PCR_SELECT_sizeofSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_sizeofSelect_set" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PCR_SELECT_sizeofSelect_set" "', argument " "2"" of type '" "UINT8""'");
  } 
  arg2 = (UINT8)(val2);
  if (arg1) (arg1)->sizeofSelect = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_sizeofSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_sizeofSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_sizeofSelect_get" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  result = (UINT8) ((arg1)->sizeofSelect);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_pcrSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PCR_SELECT_pcrSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_pcrSelect_set" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_PCR_SELECT_pcrSelect_set" "', argument " "2"" of type '" "BYTE [((32+7)/8)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((32+7)/8); ++ii) *(BYTE *)&arg1->pcrSelect[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""pcrSelect""' of type '""BYTE [((32+7)/8)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_pcrSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_pcrSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_pcrSelect_get" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->pcrSelect);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PCR_SELECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_PCR_SELECT")) SWIG_fail;
  result = (TPMS_TAGGED_PCR_SELECT *)calloc(1, sizeof(TPMS_TAGGED_PCR_SELECT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PCR_SELECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PCR_SELECT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PCR_SELECT" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TAGGED_PCR_SELECT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_handle_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  TPM2_HANDLE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_POLICY_handle_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_handle_set" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_POLICY_handle_set" "', argument " "2"" of type '" "TPM2_HANDLE""'");
  } 
  arg2 = (TPM2_HANDLE)(val2);
  if (arg1) (arg1)->handle = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_handle_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_HANDLE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_POLICY_handle_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_handle_get" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  result = (TPM2_HANDLE) ((arg1)->handle);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_policyHash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  TPMT_HA *arg2 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_POLICY_policyHash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_policyHash_set" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_POLICY_policyHash_set" "', argument " "2"" of type '" "TPMT_HA *""'"); 
  }
  arg2 = (TPMT_HA *)(argp2);
  if (arg1) (arg1)->policyHash = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_policyHash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_POLICY_policyHash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_policyHash_get" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  result = (TPMT_HA *)& ((arg1)->policyHash);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_POLICY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_POLICY")) SWIG_fail;
  result = (TPMS_TAGGED_POLICY *)calloc(1, sizeof(TPMS_TAGGED_POLICY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_POLICY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_POLICY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_POLICY" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TAGGED_POLICY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TAGGED_POLICY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_CC_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CC_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_count_set" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_CC_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CC_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_count_get" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_commandCodes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  TPM2_CC *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CC_commandCodes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_commandCodes_set" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_CC_commandCodes_set" "', argument " "2"" of type '" "TPM2_CC [256]""'"); 
  } 
  arg2 = (TPM2_CC *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)256; ++ii) *(TPM2_CC *)&arg1->commandCodes[ii] = *((TPM2_CC *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""commandCodes""' of type '""TPM2_CC [256]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_commandCodes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CC_commandCodes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_commandCodes_get" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  result = (TPM2_CC *)(TPM2_CC *) ((arg1)->commandCodes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_CC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_CC")) SWIG_fail;
  result = (TPML_CC *)calloc(1, sizeof(TPML_CC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CC" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_CC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_CC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_CCA_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CCA_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_count_set" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_CCA_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CCA_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_count_get" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_commandAttributes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  TPMA_CC *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CCA_commandAttributes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_commandAttributes_set" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_CCA_commandAttributes_set" "', argument " "2"" of type '" "TPMA_CC [256]""'"); 
  } 
  arg2 = (TPMA_CC *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)256; ++ii) *(TPMA_CC *)&arg1->commandAttributes[ii] = *((TPMA_CC *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""commandAttributes""' of type '""TPMA_CC [256]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_commandAttributes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CCA_commandAttributes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_commandAttributes_get" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  result = (TPMA_CC *)(TPMA_CC *) ((arg1)->commandAttributes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_CCA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_CCA")) SWIG_fail;
  result = (TPML_CCA *)calloc(1, sizeof(TPML_CCA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CCA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CCA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CCA" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_CCA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_CCA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_ALG_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_count_set" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ALG_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_count_get" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_algorithms_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  TPM2_ALG_ID *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_algorithms_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_algorithms_set" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ALG_algorithms_set" "', argument " "2"" of type '" "TPM2_ALG_ID [128]""'"); 
  } 
  arg2 = (TPM2_ALG_ID *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)128; ++ii) *(TPM2_ALG_ID *)&arg1->algorithms[ii] = *((TPM2_ALG_ID *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""algorithms""' of type '""TPM2_ALG_ID [128]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_algorithms_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ALG_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_algorithms_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_algorithms_get" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  result = (TPM2_ALG_ID *)(TPM2_ALG_ID *) ((arg1)->algorithms);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_short, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_ALG(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ALG")) SWIG_fail;
  result = (TPML_ALG *)calloc(1, sizeof(TPML_ALG));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_ALG_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_ALG, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_HANDLE_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_HANDLE_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_count_set" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_HANDLE_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_HANDLE_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_count_get" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_handle_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  TPM2_HANDLE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_HANDLE_handle_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_handle_set" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_HANDLE_handle_set" "', argument " "2"" of type '" "TPM2_HANDLE [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPM2_HANDLE))]""'"); 
  } 
  arg2 = (TPM2_HANDLE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPM2_HANDLE)); ++ii) *(TPM2_HANDLE *)&arg1->handle[ii] = *((TPM2_HANDLE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""handle""' of type '""TPM2_HANDLE [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPM2_HANDLE))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_handle_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_HANDLE_handle_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_handle_get" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  result = (TPM2_HANDLE *)(TPM2_HANDLE *) ((arg1)->handle);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_HANDLE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_HANDLE")) SWIG_fail;
  result = (TPML_HANDLE *)calloc(1, sizeof(TPML_HANDLE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_HANDLE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_HANDLE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_HANDLE" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_HANDLE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_HANDLE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_DIGEST_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_count_set" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_DIGEST_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_count_get" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_digests_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  TPM2B_DIGEST *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_digests_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_digests_set" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_DIGEST_digests_set" "', argument " "2"" of type '" "TPM2B_DIGEST [8]""'"); 
  } 
  arg2 = (TPM2B_DIGEST *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)8; ++ii) *(TPM2B_DIGEST *)&arg1->digests[ii] = *((TPM2B_DIGEST *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""digests""' of type '""TPM2B_DIGEST [8]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_digests_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_digests_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_digests_get" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  result = (TPM2B_DIGEST *)(TPM2B_DIGEST *) ((arg1)->digests);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_DIGEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_DIGEST")) SWIG_fail;
  result = (TPML_DIGEST *)calloc(1, sizeof(TPML_DIGEST));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_DIGEST_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_DIGEST, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_VALUES_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_count_set" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_DIGEST_VALUES_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_VALUES_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_count_get" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_digests_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  TPMT_HA *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_VALUES_digests_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_digests_set" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_DIGEST_VALUES_digests_set" "', argument " "2"" of type '" "TPMT_HA [16]""'"); 
  } 
  arg2 = (TPMT_HA *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(TPMT_HA *)&arg1->digests[ii] = *((TPMT_HA *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""digests""' of type '""TPMT_HA [16]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_digests_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_VALUES_digests_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_digests_get" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  result = (TPMT_HA *)(TPMT_HA *) ((arg1)->digests);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_VALUES(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_DIGEST_VALUES")) SWIG_fail;
  result = (TPML_DIGEST_VALUES *)calloc(1, sizeof(TPML_DIGEST_VALUES));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_VALUES(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_VALUES",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_VALUES" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_DIGEST_VALUES_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_DIGEST_VALUES, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_PCR_SELECTION_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_count_set" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_PCR_SELECTION_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_PCR_SELECTION_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_count_get" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_pcrSelections_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  TPMS_PCR_SELECTION *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_PCR_SELECTION_pcrSelections_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_pcrSelections_set" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_PCR_SELECTION_pcrSelections_set" "', argument " "2"" of type '" "TPMS_PCR_SELECTION [16]""'"); 
  } 
  arg2 = (TPMS_PCR_SELECTION *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)16; ++ii) *(TPMS_PCR_SELECTION *)&arg1->pcrSelections[ii] = *((TPMS_PCR_SELECTION *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""pcrSelections""' of type '""TPMS_PCR_SELECTION [16]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_pcrSelections_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_PCR_SELECTION_pcrSelections_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_pcrSelections_get" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  result = (TPMS_PCR_SELECTION *)(TPMS_PCR_SELECTION *) ((arg1)->pcrSelections);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_PCR_SELECTION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_PCR_SELECTION")) SWIG_fail;
  result = (TPML_PCR_SELECTION *)calloc(1, sizeof(TPML_PCR_SELECTION));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_PCR_SELECTION(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_PCR_SELECTION",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_PCR_SELECTION" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_PCR_SELECTION_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_PCR_SELECTION, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PROPERTY_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_count_set" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ALG_PROPERTY_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PROPERTY_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_count_get" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_algProperties_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  TPMS_ALG_PROPERTY *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PROPERTY_algProperties_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_algProperties_set" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ALG_PROPERTY_algProperties_set" "', argument " "2"" of type '" "TPMS_ALG_PROPERTY [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_ALG_PROPERTY))]""'"); 
  } 
  arg2 = (TPMS_ALG_PROPERTY *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_ALG_PROPERTY)); ++ii) *(TPMS_ALG_PROPERTY *)&arg1->algProperties[ii] = *((TPMS_ALG_PROPERTY *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""algProperties""' of type '""TPMS_ALG_PROPERTY [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_ALG_PROPERTY))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_algProperties_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PROPERTY_algProperties_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_algProperties_get" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  result = (TPMS_ALG_PROPERTY *)(TPMS_ALG_PROPERTY *) ((arg1)->algProperties);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_ALG_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ALG_PROPERTY")) SWIG_fail;
  result = (TPML_ALG_PROPERTY *)calloc(1, sizeof(TPML_ALG_PROPERTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_PROPERTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_PROPERTY" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_ALG_PROPERTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_ALG_PROPERTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_TPM_PROPERTY_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_count_set" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_TAGGED_TPM_PROPERTY_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_TPM_PROPERTY_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_count_get" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_tpmProperty_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  TPMS_TAGGED_PROPERTY *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_TPM_PROPERTY_tpmProperty_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_tpmProperty_set" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_TAGGED_TPM_PROPERTY_tpmProperty_set" "', argument " "2"" of type '" "TPMS_TAGGED_PROPERTY [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_TAGGED_PROPERTY))]""'"); 
  } 
  arg2 = (TPMS_TAGGED_PROPERTY *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_TAGGED_PROPERTY)); ++ii) *(TPMS_TAGGED_PROPERTY *)&arg1->tpmProperty[ii] = *((TPMS_TAGGED_PROPERTY *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""tpmProperty""' of type '""TPMS_TAGGED_PROPERTY [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_TAGGED_PROPERTY))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_tpmProperty_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_TPM_PROPERTY_tpmProperty_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_tpmProperty_get" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  result = (TPMS_TAGGED_PROPERTY *)(TPMS_TAGGED_PROPERTY *) ((arg1)->tpmProperty);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_TPM_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_TAGGED_TPM_PROPERTY")) SWIG_fail;
  result = (TPML_TAGGED_TPM_PROPERTY *)calloc(1, sizeof(TPML_TAGGED_TPM_PROPERTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_TPM_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_TPM_PROPERTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_TPM_PROPERTY" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_TAGGED_TPM_PROPERTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_PCR_PROPERTY_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_count_set" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_TAGGED_PCR_PROPERTY_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_PCR_PROPERTY_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_count_get" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_pcrProperty_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  TPMS_TAGGED_PCR_SELECT *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_PCR_PROPERTY_pcrProperty_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_pcrProperty_set" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_TAGGED_PCR_PROPERTY_pcrProperty_set" "', argument " "2"" of type '" "TPMS_TAGGED_PCR_SELECT [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_TAGGED_PCR_SELECT))]""'"); 
  } 
  arg2 = (TPMS_TAGGED_PCR_SELECT *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_TAGGED_PCR_SELECT)); ++ii) *(TPMS_TAGGED_PCR_SELECT *)&arg1->pcrProperty[ii] = *((TPMS_TAGGED_PCR_SELECT *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""pcrProperty""' of type '""TPMS_TAGGED_PCR_SELECT [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPMS_TAGGED_PCR_SELECT))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_pcrProperty_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_PCR_PROPERTY_pcrProperty_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_pcrProperty_get" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  result = (TPMS_TAGGED_PCR_SELECT *)(TPMS_TAGGED_PCR_SELECT *) ((arg1)->pcrProperty);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_PCR_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_TAGGED_PCR_PROPERTY")) SWIG_fail;
  result = (TPML_TAGGED_PCR_PROPERTY *)calloc(1, sizeof(TPML_TAGGED_PCR_PROPERTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_PCR_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_PCR_PROPERTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_PCR_PROPERTY" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_TAGGED_PCR_PROPERTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ECC_CURVE_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_count_set" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ECC_CURVE_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ECC_CURVE_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_count_get" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_eccCurves_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  TPM2_ECC_CURVE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ECC_CURVE_eccCurves_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_eccCurves_set" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ECC_CURVE_eccCurves_set" "', argument " "2"" of type '" "TPM2_ECC_CURVE [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPM2_ECC_CURVE))]""'"); 
  } 
  arg2 = (TPM2_ECC_CURVE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPM2_ECC_CURVE)); ++ii) *(TPM2_ECC_CURVE *)&arg1->eccCurves[ii] = *((TPM2_ECC_CURVE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""eccCurves""' of type '""TPM2_ECC_CURVE [((1024-sizeof(TPM2_CAP)-sizeof(UINT32))/sizeof(TPM2_ECC_CURVE))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_eccCurves_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ECC_CURVE_eccCurves_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_eccCurves_get" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  result = (TPM2_ECC_CURVE *)(TPM2_ECC_CURVE *) ((arg1)->eccCurves);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_short, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_ECC_CURVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ECC_CURVE")) SWIG_fail;
  result = (TPML_ECC_CURVE *)calloc(1, sizeof(TPML_ECC_CURVE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ECC_CURVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ECC_CURVE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ECC_CURVE" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_ECC_CURVE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_ECC_CURVE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_INTEL_PTT_PROPERTY_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_count_set" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_INTEL_PTT_PROPERTY_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_INTEL_PTT_PROPERTY_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_count_get" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_property_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  UINT32 *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_INTEL_PTT_PROPERTY_property_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_property_set" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_INTEL_PTT_PROPERTY_property_set" "', argument " "2"" of type '" "UINT32 [(1024/sizeof(UINT32))]""'"); 
  } 
  arg2 = (UINT32 *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)(1024/sizeof(UINT32)); ++ii) *(UINT32 *)&arg1->property[ii] = *((UINT32 *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""property""' of type '""UINT32 [(1024/sizeof(UINT32))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_property_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_INTEL_PTT_PROPERTY_property_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_property_get" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  result = (UINT32 *)(UINT32 *) ((arg1)->property);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_INTEL_PTT_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_INTEL_PTT_PROPERTY")) SWIG_fail;
  result = (TPML_INTEL_PTT_PROPERTY *)calloc(1, sizeof(TPML_INTEL_PTT_PROPERTY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_INTEL_PTT_PROPERTY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_INTEL_PTT_PROPERTY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_INTEL_PTT_PROPERTY" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_INTEL_PTT_PROPERTY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_algorithms_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_ALG_PROPERTY *arg2 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_algorithms_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_algorithms_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_algorithms_set" "', argument " "2"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg2 = (TPML_ALG_PROPERTY *)(argp2);
  if (arg1) (arg1)->algorithms = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_algorithms_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_algorithms_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_algorithms_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_ALG_PROPERTY *)& ((arg1)->algorithms);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_handles_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_HANDLE *arg2 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_handles_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_handles_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_handles_set" "', argument " "2"" of type '" "TPML_HANDLE *""'"); 
  }
  arg2 = (TPML_HANDLE *)(argp2);
  if (arg1) (arg1)->handles = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_handles_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_handles_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_handles_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_HANDLE *)& ((arg1)->handles);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_command_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_CCA *arg2 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_command_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_command_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_command_set" "', argument " "2"" of type '" "TPML_CCA *""'"); 
  }
  arg2 = (TPML_CCA *)(argp2);
  if (arg1) (arg1)->command = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_command_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_command_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_command_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_CCA *)& ((arg1)->command);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_ppCommands_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_CC *arg2 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_ppCommands_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_ppCommands_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_ppCommands_set" "', argument " "2"" of type '" "TPML_CC *""'"); 
  }
  arg2 = (TPML_CC *)(argp2);
  if (arg1) (arg1)->ppCommands = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_ppCommands_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_ppCommands_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_ppCommands_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_CC *)& ((arg1)->ppCommands);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_auditCommands_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_CC *arg2 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_auditCommands_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_auditCommands_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_auditCommands_set" "', argument " "2"" of type '" "TPML_CC *""'"); 
  }
  arg2 = (TPML_CC *)(argp2);
  if (arg1) (arg1)->auditCommands = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_auditCommands_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_auditCommands_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_auditCommands_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_CC *)& ((arg1)->auditCommands);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_assignedPCR_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_PCR_SELECTION *arg2 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_assignedPCR_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_assignedPCR_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_assignedPCR_set" "', argument " "2"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg2 = (TPML_PCR_SELECTION *)(argp2);
  if (arg1) (arg1)->assignedPCR = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_assignedPCR_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_assignedPCR_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_assignedPCR_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_PCR_SELECTION *)& ((arg1)->assignedPCR);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_tpmProperties_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_TAGGED_TPM_PROPERTY *arg2 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_tpmProperties_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_tpmProperties_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_tpmProperties_set" "', argument " "2"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg2 = (TPML_TAGGED_TPM_PROPERTY *)(argp2);
  if (arg1) (arg1)->tpmProperties = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_tpmProperties_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_tpmProperties_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_tpmProperties_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_TAGGED_TPM_PROPERTY *)& ((arg1)->tpmProperties);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_pcrProperties_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_TAGGED_PCR_PROPERTY *arg2 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_pcrProperties_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_pcrProperties_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_pcrProperties_set" "', argument " "2"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg2 = (TPML_TAGGED_PCR_PROPERTY *)(argp2);
  if (arg1) (arg1)->pcrProperties = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_pcrProperties_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_pcrProperties_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_pcrProperties_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_TAGGED_PCR_PROPERTY *)& ((arg1)->pcrProperties);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_eccCurves_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_ECC_CURVE *arg2 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_eccCurves_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_eccCurves_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_eccCurves_set" "', argument " "2"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg2 = (TPML_ECC_CURVE *)(argp2);
  if (arg1) (arg1)->eccCurves = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_eccCurves_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_eccCurves_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_eccCurves_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_ECC_CURVE *)& ((arg1)->eccCurves);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_intelPttProperty_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPML_INTEL_PTT_PROPERTY *arg2 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_intelPttProperty_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_intelPttProperty_set" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_intelPttProperty_set" "', argument " "2"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg2 = (TPML_INTEL_PTT_PROPERTY *)(argp2);
  if (arg1) (arg1)->intelPttProperty = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_intelPttProperty_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_intelPttProperty_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_intelPttProperty_get" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPML_INTEL_PTT_PROPERTY *)& ((arg1)->intelPttProperty);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_CAPABILITIES(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_CAPABILITIES")) SWIG_fail;
  result = (TPMU_CAPABILITIES *)calloc(1, sizeof(TPMU_CAPABILITIES));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_CAPABILITIES(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_CAPABILITIES",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_CAPABILITIES" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_CAPABILITIES_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_CAPABILITIES, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_capability_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  TPM2_CAP arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CAPABILITY_DATA_capability_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_capability_set" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CAPABILITY_DATA_capability_set" "', argument " "2"" of type '" "TPM2_CAP""'");
  } 
  arg2 = (TPM2_CAP)(val2);
  if (arg1) (arg1)->capability = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_capability_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_CAP result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CAPABILITY_DATA_capability_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_capability_get" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  result = (TPM2_CAP) ((arg1)->capability);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_data_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  TPMU_CAPABILITIES *arg2 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CAPABILITY_DATA_data_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_data_set" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CAPABILITY_DATA_data_set" "', argument " "2"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg2 = (TPMU_CAPABILITIES *)(argp2);
  if (arg1) (arg1)->data = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_data_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CAPABILITY_DATA_data_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_data_get" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  result = (TPMU_CAPABILITIES *)& ((arg1)->data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CAPABILITY_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CAPABILITY_DATA")) SWIG_fail;
  result = (TPMS_CAPABILITY_DATA *)calloc(1, sizeof(TPMS_CAPABILITY_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CAPABILITY_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CAPABILITY_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CAPABILITY_DATA" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CAPABILITY_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CAPABILITY_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_clock_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  UINT64 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_clock_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_clock_set" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CLOCK_INFO_clock_set" "', argument " "2"" of type '" "UINT64""'");
  } 
  arg2 = (UINT64)(val2);
  if (arg1) (arg1)->clock = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_clock_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT64 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_clock_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_clock_get" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = (UINT64) ((arg1)->clock);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_resetCount_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_resetCount_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_resetCount_set" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CLOCK_INFO_resetCount_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->resetCount = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_resetCount_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_resetCount_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_resetCount_get" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = (UINT32) ((arg1)->resetCount);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_restartCount_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_restartCount_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_restartCount_set" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CLOCK_INFO_restartCount_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->restartCount = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_restartCount_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_restartCount_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_restartCount_get" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = (UINT32) ((arg1)->restartCount);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_safe_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  TPMI_YES_NO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_safe_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_safe_set" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CLOCK_INFO_safe_set" "', argument " "2"" of type '" "TPMI_YES_NO""'");
  } 
  arg2 = (TPMI_YES_NO)(val2);
  if (arg1) (arg1)->safe = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_safe_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_YES_NO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_safe_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_safe_get" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = (TPMI_YES_NO) ((arg1)->safe);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CLOCK_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CLOCK_INFO")) SWIG_fail;
  result = (TPMS_CLOCK_INFO *)calloc(1, sizeof(TPMS_CLOCK_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CLOCK_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CLOCK_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CLOCK_INFO" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CLOCK_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CLOCK_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  UINT64 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_INFO_time_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_time_set" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TIME_INFO_time_set" "', argument " "2"" of type '" "UINT64""'");
  } 
  arg2 = (UINT64)(val2);
  if (arg1) (arg1)->time = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT64 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_INFO_time_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_time_get" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  result = (UINT64) ((arg1)->time);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_clockInfo_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  TPMS_CLOCK_INFO *arg2 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_INFO_clockInfo_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_clockInfo_set" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TIME_INFO_clockInfo_set" "', argument " "2"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg2 = (TPMS_CLOCK_INFO *)(argp2);
  if (arg1) (arg1)->clockInfo = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_clockInfo_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_INFO_clockInfo_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_clockInfo_get" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  result = (TPMS_CLOCK_INFO *)& ((arg1)->clockInfo);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TIME_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TIME_INFO")) SWIG_fail;
  result = (TPMS_TIME_INFO *)calloc(1, sizeof(TPMS_TIME_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_INFO" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TIME_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TIME_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  TPMS_TIME_INFO *arg2 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_ATTEST_INFO_time_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_time_set" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TIME_ATTEST_INFO_time_set" "', argument " "2"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg2 = (TPMS_TIME_INFO *)(argp2);
  if (arg1) (arg1)->time = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_ATTEST_INFO_time_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_time_get" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  result = (TPMS_TIME_INFO *)& ((arg1)->time);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_firmwareVersion_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  UINT64 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_ATTEST_INFO_firmwareVersion_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_firmwareVersion_set" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TIME_ATTEST_INFO_firmwareVersion_set" "', argument " "2"" of type '" "UINT64""'");
  } 
  arg2 = (UINT64)(val2);
  if (arg1) (arg1)->firmwareVersion = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_firmwareVersion_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT64 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_ATTEST_INFO_firmwareVersion_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_firmwareVersion_get" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  result = (UINT64) ((arg1)->firmwareVersion);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TIME_ATTEST_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TIME_ATTEST_INFO")) SWIG_fail;
  result = (TPMS_TIME_ATTEST_INFO *)calloc(1, sizeof(TPMS_TIME_ATTEST_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_ATTEST_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_ATTEST_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_ATTEST_INFO" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TIME_ATTEST_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_name_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CERTIFY_INFO_name_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_name_set" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CERTIFY_INFO_name_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->name = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_name_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CERTIFY_INFO_name_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_name_get" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->name);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_qualifiedName_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CERTIFY_INFO_qualifiedName_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_qualifiedName_set" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CERTIFY_INFO_qualifiedName_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->qualifiedName = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_qualifiedName_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CERTIFY_INFO_qualifiedName_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_qualifiedName_get" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->qualifiedName);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CERTIFY_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CERTIFY_INFO")) SWIG_fail;
  result = (TPMS_CERTIFY_INFO *)calloc(1, sizeof(TPMS_CERTIFY_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CERTIFY_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CERTIFY_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CERTIFY_INFO" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CERTIFY_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CERTIFY_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_pcrSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  TPML_PCR_SELECTION *arg2 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_QUOTE_INFO_pcrSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_pcrSelect_set" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_QUOTE_INFO_pcrSelect_set" "', argument " "2"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg2 = (TPML_PCR_SELECTION *)(argp2);
  if (arg1) (arg1)->pcrSelect = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_pcrSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_QUOTE_INFO_pcrSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_pcrSelect_get" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  result = (TPML_PCR_SELECTION *)& ((arg1)->pcrSelect);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_pcrDigest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_QUOTE_INFO_pcrDigest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_pcrDigest_set" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_QUOTE_INFO_pcrDigest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->pcrDigest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_pcrDigest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_QUOTE_INFO_pcrDigest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_pcrDigest_get" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->pcrDigest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_QUOTE_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_QUOTE_INFO")) SWIG_fail;
  result = (TPMS_QUOTE_INFO *)calloc(1, sizeof(TPMS_QUOTE_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_QUOTE_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_QUOTE_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_QUOTE_INFO" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_QUOTE_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_QUOTE_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_auditCounter_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  UINT64 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_auditCounter_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_auditCounter_set" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_COMMAND_AUDIT_INFO_auditCounter_set" "', argument " "2"" of type '" "UINT64""'");
  } 
  arg2 = (UINT64)(val2);
  if (arg1) (arg1)->auditCounter = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_auditCounter_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT64 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_auditCounter_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_auditCounter_get" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = (UINT64) ((arg1)->auditCounter);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_digestAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  TPM2_ALG_ID arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_digestAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_digestAlg_set" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_COMMAND_AUDIT_INFO_digestAlg_set" "', argument " "2"" of type '" "TPM2_ALG_ID""'");
  } 
  arg2 = (TPM2_ALG_ID)(val2);
  if (arg1) (arg1)->digestAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_digestAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ALG_ID result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_digestAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_digestAlg_get" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = (TPM2_ALG_ID) ((arg1)->digestAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_auditDigest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_auditDigest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_auditDigest_set" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_COMMAND_AUDIT_INFO_auditDigest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->auditDigest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_auditDigest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_auditDigest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_auditDigest_get" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->auditDigest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_commandDigest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_commandDigest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_commandDigest_set" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_COMMAND_AUDIT_INFO_commandDigest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->commandDigest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_commandDigest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_commandDigest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_commandDigest_get" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->commandDigest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_COMMAND_AUDIT_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_COMMAND_AUDIT_INFO")) SWIG_fail;
  result = (TPMS_COMMAND_AUDIT_INFO *)calloc(1, sizeof(TPMS_COMMAND_AUDIT_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_COMMAND_AUDIT_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_COMMAND_AUDIT_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_COMMAND_AUDIT_INFO" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_COMMAND_AUDIT_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_exclusiveSession_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  TPMI_YES_NO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SESSION_AUDIT_INFO_exclusiveSession_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_exclusiveSession_set" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SESSION_AUDIT_INFO_exclusiveSession_set" "', argument " "2"" of type '" "TPMI_YES_NO""'");
  } 
  arg2 = (TPMI_YES_NO)(val2);
  if (arg1) (arg1)->exclusiveSession = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_exclusiveSession_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_YES_NO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SESSION_AUDIT_INFO_exclusiveSession_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_exclusiveSession_get" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  result = (TPMI_YES_NO) ((arg1)->exclusiveSession);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_sessionDigest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SESSION_AUDIT_INFO_sessionDigest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_sessionDigest_set" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SESSION_AUDIT_INFO_sessionDigest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->sessionDigest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_sessionDigest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SESSION_AUDIT_INFO_sessionDigest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_sessionDigest_get" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->sessionDigest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SESSION_AUDIT_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SESSION_AUDIT_INFO")) SWIG_fail;
  result = (TPMS_SESSION_AUDIT_INFO *)calloc(1, sizeof(TPMS_SESSION_AUDIT_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SESSION_AUDIT_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SESSION_AUDIT_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SESSION_AUDIT_INFO" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SESSION_AUDIT_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_objectName_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_INFO_objectName_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_objectName_set" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_INFO_objectName_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->objectName = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_objectName_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_INFO_objectName_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_objectName_get" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->objectName);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_creationHash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_INFO_creationHash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_creationHash_set" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_INFO_creationHash_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->creationHash = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_creationHash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_INFO_creationHash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_creationHash_get" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->creationHash);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CREATION_INFO")) SWIG_fail;
  result = (TPMS_CREATION_INFO *)calloc(1, sizeof(TPMS_CREATION_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_INFO" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CREATION_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CREATION_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_indexName_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_CERTIFY_INFO_indexName_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_indexName_set" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_CERTIFY_INFO_indexName_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->indexName = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_indexName_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_indexName_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_indexName_get" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->indexName);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_offset_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_CERTIFY_INFO_offset_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_offset_set" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_CERTIFY_INFO_offset_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->offset = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_offset_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_offset_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_offset_get" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  result = (UINT16) ((arg1)->offset);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_nvContents_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  TPM2B_MAX_NV_BUFFER *arg2 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_CERTIFY_INFO_nvContents_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_nvContents_set" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_CERTIFY_INFO_nvContents_set" "', argument " "2"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg2 = (TPM2B_MAX_NV_BUFFER *)(argp2);
  if (arg1) (arg1)->nvContents = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_nvContents_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_nvContents_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_nvContents_get" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  result = (TPM2B_MAX_NV_BUFFER *)& ((arg1)->nvContents);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_CERTIFY_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_CERTIFY_INFO")) SWIG_fail;
  result = (TPMS_NV_CERTIFY_INFO *)calloc(1, sizeof(TPMS_NV_CERTIFY_INFO));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_CERTIFY_INFO(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_CERTIFY_INFO",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_CERTIFY_INFO" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_NV_CERTIFY_INFO_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_ATTEST_certify_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_CERTIFY_INFO *arg2 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_certify_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_certify_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_certify_set" "', argument " "2"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg2 = (TPMS_CERTIFY_INFO *)(argp2);
  if (arg1) (arg1)->certify = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_certify_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_certify_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_certify_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_CERTIFY_INFO *)& ((arg1)->certify);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_creation_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_CREATION_INFO *arg2 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_creation_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_creation_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_creation_set" "', argument " "2"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg2 = (TPMS_CREATION_INFO *)(argp2);
  if (arg1) (arg1)->creation = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_creation_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_creation_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_creation_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_CREATION_INFO *)& ((arg1)->creation);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_quote_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_QUOTE_INFO *arg2 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_quote_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_quote_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_quote_set" "', argument " "2"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg2 = (TPMS_QUOTE_INFO *)(argp2);
  if (arg1) (arg1)->quote = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_quote_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_quote_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_quote_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_QUOTE_INFO *)& ((arg1)->quote);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_commandAudit_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_COMMAND_AUDIT_INFO *arg2 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_commandAudit_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_commandAudit_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_commandAudit_set" "', argument " "2"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg2 = (TPMS_COMMAND_AUDIT_INFO *)(argp2);
  if (arg1) (arg1)->commandAudit = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_commandAudit_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_commandAudit_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_commandAudit_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_COMMAND_AUDIT_INFO *)& ((arg1)->commandAudit);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_sessionAudit_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_SESSION_AUDIT_INFO *arg2 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_sessionAudit_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_sessionAudit_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_sessionAudit_set" "', argument " "2"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg2 = (TPMS_SESSION_AUDIT_INFO *)(argp2);
  if (arg1) (arg1)->sessionAudit = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_sessionAudit_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_sessionAudit_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_sessionAudit_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_SESSION_AUDIT_INFO *)& ((arg1)->sessionAudit);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_TIME_ATTEST_INFO *arg2 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_time_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_time_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_time_set" "', argument " "2"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg2 = (TPMS_TIME_ATTEST_INFO *)(argp2);
  if (arg1) (arg1)->time = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_time_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_time_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_TIME_ATTEST_INFO *)& ((arg1)->time);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_nv_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMS_NV_CERTIFY_INFO *arg2 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_nv_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_nv_set" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_nv_set" "', argument " "2"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg2 = (TPMS_NV_CERTIFY_INFO *)(argp2);
  if (arg1) (arg1)->nv = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_nv_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_nv_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_nv_get" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMS_NV_CERTIFY_INFO *)& ((arg1)->nv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_ATTEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ATTEST")) SWIG_fail;
  result = (TPMU_ATTEST *)calloc(1, sizeof(TPMU_ATTEST));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ATTEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ATTEST",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ATTEST" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_ATTEST_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_ATTEST, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ATTEST_magic_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPM2_GENERATED arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_magic_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_magic_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ATTEST_magic_set" "', argument " "2"" of type '" "TPM2_GENERATED""'");
  } 
  arg2 = (TPM2_GENERATED)(val2);
  if (arg1) (arg1)->magic = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_magic_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_GENERATED result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_magic_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_magic_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPM2_GENERATED) ((arg1)->magic);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_type_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPMI_ST_ATTEST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_type_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_type_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ATTEST_type_set" "', argument " "2"" of type '" "TPMI_ST_ATTEST""'");
  } 
  arg2 = (TPMI_ST_ATTEST)(val2);
  if (arg1) (arg1)->type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_type_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ST_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_type_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_type_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPMI_ST_ATTEST) ((arg1)->type);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_qualifiedSigner_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_qualifiedSigner_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_qualifiedSigner_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ATTEST_qualifiedSigner_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->qualifiedSigner = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_qualifiedSigner_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_qualifiedSigner_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_qualifiedSigner_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->qualifiedSigner);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_extraData_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPM2B_DATA *arg2 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_extraData_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_extraData_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ATTEST_extraData_set" "', argument " "2"" of type '" "TPM2B_DATA *""'"); 
  }
  arg2 = (TPM2B_DATA *)(argp2);
  if (arg1) (arg1)->extraData = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_extraData_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_extraData_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_extraData_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPM2B_DATA *)& ((arg1)->extraData);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_clockInfo_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPMS_CLOCK_INFO *arg2 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_clockInfo_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_clockInfo_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ATTEST_clockInfo_set" "', argument " "2"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg2 = (TPMS_CLOCK_INFO *)(argp2);
  if (arg1) (arg1)->clockInfo = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_clockInfo_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_clockInfo_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_clockInfo_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPMS_CLOCK_INFO *)& ((arg1)->clockInfo);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_firmwareVersion_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  UINT64 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_firmwareVersion_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_firmwareVersion_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ATTEST_firmwareVersion_set" "', argument " "2"" of type '" "UINT64""'");
  } 
  arg2 = (UINT64)(val2);
  if (arg1) (arg1)->firmwareVersion = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_firmwareVersion_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT64 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_firmwareVersion_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_firmwareVersion_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (UINT64) ((arg1)->firmwareVersion);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_attested_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPMU_ATTEST *arg2 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_attested_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_attested_set" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ATTEST_attested_set" "', argument " "2"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg2 = (TPMU_ATTEST *)(argp2);
  if (arg1) (arg1)->attested = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_attested_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_attested_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_attested_get" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPMU_ATTEST *)& ((arg1)->attested);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ATTEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ATTEST")) SWIG_fail;
  result = (TPMS_ATTEST *)calloc(1, sizeof(TPMS_ATTEST));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ATTEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ATTEST",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ATTEST" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ATTEST_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ATTEST, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ATTEST_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_size_set" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ATTEST_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ATTEST_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_size_get" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_attestationData_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ATTEST_attestationData_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_attestationData_set" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ATTEST_attestationData_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMS_ATTEST)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMS_ATTEST); ++ii) *(BYTE *)&arg1->attestationData[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""attestationData""' of type '""BYTE [sizeof(TPMS_ATTEST)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_attestationData_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ATTEST_attestationData_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_attestationData_get" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->attestationData);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ATTEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ATTEST")) SWIG_fail;
  result = (TPM2B_ATTEST *)calloc(1, sizeof(TPM2B_ATTEST));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ATTEST(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ATTEST",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ATTEST" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ATTEST_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ATTEST, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_sessionHandle_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  TPMI_SH_AUTH_SESSION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_sessionHandle_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_sessionHandle_set" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_COMMAND_sessionHandle_set" "', argument " "2"" of type '" "TPMI_SH_AUTH_SESSION""'");
  } 
  arg2 = (TPMI_SH_AUTH_SESSION)(val2);
  if (arg1) (arg1)->sessionHandle = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_sessionHandle_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_SH_AUTH_SESSION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_sessionHandle_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_sessionHandle_get" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = (TPMI_SH_AUTH_SESSION) ((arg1)->sessionHandle);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_nonce_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  TPM2B_NONCE *arg2 = (TPM2B_NONCE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_nonce_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_nonce_set" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_COMMAND_nonce_set" "', argument " "2"" of type '" "TPM2B_NONCE *""'"); 
  }
  arg2 = (TPM2B_NONCE *)(argp2);
  if (arg1) (arg1)->nonce = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_nonce_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NONCE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_nonce_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_nonce_get" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = (TPM2B_NONCE *)& ((arg1)->nonce);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_sessionAttributes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  TPMA_SESSION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_sessionAttributes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_sessionAttributes_set" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_COMMAND_sessionAttributes_set" "', argument " "2"" of type '" "TPMA_SESSION""'");
  } 
  arg2 = (TPMA_SESSION)(val2);
  if (arg1) (arg1)->sessionAttributes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_sessionAttributes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_SESSION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_sessionAttributes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_sessionAttributes_get" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = (TPMA_SESSION) ((arg1)->sessionAttributes);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_hmac_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  TPM2B_AUTH *arg2 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_hmac_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_hmac_set" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_COMMAND_hmac_set" "', argument " "2"" of type '" "TPM2B_AUTH *""'"); 
  }
  arg2 = (TPM2B_AUTH *)(argp2);
  if (arg1) (arg1)->hmac = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_hmac_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_hmac_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_hmac_get" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = (TPM2B_AUTH *)& ((arg1)->hmac);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_COMMAND(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AUTH_COMMAND")) SWIG_fail;
  result = (TPMS_AUTH_COMMAND *)calloc(1, sizeof(TPMS_AUTH_COMMAND));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_COMMAND(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_COMMAND",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_COMMAND" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_AUTH_COMMAND_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_AUTH_COMMAND, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_nonce_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  TPM2B_NONCE *arg2 = (TPM2B_NONCE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_RESPONSE_nonce_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_nonce_set" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_RESPONSE_nonce_set" "', argument " "2"" of type '" "TPM2B_NONCE *""'"); 
  }
  arg2 = (TPM2B_NONCE *)(argp2);
  if (arg1) (arg1)->nonce = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_nonce_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NONCE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_nonce_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_nonce_get" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  result = (TPM2B_NONCE *)& ((arg1)->nonce);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_sessionAttributes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  TPMA_SESSION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_RESPONSE_sessionAttributes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_sessionAttributes_set" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_RESPONSE_sessionAttributes_set" "', argument " "2"" of type '" "TPMA_SESSION""'");
  } 
  arg2 = (TPMA_SESSION)(val2);
  if (arg1) (arg1)->sessionAttributes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_sessionAttributes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_SESSION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_sessionAttributes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_sessionAttributes_get" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  result = (TPMA_SESSION) ((arg1)->sessionAttributes);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_hmac_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  TPM2B_AUTH *arg2 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_RESPONSE_hmac_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_hmac_set" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_RESPONSE_hmac_set" "', argument " "2"" of type '" "TPM2B_AUTH *""'"); 
  }
  arg2 = (TPM2B_AUTH *)(argp2);
  if (arg1) (arg1)->hmac = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_hmac_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_hmac_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_hmac_get" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  result = (TPM2B_AUTH *)& ((arg1)->hmac);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_RESPONSE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AUTH_RESPONSE")) SWIG_fail;
  result = (TPMS_AUTH_RESPONSE *)calloc(1, sizeof(TPMS_AUTH_RESPONSE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_RESPONSE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_RESPONSE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_RESPONSE" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_AUTH_RESPONSE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_AUTH_RESPONSE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_aes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  TPMI_AES_KEY_BITS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_aes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_aes_set" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_aes_set" "', argument " "2"" of type '" "TPMI_AES_KEY_BITS""'");
  } 
  arg2 = (TPMI_AES_KEY_BITS)(val2);
  if (arg1) (arg1)->aes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_aes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_AES_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_aes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_aes_get" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPMI_AES_KEY_BITS) ((arg1)->aes);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_sm4_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  TPMI_SM4_KEY_BITS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_sm4_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_sm4_set" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_sm4_set" "', argument " "2"" of type '" "TPMI_SM4_KEY_BITS""'");
  } 
  arg2 = (TPMI_SM4_KEY_BITS)(val2);
  if (arg1) (arg1)->sm4 = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_sm4_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_SM4_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_sm4_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_sm4_get" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPMI_SM4_KEY_BITS) ((arg1)->sm4);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_camellia_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  TPMI_CAMELLIA_KEY_BITS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_camellia_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_camellia_set" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_camellia_set" "', argument " "2"" of type '" "TPMI_CAMELLIA_KEY_BITS""'");
  } 
  arg2 = (TPMI_CAMELLIA_KEY_BITS)(val2);
  if (arg1) (arg1)->camellia = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_camellia_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_CAMELLIA_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_camellia_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_camellia_get" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPMI_CAMELLIA_KEY_BITS) ((arg1)->camellia);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_sym_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  TPM2_KEY_BITS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_sym_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_sym_set" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_sym_set" "', argument " "2"" of type '" "TPM2_KEY_BITS""'");
  } 
  arg2 = (TPM2_KEY_BITS)(val2);
  if (arg1) (arg1)->sym = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_sym_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_sym_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_sym_get" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPM2_KEY_BITS) ((arg1)->sym);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_exclusiveOr_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_exclusiveOr_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_exclusiveOr_set" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_exclusiveOr_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->exclusiveOr = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_exclusiveOr_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_exclusiveOr_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_exclusiveOr_get" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->exclusiveOr);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SYM_KEY_BITS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SYM_KEY_BITS")) SWIG_fail;
  result = (TPMU_SYM_KEY_BITS *)calloc(1, sizeof(TPMU_SYM_KEY_BITS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_KEY_BITS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_KEY_BITS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_KEY_BITS" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SYM_KEY_BITS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SYM_KEY_BITS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_aes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  TPMI_ALG_SYM_MODE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_aes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_aes_set" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_MODE_aes_set" "', argument " "2"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg2 = (TPMI_ALG_SYM_MODE)(val2);
  if (arg1) (arg1)->aes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_aes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SYM_MODE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_aes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_aes_get" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = (TPMI_ALG_SYM_MODE) ((arg1)->aes);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_sm4_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  TPMI_ALG_SYM_MODE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_sm4_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_sm4_set" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_MODE_sm4_set" "', argument " "2"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg2 = (TPMI_ALG_SYM_MODE)(val2);
  if (arg1) (arg1)->sm4 = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_sm4_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SYM_MODE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_sm4_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_sm4_get" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = (TPMI_ALG_SYM_MODE) ((arg1)->sm4);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_camellia_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  TPMI_ALG_SYM_MODE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_camellia_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_camellia_set" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_MODE_camellia_set" "', argument " "2"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg2 = (TPMI_ALG_SYM_MODE)(val2);
  if (arg1) (arg1)->camellia = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_camellia_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SYM_MODE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_camellia_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_camellia_get" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = (TPMI_ALG_SYM_MODE) ((arg1)->camellia);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_sym_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  TPMI_ALG_SYM_MODE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_sym_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_sym_set" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_MODE_sym_set" "', argument " "2"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg2 = (TPMI_ALG_SYM_MODE)(val2);
  if (arg1) (arg1)->sym = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_sym_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SYM_MODE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_sym_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_sym_get" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = (TPMI_ALG_SYM_MODE) ((arg1)->sym);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SYM_MODE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SYM_MODE")) SWIG_fail;
  result = (TPMU_SYM_MODE *)calloc(1, sizeof(TPMU_SYM_MODE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_MODE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_MODE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_MODE" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SYM_MODE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SYM_MODE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_algorithm_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  TPMI_ALG_SYM arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_algorithm_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_algorithm_set" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SYM_DEF_algorithm_set" "', argument " "2"" of type '" "TPMI_ALG_SYM""'");
  } 
  arg2 = (TPMI_ALG_SYM)(val2);
  if (arg1) (arg1)->algorithm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_algorithm_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SYM result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_algorithm_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_algorithm_get" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  result = (TPMI_ALG_SYM) ((arg1)->algorithm);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_keyBits_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  TPMU_SYM_KEY_BITS *arg2 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_keyBits_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_keyBits_set" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_keyBits_set" "', argument " "2"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg2 = (TPMU_SYM_KEY_BITS *)(argp2);
  if (arg1) (arg1)->keyBits = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_keyBits_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_keyBits_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_keyBits_get" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  result = (TPMU_SYM_KEY_BITS *)& ((arg1)->keyBits);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_mode_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  TPMU_SYM_MODE *arg2 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_mode_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_mode_set" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_mode_set" "', argument " "2"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg2 = (TPMU_SYM_MODE *)(argp2);
  if (arg1) (arg1)->mode = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_mode_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_mode_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_mode_get" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  result = (TPMU_SYM_MODE *)& ((arg1)->mode);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SYM_DEF")) SWIG_fail;
  result = (TPMT_SYM_DEF *)calloc(1, sizeof(TPMT_SYM_DEF));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SYM_DEF_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SYM_DEF, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_algorithm_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  TPMI_ALG_SYM_OBJECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_OBJECT_algorithm_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_algorithm_set" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SYM_DEF_OBJECT_algorithm_set" "', argument " "2"" of type '" "TPMI_ALG_SYM_OBJECT""'");
  } 
  arg2 = (TPMI_ALG_SYM_OBJECT)(val2);
  if (arg1) (arg1)->algorithm = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_algorithm_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SYM_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_algorithm_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_algorithm_get" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  result = (TPMI_ALG_SYM_OBJECT) ((arg1)->algorithm);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_keyBits_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  TPMU_SYM_KEY_BITS *arg2 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_OBJECT_keyBits_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_keyBits_set" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_OBJECT_keyBits_set" "', argument " "2"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg2 = (TPMU_SYM_KEY_BITS *)(argp2);
  if (arg1) (arg1)->keyBits = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_keyBits_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_keyBits_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_keyBits_get" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  result = (TPMU_SYM_KEY_BITS *)& ((arg1)->keyBits);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_mode_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  TPMU_SYM_MODE *arg2 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_OBJECT_mode_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_mode_set" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_OBJECT_mode_set" "', argument " "2"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg2 = (TPMU_SYM_MODE *)(argp2);
  if (arg1) (arg1)->mode = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_mode_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_mode_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_mode_get" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  result = (TPMU_SYM_MODE *)& ((arg1)->mode);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_OBJECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SYM_DEF_OBJECT")) SWIG_fail;
  result = (TPMT_SYM_DEF_OBJECT *)calloc(1, sizeof(TPMT_SYM_DEF_OBJECT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_OBJECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_OBJECT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_OBJECT" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SYM_DEF_OBJECT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SYM_KEY_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_size_set" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SYM_KEY_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SYM_KEY_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_size_get" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SYM_KEY_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_buffer_set" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SYM_KEY_buffer_set" "', argument " "2"" of type '" "BYTE [32]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)32; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [32]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SYM_KEY_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_buffer_get" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SYM_KEY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SYM_KEY")) SWIG_fail;
  result = (TPM2B_SYM_KEY *)calloc(1, sizeof(TPM2B_SYM_KEY));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SYM_KEY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SYM_KEY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SYM_KEY" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SYM_KEY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SYM_KEY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_sym_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg2 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SYMCIPHER_PARMS_sym_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_sym_set" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SYMCIPHER_PARMS_sym_set" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg2 = (TPMT_SYM_DEF_OBJECT *)(argp2);
  if (arg1) (arg1)->sym = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_sym_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SYMCIPHER_PARMS_sym_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_sym_get" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT *)& ((arg1)->sym);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SYMCIPHER_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SYMCIPHER_PARMS")) SWIG_fail;
  result = (TPMS_SYMCIPHER_PARMS *)calloc(1, sizeof(TPMS_SYMCIPHER_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SYMCIPHER_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SYMCIPHER_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SYMCIPHER_PARMS" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SYMCIPHER_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_DATA_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_size_set" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_DATA_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_DATA_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_size_get" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_DATA_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_buffer_set" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_DATA_buffer_set" "', argument " "2"" of type '" "BYTE [256]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)256; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [256]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_DATA_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_buffer_get" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_DATA")) SWIG_fail;
  result = (TPM2B_SENSITIVE_DATA *)calloc(1, sizeof(TPM2B_SENSITIVE_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_DATA" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SENSITIVE_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SENSITIVE_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_userAuth_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  TPM2B_AUTH *arg2 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SENSITIVE_CREATE_userAuth_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_userAuth_set" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SENSITIVE_CREATE_userAuth_set" "', argument " "2"" of type '" "TPM2B_AUTH *""'"); 
  }
  arg2 = (TPM2B_AUTH *)(argp2);
  if (arg1) (arg1)->userAuth = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_userAuth_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SENSITIVE_CREATE_userAuth_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_userAuth_get" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  result = (TPM2B_AUTH *)& ((arg1)->userAuth);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_data_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  TPM2B_SENSITIVE_DATA *arg2 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SENSITIVE_CREATE_data_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_data_set" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SENSITIVE_CREATE_data_set" "', argument " "2"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg2 = (TPM2B_SENSITIVE_DATA *)(argp2);
  if (arg1) (arg1)->data = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_data_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SENSITIVE_CREATE_data_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_data_get" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  result = (TPM2B_SENSITIVE_DATA *)& ((arg1)->data);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SENSITIVE_CREATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SENSITIVE_CREATE")) SWIG_fail;
  result = (TPMS_SENSITIVE_CREATE *)calloc(1, sizeof(TPMS_SENSITIVE_CREATE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SENSITIVE_CREATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SENSITIVE_CREATE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SENSITIVE_CREATE" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SENSITIVE_CREATE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SENSITIVE_CREATE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_CREATE_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_size_set" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_CREATE_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_CREATE_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_size_get" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_sensitive_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPMS_SENSITIVE_CREATE *arg2 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_CREATE_sensitive_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_sensitive_set" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_CREATE_sensitive_set" "', argument " "2"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg2 = (TPMS_SENSITIVE_CREATE *)(argp2);
  if (arg1) (arg1)->sensitive = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_sensitive_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_CREATE_sensitive_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_sensitive_get" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  result = (TPMS_SENSITIVE_CREATE *)& ((arg1)->sensitive);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_CREATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_CREATE")) SWIG_fail;
  result = (TPM2B_SENSITIVE_CREATE *)calloc(1, sizeof(TPM2B_SENSITIVE_CREATE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_CREATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_CREATE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_CREATE" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SENSITIVE_CREATE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_hashAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_HASH_hashAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_hashAlg_set" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_HASH_hashAlg_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hashAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_hashAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_HASH_hashAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_hashAlg_get" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hashAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_HASH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_HASH")) SWIG_fail;
  result = (TPMS_SCHEME_HASH *)calloc(1, sizeof(TPMS_SCHEME_HASH));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_HASH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_HASH",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_HASH" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SCHEME_HASH_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SCHEME_HASH, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_hashAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_ECDAA_hashAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_hashAlg_set" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_ECDAA_hashAlg_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hashAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_hashAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_ECDAA_hashAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_hashAlg_get" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hashAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_ECDAA_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_count_set" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_ECDAA_count_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_ECDAA_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_count_get" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  result = (UINT16) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_ECDAA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_ECDAA")) SWIG_fail;
  result = (TPMS_SCHEME_ECDAA *)calloc(1, sizeof(TPMS_SCHEME_ECDAA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_ECDAA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_ECDAA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_ECDAA" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SCHEME_ECDAA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SCHEME_ECDAA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_hashAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_XOR_hashAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_hashAlg_set" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_XOR_hashAlg_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hashAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_hashAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_XOR_hashAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_hashAlg_get" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hashAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_kdf_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  TPMI_ALG_KDF arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_XOR_kdf_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_kdf_set" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_XOR_kdf_set" "', argument " "2"" of type '" "TPMI_ALG_KDF""'");
  } 
  arg2 = (TPMI_ALG_KDF)(val2);
  if (arg1) (arg1)->kdf = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_kdf_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_KDF result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_XOR_kdf_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_kdf_get" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  result = (TPMI_ALG_KDF) ((arg1)->kdf);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_XOR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_XOR")) SWIG_fail;
  result = (TPMS_SCHEME_XOR *)calloc(1, sizeof(TPMS_SCHEME_XOR));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_XOR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_XOR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_XOR" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SCHEME_XOR_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SCHEME_XOR, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_hmac_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  TPMS_SCHEME_HMAC *arg2 = (TPMS_SCHEME_HMAC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SCHEME_KEYEDHASH_hmac_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_hmac_set" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SCHEME_KEYEDHASH_hmac_set" "', argument " "2"" of type '" "TPMS_SCHEME_HMAC *""'"); 
  }
  arg2 = (TPMS_SCHEME_HMAC *)(argp2);
  if (arg1) (arg1)->hmac = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_hmac_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HMAC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SCHEME_KEYEDHASH_hmac_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_hmac_get" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  result = (TPMS_SCHEME_HMAC *)& ((arg1)->hmac);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_exclusiveOr_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  TPMS_SCHEME_XOR *arg2 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SCHEME_KEYEDHASH_exclusiveOr_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_exclusiveOr_set" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SCHEME_KEYEDHASH_exclusiveOr_set" "', argument " "2"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg2 = (TPMS_SCHEME_XOR *)(argp2);
  if (arg1) (arg1)->exclusiveOr = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_exclusiveOr_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SCHEME_KEYEDHASH_exclusiveOr_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_exclusiveOr_get" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  result = (TPMS_SCHEME_XOR *)& ((arg1)->exclusiveOr);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SCHEME_KEYEDHASH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SCHEME_KEYEDHASH")) SWIG_fail;
  result = (TPMU_SCHEME_KEYEDHASH *)calloc(1, sizeof(TPMU_SCHEME_KEYEDHASH));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SCHEME_KEYEDHASH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SCHEME_KEYEDHASH",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SCHEME_KEYEDHASH" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SCHEME_KEYEDHASH_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  TPMI_ALG_KEYEDHASH_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KEYEDHASH_SCHEME_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_scheme_set" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_KEYEDHASH_SCHEME_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_KEYEDHASH_SCHEME""'");
  } 
  arg2 = (TPMI_ALG_KEYEDHASH_SCHEME)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_KEYEDHASH_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KEYEDHASH_SCHEME_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_scheme_get" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  result = (TPMI_ALG_KEYEDHASH_SCHEME) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  TPMU_SCHEME_KEYEDHASH *arg2 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KEYEDHASH_SCHEME_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_details_set" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_KEYEDHASH_SCHEME_details_set" "', argument " "2"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg2 = (TPMU_SCHEME_KEYEDHASH *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KEYEDHASH_SCHEME_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_details_get" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  result = (TPMU_SCHEME_KEYEDHASH *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_KEYEDHASH_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_KEYEDHASH_SCHEME")) SWIG_fail;
  result = (TPMT_KEYEDHASH_SCHEME *)calloc(1, sizeof(TPMT_KEYEDHASH_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KEYEDHASH_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KEYEDHASH_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KEYEDHASH_SCHEME" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_KEYEDHASH_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_rsassa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_RSASSA *arg2 = (TPMS_SIG_SCHEME_RSASSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_rsassa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_rsassa_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_rsassa_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_RSASSA *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_RSASSA *)(argp2);
  if (arg1) (arg1)->rsassa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_rsassa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_RSASSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_rsassa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_rsassa_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_RSASSA *)& ((arg1)->rsassa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_rsapss_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_RSAPSS *arg2 = (TPMS_SIG_SCHEME_RSAPSS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_rsapss_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_rsapss_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_rsapss_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_RSAPSS *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_RSAPSS *)(argp2);
  if (arg1) (arg1)->rsapss = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_rsapss_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_RSAPSS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_rsapss_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_rsapss_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_RSAPSS *)& ((arg1)->rsapss);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ecdsa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_ECDSA *arg2 = (TPMS_SIG_SCHEME_ECDSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_ecdsa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ecdsa_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_ecdsa_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_ECDSA *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_ECDSA *)(argp2);
  if (arg1) (arg1)->ecdsa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ecdsa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_ECDSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_ecdsa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ecdsa_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_ECDSA *)& ((arg1)->ecdsa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ecdaa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_ECDAA *arg2 = (TPMS_SIG_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_ecdaa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ecdaa_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_ecdaa_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_ECDAA *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_ECDAA *)(argp2);
  if (arg1) (arg1)->ecdaa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ecdaa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_ecdaa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ecdaa_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_ECDAA *)& ((arg1)->ecdaa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_sm2_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_SM2 *arg2 = (TPMS_SIG_SCHEME_SM2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_sm2_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_sm2_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_sm2_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_SM2 *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_SM2 *)(argp2);
  if (arg1) (arg1)->sm2 = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_sm2_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_SM2 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_sm2_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_sm2_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_SM2 *)& ((arg1)->sm2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ecschnorr_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_ECSCHNORR *arg2 = (TPMS_SIG_SCHEME_ECSCHNORR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_ecschnorr_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ecschnorr_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_ecschnorr_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_ECSCHNORR *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_ECSCHNORR *)(argp2);
  if (arg1) (arg1)->ecschnorr = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ecschnorr_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_ECSCHNORR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_ecschnorr_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ecschnorr_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_ECSCHNORR *)& ((arg1)->ecschnorr);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_hmac_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SCHEME_HMAC *arg2 = (TPMS_SCHEME_HMAC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_hmac_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_hmac_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_hmac_set" "', argument " "2"" of type '" "TPMS_SCHEME_HMAC *""'"); 
  }
  arg2 = (TPMS_SCHEME_HMAC *)(argp2);
  if (arg1) (arg1)->hmac = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_hmac_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HMAC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_hmac_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_hmac_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SCHEME_HMAC *)& ((arg1)->hmac);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_any_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMS_SCHEME_HASH *arg2 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_any_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_any_set" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_any_set" "', argument " "2"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg2 = (TPMS_SCHEME_HASH *)(argp2);
  if (arg1) (arg1)->any = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_any_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_any_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_any_get" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMS_SCHEME_HASH *)& ((arg1)->any);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SIG_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SIG_SCHEME")) SWIG_fail;
  result = (TPMU_SIG_SCHEME *)calloc(1, sizeof(TPMU_SIG_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIG_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIG_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIG_SCHEME" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SIG_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SIG_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  TPMI_ALG_SIG_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIG_SCHEME_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_scheme_set" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SIG_SCHEME_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_SIG_SCHEME""'");
  } 
  arg2 = (TPMI_ALG_SIG_SCHEME)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SIG_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIG_SCHEME_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_scheme_get" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  result = (TPMI_ALG_SIG_SCHEME) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  TPMU_SIG_SCHEME *arg2 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIG_SCHEME_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_details_set" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SIG_SCHEME_details_set" "', argument " "2"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg2 = (TPMU_SIG_SCHEME *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIG_SCHEME_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_details_get" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  result = (TPMU_SIG_SCHEME *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SIG_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SIG_SCHEME")) SWIG_fail;
  result = (TPMT_SIG_SCHEME *)calloc(1, sizeof(TPMT_SIG_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIG_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIG_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIG_SCHEME" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SIG_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SIG_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_mgf1_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  TPMS_SCHEME_MGF1 *arg2 = (TPMS_SCHEME_MGF1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_mgf1_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_mgf1_set" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_KDF_SCHEME_mgf1_set" "', argument " "2"" of type '" "TPMS_SCHEME_MGF1 *""'"); 
  }
  arg2 = (TPMS_SCHEME_MGF1 *)(argp2);
  if (arg1) (arg1)->mgf1 = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_mgf1_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_MGF1 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_mgf1_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_mgf1_get" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = (TPMS_SCHEME_MGF1 *)& ((arg1)->mgf1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_kdf1_sp800_56a_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  TPMS_SCHEME_KDF1_SP800_56A *arg2 = (TPMS_SCHEME_KDF1_SP800_56A *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_kdf1_sp800_56a_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_kdf1_sp800_56a_set" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_KDF_SCHEME_kdf1_sp800_56a_set" "', argument " "2"" of type '" "TPMS_SCHEME_KDF1_SP800_56A *""'"); 
  }
  arg2 = (TPMS_SCHEME_KDF1_SP800_56A *)(argp2);
  if (arg1) (arg1)->kdf1_sp800_56a = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_kdf1_sp800_56a_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_KDF1_SP800_56A *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_kdf1_sp800_56a_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_kdf1_sp800_56a_get" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = (TPMS_SCHEME_KDF1_SP800_56A *)& ((arg1)->kdf1_sp800_56a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_kdf2_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  TPMS_SCHEME_KDF2 *arg2 = (TPMS_SCHEME_KDF2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_kdf2_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_kdf2_set" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_KDF_SCHEME_kdf2_set" "', argument " "2"" of type '" "TPMS_SCHEME_KDF2 *""'"); 
  }
  arg2 = (TPMS_SCHEME_KDF2 *)(argp2);
  if (arg1) (arg1)->kdf2 = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_kdf2_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_KDF2 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_kdf2_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_kdf2_get" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = (TPMS_SCHEME_KDF2 *)& ((arg1)->kdf2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_kdf1_sp800_108_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  TPMS_SCHEME_KDF1_SP800_108 *arg2 = (TPMS_SCHEME_KDF1_SP800_108 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_kdf1_sp800_108_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_kdf1_sp800_108_set" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_KDF_SCHEME_kdf1_sp800_108_set" "', argument " "2"" of type '" "TPMS_SCHEME_KDF1_SP800_108 *""'"); 
  }
  arg2 = (TPMS_SCHEME_KDF1_SP800_108 *)(argp2);
  if (arg1) (arg1)->kdf1_sp800_108 = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_kdf1_sp800_108_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_KDF1_SP800_108 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_kdf1_sp800_108_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_kdf1_sp800_108_get" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = (TPMS_SCHEME_KDF1_SP800_108 *)& ((arg1)->kdf1_sp800_108);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_KDF_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_KDF_SCHEME")) SWIG_fail;
  result = (TPMU_KDF_SCHEME *)calloc(1, sizeof(TPMU_KDF_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_KDF_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_KDF_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_KDF_SCHEME" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_KDF_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_KDF_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  TPMI_ALG_KDF arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KDF_SCHEME_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_scheme_set" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_KDF_SCHEME_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_KDF""'");
  } 
  arg2 = (TPMI_ALG_KDF)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_KDF result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KDF_SCHEME_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_scheme_get" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  result = (TPMI_ALG_KDF) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  TPMU_KDF_SCHEME *arg2 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KDF_SCHEME_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_details_set" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_KDF_SCHEME_details_set" "', argument " "2"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg2 = (TPMU_KDF_SCHEME *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KDF_SCHEME_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_details_get" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  result = (TPMU_KDF_SCHEME *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_KDF_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_KDF_SCHEME")) SWIG_fail;
  result = (TPMT_KDF_SCHEME *)calloc(1, sizeof(TPMT_KDF_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KDF_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KDF_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KDF_SCHEME" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_KDF_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_KDF_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecdh_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_KEY_SCHEME_ECDH *arg2 = (TPMS_KEY_SCHEME_ECDH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_ecdh_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecdh_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_ecdh_set" "', argument " "2"" of type '" "TPMS_KEY_SCHEME_ECDH *""'"); 
  }
  arg2 = (TPMS_KEY_SCHEME_ECDH *)(argp2);
  if (arg1) (arg1)->ecdh = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecdh_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEY_SCHEME_ECDH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ecdh_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecdh_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_KEY_SCHEME_ECDH *)& ((arg1)->ecdh);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecmqv_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_KEY_SCHEME_ECMQV *arg2 = (TPMS_KEY_SCHEME_ECMQV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_ecmqv_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecmqv_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_ecmqv_set" "', argument " "2"" of type '" "TPMS_KEY_SCHEME_ECMQV *""'"); 
  }
  arg2 = (TPMS_KEY_SCHEME_ECMQV *)(argp2);
  if (arg1) (arg1)->ecmqv = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecmqv_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEY_SCHEME_ECMQV *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ecmqv_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecmqv_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_KEY_SCHEME_ECMQV *)& ((arg1)->ecmqv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_rsassa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_RSASSA *arg2 = (TPMS_SIG_SCHEME_RSASSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_rsassa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_rsassa_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_rsassa_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_RSASSA *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_RSASSA *)(argp2);
  if (arg1) (arg1)->rsassa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_rsassa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_RSASSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_rsassa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_rsassa_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_RSASSA *)& ((arg1)->rsassa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_rsapss_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_RSAPSS *arg2 = (TPMS_SIG_SCHEME_RSAPSS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_rsapss_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_rsapss_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_rsapss_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_RSAPSS *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_RSAPSS *)(argp2);
  if (arg1) (arg1)->rsapss = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_rsapss_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_RSAPSS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_rsapss_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_rsapss_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_RSAPSS *)& ((arg1)->rsapss);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecdsa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_ECDSA *arg2 = (TPMS_SIG_SCHEME_ECDSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_ecdsa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecdsa_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_ecdsa_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_ECDSA *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_ECDSA *)(argp2);
  if (arg1) (arg1)->ecdsa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecdsa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_ECDSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ecdsa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecdsa_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_ECDSA *)& ((arg1)->ecdsa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecdaa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_ECDAA *arg2 = (TPMS_SIG_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_ecdaa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecdaa_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_ecdaa_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_ECDAA *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_ECDAA *)(argp2);
  if (arg1) (arg1)->ecdaa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecdaa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ecdaa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecdaa_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_ECDAA *)& ((arg1)->ecdaa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_sm2_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_SM2 *arg2 = (TPMS_SIG_SCHEME_SM2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_sm2_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_sm2_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_sm2_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_SM2 *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_SM2 *)(argp2);
  if (arg1) (arg1)->sm2 = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_sm2_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_SM2 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_sm2_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_sm2_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_SM2 *)& ((arg1)->sm2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecschnorr_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SIG_SCHEME_ECSCHNORR *arg2 = (TPMS_SIG_SCHEME_ECSCHNORR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_ecschnorr_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecschnorr_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_ecschnorr_set" "', argument " "2"" of type '" "TPMS_SIG_SCHEME_ECSCHNORR *""'"); 
  }
  arg2 = (TPMS_SIG_SCHEME_ECSCHNORR *)(argp2);
  if (arg1) (arg1)->ecschnorr = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ecschnorr_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIG_SCHEME_ECSCHNORR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ecschnorr_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ecschnorr_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SIG_SCHEME_ECSCHNORR *)& ((arg1)->ecschnorr);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_rsaes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_ENC_SCHEME_RSAES *arg2 = (TPMS_ENC_SCHEME_RSAES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_rsaes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_rsaes_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_rsaes_set" "', argument " "2"" of type '" "TPMS_ENC_SCHEME_RSAES *""'"); 
  }
  arg2 = (TPMS_ENC_SCHEME_RSAES *)(argp2);
  if (arg1) (arg1)->rsaes = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_rsaes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ENC_SCHEME_RSAES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_rsaes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_rsaes_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_ENC_SCHEME_RSAES *)& ((arg1)->rsaes);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_oaep_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_ENC_SCHEME_OAEP *arg2 = (TPMS_ENC_SCHEME_OAEP *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_oaep_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_oaep_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_oaep_set" "', argument " "2"" of type '" "TPMS_ENC_SCHEME_OAEP *""'"); 
  }
  arg2 = (TPMS_ENC_SCHEME_OAEP *)(argp2);
  if (arg1) (arg1)->oaep = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_oaep_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ENC_SCHEME_OAEP *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_oaep_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_oaep_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_ENC_SCHEME_OAEP *)& ((arg1)->oaep);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_anySig_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMS_SCHEME_HASH *arg2 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_anySig_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_anySig_set" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_anySig_set" "', argument " "2"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg2 = (TPMS_SCHEME_HASH *)(argp2);
  if (arg1) (arg1)->anySig = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_anySig_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_anySig_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_anySig_get" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMS_SCHEME_HASH *)& ((arg1)->anySig);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_ASYM_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ASYM_SCHEME")) SWIG_fail;
  result = (TPMU_ASYM_SCHEME *)calloc(1, sizeof(TPMU_ASYM_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ASYM_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ASYM_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ASYM_SCHEME" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_ASYM_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_ASYM_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  TPMI_ALG_ASYM_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ASYM_SCHEME_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_scheme_set" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_ASYM_SCHEME_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_ASYM_SCHEME""'");
  } 
  arg2 = (TPMI_ALG_ASYM_SCHEME)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_ASYM_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ASYM_SCHEME_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_scheme_get" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  result = (TPMI_ALG_ASYM_SCHEME) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  TPMU_ASYM_SCHEME *arg2 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ASYM_SCHEME_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_details_set" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_ASYM_SCHEME_details_set" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMU_ASYM_SCHEME *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ASYM_SCHEME_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_details_get" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  result = (TPMU_ASYM_SCHEME *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_ASYM_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_ASYM_SCHEME")) SWIG_fail;
  result = (TPMT_ASYM_SCHEME *)calloc(1, sizeof(TPMT_ASYM_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ASYM_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ASYM_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ASYM_SCHEME" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_ASYM_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_ASYM_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  TPMI_ALG_RSA_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_SCHEME_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_scheme_set" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_RSA_SCHEME_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_RSA_SCHEME""'");
  } 
  arg2 = (TPMI_ALG_RSA_SCHEME)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_RSA_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_SCHEME_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_scheme_get" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  result = (TPMI_ALG_RSA_SCHEME) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  TPMU_ASYM_SCHEME *arg2 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_SCHEME_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_details_set" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_RSA_SCHEME_details_set" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMU_ASYM_SCHEME *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_SCHEME_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_details_get" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  result = (TPMU_ASYM_SCHEME *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_RSA_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_RSA_SCHEME")) SWIG_fail;
  result = (TPMT_RSA_SCHEME *)calloc(1, sizeof(TPMT_RSA_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_SCHEME" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_RSA_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_RSA_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  TPMI_ALG_RSA_DECRYPT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_DECRYPT_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_scheme_set" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_RSA_DECRYPT_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_RSA_DECRYPT""'");
  } 
  arg2 = (TPMI_ALG_RSA_DECRYPT)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_RSA_DECRYPT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_DECRYPT_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_scheme_get" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  result = (TPMI_ALG_RSA_DECRYPT) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  TPMU_ASYM_SCHEME *arg2 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_DECRYPT_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_details_set" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_RSA_DECRYPT_details_set" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMU_ASYM_SCHEME *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_DECRYPT_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_details_get" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  result = (TPMU_ASYM_SCHEME *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_RSA_DECRYPT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_RSA_DECRYPT")) SWIG_fail;
  result = (TPMT_RSA_DECRYPT *)calloc(1, sizeof(TPMT_RSA_DECRYPT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_DECRYPT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_DECRYPT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_DECRYPT" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_RSA_DECRYPT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_RSA_DECRYPT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_KEY_RSA_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_size_set" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PUBLIC_KEY_RSA_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_KEY_RSA_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_size_get" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_KEY_RSA_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_buffer_set" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PUBLIC_KEY_RSA_buffer_set" "', argument " "2"" of type '" "BYTE [512]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)512; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [512]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_KEY_RSA_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_buffer_get" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_KEY_RSA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PUBLIC_KEY_RSA")) SWIG_fail;
  result = (TPM2B_PUBLIC_KEY_RSA *)calloc(1, sizeof(TPM2B_PUBLIC_KEY_RSA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_KEY_RSA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_KEY_RSA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_KEY_RSA" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PUBLIC_KEY_RSA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_KEY_RSA_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_size_set" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_KEY_RSA_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_KEY_RSA_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_size_get" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_KEY_RSA_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_buffer_set" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_KEY_RSA_buffer_set" "', argument " "2"" of type '" "BYTE [512/2]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)512/2; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [512/2]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_KEY_RSA_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_buffer_get" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_KEY_RSA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_KEY_RSA")) SWIG_fail;
  result = (TPM2B_PRIVATE_KEY_RSA *)calloc(1, sizeof(TPM2B_PRIVATE_KEY_RSA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_KEY_RSA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_KEY_RSA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_KEY_RSA" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PRIVATE_KEY_RSA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_PARAMETER_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_size_set" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ECC_PARAMETER_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_PARAMETER_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_size_get" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_PARAMETER_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_buffer_set" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ECC_PARAMETER_buffer_set" "', argument " "2"" of type '" "BYTE [128]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)128; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [128]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_PARAMETER_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_buffer_get" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_PARAMETER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ECC_PARAMETER")) SWIG_fail;
  result = (TPM2B_ECC_PARAMETER *)calloc(1, sizeof(TPM2B_ECC_PARAMETER));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_PARAMETER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_PARAMETER",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_PARAMETER" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ECC_PARAMETER_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ECC_PARAMETER, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_x_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_POINT_x_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_x_set" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_POINT_x_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->x = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_x_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_POINT_x_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_x_get" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->x);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_y_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_POINT_y_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_y_set" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_POINT_y_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->y = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_y_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_POINT_y_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_y_get" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->y);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ECC_POINT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ECC_POINT")) SWIG_fail;
  result = (TPMS_ECC_POINT *)calloc(1, sizeof(TPMS_ECC_POINT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_POINT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_POINT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_POINT" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ECC_POINT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ECC_POINT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_POINT_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_size_set" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ECC_POINT_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_POINT_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_size_get" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_point_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  TPMS_ECC_POINT *arg2 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_POINT_point_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_point_set" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ECC_POINT_point_set" "', argument " "2"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg2 = (TPMS_ECC_POINT *)(argp2);
  if (arg1) (arg1)->point = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_point_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_POINT_point_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_point_get" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  result = (TPMS_ECC_POINT *)& ((arg1)->point);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_POINT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ECC_POINT")) SWIG_fail;
  result = (TPM2B_ECC_POINT *)calloc(1, sizeof(TPM2B_ECC_POINT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_POINT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_POINT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_POINT" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ECC_POINT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ECC_POINT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  TPMI_ALG_ECC_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ECC_SCHEME_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_scheme_set" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_ECC_SCHEME_scheme_set" "', argument " "2"" of type '" "TPMI_ALG_ECC_SCHEME""'");
  } 
  arg2 = (TPMI_ALG_ECC_SCHEME)(val2);
  if (arg1) (arg1)->scheme = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_ECC_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ECC_SCHEME_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_scheme_get" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  result = (TPMI_ALG_ECC_SCHEME) ((arg1)->scheme);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_details_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  TPMU_ASYM_SCHEME *arg2 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ECC_SCHEME_details_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_details_set" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_ECC_SCHEME_details_set" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMU_ASYM_SCHEME *)(argp2);
  if (arg1) (arg1)->details = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_details_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ECC_SCHEME_details_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_details_get" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  result = (TPMU_ASYM_SCHEME *)& ((arg1)->details);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_ECC_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_ECC_SCHEME")) SWIG_fail;
  result = (TPMT_ECC_SCHEME *)calloc(1, sizeof(TPMT_ECC_SCHEME));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ECC_SCHEME(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ECC_SCHEME",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ECC_SCHEME" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_ECC_SCHEME_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_ECC_SCHEME, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_curveID_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2_ECC_CURVE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_curveID_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_curveID_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_curveID_set" "', argument " "2"" of type '" "TPM2_ECC_CURVE""'");
  } 
  arg2 = (TPM2_ECC_CURVE)(val2);
  if (arg1) (arg1)->curveID = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_curveID_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ECC_CURVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_curveID_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_curveID_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2_ECC_CURVE) ((arg1)->curveID);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_keySize_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_keySize_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_keySize_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_keySize_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->keySize = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_keySize_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_keySize_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_keySize_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (UINT16) ((arg1)->keySize);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_kdf_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPMT_KDF_SCHEME *arg2 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_kdf_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_kdf_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_kdf_set" "', argument " "2"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg2 = (TPMT_KDF_SCHEME *)(argp2);
  if (arg1) (arg1)->kdf = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_kdf_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_kdf_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_kdf_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPMT_KDF_SCHEME *)& ((arg1)->kdf);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_sign_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPMT_ECC_SCHEME *arg2 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_sign_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_sign_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_sign_set" "', argument " "2"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg2 = (TPMT_ECC_SCHEME *)(argp2);
  if (arg1) (arg1)->sign = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_sign_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_sign_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_sign_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPMT_ECC_SCHEME *)& ((arg1)->sign);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_p_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_p_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_p_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_p_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->p = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_p_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_p_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_p_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->p);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_a_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_a_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_a_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_a_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->a = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_a_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_a_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_a_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_b_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_b_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_b_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_b_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->b = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_b_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_b_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_b_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_gX_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_gX_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_gX_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_gX_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->gX = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_gX_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_gX_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_gX_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->gX);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_gY_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_gY_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_gY_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_gY_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->gY = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_gY_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_gY_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_gY_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->gY);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_n_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_n_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_n_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_n_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->n = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_n_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_n_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_n_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->n);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_h_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_h_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_h_set" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_h_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->h = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_h_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_h_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_h_get" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->h);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DETAIL_ECC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALGORITHM_DETAIL_ECC")) SWIG_fail;
  result = (TPMS_ALGORITHM_DETAIL_ECC *)calloc(1, sizeof(TPMS_ALGORITHM_DETAIL_ECC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DETAIL_ECC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DETAIL_ECC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DETAIL_ECC" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ALGORITHM_DETAIL_ECC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_hash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_RSA_hash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_hash_set" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SIGNATURE_RSA_hash_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hash = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_hash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_RSA_hash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_hash_get" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hash);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_sig_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  TPM2B_PUBLIC_KEY_RSA *arg2 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_RSA_sig_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_sig_set" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_RSA_sig_set" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg2 = (TPM2B_PUBLIC_KEY_RSA *)(argp2);
  if (arg1) (arg1)->sig = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_sig_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_RSA_sig_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_sig_get" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  result = (TPM2B_PUBLIC_KEY_RSA *)& ((arg1)->sig);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_RSA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SIGNATURE_RSA")) SWIG_fail;
  result = (TPMS_SIGNATURE_RSA *)calloc(1, sizeof(TPMS_SIGNATURE_RSA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_RSA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_RSA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_RSA" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SIGNATURE_RSA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SIGNATURE_RSA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_hash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_ECC_hash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_hash_set" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SIGNATURE_ECC_hash_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->hash = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_hash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_hash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_hash_get" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->hash);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_signatureR_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_ECC_signatureR_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_signatureR_set" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_ECC_signatureR_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->signatureR = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_signatureR_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_signatureR_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_signatureR_get" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->signatureR);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_signatureS_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_ECC_signatureS_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_signatureS_set" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_ECC_signatureS_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->signatureS = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_signatureS_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_signatureS_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_signatureS_get" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->signatureS);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_ECC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SIGNATURE_ECC")) SWIG_fail;
  result = (TPMS_SIGNATURE_ECC *)calloc(1, sizeof(TPMS_SIGNATURE_ECC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_ECC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_ECC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_ECC" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SIGNATURE_ECC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SIGNATURE_ECC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_rsassa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SIGNATURE_RSASSA *arg2 = (TPMS_SIGNATURE_RSASSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_rsassa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_rsassa_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_rsassa_set" "', argument " "2"" of type '" "TPMS_SIGNATURE_RSASSA *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_RSASSA *)(argp2);
  if (arg1) (arg1)->rsassa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_rsassa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSASSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_rsassa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_rsassa_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SIGNATURE_RSASSA *)& ((arg1)->rsassa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_rsapss_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SIGNATURE_RSAPSS *arg2 = (TPMS_SIGNATURE_RSAPSS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_rsapss_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_rsapss_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_rsapss_set" "', argument " "2"" of type '" "TPMS_SIGNATURE_RSAPSS *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_RSAPSS *)(argp2);
  if (arg1) (arg1)->rsapss = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_rsapss_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSAPSS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_rsapss_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_rsapss_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SIGNATURE_RSAPSS *)& ((arg1)->rsapss);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ecdsa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SIGNATURE_ECDSA *arg2 = (TPMS_SIGNATURE_ECDSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_ecdsa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ecdsa_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_ecdsa_set" "', argument " "2"" of type '" "TPMS_SIGNATURE_ECDSA *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_ECDSA *)(argp2);
  if (arg1) (arg1)->ecdsa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ecdsa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECDSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_ecdsa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ecdsa_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SIGNATURE_ECDSA *)& ((arg1)->ecdsa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ecdaa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SIGNATURE_ECDAA *arg2 = (TPMS_SIGNATURE_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_ecdaa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ecdaa_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_ecdaa_set" "', argument " "2"" of type '" "TPMS_SIGNATURE_ECDAA *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_ECDAA *)(argp2);
  if (arg1) (arg1)->ecdaa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ecdaa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_ecdaa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ecdaa_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SIGNATURE_ECDAA *)& ((arg1)->ecdaa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_sm2_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SIGNATURE_SM2 *arg2 = (TPMS_SIGNATURE_SM2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_sm2_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_sm2_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_sm2_set" "', argument " "2"" of type '" "TPMS_SIGNATURE_SM2 *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_SM2 *)(argp2);
  if (arg1) (arg1)->sm2 = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_sm2_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_SM2 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_sm2_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_sm2_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SIGNATURE_SM2 *)& ((arg1)->sm2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ecschnorr_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SIGNATURE_ECSCHNORR *arg2 = (TPMS_SIGNATURE_ECSCHNORR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_ecschnorr_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ecschnorr_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_ecschnorr_set" "', argument " "2"" of type '" "TPMS_SIGNATURE_ECSCHNORR *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_ECSCHNORR *)(argp2);
  if (arg1) (arg1)->ecschnorr = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ecschnorr_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECSCHNORR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_ecschnorr_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ecschnorr_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SIGNATURE_ECSCHNORR *)& ((arg1)->ecschnorr);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_hmac_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMT_HA *arg2 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_hmac_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_hmac_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_hmac_set" "', argument " "2"" of type '" "TPMT_HA *""'"); 
  }
  arg2 = (TPMT_HA *)(argp2);
  if (arg1) (arg1)->hmac = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_hmac_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_hmac_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_hmac_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMT_HA *)& ((arg1)->hmac);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_any_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMS_SCHEME_HASH *arg2 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_any_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_any_set" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_any_set" "', argument " "2"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg2 = (TPMS_SCHEME_HASH *)(argp2);
  if (arg1) (arg1)->any = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_any_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_any_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_any_get" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMS_SCHEME_HASH *)& ((arg1)->any);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SIGNATURE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SIGNATURE")) SWIG_fail;
  result = (TPMU_SIGNATURE *)calloc(1, sizeof(TPMU_SIGNATURE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIGNATURE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIGNATURE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIGNATURE" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SIGNATURE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SIGNATURE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_sigAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  TPMI_ALG_SIG_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIGNATURE_sigAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_sigAlg_set" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SIGNATURE_sigAlg_set" "', argument " "2"" of type '" "TPMI_ALG_SIG_SCHEME""'");
  } 
  arg2 = (TPMI_ALG_SIG_SCHEME)(val2);
  if (arg1) (arg1)->sigAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_sigAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_SIG_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIGNATURE_sigAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_sigAlg_get" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  result = (TPMI_ALG_SIG_SCHEME) ((arg1)->sigAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_signature_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  TPMU_SIGNATURE *arg2 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIGNATURE_signature_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_signature_set" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SIGNATURE_signature_set" "', argument " "2"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg2 = (TPMU_SIGNATURE *)(argp2);
  if (arg1) (arg1)->signature = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_signature_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIGNATURE_signature_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_signature_get" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  result = (TPMU_SIGNATURE *)& ((arg1)->signature);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SIGNATURE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SIGNATURE")) SWIG_fail;
  result = (TPMT_SIGNATURE *)calloc(1, sizeof(TPMT_SIGNATURE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIGNATURE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIGNATURE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIGNATURE" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SIGNATURE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SIGNATURE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_ecc_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_ecc_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_ecc_set" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ENCRYPTED_SECRET_ecc_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMS_ECC_POINT)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMS_ECC_POINT); ++ii) *(BYTE *)&arg1->ecc[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""ecc""' of type '""BYTE [sizeof(TPMS_ECC_POINT)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_ecc_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_ecc_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_ecc_get" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->ecc);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_rsa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_rsa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_rsa_set" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ENCRYPTED_SECRET_rsa_set" "', argument " "2"" of type '" "BYTE [512]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)512; ++ii) *(BYTE *)&arg1->rsa[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""rsa""' of type '""BYTE [512]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_rsa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_rsa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_rsa_get" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->rsa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_symmetric_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_symmetric_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_symmetric_set" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ENCRYPTED_SECRET_symmetric_set" "', argument " "2"" of type '" "BYTE [sizeof(TPM2B_DIGEST)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPM2B_DIGEST); ++ii) *(BYTE *)&arg1->symmetric[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""symmetric""' of type '""BYTE [sizeof(TPM2B_DIGEST)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_symmetric_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_symmetric_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_symmetric_get" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->symmetric);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_keyedHash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_keyedHash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_keyedHash_set" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ENCRYPTED_SECRET_keyedHash_set" "', argument " "2"" of type '" "BYTE [sizeof(TPM2B_DIGEST)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPM2B_DIGEST); ++ii) *(BYTE *)&arg1->keyedHash[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""keyedHash""' of type '""BYTE [sizeof(TPM2B_DIGEST)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_keyedHash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_keyedHash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_keyedHash_get" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->keyedHash);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_ENCRYPTED_SECRET(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ENCRYPTED_SECRET")) SWIG_fail;
  result = (TPMU_ENCRYPTED_SECRET *)calloc(1, sizeof(TPMU_ENCRYPTED_SECRET));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ENCRYPTED_SECRET(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ENCRYPTED_SECRET",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ENCRYPTED_SECRET" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_ENCRYPTED_SECRET_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ENCRYPTED_SECRET_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_size_set" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ENCRYPTED_SECRET_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ENCRYPTED_SECRET_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_size_get" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_secret_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ENCRYPTED_SECRET_secret_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_secret_set" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ENCRYPTED_SECRET_secret_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMU_ENCRYPTED_SECRET)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMU_ENCRYPTED_SECRET); ++ii) *(BYTE *)&arg1->secret[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""secret""' of type '""BYTE [sizeof(TPMU_ENCRYPTED_SECRET)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_secret_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ENCRYPTED_SECRET_secret_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_secret_get" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->secret);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ENCRYPTED_SECRET(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ENCRYPTED_SECRET")) SWIG_fail;
  result = (TPM2B_ENCRYPTED_SECRET *)calloc(1, sizeof(TPM2B_ENCRYPTED_SECRET));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ENCRYPTED_SECRET(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ENCRYPTED_SECRET",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ENCRYPTED_SECRET" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ENCRYPTED_SECRET_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_keyedHash_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_keyedHash_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_keyedHash_set" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_ID_keyedHash_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->keyedHash = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_keyedHash_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_keyedHash_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_keyedHash_get" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->keyedHash);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_sym_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_sym_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_sym_set" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_ID_sym_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->sym = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_sym_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_sym_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_sym_get" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->sym);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_rsa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  TPM2B_PUBLIC_KEY_RSA *arg2 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_rsa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_rsa_set" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_ID_rsa_set" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg2 = (TPM2B_PUBLIC_KEY_RSA *)(argp2);
  if (arg1) (arg1)->rsa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_rsa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_rsa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_rsa_get" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = (TPM2B_PUBLIC_KEY_RSA *)& ((arg1)->rsa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_ecc_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  TPMS_ECC_POINT *arg2 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_ecc_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_ecc_set" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_ID_ecc_set" "', argument " "2"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg2 = (TPMS_ECC_POINT *)(argp2);
  if (arg1) (arg1)->ecc = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_ecc_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_ecc_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_ecc_get" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = (TPMS_ECC_POINT *)& ((arg1)->ecc);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_ID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_PUBLIC_ID")) SWIG_fail;
  result = (TPMU_PUBLIC_ID *)calloc(1, sizeof(TPMU_PUBLIC_ID));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_ID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_ID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_ID" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_PUBLIC_ID_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_PUBLIC_ID, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  TPMT_KEYEDHASH_SCHEME *arg2 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_KEYEDHASH_PARMS_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_scheme_set" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_KEYEDHASH_PARMS_scheme_set" "', argument " "2"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg2 = (TPMT_KEYEDHASH_SCHEME *)(argp2);
  if (arg1) (arg1)->scheme = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_KEYEDHASH_PARMS_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_scheme_get" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  result = (TPMT_KEYEDHASH_SCHEME *)& ((arg1)->scheme);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_KEYEDHASH_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_KEYEDHASH_PARMS")) SWIG_fail;
  result = (TPMS_KEYEDHASH_PARMS *)calloc(1, sizeof(TPMS_KEYEDHASH_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_KEYEDHASH_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_KEYEDHASH_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_KEYEDHASH_PARMS" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_KEYEDHASH_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_symmetric_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg2 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ASYM_PARMS_symmetric_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_symmetric_set" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ASYM_PARMS_symmetric_set" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg2 = (TPMT_SYM_DEF_OBJECT *)(argp2);
  if (arg1) (arg1)->symmetric = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_symmetric_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ASYM_PARMS_symmetric_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_symmetric_get" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT *)& ((arg1)->symmetric);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  TPMT_ASYM_SCHEME *arg2 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ASYM_PARMS_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_scheme_set" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ASYM_PARMS_scheme_set" "', argument " "2"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMT_ASYM_SCHEME *)(argp2);
  if (arg1) (arg1)->scheme = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ASYM_PARMS_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_scheme_get" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  result = (TPMT_ASYM_SCHEME *)& ((arg1)->scheme);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ASYM_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ASYM_PARMS")) SWIG_fail;
  result = (TPMS_ASYM_PARMS *)calloc(1, sizeof(TPMS_ASYM_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ASYM_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ASYM_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ASYM_PARMS" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ASYM_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ASYM_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_symmetric_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg2 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_symmetric_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_symmetric_set" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_RSA_PARMS_symmetric_set" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg2 = (TPMT_SYM_DEF_OBJECT *)(argp2);
  if (arg1) (arg1)->symmetric = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_symmetric_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_symmetric_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_symmetric_get" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT *)& ((arg1)->symmetric);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  TPMT_RSA_SCHEME *arg2 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_scheme_set" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_RSA_PARMS_scheme_set" "', argument " "2"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg2 = (TPMT_RSA_SCHEME *)(argp2);
  if (arg1) (arg1)->scheme = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_scheme_get" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = (TPMT_RSA_SCHEME *)& ((arg1)->scheme);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_keyBits_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  TPMI_RSA_KEY_BITS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_keyBits_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_keyBits_set" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_RSA_PARMS_keyBits_set" "', argument " "2"" of type '" "TPMI_RSA_KEY_BITS""'");
  } 
  arg2 = (TPMI_RSA_KEY_BITS)(val2);
  if (arg1) (arg1)->keyBits = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_keyBits_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RSA_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_keyBits_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_keyBits_get" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = (TPMI_RSA_KEY_BITS) ((arg1)->keyBits);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_exponent_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_exponent_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_exponent_set" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_RSA_PARMS_exponent_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->exponent = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_exponent_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_exponent_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_exponent_get" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = (UINT32) ((arg1)->exponent);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_RSA_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_RSA_PARMS")) SWIG_fail;
  result = (TPMS_RSA_PARMS *)calloc(1, sizeof(TPMS_RSA_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_RSA_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_RSA_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_RSA_PARMS" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_RSA_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_RSA_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_symmetric_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg2 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_symmetric_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_symmetric_set" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_PARMS_symmetric_set" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg2 = (TPMT_SYM_DEF_OBJECT *)(argp2);
  if (arg1) (arg1)->symmetric = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_symmetric_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_symmetric_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_symmetric_get" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT *)& ((arg1)->symmetric);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_scheme_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  TPMT_ECC_SCHEME *arg2 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_scheme_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_scheme_set" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_PARMS_scheme_set" "', argument " "2"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg2 = (TPMT_ECC_SCHEME *)(argp2);
  if (arg1) (arg1)->scheme = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_scheme_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_scheme_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_scheme_get" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = (TPMT_ECC_SCHEME *)& ((arg1)->scheme);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_curveID_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  TPMI_ECC_CURVE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_curveID_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_curveID_set" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ECC_PARMS_curveID_set" "', argument " "2"" of type '" "TPMI_ECC_CURVE""'");
  } 
  arg2 = (TPMI_ECC_CURVE)(val2);
  if (arg1) (arg1)->curveID = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_curveID_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ECC_CURVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_curveID_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_curveID_get" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = (TPMI_ECC_CURVE) ((arg1)->curveID);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_kdf_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  TPMT_KDF_SCHEME *arg2 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_kdf_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_kdf_set" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_PARMS_kdf_set" "', argument " "2"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg2 = (TPMT_KDF_SCHEME *)(argp2);
  if (arg1) (arg1)->kdf = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_kdf_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_kdf_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_kdf_get" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = (TPMT_KDF_SCHEME *)& ((arg1)->kdf);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ECC_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ECC_PARMS")) SWIG_fail;
  result = (TPMS_ECC_PARMS *)calloc(1, sizeof(TPMS_ECC_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_PARMS" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ECC_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ECC_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_keyedHashDetail_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  TPMS_KEYEDHASH_PARMS *arg2 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_keyedHashDetail_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_keyedHashDetail_set" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_keyedHashDetail_set" "', argument " "2"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg2 = (TPMS_KEYEDHASH_PARMS *)(argp2);
  if (arg1) (arg1)->keyedHashDetail = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_keyedHashDetail_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_keyedHashDetail_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_keyedHashDetail_get" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMS_KEYEDHASH_PARMS *)& ((arg1)->keyedHashDetail);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_symDetail_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  TPMS_SYMCIPHER_PARMS *arg2 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_symDetail_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_symDetail_set" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_symDetail_set" "', argument " "2"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg2 = (TPMS_SYMCIPHER_PARMS *)(argp2);
  if (arg1) (arg1)->symDetail = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_symDetail_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_symDetail_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_symDetail_get" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMS_SYMCIPHER_PARMS *)& ((arg1)->symDetail);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_rsaDetail_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  TPMS_RSA_PARMS *arg2 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_rsaDetail_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_rsaDetail_set" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_rsaDetail_set" "', argument " "2"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg2 = (TPMS_RSA_PARMS *)(argp2);
  if (arg1) (arg1)->rsaDetail = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_rsaDetail_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_rsaDetail_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_rsaDetail_get" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMS_RSA_PARMS *)& ((arg1)->rsaDetail);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_eccDetail_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  TPMS_ECC_PARMS *arg2 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_eccDetail_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_eccDetail_set" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_eccDetail_set" "', argument " "2"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg2 = (TPMS_ECC_PARMS *)(argp2);
  if (arg1) (arg1)->eccDetail = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_eccDetail_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_eccDetail_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_eccDetail_get" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMS_ECC_PARMS *)& ((arg1)->eccDetail);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_asymDetail_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  TPMS_ASYM_PARMS *arg2 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_asymDetail_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_asymDetail_set" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_asymDetail_set" "', argument " "2"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg2 = (TPMS_ASYM_PARMS *)(argp2);
  if (arg1) (arg1)->asymDetail = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_asymDetail_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_asymDetail_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_asymDetail_get" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMS_ASYM_PARMS *)& ((arg1)->asymDetail);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_PUBLIC_PARMS")) SWIG_fail;
  result = (TPMU_PUBLIC_PARMS *)calloc(1, sizeof(TPMU_PUBLIC_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_PARMS" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_PUBLIC_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_PUBLIC_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_type_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  TPMI_ALG_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PARMS_type_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_type_set" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_PARMS_type_set" "', argument " "2"" of type '" "TPMI_ALG_PUBLIC""'");
  } 
  arg2 = (TPMI_ALG_PUBLIC)(val2);
  if (arg1) (arg1)->type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_type_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PARMS_type_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_type_get" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  result = (TPMI_ALG_PUBLIC) ((arg1)->type);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_parameters_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  TPMU_PUBLIC_PARMS *arg2 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PARMS_parameters_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_parameters_set" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_PARMS_parameters_set" "', argument " "2"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg2 = (TPMU_PUBLIC_PARMS *)(argp2);
  if (arg1) (arg1)->parameters = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_parameters_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PARMS_parameters_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_parameters_get" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  result = (TPMU_PUBLIC_PARMS *)& ((arg1)->parameters);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_PUBLIC_PARMS")) SWIG_fail;
  result = (TPMT_PUBLIC_PARMS *)calloc(1, sizeof(TPMT_PUBLIC_PARMS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_PARMS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_PARMS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_PARMS" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_PUBLIC_PARMS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_PUBLIC_PARMS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_type_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPMI_ALG_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_type_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_type_set" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_type_set" "', argument " "2"" of type '" "TPMI_ALG_PUBLIC""'");
  } 
  arg2 = (TPMI_ALG_PUBLIC)(val2);
  if (arg1) (arg1)->type = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_type_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_type_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_type_get" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMI_ALG_PUBLIC) ((arg1)->type);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_nameAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_nameAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_nameAlg_set" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_nameAlg_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->nameAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_nameAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_nameAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_nameAlg_get" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->nameAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_objectAttributes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPMA_OBJECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_objectAttributes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_objectAttributes_set" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_objectAttributes_set" "', argument " "2"" of type '" "TPMA_OBJECT""'");
  } 
  arg2 = (TPMA_OBJECT)(val2);
  if (arg1) (arg1)->objectAttributes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_objectAttributes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_objectAttributes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_objectAttributes_get" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMA_OBJECT) ((arg1)->objectAttributes);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_authPolicy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_authPolicy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_authPolicy_set" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_authPolicy_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->authPolicy = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_authPolicy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_authPolicy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_authPolicy_get" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->authPolicy);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_parameters_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPMU_PUBLIC_PARMS *arg2 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_parameters_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_parameters_set" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_parameters_set" "', argument " "2"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg2 = (TPMU_PUBLIC_PARMS *)(argp2);
  if (arg1) (arg1)->parameters = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_parameters_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_parameters_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_parameters_get" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMU_PUBLIC_PARMS *)& ((arg1)->parameters);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_unique_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPMU_PUBLIC_ID *arg2 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_unique_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_unique_set" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_unique_set" "', argument " "2"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg2 = (TPMU_PUBLIC_ID *)(argp2);
  if (arg1) (arg1)->unique = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_unique_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_unique_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_unique_get" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMU_PUBLIC_ID *)& ((arg1)->unique);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_PUBLIC")) SWIG_fail;
  result = (TPMT_PUBLIC *)calloc(1, sizeof(TPMT_PUBLIC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_PUBLIC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_PUBLIC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_size_set" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PUBLIC_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_size_get" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_publicArea_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  TPMT_PUBLIC *arg2 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_publicArea_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_publicArea_set" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PUBLIC_publicArea_set" "', argument " "2"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg2 = (TPMT_PUBLIC *)(argp2);
  if (arg1) (arg1)->publicArea = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_publicArea_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_publicArea_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_publicArea_get" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  result = (TPMT_PUBLIC *)& ((arg1)->publicArea);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PUBLIC")) SWIG_fail;
  result = (TPM2B_PUBLIC *)calloc(1, sizeof(TPM2B_PUBLIC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PUBLIC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PUBLIC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_TEMPLATE_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_size_set" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_TEMPLATE_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_TEMPLATE_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_size_get" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_TEMPLATE_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_buffer_set" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_TEMPLATE_buffer_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMT_PUBLIC)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMT_PUBLIC); ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [sizeof(TPMT_PUBLIC)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_TEMPLATE_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_buffer_get" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_TEMPLATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_TEMPLATE")) SWIG_fail;
  result = (TPM2B_TEMPLATE *)calloc(1, sizeof(TPM2B_TEMPLATE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_TEMPLATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_TEMPLATE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_TEMPLATE" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_TEMPLATE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_TEMPLATE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_VENDOR_SPECIFIC_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_size_set" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_VENDOR_SPECIFIC_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_size_get" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_set" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_set" "', argument " "2"" of type '" "BYTE [((512/2) *(3+2))]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)((512/2) *(3+2)); ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [((512/2) *(3+2))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_get" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_VENDOR_SPECIFIC")) SWIG_fail;
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)calloc(1, sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_VENDOR_SPECIFIC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_VENDOR_SPECIFIC" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PRIVATE_VENDOR_SPECIFIC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_rsa_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  TPM2B_PRIVATE_KEY_RSA *arg2 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_rsa_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_rsa_set" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_rsa_set" "', argument " "2"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg2 = (TPM2B_PRIVATE_KEY_RSA *)(argp2);
  if (arg1) (arg1)->rsa = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_rsa_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_rsa_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_rsa_get" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPM2B_PRIVATE_KEY_RSA *)& ((arg1)->rsa);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_ecc_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_ecc_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_ecc_set" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_ecc_set" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  if (arg1) (arg1)->ecc = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_ecc_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_ecc_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_ecc_get" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)& ((arg1)->ecc);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_bits_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  TPM2B_SENSITIVE_DATA *arg2 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_bits_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_bits_set" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_bits_set" "', argument " "2"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg2 = (TPM2B_SENSITIVE_DATA *)(argp2);
  if (arg1) (arg1)->bits = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_bits_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_bits_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_bits_get" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPM2B_SENSITIVE_DATA *)& ((arg1)->bits);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_sym_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  TPM2B_SYM_KEY *arg2 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_sym_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_sym_set" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_sym_set" "', argument " "2"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg2 = (TPM2B_SYM_KEY *)(argp2);
  if (arg1) (arg1)->sym = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_sym_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_sym_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_sym_get" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPM2B_SYM_KEY *)& ((arg1)->sym);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_any_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg2 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_any_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_any_set" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_any_set" "', argument " "2"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg2 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp2);
  if (arg1) (arg1)->any = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_any_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_any_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_any_get" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)& ((arg1)->any);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SENSITIVE_COMPOSITE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SENSITIVE_COMPOSITE")) SWIG_fail;
  result = (TPMU_SENSITIVE_COMPOSITE *)calloc(1, sizeof(TPMU_SENSITIVE_COMPOSITE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SENSITIVE_COMPOSITE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SENSITIVE_COMPOSITE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SENSITIVE_COMPOSITE" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SENSITIVE_COMPOSITE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_sensitiveType_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  TPMI_ALG_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_sensitiveType_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_sensitiveType_set" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SENSITIVE_sensitiveType_set" "', argument " "2"" of type '" "TPMI_ALG_PUBLIC""'");
  } 
  arg2 = (TPMI_ALG_PUBLIC)(val2);
  if (arg1) (arg1)->sensitiveType = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_sensitiveType_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_sensitiveType_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_sensitiveType_get" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = (TPMI_ALG_PUBLIC) ((arg1)->sensitiveType);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_authValue_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  TPM2B_AUTH *arg2 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_authValue_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_authValue_set" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SENSITIVE_authValue_set" "', argument " "2"" of type '" "TPM2B_AUTH *""'"); 
  }
  arg2 = (TPM2B_AUTH *)(argp2);
  if (arg1) (arg1)->authValue = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_authValue_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_authValue_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_authValue_get" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = (TPM2B_AUTH *)& ((arg1)->authValue);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_seedValue_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_seedValue_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_seedValue_set" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SENSITIVE_seedValue_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->seedValue = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_seedValue_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_seedValue_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_seedValue_get" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->seedValue);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_sensitive_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  TPMU_SENSITIVE_COMPOSITE *arg2 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_sensitive_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_sensitive_set" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SENSITIVE_sensitive_set" "', argument " "2"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg2 = (TPMU_SENSITIVE_COMPOSITE *)(argp2);
  if (arg1) (arg1)->sensitive = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_sensitive_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_sensitive_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_sensitive_get" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = (TPMU_SENSITIVE_COMPOSITE *)& ((arg1)->sensitive);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SENSITIVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SENSITIVE")) SWIG_fail;
  result = (TPMT_SENSITIVE *)calloc(1, sizeof(TPMT_SENSITIVE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SENSITIVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SENSITIVE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SENSITIVE" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SENSITIVE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SENSITIVE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_size_set" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_size_get" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_sensitiveArea_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  TPMT_SENSITIVE *arg2 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_sensitiveArea_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_sensitiveArea_set" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_sensitiveArea_set" "', argument " "2"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg2 = (TPMT_SENSITIVE *)(argp2);
  if (arg1) (arg1)->sensitiveArea = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_sensitiveArea_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_sensitiveArea_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_sensitiveArea_get" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  result = (TPMT_SENSITIVE *)& ((arg1)->sensitiveArea);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE")) SWIG_fail;
  result = (TPM2B_SENSITIVE *)calloc(1, sizeof(TPM2B_SENSITIVE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SENSITIVE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SENSITIVE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap__PRIVATE_integrityOuter_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:_PRIVATE_integrityOuter_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_integrityOuter_set" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_PRIVATE_integrityOuter_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->integrityOuter = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_integrityOuter_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_integrityOuter_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_integrityOuter_get" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->integrityOuter);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_integrityInner_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:_PRIVATE_integrityInner_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_integrityInner_set" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_PRIVATE_integrityInner_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->integrityInner = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_integrityInner_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_integrityInner_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_integrityInner_get" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->integrityInner);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_sensitive_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  TPM2B_SENSITIVE *arg2 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:_PRIVATE_sensitive_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_sensitive_set" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_PRIVATE_sensitive_set" "', argument " "2"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg2 = (TPM2B_SENSITIVE *)(argp2);
  if (arg1) (arg1)->sensitive = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_sensitive_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_sensitive_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_sensitive_get" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  result = (TPM2B_SENSITIVE *)& ((arg1)->sensitive);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new__PRIVATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new__PRIVATE")) SWIG_fail;
  result = (_PRIVATE *)calloc(1, sizeof(_PRIVATE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete__PRIVATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete__PRIVATE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__PRIVATE" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_PRIVATE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p__PRIVATE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_size_set" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_size_get" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_buffer_set" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_buffer_set" "', argument " "2"" of type '" "BYTE [sizeof(_PRIVATE)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(_PRIVATE); ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [sizeof(_PRIVATE)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_buffer_get" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE")) SWIG_fail;
  result = (TPM2B_PRIVATE *)calloc(1, sizeof(TPM2B_PRIVATE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PRIVATE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PRIVATE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_integrityHMAC_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ID_OBJECT_integrityHMAC_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_integrityHMAC_set" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ID_OBJECT_integrityHMAC_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->integrityHMAC = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_integrityHMAC_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ID_OBJECT_integrityHMAC_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_integrityHMAC_get" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->integrityHMAC);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_encIdentity_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ID_OBJECT_encIdentity_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_encIdentity_set" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ID_OBJECT_encIdentity_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->encIdentity = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_encIdentity_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ID_OBJECT_encIdentity_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_encIdentity_get" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->encIdentity);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ID_OBJECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ID_OBJECT")) SWIG_fail;
  result = (TPMS_ID_OBJECT *)calloc(1, sizeof(TPMS_ID_OBJECT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ID_OBJECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ID_OBJECT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ID_OBJECT" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ID_OBJECT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ID_OBJECT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ID_OBJECT_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_size_set" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ID_OBJECT_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ID_OBJECT_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_size_get" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_credential_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ID_OBJECT_credential_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_credential_set" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ID_OBJECT_credential_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMS_ID_OBJECT)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMS_ID_OBJECT); ++ii) *(BYTE *)&arg1->credential[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""credential""' of type '""BYTE [sizeof(TPMS_ID_OBJECT)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_credential_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ID_OBJECT_credential_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_credential_get" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->credential);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ID_OBJECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ID_OBJECT")) SWIG_fail;
  result = (TPM2B_ID_OBJECT *)calloc(1, sizeof(TPM2B_ID_OBJECT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ID_OBJECT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ID_OBJECT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ID_OBJECT" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ID_OBJECT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ID_OBJECT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *TPM2_NV_INDEX_INDEX_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NV_INDEX_INDEX_SHIFT",SWIG_From_int((int)((0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NV_INDEX_RH_NV_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NV_INDEX_RH_NV_SHIFT",SWIG_From_int((int)((24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_set" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->pinCount = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_get" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  result = (UINT32) ((arg1)->pinCount);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_set" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->pinLimit = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_get" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  result = (UINT32) ((arg1)->pinLimit);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_PIN_COUNTER_PARAMETERS")) SWIG_fail;
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS *)calloc(1, sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PIN_COUNTER_PARAMETERS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PIN_COUNTER_PARAMETERS" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_NV_PIN_COUNTER_PARAMETERS_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *TPMA_NV_TPM2_NT_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_TPM2_NT_SHIFT",SWIG_From_int((int)((4))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_nvIndex_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  TPMI_RH_NV_INDEX arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_nvIndex_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_nvIndex_set" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PUBLIC_nvIndex_set" "', argument " "2"" of type '" "TPMI_RH_NV_INDEX""'");
  } 
  arg2 = (TPMI_RH_NV_INDEX)(val2);
  if (arg1) (arg1)->nvIndex = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_nvIndex_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RH_NV_INDEX result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_nvIndex_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_nvIndex_get" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (TPMI_RH_NV_INDEX) ((arg1)->nvIndex);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_nameAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  TPMI_ALG_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_nameAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_nameAlg_set" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PUBLIC_nameAlg_set" "', argument " "2"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg2 = (TPMI_ALG_HASH)(val2);
  if (arg1) (arg1)->nameAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_nameAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_ALG_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_nameAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_nameAlg_get" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (TPMI_ALG_HASH) ((arg1)->nameAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_attributes_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  TPMA_NV arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_attributes_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_attributes_set" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PUBLIC_attributes_set" "', argument " "2"" of type '" "TPMA_NV""'");
  } 
  arg2 = (TPMA_NV)(val2);
  if (arg1) (arg1)->attributes = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_attributes_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_NV result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_attributes_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_attributes_get" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (TPMA_NV) ((arg1)->attributes);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_authPolicy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_authPolicy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_authPolicy_set" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_PUBLIC_authPolicy_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->authPolicy = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_authPolicy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_authPolicy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_authPolicy_get" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->authPolicy);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_dataSize_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_dataSize_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_dataSize_set" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PUBLIC_dataSize_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->dataSize = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_dataSize_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_dataSize_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_dataSize_get" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (UINT16) ((arg1)->dataSize);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_PUBLIC")) SWIG_fail;
  result = (TPMS_NV_PUBLIC *)calloc(1, sizeof(TPMS_NV_PUBLIC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PUBLIC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PUBLIC" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_NV_PUBLIC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_NV_PUBLIC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NV_PUBLIC_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_size_set" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_NV_PUBLIC_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NV_PUBLIC_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_size_get" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_nvPublic_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  TPMS_NV_PUBLIC *arg2 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NV_PUBLIC_nvPublic_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_nvPublic_set" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_NV_PUBLIC_nvPublic_set" "', argument " "2"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg2 = (TPMS_NV_PUBLIC *)(argp2);
  if (arg1) (arg1)->nvPublic = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_nvPublic_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NV_PUBLIC_nvPublic_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_nvPublic_get" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  result = (TPMS_NV_PUBLIC *)& ((arg1)->nvPublic);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_NV_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_NV_PUBLIC")) SWIG_fail;
  result = (TPM2B_NV_PUBLIC *)calloc(1, sizeof(TPM2B_NV_PUBLIC));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NV_PUBLIC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NV_PUBLIC",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NV_PUBLIC" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_NV_PUBLIC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_NV_PUBLIC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_SENSITIVE_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_size_set" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CONTEXT_SENSITIVE_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_SENSITIVE_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_size_get" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_SENSITIVE_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_buffer_set" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CONTEXT_SENSITIVE_buffer_set" "', argument " "2"" of type '" "BYTE [5120]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)5120; ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [5120]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_SENSITIVE_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_buffer_get" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_SENSITIVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CONTEXT_SENSITIVE")) SWIG_fail;
  result = (TPM2B_CONTEXT_SENSITIVE *)calloc(1, sizeof(TPM2B_CONTEXT_SENSITIVE));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_SENSITIVE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_SENSITIVE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_SENSITIVE" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_CONTEXT_SENSITIVE_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_integrity_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_DATA_integrity_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_integrity_set" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_DATA_integrity_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->integrity = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_integrity_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_DATA_integrity_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_integrity_get" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->integrity);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_encrypted_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  TPM2B_CONTEXT_SENSITIVE *arg2 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_DATA_encrypted_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_encrypted_set" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_DATA_encrypted_set" "', argument " "2"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg2 = (TPM2B_CONTEXT_SENSITIVE *)(argp2);
  if (arg1) (arg1)->encrypted = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_encrypted_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_DATA_encrypted_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_encrypted_get" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  result = (TPM2B_CONTEXT_SENSITIVE *)& ((arg1)->encrypted);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CONTEXT_DATA")) SWIG_fail;
  result = (TPMS_CONTEXT_DATA *)calloc(1, sizeof(TPMS_CONTEXT_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_DATA" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CONTEXT_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CONTEXT_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_DATA_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_size_set" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CONTEXT_DATA_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_DATA_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_size_get" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_buffer_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  BYTE *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_DATA_buffer_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_buffer_set" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CONTEXT_DATA_buffer_set" "', argument " "2"" of type '" "BYTE [sizeof(TPMS_CONTEXT_DATA)]""'"); 
  } 
  arg2 = (BYTE *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)sizeof(TPMS_CONTEXT_DATA); ++ii) *(BYTE *)&arg1->buffer[ii] = *((BYTE *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""buffer""' of type '""BYTE [sizeof(TPMS_CONTEXT_DATA)]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_buffer_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_DATA_buffer_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_buffer_get" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  result = (BYTE *)(BYTE *) ((arg1)->buffer);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CONTEXT_DATA")) SWIG_fail;
  result = (TPM2B_CONTEXT_DATA *)calloc(1, sizeof(TPM2B_CONTEXT_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_DATA" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_CONTEXT_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_CONTEXT_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_sequence_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  UINT64 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_sequence_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_sequence_set" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_sequence_set" "', argument " "2"" of type '" "UINT64""'");
  } 
  arg2 = (UINT64)(val2);
  if (arg1) (arg1)->sequence = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_sequence_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT64 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_sequence_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_sequence_get" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = (UINT64) ((arg1)->sequence);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_savedHandle_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  TPMI_DH_CONTEXT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_savedHandle_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_savedHandle_set" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_savedHandle_set" "', argument " "2"" of type '" "TPMI_DH_CONTEXT""'");
  } 
  arg2 = (TPMI_DH_CONTEXT)(val2);
  if (arg1) (arg1)->savedHandle = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_savedHandle_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_DH_CONTEXT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_savedHandle_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_savedHandle_get" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = (TPMI_DH_CONTEXT) ((arg1)->savedHandle);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_hierarchy_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  TPMI_RH_HIERARCHY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_hierarchy_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_hierarchy_set" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_hierarchy_set" "', argument " "2"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg2 = (TPMI_RH_HIERARCHY)(val2);
  if (arg1) (arg1)->hierarchy = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_hierarchy_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_RH_HIERARCHY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_hierarchy_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_hierarchy_get" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = (TPMI_RH_HIERARCHY) ((arg1)->hierarchy);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_contextBlob_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  TPM2B_CONTEXT_DATA *arg2 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_contextBlob_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_contextBlob_set" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_contextBlob_set" "', argument " "2"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg2 = (TPM2B_CONTEXT_DATA *)(argp2);
  if (arg1) (arg1)->contextBlob = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_contextBlob_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_contextBlob_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_contextBlob_get" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = (TPM2B_CONTEXT_DATA *)& ((arg1)->contextBlob);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CONTEXT")) SWIG_fail;
  result = (TPMS_CONTEXT *)calloc(1, sizeof(TPMS_CONTEXT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CONTEXT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CONTEXT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_pcrSelect_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPML_PCR_SELECTION *arg2 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_pcrSelect_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_pcrSelect_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_pcrSelect_set" "', argument " "2"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg2 = (TPML_PCR_SELECTION *)(argp2);
  if (arg1) (arg1)->pcrSelect = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_pcrSelect_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_pcrSelect_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_pcrSelect_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPML_PCR_SELECTION *)& ((arg1)->pcrSelect);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_pcrDigest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_pcrDigest_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_pcrDigest_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_pcrDigest_set" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  if (arg1) (arg1)->pcrDigest = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_pcrDigest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_pcrDigest_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_pcrDigest_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPM2B_DIGEST *)& ((arg1)->pcrDigest);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_locality_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPMA_LOCALITY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_locality_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_locality_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CREATION_DATA_locality_set" "', argument " "2"" of type '" "TPMA_LOCALITY""'");
  } 
  arg2 = (TPMA_LOCALITY)(val2);
  if (arg1) (arg1)->locality = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_locality_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMA_LOCALITY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_locality_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_locality_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPMA_LOCALITY) ((arg1)->locality);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_parentNameAlg_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPM2_ALG_ID arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_parentNameAlg_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_parentNameAlg_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CREATION_DATA_parentNameAlg_set" "', argument " "2"" of type '" "TPM2_ALG_ID""'");
  } 
  arg2 = (TPM2_ALG_ID)(val2);
  if (arg1) (arg1)->parentNameAlg = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_parentNameAlg_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2_ALG_ID result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_parentNameAlg_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_parentNameAlg_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPM2_ALG_ID) ((arg1)->parentNameAlg);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_parentName_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_parentName_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_parentName_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_parentName_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->parentName = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_parentName_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_parentName_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_parentName_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->parentName);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_parentQualifiedName_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_parentQualifiedName_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_parentQualifiedName_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_parentQualifiedName_set" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  if (arg1) (arg1)->parentQualifiedName = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_parentQualifiedName_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_parentQualifiedName_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_parentQualifiedName_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPM2B_NAME *)& ((arg1)->parentQualifiedName);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_outsideInfo_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPM2B_DATA *arg2 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_outsideInfo_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_outsideInfo_set" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_outsideInfo_set" "', argument " "2"" of type '" "TPM2B_DATA *""'"); 
  }
  arg2 = (TPM2B_DATA *)(argp2);
  if (arg1) (arg1)->outsideInfo = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_outsideInfo_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_outsideInfo_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_outsideInfo_get" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPM2B_DATA *)& ((arg1)->outsideInfo);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CREATION_DATA")) SWIG_fail;
  result = (TPMS_CREATION_DATA *)calloc(1, sizeof(TPMS_CREATION_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_DATA" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CREATION_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CREATION_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_size_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CREATION_DATA_size_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_size_set" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CREATION_DATA_size_set" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  if (arg1) (arg1)->size = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_size_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CREATION_DATA_size_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_size_get" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  result = (UINT16) ((arg1)->size);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_creationData_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  TPMS_CREATION_DATA *arg2 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CREATION_DATA_creationData_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_creationData_set" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CREATION_DATA_creationData_set" "', argument " "2"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg2 = (TPMS_CREATION_DATA *)(argp2);
  if (arg1) (arg1)->creationData = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_creationData_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CREATION_DATA_creationData_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_creationData_get" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  result = (TPMS_CREATION_DATA *)& ((arg1)->creationData);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CREATION_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CREATION_DATA")) SWIG_fail;
  result = (TPM2B_CREATION_DATA *)calloc(1, sizeof(TPM2B_CREATION_DATA));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CREATION_DATA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CREATION_DATA",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CREATION_DATA" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_CREATION_DATA_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_CREATION_DATA, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_tag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  TPM_AT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AC_OUTPUT_tag_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_tag_set" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AC_OUTPUT_tag_set" "', argument " "2"" of type '" "TPM_AT""'");
  } 
  arg2 = (TPM_AT)(val2);
  if (arg1) (arg1)->tag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_tag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM_AT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AC_OUTPUT_tag_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_tag_get" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  result = (TPM_AT) ((arg1)->tag);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_data_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AC_OUTPUT_data_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_data_set" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AC_OUTPUT_data_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->data = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_data_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AC_OUTPUT_data_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_data_get" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  result = (UINT32) ((arg1)->data);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_AC_OUTPUT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AC_OUTPUT")) SWIG_fail;
  result = (TPMS_AC_OUTPUT *)calloc(1, sizeof(TPMS_AC_OUTPUT));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AC_OUTPUT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AC_OUTPUT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AC_OUTPUT" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_AC_OUTPUT_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_AC_OUTPUT, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_count_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_AC_CAPABILITIES_count_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_count_set" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_AC_CAPABILITIES_count_set" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  if (arg1) (arg1)->count = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_count_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_AC_CAPABILITIES_count_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_count_get" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  result = (UINT32) ((arg1)->count);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_acCapabilities_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  TPMS_AC_OUTPUT *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_AC_CAPABILITIES_acCapabilities_set",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_acCapabilities_set" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_AC_CAPABILITIES_acCapabilities_set" "', argument " "2"" of type '" "TPMS_AC_OUTPUT [(1024/sizeof(TPMS_AC_OUTPUT))]""'"); 
  } 
  arg2 = (TPMS_AC_OUTPUT *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)(1024/sizeof(TPMS_AC_OUTPUT)); ++ii) *(TPMS_AC_OUTPUT *)&arg1->acCapabilities[ii] = *((TPMS_AC_OUTPUT *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""acCapabilities""' of type '""TPMS_AC_OUTPUT [(1024/sizeof(TPMS_AC_OUTPUT))]""'");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_acCapabilities_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_AC_CAPABILITIES_acCapabilities_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_acCapabilities_get" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  result = (TPMS_AC_OUTPUT *)(TPMS_AC_OUTPUT *) ((arg1)->acCapabilities);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_AC_CAPABILITIES(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_AC_CAPABILITIES")) SWIG_fail;
  result = (TPML_AC_CAPABILITIES *)calloc(1, sizeof(TPML_AC_CAPABILITIES));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_AC_CAPABILITIES(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_AC_CAPABILITIES",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_AC_CAPABILITIES" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_AC_CAPABILITIES_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_AC_CAPABILITIES, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_Tss2_RC_Decode(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TSS2_RC arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tss2_RC_Decode",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_RC_Decode" "', argument " "1"" of type '" "TSS2_RC""'");
  } 
  arg1 = (TSS2_RC)(val1);
  result = (char *)Tss2_RC_Decode(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMI_YES_NO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMI_YES_NO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMI_YES_NO_PTR")) SWIG_fail;
  result = (TPMI_YES_NO *)new_TPMI_YES_NO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMI_YES_NO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMI_YES_NO arg1 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_YES_NO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMI_YES_NO_PTR",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "copy_TPMI_YES_NO_PTR" "', argument " "1"" of type '" "TPMI_YES_NO""'");
  } 
  arg1 = (TPMI_YES_NO)(val1);
  result = (TPMI_YES_NO *)copy_TPMI_YES_NO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMI_YES_NO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMI_YES_NO *arg1 = (TPMI_YES_NO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMI_YES_NO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMI_YES_NO_PTR" "', argument " "1"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg1 = (TPMI_YES_NO *)(argp1);
  delete_TPMI_YES_NO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMI_YES_NO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMI_YES_NO *arg1 = (TPMI_YES_NO *) 0 ;
  TPMI_YES_NO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMI_YES_NO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMI_YES_NO_PTR_assign" "', argument " "1"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg1 = (TPMI_YES_NO *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMI_YES_NO_PTR_assign" "', argument " "2"" of type '" "TPMI_YES_NO""'");
  } 
  arg2 = (TPMI_YES_NO)(val2);
  TPMI_YES_NO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMI_YES_NO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMI_YES_NO *arg1 = (TPMI_YES_NO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMI_YES_NO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMI_YES_NO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMI_YES_NO_PTR_value" "', argument " "1"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg1 = (TPMI_YES_NO *)(argp1);
  result = (TPMI_YES_NO)TPMI_YES_NO_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_EMPTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_EMPTY",SWIG_From_size_t((size_t)(sizeof(TPMS_EMPTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_EMPTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_EMPTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_EMPTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_EMPTY_ARRAY *)new_TPMS_EMPTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_EMPTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY_ARRAY *arg1 = (TPMS_EMPTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_EMPTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_EMPTY_ARRAY" "', argument " "1"" of type '" "TPMS_EMPTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_EMPTY_ARRAY *)(argp1);
  delete_TPMS_EMPTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY_ARRAY *arg1 = (TPMS_EMPTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_EMPTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_EMPTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_EMPTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_EMPTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_EMPTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_EMPTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_EMPTY *)memcpy((TPMS_EMPTY *)malloc(sizeof(TPMS_EMPTY)),&result,sizeof(TPMS_EMPTY)), SWIGTYPE_p_TPMS_EMPTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY_ARRAY *arg1 = (TPMS_EMPTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_EMPTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_EMPTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_EMPTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_EMPTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_EMPTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_EMPTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_EMPTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_EMPTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_EMPTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_EMPTY""'");
    } else {
      arg3 = *((TPMS_EMPTY *)(argp3));
    }
  }
  TPMS_EMPTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY_ARRAY *arg1 = (TPMS_EMPTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_EMPTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_ARRAY_cast" "', argument " "1"" of type '" "TPMS_EMPTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_EMPTY_ARRAY *)(argp1);
  result = (TPMS_EMPTY *)TPMS_EMPTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_EMPTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  result = (TPMS_EMPTY_ARRAY *)TPMS_EMPTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_EMPTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_EMPTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_EMPTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_EMPTY_PTR")) SWIG_fail;
  result = (TPMS_EMPTY *)new_TPMS_EMPTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_EMPTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_EMPTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_EMPTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_EMPTY_PTR" "', argument " "1"" of type '" "TPMS_EMPTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_EMPTY_PTR" "', argument " "1"" of type '" "TPMS_EMPTY""'");
    } else {
      arg1 = *((TPMS_EMPTY *)(argp1));
    }
  }
  result = (TPMS_EMPTY *)copy_TPMS_EMPTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_EMPTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_EMPTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_EMPTY_PTR" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  delete_TPMS_EMPTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  TPMS_EMPTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_EMPTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_PTR_assign" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_EMPTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_EMPTY_PTR_assign" "', argument " "2"" of type '" "TPMS_EMPTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_EMPTY_PTR_assign" "', argument " "2"" of type '" "TPMS_EMPTY""'");
    } else {
      arg2 = *((TPMS_EMPTY *)(argp2));
    }
  }
  TPMS_EMPTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_EMPTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_PTR_value" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  result = TPMS_EMPTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_EMPTY *)memcpy((TPMS_EMPTY *)malloc(sizeof(TPMS_EMPTY)),&result,sizeof(TPMS_EMPTY)), SWIGTYPE_p_TPMS_EMPTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_EMPTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_EMPTY_PTR_PTR")) SWIG_fail;
  result = (TPMS_EMPTY **)new_TPMS_EMPTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_EMPTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_EMPTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_EMPTY_PTR_PTR" "', argument " "1"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  result = (TPMS_EMPTY **)copy_TPMS_EMPTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_EMPTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY **arg1 = (TPMS_EMPTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_EMPTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_EMPTY_PTR_PTR" "', argument " "1"" of type '" "TPMS_EMPTY **""'"); 
  }
  arg1 = (TPMS_EMPTY **)(argp1);
  delete_TPMS_EMPTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY **arg1 = (TPMS_EMPTY **) 0 ;
  TPMS_EMPTY *arg2 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_EMPTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_EMPTY **""'"); 
  }
  arg1 = (TPMS_EMPTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_EMPTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg2 = (TPMS_EMPTY *)(argp2);
  TPMS_EMPTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_EMPTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY **arg1 = (TPMS_EMPTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_EMPTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_EMPTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_EMPTY_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_EMPTY **""'"); 
  }
  arg1 = (TPMS_EMPTY **)(argp1);
  result = (TPMS_EMPTY *)TPMS_EMPTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ALGORITHM_DESCRIPTION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ALGORITHM_DESCRIPTION",SWIG_From_size_t((size_t)(sizeof(TPMS_ALGORITHM_DESCRIPTION))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DESCRIPTION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ALGORITHM_DESCRIPTION_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ALGORITHM_DESCRIPTION_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)new_TPMS_ALGORITHM_DESCRIPTION_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION_ARRAY *arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)(argp1);
  delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION_ARRAY *arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ALGORITHM_DESCRIPTION *)memcpy((TPMS_ALGORITHM_DESCRIPTION *)malloc(sizeof(TPMS_ALGORITHM_DESCRIPTION)),&result,sizeof(TPMS_ALGORITHM_DESCRIPTION)), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION_ARRAY *arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ALGORITHM_DESCRIPTION arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ALGORITHM_DESCRIPTION""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ALGORITHM_DESCRIPTION""'");
    } else {
      arg3 = *((TPMS_ALGORITHM_DESCRIPTION *)(argp3));
    }
  }
  TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION_ARRAY *arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)(argp1);
  result = (TPMS_ALGORITHM_DESCRIPTION *)TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  result = (TPMS_ALGORITHM_DESCRIPTION_ARRAY *)TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ALGORITHM_DESCRIPTION_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DESCRIPTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALGORITHM_DESCRIPTION_PTR")) SWIG_fail;
  result = (TPMS_ALGORITHM_DESCRIPTION *)new_TPMS_ALGORITHM_DESCRIPTION_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ALGORITHM_DESCRIPTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ALGORITHM_DESCRIPTION_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ALGORITHM_DESCRIPTION_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ALGORITHM_DESCRIPTION_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION""'");
    } else {
      arg1 = *((TPMS_ALGORITHM_DESCRIPTION *)(argp1));
    }
  }
  result = (TPMS_ALGORITHM_DESCRIPTION *)copy_TPMS_ALGORITHM_DESCRIPTION_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DESCRIPTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DESCRIPTION_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DESCRIPTION_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  delete_TPMS_ALGORITHM_DESCRIPTION_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  TPMS_ALGORITHM_DESCRIPTION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DESCRIPTION_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_assign" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_assign" "', argument " "2"" of type '" "TPMS_ALGORITHM_DESCRIPTION""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_assign" "', argument " "2"" of type '" "TPMS_ALGORITHM_DESCRIPTION""'");
    } else {
      arg2 = *((TPMS_ALGORITHM_DESCRIPTION *)(argp2));
    }
  }
  TPMS_ALGORITHM_DESCRIPTION_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DESCRIPTION_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_value" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  result = TPMS_ALGORITHM_DESCRIPTION_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ALGORITHM_DESCRIPTION *)memcpy((TPMS_ALGORITHM_DESCRIPTION *)malloc(sizeof(TPMS_ALGORITHM_DESCRIPTION)),&result,sizeof(TPMS_ALGORITHM_DESCRIPTION)), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR")) SWIG_fail;
  result = (TPMS_ALGORITHM_DESCRIPTION **)new_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  result = (TPMS_ALGORITHM_DESCRIPTION **)copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION **arg1 = (TPMS_ALGORITHM_DESCRIPTION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION **""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION **)(argp1);
  delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION **arg1 = (TPMS_ALGORITHM_DESCRIPTION **) 0 ;
  TPMS_ALGORITHM_DESCRIPTION *arg2 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION **""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg2 = (TPMS_ALGORITHM_DESCRIPTION *)(argp2);
  TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION **arg1 = (TPMS_ALGORITHM_DESCRIPTION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DESCRIPTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION **""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION **)(argp1);
  result = (TPMS_ALGORITHM_DESCRIPTION *)TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_HA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_HA",SWIG_From_size_t((size_t)(sizeof(TPMU_HA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_HA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_HA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_HA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_HA_ARRAY *)new_TPMU_HA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_HA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA_ARRAY *arg1 = (TPMU_HA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_HA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_HA_ARRAY" "', argument " "1"" of type '" "TPMU_HA_ARRAY *""'"); 
  }
  arg1 = (TPMU_HA_ARRAY *)(argp1);
  delete_TPMU_HA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA_ARRAY *arg1 = (TPMU_HA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_HA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_HA_ARRAY *""'"); 
  }
  arg1 = (TPMU_HA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_HA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_HA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_HA *)memcpy((TPMU_HA *)malloc(sizeof(TPMU_HA)),&result,sizeof(TPMU_HA)), SWIGTYPE_p_TPMU_HA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA_ARRAY *arg1 = (TPMU_HA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_HA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_HA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_HA_ARRAY *""'"); 
  }
  arg1 = (TPMU_HA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_HA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_HA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_HA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_HA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_HA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_HA""'");
    } else {
      arg3 = *((TPMU_HA *)(argp3));
    }
  }
  TPMU_HA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA_ARRAY *arg1 = (TPMU_HA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_ARRAY_cast" "', argument " "1"" of type '" "TPMU_HA_ARRAY *""'"); 
  }
  arg1 = (TPMU_HA_ARRAY *)(argp1);
  result = (TPMU_HA *)TPMU_HA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (TPMU_HA_ARRAY *)TPMU_HA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_HA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_HA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_HA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_HA_PTR")) SWIG_fail;
  result = (TPMU_HA *)new_TPMU_HA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_HA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_HA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_HA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_HA_PTR" "', argument " "1"" of type '" "TPMU_HA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_HA_PTR" "', argument " "1"" of type '" "TPMU_HA""'");
    } else {
      arg1 = *((TPMU_HA *)(argp1));
    }
  }
  result = (TPMU_HA *)copy_TPMU_HA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_HA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_HA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_HA_PTR" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  delete_TPMU_HA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  TPMU_HA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_PTR_assign" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_HA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_PTR_assign" "', argument " "2"" of type '" "TPMU_HA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_HA_PTR_assign" "', argument " "2"" of type '" "TPMU_HA""'");
    } else {
      arg2 = *((TPMU_HA *)(argp2));
    }
  }
  TPMU_HA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_PTR_value" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = TPMU_HA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_HA *)memcpy((TPMU_HA *)malloc(sizeof(TPMU_HA)),&result,sizeof(TPMU_HA)), SWIGTYPE_p_TPMU_HA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_HA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_HA_PTR_PTR")) SWIG_fail;
  result = (TPMU_HA **)new_TPMU_HA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_HA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_HA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_HA_PTR_PTR" "', argument " "1"" of type '" "TPMU_HA *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  result = (TPMU_HA **)copy_TPMU_HA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_HA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA **arg1 = (TPMU_HA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_HA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_HA_PTR_PTR" "', argument " "1"" of type '" "TPMU_HA **""'"); 
  }
  arg1 = (TPMU_HA **)(argp1);
  delete_TPMU_HA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA **arg1 = (TPMU_HA **) 0 ;
  TPMU_HA *arg2 = (TPMU_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_HA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_HA **""'"); 
  }
  arg1 = (TPMU_HA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_HA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_HA *""'"); 
  }
  arg2 = (TPMU_HA *)(argp2);
  TPMU_HA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_HA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA **arg1 = (TPMU_HA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_HA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_HA_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_HA **""'"); 
  }
  arg1 = (TPMU_HA **)(argp1);
  result = (TPMU_HA *)TPMU_HA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_HA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_HA",SWIG_From_size_t((size_t)(sizeof(TPMT_HA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_HA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_HA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_HA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_HA_ARRAY *)new_TPMT_HA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_HA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA_ARRAY *arg1 = (TPMT_HA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_HA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_HA_ARRAY" "', argument " "1"" of type '" "TPMT_HA_ARRAY *""'"); 
  }
  arg1 = (TPMT_HA_ARRAY *)(argp1);
  delete_TPMT_HA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA_ARRAY *arg1 = (TPMT_HA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_HA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_HA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_HA_ARRAY *""'"); 
  }
  arg1 = (TPMT_HA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_HA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_HA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_HA *)memcpy((TPMT_HA *)malloc(sizeof(TPMT_HA)),&result,sizeof(TPMT_HA)), SWIGTYPE_p_TPMT_HA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA_ARRAY *arg1 = (TPMT_HA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_HA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_HA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_HA_ARRAY *""'"); 
  }
  arg1 = (TPMT_HA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_HA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_HA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_HA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_HA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_HA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_HA""'");
    } else {
      arg3 = *((TPMT_HA *)(argp3));
    }
  }
  TPMT_HA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA_ARRAY *arg1 = (TPMT_HA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_HA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_ARRAY_cast" "', argument " "1"" of type '" "TPMT_HA_ARRAY *""'"); 
  }
  arg1 = (TPMT_HA_ARRAY *)(argp1);
  result = (TPMT_HA *)TPMT_HA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_HA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  result = (TPMT_HA_ARRAY *)TPMT_HA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_HA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_HA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_HA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_HA_PTR")) SWIG_fail;
  result = (TPMT_HA *)new_TPMT_HA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_HA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_HA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_HA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_HA_PTR" "', argument " "1"" of type '" "TPMT_HA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_HA_PTR" "', argument " "1"" of type '" "TPMT_HA""'");
    } else {
      arg1 = *((TPMT_HA *)(argp1));
    }
  }
  result = (TPMT_HA *)copy_TPMT_HA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_HA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_HA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_HA_PTR" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  delete_TPMT_HA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  TPMT_HA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_HA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_PTR_assign" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_HA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_HA_PTR_assign" "', argument " "2"" of type '" "TPMT_HA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_HA_PTR_assign" "', argument " "2"" of type '" "TPMT_HA""'");
    } else {
      arg2 = *((TPMT_HA *)(argp2));
    }
  }
  TPMT_HA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_HA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_PTR_value" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  result = TPMT_HA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_HA *)memcpy((TPMT_HA *)malloc(sizeof(TPMT_HA)),&result,sizeof(TPMT_HA)), SWIGTYPE_p_TPMT_HA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_HA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_HA_PTR_PTR")) SWIG_fail;
  result = (TPMT_HA **)new_TPMT_HA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_HA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_HA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_HA_PTR_PTR" "', argument " "1"" of type '" "TPMT_HA *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  result = (TPMT_HA **)copy_TPMT_HA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_HA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA **arg1 = (TPMT_HA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_HA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_HA_PTR_PTR" "', argument " "1"" of type '" "TPMT_HA **""'"); 
  }
  arg1 = (TPMT_HA **)(argp1);
  delete_TPMT_HA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA **arg1 = (TPMT_HA **) 0 ;
  TPMT_HA *arg2 = (TPMT_HA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_HA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_HA **""'"); 
  }
  arg1 = (TPMT_HA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_HA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_HA *""'"); 
  }
  arg2 = (TPMT_HA *)(argp2);
  TPMT_HA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_HA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA **arg1 = (TPMT_HA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_HA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_HA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_HA_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_HA **""'"); 
  }
  arg1 = (TPMT_HA **)(argp1);
  result = (TPMT_HA *)TPMT_HA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_HA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_DIGEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_DIGEST",SWIG_From_size_t((size_t)(sizeof(TPM2B_DIGEST))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_DIGEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_DIGEST_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_DIGEST_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_DIGEST_ARRAY *)new_TPM2B_DIGEST_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DIGEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST_ARRAY *arg1 = (TPM2B_DIGEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DIGEST_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DIGEST_ARRAY" "', argument " "1"" of type '" "TPM2B_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DIGEST_ARRAY *)(argp1);
  delete_TPM2B_DIGEST_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST_ARRAY *arg1 = (TPM2B_DIGEST_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_DIGEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DIGEST_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DIGEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_DIGEST_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_DIGEST_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_DIGEST *)memcpy((TPM2B_DIGEST *)malloc(sizeof(TPM2B_DIGEST)),&result,sizeof(TPM2B_DIGEST)), SWIGTYPE_p_TPM2B_DIGEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST_ARRAY *arg1 = (TPM2B_DIGEST_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_DIGEST arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_DIGEST_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DIGEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_DIGEST_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_DIGEST,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_DIGEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_DIGEST""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_DIGEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_DIGEST""'");
    } else {
      arg3 = *((TPM2B_DIGEST *)(argp3));
    }
  }
  TPM2B_DIGEST_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST_ARRAY *arg1 = (TPM2B_DIGEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DIGEST_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DIGEST_ARRAY *)(argp1);
  result = (TPM2B_DIGEST *)TPM2B_DIGEST_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DIGEST_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  result = (TPM2B_DIGEST_ARRAY *)TPM2B_DIGEST_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_DIGEST_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_DIGEST_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_DIGEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_DIGEST_PTR")) SWIG_fail;
  result = (TPM2B_DIGEST *)new_TPM2B_DIGEST_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_DIGEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_DIGEST_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_DIGEST,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_DIGEST_PTR" "', argument " "1"" of type '" "TPM2B_DIGEST""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_DIGEST_PTR" "', argument " "1"" of type '" "TPM2B_DIGEST""'");
    } else {
      arg1 = *((TPM2B_DIGEST *)(argp1));
    }
  }
  result = (TPM2B_DIGEST *)copy_TPM2B_DIGEST_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DIGEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DIGEST_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DIGEST_PTR" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  delete_TPM2B_DIGEST_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  TPM2B_DIGEST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DIGEST_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_PTR_assign" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_DIGEST,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_DIGEST_PTR_assign" "', argument " "2"" of type '" "TPM2B_DIGEST""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_DIGEST_PTR_assign" "', argument " "2"" of type '" "TPM2B_DIGEST""'");
    } else {
      arg2 = *((TPM2B_DIGEST *)(argp2));
    }
  }
  TPM2B_DIGEST_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DIGEST_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_PTR_value" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  result = TPM2B_DIGEST_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_DIGEST *)memcpy((TPM2B_DIGEST *)malloc(sizeof(TPM2B_DIGEST)),&result,sizeof(TPM2B_DIGEST)), SWIGTYPE_p_TPM2B_DIGEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_DIGEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_DIGEST_PTR_PTR")) SWIG_fail;
  result = (TPM2B_DIGEST **)new_TPM2B_DIGEST_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_DIGEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_DIGEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_DIGEST_PTR_PTR" "', argument " "1"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  result = (TPM2B_DIGEST **)copy_TPM2B_DIGEST_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DIGEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST **arg1 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DIGEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DIGEST_PTR_PTR" "', argument " "1"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg1 = (TPM2B_DIGEST **)(argp1);
  delete_TPM2B_DIGEST_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST **arg1 = (TPM2B_DIGEST **) 0 ;
  TPM2B_DIGEST *arg2 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DIGEST_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg1 = (TPM2B_DIGEST **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_DIGEST_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg2 = (TPM2B_DIGEST *)(argp2);
  TPM2B_DIGEST_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DIGEST_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST **arg1 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DIGEST_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DIGEST_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg1 = (TPM2B_DIGEST **)(argp1);
  result = (TPM2B_DIGEST *)TPM2B_DIGEST_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_DATA",SWIG_From_size_t((size_t)(sizeof(TPM2B_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_DATA_ARRAY *)new_TPM2B_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA_ARRAY *arg1 = (TPM2B_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DATA_ARRAY" "', argument " "1"" of type '" "TPM2B_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DATA_ARRAY *)(argp1);
  delete_TPM2B_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA_ARRAY *arg1 = (TPM2B_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_DATA *)memcpy((TPM2B_DATA *)malloc(sizeof(TPM2B_DATA)),&result,sizeof(TPM2B_DATA)), SWIGTYPE_p_TPM2B_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA_ARRAY *arg1 = (TPM2B_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_DATA""'");
    } else {
      arg3 = *((TPM2B_DATA *)(argp3));
    }
  }
  TPM2B_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA_ARRAY *arg1 = (TPM2B_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_DATA_ARRAY *)(argp1);
  result = (TPM2B_DATA *)TPM2B_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  result = (TPM2B_DATA_ARRAY *)TPM2B_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_DATA_PTR")) SWIG_fail;
  result = (TPM2B_DATA *)new_TPM2B_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_DATA_PTR" "', argument " "1"" of type '" "TPM2B_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_DATA_PTR" "', argument " "1"" of type '" "TPM2B_DATA""'");
    } else {
      arg1 = *((TPM2B_DATA *)(argp1));
    }
  }
  result = (TPM2B_DATA *)copy_TPM2B_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DATA_PTR" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  delete_TPM2B_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  TPM2B_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_PTR_assign" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_DATA""'");
    } else {
      arg2 = *((TPM2B_DATA *)(argp2));
    }
  }
  TPM2B_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_PTR_value" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  result = TPM2B_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_DATA *)memcpy((TPM2B_DATA *)malloc(sizeof(TPM2B_DATA)),&result,sizeof(TPM2B_DATA)), SWIGTYPE_p_TPM2B_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_DATA_PTR_PTR")) SWIG_fail;
  result = (TPM2B_DATA **)new_TPM2B_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_DATA *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  result = (TPM2B_DATA **)copy_TPM2B_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA **arg1 = (TPM2B_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_DATA **""'"); 
  }
  arg1 = (TPM2B_DATA **)(argp1);
  delete_TPM2B_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA **arg1 = (TPM2B_DATA **) 0 ;
  TPM2B_DATA *arg2 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_DATA **""'"); 
  }
  arg1 = (TPM2B_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_DATA *""'"); 
  }
  arg2 = (TPM2B_DATA *)(argp2);
  TPM2B_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA **arg1 = (TPM2B_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_DATA **""'"); 
  }
  arg1 = (TPM2B_DATA **)(argp1);
  result = (TPM2B_DATA *)TPM2B_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_EVENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_EVENT",SWIG_From_size_t((size_t)(sizeof(TPM2B_EVENT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_EVENT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_EVENT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_EVENT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_EVENT_ARRAY *)new_TPM2B_EVENT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_EVENT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT_ARRAY *arg1 = (TPM2B_EVENT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_EVENT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_EVENT_ARRAY" "', argument " "1"" of type '" "TPM2B_EVENT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_EVENT_ARRAY *)(argp1);
  delete_TPM2B_EVENT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT_ARRAY *arg1 = (TPM2B_EVENT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_EVENT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_EVENT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_EVENT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_EVENT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_EVENT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_EVENT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_EVENT *)memcpy((TPM2B_EVENT *)malloc(sizeof(TPM2B_EVENT)),&result,sizeof(TPM2B_EVENT)), SWIGTYPE_p_TPM2B_EVENT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT_ARRAY *arg1 = (TPM2B_EVENT_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_EVENT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_EVENT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_EVENT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_EVENT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_EVENT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_EVENT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_EVENT_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_EVENT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_EVENT_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_EVENT""'");
    } else {
      arg3 = *((TPM2B_EVENT *)(argp3));
    }
  }
  TPM2B_EVENT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT_ARRAY *arg1 = (TPM2B_EVENT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_EVENT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_EVENT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_EVENT_ARRAY *)(argp1);
  result = (TPM2B_EVENT *)TPM2B_EVENT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_EVENT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  result = (TPM2B_EVENT_ARRAY *)TPM2B_EVENT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_EVENT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_EVENT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_EVENT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_EVENT_PTR")) SWIG_fail;
  result = (TPM2B_EVENT *)new_TPM2B_EVENT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_EVENT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_EVENT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_EVENT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_EVENT_PTR" "', argument " "1"" of type '" "TPM2B_EVENT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_EVENT_PTR" "', argument " "1"" of type '" "TPM2B_EVENT""'");
    } else {
      arg1 = *((TPM2B_EVENT *)(argp1));
    }
  }
  result = (TPM2B_EVENT *)copy_TPM2B_EVENT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_EVENT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_EVENT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_EVENT_PTR" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  delete_TPM2B_EVENT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  TPM2B_EVENT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_EVENT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_PTR_assign" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_EVENT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_EVENT_PTR_assign" "', argument " "2"" of type '" "TPM2B_EVENT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_EVENT_PTR_assign" "', argument " "2"" of type '" "TPM2B_EVENT""'");
    } else {
      arg2 = *((TPM2B_EVENT *)(argp2));
    }
  }
  TPM2B_EVENT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_EVENT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_PTR_value" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  result = TPM2B_EVENT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_EVENT *)memcpy((TPM2B_EVENT *)malloc(sizeof(TPM2B_EVENT)),&result,sizeof(TPM2B_EVENT)), SWIGTYPE_p_TPM2B_EVENT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_EVENT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_EVENT_PTR_PTR")) SWIG_fail;
  result = (TPM2B_EVENT **)new_TPM2B_EVENT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_EVENT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_EVENT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_EVENT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_EVENT_PTR_PTR" "', argument " "1"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  result = (TPM2B_EVENT **)copy_TPM2B_EVENT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_EVENT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_EVENT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT **arg1 = (TPM2B_EVENT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_EVENT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_EVENT_PTR_PTR" "', argument " "1"" of type '" "TPM2B_EVENT **""'"); 
  }
  arg1 = (TPM2B_EVENT **)(argp1);
  delete_TPM2B_EVENT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT **arg1 = (TPM2B_EVENT **) 0 ;
  TPM2B_EVENT *arg2 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_EVENT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_EVENT **""'"); 
  }
  arg1 = (TPM2B_EVENT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_EVENT_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg2 = (TPM2B_EVENT *)(argp2);
  TPM2B_EVENT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_EVENT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT **arg1 = (TPM2B_EVENT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_EVENT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_EVENT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_EVENT_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_EVENT **""'"); 
  }
  arg1 = (TPM2B_EVENT **)(argp1);
  result = (TPM2B_EVENT *)TPM2B_EVENT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_MAX_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_MAX_BUFFER",SWIG_From_size_t((size_t)(sizeof(TPM2B_MAX_BUFFER))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_BUFFER_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_MAX_BUFFER_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_MAX_BUFFER_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_MAX_BUFFER_ARRAY *)new_TPM2B_MAX_BUFFER_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_BUFFER_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER_ARRAY *arg1 = (TPM2B_MAX_BUFFER_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_BUFFER_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_BUFFER_ARRAY" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER_ARRAY *)(argp1);
  delete_TPM2B_MAX_BUFFER_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER_ARRAY *arg1 = (TPM2B_MAX_BUFFER_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_MAX_BUFFER result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_BUFFER_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_MAX_BUFFER_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_MAX_BUFFER_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_MAX_BUFFER *)memcpy((TPM2B_MAX_BUFFER *)malloc(sizeof(TPM2B_MAX_BUFFER)),&result,sizeof(TPM2B_MAX_BUFFER)), SWIGTYPE_p_TPM2B_MAX_BUFFER, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER_ARRAY *arg1 = (TPM2B_MAX_BUFFER_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_MAX_BUFFER arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_MAX_BUFFER_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_MAX_BUFFER_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_MAX_BUFFER,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_MAX_BUFFER_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_MAX_BUFFER""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_MAX_BUFFER_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_MAX_BUFFER""'");
    } else {
      arg3 = *((TPM2B_MAX_BUFFER *)(argp3));
    }
  }
  TPM2B_MAX_BUFFER_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER_ARRAY *arg1 = (TPM2B_MAX_BUFFER_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_BUFFER_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER_ARRAY *)(argp1);
  result = (TPM2B_MAX_BUFFER *)TPM2B_MAX_BUFFER_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_BUFFER_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  result = (TPM2B_MAX_BUFFER_ARRAY *)TPM2B_MAX_BUFFER_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_MAX_BUFFER_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_MAX_BUFFER_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_BUFFER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_MAX_BUFFER_PTR")) SWIG_fail;
  result = (TPM2B_MAX_BUFFER *)new_TPM2B_MAX_BUFFER_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_MAX_BUFFER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_MAX_BUFFER_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_MAX_BUFFER,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_MAX_BUFFER_PTR" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_MAX_BUFFER_PTR" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER""'");
    } else {
      arg1 = *((TPM2B_MAX_BUFFER *)(argp1));
    }
  }
  result = (TPM2B_MAX_BUFFER *)copy_TPM2B_MAX_BUFFER_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_BUFFER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_BUFFER_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_BUFFER_PTR" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  delete_TPM2B_MAX_BUFFER_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  TPM2B_MAX_BUFFER arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_BUFFER_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_PTR_assign" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_MAX_BUFFER,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_MAX_BUFFER_PTR_assign" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_MAX_BUFFER_PTR_assign" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER""'");
    } else {
      arg2 = *((TPM2B_MAX_BUFFER *)(argp2));
    }
  }
  TPM2B_MAX_BUFFER_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_BUFFER_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_PTR_value" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  result = TPM2B_MAX_BUFFER_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_MAX_BUFFER *)memcpy((TPM2B_MAX_BUFFER *)malloc(sizeof(TPM2B_MAX_BUFFER)),&result,sizeof(TPM2B_MAX_BUFFER)), SWIGTYPE_p_TPM2B_MAX_BUFFER, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_BUFFER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_MAX_BUFFER_PTR_PTR")) SWIG_fail;
  result = (TPM2B_MAX_BUFFER **)new_TPM2B_MAX_BUFFER_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_MAX_BUFFER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_MAX_BUFFER_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_MAX_BUFFER_PTR_PTR" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  result = (TPM2B_MAX_BUFFER **)copy_TPM2B_MAX_BUFFER_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_BUFFER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER **arg1 = (TPM2B_MAX_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_BUFFER_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_BUFFER_PTR_PTR" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER **)(argp1);
  delete_TPM2B_MAX_BUFFER_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER **arg1 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2B_MAX_BUFFER *arg2 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_BUFFER_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_MAX_BUFFER_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg2 = (TPM2B_MAX_BUFFER *)(argp2);
  TPM2B_MAX_BUFFER_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_BUFFER_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER **arg1 = (TPM2B_MAX_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_BUFFER_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_BUFFER_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER **)(argp1);
  result = (TPM2B_MAX_BUFFER *)TPM2B_MAX_BUFFER_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_MAX_NV_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_MAX_NV_BUFFER",SWIG_From_size_t((size_t)(sizeof(TPM2B_MAX_NV_BUFFER))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_NV_BUFFER_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_MAX_NV_BUFFER_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_MAX_NV_BUFFER_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_MAX_NV_BUFFER_ARRAY *)new_TPM2B_MAX_NV_BUFFER_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_NV_BUFFER_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER_ARRAY *arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_NV_BUFFER_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_NV_BUFFER_ARRAY" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *)(argp1);
  delete_TPM2B_MAX_NV_BUFFER_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER_ARRAY *arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_MAX_NV_BUFFER result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_NV_BUFFER_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_MAX_NV_BUFFER_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_MAX_NV_BUFFER *)memcpy((TPM2B_MAX_NV_BUFFER *)malloc(sizeof(TPM2B_MAX_NV_BUFFER)),&result,sizeof(TPM2B_MAX_NV_BUFFER)), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER_ARRAY *arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_MAX_NV_BUFFER arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_MAX_NV_BUFFER_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_MAX_NV_BUFFER,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_MAX_NV_BUFFER""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_MAX_NV_BUFFER""'");
    } else {
      arg3 = *((TPM2B_MAX_NV_BUFFER *)(argp3));
    }
  }
  TPM2B_MAX_NV_BUFFER_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER_ARRAY *arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_NV_BUFFER_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER_ARRAY *)(argp1);
  result = (TPM2B_MAX_NV_BUFFER *)TPM2B_MAX_NV_BUFFER_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_NV_BUFFER_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  result = (TPM2B_MAX_NV_BUFFER_ARRAY *)TPM2B_MAX_NV_BUFFER_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_MAX_NV_BUFFER_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_MAX_NV_BUFFER_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_NV_BUFFER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_MAX_NV_BUFFER_PTR")) SWIG_fail;
  result = (TPM2B_MAX_NV_BUFFER *)new_TPM2B_MAX_NV_BUFFER_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_MAX_NV_BUFFER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_MAX_NV_BUFFER_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_MAX_NV_BUFFER,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_MAX_NV_BUFFER_PTR" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_MAX_NV_BUFFER_PTR" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER""'");
    } else {
      arg1 = *((TPM2B_MAX_NV_BUFFER *)(argp1));
    }
  }
  result = (TPM2B_MAX_NV_BUFFER *)copy_TPM2B_MAX_NV_BUFFER_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_NV_BUFFER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_NV_BUFFER_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_NV_BUFFER_PTR" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  delete_TPM2B_MAX_NV_BUFFER_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  TPM2B_MAX_NV_BUFFER arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_NV_BUFFER_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_PTR_assign" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_MAX_NV_BUFFER,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_MAX_NV_BUFFER_PTR_assign" "', argument " "2"" of type '" "TPM2B_MAX_NV_BUFFER""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_MAX_NV_BUFFER_PTR_assign" "', argument " "2"" of type '" "TPM2B_MAX_NV_BUFFER""'");
    } else {
      arg2 = *((TPM2B_MAX_NV_BUFFER *)(argp2));
    }
  }
  TPM2B_MAX_NV_BUFFER_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_NV_BUFFER_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_PTR_value" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  result = TPM2B_MAX_NV_BUFFER_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_MAX_NV_BUFFER *)memcpy((TPM2B_MAX_NV_BUFFER *)malloc(sizeof(TPM2B_MAX_NV_BUFFER)),&result,sizeof(TPM2B_MAX_NV_BUFFER)), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_MAX_NV_BUFFER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_MAX_NV_BUFFER_PTR_PTR")) SWIG_fail;
  result = (TPM2B_MAX_NV_BUFFER **)new_TPM2B_MAX_NV_BUFFER_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_MAX_NV_BUFFER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_MAX_NV_BUFFER_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_MAX_NV_BUFFER_PTR_PTR" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  result = (TPM2B_MAX_NV_BUFFER **)copy_TPM2B_MAX_NV_BUFFER_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_MAX_NV_BUFFER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER **arg1 = (TPM2B_MAX_NV_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_MAX_NV_BUFFER_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_MAX_NV_BUFFER_PTR_PTR" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER **""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER **)(argp1);
  delete_TPM2B_MAX_NV_BUFFER_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER **arg1 = (TPM2B_MAX_NV_BUFFER **) 0 ;
  TPM2B_MAX_NV_BUFFER *arg2 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_MAX_NV_BUFFER_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER **""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_MAX_NV_BUFFER_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg2 = (TPM2B_MAX_NV_BUFFER *)(argp2);
  TPM2B_MAX_NV_BUFFER_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_MAX_NV_BUFFER_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER **arg1 = (TPM2B_MAX_NV_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_MAX_NV_BUFFER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_MAX_NV_BUFFER_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_MAX_NV_BUFFER_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER **""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER **)(argp1);
  result = (TPM2B_MAX_NV_BUFFER *)TPM2B_MAX_NV_BUFFER_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_IV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_IV",SWIG_From_size_t((size_t)(sizeof(TPM2B_IV))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_IV_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_IV_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_IV_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_IV_ARRAY *)new_TPM2B_IV_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_IV_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV_ARRAY *arg1 = (TPM2B_IV_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_IV_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_IV_ARRAY" "', argument " "1"" of type '" "TPM2B_IV_ARRAY *""'"); 
  }
  arg1 = (TPM2B_IV_ARRAY *)(argp1);
  delete_TPM2B_IV_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV_ARRAY *arg1 = (TPM2B_IV_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_IV result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_IV_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_IV_ARRAY *""'"); 
  }
  arg1 = (TPM2B_IV_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_IV_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_IV_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_IV *)memcpy((TPM2B_IV *)malloc(sizeof(TPM2B_IV)),&result,sizeof(TPM2B_IV)), SWIGTYPE_p_TPM2B_IV, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV_ARRAY *arg1 = (TPM2B_IV_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_IV arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_IV_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_IV_ARRAY *""'"); 
  }
  arg1 = (TPM2B_IV_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_IV_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_IV,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_IV_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_IV""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_IV_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_IV""'");
    } else {
      arg3 = *((TPM2B_IV *)(argp3));
    }
  }
  TPM2B_IV_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV_ARRAY *arg1 = (TPM2B_IV_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_IV_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_IV_ARRAY *""'"); 
  }
  arg1 = (TPM2B_IV_ARRAY *)(argp1);
  result = (TPM2B_IV *)TPM2B_IV_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_IV_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  result = (TPM2B_IV_ARRAY *)TPM2B_IV_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_IV_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_IV_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_IV_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_IV_PTR")) SWIG_fail;
  result = (TPM2B_IV *)new_TPM2B_IV_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_IV_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_IV_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_IV,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_IV_PTR" "', argument " "1"" of type '" "TPM2B_IV""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_IV_PTR" "', argument " "1"" of type '" "TPM2B_IV""'");
    } else {
      arg1 = *((TPM2B_IV *)(argp1));
    }
  }
  result = (TPM2B_IV *)copy_TPM2B_IV_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_IV_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_IV_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_IV_PTR" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  delete_TPM2B_IV_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  TPM2B_IV arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_IV_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_PTR_assign" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_IV,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_IV_PTR_assign" "', argument " "2"" of type '" "TPM2B_IV""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_IV_PTR_assign" "', argument " "2"" of type '" "TPM2B_IV""'");
    } else {
      arg2 = *((TPM2B_IV *)(argp2));
    }
  }
  TPM2B_IV_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_IV_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_PTR_value" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  result = TPM2B_IV_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_IV *)memcpy((TPM2B_IV *)malloc(sizeof(TPM2B_IV)),&result,sizeof(TPM2B_IV)), SWIGTYPE_p_TPM2B_IV, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_IV_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_IV_PTR_PTR")) SWIG_fail;
  result = (TPM2B_IV **)new_TPM2B_IV_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_IV_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_IV_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_IV_PTR_PTR" "', argument " "1"" of type '" "TPM2B_IV *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  result = (TPM2B_IV **)copy_TPM2B_IV_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_IV_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV **arg1 = (TPM2B_IV **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_IV_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_IV_PTR_PTR" "', argument " "1"" of type '" "TPM2B_IV **""'"); 
  }
  arg1 = (TPM2B_IV **)(argp1);
  delete_TPM2B_IV_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV **arg1 = (TPM2B_IV **) 0 ;
  TPM2B_IV *arg2 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_IV_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_IV **""'"); 
  }
  arg1 = (TPM2B_IV **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_IV_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_IV *""'"); 
  }
  arg2 = (TPM2B_IV *)(argp2);
  TPM2B_IV_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_IV_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV **arg1 = (TPM2B_IV **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_IV *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_IV_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_IV_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_IV **""'"); 
  }
  arg1 = (TPM2B_IV **)(argp1);
  result = (TPM2B_IV *)TPM2B_IV_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_NAME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_NAME",SWIG_From_size_t((size_t)(sizeof(TPMU_NAME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_NAME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_NAME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_NAME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_NAME_ARRAY *)new_TPMU_NAME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_NAME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME_ARRAY *arg1 = (TPMU_NAME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_NAME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_NAME_ARRAY" "', argument " "1"" of type '" "TPMU_NAME_ARRAY *""'"); 
  }
  arg1 = (TPMU_NAME_ARRAY *)(argp1);
  delete_TPMU_NAME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME_ARRAY *arg1 = (TPMU_NAME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_NAME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_NAME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_NAME_ARRAY *""'"); 
  }
  arg1 = (TPMU_NAME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_NAME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_NAME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_NAME *)memcpy((TPMU_NAME *)malloc(sizeof(TPMU_NAME)),&result,sizeof(TPMU_NAME)), SWIGTYPE_p_TPMU_NAME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME_ARRAY *arg1 = (TPMU_NAME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_NAME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_NAME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_NAME_ARRAY *""'"); 
  }
  arg1 = (TPMU_NAME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_NAME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_NAME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_NAME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_NAME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_NAME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_NAME""'");
    } else {
      arg3 = *((TPMU_NAME *)(argp3));
    }
  }
  TPMU_NAME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME_ARRAY *arg1 = (TPMU_NAME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_NAME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_ARRAY_cast" "', argument " "1"" of type '" "TPMU_NAME_ARRAY *""'"); 
  }
  arg1 = (TPMU_NAME_ARRAY *)(argp1);
  result = (TPMU_NAME *)TPMU_NAME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_NAME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  result = (TPMU_NAME_ARRAY *)TPMU_NAME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_NAME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_NAME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_NAME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_NAME_PTR")) SWIG_fail;
  result = (TPMU_NAME *)new_TPMU_NAME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_NAME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_NAME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_NAME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_NAME_PTR" "', argument " "1"" of type '" "TPMU_NAME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_NAME_PTR" "', argument " "1"" of type '" "TPMU_NAME""'");
    } else {
      arg1 = *((TPMU_NAME *)(argp1));
    }
  }
  result = (TPMU_NAME *)copy_TPMU_NAME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_NAME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_NAME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_NAME_PTR" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  delete_TPMU_NAME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  TPMU_NAME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_NAME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_PTR_assign" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_NAME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_NAME_PTR_assign" "', argument " "2"" of type '" "TPMU_NAME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_NAME_PTR_assign" "', argument " "2"" of type '" "TPMU_NAME""'");
    } else {
      arg2 = *((TPMU_NAME *)(argp2));
    }
  }
  TPMU_NAME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_NAME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_PTR_value" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  result = TPMU_NAME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_NAME *)memcpy((TPMU_NAME *)malloc(sizeof(TPMU_NAME)),&result,sizeof(TPMU_NAME)), SWIGTYPE_p_TPMU_NAME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_NAME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_NAME_PTR_PTR")) SWIG_fail;
  result = (TPMU_NAME **)new_TPMU_NAME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_NAME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_NAME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_NAME_PTR_PTR" "', argument " "1"" of type '" "TPMU_NAME *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  result = (TPMU_NAME **)copy_TPMU_NAME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_NAME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME **arg1 = (TPMU_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_NAME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_NAME_PTR_PTR" "', argument " "1"" of type '" "TPMU_NAME **""'"); 
  }
  arg1 = (TPMU_NAME **)(argp1);
  delete_TPMU_NAME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME **arg1 = (TPMU_NAME **) 0 ;
  TPMU_NAME *arg2 = (TPMU_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_NAME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_NAME **""'"); 
  }
  arg1 = (TPMU_NAME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_NAME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_NAME *""'"); 
  }
  arg2 = (TPMU_NAME *)(argp2);
  TPMU_NAME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_NAME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME **arg1 = (TPMU_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_NAME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_NAME_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_NAME **""'"); 
  }
  arg1 = (TPMU_NAME **)(argp1);
  result = (TPMU_NAME *)TPMU_NAME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_NAME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_NAME",SWIG_From_size_t((size_t)(sizeof(TPM2B_NAME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_NAME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_NAME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_NAME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_NAME_ARRAY *)new_TPM2B_NAME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NAME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME_ARRAY *arg1 = (TPM2B_NAME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NAME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NAME_ARRAY" "', argument " "1"" of type '" "TPM2B_NAME_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NAME_ARRAY *)(argp1);
  delete_TPM2B_NAME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME_ARRAY *arg1 = (TPM2B_NAME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_NAME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NAME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_NAME_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NAME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_NAME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_NAME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_NAME *)memcpy((TPM2B_NAME *)malloc(sizeof(TPM2B_NAME)),&result,sizeof(TPM2B_NAME)), SWIGTYPE_p_TPM2B_NAME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME_ARRAY *arg1 = (TPM2B_NAME_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_NAME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_NAME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_NAME_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NAME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_NAME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_NAME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_NAME_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_NAME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_NAME_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_NAME""'");
    } else {
      arg3 = *((TPM2B_NAME *)(argp3));
    }
  }
  TPM2B_NAME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME_ARRAY *arg1 = (TPM2B_NAME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NAME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_NAME_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NAME_ARRAY *)(argp1);
  result = (TPM2B_NAME *)TPM2B_NAME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NAME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  result = (TPM2B_NAME_ARRAY *)TPM2B_NAME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_NAME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_NAME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_NAME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_NAME_PTR")) SWIG_fail;
  result = (TPM2B_NAME *)new_TPM2B_NAME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_NAME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_NAME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_NAME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_NAME_PTR" "', argument " "1"" of type '" "TPM2B_NAME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_NAME_PTR" "', argument " "1"" of type '" "TPM2B_NAME""'");
    } else {
      arg1 = *((TPM2B_NAME *)(argp1));
    }
  }
  result = (TPM2B_NAME *)copy_TPM2B_NAME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NAME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NAME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NAME_PTR" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  delete_TPM2B_NAME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  TPM2B_NAME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NAME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_PTR_assign" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_NAME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_NAME_PTR_assign" "', argument " "2"" of type '" "TPM2B_NAME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_NAME_PTR_assign" "', argument " "2"" of type '" "TPM2B_NAME""'");
    } else {
      arg2 = *((TPM2B_NAME *)(argp2));
    }
  }
  TPM2B_NAME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NAME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_PTR_value" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  result = TPM2B_NAME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_NAME *)memcpy((TPM2B_NAME *)malloc(sizeof(TPM2B_NAME)),&result,sizeof(TPM2B_NAME)), SWIGTYPE_p_TPM2B_NAME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_NAME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_NAME_PTR_PTR")) SWIG_fail;
  result = (TPM2B_NAME **)new_TPM2B_NAME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_NAME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_NAME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_NAME_PTR_PTR" "', argument " "1"" of type '" "TPM2B_NAME *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  result = (TPM2B_NAME **)copy_TPM2B_NAME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NAME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME **arg1 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NAME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NAME_PTR_PTR" "', argument " "1"" of type '" "TPM2B_NAME **""'"); 
  }
  arg1 = (TPM2B_NAME **)(argp1);
  delete_TPM2B_NAME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME **arg1 = (TPM2B_NAME **) 0 ;
  TPM2B_NAME *arg2 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NAME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_NAME **""'"); 
  }
  arg1 = (TPM2B_NAME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_NAME_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_NAME *""'"); 
  }
  arg2 = (TPM2B_NAME *)(argp2);
  TPM2B_NAME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NAME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME **arg1 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NAME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NAME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NAME_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_NAME **""'"); 
  }
  arg1 = (TPM2B_NAME **)(argp1);
  result = (TPM2B_NAME *)TPM2B_NAME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_PCR_SELECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_PCR_SELECT",SWIG_From_size_t((size_t)(sizeof(TPMS_PCR_SELECT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_PCR_SELECT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_PCR_SELECT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_PCR_SELECT_ARRAY *)new_TPMS_PCR_SELECT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT_ARRAY *arg1 = (TPMS_PCR_SELECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECT_ARRAY" "', argument " "1"" of type '" "TPMS_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT_ARRAY *)(argp1);
  delete_TPMS_PCR_SELECT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT_ARRAY *arg1 = (TPMS_PCR_SELECT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_PCR_SELECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_PCR_SELECT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_PCR_SELECT *)memcpy((TPMS_PCR_SELECT *)malloc(sizeof(TPMS_PCR_SELECT)),&result,sizeof(TPMS_PCR_SELECT)), SWIGTYPE_p_TPMS_PCR_SELECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT_ARRAY *arg1 = (TPMS_PCR_SELECT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_PCR_SELECT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_PCR_SELECT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_PCR_SELECT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_PCR_SELECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_PCR_SELECT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_PCR_SELECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_PCR_SELECT""'");
    } else {
      arg3 = *((TPMS_PCR_SELECT *)(argp3));
    }
  }
  TPMS_PCR_SELECT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT_ARRAY *arg1 = (TPMS_PCR_SELECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_ARRAY_cast" "', argument " "1"" of type '" "TPMS_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT_ARRAY *)(argp1);
  result = (TPMS_PCR_SELECT *)TPMS_PCR_SELECT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  result = (TPMS_PCR_SELECT_ARRAY *)TPMS_PCR_SELECT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_PCR_SELECT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_PCR_SELECT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_PCR_SELECT_PTR")) SWIG_fail;
  result = (TPMS_PCR_SELECT *)new_TPMS_PCR_SELECT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_PCR_SELECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_PCR_SELECT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_PCR_SELECT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_PCR_SELECT_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_PCR_SELECT_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECT""'");
    } else {
      arg1 = *((TPMS_PCR_SELECT *)(argp1));
    }
  }
  result = (TPMS_PCR_SELECT *)copy_TPMS_PCR_SELECT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECT_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  delete_TPMS_PCR_SELECT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  TPMS_PCR_SELECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_PTR_assign" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_PCR_SELECT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_PCR_SELECT_PTR_assign" "', argument " "2"" of type '" "TPMS_PCR_SELECT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_PCR_SELECT_PTR_assign" "', argument " "2"" of type '" "TPMS_PCR_SELECT""'");
    } else {
      arg2 = *((TPMS_PCR_SELECT *)(argp2));
    }
  }
  TPMS_PCR_SELECT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_PTR_value" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  result = TPMS_PCR_SELECT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_PCR_SELECT *)memcpy((TPMS_PCR_SELECT *)malloc(sizeof(TPMS_PCR_SELECT)),&result,sizeof(TPMS_PCR_SELECT)), SWIGTYPE_p_TPMS_PCR_SELECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_PCR_SELECT_PTR_PTR")) SWIG_fail;
  result = (TPMS_PCR_SELECT **)new_TPMS_PCR_SELECT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_PCR_SELECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_PCR_SELECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_PCR_SELECT_PTR_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  result = (TPMS_PCR_SELECT **)copy_TPMS_PCR_SELECT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT **arg1 = (TPMS_PCR_SELECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECT_PTR_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECT **""'"); 
  }
  arg1 = (TPMS_PCR_SELECT **)(argp1);
  delete_TPMS_PCR_SELECT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT **arg1 = (TPMS_PCR_SELECT **) 0 ;
  TPMS_PCR_SELECT *arg2 = (TPMS_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_PCR_SELECT **""'"); 
  }
  arg1 = (TPMS_PCR_SELECT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_PCR_SELECT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg2 = (TPMS_PCR_SELECT *)(argp2);
  TPMS_PCR_SELECT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT **arg1 = (TPMS_PCR_SELECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECT_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_PCR_SELECT **""'"); 
  }
  arg1 = (TPMS_PCR_SELECT **)(argp1);
  result = (TPMS_PCR_SELECT *)TPMS_PCR_SELECT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_PCR_SELECTION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_PCR_SELECTION",SWIG_From_size_t((size_t)(sizeof(TPMS_PCR_SELECTION))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECTION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_PCR_SELECTION_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_PCR_SELECTION_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_PCR_SELECTION_ARRAY *)new_TPMS_PCR_SELECTION_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECTION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION_ARRAY *arg1 = (TPMS_PCR_SELECTION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECTION_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECTION_ARRAY" "', argument " "1"" of type '" "TPMS_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION_ARRAY *)(argp1);
  delete_TPMS_PCR_SELECTION_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION_ARRAY *arg1 = (TPMS_PCR_SELECTION_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_PCR_SELECTION result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECTION_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECTION_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_PCR_SELECTION_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_PCR_SELECTION *)memcpy((TPMS_PCR_SELECTION *)malloc(sizeof(TPMS_PCR_SELECTION)),&result,sizeof(TPMS_PCR_SELECTION)), SWIGTYPE_p_TPMS_PCR_SELECTION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION_ARRAY *arg1 = (TPMS_PCR_SELECTION_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_PCR_SELECTION arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_PCR_SELECTION_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_PCR_SELECTION_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_PCR_SELECTION,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_PCR_SELECTION_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_PCR_SELECTION""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_PCR_SELECTION_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_PCR_SELECTION""'");
    } else {
      arg3 = *((TPMS_PCR_SELECTION *)(argp3));
    }
  }
  TPMS_PCR_SELECTION_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION_ARRAY *arg1 = (TPMS_PCR_SELECTION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_ARRAY_cast" "', argument " "1"" of type '" "TPMS_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION_ARRAY *)(argp1);
  result = (TPMS_PCR_SELECTION *)TPMS_PCR_SELECTION_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  result = (TPMS_PCR_SELECTION_ARRAY *)TPMS_PCR_SELECTION_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_PCR_SELECTION_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_PCR_SELECTION_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_PCR_SELECTION_PTR")) SWIG_fail;
  result = (TPMS_PCR_SELECTION *)new_TPMS_PCR_SELECTION_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_PCR_SELECTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_PCR_SELECTION_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_PCR_SELECTION,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_PCR_SELECTION_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECTION""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_PCR_SELECTION_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECTION""'");
    } else {
      arg1 = *((TPMS_PCR_SELECTION *)(argp1));
    }
  }
  result = (TPMS_PCR_SELECTION *)copy_TPMS_PCR_SELECTION_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECTION_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECTION_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  delete_TPMS_PCR_SELECTION_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  TPMS_PCR_SELECTION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECTION_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_PTR_assign" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_PCR_SELECTION,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_PCR_SELECTION_PTR_assign" "', argument " "2"" of type '" "TPMS_PCR_SELECTION""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_PCR_SELECTION_PTR_assign" "', argument " "2"" of type '" "TPMS_PCR_SELECTION""'");
    } else {
      arg2 = *((TPMS_PCR_SELECTION *)(argp2));
    }
  }
  TPMS_PCR_SELECTION_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_PTR_value" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  result = TPMS_PCR_SELECTION_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_PCR_SELECTION *)memcpy((TPMS_PCR_SELECTION *)malloc(sizeof(TPMS_PCR_SELECTION)),&result,sizeof(TPMS_PCR_SELECTION)), SWIGTYPE_p_TPMS_PCR_SELECTION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_PCR_SELECTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_PCR_SELECTION_PTR_PTR")) SWIG_fail;
  result = (TPMS_PCR_SELECTION **)new_TPMS_PCR_SELECTION_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_PCR_SELECTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_PCR_SELECTION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_PCR_SELECTION_PTR_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  result = (TPMS_PCR_SELECTION **)copy_TPMS_PCR_SELECTION_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_PCR_SELECTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION **arg1 = (TPMS_PCR_SELECTION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_PCR_SELECTION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_PCR_SELECTION_PTR_PTR" "', argument " "1"" of type '" "TPMS_PCR_SELECTION **""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION **)(argp1);
  delete_TPMS_PCR_SELECTION_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION **arg1 = (TPMS_PCR_SELECTION **) 0 ;
  TPMS_PCR_SELECTION *arg2 = (TPMS_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_PCR_SELECTION_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_PCR_SELECTION **""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_PCR_SELECTION_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg2 = (TPMS_PCR_SELECTION *)(argp2);
  TPMS_PCR_SELECTION_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_PCR_SELECTION_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION **arg1 = (TPMS_PCR_SELECTION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_PCR_SELECTION_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_PCR_SELECTION_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_PCR_SELECTION **""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION **)(argp1);
  result = (TPMS_PCR_SELECTION *)TPMS_PCR_SELECTION_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_TK_CREATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_TK_CREATION",SWIG_From_size_t((size_t)(sizeof(TPMT_TK_CREATION))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_CREATION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_TK_CREATION_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_TK_CREATION_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_TK_CREATION_ARRAY *)new_TPMT_TK_CREATION_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_CREATION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION_ARRAY *arg1 = (TPMT_TK_CREATION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_CREATION_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_CREATION_ARRAY" "', argument " "1"" of type '" "TPMT_TK_CREATION_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_CREATION_ARRAY *)(argp1);
  delete_TPMT_TK_CREATION_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION_ARRAY *arg1 = (TPMT_TK_CREATION_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_TK_CREATION result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_CREATION_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_TK_CREATION_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_CREATION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_CREATION_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_TK_CREATION_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_TK_CREATION *)memcpy((TPMT_TK_CREATION *)malloc(sizeof(TPMT_TK_CREATION)),&result,sizeof(TPMT_TK_CREATION)), SWIGTYPE_p_TPMT_TK_CREATION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION_ARRAY *arg1 = (TPMT_TK_CREATION_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_TK_CREATION arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_TK_CREATION_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_TK_CREATION_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_CREATION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_CREATION_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_TK_CREATION,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_TK_CREATION_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_CREATION""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_CREATION_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_CREATION""'");
    } else {
      arg3 = *((TPMT_TK_CREATION *)(argp3));
    }
  }
  TPMT_TK_CREATION_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION_ARRAY *arg1 = (TPMT_TK_CREATION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_ARRAY_cast" "', argument " "1"" of type '" "TPMT_TK_CREATION_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_CREATION_ARRAY *)(argp1);
  result = (TPMT_TK_CREATION *)TPMT_TK_CREATION_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  result = (TPMT_TK_CREATION_ARRAY *)TPMT_TK_CREATION_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_CREATION_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_CREATION_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_TK_CREATION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_CREATION_PTR")) SWIG_fail;
  result = (TPMT_TK_CREATION *)new_TPMT_TK_CREATION_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_CREATION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_CREATION_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_TK_CREATION,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_CREATION_PTR" "', argument " "1"" of type '" "TPMT_TK_CREATION""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_TK_CREATION_PTR" "', argument " "1"" of type '" "TPMT_TK_CREATION""'");
    } else {
      arg1 = *((TPMT_TK_CREATION *)(argp1));
    }
  }
  result = (TPMT_TK_CREATION *)copy_TPMT_TK_CREATION_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_CREATION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_CREATION_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_CREATION_PTR" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  delete_TPMT_TK_CREATION_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  TPMT_TK_CREATION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_CREATION_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_TK_CREATION,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_CREATION_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_CREATION""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_CREATION_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_CREATION""'");
    } else {
      arg2 = *((TPMT_TK_CREATION *)(argp2));
    }
  }
  TPMT_TK_CREATION_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_PTR_value" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  result = TPMT_TK_CREATION_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_TK_CREATION *)memcpy((TPMT_TK_CREATION *)malloc(sizeof(TPMT_TK_CREATION)),&result,sizeof(TPMT_TK_CREATION)), SWIGTYPE_p_TPMT_TK_CREATION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_CREATION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_CREATION_PTR_PTR")) SWIG_fail;
  result = (TPMT_TK_CREATION **)new_TPMT_TK_CREATION_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_CREATION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_CREATION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_CREATION_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  result = (TPMT_TK_CREATION **)copy_TPMT_TK_CREATION_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_CREATION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION **arg1 = (TPMT_TK_CREATION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_CREATION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_CREATION_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg1 = (TPMT_TK_CREATION **)(argp1);
  delete_TPMT_TK_CREATION_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION **arg1 = (TPMT_TK_CREATION **) 0 ;
  TPMT_TK_CREATION *arg2 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_CREATION_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg1 = (TPMT_TK_CREATION **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_CREATION_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg2 = (TPMT_TK_CREATION *)(argp2);
  TPMT_TK_CREATION_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_CREATION_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION **arg1 = (TPMT_TK_CREATION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_CREATION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_CREATION_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_CREATION_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg1 = (TPMT_TK_CREATION **)(argp1);
  result = (TPMT_TK_CREATION *)TPMT_TK_CREATION_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_TK_VERIFIED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_TK_VERIFIED",SWIG_From_size_t((size_t)(sizeof(TPMT_TK_VERIFIED))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_VERIFIED_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_TK_VERIFIED_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_TK_VERIFIED_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_TK_VERIFIED_ARRAY *)new_TPMT_TK_VERIFIED_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_VERIFIED_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED_ARRAY *arg1 = (TPMT_TK_VERIFIED_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_VERIFIED_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_VERIFIED_ARRAY" "', argument " "1"" of type '" "TPMT_TK_VERIFIED_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED_ARRAY *)(argp1);
  delete_TPMT_TK_VERIFIED_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED_ARRAY *arg1 = (TPMT_TK_VERIFIED_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_TK_VERIFIED result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_VERIFIED_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_TK_VERIFIED_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_VERIFIED_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_TK_VERIFIED_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_TK_VERIFIED *)memcpy((TPMT_TK_VERIFIED *)malloc(sizeof(TPMT_TK_VERIFIED)),&result,sizeof(TPMT_TK_VERIFIED)), SWIGTYPE_p_TPMT_TK_VERIFIED, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED_ARRAY *arg1 = (TPMT_TK_VERIFIED_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_TK_VERIFIED arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_TK_VERIFIED_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_TK_VERIFIED_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_VERIFIED_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_TK_VERIFIED,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_TK_VERIFIED_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_VERIFIED""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_VERIFIED_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_VERIFIED""'");
    } else {
      arg3 = *((TPMT_TK_VERIFIED *)(argp3));
    }
  }
  TPMT_TK_VERIFIED_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED_ARRAY *arg1 = (TPMT_TK_VERIFIED_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_ARRAY_cast" "', argument " "1"" of type '" "TPMT_TK_VERIFIED_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED_ARRAY *)(argp1);
  result = (TPMT_TK_VERIFIED *)TPMT_TK_VERIFIED_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  result = (TPMT_TK_VERIFIED_ARRAY *)TPMT_TK_VERIFIED_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_VERIFIED_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_VERIFIED_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_TK_VERIFIED_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_VERIFIED_PTR")) SWIG_fail;
  result = (TPMT_TK_VERIFIED *)new_TPMT_TK_VERIFIED_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_VERIFIED_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_VERIFIED_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_TK_VERIFIED,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_VERIFIED_PTR" "', argument " "1"" of type '" "TPMT_TK_VERIFIED""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_TK_VERIFIED_PTR" "', argument " "1"" of type '" "TPMT_TK_VERIFIED""'");
    } else {
      arg1 = *((TPMT_TK_VERIFIED *)(argp1));
    }
  }
  result = (TPMT_TK_VERIFIED *)copy_TPMT_TK_VERIFIED_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_VERIFIED_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_VERIFIED_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_VERIFIED_PTR" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  delete_TPMT_TK_VERIFIED_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  TPMT_TK_VERIFIED arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_VERIFIED_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_TK_VERIFIED,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_VERIFIED_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_VERIFIED""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_VERIFIED_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_VERIFIED""'");
    } else {
      arg2 = *((TPMT_TK_VERIFIED *)(argp2));
    }
  }
  TPMT_TK_VERIFIED_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_PTR_value" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  result = TPMT_TK_VERIFIED_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_TK_VERIFIED *)memcpy((TPMT_TK_VERIFIED *)malloc(sizeof(TPMT_TK_VERIFIED)),&result,sizeof(TPMT_TK_VERIFIED)), SWIGTYPE_p_TPMT_TK_VERIFIED, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_VERIFIED_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_VERIFIED_PTR_PTR")) SWIG_fail;
  result = (TPMT_TK_VERIFIED **)new_TPMT_TK_VERIFIED_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_VERIFIED_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_VERIFIED_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_VERIFIED_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  result = (TPMT_TK_VERIFIED **)copy_TPMT_TK_VERIFIED_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_VERIFIED_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED **arg1 = (TPMT_TK_VERIFIED **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_VERIFIED_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_VERIFIED_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_VERIFIED **""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED **)(argp1);
  delete_TPMT_TK_VERIFIED_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED **arg1 = (TPMT_TK_VERIFIED **) 0 ;
  TPMT_TK_VERIFIED *arg2 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_VERIFIED_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_VERIFIED **""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_VERIFIED_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg2 = (TPMT_TK_VERIFIED *)(argp2);
  TPMT_TK_VERIFIED_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_VERIFIED_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED **arg1 = (TPMT_TK_VERIFIED **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_VERIFIED *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_VERIFIED_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_VERIFIED_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_TK_VERIFIED **""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED **)(argp1);
  result = (TPMT_TK_VERIFIED *)TPMT_TK_VERIFIED_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_TK_AUTH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_TK_AUTH",SWIG_From_size_t((size_t)(sizeof(TPMT_TK_AUTH))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_AUTH_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_TK_AUTH_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_TK_AUTH_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_TK_AUTH_ARRAY *)new_TPMT_TK_AUTH_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_AUTH_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH_ARRAY *arg1 = (TPMT_TK_AUTH_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_AUTH_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_AUTH_ARRAY" "', argument " "1"" of type '" "TPMT_TK_AUTH_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_AUTH_ARRAY *)(argp1);
  delete_TPMT_TK_AUTH_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH_ARRAY *arg1 = (TPMT_TK_AUTH_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_TK_AUTH result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_AUTH_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_TK_AUTH_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_AUTH_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_AUTH_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_TK_AUTH_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_TK_AUTH *)memcpy((TPMT_TK_AUTH *)malloc(sizeof(TPMT_TK_AUTH)),&result,sizeof(TPMT_TK_AUTH)), SWIGTYPE_p_TPMT_TK_AUTH, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH_ARRAY *arg1 = (TPMT_TK_AUTH_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_TK_AUTH arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_TK_AUTH_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_TK_AUTH_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_AUTH_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_AUTH_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_TK_AUTH,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_TK_AUTH_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_AUTH""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_AUTH_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_AUTH""'");
    } else {
      arg3 = *((TPMT_TK_AUTH *)(argp3));
    }
  }
  TPMT_TK_AUTH_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH_ARRAY *arg1 = (TPMT_TK_AUTH_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_ARRAY_cast" "', argument " "1"" of type '" "TPMT_TK_AUTH_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_AUTH_ARRAY *)(argp1);
  result = (TPMT_TK_AUTH *)TPMT_TK_AUTH_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  result = (TPMT_TK_AUTH_ARRAY *)TPMT_TK_AUTH_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_AUTH_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_AUTH_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_TK_AUTH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_AUTH_PTR")) SWIG_fail;
  result = (TPMT_TK_AUTH *)new_TPMT_TK_AUTH_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_AUTH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_AUTH_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_TK_AUTH,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_AUTH_PTR" "', argument " "1"" of type '" "TPMT_TK_AUTH""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_TK_AUTH_PTR" "', argument " "1"" of type '" "TPMT_TK_AUTH""'");
    } else {
      arg1 = *((TPMT_TK_AUTH *)(argp1));
    }
  }
  result = (TPMT_TK_AUTH *)copy_TPMT_TK_AUTH_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_AUTH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_AUTH_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_AUTH_PTR" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  delete_TPMT_TK_AUTH_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  TPMT_TK_AUTH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_AUTH_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_TK_AUTH,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_AUTH_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_AUTH""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_AUTH_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_AUTH""'");
    } else {
      arg2 = *((TPMT_TK_AUTH *)(argp2));
    }
  }
  TPMT_TK_AUTH_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_PTR_value" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  result = TPMT_TK_AUTH_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_TK_AUTH *)memcpy((TPMT_TK_AUTH *)malloc(sizeof(TPMT_TK_AUTH)),&result,sizeof(TPMT_TK_AUTH)), SWIGTYPE_p_TPMT_TK_AUTH, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_AUTH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_AUTH_PTR_PTR")) SWIG_fail;
  result = (TPMT_TK_AUTH **)new_TPMT_TK_AUTH_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_AUTH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_AUTH_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_AUTH_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  result = (TPMT_TK_AUTH **)copy_TPMT_TK_AUTH_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_AUTH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH **arg1 = (TPMT_TK_AUTH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_AUTH_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_AUTH_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg1 = (TPMT_TK_AUTH **)(argp1);
  delete_TPMT_TK_AUTH_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH **arg1 = (TPMT_TK_AUTH **) 0 ;
  TPMT_TK_AUTH *arg2 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_AUTH_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg1 = (TPMT_TK_AUTH **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_AUTH_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg2 = (TPMT_TK_AUTH *)(argp2);
  TPMT_TK_AUTH_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_AUTH_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH **arg1 = (TPMT_TK_AUTH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_AUTH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_AUTH_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_AUTH_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg1 = (TPMT_TK_AUTH **)(argp1);
  result = (TPMT_TK_AUTH *)TPMT_TK_AUTH_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_TK_HASHCHECK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_TK_HASHCHECK",SWIG_From_size_t((size_t)(sizeof(TPMT_TK_HASHCHECK))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_HASHCHECK_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_TK_HASHCHECK_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_TK_HASHCHECK_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_TK_HASHCHECK_ARRAY *)new_TPMT_TK_HASHCHECK_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_HASHCHECK_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK_ARRAY *arg1 = (TPMT_TK_HASHCHECK_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_HASHCHECK_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_HASHCHECK_ARRAY" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK_ARRAY *)(argp1);
  delete_TPMT_TK_HASHCHECK_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK_ARRAY *arg1 = (TPMT_TK_HASHCHECK_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_TK_HASHCHECK result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_HASHCHECK_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_HASHCHECK_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_TK_HASHCHECK_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_TK_HASHCHECK *)memcpy((TPMT_TK_HASHCHECK *)malloc(sizeof(TPMT_TK_HASHCHECK)),&result,sizeof(TPMT_TK_HASHCHECK)), SWIGTYPE_p_TPMT_TK_HASHCHECK, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK_ARRAY *arg1 = (TPMT_TK_HASHCHECK_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_TK_HASHCHECK arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_TK_HASHCHECK_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_TK_HASHCHECK_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_TK_HASHCHECK,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_TK_HASHCHECK_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_HASHCHECK""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_HASHCHECK_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_TK_HASHCHECK""'");
    } else {
      arg3 = *((TPMT_TK_HASHCHECK *)(argp3));
    }
  }
  TPMT_TK_HASHCHECK_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK_ARRAY *arg1 = (TPMT_TK_HASHCHECK_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_ARRAY_cast" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK_ARRAY *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK_ARRAY *)(argp1);
  result = (TPMT_TK_HASHCHECK *)TPMT_TK_HASHCHECK_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  result = (TPMT_TK_HASHCHECK_ARRAY *)TPMT_TK_HASHCHECK_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_TK_HASHCHECK_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_TK_HASHCHECK_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_TK_HASHCHECK_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_HASHCHECK_PTR")) SWIG_fail;
  result = (TPMT_TK_HASHCHECK *)new_TPMT_TK_HASHCHECK_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_HASHCHECK_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_HASHCHECK_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_TK_HASHCHECK,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_HASHCHECK_PTR" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_TK_HASHCHECK_PTR" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK""'");
    } else {
      arg1 = *((TPMT_TK_HASHCHECK *)(argp1));
    }
  }
  result = (TPMT_TK_HASHCHECK *)copy_TPMT_TK_HASHCHECK_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_HASHCHECK_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_HASHCHECK_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_HASHCHECK_PTR" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  delete_TPMT_TK_HASHCHECK_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  TPMT_TK_HASHCHECK arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_HASHCHECK_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_TK_HASHCHECK,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_HASHCHECK_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_HASHCHECK""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_TK_HASHCHECK_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_HASHCHECK""'");
    } else {
      arg2 = *((TPMT_TK_HASHCHECK *)(argp2));
    }
  }
  TPMT_TK_HASHCHECK_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_PTR_value" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  result = TPMT_TK_HASHCHECK_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_TK_HASHCHECK *)memcpy((TPMT_TK_HASHCHECK *)malloc(sizeof(TPMT_TK_HASHCHECK)),&result,sizeof(TPMT_TK_HASHCHECK)), SWIGTYPE_p_TPMT_TK_HASHCHECK, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_TK_HASHCHECK_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_TK_HASHCHECK_PTR_PTR")) SWIG_fail;
  result = (TPMT_TK_HASHCHECK **)new_TPMT_TK_HASHCHECK_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_TK_HASHCHECK_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_TK_HASHCHECK_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_TK_HASHCHECK_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  result = (TPMT_TK_HASHCHECK **)copy_TPMT_TK_HASHCHECK_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_TK_HASHCHECK_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK **arg1 = (TPMT_TK_HASHCHECK **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_TK_HASHCHECK_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_TK_HASHCHECK_PTR_PTR" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK **)(argp1);
  delete_TPMT_TK_HASHCHECK_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK **arg1 = (TPMT_TK_HASHCHECK **) 0 ;
  TPMT_TK_HASHCHECK *arg2 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_TK_HASHCHECK_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_TK_HASHCHECK_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg2 = (TPMT_TK_HASHCHECK *)(argp2);
  TPMT_TK_HASHCHECK_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_TK_HASHCHECK_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK **arg1 = (TPMT_TK_HASHCHECK **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_TK_HASHCHECK *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_TK_HASHCHECK_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_TK_HASHCHECK_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK **)(argp1);
  result = (TPMT_TK_HASHCHECK *)TPMT_TK_HASHCHECK_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ALG_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ALG_PROPERTY",SWIG_From_size_t((size_t)(sizeof(TPMS_ALG_PROPERTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALG_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ALG_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ALG_PROPERTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ALG_PROPERTY_ARRAY *)new_TPMS_ALG_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALG_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY_ARRAY *arg1 = (TPMS_ALG_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALG_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALG_PROPERTY_ARRAY" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY_ARRAY *)(argp1);
  delete_TPMS_ALG_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY_ARRAY *arg1 = (TPMS_ALG_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ALG_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALG_PROPERTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALG_PROPERTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ALG_PROPERTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ALG_PROPERTY *)memcpy((TPMS_ALG_PROPERTY *)malloc(sizeof(TPMS_ALG_PROPERTY)),&result,sizeof(TPMS_ALG_PROPERTY)), SWIGTYPE_p_TPMS_ALG_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY_ARRAY *arg1 = (TPMS_ALG_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ALG_PROPERTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ALG_PROPERTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALG_PROPERTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ALG_PROPERTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ALG_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ALG_PROPERTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ALG_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ALG_PROPERTY""'");
    } else {
      arg3 = *((TPMS_ALG_PROPERTY *)(argp3));
    }
  }
  TPMS_ALG_PROPERTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY_ARRAY *arg1 = (TPMS_ALG_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALG_PROPERTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY_ARRAY *)(argp1);
  result = (TPMS_ALG_PROPERTY *)TPMS_ALG_PROPERTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALG_PROPERTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  result = (TPMS_ALG_PROPERTY_ARRAY *)TPMS_ALG_PROPERTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ALG_PROPERTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ALG_PROPERTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ALG_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALG_PROPERTY_PTR")) SWIG_fail;
  result = (TPMS_ALG_PROPERTY *)new_TPMS_ALG_PROPERTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ALG_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ALG_PROPERTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ALG_PROPERTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ALG_PROPERTY_PTR" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ALG_PROPERTY_PTR" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY""'");
    } else {
      arg1 = *((TPMS_ALG_PROPERTY *)(argp1));
    }
  }
  result = (TPMS_ALG_PROPERTY *)copy_TPMS_ALG_PROPERTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALG_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALG_PROPERTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALG_PROPERTY_PTR" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  delete_TPMS_ALG_PROPERTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  TPMS_ALG_PROPERTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALG_PROPERTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_PTR_assign" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ALG_PROPERTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALG_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPMS_ALG_PROPERTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ALG_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPMS_ALG_PROPERTY""'");
    } else {
      arg2 = *((TPMS_ALG_PROPERTY *)(argp2));
    }
  }
  TPMS_ALG_PROPERTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALG_PROPERTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_PTR_value" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  result = TPMS_ALG_PROPERTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ALG_PROPERTY *)memcpy((TPMS_ALG_PROPERTY *)malloc(sizeof(TPMS_ALG_PROPERTY)),&result,sizeof(TPMS_ALG_PROPERTY)), SWIGTYPE_p_TPMS_ALG_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALG_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALG_PROPERTY_PTR_PTR")) SWIG_fail;
  result = (TPMS_ALG_PROPERTY **)new_TPMS_ALG_PROPERTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ALG_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ALG_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ALG_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  result = (TPMS_ALG_PROPERTY **)copy_TPMS_ALG_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALG_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY **arg1 = (TPMS_ALG_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALG_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALG_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY **""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY **)(argp1);
  delete_TPMS_ALG_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY **arg1 = (TPMS_ALG_PROPERTY **) 0 ;
  TPMS_ALG_PROPERTY *arg2 = (TPMS_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALG_PROPERTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY **""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALG_PROPERTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg2 = (TPMS_ALG_PROPERTY *)(argp2);
  TPMS_ALG_PROPERTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALG_PROPERTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY **arg1 = (TPMS_ALG_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALG_PROPERTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALG_PROPERTY_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY **""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY **)(argp1);
  result = (TPMS_ALG_PROPERTY *)TPMS_ALG_PROPERTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_TAGGED_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_TAGGED_PROPERTY",SWIG_From_size_t((size_t)(sizeof(TPMS_TAGGED_PROPERTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_TAGGED_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_TAGGED_PROPERTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_TAGGED_PROPERTY_ARRAY *)new_TPMS_TAGGED_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY_ARRAY *arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PROPERTY_ARRAY" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *)(argp1);
  delete_TPMS_TAGGED_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY_ARRAY *arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_TAGGED_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PROPERTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_TAGGED_PROPERTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_TAGGED_PROPERTY *)memcpy((TPMS_TAGGED_PROPERTY *)malloc(sizeof(TPMS_TAGGED_PROPERTY)),&result,sizeof(TPMS_TAGGED_PROPERTY)), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY_ARRAY *arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_TAGGED_PROPERTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_TAGGED_PROPERTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_TAGGED_PROPERTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TAGGED_PROPERTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TAGGED_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TAGGED_PROPERTY""'");
    } else {
      arg3 = *((TPMS_TAGGED_PROPERTY *)(argp3));
    }
  }
  TPMS_TAGGED_PROPERTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY_ARRAY *arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PROPERTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY_cast" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY_ARRAY *)(argp1);
  result = (TPMS_TAGGED_PROPERTY *)TPMS_TAGGED_PROPERTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PROPERTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  result = (TPMS_TAGGED_PROPERTY_ARRAY *)TPMS_TAGGED_PROPERTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TAGGED_PROPERTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TAGGED_PROPERTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_PROPERTY_PTR")) SWIG_fail;
  result = (TPMS_TAGGED_PROPERTY *)new_TPMS_TAGGED_PROPERTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TAGGED_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TAGGED_PROPERTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_TAGGED_PROPERTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TAGGED_PROPERTY_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_TAGGED_PROPERTY_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY""'");
    } else {
      arg1 = *((TPMS_TAGGED_PROPERTY *)(argp1));
    }
  }
  result = (TPMS_TAGGED_PROPERTY *)copy_TPMS_TAGGED_PROPERTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PROPERTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PROPERTY_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  delete_TPMS_TAGGED_PROPERTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  TPMS_TAGGED_PROPERTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PROPERTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_PTR_assign" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_TAGGED_PROPERTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_PROPERTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TAGGED_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_PROPERTY""'");
    } else {
      arg2 = *((TPMS_TAGGED_PROPERTY *)(argp2));
    }
  }
  TPMS_TAGGED_PROPERTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PROPERTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_PTR_value" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  result = TPMS_TAGGED_PROPERTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_TAGGED_PROPERTY *)memcpy((TPMS_TAGGED_PROPERTY *)malloc(sizeof(TPMS_TAGGED_PROPERTY)),&result,sizeof(TPMS_TAGGED_PROPERTY)), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_PROPERTY_PTR_PTR")) SWIG_fail;
  result = (TPMS_TAGGED_PROPERTY **)new_TPMS_TAGGED_PROPERTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TAGGED_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TAGGED_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TAGGED_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  result = (TPMS_TAGGED_PROPERTY **)copy_TPMS_TAGGED_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY **arg1 = (TPMS_TAGGED_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY **""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY **)(argp1);
  delete_TPMS_TAGGED_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY **arg1 = (TPMS_TAGGED_PROPERTY **) 0 ;
  TPMS_TAGGED_PROPERTY *arg2 = (TPMS_TAGGED_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PROPERTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY **""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_PROPERTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg2 = (TPMS_TAGGED_PROPERTY *)(argp2);
  TPMS_TAGGED_PROPERTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PROPERTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY **arg1 = (TPMS_TAGGED_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PROPERTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PROPERTY_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY **""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY **)(argp1);
  result = (TPMS_TAGGED_PROPERTY *)TPMS_TAGGED_PROPERTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_TAGGED_PCR_SELECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_TAGGED_PCR_SELECT",SWIG_From_size_t((size_t)(sizeof(TPMS_TAGGED_PCR_SELECT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PCR_SELECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_TAGGED_PCR_SELECT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_TAGGED_PCR_SELECT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_TAGGED_PCR_SELECT_ARRAY *)new_TPMS_TAGGED_PCR_SELECT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PCR_SELECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT_ARRAY *arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PCR_SELECT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PCR_SELECT_ARRAY" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *)(argp1);
  delete_TPMS_TAGGED_PCR_SELECT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT_ARRAY *arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_TAGGED_PCR_SELECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_TAGGED_PCR_SELECT *)memcpy((TPMS_TAGGED_PCR_SELECT *)malloc(sizeof(TPMS_TAGGED_PCR_SELECT)),&result,sizeof(TPMS_TAGGED_PCR_SELECT)), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT_ARRAY *arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_TAGGED_PCR_SELECT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TAGGED_PCR_SELECT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TAGGED_PCR_SELECT""'");
    } else {
      arg3 = *((TPMS_TAGGED_PCR_SELECT *)(argp3));
    }
  }
  TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT_ARRAY *arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY_cast" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT_ARRAY *)(argp1);
  result = (TPMS_TAGGED_PCR_SELECT *)TPMS_TAGGED_PCR_SELECT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  result = (TPMS_TAGGED_PCR_SELECT_ARRAY *)TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TAGGED_PCR_SELECT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PCR_SELECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_PCR_SELECT_PTR")) SWIG_fail;
  result = (TPMS_TAGGED_PCR_SELECT *)new_TPMS_TAGGED_PCR_SELECT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TAGGED_PCR_SELECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TAGGED_PCR_SELECT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TAGGED_PCR_SELECT_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_TAGGED_PCR_SELECT_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT""'");
    } else {
      arg1 = *((TPMS_TAGGED_PCR_SELECT *)(argp1));
    }
  }
  result = (TPMS_TAGGED_PCR_SELECT *)copy_TPMS_TAGGED_PCR_SELECT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PCR_SELECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PCR_SELECT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PCR_SELECT_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  delete_TPMS_TAGGED_PCR_SELECT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  TPMS_TAGGED_PCR_SELECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PCR_SELECT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_assign" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_PCR_SELECT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_PCR_SELECT""'");
    } else {
      arg2 = *((TPMS_TAGGED_PCR_SELECT *)(argp2));
    }
  }
  TPMS_TAGGED_PCR_SELECT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_value" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  result = TPMS_TAGGED_PCR_SELECT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_TAGGED_PCR_SELECT *)memcpy((TPMS_TAGGED_PCR_SELECT *)malloc(sizeof(TPMS_TAGGED_PCR_SELECT)),&result,sizeof(TPMS_TAGGED_PCR_SELECT)), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_PCR_SELECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_PCR_SELECT_PTR_PTR")) SWIG_fail;
  result = (TPMS_TAGGED_PCR_SELECT **)new_TPMS_TAGGED_PCR_SELECT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  result = (TPMS_TAGGED_PCR_SELECT **)copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT **arg1 = (TPMS_TAGGED_PCR_SELECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT **""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT **)(argp1);
  delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT **arg1 = (TPMS_TAGGED_PCR_SELECT **) 0 ;
  TPMS_TAGGED_PCR_SELECT *arg2 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT **""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg2 = (TPMS_TAGGED_PCR_SELECT *)(argp2);
  TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_PCR_SELECT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT **arg1 = (TPMS_TAGGED_PCR_SELECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_PCR_SELECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_PCR_SELECT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_PCR_SELECT_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT **""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT **)(argp1);
  result = (TPMS_TAGGED_PCR_SELECT *)TPMS_TAGGED_PCR_SELECT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_TAGGED_POLICY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_TAGGED_POLICY",SWIG_From_size_t((size_t)(sizeof(TPMS_TAGGED_POLICY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_POLICY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_TAGGED_POLICY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_TAGGED_POLICY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_TAGGED_POLICY_ARRAY *)new_TPMS_TAGGED_POLICY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_POLICY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY_ARRAY *arg1 = (TPMS_TAGGED_POLICY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_POLICY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_POLICY_ARRAY" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY_ARRAY *)(argp1);
  delete_TPMS_TAGGED_POLICY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY_ARRAY *arg1 = (TPMS_TAGGED_POLICY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_TAGGED_POLICY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_POLICY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_POLICY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_TAGGED_POLICY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_TAGGED_POLICY *)memcpy((TPMS_TAGGED_POLICY *)malloc(sizeof(TPMS_TAGGED_POLICY)),&result,sizeof(TPMS_TAGGED_POLICY)), SWIGTYPE_p_TPMS_TAGGED_POLICY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY_ARRAY *arg1 = (TPMS_TAGGED_POLICY_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_TAGGED_POLICY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_TAGGED_POLICY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TAGGED_POLICY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_TAGGED_POLICY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_TAGGED_POLICY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TAGGED_POLICY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TAGGED_POLICY_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TAGGED_POLICY""'");
    } else {
      arg3 = *((TPMS_TAGGED_POLICY *)(argp3));
    }
  }
  TPMS_TAGGED_POLICY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY_ARRAY *arg1 = (TPMS_TAGGED_POLICY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_POLICY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_ARRAY_cast" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY_ARRAY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY_ARRAY *)(argp1);
  result = (TPMS_TAGGED_POLICY *)TPMS_TAGGED_POLICY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_POLICY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  result = (TPMS_TAGGED_POLICY_ARRAY *)TPMS_TAGGED_POLICY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TAGGED_POLICY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TAGGED_POLICY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_POLICY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_POLICY_PTR")) SWIG_fail;
  result = (TPMS_TAGGED_POLICY *)new_TPMS_TAGGED_POLICY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TAGGED_POLICY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TAGGED_POLICY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_TAGGED_POLICY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TAGGED_POLICY_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_TAGGED_POLICY_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY""'");
    } else {
      arg1 = *((TPMS_TAGGED_POLICY *)(argp1));
    }
  }
  result = (TPMS_TAGGED_POLICY *)copy_TPMS_TAGGED_POLICY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_POLICY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_POLICY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_POLICY_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  delete_TPMS_TAGGED_POLICY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  TPMS_TAGGED_POLICY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_POLICY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_PTR_assign" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_TAGGED_POLICY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_POLICY_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_POLICY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TAGGED_POLICY_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_POLICY""'");
    } else {
      arg2 = *((TPMS_TAGGED_POLICY *)(argp2));
    }
  }
  TPMS_TAGGED_POLICY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_POLICY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_PTR_value" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  result = TPMS_TAGGED_POLICY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_TAGGED_POLICY *)memcpy((TPMS_TAGGED_POLICY *)malloc(sizeof(TPMS_TAGGED_POLICY)),&result,sizeof(TPMS_TAGGED_POLICY)), SWIGTYPE_p_TPMS_TAGGED_POLICY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TAGGED_POLICY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TAGGED_POLICY_PTR_PTR")) SWIG_fail;
  result = (TPMS_TAGGED_POLICY **)new_TPMS_TAGGED_POLICY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TAGGED_POLICY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TAGGED_POLICY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TAGGED_POLICY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TAGGED_POLICY_PTR_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  result = (TPMS_TAGGED_POLICY **)copy_TPMS_TAGGED_POLICY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TAGGED_POLICY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TAGGED_POLICY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY **arg1 = (TPMS_TAGGED_POLICY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TAGGED_POLICY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TAGGED_POLICY_PTR_PTR" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY **""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY **)(argp1);
  delete_TPMS_TAGGED_POLICY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY **arg1 = (TPMS_TAGGED_POLICY **) 0 ;
  TPMS_TAGGED_POLICY *arg2 = (TPMS_TAGGED_POLICY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TAGGED_POLICY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY **""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TAGGED_POLICY_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg2 = (TPMS_TAGGED_POLICY *)(argp2);
  TPMS_TAGGED_POLICY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TAGGED_POLICY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY **arg1 = (TPMS_TAGGED_POLICY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TAGGED_POLICY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TAGGED_POLICY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TAGGED_POLICY_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY **""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY **)(argp1);
  result = (TPMS_TAGGED_POLICY *)TPMS_TAGGED_POLICY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_CC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_CC",SWIG_From_size_t((size_t)(sizeof(TPML_CC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_CC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_CC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_CC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_CC_ARRAY *)new_TPML_CC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC_ARRAY *arg1 = (TPML_CC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CC_ARRAY" "', argument " "1"" of type '" "TPML_CC_ARRAY *""'"); 
  }
  arg1 = (TPML_CC_ARRAY *)(argp1);
  delete_TPML_CC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC_ARRAY *arg1 = (TPML_CC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_CC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_CC_ARRAY *""'"); 
  }
  arg1 = (TPML_CC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_CC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_CC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_CC *)memcpy((TPML_CC *)malloc(sizeof(TPML_CC)),&result,sizeof(TPML_CC)), SWIGTYPE_p_TPML_CC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC_ARRAY *arg1 = (TPML_CC_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_CC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_CC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_CC_ARRAY *""'"); 
  }
  arg1 = (TPML_CC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_CC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_CC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_CC_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_CC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_CC_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_CC""'");
    } else {
      arg3 = *((TPML_CC *)(argp3));
    }
  }
  TPML_CC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC_ARRAY *arg1 = (TPML_CC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_ARRAY_cast" "', argument " "1"" of type '" "TPML_CC_ARRAY *""'"); 
  }
  arg1 = (TPML_CC_ARRAY *)(argp1);
  result = (TPML_CC *)TPML_CC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  result = (TPML_CC_ARRAY *)TPML_CC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_CC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_CC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_CC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_CC_PTR")) SWIG_fail;
  result = (TPML_CC *)new_TPML_CC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_CC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_CC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_CC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_CC_PTR" "', argument " "1"" of type '" "TPML_CC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_CC_PTR" "', argument " "1"" of type '" "TPML_CC""'");
    } else {
      arg1 = *((TPML_CC *)(argp1));
    }
  }
  result = (TPML_CC *)copy_TPML_CC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CC_PTR" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  delete_TPML_CC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  TPML_CC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_PTR_assign" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_CC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_CC_PTR_assign" "', argument " "2"" of type '" "TPML_CC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_CC_PTR_assign" "', argument " "2"" of type '" "TPML_CC""'");
    } else {
      arg2 = *((TPML_CC *)(argp2));
    }
  }
  TPML_CC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_PTR_value" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  result = TPML_CC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_CC *)memcpy((TPML_CC *)malloc(sizeof(TPML_CC)),&result,sizeof(TPML_CC)), SWIGTYPE_p_TPML_CC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_CC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_CC_PTR_PTR")) SWIG_fail;
  result = (TPML_CC **)new_TPML_CC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_CC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_CC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_CC_PTR_PTR" "', argument " "1"" of type '" "TPML_CC *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  result = (TPML_CC **)copy_TPML_CC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC **arg1 = (TPML_CC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CC_PTR_PTR" "', argument " "1"" of type '" "TPML_CC **""'"); 
  }
  arg1 = (TPML_CC **)(argp1);
  delete_TPML_CC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC **arg1 = (TPML_CC **) 0 ;
  TPML_CC *arg2 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_CC **""'"); 
  }
  arg1 = (TPML_CC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_CC_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_CC *""'"); 
  }
  arg2 = (TPML_CC *)(argp2);
  TPML_CC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC **arg1 = (TPML_CC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CC_PTR_PTR_value" "', argument " "1"" of type '" "TPML_CC **""'"); 
  }
  arg1 = (TPML_CC **)(argp1);
  result = (TPML_CC *)TPML_CC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_CCA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_CCA",SWIG_From_size_t((size_t)(sizeof(TPML_CCA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_CCA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_CCA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_CCA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_CCA_ARRAY *)new_TPML_CCA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CCA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA_ARRAY *arg1 = (TPML_CCA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CCA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CCA_ARRAY" "', argument " "1"" of type '" "TPML_CCA_ARRAY *""'"); 
  }
  arg1 = (TPML_CCA_ARRAY *)(argp1);
  delete_TPML_CCA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA_ARRAY *arg1 = (TPML_CCA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_CCA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CCA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_CCA_ARRAY *""'"); 
  }
  arg1 = (TPML_CCA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_CCA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_CCA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_CCA *)memcpy((TPML_CCA *)malloc(sizeof(TPML_CCA)),&result,sizeof(TPML_CCA)), SWIGTYPE_p_TPML_CCA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA_ARRAY *arg1 = (TPML_CCA_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_CCA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_CCA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_CCA_ARRAY *""'"); 
  }
  arg1 = (TPML_CCA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_CCA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_CCA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_CCA_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_CCA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_CCA_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_CCA""'");
    } else {
      arg3 = *((TPML_CCA *)(argp3));
    }
  }
  TPML_CCA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA_ARRAY *arg1 = (TPML_CCA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CCA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_ARRAY_cast" "', argument " "1"" of type '" "TPML_CCA_ARRAY *""'"); 
  }
  arg1 = (TPML_CCA_ARRAY *)(argp1);
  result = (TPML_CCA *)TPML_CCA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CCA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  result = (TPML_CCA_ARRAY *)TPML_CCA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_CCA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_CCA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_CCA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_CCA_PTR")) SWIG_fail;
  result = (TPML_CCA *)new_TPML_CCA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_CCA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_CCA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_CCA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_CCA_PTR" "', argument " "1"" of type '" "TPML_CCA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_CCA_PTR" "', argument " "1"" of type '" "TPML_CCA""'");
    } else {
      arg1 = *((TPML_CCA *)(argp1));
    }
  }
  result = (TPML_CCA *)copy_TPML_CCA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CCA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CCA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CCA_PTR" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  delete_TPML_CCA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  TPML_CCA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CCA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_PTR_assign" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_CCA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_CCA_PTR_assign" "', argument " "2"" of type '" "TPML_CCA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_CCA_PTR_assign" "', argument " "2"" of type '" "TPML_CCA""'");
    } else {
      arg2 = *((TPML_CCA *)(argp2));
    }
  }
  TPML_CCA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CCA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_PTR_value" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  result = TPML_CCA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_CCA *)memcpy((TPML_CCA *)malloc(sizeof(TPML_CCA)),&result,sizeof(TPML_CCA)), SWIGTYPE_p_TPML_CCA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_CCA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_CCA_PTR_PTR")) SWIG_fail;
  result = (TPML_CCA **)new_TPML_CCA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_CCA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_CCA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_CCA_PTR_PTR" "', argument " "1"" of type '" "TPML_CCA *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  result = (TPML_CCA **)copy_TPML_CCA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_CCA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA **arg1 = (TPML_CCA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_CCA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_CCA_PTR_PTR" "', argument " "1"" of type '" "TPML_CCA **""'"); 
  }
  arg1 = (TPML_CCA **)(argp1);
  delete_TPML_CCA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA **arg1 = (TPML_CCA **) 0 ;
  TPML_CCA *arg2 = (TPML_CCA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_CCA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_CCA **""'"); 
  }
  arg1 = (TPML_CCA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_CCA_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_CCA *""'"); 
  }
  arg2 = (TPML_CCA *)(argp2);
  TPML_CCA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_CCA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA **arg1 = (TPML_CCA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_CCA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_CCA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_CCA_PTR_PTR_value" "', argument " "1"" of type '" "TPML_CCA **""'"); 
  }
  arg1 = (TPML_CCA **)(argp1);
  result = (TPML_CCA *)TPML_CCA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_CCA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_ALG_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_ALG",SWIG_From_size_t((size_t)(sizeof(TPML_ALG))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_ALG_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_ALG_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_ALG_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_ALG_ARRAY *)new_TPML_ALG_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_ARRAY *arg1 = (TPML_ALG_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_ARRAY" "', argument " "1"" of type '" "TPML_ALG_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_ARRAY *)(argp1);
  delete_TPML_ALG_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_ARRAY *arg1 = (TPML_ALG_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_ALG result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_ALG_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ALG_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_ALG_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_ALG *)memcpy((TPML_ALG *)malloc(sizeof(TPML_ALG)),&result,sizeof(TPML_ALG)), SWIGTYPE_p_TPML_ALG, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_ARRAY *arg1 = (TPML_ALG_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_ALG arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_ALG_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_ALG_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ALG_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_ALG,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_ALG_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_ALG""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_ALG_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_ALG""'");
    } else {
      arg3 = *((TPML_ALG *)(argp3));
    }
  }
  TPML_ALG_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_ARRAY *arg1 = (TPML_ALG_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_ARRAY_cast" "', argument " "1"" of type '" "TPML_ALG_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_ARRAY *)(argp1);
  result = (TPML_ALG *)TPML_ALG_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  result = (TPML_ALG_ARRAY *)TPML_ALG_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_ALG_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_ALG_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_ALG_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ALG_PTR")) SWIG_fail;
  result = (TPML_ALG *)new_TPML_ALG_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_ALG_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_ALG_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_ALG,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_ALG_PTR" "', argument " "1"" of type '" "TPML_ALG""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_ALG_PTR" "', argument " "1"" of type '" "TPML_ALG""'");
    } else {
      arg1 = *((TPML_ALG *)(argp1));
    }
  }
  result = (TPML_ALG *)copy_TPML_ALG_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_PTR" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  delete_TPML_ALG_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  TPML_ALG arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PTR_assign" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_ALG,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ALG_PTR_assign" "', argument " "2"" of type '" "TPML_ALG""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_ALG_PTR_assign" "', argument " "2"" of type '" "TPML_ALG""'");
    } else {
      arg2 = *((TPML_ALG *)(argp2));
    }
  }
  TPML_ALG_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PTR_value" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  result = TPML_ALG_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_ALG *)memcpy((TPML_ALG *)malloc(sizeof(TPML_ALG)),&result,sizeof(TPML_ALG)), SWIGTYPE_p_TPML_ALG, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_ALG_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ALG_PTR_PTR")) SWIG_fail;
  result = (TPML_ALG **)new_TPML_ALG_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_ALG_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_ALG_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_ALG_PTR_PTR" "', argument " "1"" of type '" "TPML_ALG *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  result = (TPML_ALG **)copy_TPML_ALG_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG **arg1 = (TPML_ALG **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_PTR_PTR" "', argument " "1"" of type '" "TPML_ALG **""'"); 
  }
  arg1 = (TPML_ALG **)(argp1);
  delete_TPML_ALG_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG **arg1 = (TPML_ALG **) 0 ;
  TPML_ALG *arg2 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_ALG **""'"); 
  }
  arg1 = (TPML_ALG **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ALG_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_ALG *""'"); 
  }
  arg2 = (TPML_ALG *)(argp2);
  TPML_ALG_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG **arg1 = (TPML_ALG **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PTR_PTR_value" "', argument " "1"" of type '" "TPML_ALG **""'"); 
  }
  arg1 = (TPML_ALG **)(argp1);
  result = (TPML_ALG *)TPML_ALG_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_HANDLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_HANDLE",SWIG_From_size_t((size_t)(sizeof(TPML_HANDLE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_HANDLE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_HANDLE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_HANDLE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_HANDLE_ARRAY *)new_TPML_HANDLE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_HANDLE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE_ARRAY *arg1 = (TPML_HANDLE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_HANDLE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_HANDLE_ARRAY" "', argument " "1"" of type '" "TPML_HANDLE_ARRAY *""'"); 
  }
  arg1 = (TPML_HANDLE_ARRAY *)(argp1);
  delete_TPML_HANDLE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE_ARRAY *arg1 = (TPML_HANDLE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_HANDLE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_HANDLE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_HANDLE_ARRAY *""'"); 
  }
  arg1 = (TPML_HANDLE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_HANDLE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_HANDLE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_HANDLE *)memcpy((TPML_HANDLE *)malloc(sizeof(TPML_HANDLE)),&result,sizeof(TPML_HANDLE)), SWIGTYPE_p_TPML_HANDLE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE_ARRAY *arg1 = (TPML_HANDLE_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_HANDLE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_HANDLE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_HANDLE_ARRAY *""'"); 
  }
  arg1 = (TPML_HANDLE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_HANDLE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_HANDLE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_HANDLE_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_HANDLE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_HANDLE_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_HANDLE""'");
    } else {
      arg3 = *((TPML_HANDLE *)(argp3));
    }
  }
  TPML_HANDLE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE_ARRAY *arg1 = (TPML_HANDLE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_HANDLE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_ARRAY_cast" "', argument " "1"" of type '" "TPML_HANDLE_ARRAY *""'"); 
  }
  arg1 = (TPML_HANDLE_ARRAY *)(argp1);
  result = (TPML_HANDLE *)TPML_HANDLE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_HANDLE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  result = (TPML_HANDLE_ARRAY *)TPML_HANDLE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_HANDLE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_HANDLE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_HANDLE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_HANDLE_PTR")) SWIG_fail;
  result = (TPML_HANDLE *)new_TPML_HANDLE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_HANDLE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_HANDLE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_HANDLE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_HANDLE_PTR" "', argument " "1"" of type '" "TPML_HANDLE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_HANDLE_PTR" "', argument " "1"" of type '" "TPML_HANDLE""'");
    } else {
      arg1 = *((TPML_HANDLE *)(argp1));
    }
  }
  result = (TPML_HANDLE *)copy_TPML_HANDLE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_HANDLE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_HANDLE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_HANDLE_PTR" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  delete_TPML_HANDLE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  TPML_HANDLE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_HANDLE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_PTR_assign" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_HANDLE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_HANDLE_PTR_assign" "', argument " "2"" of type '" "TPML_HANDLE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_HANDLE_PTR_assign" "', argument " "2"" of type '" "TPML_HANDLE""'");
    } else {
      arg2 = *((TPML_HANDLE *)(argp2));
    }
  }
  TPML_HANDLE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_HANDLE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_PTR_value" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  result = TPML_HANDLE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_HANDLE *)memcpy((TPML_HANDLE *)malloc(sizeof(TPML_HANDLE)),&result,sizeof(TPML_HANDLE)), SWIGTYPE_p_TPML_HANDLE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_HANDLE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_HANDLE_PTR_PTR")) SWIG_fail;
  result = (TPML_HANDLE **)new_TPML_HANDLE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_HANDLE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_HANDLE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_HANDLE_PTR_PTR" "', argument " "1"" of type '" "TPML_HANDLE *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  result = (TPML_HANDLE **)copy_TPML_HANDLE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_HANDLE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE **arg1 = (TPML_HANDLE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_HANDLE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_HANDLE_PTR_PTR" "', argument " "1"" of type '" "TPML_HANDLE **""'"); 
  }
  arg1 = (TPML_HANDLE **)(argp1);
  delete_TPML_HANDLE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE **arg1 = (TPML_HANDLE **) 0 ;
  TPML_HANDLE *arg2 = (TPML_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_HANDLE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_HANDLE **""'"); 
  }
  arg1 = (TPML_HANDLE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_HANDLE_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_HANDLE *""'"); 
  }
  arg2 = (TPML_HANDLE *)(argp2);
  TPML_HANDLE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_HANDLE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE **arg1 = (TPML_HANDLE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_HANDLE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_HANDLE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_HANDLE_PTR_PTR_value" "', argument " "1"" of type '" "TPML_HANDLE **""'"); 
  }
  arg1 = (TPML_HANDLE **)(argp1);
  result = (TPML_HANDLE *)TPML_HANDLE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_DIGEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_DIGEST",SWIG_From_size_t((size_t)(sizeof(TPML_DIGEST))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_DIGEST_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_DIGEST_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_DIGEST_ARRAY *)new_TPML_DIGEST_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_ARRAY *arg1 = (TPML_DIGEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_ARRAY" "', argument " "1"" of type '" "TPML_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_ARRAY *)(argp1);
  delete_TPML_DIGEST_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_ARRAY *arg1 = (TPML_DIGEST_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_DIGEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_DIGEST_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_DIGEST_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_DIGEST *)memcpy((TPML_DIGEST *)malloc(sizeof(TPML_DIGEST)),&result,sizeof(TPML_DIGEST)), SWIGTYPE_p_TPML_DIGEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_ARRAY *arg1 = (TPML_DIGEST_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_DIGEST arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_DIGEST_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_DIGEST_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_DIGEST,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_DIGEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_DIGEST""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_DIGEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_DIGEST""'");
    } else {
      arg3 = *((TPML_DIGEST *)(argp3));
    }
  }
  TPML_DIGEST_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_ARRAY *arg1 = (TPML_DIGEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_ARRAY_cast" "', argument " "1"" of type '" "TPML_DIGEST_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_ARRAY *)(argp1);
  result = (TPML_DIGEST *)TPML_DIGEST_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  result = (TPML_DIGEST_ARRAY *)TPML_DIGEST_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_DIGEST_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_DIGEST_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_DIGEST_PTR")) SWIG_fail;
  result = (TPML_DIGEST *)new_TPML_DIGEST_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_DIGEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_DIGEST_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_DIGEST,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_DIGEST_PTR" "', argument " "1"" of type '" "TPML_DIGEST""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_DIGEST_PTR" "', argument " "1"" of type '" "TPML_DIGEST""'");
    } else {
      arg1 = *((TPML_DIGEST *)(argp1));
    }
  }
  result = (TPML_DIGEST *)copy_TPML_DIGEST_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_PTR" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  delete_TPML_DIGEST_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  TPML_DIGEST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_PTR_assign" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_DIGEST,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_DIGEST_PTR_assign" "', argument " "2"" of type '" "TPML_DIGEST""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_DIGEST_PTR_assign" "', argument " "2"" of type '" "TPML_DIGEST""'");
    } else {
      arg2 = *((TPML_DIGEST *)(argp2));
    }
  }
  TPML_DIGEST_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_PTR_value" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  result = TPML_DIGEST_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_DIGEST *)memcpy((TPML_DIGEST *)malloc(sizeof(TPML_DIGEST)),&result,sizeof(TPML_DIGEST)), SWIGTYPE_p_TPML_DIGEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_DIGEST_PTR_PTR")) SWIG_fail;
  result = (TPML_DIGEST **)new_TPML_DIGEST_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_DIGEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_DIGEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_DIGEST_PTR_PTR" "', argument " "1"" of type '" "TPML_DIGEST *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  result = (TPML_DIGEST **)copy_TPML_DIGEST_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST **arg1 = (TPML_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_PTR_PTR" "', argument " "1"" of type '" "TPML_DIGEST **""'"); 
  }
  arg1 = (TPML_DIGEST **)(argp1);
  delete_TPML_DIGEST_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST **arg1 = (TPML_DIGEST **) 0 ;
  TPML_DIGEST *arg2 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_DIGEST **""'"); 
  }
  arg1 = (TPML_DIGEST **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_DIGEST_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_DIGEST *""'"); 
  }
  arg2 = (TPML_DIGEST *)(argp2);
  TPML_DIGEST_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST **arg1 = (TPML_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_PTR_PTR_value" "', argument " "1"" of type '" "TPML_DIGEST **""'"); 
  }
  arg1 = (TPML_DIGEST **)(argp1);
  result = (TPML_DIGEST *)TPML_DIGEST_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_DIGEST_VALUES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_DIGEST_VALUES",SWIG_From_size_t((size_t)(sizeof(TPML_DIGEST_VALUES))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_VALUES_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_DIGEST_VALUES_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_DIGEST_VALUES_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_DIGEST_VALUES_ARRAY *)new_TPML_DIGEST_VALUES_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_VALUES_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES_ARRAY *arg1 = (TPML_DIGEST_VALUES_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_VALUES_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_VALUES_ARRAY" "', argument " "1"" of type '" "TPML_DIGEST_VALUES_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES_ARRAY *)(argp1);
  delete_TPML_DIGEST_VALUES_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES_ARRAY *arg1 = (TPML_DIGEST_VALUES_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_DIGEST_VALUES result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_VALUES_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_DIGEST_VALUES_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_DIGEST_VALUES_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_DIGEST_VALUES_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_DIGEST_VALUES *)memcpy((TPML_DIGEST_VALUES *)malloc(sizeof(TPML_DIGEST_VALUES)),&result,sizeof(TPML_DIGEST_VALUES)), SWIGTYPE_p_TPML_DIGEST_VALUES, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES_ARRAY *arg1 = (TPML_DIGEST_VALUES_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_DIGEST_VALUES arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_DIGEST_VALUES_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_DIGEST_VALUES_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_DIGEST_VALUES_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_DIGEST_VALUES,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_DIGEST_VALUES_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_DIGEST_VALUES""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_DIGEST_VALUES_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_DIGEST_VALUES""'");
    } else {
      arg3 = *((TPML_DIGEST_VALUES *)(argp3));
    }
  }
  TPML_DIGEST_VALUES_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES_ARRAY *arg1 = (TPML_DIGEST_VALUES_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_VALUES_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_ARRAY_cast" "', argument " "1"" of type '" "TPML_DIGEST_VALUES_ARRAY *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES_ARRAY *)(argp1);
  result = (TPML_DIGEST_VALUES *)TPML_DIGEST_VALUES_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_VALUES_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  result = (TPML_DIGEST_VALUES_ARRAY *)TPML_DIGEST_VALUES_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_DIGEST_VALUES_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_DIGEST_VALUES_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_VALUES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_DIGEST_VALUES_PTR")) SWIG_fail;
  result = (TPML_DIGEST_VALUES *)new_TPML_DIGEST_VALUES_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_DIGEST_VALUES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_DIGEST_VALUES_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_DIGEST_VALUES,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_DIGEST_VALUES_PTR" "', argument " "1"" of type '" "TPML_DIGEST_VALUES""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_DIGEST_VALUES_PTR" "', argument " "1"" of type '" "TPML_DIGEST_VALUES""'");
    } else {
      arg1 = *((TPML_DIGEST_VALUES *)(argp1));
    }
  }
  result = (TPML_DIGEST_VALUES *)copy_TPML_DIGEST_VALUES_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_VALUES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_VALUES_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_VALUES_PTR" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  delete_TPML_DIGEST_VALUES_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  TPML_DIGEST_VALUES arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_VALUES_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_PTR_assign" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_DIGEST_VALUES,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_DIGEST_VALUES_PTR_assign" "', argument " "2"" of type '" "TPML_DIGEST_VALUES""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_DIGEST_VALUES_PTR_assign" "', argument " "2"" of type '" "TPML_DIGEST_VALUES""'");
    } else {
      arg2 = *((TPML_DIGEST_VALUES *)(argp2));
    }
  }
  TPML_DIGEST_VALUES_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_VALUES_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_PTR_value" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  result = TPML_DIGEST_VALUES_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_DIGEST_VALUES *)memcpy((TPML_DIGEST_VALUES *)malloc(sizeof(TPML_DIGEST_VALUES)),&result,sizeof(TPML_DIGEST_VALUES)), SWIGTYPE_p_TPML_DIGEST_VALUES, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_DIGEST_VALUES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_DIGEST_VALUES_PTR_PTR")) SWIG_fail;
  result = (TPML_DIGEST_VALUES **)new_TPML_DIGEST_VALUES_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_DIGEST_VALUES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_DIGEST_VALUES_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_DIGEST_VALUES_PTR_PTR" "', argument " "1"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  result = (TPML_DIGEST_VALUES **)copy_TPML_DIGEST_VALUES_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_DIGEST_VALUES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES **arg1 = (TPML_DIGEST_VALUES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_DIGEST_VALUES_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_DIGEST_VALUES_PTR_PTR" "', argument " "1"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES **)(argp1);
  delete_TPML_DIGEST_VALUES_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES **arg1 = (TPML_DIGEST_VALUES **) 0 ;
  TPML_DIGEST_VALUES *arg2 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_DIGEST_VALUES_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_DIGEST_VALUES_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg2 = (TPML_DIGEST_VALUES *)(argp2);
  TPML_DIGEST_VALUES_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_DIGEST_VALUES_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES **arg1 = (TPML_DIGEST_VALUES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_DIGEST_VALUES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_DIGEST_VALUES_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_DIGEST_VALUES_PTR_PTR_value" "', argument " "1"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES **)(argp1);
  result = (TPML_DIGEST_VALUES *)TPML_DIGEST_VALUES_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_PCR_SELECTION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_PCR_SELECTION",SWIG_From_size_t((size_t)(sizeof(TPML_PCR_SELECTION))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_PCR_SELECTION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_PCR_SELECTION_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_PCR_SELECTION_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_PCR_SELECTION_ARRAY *)new_TPML_PCR_SELECTION_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_PCR_SELECTION_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION_ARRAY *arg1 = (TPML_PCR_SELECTION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_PCR_SELECTION_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_PCR_SELECTION_ARRAY" "', argument " "1"" of type '" "TPML_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION_ARRAY *)(argp1);
  delete_TPML_PCR_SELECTION_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION_ARRAY *arg1 = (TPML_PCR_SELECTION_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_PCR_SELECTION result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_PCR_SELECTION_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_PCR_SELECTION_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_PCR_SELECTION_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_PCR_SELECTION *)memcpy((TPML_PCR_SELECTION *)malloc(sizeof(TPML_PCR_SELECTION)),&result,sizeof(TPML_PCR_SELECTION)), SWIGTYPE_p_TPML_PCR_SELECTION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION_ARRAY *arg1 = (TPML_PCR_SELECTION_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_PCR_SELECTION arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_PCR_SELECTION_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_PCR_SELECTION_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_PCR_SELECTION,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_PCR_SELECTION_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_PCR_SELECTION""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_PCR_SELECTION_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_PCR_SELECTION""'");
    } else {
      arg3 = *((TPML_PCR_SELECTION *)(argp3));
    }
  }
  TPML_PCR_SELECTION_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION_ARRAY *arg1 = (TPML_PCR_SELECTION_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_PCR_SELECTION_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_ARRAY_cast" "', argument " "1"" of type '" "TPML_PCR_SELECTION_ARRAY *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION_ARRAY *)(argp1);
  result = (TPML_PCR_SELECTION *)TPML_PCR_SELECTION_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_PCR_SELECTION_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  result = (TPML_PCR_SELECTION_ARRAY *)TPML_PCR_SELECTION_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_PCR_SELECTION_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_PCR_SELECTION_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_PCR_SELECTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_PCR_SELECTION_PTR")) SWIG_fail;
  result = (TPML_PCR_SELECTION *)new_TPML_PCR_SELECTION_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_PCR_SELECTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_PCR_SELECTION_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_PCR_SELECTION,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_PCR_SELECTION_PTR" "', argument " "1"" of type '" "TPML_PCR_SELECTION""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_PCR_SELECTION_PTR" "', argument " "1"" of type '" "TPML_PCR_SELECTION""'");
    } else {
      arg1 = *((TPML_PCR_SELECTION *)(argp1));
    }
  }
  result = (TPML_PCR_SELECTION *)copy_TPML_PCR_SELECTION_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_PCR_SELECTION_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_PCR_SELECTION_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_PCR_SELECTION_PTR" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  delete_TPML_PCR_SELECTION_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  TPML_PCR_SELECTION arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_PCR_SELECTION_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_PTR_assign" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_PCR_SELECTION,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_PCR_SELECTION_PTR_assign" "', argument " "2"" of type '" "TPML_PCR_SELECTION""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_PCR_SELECTION_PTR_assign" "', argument " "2"" of type '" "TPML_PCR_SELECTION""'");
    } else {
      arg2 = *((TPML_PCR_SELECTION *)(argp2));
    }
  }
  TPML_PCR_SELECTION_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_PCR_SELECTION_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_PTR_value" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  result = TPML_PCR_SELECTION_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_PCR_SELECTION *)memcpy((TPML_PCR_SELECTION *)malloc(sizeof(TPML_PCR_SELECTION)),&result,sizeof(TPML_PCR_SELECTION)), SWIGTYPE_p_TPML_PCR_SELECTION, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_PCR_SELECTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_PCR_SELECTION_PTR_PTR")) SWIG_fail;
  result = (TPML_PCR_SELECTION **)new_TPML_PCR_SELECTION_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_PCR_SELECTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_PCR_SELECTION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_PCR_SELECTION_PTR_PTR" "', argument " "1"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  result = (TPML_PCR_SELECTION **)copy_TPML_PCR_SELECTION_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_PCR_SELECTION_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION **arg1 = (TPML_PCR_SELECTION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_PCR_SELECTION_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_PCR_SELECTION_PTR_PTR" "', argument " "1"" of type '" "TPML_PCR_SELECTION **""'"); 
  }
  arg1 = (TPML_PCR_SELECTION **)(argp1);
  delete_TPML_PCR_SELECTION_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION **arg1 = (TPML_PCR_SELECTION **) 0 ;
  TPML_PCR_SELECTION *arg2 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_PCR_SELECTION_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_PCR_SELECTION **""'"); 
  }
  arg1 = (TPML_PCR_SELECTION **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_PCR_SELECTION_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg2 = (TPML_PCR_SELECTION *)(argp2);
  TPML_PCR_SELECTION_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_PCR_SELECTION_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION **arg1 = (TPML_PCR_SELECTION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_PCR_SELECTION *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_PCR_SELECTION_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_PCR_SELECTION_PTR_PTR_value" "', argument " "1"" of type '" "TPML_PCR_SELECTION **""'"); 
  }
  arg1 = (TPML_PCR_SELECTION **)(argp1);
  result = (TPML_PCR_SELECTION *)TPML_PCR_SELECTION_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_ALG_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_ALG_PROPERTY",SWIG_From_size_t((size_t)(sizeof(TPML_ALG_PROPERTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_ALG_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_ALG_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_ALG_PROPERTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_ALG_PROPERTY_ARRAY *)new_TPML_ALG_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY_ARRAY *arg1 = (TPML_ALG_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_PROPERTY_ARRAY" "', argument " "1"" of type '" "TPML_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY_ARRAY *)(argp1);
  delete_TPML_ALG_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY_ARRAY *arg1 = (TPML_ALG_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_ALG_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PROPERTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ALG_PROPERTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_ALG_PROPERTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_ALG_PROPERTY *)memcpy((TPML_ALG_PROPERTY *)malloc(sizeof(TPML_ALG_PROPERTY)),&result,sizeof(TPML_ALG_PROPERTY)), SWIGTYPE_p_TPML_ALG_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY_ARRAY *arg1 = (TPML_ALG_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_ALG_PROPERTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_ALG_PROPERTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ALG_PROPERTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_ALG_PROPERTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_ALG_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_ALG_PROPERTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_ALG_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_ALG_PROPERTY""'");
    } else {
      arg3 = *((TPML_ALG_PROPERTY *)(argp3));
    }
  }
  TPML_ALG_PROPERTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY_ARRAY *arg1 = (TPML_ALG_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PROPERTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_ARRAY_cast" "', argument " "1"" of type '" "TPML_ALG_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY_ARRAY *)(argp1);
  result = (TPML_ALG_PROPERTY *)TPML_ALG_PROPERTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PROPERTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  result = (TPML_ALG_PROPERTY_ARRAY *)TPML_ALG_PROPERTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_ALG_PROPERTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_ALG_PROPERTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_ALG_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ALG_PROPERTY_PTR")) SWIG_fail;
  result = (TPML_ALG_PROPERTY *)new_TPML_ALG_PROPERTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_ALG_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_ALG_PROPERTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_ALG_PROPERTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_ALG_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_ALG_PROPERTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_ALG_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_ALG_PROPERTY""'");
    } else {
      arg1 = *((TPML_ALG_PROPERTY *)(argp1));
    }
  }
  result = (TPML_ALG_PROPERTY *)copy_TPML_ALG_PROPERTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_PROPERTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  delete_TPML_ALG_PROPERTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  TPML_ALG_PROPERTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PROPERTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_PTR_assign" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_ALG_PROPERTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ALG_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_ALG_PROPERTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_ALG_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_ALG_PROPERTY""'");
    } else {
      arg2 = *((TPML_ALG_PROPERTY *)(argp2));
    }
  }
  TPML_ALG_PROPERTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PROPERTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_PTR_value" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  result = TPML_ALG_PROPERTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_ALG_PROPERTY *)memcpy((TPML_ALG_PROPERTY *)malloc(sizeof(TPML_ALG_PROPERTY)),&result,sizeof(TPML_ALG_PROPERTY)), SWIGTYPE_p_TPML_ALG_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_ALG_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ALG_PROPERTY_PTR_PTR")) SWIG_fail;
  result = (TPML_ALG_PROPERTY **)new_TPML_ALG_PROPERTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_ALG_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_ALG_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_ALG_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  result = (TPML_ALG_PROPERTY **)copy_TPML_ALG_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ALG_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY **arg1 = (TPML_ALG_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ALG_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ALG_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_ALG_PROPERTY **""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY **)(argp1);
  delete_TPML_ALG_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY **arg1 = (TPML_ALG_PROPERTY **) 0 ;
  TPML_ALG_PROPERTY *arg2 = (TPML_ALG_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ALG_PROPERTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_ALG_PROPERTY **""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ALG_PROPERTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg2 = (TPML_ALG_PROPERTY *)(argp2);
  TPML_ALG_PROPERTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ALG_PROPERTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY **arg1 = (TPML_ALG_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ALG_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ALG_PROPERTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ALG_PROPERTY_PTR_PTR_value" "', argument " "1"" of type '" "TPML_ALG_PROPERTY **""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY **)(argp1);
  result = (TPML_ALG_PROPERTY *)TPML_ALG_PROPERTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_TAGGED_TPM_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_TAGGED_TPM_PROPERTY",SWIG_From_size_t((size_t)(sizeof(TPML_TAGGED_TPM_PROPERTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_TPM_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_TAGGED_TPM_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_TAGGED_TPM_PROPERTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_TAGGED_TPM_PROPERTY_ARRAY *)new_TPML_TAGGED_TPM_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_TPM_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_TPM_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_TPM_PROPERTY_ARRAY" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *)(argp1);
  delete_TPML_TAGGED_TPM_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_TAGGED_TPM_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_TAGGED_TPM_PROPERTY *)memcpy((TPML_TAGGED_TPM_PROPERTY *)malloc(sizeof(TPML_TAGGED_TPM_PROPERTY)),&result,sizeof(TPML_TAGGED_TPM_PROPERTY)), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_TAGGED_TPM_PROPERTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_TAGGED_TPM_PROPERTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_TAGGED_TPM_PROPERTY""'");
    } else {
      arg3 = *((TPML_TAGGED_TPM_PROPERTY *)(argp3));
    }
  }
  TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_TPM_PROPERTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY_cast" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY_ARRAY *)(argp1);
  result = (TPML_TAGGED_TPM_PROPERTY *)TPML_TAGGED_TPM_PROPERTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  result = (TPML_TAGGED_TPM_PROPERTY_ARRAY *)TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_TAGGED_TPM_PROPERTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_TPM_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_TAGGED_TPM_PROPERTY_PTR")) SWIG_fail;
  result = (TPML_TAGGED_TPM_PROPERTY *)new_TPML_TAGGED_TPM_PROPERTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_TAGGED_TPM_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_TAGGED_TPM_PROPERTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_TAGGED_TPM_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_TAGGED_TPM_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY""'");
    } else {
      arg1 = *((TPML_TAGGED_TPM_PROPERTY *)(argp1));
    }
  }
  result = (TPML_TAGGED_TPM_PROPERTY *)copy_TPML_TAGGED_TPM_PROPERTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_TPM_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_TPM_PROPERTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_TPM_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  delete_TPML_TAGGED_TPM_PROPERTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  TPML_TAGGED_TPM_PROPERTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_TPM_PROPERTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_assign" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_TAGGED_TPM_PROPERTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_TAGGED_TPM_PROPERTY""'");
    } else {
      arg2 = *((TPML_TAGGED_TPM_PROPERTY *)(argp2));
    }
  }
  TPML_TAGGED_TPM_PROPERTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_TPM_PROPERTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_value" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  result = TPML_TAGGED_TPM_PROPERTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_TAGGED_TPM_PROPERTY *)memcpy((TPML_TAGGED_TPM_PROPERTY *)malloc(sizeof(TPML_TAGGED_TPM_PROPERTY)),&result,sizeof(TPML_TAGGED_TPM_PROPERTY)), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_TAGGED_TPM_PROPERTY_PTR_PTR")) SWIG_fail;
  result = (TPML_TAGGED_TPM_PROPERTY **)new_TPML_TAGGED_TPM_PROPERTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  result = (TPML_TAGGED_TPM_PROPERTY **)copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY **arg1 = (TPML_TAGGED_TPM_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY **""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY **)(argp1);
  delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY **arg1 = (TPML_TAGGED_TPM_PROPERTY **) 0 ;
  TPML_TAGGED_TPM_PROPERTY *arg2 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY **""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg2 = (TPML_TAGGED_TPM_PROPERTY *)(argp2);
  TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY **arg1 = (TPML_TAGGED_TPM_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_TPM_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY **""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY **)(argp1);
  result = (TPML_TAGGED_TPM_PROPERTY *)TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_TAGGED_PCR_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_TAGGED_PCR_PROPERTY",SWIG_From_size_t((size_t)(sizeof(TPML_TAGGED_PCR_PROPERTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_PCR_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_TAGGED_PCR_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_TAGGED_PCR_PROPERTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_TAGGED_PCR_PROPERTY_ARRAY *)new_TPML_TAGGED_PCR_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_PCR_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_PCR_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_PCR_PROPERTY_ARRAY" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *)(argp1);
  delete_TPML_TAGGED_PCR_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_TAGGED_PCR_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_TAGGED_PCR_PROPERTY *)memcpy((TPML_TAGGED_PCR_PROPERTY *)malloc(sizeof(TPML_TAGGED_PCR_PROPERTY)),&result,sizeof(TPML_TAGGED_PCR_PROPERTY)), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_TAGGED_PCR_PROPERTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_TAGGED_PCR_PROPERTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_TAGGED_PCR_PROPERTY""'");
    } else {
      arg3 = *((TPML_TAGGED_PCR_PROPERTY *)(argp3));
    }
  }
  TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY_ARRAY *arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_PCR_PROPERTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY_cast" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY_ARRAY *)(argp1);
  result = (TPML_TAGGED_PCR_PROPERTY *)TPML_TAGGED_PCR_PROPERTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  result = (TPML_TAGGED_PCR_PROPERTY_ARRAY *)TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_TAGGED_PCR_PROPERTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_PCR_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_TAGGED_PCR_PROPERTY_PTR")) SWIG_fail;
  result = (TPML_TAGGED_PCR_PROPERTY *)new_TPML_TAGGED_PCR_PROPERTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_TAGGED_PCR_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_TAGGED_PCR_PROPERTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_TAGGED_PCR_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_TAGGED_PCR_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY""'");
    } else {
      arg1 = *((TPML_TAGGED_PCR_PROPERTY *)(argp1));
    }
  }
  result = (TPML_TAGGED_PCR_PROPERTY *)copy_TPML_TAGGED_PCR_PROPERTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_PCR_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_PCR_PROPERTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_PCR_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  delete_TPML_TAGGED_PCR_PROPERTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  TPML_TAGGED_PCR_PROPERTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_PCR_PROPERTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_assign" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_TAGGED_PCR_PROPERTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_TAGGED_PCR_PROPERTY""'");
    } else {
      arg2 = *((TPML_TAGGED_PCR_PROPERTY *)(argp2));
    }
  }
  TPML_TAGGED_PCR_PROPERTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_PCR_PROPERTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_value" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  result = TPML_TAGGED_PCR_PROPERTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_TAGGED_PCR_PROPERTY *)memcpy((TPML_TAGGED_PCR_PROPERTY *)malloc(sizeof(TPML_TAGGED_PCR_PROPERTY)),&result,sizeof(TPML_TAGGED_PCR_PROPERTY)), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_TAGGED_PCR_PROPERTY_PTR_PTR")) SWIG_fail;
  result = (TPML_TAGGED_PCR_PROPERTY **)new_TPML_TAGGED_PCR_PROPERTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  result = (TPML_TAGGED_PCR_PROPERTY **)copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY **arg1 = (TPML_TAGGED_PCR_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY **""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY **)(argp1);
  delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY **arg1 = (TPML_TAGGED_PCR_PROPERTY **) 0 ;
  TPML_TAGGED_PCR_PROPERTY *arg2 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY **""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg2 = (TPML_TAGGED_PCR_PROPERTY *)(argp2);
  TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY **arg1 = (TPML_TAGGED_PCR_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_TAGGED_PCR_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY **""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY **)(argp1);
  result = (TPML_TAGGED_PCR_PROPERTY *)TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_ECC_CURVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_ECC_CURVE",SWIG_From_size_t((size_t)(sizeof(TPML_ECC_CURVE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_ECC_CURVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_ECC_CURVE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_ECC_CURVE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_ECC_CURVE_ARRAY *)new_TPML_ECC_CURVE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ECC_CURVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE_ARRAY *arg1 = (TPML_ECC_CURVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ECC_CURVE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ECC_CURVE_ARRAY" "', argument " "1"" of type '" "TPML_ECC_CURVE_ARRAY *""'"); 
  }
  arg1 = (TPML_ECC_CURVE_ARRAY *)(argp1);
  delete_TPML_ECC_CURVE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE_ARRAY *arg1 = (TPML_ECC_CURVE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_ECC_CURVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ECC_CURVE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_ECC_CURVE_ARRAY *""'"); 
  }
  arg1 = (TPML_ECC_CURVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ECC_CURVE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_ECC_CURVE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_ECC_CURVE *)memcpy((TPML_ECC_CURVE *)malloc(sizeof(TPML_ECC_CURVE)),&result,sizeof(TPML_ECC_CURVE)), SWIGTYPE_p_TPML_ECC_CURVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE_ARRAY *arg1 = (TPML_ECC_CURVE_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_ECC_CURVE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_ECC_CURVE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_ECC_CURVE_ARRAY *""'"); 
  }
  arg1 = (TPML_ECC_CURVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_ECC_CURVE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_ECC_CURVE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_ECC_CURVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_ECC_CURVE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_ECC_CURVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_ECC_CURVE""'");
    } else {
      arg3 = *((TPML_ECC_CURVE *)(argp3));
    }
  }
  TPML_ECC_CURVE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE_ARRAY *arg1 = (TPML_ECC_CURVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ECC_CURVE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_ARRAY_cast" "', argument " "1"" of type '" "TPML_ECC_CURVE_ARRAY *""'"); 
  }
  arg1 = (TPML_ECC_CURVE_ARRAY *)(argp1);
  result = (TPML_ECC_CURVE *)TPML_ECC_CURVE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ECC_CURVE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  result = (TPML_ECC_CURVE_ARRAY *)TPML_ECC_CURVE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_ECC_CURVE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_ECC_CURVE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_ECC_CURVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ECC_CURVE_PTR")) SWIG_fail;
  result = (TPML_ECC_CURVE *)new_TPML_ECC_CURVE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_ECC_CURVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_ECC_CURVE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_ECC_CURVE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_ECC_CURVE_PTR" "', argument " "1"" of type '" "TPML_ECC_CURVE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_ECC_CURVE_PTR" "', argument " "1"" of type '" "TPML_ECC_CURVE""'");
    } else {
      arg1 = *((TPML_ECC_CURVE *)(argp1));
    }
  }
  result = (TPML_ECC_CURVE *)copy_TPML_ECC_CURVE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ECC_CURVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ECC_CURVE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ECC_CURVE_PTR" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  delete_TPML_ECC_CURVE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  TPML_ECC_CURVE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ECC_CURVE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_PTR_assign" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_ECC_CURVE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ECC_CURVE_PTR_assign" "', argument " "2"" of type '" "TPML_ECC_CURVE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_ECC_CURVE_PTR_assign" "', argument " "2"" of type '" "TPML_ECC_CURVE""'");
    } else {
      arg2 = *((TPML_ECC_CURVE *)(argp2));
    }
  }
  TPML_ECC_CURVE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ECC_CURVE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_PTR_value" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  result = TPML_ECC_CURVE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_ECC_CURVE *)memcpy((TPML_ECC_CURVE *)malloc(sizeof(TPML_ECC_CURVE)),&result,sizeof(TPML_ECC_CURVE)), SWIGTYPE_p_TPML_ECC_CURVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_ECC_CURVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_ECC_CURVE_PTR_PTR")) SWIG_fail;
  result = (TPML_ECC_CURVE **)new_TPML_ECC_CURVE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_ECC_CURVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_ECC_CURVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_ECC_CURVE_PTR_PTR" "', argument " "1"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  result = (TPML_ECC_CURVE **)copy_TPML_ECC_CURVE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_ECC_CURVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE **arg1 = (TPML_ECC_CURVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_ECC_CURVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_ECC_CURVE_PTR_PTR" "', argument " "1"" of type '" "TPML_ECC_CURVE **""'"); 
  }
  arg1 = (TPML_ECC_CURVE **)(argp1);
  delete_TPML_ECC_CURVE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE **arg1 = (TPML_ECC_CURVE **) 0 ;
  TPML_ECC_CURVE *arg2 = (TPML_ECC_CURVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_ECC_CURVE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_ECC_CURVE **""'"); 
  }
  arg1 = (TPML_ECC_CURVE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_ECC_CURVE_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg2 = (TPML_ECC_CURVE *)(argp2);
  TPML_ECC_CURVE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_ECC_CURVE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE **arg1 = (TPML_ECC_CURVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_ECC_CURVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_ECC_CURVE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_ECC_CURVE_PTR_PTR_value" "', argument " "1"" of type '" "TPML_ECC_CURVE **""'"); 
  }
  arg1 = (TPML_ECC_CURVE **)(argp1);
  result = (TPML_ECC_CURVE *)TPML_ECC_CURVE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_INTEL_PTT_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_INTEL_PTT_PROPERTY",SWIG_From_size_t((size_t)(sizeof(TPML_INTEL_PTT_PROPERTY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_INTEL_PTT_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_INTEL_PTT_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_INTEL_PTT_PROPERTY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_INTEL_PTT_PROPERTY_ARRAY *)new_TPML_INTEL_PTT_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_INTEL_PTT_PROPERTY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY_ARRAY *arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_INTEL_PTT_PROPERTY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_INTEL_PTT_PROPERTY_ARRAY" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *)(argp1);
  delete_TPML_INTEL_PTT_PROPERTY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY_ARRAY *arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_INTEL_PTT_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_INTEL_PTT_PROPERTY *)memcpy((TPML_INTEL_PTT_PROPERTY *)malloc(sizeof(TPML_INTEL_PTT_PROPERTY)),&result,sizeof(TPML_INTEL_PTT_PROPERTY)), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY_ARRAY *arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_INTEL_PTT_PROPERTY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_INTEL_PTT_PROPERTY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_INTEL_PTT_PROPERTY""'");
    } else {
      arg3 = *((TPML_INTEL_PTT_PROPERTY *)(argp3));
    }
  }
  TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY_ARRAY *arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_INTEL_PTT_PROPERTY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY_cast" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY_ARRAY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY_ARRAY *)(argp1);
  result = (TPML_INTEL_PTT_PROPERTY *)TPML_INTEL_PTT_PROPERTY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  result = (TPML_INTEL_PTT_PROPERTY_ARRAY *)TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_INTEL_PTT_PROPERTY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_INTEL_PTT_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_INTEL_PTT_PROPERTY_PTR")) SWIG_fail;
  result = (TPML_INTEL_PTT_PROPERTY *)new_TPML_INTEL_PTT_PROPERTY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_INTEL_PTT_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_INTEL_PTT_PROPERTY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_INTEL_PTT_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_INTEL_PTT_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY""'");
    } else {
      arg1 = *((TPML_INTEL_PTT_PROPERTY *)(argp1));
    }
  }
  result = (TPML_INTEL_PTT_PROPERTY *)copy_TPML_INTEL_PTT_PROPERTY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_INTEL_PTT_PROPERTY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_INTEL_PTT_PROPERTY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_INTEL_PTT_PROPERTY_PTR" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  delete_TPML_INTEL_PTT_PROPERTY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  TPML_INTEL_PTT_PROPERTY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_INTEL_PTT_PROPERTY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_assign" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_INTEL_PTT_PROPERTY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_assign" "', argument " "2"" of type '" "TPML_INTEL_PTT_PROPERTY""'");
    } else {
      arg2 = *((TPML_INTEL_PTT_PROPERTY *)(argp2));
    }
  }
  TPML_INTEL_PTT_PROPERTY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_INTEL_PTT_PROPERTY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_value" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  result = TPML_INTEL_PTT_PROPERTY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_INTEL_PTT_PROPERTY *)memcpy((TPML_INTEL_PTT_PROPERTY *)malloc(sizeof(TPML_INTEL_PTT_PROPERTY)),&result,sizeof(TPML_INTEL_PTT_PROPERTY)), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_INTEL_PTT_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_INTEL_PTT_PROPERTY_PTR_PTR")) SWIG_fail;
  result = (TPML_INTEL_PTT_PROPERTY **)new_TPML_INTEL_PTT_PROPERTY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  result = (TPML_INTEL_PTT_PROPERTY **)copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY **arg1 = (TPML_INTEL_PTT_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY **""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY **)(argp1);
  delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY **arg1 = (TPML_INTEL_PTT_PROPERTY **) 0 ;
  TPML_INTEL_PTT_PROPERTY *arg2 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY **""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg2 = (TPML_INTEL_PTT_PROPERTY *)(argp2);
  TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_INTEL_PTT_PROPERTY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY **arg1 = (TPML_INTEL_PTT_PROPERTY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_INTEL_PTT_PROPERTY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_INTEL_PTT_PROPERTY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_INTEL_PTT_PROPERTY_PTR_PTR_value" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY **""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY **)(argp1);
  result = (TPML_INTEL_PTT_PROPERTY *)TPML_INTEL_PTT_PROPERTY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_CAPABILITIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_CAPABILITIES",SWIG_From_size_t((size_t)(sizeof(TPMU_CAPABILITIES))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_CAPABILITIES_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_CAPABILITIES_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_CAPABILITIES_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_CAPABILITIES_ARRAY *)new_TPMU_CAPABILITIES_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_CAPABILITIES_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES_ARRAY *arg1 = (TPMU_CAPABILITIES_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_CAPABILITIES_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_CAPABILITIES_ARRAY" "', argument " "1"" of type '" "TPMU_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES_ARRAY *)(argp1);
  delete_TPMU_CAPABILITIES_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES_ARRAY *arg1 = (TPMU_CAPABILITIES_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_CAPABILITIES result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_CAPABILITIES_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_CAPABILITIES_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_CAPABILITIES *)memcpy((TPMU_CAPABILITIES *)malloc(sizeof(TPMU_CAPABILITIES)),&result,sizeof(TPMU_CAPABILITIES)), SWIGTYPE_p_TPMU_CAPABILITIES, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES_ARRAY *arg1 = (TPMU_CAPABILITIES_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_CAPABILITIES arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_CAPABILITIES_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_CAPABILITIES_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_CAPABILITIES,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_CAPABILITIES_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_CAPABILITIES""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_CAPABILITIES_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_CAPABILITIES""'");
    } else {
      arg3 = *((TPMU_CAPABILITIES *)(argp3));
    }
  }
  TPMU_CAPABILITIES_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES_ARRAY *arg1 = (TPMU_CAPABILITIES_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_ARRAY_cast" "', argument " "1"" of type '" "TPMU_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES_ARRAY *)(argp1);
  result = (TPMU_CAPABILITIES *)TPMU_CAPABILITIES_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPMU_CAPABILITIES_ARRAY *)TPMU_CAPABILITIES_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_CAPABILITIES_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_CAPABILITIES_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_CAPABILITIES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_CAPABILITIES_PTR")) SWIG_fail;
  result = (TPMU_CAPABILITIES *)new_TPMU_CAPABILITIES_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_CAPABILITIES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_CAPABILITIES_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_CAPABILITIES,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_CAPABILITIES_PTR" "', argument " "1"" of type '" "TPMU_CAPABILITIES""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_CAPABILITIES_PTR" "', argument " "1"" of type '" "TPMU_CAPABILITIES""'");
    } else {
      arg1 = *((TPMU_CAPABILITIES *)(argp1));
    }
  }
  result = (TPMU_CAPABILITIES *)copy_TPMU_CAPABILITIES_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_CAPABILITIES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_CAPABILITIES_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_CAPABILITIES_PTR" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  delete_TPMU_CAPABILITIES_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  TPMU_CAPABILITIES arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_PTR_assign" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_CAPABILITIES,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_PTR_assign" "', argument " "2"" of type '" "TPMU_CAPABILITIES""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_CAPABILITIES_PTR_assign" "', argument " "2"" of type '" "TPMU_CAPABILITIES""'");
    } else {
      arg2 = *((TPMU_CAPABILITIES *)(argp2));
    }
  }
  TPMU_CAPABILITIES_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_PTR_value" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = TPMU_CAPABILITIES_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_CAPABILITIES *)memcpy((TPMU_CAPABILITIES *)malloc(sizeof(TPMU_CAPABILITIES)),&result,sizeof(TPMU_CAPABILITIES)), SWIGTYPE_p_TPMU_CAPABILITIES, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_CAPABILITIES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_CAPABILITIES_PTR_PTR")) SWIG_fail;
  result = (TPMU_CAPABILITIES **)new_TPMU_CAPABILITIES_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_CAPABILITIES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_CAPABILITIES_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_CAPABILITIES_PTR_PTR" "', argument " "1"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  result = (TPMU_CAPABILITIES **)copy_TPMU_CAPABILITIES_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_CAPABILITIES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES **arg1 = (TPMU_CAPABILITIES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_CAPABILITIES_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_CAPABILITIES_PTR_PTR" "', argument " "1"" of type '" "TPMU_CAPABILITIES **""'"); 
  }
  arg1 = (TPMU_CAPABILITIES **)(argp1);
  delete_TPMU_CAPABILITIES_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES **arg1 = (TPMU_CAPABILITIES **) 0 ;
  TPMU_CAPABILITIES *arg2 = (TPMU_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_CAPABILITIES_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_CAPABILITIES **""'"); 
  }
  arg1 = (TPMU_CAPABILITIES **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_CAPABILITIES_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg2 = (TPMU_CAPABILITIES *)(argp2);
  TPMU_CAPABILITIES_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_CAPABILITIES_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES **arg1 = (TPMU_CAPABILITIES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_CAPABILITIES_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_CAPABILITIES_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_CAPABILITIES **""'"); 
  }
  arg1 = (TPMU_CAPABILITIES **)(argp1);
  result = (TPMU_CAPABILITIES *)TPMU_CAPABILITIES_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CAPABILITY_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CAPABILITY_DATA",SWIG_From_size_t((size_t)(sizeof(TPMS_CAPABILITY_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CAPABILITY_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CAPABILITY_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CAPABILITY_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CAPABILITY_DATA_ARRAY *)new_TPMS_CAPABILITY_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CAPABILITY_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA_ARRAY *arg1 = (TPMS_CAPABILITY_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CAPABILITY_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CAPABILITY_DATA_ARRAY" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA_ARRAY *)(argp1);
  delete_TPMS_CAPABILITY_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA_ARRAY *arg1 = (TPMS_CAPABILITY_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CAPABILITY_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CAPABILITY_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CAPABILITY_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CAPABILITY_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CAPABILITY_DATA *)memcpy((TPMS_CAPABILITY_DATA *)malloc(sizeof(TPMS_CAPABILITY_DATA)),&result,sizeof(TPMS_CAPABILITY_DATA)), SWIGTYPE_p_TPMS_CAPABILITY_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA_ARRAY *arg1 = (TPMS_CAPABILITY_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CAPABILITY_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CAPABILITY_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CAPABILITY_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CAPABILITY_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CAPABILITY_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CAPABILITY_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CAPABILITY_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CAPABILITY_DATA""'");
    } else {
      arg3 = *((TPMS_CAPABILITY_DATA *)(argp3));
    }
  }
  TPMS_CAPABILITY_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA_ARRAY *arg1 = (TPMS_CAPABILITY_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CAPABILITY_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA_ARRAY *)(argp1);
  result = (TPMS_CAPABILITY_DATA *)TPMS_CAPABILITY_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CAPABILITY_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  result = (TPMS_CAPABILITY_DATA_ARRAY *)TPMS_CAPABILITY_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CAPABILITY_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CAPABILITY_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CAPABILITY_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CAPABILITY_DATA_PTR")) SWIG_fail;
  result = (TPMS_CAPABILITY_DATA *)new_TPMS_CAPABILITY_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CAPABILITY_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CAPABILITY_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CAPABILITY_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CAPABILITY_DATA_PTR" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CAPABILITY_DATA_PTR" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA""'");
    } else {
      arg1 = *((TPMS_CAPABILITY_DATA *)(argp1));
    }
  }
  result = (TPMS_CAPABILITY_DATA *)copy_TPMS_CAPABILITY_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CAPABILITY_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CAPABILITY_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CAPABILITY_DATA_PTR" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  delete_TPMS_CAPABILITY_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  TPMS_CAPABILITY_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CAPABILITY_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_PTR_assign" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CAPABILITY_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CAPABILITY_DATA_PTR_assign" "', argument " "2"" of type '" "TPMS_CAPABILITY_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CAPABILITY_DATA_PTR_assign" "', argument " "2"" of type '" "TPMS_CAPABILITY_DATA""'");
    } else {
      arg2 = *((TPMS_CAPABILITY_DATA *)(argp2));
    }
  }
  TPMS_CAPABILITY_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CAPABILITY_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_PTR_value" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  result = TPMS_CAPABILITY_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CAPABILITY_DATA *)memcpy((TPMS_CAPABILITY_DATA *)malloc(sizeof(TPMS_CAPABILITY_DATA)),&result,sizeof(TPMS_CAPABILITY_DATA)), SWIGTYPE_p_TPMS_CAPABILITY_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CAPABILITY_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CAPABILITY_DATA_PTR_PTR")) SWIG_fail;
  result = (TPMS_CAPABILITY_DATA **)new_TPMS_CAPABILITY_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CAPABILITY_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CAPABILITY_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CAPABILITY_DATA_PTR_PTR" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  result = (TPMS_CAPABILITY_DATA **)copy_TPMS_CAPABILITY_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CAPABILITY_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA **arg1 = (TPMS_CAPABILITY_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CAPABILITY_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CAPABILITY_DATA_PTR_PTR" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA **""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA **)(argp1);
  delete_TPMS_CAPABILITY_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA **arg1 = (TPMS_CAPABILITY_DATA **) 0 ;
  TPMS_CAPABILITY_DATA *arg2 = (TPMS_CAPABILITY_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CAPABILITY_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA **""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CAPABILITY_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg2 = (TPMS_CAPABILITY_DATA *)(argp2);
  TPMS_CAPABILITY_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CAPABILITY_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA **arg1 = (TPMS_CAPABILITY_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CAPABILITY_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CAPABILITY_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CAPABILITY_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA **""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA **)(argp1);
  result = (TPMS_CAPABILITY_DATA *)TPMS_CAPABILITY_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CLOCK_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CLOCK_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_CLOCK_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CLOCK_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CLOCK_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CLOCK_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CLOCK_INFO_ARRAY *)new_TPMS_CLOCK_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CLOCK_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO_ARRAY *arg1 = (TPMS_CLOCK_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CLOCK_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CLOCK_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_CLOCK_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO_ARRAY *)(argp1);
  delete_TPMS_CLOCK_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO_ARRAY *arg1 = (TPMS_CLOCK_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CLOCK_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CLOCK_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CLOCK_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CLOCK_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CLOCK_INFO *)memcpy((TPMS_CLOCK_INFO *)malloc(sizeof(TPMS_CLOCK_INFO)),&result,sizeof(TPMS_CLOCK_INFO)), SWIGTYPE_p_TPMS_CLOCK_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO_ARRAY *arg1 = (TPMS_CLOCK_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CLOCK_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CLOCK_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CLOCK_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CLOCK_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CLOCK_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CLOCK_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CLOCK_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CLOCK_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CLOCK_INFO""'");
    } else {
      arg3 = *((TPMS_CLOCK_INFO *)(argp3));
    }
  }
  TPMS_CLOCK_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO_ARRAY *arg1 = (TPMS_CLOCK_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CLOCK_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO_ARRAY *)(argp1);
  result = (TPMS_CLOCK_INFO *)TPMS_CLOCK_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = (TPMS_CLOCK_INFO_ARRAY *)TPMS_CLOCK_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CLOCK_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CLOCK_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CLOCK_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CLOCK_INFO_PTR")) SWIG_fail;
  result = (TPMS_CLOCK_INFO *)new_TPMS_CLOCK_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CLOCK_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CLOCK_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CLOCK_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CLOCK_INFO_PTR" "', argument " "1"" of type '" "TPMS_CLOCK_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CLOCK_INFO_PTR" "', argument " "1"" of type '" "TPMS_CLOCK_INFO""'");
    } else {
      arg1 = *((TPMS_CLOCK_INFO *)(argp1));
    }
  }
  result = (TPMS_CLOCK_INFO *)copy_TPMS_CLOCK_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CLOCK_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CLOCK_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CLOCK_INFO_PTR" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  delete_TPMS_CLOCK_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  TPMS_CLOCK_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CLOCK_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CLOCK_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_CLOCK_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CLOCK_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_CLOCK_INFO""'");
    } else {
      arg2 = *((TPMS_CLOCK_INFO *)(argp2));
    }
  }
  TPMS_CLOCK_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = TPMS_CLOCK_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CLOCK_INFO *)memcpy((TPMS_CLOCK_INFO *)malloc(sizeof(TPMS_CLOCK_INFO)),&result,sizeof(TPMS_CLOCK_INFO)), SWIGTYPE_p_TPMS_CLOCK_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CLOCK_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CLOCK_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_CLOCK_INFO **)new_TPMS_CLOCK_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CLOCK_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CLOCK_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CLOCK_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  result = (TPMS_CLOCK_INFO **)copy_TPMS_CLOCK_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CLOCK_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO **arg1 = (TPMS_CLOCK_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CLOCK_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CLOCK_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_CLOCK_INFO **""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO **)(argp1);
  delete_TPMS_CLOCK_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO **arg1 = (TPMS_CLOCK_INFO **) 0 ;
  TPMS_CLOCK_INFO *arg2 = (TPMS_CLOCK_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CLOCK_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CLOCK_INFO **""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CLOCK_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg2 = (TPMS_CLOCK_INFO *)(argp2);
  TPMS_CLOCK_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CLOCK_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO **arg1 = (TPMS_CLOCK_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CLOCK_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CLOCK_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CLOCK_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CLOCK_INFO **""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO **)(argp1);
  result = (TPMS_CLOCK_INFO *)TPMS_CLOCK_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_TIME_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_TIME_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_TIME_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_TIME_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_TIME_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_TIME_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_TIME_INFO_ARRAY *)new_TPMS_TIME_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO_ARRAY *arg1 = (TPMS_TIME_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_TIME_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_INFO_ARRAY *)(argp1);
  delete_TPMS_TIME_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO_ARRAY *arg1 = (TPMS_TIME_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_TIME_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_TIME_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TIME_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_TIME_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_TIME_INFO *)memcpy((TPMS_TIME_INFO *)malloc(sizeof(TPMS_TIME_INFO)),&result,sizeof(TPMS_TIME_INFO)), SWIGTYPE_p_TPMS_TIME_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO_ARRAY *arg1 = (TPMS_TIME_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_TIME_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_TIME_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_TIME_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TIME_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_TIME_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_TIME_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TIME_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TIME_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TIME_INFO""'");
    } else {
      arg3 = *((TPMS_TIME_INFO *)(argp3));
    }
  }
  TPMS_TIME_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO_ARRAY *arg1 = (TPMS_TIME_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_TIME_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_INFO_ARRAY *)(argp1);
  result = (TPMS_TIME_INFO *)TPMS_TIME_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  result = (TPMS_TIME_INFO_ARRAY *)TPMS_TIME_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TIME_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TIME_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_TIME_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TIME_INFO_PTR")) SWIG_fail;
  result = (TPMS_TIME_INFO *)new_TPMS_TIME_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TIME_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TIME_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_TIME_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TIME_INFO_PTR" "', argument " "1"" of type '" "TPMS_TIME_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_TIME_INFO_PTR" "', argument " "1"" of type '" "TPMS_TIME_INFO""'");
    } else {
      arg1 = *((TPMS_TIME_INFO *)(argp1));
    }
  }
  result = (TPMS_TIME_INFO *)copy_TPMS_TIME_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_INFO_PTR" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  delete_TPMS_TIME_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  TPMS_TIME_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_TIME_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TIME_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_TIME_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TIME_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_TIME_INFO""'");
    } else {
      arg2 = *((TPMS_TIME_INFO *)(argp2));
    }
  }
  TPMS_TIME_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  result = TPMS_TIME_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_TIME_INFO *)memcpy((TPMS_TIME_INFO *)malloc(sizeof(TPMS_TIME_INFO)),&result,sizeof(TPMS_TIME_INFO)), SWIGTYPE_p_TPMS_TIME_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TIME_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TIME_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_TIME_INFO **)new_TPMS_TIME_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TIME_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TIME_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TIME_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  result = (TPMS_TIME_INFO **)copy_TPMS_TIME_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO **arg1 = (TPMS_TIME_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_TIME_INFO **""'"); 
  }
  arg1 = (TPMS_TIME_INFO **)(argp1);
  delete_TPMS_TIME_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO **arg1 = (TPMS_TIME_INFO **) 0 ;
  TPMS_TIME_INFO *arg2 = (TPMS_TIME_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_TIME_INFO **""'"); 
  }
  arg1 = (TPMS_TIME_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TIME_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg2 = (TPMS_TIME_INFO *)(argp2);
  TPMS_TIME_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO **arg1 = (TPMS_TIME_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_TIME_INFO **""'"); 
  }
  arg1 = (TPMS_TIME_INFO **)(argp1);
  result = (TPMS_TIME_INFO *)TPMS_TIME_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_TIME_ATTEST_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_TIME_ATTEST_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_TIME_ATTEST_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_TIME_ATTEST_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_TIME_ATTEST_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_TIME_ATTEST_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_TIME_ATTEST_INFO_ARRAY *)new_TPMS_TIME_ATTEST_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_ATTEST_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO_ARRAY *arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_ATTEST_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_ATTEST_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *)(argp1);
  delete_TPMS_TIME_ATTEST_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO_ARRAY *arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_TIME_ATTEST_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_ATTEST_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_TIME_ATTEST_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_TIME_ATTEST_INFO *)memcpy((TPMS_TIME_ATTEST_INFO *)malloc(sizeof(TPMS_TIME_ATTEST_INFO)),&result,sizeof(TPMS_TIME_ATTEST_INFO)), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO_ARRAY *arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_TIME_ATTEST_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_TIME_ATTEST_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_TIME_ATTEST_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TIME_ATTEST_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_TIME_ATTEST_INFO""'");
    } else {
      arg3 = *((TPMS_TIME_ATTEST_INFO *)(argp3));
    }
  }
  TPMS_TIME_ATTEST_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO_ARRAY *arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_ATTEST_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO_ARRAY *)(argp1);
  result = (TPMS_TIME_ATTEST_INFO *)TPMS_TIME_ATTEST_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_ATTEST_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  result = (TPMS_TIME_ATTEST_INFO_ARRAY *)TPMS_TIME_ATTEST_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_TIME_ATTEST_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_TIME_ATTEST_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_TIME_ATTEST_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TIME_ATTEST_INFO_PTR")) SWIG_fail;
  result = (TPMS_TIME_ATTEST_INFO *)new_TPMS_TIME_ATTEST_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TIME_ATTEST_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TIME_ATTEST_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_TIME_ATTEST_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TIME_ATTEST_INFO_PTR" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_TIME_ATTEST_INFO_PTR" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO""'");
    } else {
      arg1 = *((TPMS_TIME_ATTEST_INFO *)(argp1));
    }
  }
  result = (TPMS_TIME_ATTEST_INFO *)copy_TPMS_TIME_ATTEST_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_ATTEST_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_ATTEST_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_ATTEST_INFO_PTR" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  delete_TPMS_TIME_ATTEST_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  TPMS_TIME_ATTEST_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_ATTEST_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_TIME_ATTEST_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TIME_ATTEST_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_TIME_ATTEST_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_TIME_ATTEST_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_TIME_ATTEST_INFO""'");
    } else {
      arg2 = *((TPMS_TIME_ATTEST_INFO *)(argp2));
    }
  }
  TPMS_TIME_ATTEST_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_ATTEST_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  result = TPMS_TIME_ATTEST_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_TIME_ATTEST_INFO *)memcpy((TPMS_TIME_ATTEST_INFO *)malloc(sizeof(TPMS_TIME_ATTEST_INFO)),&result,sizeof(TPMS_TIME_ATTEST_INFO)), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_TIME_ATTEST_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_TIME_ATTEST_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_TIME_ATTEST_INFO **)new_TPMS_TIME_ATTEST_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_TIME_ATTEST_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_TIME_ATTEST_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_TIME_ATTEST_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  result = (TPMS_TIME_ATTEST_INFO **)copy_TPMS_TIME_ATTEST_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_TIME_ATTEST_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO **arg1 = (TPMS_TIME_ATTEST_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_TIME_ATTEST_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_TIME_ATTEST_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO **""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO **)(argp1);
  delete_TPMS_TIME_ATTEST_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO **arg1 = (TPMS_TIME_ATTEST_INFO **) 0 ;
  TPMS_TIME_ATTEST_INFO *arg2 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_TIME_ATTEST_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO **""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_TIME_ATTEST_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg2 = (TPMS_TIME_ATTEST_INFO *)(argp2);
  TPMS_TIME_ATTEST_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_TIME_ATTEST_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO **arg1 = (TPMS_TIME_ATTEST_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_TIME_ATTEST_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_TIME_ATTEST_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_TIME_ATTEST_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO **""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO **)(argp1);
  result = (TPMS_TIME_ATTEST_INFO *)TPMS_TIME_ATTEST_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CERTIFY_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CERTIFY_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_CERTIFY_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CERTIFY_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CERTIFY_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CERTIFY_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CERTIFY_INFO_ARRAY *)new_TPMS_CERTIFY_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CERTIFY_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO_ARRAY *arg1 = (TPMS_CERTIFY_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CERTIFY_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CERTIFY_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO_ARRAY *)(argp1);
  delete_TPMS_CERTIFY_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO_ARRAY *arg1 = (TPMS_CERTIFY_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CERTIFY_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CERTIFY_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CERTIFY_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CERTIFY_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CERTIFY_INFO *)memcpy((TPMS_CERTIFY_INFO *)malloc(sizeof(TPMS_CERTIFY_INFO)),&result,sizeof(TPMS_CERTIFY_INFO)), SWIGTYPE_p_TPMS_CERTIFY_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO_ARRAY *arg1 = (TPMS_CERTIFY_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CERTIFY_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CERTIFY_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CERTIFY_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CERTIFY_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CERTIFY_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CERTIFY_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CERTIFY_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CERTIFY_INFO""'");
    } else {
      arg3 = *((TPMS_CERTIFY_INFO *)(argp3));
    }
  }
  TPMS_CERTIFY_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO_ARRAY *arg1 = (TPMS_CERTIFY_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CERTIFY_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO_ARRAY *)(argp1);
  result = (TPMS_CERTIFY_INFO *)TPMS_CERTIFY_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CERTIFY_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  result = (TPMS_CERTIFY_INFO_ARRAY *)TPMS_CERTIFY_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CERTIFY_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CERTIFY_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CERTIFY_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CERTIFY_INFO_PTR")) SWIG_fail;
  result = (TPMS_CERTIFY_INFO *)new_TPMS_CERTIFY_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CERTIFY_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CERTIFY_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CERTIFY_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CERTIFY_INFO_PTR" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CERTIFY_INFO_PTR" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO""'");
    } else {
      arg1 = *((TPMS_CERTIFY_INFO *)(argp1));
    }
  }
  result = (TPMS_CERTIFY_INFO *)copy_TPMS_CERTIFY_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CERTIFY_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CERTIFY_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CERTIFY_INFO_PTR" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  delete_TPMS_CERTIFY_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  TPMS_CERTIFY_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CERTIFY_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CERTIFY_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CERTIFY_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_CERTIFY_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CERTIFY_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_CERTIFY_INFO""'");
    } else {
      arg2 = *((TPMS_CERTIFY_INFO *)(argp2));
    }
  }
  TPMS_CERTIFY_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CERTIFY_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  result = TPMS_CERTIFY_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CERTIFY_INFO *)memcpy((TPMS_CERTIFY_INFO *)malloc(sizeof(TPMS_CERTIFY_INFO)),&result,sizeof(TPMS_CERTIFY_INFO)), SWIGTYPE_p_TPMS_CERTIFY_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CERTIFY_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CERTIFY_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_CERTIFY_INFO **)new_TPMS_CERTIFY_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CERTIFY_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CERTIFY_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CERTIFY_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  result = (TPMS_CERTIFY_INFO **)copy_TPMS_CERTIFY_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CERTIFY_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO **arg1 = (TPMS_CERTIFY_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CERTIFY_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CERTIFY_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO **""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO **)(argp1);
  delete_TPMS_CERTIFY_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO **arg1 = (TPMS_CERTIFY_INFO **) 0 ;
  TPMS_CERTIFY_INFO *arg2 = (TPMS_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CERTIFY_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO **""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CERTIFY_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg2 = (TPMS_CERTIFY_INFO *)(argp2);
  TPMS_CERTIFY_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CERTIFY_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO **arg1 = (TPMS_CERTIFY_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CERTIFY_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CERTIFY_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO **""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO **)(argp1);
  result = (TPMS_CERTIFY_INFO *)TPMS_CERTIFY_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_QUOTE_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_QUOTE_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_QUOTE_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_QUOTE_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_QUOTE_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_QUOTE_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_QUOTE_INFO_ARRAY *)new_TPMS_QUOTE_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_QUOTE_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO_ARRAY *arg1 = (TPMS_QUOTE_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_QUOTE_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_QUOTE_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_QUOTE_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO_ARRAY *)(argp1);
  delete_TPMS_QUOTE_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO_ARRAY *arg1 = (TPMS_QUOTE_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_QUOTE_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_QUOTE_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_QUOTE_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_QUOTE_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_QUOTE_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_QUOTE_INFO *)memcpy((TPMS_QUOTE_INFO *)malloc(sizeof(TPMS_QUOTE_INFO)),&result,sizeof(TPMS_QUOTE_INFO)), SWIGTYPE_p_TPMS_QUOTE_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO_ARRAY *arg1 = (TPMS_QUOTE_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_QUOTE_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_QUOTE_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_QUOTE_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_QUOTE_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_QUOTE_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_QUOTE_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_QUOTE_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_QUOTE_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_QUOTE_INFO""'");
    } else {
      arg3 = *((TPMS_QUOTE_INFO *)(argp3));
    }
  }
  TPMS_QUOTE_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO_ARRAY *arg1 = (TPMS_QUOTE_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_QUOTE_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_QUOTE_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO_ARRAY *)(argp1);
  result = (TPMS_QUOTE_INFO *)TPMS_QUOTE_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_QUOTE_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  result = (TPMS_QUOTE_INFO_ARRAY *)TPMS_QUOTE_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_QUOTE_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_QUOTE_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_QUOTE_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_QUOTE_INFO_PTR")) SWIG_fail;
  result = (TPMS_QUOTE_INFO *)new_TPMS_QUOTE_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_QUOTE_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_QUOTE_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_QUOTE_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_QUOTE_INFO_PTR" "', argument " "1"" of type '" "TPMS_QUOTE_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_QUOTE_INFO_PTR" "', argument " "1"" of type '" "TPMS_QUOTE_INFO""'");
    } else {
      arg1 = *((TPMS_QUOTE_INFO *)(argp1));
    }
  }
  result = (TPMS_QUOTE_INFO *)copy_TPMS_QUOTE_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_QUOTE_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_QUOTE_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_QUOTE_INFO_PTR" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  delete_TPMS_QUOTE_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  TPMS_QUOTE_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_QUOTE_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_QUOTE_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_QUOTE_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_QUOTE_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_QUOTE_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_QUOTE_INFO""'");
    } else {
      arg2 = *((TPMS_QUOTE_INFO *)(argp2));
    }
  }
  TPMS_QUOTE_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_QUOTE_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  result = TPMS_QUOTE_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_QUOTE_INFO *)memcpy((TPMS_QUOTE_INFO *)malloc(sizeof(TPMS_QUOTE_INFO)),&result,sizeof(TPMS_QUOTE_INFO)), SWIGTYPE_p_TPMS_QUOTE_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_QUOTE_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_QUOTE_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_QUOTE_INFO **)new_TPMS_QUOTE_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_QUOTE_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_QUOTE_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_QUOTE_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  result = (TPMS_QUOTE_INFO **)copy_TPMS_QUOTE_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_QUOTE_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO **arg1 = (TPMS_QUOTE_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_QUOTE_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_QUOTE_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_QUOTE_INFO **""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO **)(argp1);
  delete_TPMS_QUOTE_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO **arg1 = (TPMS_QUOTE_INFO **) 0 ;
  TPMS_QUOTE_INFO *arg2 = (TPMS_QUOTE_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_QUOTE_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_QUOTE_INFO **""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_QUOTE_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg2 = (TPMS_QUOTE_INFO *)(argp2);
  TPMS_QUOTE_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_QUOTE_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO **arg1 = (TPMS_QUOTE_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_QUOTE_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_QUOTE_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_QUOTE_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_QUOTE_INFO **""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO **)(argp1);
  result = (TPMS_QUOTE_INFO *)TPMS_QUOTE_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_COMMAND_AUDIT_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_COMMAND_AUDIT_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_COMMAND_AUDIT_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_COMMAND_AUDIT_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_COMMAND_AUDIT_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_COMMAND_AUDIT_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_COMMAND_AUDIT_INFO_ARRAY *)new_TPMS_COMMAND_AUDIT_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_COMMAND_AUDIT_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO_ARRAY *arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_COMMAND_AUDIT_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_COMMAND_AUDIT_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *)(argp1);
  delete_TPMS_COMMAND_AUDIT_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO_ARRAY *arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_COMMAND_AUDIT_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_COMMAND_AUDIT_INFO *)memcpy((TPMS_COMMAND_AUDIT_INFO *)malloc(sizeof(TPMS_COMMAND_AUDIT_INFO)),&result,sizeof(TPMS_COMMAND_AUDIT_INFO)), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO_ARRAY *arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_COMMAND_AUDIT_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_COMMAND_AUDIT_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_COMMAND_AUDIT_INFO""'");
    } else {
      arg3 = *((TPMS_COMMAND_AUDIT_INFO *)(argp3));
    }
  }
  TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO_ARRAY *arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO_ARRAY *)(argp1);
  result = (TPMS_COMMAND_AUDIT_INFO *)TPMS_COMMAND_AUDIT_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = (TPMS_COMMAND_AUDIT_INFO_ARRAY *)TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_COMMAND_AUDIT_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_COMMAND_AUDIT_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_COMMAND_AUDIT_INFO_PTR")) SWIG_fail;
  result = (TPMS_COMMAND_AUDIT_INFO *)new_TPMS_COMMAND_AUDIT_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_COMMAND_AUDIT_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_COMMAND_AUDIT_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_COMMAND_AUDIT_INFO_PTR" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_COMMAND_AUDIT_INFO_PTR" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO""'");
    } else {
      arg1 = *((TPMS_COMMAND_AUDIT_INFO *)(argp1));
    }
  }
  result = (TPMS_COMMAND_AUDIT_INFO *)copy_TPMS_COMMAND_AUDIT_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_COMMAND_AUDIT_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_COMMAND_AUDIT_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_COMMAND_AUDIT_INFO_PTR" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  delete_TPMS_COMMAND_AUDIT_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  TPMS_COMMAND_AUDIT_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_COMMAND_AUDIT_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_COMMAND_AUDIT_INFO""'");
    } else {
      arg2 = *((TPMS_COMMAND_AUDIT_INFO *)(argp2));
    }
  }
  TPMS_COMMAND_AUDIT_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = TPMS_COMMAND_AUDIT_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_COMMAND_AUDIT_INFO *)memcpy((TPMS_COMMAND_AUDIT_INFO *)malloc(sizeof(TPMS_COMMAND_AUDIT_INFO)),&result,sizeof(TPMS_COMMAND_AUDIT_INFO)), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_COMMAND_AUDIT_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_COMMAND_AUDIT_INFO **)new_TPMS_COMMAND_AUDIT_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  result = (TPMS_COMMAND_AUDIT_INFO **)copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO **arg1 = (TPMS_COMMAND_AUDIT_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO **""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO **)(argp1);
  delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO **arg1 = (TPMS_COMMAND_AUDIT_INFO **) 0 ;
  TPMS_COMMAND_AUDIT_INFO *arg2 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO **""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg2 = (TPMS_COMMAND_AUDIT_INFO *)(argp2);
  TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO **arg1 = (TPMS_COMMAND_AUDIT_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_COMMAND_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO **""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO **)(argp1);
  result = (TPMS_COMMAND_AUDIT_INFO *)TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SESSION_AUDIT_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SESSION_AUDIT_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_SESSION_AUDIT_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SESSION_AUDIT_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SESSION_AUDIT_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SESSION_AUDIT_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SESSION_AUDIT_INFO_ARRAY *)new_TPMS_SESSION_AUDIT_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SESSION_AUDIT_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO_ARRAY *arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SESSION_AUDIT_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SESSION_AUDIT_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *)(argp1);
  delete_TPMS_SESSION_AUDIT_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO_ARRAY *arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SESSION_AUDIT_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SESSION_AUDIT_INFO *)memcpy((TPMS_SESSION_AUDIT_INFO *)malloc(sizeof(TPMS_SESSION_AUDIT_INFO)),&result,sizeof(TPMS_SESSION_AUDIT_INFO)), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO_ARRAY *arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SESSION_AUDIT_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SESSION_AUDIT_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SESSION_AUDIT_INFO""'");
    } else {
      arg3 = *((TPMS_SESSION_AUDIT_INFO *)(argp3));
    }
  }
  TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO_ARRAY *arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SESSION_AUDIT_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO_ARRAY *)(argp1);
  result = (TPMS_SESSION_AUDIT_INFO *)TPMS_SESSION_AUDIT_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  result = (TPMS_SESSION_AUDIT_INFO_ARRAY *)TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SESSION_AUDIT_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SESSION_AUDIT_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SESSION_AUDIT_INFO_PTR")) SWIG_fail;
  result = (TPMS_SESSION_AUDIT_INFO *)new_TPMS_SESSION_AUDIT_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SESSION_AUDIT_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SESSION_AUDIT_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SESSION_AUDIT_INFO_PTR" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SESSION_AUDIT_INFO_PTR" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO""'");
    } else {
      arg1 = *((TPMS_SESSION_AUDIT_INFO *)(argp1));
    }
  }
  result = (TPMS_SESSION_AUDIT_INFO *)copy_TPMS_SESSION_AUDIT_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SESSION_AUDIT_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SESSION_AUDIT_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SESSION_AUDIT_INFO_PTR" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  delete_TPMS_SESSION_AUDIT_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  TPMS_SESSION_AUDIT_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SESSION_AUDIT_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_SESSION_AUDIT_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_SESSION_AUDIT_INFO""'");
    } else {
      arg2 = *((TPMS_SESSION_AUDIT_INFO *)(argp2));
    }
  }
  TPMS_SESSION_AUDIT_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SESSION_AUDIT_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  result = TPMS_SESSION_AUDIT_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SESSION_AUDIT_INFO *)memcpy((TPMS_SESSION_AUDIT_INFO *)malloc(sizeof(TPMS_SESSION_AUDIT_INFO)),&result,sizeof(TPMS_SESSION_AUDIT_INFO)), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SESSION_AUDIT_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SESSION_AUDIT_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_SESSION_AUDIT_INFO **)new_TPMS_SESSION_AUDIT_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  result = (TPMS_SESSION_AUDIT_INFO **)copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO **arg1 = (TPMS_SESSION_AUDIT_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO **""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO **)(argp1);
  delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO **arg1 = (TPMS_SESSION_AUDIT_INFO **) 0 ;
  TPMS_SESSION_AUDIT_INFO *arg2 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO **""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg2 = (TPMS_SESSION_AUDIT_INFO *)(argp2);
  TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SESSION_AUDIT_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO **arg1 = (TPMS_SESSION_AUDIT_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SESSION_AUDIT_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SESSION_AUDIT_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SESSION_AUDIT_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO **""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO **)(argp1);
  result = (TPMS_SESSION_AUDIT_INFO *)TPMS_SESSION_AUDIT_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CREATION_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CREATION_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_CREATION_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CREATION_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CREATION_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CREATION_INFO_ARRAY *)new_TPMS_CREATION_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO_ARRAY *arg1 = (TPMS_CREATION_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_CREATION_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO_ARRAY *)(argp1);
  delete_TPMS_CREATION_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO_ARRAY *arg1 = (TPMS_CREATION_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CREATION_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CREATION_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CREATION_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CREATION_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CREATION_INFO *)memcpy((TPMS_CREATION_INFO *)malloc(sizeof(TPMS_CREATION_INFO)),&result,sizeof(TPMS_CREATION_INFO)), SWIGTYPE_p_TPMS_CREATION_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO_ARRAY *arg1 = (TPMS_CREATION_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CREATION_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CREATION_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CREATION_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CREATION_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CREATION_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CREATION_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CREATION_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CREATION_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CREATION_INFO""'");
    } else {
      arg3 = *((TPMS_CREATION_INFO *)(argp3));
    }
  }
  TPMS_CREATION_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO_ARRAY *arg1 = (TPMS_CREATION_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CREATION_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO_ARRAY *)(argp1);
  result = (TPMS_CREATION_INFO *)TPMS_CREATION_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  result = (TPMS_CREATION_INFO_ARRAY *)TPMS_CREATION_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CREATION_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CREATION_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CREATION_INFO_PTR")) SWIG_fail;
  result = (TPMS_CREATION_INFO *)new_TPMS_CREATION_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CREATION_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CREATION_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CREATION_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CREATION_INFO_PTR" "', argument " "1"" of type '" "TPMS_CREATION_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CREATION_INFO_PTR" "', argument " "1"" of type '" "TPMS_CREATION_INFO""'");
    } else {
      arg1 = *((TPMS_CREATION_INFO *)(argp1));
    }
  }
  result = (TPMS_CREATION_INFO *)copy_TPMS_CREATION_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_INFO_PTR" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  delete_TPMS_CREATION_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  TPMS_CREATION_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CREATION_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_CREATION_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CREATION_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_CREATION_INFO""'");
    } else {
      arg2 = *((TPMS_CREATION_INFO *)(argp2));
    }
  }
  TPMS_CREATION_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  result = TPMS_CREATION_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CREATION_INFO *)memcpy((TPMS_CREATION_INFO *)malloc(sizeof(TPMS_CREATION_INFO)),&result,sizeof(TPMS_CREATION_INFO)), SWIGTYPE_p_TPMS_CREATION_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CREATION_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_CREATION_INFO **)new_TPMS_CREATION_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CREATION_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CREATION_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CREATION_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  result = (TPMS_CREATION_INFO **)copy_TPMS_CREATION_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO **arg1 = (TPMS_CREATION_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_CREATION_INFO **""'"); 
  }
  arg1 = (TPMS_CREATION_INFO **)(argp1);
  delete_TPMS_CREATION_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO **arg1 = (TPMS_CREATION_INFO **) 0 ;
  TPMS_CREATION_INFO *arg2 = (TPMS_CREATION_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CREATION_INFO **""'"); 
  }
  arg1 = (TPMS_CREATION_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg2 = (TPMS_CREATION_INFO *)(argp2);
  TPMS_CREATION_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO **arg1 = (TPMS_CREATION_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CREATION_INFO **""'"); 
  }
  arg1 = (TPMS_CREATION_INFO **)(argp1);
  result = (TPMS_CREATION_INFO *)TPMS_CREATION_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_NV_CERTIFY_INFO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_NV_CERTIFY_INFO",SWIG_From_size_t((size_t)(sizeof(TPMS_NV_CERTIFY_INFO))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_CERTIFY_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_NV_CERTIFY_INFO_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_NV_CERTIFY_INFO_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_NV_CERTIFY_INFO_ARRAY *)new_TPMS_NV_CERTIFY_INFO_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_CERTIFY_INFO_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO_ARRAY *arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_CERTIFY_INFO_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_CERTIFY_INFO_ARRAY" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *)(argp1);
  delete_TPMS_NV_CERTIFY_INFO_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO_ARRAY *arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_NV_CERTIFY_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_CERTIFY_INFO_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_NV_CERTIFY_INFO_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_NV_CERTIFY_INFO *)memcpy((TPMS_NV_CERTIFY_INFO *)malloc(sizeof(TPMS_NV_CERTIFY_INFO)),&result,sizeof(TPMS_NV_CERTIFY_INFO)), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO_ARRAY *arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_NV_CERTIFY_INFO arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_NV_CERTIFY_INFO_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_NV_CERTIFY_INFO,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_NV_CERTIFY_INFO""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_NV_CERTIFY_INFO""'");
    } else {
      arg3 = *((TPMS_NV_CERTIFY_INFO *)(argp3));
    }
  }
  TPMS_NV_CERTIFY_INFO_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO_ARRAY *arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY_cast" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO_ARRAY *)(argp1);
  result = (TPMS_NV_CERTIFY_INFO *)TPMS_NV_CERTIFY_INFO_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  result = (TPMS_NV_CERTIFY_INFO_ARRAY *)TPMS_NV_CERTIFY_INFO_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_NV_CERTIFY_INFO_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_NV_CERTIFY_INFO_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_NV_CERTIFY_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_CERTIFY_INFO_PTR")) SWIG_fail;
  result = (TPMS_NV_CERTIFY_INFO *)new_TPMS_NV_CERTIFY_INFO_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_NV_CERTIFY_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_NV_CERTIFY_INFO_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_NV_CERTIFY_INFO,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_NV_CERTIFY_INFO_PTR" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_NV_CERTIFY_INFO_PTR" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO""'");
    } else {
      arg1 = *((TPMS_NV_CERTIFY_INFO *)(argp1));
    }
  }
  result = (TPMS_NV_CERTIFY_INFO *)copy_TPMS_NV_CERTIFY_INFO_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_CERTIFY_INFO_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_CERTIFY_INFO_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_CERTIFY_INFO_PTR" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  delete_TPMS_NV_CERTIFY_INFO_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  TPMS_NV_CERTIFY_INFO arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_CERTIFY_INFO_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_PTR_assign" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_NV_CERTIFY_INFO,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_CERTIFY_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_CERTIFY_INFO""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_NV_CERTIFY_INFO_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_CERTIFY_INFO""'");
    } else {
      arg2 = *((TPMS_NV_CERTIFY_INFO *)(argp2));
    }
  }
  TPMS_NV_CERTIFY_INFO_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_PTR_value" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  result = TPMS_NV_CERTIFY_INFO_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_NV_CERTIFY_INFO *)memcpy((TPMS_NV_CERTIFY_INFO *)malloc(sizeof(TPMS_NV_CERTIFY_INFO)),&result,sizeof(TPMS_NV_CERTIFY_INFO)), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_CERTIFY_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_CERTIFY_INFO_PTR_PTR")) SWIG_fail;
  result = (TPMS_NV_CERTIFY_INFO **)new_TPMS_NV_CERTIFY_INFO_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_NV_CERTIFY_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_NV_CERTIFY_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_NV_CERTIFY_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  result = (TPMS_NV_CERTIFY_INFO **)copy_TPMS_NV_CERTIFY_INFO_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_CERTIFY_INFO_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO **arg1 = (TPMS_NV_CERTIFY_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_CERTIFY_INFO_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_CERTIFY_INFO_PTR_PTR" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO **""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO **)(argp1);
  delete_TPMS_NV_CERTIFY_INFO_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO **arg1 = (TPMS_NV_CERTIFY_INFO **) 0 ;
  TPMS_NV_CERTIFY_INFO *arg2 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_CERTIFY_INFO_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO **""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_CERTIFY_INFO_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg2 = (TPMS_NV_CERTIFY_INFO *)(argp2);
  TPMS_NV_CERTIFY_INFO_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_CERTIFY_INFO_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO **arg1 = (TPMS_NV_CERTIFY_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_CERTIFY_INFO *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_CERTIFY_INFO_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_CERTIFY_INFO_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO **""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO **)(argp1);
  result = (TPMS_NV_CERTIFY_INFO *)TPMS_NV_CERTIFY_INFO_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_ATTEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_ATTEST",SWIG_From_size_t((size_t)(sizeof(TPMU_ATTEST))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_ATTEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_ATTEST_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_ATTEST_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_ATTEST_ARRAY *)new_TPMU_ATTEST_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ATTEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST_ARRAY *arg1 = (TPMU_ATTEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ATTEST_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ATTEST_ARRAY" "', argument " "1"" of type '" "TPMU_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMU_ATTEST_ARRAY *)(argp1);
  delete_TPMU_ATTEST_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST_ARRAY *arg1 = (TPMU_ATTEST_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMU_ATTEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_ATTEST_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_ATTEST_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_ATTEST *)memcpy((TPMU_ATTEST *)malloc(sizeof(TPMU_ATTEST)),&result,sizeof(TPMU_ATTEST)), SWIGTYPE_p_TPMU_ATTEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST_ARRAY *arg1 = (TPMU_ATTEST_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_ATTEST arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_ATTEST_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMU_ATTEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_ATTEST_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_ATTEST,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_ATTEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_ATTEST""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_ATTEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_ATTEST""'");
    } else {
      arg3 = *((TPMU_ATTEST *)(argp3));
    }
  }
  TPMU_ATTEST_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST_ARRAY *arg1 = (TPMU_ATTEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_ARRAY_cast" "', argument " "1"" of type '" "TPMU_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMU_ATTEST_ARRAY *)(argp1);
  result = (TPMU_ATTEST *)TPMU_ATTEST_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMU_ATTEST_ARRAY *)TPMU_ATTEST_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_ATTEST_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_ATTEST_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ATTEST_PTR")) SWIG_fail;
  result = (TPMU_ATTEST *)new_TPMU_ATTEST_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_ATTEST_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_ATTEST,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_ATTEST_PTR" "', argument " "1"" of type '" "TPMU_ATTEST""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_ATTEST_PTR" "', argument " "1"" of type '" "TPMU_ATTEST""'");
    } else {
      arg1 = *((TPMU_ATTEST *)(argp1));
    }
  }
  result = (TPMU_ATTEST *)copy_TPMU_ATTEST_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ATTEST_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ATTEST_PTR" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  delete_TPMU_ATTEST_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  TPMU_ATTEST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_PTR_assign" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_ATTEST,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_PTR_assign" "', argument " "2"" of type '" "TPMU_ATTEST""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_ATTEST_PTR_assign" "', argument " "2"" of type '" "TPMU_ATTEST""'");
    } else {
      arg2 = *((TPMU_ATTEST *)(argp2));
    }
  }
  TPMU_ATTEST_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_PTR_value" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = TPMU_ATTEST_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_ATTEST *)memcpy((TPMU_ATTEST *)malloc(sizeof(TPMU_ATTEST)),&result,sizeof(TPMU_ATTEST)), SWIGTYPE_p_TPMU_ATTEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ATTEST_PTR_PTR")) SWIG_fail;
  result = (TPMU_ATTEST **)new_TPMU_ATTEST_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_ATTEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_ATTEST_PTR_PTR" "', argument " "1"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  result = (TPMU_ATTEST **)copy_TPMU_ATTEST_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST **arg1 = (TPMU_ATTEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ATTEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ATTEST_PTR_PTR" "', argument " "1"" of type '" "TPMU_ATTEST **""'"); 
  }
  arg1 = (TPMU_ATTEST **)(argp1);
  delete_TPMU_ATTEST_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST **arg1 = (TPMU_ATTEST **) 0 ;
  TPMU_ATTEST *arg2 = (TPMU_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ATTEST_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_ATTEST **""'"); 
  }
  arg1 = (TPMU_ATTEST **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ATTEST_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg2 = (TPMU_ATTEST *)(argp2);
  TPMU_ATTEST_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ATTEST_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST **arg1 = (TPMU_ATTEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ATTEST_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ATTEST_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_ATTEST **""'"); 
  }
  arg1 = (TPMU_ATTEST **)(argp1);
  result = (TPMU_ATTEST *)TPMU_ATTEST_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ATTEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ATTEST",SWIG_From_size_t((size_t)(sizeof(TPMS_ATTEST))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ATTEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ATTEST_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ATTEST_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ATTEST_ARRAY *)new_TPMS_ATTEST_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ATTEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST_ARRAY *arg1 = (TPMS_ATTEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ATTEST_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ATTEST_ARRAY" "', argument " "1"" of type '" "TPMS_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMS_ATTEST_ARRAY *)(argp1);
  delete_TPMS_ATTEST_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST_ARRAY *arg1 = (TPMS_ATTEST_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMS_ATTEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ATTEST_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ATTEST_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ATTEST *)memcpy((TPMS_ATTEST *)malloc(sizeof(TPMS_ATTEST)),&result,sizeof(TPMS_ATTEST)), SWIGTYPE_p_TPMS_ATTEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST_ARRAY *arg1 = (TPMS_ATTEST_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ATTEST arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ATTEST_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMS_ATTEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ATTEST_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ATTEST,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ATTEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ATTEST""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ATTEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ATTEST""'");
    } else {
      arg3 = *((TPMS_ATTEST *)(argp3));
    }
  }
  TPMS_ATTEST_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST_ARRAY *arg1 = (TPMS_ATTEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPMS_ATTEST_ARRAY *)(argp1);
  result = (TPMS_ATTEST *)TPMS_ATTEST_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPMS_ATTEST_ARRAY *)TPMS_ATTEST_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ATTEST_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ATTEST_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ATTEST_PTR")) SWIG_fail;
  result = (TPMS_ATTEST *)new_TPMS_ATTEST_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ATTEST_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ATTEST,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ATTEST_PTR" "', argument " "1"" of type '" "TPMS_ATTEST""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ATTEST_PTR" "', argument " "1"" of type '" "TPMS_ATTEST""'");
    } else {
      arg1 = *((TPMS_ATTEST *)(argp1));
    }
  }
  result = (TPMS_ATTEST *)copy_TPMS_ATTEST_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ATTEST_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ATTEST_PTR" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  delete_TPMS_ATTEST_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  TPMS_ATTEST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_PTR_assign" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ATTEST,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ATTEST_PTR_assign" "', argument " "2"" of type '" "TPMS_ATTEST""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ATTEST_PTR_assign" "', argument " "2"" of type '" "TPMS_ATTEST""'");
    } else {
      arg2 = *((TPMS_ATTEST *)(argp2));
    }
  }
  TPMS_ATTEST_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_PTR_value" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = TPMS_ATTEST_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ATTEST *)memcpy((TPMS_ATTEST *)malloc(sizeof(TPMS_ATTEST)),&result,sizeof(TPMS_ATTEST)), SWIGTYPE_p_TPMS_ATTEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ATTEST_PTR_PTR")) SWIG_fail;
  result = (TPMS_ATTEST **)new_TPMS_ATTEST_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ATTEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ATTEST_PTR_PTR" "', argument " "1"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  result = (TPMS_ATTEST **)copy_TPMS_ATTEST_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST **arg1 = (TPMS_ATTEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ATTEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ATTEST_PTR_PTR" "', argument " "1"" of type '" "TPMS_ATTEST **""'"); 
  }
  arg1 = (TPMS_ATTEST **)(argp1);
  delete_TPMS_ATTEST_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST **arg1 = (TPMS_ATTEST **) 0 ;
  TPMS_ATTEST *arg2 = (TPMS_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ATTEST_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ATTEST **""'"); 
  }
  arg1 = (TPMS_ATTEST **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ATTEST_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg2 = (TPMS_ATTEST *)(argp2);
  TPMS_ATTEST_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ATTEST_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST **arg1 = (TPMS_ATTEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ATTEST_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ATTEST_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ATTEST **""'"); 
  }
  arg1 = (TPMS_ATTEST **)(argp1);
  result = (TPMS_ATTEST *)TPMS_ATTEST_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_ATTEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_ATTEST",SWIG_From_size_t((size_t)(sizeof(TPM2B_ATTEST))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ATTEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_ATTEST_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_ATTEST_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_ATTEST_ARRAY *)new_TPM2B_ATTEST_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ATTEST_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST_ARRAY *arg1 = (TPM2B_ATTEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ATTEST_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ATTEST_ARRAY" "', argument " "1"" of type '" "TPM2B_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ATTEST_ARRAY *)(argp1);
  delete_TPM2B_ATTEST_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST_ARRAY *arg1 = (TPM2B_ATTEST_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ATTEST_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ATTEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ATTEST_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_ATTEST_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_ATTEST *)memcpy((TPM2B_ATTEST *)malloc(sizeof(TPM2B_ATTEST)),&result,sizeof(TPM2B_ATTEST)), SWIGTYPE_p_TPM2B_ATTEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST_ARRAY *arg1 = (TPM2B_ATTEST_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_ATTEST arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_ATTEST_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ATTEST_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ATTEST_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_ATTEST,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_ATTEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ATTEST""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ATTEST_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ATTEST""'");
    } else {
      arg3 = *((TPM2B_ATTEST *)(argp3));
    }
  }
  TPM2B_ATTEST_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST_ARRAY *arg1 = (TPM2B_ATTEST_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ATTEST_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_ATTEST_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ATTEST_ARRAY *)(argp1);
  result = (TPM2B_ATTEST *)TPM2B_ATTEST_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ATTEST_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  result = (TPM2B_ATTEST_ARRAY *)TPM2B_ATTEST_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ATTEST_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ATTEST_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ATTEST_PTR")) SWIG_fail;
  result = (TPM2B_ATTEST *)new_TPM2B_ATTEST_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ATTEST_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_ATTEST,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ATTEST_PTR" "', argument " "1"" of type '" "TPM2B_ATTEST""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_ATTEST_PTR" "', argument " "1"" of type '" "TPM2B_ATTEST""'");
    } else {
      arg1 = *((TPM2B_ATTEST *)(argp1));
    }
  }
  result = (TPM2B_ATTEST *)copy_TPM2B_ATTEST_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ATTEST_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ATTEST_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ATTEST_PTR" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  delete_TPM2B_ATTEST_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  TPM2B_ATTEST arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ATTEST_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_PTR_assign" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_ATTEST,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ATTEST_PTR_assign" "', argument " "2"" of type '" "TPM2B_ATTEST""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ATTEST_PTR_assign" "', argument " "2"" of type '" "TPM2B_ATTEST""'");
    } else {
      arg2 = *((TPM2B_ATTEST *)(argp2));
    }
  }
  TPM2B_ATTEST_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ATTEST_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_PTR_value" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  result = TPM2B_ATTEST_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_ATTEST *)memcpy((TPM2B_ATTEST *)malloc(sizeof(TPM2B_ATTEST)),&result,sizeof(TPM2B_ATTEST)), SWIGTYPE_p_TPM2B_ATTEST, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ATTEST_PTR_PTR")) SWIG_fail;
  result = (TPM2B_ATTEST **)new_TPM2B_ATTEST_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ATTEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ATTEST_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  result = (TPM2B_ATTEST **)copy_TPM2B_ATTEST_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ATTEST_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST **arg1 = (TPM2B_ATTEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ATTEST_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ATTEST_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg1 = (TPM2B_ATTEST **)(argp1);
  delete_TPM2B_ATTEST_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST **arg1 = (TPM2B_ATTEST **) 0 ;
  TPM2B_ATTEST *arg2 = (TPM2B_ATTEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ATTEST_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg1 = (TPM2B_ATTEST **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ATTEST_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg2 = (TPM2B_ATTEST *)(argp2);
  TPM2B_ATTEST_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ATTEST_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST **arg1 = (TPM2B_ATTEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ATTEST *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ATTEST_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ATTEST_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg1 = (TPM2B_ATTEST **)(argp1);
  result = (TPM2B_ATTEST *)TPM2B_ATTEST_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_AUTH_COMMAND_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_AUTH_COMMAND",SWIG_From_size_t((size_t)(sizeof(TPMS_AUTH_COMMAND))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_COMMAND_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_AUTH_COMMAND_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_AUTH_COMMAND_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_AUTH_COMMAND_ARRAY *)new_TPMS_AUTH_COMMAND_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_COMMAND_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND_ARRAY *arg1 = (TPMS_AUTH_COMMAND_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_COMMAND_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_COMMAND_ARRAY" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND_ARRAY *)(argp1);
  delete_TPMS_AUTH_COMMAND_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND_ARRAY *arg1 = (TPMS_AUTH_COMMAND_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_AUTH_COMMAND result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_COMMAND_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_AUTH_COMMAND_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_AUTH_COMMAND *)memcpy((TPMS_AUTH_COMMAND *)malloc(sizeof(TPMS_AUTH_COMMAND)),&result,sizeof(TPMS_AUTH_COMMAND)), SWIGTYPE_p_TPMS_AUTH_COMMAND, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND_ARRAY *arg1 = (TPMS_AUTH_COMMAND_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_AUTH_COMMAND arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_AUTH_COMMAND_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_COMMAND_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_AUTH_COMMAND,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_AUTH_COMMAND_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_AUTH_COMMAND""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_AUTH_COMMAND_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_AUTH_COMMAND""'");
    } else {
      arg3 = *((TPMS_AUTH_COMMAND *)(argp3));
    }
  }
  TPMS_AUTH_COMMAND_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND_ARRAY *arg1 = (TPMS_AUTH_COMMAND_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_ARRAY_cast" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND_ARRAY *)(argp1);
  result = (TPMS_AUTH_COMMAND *)TPMS_AUTH_COMMAND_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = (TPMS_AUTH_COMMAND_ARRAY *)TPMS_AUTH_COMMAND_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_AUTH_COMMAND_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_AUTH_COMMAND_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_COMMAND_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AUTH_COMMAND_PTR")) SWIG_fail;
  result = (TPMS_AUTH_COMMAND *)new_TPMS_AUTH_COMMAND_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_AUTH_COMMAND_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_AUTH_COMMAND_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_AUTH_COMMAND,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_AUTH_COMMAND_PTR" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_AUTH_COMMAND_PTR" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND""'");
    } else {
      arg1 = *((TPMS_AUTH_COMMAND *)(argp1));
    }
  }
  result = (TPMS_AUTH_COMMAND *)copy_TPMS_AUTH_COMMAND_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_COMMAND_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_COMMAND_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_COMMAND_PTR" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  delete_TPMS_AUTH_COMMAND_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  TPMS_AUTH_COMMAND arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_PTR_assign" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_AUTH_COMMAND,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_COMMAND_PTR_assign" "', argument " "2"" of type '" "TPMS_AUTH_COMMAND""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_AUTH_COMMAND_PTR_assign" "', argument " "2"" of type '" "TPMS_AUTH_COMMAND""'");
    } else {
      arg2 = *((TPMS_AUTH_COMMAND *)(argp2));
    }
  }
  TPMS_AUTH_COMMAND_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_PTR_value" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = TPMS_AUTH_COMMAND_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_AUTH_COMMAND *)memcpy((TPMS_AUTH_COMMAND *)malloc(sizeof(TPMS_AUTH_COMMAND)),&result,sizeof(TPMS_AUTH_COMMAND)), SWIGTYPE_p_TPMS_AUTH_COMMAND, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_COMMAND_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AUTH_COMMAND_PTR_PTR")) SWIG_fail;
  result = (TPMS_AUTH_COMMAND **)new_TPMS_AUTH_COMMAND_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_AUTH_COMMAND, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_AUTH_COMMAND_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_AUTH_COMMAND_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_AUTH_COMMAND_PTR_PTR" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  result = (TPMS_AUTH_COMMAND **)copy_TPMS_AUTH_COMMAND_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_AUTH_COMMAND, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_COMMAND_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND **arg1 = (TPMS_AUTH_COMMAND **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_COMMAND_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_COMMAND_PTR_PTR" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND **""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND **)(argp1);
  delete_TPMS_AUTH_COMMAND_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND **arg1 = (TPMS_AUTH_COMMAND **) 0 ;
  TPMS_AUTH_COMMAND *arg2 = (TPMS_AUTH_COMMAND *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_COMMAND_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND **""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_COMMAND_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg2 = (TPMS_AUTH_COMMAND *)(argp2);
  TPMS_AUTH_COMMAND_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_COMMAND_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND **arg1 = (TPMS_AUTH_COMMAND **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_COMMAND *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_COMMAND_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_COMMAND_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND **""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND **)(argp1);
  result = (TPMS_AUTH_COMMAND *)TPMS_AUTH_COMMAND_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_AUTH_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_AUTH_RESPONSE",SWIG_From_size_t((size_t)(sizeof(TPMS_AUTH_RESPONSE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_RESPONSE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_AUTH_RESPONSE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_AUTH_RESPONSE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_AUTH_RESPONSE_ARRAY *)new_TPMS_AUTH_RESPONSE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_RESPONSE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE_ARRAY *arg1 = (TPMS_AUTH_RESPONSE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_RESPONSE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_RESPONSE_ARRAY" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE_ARRAY *)(argp1);
  delete_TPMS_AUTH_RESPONSE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE_ARRAY *arg1 = (TPMS_AUTH_RESPONSE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_AUTH_RESPONSE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_RESPONSE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_RESPONSE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_AUTH_RESPONSE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_AUTH_RESPONSE *)memcpy((TPMS_AUTH_RESPONSE *)malloc(sizeof(TPMS_AUTH_RESPONSE)),&result,sizeof(TPMS_AUTH_RESPONSE)), SWIGTYPE_p_TPMS_AUTH_RESPONSE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE_ARRAY *arg1 = (TPMS_AUTH_RESPONSE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_AUTH_RESPONSE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_AUTH_RESPONSE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AUTH_RESPONSE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_AUTH_RESPONSE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_AUTH_RESPONSE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_AUTH_RESPONSE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_AUTH_RESPONSE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_AUTH_RESPONSE""'");
    } else {
      arg3 = *((TPMS_AUTH_RESPONSE *)(argp3));
    }
  }
  TPMS_AUTH_RESPONSE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE_ARRAY *arg1 = (TPMS_AUTH_RESPONSE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_ARRAY_cast" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE_ARRAY *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE_ARRAY *)(argp1);
  result = (TPMS_AUTH_RESPONSE *)TPMS_AUTH_RESPONSE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  result = (TPMS_AUTH_RESPONSE_ARRAY *)TPMS_AUTH_RESPONSE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_AUTH_RESPONSE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_AUTH_RESPONSE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_RESPONSE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AUTH_RESPONSE_PTR")) SWIG_fail;
  result = (TPMS_AUTH_RESPONSE *)new_TPMS_AUTH_RESPONSE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_AUTH_RESPONSE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_AUTH_RESPONSE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_AUTH_RESPONSE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_AUTH_RESPONSE_PTR" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_AUTH_RESPONSE_PTR" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE""'");
    } else {
      arg1 = *((TPMS_AUTH_RESPONSE *)(argp1));
    }
  }
  result = (TPMS_AUTH_RESPONSE *)copy_TPMS_AUTH_RESPONSE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_RESPONSE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_RESPONSE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_RESPONSE_PTR" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  delete_TPMS_AUTH_RESPONSE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  TPMS_AUTH_RESPONSE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_RESPONSE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_PTR_assign" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_AUTH_RESPONSE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_RESPONSE_PTR_assign" "', argument " "2"" of type '" "TPMS_AUTH_RESPONSE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_AUTH_RESPONSE_PTR_assign" "', argument " "2"" of type '" "TPMS_AUTH_RESPONSE""'");
    } else {
      arg2 = *((TPMS_AUTH_RESPONSE *)(argp2));
    }
  }
  TPMS_AUTH_RESPONSE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_PTR_value" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  result = TPMS_AUTH_RESPONSE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_AUTH_RESPONSE *)memcpy((TPMS_AUTH_RESPONSE *)malloc(sizeof(TPMS_AUTH_RESPONSE)),&result,sizeof(TPMS_AUTH_RESPONSE)), SWIGTYPE_p_TPMS_AUTH_RESPONSE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_AUTH_RESPONSE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AUTH_RESPONSE_PTR_PTR")) SWIG_fail;
  result = (TPMS_AUTH_RESPONSE **)new_TPMS_AUTH_RESPONSE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_AUTH_RESPONSE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_AUTH_RESPONSE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_AUTH_RESPONSE_PTR_PTR" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  result = (TPMS_AUTH_RESPONSE **)copy_TPMS_AUTH_RESPONSE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AUTH_RESPONSE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE **arg1 = (TPMS_AUTH_RESPONSE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AUTH_RESPONSE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AUTH_RESPONSE_PTR_PTR" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE **""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE **)(argp1);
  delete_TPMS_AUTH_RESPONSE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE **arg1 = (TPMS_AUTH_RESPONSE **) 0 ;
  TPMS_AUTH_RESPONSE *arg2 = (TPMS_AUTH_RESPONSE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AUTH_RESPONSE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE **""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AUTH_RESPONSE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg2 = (TPMS_AUTH_RESPONSE *)(argp2);
  TPMS_AUTH_RESPONSE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AUTH_RESPONSE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE **arg1 = (TPMS_AUTH_RESPONSE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AUTH_RESPONSE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AUTH_RESPONSE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AUTH_RESPONSE_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE **""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE **)(argp1);
  result = (TPMS_AUTH_RESPONSE *)TPMS_AUTH_RESPONSE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_SYM_KEY_BITS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_SYM_KEY_BITS",SWIG_From_size_t((size_t)(sizeof(TPMU_SYM_KEY_BITS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_SYM_KEY_BITS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_SYM_KEY_BITS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_SYM_KEY_BITS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_SYM_KEY_BITS_ARRAY *)new_TPMU_SYM_KEY_BITS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_KEY_BITS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS_ARRAY *arg1 = (TPMU_SYM_KEY_BITS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_KEY_BITS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_KEY_BITS_ARRAY" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS_ARRAY *)(argp1);
  delete_TPMU_SYM_KEY_BITS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS_ARRAY *arg1 = (TPMU_SYM_KEY_BITS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_SYM_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_SYM_KEY_BITS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_SYM_KEY_BITS *)memcpy((TPMU_SYM_KEY_BITS *)malloc(sizeof(TPMU_SYM_KEY_BITS)),&result,sizeof(TPMU_SYM_KEY_BITS)), SWIGTYPE_p_TPMU_SYM_KEY_BITS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS_ARRAY *arg1 = (TPMU_SYM_KEY_BITS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_SYM_KEY_BITS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_SYM_KEY_BITS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_KEY_BITS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_SYM_KEY_BITS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_SYM_KEY_BITS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SYM_KEY_BITS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SYM_KEY_BITS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SYM_KEY_BITS""'");
    } else {
      arg3 = *((TPMU_SYM_KEY_BITS *)(argp3));
    }
  }
  TPMU_SYM_KEY_BITS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS_ARRAY *arg1 = (TPMU_SYM_KEY_BITS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_ARRAY_cast" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS_ARRAY *)(argp1);
  result = (TPMU_SYM_KEY_BITS *)TPMU_SYM_KEY_BITS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPMU_SYM_KEY_BITS_ARRAY *)TPMU_SYM_KEY_BITS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SYM_KEY_BITS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SYM_KEY_BITS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_SYM_KEY_BITS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SYM_KEY_BITS_PTR")) SWIG_fail;
  result = (TPMU_SYM_KEY_BITS *)new_TPMU_SYM_KEY_BITS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SYM_KEY_BITS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SYM_KEY_BITS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_SYM_KEY_BITS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SYM_KEY_BITS_PTR" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_SYM_KEY_BITS_PTR" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS""'");
    } else {
      arg1 = *((TPMU_SYM_KEY_BITS *)(argp1));
    }
  }
  result = (TPMU_SYM_KEY_BITS *)copy_TPMU_SYM_KEY_BITS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_KEY_BITS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_KEY_BITS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_KEY_BITS_PTR" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  delete_TPMU_SYM_KEY_BITS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  TPMU_SYM_KEY_BITS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_PTR_assign" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_SYM_KEY_BITS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SYM_KEY_BITS_PTR_assign" "', argument " "2"" of type '" "TPMU_SYM_KEY_BITS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SYM_KEY_BITS_PTR_assign" "', argument " "2"" of type '" "TPMU_SYM_KEY_BITS""'");
    } else {
      arg2 = *((TPMU_SYM_KEY_BITS *)(argp2));
    }
  }
  TPMU_SYM_KEY_BITS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_PTR_value" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = TPMU_SYM_KEY_BITS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_SYM_KEY_BITS *)memcpy((TPMU_SYM_KEY_BITS *)malloc(sizeof(TPMU_SYM_KEY_BITS)),&result,sizeof(TPMU_SYM_KEY_BITS)), SWIGTYPE_p_TPMU_SYM_KEY_BITS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SYM_KEY_BITS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SYM_KEY_BITS_PTR_PTR")) SWIG_fail;
  result = (TPMU_SYM_KEY_BITS **)new_TPMU_SYM_KEY_BITS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SYM_KEY_BITS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SYM_KEY_BITS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SYM_KEY_BITS_PTR_PTR" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  result = (TPMU_SYM_KEY_BITS **)copy_TPMU_SYM_KEY_BITS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_KEY_BITS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS **arg1 = (TPMU_SYM_KEY_BITS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_KEY_BITS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_KEY_BITS_PTR_PTR" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS **""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS **)(argp1);
  delete_TPMU_SYM_KEY_BITS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS **arg1 = (TPMU_SYM_KEY_BITS **) 0 ;
  TPMU_SYM_KEY_BITS *arg2 = (TPMU_SYM_KEY_BITS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_KEY_BITS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS **""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SYM_KEY_BITS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg2 = (TPMU_SYM_KEY_BITS *)(argp2);
  TPMU_SYM_KEY_BITS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_KEY_BITS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS **arg1 = (TPMU_SYM_KEY_BITS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_KEY_BITS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_KEY_BITS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_KEY_BITS_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS **""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS **)(argp1);
  result = (TPMU_SYM_KEY_BITS *)TPMU_SYM_KEY_BITS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_SYM_MODE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_SYM_MODE",SWIG_From_size_t((size_t)(sizeof(TPMU_SYM_MODE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_SYM_MODE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_SYM_MODE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_SYM_MODE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_SYM_MODE_ARRAY *)new_TPMU_SYM_MODE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_MODE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE_ARRAY *arg1 = (TPMU_SYM_MODE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_MODE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_MODE_ARRAY" "', argument " "1"" of type '" "TPMU_SYM_MODE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_MODE_ARRAY *)(argp1);
  delete_TPMU_SYM_MODE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE_ARRAY *arg1 = (TPMU_SYM_MODE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_SYM_MODE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_SYM_MODE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_MODE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_MODE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_SYM_MODE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_SYM_MODE *)memcpy((TPMU_SYM_MODE *)malloc(sizeof(TPMU_SYM_MODE)),&result,sizeof(TPMU_SYM_MODE)), SWIGTYPE_p_TPMU_SYM_MODE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE_ARRAY *arg1 = (TPMU_SYM_MODE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_SYM_MODE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_SYM_MODE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_SYM_MODE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_MODE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SYM_MODE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_SYM_MODE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_SYM_MODE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SYM_MODE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SYM_MODE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SYM_MODE""'");
    } else {
      arg3 = *((TPMU_SYM_MODE *)(argp3));
    }
  }
  TPMU_SYM_MODE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE_ARRAY *arg1 = (TPMU_SYM_MODE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_ARRAY_cast" "', argument " "1"" of type '" "TPMU_SYM_MODE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SYM_MODE_ARRAY *)(argp1);
  result = (TPMU_SYM_MODE *)TPMU_SYM_MODE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = (TPMU_SYM_MODE_ARRAY *)TPMU_SYM_MODE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SYM_MODE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SYM_MODE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_SYM_MODE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SYM_MODE_PTR")) SWIG_fail;
  result = (TPMU_SYM_MODE *)new_TPMU_SYM_MODE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SYM_MODE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SYM_MODE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_SYM_MODE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SYM_MODE_PTR" "', argument " "1"" of type '" "TPMU_SYM_MODE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_SYM_MODE_PTR" "', argument " "1"" of type '" "TPMU_SYM_MODE""'");
    } else {
      arg1 = *((TPMU_SYM_MODE *)(argp1));
    }
  }
  result = (TPMU_SYM_MODE *)copy_TPMU_SYM_MODE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_MODE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_MODE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_MODE_PTR" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  delete_TPMU_SYM_MODE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  TPMU_SYM_MODE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_PTR_assign" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_SYM_MODE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SYM_MODE_PTR_assign" "', argument " "2"" of type '" "TPMU_SYM_MODE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SYM_MODE_PTR_assign" "', argument " "2"" of type '" "TPMU_SYM_MODE""'");
    } else {
      arg2 = *((TPMU_SYM_MODE *)(argp2));
    }
  }
  TPMU_SYM_MODE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_PTR_value" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = TPMU_SYM_MODE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_SYM_MODE *)memcpy((TPMU_SYM_MODE *)malloc(sizeof(TPMU_SYM_MODE)),&result,sizeof(TPMU_SYM_MODE)), SWIGTYPE_p_TPMU_SYM_MODE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SYM_MODE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SYM_MODE_PTR_PTR")) SWIG_fail;
  result = (TPMU_SYM_MODE **)new_TPMU_SYM_MODE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SYM_MODE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SYM_MODE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SYM_MODE_PTR_PTR" "', argument " "1"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  result = (TPMU_SYM_MODE **)copy_TPMU_SYM_MODE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SYM_MODE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE **arg1 = (TPMU_SYM_MODE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SYM_MODE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SYM_MODE_PTR_PTR" "', argument " "1"" of type '" "TPMU_SYM_MODE **""'"); 
  }
  arg1 = (TPMU_SYM_MODE **)(argp1);
  delete_TPMU_SYM_MODE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE **arg1 = (TPMU_SYM_MODE **) 0 ;
  TPMU_SYM_MODE *arg2 = (TPMU_SYM_MODE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SYM_MODE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_SYM_MODE **""'"); 
  }
  arg1 = (TPMU_SYM_MODE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SYM_MODE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg2 = (TPMU_SYM_MODE *)(argp2);
  TPMU_SYM_MODE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SYM_MODE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE **arg1 = (TPMU_SYM_MODE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SYM_MODE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SYM_MODE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SYM_MODE_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_SYM_MODE **""'"); 
  }
  arg1 = (TPMU_SYM_MODE **)(argp1);
  result = (TPMU_SYM_MODE *)TPMU_SYM_MODE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_SYM_DEF_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_SYM_DEF",SWIG_From_size_t((size_t)(sizeof(TPMT_SYM_DEF))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_SYM_DEF_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_SYM_DEF_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_SYM_DEF_ARRAY *)new_TPMT_SYM_DEF_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_ARRAY *arg1 = (TPMT_SYM_DEF_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_ARRAY" "', argument " "1"" of type '" "TPMT_SYM_DEF_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_ARRAY *)(argp1);
  delete_TPMT_SYM_DEF_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_ARRAY *arg1 = (TPMT_SYM_DEF_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_SYM_DEF result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_SYM_DEF_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SYM_DEF_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_SYM_DEF_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_SYM_DEF *)memcpy((TPMT_SYM_DEF *)malloc(sizeof(TPMT_SYM_DEF)),&result,sizeof(TPMT_SYM_DEF)), SWIGTYPE_p_TPMT_SYM_DEF, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_ARRAY *arg1 = (TPMT_SYM_DEF_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_SYM_DEF arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_SYM_DEF_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_SYM_DEF_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SYM_DEF_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_SYM_DEF,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_SYM_DEF_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SYM_DEF""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SYM_DEF_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SYM_DEF""'");
    } else {
      arg3 = *((TPMT_SYM_DEF *)(argp3));
    }
  }
  TPMT_SYM_DEF_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_ARRAY *arg1 = (TPMT_SYM_DEF_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_ARRAY_cast" "', argument " "1"" of type '" "TPMT_SYM_DEF_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_ARRAY *)(argp1);
  result = (TPMT_SYM_DEF *)TPMT_SYM_DEF_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  result = (TPMT_SYM_DEF_ARRAY *)TPMT_SYM_DEF_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SYM_DEF_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SYM_DEF_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SYM_DEF_PTR")) SWIG_fail;
  result = (TPMT_SYM_DEF *)new_TPMT_SYM_DEF_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SYM_DEF_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SYM_DEF_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_SYM_DEF,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SYM_DEF_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_SYM_DEF_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF""'");
    } else {
      arg1 = *((TPMT_SYM_DEF *)(argp1));
    }
  }
  result = (TPMT_SYM_DEF *)copy_TPMT_SYM_DEF_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  delete_TPMT_SYM_DEF_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  TPMT_SYM_DEF arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_PTR_assign" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_SYM_DEF,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_PTR_assign" "', argument " "2"" of type '" "TPMT_SYM_DEF""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SYM_DEF_PTR_assign" "', argument " "2"" of type '" "TPMT_SYM_DEF""'");
    } else {
      arg2 = *((TPMT_SYM_DEF *)(argp2));
    }
  }
  TPMT_SYM_DEF_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_PTR_value" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  result = TPMT_SYM_DEF_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_SYM_DEF *)memcpy((TPMT_SYM_DEF *)malloc(sizeof(TPMT_SYM_DEF)),&result,sizeof(TPMT_SYM_DEF)), SWIGTYPE_p_TPMT_SYM_DEF, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SYM_DEF_PTR_PTR")) SWIG_fail;
  result = (TPMT_SYM_DEF **)new_TPMT_SYM_DEF_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SYM_DEF, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SYM_DEF_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SYM_DEF_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SYM_DEF_PTR_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  result = (TPMT_SYM_DEF **)copy_TPMT_SYM_DEF_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SYM_DEF, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF **arg1 = (TPMT_SYM_DEF **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_PTR_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF **""'"); 
  }
  arg1 = (TPMT_SYM_DEF **)(argp1);
  delete_TPMT_SYM_DEF_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF **arg1 = (TPMT_SYM_DEF **) 0 ;
  TPMT_SYM_DEF *arg2 = (TPMT_SYM_DEF *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_SYM_DEF **""'"); 
  }
  arg1 = (TPMT_SYM_DEF **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg2 = (TPMT_SYM_DEF *)(argp2);
  TPMT_SYM_DEF_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF **arg1 = (TPMT_SYM_DEF **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_SYM_DEF **""'"); 
  }
  arg1 = (TPMT_SYM_DEF **)(argp1);
  result = (TPMT_SYM_DEF *)TPMT_SYM_DEF_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_SYM_DEF_OBJECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_SYM_DEF_OBJECT",SWIG_From_size_t((size_t)(sizeof(TPMT_SYM_DEF_OBJECT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_OBJECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_SYM_DEF_OBJECT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_SYM_DEF_OBJECT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_SYM_DEF_OBJECT_ARRAY *)new_TPMT_SYM_DEF_OBJECT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_OBJECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT_ARRAY *arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_OBJECT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_OBJECT_ARRAY" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *)(argp1);
  delete_TPMT_SYM_DEF_OBJECT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT_ARRAY *arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_SYM_DEF_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_OBJECT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_SYM_DEF_OBJECT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_SYM_DEF_OBJECT *)memcpy((TPMT_SYM_DEF_OBJECT *)malloc(sizeof(TPMT_SYM_DEF_OBJECT)),&result,sizeof(TPMT_SYM_DEF_OBJECT)), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT_ARRAY *arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_SYM_DEF_OBJECT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_SYM_DEF_OBJECT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_SYM_DEF_OBJECT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SYM_DEF_OBJECT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SYM_DEF_OBJECT""'");
    } else {
      arg3 = *((TPMT_SYM_DEF_OBJECT *)(argp3));
    }
  }
  TPMT_SYM_DEF_OBJECT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT_ARRAY *arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY_cast" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT_ARRAY *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT *)TPMT_SYM_DEF_OBJECT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT_ARRAY *)TPMT_SYM_DEF_OBJECT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SYM_DEF_OBJECT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SYM_DEF_OBJECT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SYM_DEF_OBJECT_PTR")) SWIG_fail;
  result = (TPMT_SYM_DEF_OBJECT *)new_TPMT_SYM_DEF_OBJECT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SYM_DEF_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SYM_DEF_OBJECT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_SYM_DEF_OBJECT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SYM_DEF_OBJECT_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_SYM_DEF_OBJECT_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT""'");
    } else {
      arg1 = *((TPMT_SYM_DEF_OBJECT *)(argp1));
    }
  }
  result = (TPMT_SYM_DEF_OBJECT *)copy_TPMT_SYM_DEF_OBJECT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_OBJECT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_OBJECT_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  delete_TPMT_SYM_DEF_OBJECT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  TPMT_SYM_DEF_OBJECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_OBJECT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_PTR_assign" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_SYM_DEF_OBJECT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_OBJECT_PTR_assign" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SYM_DEF_OBJECT_PTR_assign" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT""'");
    } else {
      arg2 = *((TPMT_SYM_DEF_OBJECT *)(argp2));
    }
  }
  TPMT_SYM_DEF_OBJECT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_PTR_value" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  result = TPMT_SYM_DEF_OBJECT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_SYM_DEF_OBJECT *)memcpy((TPMT_SYM_DEF_OBJECT *)malloc(sizeof(TPMT_SYM_DEF_OBJECT)),&result,sizeof(TPMT_SYM_DEF_OBJECT)), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SYM_DEF_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SYM_DEF_OBJECT_PTR_PTR")) SWIG_fail;
  result = (TPMT_SYM_DEF_OBJECT **)new_TPMT_SYM_DEF_OBJECT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SYM_DEF_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SYM_DEF_OBJECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SYM_DEF_OBJECT_PTR_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  result = (TPMT_SYM_DEF_OBJECT **)copy_TPMT_SYM_DEF_OBJECT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SYM_DEF_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT **arg1 = (TPMT_SYM_DEF_OBJECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SYM_DEF_OBJECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SYM_DEF_OBJECT_PTR_PTR" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT **""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT **)(argp1);
  delete_TPMT_SYM_DEF_OBJECT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT **arg1 = (TPMT_SYM_DEF_OBJECT **) 0 ;
  TPMT_SYM_DEF_OBJECT *arg2 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SYM_DEF_OBJECT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT **""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SYM_DEF_OBJECT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg2 = (TPMT_SYM_DEF_OBJECT *)(argp2);
  TPMT_SYM_DEF_OBJECT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SYM_DEF_OBJECT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT **arg1 = (TPMT_SYM_DEF_OBJECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SYM_DEF_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SYM_DEF_OBJECT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SYM_DEF_OBJECT_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT **""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT **)(argp1);
  result = (TPMT_SYM_DEF_OBJECT *)TPMT_SYM_DEF_OBJECT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_SYM_KEY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_SYM_KEY",SWIG_From_size_t((size_t)(sizeof(TPM2B_SYM_KEY))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SYM_KEY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_SYM_KEY_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_SYM_KEY_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_SYM_KEY_ARRAY *)new_TPM2B_SYM_KEY_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SYM_KEY_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY_ARRAY *arg1 = (TPM2B_SYM_KEY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SYM_KEY_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SYM_KEY_ARRAY" "', argument " "1"" of type '" "TPM2B_SYM_KEY_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY_ARRAY *)(argp1);
  delete_TPM2B_SYM_KEY_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY_ARRAY *arg1 = (TPM2B_SYM_KEY_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_SYM_KEY result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SYM_KEY_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_SYM_KEY_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SYM_KEY_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_SYM_KEY_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_SYM_KEY *)memcpy((TPM2B_SYM_KEY *)malloc(sizeof(TPM2B_SYM_KEY)),&result,sizeof(TPM2B_SYM_KEY)), SWIGTYPE_p_TPM2B_SYM_KEY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY_ARRAY *arg1 = (TPM2B_SYM_KEY_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_SYM_KEY arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_SYM_KEY_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_SYM_KEY_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SYM_KEY_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_SYM_KEY,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_SYM_KEY_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SYM_KEY""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SYM_KEY_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SYM_KEY""'");
    } else {
      arg3 = *((TPM2B_SYM_KEY *)(argp3));
    }
  }
  TPM2B_SYM_KEY_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY_ARRAY *arg1 = (TPM2B_SYM_KEY_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SYM_KEY_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_SYM_KEY_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY_ARRAY *)(argp1);
  result = (TPM2B_SYM_KEY *)TPM2B_SYM_KEY_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SYM_KEY_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  result = (TPM2B_SYM_KEY_ARRAY *)TPM2B_SYM_KEY_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SYM_KEY_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SYM_KEY_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_SYM_KEY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SYM_KEY_PTR")) SWIG_fail;
  result = (TPM2B_SYM_KEY *)new_TPM2B_SYM_KEY_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SYM_KEY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SYM_KEY_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_SYM_KEY,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SYM_KEY_PTR" "', argument " "1"" of type '" "TPM2B_SYM_KEY""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_SYM_KEY_PTR" "', argument " "1"" of type '" "TPM2B_SYM_KEY""'");
    } else {
      arg1 = *((TPM2B_SYM_KEY *)(argp1));
    }
  }
  result = (TPM2B_SYM_KEY *)copy_TPM2B_SYM_KEY_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SYM_KEY_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SYM_KEY_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SYM_KEY_PTR" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  delete_TPM2B_SYM_KEY_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  TPM2B_SYM_KEY arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SYM_KEY_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_PTR_assign" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_SYM_KEY,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SYM_KEY_PTR_assign" "', argument " "2"" of type '" "TPM2B_SYM_KEY""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SYM_KEY_PTR_assign" "', argument " "2"" of type '" "TPM2B_SYM_KEY""'");
    } else {
      arg2 = *((TPM2B_SYM_KEY *)(argp2));
    }
  }
  TPM2B_SYM_KEY_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SYM_KEY_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_PTR_value" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  result = TPM2B_SYM_KEY_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_SYM_KEY *)memcpy((TPM2B_SYM_KEY *)malloc(sizeof(TPM2B_SYM_KEY)),&result,sizeof(TPM2B_SYM_KEY)), SWIGTYPE_p_TPM2B_SYM_KEY, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SYM_KEY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SYM_KEY_PTR_PTR")) SWIG_fail;
  result = (TPM2B_SYM_KEY **)new_TPM2B_SYM_KEY_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SYM_KEY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SYM_KEY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SYM_KEY_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  result = (TPM2B_SYM_KEY **)copy_TPM2B_SYM_KEY_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SYM_KEY_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY **arg1 = (TPM2B_SYM_KEY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SYM_KEY_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SYM_KEY_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SYM_KEY **""'"); 
  }
  arg1 = (TPM2B_SYM_KEY **)(argp1);
  delete_TPM2B_SYM_KEY_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY **arg1 = (TPM2B_SYM_KEY **) 0 ;
  TPM2B_SYM_KEY *arg2 = (TPM2B_SYM_KEY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SYM_KEY_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_SYM_KEY **""'"); 
  }
  arg1 = (TPM2B_SYM_KEY **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SYM_KEY_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg2 = (TPM2B_SYM_KEY *)(argp2);
  TPM2B_SYM_KEY_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SYM_KEY_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY **arg1 = (TPM2B_SYM_KEY **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SYM_KEY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SYM_KEY_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SYM_KEY_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_SYM_KEY **""'"); 
  }
  arg1 = (TPM2B_SYM_KEY **)(argp1);
  result = (TPM2B_SYM_KEY *)TPM2B_SYM_KEY_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SYMCIPHER_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SYMCIPHER_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMS_SYMCIPHER_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SYMCIPHER_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SYMCIPHER_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SYMCIPHER_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SYMCIPHER_PARMS_ARRAY *)new_TPMS_SYMCIPHER_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SYMCIPHER_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS_ARRAY *arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SYMCIPHER_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SYMCIPHER_PARMS_ARRAY" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *)(argp1);
  delete_TPMS_SYMCIPHER_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS_ARRAY *arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SYMCIPHER_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SYMCIPHER_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SYMCIPHER_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SYMCIPHER_PARMS *)memcpy((TPMS_SYMCIPHER_PARMS *)malloc(sizeof(TPMS_SYMCIPHER_PARMS)),&result,sizeof(TPMS_SYMCIPHER_PARMS)), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS_ARRAY *arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SYMCIPHER_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SYMCIPHER_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SYMCIPHER_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SYMCIPHER_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SYMCIPHER_PARMS""'");
    } else {
      arg3 = *((TPMS_SYMCIPHER_PARMS *)(argp3));
    }
  }
  TPMS_SYMCIPHER_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS_ARRAY *arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SYMCIPHER_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS_ARRAY *)(argp1);
  result = (TPMS_SYMCIPHER_PARMS *)TPMS_SYMCIPHER_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SYMCIPHER_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  result = (TPMS_SYMCIPHER_PARMS_ARRAY *)TPMS_SYMCIPHER_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SYMCIPHER_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SYMCIPHER_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SYMCIPHER_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SYMCIPHER_PARMS_PTR")) SWIG_fail;
  result = (TPMS_SYMCIPHER_PARMS *)new_TPMS_SYMCIPHER_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SYMCIPHER_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SYMCIPHER_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SYMCIPHER_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SYMCIPHER_PARMS_PTR" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SYMCIPHER_PARMS_PTR" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS""'");
    } else {
      arg1 = *((TPMS_SYMCIPHER_PARMS *)(argp1));
    }
  }
  result = (TPMS_SYMCIPHER_PARMS *)copy_TPMS_SYMCIPHER_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SYMCIPHER_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SYMCIPHER_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SYMCIPHER_PARMS_PTR" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  delete_TPMS_SYMCIPHER_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  TPMS_SYMCIPHER_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SYMCIPHER_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SYMCIPHER_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SYMCIPHER_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_SYMCIPHER_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SYMCIPHER_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_SYMCIPHER_PARMS""'");
    } else {
      arg2 = *((TPMS_SYMCIPHER_PARMS *)(argp2));
    }
  }
  TPMS_SYMCIPHER_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SYMCIPHER_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_PTR_value" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  result = TPMS_SYMCIPHER_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SYMCIPHER_PARMS *)memcpy((TPMS_SYMCIPHER_PARMS *)malloc(sizeof(TPMS_SYMCIPHER_PARMS)),&result,sizeof(TPMS_SYMCIPHER_PARMS)), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SYMCIPHER_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SYMCIPHER_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMS_SYMCIPHER_PARMS **)new_TPMS_SYMCIPHER_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SYMCIPHER_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SYMCIPHER_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SYMCIPHER_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  result = (TPMS_SYMCIPHER_PARMS **)copy_TPMS_SYMCIPHER_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SYMCIPHER_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS **arg1 = (TPMS_SYMCIPHER_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SYMCIPHER_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SYMCIPHER_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS **""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS **)(argp1);
  delete_TPMS_SYMCIPHER_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS **arg1 = (TPMS_SYMCIPHER_PARMS **) 0 ;
  TPMS_SYMCIPHER_PARMS *arg2 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SYMCIPHER_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS **""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SYMCIPHER_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg2 = (TPMS_SYMCIPHER_PARMS *)(argp2);
  TPMS_SYMCIPHER_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SYMCIPHER_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS **arg1 = (TPMS_SYMCIPHER_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SYMCIPHER_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SYMCIPHER_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SYMCIPHER_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS **""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS **)(argp1);
  result = (TPMS_SYMCIPHER_PARMS *)TPMS_SYMCIPHER_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_SENSITIVE_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_SENSITIVE_DATA",SWIG_From_size_t((size_t)(sizeof(TPM2B_SENSITIVE_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_SENSITIVE_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_SENSITIVE_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_SENSITIVE_DATA_ARRAY *)new_TPM2B_SENSITIVE_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA_ARRAY *arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_DATA_ARRAY" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *)(argp1);
  delete_TPM2B_SENSITIVE_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA_ARRAY *arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_SENSITIVE_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_SENSITIVE_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_SENSITIVE_DATA *)memcpy((TPM2B_SENSITIVE_DATA *)malloc(sizeof(TPM2B_SENSITIVE_DATA)),&result,sizeof(TPM2B_SENSITIVE_DATA)), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA_ARRAY *arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_SENSITIVE_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_SENSITIVE_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_SENSITIVE_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SENSITIVE_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SENSITIVE_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SENSITIVE_DATA""'");
    } else {
      arg3 = *((TPM2B_SENSITIVE_DATA *)(argp3));
    }
  }
  TPM2B_SENSITIVE_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA_ARRAY *arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA_ARRAY *)(argp1);
  result = (TPM2B_SENSITIVE_DATA *)TPM2B_SENSITIVE_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  result = (TPM2B_SENSITIVE_DATA_ARRAY *)TPM2B_SENSITIVE_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SENSITIVE_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SENSITIVE_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_DATA_PTR")) SWIG_fail;
  result = (TPM2B_SENSITIVE_DATA *)new_TPM2B_SENSITIVE_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SENSITIVE_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SENSITIVE_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_SENSITIVE_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SENSITIVE_DATA_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_SENSITIVE_DATA_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA""'");
    } else {
      arg1 = *((TPM2B_SENSITIVE_DATA *)(argp1));
    }
  }
  result = (TPM2B_SENSITIVE_DATA *)copy_TPM2B_SENSITIVE_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_DATA_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  delete_TPM2B_SENSITIVE_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  TPM2B_SENSITIVE_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_PTR_assign" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_SENSITIVE_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SENSITIVE_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE_DATA""'");
    } else {
      arg2 = *((TPM2B_SENSITIVE_DATA *)(argp2));
    }
  }
  TPM2B_SENSITIVE_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_PTR_value" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  result = TPM2B_SENSITIVE_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_SENSITIVE_DATA *)memcpy((TPM2B_SENSITIVE_DATA *)malloc(sizeof(TPM2B_SENSITIVE_DATA)),&result,sizeof(TPM2B_SENSITIVE_DATA)), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_DATA_PTR_PTR")) SWIG_fail;
  result = (TPM2B_SENSITIVE_DATA **)new_TPM2B_SENSITIVE_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SENSITIVE_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SENSITIVE_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SENSITIVE_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  result = (TPM2B_SENSITIVE_DATA **)copy_TPM2B_SENSITIVE_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA **arg1 = (TPM2B_SENSITIVE_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA **)(argp1);
  delete_TPM2B_SENSITIVE_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA **arg1 = (TPM2B_SENSITIVE_DATA **) 0 ;
  TPM2B_SENSITIVE_DATA *arg2 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg2 = (TPM2B_SENSITIVE_DATA *)(argp2);
  TPM2B_SENSITIVE_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA **arg1 = (TPM2B_SENSITIVE_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA **)(argp1);
  result = (TPM2B_SENSITIVE_DATA *)TPM2B_SENSITIVE_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SENSITIVE_CREATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SENSITIVE_CREATE",SWIG_From_size_t((size_t)(sizeof(TPMS_SENSITIVE_CREATE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SENSITIVE_CREATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SENSITIVE_CREATE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SENSITIVE_CREATE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SENSITIVE_CREATE_ARRAY *)new_TPMS_SENSITIVE_CREATE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SENSITIVE_CREATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE_ARRAY *arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SENSITIVE_CREATE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SENSITIVE_CREATE_ARRAY" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *)(argp1);
  delete_TPMS_SENSITIVE_CREATE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE_ARRAY *arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SENSITIVE_CREATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SENSITIVE_CREATE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SENSITIVE_CREATE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SENSITIVE_CREATE *)memcpy((TPMS_SENSITIVE_CREATE *)malloc(sizeof(TPMS_SENSITIVE_CREATE)),&result,sizeof(TPMS_SENSITIVE_CREATE)), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE_ARRAY *arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SENSITIVE_CREATE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SENSITIVE_CREATE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SENSITIVE_CREATE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SENSITIVE_CREATE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SENSITIVE_CREATE""'");
    } else {
      arg3 = *((TPMS_SENSITIVE_CREATE *)(argp3));
    }
  }
  TPMS_SENSITIVE_CREATE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE_ARRAY *arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SENSITIVE_CREATE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE_ARRAY *)(argp1);
  result = (TPMS_SENSITIVE_CREATE *)TPMS_SENSITIVE_CREATE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SENSITIVE_CREATE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  result = (TPMS_SENSITIVE_CREATE_ARRAY *)TPMS_SENSITIVE_CREATE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SENSITIVE_CREATE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SENSITIVE_CREATE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SENSITIVE_CREATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SENSITIVE_CREATE_PTR")) SWIG_fail;
  result = (TPMS_SENSITIVE_CREATE *)new_TPMS_SENSITIVE_CREATE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SENSITIVE_CREATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SENSITIVE_CREATE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SENSITIVE_CREATE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SENSITIVE_CREATE_PTR" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SENSITIVE_CREATE_PTR" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE""'");
    } else {
      arg1 = *((TPMS_SENSITIVE_CREATE *)(argp1));
    }
  }
  result = (TPMS_SENSITIVE_CREATE *)copy_TPMS_SENSITIVE_CREATE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SENSITIVE_CREATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SENSITIVE_CREATE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SENSITIVE_CREATE_PTR" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  delete_TPMS_SENSITIVE_CREATE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  TPMS_SENSITIVE_CREATE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SENSITIVE_CREATE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_PTR_assign" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SENSITIVE_CREATE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SENSITIVE_CREATE_PTR_assign" "', argument " "2"" of type '" "TPMS_SENSITIVE_CREATE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SENSITIVE_CREATE_PTR_assign" "', argument " "2"" of type '" "TPMS_SENSITIVE_CREATE""'");
    } else {
      arg2 = *((TPMS_SENSITIVE_CREATE *)(argp2));
    }
  }
  TPMS_SENSITIVE_CREATE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SENSITIVE_CREATE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_PTR_value" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  result = TPMS_SENSITIVE_CREATE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SENSITIVE_CREATE *)memcpy((TPMS_SENSITIVE_CREATE *)malloc(sizeof(TPMS_SENSITIVE_CREATE)),&result,sizeof(TPMS_SENSITIVE_CREATE)), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SENSITIVE_CREATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SENSITIVE_CREATE_PTR_PTR")) SWIG_fail;
  result = (TPMS_SENSITIVE_CREATE **)new_TPMS_SENSITIVE_CREATE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SENSITIVE_CREATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SENSITIVE_CREATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SENSITIVE_CREATE_PTR_PTR" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  result = (TPMS_SENSITIVE_CREATE **)copy_TPMS_SENSITIVE_CREATE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SENSITIVE_CREATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE **arg1 = (TPMS_SENSITIVE_CREATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SENSITIVE_CREATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SENSITIVE_CREATE_PTR_PTR" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE **""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE **)(argp1);
  delete_TPMS_SENSITIVE_CREATE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE **arg1 = (TPMS_SENSITIVE_CREATE **) 0 ;
  TPMS_SENSITIVE_CREATE *arg2 = (TPMS_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SENSITIVE_CREATE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE **""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SENSITIVE_CREATE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg2 = (TPMS_SENSITIVE_CREATE *)(argp2);
  TPMS_SENSITIVE_CREATE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SENSITIVE_CREATE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE **arg1 = (TPMS_SENSITIVE_CREATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SENSITIVE_CREATE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SENSITIVE_CREATE_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE **""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE **)(argp1);
  result = (TPMS_SENSITIVE_CREATE *)TPMS_SENSITIVE_CREATE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_SENSITIVE_CREATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_SENSITIVE_CREATE",SWIG_From_size_t((size_t)(sizeof(TPM2B_SENSITIVE_CREATE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_CREATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_SENSITIVE_CREATE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_SENSITIVE_CREATE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_SENSITIVE_CREATE_ARRAY *)new_TPM2B_SENSITIVE_CREATE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_CREATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE_ARRAY *arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_CREATE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_CREATE_ARRAY" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *)(argp1);
  delete_TPM2B_SENSITIVE_CREATE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE_ARRAY *arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_SENSITIVE_CREATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_CREATE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_SENSITIVE_CREATE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_SENSITIVE_CREATE *)memcpy((TPM2B_SENSITIVE_CREATE *)malloc(sizeof(TPM2B_SENSITIVE_CREATE)),&result,sizeof(TPM2B_SENSITIVE_CREATE)), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE_ARRAY *arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_SENSITIVE_CREATE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_SENSITIVE_CREATE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_SENSITIVE_CREATE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SENSITIVE_CREATE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SENSITIVE_CREATE""'");
    } else {
      arg3 = *((TPM2B_SENSITIVE_CREATE *)(argp3));
    }
  }
  TPM2B_SENSITIVE_CREATE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE_ARRAY *arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_CREATE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE_ARRAY *)(argp1);
  result = (TPM2B_SENSITIVE_CREATE *)TPM2B_SENSITIVE_CREATE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_CREATE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  result = (TPM2B_SENSITIVE_CREATE_ARRAY *)TPM2B_SENSITIVE_CREATE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SENSITIVE_CREATE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SENSITIVE_CREATE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_CREATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_CREATE_PTR")) SWIG_fail;
  result = (TPM2B_SENSITIVE_CREATE *)new_TPM2B_SENSITIVE_CREATE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SENSITIVE_CREATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SENSITIVE_CREATE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_SENSITIVE_CREATE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SENSITIVE_CREATE_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_SENSITIVE_CREATE_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE""'");
    } else {
      arg1 = *((TPM2B_SENSITIVE_CREATE *)(argp1));
    }
  }
  result = (TPM2B_SENSITIVE_CREATE *)copy_TPM2B_SENSITIVE_CREATE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_CREATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_CREATE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_CREATE_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  delete_TPM2B_SENSITIVE_CREATE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_SENSITIVE_CREATE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_CREATE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_PTR_assign" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_SENSITIVE_CREATE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_CREATE_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE_CREATE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SENSITIVE_CREATE_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE_CREATE""'");
    } else {
      arg2 = *((TPM2B_SENSITIVE_CREATE *)(argp2));
    }
  }
  TPM2B_SENSITIVE_CREATE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_CREATE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_PTR_value" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  result = TPM2B_SENSITIVE_CREATE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_SENSITIVE_CREATE *)memcpy((TPM2B_SENSITIVE_CREATE *)malloc(sizeof(TPM2B_SENSITIVE_CREATE)),&result,sizeof(TPM2B_SENSITIVE_CREATE)), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_CREATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_CREATE_PTR_PTR")) SWIG_fail;
  result = (TPM2B_SENSITIVE_CREATE **)new_TPM2B_SENSITIVE_CREATE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SENSITIVE_CREATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SENSITIVE_CREATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SENSITIVE_CREATE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  result = (TPM2B_SENSITIVE_CREATE **)copy_TPM2B_SENSITIVE_CREATE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_CREATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE **arg1 = (TPM2B_SENSITIVE_CREATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_CREATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_CREATE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE **)(argp1);
  delete_TPM2B_SENSITIVE_CREATE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE **arg1 = (TPM2B_SENSITIVE_CREATE **) 0 ;
  TPM2B_SENSITIVE_CREATE *arg2 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_CREATE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_CREATE_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg2 = (TPM2B_SENSITIVE_CREATE *)(argp2);
  TPM2B_SENSITIVE_CREATE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_CREATE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE **arg1 = (TPM2B_SENSITIVE_CREATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_CREATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_CREATE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_CREATE_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE **)(argp1);
  result = (TPM2B_SENSITIVE_CREATE *)TPM2B_SENSITIVE_CREATE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SCHEME_HASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SCHEME_HASH",SWIG_From_size_t((size_t)(sizeof(TPMS_SCHEME_HASH))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_HASH_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SCHEME_HASH_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SCHEME_HASH_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SCHEME_HASH_ARRAY *)new_TPMS_SCHEME_HASH_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_HASH_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH_ARRAY *arg1 = (TPMS_SCHEME_HASH_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_HASH_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_HASH_ARRAY" "', argument " "1"" of type '" "TPMS_SCHEME_HASH_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH_ARRAY *)(argp1);
  delete_TPMS_SCHEME_HASH_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH_ARRAY *arg1 = (TPMS_SCHEME_HASH_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SCHEME_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_HASH_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SCHEME_HASH_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_HASH_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SCHEME_HASH_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SCHEME_HASH *)memcpy((TPMS_SCHEME_HASH *)malloc(sizeof(TPMS_SCHEME_HASH)),&result,sizeof(TPMS_SCHEME_HASH)), SWIGTYPE_p_TPMS_SCHEME_HASH, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH_ARRAY *arg1 = (TPMS_SCHEME_HASH_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SCHEME_HASH arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SCHEME_HASH_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SCHEME_HASH_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_HASH_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SCHEME_HASH,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SCHEME_HASH_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SCHEME_HASH""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SCHEME_HASH_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SCHEME_HASH""'");
    } else {
      arg3 = *((TPMS_SCHEME_HASH *)(argp3));
    }
  }
  TPMS_SCHEME_HASH_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH_ARRAY *arg1 = (TPMS_SCHEME_HASH_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_HASH_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SCHEME_HASH_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH_ARRAY *)(argp1);
  result = (TPMS_SCHEME_HASH *)TPMS_SCHEME_HASH_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_HASH_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  result = (TPMS_SCHEME_HASH_ARRAY *)TPMS_SCHEME_HASH_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SCHEME_HASH_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SCHEME_HASH_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_HASH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_HASH_PTR")) SWIG_fail;
  result = (TPMS_SCHEME_HASH *)new_TPMS_SCHEME_HASH_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SCHEME_HASH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SCHEME_HASH_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SCHEME_HASH,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SCHEME_HASH_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_HASH""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SCHEME_HASH_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_HASH""'");
    } else {
      arg1 = *((TPMS_SCHEME_HASH *)(argp1));
    }
  }
  result = (TPMS_SCHEME_HASH *)copy_TPMS_SCHEME_HASH_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_HASH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_HASH_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_HASH_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  delete_TPMS_SCHEME_HASH_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  TPMS_SCHEME_HASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_HASH_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_PTR_assign" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SCHEME_HASH,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SCHEME_HASH_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_HASH""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SCHEME_HASH_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_HASH""'");
    } else {
      arg2 = *((TPMS_SCHEME_HASH *)(argp2));
    }
  }
  TPMS_SCHEME_HASH_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_HASH_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_PTR_value" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  result = TPMS_SCHEME_HASH_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SCHEME_HASH *)memcpy((TPMS_SCHEME_HASH *)malloc(sizeof(TPMS_SCHEME_HASH)),&result,sizeof(TPMS_SCHEME_HASH)), SWIGTYPE_p_TPMS_SCHEME_HASH, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_HASH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_HASH_PTR_PTR")) SWIG_fail;
  result = (TPMS_SCHEME_HASH **)new_TPMS_SCHEME_HASH_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SCHEME_HASH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SCHEME_HASH_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SCHEME_HASH_PTR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  result = (TPMS_SCHEME_HASH **)copy_TPMS_SCHEME_HASH_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_HASH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH **arg1 = (TPMS_SCHEME_HASH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_HASH_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_HASH_PTR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_HASH **""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH **)(argp1);
  delete_TPMS_SCHEME_HASH_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH **arg1 = (TPMS_SCHEME_HASH **) 0 ;
  TPMS_SCHEME_HASH *arg2 = (TPMS_SCHEME_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_HASH_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SCHEME_HASH **""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SCHEME_HASH_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg2 = (TPMS_SCHEME_HASH *)(argp2);
  TPMS_SCHEME_HASH_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_HASH_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH **arg1 = (TPMS_SCHEME_HASH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_HASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_HASH_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_HASH_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SCHEME_HASH **""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH **)(argp1);
  result = (TPMS_SCHEME_HASH *)TPMS_SCHEME_HASH_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SCHEME_ECDAA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SCHEME_ECDAA",SWIG_From_size_t((size_t)(sizeof(TPMS_SCHEME_ECDAA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_ECDAA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SCHEME_ECDAA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SCHEME_ECDAA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SCHEME_ECDAA_ARRAY *)new_TPMS_SCHEME_ECDAA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_ECDAA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA_ARRAY *arg1 = (TPMS_SCHEME_ECDAA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_ECDAA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_ECDAA_ARRAY" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA_ARRAY *)(argp1);
  delete_TPMS_SCHEME_ECDAA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA_ARRAY *arg1 = (TPMS_SCHEME_ECDAA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SCHEME_ECDAA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_ECDAA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_ECDAA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SCHEME_ECDAA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SCHEME_ECDAA *)memcpy((TPMS_SCHEME_ECDAA *)malloc(sizeof(TPMS_SCHEME_ECDAA)),&result,sizeof(TPMS_SCHEME_ECDAA)), SWIGTYPE_p_TPMS_SCHEME_ECDAA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA_ARRAY *arg1 = (TPMS_SCHEME_ECDAA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SCHEME_ECDAA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SCHEME_ECDAA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_ECDAA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SCHEME_ECDAA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SCHEME_ECDAA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SCHEME_ECDAA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SCHEME_ECDAA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SCHEME_ECDAA""'");
    } else {
      arg3 = *((TPMS_SCHEME_ECDAA *)(argp3));
    }
  }
  TPMS_SCHEME_ECDAA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA_ARRAY *arg1 = (TPMS_SCHEME_ECDAA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_ECDAA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA_ARRAY *)(argp1);
  result = (TPMS_SCHEME_ECDAA *)TPMS_SCHEME_ECDAA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_ECDAA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  result = (TPMS_SCHEME_ECDAA_ARRAY *)TPMS_SCHEME_ECDAA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SCHEME_ECDAA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SCHEME_ECDAA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_ECDAA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_ECDAA_PTR")) SWIG_fail;
  result = (TPMS_SCHEME_ECDAA *)new_TPMS_SCHEME_ECDAA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SCHEME_ECDAA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SCHEME_ECDAA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SCHEME_ECDAA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SCHEME_ECDAA_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SCHEME_ECDAA_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA""'");
    } else {
      arg1 = *((TPMS_SCHEME_ECDAA *)(argp1));
    }
  }
  result = (TPMS_SCHEME_ECDAA *)copy_TPMS_SCHEME_ECDAA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_ECDAA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_ECDAA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_ECDAA_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  delete_TPMS_SCHEME_ECDAA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  TPMS_SCHEME_ECDAA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_ECDAA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_PTR_assign" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SCHEME_ECDAA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SCHEME_ECDAA_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_ECDAA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SCHEME_ECDAA_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_ECDAA""'");
    } else {
      arg2 = *((TPMS_SCHEME_ECDAA *)(argp2));
    }
  }
  TPMS_SCHEME_ECDAA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_ECDAA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_PTR_value" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  result = TPMS_SCHEME_ECDAA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SCHEME_ECDAA *)memcpy((TPMS_SCHEME_ECDAA *)malloc(sizeof(TPMS_SCHEME_ECDAA)),&result,sizeof(TPMS_SCHEME_ECDAA)), SWIGTYPE_p_TPMS_SCHEME_ECDAA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_ECDAA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_ECDAA_PTR_PTR")) SWIG_fail;
  result = (TPMS_SCHEME_ECDAA **)new_TPMS_SCHEME_ECDAA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SCHEME_ECDAA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SCHEME_ECDAA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SCHEME_ECDAA_PTR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  result = (TPMS_SCHEME_ECDAA **)copy_TPMS_SCHEME_ECDAA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_ECDAA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA **arg1 = (TPMS_SCHEME_ECDAA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_ECDAA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_ECDAA_PTR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA **""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA **)(argp1);
  delete_TPMS_SCHEME_ECDAA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA **arg1 = (TPMS_SCHEME_ECDAA **) 0 ;
  TPMS_SCHEME_ECDAA *arg2 = (TPMS_SCHEME_ECDAA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_ECDAA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA **""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SCHEME_ECDAA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg2 = (TPMS_SCHEME_ECDAA *)(argp2);
  TPMS_SCHEME_ECDAA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_ECDAA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA **arg1 = (TPMS_SCHEME_ECDAA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_ECDAA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_ECDAA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_ECDAA_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA **""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA **)(argp1);
  result = (TPMS_SCHEME_ECDAA *)TPMS_SCHEME_ECDAA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SCHEME_XOR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SCHEME_XOR",SWIG_From_size_t((size_t)(sizeof(TPMS_SCHEME_XOR))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_XOR_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SCHEME_XOR_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SCHEME_XOR_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SCHEME_XOR_ARRAY *)new_TPMS_SCHEME_XOR_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_XOR_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR_ARRAY *arg1 = (TPMS_SCHEME_XOR_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_XOR_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_XOR_ARRAY" "', argument " "1"" of type '" "TPMS_SCHEME_XOR_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR_ARRAY *)(argp1);
  delete_TPMS_SCHEME_XOR_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR_ARRAY *arg1 = (TPMS_SCHEME_XOR_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SCHEME_XOR result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_XOR_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SCHEME_XOR_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_XOR_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SCHEME_XOR_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SCHEME_XOR *)memcpy((TPMS_SCHEME_XOR *)malloc(sizeof(TPMS_SCHEME_XOR)),&result,sizeof(TPMS_SCHEME_XOR)), SWIGTYPE_p_TPMS_SCHEME_XOR, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR_ARRAY *arg1 = (TPMS_SCHEME_XOR_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SCHEME_XOR arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SCHEME_XOR_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SCHEME_XOR_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SCHEME_XOR_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SCHEME_XOR,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SCHEME_XOR_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SCHEME_XOR""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SCHEME_XOR_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SCHEME_XOR""'");
    } else {
      arg3 = *((TPMS_SCHEME_XOR *)(argp3));
    }
  }
  TPMS_SCHEME_XOR_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR_ARRAY *arg1 = (TPMS_SCHEME_XOR_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_XOR_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SCHEME_XOR_ARRAY *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR_ARRAY *)(argp1);
  result = (TPMS_SCHEME_XOR *)TPMS_SCHEME_XOR_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_XOR_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  result = (TPMS_SCHEME_XOR_ARRAY *)TPMS_SCHEME_XOR_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SCHEME_XOR_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SCHEME_XOR_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_XOR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_XOR_PTR")) SWIG_fail;
  result = (TPMS_SCHEME_XOR *)new_TPMS_SCHEME_XOR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SCHEME_XOR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SCHEME_XOR_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SCHEME_XOR,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SCHEME_XOR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_XOR""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SCHEME_XOR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_XOR""'");
    } else {
      arg1 = *((TPMS_SCHEME_XOR *)(argp1));
    }
  }
  result = (TPMS_SCHEME_XOR *)copy_TPMS_SCHEME_XOR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_XOR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_XOR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_XOR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  delete_TPMS_SCHEME_XOR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  TPMS_SCHEME_XOR arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_XOR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_PTR_assign" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SCHEME_XOR,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SCHEME_XOR_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_XOR""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SCHEME_XOR_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_XOR""'");
    } else {
      arg2 = *((TPMS_SCHEME_XOR *)(argp2));
    }
  }
  TPMS_SCHEME_XOR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_XOR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_PTR_value" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  result = TPMS_SCHEME_XOR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SCHEME_XOR *)memcpy((TPMS_SCHEME_XOR *)malloc(sizeof(TPMS_SCHEME_XOR)),&result,sizeof(TPMS_SCHEME_XOR)), SWIGTYPE_p_TPMS_SCHEME_XOR, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SCHEME_XOR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SCHEME_XOR_PTR_PTR")) SWIG_fail;
  result = (TPMS_SCHEME_XOR **)new_TPMS_SCHEME_XOR_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SCHEME_XOR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SCHEME_XOR_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SCHEME_XOR_PTR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  result = (TPMS_SCHEME_XOR **)copy_TPMS_SCHEME_XOR_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SCHEME_XOR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR **arg1 = (TPMS_SCHEME_XOR **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SCHEME_XOR_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SCHEME_XOR_PTR_PTR" "', argument " "1"" of type '" "TPMS_SCHEME_XOR **""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR **)(argp1);
  delete_TPMS_SCHEME_XOR_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR **arg1 = (TPMS_SCHEME_XOR **) 0 ;
  TPMS_SCHEME_XOR *arg2 = (TPMS_SCHEME_XOR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SCHEME_XOR_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SCHEME_XOR **""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SCHEME_XOR_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg2 = (TPMS_SCHEME_XOR *)(argp2);
  TPMS_SCHEME_XOR_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SCHEME_XOR_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR **arg1 = (TPMS_SCHEME_XOR **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SCHEME_XOR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SCHEME_XOR_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SCHEME_XOR_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SCHEME_XOR **""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR **)(argp1);
  result = (TPMS_SCHEME_XOR *)TPMS_SCHEME_XOR_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_SCHEME_KEYEDHASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_SCHEME_KEYEDHASH",SWIG_From_size_t((size_t)(sizeof(TPMU_SCHEME_KEYEDHASH))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_SCHEME_KEYEDHASH_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_SCHEME_KEYEDHASH_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_SCHEME_KEYEDHASH_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_SCHEME_KEYEDHASH_ARRAY *)new_TPMU_SCHEME_KEYEDHASH_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SCHEME_KEYEDHASH_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH_ARRAY *arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SCHEME_KEYEDHASH_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SCHEME_KEYEDHASH_ARRAY" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH_ARRAY *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *)(argp1);
  delete_TPMU_SCHEME_KEYEDHASH_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH_ARRAY *arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_SCHEME_KEYEDHASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH_ARRAY *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_SCHEME_KEYEDHASH *)memcpy((TPMU_SCHEME_KEYEDHASH *)malloc(sizeof(TPMU_SCHEME_KEYEDHASH)),&result,sizeof(TPMU_SCHEME_KEYEDHASH)), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH_ARRAY *arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_SCHEME_KEYEDHASH arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH_ARRAY *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SCHEME_KEYEDHASH""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SCHEME_KEYEDHASH""'");
    } else {
      arg3 = *((TPMU_SCHEME_KEYEDHASH *)(argp3));
    }
  }
  TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH_ARRAY *arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SCHEME_KEYEDHASH_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY_cast" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH_ARRAY *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH_ARRAY *)(argp1);
  result = (TPMU_SCHEME_KEYEDHASH *)TPMU_SCHEME_KEYEDHASH_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  result = (TPMU_SCHEME_KEYEDHASH_ARRAY *)TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SCHEME_KEYEDHASH_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_SCHEME_KEYEDHASH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SCHEME_KEYEDHASH_PTR")) SWIG_fail;
  result = (TPMU_SCHEME_KEYEDHASH *)new_TPMU_SCHEME_KEYEDHASH_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SCHEME_KEYEDHASH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SCHEME_KEYEDHASH_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SCHEME_KEYEDHASH_PTR" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_SCHEME_KEYEDHASH_PTR" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH""'");
    } else {
      arg1 = *((TPMU_SCHEME_KEYEDHASH *)(argp1));
    }
  }
  result = (TPMU_SCHEME_KEYEDHASH *)copy_TPMU_SCHEME_KEYEDHASH_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SCHEME_KEYEDHASH_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SCHEME_KEYEDHASH_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SCHEME_KEYEDHASH_PTR" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  delete_TPMU_SCHEME_KEYEDHASH_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  TPMU_SCHEME_KEYEDHASH arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SCHEME_KEYEDHASH_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_assign" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_assign" "', argument " "2"" of type '" "TPMU_SCHEME_KEYEDHASH""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_assign" "', argument " "2"" of type '" "TPMU_SCHEME_KEYEDHASH""'");
    } else {
      arg2 = *((TPMU_SCHEME_KEYEDHASH *)(argp2));
    }
  }
  TPMU_SCHEME_KEYEDHASH_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SCHEME_KEYEDHASH_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_value" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  result = TPMU_SCHEME_KEYEDHASH_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_SCHEME_KEYEDHASH *)memcpy((TPMU_SCHEME_KEYEDHASH *)malloc(sizeof(TPMU_SCHEME_KEYEDHASH)),&result,sizeof(TPMU_SCHEME_KEYEDHASH)), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SCHEME_KEYEDHASH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SCHEME_KEYEDHASH_PTR_PTR")) SWIG_fail;
  result = (TPMU_SCHEME_KEYEDHASH **)new_TPMU_SCHEME_KEYEDHASH_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  result = (TPMU_SCHEME_KEYEDHASH **)copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH **arg1 = (TPMU_SCHEME_KEYEDHASH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH **""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH **)(argp1);
  delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH **arg1 = (TPMU_SCHEME_KEYEDHASH **) 0 ;
  TPMU_SCHEME_KEYEDHASH *arg2 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH **""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg2 = (TPMU_SCHEME_KEYEDHASH *)(argp2);
  TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SCHEME_KEYEDHASH_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH **arg1 = (TPMU_SCHEME_KEYEDHASH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SCHEME_KEYEDHASH *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SCHEME_KEYEDHASH_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SCHEME_KEYEDHASH_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH **""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH **)(argp1);
  result = (TPMU_SCHEME_KEYEDHASH *)TPMU_SCHEME_KEYEDHASH_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_KEYEDHASH_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_KEYEDHASH_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMT_KEYEDHASH_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_KEYEDHASH_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_KEYEDHASH_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_KEYEDHASH_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_KEYEDHASH_SCHEME_ARRAY *)new_TPMT_KEYEDHASH_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KEYEDHASH_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME_ARRAY *arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KEYEDHASH_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KEYEDHASH_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *)(argp1);
  delete_TPMT_KEYEDHASH_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME_ARRAY *arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_KEYEDHASH_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_KEYEDHASH_SCHEME *)memcpy((TPMT_KEYEDHASH_SCHEME *)malloc(sizeof(TPMT_KEYEDHASH_SCHEME)),&result,sizeof(TPMT_KEYEDHASH_SCHEME)), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME_ARRAY *arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_KEYEDHASH_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_KEYEDHASH_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_KEYEDHASH_SCHEME""'");
    } else {
      arg3 = *((TPMT_KEYEDHASH_SCHEME *)(argp3));
    }
  }
  TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME_ARRAY *arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KEYEDHASH_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME_ARRAY *)(argp1);
  result = (TPMT_KEYEDHASH_SCHEME *)TPMT_KEYEDHASH_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  result = (TPMT_KEYEDHASH_SCHEME_ARRAY *)TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_KEYEDHASH_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_KEYEDHASH_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_KEYEDHASH_SCHEME_PTR")) SWIG_fail;
  result = (TPMT_KEYEDHASH_SCHEME *)new_TPMT_KEYEDHASH_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_KEYEDHASH_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_KEYEDHASH_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_KEYEDHASH_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_KEYEDHASH_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME""'");
    } else {
      arg1 = *((TPMT_KEYEDHASH_SCHEME *)(argp1));
    }
  }
  result = (TPMT_KEYEDHASH_SCHEME *)copy_TPMT_KEYEDHASH_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KEYEDHASH_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KEYEDHASH_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KEYEDHASH_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  delete_TPMT_KEYEDHASH_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  TPMT_KEYEDHASH_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KEYEDHASH_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_KEYEDHASH_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_KEYEDHASH_SCHEME""'");
    } else {
      arg2 = *((TPMT_KEYEDHASH_SCHEME *)(argp2));
    }
  }
  TPMT_KEYEDHASH_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KEYEDHASH_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  result = TPMT_KEYEDHASH_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_KEYEDHASH_SCHEME *)memcpy((TPMT_KEYEDHASH_SCHEME *)malloc(sizeof(TPMT_KEYEDHASH_SCHEME)),&result,sizeof(TPMT_KEYEDHASH_SCHEME)), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_KEYEDHASH_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_KEYEDHASH_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMT_KEYEDHASH_SCHEME **)new_TPMT_KEYEDHASH_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  result = (TPMT_KEYEDHASH_SCHEME **)copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME **arg1 = (TPMT_KEYEDHASH_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME **""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME **)(argp1);
  delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME **arg1 = (TPMT_KEYEDHASH_SCHEME **) 0 ;
  TPMT_KEYEDHASH_SCHEME *arg2 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME **""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg2 = (TPMT_KEYEDHASH_SCHEME *)(argp2);
  TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KEYEDHASH_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME **arg1 = (TPMT_KEYEDHASH_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KEYEDHASH_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KEYEDHASH_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KEYEDHASH_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME **""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME **)(argp1);
  result = (TPMT_KEYEDHASH_SCHEME *)TPMT_KEYEDHASH_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_SIG_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_SIG_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMU_SIG_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_SIG_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_SIG_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_SIG_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_SIG_SCHEME_ARRAY *)new_TPMU_SIG_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIG_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME_ARRAY *arg1 = (TPMU_SIG_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIG_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIG_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMU_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME_ARRAY *)(argp1);
  delete_TPMU_SIG_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME_ARRAY *arg1 = (TPMU_SIG_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_SIG_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SIG_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_SIG_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_SIG_SCHEME *)memcpy((TPMU_SIG_SCHEME *)malloc(sizeof(TPMU_SIG_SCHEME)),&result,sizeof(TPMU_SIG_SCHEME)), SWIGTYPE_p_TPMU_SIG_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME_ARRAY *arg1 = (TPMU_SIG_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_SIG_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_SIG_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SIG_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_SIG_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_SIG_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SIG_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SIG_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SIG_SCHEME""'");
    } else {
      arg3 = *((TPMU_SIG_SCHEME *)(argp3));
    }
  }
  TPMU_SIG_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME_ARRAY *arg1 = (TPMU_SIG_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMU_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME_ARRAY *)(argp1);
  result = (TPMU_SIG_SCHEME *)TPMU_SIG_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMU_SIG_SCHEME_ARRAY *)TPMU_SIG_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SIG_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SIG_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_SIG_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SIG_SCHEME_PTR")) SWIG_fail;
  result = (TPMU_SIG_SCHEME *)new_TPMU_SIG_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SIG_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SIG_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_SIG_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SIG_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_SIG_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_SIG_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_SIG_SCHEME""'");
    } else {
      arg1 = *((TPMU_SIG_SCHEME *)(argp1));
    }
  }
  result = (TPMU_SIG_SCHEME *)copy_TPMU_SIG_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIG_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIG_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIG_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  delete_TPMU_SIG_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  TPMU_SIG_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_SIG_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMU_SIG_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SIG_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMU_SIG_SCHEME""'");
    } else {
      arg2 = *((TPMU_SIG_SCHEME *)(argp2));
    }
  }
  TPMU_SIG_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = TPMU_SIG_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_SIG_SCHEME *)memcpy((TPMU_SIG_SCHEME *)malloc(sizeof(TPMU_SIG_SCHEME)),&result,sizeof(TPMU_SIG_SCHEME)), SWIGTYPE_p_TPMU_SIG_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SIG_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SIG_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMU_SIG_SCHEME **)new_TPMU_SIG_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SIG_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SIG_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SIG_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  result = (TPMU_SIG_SCHEME **)copy_TPMU_SIG_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIG_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME **arg1 = (TPMU_SIG_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIG_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIG_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMU_SIG_SCHEME **""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME **)(argp1);
  delete_TPMU_SIG_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME **arg1 = (TPMU_SIG_SCHEME **) 0 ;
  TPMU_SIG_SCHEME *arg2 = (TPMU_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIG_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_SIG_SCHEME **""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIG_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg2 = (TPMU_SIG_SCHEME *)(argp2);
  TPMU_SIG_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIG_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME **arg1 = (TPMU_SIG_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIG_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIG_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_SIG_SCHEME **""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME **)(argp1);
  result = (TPMU_SIG_SCHEME *)TPMU_SIG_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_SIG_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_SIG_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMT_SIG_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_SIG_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_SIG_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_SIG_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_SIG_SCHEME_ARRAY *)new_TPMT_SIG_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIG_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME_ARRAY *arg1 = (TPMT_SIG_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIG_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIG_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMT_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME_ARRAY *)(argp1);
  delete_TPMT_SIG_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME_ARRAY *arg1 = (TPMT_SIG_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_SIG_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIG_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SIG_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_SIG_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_SIG_SCHEME *)memcpy((TPMT_SIG_SCHEME *)malloc(sizeof(TPMT_SIG_SCHEME)),&result,sizeof(TPMT_SIG_SCHEME)), SWIGTYPE_p_TPMT_SIG_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME_ARRAY *arg1 = (TPMT_SIG_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_SIG_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_SIG_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SIG_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_SIG_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_SIG_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SIG_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SIG_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SIG_SCHEME""'");
    } else {
      arg3 = *((TPMT_SIG_SCHEME *)(argp3));
    }
  }
  TPMT_SIG_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME_ARRAY *arg1 = (TPMT_SIG_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIG_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMT_SIG_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME_ARRAY *)(argp1);
  result = (TPMT_SIG_SCHEME *)TPMT_SIG_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIG_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  result = (TPMT_SIG_SCHEME_ARRAY *)TPMT_SIG_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SIG_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SIG_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_SIG_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SIG_SCHEME_PTR")) SWIG_fail;
  result = (TPMT_SIG_SCHEME *)new_TPMT_SIG_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SIG_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SIG_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_SIG_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SIG_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_SIG_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_SIG_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_SIG_SCHEME""'");
    } else {
      arg1 = *((TPMT_SIG_SCHEME *)(argp1));
    }
  }
  result = (TPMT_SIG_SCHEME *)copy_TPMT_SIG_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIG_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIG_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIG_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  delete_TPMT_SIG_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  TPMT_SIG_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIG_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_SIG_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SIG_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_SIG_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SIG_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_SIG_SCHEME""'");
    } else {
      arg2 = *((TPMT_SIG_SCHEME *)(argp2));
    }
  }
  TPMT_SIG_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIG_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  result = TPMT_SIG_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_SIG_SCHEME *)memcpy((TPMT_SIG_SCHEME *)malloc(sizeof(TPMT_SIG_SCHEME)),&result,sizeof(TPMT_SIG_SCHEME)), SWIGTYPE_p_TPMT_SIG_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SIG_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SIG_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMT_SIG_SCHEME **)new_TPMT_SIG_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SIG_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SIG_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SIG_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  result = (TPMT_SIG_SCHEME **)copy_TPMT_SIG_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIG_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME **arg1 = (TPMT_SIG_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIG_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIG_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_SIG_SCHEME **""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME **)(argp1);
  delete_TPMT_SIG_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME **arg1 = (TPMT_SIG_SCHEME **) 0 ;
  TPMT_SIG_SCHEME *arg2 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIG_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_SIG_SCHEME **""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SIG_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg2 = (TPMT_SIG_SCHEME *)(argp2);
  TPMT_SIG_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIG_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME **arg1 = (TPMT_SIG_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIG_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIG_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIG_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_SIG_SCHEME **""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME **)(argp1);
  result = (TPMT_SIG_SCHEME *)TPMT_SIG_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_KDF_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_KDF_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMU_KDF_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_KDF_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_KDF_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_KDF_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_KDF_SCHEME_ARRAY *)new_TPMU_KDF_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_KDF_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME_ARRAY *arg1 = (TPMU_KDF_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_KDF_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_KDF_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMU_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME_ARRAY *)(argp1);
  delete_TPMU_KDF_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME_ARRAY *arg1 = (TPMU_KDF_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_KDF_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_KDF_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_KDF_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_KDF_SCHEME *)memcpy((TPMU_KDF_SCHEME *)malloc(sizeof(TPMU_KDF_SCHEME)),&result,sizeof(TPMU_KDF_SCHEME)), SWIGTYPE_p_TPMU_KDF_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME_ARRAY *arg1 = (TPMU_KDF_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_KDF_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_KDF_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_KDF_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_KDF_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_KDF_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_KDF_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_KDF_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_KDF_SCHEME""'");
    } else {
      arg3 = *((TPMU_KDF_SCHEME *)(argp3));
    }
  }
  TPMU_KDF_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME_ARRAY *arg1 = (TPMU_KDF_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMU_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME_ARRAY *)(argp1);
  result = (TPMU_KDF_SCHEME *)TPMU_KDF_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = (TPMU_KDF_SCHEME_ARRAY *)TPMU_KDF_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_KDF_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_KDF_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_KDF_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_KDF_SCHEME_PTR")) SWIG_fail;
  result = (TPMU_KDF_SCHEME *)new_TPMU_KDF_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_KDF_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_KDF_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_KDF_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_KDF_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_KDF_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_KDF_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_KDF_SCHEME""'");
    } else {
      arg1 = *((TPMU_KDF_SCHEME *)(argp1));
    }
  }
  result = (TPMU_KDF_SCHEME *)copy_TPMU_KDF_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_KDF_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_KDF_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_KDF_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  delete_TPMU_KDF_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  TPMU_KDF_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_KDF_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_KDF_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMU_KDF_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_KDF_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMU_KDF_SCHEME""'");
    } else {
      arg2 = *((TPMU_KDF_SCHEME *)(argp2));
    }
  }
  TPMU_KDF_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = TPMU_KDF_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_KDF_SCHEME *)memcpy((TPMU_KDF_SCHEME *)malloc(sizeof(TPMU_KDF_SCHEME)),&result,sizeof(TPMU_KDF_SCHEME)), SWIGTYPE_p_TPMU_KDF_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_KDF_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_KDF_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMU_KDF_SCHEME **)new_TPMU_KDF_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_KDF_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_KDF_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_KDF_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  result = (TPMU_KDF_SCHEME **)copy_TPMU_KDF_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_KDF_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME **arg1 = (TPMU_KDF_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_KDF_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_KDF_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMU_KDF_SCHEME **""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME **)(argp1);
  delete_TPMU_KDF_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME **arg1 = (TPMU_KDF_SCHEME **) 0 ;
  TPMU_KDF_SCHEME *arg2 = (TPMU_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_KDF_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_KDF_SCHEME **""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_KDF_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg2 = (TPMU_KDF_SCHEME *)(argp2);
  TPMU_KDF_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_KDF_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME **arg1 = (TPMU_KDF_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_KDF_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_KDF_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_KDF_SCHEME **""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME **)(argp1);
  result = (TPMU_KDF_SCHEME *)TPMU_KDF_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_KDF_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_KDF_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMT_KDF_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_KDF_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_KDF_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_KDF_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_KDF_SCHEME_ARRAY *)new_TPMT_KDF_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KDF_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME_ARRAY *arg1 = (TPMT_KDF_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KDF_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KDF_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMT_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME_ARRAY *)(argp1);
  delete_TPMT_KDF_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME_ARRAY *arg1 = (TPMT_KDF_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_KDF_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KDF_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_KDF_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_KDF_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_KDF_SCHEME *)memcpy((TPMT_KDF_SCHEME *)malloc(sizeof(TPMT_KDF_SCHEME)),&result,sizeof(TPMT_KDF_SCHEME)), SWIGTYPE_p_TPMT_KDF_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME_ARRAY *arg1 = (TPMT_KDF_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_KDF_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_KDF_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_KDF_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_KDF_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_KDF_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_KDF_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_KDF_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_KDF_SCHEME""'");
    } else {
      arg3 = *((TPMT_KDF_SCHEME *)(argp3));
    }
  }
  TPMT_KDF_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME_ARRAY *arg1 = (TPMT_KDF_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KDF_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMT_KDF_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME_ARRAY *)(argp1);
  result = (TPMT_KDF_SCHEME *)TPMT_KDF_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KDF_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  result = (TPMT_KDF_SCHEME_ARRAY *)TPMT_KDF_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_KDF_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_KDF_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_KDF_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_KDF_SCHEME_PTR")) SWIG_fail;
  result = (TPMT_KDF_SCHEME *)new_TPMT_KDF_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_KDF_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_KDF_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_KDF_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_KDF_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_KDF_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_KDF_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_KDF_SCHEME""'");
    } else {
      arg1 = *((TPMT_KDF_SCHEME *)(argp1));
    }
  }
  result = (TPMT_KDF_SCHEME *)copy_TPMT_KDF_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KDF_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KDF_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KDF_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  delete_TPMT_KDF_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  TPMT_KDF_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KDF_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_KDF_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_KDF_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_KDF_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_KDF_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_KDF_SCHEME""'");
    } else {
      arg2 = *((TPMT_KDF_SCHEME *)(argp2));
    }
  }
  TPMT_KDF_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KDF_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  result = TPMT_KDF_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_KDF_SCHEME *)memcpy((TPMT_KDF_SCHEME *)malloc(sizeof(TPMT_KDF_SCHEME)),&result,sizeof(TPMT_KDF_SCHEME)), SWIGTYPE_p_TPMT_KDF_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_KDF_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_KDF_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMT_KDF_SCHEME **)new_TPMT_KDF_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_KDF_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_KDF_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_KDF_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  result = (TPMT_KDF_SCHEME **)copy_TPMT_KDF_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_KDF_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME **arg1 = (TPMT_KDF_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_KDF_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_KDF_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_KDF_SCHEME **""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME **)(argp1);
  delete_TPMT_KDF_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME **arg1 = (TPMT_KDF_SCHEME **) 0 ;
  TPMT_KDF_SCHEME *arg2 = (TPMT_KDF_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_KDF_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_KDF_SCHEME **""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_KDF_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg2 = (TPMT_KDF_SCHEME *)(argp2);
  TPMT_KDF_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_KDF_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME **arg1 = (TPMT_KDF_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_KDF_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_KDF_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_KDF_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_KDF_SCHEME **""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME **)(argp1);
  result = (TPMT_KDF_SCHEME *)TPMT_KDF_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_ASYM_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_ASYM_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMU_ASYM_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_ASYM_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_ASYM_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_ASYM_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_ASYM_SCHEME_ARRAY *)new_TPMU_ASYM_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ASYM_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME_ARRAY *arg1 = (TPMU_ASYM_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ASYM_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ASYM_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME_ARRAY *)(argp1);
  delete_TPMU_ASYM_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME_ARRAY *arg1 = (TPMU_ASYM_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_ASYM_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_ASYM_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_ASYM_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_ASYM_SCHEME *)memcpy((TPMU_ASYM_SCHEME *)malloc(sizeof(TPMU_ASYM_SCHEME)),&result,sizeof(TPMU_ASYM_SCHEME)), SWIGTYPE_p_TPMU_ASYM_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME_ARRAY *arg1 = (TPMU_ASYM_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_ASYM_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_ASYM_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_ASYM_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_ASYM_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_ASYM_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_ASYM_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_ASYM_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_ASYM_SCHEME""'");
    } else {
      arg3 = *((TPMU_ASYM_SCHEME *)(argp3));
    }
  }
  TPMU_ASYM_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME_ARRAY *arg1 = (TPMU_ASYM_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME_ARRAY *)(argp1);
  result = (TPMU_ASYM_SCHEME *)TPMU_ASYM_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMU_ASYM_SCHEME_ARRAY *)TPMU_ASYM_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_ASYM_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_ASYM_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_ASYM_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ASYM_SCHEME_PTR")) SWIG_fail;
  result = (TPMU_ASYM_SCHEME *)new_TPMU_ASYM_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_ASYM_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_ASYM_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_ASYM_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_ASYM_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_ASYM_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME""'");
    } else {
      arg1 = *((TPMU_ASYM_SCHEME *)(argp1));
    }
  }
  result = (TPMU_ASYM_SCHEME *)copy_TPMU_ASYM_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ASYM_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ASYM_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ASYM_SCHEME_PTR" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  delete_TPMU_ASYM_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  TPMU_ASYM_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_ASYM_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_ASYM_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME""'");
    } else {
      arg2 = *((TPMU_ASYM_SCHEME *)(argp2));
    }
  }
  TPMU_ASYM_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = TPMU_ASYM_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_ASYM_SCHEME *)memcpy((TPMU_ASYM_SCHEME *)malloc(sizeof(TPMU_ASYM_SCHEME)),&result,sizeof(TPMU_ASYM_SCHEME)), SWIGTYPE_p_TPMU_ASYM_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_ASYM_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ASYM_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMU_ASYM_SCHEME **)new_TPMU_ASYM_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_ASYM_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_ASYM_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_ASYM_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  result = (TPMU_ASYM_SCHEME **)copy_TPMU_ASYM_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ASYM_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME **arg1 = (TPMU_ASYM_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ASYM_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ASYM_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME **""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME **)(argp1);
  delete_TPMU_ASYM_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME **arg1 = (TPMU_ASYM_SCHEME **) 0 ;
  TPMU_ASYM_SCHEME *arg2 = (TPMU_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ASYM_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME **""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ASYM_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMU_ASYM_SCHEME *)(argp2);
  TPMU_ASYM_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ASYM_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME **arg1 = (TPMU_ASYM_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ASYM_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ASYM_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME **""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME **)(argp1);
  result = (TPMU_ASYM_SCHEME *)TPMU_ASYM_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_ASYM_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_ASYM_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMT_ASYM_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_ASYM_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_ASYM_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_ASYM_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_ASYM_SCHEME_ARRAY *)new_TPMT_ASYM_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ASYM_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME_ARRAY *arg1 = (TPMT_ASYM_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ASYM_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ASYM_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME_ARRAY *)(argp1);
  delete_TPMT_ASYM_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME_ARRAY *arg1 = (TPMT_ASYM_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_ASYM_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ASYM_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_ASYM_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_ASYM_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_ASYM_SCHEME *)memcpy((TPMT_ASYM_SCHEME *)malloc(sizeof(TPMT_ASYM_SCHEME)),&result,sizeof(TPMT_ASYM_SCHEME)), SWIGTYPE_p_TPMT_ASYM_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME_ARRAY *arg1 = (TPMT_ASYM_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_ASYM_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_ASYM_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_ASYM_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_ASYM_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_ASYM_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_ASYM_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_ASYM_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_ASYM_SCHEME""'");
    } else {
      arg3 = *((TPMT_ASYM_SCHEME *)(argp3));
    }
  }
  TPMT_ASYM_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME_ARRAY *arg1 = (TPMT_ASYM_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ASYM_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME_ARRAY *)(argp1);
  result = (TPMT_ASYM_SCHEME *)TPMT_ASYM_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ASYM_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  result = (TPMT_ASYM_SCHEME_ARRAY *)TPMT_ASYM_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_ASYM_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_ASYM_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_ASYM_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_ASYM_SCHEME_PTR")) SWIG_fail;
  result = (TPMT_ASYM_SCHEME *)new_TPMT_ASYM_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_ASYM_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_ASYM_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_ASYM_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_ASYM_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_ASYM_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME""'");
    } else {
      arg1 = *((TPMT_ASYM_SCHEME *)(argp1));
    }
  }
  result = (TPMT_ASYM_SCHEME *)copy_TPMT_ASYM_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ASYM_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ASYM_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ASYM_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  delete_TPMT_ASYM_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  TPMT_ASYM_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ASYM_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_ASYM_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_ASYM_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_ASYM_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_ASYM_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_ASYM_SCHEME""'");
    } else {
      arg2 = *((TPMT_ASYM_SCHEME *)(argp2));
    }
  }
  TPMT_ASYM_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ASYM_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  result = TPMT_ASYM_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_ASYM_SCHEME *)memcpy((TPMT_ASYM_SCHEME *)malloc(sizeof(TPMT_ASYM_SCHEME)),&result,sizeof(TPMT_ASYM_SCHEME)), SWIGTYPE_p_TPMT_ASYM_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_ASYM_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_ASYM_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMT_ASYM_SCHEME **)new_TPMT_ASYM_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_ASYM_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_ASYM_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_ASYM_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  result = (TPMT_ASYM_SCHEME **)copy_TPMT_ASYM_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ASYM_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME **arg1 = (TPMT_ASYM_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ASYM_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ASYM_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME **""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME **)(argp1);
  delete_TPMT_ASYM_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME **arg1 = (TPMT_ASYM_SCHEME **) 0 ;
  TPMT_ASYM_SCHEME *arg2 = (TPMT_ASYM_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ASYM_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME **""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_ASYM_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg2 = (TPMT_ASYM_SCHEME *)(argp2);
  TPMT_ASYM_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ASYM_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME **arg1 = (TPMT_ASYM_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ASYM_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ASYM_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ASYM_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME **""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME **)(argp1);
  result = (TPMT_ASYM_SCHEME *)TPMT_ASYM_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_RSA_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_RSA_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMT_RSA_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_RSA_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_RSA_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_RSA_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_RSA_SCHEME_ARRAY *)new_TPMT_RSA_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME_ARRAY *arg1 = (TPMT_RSA_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMT_RSA_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME_ARRAY *)(argp1);
  delete_TPMT_RSA_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME_ARRAY *arg1 = (TPMT_RSA_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_RSA_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_RSA_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_RSA_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_RSA_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_RSA_SCHEME *)memcpy((TPMT_RSA_SCHEME *)malloc(sizeof(TPMT_RSA_SCHEME)),&result,sizeof(TPMT_RSA_SCHEME)), SWIGTYPE_p_TPMT_RSA_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME_ARRAY *arg1 = (TPMT_RSA_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_RSA_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_RSA_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_RSA_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_RSA_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_RSA_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_RSA_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_RSA_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_RSA_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_RSA_SCHEME""'");
    } else {
      arg3 = *((TPMT_RSA_SCHEME *)(argp3));
    }
  }
  TPMT_RSA_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME_ARRAY *arg1 = (TPMT_RSA_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMT_RSA_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME_ARRAY *)(argp1);
  result = (TPMT_RSA_SCHEME *)TPMT_RSA_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  result = (TPMT_RSA_SCHEME_ARRAY *)TPMT_RSA_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_RSA_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_RSA_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_RSA_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_RSA_SCHEME_PTR")) SWIG_fail;
  result = (TPMT_RSA_SCHEME *)new_TPMT_RSA_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_RSA_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_RSA_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_RSA_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_RSA_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_RSA_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_RSA_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_RSA_SCHEME""'");
    } else {
      arg1 = *((TPMT_RSA_SCHEME *)(argp1));
    }
  }
  result = (TPMT_RSA_SCHEME *)copy_TPMT_RSA_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  delete_TPMT_RSA_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  TPMT_RSA_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_RSA_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_RSA_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_RSA_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_RSA_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_RSA_SCHEME""'");
    } else {
      arg2 = *((TPMT_RSA_SCHEME *)(argp2));
    }
  }
  TPMT_RSA_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  result = TPMT_RSA_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_RSA_SCHEME *)memcpy((TPMT_RSA_SCHEME *)malloc(sizeof(TPMT_RSA_SCHEME)),&result,sizeof(TPMT_RSA_SCHEME)), SWIGTYPE_p_TPMT_RSA_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_RSA_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_RSA_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMT_RSA_SCHEME **)new_TPMT_RSA_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_RSA_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_RSA_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_RSA_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  result = (TPMT_RSA_SCHEME **)copy_TPMT_RSA_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME **arg1 = (TPMT_RSA_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_RSA_SCHEME **""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME **)(argp1);
  delete_TPMT_RSA_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME **arg1 = (TPMT_RSA_SCHEME **) 0 ;
  TPMT_RSA_SCHEME *arg2 = (TPMT_RSA_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_RSA_SCHEME **""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_RSA_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg2 = (TPMT_RSA_SCHEME *)(argp2);
  TPMT_RSA_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME **arg1 = (TPMT_RSA_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_RSA_SCHEME **""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME **)(argp1);
  result = (TPMT_RSA_SCHEME *)TPMT_RSA_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_RSA_DECRYPT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_RSA_DECRYPT",SWIG_From_size_t((size_t)(sizeof(TPMT_RSA_DECRYPT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_RSA_DECRYPT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_RSA_DECRYPT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_RSA_DECRYPT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_RSA_DECRYPT_ARRAY *)new_TPMT_RSA_DECRYPT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_DECRYPT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT_ARRAY *arg1 = (TPMT_RSA_DECRYPT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_DECRYPT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_DECRYPT_ARRAY" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT_ARRAY *)(argp1);
  delete_TPMT_RSA_DECRYPT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT_ARRAY *arg1 = (TPMT_RSA_DECRYPT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_RSA_DECRYPT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_DECRYPT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_RSA_DECRYPT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_RSA_DECRYPT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_RSA_DECRYPT *)memcpy((TPMT_RSA_DECRYPT *)malloc(sizeof(TPMT_RSA_DECRYPT)),&result,sizeof(TPMT_RSA_DECRYPT)), SWIGTYPE_p_TPMT_RSA_DECRYPT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT_ARRAY *arg1 = (TPMT_RSA_DECRYPT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_RSA_DECRYPT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_RSA_DECRYPT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_RSA_DECRYPT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_RSA_DECRYPT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_RSA_DECRYPT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_RSA_DECRYPT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_RSA_DECRYPT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_RSA_DECRYPT""'");
    } else {
      arg3 = *((TPMT_RSA_DECRYPT *)(argp3));
    }
  }
  TPMT_RSA_DECRYPT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT_ARRAY *arg1 = (TPMT_RSA_DECRYPT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_DECRYPT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_ARRAY_cast" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT_ARRAY *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT_ARRAY *)(argp1);
  result = (TPMT_RSA_DECRYPT *)TPMT_RSA_DECRYPT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_DECRYPT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  result = (TPMT_RSA_DECRYPT_ARRAY *)TPMT_RSA_DECRYPT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_RSA_DECRYPT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_RSA_DECRYPT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_RSA_DECRYPT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_RSA_DECRYPT_PTR")) SWIG_fail;
  result = (TPMT_RSA_DECRYPT *)new_TPMT_RSA_DECRYPT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_RSA_DECRYPT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_RSA_DECRYPT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_RSA_DECRYPT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_RSA_DECRYPT_PTR" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_RSA_DECRYPT_PTR" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT""'");
    } else {
      arg1 = *((TPMT_RSA_DECRYPT *)(argp1));
    }
  }
  result = (TPMT_RSA_DECRYPT *)copy_TPMT_RSA_DECRYPT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_DECRYPT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_DECRYPT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_DECRYPT_PTR" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  delete_TPMT_RSA_DECRYPT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  TPMT_RSA_DECRYPT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_DECRYPT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_PTR_assign" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_RSA_DECRYPT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_RSA_DECRYPT_PTR_assign" "', argument " "2"" of type '" "TPMT_RSA_DECRYPT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_RSA_DECRYPT_PTR_assign" "', argument " "2"" of type '" "TPMT_RSA_DECRYPT""'");
    } else {
      arg2 = *((TPMT_RSA_DECRYPT *)(argp2));
    }
  }
  TPMT_RSA_DECRYPT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_DECRYPT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_PTR_value" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  result = TPMT_RSA_DECRYPT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_RSA_DECRYPT *)memcpy((TPMT_RSA_DECRYPT *)malloc(sizeof(TPMT_RSA_DECRYPT)),&result,sizeof(TPMT_RSA_DECRYPT)), SWIGTYPE_p_TPMT_RSA_DECRYPT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_RSA_DECRYPT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_RSA_DECRYPT_PTR_PTR")) SWIG_fail;
  result = (TPMT_RSA_DECRYPT **)new_TPMT_RSA_DECRYPT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_RSA_DECRYPT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_RSA_DECRYPT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_RSA_DECRYPT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_RSA_DECRYPT_PTR_PTR" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  result = (TPMT_RSA_DECRYPT **)copy_TPMT_RSA_DECRYPT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_RSA_DECRYPT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_RSA_DECRYPT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT **arg1 = (TPMT_RSA_DECRYPT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_RSA_DECRYPT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_RSA_DECRYPT_PTR_PTR" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT **""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT **)(argp1);
  delete_TPMT_RSA_DECRYPT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT **arg1 = (TPMT_RSA_DECRYPT **) 0 ;
  TPMT_RSA_DECRYPT *arg2 = (TPMT_RSA_DECRYPT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_RSA_DECRYPT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT **""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_RSA_DECRYPT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg2 = (TPMT_RSA_DECRYPT *)(argp2);
  TPMT_RSA_DECRYPT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_RSA_DECRYPT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT **arg1 = (TPMT_RSA_DECRYPT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_RSA_DECRYPT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_RSA_DECRYPT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_RSA_DECRYPT_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT **""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT **)(argp1);
  result = (TPMT_RSA_DECRYPT *)TPMT_RSA_DECRYPT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_PUBLIC_KEY_RSA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_PUBLIC_KEY_RSA",SWIG_From_size_t((size_t)(sizeof(TPM2B_PUBLIC_KEY_RSA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_KEY_RSA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_PUBLIC_KEY_RSA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_PUBLIC_KEY_RSA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_PUBLIC_KEY_RSA_ARRAY *)new_TPM2B_PUBLIC_KEY_RSA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_KEY_RSA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA_ARRAY *arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_KEY_RSA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_KEY_RSA_ARRAY" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *)(argp1);
  delete_TPM2B_PUBLIC_KEY_RSA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA_ARRAY *arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_PUBLIC_KEY_RSA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_PUBLIC_KEY_RSA *)memcpy((TPM2B_PUBLIC_KEY_RSA *)malloc(sizeof(TPM2B_PUBLIC_KEY_RSA)),&result,sizeof(TPM2B_PUBLIC_KEY_RSA)), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA_ARRAY *arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_PUBLIC_KEY_RSA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PUBLIC_KEY_RSA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PUBLIC_KEY_RSA""'");
    } else {
      arg3 = *((TPM2B_PUBLIC_KEY_RSA *)(argp3));
    }
  }
  TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA_ARRAY *arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_KEY_RSA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA_ARRAY *)(argp1);
  result = (TPM2B_PUBLIC_KEY_RSA *)TPM2B_PUBLIC_KEY_RSA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  result = (TPM2B_PUBLIC_KEY_RSA_ARRAY *)TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PUBLIC_KEY_RSA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_KEY_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PUBLIC_KEY_RSA_PTR")) SWIG_fail;
  result = (TPM2B_PUBLIC_KEY_RSA *)new_TPM2B_PUBLIC_KEY_RSA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PUBLIC_KEY_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PUBLIC_KEY_RSA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PUBLIC_KEY_RSA_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_PUBLIC_KEY_RSA_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA""'");
    } else {
      arg1 = *((TPM2B_PUBLIC_KEY_RSA *)(argp1));
    }
  }
  result = (TPM2B_PUBLIC_KEY_RSA *)copy_TPM2B_PUBLIC_KEY_RSA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_KEY_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_KEY_RSA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_KEY_RSA_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  delete_TPM2B_PUBLIC_KEY_RSA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  TPM2B_PUBLIC_KEY_RSA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_KEY_RSA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_assign" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_assign" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_assign" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA""'");
    } else {
      arg2 = *((TPM2B_PUBLIC_KEY_RSA *)(argp2));
    }
  }
  TPM2B_PUBLIC_KEY_RSA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_KEY_RSA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_value" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  result = TPM2B_PUBLIC_KEY_RSA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_PUBLIC_KEY_RSA *)memcpy((TPM2B_PUBLIC_KEY_RSA *)malloc(sizeof(TPM2B_PUBLIC_KEY_RSA)),&result,sizeof(TPM2B_PUBLIC_KEY_RSA)), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PUBLIC_KEY_RSA_PTR_PTR")) SWIG_fail;
  result = (TPM2B_PUBLIC_KEY_RSA **)new_TPM2B_PUBLIC_KEY_RSA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  result = (TPM2B_PUBLIC_KEY_RSA **)copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA **arg1 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA **)(argp1);
  delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA **arg1 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  TPM2B_PUBLIC_KEY_RSA *arg2 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg2 = (TPM2B_PUBLIC_KEY_RSA *)(argp2);
  TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA **arg1 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA **)(argp1);
  result = (TPM2B_PUBLIC_KEY_RSA *)TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_PRIVATE_KEY_RSA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_PRIVATE_KEY_RSA",SWIG_From_size_t((size_t)(sizeof(TPM2B_PRIVATE_KEY_RSA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_KEY_RSA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_PRIVATE_KEY_RSA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_PRIVATE_KEY_RSA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_PRIVATE_KEY_RSA_ARRAY *)new_TPM2B_PRIVATE_KEY_RSA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_KEY_RSA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA_ARRAY *arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_KEY_RSA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_KEY_RSA_ARRAY" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *)(argp1);
  delete_TPM2B_PRIVATE_KEY_RSA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA_ARRAY *arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_PRIVATE_KEY_RSA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_PRIVATE_KEY_RSA *)memcpy((TPM2B_PRIVATE_KEY_RSA *)malloc(sizeof(TPM2B_PRIVATE_KEY_RSA)),&result,sizeof(TPM2B_PRIVATE_KEY_RSA)), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA_ARRAY *arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_PRIVATE_KEY_RSA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PRIVATE_KEY_RSA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PRIVATE_KEY_RSA""'");
    } else {
      arg3 = *((TPM2B_PRIVATE_KEY_RSA *)(argp3));
    }
  }
  TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA_ARRAY *arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_KEY_RSA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA_ARRAY *)(argp1);
  result = (TPM2B_PRIVATE_KEY_RSA *)TPM2B_PRIVATE_KEY_RSA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  result = (TPM2B_PRIVATE_KEY_RSA_ARRAY *)TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PRIVATE_KEY_RSA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_KEY_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_KEY_RSA_PTR")) SWIG_fail;
  result = (TPM2B_PRIVATE_KEY_RSA *)new_TPM2B_PRIVATE_KEY_RSA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PRIVATE_KEY_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PRIVATE_KEY_RSA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PRIVATE_KEY_RSA_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_PRIVATE_KEY_RSA_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA""'");
    } else {
      arg1 = *((TPM2B_PRIVATE_KEY_RSA *)(argp1));
    }
  }
  result = (TPM2B_PRIVATE_KEY_RSA *)copy_TPM2B_PRIVATE_KEY_RSA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_KEY_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_KEY_RSA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_KEY_RSA_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  delete_TPM2B_PRIVATE_KEY_RSA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  TPM2B_PRIVATE_KEY_RSA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_KEY_RSA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_assign" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE_KEY_RSA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE_KEY_RSA""'");
    } else {
      arg2 = *((TPM2B_PRIVATE_KEY_RSA *)(argp2));
    }
  }
  TPM2B_PRIVATE_KEY_RSA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_KEY_RSA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_value" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  result = TPM2B_PRIVATE_KEY_RSA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_PRIVATE_KEY_RSA *)memcpy((TPM2B_PRIVATE_KEY_RSA *)malloc(sizeof(TPM2B_PRIVATE_KEY_RSA)),&result,sizeof(TPM2B_PRIVATE_KEY_RSA)), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_KEY_RSA_PTR_PTR")) SWIG_fail;
  result = (TPM2B_PRIVATE_KEY_RSA **)new_TPM2B_PRIVATE_KEY_RSA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  result = (TPM2B_PRIVATE_KEY_RSA **)copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA **arg1 = (TPM2B_PRIVATE_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA **""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA **)(argp1);
  delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA **arg1 = (TPM2B_PRIVATE_KEY_RSA **) 0 ;
  TPM2B_PRIVATE_KEY_RSA *arg2 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA **""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg2 = (TPM2B_PRIVATE_KEY_RSA *)(argp2);
  TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA **arg1 = (TPM2B_PRIVATE_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_KEY_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA **""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA **)(argp1);
  result = (TPM2B_PRIVATE_KEY_RSA *)TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_ECC_PARAMETER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_ECC_PARAMETER",SWIG_From_size_t((size_t)(sizeof(TPM2B_ECC_PARAMETER))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_PARAMETER_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_ECC_PARAMETER_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_ECC_PARAMETER_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_ECC_PARAMETER_ARRAY *)new_TPM2B_ECC_PARAMETER_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_PARAMETER_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER_ARRAY *arg1 = (TPM2B_ECC_PARAMETER_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_PARAMETER_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_PARAMETER_ARRAY" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER_ARRAY *)(argp1);
  delete_TPM2B_ECC_PARAMETER_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER_ARRAY *arg1 = (TPM2B_ECC_PARAMETER_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_ECC_PARAMETER result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_PARAMETER_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ECC_PARAMETER_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_ECC_PARAMETER_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_ECC_PARAMETER *)memcpy((TPM2B_ECC_PARAMETER *)malloc(sizeof(TPM2B_ECC_PARAMETER)),&result,sizeof(TPM2B_ECC_PARAMETER)), SWIGTYPE_p_TPM2B_ECC_PARAMETER, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER_ARRAY *arg1 = (TPM2B_ECC_PARAMETER_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_ECC_PARAMETER arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_ECC_PARAMETER_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ECC_PARAMETER_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_ECC_PARAMETER,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_ECC_PARAMETER_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ECC_PARAMETER""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ECC_PARAMETER_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ECC_PARAMETER""'");
    } else {
      arg3 = *((TPM2B_ECC_PARAMETER *)(argp3));
    }
  }
  TPM2B_ECC_PARAMETER_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER_ARRAY *arg1 = (TPM2B_ECC_PARAMETER_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_PARAMETER_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER_ARRAY *)(argp1);
  result = (TPM2B_ECC_PARAMETER *)TPM2B_ECC_PARAMETER_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_PARAMETER_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  result = (TPM2B_ECC_PARAMETER_ARRAY *)TPM2B_ECC_PARAMETER_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ECC_PARAMETER_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ECC_PARAMETER_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_PARAMETER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ECC_PARAMETER_PTR")) SWIG_fail;
  result = (TPM2B_ECC_PARAMETER *)new_TPM2B_ECC_PARAMETER_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ECC_PARAMETER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ECC_PARAMETER_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_ECC_PARAMETER,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ECC_PARAMETER_PTR" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_ECC_PARAMETER_PTR" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER""'");
    } else {
      arg1 = *((TPM2B_ECC_PARAMETER *)(argp1));
    }
  }
  result = (TPM2B_ECC_PARAMETER *)copy_TPM2B_ECC_PARAMETER_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_PARAMETER_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_PARAMETER_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_PARAMETER_PTR" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  delete_TPM2B_ECC_PARAMETER_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  TPM2B_ECC_PARAMETER arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_PARAMETER_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_PTR_assign" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_ECC_PARAMETER,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ECC_PARAMETER_PTR_assign" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ECC_PARAMETER_PTR_assign" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER""'");
    } else {
      arg2 = *((TPM2B_ECC_PARAMETER *)(argp2));
    }
  }
  TPM2B_ECC_PARAMETER_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_PARAMETER_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_PTR_value" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  result = TPM2B_ECC_PARAMETER_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_ECC_PARAMETER *)memcpy((TPM2B_ECC_PARAMETER *)malloc(sizeof(TPM2B_ECC_PARAMETER)),&result,sizeof(TPM2B_ECC_PARAMETER)), SWIGTYPE_p_TPM2B_ECC_PARAMETER, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_PARAMETER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ECC_PARAMETER_PTR_PTR")) SWIG_fail;
  result = (TPM2B_ECC_PARAMETER **)new_TPM2B_ECC_PARAMETER_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ECC_PARAMETER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ECC_PARAMETER_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ECC_PARAMETER_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  result = (TPM2B_ECC_PARAMETER **)copy_TPM2B_ECC_PARAMETER_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_PARAMETER_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER **arg1 = (TPM2B_ECC_PARAMETER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_PARAMETER_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_PARAMETER_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER **""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER **)(argp1);
  delete_TPM2B_ECC_PARAMETER_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER **arg1 = (TPM2B_ECC_PARAMETER **) 0 ;
  TPM2B_ECC_PARAMETER *arg2 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_PARAMETER_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER **""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ECC_PARAMETER_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg2 = (TPM2B_ECC_PARAMETER *)(argp2);
  TPM2B_ECC_PARAMETER_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_PARAMETER_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER **arg1 = (TPM2B_ECC_PARAMETER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_PARAMETER *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_PARAMETER_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_PARAMETER_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER **""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER **)(argp1);
  result = (TPM2B_ECC_PARAMETER *)TPM2B_ECC_PARAMETER_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ECC_POINT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ECC_POINT",SWIG_From_size_t((size_t)(sizeof(TPMS_ECC_POINT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ECC_POINT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ECC_POINT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ECC_POINT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ECC_POINT_ARRAY *)new_TPMS_ECC_POINT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_POINT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT_ARRAY *arg1 = (TPMS_ECC_POINT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_POINT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_POINT_ARRAY" "', argument " "1"" of type '" "TPMS_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_POINT_ARRAY *)(argp1);
  delete_TPMS_ECC_POINT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT_ARRAY *arg1 = (TPMS_ECC_POINT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ECC_POINT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_POINT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_POINT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ECC_POINT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ECC_POINT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ECC_POINT *)memcpy((TPMS_ECC_POINT *)malloc(sizeof(TPMS_ECC_POINT)),&result,sizeof(TPMS_ECC_POINT)), SWIGTYPE_p_TPMS_ECC_POINT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT_ARRAY *arg1 = (TPMS_ECC_POINT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ECC_POINT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ECC_POINT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_POINT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ECC_POINT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ECC_POINT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ECC_POINT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ECC_POINT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ECC_POINT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ECC_POINT""'");
    } else {
      arg3 = *((TPMS_ECC_POINT *)(argp3));
    }
  }
  TPMS_ECC_POINT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT_ARRAY *arg1 = (TPMS_ECC_POINT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_POINT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_POINT_ARRAY *)(argp1);
  result = (TPMS_ECC_POINT *)TPMS_ECC_POINT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_POINT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  result = (TPMS_ECC_POINT_ARRAY *)TPMS_ECC_POINT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ECC_POINT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ECC_POINT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ECC_POINT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ECC_POINT_PTR")) SWIG_fail;
  result = (TPMS_ECC_POINT *)new_TPMS_ECC_POINT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ECC_POINT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ECC_POINT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ECC_POINT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ECC_POINT_PTR" "', argument " "1"" of type '" "TPMS_ECC_POINT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ECC_POINT_PTR" "', argument " "1"" of type '" "TPMS_ECC_POINT""'");
    } else {
      arg1 = *((TPMS_ECC_POINT *)(argp1));
    }
  }
  result = (TPMS_ECC_POINT *)copy_TPMS_ECC_POINT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_POINT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_POINT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_POINT_PTR" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  delete_TPMS_ECC_POINT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  TPMS_ECC_POINT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_POINT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_PTR_assign" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ECC_POINT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_POINT_PTR_assign" "', argument " "2"" of type '" "TPMS_ECC_POINT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ECC_POINT_PTR_assign" "', argument " "2"" of type '" "TPMS_ECC_POINT""'");
    } else {
      arg2 = *((TPMS_ECC_POINT *)(argp2));
    }
  }
  TPMS_ECC_POINT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_POINT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_PTR_value" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  result = TPMS_ECC_POINT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ECC_POINT *)memcpy((TPMS_ECC_POINT *)malloc(sizeof(TPMS_ECC_POINT)),&result,sizeof(TPMS_ECC_POINT)), SWIGTYPE_p_TPMS_ECC_POINT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ECC_POINT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ECC_POINT_PTR_PTR")) SWIG_fail;
  result = (TPMS_ECC_POINT **)new_TPMS_ECC_POINT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ECC_POINT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ECC_POINT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ECC_POINT_PTR_PTR" "', argument " "1"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  result = (TPMS_ECC_POINT **)copy_TPMS_ECC_POINT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_POINT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT **arg1 = (TPMS_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_POINT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_POINT_PTR_PTR" "', argument " "1"" of type '" "TPMS_ECC_POINT **""'"); 
  }
  arg1 = (TPMS_ECC_POINT **)(argp1);
  delete_TPMS_ECC_POINT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT **arg1 = (TPMS_ECC_POINT **) 0 ;
  TPMS_ECC_POINT *arg2 = (TPMS_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_POINT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ECC_POINT **""'"); 
  }
  arg1 = (TPMS_ECC_POINT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_POINT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg2 = (TPMS_ECC_POINT *)(argp2);
  TPMS_ECC_POINT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_POINT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT **arg1 = (TPMS_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_POINT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_POINT_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ECC_POINT **""'"); 
  }
  arg1 = (TPMS_ECC_POINT **)(argp1);
  result = (TPMS_ECC_POINT *)TPMS_ECC_POINT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_ECC_POINT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_ECC_POINT",SWIG_From_size_t((size_t)(sizeof(TPM2B_ECC_POINT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_POINT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_ECC_POINT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_ECC_POINT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_ECC_POINT_ARRAY *)new_TPM2B_ECC_POINT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_POINT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT_ARRAY *arg1 = (TPM2B_ECC_POINT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_POINT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_POINT_ARRAY" "', argument " "1"" of type '" "TPM2B_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT_ARRAY *)(argp1);
  delete_TPM2B_ECC_POINT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT_ARRAY *arg1 = (TPM2B_ECC_POINT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_ECC_POINT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_POINT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ECC_POINT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_ECC_POINT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_ECC_POINT *)memcpy((TPM2B_ECC_POINT *)malloc(sizeof(TPM2B_ECC_POINT)),&result,sizeof(TPM2B_ECC_POINT)), SWIGTYPE_p_TPM2B_ECC_POINT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT_ARRAY *arg1 = (TPM2B_ECC_POINT_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_ECC_POINT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_ECC_POINT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ECC_POINT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_ECC_POINT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_ECC_POINT_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ECC_POINT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ECC_POINT_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ECC_POINT""'");
    } else {
      arg3 = *((TPM2B_ECC_POINT *)(argp3));
    }
  }
  TPM2B_ECC_POINT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT_ARRAY *arg1 = (TPM2B_ECC_POINT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_POINT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_ECC_POINT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT_ARRAY *)(argp1);
  result = (TPM2B_ECC_POINT *)TPM2B_ECC_POINT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_POINT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  result = (TPM2B_ECC_POINT_ARRAY *)TPM2B_ECC_POINT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ECC_POINT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ECC_POINT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_POINT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ECC_POINT_PTR")) SWIG_fail;
  result = (TPM2B_ECC_POINT *)new_TPM2B_ECC_POINT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ECC_POINT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ECC_POINT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_ECC_POINT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ECC_POINT_PTR" "', argument " "1"" of type '" "TPM2B_ECC_POINT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_ECC_POINT_PTR" "', argument " "1"" of type '" "TPM2B_ECC_POINT""'");
    } else {
      arg1 = *((TPM2B_ECC_POINT *)(argp1));
    }
  }
  result = (TPM2B_ECC_POINT *)copy_TPM2B_ECC_POINT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_POINT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_POINT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_POINT_PTR" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  delete_TPM2B_ECC_POINT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  TPM2B_ECC_POINT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_POINT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_PTR_assign" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_ECC_POINT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ECC_POINT_PTR_assign" "', argument " "2"" of type '" "TPM2B_ECC_POINT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ECC_POINT_PTR_assign" "', argument " "2"" of type '" "TPM2B_ECC_POINT""'");
    } else {
      arg2 = *((TPM2B_ECC_POINT *)(argp2));
    }
  }
  TPM2B_ECC_POINT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_POINT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_PTR_value" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  result = TPM2B_ECC_POINT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_ECC_POINT *)memcpy((TPM2B_ECC_POINT *)malloc(sizeof(TPM2B_ECC_POINT)),&result,sizeof(TPM2B_ECC_POINT)), SWIGTYPE_p_TPM2B_ECC_POINT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ECC_POINT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ECC_POINT_PTR_PTR")) SWIG_fail;
  result = (TPM2B_ECC_POINT **)new_TPM2B_ECC_POINT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ECC_POINT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ECC_POINT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ECC_POINT_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  result = (TPM2B_ECC_POINT **)copy_TPM2B_ECC_POINT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ECC_POINT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT **arg1 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ECC_POINT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ECC_POINT_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg1 = (TPM2B_ECC_POINT **)(argp1);
  delete_TPM2B_ECC_POINT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT **arg1 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT *arg2 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ECC_POINT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg1 = (TPM2B_ECC_POINT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ECC_POINT_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg2 = (TPM2B_ECC_POINT *)(argp2);
  TPM2B_ECC_POINT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ECC_POINT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT **arg1 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ECC_POINT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ECC_POINT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ECC_POINT_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg1 = (TPM2B_ECC_POINT **)(argp1);
  result = (TPM2B_ECC_POINT *)TPM2B_ECC_POINT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_ECC_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_ECC_SCHEME",SWIG_From_size_t((size_t)(sizeof(TPMT_ECC_SCHEME))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_ECC_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_ECC_SCHEME_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_ECC_SCHEME_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_ECC_SCHEME_ARRAY *)new_TPMT_ECC_SCHEME_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ECC_SCHEME_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME_ARRAY *arg1 = (TPMT_ECC_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ECC_SCHEME_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ECC_SCHEME_ARRAY" "', argument " "1"" of type '" "TPMT_ECC_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME_ARRAY *)(argp1);
  delete_TPMT_ECC_SCHEME_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME_ARRAY *arg1 = (TPMT_ECC_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_ECC_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ECC_SCHEME_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_ECC_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_ECC_SCHEME_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_ECC_SCHEME_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_ECC_SCHEME *)memcpy((TPMT_ECC_SCHEME *)malloc(sizeof(TPMT_ECC_SCHEME)),&result,sizeof(TPMT_ECC_SCHEME)), SWIGTYPE_p_TPMT_ECC_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME_ARRAY *arg1 = (TPMT_ECC_SCHEME_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_ECC_SCHEME arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_ECC_SCHEME_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_ECC_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_ECC_SCHEME_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_ECC_SCHEME,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_ECC_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_ECC_SCHEME""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_ECC_SCHEME_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_ECC_SCHEME""'");
    } else {
      arg3 = *((TPMT_ECC_SCHEME *)(argp3));
    }
  }
  TPMT_ECC_SCHEME_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME_ARRAY *arg1 = (TPMT_ECC_SCHEME_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ECC_SCHEME_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_ARRAY_cast" "', argument " "1"" of type '" "TPMT_ECC_SCHEME_ARRAY *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME_ARRAY *)(argp1);
  result = (TPMT_ECC_SCHEME *)TPMT_ECC_SCHEME_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ECC_SCHEME_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  result = (TPMT_ECC_SCHEME_ARRAY *)TPMT_ECC_SCHEME_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_ECC_SCHEME_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_ECC_SCHEME_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_ECC_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_ECC_SCHEME_PTR")) SWIG_fail;
  result = (TPMT_ECC_SCHEME *)new_TPMT_ECC_SCHEME_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_ECC_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_ECC_SCHEME_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_ECC_SCHEME,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_ECC_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_ECC_SCHEME""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_ECC_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_ECC_SCHEME""'");
    } else {
      arg1 = *((TPMT_ECC_SCHEME *)(argp1));
    }
  }
  result = (TPMT_ECC_SCHEME *)copy_TPMT_ECC_SCHEME_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ECC_SCHEME_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ECC_SCHEME_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ECC_SCHEME_PTR" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  delete_TPMT_ECC_SCHEME_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  TPMT_ECC_SCHEME arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ECC_SCHEME_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_PTR_assign" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_ECC_SCHEME,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_ECC_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_ECC_SCHEME""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_ECC_SCHEME_PTR_assign" "', argument " "2"" of type '" "TPMT_ECC_SCHEME""'");
    } else {
      arg2 = *((TPMT_ECC_SCHEME *)(argp2));
    }
  }
  TPMT_ECC_SCHEME_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ECC_SCHEME_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_PTR_value" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  result = TPMT_ECC_SCHEME_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_ECC_SCHEME *)memcpy((TPMT_ECC_SCHEME *)malloc(sizeof(TPMT_ECC_SCHEME)),&result,sizeof(TPMT_ECC_SCHEME)), SWIGTYPE_p_TPMT_ECC_SCHEME, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_ECC_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_ECC_SCHEME_PTR_PTR")) SWIG_fail;
  result = (TPMT_ECC_SCHEME **)new_TPMT_ECC_SCHEME_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_ECC_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_ECC_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_ECC_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  result = (TPMT_ECC_SCHEME **)copy_TPMT_ECC_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_ECC_SCHEME_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME **arg1 = (TPMT_ECC_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_ECC_SCHEME_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_ECC_SCHEME_PTR_PTR" "', argument " "1"" of type '" "TPMT_ECC_SCHEME **""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME **)(argp1);
  delete_TPMT_ECC_SCHEME_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME **arg1 = (TPMT_ECC_SCHEME **) 0 ;
  TPMT_ECC_SCHEME *arg2 = (TPMT_ECC_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_ECC_SCHEME_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_ECC_SCHEME **""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_ECC_SCHEME_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg2 = (TPMT_ECC_SCHEME *)(argp2);
  TPMT_ECC_SCHEME_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_ECC_SCHEME_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME **arg1 = (TPMT_ECC_SCHEME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_ECC_SCHEME *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_ECC_SCHEME_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_ECC_SCHEME_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_ECC_SCHEME **""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME **)(argp1);
  result = (TPMT_ECC_SCHEME *)TPMT_ECC_SCHEME_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ALGORITHM_DETAIL_ECC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ALGORITHM_DETAIL_ECC",SWIG_From_size_t((size_t)(sizeof(TPMS_ALGORITHM_DETAIL_ECC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY *arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)(argp1);
  delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY *arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ALGORITHM_DETAIL_ECC *)memcpy((TPMS_ALGORITHM_DETAIL_ECC *)malloc(sizeof(TPMS_ALGORITHM_DETAIL_ECC)),&result,sizeof(TPMS_ALGORITHM_DETAIL_ECC)), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY *arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ALGORITHM_DETAIL_ECC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ALGORITHM_DETAIL_ECC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ALGORITHM_DETAIL_ECC""'");
    } else {
      arg3 = *((TPMS_ALGORITHM_DETAIL_ECC *)(argp3));
    }
  }
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY *arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)(argp1);
  result = (TPMS_ALGORITHM_DETAIL_ECC *)TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPMS_ALGORITHM_DETAIL_ECC_ARRAY *)TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ALGORITHM_DETAIL_ECC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DETAIL_ECC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALGORITHM_DETAIL_ECC_PTR")) SWIG_fail;
  result = (TPMS_ALGORITHM_DETAIL_ECC *)new_TPMS_ALGORITHM_DETAIL_ECC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ALGORITHM_DETAIL_ECC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ALGORITHM_DETAIL_ECC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ALGORITHM_DETAIL_ECC_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ALGORITHM_DETAIL_ECC_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC""'");
    } else {
      arg1 = *((TPMS_ALGORITHM_DETAIL_ECC *)(argp1));
    }
  }
  result = (TPMS_ALGORITHM_DETAIL_ECC *)copy_TPMS_ALGORITHM_DETAIL_ECC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DETAIL_ECC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DETAIL_ECC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DETAIL_ECC_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  delete_TPMS_ALGORITHM_DETAIL_ECC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  TPMS_ALGORITHM_DETAIL_ECC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_assign" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_assign" "', argument " "2"" of type '" "TPMS_ALGORITHM_DETAIL_ECC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_assign" "', argument " "2"" of type '" "TPMS_ALGORITHM_DETAIL_ECC""'");
    } else {
      arg2 = *((TPMS_ALGORITHM_DETAIL_ECC *)(argp2));
    }
  }
  TPMS_ALGORITHM_DETAIL_ECC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_value" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = TPMS_ALGORITHM_DETAIL_ECC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ALGORITHM_DETAIL_ECC *)memcpy((TPMS_ALGORITHM_DETAIL_ECC *)malloc(sizeof(TPMS_ALGORITHM_DETAIL_ECC)),&result,sizeof(TPMS_ALGORITHM_DETAIL_ECC)), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR")) SWIG_fail;
  result = (TPMS_ALGORITHM_DETAIL_ECC **)new_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  result = (TPMS_ALGORITHM_DETAIL_ECC **)copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC **arg1 = (TPMS_ALGORITHM_DETAIL_ECC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC **""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC **)(argp1);
  delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC **arg1 = (TPMS_ALGORITHM_DETAIL_ECC **) 0 ;
  TPMS_ALGORITHM_DETAIL_ECC *arg2 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC **""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg2 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp2);
  TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC **arg1 = (TPMS_ALGORITHM_DETAIL_ECC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ALGORITHM_DETAIL_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC **""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC **)(argp1);
  result = (TPMS_ALGORITHM_DETAIL_ECC *)TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SIGNATURE_RSA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SIGNATURE_RSA",SWIG_From_size_t((size_t)(sizeof(TPMS_SIGNATURE_RSA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_RSA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SIGNATURE_RSA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SIGNATURE_RSA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SIGNATURE_RSA_ARRAY *)new_TPMS_SIGNATURE_RSA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_RSA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA_ARRAY *arg1 = (TPMS_SIGNATURE_RSA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_RSA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_RSA_ARRAY" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA_ARRAY *)(argp1);
  delete_TPMS_SIGNATURE_RSA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA_ARRAY *arg1 = (TPMS_SIGNATURE_RSA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SIGNATURE_RSA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_RSA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SIGNATURE_RSA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SIGNATURE_RSA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SIGNATURE_RSA *)memcpy((TPMS_SIGNATURE_RSA *)malloc(sizeof(TPMS_SIGNATURE_RSA)),&result,sizeof(TPMS_SIGNATURE_RSA)), SWIGTYPE_p_TPMS_SIGNATURE_RSA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA_ARRAY *arg1 = (TPMS_SIGNATURE_RSA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SIGNATURE_RSA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SIGNATURE_RSA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SIGNATURE_RSA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SIGNATURE_RSA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SIGNATURE_RSA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SIGNATURE_RSA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SIGNATURE_RSA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SIGNATURE_RSA""'");
    } else {
      arg3 = *((TPMS_SIGNATURE_RSA *)(argp3));
    }
  }
  TPMS_SIGNATURE_RSA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA_ARRAY *arg1 = (TPMS_SIGNATURE_RSA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_RSA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA_ARRAY *)(argp1);
  result = (TPMS_SIGNATURE_RSA *)TPMS_SIGNATURE_RSA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_RSA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  result = (TPMS_SIGNATURE_RSA_ARRAY *)TPMS_SIGNATURE_RSA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SIGNATURE_RSA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SIGNATURE_RSA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SIGNATURE_RSA_PTR")) SWIG_fail;
  result = (TPMS_SIGNATURE_RSA *)new_TPMS_SIGNATURE_RSA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SIGNATURE_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SIGNATURE_RSA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SIGNATURE_RSA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SIGNATURE_RSA_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SIGNATURE_RSA_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA""'");
    } else {
      arg1 = *((TPMS_SIGNATURE_RSA *)(argp1));
    }
  }
  result = (TPMS_SIGNATURE_RSA *)copy_TPMS_SIGNATURE_RSA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_RSA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_RSA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_RSA_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  delete_TPMS_SIGNATURE_RSA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  TPMS_SIGNATURE_RSA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_RSA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_PTR_assign" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SIGNATURE_RSA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_RSA_PTR_assign" "', argument " "2"" of type '" "TPMS_SIGNATURE_RSA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SIGNATURE_RSA_PTR_assign" "', argument " "2"" of type '" "TPMS_SIGNATURE_RSA""'");
    } else {
      arg2 = *((TPMS_SIGNATURE_RSA *)(argp2));
    }
  }
  TPMS_SIGNATURE_RSA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_RSA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_PTR_value" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  result = TPMS_SIGNATURE_RSA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SIGNATURE_RSA *)memcpy((TPMS_SIGNATURE_RSA *)malloc(sizeof(TPMS_SIGNATURE_RSA)),&result,sizeof(TPMS_SIGNATURE_RSA)), SWIGTYPE_p_TPMS_SIGNATURE_RSA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SIGNATURE_RSA_PTR_PTR")) SWIG_fail;
  result = (TPMS_SIGNATURE_RSA **)new_TPMS_SIGNATURE_RSA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SIGNATURE_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SIGNATURE_RSA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SIGNATURE_RSA_PTR_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  result = (TPMS_SIGNATURE_RSA **)copy_TPMS_SIGNATURE_RSA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_RSA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA **arg1 = (TPMS_SIGNATURE_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_RSA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_RSA_PTR_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA **""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA **)(argp1);
  delete_TPMS_SIGNATURE_RSA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA **arg1 = (TPMS_SIGNATURE_RSA **) 0 ;
  TPMS_SIGNATURE_RSA *arg2 = (TPMS_SIGNATURE_RSA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_RSA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA **""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_RSA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_RSA *)(argp2);
  TPMS_SIGNATURE_RSA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_RSA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA **arg1 = (TPMS_SIGNATURE_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_RSA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_RSA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_RSA_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA **""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA **)(argp1);
  result = (TPMS_SIGNATURE_RSA *)TPMS_SIGNATURE_RSA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_SIGNATURE_ECC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_SIGNATURE_ECC",SWIG_From_size_t((size_t)(sizeof(TPMS_SIGNATURE_ECC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_ECC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_SIGNATURE_ECC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_SIGNATURE_ECC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_SIGNATURE_ECC_ARRAY *)new_TPMS_SIGNATURE_ECC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_ECC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC_ARRAY *arg1 = (TPMS_SIGNATURE_ECC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_ECC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_ECC_ARRAY" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC_ARRAY *)(argp1);
  delete_TPMS_SIGNATURE_ECC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC_ARRAY *arg1 = (TPMS_SIGNATURE_ECC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_SIGNATURE_ECC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_ECC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SIGNATURE_ECC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_SIGNATURE_ECC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_SIGNATURE_ECC *)memcpy((TPMS_SIGNATURE_ECC *)malloc(sizeof(TPMS_SIGNATURE_ECC)),&result,sizeof(TPMS_SIGNATURE_ECC)), SWIGTYPE_p_TPMS_SIGNATURE_ECC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC_ARRAY *arg1 = (TPMS_SIGNATURE_ECC_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_SIGNATURE_ECC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_SIGNATURE_ECC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_SIGNATURE_ECC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_SIGNATURE_ECC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_SIGNATURE_ECC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SIGNATURE_ECC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SIGNATURE_ECC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_SIGNATURE_ECC""'");
    } else {
      arg3 = *((TPMS_SIGNATURE_ECC *)(argp3));
    }
  }
  TPMS_SIGNATURE_ECC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC_ARRAY *arg1 = (TPMS_SIGNATURE_ECC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_ARRAY_cast" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC_ARRAY *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC_ARRAY *)(argp1);
  result = (TPMS_SIGNATURE_ECC *)TPMS_SIGNATURE_ECC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  result = (TPMS_SIGNATURE_ECC_ARRAY *)TPMS_SIGNATURE_ECC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_SIGNATURE_ECC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_SIGNATURE_ECC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_ECC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SIGNATURE_ECC_PTR")) SWIG_fail;
  result = (TPMS_SIGNATURE_ECC *)new_TPMS_SIGNATURE_ECC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SIGNATURE_ECC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SIGNATURE_ECC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_SIGNATURE_ECC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SIGNATURE_ECC_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_SIGNATURE_ECC_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC""'");
    } else {
      arg1 = *((TPMS_SIGNATURE_ECC *)(argp1));
    }
  }
  result = (TPMS_SIGNATURE_ECC *)copy_TPMS_SIGNATURE_ECC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_ECC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_ECC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_ECC_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  delete_TPMS_SIGNATURE_ECC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  TPMS_SIGNATURE_ECC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_ECC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_PTR_assign" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_SIGNATURE_ECC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_ECC_PTR_assign" "', argument " "2"" of type '" "TPMS_SIGNATURE_ECC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_SIGNATURE_ECC_PTR_assign" "', argument " "2"" of type '" "TPMS_SIGNATURE_ECC""'");
    } else {
      arg2 = *((TPMS_SIGNATURE_ECC *)(argp2));
    }
  }
  TPMS_SIGNATURE_ECC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_PTR_value" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  result = TPMS_SIGNATURE_ECC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_SIGNATURE_ECC *)memcpy((TPMS_SIGNATURE_ECC *)malloc(sizeof(TPMS_SIGNATURE_ECC)),&result,sizeof(TPMS_SIGNATURE_ECC)), SWIGTYPE_p_TPMS_SIGNATURE_ECC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_SIGNATURE_ECC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_SIGNATURE_ECC_PTR_PTR")) SWIG_fail;
  result = (TPMS_SIGNATURE_ECC **)new_TPMS_SIGNATURE_ECC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_SIGNATURE_ECC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_SIGNATURE_ECC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_SIGNATURE_ECC_PTR_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  result = (TPMS_SIGNATURE_ECC **)copy_TPMS_SIGNATURE_ECC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_SIGNATURE_ECC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC **arg1 = (TPMS_SIGNATURE_ECC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_SIGNATURE_ECC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_SIGNATURE_ECC_PTR_PTR" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC **""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC **)(argp1);
  delete_TPMS_SIGNATURE_ECC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC **arg1 = (TPMS_SIGNATURE_ECC **) 0 ;
  TPMS_SIGNATURE_ECC *arg2 = (TPMS_SIGNATURE_ECC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_SIGNATURE_ECC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC **""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_SIGNATURE_ECC_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg2 = (TPMS_SIGNATURE_ECC *)(argp2);
  TPMS_SIGNATURE_ECC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_SIGNATURE_ECC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC **arg1 = (TPMS_SIGNATURE_ECC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_SIGNATURE_ECC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_SIGNATURE_ECC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_SIGNATURE_ECC_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC **""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC **)(argp1);
  result = (TPMS_SIGNATURE_ECC *)TPMS_SIGNATURE_ECC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_SIGNATURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_SIGNATURE",SWIG_From_size_t((size_t)(sizeof(TPMU_SIGNATURE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_SIGNATURE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_SIGNATURE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_SIGNATURE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_SIGNATURE_ARRAY *)new_TPMU_SIGNATURE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIGNATURE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE_ARRAY *arg1 = (TPMU_SIGNATURE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIGNATURE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIGNATURE_ARRAY" "', argument " "1"" of type '" "TPMU_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIGNATURE_ARRAY *)(argp1);
  delete_TPMU_SIGNATURE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE_ARRAY *arg1 = (TPMU_SIGNATURE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_SIGNATURE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIGNATURE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SIGNATURE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_SIGNATURE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_SIGNATURE *)memcpy((TPMU_SIGNATURE *)malloc(sizeof(TPMU_SIGNATURE)),&result,sizeof(TPMU_SIGNATURE)), SWIGTYPE_p_TPMU_SIGNATURE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE_ARRAY *arg1 = (TPMU_SIGNATURE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_SIGNATURE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_SIGNATURE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIGNATURE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SIGNATURE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_SIGNATURE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_SIGNATURE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SIGNATURE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SIGNATURE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SIGNATURE""'");
    } else {
      arg3 = *((TPMU_SIGNATURE *)(argp3));
    }
  }
  TPMU_SIGNATURE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE_ARRAY *arg1 = (TPMU_SIGNATURE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ARRAY_cast" "', argument " "1"" of type '" "TPMU_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SIGNATURE_ARRAY *)(argp1);
  result = (TPMU_SIGNATURE *)TPMU_SIGNATURE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMU_SIGNATURE_ARRAY *)TPMU_SIGNATURE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SIGNATURE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SIGNATURE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_SIGNATURE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SIGNATURE_PTR")) SWIG_fail;
  result = (TPMU_SIGNATURE *)new_TPMU_SIGNATURE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SIGNATURE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SIGNATURE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_SIGNATURE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SIGNATURE_PTR" "', argument " "1"" of type '" "TPMU_SIGNATURE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_SIGNATURE_PTR" "', argument " "1"" of type '" "TPMU_SIGNATURE""'");
    } else {
      arg1 = *((TPMU_SIGNATURE *)(argp1));
    }
  }
  result = (TPMU_SIGNATURE *)copy_TPMU_SIGNATURE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIGNATURE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIGNATURE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIGNATURE_PTR" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  delete_TPMU_SIGNATURE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  TPMU_SIGNATURE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_PTR_assign" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_SIGNATURE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_PTR_assign" "', argument " "2"" of type '" "TPMU_SIGNATURE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SIGNATURE_PTR_assign" "', argument " "2"" of type '" "TPMU_SIGNATURE""'");
    } else {
      arg2 = *((TPMU_SIGNATURE *)(argp2));
    }
  }
  TPMU_SIGNATURE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_PTR_value" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = TPMU_SIGNATURE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_SIGNATURE *)memcpy((TPMU_SIGNATURE *)malloc(sizeof(TPMU_SIGNATURE)),&result,sizeof(TPMU_SIGNATURE)), SWIGTYPE_p_TPMU_SIGNATURE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SIGNATURE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SIGNATURE_PTR_PTR")) SWIG_fail;
  result = (TPMU_SIGNATURE **)new_TPMU_SIGNATURE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SIGNATURE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SIGNATURE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SIGNATURE_PTR_PTR" "', argument " "1"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  result = (TPMU_SIGNATURE **)copy_TPMU_SIGNATURE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SIGNATURE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE **arg1 = (TPMU_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SIGNATURE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SIGNATURE_PTR_PTR" "', argument " "1"" of type '" "TPMU_SIGNATURE **""'"); 
  }
  arg1 = (TPMU_SIGNATURE **)(argp1);
  delete_TPMU_SIGNATURE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE **arg1 = (TPMU_SIGNATURE **) 0 ;
  TPMU_SIGNATURE *arg2 = (TPMU_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SIGNATURE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_SIGNATURE **""'"); 
  }
  arg1 = (TPMU_SIGNATURE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SIGNATURE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg2 = (TPMU_SIGNATURE *)(argp2);
  TPMU_SIGNATURE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SIGNATURE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE **arg1 = (TPMU_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SIGNATURE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SIGNATURE_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_SIGNATURE **""'"); 
  }
  arg1 = (TPMU_SIGNATURE **)(argp1);
  result = (TPMU_SIGNATURE *)TPMU_SIGNATURE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_SIGNATURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_SIGNATURE",SWIG_From_size_t((size_t)(sizeof(TPMT_SIGNATURE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_SIGNATURE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_SIGNATURE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_SIGNATURE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_SIGNATURE_ARRAY *)new_TPMT_SIGNATURE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIGNATURE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE_ARRAY *arg1 = (TPMT_SIGNATURE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIGNATURE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIGNATURE_ARRAY" "', argument " "1"" of type '" "TPMT_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIGNATURE_ARRAY *)(argp1);
  delete_TPMT_SIGNATURE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE_ARRAY *arg1 = (TPMT_SIGNATURE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_SIGNATURE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIGNATURE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIGNATURE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SIGNATURE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_SIGNATURE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_SIGNATURE *)memcpy((TPMT_SIGNATURE *)malloc(sizeof(TPMT_SIGNATURE)),&result,sizeof(TPMT_SIGNATURE)), SWIGTYPE_p_TPMT_SIGNATURE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE_ARRAY *arg1 = (TPMT_SIGNATURE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_SIGNATURE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_SIGNATURE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIGNATURE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SIGNATURE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_SIGNATURE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_SIGNATURE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SIGNATURE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SIGNATURE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SIGNATURE""'");
    } else {
      arg3 = *((TPMT_SIGNATURE *)(argp3));
    }
  }
  TPMT_SIGNATURE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE_ARRAY *arg1 = (TPMT_SIGNATURE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIGNATURE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_ARRAY_cast" "', argument " "1"" of type '" "TPMT_SIGNATURE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SIGNATURE_ARRAY *)(argp1);
  result = (TPMT_SIGNATURE *)TPMT_SIGNATURE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIGNATURE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  result = (TPMT_SIGNATURE_ARRAY *)TPMT_SIGNATURE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SIGNATURE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SIGNATURE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_SIGNATURE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SIGNATURE_PTR")) SWIG_fail;
  result = (TPMT_SIGNATURE *)new_TPMT_SIGNATURE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SIGNATURE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SIGNATURE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_SIGNATURE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SIGNATURE_PTR" "', argument " "1"" of type '" "TPMT_SIGNATURE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_SIGNATURE_PTR" "', argument " "1"" of type '" "TPMT_SIGNATURE""'");
    } else {
      arg1 = *((TPMT_SIGNATURE *)(argp1));
    }
  }
  result = (TPMT_SIGNATURE *)copy_TPMT_SIGNATURE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIGNATURE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIGNATURE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIGNATURE_PTR" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  delete_TPMT_SIGNATURE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  TPMT_SIGNATURE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIGNATURE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_PTR_assign" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_SIGNATURE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SIGNATURE_PTR_assign" "', argument " "2"" of type '" "TPMT_SIGNATURE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SIGNATURE_PTR_assign" "', argument " "2"" of type '" "TPMT_SIGNATURE""'");
    } else {
      arg2 = *((TPMT_SIGNATURE *)(argp2));
    }
  }
  TPMT_SIGNATURE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIGNATURE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_PTR_value" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  result = TPMT_SIGNATURE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_SIGNATURE *)memcpy((TPMT_SIGNATURE *)malloc(sizeof(TPMT_SIGNATURE)),&result,sizeof(TPMT_SIGNATURE)), SWIGTYPE_p_TPMT_SIGNATURE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SIGNATURE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SIGNATURE_PTR_PTR")) SWIG_fail;
  result = (TPMT_SIGNATURE **)new_TPMT_SIGNATURE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SIGNATURE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SIGNATURE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SIGNATURE_PTR_PTR" "', argument " "1"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  result = (TPMT_SIGNATURE **)copy_TPMT_SIGNATURE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SIGNATURE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE **arg1 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SIGNATURE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SIGNATURE_PTR_PTR" "', argument " "1"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg1 = (TPMT_SIGNATURE **)(argp1);
  delete_TPMT_SIGNATURE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE **arg1 = (TPMT_SIGNATURE **) 0 ;
  TPMT_SIGNATURE *arg2 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SIGNATURE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg1 = (TPMT_SIGNATURE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SIGNATURE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg2 = (TPMT_SIGNATURE *)(argp2);
  TPMT_SIGNATURE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SIGNATURE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE **arg1 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SIGNATURE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SIGNATURE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SIGNATURE_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg1 = (TPMT_SIGNATURE **)(argp1);
  result = (TPMT_SIGNATURE *)TPMT_SIGNATURE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_ENCRYPTED_SECRET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_ENCRYPTED_SECRET",SWIG_From_size_t((size_t)(sizeof(TPMU_ENCRYPTED_SECRET))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_ENCRYPTED_SECRET_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_ENCRYPTED_SECRET_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_ENCRYPTED_SECRET_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_ENCRYPTED_SECRET_ARRAY *)new_TPMU_ENCRYPTED_SECRET_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ENCRYPTED_SECRET_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET_ARRAY *arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ENCRYPTED_SECRET_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ENCRYPTED_SECRET_ARRAY" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *)(argp1);
  delete_TPMU_ENCRYPTED_SECRET_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET_ARRAY *arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_ENCRYPTED_SECRET result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_ENCRYPTED_SECRET_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_ENCRYPTED_SECRET *)memcpy((TPMU_ENCRYPTED_SECRET *)malloc(sizeof(TPMU_ENCRYPTED_SECRET)),&result,sizeof(TPMU_ENCRYPTED_SECRET)), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET_ARRAY *arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_ENCRYPTED_SECRET arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_ENCRYPTED_SECRET_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_ENCRYPTED_SECRET,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_ENCRYPTED_SECRET""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_ENCRYPTED_SECRET""'");
    } else {
      arg3 = *((TPMU_ENCRYPTED_SECRET *)(argp3));
    }
  }
  TPMU_ENCRYPTED_SECRET_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET_ARRAY *arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY_cast" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET_ARRAY *)(argp1);
  result = (TPMU_ENCRYPTED_SECRET *)TPMU_ENCRYPTED_SECRET_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = (TPMU_ENCRYPTED_SECRET_ARRAY *)TPMU_ENCRYPTED_SECRET_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_ENCRYPTED_SECRET_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_ENCRYPTED_SECRET_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_ENCRYPTED_SECRET_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ENCRYPTED_SECRET_PTR")) SWIG_fail;
  result = (TPMU_ENCRYPTED_SECRET *)new_TPMU_ENCRYPTED_SECRET_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_ENCRYPTED_SECRET_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_ENCRYPTED_SECRET_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_ENCRYPTED_SECRET,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_ENCRYPTED_SECRET_PTR" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_ENCRYPTED_SECRET_PTR" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET""'");
    } else {
      arg1 = *((TPMU_ENCRYPTED_SECRET *)(argp1));
    }
  }
  result = (TPMU_ENCRYPTED_SECRET *)copy_TPMU_ENCRYPTED_SECRET_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ENCRYPTED_SECRET_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ENCRYPTED_SECRET_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ENCRYPTED_SECRET_PTR" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  delete_TPMU_ENCRYPTED_SECRET_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  TPMU_ENCRYPTED_SECRET arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_PTR_assign" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_ENCRYPTED_SECRET,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ENCRYPTED_SECRET_PTR_assign" "', argument " "2"" of type '" "TPMU_ENCRYPTED_SECRET""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_ENCRYPTED_SECRET_PTR_assign" "', argument " "2"" of type '" "TPMU_ENCRYPTED_SECRET""'");
    } else {
      arg2 = *((TPMU_ENCRYPTED_SECRET *)(argp2));
    }
  }
  TPMU_ENCRYPTED_SECRET_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_PTR_value" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = TPMU_ENCRYPTED_SECRET_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_ENCRYPTED_SECRET *)memcpy((TPMU_ENCRYPTED_SECRET *)malloc(sizeof(TPMU_ENCRYPTED_SECRET)),&result,sizeof(TPMU_ENCRYPTED_SECRET)), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_ENCRYPTED_SECRET_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_ENCRYPTED_SECRET_PTR_PTR")) SWIG_fail;
  result = (TPMU_ENCRYPTED_SECRET **)new_TPMU_ENCRYPTED_SECRET_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_ENCRYPTED_SECRET_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_ENCRYPTED_SECRET_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_ENCRYPTED_SECRET_PTR_PTR" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  result = (TPMU_ENCRYPTED_SECRET **)copy_TPMU_ENCRYPTED_SECRET_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_ENCRYPTED_SECRET_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET **arg1 = (TPMU_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_ENCRYPTED_SECRET_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_ENCRYPTED_SECRET_PTR_PTR" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET **""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET **)(argp1);
  delete_TPMU_ENCRYPTED_SECRET_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET **arg1 = (TPMU_ENCRYPTED_SECRET **) 0 ;
  TPMU_ENCRYPTED_SECRET *arg2 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_ENCRYPTED_SECRET_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET **""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_ENCRYPTED_SECRET_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg2 = (TPMU_ENCRYPTED_SECRET *)(argp2);
  TPMU_ENCRYPTED_SECRET_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_ENCRYPTED_SECRET_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET **arg1 = (TPMU_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_ENCRYPTED_SECRET_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_ENCRYPTED_SECRET_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET **""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET **)(argp1);
  result = (TPMU_ENCRYPTED_SECRET *)TPMU_ENCRYPTED_SECRET_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_ENCRYPTED_SECRET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_ENCRYPTED_SECRET",SWIG_From_size_t((size_t)(sizeof(TPM2B_ENCRYPTED_SECRET))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ENCRYPTED_SECRET_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_ENCRYPTED_SECRET_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_ENCRYPTED_SECRET_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_ENCRYPTED_SECRET_ARRAY *)new_TPM2B_ENCRYPTED_SECRET_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ENCRYPTED_SECRET_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET_ARRAY *arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ENCRYPTED_SECRET_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ENCRYPTED_SECRET_ARRAY" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *)(argp1);
  delete_TPM2B_ENCRYPTED_SECRET_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET_ARRAY *arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_ENCRYPTED_SECRET result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_ENCRYPTED_SECRET *)memcpy((TPM2B_ENCRYPTED_SECRET *)malloc(sizeof(TPM2B_ENCRYPTED_SECRET)),&result,sizeof(TPM2B_ENCRYPTED_SECRET)), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET_ARRAY *arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_ENCRYPTED_SECRET arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ENCRYPTED_SECRET""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ENCRYPTED_SECRET""'");
    } else {
      arg3 = *((TPM2B_ENCRYPTED_SECRET *)(argp3));
    }
  }
  TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET_ARRAY *arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ENCRYPTED_SECRET_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET_ARRAY *)(argp1);
  result = (TPM2B_ENCRYPTED_SECRET *)TPM2B_ENCRYPTED_SECRET_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  result = (TPM2B_ENCRYPTED_SECRET_ARRAY *)TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ENCRYPTED_SECRET_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_ENCRYPTED_SECRET_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ENCRYPTED_SECRET_PTR")) SWIG_fail;
  result = (TPM2B_ENCRYPTED_SECRET *)new_TPM2B_ENCRYPTED_SECRET_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ENCRYPTED_SECRET_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ENCRYPTED_SECRET_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ENCRYPTED_SECRET_PTR" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_ENCRYPTED_SECRET_PTR" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET""'");
    } else {
      arg1 = *((TPM2B_ENCRYPTED_SECRET *)(argp1));
    }
  }
  result = (TPM2B_ENCRYPTED_SECRET *)copy_TPM2B_ENCRYPTED_SECRET_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ENCRYPTED_SECRET_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ENCRYPTED_SECRET_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ENCRYPTED_SECRET_PTR" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  delete_TPM2B_ENCRYPTED_SECRET_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  TPM2B_ENCRYPTED_SECRET arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ENCRYPTED_SECRET_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_assign" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_assign" "', argument " "2"" of type '" "TPM2B_ENCRYPTED_SECRET""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_assign" "', argument " "2"" of type '" "TPM2B_ENCRYPTED_SECRET""'");
    } else {
      arg2 = *((TPM2B_ENCRYPTED_SECRET *)(argp2));
    }
  }
  TPM2B_ENCRYPTED_SECRET_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ENCRYPTED_SECRET_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_value" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  result = TPM2B_ENCRYPTED_SECRET_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_ENCRYPTED_SECRET *)memcpy((TPM2B_ENCRYPTED_SECRET *)malloc(sizeof(TPM2B_ENCRYPTED_SECRET)),&result,sizeof(TPM2B_ENCRYPTED_SECRET)), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ENCRYPTED_SECRET_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ENCRYPTED_SECRET_PTR_PTR")) SWIG_fail;
  result = (TPM2B_ENCRYPTED_SECRET **)new_TPM2B_ENCRYPTED_SECRET_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  result = (TPM2B_ENCRYPTED_SECRET **)copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET **arg1 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET **)(argp1);
  delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET **arg1 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg2 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg2 = (TPM2B_ENCRYPTED_SECRET *)(argp2);
  TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ENCRYPTED_SECRET_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET **arg1 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ENCRYPTED_SECRET *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ENCRYPTED_SECRET_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ENCRYPTED_SECRET_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET **)(argp1);
  result = (TPM2B_ENCRYPTED_SECRET *)TPM2B_ENCRYPTED_SECRET_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_PUBLIC_ID_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_PUBLIC_ID",SWIG_From_size_t((size_t)(sizeof(TPMU_PUBLIC_ID))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_ID_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_PUBLIC_ID_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_PUBLIC_ID_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_PUBLIC_ID_ARRAY *)new_TPMU_PUBLIC_ID_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_ID_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID_ARRAY *arg1 = (TPMU_PUBLIC_ID_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_ID_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_ID_ARRAY" "', argument " "1"" of type '" "TPMU_PUBLIC_ID_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID_ARRAY *)(argp1);
  delete_TPMU_PUBLIC_ID_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID_ARRAY *arg1 = (TPMU_PUBLIC_ID_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_PUBLIC_ID result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_PUBLIC_ID_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_PUBLIC_ID_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_PUBLIC_ID_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_PUBLIC_ID *)memcpy((TPMU_PUBLIC_ID *)malloc(sizeof(TPMU_PUBLIC_ID)),&result,sizeof(TPMU_PUBLIC_ID)), SWIGTYPE_p_TPMU_PUBLIC_ID, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID_ARRAY *arg1 = (TPMU_PUBLIC_ID_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_PUBLIC_ID arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_PUBLIC_ID_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_PUBLIC_ID_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_PUBLIC_ID_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_PUBLIC_ID,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_PUBLIC_ID_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_PUBLIC_ID""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_PUBLIC_ID_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_PUBLIC_ID""'");
    } else {
      arg3 = *((TPMU_PUBLIC_ID *)(argp3));
    }
  }
  TPMU_PUBLIC_ID_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID_ARRAY *arg1 = (TPMU_PUBLIC_ID_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_ARRAY_cast" "', argument " "1"" of type '" "TPMU_PUBLIC_ID_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID_ARRAY *)(argp1);
  result = (TPMU_PUBLIC_ID *)TPMU_PUBLIC_ID_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = (TPMU_PUBLIC_ID_ARRAY *)TPMU_PUBLIC_ID_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_PUBLIC_ID_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_PUBLIC_ID_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_ID_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_PUBLIC_ID_PTR")) SWIG_fail;
  result = (TPMU_PUBLIC_ID *)new_TPMU_PUBLIC_ID_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_PUBLIC_ID_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_PUBLIC_ID_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_PUBLIC_ID,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_PUBLIC_ID_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_ID""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_PUBLIC_ID_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_ID""'");
    } else {
      arg1 = *((TPMU_PUBLIC_ID *)(argp1));
    }
  }
  result = (TPMU_PUBLIC_ID *)copy_TPMU_PUBLIC_ID_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_ID_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_ID_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_ID_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  delete_TPMU_PUBLIC_ID_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  TPMU_PUBLIC_ID arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_PTR_assign" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_PUBLIC_ID,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_ID_PTR_assign" "', argument " "2"" of type '" "TPMU_PUBLIC_ID""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_PUBLIC_ID_PTR_assign" "', argument " "2"" of type '" "TPMU_PUBLIC_ID""'");
    } else {
      arg2 = *((TPMU_PUBLIC_ID *)(argp2));
    }
  }
  TPMU_PUBLIC_ID_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_PTR_value" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = TPMU_PUBLIC_ID_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_PUBLIC_ID *)memcpy((TPMU_PUBLIC_ID *)malloc(sizeof(TPMU_PUBLIC_ID)),&result,sizeof(TPMU_PUBLIC_ID)), SWIGTYPE_p_TPMU_PUBLIC_ID, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_ID_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_PUBLIC_ID_PTR_PTR")) SWIG_fail;
  result = (TPMU_PUBLIC_ID **)new_TPMU_PUBLIC_ID_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_PUBLIC_ID_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_PUBLIC_ID_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_PUBLIC_ID_PTR_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  result = (TPMU_PUBLIC_ID **)copy_TPMU_PUBLIC_ID_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_ID_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID **arg1 = (TPMU_PUBLIC_ID **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_ID_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_ID_PTR_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_ID **""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID **)(argp1);
  delete_TPMU_PUBLIC_ID_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID **arg1 = (TPMU_PUBLIC_ID **) 0 ;
  TPMU_PUBLIC_ID *arg2 = (TPMU_PUBLIC_ID *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_ID_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_PUBLIC_ID **""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_ID_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg2 = (TPMU_PUBLIC_ID *)(argp2);
  TPMU_PUBLIC_ID_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_ID_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID **arg1 = (TPMU_PUBLIC_ID **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_ID *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_ID_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_ID_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_PUBLIC_ID **""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID **)(argp1);
  result = (TPMU_PUBLIC_ID *)TPMU_PUBLIC_ID_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_KEYEDHASH_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_KEYEDHASH_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMS_KEYEDHASH_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_KEYEDHASH_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_KEYEDHASH_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_KEYEDHASH_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_KEYEDHASH_PARMS_ARRAY *)new_TPMS_KEYEDHASH_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_KEYEDHASH_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS_ARRAY *arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_KEYEDHASH_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_KEYEDHASH_PARMS_ARRAY" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *)(argp1);
  delete_TPMS_KEYEDHASH_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS_ARRAY *arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_KEYEDHASH_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_KEYEDHASH_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_KEYEDHASH_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_KEYEDHASH_PARMS *)memcpy((TPMS_KEYEDHASH_PARMS *)malloc(sizeof(TPMS_KEYEDHASH_PARMS)),&result,sizeof(TPMS_KEYEDHASH_PARMS)), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS_ARRAY *arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_KEYEDHASH_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_KEYEDHASH_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_KEYEDHASH_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_KEYEDHASH_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_KEYEDHASH_PARMS""'");
    } else {
      arg3 = *((TPMS_KEYEDHASH_PARMS *)(argp3));
    }
  }
  TPMS_KEYEDHASH_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS_ARRAY *arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_KEYEDHASH_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS_ARRAY *)(argp1);
  result = (TPMS_KEYEDHASH_PARMS *)TPMS_KEYEDHASH_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_KEYEDHASH_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  result = (TPMS_KEYEDHASH_PARMS_ARRAY *)TPMS_KEYEDHASH_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_KEYEDHASH_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_KEYEDHASH_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_KEYEDHASH_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_KEYEDHASH_PARMS_PTR")) SWIG_fail;
  result = (TPMS_KEYEDHASH_PARMS *)new_TPMS_KEYEDHASH_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_KEYEDHASH_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_KEYEDHASH_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_KEYEDHASH_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_KEYEDHASH_PARMS_PTR" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_KEYEDHASH_PARMS_PTR" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS""'");
    } else {
      arg1 = *((TPMS_KEYEDHASH_PARMS *)(argp1));
    }
  }
  result = (TPMS_KEYEDHASH_PARMS *)copy_TPMS_KEYEDHASH_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_KEYEDHASH_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_KEYEDHASH_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_KEYEDHASH_PARMS_PTR" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  delete_TPMS_KEYEDHASH_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  TPMS_KEYEDHASH_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_KEYEDHASH_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_KEYEDHASH_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_KEYEDHASH_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_KEYEDHASH_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_KEYEDHASH_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_KEYEDHASH_PARMS""'");
    } else {
      arg2 = *((TPMS_KEYEDHASH_PARMS *)(argp2));
    }
  }
  TPMS_KEYEDHASH_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_KEYEDHASH_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_PTR_value" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  result = TPMS_KEYEDHASH_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_KEYEDHASH_PARMS *)memcpy((TPMS_KEYEDHASH_PARMS *)malloc(sizeof(TPMS_KEYEDHASH_PARMS)),&result,sizeof(TPMS_KEYEDHASH_PARMS)), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_KEYEDHASH_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_KEYEDHASH_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMS_KEYEDHASH_PARMS **)new_TPMS_KEYEDHASH_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_KEYEDHASH_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_KEYEDHASH_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_KEYEDHASH_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  result = (TPMS_KEYEDHASH_PARMS **)copy_TPMS_KEYEDHASH_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_KEYEDHASH_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS **arg1 = (TPMS_KEYEDHASH_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_KEYEDHASH_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_KEYEDHASH_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS **""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS **)(argp1);
  delete_TPMS_KEYEDHASH_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS **arg1 = (TPMS_KEYEDHASH_PARMS **) 0 ;
  TPMS_KEYEDHASH_PARMS *arg2 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_KEYEDHASH_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS **""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_KEYEDHASH_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg2 = (TPMS_KEYEDHASH_PARMS *)(argp2);
  TPMS_KEYEDHASH_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_KEYEDHASH_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS **arg1 = (TPMS_KEYEDHASH_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_KEYEDHASH_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_KEYEDHASH_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_KEYEDHASH_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS **""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS **)(argp1);
  result = (TPMS_KEYEDHASH_PARMS *)TPMS_KEYEDHASH_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ASYM_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ASYM_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMS_ASYM_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ASYM_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ASYM_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ASYM_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ASYM_PARMS_ARRAY *)new_TPMS_ASYM_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ASYM_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS_ARRAY *arg1 = (TPMS_ASYM_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ASYM_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ASYM_PARMS_ARRAY" "', argument " "1"" of type '" "TPMS_ASYM_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS_ARRAY *)(argp1);
  delete_TPMS_ASYM_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS_ARRAY *arg1 = (TPMS_ASYM_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ASYM_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ASYM_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ASYM_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ASYM_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ASYM_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ASYM_PARMS *)memcpy((TPMS_ASYM_PARMS *)malloc(sizeof(TPMS_ASYM_PARMS)),&result,sizeof(TPMS_ASYM_PARMS)), SWIGTYPE_p_TPMS_ASYM_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS_ARRAY *arg1 = (TPMS_ASYM_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ASYM_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ASYM_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ASYM_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ASYM_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ASYM_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ASYM_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ASYM_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ASYM_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ASYM_PARMS""'");
    } else {
      arg3 = *((TPMS_ASYM_PARMS *)(argp3));
    }
  }
  TPMS_ASYM_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS_ARRAY *arg1 = (TPMS_ASYM_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ASYM_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ASYM_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS_ARRAY *)(argp1);
  result = (TPMS_ASYM_PARMS *)TPMS_ASYM_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ASYM_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  result = (TPMS_ASYM_PARMS_ARRAY *)TPMS_ASYM_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ASYM_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ASYM_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ASYM_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ASYM_PARMS_PTR")) SWIG_fail;
  result = (TPMS_ASYM_PARMS *)new_TPMS_ASYM_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ASYM_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ASYM_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ASYM_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ASYM_PARMS_PTR" "', argument " "1"" of type '" "TPMS_ASYM_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ASYM_PARMS_PTR" "', argument " "1"" of type '" "TPMS_ASYM_PARMS""'");
    } else {
      arg1 = *((TPMS_ASYM_PARMS *)(argp1));
    }
  }
  result = (TPMS_ASYM_PARMS *)copy_TPMS_ASYM_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ASYM_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ASYM_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ASYM_PARMS_PTR" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  delete_TPMS_ASYM_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  TPMS_ASYM_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ASYM_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ASYM_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ASYM_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_ASYM_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ASYM_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_ASYM_PARMS""'");
    } else {
      arg2 = *((TPMS_ASYM_PARMS *)(argp2));
    }
  }
  TPMS_ASYM_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ASYM_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_PTR_value" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  result = TPMS_ASYM_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ASYM_PARMS *)memcpy((TPMS_ASYM_PARMS *)malloc(sizeof(TPMS_ASYM_PARMS)),&result,sizeof(TPMS_ASYM_PARMS)), SWIGTYPE_p_TPMS_ASYM_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ASYM_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ASYM_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMS_ASYM_PARMS **)new_TPMS_ASYM_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ASYM_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS *arg1 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ASYM_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ASYM_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS *)(argp1);
  result = (TPMS_ASYM_PARMS **)copy_TPMS_ASYM_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ASYM_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS **arg1 = (TPMS_ASYM_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ASYM_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ASYM_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_ASYM_PARMS **""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS **)(argp1);
  delete_TPMS_ASYM_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS **arg1 = (TPMS_ASYM_PARMS **) 0 ;
  TPMS_ASYM_PARMS *arg2 = (TPMS_ASYM_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ASYM_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ASYM_PARMS **""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ASYM_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ASYM_PARMS *""'"); 
  }
  arg2 = (TPMS_ASYM_PARMS *)(argp2);
  TPMS_ASYM_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ASYM_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ASYM_PARMS **arg1 = (TPMS_ASYM_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ASYM_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ASYM_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ASYM_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ASYM_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ASYM_PARMS **""'"); 
  }
  arg1 = (TPMS_ASYM_PARMS **)(argp1);
  result = (TPMS_ASYM_PARMS *)TPMS_ASYM_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ASYM_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_RSA_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_RSA_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMS_RSA_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_RSA_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_RSA_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_RSA_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_RSA_PARMS_ARRAY *)new_TPMS_RSA_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_RSA_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS_ARRAY *arg1 = (TPMS_RSA_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_RSA_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_RSA_PARMS_ARRAY" "', argument " "1"" of type '" "TPMS_RSA_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS_ARRAY *)(argp1);
  delete_TPMS_RSA_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS_ARRAY *arg1 = (TPMS_RSA_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_RSA_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_RSA_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_RSA_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_RSA_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_RSA_PARMS *)memcpy((TPMS_RSA_PARMS *)malloc(sizeof(TPMS_RSA_PARMS)),&result,sizeof(TPMS_RSA_PARMS)), SWIGTYPE_p_TPMS_RSA_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS_ARRAY *arg1 = (TPMS_RSA_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_RSA_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_RSA_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_RSA_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_RSA_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_RSA_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_RSA_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_RSA_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_RSA_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_RSA_PARMS""'");
    } else {
      arg3 = *((TPMS_RSA_PARMS *)(argp3));
    }
  }
  TPMS_RSA_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS_ARRAY *arg1 = (TPMS_RSA_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMS_RSA_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS_ARRAY *)(argp1);
  result = (TPMS_RSA_PARMS *)TPMS_RSA_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = (TPMS_RSA_PARMS_ARRAY *)TPMS_RSA_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_RSA_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_RSA_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_RSA_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_RSA_PARMS_PTR")) SWIG_fail;
  result = (TPMS_RSA_PARMS *)new_TPMS_RSA_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_RSA_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_RSA_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_RSA_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_RSA_PARMS_PTR" "', argument " "1"" of type '" "TPMS_RSA_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_RSA_PARMS_PTR" "', argument " "1"" of type '" "TPMS_RSA_PARMS""'");
    } else {
      arg1 = *((TPMS_RSA_PARMS *)(argp1));
    }
  }
  result = (TPMS_RSA_PARMS *)copy_TPMS_RSA_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_RSA_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_RSA_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_RSA_PARMS_PTR" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  delete_TPMS_RSA_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  TPMS_RSA_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_RSA_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_RSA_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_RSA_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_RSA_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_RSA_PARMS""'");
    } else {
      arg2 = *((TPMS_RSA_PARMS *)(argp2));
    }
  }
  TPMS_RSA_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_PTR_value" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = TPMS_RSA_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_RSA_PARMS *)memcpy((TPMS_RSA_PARMS *)malloc(sizeof(TPMS_RSA_PARMS)),&result,sizeof(TPMS_RSA_PARMS)), SWIGTYPE_p_TPMS_RSA_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_RSA_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_RSA_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMS_RSA_PARMS **)new_TPMS_RSA_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_RSA_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_RSA_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_RSA_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  result = (TPMS_RSA_PARMS **)copy_TPMS_RSA_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_RSA_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS **arg1 = (TPMS_RSA_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_RSA_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_RSA_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_RSA_PARMS **""'"); 
  }
  arg1 = (TPMS_RSA_PARMS **)(argp1);
  delete_TPMS_RSA_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS **arg1 = (TPMS_RSA_PARMS **) 0 ;
  TPMS_RSA_PARMS *arg2 = (TPMS_RSA_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_RSA_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_RSA_PARMS **""'"); 
  }
  arg1 = (TPMS_RSA_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_RSA_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg2 = (TPMS_RSA_PARMS *)(argp2);
  TPMS_RSA_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_RSA_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS **arg1 = (TPMS_RSA_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_RSA_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_RSA_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_RSA_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_RSA_PARMS **""'"); 
  }
  arg1 = (TPMS_RSA_PARMS **)(argp1);
  result = (TPMS_RSA_PARMS *)TPMS_RSA_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ECC_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ECC_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMS_ECC_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ECC_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ECC_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ECC_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ECC_PARMS_ARRAY *)new_TPMS_ECC_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS_ARRAY *arg1 = (TPMS_ECC_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_PARMS_ARRAY" "', argument " "1"" of type '" "TPMS_ECC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS_ARRAY *)(argp1);
  delete_TPMS_ECC_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS_ARRAY *arg1 = (TPMS_ECC_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ECC_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ECC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ECC_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ECC_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ECC_PARMS *)memcpy((TPMS_ECC_PARMS *)malloc(sizeof(TPMS_ECC_PARMS)),&result,sizeof(TPMS_ECC_PARMS)), SWIGTYPE_p_TPMS_ECC_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS_ARRAY *arg1 = (TPMS_ECC_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ECC_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ECC_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ECC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ECC_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ECC_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ECC_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ECC_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ECC_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ECC_PARMS""'");
    } else {
      arg3 = *((TPMS_ECC_PARMS *)(argp3));
    }
  }
  TPMS_ECC_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS_ARRAY *arg1 = (TPMS_ECC_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ECC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS_ARRAY *)(argp1);
  result = (TPMS_ECC_PARMS *)TPMS_ECC_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = (TPMS_ECC_PARMS_ARRAY *)TPMS_ECC_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ECC_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ECC_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ECC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ECC_PARMS_PTR")) SWIG_fail;
  result = (TPMS_ECC_PARMS *)new_TPMS_ECC_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ECC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ECC_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ECC_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ECC_PARMS_PTR" "', argument " "1"" of type '" "TPMS_ECC_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ECC_PARMS_PTR" "', argument " "1"" of type '" "TPMS_ECC_PARMS""'");
    } else {
      arg1 = *((TPMS_ECC_PARMS *)(argp1));
    }
  }
  result = (TPMS_ECC_PARMS *)copy_TPMS_ECC_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_PARMS_PTR" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  delete_TPMS_ECC_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  TPMS_ECC_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ECC_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_ECC_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ECC_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMS_ECC_PARMS""'");
    } else {
      arg2 = *((TPMS_ECC_PARMS *)(argp2));
    }
  }
  TPMS_ECC_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_PTR_value" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = TPMS_ECC_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ECC_PARMS *)memcpy((TPMS_ECC_PARMS *)malloc(sizeof(TPMS_ECC_PARMS)),&result,sizeof(TPMS_ECC_PARMS)), SWIGTYPE_p_TPMS_ECC_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ECC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ECC_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMS_ECC_PARMS **)new_TPMS_ECC_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ECC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ECC_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ECC_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  result = (TPMS_ECC_PARMS **)copy_TPMS_ECC_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ECC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS **arg1 = (TPMS_ECC_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ECC_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ECC_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMS_ECC_PARMS **""'"); 
  }
  arg1 = (TPMS_ECC_PARMS **)(argp1);
  delete_TPMS_ECC_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS **arg1 = (TPMS_ECC_PARMS **) 0 ;
  TPMS_ECC_PARMS *arg2 = (TPMS_ECC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ECC_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ECC_PARMS **""'"); 
  }
  arg1 = (TPMS_ECC_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ECC_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg2 = (TPMS_ECC_PARMS *)(argp2);
  TPMS_ECC_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ECC_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS **arg1 = (TPMS_ECC_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ECC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ECC_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ECC_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ECC_PARMS **""'"); 
  }
  arg1 = (TPMS_ECC_PARMS **)(argp1);
  result = (TPMS_ECC_PARMS *)TPMS_ECC_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_PUBLIC_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_PUBLIC_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMU_PUBLIC_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_PUBLIC_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_PUBLIC_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_PUBLIC_PARMS_ARRAY *)new_TPMU_PUBLIC_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS_ARRAY *arg1 = (TPMU_PUBLIC_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_PARMS_ARRAY" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS_ARRAY *)(argp1);
  delete_TPMU_PUBLIC_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS_ARRAY *arg1 = (TPMU_PUBLIC_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_PUBLIC_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_PUBLIC_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_PUBLIC_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_PUBLIC_PARMS *)memcpy((TPMU_PUBLIC_PARMS *)malloc(sizeof(TPMU_PUBLIC_PARMS)),&result,sizeof(TPMU_PUBLIC_PARMS)), SWIGTYPE_p_TPMU_PUBLIC_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS_ARRAY *arg1 = (TPMU_PUBLIC_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_PUBLIC_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_PUBLIC_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_PUBLIC_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_PUBLIC_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_PUBLIC_PARMS""'");
    } else {
      arg3 = *((TPMU_PUBLIC_PARMS *)(argp3));
    }
  }
  TPMU_PUBLIC_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS_ARRAY *arg1 = (TPMU_PUBLIC_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS_ARRAY *)(argp1);
  result = (TPMU_PUBLIC_PARMS *)TPMU_PUBLIC_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMU_PUBLIC_PARMS_ARRAY *)TPMU_PUBLIC_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_PUBLIC_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_PUBLIC_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_PUBLIC_PARMS_PTR")) SWIG_fail;
  result = (TPMU_PUBLIC_PARMS *)new_TPMU_PUBLIC_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_PUBLIC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_PUBLIC_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_PUBLIC_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_PUBLIC_PARMS_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_PUBLIC_PARMS_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS""'");
    } else {
      arg1 = *((TPMU_PUBLIC_PARMS *)(argp1));
    }
  }
  result = (TPMU_PUBLIC_PARMS *)copy_TPMU_PUBLIC_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_PARMS_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  delete_TPMU_PUBLIC_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  TPMU_PUBLIC_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_PUBLIC_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMU_PUBLIC_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_PUBLIC_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMU_PUBLIC_PARMS""'");
    } else {
      arg2 = *((TPMU_PUBLIC_PARMS *)(argp2));
    }
  }
  TPMU_PUBLIC_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_PTR_value" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = TPMU_PUBLIC_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_PUBLIC_PARMS *)memcpy((TPMU_PUBLIC_PARMS *)malloc(sizeof(TPMU_PUBLIC_PARMS)),&result,sizeof(TPMU_PUBLIC_PARMS)), SWIGTYPE_p_TPMU_PUBLIC_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_PUBLIC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_PUBLIC_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMU_PUBLIC_PARMS **)new_TPMU_PUBLIC_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_PUBLIC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_PUBLIC_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_PUBLIC_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  result = (TPMU_PUBLIC_PARMS **)copy_TPMU_PUBLIC_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_PUBLIC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS **arg1 = (TPMU_PUBLIC_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_PUBLIC_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_PUBLIC_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS **""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS **)(argp1);
  delete_TPMU_PUBLIC_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS **arg1 = (TPMU_PUBLIC_PARMS **) 0 ;
  TPMU_PUBLIC_PARMS *arg2 = (TPMU_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_PUBLIC_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS **""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_PUBLIC_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg2 = (TPMU_PUBLIC_PARMS *)(argp2);
  TPMU_PUBLIC_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_PUBLIC_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS **arg1 = (TPMU_PUBLIC_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_PUBLIC_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_PUBLIC_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS **""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS **)(argp1);
  result = (TPMU_PUBLIC_PARMS *)TPMU_PUBLIC_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_PUBLIC_PARMS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_PUBLIC_PARMS",SWIG_From_size_t((size_t)(sizeof(TPMT_PUBLIC_PARMS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_PUBLIC_PARMS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_PUBLIC_PARMS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_PUBLIC_PARMS_ARRAY *)new_TPMT_PUBLIC_PARMS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_PARMS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS_ARRAY *arg1 = (TPMT_PUBLIC_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_PARMS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_PARMS_ARRAY" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS_ARRAY *)(argp1);
  delete_TPMT_PUBLIC_PARMS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS_ARRAY *arg1 = (TPMT_PUBLIC_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_PUBLIC_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PARMS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_PARMS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_PUBLIC_PARMS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_PUBLIC_PARMS *)memcpy((TPMT_PUBLIC_PARMS *)malloc(sizeof(TPMT_PUBLIC_PARMS)),&result,sizeof(TPMT_PUBLIC_PARMS)), SWIGTYPE_p_TPMT_PUBLIC_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS_ARRAY *arg1 = (TPMT_PUBLIC_PARMS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_PUBLIC_PARMS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_PUBLIC_PARMS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_PUBLIC_PARMS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_PUBLIC_PARMS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_PUBLIC_PARMS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_PUBLIC_PARMS""'");
    } else {
      arg3 = *((TPMT_PUBLIC_PARMS *)(argp3));
    }
  }
  TPMT_PUBLIC_PARMS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS_ARRAY *arg1 = (TPMT_PUBLIC_PARMS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PARMS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_ARRAY_cast" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS_ARRAY *)(argp1);
  result = (TPMT_PUBLIC_PARMS *)TPMT_PUBLIC_PARMS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PARMS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  result = (TPMT_PUBLIC_PARMS_ARRAY *)TPMT_PUBLIC_PARMS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_PUBLIC_PARMS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_PUBLIC_PARMS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_PUBLIC_PARMS_PTR")) SWIG_fail;
  result = (TPMT_PUBLIC_PARMS *)new_TPMT_PUBLIC_PARMS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_PUBLIC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_PUBLIC_PARMS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_PUBLIC_PARMS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_PUBLIC_PARMS_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_PUBLIC_PARMS_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS""'");
    } else {
      arg1 = *((TPMT_PUBLIC_PARMS *)(argp1));
    }
  }
  result = (TPMT_PUBLIC_PARMS *)copy_TPMT_PUBLIC_PARMS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_PARMS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_PARMS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_PARMS_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  delete_TPMT_PUBLIC_PARMS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  TPMT_PUBLIC_PARMS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PARMS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_PTR_assign" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_PUBLIC_PARMS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMT_PUBLIC_PARMS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_PUBLIC_PARMS_PTR_assign" "', argument " "2"" of type '" "TPMT_PUBLIC_PARMS""'");
    } else {
      arg2 = *((TPMT_PUBLIC_PARMS *)(argp2));
    }
  }
  TPMT_PUBLIC_PARMS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PARMS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_PTR_value" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  result = TPMT_PUBLIC_PARMS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_PUBLIC_PARMS *)memcpy((TPMT_PUBLIC_PARMS *)malloc(sizeof(TPMT_PUBLIC_PARMS)),&result,sizeof(TPMT_PUBLIC_PARMS)), SWIGTYPE_p_TPMT_PUBLIC_PARMS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_PUBLIC_PARMS_PTR_PTR")) SWIG_fail;
  result = (TPMT_PUBLIC_PARMS **)new_TPMT_PUBLIC_PARMS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_PUBLIC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_PUBLIC_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_PUBLIC_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  result = (TPMT_PUBLIC_PARMS **)copy_TPMT_PUBLIC_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_PARMS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS **arg1 = (TPMT_PUBLIC_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_PARMS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_PARMS_PTR_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS **""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS **)(argp1);
  delete_TPMT_PUBLIC_PARMS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS **arg1 = (TPMT_PUBLIC_PARMS **) 0 ;
  TPMT_PUBLIC_PARMS *arg2 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PARMS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS **""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_PARMS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg2 = (TPMT_PUBLIC_PARMS *)(argp2);
  TPMT_PUBLIC_PARMS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PARMS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS **arg1 = (TPMT_PUBLIC_PARMS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_PARMS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PARMS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PARMS_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS **""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS **)(argp1);
  result = (TPMT_PUBLIC_PARMS *)TPMT_PUBLIC_PARMS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_PUBLIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_PUBLIC",SWIG_From_size_t((size_t)(sizeof(TPMT_PUBLIC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_PUBLIC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_PUBLIC_ARRAY *)new_TPMT_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_ARRAY *arg1 = (TPMT_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_ARRAY" "', argument " "1"" of type '" "TPMT_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_ARRAY *)(argp1);
  delete_TPMT_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_ARRAY *arg1 = (TPMT_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_PUBLIC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_PUBLIC *)memcpy((TPMT_PUBLIC *)malloc(sizeof(TPMT_PUBLIC)),&result,sizeof(TPMT_PUBLIC)), SWIGTYPE_p_TPMT_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_ARRAY *arg1 = (TPMT_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_PUBLIC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_PUBLIC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_PUBLIC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_PUBLIC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_PUBLIC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_PUBLIC""'");
    } else {
      arg3 = *((TPMT_PUBLIC *)(argp3));
    }
  }
  TPMT_PUBLIC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_ARRAY *arg1 = (TPMT_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_ARRAY_cast" "', argument " "1"" of type '" "TPMT_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMT_PUBLIC_ARRAY *)(argp1);
  result = (TPMT_PUBLIC *)TPMT_PUBLIC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMT_PUBLIC_ARRAY *)TPMT_PUBLIC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_PUBLIC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_PUBLIC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_PUBLIC_PTR")) SWIG_fail;
  result = (TPMT_PUBLIC *)new_TPMT_PUBLIC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_PUBLIC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_PUBLIC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_PUBLIC_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_PUBLIC_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC""'");
    } else {
      arg1 = *((TPMT_PUBLIC *)(argp1));
    }
  }
  result = (TPMT_PUBLIC *)copy_TPMT_PUBLIC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  delete_TPMT_PUBLIC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  TPMT_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PTR_assign" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_PUBLIC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPMT_PUBLIC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPMT_PUBLIC""'");
    } else {
      arg2 = *((TPMT_PUBLIC *)(argp2));
    }
  }
  TPMT_PUBLIC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PTR_value" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = TPMT_PUBLIC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_PUBLIC *)memcpy((TPMT_PUBLIC *)malloc(sizeof(TPMT_PUBLIC)),&result,sizeof(TPMT_PUBLIC)), SWIGTYPE_p_TPMT_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_PUBLIC_PTR_PTR")) SWIG_fail;
  result = (TPMT_PUBLIC **)new_TPMT_PUBLIC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  result = (TPMT_PUBLIC **)copy_TPMT_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC **arg1 = (TPMT_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPMT_PUBLIC **""'"); 
  }
  arg1 = (TPMT_PUBLIC **)(argp1);
  delete_TPMT_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC **arg1 = (TPMT_PUBLIC **) 0 ;
  TPMT_PUBLIC *arg2 = (TPMT_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_PUBLIC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_PUBLIC **""'"); 
  }
  arg1 = (TPMT_PUBLIC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_PUBLIC_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg2 = (TPMT_PUBLIC *)(argp2);
  TPMT_PUBLIC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_PUBLIC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC **arg1 = (TPMT_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_PUBLIC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_PUBLIC_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_PUBLIC **""'"); 
  }
  arg1 = (TPMT_PUBLIC **)(argp1);
  result = (TPMT_PUBLIC *)TPMT_PUBLIC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_PUBLIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_PUBLIC",SWIG_From_size_t((size_t)(sizeof(TPM2B_PUBLIC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_PUBLIC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_PUBLIC_ARRAY *)new_TPM2B_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_ARRAY *arg1 = (TPM2B_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_ARRAY" "', argument " "1"" of type '" "TPM2B_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_ARRAY *)(argp1);
  delete_TPM2B_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_ARRAY *arg1 = (TPM2B_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PUBLIC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_PUBLIC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_PUBLIC *)memcpy((TPM2B_PUBLIC *)malloc(sizeof(TPM2B_PUBLIC)),&result,sizeof(TPM2B_PUBLIC)), SWIGTYPE_p_TPM2B_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_ARRAY *arg1 = (TPM2B_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_PUBLIC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_PUBLIC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PUBLIC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_PUBLIC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PUBLIC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PUBLIC""'");
    } else {
      arg3 = *((TPM2B_PUBLIC *)(argp3));
    }
  }
  TPM2B_PUBLIC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_ARRAY *arg1 = (TPM2B_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_ARRAY *)(argp1);
  result = (TPM2B_PUBLIC *)TPM2B_PUBLIC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  result = (TPM2B_PUBLIC_ARRAY *)TPM2B_PUBLIC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PUBLIC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PUBLIC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PUBLIC_PTR")) SWIG_fail;
  result = (TPM2B_PUBLIC *)new_TPM2B_PUBLIC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PUBLIC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_PUBLIC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PUBLIC_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_PUBLIC_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC""'");
    } else {
      arg1 = *((TPM2B_PUBLIC *)(argp1));
    }
  }
  result = (TPM2B_PUBLIC *)copy_TPM2B_PUBLIC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  delete_TPM2B_PUBLIC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_PTR_assign" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_PUBLIC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPM2B_PUBLIC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPM2B_PUBLIC""'");
    } else {
      arg2 = *((TPM2B_PUBLIC *)(argp2));
    }
  }
  TPM2B_PUBLIC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_PTR_value" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  result = TPM2B_PUBLIC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_PUBLIC *)memcpy((TPM2B_PUBLIC *)malloc(sizeof(TPM2B_PUBLIC)),&result,sizeof(TPM2B_PUBLIC)), SWIGTYPE_p_TPM2B_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PUBLIC_PTR_PTR")) SWIG_fail;
  result = (TPM2B_PUBLIC **)new_TPM2B_PUBLIC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  result = (TPM2B_PUBLIC **)copy_TPM2B_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC **arg1 = (TPM2B_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg1 = (TPM2B_PUBLIC **)(argp1);
  delete_TPM2B_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC **arg1 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_PUBLIC *arg2 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PUBLIC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg1 = (TPM2B_PUBLIC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PUBLIC_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg2 = (TPM2B_PUBLIC *)(argp2);
  TPM2B_PUBLIC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PUBLIC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC **arg1 = (TPM2B_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PUBLIC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PUBLIC_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg1 = (TPM2B_PUBLIC **)(argp1);
  result = (TPM2B_PUBLIC *)TPM2B_PUBLIC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_TEMPLATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_TEMPLATE",SWIG_From_size_t((size_t)(sizeof(TPM2B_TEMPLATE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_TEMPLATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_TEMPLATE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_TEMPLATE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_TEMPLATE_ARRAY *)new_TPM2B_TEMPLATE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_TEMPLATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE_ARRAY *arg1 = (TPM2B_TEMPLATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_TEMPLATE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_TEMPLATE_ARRAY" "', argument " "1"" of type '" "TPM2B_TEMPLATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE_ARRAY *)(argp1);
  delete_TPM2B_TEMPLATE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE_ARRAY *arg1 = (TPM2B_TEMPLATE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_TEMPLATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_TEMPLATE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_TEMPLATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_TEMPLATE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_TEMPLATE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_TEMPLATE *)memcpy((TPM2B_TEMPLATE *)malloc(sizeof(TPM2B_TEMPLATE)),&result,sizeof(TPM2B_TEMPLATE)), SWIGTYPE_p_TPM2B_TEMPLATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE_ARRAY *arg1 = (TPM2B_TEMPLATE_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_TEMPLATE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_TEMPLATE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_TEMPLATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_TEMPLATE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_TEMPLATE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_TEMPLATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_TEMPLATE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_TEMPLATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_TEMPLATE""'");
    } else {
      arg3 = *((TPM2B_TEMPLATE *)(argp3));
    }
  }
  TPM2B_TEMPLATE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE_ARRAY *arg1 = (TPM2B_TEMPLATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_TEMPLATE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_TEMPLATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE_ARRAY *)(argp1);
  result = (TPM2B_TEMPLATE *)TPM2B_TEMPLATE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_TEMPLATE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  result = (TPM2B_TEMPLATE_ARRAY *)TPM2B_TEMPLATE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_TEMPLATE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_TEMPLATE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_TEMPLATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_TEMPLATE_PTR")) SWIG_fail;
  result = (TPM2B_TEMPLATE *)new_TPM2B_TEMPLATE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_TEMPLATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_TEMPLATE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_TEMPLATE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_TEMPLATE_PTR" "', argument " "1"" of type '" "TPM2B_TEMPLATE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_TEMPLATE_PTR" "', argument " "1"" of type '" "TPM2B_TEMPLATE""'");
    } else {
      arg1 = *((TPM2B_TEMPLATE *)(argp1));
    }
  }
  result = (TPM2B_TEMPLATE *)copy_TPM2B_TEMPLATE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_TEMPLATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_TEMPLATE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_TEMPLATE_PTR" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  delete_TPM2B_TEMPLATE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  TPM2B_TEMPLATE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_TEMPLATE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_PTR_assign" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_TEMPLATE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_TEMPLATE_PTR_assign" "', argument " "2"" of type '" "TPM2B_TEMPLATE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_TEMPLATE_PTR_assign" "', argument " "2"" of type '" "TPM2B_TEMPLATE""'");
    } else {
      arg2 = *((TPM2B_TEMPLATE *)(argp2));
    }
  }
  TPM2B_TEMPLATE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_TEMPLATE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_PTR_value" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  result = TPM2B_TEMPLATE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_TEMPLATE *)memcpy((TPM2B_TEMPLATE *)malloc(sizeof(TPM2B_TEMPLATE)),&result,sizeof(TPM2B_TEMPLATE)), SWIGTYPE_p_TPM2B_TEMPLATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_TEMPLATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_TEMPLATE_PTR_PTR")) SWIG_fail;
  result = (TPM2B_TEMPLATE **)new_TPM2B_TEMPLATE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_TEMPLATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_TEMPLATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_TEMPLATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_TEMPLATE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  result = (TPM2B_TEMPLATE **)copy_TPM2B_TEMPLATE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_TEMPLATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_TEMPLATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE **arg1 = (TPM2B_TEMPLATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_TEMPLATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_TEMPLATE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_TEMPLATE **""'"); 
  }
  arg1 = (TPM2B_TEMPLATE **)(argp1);
  delete_TPM2B_TEMPLATE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE **arg1 = (TPM2B_TEMPLATE **) 0 ;
  TPM2B_TEMPLATE *arg2 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_TEMPLATE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_TEMPLATE **""'"); 
  }
  arg1 = (TPM2B_TEMPLATE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_TEMPLATE_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg2 = (TPM2B_TEMPLATE *)(argp2);
  TPM2B_TEMPLATE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_TEMPLATE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE **arg1 = (TPM2B_TEMPLATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_TEMPLATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_TEMPLATE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_TEMPLATE_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_TEMPLATE **""'"); 
  }
  arg1 = (TPM2B_TEMPLATE **)(argp1);
  result = (TPM2B_TEMPLATE *)TPM2B_TEMPLATE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_PRIVATE_VENDOR_SPECIFIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_PRIVATE_VENDOR_SPECIFIC",SWIG_From_size_t((size_t)(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)(argp1);
  delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_PRIVATE_VENDOR_SPECIFIC *)memcpy((TPM2B_PRIVATE_VENDOR_SPECIFIC *)malloc(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC)),&result,sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC)), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC""'");
    } else {
      arg3 = *((TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp3));
    }
  }
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)(argp1);
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *)TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR")) SWIG_fail;
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC""'");
    } else {
      arg1 = *((TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1));
    }
  }
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC""'");
    } else {
      arg2 = *((TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp2));
    }
  }
  TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  result = TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_PRIVATE_VENDOR_SPECIFIC *)memcpy((TPM2B_PRIVATE_VENDOR_SPECIFIC *)malloc(sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC)),&result,sizeof(TPM2B_PRIVATE_VENDOR_SPECIFIC)), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR")) SWIG_fail;
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC **)new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp1);
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC **)copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC **arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC **""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC **)(argp1);
  delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC **arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC **) 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *arg2 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC **""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC *""'"); 
  }
  arg2 = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)(argp2);
  TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_VENDOR_SPECIFIC **arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_VENDOR_SPECIFIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_PRIVATE_VENDOR_SPECIFIC **""'"); 
  }
  arg1 = (TPM2B_PRIVATE_VENDOR_SPECIFIC **)(argp1);
  result = (TPM2B_PRIVATE_VENDOR_SPECIFIC *)TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMU_SENSITIVE_COMPOSITE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMU_SENSITIVE_COMPOSITE",SWIG_From_size_t((size_t)(sizeof(TPMU_SENSITIVE_COMPOSITE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMU_SENSITIVE_COMPOSITE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMU_SENSITIVE_COMPOSITE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMU_SENSITIVE_COMPOSITE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMU_SENSITIVE_COMPOSITE_ARRAY *)new_TPMU_SENSITIVE_COMPOSITE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SENSITIVE_COMPOSITE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE_ARRAY *arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SENSITIVE_COMPOSITE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SENSITIVE_COMPOSITE_ARRAY" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *)(argp1);
  delete_TPMU_SENSITIVE_COMPOSITE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE_ARRAY *arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMU_SENSITIVE_COMPOSITE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMU_SENSITIVE_COMPOSITE *)memcpy((TPMU_SENSITIVE_COMPOSITE *)malloc(sizeof(TPMU_SENSITIVE_COMPOSITE)),&result,sizeof(TPMU_SENSITIVE_COMPOSITE)), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE_ARRAY *arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMU_SENSITIVE_COMPOSITE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SENSITIVE_COMPOSITE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMU_SENSITIVE_COMPOSITE""'");
    } else {
      arg3 = *((TPMU_SENSITIVE_COMPOSITE *)(argp3));
    }
  }
  TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE_ARRAY *arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY_cast" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE_ARRAY *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE_ARRAY *)(argp1);
  result = (TPMU_SENSITIVE_COMPOSITE *)TPMU_SENSITIVE_COMPOSITE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPMU_SENSITIVE_COMPOSITE_ARRAY *)TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMU_SENSITIVE_COMPOSITE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMU_SENSITIVE_COMPOSITE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SENSITIVE_COMPOSITE_PTR")) SWIG_fail;
  result = (TPMU_SENSITIVE_COMPOSITE *)new_TPMU_SENSITIVE_COMPOSITE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SENSITIVE_COMPOSITE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SENSITIVE_COMPOSITE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SENSITIVE_COMPOSITE_PTR" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMU_SENSITIVE_COMPOSITE_PTR" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE""'");
    } else {
      arg1 = *((TPMU_SENSITIVE_COMPOSITE *)(argp1));
    }
  }
  result = (TPMU_SENSITIVE_COMPOSITE *)copy_TPMU_SENSITIVE_COMPOSITE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SENSITIVE_COMPOSITE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SENSITIVE_COMPOSITE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SENSITIVE_COMPOSITE_PTR" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  delete_TPMU_SENSITIVE_COMPOSITE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  TPMU_SENSITIVE_COMPOSITE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_assign" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_assign" "', argument " "2"" of type '" "TPMU_SENSITIVE_COMPOSITE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_assign" "', argument " "2"" of type '" "TPMU_SENSITIVE_COMPOSITE""'");
    } else {
      arg2 = *((TPMU_SENSITIVE_COMPOSITE *)(argp2));
    }
  }
  TPMU_SENSITIVE_COMPOSITE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_value" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = TPMU_SENSITIVE_COMPOSITE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMU_SENSITIVE_COMPOSITE *)memcpy((TPMU_SENSITIVE_COMPOSITE *)malloc(sizeof(TPMU_SENSITIVE_COMPOSITE)),&result,sizeof(TPMU_SENSITIVE_COMPOSITE)), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMU_SENSITIVE_COMPOSITE_PTR_PTR")) SWIG_fail;
  result = (TPMU_SENSITIVE_COMPOSITE **)new_TPMU_SENSITIVE_COMPOSITE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  result = (TPMU_SENSITIVE_COMPOSITE **)copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE **arg1 = (TPMU_SENSITIVE_COMPOSITE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE **""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE **)(argp1);
  delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE **arg1 = (TPMU_SENSITIVE_COMPOSITE **) 0 ;
  TPMU_SENSITIVE_COMPOSITE *arg2 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE **""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg2 = (TPMU_SENSITIVE_COMPOSITE *)(argp2);
  TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE **arg1 = (TPMU_SENSITIVE_COMPOSITE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMU_SENSITIVE_COMPOSITE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE **""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE **)(argp1);
  result = (TPMU_SENSITIVE_COMPOSITE *)TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMT_SENSITIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMT_SENSITIVE",SWIG_From_size_t((size_t)(sizeof(TPMT_SENSITIVE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMT_SENSITIVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMT_SENSITIVE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMT_SENSITIVE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMT_SENSITIVE_ARRAY *)new_TPMT_SENSITIVE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SENSITIVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE_ARRAY *arg1 = (TPMT_SENSITIVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SENSITIVE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SENSITIVE_ARRAY" "', argument " "1"" of type '" "TPMT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SENSITIVE_ARRAY *)(argp1);
  delete_TPMT_SENSITIVE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE_ARRAY *arg1 = (TPMT_SENSITIVE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMT_SENSITIVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_ARRAY___getitem__" "', argument " "1"" of type '" "TPMT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SENSITIVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SENSITIVE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMT_SENSITIVE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMT_SENSITIVE *)memcpy((TPMT_SENSITIVE *)malloc(sizeof(TPMT_SENSITIVE)),&result,sizeof(TPMT_SENSITIVE)), SWIGTYPE_p_TPMT_SENSITIVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE_ARRAY *arg1 = (TPMT_SENSITIVE_ARRAY *) 0 ;
  size_t arg2 ;
  TPMT_SENSITIVE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMT_SENSITIVE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_ARRAY___setitem__" "', argument " "1"" of type '" "TPMT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SENSITIVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMT_SENSITIVE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMT_SENSITIVE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMT_SENSITIVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SENSITIVE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SENSITIVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPMT_SENSITIVE""'");
    } else {
      arg3 = *((TPMT_SENSITIVE *)(argp3));
    }
  }
  TPMT_SENSITIVE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE_ARRAY *arg1 = (TPMT_SENSITIVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_ARRAY_cast" "', argument " "1"" of type '" "TPMT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPMT_SENSITIVE_ARRAY *)(argp1);
  result = (TPMT_SENSITIVE *)TPMT_SENSITIVE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_ARRAY_frompointer" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = (TPMT_SENSITIVE_ARRAY *)TPMT_SENSITIVE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMT_SENSITIVE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMT_SENSITIVE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMT_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SENSITIVE_PTR")) SWIG_fail;
  result = (TPMT_SENSITIVE *)new_TPMT_SENSITIVE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SENSITIVE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMT_SENSITIVE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SENSITIVE_PTR" "', argument " "1"" of type '" "TPMT_SENSITIVE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMT_SENSITIVE_PTR" "', argument " "1"" of type '" "TPMT_SENSITIVE""'");
    } else {
      arg1 = *((TPMT_SENSITIVE *)(argp1));
    }
  }
  result = (TPMT_SENSITIVE *)copy_TPMT_SENSITIVE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SENSITIVE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SENSITIVE_PTR" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  delete_TPMT_SENSITIVE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  TPMT_SENSITIVE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_PTR_assign" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMT_SENSITIVE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SENSITIVE_PTR_assign" "', argument " "2"" of type '" "TPMT_SENSITIVE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMT_SENSITIVE_PTR_assign" "', argument " "2"" of type '" "TPMT_SENSITIVE""'");
    } else {
      arg2 = *((TPMT_SENSITIVE *)(argp2));
    }
  }
  TPMT_SENSITIVE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_PTR_value" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = TPMT_SENSITIVE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMT_SENSITIVE *)memcpy((TPMT_SENSITIVE *)malloc(sizeof(TPMT_SENSITIVE)),&result,sizeof(TPMT_SENSITIVE)), SWIGTYPE_p_TPMT_SENSITIVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMT_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMT_SENSITIVE_PTR_PTR")) SWIG_fail;
  result = (TPMT_SENSITIVE **)new_TPMT_SENSITIVE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMT_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMT_SENSITIVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMT_SENSITIVE_PTR_PTR" "', argument " "1"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  result = (TPMT_SENSITIVE **)copy_TPMT_SENSITIVE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMT_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE **arg1 = (TPMT_SENSITIVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMT_SENSITIVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMT_SENSITIVE_PTR_PTR" "', argument " "1"" of type '" "TPMT_SENSITIVE **""'"); 
  }
  arg1 = (TPMT_SENSITIVE **)(argp1);
  delete_TPMT_SENSITIVE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE **arg1 = (TPMT_SENSITIVE **) 0 ;
  TPMT_SENSITIVE *arg2 = (TPMT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMT_SENSITIVE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_PTR_PTR_assign" "', argument " "1"" of type '" "TPMT_SENSITIVE **""'"); 
  }
  arg1 = (TPMT_SENSITIVE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMT_SENSITIVE_PTR_PTR_assign" "', argument " "2"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg2 = (TPMT_SENSITIVE *)(argp2);
  TPMT_SENSITIVE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMT_SENSITIVE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE **arg1 = (TPMT_SENSITIVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMT_SENSITIVE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMT_SENSITIVE_PTR_PTR_value" "', argument " "1"" of type '" "TPMT_SENSITIVE **""'"); 
  }
  arg1 = (TPMT_SENSITIVE **)(argp1);
  result = (TPMT_SENSITIVE *)TPMT_SENSITIVE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_SENSITIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_SENSITIVE",SWIG_From_size_t((size_t)(sizeof(TPM2B_SENSITIVE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_SENSITIVE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_SENSITIVE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_SENSITIVE_ARRAY *)new_TPM2B_SENSITIVE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_ARRAY *arg1 = (TPM2B_SENSITIVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_ARRAY" "', argument " "1"" of type '" "TPM2B_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_ARRAY *)(argp1);
  delete_TPM2B_SENSITIVE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_ARRAY *arg1 = (TPM2B_SENSITIVE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_SENSITIVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_SENSITIVE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_SENSITIVE *)memcpy((TPM2B_SENSITIVE *)malloc(sizeof(TPM2B_SENSITIVE)),&result,sizeof(TPM2B_SENSITIVE)), SWIGTYPE_p_TPM2B_SENSITIVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_ARRAY *arg1 = (TPM2B_SENSITIVE_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_SENSITIVE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_SENSITIVE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_SENSITIVE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_SENSITIVE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_SENSITIVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SENSITIVE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SENSITIVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_SENSITIVE""'");
    } else {
      arg3 = *((TPM2B_SENSITIVE *)(argp3));
    }
  }
  TPM2B_SENSITIVE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_ARRAY *arg1 = (TPM2B_SENSITIVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_ARRAY *)(argp1);
  result = (TPM2B_SENSITIVE *)TPM2B_SENSITIVE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  result = (TPM2B_SENSITIVE_ARRAY *)TPM2B_SENSITIVE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_SENSITIVE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_SENSITIVE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_PTR")) SWIG_fail;
  result = (TPM2B_SENSITIVE *)new_TPM2B_SENSITIVE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SENSITIVE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_SENSITIVE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SENSITIVE_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_SENSITIVE_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE""'");
    } else {
      arg1 = *((TPM2B_SENSITIVE *)(argp1));
    }
  }
  result = (TPM2B_SENSITIVE *)copy_TPM2B_SENSITIVE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  delete_TPM2B_SENSITIVE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  TPM2B_SENSITIVE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_PTR_assign" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_SENSITIVE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_SENSITIVE_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE""'");
    } else {
      arg2 = *((TPM2B_SENSITIVE *)(argp2));
    }
  }
  TPM2B_SENSITIVE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_PTR_value" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  result = TPM2B_SENSITIVE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_SENSITIVE *)memcpy((TPM2B_SENSITIVE *)malloc(sizeof(TPM2B_SENSITIVE)),&result,sizeof(TPM2B_SENSITIVE)), SWIGTYPE_p_TPM2B_SENSITIVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_SENSITIVE_PTR_PTR")) SWIG_fail;
  result = (TPM2B_SENSITIVE **)new_TPM2B_SENSITIVE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_SENSITIVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_SENSITIVE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  result = (TPM2B_SENSITIVE **)copy_TPM2B_SENSITIVE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE **arg1 = (TPM2B_SENSITIVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_SENSITIVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_SENSITIVE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_SENSITIVE **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE **)(argp1);
  delete_TPM2B_SENSITIVE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE **arg1 = (TPM2B_SENSITIVE **) 0 ;
  TPM2B_SENSITIVE *arg2 = (TPM2B_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_SENSITIVE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_SENSITIVE **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_SENSITIVE_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg2 = (TPM2B_SENSITIVE *)(argp2);
  TPM2B_SENSITIVE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_SENSITIVE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE **arg1 = (TPM2B_SENSITIVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_SENSITIVE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_SENSITIVE_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_SENSITIVE **""'"); 
  }
  arg1 = (TPM2B_SENSITIVE **)(argp1);
  result = (TPM2B_SENSITIVE *)TPM2B_SENSITIVE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof__PRIVATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof__PRIVATE",SWIG_From_size_t((size_t)(sizeof(_PRIVATE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new__PRIVATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new__PRIVATE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new__PRIVATE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (_PRIVATE_ARRAY *)new__PRIVATE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete__PRIVATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE_ARRAY *arg1 = (_PRIVATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete__PRIVATE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__PRIVATE_ARRAY" "', argument " "1"" of type '" "_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (_PRIVATE_ARRAY *)(argp1);
  delete__PRIVATE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE_ARRAY *arg1 = (_PRIVATE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  _PRIVATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:_PRIVATE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_ARRAY___getitem__" "', argument " "1"" of type '" "_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (_PRIVATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_PRIVATE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = _PRIVATE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((_PRIVATE *)memcpy((_PRIVATE *)malloc(sizeof(_PRIVATE)),&result,sizeof(_PRIVATE)), SWIGTYPE_p__PRIVATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE_ARRAY *arg1 = (_PRIVATE_ARRAY *) 0 ;
  size_t arg2 ;
  _PRIVATE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:_PRIVATE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_ARRAY___setitem__" "', argument " "1"" of type '" "_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (_PRIVATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_PRIVATE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p__PRIVATE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "_PRIVATE_ARRAY___setitem__" "', argument " "3"" of type '" "_PRIVATE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_PRIVATE_ARRAY___setitem__" "', argument " "3"" of type '" "_PRIVATE""'");
    } else {
      arg3 = *((_PRIVATE *)(argp3));
    }
  }
  _PRIVATE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE_ARRAY *arg1 = (_PRIVATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_ARRAY_cast" "', argument " "1"" of type '" "_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (_PRIVATE_ARRAY *)(argp1);
  result = (_PRIVATE *)_PRIVATE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_ARRAY_frompointer" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  result = (_PRIVATE_ARRAY *)_PRIVATE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_PRIVATE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p__PRIVATE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new__PRIVATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new__PRIVATE_PTR")) SWIG_fail;
  result = (_PRIVATE *)new__PRIVATE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy__PRIVATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy__PRIVATE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p__PRIVATE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy__PRIVATE_PTR" "', argument " "1"" of type '" "_PRIVATE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy__PRIVATE_PTR" "', argument " "1"" of type '" "_PRIVATE""'");
    } else {
      arg1 = *((_PRIVATE *)(argp1));
    }
  }
  result = (_PRIVATE *)copy__PRIVATE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete__PRIVATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete__PRIVATE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__PRIVATE_PTR" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  delete__PRIVATE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  _PRIVATE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:_PRIVATE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_PTR_assign" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p__PRIVATE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_PRIVATE_PTR_assign" "', argument " "2"" of type '" "_PRIVATE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_PRIVATE_PTR_assign" "', argument " "2"" of type '" "_PRIVATE""'");
    } else {
      arg2 = *((_PRIVATE *)(argp2));
    }
  }
  _PRIVATE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_PTR_value" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  result = _PRIVATE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((_PRIVATE *)memcpy((_PRIVATE *)malloc(sizeof(_PRIVATE)),&result,sizeof(_PRIVATE)), SWIGTYPE_p__PRIVATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new__PRIVATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new__PRIVATE_PTR_PTR")) SWIG_fail;
  result = (_PRIVATE **)new__PRIVATE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p__PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy__PRIVATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE *arg1 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy__PRIVATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy__PRIVATE_PTR_PTR" "', argument " "1"" of type '" "_PRIVATE *""'"); 
  }
  arg1 = (_PRIVATE *)(argp1);
  result = (_PRIVATE **)copy__PRIVATE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p__PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete__PRIVATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE **arg1 = (_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete__PRIVATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__PRIVATE_PTR_PTR" "', argument " "1"" of type '" "_PRIVATE **""'"); 
  }
  arg1 = (_PRIVATE **)(argp1);
  delete__PRIVATE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE **arg1 = (_PRIVATE **) 0 ;
  _PRIVATE *arg2 = (_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:_PRIVATE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_PTR_PTR_assign" "', argument " "1"" of type '" "_PRIVATE **""'"); 
  }
  arg1 = (_PRIVATE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_PRIVATE_PTR_PTR_assign" "', argument " "2"" of type '" "_PRIVATE *""'"); 
  }
  arg2 = (_PRIVATE *)(argp2);
  _PRIVATE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__PRIVATE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  _PRIVATE **arg1 = (_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  _PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:_PRIVATE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p__PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_PRIVATE_PTR_PTR_value" "', argument " "1"" of type '" "_PRIVATE **""'"); 
  }
  arg1 = (_PRIVATE **)(argp1);
  result = (_PRIVATE *)_PRIVATE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_PRIVATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_PRIVATE",SWIG_From_size_t((size_t)(sizeof(TPM2B_PRIVATE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_PRIVATE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_PRIVATE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_PRIVATE_ARRAY *)new_TPM2B_PRIVATE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_ARRAY *arg1 = (TPM2B_PRIVATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_ARRAY" "', argument " "1"" of type '" "TPM2B_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_ARRAY *)(argp1);
  delete_TPM2B_PRIVATE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_ARRAY *arg1 = (TPM2B_PRIVATE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_PRIVATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_PRIVATE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_PRIVATE *)memcpy((TPM2B_PRIVATE *)malloc(sizeof(TPM2B_PRIVATE)),&result,sizeof(TPM2B_PRIVATE)), SWIGTYPE_p_TPM2B_PRIVATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_ARRAY *arg1 = (TPM2B_PRIVATE_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_PRIVATE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_PRIVATE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_PRIVATE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_PRIVATE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_PRIVATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PRIVATE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PRIVATE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_PRIVATE""'");
    } else {
      arg3 = *((TPM2B_PRIVATE *)(argp3));
    }
  }
  TPM2B_PRIVATE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_ARRAY *arg1 = (TPM2B_PRIVATE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_PRIVATE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_ARRAY *)(argp1);
  result = (TPM2B_PRIVATE *)TPM2B_PRIVATE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  result = (TPM2B_PRIVATE_ARRAY *)TPM2B_PRIVATE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_PRIVATE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_PRIVATE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_PTR")) SWIG_fail;
  result = (TPM2B_PRIVATE *)new_TPM2B_PRIVATE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PRIVATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PRIVATE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_PRIVATE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PRIVATE_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_PRIVATE_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE""'");
    } else {
      arg1 = *((TPM2B_PRIVATE *)(argp1));
    }
  }
  result = (TPM2B_PRIVATE *)copy_TPM2B_PRIVATE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  delete_TPM2B_PRIVATE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_PRIVATE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_PTR_assign" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_PRIVATE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_PRIVATE_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE""'");
    } else {
      arg2 = *((TPM2B_PRIVATE *)(argp2));
    }
  }
  TPM2B_PRIVATE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_PTR_value" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  result = TPM2B_PRIVATE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_PRIVATE *)memcpy((TPM2B_PRIVATE *)malloc(sizeof(TPM2B_PRIVATE)),&result,sizeof(TPM2B_PRIVATE)), SWIGTYPE_p_TPM2B_PRIVATE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_PRIVATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_PRIVATE_PTR_PTR")) SWIG_fail;
  result = (TPM2B_PRIVATE **)new_TPM2B_PRIVATE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_PRIVATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_PRIVATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_PRIVATE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  result = (TPM2B_PRIVATE **)copy_TPM2B_PRIVATE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_PRIVATE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE **arg1 = (TPM2B_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_PRIVATE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_PRIVATE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg1 = (TPM2B_PRIVATE **)(argp1);
  delete_TPM2B_PRIVATE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE **arg1 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_PRIVATE *arg2 = (TPM2B_PRIVATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_PRIVATE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg1 = (TPM2B_PRIVATE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_PRIVATE_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg2 = (TPM2B_PRIVATE *)(argp2);
  TPM2B_PRIVATE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_PRIVATE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE **arg1 = (TPM2B_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_PRIVATE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_PRIVATE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_PRIVATE_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg1 = (TPM2B_PRIVATE **)(argp1);
  result = (TPM2B_PRIVATE *)TPM2B_PRIVATE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_ID_OBJECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_ID_OBJECT",SWIG_From_size_t((size_t)(sizeof(TPMS_ID_OBJECT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_ID_OBJECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_ID_OBJECT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_ID_OBJECT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_ID_OBJECT_ARRAY *)new_TPMS_ID_OBJECT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ID_OBJECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT_ARRAY *arg1 = (TPMS_ID_OBJECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ID_OBJECT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ID_OBJECT_ARRAY" "', argument " "1"" of type '" "TPMS_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT_ARRAY *)(argp1);
  delete_TPMS_ID_OBJECT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT_ARRAY *arg1 = (TPMS_ID_OBJECT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_ID_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ID_OBJECT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ID_OBJECT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_ID_OBJECT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_ID_OBJECT *)memcpy((TPMS_ID_OBJECT *)malloc(sizeof(TPMS_ID_OBJECT)),&result,sizeof(TPMS_ID_OBJECT)), SWIGTYPE_p_TPMS_ID_OBJECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT_ARRAY *arg1 = (TPMS_ID_OBJECT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_ID_OBJECT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_ID_OBJECT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_ID_OBJECT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_ID_OBJECT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_ID_OBJECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ID_OBJECT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ID_OBJECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_ID_OBJECT""'");
    } else {
      arg3 = *((TPMS_ID_OBJECT *)(argp3));
    }
  }
  TPMS_ID_OBJECT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT_ARRAY *arg1 = (TPMS_ID_OBJECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ID_OBJECT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_ARRAY_cast" "', argument " "1"" of type '" "TPMS_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT_ARRAY *)(argp1);
  result = (TPMS_ID_OBJECT *)TPMS_ID_OBJECT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ID_OBJECT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  result = (TPMS_ID_OBJECT_ARRAY *)TPMS_ID_OBJECT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_ID_OBJECT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_ID_OBJECT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_ID_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ID_OBJECT_PTR")) SWIG_fail;
  result = (TPMS_ID_OBJECT *)new_TPMS_ID_OBJECT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ID_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ID_OBJECT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_ID_OBJECT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ID_OBJECT_PTR" "', argument " "1"" of type '" "TPMS_ID_OBJECT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_ID_OBJECT_PTR" "', argument " "1"" of type '" "TPMS_ID_OBJECT""'");
    } else {
      arg1 = *((TPMS_ID_OBJECT *)(argp1));
    }
  }
  result = (TPMS_ID_OBJECT *)copy_TPMS_ID_OBJECT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ID_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ID_OBJECT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ID_OBJECT_PTR" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  delete_TPMS_ID_OBJECT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  TPMS_ID_OBJECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ID_OBJECT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_PTR_assign" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_ID_OBJECT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ID_OBJECT_PTR_assign" "', argument " "2"" of type '" "TPMS_ID_OBJECT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_ID_OBJECT_PTR_assign" "', argument " "2"" of type '" "TPMS_ID_OBJECT""'");
    } else {
      arg2 = *((TPMS_ID_OBJECT *)(argp2));
    }
  }
  TPMS_ID_OBJECT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ID_OBJECT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_PTR_value" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  result = TPMS_ID_OBJECT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_ID_OBJECT *)memcpy((TPMS_ID_OBJECT *)malloc(sizeof(TPMS_ID_OBJECT)),&result,sizeof(TPMS_ID_OBJECT)), SWIGTYPE_p_TPMS_ID_OBJECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_ID_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_ID_OBJECT_PTR_PTR")) SWIG_fail;
  result = (TPMS_ID_OBJECT **)new_TPMS_ID_OBJECT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_ID_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_ID_OBJECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_ID_OBJECT_PTR_PTR" "', argument " "1"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  result = (TPMS_ID_OBJECT **)copy_TPMS_ID_OBJECT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_ID_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT **arg1 = (TPMS_ID_OBJECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_ID_OBJECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_ID_OBJECT_PTR_PTR" "', argument " "1"" of type '" "TPMS_ID_OBJECT **""'"); 
  }
  arg1 = (TPMS_ID_OBJECT **)(argp1);
  delete_TPMS_ID_OBJECT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT **arg1 = (TPMS_ID_OBJECT **) 0 ;
  TPMS_ID_OBJECT *arg2 = (TPMS_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_ID_OBJECT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_ID_OBJECT **""'"); 
  }
  arg1 = (TPMS_ID_OBJECT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_ID_OBJECT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg2 = (TPMS_ID_OBJECT *)(argp2);
  TPMS_ID_OBJECT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_ID_OBJECT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT **arg1 = (TPMS_ID_OBJECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_ID_OBJECT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_ID_OBJECT_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_ID_OBJECT **""'"); 
  }
  arg1 = (TPMS_ID_OBJECT **)(argp1);
  result = (TPMS_ID_OBJECT *)TPMS_ID_OBJECT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_ID_OBJECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_ID_OBJECT",SWIG_From_size_t((size_t)(sizeof(TPM2B_ID_OBJECT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ID_OBJECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_ID_OBJECT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_ID_OBJECT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_ID_OBJECT_ARRAY *)new_TPM2B_ID_OBJECT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ID_OBJECT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT_ARRAY *arg1 = (TPM2B_ID_OBJECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ID_OBJECT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ID_OBJECT_ARRAY" "', argument " "1"" of type '" "TPM2B_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT_ARRAY *)(argp1);
  delete_TPM2B_ID_OBJECT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT_ARRAY *arg1 = (TPM2B_ID_OBJECT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_ID_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ID_OBJECT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ID_OBJECT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_ID_OBJECT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_ID_OBJECT *)memcpy((TPM2B_ID_OBJECT *)malloc(sizeof(TPM2B_ID_OBJECT)),&result,sizeof(TPM2B_ID_OBJECT)), SWIGTYPE_p_TPM2B_ID_OBJECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT_ARRAY *arg1 = (TPM2B_ID_OBJECT_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_ID_OBJECT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_ID_OBJECT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_ID_OBJECT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_ID_OBJECT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_ID_OBJECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ID_OBJECT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ID_OBJECT_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_ID_OBJECT""'");
    } else {
      arg3 = *((TPM2B_ID_OBJECT *)(argp3));
    }
  }
  TPM2B_ID_OBJECT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT_ARRAY *arg1 = (TPM2B_ID_OBJECT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ID_OBJECT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_ID_OBJECT_ARRAY *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT_ARRAY *)(argp1);
  result = (TPM2B_ID_OBJECT *)TPM2B_ID_OBJECT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ID_OBJECT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  result = (TPM2B_ID_OBJECT_ARRAY *)TPM2B_ID_OBJECT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_ID_OBJECT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_ID_OBJECT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_ID_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ID_OBJECT_PTR")) SWIG_fail;
  result = (TPM2B_ID_OBJECT *)new_TPM2B_ID_OBJECT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ID_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ID_OBJECT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_ID_OBJECT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ID_OBJECT_PTR" "', argument " "1"" of type '" "TPM2B_ID_OBJECT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_ID_OBJECT_PTR" "', argument " "1"" of type '" "TPM2B_ID_OBJECT""'");
    } else {
      arg1 = *((TPM2B_ID_OBJECT *)(argp1));
    }
  }
  result = (TPM2B_ID_OBJECT *)copy_TPM2B_ID_OBJECT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ID_OBJECT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ID_OBJECT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ID_OBJECT_PTR" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  delete_TPM2B_ID_OBJECT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  TPM2B_ID_OBJECT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ID_OBJECT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_PTR_assign" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_ID_OBJECT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ID_OBJECT_PTR_assign" "', argument " "2"" of type '" "TPM2B_ID_OBJECT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_ID_OBJECT_PTR_assign" "', argument " "2"" of type '" "TPM2B_ID_OBJECT""'");
    } else {
      arg2 = *((TPM2B_ID_OBJECT *)(argp2));
    }
  }
  TPM2B_ID_OBJECT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ID_OBJECT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_PTR_value" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  result = TPM2B_ID_OBJECT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_ID_OBJECT *)memcpy((TPM2B_ID_OBJECT *)malloc(sizeof(TPM2B_ID_OBJECT)),&result,sizeof(TPM2B_ID_OBJECT)), SWIGTYPE_p_TPM2B_ID_OBJECT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_ID_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_ID_OBJECT_PTR_PTR")) SWIG_fail;
  result = (TPM2B_ID_OBJECT **)new_TPM2B_ID_OBJECT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_ID_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_ID_OBJECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_ID_OBJECT_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  result = (TPM2B_ID_OBJECT **)copy_TPM2B_ID_OBJECT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_ID_OBJECT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT **arg1 = (TPM2B_ID_OBJECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_ID_OBJECT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_ID_OBJECT_PTR_PTR" "', argument " "1"" of type '" "TPM2B_ID_OBJECT **""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT **)(argp1);
  delete_TPM2B_ID_OBJECT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT **arg1 = (TPM2B_ID_OBJECT **) 0 ;
  TPM2B_ID_OBJECT *arg2 = (TPM2B_ID_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_ID_OBJECT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_ID_OBJECT **""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_ID_OBJECT_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg2 = (TPM2B_ID_OBJECT *)(argp2);
  TPM2B_ID_OBJECT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_ID_OBJECT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT **arg1 = (TPM2B_ID_OBJECT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_ID_OBJECT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_ID_OBJECT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_ID_OBJECT_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_ID_OBJECT **""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT **)(argp1);
  result = (TPM2B_ID_OBJECT *)TPM2B_ID_OBJECT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_NV_PIN_COUNTER_PARAMETERS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_NV_PIN_COUNTER_PARAMETERS",SWIG_From_size_t((size_t)(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)(argp1);
  delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_NV_PIN_COUNTER_PARAMETERS *)memcpy((TPMS_NV_PIN_COUNTER_PARAMETERS *)malloc(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS)),&result,sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS)), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS""'");
    } else {
      arg3 = *((TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp3));
    }
  }
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)(argp1);
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS *)TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *)TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR")) SWIG_fail;
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS *)new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS""'");
    } else {
      arg1 = *((TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1));
    }
  }
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS *)copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS""'");
    } else {
      arg2 = *((TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp2));
    }
  }
  TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  result = TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_NV_PIN_COUNTER_PARAMETERS *)memcpy((TPMS_NV_PIN_COUNTER_PARAMETERS *)malloc(sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS)),&result,sizeof(TPMS_NV_PIN_COUNTER_PARAMETERS)), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR")) SWIG_fail;
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS **)new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS **)copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS **arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS **""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS **)(argp1);
  delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS **arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS **) 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg2 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS **""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg2 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp2);
  TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS **arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS **""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS **)(argp1);
  result = (TPMS_NV_PIN_COUNTER_PARAMETERS *)TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_NV_PUBLIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_NV_PUBLIC",SWIG_From_size_t((size_t)(sizeof(TPMS_NV_PUBLIC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_NV_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_NV_PUBLIC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_NV_PUBLIC_ARRAY *)new_TPMS_NV_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC_ARRAY *arg1 = (TPMS_NV_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PUBLIC_ARRAY" "', argument " "1"" of type '" "TPMS_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC_ARRAY *)(argp1);
  delete_TPMS_NV_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC_ARRAY *arg1 = (TPMS_NV_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_NV_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PUBLIC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_NV_PUBLIC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_NV_PUBLIC *)memcpy((TPMS_NV_PUBLIC *)malloc(sizeof(TPMS_NV_PUBLIC)),&result,sizeof(TPMS_NV_PUBLIC)), SWIGTYPE_p_TPMS_NV_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC_ARRAY *arg1 = (TPMS_NV_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_NV_PUBLIC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_NV_PUBLIC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_NV_PUBLIC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_NV_PUBLIC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_NV_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_NV_PUBLIC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_NV_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_NV_PUBLIC""'");
    } else {
      arg3 = *((TPMS_NV_PUBLIC *)(argp3));
    }
  }
  TPMS_NV_PUBLIC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC_ARRAY *arg1 = (TPMS_NV_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_ARRAY_cast" "', argument " "1"" of type '" "TPMS_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC_ARRAY *)(argp1);
  result = (TPMS_NV_PUBLIC *)TPMS_NV_PUBLIC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (TPMS_NV_PUBLIC_ARRAY *)TPMS_NV_PUBLIC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_NV_PUBLIC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_NV_PUBLIC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_NV_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_PUBLIC_PTR")) SWIG_fail;
  result = (TPMS_NV_PUBLIC *)new_TPMS_NV_PUBLIC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_NV_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_NV_PUBLIC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_NV_PUBLIC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_NV_PUBLIC_PTR" "', argument " "1"" of type '" "TPMS_NV_PUBLIC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_NV_PUBLIC_PTR" "', argument " "1"" of type '" "TPMS_NV_PUBLIC""'");
    } else {
      arg1 = *((TPMS_NV_PUBLIC *)(argp1));
    }
  }
  result = (TPMS_NV_PUBLIC *)copy_TPMS_NV_PUBLIC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PUBLIC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PUBLIC_PTR" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  delete_TPMS_NV_PUBLIC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  TPMS_NV_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_PTR_assign" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_NV_PUBLIC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_PUBLIC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_NV_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_PUBLIC""'");
    } else {
      arg2 = *((TPMS_NV_PUBLIC *)(argp2));
    }
  }
  TPMS_NV_PUBLIC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_PTR_value" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = TPMS_NV_PUBLIC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_NV_PUBLIC *)memcpy((TPMS_NV_PUBLIC *)malloc(sizeof(TPMS_NV_PUBLIC)),&result,sizeof(TPMS_NV_PUBLIC)), SWIGTYPE_p_TPMS_NV_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_NV_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_NV_PUBLIC_PTR_PTR")) SWIG_fail;
  result = (TPMS_NV_PUBLIC **)new_TPMS_NV_PUBLIC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_NV_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_NV_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_NV_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  result = (TPMS_NV_PUBLIC **)copy_TPMS_NV_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_NV_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC **arg1 = (TPMS_NV_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_NV_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_NV_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPMS_NV_PUBLIC **""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC **)(argp1);
  delete_TPMS_NV_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC **arg1 = (TPMS_NV_PUBLIC **) 0 ;
  TPMS_NV_PUBLIC *arg2 = (TPMS_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_NV_PUBLIC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_NV_PUBLIC **""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_NV_PUBLIC_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg2 = (TPMS_NV_PUBLIC *)(argp2);
  TPMS_NV_PUBLIC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_NV_PUBLIC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC **arg1 = (TPMS_NV_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_NV_PUBLIC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_NV_PUBLIC_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_NV_PUBLIC **""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC **)(argp1);
  result = (TPMS_NV_PUBLIC *)TPMS_NV_PUBLIC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_NV_PUBLIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_NV_PUBLIC",SWIG_From_size_t((size_t)(sizeof(TPM2B_NV_PUBLIC))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_NV_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_NV_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_NV_PUBLIC_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_NV_PUBLIC_ARRAY *)new_TPM2B_NV_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NV_PUBLIC_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC_ARRAY *arg1 = (TPM2B_NV_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NV_PUBLIC_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NV_PUBLIC_ARRAY" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC_ARRAY *)(argp1);
  delete_TPM2B_NV_PUBLIC_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC_ARRAY *arg1 = (TPM2B_NV_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_NV_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NV_PUBLIC_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_NV_PUBLIC_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_NV_PUBLIC_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_NV_PUBLIC *)memcpy((TPM2B_NV_PUBLIC *)malloc(sizeof(TPM2B_NV_PUBLIC)),&result,sizeof(TPM2B_NV_PUBLIC)), SWIGTYPE_p_TPM2B_NV_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC_ARRAY *arg1 = (TPM2B_NV_PUBLIC_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_NV_PUBLIC arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_NV_PUBLIC_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_NV_PUBLIC_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_NV_PUBLIC,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_NV_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_NV_PUBLIC""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_NV_PUBLIC_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_NV_PUBLIC""'");
    } else {
      arg3 = *((TPM2B_NV_PUBLIC *)(argp3));
    }
  }
  TPM2B_NV_PUBLIC_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC_ARRAY *arg1 = (TPM2B_NV_PUBLIC_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NV_PUBLIC_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC_ARRAY *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC_ARRAY *)(argp1);
  result = (TPM2B_NV_PUBLIC *)TPM2B_NV_PUBLIC_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NV_PUBLIC_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  result = (TPM2B_NV_PUBLIC_ARRAY *)TPM2B_NV_PUBLIC_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_NV_PUBLIC_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_NV_PUBLIC_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_NV_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_NV_PUBLIC_PTR")) SWIG_fail;
  result = (TPM2B_NV_PUBLIC *)new_TPM2B_NV_PUBLIC_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_NV_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_NV_PUBLIC_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_NV_PUBLIC,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_NV_PUBLIC_PTR" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_NV_PUBLIC_PTR" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC""'");
    } else {
      arg1 = *((TPM2B_NV_PUBLIC *)(argp1));
    }
  }
  result = (TPM2B_NV_PUBLIC *)copy_TPM2B_NV_PUBLIC_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NV_PUBLIC_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NV_PUBLIC_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NV_PUBLIC_PTR" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  delete_TPM2B_NV_PUBLIC_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  TPM2B_NV_PUBLIC arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NV_PUBLIC_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_PTR_assign" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_NV_PUBLIC,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_NV_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPM2B_NV_PUBLIC""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_NV_PUBLIC_PTR_assign" "', argument " "2"" of type '" "TPM2B_NV_PUBLIC""'");
    } else {
      arg2 = *((TPM2B_NV_PUBLIC *)(argp2));
    }
  }
  TPM2B_NV_PUBLIC_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NV_PUBLIC_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_PTR_value" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  result = TPM2B_NV_PUBLIC_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_NV_PUBLIC *)memcpy((TPM2B_NV_PUBLIC *)malloc(sizeof(TPM2B_NV_PUBLIC)),&result,sizeof(TPM2B_NV_PUBLIC)), SWIGTYPE_p_TPM2B_NV_PUBLIC, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_NV_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_NV_PUBLIC_PTR_PTR")) SWIG_fail;
  result = (TPM2B_NV_PUBLIC **)new_TPM2B_NV_PUBLIC_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_NV_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_NV_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_NV_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  result = (TPM2B_NV_PUBLIC **)copy_TPM2B_NV_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_NV_PUBLIC_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC **arg1 = (TPM2B_NV_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_NV_PUBLIC_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_NV_PUBLIC_PTR_PTR" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC **""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC **)(argp1);
  delete_TPM2B_NV_PUBLIC_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC **arg1 = (TPM2B_NV_PUBLIC **) 0 ;
  TPM2B_NV_PUBLIC *arg2 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_NV_PUBLIC_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC **""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_NV_PUBLIC_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg2 = (TPM2B_NV_PUBLIC *)(argp2);
  TPM2B_NV_PUBLIC_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_NV_PUBLIC_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC **arg1 = (TPM2B_NV_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_NV_PUBLIC *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_NV_PUBLIC_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_NV_PUBLIC_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC **""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC **)(argp1);
  result = (TPM2B_NV_PUBLIC *)TPM2B_NV_PUBLIC_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_CONTEXT_SENSITIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_CONTEXT_SENSITIVE",SWIG_From_size_t((size_t)(sizeof(TPM2B_CONTEXT_SENSITIVE))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_SENSITIVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_CONTEXT_SENSITIVE_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_CONTEXT_SENSITIVE_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_CONTEXT_SENSITIVE_ARRAY *)new_TPM2B_CONTEXT_SENSITIVE_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_SENSITIVE_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE_ARRAY *arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_SENSITIVE_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_SENSITIVE_ARRAY" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *)(argp1);
  delete_TPM2B_CONTEXT_SENSITIVE_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE_ARRAY *arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_CONTEXT_SENSITIVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_CONTEXT_SENSITIVE *)memcpy((TPM2B_CONTEXT_SENSITIVE *)malloc(sizeof(TPM2B_CONTEXT_SENSITIVE)),&result,sizeof(TPM2B_CONTEXT_SENSITIVE)), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE_ARRAY *arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_CONTEXT_SENSITIVE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_CONTEXT_SENSITIVE""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_CONTEXT_SENSITIVE""'");
    } else {
      arg3 = *((TPM2B_CONTEXT_SENSITIVE *)(argp3));
    }
  }
  TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE_ARRAY *arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_SENSITIVE_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE_ARRAY *)(argp1);
  result = (TPM2B_CONTEXT_SENSITIVE *)TPM2B_CONTEXT_SENSITIVE_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  result = (TPM2B_CONTEXT_SENSITIVE_ARRAY *)TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_CONTEXT_SENSITIVE_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CONTEXT_SENSITIVE_PTR")) SWIG_fail;
  result = (TPM2B_CONTEXT_SENSITIVE *)new_TPM2B_CONTEXT_SENSITIVE_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_CONTEXT_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_CONTEXT_SENSITIVE_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_CONTEXT_SENSITIVE_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_CONTEXT_SENSITIVE_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE""'");
    } else {
      arg1 = *((TPM2B_CONTEXT_SENSITIVE *)(argp1));
    }
  }
  result = (TPM2B_CONTEXT_SENSITIVE *)copy_TPM2B_CONTEXT_SENSITIVE_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_SENSITIVE_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_SENSITIVE_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_SENSITIVE_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  delete_TPM2B_CONTEXT_SENSITIVE_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  TPM2B_CONTEXT_SENSITIVE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_SENSITIVE_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_assign" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_assign" "', argument " "2"" of type '" "TPM2B_CONTEXT_SENSITIVE""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_assign" "', argument " "2"" of type '" "TPM2B_CONTEXT_SENSITIVE""'");
    } else {
      arg2 = *((TPM2B_CONTEXT_SENSITIVE *)(argp2));
    }
  }
  TPM2B_CONTEXT_SENSITIVE_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_SENSITIVE_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_value" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  result = TPM2B_CONTEXT_SENSITIVE_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_CONTEXT_SENSITIVE *)memcpy((TPM2B_CONTEXT_SENSITIVE *)malloc(sizeof(TPM2B_CONTEXT_SENSITIVE)),&result,sizeof(TPM2B_CONTEXT_SENSITIVE)), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CONTEXT_SENSITIVE_PTR_PTR")) SWIG_fail;
  result = (TPM2B_CONTEXT_SENSITIVE **)new_TPM2B_CONTEXT_SENSITIVE_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  result = (TPM2B_CONTEXT_SENSITIVE **)copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE **arg1 = (TPM2B_CONTEXT_SENSITIVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE **""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE **)(argp1);
  delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE **arg1 = (TPM2B_CONTEXT_SENSITIVE **) 0 ;
  TPM2B_CONTEXT_SENSITIVE *arg2 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE **""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg2 = (TPM2B_CONTEXT_SENSITIVE *)(argp2);
  TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE **arg1 = (TPM2B_CONTEXT_SENSITIVE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_SENSITIVE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE **""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE **)(argp1);
  result = (TPM2B_CONTEXT_SENSITIVE *)TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CONTEXT_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CONTEXT_DATA",SWIG_From_size_t((size_t)(sizeof(TPMS_CONTEXT_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CONTEXT_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CONTEXT_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CONTEXT_DATA_ARRAY *)new_TPMS_CONTEXT_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA_ARRAY *arg1 = (TPMS_CONTEXT_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_DATA_ARRAY" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA_ARRAY *)(argp1);
  delete_TPMS_CONTEXT_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA_ARRAY *arg1 = (TPMS_CONTEXT_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CONTEXT_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CONTEXT_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CONTEXT_DATA *)memcpy((TPMS_CONTEXT_DATA *)malloc(sizeof(TPMS_CONTEXT_DATA)),&result,sizeof(TPMS_CONTEXT_DATA)), SWIGTYPE_p_TPMS_CONTEXT_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA_ARRAY *arg1 = (TPMS_CONTEXT_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CONTEXT_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CONTEXT_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CONTEXT_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CONTEXT_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CONTEXT_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CONTEXT_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CONTEXT_DATA""'");
    } else {
      arg3 = *((TPMS_CONTEXT_DATA *)(argp3));
    }
  }
  TPMS_CONTEXT_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA_ARRAY *arg1 = (TPMS_CONTEXT_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA_ARRAY *)(argp1);
  result = (TPMS_CONTEXT_DATA *)TPMS_CONTEXT_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  result = (TPMS_CONTEXT_DATA_ARRAY *)TPMS_CONTEXT_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CONTEXT_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CONTEXT_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CONTEXT_DATA_PTR")) SWIG_fail;
  result = (TPMS_CONTEXT_DATA *)new_TPMS_CONTEXT_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CONTEXT_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CONTEXT_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CONTEXT_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CONTEXT_DATA_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CONTEXT_DATA_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA""'");
    } else {
      arg1 = *((TPMS_CONTEXT_DATA *)(argp1));
    }
  }
  result = (TPMS_CONTEXT_DATA *)copy_TPMS_CONTEXT_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_DATA_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  delete_TPMS_CONTEXT_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  TPMS_CONTEXT_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_PTR_assign" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CONTEXT_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_DATA_PTR_assign" "', argument " "2"" of type '" "TPMS_CONTEXT_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CONTEXT_DATA_PTR_assign" "', argument " "2"" of type '" "TPMS_CONTEXT_DATA""'");
    } else {
      arg2 = *((TPMS_CONTEXT_DATA *)(argp2));
    }
  }
  TPMS_CONTEXT_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_PTR_value" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  result = TPMS_CONTEXT_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CONTEXT_DATA *)memcpy((TPMS_CONTEXT_DATA *)malloc(sizeof(TPMS_CONTEXT_DATA)),&result,sizeof(TPMS_CONTEXT_DATA)), SWIGTYPE_p_TPMS_CONTEXT_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CONTEXT_DATA_PTR_PTR")) SWIG_fail;
  result = (TPMS_CONTEXT_DATA **)new_TPMS_CONTEXT_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CONTEXT_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CONTEXT_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CONTEXT_DATA_PTR_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  result = (TPMS_CONTEXT_DATA **)copy_TPMS_CONTEXT_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA **arg1 = (TPMS_CONTEXT_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_DATA_PTR_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA **""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA **)(argp1);
  delete_TPMS_CONTEXT_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA **arg1 = (TPMS_CONTEXT_DATA **) 0 ;
  TPMS_CONTEXT_DATA *arg2 = (TPMS_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA **""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg2 = (TPMS_CONTEXT_DATA *)(argp2);
  TPMS_CONTEXT_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA **arg1 = (TPMS_CONTEXT_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA **""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA **)(argp1);
  result = (TPMS_CONTEXT_DATA *)TPMS_CONTEXT_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_CONTEXT_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_CONTEXT_DATA",SWIG_From_size_t((size_t)(sizeof(TPM2B_CONTEXT_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_CONTEXT_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_CONTEXT_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_CONTEXT_DATA_ARRAY *)new_TPM2B_CONTEXT_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA_ARRAY *arg1 = (TPM2B_CONTEXT_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_DATA_ARRAY" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA_ARRAY *)(argp1);
  delete_TPM2B_CONTEXT_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA_ARRAY *arg1 = (TPM2B_CONTEXT_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_CONTEXT_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CONTEXT_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_CONTEXT_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_CONTEXT_DATA *)memcpy((TPM2B_CONTEXT_DATA *)malloc(sizeof(TPM2B_CONTEXT_DATA)),&result,sizeof(TPM2B_CONTEXT_DATA)), SWIGTYPE_p_TPM2B_CONTEXT_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA_ARRAY *arg1 = (TPM2B_CONTEXT_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_CONTEXT_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_CONTEXT_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CONTEXT_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_CONTEXT_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_CONTEXT_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_CONTEXT_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_CONTEXT_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_CONTEXT_DATA""'");
    } else {
      arg3 = *((TPM2B_CONTEXT_DATA *)(argp3));
    }
  }
  TPM2B_CONTEXT_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA_ARRAY *arg1 = (TPM2B_CONTEXT_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA_ARRAY *)(argp1);
  result = (TPM2B_CONTEXT_DATA *)TPM2B_CONTEXT_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  result = (TPM2B_CONTEXT_DATA_ARRAY *)TPM2B_CONTEXT_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_CONTEXT_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_CONTEXT_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CONTEXT_DATA_PTR")) SWIG_fail;
  result = (TPM2B_CONTEXT_DATA *)new_TPM2B_CONTEXT_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_CONTEXT_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_CONTEXT_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_CONTEXT_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_CONTEXT_DATA_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_CONTEXT_DATA_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA""'");
    } else {
      arg1 = *((TPM2B_CONTEXT_DATA *)(argp1));
    }
  }
  result = (TPM2B_CONTEXT_DATA *)copy_TPM2B_CONTEXT_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_DATA_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  delete_TPM2B_CONTEXT_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  TPM2B_CONTEXT_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_PTR_assign" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_CONTEXT_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CONTEXT_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_CONTEXT_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_CONTEXT_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_CONTEXT_DATA""'");
    } else {
      arg2 = *((TPM2B_CONTEXT_DATA *)(argp2));
    }
  }
  TPM2B_CONTEXT_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_PTR_value" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  result = TPM2B_CONTEXT_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_CONTEXT_DATA *)memcpy((TPM2B_CONTEXT_DATA *)malloc(sizeof(TPM2B_CONTEXT_DATA)),&result,sizeof(TPM2B_CONTEXT_DATA)), SWIGTYPE_p_TPM2B_CONTEXT_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CONTEXT_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CONTEXT_DATA_PTR_PTR")) SWIG_fail;
  result = (TPM2B_CONTEXT_DATA **)new_TPM2B_CONTEXT_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_CONTEXT_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_CONTEXT_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_CONTEXT_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  result = (TPM2B_CONTEXT_DATA **)copy_TPM2B_CONTEXT_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CONTEXT_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA **arg1 = (TPM2B_CONTEXT_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CONTEXT_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CONTEXT_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA **""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA **)(argp1);
  delete_TPM2B_CONTEXT_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA **arg1 = (TPM2B_CONTEXT_DATA **) 0 ;
  TPM2B_CONTEXT_DATA *arg2 = (TPM2B_CONTEXT_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CONTEXT_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA **""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CONTEXT_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg2 = (TPM2B_CONTEXT_DATA *)(argp2);
  TPM2B_CONTEXT_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CONTEXT_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA **arg1 = (TPM2B_CONTEXT_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CONTEXT_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CONTEXT_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CONTEXT_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA **""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA **)(argp1);
  result = (TPM2B_CONTEXT_DATA *)TPM2B_CONTEXT_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CONTEXT",SWIG_From_size_t((size_t)(sizeof(TPMS_CONTEXT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CONTEXT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CONTEXT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CONTEXT_ARRAY *)new_TPMS_CONTEXT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_ARRAY *arg1 = (TPMS_CONTEXT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_ARRAY" "', argument " "1"" of type '" "TPMS_CONTEXT_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_ARRAY *)(argp1);
  delete_TPMS_CONTEXT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_ARRAY *arg1 = (TPMS_CONTEXT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CONTEXT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CONTEXT_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CONTEXT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CONTEXT *)memcpy((TPMS_CONTEXT *)malloc(sizeof(TPMS_CONTEXT)),&result,sizeof(TPMS_CONTEXT)), SWIGTYPE_p_TPMS_CONTEXT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_ARRAY *arg1 = (TPMS_CONTEXT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CONTEXT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CONTEXT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CONTEXT_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CONTEXT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CONTEXT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CONTEXT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CONTEXT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CONTEXT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CONTEXT""'");
    } else {
      arg3 = *((TPMS_CONTEXT *)(argp3));
    }
  }
  TPMS_CONTEXT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_ARRAY *arg1 = (TPMS_CONTEXT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CONTEXT_ARRAY *""'"); 
  }
  arg1 = (TPMS_CONTEXT_ARRAY *)(argp1);
  result = (TPMS_CONTEXT *)TPMS_CONTEXT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = (TPMS_CONTEXT_ARRAY *)TPMS_CONTEXT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CONTEXT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CONTEXT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CONTEXT_PTR")) SWIG_fail;
  result = (TPMS_CONTEXT *)new_TPMS_CONTEXT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CONTEXT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CONTEXT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CONTEXT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CONTEXT_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CONTEXT_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT""'");
    } else {
      arg1 = *((TPMS_CONTEXT *)(argp1));
    }
  }
  result = (TPMS_CONTEXT *)copy_TPMS_CONTEXT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  delete_TPMS_CONTEXT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  TPMS_CONTEXT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_PTR_assign" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CONTEXT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_PTR_assign" "', argument " "2"" of type '" "TPMS_CONTEXT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CONTEXT_PTR_assign" "', argument " "2"" of type '" "TPMS_CONTEXT""'");
    } else {
      arg2 = *((TPMS_CONTEXT *)(argp2));
    }
  }
  TPMS_CONTEXT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_PTR_value" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = TPMS_CONTEXT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CONTEXT *)memcpy((TPMS_CONTEXT *)malloc(sizeof(TPMS_CONTEXT)),&result,sizeof(TPMS_CONTEXT)), SWIGTYPE_p_TPMS_CONTEXT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CONTEXT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CONTEXT_PTR_PTR")) SWIG_fail;
  result = (TPMS_CONTEXT **)new_TPMS_CONTEXT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CONTEXT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CONTEXT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CONTEXT_PTR_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  result = (TPMS_CONTEXT **)copy_TPMS_CONTEXT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CONTEXT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT **arg1 = (TPMS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CONTEXT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CONTEXT_PTR_PTR" "', argument " "1"" of type '" "TPMS_CONTEXT **""'"); 
  }
  arg1 = (TPMS_CONTEXT **)(argp1);
  delete_TPMS_CONTEXT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT **arg1 = (TPMS_CONTEXT **) 0 ;
  TPMS_CONTEXT *arg2 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CONTEXT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CONTEXT **""'"); 
  }
  arg1 = (TPMS_CONTEXT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CONTEXT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg2 = (TPMS_CONTEXT *)(argp2);
  TPMS_CONTEXT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CONTEXT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT **arg1 = (TPMS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CONTEXT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CONTEXT_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CONTEXT **""'"); 
  }
  arg1 = (TPMS_CONTEXT **)(argp1);
  result = (TPMS_CONTEXT *)TPMS_CONTEXT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_CREATION_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_CREATION_DATA",SWIG_From_size_t((size_t)(sizeof(TPMS_CREATION_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_CREATION_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_CREATION_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_CREATION_DATA_ARRAY *)new_TPMS_CREATION_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA_ARRAY *arg1 = (TPMS_CREATION_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_DATA_ARRAY" "', argument " "1"" of type '" "TPMS_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA_ARRAY *)(argp1);
  delete_TPMS_CREATION_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA_ARRAY *arg1 = (TPMS_CREATION_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_CREATION_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CREATION_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_CREATION_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_CREATION_DATA *)memcpy((TPMS_CREATION_DATA *)malloc(sizeof(TPMS_CREATION_DATA)),&result,sizeof(TPMS_CREATION_DATA)), SWIGTYPE_p_TPMS_CREATION_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA_ARRAY *arg1 = (TPMS_CREATION_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_CREATION_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_CREATION_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_CREATION_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_CREATION_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_CREATION_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CREATION_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CREATION_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_CREATION_DATA""'");
    } else {
      arg3 = *((TPMS_CREATION_DATA *)(argp3));
    }
  }
  TPMS_CREATION_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA_ARRAY *arg1 = (TPMS_CREATION_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPMS_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA_ARRAY *)(argp1);
  result = (TPMS_CREATION_DATA *)TPMS_CREATION_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPMS_CREATION_DATA_ARRAY *)TPMS_CREATION_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_CREATION_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_CREATION_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CREATION_DATA_PTR")) SWIG_fail;
  result = (TPMS_CREATION_DATA *)new_TPMS_CREATION_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CREATION_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CREATION_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_CREATION_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CREATION_DATA_PTR" "', argument " "1"" of type '" "TPMS_CREATION_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_CREATION_DATA_PTR" "', argument " "1"" of type '" "TPMS_CREATION_DATA""'");
    } else {
      arg1 = *((TPMS_CREATION_DATA *)(argp1));
    }
  }
  result = (TPMS_CREATION_DATA *)copy_TPMS_CREATION_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_DATA_PTR" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  delete_TPMS_CREATION_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  TPMS_CREATION_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_PTR_assign" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_CREATION_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_PTR_assign" "', argument " "2"" of type '" "TPMS_CREATION_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_CREATION_DATA_PTR_assign" "', argument " "2"" of type '" "TPMS_CREATION_DATA""'");
    } else {
      arg2 = *((TPMS_CREATION_DATA *)(argp2));
    }
  }
  TPMS_CREATION_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_PTR_value" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = TPMS_CREATION_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_CREATION_DATA *)memcpy((TPMS_CREATION_DATA *)malloc(sizeof(TPMS_CREATION_DATA)),&result,sizeof(TPMS_CREATION_DATA)), SWIGTYPE_p_TPMS_CREATION_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_CREATION_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_CREATION_DATA_PTR_PTR")) SWIG_fail;
  result = (TPMS_CREATION_DATA **)new_TPMS_CREATION_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_CREATION_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_CREATION_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_CREATION_DATA_PTR_PTR" "', argument " "1"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  result = (TPMS_CREATION_DATA **)copy_TPMS_CREATION_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_CREATION_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA **arg1 = (TPMS_CREATION_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_CREATION_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_CREATION_DATA_PTR_PTR" "', argument " "1"" of type '" "TPMS_CREATION_DATA **""'"); 
  }
  arg1 = (TPMS_CREATION_DATA **)(argp1);
  delete_TPMS_CREATION_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA **arg1 = (TPMS_CREATION_DATA **) 0 ;
  TPMS_CREATION_DATA *arg2 = (TPMS_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_CREATION_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_CREATION_DATA **""'"); 
  }
  arg1 = (TPMS_CREATION_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_CREATION_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg2 = (TPMS_CREATION_DATA *)(argp2);
  TPMS_CREATION_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_CREATION_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA **arg1 = (TPMS_CREATION_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_CREATION_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_CREATION_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_CREATION_DATA **""'"); 
  }
  arg1 = (TPMS_CREATION_DATA **)(argp1);
  result = (TPMS_CREATION_DATA *)TPMS_CREATION_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPM2B_CREATION_DATA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPM2B_CREATION_DATA",SWIG_From_size_t((size_t)(sizeof(TPM2B_CREATION_DATA))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CREATION_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPM2B_CREATION_DATA_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPM2B_CREATION_DATA_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPM2B_CREATION_DATA_ARRAY *)new_TPM2B_CREATION_DATA_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CREATION_DATA_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA_ARRAY *arg1 = (TPM2B_CREATION_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CREATION_DATA_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CREATION_DATA_ARRAY" "', argument " "1"" of type '" "TPM2B_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA_ARRAY *)(argp1);
  delete_TPM2B_CREATION_DATA_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA_ARRAY *arg1 = (TPM2B_CREATION_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPM2B_CREATION_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CREATION_DATA_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_ARRAY___getitem__" "', argument " "1"" of type '" "TPM2B_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CREATION_DATA_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPM2B_CREATION_DATA_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPM2B_CREATION_DATA *)memcpy((TPM2B_CREATION_DATA *)malloc(sizeof(TPM2B_CREATION_DATA)),&result,sizeof(TPM2B_CREATION_DATA)), SWIGTYPE_p_TPM2B_CREATION_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA_ARRAY *arg1 = (TPM2B_CREATION_DATA_ARRAY *) 0 ;
  size_t arg2 ;
  TPM2B_CREATION_DATA arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPM2B_CREATION_DATA_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_ARRAY___setitem__" "', argument " "1"" of type '" "TPM2B_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPM2B_CREATION_DATA_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPM2B_CREATION_DATA,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPM2B_CREATION_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_CREATION_DATA""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_CREATION_DATA_ARRAY___setitem__" "', argument " "3"" of type '" "TPM2B_CREATION_DATA""'");
    } else {
      arg3 = *((TPM2B_CREATION_DATA *)(argp3));
    }
  }
  TPM2B_CREATION_DATA_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA_ARRAY *arg1 = (TPM2B_CREATION_DATA_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CREATION_DATA_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_ARRAY_cast" "', argument " "1"" of type '" "TPM2B_CREATION_DATA_ARRAY *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA_ARRAY *)(argp1);
  result = (TPM2B_CREATION_DATA *)TPM2B_CREATION_DATA_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CREATION_DATA_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_ARRAY_frompointer" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  result = (TPM2B_CREATION_DATA_ARRAY *)TPM2B_CREATION_DATA_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPM2B_CREATION_DATA_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPM2B_CREATION_DATA_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPM2B_CREATION_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CREATION_DATA_PTR")) SWIG_fail;
  result = (TPM2B_CREATION_DATA *)new_TPM2B_CREATION_DATA_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_CREATION_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_CREATION_DATA_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPM2B_CREATION_DATA,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_CREATION_DATA_PTR" "', argument " "1"" of type '" "TPM2B_CREATION_DATA""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPM2B_CREATION_DATA_PTR" "', argument " "1"" of type '" "TPM2B_CREATION_DATA""'");
    } else {
      arg1 = *((TPM2B_CREATION_DATA *)(argp1));
    }
  }
  result = (TPM2B_CREATION_DATA *)copy_TPM2B_CREATION_DATA_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CREATION_DATA_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CREATION_DATA_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CREATION_DATA_PTR" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  delete_TPM2B_CREATION_DATA_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  TPM2B_CREATION_DATA arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CREATION_DATA_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_PTR_assign" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPM2B_CREATION_DATA,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CREATION_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_CREATION_DATA""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPM2B_CREATION_DATA_PTR_assign" "', argument " "2"" of type '" "TPM2B_CREATION_DATA""'");
    } else {
      arg2 = *((TPM2B_CREATION_DATA *)(argp2));
    }
  }
  TPM2B_CREATION_DATA_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CREATION_DATA_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_PTR_value" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  result = TPM2B_CREATION_DATA_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPM2B_CREATION_DATA *)memcpy((TPM2B_CREATION_DATA *)malloc(sizeof(TPM2B_CREATION_DATA)),&result,sizeof(TPM2B_CREATION_DATA)), SWIGTYPE_p_TPM2B_CREATION_DATA, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPM2B_CREATION_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPM2B_CREATION_DATA_PTR_PTR")) SWIG_fail;
  result = (TPM2B_CREATION_DATA **)new_TPM2B_CREATION_DATA_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPM2B_CREATION_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPM2B_CREATION_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPM2B_CREATION_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  result = (TPM2B_CREATION_DATA **)copy_TPM2B_CREATION_DATA_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPM2B_CREATION_DATA_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA **arg1 = (TPM2B_CREATION_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPM2B_CREATION_DATA_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPM2B_CREATION_DATA_PTR_PTR" "', argument " "1"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA **)(argp1);
  delete_TPM2B_CREATION_DATA_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA **arg1 = (TPM2B_CREATION_DATA **) 0 ;
  TPM2B_CREATION_DATA *arg2 = (TPM2B_CREATION_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPM2B_CREATION_DATA_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_PTR_PTR_assign" "', argument " "1"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPM2B_CREATION_DATA_PTR_PTR_assign" "', argument " "2"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg2 = (TPM2B_CREATION_DATA *)(argp2);
  TPM2B_CREATION_DATA_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPM2B_CREATION_DATA_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA **arg1 = (TPM2B_CREATION_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPM2B_CREATION_DATA *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPM2B_CREATION_DATA_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPM2B_CREATION_DATA_PTR_PTR_value" "', argument " "1"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA **)(argp1);
  result = (TPM2B_CREATION_DATA *)TPM2B_CREATION_DATA_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPMS_AC_OUTPUT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPMS_AC_OUTPUT",SWIG_From_size_t((size_t)(sizeof(TPMS_AC_OUTPUT))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPMS_AC_OUTPUT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPMS_AC_OUTPUT_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPMS_AC_OUTPUT_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPMS_AC_OUTPUT_ARRAY *)new_TPMS_AC_OUTPUT_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AC_OUTPUT_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT_ARRAY *arg1 = (TPMS_AC_OUTPUT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AC_OUTPUT_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AC_OUTPUT_ARRAY" "', argument " "1"" of type '" "TPMS_AC_OUTPUT_ARRAY *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT_ARRAY *)(argp1);
  delete_TPMS_AC_OUTPUT_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT_ARRAY *arg1 = (TPMS_AC_OUTPUT_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPMS_AC_OUTPUT result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AC_OUTPUT_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_ARRAY___getitem__" "', argument " "1"" of type '" "TPMS_AC_OUTPUT_ARRAY *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AC_OUTPUT_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPMS_AC_OUTPUT_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPMS_AC_OUTPUT *)memcpy((TPMS_AC_OUTPUT *)malloc(sizeof(TPMS_AC_OUTPUT)),&result,sizeof(TPMS_AC_OUTPUT)), SWIGTYPE_p_TPMS_AC_OUTPUT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT_ARRAY *arg1 = (TPMS_AC_OUTPUT_ARRAY *) 0 ;
  size_t arg2 ;
  TPMS_AC_OUTPUT arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPMS_AC_OUTPUT_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_ARRAY___setitem__" "', argument " "1"" of type '" "TPMS_AC_OUTPUT_ARRAY *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPMS_AC_OUTPUT_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPMS_AC_OUTPUT,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPMS_AC_OUTPUT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_AC_OUTPUT""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_AC_OUTPUT_ARRAY___setitem__" "', argument " "3"" of type '" "TPMS_AC_OUTPUT""'");
    } else {
      arg3 = *((TPMS_AC_OUTPUT *)(argp3));
    }
  }
  TPMS_AC_OUTPUT_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT_ARRAY *arg1 = (TPMS_AC_OUTPUT_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AC_OUTPUT_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_ARRAY_cast" "', argument " "1"" of type '" "TPMS_AC_OUTPUT_ARRAY *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT_ARRAY *)(argp1);
  result = (TPMS_AC_OUTPUT *)TPMS_AC_OUTPUT_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AC_OUTPUT_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_ARRAY_frompointer" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  result = (TPMS_AC_OUTPUT_ARRAY *)TPMS_AC_OUTPUT_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPMS_AC_OUTPUT_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPMS_AC_OUTPUT_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPMS_AC_OUTPUT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AC_OUTPUT_PTR")) SWIG_fail;
  result = (TPMS_AC_OUTPUT *)new_TPMS_AC_OUTPUT_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_AC_OUTPUT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_AC_OUTPUT_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPMS_AC_OUTPUT,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_AC_OUTPUT_PTR" "', argument " "1"" of type '" "TPMS_AC_OUTPUT""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPMS_AC_OUTPUT_PTR" "', argument " "1"" of type '" "TPMS_AC_OUTPUT""'");
    } else {
      arg1 = *((TPMS_AC_OUTPUT *)(argp1));
    }
  }
  result = (TPMS_AC_OUTPUT *)copy_TPMS_AC_OUTPUT_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AC_OUTPUT_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AC_OUTPUT_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AC_OUTPUT_PTR" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  delete_TPMS_AC_OUTPUT_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  TPMS_AC_OUTPUT arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AC_OUTPUT_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_PTR_assign" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPMS_AC_OUTPUT,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AC_OUTPUT_PTR_assign" "', argument " "2"" of type '" "TPMS_AC_OUTPUT""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPMS_AC_OUTPUT_PTR_assign" "', argument " "2"" of type '" "TPMS_AC_OUTPUT""'");
    } else {
      arg2 = *((TPMS_AC_OUTPUT *)(argp2));
    }
  }
  TPMS_AC_OUTPUT_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AC_OUTPUT_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_PTR_value" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  result = TPMS_AC_OUTPUT_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPMS_AC_OUTPUT *)memcpy((TPMS_AC_OUTPUT *)malloc(sizeof(TPMS_AC_OUTPUT)),&result,sizeof(TPMS_AC_OUTPUT)), SWIGTYPE_p_TPMS_AC_OUTPUT, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPMS_AC_OUTPUT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPMS_AC_OUTPUT_PTR_PTR")) SWIG_fail;
  result = (TPMS_AC_OUTPUT **)new_TPMS_AC_OUTPUT_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPMS_AC_OUTPUT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPMS_AC_OUTPUT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPMS_AC_OUTPUT_PTR_PTR" "', argument " "1"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  result = (TPMS_AC_OUTPUT **)copy_TPMS_AC_OUTPUT_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPMS_AC_OUTPUT_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT **arg1 = (TPMS_AC_OUTPUT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPMS_AC_OUTPUT_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPMS_AC_OUTPUT_PTR_PTR" "', argument " "1"" of type '" "TPMS_AC_OUTPUT **""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT **)(argp1);
  delete_TPMS_AC_OUTPUT_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT **arg1 = (TPMS_AC_OUTPUT **) 0 ;
  TPMS_AC_OUTPUT *arg2 = (TPMS_AC_OUTPUT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPMS_AC_OUTPUT_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_PTR_PTR_assign" "', argument " "1"" of type '" "TPMS_AC_OUTPUT **""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPMS_AC_OUTPUT_PTR_PTR_assign" "', argument " "2"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg2 = (TPMS_AC_OUTPUT *)(argp2);
  TPMS_AC_OUTPUT_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPMS_AC_OUTPUT_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT **arg1 = (TPMS_AC_OUTPUT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPMS_AC_OUTPUT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPMS_AC_OUTPUT_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPMS_AC_OUTPUT_PTR_PTR_value" "', argument " "1"" of type '" "TPMS_AC_OUTPUT **""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT **)(argp1);
  result = (TPMS_AC_OUTPUT *)TPMS_AC_OUTPUT_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *sizeof_TPML_AC_CAPABILITIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "sizeof_TPML_AC_CAPABILITIES",SWIG_From_size_t((size_t)(sizeof(TPML_AC_CAPABILITIES))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_new_TPML_AC_CAPABILITIES_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_TPML_AC_CAPABILITIES_ARRAY",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_TPML_AC_CAPABILITIES_ARRAY" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (TPML_AC_CAPABILITIES_ARRAY *)new_TPML_AC_CAPABILITIES_ARRAY(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_AC_CAPABILITIES_ARRAY(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES_ARRAY *arg1 = (TPML_AC_CAPABILITIES_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_AC_CAPABILITIES_ARRAY",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_AC_CAPABILITIES_ARRAY" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES_ARRAY *)(argp1);
  delete_TPML_AC_CAPABILITIES_ARRAY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_ARRAY___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES_ARRAY *arg1 = (TPML_AC_CAPABILITIES_ARRAY *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TPML_AC_CAPABILITIES result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_AC_CAPABILITIES_ARRAY___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_ARRAY___getitem__" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_AC_CAPABILITIES_ARRAY___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = TPML_AC_CAPABILITIES_ARRAY___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj((TPML_AC_CAPABILITIES *)memcpy((TPML_AC_CAPABILITIES *)malloc(sizeof(TPML_AC_CAPABILITIES)),&result,sizeof(TPML_AC_CAPABILITIES)), SWIGTYPE_p_TPML_AC_CAPABILITIES, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_ARRAY___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES_ARRAY *arg1 = (TPML_AC_CAPABILITIES_ARRAY *) 0 ;
  size_t arg2 ;
  TPML_AC_CAPABILITIES arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:TPML_AC_CAPABILITIES_ARRAY___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_ARRAY___setitem__" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES_ARRAY *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TPML_AC_CAPABILITIES_ARRAY___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  {
    res3 = SWIG_ConvertPtr(obj2, &argp3, SWIGTYPE_p_TPML_AC_CAPABILITIES,  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TPML_AC_CAPABILITIES_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_AC_CAPABILITIES""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_AC_CAPABILITIES_ARRAY___setitem__" "', argument " "3"" of type '" "TPML_AC_CAPABILITIES""'");
    } else {
      arg3 = *((TPML_AC_CAPABILITIES *)(argp3));
    }
  }
  TPML_AC_CAPABILITIES_ARRAY___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_ARRAY_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES_ARRAY *arg1 = (TPML_AC_CAPABILITIES_ARRAY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_AC_CAPABILITIES_ARRAY_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_ARRAY_cast" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES_ARRAY *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES_ARRAY *)(argp1);
  result = (TPML_AC_CAPABILITIES *)TPML_AC_CAPABILITIES_ARRAY_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_ARRAY_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES_ARRAY *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_AC_CAPABILITIES_ARRAY_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_ARRAY_frompointer" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  result = (TPML_AC_CAPABILITIES_ARRAY *)TPML_AC_CAPABILITIES_ARRAY_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TPML_AC_CAPABILITIES_ARRAY_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TPML_AC_CAPABILITIES_ARRAY, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TPML_AC_CAPABILITIES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_AC_CAPABILITIES_PTR")) SWIG_fail;
  result = (TPML_AC_CAPABILITIES *)new_TPML_AC_CAPABILITIES_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_AC_CAPABILITIES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_AC_CAPABILITIES_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_TPML_AC_CAPABILITIES,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_AC_CAPABILITIES_PTR" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_TPML_AC_CAPABILITIES_PTR" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES""'");
    } else {
      arg1 = *((TPML_AC_CAPABILITIES *)(argp1));
    }
  }
  result = (TPML_AC_CAPABILITIES *)copy_TPML_AC_CAPABILITIES_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_AC_CAPABILITIES_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_AC_CAPABILITIES_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_AC_CAPABILITIES_PTR" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  delete_TPML_AC_CAPABILITIES_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  TPML_AC_CAPABILITIES arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_AC_CAPABILITIES_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_PTR_assign" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_TPML_AC_CAPABILITIES,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_AC_CAPABILITIES_PTR_assign" "', argument " "2"" of type '" "TPML_AC_CAPABILITIES""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "TPML_AC_CAPABILITIES_PTR_assign" "', argument " "2"" of type '" "TPML_AC_CAPABILITIES""'");
    } else {
      arg2 = *((TPML_AC_CAPABILITIES *)(argp2));
    }
  }
  TPML_AC_CAPABILITIES_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_AC_CAPABILITIES_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_PTR_value" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  result = TPML_AC_CAPABILITIES_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((TPML_AC_CAPABILITIES *)memcpy((TPML_AC_CAPABILITIES *)malloc(sizeof(TPML_AC_CAPABILITIES)),&result,sizeof(TPML_AC_CAPABILITIES)), SWIGTYPE_p_TPML_AC_CAPABILITIES, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TPML_AC_CAPABILITIES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_TPML_AC_CAPABILITIES_PTR_PTR")) SWIG_fail;
  result = (TPML_AC_CAPABILITIES **)new_TPML_AC_CAPABILITIES_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_AC_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_TPML_AC_CAPABILITIES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_TPML_AC_CAPABILITIES_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_TPML_AC_CAPABILITIES_PTR_PTR" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  result = (TPML_AC_CAPABILITIES **)copy_TPML_AC_CAPABILITIES_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TPML_AC_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TPML_AC_CAPABILITIES_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES **arg1 = (TPML_AC_CAPABILITIES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TPML_AC_CAPABILITIES_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TPML_AC_CAPABILITIES_PTR_PTR" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES **""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES **)(argp1);
  delete_TPML_AC_CAPABILITIES_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES **arg1 = (TPML_AC_CAPABILITIES **) 0 ;
  TPML_AC_CAPABILITIES *arg2 = (TPML_AC_CAPABILITIES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TPML_AC_CAPABILITIES_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_PTR_PTR_assign" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES **""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TPML_AC_CAPABILITIES_PTR_PTR_assign" "', argument " "2"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg2 = (TPML_AC_CAPABILITIES *)(argp2);
  TPML_AC_CAPABILITIES_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TPML_AC_CAPABILITIES_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES **arg1 = (TPML_AC_CAPABILITIES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TPML_AC_CAPABILITIES *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TPML_AC_CAPABILITIES_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TPML_AC_CAPABILITIES_PTR_PTR_value" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES **""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES **)(argp1);
  result = (TPML_AC_CAPABILITIES *)TPML_AC_CAPABILITIES_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VOID_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_VOID_PTR_PTR")) SWIG_fail;
  result = (void **)new_VOID_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_VOID_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void *arg1 = (void *) 0 ;
  int res1 ;
  PyObject * obj0 = 0 ;
  void **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_VOID_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0,SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_VOID_PTR_PTR" "', argument " "1"" of type '" "void *""'"); 
  }
  result = (void **)copy_VOID_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VOID_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void **arg1 = (void **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_VOID_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_void, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VOID_PTR_PTR" "', argument " "1"" of type '" "void **""'"); 
  }
  arg1 = (void **)(argp1);
  delete_VOID_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VOID_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void **arg1 = (void **) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:VOID_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_void, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VOID_PTR_PTR_assign" "', argument " "1"" of type '" "void **""'"); 
  }
  arg1 = (void **)(argp1);
  res2 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VOID_PTR_PTR_assign" "', argument " "2"" of type '" "void *""'"); 
  }
  VOID_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VOID_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void **arg1 = (void **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  void *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:VOID_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_void, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VOID_PTR_PTR_value" "', argument " "1"" of type '" "void **""'"); 
  }
  arg1 = (void **)(argp1);
  result = (void *)VOID_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SIZE_T_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_SIZE_T_PTR")) SWIG_fail;
  result = (size_t *)new_SIZE_T_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_size_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_SIZE_T_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  size_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_SIZE_T_PTR",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "copy_SIZE_T_PTR" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (size_t *)copy_SIZE_T_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_size_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_SIZE_T_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t *arg1 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_SIZE_T_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SIZE_T_PTR" "', argument " "1"" of type '" "size_t *""'"); 
  }
  arg1 = (size_t *)(argp1);
  delete_SIZE_T_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SIZE_T_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t *arg1 = (size_t *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SIZE_T_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SIZE_T_PTR_assign" "', argument " "1"" of type '" "size_t *""'"); 
  }
  arg1 = (size_t *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SIZE_T_PTR_assign" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  SIZE_T_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SIZE_T_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t *arg1 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  size_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SIZE_T_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SIZE_T_PTR_value" "', argument " "1"" of type '" "size_t *""'"); 
  }
  arg1 = (size_t *)(argp1);
  result = SIZE_T_PTR_value(arg1);
  resultobj = SWIG_From_size_t((size_t)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SIZE_T_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_SIZE_T_PTR_PTR")) SWIG_fail;
  result = (size_t **)new_SIZE_T_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_size_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_SIZE_T_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t *arg1 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  size_t **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_SIZE_T_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_SIZE_T_PTR_PTR" "', argument " "1"" of type '" "size_t *""'"); 
  }
  arg1 = (size_t *)(argp1);
  result = (size_t **)copy_SIZE_T_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_size_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_SIZE_T_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t **arg1 = (size_t **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_SIZE_T_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SIZE_T_PTR_PTR" "', argument " "1"" of type '" "size_t **""'"); 
  }
  arg1 = (size_t **)(argp1);
  delete_SIZE_T_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SIZE_T_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t **arg1 = (size_t **) 0 ;
  size_t *arg2 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SIZE_T_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SIZE_T_PTR_PTR_assign" "', argument " "1"" of type '" "size_t **""'"); 
  }
  arg1 = (size_t **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SIZE_T_PTR_PTR_assign" "', argument " "2"" of type '" "size_t *""'"); 
  }
  arg2 = (size_t *)(argp2);
  SIZE_T_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SIZE_T_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t **arg1 = (size_t **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  size_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SIZE_T_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SIZE_T_PTR_PTR_value" "', argument " "1"" of type '" "size_t **""'"); 
  }
  arg1 = (size_t **)(argp1);
  result = (size_t *)SIZE_T_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_size_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CHAR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_CHAR_PTR")) SWIG_fail;
  result = (char *)new_CHAR_PTR();
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_CHAR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char arg1 ;
  char val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_CHAR_PTR",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "copy_CHAR_PTR" "', argument " "1"" of type '" "char""'");
  } 
  arg1 = (char)(val1);
  result = (char *)copy_CHAR_PTR(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CHAR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_CHAR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(obj0, &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CHAR_PTR" "', argument " "1"" of type '" "char *""'");
  }
  arg1 = (char *)(buf1);
  delete_CHAR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHAR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char arg2 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:CHAR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(obj0, &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHAR_PTR_assign" "', argument " "1"" of type '" "char *""'");
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHAR_PTR_assign" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = (char)(val2);
  CHAR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHAR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject * obj0 = 0 ;
  char result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:CHAR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(obj0, &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHAR_PTR_value" "', argument " "1"" of type '" "char *""'");
  }
  arg1 = (char *)(buf1);
  result = (char)CHAR_PTR_value(arg1);
  resultobj = SWIG_From_char((char)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CHAR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_CHAR_PTR_PTR")) SWIG_fail;
  result = (char **)new_CHAR_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_CHAR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject * obj0 = 0 ;
  char **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_CHAR_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(obj0, &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_CHAR_PTR_PTR" "', argument " "1"" of type '" "char *""'");
  }
  arg1 = (char *)(buf1);
  result = (char **)copy_CHAR_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CHAR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_CHAR_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CHAR_PTR_PTR" "', argument " "1"" of type '" "char **""'"); 
  }
  arg1 = (char **)(argp1);
  delete_CHAR_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHAR_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:CHAR_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHAR_PTR_PTR_assign" "', argument " "1"" of type '" "char **""'"); 
  }
  arg1 = (char **)(argp1);
  res2 = SWIG_AsCharPtrAndSize(obj1, &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHAR_PTR_PTR_assign" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = (char *)(buf2);
  CHAR_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHAR_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char **arg1 = (char **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:CHAR_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHAR_PTR_PTR_value" "', argument " "1"" of type '" "char **""'"); 
  }
  arg1 = (char **)(argp1);
  result = (char *)CHAR_PTR_PTR_value(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_UINT8_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_UINT8_PTR")) SWIG_fail;
  result = (UINT8 *)new_UINT8_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_UINT8_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 arg1 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_UINT8_PTR",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "copy_UINT8_PTR" "', argument " "1"" of type '" "UINT8""'");
  } 
  arg1 = (UINT8)(val1);
  result = (UINT8 *)copy_UINT8_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_UINT8_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 *arg1 = (UINT8 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_UINT8_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_UINT8_PTR" "', argument " "1"" of type '" "UINT8 *""'"); 
  }
  arg1 = (UINT8 *)(argp1);
  delete_UINT8_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT8_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 *arg1 = (UINT8 *) 0 ;
  UINT8 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:UINT8_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT8_PTR_assign" "', argument " "1"" of type '" "UINT8 *""'"); 
  }
  arg1 = (UINT8 *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_char(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "UINT8_PTR_assign" "', argument " "2"" of type '" "UINT8""'");
  } 
  arg2 = (UINT8)(val2);
  UINT8_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT8_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 *arg1 = (UINT8 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT8_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT8_PTR_value" "', argument " "1"" of type '" "UINT8 *""'"); 
  }
  arg1 = (UINT8 *)(argp1);
  result = (UINT8)UINT8_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_UINT8_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_UINT8_PTR_PTR")) SWIG_fail;
  result = (UINT8 **)new_UINT8_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_UINT8_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 *arg1 = (UINT8 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_UINT8_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_UINT8_PTR_PTR" "', argument " "1"" of type '" "UINT8 *""'"); 
  }
  arg1 = (UINT8 *)(argp1);
  result = (UINT8 **)copy_UINT8_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_UINT8_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 **arg1 = (UINT8 **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_UINT8_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_UINT8_PTR_PTR" "', argument " "1"" of type '" "UINT8 **""'"); 
  }
  arg1 = (UINT8 **)(argp1);
  delete_UINT8_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT8_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 **arg1 = (UINT8 **) 0 ;
  UINT8 *arg2 = (UINT8 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:UINT8_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT8_PTR_PTR_assign" "', argument " "1"" of type '" "UINT8 **""'"); 
  }
  arg1 = (UINT8 **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "UINT8_PTR_PTR_assign" "', argument " "2"" of type '" "UINT8 *""'"); 
  }
  arg2 = (UINT8 *)(argp2);
  UINT8_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT8_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 **arg1 = (UINT8 **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT8 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT8_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT8_PTR_PTR_value" "', argument " "1"" of type '" "UINT8 **""'"); 
  }
  arg1 = (UINT8 **)(argp1);
  result = (UINT8 *)UINT8_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_uint32_t_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_uint32_t_PTR")) SWIG_fail;
  result = (uint32_t_PTR *)new_uint32_t_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint32_t_PTR, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_uint32_t_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_uint32_t_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_uint32_t_PTR" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  delete_uint32_t_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint32_t_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:uint32_t_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint32_t_PTR_assign" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "uint32_t_PTR_assign" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  uint32_t_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint32_t_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint32_t_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint32_t_PTR_value" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  result = (uint32_t)uint32_t_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint32_t_PTR_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint32_t_PTR_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint32_t_PTR_cast" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  result = (uint32_t *)uint32_t_PTR_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint32_t_PTR_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t *arg1 = (uint32_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint32_t_PTR_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint32_t_PTR_frompointer" "', argument " "1"" of type '" "uint32_t *""'"); 
  }
  arg1 = (uint32_t *)(argp1);
  result = (uint32_t_PTR *)uint32_t_PTR_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *uint32_t_PTR_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_uint32_t_PTR, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_uint32_t_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_uint32_t_PTR_PTR")) SWIG_fail;
  result = (uint32_t_PTR *)new_uint32_t_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_uint32_t_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_uint32_t_PTR_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_uint32_t_PTR,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_uint32_t_PTR_PTR" "', argument " "1"" of type '" "uint32_t_PTR""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_uint32_t_PTR_PTR" "', argument " "1"" of type '" "uint32_t_PTR""'");
    } else {
      arg1 = *((uint32_t_PTR *)(argp1));
    }
  }
  result = (uint32_t_PTR *)copy_uint32_t_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_uint32_t_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_uint32_t_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_uint32_t_PTR_PTR" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  delete_uint32_t_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint32_t_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  uint32_t_PTR arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:uint32_t_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint32_t_PTR_PTR_assign" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_uint32_t_PTR,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "uint32_t_PTR_PTR_assign" "', argument " "2"" of type '" "uint32_t_PTR""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "uint32_t_PTR_PTR_assign" "', argument " "2"" of type '" "uint32_t_PTR""'");
    } else {
      arg2 = *((uint32_t_PTR *)(argp2));
    }
  }
  uint32_t_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint32_t_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint32_t_PTR *arg1 = (uint32_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint32_t_PTR result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint32_t_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint32_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint32_t_PTR_PTR_value" "', argument " "1"" of type '" "uint32_t_PTR *""'"); 
  }
  arg1 = (uint32_t_PTR *)(argp1);
  result = uint32_t_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((uint32_t_PTR *)memcpy((uint32_t_PTR *)malloc(sizeof(uint32_t_PTR)),&result,sizeof(uint32_t_PTR)), SWIGTYPE_p_uint32_t_PTR, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_UINT32_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_UINT32_PTR")) SWIG_fail;
  result = (UINT32_PTR *)new_UINT32_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT32_PTR, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_UINT32_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_UINT32_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_UINT32_PTR" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  delete_UINT32_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT32_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  UINT32 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:UINT32_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT32_PTR_assign" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "UINT32_PTR_assign" "', argument " "2"" of type '" "UINT32""'");
  } 
  arg2 = (UINT32)(val2);
  UINT32_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT32_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT32_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT32_PTR_value" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  result = (UINT32)UINT32_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT32_PTR_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT32_PTR_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT32_PTR_cast" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  result = (UINT32 *)UINT32_PTR_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT32_PTR_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32 *arg1 = (UINT32 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT32_PTR_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT32_PTR_frompointer" "', argument " "1"" of type '" "UINT32 *""'"); 
  }
  arg1 = (UINT32 *)(argp1);
  result = (UINT32_PTR *)UINT32_PTR_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *UINT32_PTR_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_UINT32_PTR, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_UINT32_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_UINT32_PTR_PTR")) SWIG_fail;
  result = (UINT32_PTR *)new_UINT32_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_UINT32_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_UINT32_PTR_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_UINT32_PTR,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_UINT32_PTR_PTR" "', argument " "1"" of type '" "UINT32_PTR""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_UINT32_PTR_PTR" "', argument " "1"" of type '" "UINT32_PTR""'");
    } else {
      arg1 = *((UINT32_PTR *)(argp1));
    }
  }
  result = (UINT32_PTR *)copy_UINT32_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_UINT32_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_UINT32_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_UINT32_PTR_PTR" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  delete_UINT32_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT32_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  UINT32_PTR arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:UINT32_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT32_PTR_PTR_assign" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_UINT32_PTR,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "UINT32_PTR_PTR_assign" "', argument " "2"" of type '" "UINT32_PTR""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "UINT32_PTR_PTR_assign" "', argument " "2"" of type '" "UINT32_PTR""'");
    } else {
      arg2 = *((UINT32_PTR *)(argp2));
    }
  }
  UINT32_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT32_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32_PTR *arg1 = (UINT32_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT32_PTR result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT32_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT32_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT32_PTR_PTR_value" "', argument " "1"" of type '" "UINT32_PTR *""'"); 
  }
  arg1 = (UINT32_PTR *)(argp1);
  result = UINT32_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((UINT32_PTR *)memcpy((UINT32_PTR *)malloc(sizeof(UINT32_PTR)),&result,sizeof(UINT32_PTR)), SWIGTYPE_p_UINT32_PTR, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_uint16_t_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_uint16_t_PTR")) SWIG_fail;
  result = (uint16_t_PTR *)new_uint16_t_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint16_t_PTR, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_uint16_t_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_uint16_t_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_uint16_t_PTR" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  delete_uint16_t_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint16_t_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:uint16_t_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint16_t_PTR_assign" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "uint16_t_PTR_assign" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = (uint16_t)(val2);
  uint16_t_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint16_t_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint16_t result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint16_t_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint16_t_PTR_value" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  result = (uint16_t)uint16_t_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint16_t_PTR_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint16_t *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint16_t_PTR_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint16_t_PTR_cast" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  result = (uint16_t *)uint16_t_PTR_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_short, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint16_t_PTR_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t *arg1 = (uint16_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint16_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint16_t_PTR_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint16_t_PTR_frompointer" "', argument " "1"" of type '" "uint16_t *""'"); 
  }
  arg1 = (uint16_t *)(argp1);
  result = (uint16_t_PTR *)uint16_t_PTR_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *uint16_t_PTR_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_uint16_t_PTR, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_uint16_t_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_uint16_t_PTR_PTR")) SWIG_fail;
  result = (uint16_t_PTR *)new_uint16_t_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_uint16_t_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint16_t_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_uint16_t_PTR_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_uint16_t_PTR,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_uint16_t_PTR_PTR" "', argument " "1"" of type '" "uint16_t_PTR""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_uint16_t_PTR_PTR" "', argument " "1"" of type '" "uint16_t_PTR""'");
    } else {
      arg1 = *((uint16_t_PTR *)(argp1));
    }
  }
  result = (uint16_t_PTR *)copy_uint16_t_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_uint16_t_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_uint16_t_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_uint16_t_PTR_PTR" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  delete_uint16_t_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint16_t_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  uint16_t_PTR arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:uint16_t_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint16_t_PTR_PTR_assign" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_uint16_t_PTR,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "uint16_t_PTR_PTR_assign" "', argument " "2"" of type '" "uint16_t_PTR""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "uint16_t_PTR_PTR_assign" "', argument " "2"" of type '" "uint16_t_PTR""'");
    } else {
      arg2 = *((uint16_t_PTR *)(argp2));
    }
  }
  uint16_t_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_uint16_t_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint16_t_PTR *arg1 = (uint16_t_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  uint16_t_PTR result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:uint16_t_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_uint16_t_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uint16_t_PTR_PTR_value" "', argument " "1"" of type '" "uint16_t_PTR *""'"); 
  }
  arg1 = (uint16_t_PTR *)(argp1);
  result = uint16_t_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((uint16_t_PTR *)memcpy((uint16_t_PTR *)malloc(sizeof(uint16_t_PTR)),&result,sizeof(uint16_t_PTR)), SWIGTYPE_p_uint16_t_PTR, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_UINT16_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_UINT16_PTR")) SWIG_fail;
  result = (UINT16_PTR *)new_UINT16_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT16_PTR, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_UINT16_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_UINT16_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_UINT16_PTR" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  delete_UINT16_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT16_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  UINT16 arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:UINT16_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT16_PTR_assign" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "UINT16_PTR_assign" "', argument " "2"" of type '" "UINT16""'");
  } 
  arg2 = (UINT16)(val2);
  UINT16_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT16_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT16_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT16_PTR_value" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  result = (UINT16)UINT16_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_short((unsigned short)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT16_PTR_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16 *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT16_PTR_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT16_PTR_cast" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  result = (UINT16 *)UINT16_PTR_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_short, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT16_PTR_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16 *arg1 = (UINT16 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT16_PTR_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT16_PTR_frompointer" "', argument " "1"" of type '" "UINT16 *""'"); 
  }
  arg1 = (UINT16 *)(argp1);
  result = (UINT16_PTR *)UINT16_PTR_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *UINT16_PTR_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_UINT16_PTR, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_UINT16_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_UINT16_PTR_PTR")) SWIG_fail;
  result = (UINT16_PTR *)new_UINT16_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_UINT16_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_UINT16_PTR_PTR",&obj0)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(obj0, &argp1, SWIGTYPE_p_UINT16_PTR,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_UINT16_PTR_PTR" "', argument " "1"" of type '" "UINT16_PTR""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "copy_UINT16_PTR_PTR" "', argument " "1"" of type '" "UINT16_PTR""'");
    } else {
      arg1 = *((UINT16_PTR *)(argp1));
    }
  }
  result = (UINT16_PTR *)copy_UINT16_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_UINT16_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_UINT16_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_UINT16_PTR_PTR" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  delete_UINT16_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT16_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  UINT16_PTR arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:UINT16_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT16_PTR_PTR_assign" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  {
    res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_UINT16_PTR,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "UINT16_PTR_PTR_assign" "', argument " "2"" of type '" "UINT16_PTR""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "UINT16_PTR_PTR_assign" "', argument " "2"" of type '" "UINT16_PTR""'");
    } else {
      arg2 = *((UINT16_PTR *)(argp2));
    }
  }
  UINT16_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_UINT16_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16_PTR *arg1 = (UINT16_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  UINT16_PTR result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:UINT16_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_UINT16_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "UINT16_PTR_PTR_value" "', argument " "1"" of type '" "UINT16_PTR *""'"); 
  }
  arg1 = (UINT16_PTR *)(argp1);
  result = UINT16_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj((UINT16_PTR *)memcpy((UINT16_PTR *)malloc(sizeof(UINT16_PTR)),&result,sizeof(UINT16_PTR)), SWIGTYPE_p_UINT16_PTR, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_ByteArray(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  ByteArray *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_ByteArray",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_ByteArray" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = (size_t)(val1);
  result = (ByteArray *)new_ByteArray(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ByteArray, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ByteArray(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ByteArray *arg1 = (ByteArray *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_ByteArray",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ByteArray, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ByteArray" "', argument " "1"" of type '" "ByteArray *""'"); 
  }
  arg1 = (ByteArray *)(argp1);
  delete_ByteArray(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ByteArray___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ByteArray *arg1 = (ByteArray *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  BYTE result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:ByteArray___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ByteArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ByteArray___getitem__" "', argument " "1"" of type '" "ByteArray *""'"); 
  }
  arg1 = (ByteArray *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ByteArray___getitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  result = (BYTE)ByteArray___getitem__(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_char((unsigned char)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ByteArray___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ByteArray *arg1 = (ByteArray *) 0 ;
  size_t arg2 ;
  BYTE arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  unsigned char val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:ByteArray___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ByteArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ByteArray___setitem__" "', argument " "1"" of type '" "ByteArray *""'"); 
  }
  arg1 = (ByteArray *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ByteArray___setitem__" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_char(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ByteArray___setitem__" "', argument " "3"" of type '" "BYTE""'");
  } 
  arg3 = (BYTE)(val3);
  ByteArray___setitem__(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ByteArray_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ByteArray *arg1 = (ByteArray *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  BYTE *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ByteArray_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ByteArray, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ByteArray_cast" "', argument " "1"" of type '" "ByteArray *""'"); 
  }
  arg1 = (ByteArray *)(argp1);
  result = (BYTE *)ByteArray_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ByteArray_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  BYTE *arg1 = (BYTE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  ByteArray *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ByteArray_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ByteArray_frompointer" "', argument " "1"" of type '" "BYTE *""'"); 
  }
  arg1 = (BYTE *)(argp1);
  result = (ByteArray *)ByteArray_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ByteArray, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *ByteArray_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_ByteArray, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *TSS2_TCTI_TIMEOUT_BLOCK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_TIMEOUT_BLOCK",SWIG_From_int((int)(-1)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_TIMEOUT_NONE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_TIMEOUT_NONE",SWIG_From_int((int)(0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_NONE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_NONE",SWIG_From_unsigned_SS_int((unsigned int)(0xfffU)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PASSWORD_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PASSWORD",SWIG_From_unsigned_SS_int((unsigned int)(0x0ffU)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR0",SWIG_From_unsigned_SS_int((unsigned int)(0U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR1",SWIG_From_unsigned_SS_int((unsigned int)(1U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR2",SWIG_From_unsigned_SS_int((unsigned int)(2U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR3",SWIG_From_unsigned_SS_int((unsigned int)(3U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR4",SWIG_From_unsigned_SS_int((unsigned int)(4U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR5_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR5",SWIG_From_unsigned_SS_int((unsigned int)(5U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR6_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR6",SWIG_From_unsigned_SS_int((unsigned int)(6U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR7_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR7",SWIG_From_unsigned_SS_int((unsigned int)(7U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR8_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR8",SWIG_From_unsigned_SS_int((unsigned int)(8U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR9_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR9",SWIG_From_unsigned_SS_int((unsigned int)(9U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR10_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR10",SWIG_From_unsigned_SS_int((unsigned int)(10U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR11_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR11",SWIG_From_unsigned_SS_int((unsigned int)(11U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR12_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR12",SWIG_From_unsigned_SS_int((unsigned int)(12U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR13_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR13",SWIG_From_unsigned_SS_int((unsigned int)(13U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR14_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR14",SWIG_From_unsigned_SS_int((unsigned int)(14U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR15_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR15",SWIG_From_unsigned_SS_int((unsigned int)(15U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR16_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR16",SWIG_From_unsigned_SS_int((unsigned int)(16U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR17_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR17",SWIG_From_unsigned_SS_int((unsigned int)(17U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR18_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR18",SWIG_From_unsigned_SS_int((unsigned int)(18U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR19_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR19",SWIG_From_unsigned_SS_int((unsigned int)(19U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR20_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR20",SWIG_From_unsigned_SS_int((unsigned int)(20U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR21_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR21",SWIG_From_unsigned_SS_int((unsigned int)(21U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR22_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR22",SWIG_From_unsigned_SS_int((unsigned int)(22U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR23_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR23",SWIG_From_unsigned_SS_int((unsigned int)(23U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR24_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR24",SWIG_From_unsigned_SS_int((unsigned int)(24U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR25_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR25",SWIG_From_unsigned_SS_int((unsigned int)(25U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR26_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR26",SWIG_From_unsigned_SS_int((unsigned int)(26U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR27_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR27",SWIG_From_unsigned_SS_int((unsigned int)(27U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR28_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR28",SWIG_From_unsigned_SS_int((unsigned int)(28U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR29_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR29",SWIG_From_unsigned_SS_int((unsigned int)(29U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR30_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR30",SWIG_From_unsigned_SS_int((unsigned int)(30U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_PCR31_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_PCR31",SWIG_From_unsigned_SS_int((unsigned int)(31U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_OWNER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_OWNER",SWIG_From_unsigned_SS_int((unsigned int)(0x101U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_NULL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_NULL",SWIG_From_unsigned_SS_int((unsigned int)(0x107U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_LOCKOUT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_LOCKOUT",SWIG_From_unsigned_SS_int((unsigned int)(0x10AU)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_ENDORSEMENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_ENDORSEMENT",SWIG_From_unsigned_SS_int((unsigned int)(0x10BU)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_PLATFORM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_PLATFORM",SWIG_From_unsigned_SS_int((unsigned int)(0x10CU)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_PLATFORM_NV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_PLATFORM_NV",SWIG_From_unsigned_SS_int((unsigned int)(0x10DU)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_RH_AUTH_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_RH_AUTH_FIRST",SWIG_From_unsigned_SS_int((unsigned int)(0x110U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *ESYS_TR_MIN_OBJECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "ESYS_TR_MIN_OBJECT",SWIG_From_unsigned_SS_int((unsigned int)(0x1000U)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_CAP_HANDLES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_CAP_HANDLES",SWIG_From_int((int)(((1024-8)/4))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_CAP_ALGS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_CAP_ALGS",SWIG_From_int((int)(((1024-8)/8))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_TPM_PROPERTIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_TPM_PROPERTIES",SWIG_From_int((int)(((1024-8)/8))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_PCR_PROPERTIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_PCR_PROPERTIES",SWIG_From_int((int)(((1024-8)/12))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_ECC_CURVES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_ECC_CURVES",SWIG_From_int((int)(((1024-8)/2))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_TAGGED_POLICIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_TAGGED_POLICIES",SWIG_From_int((int)(((1024-8)/72))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_PTT_PROPERTIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_PTT_PROPERTIES",SWIG_From_int((int)((1024/4))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_MAX_AC_CAPABILITIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_MAX_AC_CAPABILITIES",SWIG_From_int((int)((1024/8))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ERROR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ERROR",SWIG_From_int((int)(0x0000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_RSA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_RSA",SWIG_From_int((int)(0x0001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA",SWIG_From_int((int)(0x0004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA1",SWIG_From_int((int)(0x0004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_HMAC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_HMAC",SWIG_From_int((int)(0x0005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_AES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_AES",SWIG_From_int((int)(0x0006)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_MGF1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_MGF1",SWIG_From_int((int)(0x0007)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_KEYEDHASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_KEYEDHASH",SWIG_From_int((int)(0x0008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_XOR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_XOR",SWIG_From_int((int)(0x000A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA256_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA256",SWIG_From_int((int)(0x000B)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA384_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA384",SWIG_From_int((int)(0x000C)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA512_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA512",SWIG_From_int((int)(0x000D)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_NULL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_NULL",SWIG_From_int((int)(0x0010)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SM3_256_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SM3_256",SWIG_From_int((int)(0x0012)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SM4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SM4",SWIG_From_int((int)(0x0013)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_RSASSA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_RSASSA",SWIG_From_int((int)(0x0014)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_RSAES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_RSAES",SWIG_From_int((int)(0x0015)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_RSAPSS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_RSAPSS",SWIG_From_int((int)(0x0016)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_OAEP_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_OAEP",SWIG_From_int((int)(0x0017)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECDSA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECDSA",SWIG_From_int((int)(0x0018)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECDH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECDH",SWIG_From_int((int)(0x0019)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECDAA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECDAA",SWIG_From_int((int)(0x001A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SM2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SM2",SWIG_From_int((int)(0x001B)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECSCHNORR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECSCHNORR",SWIG_From_int((int)(0x001C)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECMQV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECMQV",SWIG_From_int((int)(0x001D)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_KDF1_SP800_56A_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_KDF1_SP800_56A",SWIG_From_int((int)(0x0020)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_KDF2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_KDF2",SWIG_From_int((int)(0x0021)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_KDF1_SP800_108_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_KDF1_SP800_108",SWIG_From_int((int)(0x0022)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECC",SWIG_From_int((int)(0x0023)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SYMCIPHER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SYMCIPHER",SWIG_From_int((int)(0x0025)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_CAMELLIA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_CAMELLIA",SWIG_From_int((int)(0x0026)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_CMAC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_CMAC",SWIG_From_int((int)(0x003F)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_CTR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_CTR",SWIG_From_int((int)(0x0040)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA3_256_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA3_256",SWIG_From_int((int)(0x0027)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA3_384_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA3_384",SWIG_From_int((int)(0x0028)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_SHA3_512_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_SHA3_512",SWIG_From_int((int)(0x0029)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_OFB_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_OFB",SWIG_From_int((int)(0x0041)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_CBC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_CBC",SWIG_From_int((int)(0x0042)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_CFB_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_CFB",SWIG_From_int((int)(0x0043)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_ECB_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_ECB",SWIG_From_int((int)(0x0044)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_FIRST",SWIG_From_int((int)(0x0001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ALG_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ALG_LAST",SWIG_From_int((int)(0x0044)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_NONE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_NONE",SWIG_From_int((int)(0x0000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_NIST_P192_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_NIST_P192",SWIG_From_int((int)(0x0001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_NIST_P224_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_NIST_P224",SWIG_From_int((int)(0x0002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_NIST_P256_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_NIST_P256",SWIG_From_int((int)(0x0003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_NIST_P384_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_NIST_P384",SWIG_From_int((int)(0x0004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_NIST_P521_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_NIST_P521",SWIG_From_int((int)(0x0005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_BN_P256_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_BN_P256",SWIG_From_int((int)(0x0010)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_BN_P638_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_BN_P638",SWIG_From_int((int)(0x0011)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ECC_SM2_P256_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ECC_SM2_P256",SWIG_From_int((int)(0x0020)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SPEC_FAMILY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SPEC_FAMILY",SWIG_From_int((int)(0x322E3000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SPEC_LEVEL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SPEC_LEVEL",SWIG_From_int((int)(00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SPEC_VERSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SPEC_VERSION",SWIG_From_int((int)(126)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SPEC_YEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SPEC_YEAR",SWIG_From_int((int)(2015)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SPEC_DAY_OF_YEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SPEC_DAY_OF_YEAR",SWIG_From_int((int)(233)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_GENERATED_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_GENERATED_VALUE",SWIG_From_int((int)(0xff544347)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SUCCESS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SUCCESS",SWIG_From_int((int)(0x000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_BAD_TAG_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_BAD_TAG",SWIG_From_int((int)(0x01E)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_VER1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_VER1",SWIG_From_int((int)(0x100)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_INITIALIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_INITIALIZE",SWIG_From_int((int)((0x100+0x000))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_FAILURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_FAILURE",SWIG_From_int((int)((0x100+0x001))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SEQUENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SEQUENCE",SWIG_From_int((int)((0x100+0x003))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_PRIVATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_PRIVATE",SWIG_From_int((int)((0x100+0x00B))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_HMAC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_HMAC",SWIG_From_int((int)((0x100+0x019))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_DISABLED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_DISABLED",SWIG_From_int((int)((0x100+0x020))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_EXCLUSIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_EXCLUSIVE",SWIG_From_int((int)((0x100+0x021))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_AUTH_TYPE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_AUTH_TYPE",SWIG_From_int((int)((0x100+0x024))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_AUTH_MISSING_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_AUTH_MISSING",SWIG_From_int((int)((0x100+0x025))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_POLICY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_POLICY",SWIG_From_int((int)((0x100+0x026))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_PCR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_PCR",SWIG_From_int((int)((0x100+0x027))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_PCR_CHANGED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_PCR_CHANGED",SWIG_From_int((int)((0x100+0x028))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_UPGRADE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_UPGRADE",SWIG_From_int((int)((0x100+0x02D))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_TOO_MANY_CONTEXTS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_TOO_MANY_CONTEXTS",SWIG_From_int((int)((0x100+0x02E))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_AUTH_UNAVAILABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_AUTH_UNAVAILABLE",SWIG_From_int((int)((0x100+0x02F))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REBOOT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REBOOT",SWIG_From_int((int)((0x100+0x030))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_UNBALANCED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_UNBALANCED",SWIG_From_int((int)((0x100+0x031))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_COMMAND_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_COMMAND_SIZE",SWIG_From_int((int)((0x100+0x042))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_COMMAND_CODE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_COMMAND_CODE",SWIG_From_int((int)((0x100+0x043))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_AUTHSIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_AUTHSIZE",SWIG_From_int((int)((0x100+0x044))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_AUTH_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_AUTH_CONTEXT",SWIG_From_int((int)((0x100+0x045))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_RANGE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_RANGE",SWIG_From_int((int)((0x100+0x046))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_SIZE",SWIG_From_int((int)((0x100+0x047))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_LOCKED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_LOCKED",SWIG_From_int((int)((0x100+0x048))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_AUTHORIZATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_AUTHORIZATION",SWIG_From_int((int)((0x100+0x049))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_UNINITIALIZED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_UNINITIALIZED",SWIG_From_int((int)((0x100+0x04A))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_SPACE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_SPACE",SWIG_From_int((int)((0x100+0x04B))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_DEFINED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_DEFINED",SWIG_From_int((int)((0x100+0x04C))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_BAD_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_BAD_CONTEXT",SWIG_From_int((int)((0x100+0x050))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_CPHASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_CPHASH",SWIG_From_int((int)((0x100+0x051))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_PARENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_PARENT",SWIG_From_int((int)((0x100+0x052))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NEEDS_TEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NEEDS_TEST",SWIG_From_int((int)((0x100+0x053))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NO_RESULT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NO_RESULT",SWIG_From_int((int)((0x100+0x054))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SENSITIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SENSITIVE",SWIG_From_int((int)((0x100+0x055))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_MAX_FM0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_MAX_FM0",SWIG_From_int((int)((0x100+0x07F))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_FMT1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_FMT1",SWIG_From_int((int)(0x080)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_ASYMMETRIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_ASYMMETRIC",SWIG_From_int((int)((0x080+0x001))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_ATTRIBUTES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_ATTRIBUTES",SWIG_From_int((int)((0x080+0x002))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_HASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_HASH",SWIG_From_int((int)((0x080+0x003))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_VALUE",SWIG_From_int((int)((0x080+0x004))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_HIERARCHY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_HIERARCHY",SWIG_From_int((int)((0x080+0x005))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_KEY_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_KEY_SIZE",SWIG_From_int((int)((0x080+0x007))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_MGF_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_MGF",SWIG_From_int((int)((0x080+0x008))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_MODE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_MODE",SWIG_From_int((int)((0x080+0x009))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_TYPE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_TYPE",SWIG_From_int((int)((0x080+0x00A))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_HANDLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_HANDLE",SWIG_From_int((int)((0x080+0x00B))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_KDF_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_KDF",SWIG_From_int((int)((0x080+0x00C))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_RANGE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_RANGE",SWIG_From_int((int)((0x080+0x00D))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_AUTH_FAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_AUTH_FAIL",SWIG_From_int((int)((0x080+0x00E))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NONCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NONCE",SWIG_From_int((int)((0x080+0x00F))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_PP_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_PP",SWIG_From_int((int)((0x080+0x010))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SCHEME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SCHEME",SWIG_From_int((int)((0x080+0x012))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SIZE",SWIG_From_int((int)((0x080+0x015))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SYMMETRIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SYMMETRIC",SWIG_From_int((int)((0x080+0x016))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_TAG_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_TAG",SWIG_From_int((int)((0x080+0x017))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SELECTOR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SELECTOR",SWIG_From_int((int)((0x080+0x018))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_INSUFFICIENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_INSUFFICIENT",SWIG_From_int((int)((0x080+0x01A))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SIGNATURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SIGNATURE",SWIG_From_int((int)((0x080+0x01B))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_KEY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_KEY",SWIG_From_int((int)((0x080+0x01C))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_POLICY_FAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_POLICY_FAIL",SWIG_From_int((int)((0x080+0x01D))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_INTEGRITY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_INTEGRITY",SWIG_From_int((int)((0x080+0x01F))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_TICKET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_TICKET",SWIG_From_int((int)((0x080+0x020))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_RESERVED_BITS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_RESERVED_BITS",SWIG_From_int((int)((0x080+0x021))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_BAD_AUTH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_BAD_AUTH",SWIG_From_int((int)((0x080+0x022))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_EXPIRED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_EXPIRED",SWIG_From_int((int)((0x080+0x023))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_POLICY_CC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_POLICY_CC",SWIG_From_int((int)((0x080+0x024))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_BINDING_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_BINDING",SWIG_From_int((int)((0x080+0x025))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_CURVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_CURVE",SWIG_From_int((int)((0x080+0x026))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_ECC_POINT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_ECC_POINT",SWIG_From_int((int)((0x080+0x027))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_WARN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_WARN",SWIG_From_int((int)(0x900)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_CONTEXT_GAP_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_CONTEXT_GAP",SWIG_From_int((int)((0x900+0x001))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_OBJECT_MEMORY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_OBJECT_MEMORY",SWIG_From_int((int)((0x900+0x002))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SESSION_MEMORY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SESSION_MEMORY",SWIG_From_int((int)((0x900+0x003))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_MEMORY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_MEMORY",SWIG_From_int((int)((0x900+0x004))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_SESSION_HANDLES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_SESSION_HANDLES",SWIG_From_int((int)((0x900+0x005))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_OBJECT_HANDLES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_OBJECT_HANDLES",SWIG_From_int((int)((0x900+0x006))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_LOCALITY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_LOCALITY",SWIG_From_int((int)((0x900+0x007))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_YIELDED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_YIELDED",SWIG_From_int((int)((0x900+0x008))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_CANCELED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_CANCELED",SWIG_From_int((int)((0x900+0x009))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_TESTING_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_TESTING",SWIG_From_int((int)((0x900+0x00A))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H0",SWIG_From_int((int)((0x900+0x010))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H1",SWIG_From_int((int)((0x900+0x011))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H2",SWIG_From_int((int)((0x900+0x012))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H3",SWIG_From_int((int)((0x900+0x013))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H4",SWIG_From_int((int)((0x900+0x014))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H5_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H5",SWIG_From_int((int)((0x900+0x015))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_H6_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_H6",SWIG_From_int((int)((0x900+0x016))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S0",SWIG_From_int((int)((0x900+0x018))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S1",SWIG_From_int((int)((0x900+0x019))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S2",SWIG_From_int((int)((0x900+0x01A))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S3",SWIG_From_int((int)((0x900+0x01B))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S4",SWIG_From_int((int)((0x900+0x01C))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S5_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S5",SWIG_From_int((int)((0x900+0x01D))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_REFERENCE_S6_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_REFERENCE_S6",SWIG_From_int((int)((0x900+0x01E))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_RATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_RATE",SWIG_From_int((int)((0x900+0x020))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_LOCKOUT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_LOCKOUT",SWIG_From_int((int)((0x900+0x021))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_RETRY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_RETRY",SWIG_From_int((int)((0x900+0x022))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NV_UNAVAILABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NV_UNAVAILABLE",SWIG_From_int((int)((0x900+0x023))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_NOT_USED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_NOT_USED",SWIG_From_int((int)((0x900+0x07F))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_H_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_H",SWIG_From_int((int)(0x000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_P_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_P",SWIG_From_int((int)(0x040)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_S_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_S",SWIG_From_int((int)(0x800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_1",SWIG_From_int((int)(0x100)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_2",SWIG_From_int((int)(0x200)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_3",SWIG_From_int((int)(0x300)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_4",SWIG_From_int((int)(0x400)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_5_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_5",SWIG_From_int((int)(0x500)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_6_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_6",SWIG_From_int((int)(0x600)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_7_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_7",SWIG_From_int((int)(0x700)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_8_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_8",SWIG_From_int((int)(0x800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_9_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_9",SWIG_From_int((int)(0x900)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_A_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_A",SWIG_From_int((int)(0xA00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_B_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_B",SWIG_From_int((int)(0xB00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_C_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_C",SWIG_From_int((int)(0xC00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_D_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_D",SWIG_From_int((int)(0xD00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_E_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_E",SWIG_From_int((int)(0xE00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_F_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_F",SWIG_From_int((int)(0xF00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RC_N_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RC_N_MASK",SWIG_From_int((int)(0xF00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_COARSE_SLOWER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_COARSE_SLOWER",SWIG_From_int((int)(-3)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_MEDIUM_SLOWER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_MEDIUM_SLOWER",SWIG_From_int((int)(-2)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_FINE_SLOWER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_FINE_SLOWER",SWIG_From_int((int)(-1)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_NO_CHANGE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_NO_CHANGE",SWIG_From_int((int)(0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_FINE_FASTER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_FINE_FASTER",SWIG_From_int((int)(1)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_MEDIUM_FASTER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_MEDIUM_FASTER",SWIG_From_int((int)(2)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CLOCK_COARSE_FASTER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CLOCK_COARSE_FASTER",SWIG_From_int((int)(3)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_EQ_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_EQ",SWIG_From_int((int)(0x0000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_NEQ_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_NEQ",SWIG_From_int((int)(0x0001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_SIGNED_GT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_SIGNED_GT",SWIG_From_int((int)(0x0002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_UNSIGNED_GT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_UNSIGNED_GT",SWIG_From_int((int)(0x0003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_SIGNED_LT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_SIGNED_LT",SWIG_From_int((int)(0x0004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_UNSIGNED_LT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_UNSIGNED_LT",SWIG_From_int((int)(0x0005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_SIGNED_GE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_SIGNED_GE",SWIG_From_int((int)(0x0006)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_UNSIGNED_GE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_UNSIGNED_GE",SWIG_From_int((int)(0x0007)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_SIGNED_LE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_SIGNED_LE",SWIG_From_int((int)(0x0008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_UNSIGNED_LE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_UNSIGNED_LE",SWIG_From_int((int)(0x0009)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_BITSET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_BITSET",SWIG_From_int((int)(0x000A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_EO_BITCLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_EO_BITCLEAR",SWIG_From_int((int)(0x000B)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_RSP_COMMAND_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_RSP_COMMAND",SWIG_From_int((int)(0x00C4)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_NULL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_NULL",SWIG_From_int((int)(0X8000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_NO_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_NO_SESSIONS",SWIG_From_int((int)(0x8001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_SESSIONS",SWIG_From_int((int)(0x8002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_RESERVED1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_RESERVED1",SWIG_From_int((int)(0x8003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_RESERVED2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_RESERVED2",SWIG_From_int((int)(0x8004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_NV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_NV",SWIG_From_int((int)(0x8014)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_COMMAND_AUDIT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_COMMAND_AUDIT",SWIG_From_int((int)(0x8015)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_SESSION_AUDIT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_SESSION_AUDIT",SWIG_From_int((int)(0x8016)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_CERTIFY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_CERTIFY",SWIG_From_int((int)(0x8017)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_QUOTE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_QUOTE",SWIG_From_int((int)(0x8018)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_TIME_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_TIME",SWIG_From_int((int)(0x8019)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_ATTEST_CREATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_ATTEST_CREATION",SWIG_From_int((int)(0x801A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_RESERVED3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_RESERVED3",SWIG_From_int((int)(0x801B)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_CREATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_CREATION",SWIG_From_int((int)(0x8021)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_VERIFIED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_VERIFIED",SWIG_From_int((int)(0x8022)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_AUTH_SECRET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_AUTH_SECRET",SWIG_From_int((int)(0x8023)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_HASHCHECK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_HASHCHECK",SWIG_From_int((int)(0x8024)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_AUTH_SIGNED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_AUTH_SIGNED",SWIG_From_int((int)(0x8025)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ST_FU_MANIFEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ST_FU_MANIFEST",SWIG_From_int((int)(0x8029)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SU_CLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SU_CLEAR",SWIG_From_int((int)(0x0000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SU_STATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SU_STATE",SWIG_From_int((int)(0x0001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SE_HMAC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SE_HMAC",SWIG_From_int((int)(0x00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SE_POLICY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SE_POLICY",SWIG_From_int((int)(0x01)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_SE_TRIAL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_SE_TRIAL",SWIG_From_int((int)(0x03)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_FIRST",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_ALGS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_ALGS",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_HANDLES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_HANDLES",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_COMMANDS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_COMMANDS",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_PP_COMMANDS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_PP_COMMANDS",SWIG_From_int((int)(0x00000003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_AUDIT_COMMANDS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_AUDIT_COMMANDS",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_PCRS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_PCRS",SWIG_From_int((int)(0x00000005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_TPM_PROPERTIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_TPM_PROPERTIES",SWIG_From_int((int)(0x00000006)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_PCR_PROPERTIES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_PCR_PROPERTIES",SWIG_From_int((int)(0x00000007)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_ECC_CURVES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_ECC_CURVES",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_LAST",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_CAP_VENDOR_PROPERTY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_CAP_VENDOR_PROPERTY",SWIG_From_int((int)(0x00000100)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NONE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NONE",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_GROUP_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_GROUP",SWIG_From_int((int)(0x00000100)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_FIXED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_FIXED",SWIG_From_int((int)((0x00000100*1))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_FAMILY_INDICATOR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_FAMILY_INDICATOR",SWIG_From_int((int)(((0x00000100*1) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_LEVEL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_LEVEL",SWIG_From_int((int)(((0x00000100*1) +1))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_REVISION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_REVISION",SWIG_From_int((int)(((0x00000100*1) +2))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_DAY_OF_YEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_DAY_OF_YEAR",SWIG_From_int((int)(((0x00000100*1) +3))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_YEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_YEAR",SWIG_From_int((int)(((0x00000100*1) +4))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MANUFACTURER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MANUFACTURER",SWIG_From_int((int)(((0x00000100*1) +5))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VENDOR_STRING_1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VENDOR_STRING_1",SWIG_From_int((int)(((0x00000100*1) +6))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VENDOR_STRING_2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VENDOR_STRING_2",SWIG_From_int((int)(((0x00000100*1) +7))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VENDOR_STRING_3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VENDOR_STRING_3",SWIG_From_int((int)(((0x00000100*1) +8))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VENDOR_STRING_4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VENDOR_STRING_4",SWIG_From_int((int)(((0x00000100*1) +9))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VENDOR_TPM_TYPE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VENDOR_TPM_TYPE",SWIG_From_int((int)(((0x00000100*1) +10))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_FIRMWARE_VERSION_1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_FIRMWARE_VERSION_1",SWIG_From_int((int)(((0x00000100*1) +11))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_FIRMWARE_VERSION_2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_FIRMWARE_VERSION_2",SWIG_From_int((int)(((0x00000100*1) +12))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_INPUT_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_INPUT_BUFFER",SWIG_From_int((int)(((0x00000100*1) +13))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_HR_TRANSIENT_MIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_HR_TRANSIENT_MIN",SWIG_From_int((int)(((0x00000100*1) +14))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_HR_PERSISTENT_MIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_HR_PERSISTENT_MIN",SWIG_From_int((int)(((0x00000100*1) +15))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_HR_LOADED_MIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_HR_LOADED_MIN",SWIG_From_int((int)(((0x00000100*1) +16))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_ACTIVE_SESSIONS_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_ACTIVE_SESSIONS_MAX",SWIG_From_int((int)(((0x00000100*1) +17))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_COUNT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_COUNT",SWIG_From_int((int)(((0x00000100*1) +18))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_SELECT_MIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_SELECT_MIN",SWIG_From_int((int)(((0x00000100*1) +19))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_CONTEXT_GAP_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_CONTEXT_GAP_MAX",SWIG_From_int((int)(((0x00000100*1) +20))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NV_COUNTERS_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NV_COUNTERS_MAX",SWIG_From_int((int)(((0x00000100*1) +22))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NV_INDEX_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NV_INDEX_MAX",SWIG_From_int((int)(((0x00000100*1) +23))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MEMORY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MEMORY",SWIG_From_int((int)(((0x00000100*1) +24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_CLOCK_UPDATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_CLOCK_UPDATE",SWIG_From_int((int)(((0x00000100*1) +25))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_CONTEXT_HASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_CONTEXT_HASH",SWIG_From_int((int)(((0x00000100*1) +26))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_CONTEXT_SYM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_CONTEXT_SYM",SWIG_From_int((int)(((0x00000100*1) +27))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_CONTEXT_SYM_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_CONTEXT_SYM_SIZE",SWIG_From_int((int)(((0x00000100*1) +28))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_ORDERLY_COUNT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_ORDERLY_COUNT",SWIG_From_int((int)(((0x00000100*1) +29))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MAX_COMMAND_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MAX_COMMAND_SIZE",SWIG_From_int((int)(((0x00000100*1) +30))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MAX_RESPONSE_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MAX_RESPONSE_SIZE",SWIG_From_int((int)(((0x00000100*1) +31))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MAX_DIGEST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MAX_DIGEST",SWIG_From_int((int)(((0x00000100*1) +32))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MAX_OBJECT_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MAX_OBJECT_CONTEXT",SWIG_From_int((int)(((0x00000100*1) +33))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MAX_SESSION_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MAX_SESSION_CONTEXT",SWIG_From_int((int)(((0x00000100*1) +34))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PS_FAMILY_INDICATOR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PS_FAMILY_INDICATOR",SWIG_From_int((int)(((0x00000100*1) +35))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PS_LEVEL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PS_LEVEL",SWIG_From_int((int)(((0x00000100*1) +36))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PS_REVISION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PS_REVISION",SWIG_From_int((int)(((0x00000100*1) +37))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PS_DAY_OF_YEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PS_DAY_OF_YEAR",SWIG_From_int((int)(((0x00000100*1) +38))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PS_YEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PS_YEAR",SWIG_From_int((int)(((0x00000100*1) +39))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_SPLIT_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_SPLIT_MAX",SWIG_From_int((int)(((0x00000100*1) +40))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TOTAL_COMMANDS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TOTAL_COMMANDS",SWIG_From_int((int)(((0x00000100*1) +41))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_LIBRARY_COMMANDS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_LIBRARY_COMMANDS",SWIG_From_int((int)(((0x00000100*1) +42))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VENDOR_COMMANDS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VENDOR_COMMANDS",SWIG_From_int((int)(((0x00000100*1) +43))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NV_BUFFER_MAX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NV_BUFFER_MAX",SWIG_From_int((int)(((0x00000100*1) +44))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MODES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MODES",SWIG_From_int((int)(((0x00000100*1) +45))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_VAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_VAR",SWIG_From_int((int)((0x00000100*2))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PERMANENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PERMANENT",SWIG_From_int((int)(((0x00000100*2) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_STARTUP_CLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_STARTUP_CLEAR",SWIG_From_int((int)(((0x00000100*2) +1))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_HR_NV_INDEX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_HR_NV_INDEX",SWIG_From_int((int)(((0x00000100*2) +2))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_HR_LOADED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_HR_LOADED",SWIG_From_int((int)(((0x00000100*2) +3))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_HR_LOADED_AVAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_HR_LOADED_AVAIL",SWIG_From_int((int)(((0x00000100*2) +4))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_HR_ACTIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_HR_ACTIVE",SWIG_From_int((int)(((0x00000100*2) +5))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_HR_ACTIVE_AVAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_HR_ACTIVE_AVAIL",SWIG_From_int((int)(((0x00000100*2) +6))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_HR_TRANSIENT_AVAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_HR_TRANSIENT_AVAIL",SWIG_From_int((int)(((0x00000100*2) +7))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_HR_PERSISTENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_HR_PERSISTENT",SWIG_From_int((int)(((0x00000100*2) +8))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_HR_PERSISTENT_AVAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_HR_PERSISTENT_AVAIL",SWIG_From_int((int)(((0x00000100*2) +9))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NV_COUNTERS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NV_COUNTERS",SWIG_From_int((int)(((0x00000100*2) +10))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NV_COUNTERS_AVAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NV_COUNTERS_AVAIL",SWIG_From_int((int)(((0x00000100*2) +11))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_ALGORITHM_SET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_ALGORITHM_SET",SWIG_From_int((int)(((0x00000100*2) +12))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_LOADED_CURVES_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_LOADED_CURVES",SWIG_From_int((int)(((0x00000100*2) +13))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_LOCKOUT_COUNTER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_LOCKOUT_COUNTER",SWIG_From_int((int)(((0x00000100*2) +14))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_MAX_AUTH_FAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_MAX_AUTH_FAIL",SWIG_From_int((int)(((0x00000100*2) +15))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_LOCKOUT_INTERVAL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_LOCKOUT_INTERVAL",SWIG_From_int((int)(((0x00000100*2) +16))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_LOCKOUT_RECOVERY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_LOCKOUT_RECOVERY",SWIG_From_int((int)(((0x00000100*2) +17))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_NV_WRITE_RECOVERY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_NV_WRITE_RECOVERY",SWIG_From_int((int)(((0x00000100*2) +18))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_AUDIT_COUNTER_0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_AUDIT_COUNTER_0",SWIG_From_int((int)(((0x00000100*2) +19))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_AUDIT_COUNTER_1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_AUDIT_COUNTER_1",SWIG_From_int((int)(((0x00000100*2) +20))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_PCR_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_PCR_FIRST",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_SAVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_SAVE",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_EXTEND_L0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_EXTEND_L0",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_RESET_L0_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_RESET_L0",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_EXTEND_L1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_EXTEND_L1",SWIG_From_int((int)(0x00000003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_RESET_L1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_RESET_L1",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_EXTEND_L2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_EXTEND_L2",SWIG_From_int((int)(0x00000005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_RESET_L2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_RESET_L2",SWIG_From_int((int)(0x00000006)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_EXTEND_L3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_EXTEND_L3",SWIG_From_int((int)(0x00000007)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_RESET_L3_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_RESET_L3",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_EXTEND_L4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_EXTEND_L4",SWIG_From_int((int)(0x00000009)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_RESET_L4_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_RESET_L4",SWIG_From_int((int)(0x0000000A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_NO_INCREMENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_NO_INCREMENT",SWIG_From_int((int)(0x00000011)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_DRTM_RESET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_DRTM_RESET",SWIG_From_int((int)(0x00000012)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_POLICY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_POLICY",SWIG_From_int((int)(0x00000013)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_PCR_AUTH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_PCR_AUTH",SWIG_From_int((int)(0x00000014)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PT_TPM2_PCR_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PT_TPM2_PCR_LAST",SWIG_From_int((int)(0x00000014)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_MAIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_MAIN",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_PC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_PC",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_PDA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_PDA",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_CELL_PHONE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_CELL_PHONE",SWIG_From_int((int)(0x00000003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_SERVER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_SERVER",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_PERIPHERAL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_PERIPHERAL",SWIG_From_int((int)(0x00000005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_TSS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_TSS",SWIG_From_int((int)(0x00000006)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_STORAGE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_STORAGE",SWIG_From_int((int)(0x00000007)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_AUTHENTICATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_AUTHENTICATION",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_EMBEDDED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_EMBEDDED",SWIG_From_int((int)(0x00000009)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_HARDCOPY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_HARDCOPY",SWIG_From_int((int)(0x0000000A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_INFRASTRUCTURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_INFRASTRUCTURE",SWIG_From_int((int)(0x0000000B)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_VIRTUALIZATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_VIRTUALIZATION",SWIG_From_int((int)(0x0000000C)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_TNC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_TNC",SWIG_From_int((int)(0x0000000D)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_MULTI_TENANT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_MULTI_TENANT",SWIG_From_int((int)(0x0000000E)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PS_TC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PS_TC",SWIG_From_int((int)(0x0000000F)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_PCR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_PCR",SWIG_From_int((int)(0x00)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_NV_INDEX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_NV_INDEX",SWIG_From_int((int)(0x01)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_HMAC_SESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_HMAC_SESSION",SWIG_From_int((int)(0x02)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_LOADED_SESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_LOADED_SESSION",SWIG_From_int((int)(0x02)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_POLICY_SESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_POLICY_SESSION",SWIG_From_int((int)(0x03)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_SAVED_SESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_SAVED_SESSION",SWIG_From_int((int)(0x03)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_PERMANENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_PERMANENT",SWIG_From_int((int)(0x40)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_TRANSIENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_TRANSIENT",SWIG_From_int((int)(0x80)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HT_PERSISTENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HT_PERSISTENT",SWIG_From_int((int)(0x81)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_FIRST",SWIG_From_int((int)(0x40000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_SRK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_SRK",SWIG_From_int((int)(0x40000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_OWNER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_OWNER",SWIG_From_int((int)(0x40000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_REVOKE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_REVOKE",SWIG_From_int((int)(0x40000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_TRANSPORT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_TRANSPORT",SWIG_From_int((int)(0x40000003)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_OPERATOR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_OPERATOR",SWIG_From_int((int)(0x40000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_ADMIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_ADMIN",SWIG_From_int((int)(0x40000005)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_EK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_EK",SWIG_From_int((int)(0x40000006)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_NULL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_NULL",SWIG_From_int((int)(0x40000007)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_UNASSIGNED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_UNASSIGNED",SWIG_From_int((int)(0x40000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RS_PW_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RS_PW",SWIG_From_int((int)(0x40000009)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_LOCKOUT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_LOCKOUT",SWIG_From_int((int)(0x4000000A)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_ENDORSEMENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_ENDORSEMENT",SWIG_From_int((int)(0x4000000B)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_PLATFORM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_PLATFORM",SWIG_From_int((int)(0x4000000C)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_PLATFORM_NV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_PLATFORM_NV",SWIG_From_int((int)(0x4000000D)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_AUTH_00_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_AUTH_00",SWIG_From_int((int)(0x40000010)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_AUTH_FF_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_AUTH_FF",SWIG_From_int((int)(0x4000010F)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_RH_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_RH_LAST",SWIG_From_int((int)(0x4000010F)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_HANDLE_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_HANDLE_MASK",SWIG_From_int((int)(0x00FFFFFF)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_RANGE_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_RANGE_MASK",SWIG_From_int((int)(0xFF000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_SHIFT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_SHIFT",SWIG_From_int((int)(24)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_PCR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_PCR",SWIG_From_int((int)((0x00 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_HMAC_SESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_HMAC_SESSION",SWIG_From_int((int)((0x02 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_POLICY_SESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_POLICY_SESSION",SWIG_From_int((int)((0x03 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_TRANSIENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_TRANSIENT",SWIG_From_int((int)((0x80 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_PERSISTENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_PERSISTENT",SWIG_From_int((int)((0x81 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_NV_INDEX_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_NV_INDEX",SWIG_From_int((int)((0x01 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HR_PERMANENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HR_PERMANENT",SWIG_From_int((int)((0x40 << 24))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PCR_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PCR_FIRST",SWIG_From_int((int)(((0x00 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PCR_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PCR_LAST",SWIG_From_int((int)((((0x00 << 24) +0) +32-1))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HMAC_SESSION_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HMAC_SESSION_FIRST",SWIG_From_int((int)(((0x02 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_HMAC_SESSION_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_HMAC_SESSION_LAST",SWIG_From_int((int)((((0x02 << 24) +0) +0x00fffffe))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_LOADED_SESSION_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_LOADED_SESSION_FIRST",SWIG_From_int((int)(((0x02 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_LOADED_SESSION_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_LOADED_SESSION_LAST",SWIG_From_int((int)((((0x02 << 24) +0) +0x00fffffe))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_POLICY_SESSION_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_POLICY_SESSION_FIRST",SWIG_From_int((int)(((0x03 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_POLICY_SESSION_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_POLICY_SESSION_LAST",SWIG_From_int((int)((((0x03 << 24) +0) +0x00fffffe))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_TRANSIENT_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_TRANSIENT_FIRST",SWIG_From_int((int)(((0x80 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ACTIVE_SESSION_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ACTIVE_SESSION_FIRST",SWIG_From_int((int)(((0x03 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_ACTIVE_SESSION_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_ACTIVE_SESSION_LAST",SWIG_From_int((int)((((0x03 << 24) +0) +0x00fffffe))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_TRANSIENT_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_TRANSIENT_LAST",SWIG_From_int((int)((((0x80 << 24) +0) +0x00fffffe))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PERSISTENT_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PERSISTENT_FIRST",SWIG_From_int((int)(((0x81 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PERSISTENT_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PERSISTENT_LAST",SWIG_From_int((int)((((0x81 << 24) +0) +0x00FFFFFF))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PLATFORM_PERSISTENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PLATFORM_PERSISTENT",SWIG_From_int((int)((((0x81 << 24) +0) +0x00800000))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NV_INDEX_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NV_INDEX_FIRST",SWIG_From_int((int)(((0x01 << 24) +0))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NV_INDEX_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NV_INDEX_LAST",SWIG_From_int((int)((((0x01 << 24) +0) +0x00FFFFFF))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PERMANENT_FIRST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PERMANENT_FIRST",SWIG_From_int((int)(0x40000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_PERMANENT_LAST_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_PERMANENT_LAST",SWIG_From_int((int)(0x4000010F)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_ASYMMETRIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_ASYMMETRIC",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_SYMMETRIC_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_SYMMETRIC",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_HASH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_HASH",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_OBJECT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_OBJECT",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_RESERVED1_MASK",SWIG_From_int((int)(0x000000F0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_SIGNING_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_SIGNING",SWIG_From_int((int)(0x00000100)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_ENCRYPTING_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_ENCRYPTING",SWIG_From_int((int)(0x00000200)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_METHOD_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_METHOD",SWIG_From_int((int)(0x00000400)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_ALGORITHM_RESERVED2_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_ALGORITHM_RESERVED2_MASK",SWIG_From_int((int)(0xFFFFF800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_RESERVED1_MASK",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_FIXEDTPM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_FIXEDTPM",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_STCLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_STCLEAR",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_RESERVED2_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_RESERVED2_MASK",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_FIXEDPARENT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_FIXEDPARENT",SWIG_From_int((int)(0x00000010)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_SENSITIVEDATAORIGIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_SENSITIVEDATAORIGIN",SWIG_From_int((int)(0x00000020)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_USERWITHAUTH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_USERWITHAUTH",SWIG_From_int((int)(0x00000040)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_ADMINWITHPOLICY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_ADMINWITHPOLICY",SWIG_From_int((int)(0x00000080)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_RESERVED3_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_RESERVED3_MASK",SWIG_From_int((int)(0x00000300)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_NODA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_NODA",SWIG_From_int((int)(0x00000400)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_ENCRYPTEDDUPLICATION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_ENCRYPTEDDUPLICATION",SWIG_From_int((int)(0x00000800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_RESERVED4_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_RESERVED4_MASK",SWIG_From_int((int)(0x0000F000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_RESTRICTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_RESTRICTED",SWIG_From_int((int)(0x00010000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_DECRYPT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_DECRYPT",SWIG_From_int((int)(0x00020000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_SIGN_ENCRYPT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_SIGN_ENCRYPT",SWIG_From_int((int)(0x00040000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_OBJECT_RESERVED5_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_OBJECT_RESERVED5_MASK",SWIG_From_int((int)(0xFFF80000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_CONTINUESESSION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_CONTINUESESSION",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_AUDITEXCLUSIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_AUDITEXCLUSIVE",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_AUDITRESET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_AUDITRESET",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_RESERVED1_MASK",SWIG_From_int((int)(0x00000018)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_DECRYPT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_DECRYPT",SWIG_From_int((int)(0x00000020)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_ENCRYPT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_ENCRYPT",SWIG_From_int((int)(0x00000040)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_SESSION_AUDIT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_SESSION_AUDIT",SWIG_From_int((int)(0x00000080)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_TPM2_LOC_ZERO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_TPM2_LOC_ZERO",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_TPM2_LOC_ONE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_TPM2_LOC_ONE",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_TPM2_LOC_TWO_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_TPM2_LOC_TWO",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_TPM2_LOC_THREE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_TPM2_LOC_THREE",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_TPM2_LOC_FOUR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_TPM2_LOC_FOUR",SWIG_From_int((int)(0x00000010)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_LOCALITY_EXTENDED_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_LOCALITY_EXTENDED_MASK",SWIG_From_int((int)(0x000000E0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_OWNERAUTHSET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_OWNERAUTHSET",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_ENDORSEMENTAUTHSET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_ENDORSEMENTAUTHSET",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_LOCKOUTAUTHSET_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_LOCKOUTAUTHSET",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_RESERVED1_MASK",SWIG_From_int((int)(0x000000F8)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_DISABLECLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_DISABLECLEAR",SWIG_From_int((int)(0x00000100)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_INLOCKOUT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_INLOCKOUT",SWIG_From_int((int)(0x00000200)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_TPMGENERATEDEPS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_TPMGENERATEDEPS",SWIG_From_int((int)(0x00000400)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_PERMANENT_RESERVED2_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_PERMANENT_RESERVED2_MASK",SWIG_From_int((int)(0xFFFFF800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_STARTUP_CLEAR_PHENABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_STARTUP_CLEAR_PHENABLE",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_STARTUP_CLEAR_SHENABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_STARTUP_CLEAR_SHENABLE",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_STARTUP_CLEAR_EHENABLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_STARTUP_CLEAR_EHENABLE",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_STARTUP_CLEAR_PHENABLENV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_STARTUP_CLEAR_PHENABLENV",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_STARTUP_CLEAR_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_STARTUP_CLEAR_RESERVED1_MASK",SWIG_From_int((int)(0x7FFFFFF0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_STARTUP_CLEAR_ORDERLY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_STARTUP_CLEAR_ORDERLY",SWIG_From_int((int)(0x80000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_MEMORY_SHAREDRAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_MEMORY_SHAREDRAM",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_MEMORY_SHAREDNV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_MEMORY_SHAREDNV",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_MEMORY_OBJECTCOPIEDTORAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_MEMORY_OBJECTCOPIEDTORAM",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_MEMORY_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_MEMORY_RESERVED1_MASK",SWIG_From_int((int)(0xFFFFFFF8)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_COMMANDINDEX_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_COMMANDINDEX_MASK",SWIG_From_int((int)(0x0000FFFF)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_RESERVED1_MASK",SWIG_From_int((int)(0x003F0000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_NV_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_NV",SWIG_From_int((int)(0x00400000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_EXTENSIVE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_EXTENSIVE",SWIG_From_int((int)(0x00800000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_FLUSHED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_FLUSHED",SWIG_From_int((int)(0x01000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_CHANDLES_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_CHANDLES_MASK",SWIG_From_int((int)(0x0E000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_RHANDLE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_RHANDLE",SWIG_From_int((int)(0x10000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_V_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_V",SWIG_From_int((int)(0x20000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_CC_RES_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_CC_RES_MASK",SWIG_From_int((int)(0xC0000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_MODES_FIPS_140_2_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_MODES_FIPS_140_2",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_MODES_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_MODES_RESERVED1_MASK",SWIG_From_int((int)(0xFFFFFFFE)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NV_INDEX_INDEX_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NV_INDEX_INDEX_MASK",SWIG_From_int((int)(0x00FFFFFF)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NV_INDEX_RH_NV_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NV_INDEX_RH_NV_MASK",SWIG_From_int((int)(0xFF000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NT_ORDINARY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NT_ORDINARY",SWIG_From_int((int)(0x0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NT_COUNTER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NT_COUNTER",SWIG_From_int((int)(0x1)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NT_BITS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NT_BITS",SWIG_From_int((int)(0x2)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NT_EXTEND_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NT_EXTEND",SWIG_From_int((int)(0x4)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NT_PIN_FAIL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NT_PIN_FAIL",SWIG_From_int((int)(0x8)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM2_NT_PIN_PASS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM2_NT_PIN_PASS",SWIG_From_int((int)(0x9)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_PPWRITE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_PPWRITE",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_OWNERWRITE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_OWNERWRITE",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_AUTHWRITE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_AUTHWRITE",SWIG_From_int((int)(0x00000004)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_POLICYWRITE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_POLICYWRITE",SWIG_From_int((int)(0x00000008)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_TPM2_NT_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_TPM2_NT_MASK",SWIG_From_int((int)(0x000000F0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_RESERVED1_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_RESERVED1_MASK",SWIG_From_int((int)(0x00000300)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_POLICY_DELETE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_POLICY_DELETE",SWIG_From_int((int)(0x00000400)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_WRITELOCKED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_WRITELOCKED",SWIG_From_int((int)(0x00000800)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_WRITEALL_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_WRITEALL",SWIG_From_int((int)(0x00001000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_WRITEDEFINE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_WRITEDEFINE",SWIG_From_int((int)(0x00002000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_WRITE_STCLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_WRITE_STCLEAR",SWIG_From_int((int)(0x00004000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_GLOBALLOCK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_GLOBALLOCK",SWIG_From_int((int)(0x00008000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_PPREAD_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_PPREAD",SWIG_From_int((int)(0x00010000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_OWNERREAD_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_OWNERREAD",SWIG_From_int((int)(0x00020000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_AUTHREAD_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_AUTHREAD",SWIG_From_int((int)(0x00040000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_POLICYREAD_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_POLICYREAD",SWIG_From_int((int)(0x00080000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_RESERVED2_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_RESERVED2_MASK",SWIG_From_int((int)(0x01F00000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_NO_DA_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_NO_DA",SWIG_From_int((int)(0x02000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_ORDERLY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_ORDERLY",SWIG_From_int((int)(0x04000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_CLEAR_STCLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_CLEAR_STCLEAR",SWIG_From_int((int)(0x08000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_READLOCKED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_READLOCKED",SWIG_From_int((int)(0x10000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_WRITTEN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_WRITTEN",SWIG_From_int((int)(0x20000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_PLATFORMCREATE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_PLATFORMCREATE",SWIG_From_int((int)(0x40000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPMA_NV_READ_STCLEAR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPMA_NV_READ_STCLEAR",SWIG_From_int((int)(0x80000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM_AT_ANY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM_AT_ANY",SWIG_From_int((int)(0x00000000)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM_AT_ERROR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM_AT_ERROR",SWIG_From_int((int)(0x00000001)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TPM_AT_PV1_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TPM_AT_PV1",SWIG_From_int((int)(0x00000002)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_RC_SUCCESS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_RC_SUCCESS",SWIG_From_int((int)(0)));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_RC_LAYER_MASK_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_RC_LAYER_MASK",SWIG_From_int((int)((0xff << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TPM_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TPM_RC_LAYER",SWIG_From_int((int)((0 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_FEATURE_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_FEATURE_RC_LAYER",SWIG_From_int((int)((6 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESAPI_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESAPI_RC_LAYER",SWIG_From_int((int)((7 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_LAYER",SWIG_From_int((int)((8 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_MU_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_MU_RC_LAYER",SWIG_From_int((int)((9 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_LAYER",SWIG_From_int((int)((10 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_RESMGR_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_RESMGR_RC_LAYER",SWIG_From_int((int)((11 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_RESMGR_TPM_RC_LAYER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_RESMGR_TPM_RC_LAYER",SWIG_From_int((int)((12 << (16)))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_GENERAL_FAILURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_GENERAL_FAILURE",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|1U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_NOT_IMPLEMENTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_NOT_IMPLEMENTED",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|2U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_BAD_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_BAD_CONTEXT",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|3U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_ABI_MISMATCH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_ABI_MISMATCH",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|4U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_BAD_REFERENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_BAD_REFERENCE",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|5U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_INSUFFICIENT_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_INSUFFICIENT_BUFFER",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|6U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_BAD_SEQUENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_BAD_SEQUENCE",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|7U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_NO_CONNECTION_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_NO_CONNECTION",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|8U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_IO_ERROR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_IO_ERROR",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|10U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_BAD_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_BAD_VALUE",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|11U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_NOT_PERMITTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_NOT_PERMITTED",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|12U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_MALFORMED_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_MALFORMED_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|17U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_TCTI_RC_NOT_SUPPORTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_TCTI_RC_NOT_SUPPORTED",SWIG_From_unsigned_SS_int((unsigned int)(((10 << (16))|21U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_GENERAL_FAILURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_GENERAL_FAILURE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|1U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_ABI_MISMATCH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_ABI_MISMATCH",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|4U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_BAD_REFERENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_BAD_REFERENCE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|5U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_INSUFFICIENT_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_INSUFFICIENT_BUFFER",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|6U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_BAD_SEQUENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_BAD_SEQUENCE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|7U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_BAD_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_BAD_VALUE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|11U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_INVALID_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_INVALID_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|13U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_NO_DECRYPT_PARAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_NO_DECRYPT_PARAM",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|14U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_NO_ENCRYPT_PARAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_NO_ENCRYPT_PARAM",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|15U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_BAD_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_BAD_SIZE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|16U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_MALFORMED_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_MALFORMED_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|17U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_INSUFFICIENT_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_INSUFFICIENT_CONTEXT",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|18U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_INSUFFICIENT_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_INSUFFICIENT_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|19U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_INCOMPATIBLE_TCTI_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_INCOMPATIBLE_TCTI",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|20U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_SYS_RC_BAD_TCTI_STRUCTURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_SYS_RC_BAD_TCTI_STRUCTURE",SWIG_From_unsigned_SS_int((unsigned int)(((8 << (16))|22U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_MU_RC_GENERAL_FAILURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_MU_RC_GENERAL_FAILURE",SWIG_From_unsigned_SS_int((unsigned int)(((9 << (16))|1U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_MU_RC_BAD_REFERENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_MU_RC_BAD_REFERENCE",SWIG_From_unsigned_SS_int((unsigned int)(((9 << (16))|5U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_MU_RC_BAD_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_MU_RC_BAD_SIZE",SWIG_From_unsigned_SS_int((unsigned int)(((9 << (16))|16U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_MU_RC_BAD_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_MU_RC_BAD_VALUE",SWIG_From_unsigned_SS_int((unsigned int)(((9 << (16))|11U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_MU_RC_INSUFFICIENT_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_MU_RC_INSUFFICIENT_BUFFER",SWIG_From_unsigned_SS_int((unsigned int)(((9 << (16))|6U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_GENERAL_FAILURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_GENERAL_FAILURE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|1U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_NOT_IMPLEMENTED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_NOT_IMPLEMENTED",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|2U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_ABI_MISMATCH_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_ABI_MISMATCH",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|4U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_BAD_REFERENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_BAD_REFERENCE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|5U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_INSUFFICIENT_BUFFER_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_INSUFFICIENT_BUFFER",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|6U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_BAD_SEQUENCE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_BAD_SEQUENCE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|7U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_INVALID_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_INVALID_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|13U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_TRY_AGAIN_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_TRY_AGAIN",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|9U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_IO_ERROR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_IO_ERROR",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|10U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_BAD_VALUE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_BAD_VALUE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|11U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_NO_DECRYPT_PARAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_NO_DECRYPT_PARAM",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|14U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_NO_ENCRYPT_PARAM_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_NO_ENCRYPT_PARAM",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|15U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_BAD_SIZE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_BAD_SIZE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|16U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_MALFORMED_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_MALFORMED_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|17U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_INSUFFICIENT_CONTEXT_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_INSUFFICIENT_CONTEXT",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|18U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_INSUFFICIENT_RESPONSE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_INSUFFICIENT_RESPONSE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|19U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_INCOMPATIBLE_TCTI_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_INCOMPATIBLE_TCTI",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|20U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_BAD_TCTI_STRUCTURE_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_BAD_TCTI_STRUCTURE",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|22U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_MEMORY_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_MEMORY",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|23U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_BAD_TR_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_BAD_TR",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|24U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|25U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|26U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *TSS2_ESYS_RC_RSP_AUTH_FAILED_swigconstant(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *module;
  PyObject *d;
  if (!PyArg_ParseTuple(args,(char*)"O:swigconstant", &module)) return NULL;
  d = PyModule_GetDict(module);
  if (!d) return NULL;
  SWIG_Python_SetConstant(d, "TSS2_ESYS_RC_RSP_AUTH_FAILED",SWIG_From_unsigned_SS_int((unsigned int)(((7 << (16))|27U))));
  return SWIG_Py_Void();
}


SWIGINTERN PyObject *_wrap_Tss2_MU_BYTE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  BYTE arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_BYTE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_BYTE_Marshal" "', argument " "1"" of type '" "BYTE""'");
  } 
  arg1 = (BYTE)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_BYTE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_BYTE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_BYTE_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_BYTE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  BYTE *arg4 = (BYTE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_BYTE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_BYTE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_BYTE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_BYTE_Unmarshal" "', argument " "4"" of type '" "BYTE *""'"); 
  }
  arg4 = (BYTE *)(argp4);
  result = (TSS2_RC)Tss2_MU_BYTE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT8_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  INT8 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  signed char val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT8_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_signed_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_INT8_Marshal" "', argument " "1"" of type '" "INT8""'");
  } 
  arg1 = (INT8)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_INT8_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT8_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_INT8_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT8_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  INT8 *arg4 = (INT8 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT8_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_INT8_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT8_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_signed_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_INT8_Unmarshal" "', argument " "4"" of type '" "INT8 *""'"); 
  }
  arg4 = (INT8 *)(argp4);
  result = (TSS2_RC)Tss2_MU_INT8_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT16_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  INT16 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  short val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT16_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_short(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_INT16_Marshal" "', argument " "1"" of type '" "INT16""'");
  } 
  arg1 = (INT16)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_INT16_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT16_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_INT16_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT16_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  INT16 *arg4 = (INT16 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT16_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_INT16_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT16_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_short, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_INT16_Unmarshal" "', argument " "4"" of type '" "INT16 *""'"); 
  }
  arg4 = (INT16 *)(argp4);
  result = (TSS2_RC)Tss2_MU_INT16_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT32_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  INT32 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT32_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_INT32_Marshal" "', argument " "1"" of type '" "INT32""'");
  } 
  arg1 = (INT32)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_INT32_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT32_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_INT32_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT32_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  INT32 *arg4 = (INT32 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT32_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_INT32_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT32_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_INT32_Unmarshal" "', argument " "4"" of type '" "INT32 *""'"); 
  }
  arg4 = (INT32 *)(argp4);
  result = (TSS2_RC)Tss2_MU_INT32_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT64_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  INT64 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  long long val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT64_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_long_SS_long(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_INT64_Marshal" "', argument " "1"" of type '" "INT64""'");
  } 
  arg1 = (INT64)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_INT64_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT64_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_INT64_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_INT64_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  INT64 *arg4 = (INT64 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_INT64_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_INT64_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_INT64_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_long_long, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_INT64_Unmarshal" "', argument " "4"" of type '" "INT64 *""'"); 
  }
  arg4 = (INT64 *)(argp4);
  result = (TSS2_RC)Tss2_MU_INT64_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT8_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT8 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT8_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_UINT8_Marshal" "', argument " "1"" of type '" "UINT8""'");
  } 
  arg1 = (UINT8)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_UINT8_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT8_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_UINT8_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT8_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  UINT8 *arg4 = (UINT8 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT8_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_UINT8_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT8_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_UINT8_Unmarshal" "', argument " "4"" of type '" "UINT8 *""'"); 
  }
  arg4 = (UINT8 *)(argp4);
  result = (TSS2_RC)Tss2_MU_UINT8_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT16_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT16 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned short val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT16_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_short(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_UINT16_Marshal" "', argument " "1"" of type '" "UINT16""'");
  } 
  arg1 = (UINT16)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_UINT16_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT16_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_UINT16_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT16_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  UINT16 *arg4 = (UINT16 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT16_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_UINT16_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT16_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_UINT16_Unmarshal" "', argument " "4"" of type '" "UINT16 *""'"); 
  }
  arg4 = (UINT16 *)(argp4);
  result = (TSS2_RC)Tss2_MU_UINT16_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT32_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT32 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT32_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_UINT32_Marshal" "', argument " "1"" of type '" "UINT32""'");
  } 
  arg1 = (UINT32)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_UINT32_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT32_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_UINT32_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT32_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  UINT32 *arg4 = (UINT32 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT32_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_UINT32_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT32_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_UINT32_Unmarshal" "', argument " "4"" of type '" "UINT32 *""'"); 
  }
  arg4 = (UINT32 *)(argp4);
  result = (TSS2_RC)Tss2_MU_UINT32_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT64_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  UINT64 arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned long long val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT64_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long_SS_long(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_UINT64_Marshal" "', argument " "1"" of type '" "UINT64""'");
  } 
  arg1 = (UINT64)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_UINT64_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT64_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_UINT64_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_UINT64_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  UINT64 *arg4 = (UINT64 *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_UINT64_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_UINT64_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_UINT64_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_long_long, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_UINT64_Unmarshal" "', argument " "4"" of type '" "UINT64 *""'"); 
  }
  arg4 = (UINT64 *)(argp4);
  result = (TSS2_RC)Tss2_MU_UINT64_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_CC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2_CC arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2_CC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPM2_CC_Marshal" "', argument " "1"" of type '" "TPM2_CC""'");
  } 
  arg1 = (TPM2_CC)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2_CC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_CC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2_CC_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_CC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2_CC *arg4 = (TPM2_CC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2_CC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2_CC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_CC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2_CC_Unmarshal" "', argument " "4"" of type '" "TPM2_CC *""'"); 
  }
  arg4 = (TPM2_CC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2_CC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_ST_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2_ST arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned short val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2_ST_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_short(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPM2_ST_Marshal" "', argument " "1"" of type '" "TPM2_ST""'");
  } 
  arg1 = (TPM2_ST)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2_ST_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_ST_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2_ST_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_ST_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2_ST *arg4 = (TPM2_ST *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2_ST_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2_ST_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_ST_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2_ST_Unmarshal" "', argument " "4"" of type '" "TPM2_ST *""'"); 
  }
  arg4 = (TPM2_ST *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2_ST_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_ALGORITHM_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_ALGORITHM arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_ALGORITHM_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_ALGORITHM_Marshal" "', argument " "1"" of type '" "TPMA_ALGORITHM""'");
  } 
  arg1 = (TPMA_ALGORITHM)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_ALGORITHM_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_ALGORITHM_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_ALGORITHM_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_ALGORITHM_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_ALGORITHM *arg4 = (TPMA_ALGORITHM *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_ALGORITHM_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_ALGORITHM_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_ALGORITHM_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_ALGORITHM_Unmarshal" "', argument " "4"" of type '" "TPMA_ALGORITHM *""'"); 
  }
  arg4 = (TPMA_ALGORITHM *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_ALGORITHM_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_CC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_CC arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_CC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_CC_Marshal" "', argument " "1"" of type '" "TPMA_CC""'");
  } 
  arg1 = (TPMA_CC)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_CC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_CC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_CC_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_CC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_CC *arg4 = (TPMA_CC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_CC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_CC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_CC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_CC_Unmarshal" "', argument " "4"" of type '" "TPMA_CC *""'"); 
  }
  arg4 = (TPMA_CC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_CC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_LOCALITY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_LOCALITY arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_LOCALITY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_LOCALITY_Marshal" "', argument " "1"" of type '" "TPMA_LOCALITY""'");
  } 
  arg1 = (TPMA_LOCALITY)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_LOCALITY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_LOCALITY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_LOCALITY_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_LOCALITY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_LOCALITY *arg4 = (TPMA_LOCALITY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_LOCALITY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_LOCALITY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_LOCALITY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_LOCALITY_Unmarshal" "', argument " "4"" of type '" "TPMA_LOCALITY *""'"); 
  }
  arg4 = (TPMA_LOCALITY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_LOCALITY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_NV_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_NV arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_NV_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_NV_Marshal" "', argument " "1"" of type '" "TPMA_NV""'");
  } 
  arg1 = (TPMA_NV)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_NV_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_NV_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_NV_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_NV_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_NV *arg4 = (TPMA_NV *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_NV_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_NV_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_NV_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_NV_Unmarshal" "', argument " "4"" of type '" "TPMA_NV *""'"); 
  }
  arg4 = (TPMA_NV *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_NV_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_OBJECT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_OBJECT arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_OBJECT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_OBJECT_Marshal" "', argument " "1"" of type '" "TPMA_OBJECT""'");
  } 
  arg1 = (TPMA_OBJECT)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_OBJECT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_OBJECT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_OBJECT_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_OBJECT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_OBJECT *arg4 = (TPMA_OBJECT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_OBJECT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_OBJECT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_OBJECT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_OBJECT_Unmarshal" "', argument " "4"" of type '" "TPMA_OBJECT *""'"); 
  }
  arg4 = (TPMA_OBJECT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_OBJECT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_PERMANENT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_PERMANENT arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_PERMANENT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_PERMANENT_Marshal" "', argument " "1"" of type '" "TPMA_PERMANENT""'");
  } 
  arg1 = (TPMA_PERMANENT)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_PERMANENT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_PERMANENT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_PERMANENT_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_PERMANENT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_PERMANENT *arg4 = (TPMA_PERMANENT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_PERMANENT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_PERMANENT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_PERMANENT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_PERMANENT_Unmarshal" "', argument " "4"" of type '" "TPMA_PERMANENT *""'"); 
  }
  arg4 = (TPMA_PERMANENT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_PERMANENT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_SESSION_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_SESSION arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_SESSION_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_SESSION_Marshal" "', argument " "1"" of type '" "TPMA_SESSION""'");
  } 
  arg1 = (TPMA_SESSION)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_SESSION_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_SESSION_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_SESSION_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_SESSION_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_SESSION *arg4 = (TPMA_SESSION *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_SESSION_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_SESSION_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_SESSION_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_SESSION_Unmarshal" "', argument " "4"" of type '" "TPMA_SESSION *""'"); 
  }
  arg4 = (TPMA_SESSION *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_SESSION_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_STARTUP_CLEAR_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMA_STARTUP_CLEAR arg1 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_STARTUP_CLEAR_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMA_STARTUP_CLEAR_Marshal" "', argument " "1"" of type '" "TPMA_STARTUP_CLEAR""'");
  } 
  arg1 = (TPMA_STARTUP_CLEAR)(val1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMA_STARTUP_CLEAR_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_STARTUP_CLEAR_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMA_STARTUP_CLEAR_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMA_STARTUP_CLEAR *arg4 = (TPMA_STARTUP_CLEAR *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal" "', argument " "4"" of type '" "TPMA_STARTUP_CLEAR *""'"); 
  }
  arg4 = (TPMA_STARTUP_CLEAR *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_DIGEST_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DIGEST *arg1 = (TPM2B_DIGEST *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_DIGEST_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_DIGEST_Marshal" "', argument " "1"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg1 = (TPM2B_DIGEST *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_DIGEST_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_DIGEST_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_DIGEST_Marshal((TPM2B_DIGEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_DIGEST_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_DIGEST *arg4 = (TPM2B_DIGEST *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_DIGEST_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_DIGEST_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_DIGEST_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_DIGEST_Unmarshal" "', argument " "4"" of type '" "TPM2B_DIGEST *""'"); 
  }
  arg4 = (TPM2B_DIGEST *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_DIGEST_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ATTEST_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ATTEST *arg1 = (TPM2B_ATTEST *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ATTEST_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ATTEST_Marshal" "', argument " "1"" of type '" "TPM2B_ATTEST const *""'"); 
  }
  arg1 = (TPM2B_ATTEST *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_ATTEST_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ATTEST_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_ATTEST_Marshal((TPM2B_ATTEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ATTEST_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_ATTEST *arg4 = (TPM2B_ATTEST *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ATTEST_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ATTEST_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ATTEST_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_ATTEST_Unmarshal" "', argument " "4"" of type '" "TPM2B_ATTEST *""'"); 
  }
  arg4 = (TPM2B_ATTEST *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_ATTEST_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_NAME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NAME *arg1 = (TPM2B_NAME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_NAME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_NAME_Marshal" "', argument " "1"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg1 = (TPM2B_NAME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_NAME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_NAME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_NAME_Marshal((TPM2B_NAME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_NAME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_NAME *arg4 = (TPM2B_NAME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_NAME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_NAME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_NAME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_NAME_Unmarshal" "', argument " "4"" of type '" "TPM2B_NAME *""'"); 
  }
  arg4 = (TPM2B_NAME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_NAME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_NV_BUFFER *arg1 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal" "', argument " "1"" of type '" "TPM2B_MAX_NV_BUFFER const *""'"); 
  }
  arg1 = (TPM2B_MAX_NV_BUFFER *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal((TPM2B_MAX_NV_BUFFER const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_MAX_NV_BUFFER *arg4 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal" "', argument " "4"" of type '" "TPM2B_MAX_NV_BUFFER *""'"); 
  }
  arg4 = (TPM2B_MAX_NV_BUFFER *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_DATA *arg1 = (TPM2B_SENSITIVE_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal" "', argument " "1"" of type '" "TPM2B_SENSITIVE_DATA const *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal((TPM2B_SENSITIVE_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_SENSITIVE_DATA *arg4 = (TPM2B_SENSITIVE_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal" "', argument " "4"" of type '" "TPM2B_SENSITIVE_DATA *""'"); 
  }
  arg4 = (TPM2B_SENSITIVE_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ECC_PARAMETER_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_PARAMETER *arg1 = (TPM2B_ECC_PARAMETER *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ECC_PARAMETER_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ECC_PARAMETER_Marshal" "', argument " "1"" of type '" "TPM2B_ECC_PARAMETER const *""'"); 
  }
  arg1 = (TPM2B_ECC_PARAMETER *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_ECC_PARAMETER_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ECC_PARAMETER_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_ECC_PARAMETER_Marshal((TPM2B_ECC_PARAMETER const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_ECC_PARAMETER *arg4 = (TPM2B_ECC_PARAMETER *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal" "', argument " "4"" of type '" "TPM2B_ECC_PARAMETER *""'"); 
  }
  arg4 = (TPM2B_ECC_PARAMETER *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC_KEY_RSA *arg1 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal" "', argument " "1"" of type '" "TPM2B_PUBLIC_KEY_RSA const *""'"); 
  }
  arg1 = (TPM2B_PUBLIC_KEY_RSA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal((TPM2B_PUBLIC_KEY_RSA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_PUBLIC_KEY_RSA *arg4 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal" "', argument " "4"" of type '" "TPM2B_PUBLIC_KEY_RSA *""'"); 
  }
  arg4 = (TPM2B_PUBLIC_KEY_RSA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE_KEY_RSA *arg1 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal" "', argument " "1"" of type '" "TPM2B_PRIVATE_KEY_RSA const *""'"); 
  }
  arg1 = (TPM2B_PRIVATE_KEY_RSA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal((TPM2B_PRIVATE_KEY_RSA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_PRIVATE_KEY_RSA *arg4 = (TPM2B_PRIVATE_KEY_RSA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_PRIVATE_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal" "', argument " "4"" of type '" "TPM2B_PRIVATE_KEY_RSA *""'"); 
  }
  arg4 = (TPM2B_PRIVATE_KEY_RSA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PRIVATE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PRIVATE *arg1 = (TPM2B_PRIVATE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PRIVATE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PRIVATE_Marshal" "', argument " "1"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg1 = (TPM2B_PRIVATE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_PRIVATE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PRIVATE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_PRIVATE_Marshal((TPM2B_PRIVATE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PRIVATE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_PRIVATE *arg4 = (TPM2B_PRIVATE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PRIVATE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PRIVATE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PRIVATE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_PRIVATE_Unmarshal" "', argument " "4"" of type '" "TPM2B_PRIVATE *""'"); 
  }
  arg4 = (TPM2B_PRIVATE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_PRIVATE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_SENSITIVE *arg1 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal" "', argument " "1"" of type '" "TPM2B_CONTEXT_SENSITIVE const *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_SENSITIVE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal((TPM2B_CONTEXT_SENSITIVE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_CONTEXT_SENSITIVE *arg4 = (TPM2B_CONTEXT_SENSITIVE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_CONTEXT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal" "', argument " "4"" of type '" "TPM2B_CONTEXT_SENSITIVE *""'"); 
  }
  arg4 = (TPM2B_CONTEXT_SENSITIVE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_CONTEXT_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CONTEXT_DATA *arg1 = (TPM2B_CONTEXT_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_CONTEXT_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_CONTEXT_DATA_Marshal" "', argument " "1"" of type '" "TPM2B_CONTEXT_DATA const *""'"); 
  }
  arg1 = (TPM2B_CONTEXT_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_CONTEXT_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_CONTEXT_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_CONTEXT_DATA_Marshal((TPM2B_CONTEXT_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_CONTEXT_DATA *arg4 = (TPM2B_CONTEXT_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal" "', argument " "4"" of type '" "TPM2B_CONTEXT_DATA *""'"); 
  }
  arg4 = (TPM2B_CONTEXT_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_DATA *arg1 = (TPM2B_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_DATA_Marshal" "', argument " "1"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg1 = (TPM2B_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_DATA_Marshal((TPM2B_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_DATA *arg4 = (TPM2B_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_DATA_Unmarshal" "', argument " "4"" of type '" "TPM2B_DATA *""'"); 
  }
  arg4 = (TPM2B_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SYM_KEY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SYM_KEY *arg1 = (TPM2B_SYM_KEY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SYM_KEY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SYM_KEY_Marshal" "', argument " "1"" of type '" "TPM2B_SYM_KEY const *""'"); 
  }
  arg1 = (TPM2B_SYM_KEY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_SYM_KEY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SYM_KEY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_SYM_KEY_Marshal((TPM2B_SYM_KEY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SYM_KEY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_SYM_KEY *arg4 = (TPM2B_SYM_KEY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SYM_KEY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SYM_KEY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SYM_KEY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_SYM_KEY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_SYM_KEY_Unmarshal" "', argument " "4"" of type '" "TPM2B_SYM_KEY *""'"); 
  }
  arg4 = (TPM2B_SYM_KEY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_SYM_KEY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ECC_POINT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ECC_POINT *arg1 = (TPM2B_ECC_POINT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ECC_POINT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ECC_POINT_Marshal" "', argument " "1"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg1 = (TPM2B_ECC_POINT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_ECC_POINT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ECC_POINT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_ECC_POINT_Marshal((TPM2B_ECC_POINT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ECC_POINT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_ECC_POINT *arg4 = (TPM2B_ECC_POINT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ECC_POINT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ECC_POINT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ECC_POINT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_ECC_POINT_Unmarshal" "', argument " "4"" of type '" "TPM2B_ECC_POINT *""'"); 
  }
  arg4 = (TPM2B_ECC_POINT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_ECC_POINT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_NV_PUBLIC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NV_PUBLIC *arg1 = (TPM2B_NV_PUBLIC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_NV_PUBLIC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_NV_PUBLIC_Marshal" "', argument " "1"" of type '" "TPM2B_NV_PUBLIC const *""'"); 
  }
  arg1 = (TPM2B_NV_PUBLIC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_NV_PUBLIC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_NV_PUBLIC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_NV_PUBLIC_Marshal((TPM2B_NV_PUBLIC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_NV_PUBLIC *arg4 = (TPM2B_NV_PUBLIC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal" "', argument " "4"" of type '" "TPM2B_NV_PUBLIC *""'"); 
  }
  arg4 = (TPM2B_NV_PUBLIC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SENSITIVE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE *arg1 = (TPM2B_SENSITIVE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SENSITIVE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SENSITIVE_Marshal" "', argument " "1"" of type '" "TPM2B_SENSITIVE const *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_SENSITIVE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SENSITIVE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_SENSITIVE_Marshal((TPM2B_SENSITIVE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SENSITIVE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_SENSITIVE *arg4 = (TPM2B_SENSITIVE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SENSITIVE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SENSITIVE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SENSITIVE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_SENSITIVE_Unmarshal" "', argument " "4"" of type '" "TPM2B_SENSITIVE *""'"); 
  }
  arg4 = (TPM2B_SENSITIVE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_SENSITIVE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_SENSITIVE_CREATE *arg1 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal" "', argument " "1"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg1 = (TPM2B_SENSITIVE_CREATE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal((TPM2B_SENSITIVE_CREATE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_SENSITIVE_CREATE *arg4 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal" "', argument " "4"" of type '" "TPM2B_SENSITIVE_CREATE *""'"); 
  }
  arg4 = (TPM2B_SENSITIVE_CREATE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_CREATION_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_CREATION_DATA *arg1 = (TPM2B_CREATION_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_CREATION_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_CREATION_DATA_Marshal" "', argument " "1"" of type '" "TPM2B_CREATION_DATA const *""'"); 
  }
  arg1 = (TPM2B_CREATION_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_CREATION_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_CREATION_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_CREATION_DATA_Marshal((TPM2B_CREATION_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_CREATION_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_CREATION_DATA *arg4 = (TPM2B_CREATION_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_CREATION_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_CREATION_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_CREATION_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_CREATION_DATA_Unmarshal" "', argument " "4"" of type '" "TPM2B_CREATION_DATA *""'"); 
  }
  arg4 = (TPM2B_CREATION_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_CREATION_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PUBLIC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_PUBLIC *arg1 = (TPM2B_PUBLIC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PUBLIC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PUBLIC_Marshal" "', argument " "1"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg1 = (TPM2B_PUBLIC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_PUBLIC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PUBLIC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_PUBLIC_Marshal((TPM2B_PUBLIC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_PUBLIC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_PUBLIC *arg4 = (TPM2B_PUBLIC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_PUBLIC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_PUBLIC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_PUBLIC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_PUBLIC_Unmarshal" "', argument " "4"" of type '" "TPM2B_PUBLIC *""'"); 
  }
  arg4 = (TPM2B_PUBLIC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_PUBLIC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ENCRYPTED_SECRET *arg1 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal" "', argument " "1"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg1 = (TPM2B_ENCRYPTED_SECRET *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal((TPM2B_ENCRYPTED_SECRET const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg4 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal" "', argument " "4"" of type '" "TPM2B_ENCRYPTED_SECRET *""'"); 
  }
  arg4 = (TPM2B_ENCRYPTED_SECRET *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ID_OBJECT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_ID_OBJECT *arg1 = (TPM2B_ID_OBJECT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ID_OBJECT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ID_OBJECT_Marshal" "', argument " "1"" of type '" "TPM2B_ID_OBJECT const *""'"); 
  }
  arg1 = (TPM2B_ID_OBJECT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_ID_OBJECT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ID_OBJECT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_ID_OBJECT_Marshal((TPM2B_ID_OBJECT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_ID_OBJECT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_ID_OBJECT *arg4 = (TPM2B_ID_OBJECT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_ID_OBJECT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_ID_OBJECT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_ID_OBJECT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_ID_OBJECT_Unmarshal" "', argument " "4"" of type '" "TPM2B_ID_OBJECT *""'"); 
  }
  arg4 = (TPM2B_ID_OBJECT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_ID_OBJECT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_IV_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_IV *arg1 = (TPM2B_IV *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_IV_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_IV_Marshal" "', argument " "1"" of type '" "TPM2B_IV const *""'"); 
  }
  arg1 = (TPM2B_IV *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_IV_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_IV_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_IV_Marshal((TPM2B_IV const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_IV_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_IV *arg4 = (TPM2B_IV *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_IV_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_IV_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_IV_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_IV_Unmarshal" "', argument " "4"" of type '" "TPM2B_IV *""'"); 
  }
  arg4 = (TPM2B_IV *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_IV_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_AUTH_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_AUTH *arg1 = (TPM2B_AUTH *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_AUTH_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_AUTH_Marshal" "', argument " "1"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg1 = (TPM2B_AUTH *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_AUTH_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_AUTH_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_AUTH_Marshal((TPM2B_DIGEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_AUTH_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_AUTH *arg4 = (TPM2B_AUTH *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_AUTH_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_AUTH_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_AUTH_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_AUTH_Unmarshal" "', argument " "4"" of type '" "TPM2B_AUTH *""'"); 
  }
  arg4 = (TPM2B_AUTH *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_AUTH_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_EVENT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_EVENT *arg1 = (TPM2B_EVENT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_EVENT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_EVENT_Marshal" "', argument " "1"" of type '" "TPM2B_EVENT const *""'"); 
  }
  arg1 = (TPM2B_EVENT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_EVENT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_EVENT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_EVENT_Marshal((TPM2B_EVENT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_EVENT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_EVENT *arg4 = (TPM2B_EVENT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_EVENT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_EVENT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_EVENT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_EVENT_Unmarshal" "', argument " "4"" of type '" "TPM2B_EVENT *""'"); 
  }
  arg4 = (TPM2B_EVENT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_EVENT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_MAX_BUFFER_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_MAX_BUFFER *arg1 = (TPM2B_MAX_BUFFER *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_MAX_BUFFER_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_MAX_BUFFER_Marshal" "', argument " "1"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg1 = (TPM2B_MAX_BUFFER *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_MAX_BUFFER_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_MAX_BUFFER_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_MAX_BUFFER_Marshal((TPM2B_MAX_BUFFER const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_MAX_BUFFER *arg4 = (TPM2B_MAX_BUFFER *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal" "', argument " "4"" of type '" "TPM2B_MAX_BUFFER *""'"); 
  }
  arg4 = (TPM2B_MAX_BUFFER *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_NONCE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_NONCE *arg1 = (TPM2B_NONCE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_NONCE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_NONCE_Marshal" "', argument " "1"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg1 = (TPM2B_NONCE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_NONCE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_NONCE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_NONCE_Marshal((TPM2B_DIGEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_NONCE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_NONCE *arg4 = (TPM2B_NONCE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_NONCE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_NONCE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_NONCE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_NONCE_Unmarshal" "', argument " "4"" of type '" "TPM2B_NONCE *""'"); 
  }
  arg4 = (TPM2B_NONCE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_NONCE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_OPERAND_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_OPERAND *arg1 = (TPM2B_OPERAND *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_OPERAND_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_OPERAND_Marshal" "', argument " "1"" of type '" "TPM2B_OPERAND const *""'"); 
  }
  arg1 = (TPM2B_OPERAND *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_OPERAND_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_OPERAND_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_OPERAND_Marshal((TPM2B_DIGEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_OPERAND_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_OPERAND *arg4 = (TPM2B_OPERAND *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_OPERAND_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_OPERAND_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_OPERAND_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_OPERAND_Unmarshal" "', argument " "4"" of type '" "TPM2B_OPERAND *""'"); 
  }
  arg4 = (TPM2B_OPERAND *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_OPERAND_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_TIMEOUT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TIMEOUT *arg1 = (TPM2B_TIMEOUT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_TIMEOUT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_TIMEOUT_Marshal" "', argument " "1"" of type '" "TPM2B_TIMEOUT const *""'"); 
  }
  arg1 = (TPM2B_TIMEOUT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_TIMEOUT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_TIMEOUT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_TIMEOUT_Marshal((TPM2B_DIGEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_TIMEOUT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_TIMEOUT *arg4 = (TPM2B_TIMEOUT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_TIMEOUT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_TIMEOUT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_TIMEOUT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_TIMEOUT_Unmarshal" "', argument " "4"" of type '" "TPM2B_TIMEOUT *""'"); 
  }
  arg4 = (TPM2B_TIMEOUT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_TIMEOUT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_TEMPLATE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2B_TEMPLATE *arg1 = (TPM2B_TEMPLATE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_TEMPLATE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_TEMPLATE_Marshal" "', argument " "1"" of type '" "TPM2B_TEMPLATE const *""'"); 
  }
  arg1 = (TPM2B_TEMPLATE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2B_TEMPLATE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_TEMPLATE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2B_TEMPLATE_Marshal((TPM2B_TEMPLATE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2B_TEMPLATE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2B_TEMPLATE *arg4 = (TPM2B_TEMPLATE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPM2B_TEMPLATE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2B_TEMPLATE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2B_TEMPLATE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2B_TEMPLATE_Unmarshal" "', argument " "4"" of type '" "TPM2B_TEMPLATE *""'"); 
  }
  arg4 = (TPM2B_TEMPLATE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2B_TEMPLATE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CONTEXT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT *arg1 = (TPMS_CONTEXT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CONTEXT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CONTEXT_Marshal" "', argument " "1"" of type '" "TPMS_CONTEXT const *""'"); 
  }
  arg1 = (TPMS_CONTEXT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CONTEXT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CONTEXT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CONTEXT_Marshal((TPMS_CONTEXT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CONTEXT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CONTEXT *arg4 = (TPMS_CONTEXT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CONTEXT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CONTEXT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CONTEXT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CONTEXT_Unmarshal" "', argument " "4"" of type '" "TPMS_CONTEXT *""'"); 
  }
  arg4 = (TPMS_CONTEXT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CONTEXT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TIME_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_INFO *arg1 = (TPMS_TIME_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TIME_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TIME_INFO_Marshal" "', argument " "1"" of type '" "TPMS_TIME_INFO const *""'"); 
  }
  arg1 = (TPMS_TIME_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_TIME_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TIME_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_TIME_INFO_Marshal((TPMS_TIME_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TIME_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_TIME_INFO *arg4 = (TPMS_TIME_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TIME_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TIME_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TIME_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_TIME_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_TIME_INFO *""'"); 
  }
  arg4 = (TPMS_TIME_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_TIME_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ECC_POINT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_POINT *arg1 = (TPMS_ECC_POINT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ECC_POINT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ECC_POINT_Marshal" "', argument " "1"" of type '" "TPMS_ECC_POINT const *""'"); 
  }
  arg1 = (TPMS_ECC_POINT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ECC_POINT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ECC_POINT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ECC_POINT_Marshal((TPMS_ECC_POINT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ECC_POINT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ECC_POINT *arg4 = (TPMS_ECC_POINT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ECC_POINT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ECC_POINT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ECC_POINT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ECC_POINT_Unmarshal" "', argument " "4"" of type '" "TPMS_ECC_POINT *""'"); 
  }
  arg4 = (TPMS_ECC_POINT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ECC_POINT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_NV_PUBLIC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PUBLIC *arg1 = (TPMS_NV_PUBLIC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_NV_PUBLIC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_NV_PUBLIC_Marshal" "', argument " "1"" of type '" "TPMS_NV_PUBLIC const *""'"); 
  }
  arg1 = (TPMS_NV_PUBLIC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_NV_PUBLIC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_NV_PUBLIC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_NV_PUBLIC_Marshal((TPMS_NV_PUBLIC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_NV_PUBLIC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_NV_PUBLIC *arg4 = (TPMS_NV_PUBLIC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_NV_PUBLIC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_NV_PUBLIC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_NV_PUBLIC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_NV_PUBLIC_Unmarshal" "', argument " "4"" of type '" "TPMS_NV_PUBLIC *""'"); 
  }
  arg4 = (TPMS_NV_PUBLIC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_NV_PUBLIC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ALG_PROPERTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALG_PROPERTY *arg1 = (TPMS_ALG_PROPERTY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ALG_PROPERTY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ALG_PROPERTY_Marshal" "', argument " "1"" of type '" "TPMS_ALG_PROPERTY const *""'"); 
  }
  arg1 = (TPMS_ALG_PROPERTY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ALG_PROPERTY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ALG_PROPERTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ALG_PROPERTY_Marshal((TPMS_ALG_PROPERTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ALG_PROPERTY *arg4 = (TPMS_ALG_PROPERTY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal" "', argument " "4"" of type '" "TPMS_ALG_PROPERTY *""'"); 
  }
  arg4 = (TPMS_ALG_PROPERTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DESCRIPTION *arg1 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal" "', argument " "1"" of type '" "TPMS_ALGORITHM_DESCRIPTION const *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DESCRIPTION *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal((TPMS_ALGORITHM_DESCRIPTION const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ALGORITHM_DESCRIPTION *arg4 = (TPMS_ALGORITHM_DESCRIPTION *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ALGORITHM_DESCRIPTION, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal" "', argument " "4"" of type '" "TPMS_ALGORITHM_DESCRIPTION *""'"); 
  }
  arg4 = (TPMS_ALGORITHM_DESCRIPTION *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PROPERTY *arg1 = (TPMS_TAGGED_PROPERTY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal" "', argument " "1"" of type '" "TPMS_TAGGED_PROPERTY const *""'"); 
  }
  arg1 = (TPMS_TAGGED_PROPERTY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal((TPMS_TAGGED_PROPERTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_TAGGED_PROPERTY *arg4 = (TPMS_TAGGED_PROPERTY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_TAGGED_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal" "', argument " "4"" of type '" "TPMS_TAGGED_PROPERTY *""'"); 
  }
  arg4 = (TPMS_TAGGED_PROPERTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TAGGED_POLICY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_POLICY *arg1 = (TPMS_TAGGED_POLICY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TAGGED_POLICY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TAGGED_POLICY_Marshal" "', argument " "1"" of type '" "TPMS_TAGGED_POLICY const *""'"); 
  }
  arg1 = (TPMS_TAGGED_POLICY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_TAGGED_POLICY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TAGGED_POLICY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_TAGGED_POLICY_Marshal((TPMS_TAGGED_POLICY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_TAGGED_POLICY *arg4 = (TPMS_TAGGED_POLICY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_TAGGED_POLICY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal" "', argument " "4"" of type '" "TPMS_TAGGED_POLICY *""'"); 
  }
  arg4 = (TPMS_TAGGED_POLICY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CLOCK_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CLOCK_INFO *arg1 = (TPMS_CLOCK_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CLOCK_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CLOCK_INFO_Marshal" "', argument " "1"" of type '" "TPMS_CLOCK_INFO const *""'"); 
  }
  arg1 = (TPMS_CLOCK_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CLOCK_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CLOCK_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CLOCK_INFO_Marshal((TPMS_CLOCK_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CLOCK_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CLOCK_INFO *arg4 = (TPMS_CLOCK_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CLOCK_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CLOCK_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CLOCK_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CLOCK_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CLOCK_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_CLOCK_INFO *""'"); 
  }
  arg4 = (TPMS_CLOCK_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CLOCK_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TIME_ATTEST_INFO *arg1 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal" "', argument " "1"" of type '" "TPMS_TIME_ATTEST_INFO const *""'"); 
  }
  arg1 = (TPMS_TIME_ATTEST_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal((TPMS_TIME_ATTEST_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_TIME_ATTEST_INFO *arg4 = (TPMS_TIME_ATTEST_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_TIME_ATTEST_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_TIME_ATTEST_INFO *""'"); 
  }
  arg4 = (TPMS_TIME_ATTEST_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CERTIFY_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CERTIFY_INFO *arg1 = (TPMS_CERTIFY_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CERTIFY_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CERTIFY_INFO_Marshal" "', argument " "1"" of type '" "TPMS_CERTIFY_INFO const *""'"); 
  }
  arg1 = (TPMS_CERTIFY_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CERTIFY_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CERTIFY_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CERTIFY_INFO_Marshal((TPMS_CERTIFY_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CERTIFY_INFO *arg4 = (TPMS_CERTIFY_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_CERTIFY_INFO *""'"); 
  }
  arg4 = (TPMS_CERTIFY_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_COMMAND_AUDIT_INFO *arg1 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal" "', argument " "1"" of type '" "TPMS_COMMAND_AUDIT_INFO const *""'"); 
  }
  arg1 = (TPMS_COMMAND_AUDIT_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal((TPMS_COMMAND_AUDIT_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_COMMAND_AUDIT_INFO *arg4 = (TPMS_COMMAND_AUDIT_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_COMMAND_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_COMMAND_AUDIT_INFO *""'"); 
  }
  arg4 = (TPMS_COMMAND_AUDIT_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SESSION_AUDIT_INFO *arg1 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal" "', argument " "1"" of type '" "TPMS_SESSION_AUDIT_INFO const *""'"); 
  }
  arg1 = (TPMS_SESSION_AUDIT_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal((TPMS_SESSION_AUDIT_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SESSION_AUDIT_INFO *arg4 = (TPMS_SESSION_AUDIT_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SESSION_AUDIT_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_SESSION_AUDIT_INFO *""'"); 
  }
  arg4 = (TPMS_SESSION_AUDIT_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CREATION_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_INFO *arg1 = (TPMS_CREATION_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CREATION_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CREATION_INFO_Marshal" "', argument " "1"" of type '" "TPMS_CREATION_INFO const *""'"); 
  }
  arg1 = (TPMS_CREATION_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CREATION_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CREATION_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CREATION_INFO_Marshal((TPMS_CREATION_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CREATION_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CREATION_INFO *arg4 = (TPMS_CREATION_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CREATION_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CREATION_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CREATION_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CREATION_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CREATION_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_CREATION_INFO *""'"); 
  }
  arg4 = (TPMS_CREATION_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CREATION_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_CERTIFY_INFO *arg1 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal" "', argument " "1"" of type '" "TPMS_NV_CERTIFY_INFO const *""'"); 
  }
  arg1 = (TPMS_NV_CERTIFY_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal((TPMS_NV_CERTIFY_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_NV_CERTIFY_INFO *arg4 = (TPMS_NV_CERTIFY_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_NV_CERTIFY_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_NV_CERTIFY_INFO *""'"); 
  }
  arg4 = (TPMS_NV_CERTIFY_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_AUTH_COMMAND_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_COMMAND *arg1 = (TPMS_AUTH_COMMAND *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_AUTH_COMMAND_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_AUTH_COMMAND_Marshal" "', argument " "1"" of type '" "TPMS_AUTH_COMMAND const *""'"); 
  }
  arg1 = (TPMS_AUTH_COMMAND *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_AUTH_COMMAND_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_AUTH_COMMAND_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_AUTH_COMMAND_Marshal((TPMS_AUTH_COMMAND const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_AUTH_COMMAND *arg4 = (TPMS_AUTH_COMMAND *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_AUTH_COMMAND, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal" "', argument " "4"" of type '" "TPMS_AUTH_COMMAND *""'"); 
  }
  arg4 = (TPMS_AUTH_COMMAND *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_AUTH_RESPONSE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AUTH_RESPONSE *arg1 = (TPMS_AUTH_RESPONSE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_AUTH_RESPONSE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_AUTH_RESPONSE_Marshal" "', argument " "1"" of type '" "TPMS_AUTH_RESPONSE const *""'"); 
  }
  arg1 = (TPMS_AUTH_RESPONSE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_AUTH_RESPONSE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_AUTH_RESPONSE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_AUTH_RESPONSE_Marshal((TPMS_AUTH_RESPONSE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_AUTH_RESPONSE *arg4 = (TPMS_AUTH_RESPONSE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_AUTH_RESPONSE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal" "', argument " "4"" of type '" "TPMS_AUTH_RESPONSE *""'"); 
  }
  arg4 = (TPMS_AUTH_RESPONSE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SENSITIVE_CREATE *arg1 = (TPMS_SENSITIVE_CREATE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal" "', argument " "1"" of type '" "TPMS_SENSITIVE_CREATE const *""'"); 
  }
  arg1 = (TPMS_SENSITIVE_CREATE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal((TPMS_SENSITIVE_CREATE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SENSITIVE_CREATE *arg4 = (TPMS_SENSITIVE_CREATE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal" "', argument " "4"" of type '" "TPMS_SENSITIVE_CREATE *""'"); 
  }
  arg4 = (TPMS_SENSITIVE_CREATE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SCHEME_HASH_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_HASH *arg1 = (TPMS_SCHEME_HASH *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SCHEME_HASH_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SCHEME_HASH_Marshal" "', argument " "1"" of type '" "TPMS_SCHEME_HASH const *""'"); 
  }
  arg1 = (TPMS_SCHEME_HASH *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SCHEME_HASH_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SCHEME_HASH_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SCHEME_HASH_Marshal((TPMS_SCHEME_HASH const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SCHEME_HASH_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SCHEME_HASH *arg4 = (TPMS_SCHEME_HASH *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SCHEME_HASH_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SCHEME_HASH_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SCHEME_HASH_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SCHEME_HASH, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SCHEME_HASH_Unmarshal" "', argument " "4"" of type '" "TPMS_SCHEME_HASH *""'"); 
  }
  arg4 = (TPMS_SCHEME_HASH *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SCHEME_HASH_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SCHEME_ECDAA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_ECDAA *arg1 = (TPMS_SCHEME_ECDAA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SCHEME_ECDAA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SCHEME_ECDAA_Marshal" "', argument " "1"" of type '" "TPMS_SCHEME_ECDAA const *""'"); 
  }
  arg1 = (TPMS_SCHEME_ECDAA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SCHEME_ECDAA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SCHEME_ECDAA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SCHEME_ECDAA_Marshal((TPMS_SCHEME_ECDAA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SCHEME_ECDAA *arg4 = (TPMS_SCHEME_ECDAA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SCHEME_ECDAA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal" "', argument " "4"" of type '" "TPMS_SCHEME_ECDAA *""'"); 
  }
  arg4 = (TPMS_SCHEME_ECDAA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SCHEME_XOR_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SCHEME_XOR *arg1 = (TPMS_SCHEME_XOR *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SCHEME_XOR_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SCHEME_XOR_Marshal" "', argument " "1"" of type '" "TPMS_SCHEME_XOR const *""'"); 
  }
  arg1 = (TPMS_SCHEME_XOR *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SCHEME_XOR_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SCHEME_XOR_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SCHEME_XOR_Marshal((TPMS_SCHEME_XOR const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SCHEME_XOR_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SCHEME_XOR *arg4 = (TPMS_SCHEME_XOR *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SCHEME_XOR_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SCHEME_XOR_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SCHEME_XOR_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SCHEME_XOR, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SCHEME_XOR_Unmarshal" "', argument " "4"" of type '" "TPMS_SCHEME_XOR *""'"); 
  }
  arg4 = (TPMS_SCHEME_XOR *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SCHEME_XOR_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SIGNATURE_RSA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_RSA *arg1 = (TPMS_SIGNATURE_RSA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SIGNATURE_RSA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SIGNATURE_RSA_Marshal" "', argument " "1"" of type '" "TPMS_SIGNATURE_RSA const *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_RSA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SIGNATURE_RSA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SIGNATURE_RSA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SIGNATURE_RSA_Marshal((TPMS_SIGNATURE_RSA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SIGNATURE_RSA *arg4 = (TPMS_SIGNATURE_RSA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SIGNATURE_RSA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal" "', argument " "4"" of type '" "TPMS_SIGNATURE_RSA *""'"); 
  }
  arg4 = (TPMS_SIGNATURE_RSA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SIGNATURE_ECC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SIGNATURE_ECC *arg1 = (TPMS_SIGNATURE_ECC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SIGNATURE_ECC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SIGNATURE_ECC_Marshal" "', argument " "1"" of type '" "TPMS_SIGNATURE_ECC const *""'"); 
  }
  arg1 = (TPMS_SIGNATURE_ECC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SIGNATURE_ECC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SIGNATURE_ECC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SIGNATURE_ECC_Marshal((TPMS_SIGNATURE_ECC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SIGNATURE_ECC *arg4 = (TPMS_SIGNATURE_ECC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SIGNATURE_ECC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal" "', argument " "4"" of type '" "TPMS_SIGNATURE_ECC *""'"); 
  }
  arg4 = (TPMS_SIGNATURE_ECC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal" "', argument " "1"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS const *""'"); 
  }
  arg1 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal((TPMS_NV_PIN_COUNTER_PARAMETERS const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_NV_PIN_COUNTER_PARAMETERS *arg4 = (TPMS_NV_PIN_COUNTER_PARAMETERS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal" "', argument " "4"" of type '" "TPMS_NV_PIN_COUNTER_PARAMETERS *""'"); 
  }
  arg4 = (TPMS_NV_PIN_COUNTER_PARAMETERS *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CONTEXT_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CONTEXT_DATA *arg1 = (TPMS_CONTEXT_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CONTEXT_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CONTEXT_DATA_Marshal" "', argument " "1"" of type '" "TPMS_CONTEXT_DATA const *""'"); 
  }
  arg1 = (TPMS_CONTEXT_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CONTEXT_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CONTEXT_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CONTEXT_DATA_Marshal((TPMS_CONTEXT_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CONTEXT_DATA *arg4 = (TPMS_CONTEXT_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CONTEXT_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal" "', argument " "4"" of type '" "TPMS_CONTEXT_DATA *""'"); 
  }
  arg4 = (TPMS_CONTEXT_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_PCR_SELECT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECT *arg1 = (TPMS_PCR_SELECT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_PCR_SELECT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_PCR_SELECT_Marshal" "', argument " "1"" of type '" "TPMS_PCR_SELECT const *""'"); 
  }
  arg1 = (TPMS_PCR_SELECT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_PCR_SELECT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_PCR_SELECT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_PCR_SELECT_Marshal((TPMS_PCR_SELECT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_PCR_SELECT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_PCR_SELECT *arg4 = (TPMS_PCR_SELECT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_PCR_SELECT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_PCR_SELECT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_PCR_SELECT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_PCR_SELECT_Unmarshal" "', argument " "4"" of type '" "TPMS_PCR_SELECT *""'"); 
  }
  arg4 = (TPMS_PCR_SELECT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_PCR_SELECT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_PCR_SELECTION_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_PCR_SELECTION *arg1 = (TPMS_PCR_SELECTION *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_PCR_SELECTION_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_PCR_SELECTION_Marshal" "', argument " "1"" of type '" "TPMS_PCR_SELECTION const *""'"); 
  }
  arg1 = (TPMS_PCR_SELECTION *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_PCR_SELECTION_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_PCR_SELECTION_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_PCR_SELECTION_Marshal((TPMS_PCR_SELECTION const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_PCR_SELECTION_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_PCR_SELECTION *arg4 = (TPMS_PCR_SELECTION *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_PCR_SELECTION_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_PCR_SELECTION_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_PCR_SELECTION_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_PCR_SELECTION_Unmarshal" "', argument " "4"" of type '" "TPMS_PCR_SELECTION *""'"); 
  }
  arg4 = (TPMS_PCR_SELECTION *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_PCR_SELECTION_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_TAGGED_PCR_SELECT *arg1 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal" "', argument " "1"" of type '" "TPMS_TAGGED_PCR_SELECT const *""'"); 
  }
  arg1 = (TPMS_TAGGED_PCR_SELECT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal((TPMS_TAGGED_PCR_SELECT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_TAGGED_PCR_SELECT *arg4 = (TPMS_TAGGED_PCR_SELECT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_TAGGED_PCR_SELECT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal" "', argument " "4"" of type '" "TPMS_TAGGED_PCR_SELECT *""'"); 
  }
  arg4 = (TPMS_TAGGED_PCR_SELECT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_QUOTE_INFO_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_QUOTE_INFO *arg1 = (TPMS_QUOTE_INFO *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_QUOTE_INFO_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_QUOTE_INFO_Marshal" "', argument " "1"" of type '" "TPMS_QUOTE_INFO const *""'"); 
  }
  arg1 = (TPMS_QUOTE_INFO *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_QUOTE_INFO_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_QUOTE_INFO_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_QUOTE_INFO_Marshal((TPMS_QUOTE_INFO const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_QUOTE_INFO_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_QUOTE_INFO *arg4 = (TPMS_QUOTE_INFO *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_QUOTE_INFO_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_QUOTE_INFO_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_QUOTE_INFO_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_QUOTE_INFO, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_QUOTE_INFO_Unmarshal" "', argument " "4"" of type '" "TPMS_QUOTE_INFO *""'"); 
  }
  arg4 = (TPMS_QUOTE_INFO *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_QUOTE_INFO_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CREATION_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CREATION_DATA *arg1 = (TPMS_CREATION_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CREATION_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CREATION_DATA_Marshal" "', argument " "1"" of type '" "TPMS_CREATION_DATA const *""'"); 
  }
  arg1 = (TPMS_CREATION_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CREATION_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CREATION_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CREATION_DATA_Marshal((TPMS_CREATION_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CREATION_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CREATION_DATA *arg4 = (TPMS_CREATION_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CREATION_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CREATION_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CREATION_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CREATION_DATA_Unmarshal" "', argument " "4"" of type '" "TPMS_CREATION_DATA *""'"); 
  }
  arg4 = (TPMS_CREATION_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CREATION_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ECC_PARMS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ECC_PARMS *arg1 = (TPMS_ECC_PARMS *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ECC_PARMS_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ECC_PARMS_Marshal" "', argument " "1"" of type '" "TPMS_ECC_PARMS const *""'"); 
  }
  arg1 = (TPMS_ECC_PARMS *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ECC_PARMS_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ECC_PARMS_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ECC_PARMS_Marshal((TPMS_ECC_PARMS const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ECC_PARMS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ECC_PARMS *arg4 = (TPMS_ECC_PARMS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ECC_PARMS_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ECC_PARMS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ECC_PARMS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ECC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ECC_PARMS_Unmarshal" "', argument " "4"" of type '" "TPMS_ECC_PARMS *""'"); 
  }
  arg4 = (TPMS_ECC_PARMS *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ECC_PARMS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ATTEST_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ATTEST *arg1 = (TPMS_ATTEST *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ATTEST_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ATTEST_Marshal" "', argument " "1"" of type '" "TPMS_ATTEST const *""'"); 
  }
  arg1 = (TPMS_ATTEST *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ATTEST_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ATTEST_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ATTEST_Marshal((TPMS_ATTEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ATTEST_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ATTEST *arg4 = (TPMS_ATTEST *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ATTEST_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ATTEST_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ATTEST_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ATTEST_Unmarshal" "', argument " "4"" of type '" "TPMS_ATTEST *""'"); 
  }
  arg4 = (TPMS_ATTEST *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ATTEST_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ALGORITHM_DETAIL_ECC *arg1 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal" "', argument " "1"" of type '" "TPMS_ALGORITHM_DETAIL_ECC const *""'"); 
  }
  arg1 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal((TPMS_ALGORITHM_DETAIL_ECC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ALGORITHM_DETAIL_ECC *arg4 = (TPMS_ALGORITHM_DETAIL_ECC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal" "', argument " "4"" of type '" "TPMS_ALGORITHM_DETAIL_ECC *""'"); 
  }
  arg4 = (TPMS_ALGORITHM_DETAIL_ECC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CAPABILITY_DATA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_CAPABILITY_DATA *arg1 = (TPMS_CAPABILITY_DATA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CAPABILITY_DATA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CAPABILITY_DATA_Marshal" "', argument " "1"" of type '" "TPMS_CAPABILITY_DATA const *""'"); 
  }
  arg1 = (TPMS_CAPABILITY_DATA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_CAPABILITY_DATA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CAPABILITY_DATA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_CAPABILITY_DATA_Marshal((TPMS_CAPABILITY_DATA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_CAPABILITY_DATA *arg4 = (TPMS_CAPABILITY_DATA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal" "', argument " "4"" of type '" "TPMS_CAPABILITY_DATA *""'"); 
  }
  arg4 = (TPMS_CAPABILITY_DATA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_KEYEDHASH_PARMS *arg1 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal" "', argument " "1"" of type '" "TPMS_KEYEDHASH_PARMS const *""'"); 
  }
  arg1 = (TPMS_KEYEDHASH_PARMS *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal((TPMS_KEYEDHASH_PARMS const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_KEYEDHASH_PARMS *arg4 = (TPMS_KEYEDHASH_PARMS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_KEYEDHASH_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal" "', argument " "4"" of type '" "TPMS_KEYEDHASH_PARMS *""'"); 
  }
  arg4 = (TPMS_KEYEDHASH_PARMS *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_RSA_PARMS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_RSA_PARMS *arg1 = (TPMS_RSA_PARMS *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_RSA_PARMS_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_RSA_PARMS_Marshal" "', argument " "1"" of type '" "TPMS_RSA_PARMS const *""'"); 
  }
  arg1 = (TPMS_RSA_PARMS *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_RSA_PARMS_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_RSA_PARMS_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_RSA_PARMS_Marshal((TPMS_RSA_PARMS const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_RSA_PARMS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_RSA_PARMS *arg4 = (TPMS_RSA_PARMS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_RSA_PARMS_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_RSA_PARMS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_RSA_PARMS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_RSA_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_RSA_PARMS_Unmarshal" "', argument " "4"" of type '" "TPMS_RSA_PARMS *""'"); 
  }
  arg4 = (TPMS_RSA_PARMS *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_RSA_PARMS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_SYMCIPHER_PARMS *arg1 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal" "', argument " "1"" of type '" "TPMS_SYMCIPHER_PARMS const *""'"); 
  }
  arg1 = (TPMS_SYMCIPHER_PARMS *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal((TPMS_SYMCIPHER_PARMS const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_SYMCIPHER_PARMS *arg4 = (TPMS_SYMCIPHER_PARMS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_SYMCIPHER_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal" "', argument " "4"" of type '" "TPMS_SYMCIPHER_PARMS *""'"); 
  }
  arg4 = (TPMS_SYMCIPHER_PARMS *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_AC_OUTPUT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_AC_OUTPUT *arg1 = (TPMS_AC_OUTPUT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_AC_OUTPUT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_AC_OUTPUT_Marshal" "', argument " "1"" of type '" "TPMS_AC_OUTPUT const *""'"); 
  }
  arg1 = (TPMS_AC_OUTPUT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_AC_OUTPUT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_AC_OUTPUT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_AC_OUTPUT_Marshal((TPMS_AC_OUTPUT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_AC_OUTPUT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_AC_OUTPUT *arg4 = (TPMS_AC_OUTPUT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_AC_OUTPUT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_AC_OUTPUT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_AC_OUTPUT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_AC_OUTPUT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_AC_OUTPUT_Unmarshal" "', argument " "4"" of type '" "TPMS_AC_OUTPUT *""'"); 
  }
  arg4 = (TPMS_AC_OUTPUT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_AC_OUTPUT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ID_OBJECT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_ID_OBJECT *arg1 = (TPMS_ID_OBJECT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ID_OBJECT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ID_OBJECT_Marshal" "', argument " "1"" of type '" "TPMS_ID_OBJECT const *""'"); 
  }
  arg1 = (TPMS_ID_OBJECT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_ID_OBJECT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ID_OBJECT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_ID_OBJECT_Marshal((TPMS_ID_OBJECT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_ID_OBJECT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_ID_OBJECT *arg4 = (TPMS_ID_OBJECT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMS_ID_OBJECT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_ID_OBJECT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_ID_OBJECT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMS_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_ID_OBJECT_Unmarshal" "', argument " "4"" of type '" "TPMS_ID_OBJECT *""'"); 
  }
  arg4 = (TPMS_ID_OBJECT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_ID_OBJECT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_CC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CC *arg1 = (TPML_CC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_CC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_CC_Marshal" "', argument " "1"" of type '" "TPML_CC const *""'"); 
  }
  arg1 = (TPML_CC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_CC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_CC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_CC_Marshal((TPML_CC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_CC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_CC *arg4 = (TPML_CC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_CC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_CC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_CC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_CC_Unmarshal" "', argument " "4"" of type '" "TPML_CC *""'"); 
  }
  arg4 = (TPML_CC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_CC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_CCA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_CCA *arg1 = (TPML_CCA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_CCA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_CCA_Marshal" "', argument " "1"" of type '" "TPML_CCA const *""'"); 
  }
  arg1 = (TPML_CCA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_CCA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_CCA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_CCA_Marshal((TPML_CCA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_CCA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_CCA *arg4 = (TPML_CCA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_CCA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_CCA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_CCA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_CCA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_CCA_Unmarshal" "', argument " "4"" of type '" "TPML_CCA *""'"); 
  }
  arg4 = (TPML_CCA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_CCA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_ALG_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG *arg1 = (TPML_ALG *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_ALG_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_ALG_Marshal" "', argument " "1"" of type '" "TPML_ALG const *""'"); 
  }
  arg1 = (TPML_ALG *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_ALG_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_ALG_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_ALG_Marshal((TPML_ALG const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_ALG_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_ALG *arg4 = (TPML_ALG *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_ALG_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_ALG_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_ALG_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_ALG_Unmarshal" "', argument " "4"" of type '" "TPML_ALG *""'"); 
  }
  arg4 = (TPML_ALG *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_ALG_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_HANDLE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_HANDLE *arg1 = (TPML_HANDLE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_HANDLE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_HANDLE_Marshal" "', argument " "1"" of type '" "TPML_HANDLE const *""'"); 
  }
  arg1 = (TPML_HANDLE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_HANDLE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_HANDLE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_HANDLE_Marshal((TPML_HANDLE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_HANDLE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_HANDLE *arg4 = (TPML_HANDLE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_HANDLE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_HANDLE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_HANDLE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_HANDLE_Unmarshal" "', argument " "4"" of type '" "TPML_HANDLE *""'"); 
  }
  arg4 = (TPML_HANDLE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_HANDLE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_DIGEST_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST *arg1 = (TPML_DIGEST *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_DIGEST_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_DIGEST_Marshal" "', argument " "1"" of type '" "TPML_DIGEST const *""'"); 
  }
  arg1 = (TPML_DIGEST *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_DIGEST_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_DIGEST_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_DIGEST_Marshal((TPML_DIGEST const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_DIGEST_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_DIGEST *arg4 = (TPML_DIGEST *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_DIGEST_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_DIGEST_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_DIGEST_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_DIGEST_Unmarshal" "', argument " "4"" of type '" "TPML_DIGEST *""'"); 
  }
  arg4 = (TPML_DIGEST *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_DIGEST_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_DIGEST_VALUES_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_DIGEST_VALUES *arg1 = (TPML_DIGEST_VALUES *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_DIGEST_VALUES_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_DIGEST_VALUES_Marshal" "', argument " "1"" of type '" "TPML_DIGEST_VALUES const *""'"); 
  }
  arg1 = (TPML_DIGEST_VALUES *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_DIGEST_VALUES_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_DIGEST_VALUES_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_DIGEST_VALUES_Marshal((TPML_DIGEST_VALUES const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_DIGEST_VALUES_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_DIGEST_VALUES *arg4 = (TPML_DIGEST_VALUES *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_DIGEST_VALUES_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_DIGEST_VALUES_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_DIGEST_VALUES_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_DIGEST_VALUES_Unmarshal" "', argument " "4"" of type '" "TPML_DIGEST_VALUES *""'"); 
  }
  arg4 = (TPML_DIGEST_VALUES *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_DIGEST_VALUES_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_PCR_SELECTION_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_PCR_SELECTION *arg1 = (TPML_PCR_SELECTION *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_PCR_SELECTION_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_PCR_SELECTION_Marshal" "', argument " "1"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg1 = (TPML_PCR_SELECTION *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_PCR_SELECTION_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_PCR_SELECTION_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_PCR_SELECTION_Marshal((TPML_PCR_SELECTION const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_PCR_SELECTION_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_PCR_SELECTION *arg4 = (TPML_PCR_SELECTION *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_PCR_SELECTION_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_PCR_SELECTION_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_PCR_SELECTION_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_PCR_SELECTION_Unmarshal" "', argument " "4"" of type '" "TPML_PCR_SELECTION *""'"); 
  }
  arg4 = (TPML_PCR_SELECTION *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_PCR_SELECTION_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_ALG_PROPERTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ALG_PROPERTY *arg1 = (TPML_ALG_PROPERTY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_ALG_PROPERTY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_ALG_PROPERTY_Marshal" "', argument " "1"" of type '" "TPML_ALG_PROPERTY const *""'"); 
  }
  arg1 = (TPML_ALG_PROPERTY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_ALG_PROPERTY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_ALG_PROPERTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_ALG_PROPERTY_Marshal((TPML_ALG_PROPERTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_ALG_PROPERTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_ALG_PROPERTY *arg4 = (TPML_ALG_PROPERTY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_ALG_PROPERTY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_ALG_PROPERTY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_ALG_PROPERTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_ALG_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_ALG_PROPERTY_Unmarshal" "', argument " "4"" of type '" "TPML_ALG_PROPERTY *""'"); 
  }
  arg4 = (TPML_ALG_PROPERTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_ALG_PROPERTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_ECC_CURVE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_ECC_CURVE *arg1 = (TPML_ECC_CURVE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_ECC_CURVE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_ECC_CURVE_Marshal" "', argument " "1"" of type '" "TPML_ECC_CURVE const *""'"); 
  }
  arg1 = (TPML_ECC_CURVE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_ECC_CURVE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_ECC_CURVE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_ECC_CURVE_Marshal((TPML_ECC_CURVE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_ECC_CURVE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_ECC_CURVE *arg4 = (TPML_ECC_CURVE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_ECC_CURVE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_ECC_CURVE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_ECC_CURVE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_ECC_CURVE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_ECC_CURVE_Unmarshal" "', argument " "4"" of type '" "TPML_ECC_CURVE *""'"); 
  }
  arg4 = (TPML_ECC_CURVE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_ECC_CURVE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_PCR_PROPERTY *arg1 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal" "', argument " "1"" of type '" "TPML_TAGGED_PCR_PROPERTY const *""'"); 
  }
  arg1 = (TPML_TAGGED_PCR_PROPERTY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal((TPML_TAGGED_PCR_PROPERTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_TAGGED_PCR_PROPERTY *arg4 = (TPML_TAGGED_PCR_PROPERTY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_TAGGED_PCR_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal" "', argument " "4"" of type '" "TPML_TAGGED_PCR_PROPERTY *""'"); 
  }
  arg4 = (TPML_TAGGED_PCR_PROPERTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_TAGGED_TPM_PROPERTY *arg1 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal" "', argument " "1"" of type '" "TPML_TAGGED_TPM_PROPERTY const *""'"); 
  }
  arg1 = (TPML_TAGGED_TPM_PROPERTY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal((TPML_TAGGED_TPM_PROPERTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_TAGGED_TPM_PROPERTY *arg4 = (TPML_TAGGED_TPM_PROPERTY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_TAGGED_TPM_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal" "', argument " "4"" of type '" "TPML_TAGGED_TPM_PROPERTY *""'"); 
  }
  arg4 = (TPML_TAGGED_TPM_PROPERTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_INTEL_PTT_PROPERTY *arg1 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal" "', argument " "1"" of type '" "TPML_INTEL_PTT_PROPERTY const *""'"); 
  }
  arg1 = (TPML_INTEL_PTT_PROPERTY *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal((TPML_INTEL_PTT_PROPERTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_INTEL_PTT_PROPERTY *arg4 = (TPML_INTEL_PTT_PROPERTY *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_INTEL_PTT_PROPERTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal" "', argument " "4"" of type '" "TPML_INTEL_PTT_PROPERTY *""'"); 
  }
  arg4 = (TPML_INTEL_PTT_PROPERTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_AC_CAPABILITIES_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPML_AC_CAPABILITIES *arg1 = (TPML_AC_CAPABILITIES *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_AC_CAPABILITIES_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_AC_CAPABILITIES_Marshal" "', argument " "1"" of type '" "TPML_AC_CAPABILITIES const *""'"); 
  }
  arg1 = (TPML_AC_CAPABILITIES *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPML_AC_CAPABILITIES_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_AC_CAPABILITIES_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPML_AC_CAPABILITIES_Marshal((TPML_AC_CAPABILITIES const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPML_AC_CAPABILITIES *arg4 = (TPML_AC_CAPABILITIES *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPML_AC_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal" "', argument " "4"" of type '" "TPML_AC_CAPABILITIES *""'"); 
  }
  arg4 = (TPML_AC_CAPABILITIES *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_HA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_HA *arg1 = (TPMU_HA *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_HA_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_HA_Marshal" "', argument " "1"" of type '" "TPMU_HA const *""'"); 
  }
  arg1 = (TPMU_HA *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_HA_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_HA_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_HA_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_HA_Marshal((TPMU_HA const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_HA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_HA *arg5 = (TPMU_HA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_HA_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_HA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_HA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_HA_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_HA, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_HA_Unmarshal" "', argument " "5"" of type '" "TPMU_HA *""'"); 
  }
  arg5 = (TPMU_HA *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_HA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_CAPABILITIES_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_CAPABILITIES *arg1 = (TPMU_CAPABILITIES *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_CAPABILITIES_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Marshal" "', argument " "1"" of type '" "TPMU_CAPABILITIES const *""'"); 
  }
  arg1 = (TPMU_CAPABILITIES *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_CAPABILITIES_Marshal((TPMU_CAPABILITIES const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_CAPABILITIES_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_CAPABILITIES *arg5 = (TPMU_CAPABILITIES *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_CAPABILITIES_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_CAPABILITIES, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_CAPABILITIES_Unmarshal" "', argument " "5"" of type '" "TPMU_CAPABILITIES *""'"); 
  }
  arg5 = (TPMU_CAPABILITIES *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_CAPABILITIES_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_ATTEST_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ATTEST *arg1 = (TPMU_ATTEST *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_ATTEST_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_ATTEST_Marshal" "', argument " "1"" of type '" "TPMU_ATTEST const *""'"); 
  }
  arg1 = (TPMU_ATTEST *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_ATTEST_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_ATTEST_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_ATTEST_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_ATTEST_Marshal((TPMU_ATTEST const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_ATTEST_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_ATTEST *arg5 = (TPMU_ATTEST *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_ATTEST_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_ATTEST_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_ATTEST_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_ATTEST_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_ATTEST_Unmarshal" "', argument " "5"" of type '" "TPMU_ATTEST *""'"); 
  }
  arg5 = (TPMU_ATTEST *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_ATTEST_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SYM_KEY_BITS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_KEY_BITS *arg1 = (TPMU_SYM_KEY_BITS *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SYM_KEY_BITS_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Marshal" "', argument " "1"" of type '" "TPMU_SYM_KEY_BITS const *""'"); 
  }
  arg1 = (TPMU_SYM_KEY_BITS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_SYM_KEY_BITS_Marshal((TPMU_SYM_KEY_BITS const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_SYM_KEY_BITS *arg5 = (TPMU_SYM_KEY_BITS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_SYM_KEY_BITS, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal" "', argument " "5"" of type '" "TPMU_SYM_KEY_BITS *""'"); 
  }
  arg5 = (TPMU_SYM_KEY_BITS *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SYM_MODE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SYM_MODE *arg1 = (TPMU_SYM_MODE *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SYM_MODE_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SYM_MODE_Marshal" "', argument " "1"" of type '" "TPMU_SYM_MODE const *""'"); 
  }
  arg1 = (TPMU_SYM_MODE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_SYM_MODE_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_SYM_MODE_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SYM_MODE_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_SYM_MODE_Marshal((TPMU_SYM_MODE const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SYM_MODE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_SYM_MODE *arg5 = (TPMU_SYM_MODE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SYM_MODE_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SYM_MODE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SYM_MODE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_SYM_MODE_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_SYM_MODE, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_SYM_MODE_Unmarshal" "', argument " "5"" of type '" "TPMU_SYM_MODE *""'"); 
  }
  arg5 = (TPMU_SYM_MODE *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_SYM_MODE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SIG_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIG_SCHEME *arg1 = (TPMU_SIG_SCHEME *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SIG_SCHEME_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Marshal" "', argument " "1"" of type '" "TPMU_SIG_SCHEME const *""'"); 
  }
  arg1 = (TPMU_SIG_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_SIG_SCHEME_Marshal((TPMU_SIG_SCHEME const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SIG_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_SIG_SCHEME *arg5 = (TPMU_SIG_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SIG_SCHEME_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_SIG_SCHEME_Unmarshal" "', argument " "5"" of type '" "TPMU_SIG_SCHEME *""'"); 
  }
  arg5 = (TPMU_SIG_SCHEME *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_SIG_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_KDF_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_KDF_SCHEME *arg1 = (TPMU_KDF_SCHEME *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_KDF_SCHEME_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Marshal" "', argument " "1"" of type '" "TPMU_KDF_SCHEME const *""'"); 
  }
  arg1 = (TPMU_KDF_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_KDF_SCHEME_Marshal((TPMU_KDF_SCHEME const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_KDF_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_KDF_SCHEME *arg5 = (TPMU_KDF_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_KDF_SCHEME_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_KDF_SCHEME_Unmarshal" "', argument " "5"" of type '" "TPMU_KDF_SCHEME *""'"); 
  }
  arg5 = (TPMU_KDF_SCHEME *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_KDF_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_ASYM_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ASYM_SCHEME *arg1 = (TPMU_ASYM_SCHEME *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_ASYM_SCHEME_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Marshal" "', argument " "1"" of type '" "TPMU_ASYM_SCHEME const *""'"); 
  }
  arg1 = (TPMU_ASYM_SCHEME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_ASYM_SCHEME_Marshal((TPMU_ASYM_SCHEME const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_ASYM_SCHEME *arg5 = (TPMU_ASYM_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal" "', argument " "5"" of type '" "TPMU_ASYM_SCHEME *""'"); 
  }
  arg5 = (TPMU_ASYM_SCHEME *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SCHEME_KEYEDHASH *arg1 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal" "', argument " "1"" of type '" "TPMU_SCHEME_KEYEDHASH const *""'"); 
  }
  arg1 = (TPMU_SCHEME_KEYEDHASH *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal((TPMU_SCHEME_KEYEDHASH const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_SCHEME_KEYEDHASH *arg5 = (TPMU_SCHEME_KEYEDHASH *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_SCHEME_KEYEDHASH, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal" "', argument " "5"" of type '" "TPMU_SCHEME_KEYEDHASH *""'"); 
  }
  arg5 = (TPMU_SCHEME_KEYEDHASH *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SIGNATURE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SIGNATURE *arg1 = (TPMU_SIGNATURE *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SIGNATURE_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SIGNATURE_Marshal" "', argument " "1"" of type '" "TPMU_SIGNATURE const *""'"); 
  }
  arg1 = (TPMU_SIGNATURE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_SIGNATURE_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_SIGNATURE_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SIGNATURE_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_SIGNATURE_Marshal((TPMU_SIGNATURE const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SIGNATURE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_SIGNATURE *arg5 = (TPMU_SIGNATURE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SIGNATURE_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SIGNATURE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SIGNATURE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_SIGNATURE_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_SIGNATURE_Unmarshal" "', argument " "5"" of type '" "TPMU_SIGNATURE *""'"); 
  }
  arg5 = (TPMU_SIGNATURE *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_SIGNATURE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_SENSITIVE_COMPOSITE *arg1 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal" "', argument " "1"" of type '" "TPMU_SENSITIVE_COMPOSITE const *""'"); 
  }
  arg1 = (TPMU_SENSITIVE_COMPOSITE *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal((TPMU_SENSITIVE_COMPOSITE const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_SENSITIVE_COMPOSITE *arg5 = (TPMU_SENSITIVE_COMPOSITE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_SENSITIVE_COMPOSITE, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal" "', argument " "5"" of type '" "TPMU_SENSITIVE_COMPOSITE *""'"); 
  }
  arg5 = (TPMU_SENSITIVE_COMPOSITE *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_ENCRYPTED_SECRET *arg1 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal" "', argument " "1"" of type '" "TPMU_ENCRYPTED_SECRET const *""'"); 
  }
  arg1 = (TPMU_ENCRYPTED_SECRET *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal((TPMU_ENCRYPTED_SECRET const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_ENCRYPTED_SECRET *arg5 = (TPMU_ENCRYPTED_SECRET *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal" "', argument " "5"" of type '" "TPMU_ENCRYPTED_SECRET *""'"); 
  }
  arg5 = (TPMU_ENCRYPTED_SECRET *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_PUBLIC_PARMS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_PARMS *arg1 = (TPMU_PUBLIC_PARMS *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_PUBLIC_PARMS_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Marshal" "', argument " "1"" of type '" "TPMU_PUBLIC_PARMS const *""'"); 
  }
  arg1 = (TPMU_PUBLIC_PARMS *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_PUBLIC_PARMS_Marshal((TPMU_PUBLIC_PARMS const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_PUBLIC_PARMS *arg5 = (TPMU_PUBLIC_PARMS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal" "', argument " "5"" of type '" "TPMU_PUBLIC_PARMS *""'"); 
  }
  arg5 = (TPMU_PUBLIC_PARMS *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_PUBLIC_ID_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_PUBLIC_ID *arg1 = (TPMU_PUBLIC_ID *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_PUBLIC_ID_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Marshal" "', argument " "1"" of type '" "TPMU_PUBLIC_ID const *""'"); 
  }
  arg1 = (TPMU_PUBLIC_ID *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_PUBLIC_ID_Marshal((TPMU_PUBLIC_ID const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_PUBLIC_ID_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_PUBLIC_ID *arg5 = (TPMU_PUBLIC_ID *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_PUBLIC_ID_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_PUBLIC_ID, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_PUBLIC_ID_Unmarshal" "', argument " "5"" of type '" "TPMU_PUBLIC_ID *""'"); 
  }
  arg5 = (TPMU_PUBLIC_ID *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_PUBLIC_ID_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_NAME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMU_NAME *arg1 = (TPMU_NAME *) 0 ;
  uint32_t arg2 ;
  uint8_t *arg3 ;
  size_t arg4 ;
  size_t *arg5 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  Py_ssize_t size3 = 0 ;
  void *buf3 = 0 ;
  size_t temp5 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_NAME_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_NAME_Marshal" "', argument " "1"" of type '" "TPMU_NAME const *""'"); 
  }
  arg1 = (TPMU_NAME *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMU_NAME_Marshal" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = (uint32_t)(val2);
  {
    res3 = PyObject_AsWriteBuffer(obj2, &buf3, &size3);
    if (res3<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_MU_TPMU_NAME_Marshal" "', argument " "3"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg3 = (uint8_t *) buf3;
    arg4 = (size_t) (size3/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res5 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg5),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_NAME_Marshal" "', argument " "5"" of type '" "size_t""'");
    }
    temp5 = (size_t)(val);
    arg5 = &temp5;
    res5 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMU_NAME_Marshal((TPMU_NAME const *)arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMU_NAME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  uint32_t arg4 ;
  TPMU_NAME *arg5 = (TPMU_NAME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMU_NAME_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMU_NAME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMU_NAME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj2, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tss2_MU_TPMU_NAME_Unmarshal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = (uint32_t)(val4);
  res5 = SWIG_ConvertPtr(obj3, &argp5,SWIGTYPE_p_TPMU_NAME, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Tss2_MU_TPMU_NAME_Unmarshal" "', argument " "5"" of type '" "TPMU_NAME *""'"); 
  }
  arg5 = (TPMU_NAME *)(argp5);
  result = (TSS2_RC)Tss2_MU_TPMU_NAME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_HA_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_HA *arg1 = (TPMT_HA *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_HA_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_HA_Marshal" "', argument " "1"" of type '" "TPMT_HA const *""'"); 
  }
  arg1 = (TPMT_HA *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_HA_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_HA_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_HA_Marshal((TPMT_HA const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_HA_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_HA *arg4 = (TPMT_HA *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_HA_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_HA_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_HA_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_HA_Unmarshal" "', argument " "4"" of type '" "TPMT_HA *""'"); 
  }
  arg4 = (TPMT_HA *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_HA_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SYM_DEF_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF *arg1 = (TPMT_SYM_DEF *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SYM_DEF_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SYM_DEF_Marshal" "', argument " "1"" of type '" "TPMT_SYM_DEF const *""'"); 
  }
  arg1 = (TPMT_SYM_DEF *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_SYM_DEF_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SYM_DEF_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_SYM_DEF_Marshal((TPMT_SYM_DEF const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SYM_DEF_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_SYM_DEF *arg4 = (TPMT_SYM_DEF *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SYM_DEF_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SYM_DEF_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SYM_DEF_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_SYM_DEF_Unmarshal" "', argument " "4"" of type '" "TPMT_SYM_DEF *""'"); 
  }
  arg4 = (TPMT_SYM_DEF *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_SYM_DEF_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SYM_DEF_OBJECT *arg1 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal" "', argument " "1"" of type '" "TPMT_SYM_DEF_OBJECT const *""'"); 
  }
  arg1 = (TPMT_SYM_DEF_OBJECT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal((TPMT_SYM_DEF_OBJECT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg4 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal" "', argument " "4"" of type '" "TPMT_SYM_DEF_OBJECT *""'"); 
  }
  arg4 = (TPMT_SYM_DEF_OBJECT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KEYEDHASH_SCHEME *arg1 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal" "', argument " "1"" of type '" "TPMT_KEYEDHASH_SCHEME const *""'"); 
  }
  arg1 = (TPMT_KEYEDHASH_SCHEME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal((TPMT_KEYEDHASH_SCHEME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_KEYEDHASH_SCHEME *arg4 = (TPMT_KEYEDHASH_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_KEYEDHASH_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal" "', argument " "4"" of type '" "TPMT_KEYEDHASH_SCHEME *""'"); 
  }
  arg4 = (TPMT_KEYEDHASH_SCHEME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SIG_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIG_SCHEME *arg1 = (TPMT_SIG_SCHEME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SIG_SCHEME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SIG_SCHEME_Marshal" "', argument " "1"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg1 = (TPMT_SIG_SCHEME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_SIG_SCHEME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SIG_SCHEME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_SIG_SCHEME_Marshal((TPMT_SIG_SCHEME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SIG_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_SIG_SCHEME *arg4 = (TPMT_SIG_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SIG_SCHEME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SIG_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SIG_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_SIG_SCHEME_Unmarshal" "', argument " "4"" of type '" "TPMT_SIG_SCHEME *""'"); 
  }
  arg4 = (TPMT_SIG_SCHEME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_SIG_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_KDF_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_KDF_SCHEME *arg1 = (TPMT_KDF_SCHEME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_KDF_SCHEME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_KDF_SCHEME_Marshal" "', argument " "1"" of type '" "TPMT_KDF_SCHEME const *""'"); 
  }
  arg1 = (TPMT_KDF_SCHEME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_KDF_SCHEME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_KDF_SCHEME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_KDF_SCHEME_Marshal((TPMT_KDF_SCHEME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_KDF_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_KDF_SCHEME *arg4 = (TPMT_KDF_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_KDF_SCHEME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_KDF_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_KDF_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_KDF_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_KDF_SCHEME_Unmarshal" "', argument " "4"" of type '" "TPMT_KDF_SCHEME *""'"); 
  }
  arg4 = (TPMT_KDF_SCHEME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_KDF_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_ASYM_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ASYM_SCHEME *arg1 = (TPMT_ASYM_SCHEME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_ASYM_SCHEME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_ASYM_SCHEME_Marshal" "', argument " "1"" of type '" "TPMT_ASYM_SCHEME const *""'"); 
  }
  arg1 = (TPMT_ASYM_SCHEME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_ASYM_SCHEME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_ASYM_SCHEME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_ASYM_SCHEME_Marshal((TPMT_ASYM_SCHEME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_ASYM_SCHEME *arg4 = (TPMT_ASYM_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_ASYM_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal" "', argument " "4"" of type '" "TPMT_ASYM_SCHEME *""'"); 
  }
  arg4 = (TPMT_ASYM_SCHEME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_RSA_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_SCHEME *arg1 = (TPMT_RSA_SCHEME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_RSA_SCHEME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_RSA_SCHEME_Marshal" "', argument " "1"" of type '" "TPMT_RSA_SCHEME const *""'"); 
  }
  arg1 = (TPMT_RSA_SCHEME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_RSA_SCHEME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_RSA_SCHEME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_RSA_SCHEME_Marshal((TPMT_RSA_SCHEME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_RSA_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_RSA_SCHEME *arg4 = (TPMT_RSA_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_RSA_SCHEME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_RSA_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_RSA_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_RSA_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_RSA_SCHEME_Unmarshal" "', argument " "4"" of type '" "TPMT_RSA_SCHEME *""'"); 
  }
  arg4 = (TPMT_RSA_SCHEME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_RSA_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_RSA_DECRYPT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_RSA_DECRYPT *arg1 = (TPMT_RSA_DECRYPT *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_RSA_DECRYPT_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_RSA_DECRYPT_Marshal" "', argument " "1"" of type '" "TPMT_RSA_DECRYPT const *""'"); 
  }
  arg1 = (TPMT_RSA_DECRYPT *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_RSA_DECRYPT_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_RSA_DECRYPT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_RSA_DECRYPT_Marshal((TPMT_RSA_DECRYPT const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_RSA_DECRYPT *arg4 = (TPMT_RSA_DECRYPT *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal" "', argument " "4"" of type '" "TPMT_RSA_DECRYPT *""'"); 
  }
  arg4 = (TPMT_RSA_DECRYPT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_ECC_SCHEME_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_ECC_SCHEME *arg1 = (TPMT_ECC_SCHEME *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_ECC_SCHEME_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_ECC_SCHEME_Marshal" "', argument " "1"" of type '" "TPMT_ECC_SCHEME const *""'"); 
  }
  arg1 = (TPMT_ECC_SCHEME *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_ECC_SCHEME_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_ECC_SCHEME_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_ECC_SCHEME_Marshal((TPMT_ECC_SCHEME const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_ECC_SCHEME_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_ECC_SCHEME *arg4 = (TPMT_ECC_SCHEME *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_ECC_SCHEME_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_ECC_SCHEME_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_ECC_SCHEME_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_ECC_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_ECC_SCHEME_Unmarshal" "', argument " "4"" of type '" "TPMT_ECC_SCHEME *""'"); 
  }
  arg4 = (TPMT_ECC_SCHEME *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_ECC_SCHEME_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SIGNATURE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SIGNATURE *arg1 = (TPMT_SIGNATURE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SIGNATURE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SIGNATURE_Marshal" "', argument " "1"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg1 = (TPMT_SIGNATURE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_SIGNATURE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SIGNATURE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_SIGNATURE_Marshal((TPMT_SIGNATURE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SIGNATURE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_SIGNATURE *arg4 = (TPMT_SIGNATURE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SIGNATURE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SIGNATURE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SIGNATURE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_SIGNATURE_Unmarshal" "', argument " "4"" of type '" "TPMT_SIGNATURE *""'"); 
  }
  arg4 = (TPMT_SIGNATURE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_SIGNATURE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SENSITIVE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_SENSITIVE *arg1 = (TPMT_SENSITIVE *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SENSITIVE_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SENSITIVE_Marshal" "', argument " "1"" of type '" "TPMT_SENSITIVE const *""'"); 
  }
  arg1 = (TPMT_SENSITIVE *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_SENSITIVE_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SENSITIVE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_SENSITIVE_Marshal((TPMT_SENSITIVE const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_SENSITIVE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_SENSITIVE *arg4 = (TPMT_SENSITIVE *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_SENSITIVE_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_SENSITIVE_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_SENSITIVE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_SENSITIVE_Unmarshal" "', argument " "4"" of type '" "TPMT_SENSITIVE *""'"); 
  }
  arg4 = (TPMT_SENSITIVE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_SENSITIVE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_PUBLIC_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC *arg1 = (TPMT_PUBLIC *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_PUBLIC_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_PUBLIC_Marshal" "', argument " "1"" of type '" "TPMT_PUBLIC const *""'"); 
  }
  arg1 = (TPMT_PUBLIC *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_PUBLIC_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_PUBLIC_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_PUBLIC_Marshal((TPMT_PUBLIC const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_PUBLIC_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_PUBLIC *arg4 = (TPMT_PUBLIC *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_PUBLIC_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_PUBLIC_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_PUBLIC_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_PUBLIC_Unmarshal" "', argument " "4"" of type '" "TPMT_PUBLIC *""'"); 
  }
  arg4 = (TPMT_PUBLIC *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_PUBLIC_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_PUBLIC_PARMS_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_PUBLIC_PARMS *arg1 = (TPMT_PUBLIC_PARMS *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_PUBLIC_PARMS_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_PUBLIC_PARMS_Marshal" "', argument " "1"" of type '" "TPMT_PUBLIC_PARMS const *""'"); 
  }
  arg1 = (TPMT_PUBLIC_PARMS *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_PUBLIC_PARMS_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_PUBLIC_PARMS_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_PUBLIC_PARMS_Marshal((TPMT_PUBLIC_PARMS const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_PUBLIC_PARMS *arg4 = (TPMT_PUBLIC_PARMS *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal" "', argument " "4"" of type '" "TPMT_PUBLIC_PARMS *""'"); 
  }
  arg4 = (TPMT_PUBLIC_PARMS *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_CREATION_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_CREATION *arg1 = (TPMT_TK_CREATION *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_CREATION_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_CREATION_Marshal" "', argument " "1"" of type '" "TPMT_TK_CREATION const *""'"); 
  }
  arg1 = (TPMT_TK_CREATION *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_TK_CREATION_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_CREATION_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_TK_CREATION_Marshal((TPMT_TK_CREATION const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_CREATION_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_TK_CREATION *arg4 = (TPMT_TK_CREATION *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_CREATION_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_CREATION_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_CREATION_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_TK_CREATION_Unmarshal" "', argument " "4"" of type '" "TPMT_TK_CREATION *""'"); 
  }
  arg4 = (TPMT_TK_CREATION *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_TK_CREATION_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_VERIFIED_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_VERIFIED *arg1 = (TPMT_TK_VERIFIED *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_VERIFIED_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_VERIFIED_Marshal" "', argument " "1"" of type '" "TPMT_TK_VERIFIED const *""'"); 
  }
  arg1 = (TPMT_TK_VERIFIED *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_TK_VERIFIED_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_VERIFIED_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_TK_VERIFIED_Marshal((TPMT_TK_VERIFIED const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_VERIFIED_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_TK_VERIFIED *arg4 = (TPMT_TK_VERIFIED *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_VERIFIED_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_VERIFIED_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_VERIFIED_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_TK_VERIFIED_Unmarshal" "', argument " "4"" of type '" "TPMT_TK_VERIFIED *""'"); 
  }
  arg4 = (TPMT_TK_VERIFIED *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_TK_VERIFIED_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_AUTH_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_AUTH *arg1 = (TPMT_TK_AUTH *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_AUTH_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_AUTH_Marshal" "', argument " "1"" of type '" "TPMT_TK_AUTH const *""'"); 
  }
  arg1 = (TPMT_TK_AUTH *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_TK_AUTH_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_AUTH_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_TK_AUTH_Marshal((TPMT_TK_AUTH const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_AUTH_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_TK_AUTH *arg4 = (TPMT_TK_AUTH *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_AUTH_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_AUTH_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_AUTH_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_TK_AUTH_Unmarshal" "', argument " "4"" of type '" "TPMT_TK_AUTH *""'"); 
  }
  arg4 = (TPMT_TK_AUTH *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_TK_AUTH_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_HASHCHECK_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMT_TK_HASHCHECK *arg1 = (TPMT_TK_HASHCHECK *) 0 ;
  uint8_t *arg2 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  Py_ssize_t size2 = 0 ;
  void *buf2 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_HASHCHECK_Marshal",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_HASHCHECK_Marshal" "', argument " "1"" of type '" "TPMT_TK_HASHCHECK const *""'"); 
  }
  arg1 = (TPMT_TK_HASHCHECK *)(argp1);
  {
    res2 = PyObject_AsWriteBuffer(obj1, &buf2, &size2);
    if (res2<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMT_TK_HASHCHECK_Marshal" "', argument " "2"" of type '" "(uint8_t buffer[], size_t buffer_size)""'");
    }
    arg2 = (uint8_t *) buf2;
    arg3 = (size_t) (size2/sizeof(uint8_t));
  }
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_HASHCHECK_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMT_TK_HASHCHECK_Marshal((TPMT_TK_HASHCHECK const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMT_TK_HASHCHECK *arg4 = (TPMT_TK_HASHCHECK *) 0 ;
  int res1 ;
  Py_ssize_t size1 = 0 ;
  void const *buf1 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    res1 = PyObject_AsReadBuffer(obj0, &buf1, &size1);
    if (res1<0) {
      PyErr_Clear();
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal" "', argument " "1"" of type '" "(uint8_t const buffer[], size_t buffer_size)""'");
    }
    arg1 = (uint8_t *) buf1;
    arg2 = (size_t) (size1 / sizeof(uint8_t const));
  }
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj1,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj1, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj2, &argp4,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal" "', argument " "4"" of type '" "TPMT_TK_HASHCHECK *""'"); 
  }
  arg4 = (TPMT_TK_HASHCHECK *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_HANDLE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2_HANDLE arg1 ;
  uint8_t *arg2 = (uint8_t *) 0 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPM2_HANDLE_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPM2_HANDLE_Marshal" "', argument " "1"" of type '" "TPM2_HANDLE""'");
  } 
  arg1 = (TPM2_HANDLE)(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2_HANDLE_Marshal" "', argument " "2"" of type '" "uint8_t *""'"); 
  }
  arg2 = (uint8_t *)(argp2);
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tss2_MU_TPM2_HANDLE_Marshal" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = (size_t)(val3);
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_HANDLE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2_HANDLE_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_HANDLE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2_HANDLE *arg4 = (TPM2_HANDLE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPM2_HANDLE_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2_HANDLE_Unmarshal" "', argument " "1"" of type '" "uint8_t const []""'"); 
  } 
  arg1 = (uint8_t *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPM2_HANDLE_Unmarshal" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_HANDLE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2_HANDLE_Unmarshal" "', argument " "4"" of type '" "TPM2_HANDLE *""'"); 
  }
  arg4 = (TPM2_HANDLE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2_HANDLE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMI_ALG_HASH_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMI_ALG_HASH arg1 ;
  uint8_t *arg2 = (uint8_t *) 0 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned short val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMI_ALG_HASH_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_short(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPMI_ALG_HASH_Marshal" "', argument " "1"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg1 = (TPMI_ALG_HASH)(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMI_ALG_HASH_Marshal" "', argument " "2"" of type '" "uint8_t *""'"); 
  }
  arg2 = (uint8_t *)(argp2);
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tss2_MU_TPMI_ALG_HASH_Marshal" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = (size_t)(val3);
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMI_ALG_HASH_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMI_ALG_HASH_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMI_ALG_HASH_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMI_ALG_HASH *arg4 = (TPMI_ALG_HASH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMI_ALG_HASH_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMI_ALG_HASH_Unmarshal" "', argument " "1"" of type '" "uint8_t const []""'"); 
  } 
  arg1 = (uint8_t *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMI_ALG_HASH_Unmarshal" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMI_ALG_HASH_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMI_ALG_HASH_Unmarshal" "', argument " "4"" of type '" "TPMI_ALG_HASH *""'"); 
  }
  arg4 = (TPMI_ALG_HASH *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMI_ALG_HASH_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_SE_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2_SE arg1 ;
  uint8_t *arg2 = (uint8_t *) 0 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPM2_SE_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPM2_SE_Marshal" "', argument " "1"" of type '" "TPM2_SE""'");
  } 
  arg1 = (TPM2_SE)(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2_SE_Marshal" "', argument " "2"" of type '" "uint8_t *""'"); 
  }
  arg2 = (uint8_t *)(argp2);
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tss2_MU_TPM2_SE_Marshal" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = (size_t)(val3);
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_SE_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2_SE_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_SE_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2_SE *arg4 = (TPM2_SE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPM2_SE_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2_SE_Unmarshal" "', argument " "1"" of type '" "uint8_t const []""'"); 
  } 
  arg1 = (uint8_t *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPM2_SE_Unmarshal" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_SE_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2_SE_Unmarshal" "', argument " "4"" of type '" "TPM2_SE *""'"); 
  }
  arg4 = (TPM2_SE *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2_SE_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_NT_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPM2_NT arg1 ;
  uint8_t *arg2 = (uint8_t *) 0 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  unsigned char val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPM2_NT_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_char(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Tss2_MU_TPM2_NT_Marshal" "', argument " "1"" of type '" "TPM2_NT""'");
  } 
  arg1 = (TPM2_NT)(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPM2_NT_Marshal" "', argument " "2"" of type '" "uint8_t *""'"); 
  }
  arg2 = (uint8_t *)(argp2);
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tss2_MU_TPM2_NT_Marshal" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = (size_t)(val3);
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_NT_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPM2_NT_Marshal(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPM2_NT_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPM2_NT *arg4 = (TPM2_NT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPM2_NT_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPM2_NT_Unmarshal" "', argument " "1"" of type '" "uint8_t const []""'"); 
  } 
  arg1 = (uint8_t *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPM2_NT_Unmarshal" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPM2_NT_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPM2_NT_Unmarshal" "', argument " "4"" of type '" "TPM2_NT *""'"); 
  }
  arg4 = (TPM2_NT *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPM2_NT_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_EMPTY_Marshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  TPMS_EMPTY *arg1 = (TPMS_EMPTY *) 0 ;
  uint8_t *arg2 = (uint8_t *) 0 ;
  size_t arg3 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  size_t temp4 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMS_EMPTY_Marshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_EMPTY_Marshal" "', argument " "1"" of type '" "TPMS_EMPTY const *""'"); 
  }
  arg1 = (TPMS_EMPTY *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_MU_TPMS_EMPTY_Marshal" "', argument " "2"" of type '" "uint8_t *""'"); 
  }
  arg2 = (uint8_t *)(argp2);
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tss2_MU_TPMS_EMPTY_Marshal" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = (size_t)(val3);
  if (!(SWIG_IsOK((res4 = SWIG_ConvertPtr(obj3,SWIG_as_voidptrptr(&arg4),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj3, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_EMPTY_Marshal" "', argument " "4"" of type '" "size_t""'");
    }
    temp4 = (size_t)(val);
    arg4 = &temp4;
    res4 = SWIG_AddTmpMask(ecode);
  }
  result = (TSS2_RC)Tss2_MU_TPMS_EMPTY_Marshal((TPMS_EMPTY const *)arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_MU_TPMS_EMPTY_Unmarshal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  uint8_t *arg1 ;
  size_t arg2 ;
  size_t *arg3 = (size_t *) 0 ;
  TPMS_EMPTY *arg4 = (TPMS_EMPTY *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  size_t temp3 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Tss2_MU_TPMS_EMPTY_Unmarshal",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_MU_TPMS_EMPTY_Unmarshal" "', argument " "1"" of type '" "uint8_t const []""'"); 
  } 
  arg1 = (uint8_t *)(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tss2_MU_TPMS_EMPTY_Unmarshal" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = (size_t)(val2);
  if (!(SWIG_IsOK((res3 = SWIG_ConvertPtr(obj2,SWIG_as_voidptrptr(&arg3),SWIGTYPE_p_size_t,0))))) {
    size_t val; 
    int ecode = SWIG_AsVal_size_t(obj2, &val);
    if (!SWIG_IsOK(ecode)) {
      SWIG_exception_fail(SWIG_ArgError(ecode), "in method '" "Tss2_MU_TPMS_EMPTY_Unmarshal" "', argument " "3"" of type '" "size_t""'");
    }
    temp3 = (size_t)(val);
    arg3 = &temp3;
    res3 = SWIG_AddTmpMask(ecode);
  }
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_TPMS_EMPTY, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tss2_MU_TPMS_EMPTY_Unmarshal" "', argument " "4"" of type '" "TPMS_EMPTY *""'"); 
  }
  arg4 = (TPMS_EMPTY *)(argp4);
  result = (TSS2_RC)Tss2_MU_TPMS_EMPTY_Unmarshal((unsigned char const (*))arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_size_t((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_size_t, new_flags));
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_ESYS_TR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_TR_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_ESYS_TR_PTR")) SWIG_fail;
  result = (ESYS_TR_PTR *)new_ESYS_TR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ESYS_TR_PTR, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ESYS_TR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_TR_PTR *arg1 = (ESYS_TR_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_ESYS_TR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_TR_PTR, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ESYS_TR_PTR" "', argument " "1"" of type '" "ESYS_TR_PTR *""'"); 
  }
  arg1 = (ESYS_TR_PTR *)(argp1);
  delete_ESYS_TR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ESYS_TR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_TR_PTR *arg1 = (ESYS_TR_PTR *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:ESYS_TR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_TR_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ESYS_TR_PTR_assign" "', argument " "1"" of type '" "ESYS_TR_PTR *""'"); 
  }
  arg1 = (ESYS_TR_PTR *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ESYS_TR_PTR_assign" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ESYS_TR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ESYS_TR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_TR_PTR *arg1 = (ESYS_TR_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  unsigned int result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ESYS_TR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_TR_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ESYS_TR_PTR_value" "', argument " "1"" of type '" "ESYS_TR_PTR *""'"); 
  }
  arg1 = (ESYS_TR_PTR *)(argp1);
  result = (unsigned int)ESYS_TR_PTR_value(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ESYS_TR_PTR_cast(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_TR_PTR *arg1 = (ESYS_TR_PTR *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  unsigned int *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ESYS_TR_PTR_cast",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_TR_PTR, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ESYS_TR_PTR_cast" "', argument " "1"" of type '" "ESYS_TR_PTR *""'"); 
  }
  arg1 = (ESYS_TR_PTR *)(argp1);
  result = (unsigned int *)ESYS_TR_PTR_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ESYS_TR_PTR_frompointer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int *arg1 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  ESYS_TR_PTR *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ESYS_TR_PTR_frompointer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ESYS_TR_PTR_frompointer" "', argument " "1"" of type '" "unsigned int *""'"); 
  }
  arg1 = (unsigned int *)(argp1);
  result = (ESYS_TR_PTR *)ESYS_TR_PTR_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ESYS_TR_PTR, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *ESYS_TR_PTR_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O:swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_ESYS_TR_PTR, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_ESYS_TR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_ESYS_TR_PTR_PTR")) SWIG_fail;
  result = (unsigned int **)new_ESYS_TR_PTR_PTR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_ESYS_TR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int *arg1 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  unsigned int **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_ESYS_TR_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_ESYS_TR_PTR_PTR" "', argument " "1"" of type '" "unsigned int *""'"); 
  }
  arg1 = (unsigned int *)(argp1);
  result = (unsigned int **)copy_ESYS_TR_PTR_PTR(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ESYS_TR_PTR_PTR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int **arg1 = (unsigned int **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_ESYS_TR_PTR_PTR",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ESYS_TR_PTR_PTR" "', argument " "1"" of type '" "unsigned int **""'"); 
  }
  arg1 = (unsigned int **)(argp1);
  delete_ESYS_TR_PTR_PTR(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ESYS_TR_PTR_PTR_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int **arg1 = (unsigned int **) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:ESYS_TR_PTR_PTR_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ESYS_TR_PTR_PTR_assign" "', argument " "1"" of type '" "unsigned int **""'"); 
  }
  arg1 = (unsigned int **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ESYS_TR_PTR_PTR_assign" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  ESYS_TR_PTR_PTR_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ESYS_TR_PTR_PTR_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int **arg1 = (unsigned int **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  unsigned int *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ESYS_TR_PTR_PTR_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ESYS_TR_PTR_PTR_value" "', argument " "1"" of type '" "unsigned int **""'"); 
  }
  arg1 = (unsigned int **)(argp1);
  result = (unsigned int *)ESYS_TR_PTR_PTR_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_ctx_ptr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct ESYS_CONTEXT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_ctx_ptr")) SWIG_fail;
  result = (struct ESYS_CONTEXT **)new_ctx_ptr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_ctx_ptr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct ESYS_CONTEXT *arg1 = (struct ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  struct ESYS_CONTEXT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_ctx_ptr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_ctx_ptr" "', argument " "1"" of type '" "struct ESYS_CONTEXT *""'"); 
  }
  arg1 = (struct ESYS_CONTEXT *)(argp1);
  result = (struct ESYS_CONTEXT **)copy_ctx_ptr(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ctx_ptr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct ESYS_CONTEXT **arg1 = (struct ESYS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_ctx_ptr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ctx_ptr" "', argument " "1"" of type '" "struct ESYS_CONTEXT **""'"); 
  }
  arg1 = (struct ESYS_CONTEXT **)(argp1);
  delete_ctx_ptr(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ctx_ptr_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct ESYS_CONTEXT **arg1 = (struct ESYS_CONTEXT **) 0 ;
  struct ESYS_CONTEXT *arg2 = (struct ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:ctx_ptr_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ctx_ptr_assign" "', argument " "1"" of type '" "struct ESYS_CONTEXT **""'"); 
  }
  arg1 = (struct ESYS_CONTEXT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ctx_ptr_assign" "', argument " "2"" of type '" "struct ESYS_CONTEXT *""'"); 
  }
  arg2 = (struct ESYS_CONTEXT *)(argp2);
  ctx_ptr_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ctx_ptr_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct ESYS_CONTEXT **arg1 = (struct ESYS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  struct ESYS_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:ctx_ptr_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ctx_ptr_value" "', argument " "1"" of type '" "struct ESYS_CONTEXT **""'"); 
  }
  arg1 = (struct ESYS_CONTEXT **)(argp1);
  result = (struct ESYS_CONTEXT *)ctx_ptr_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_tcti_ctx_ptr_ptr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct TSS2_TCTI_CONTEXT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_tcti_ctx_ptr_ptr")) SWIG_fail;
  result = (struct TSS2_TCTI_CONTEXT **)new_tcti_ctx_ptr_ptr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_copy_tcti_ctx_ptr_ptr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct TSS2_TCTI_CONTEXT *arg1 = (struct TSS2_TCTI_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  struct TSS2_TCTI_CONTEXT **result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:copy_tcti_ctx_ptr_ptr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "copy_tcti_ctx_ptr_ptr" "', argument " "1"" of type '" "struct TSS2_TCTI_CONTEXT *""'"); 
  }
  arg1 = (struct TSS2_TCTI_CONTEXT *)(argp1);
  result = (struct TSS2_TCTI_CONTEXT **)copy_tcti_ctx_ptr_ptr(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_tcti_ctx_ptr_ptr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct TSS2_TCTI_CONTEXT **arg1 = (struct TSS2_TCTI_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_tcti_ctx_ptr_ptr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_tcti_ctx_ptr_ptr" "', argument " "1"" of type '" "struct TSS2_TCTI_CONTEXT **""'"); 
  }
  arg1 = (struct TSS2_TCTI_CONTEXT **)(argp1);
  delete_tcti_ctx_ptr_ptr(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_tcti_ctx_ptr_ptr_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct TSS2_TCTI_CONTEXT **arg1 = (struct TSS2_TCTI_CONTEXT **) 0 ;
  struct TSS2_TCTI_CONTEXT *arg2 = (struct TSS2_TCTI_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:tcti_ctx_ptr_ptr_assign",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tcti_ctx_ptr_ptr_assign" "', argument " "1"" of type '" "struct TSS2_TCTI_CONTEXT **""'"); 
  }
  arg1 = (struct TSS2_TCTI_CONTEXT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "tcti_ctx_ptr_ptr_assign" "', argument " "2"" of type '" "struct TSS2_TCTI_CONTEXT *""'"); 
  }
  arg2 = (struct TSS2_TCTI_CONTEXT *)(argp2);
  tcti_ctx_ptr_ptr_assign(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_tcti_ctx_ptr_ptr_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct TSS2_TCTI_CONTEXT **arg1 = (struct TSS2_TCTI_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  struct TSS2_TCTI_CONTEXT *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:tcti_ctx_ptr_ptr_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "tcti_ctx_ptr_ptr_value" "', argument " "1"" of type '" "struct TSS2_TCTI_CONTEXT **""'"); 
  }
  arg1 = (struct TSS2_TCTI_CONTEXT **)(argp1);
  result = (struct TSS2_TCTI_CONTEXT *)tcti_ctx_ptr_ptr_value(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tss2_TctiLdr_Initialize_Ex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  TSS2_TCTI_CONTEXT **arg3 = (TSS2_TCTI_CONTEXT **) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Tss2_TctiLdr_Initialize_Ex",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(obj0, &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tss2_TctiLdr_Initialize_Ex" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(obj1, &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tss2_TctiLdr_Initialize_Ex" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tss2_TctiLdr_Initialize_Ex" "', argument " "3"" of type '" "TSS2_TCTI_CONTEXT **""'"); 
  }
  arg3 = (TSS2_TCTI_CONTEXT **)(argp3);
  result = (TSS2_RC)Tss2_TctiLdr_Initialize_Ex((char const *)arg1,(char const *)arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Initialize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT **arg1 = (ESYS_CONTEXT **) 0 ;
  TSS2_TCTI_CONTEXT *arg2 = (TSS2_TCTI_CONTEXT *) 0 ;
  TSS2_ABI_VERSION *arg3 = (TSS2_ABI_VERSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_Initialize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Initialize" "', argument " "1"" of type '" "ESYS_CONTEXT **""'"); 
  }
  arg1 = (ESYS_CONTEXT **)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Initialize" "', argument " "2"" of type '" "TSS2_TCTI_CONTEXT *""'"); 
  }
  arg2 = (TSS2_TCTI_CONTEXT *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_TSS2_ABI_VERSION, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Initialize" "', argument " "3"" of type '" "TSS2_ABI_VERSION *""'"); 
  }
  arg3 = (TSS2_ABI_VERSION *)(argp3);
  result = (TSS2_RC)Esys_Initialize(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Finalize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT **arg1 = (ESYS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_Finalize",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Finalize" "', argument " "1"" of type '" "ESYS_CONTEXT **""'"); 
  }
  arg1 = (ESYS_CONTEXT **)(argp1);
  Esys_Finalize(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTcti(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TSS2_TCTI_CONTEXT **arg2 = (TSS2_TCTI_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_GetTcti",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTcti" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TSS2_TCTI_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetTcti" "', argument " "2"" of type '" "TSS2_TCTI_CONTEXT **""'"); 
  }
  arg2 = (TSS2_TCTI_CONTEXT **)(argp2);
  result = (TSS2_RC)Esys_GetTcti(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetPollHandles(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TSS2_TCTI_POLL_HANDLE **arg2 = (TSS2_TCTI_POLL_HANDLE **) 0 ;
  size_t *arg3 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_GetPollHandles",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetPollHandles" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TSS2_TCTI_POLL_HANDLE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetPollHandles" "', argument " "2"" of type '" "TSS2_TCTI_POLL_HANDLE **""'"); 
  }
  arg2 = (TSS2_TCTI_POLL_HANDLE **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_GetPollHandles" "', argument " "3"" of type '" "size_t *""'"); 
  }
  arg3 = (size_t *)(argp3);
  result = (TSS2_RC)Esys_GetPollHandles(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetTimeout(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  int32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_SetTimeout",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetTimeout" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetTimeout" "', argument " "2"" of type '" "int32_t""'");
  } 
  arg2 = (int32_t)(val2);
  result = (TSS2_RC)Esys_SetTimeout(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_Serialize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  uint8_t **arg3 = (uint8_t **) 0 ;
  size_t *arg4 = (size_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_TR_Serialize",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_Serialize" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TR_Serialize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_TR_Serialize" "', argument " "3"" of type '" "uint8_t **""'"); 
  }
  arg3 = (uint8_t **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_size_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_TR_Serialize" "', argument " "4"" of type '" "size_t *""'"); 
  }
  arg4 = (size_t *)(argp4);
  result = (TSS2_RC)Esys_TR_Serialize(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_Deserialize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  uint8_t *arg2 = (uint8_t *) 0 ;
  size_t arg3 ;
  unsigned int *arg4 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_TR_Deserialize",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_Deserialize" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_TR_Deserialize" "', argument " "2"" of type '" "uint8_t const *""'"); 
  }
  arg2 = (uint8_t *)(argp2);
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_TR_Deserialize" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = (size_t)(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_TR_Deserialize" "', argument " "4"" of type '" "unsigned int *""'"); 
  }
  arg4 = (unsigned int *)(argp4);
  result = (TSS2_RC)Esys_TR_Deserialize(arg1,(unsigned char const *)arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_FromTPMPublic_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2_HANDLE arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_TR_FromTPMPublic_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_FromTPMPublic_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TR_FromTPMPublic_Async" "', argument " "2"" of type '" "TPM2_HANDLE""'");
  } 
  arg2 = (TPM2_HANDLE)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_TR_FromTPMPublic_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_TR_FromTPMPublic_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_TR_FromTPMPublic_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_TR_FromTPMPublic_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_FromTPMPublic_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_TR_FromTPMPublic_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_FromTPMPublic_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_TR_FromTPMPublic_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_TR_FromTPMPublic_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_FromTPMPublic(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2_HANDLE arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int *arg6 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_TR_FromTPMPublic",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_FromTPMPublic" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TR_FromTPMPublic" "', argument " "2"" of type '" "TPM2_HANDLE""'");
  } 
  arg2 = (TPM2_HANDLE)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_TR_FromTPMPublic" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_TR_FromTPMPublic" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_TR_FromTPMPublic" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_TR_FromTPMPublic" "', argument " "6"" of type '" "unsigned int *""'"); 
  }
  arg6 = (unsigned int *)(argp6);
  result = (TSS2_RC)Esys_TR_FromTPMPublic(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_Close(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_TR_Close",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_Close" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_TR_Close" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_TR_Close(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_SetAuth(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  TPM2B_AUTH *arg3 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_TR_SetAuth",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_SetAuth" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TR_SetAuth" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_TR_SetAuth" "', argument " "3"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg3 = (TPM2B_AUTH *)(argp3);
  result = (TSS2_RC)Esys_TR_SetAuth(arg1,arg2,(TPM2B_DIGEST const *)arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TR_GetName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  TPM2B_NAME **arg3 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_TR_GetName",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TR_GetName" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TR_GetName" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_TR_GetName" "', argument " "3"" of type '" "TPM2B_NAME **""'"); 
  }
  arg3 = (TPM2B_NAME **)(argp3);
  result = (TSS2_RC)Esys_TR_GetName(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TRSess_GetAttributes(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  TPMA_SESSION *arg3 = (TPMA_SESSION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_TRSess_GetAttributes",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TRSess_GetAttributes" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TRSess_GetAttributes" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_TRSess_GetAttributes" "', argument " "3"" of type '" "TPMA_SESSION *""'"); 
  }
  arg3 = (TPMA_SESSION *)(argp3);
  result = (TSS2_RC)Esys_TRSess_GetAttributes(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TRSess_SetAttributes(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  TPMA_SESSION arg3 ;
  TPMA_SESSION arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned char val3 ;
  int ecode3 = 0 ;
  unsigned char val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_TRSess_SetAttributes",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TRSess_SetAttributes" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TRSess_SetAttributes" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_char(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_TRSess_SetAttributes" "', argument " "3"" of type '" "TPMA_SESSION""'");
  } 
  arg3 = (TPMA_SESSION)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_char(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_TRSess_SetAttributes" "', argument " "4"" of type '" "TPMA_SESSION""'");
  } 
  arg4 = (TPMA_SESSION)(val4);
  result = (TSS2_RC)Esys_TRSess_SetAttributes(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TRSess_GetNonceTPM(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  TPM2B_NONCE **arg3 = (TPM2B_NONCE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_TRSess_GetNonceTPM",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TRSess_GetNonceTPM" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TRSess_GetNonceTPM" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_TRSess_GetNonceTPM" "', argument " "3"" of type '" "TPM2B_NONCE **""'"); 
  }
  arg3 = (TPM2B_NONCE **)(argp3);
  result = (TSS2_RC)Esys_TRSess_GetNonceTPM(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Startup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2_SU arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Startup",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Startup" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Startup" "', argument " "2"" of type '" "TPM2_SU""'");
  } 
  arg2 = (TPM2_SU)(val2);
  result = (TSS2_RC)Esys_Startup(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Startup_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2_SU arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Startup_Async",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Startup_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Startup_Async" "', argument " "2"" of type '" "TPM2_SU""'");
  } 
  arg2 = (TPM2_SU)(val2);
  result = (TSS2_RC)Esys_Startup_Async(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Startup_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_Startup_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Startup_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_Startup_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Shutdown(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2_SU arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Shutdown",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Shutdown" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Shutdown" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Shutdown" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Shutdown" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Shutdown" "', argument " "5"" of type '" "TPM2_SU""'");
  } 
  arg5 = (TPM2_SU)(val5);
  result = (TSS2_RC)Esys_Shutdown(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Shutdown_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2_SU arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Shutdown_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Shutdown_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Shutdown_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Shutdown_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Shutdown_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Shutdown_Async" "', argument " "5"" of type '" "TPM2_SU""'");
  } 
  arg5 = (TPM2_SU)(val5);
  result = (TSS2_RC)Esys_Shutdown_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Shutdown_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_Shutdown_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Shutdown_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_Shutdown_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SelfTest(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMI_YES_NO arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_SelfTest",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SelfTest" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SelfTest" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SelfTest" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SelfTest" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SelfTest" "', argument " "5"" of type '" "TPMI_YES_NO""'");
  } 
  arg5 = (TPMI_YES_NO)(val5);
  result = (TSS2_RC)Esys_SelfTest(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SelfTest_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMI_YES_NO arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_SelfTest_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SelfTest_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SelfTest_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SelfTest_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SelfTest_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SelfTest_Async" "', argument " "5"" of type '" "TPMI_YES_NO""'");
  } 
  arg5 = (TPMI_YES_NO)(val5);
  result = (TSS2_RC)Esys_SelfTest_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SelfTest_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_SelfTest_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SelfTest_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_SelfTest_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_IncrementalSelfTest(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPML_ALG *arg5 = (TPML_ALG *) 0 ;
  TPML_ALG **arg6 = (TPML_ALG **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_IncrementalSelfTest",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_IncrementalSelfTest" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_IncrementalSelfTest" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_IncrementalSelfTest" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_IncrementalSelfTest" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_IncrementalSelfTest" "', argument " "5"" of type '" "TPML_ALG const *""'"); 
  }
  arg5 = (TPML_ALG *)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_IncrementalSelfTest" "', argument " "6"" of type '" "TPML_ALG **""'"); 
  }
  arg6 = (TPML_ALG **)(argp6);
  result = (TSS2_RC)Esys_IncrementalSelfTest(arg1,arg2,arg3,arg4,(TPML_ALG const *)arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_IncrementalSelfTest_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPML_ALG *arg5 = (TPML_ALG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_IncrementalSelfTest_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_IncrementalSelfTest_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_IncrementalSelfTest_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_IncrementalSelfTest_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_IncrementalSelfTest_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_IncrementalSelfTest_Async" "', argument " "5"" of type '" "TPML_ALG const *""'"); 
  }
  arg5 = (TPML_ALG *)(argp5);
  result = (TSS2_RC)Esys_IncrementalSelfTest_Async(arg1,arg2,arg3,arg4,(TPML_ALG const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_IncrementalSelfTest_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPML_ALG **arg2 = (TPML_ALG **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_IncrementalSelfTest_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_IncrementalSelfTest_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPML_ALG, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_IncrementalSelfTest_Finish" "', argument " "2"" of type '" "TPML_ALG **""'"); 
  }
  arg2 = (TPML_ALG **)(argp2);
  result = (TSS2_RC)Esys_IncrementalSelfTest_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTestResult(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_MAX_BUFFER **arg5 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2_RC *arg6 = (TPM2_RC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_GetTestResult",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTestResult" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetTestResult" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetTestResult" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetTestResult" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_GetTestResult" "', argument " "5"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg5 = (TPM2B_MAX_BUFFER **)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_GetTestResult" "', argument " "6"" of type '" "TPM2_RC *""'"); 
  }
  arg6 = (TPM2_RC *)(argp6);
  result = (TSS2_RC)Esys_GetTestResult(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTestResult_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_GetTestResult_Async",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTestResult_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetTestResult_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetTestResult_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetTestResult_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  result = (TSS2_RC)Esys_GetTestResult_Async(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTestResult_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_MAX_BUFFER **arg2 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2_RC *arg3 = (TPM2_RC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_GetTestResult_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTestResult_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetTestResult_Finish" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg2 = (TPM2B_MAX_BUFFER **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_GetTestResult_Finish" "', argument " "3"" of type '" "TPM2_RC *""'"); 
  }
  arg3 = (TPM2_RC *)(argp3);
  result = (TSS2_RC)Esys_GetTestResult_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_StartAuthSession(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2_SE arg8 ;
  TPMT_SYM_DEF *arg9 = (TPMT_SYM_DEF *) 0 ;
  TPMI_ALG_HASH arg10 ;
  unsigned int *arg11 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned char val8 ;
  int ecode8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  unsigned short val10 ;
  int ecode10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_StartAuthSession",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_StartAuthSession" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_StartAuthSession" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_StartAuthSession" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_StartAuthSession" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_StartAuthSession" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_StartAuthSession" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_StartAuthSession" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_char(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_StartAuthSession" "', argument " "8"" of type '" "TPM2_SE""'");
  } 
  arg8 = (TPM2_SE)(val8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_StartAuthSession" "', argument " "9"" of type '" "TPMT_SYM_DEF const *""'"); 
  }
  arg9 = (TPMT_SYM_DEF *)(argp9);
  ecode10 = SWIG_AsVal_unsigned_SS_short(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_StartAuthSession" "', argument " "10"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg10 = (TPMI_ALG_HASH)(val10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_StartAuthSession" "', argument " "11"" of type '" "unsigned int *""'"); 
  }
  arg11 = (unsigned int *)(argp11);
  result = (TSS2_RC)Esys_StartAuthSession(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,arg8,(TPMT_SYM_DEF const *)arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_StartAuthSession_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2_SE arg8 ;
  TPMT_SYM_DEF *arg9 = (TPMT_SYM_DEF *) 0 ;
  TPMI_ALG_HASH arg10 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned char val8 ;
  int ecode8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  unsigned short val10 ;
  int ecode10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_StartAuthSession_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_StartAuthSession_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_StartAuthSession_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_StartAuthSession_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_StartAuthSession_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_StartAuthSession_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_StartAuthSession_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_StartAuthSession_Async" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_char(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_StartAuthSession_Async" "', argument " "8"" of type '" "TPM2_SE""'");
  } 
  arg8 = (TPM2_SE)(val8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SYM_DEF, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_StartAuthSession_Async" "', argument " "9"" of type '" "TPMT_SYM_DEF const *""'"); 
  }
  arg9 = (TPMT_SYM_DEF *)(argp9);
  ecode10 = SWIG_AsVal_unsigned_SS_short(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_StartAuthSession_Async" "', argument " "10"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg10 = (TPMI_ALG_HASH)(val10);
  result = (TSS2_RC)Esys_StartAuthSession_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,arg8,(TPMT_SYM_DEF const *)arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_StartAuthSession_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_StartAuthSession_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_StartAuthSession_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_StartAuthSession_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_StartAuthSession_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyRestart(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyRestart",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyRestart" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyRestart" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyRestart" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyRestart" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyRestart" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyRestart(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyRestart_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyRestart_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyRestart_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyRestart_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyRestart_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyRestart_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyRestart_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyRestart_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyRestart_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyRestart_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyRestart_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyRestart_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_CREATE *arg6 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPML_PCR_SELECTION *arg9 = (TPML_PCR_SELECTION *) 0 ;
  TPM2B_PRIVATE **arg10 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_PUBLIC **arg11 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_CREATION_DATA **arg12 = (TPM2B_CREATION_DATA **) 0 ;
  TPM2B_DIGEST **arg13 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_CREATION **arg14 = (TPMT_TK_CREATION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOOOO:Esys_Create",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&obj12,&obj13)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Create" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Create" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Create" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Create" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Create" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Create" "', argument " "6"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_CREATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Create" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Create" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Create" "', argument " "9"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg9 = (TPML_PCR_SELECTION *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Create" "', argument " "10"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg10 = (TPM2B_PRIVATE **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_Create" "', argument " "11"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg11 = (TPM2B_PUBLIC **)(argp11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_Create" "', argument " "12"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg12 = (TPM2B_CREATION_DATA **)(argp12);
  res13 = SWIG_ConvertPtr(obj12, &argp13,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), "in method '" "Esys_Create" "', argument " "13"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg13 = (TPM2B_DIGEST **)(argp13);
  res14 = SWIG_ConvertPtr(obj13, &argp14,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), "in method '" "Esys_Create" "', argument " "14"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg14 = (TPMT_TK_CREATION **)(argp14);
  result = (TSS2_RC)Esys_Create(arg1,arg2,arg3,arg4,arg5,(TPM2B_SENSITIVE_CREATE const *)arg6,(TPM2B_PUBLIC const *)arg7,(TPM2B_DATA const *)arg8,(TPML_PCR_SELECTION const *)arg9,arg10,arg11,arg12,arg13,arg14);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Create_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_CREATE *arg6 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPML_PCR_SELECTION *arg9 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_Create_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Create_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Create_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Create_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Create_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Create_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Create_Async" "', argument " "6"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_CREATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Create_Async" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Create_Async" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Create_Async" "', argument " "9"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg9 = (TPML_PCR_SELECTION *)(argp9);
  result = (TSS2_RC)Esys_Create_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_SENSITIVE_CREATE const *)arg6,(TPM2B_PUBLIC const *)arg7,(TPM2B_DATA const *)arg8,(TPML_PCR_SELECTION const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Create_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PRIVATE **arg2 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_PUBLIC **arg3 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_CREATION_DATA **arg4 = (TPM2B_CREATION_DATA **) 0 ;
  TPM2B_DIGEST **arg5 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_CREATION **arg6 = (TPMT_TK_CREATION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_Create_Finish",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Create_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Create_Finish" "', argument " "2"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg2 = (TPM2B_PRIVATE **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Create_Finish" "', argument " "3"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg3 = (TPM2B_PUBLIC **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_Create_Finish" "', argument " "4"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg4 = (TPM2B_CREATION_DATA **)(argp4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_Create_Finish" "', argument " "5"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg5 = (TPM2B_DIGEST **)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Create_Finish" "', argument " "6"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg6 = (TPMT_TK_CREATION **)(argp6);
  result = (TSS2_RC)Esys_Create_Finish(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Load(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PRIVATE *arg6 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  unsigned int *arg8 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_Load",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Load" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Load" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Load" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Load" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Load" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Load" "', argument " "6"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg6 = (TPM2B_PRIVATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Load" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Load" "', argument " "8"" of type '" "unsigned int *""'"); 
  }
  arg8 = (unsigned int *)(argp8);
  result = (TSS2_RC)Esys_Load(arg1,arg2,arg3,arg4,arg5,(TPM2B_PRIVATE const *)arg6,(TPM2B_PUBLIC const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Load_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PRIVATE *arg6 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_Load_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Load_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Load_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Load_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Load_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Load_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Load_Async" "', argument " "6"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg6 = (TPM2B_PRIVATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Load_Async" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  result = (TSS2_RC)Esys_Load_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_PRIVATE const *)arg6,(TPM2B_PUBLIC const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Load_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Load_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Load_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Load_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_Load_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_LoadExternal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_SENSITIVE *arg5 = (TPM2B_SENSITIVE *) 0 ;
  TPM2B_PUBLIC *arg6 = (TPM2B_PUBLIC *) 0 ;
  TPMI_RH_HIERARCHY arg7 ;
  unsigned int *arg8 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_LoadExternal",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_LoadExternal" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_LoadExternal" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_LoadExternal" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_LoadExternal" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_LoadExternal" "', argument " "5"" of type '" "TPM2B_SENSITIVE const *""'"); 
  }
  arg5 = (TPM2B_SENSITIVE *)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_LoadExternal" "', argument " "6"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg6 = (TPM2B_PUBLIC *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_LoadExternal" "', argument " "7"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg7 = (TPMI_RH_HIERARCHY)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_LoadExternal" "', argument " "8"" of type '" "unsigned int *""'"); 
  }
  arg8 = (unsigned int *)(argp8);
  result = (TSS2_RC)Esys_LoadExternal(arg1,arg2,arg3,arg4,(TPM2B_SENSITIVE const *)arg5,(TPM2B_PUBLIC const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_LoadExternal_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_SENSITIVE *arg5 = (TPM2B_SENSITIVE *) 0 ;
  TPM2B_PUBLIC *arg6 = (TPM2B_PUBLIC *) 0 ;
  TPMI_RH_HIERARCHY arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_LoadExternal_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_LoadExternal_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_LoadExternal_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_LoadExternal_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_LoadExternal_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_SENSITIVE, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_LoadExternal_Async" "', argument " "5"" of type '" "TPM2B_SENSITIVE const *""'"); 
  }
  arg5 = (TPM2B_SENSITIVE *)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_LoadExternal_Async" "', argument " "6"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg6 = (TPM2B_PUBLIC *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_LoadExternal_Async" "', argument " "7"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg7 = (TPMI_RH_HIERARCHY)(val7);
  result = (TSS2_RC)Esys_LoadExternal_Async(arg1,arg2,arg3,arg4,(TPM2B_SENSITIVE const *)arg5,(TPM2B_PUBLIC const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_LoadExternal_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_LoadExternal_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_LoadExternal_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_LoadExternal_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_LoadExternal_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ReadPublic(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PUBLIC **arg6 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_NAME **arg7 = (TPM2B_NAME **) 0 ;
  TPM2B_NAME **arg8 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_ReadPublic",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ReadPublic" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ReadPublic" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ReadPublic" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ReadPublic" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ReadPublic" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ReadPublic" "', argument " "6"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg6 = (TPM2B_PUBLIC **)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ReadPublic" "', argument " "7"" of type '" "TPM2B_NAME **""'"); 
  }
  arg7 = (TPM2B_NAME **)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_ReadPublic" "', argument " "8"" of type '" "TPM2B_NAME **""'"); 
  }
  arg8 = (TPM2B_NAME **)(argp8);
  result = (TSS2_RC)Esys_ReadPublic(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ReadPublic_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ReadPublic_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ReadPublic_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ReadPublic_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ReadPublic_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ReadPublic_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ReadPublic_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_ReadPublic_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ReadPublic_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PUBLIC **arg2 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_NAME **arg3 = (TPM2B_NAME **) 0 ;
  TPM2B_NAME **arg4 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_ReadPublic_Finish",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ReadPublic_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ReadPublic_Finish" "', argument " "2"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg2 = (TPM2B_PUBLIC **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_ReadPublic_Finish" "', argument " "3"" of type '" "TPM2B_NAME **""'"); 
  }
  arg3 = (TPM2B_NAME **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_ReadPublic_Finish" "', argument " "4"" of type '" "TPM2B_NAME **""'"); 
  }
  arg4 = (TPM2B_NAME **)(argp4);
  result = (TSS2_RC)Esys_ReadPublic_Finish(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ActivateCredential(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_ID_OBJECT *arg7 = (TPM2B_ID_OBJECT *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg8 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  TPM2B_DIGEST **arg9 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_ActivateCredential",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ActivateCredential" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ActivateCredential" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ActivateCredential" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ActivateCredential" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ActivateCredential" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ActivateCredential" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ActivateCredential" "', argument " "7"" of type '" "TPM2B_ID_OBJECT const *""'"); 
  }
  arg7 = (TPM2B_ID_OBJECT *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_ActivateCredential" "', argument " "8"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg8 = (TPM2B_ENCRYPTED_SECRET *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_ActivateCredential" "', argument " "9"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg9 = (TPM2B_DIGEST **)(argp9);
  result = (TSS2_RC)Esys_ActivateCredential(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_ID_OBJECT const *)arg7,(TPM2B_ENCRYPTED_SECRET const *)arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ActivateCredential_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_ID_OBJECT *arg7 = (TPM2B_ID_OBJECT *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg8 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_ActivateCredential_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ActivateCredential_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ActivateCredential_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ActivateCredential_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ActivateCredential_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ActivateCredential_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ActivateCredential_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ActivateCredential_Async" "', argument " "7"" of type '" "TPM2B_ID_OBJECT const *""'"); 
  }
  arg7 = (TPM2B_ID_OBJECT *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_ActivateCredential_Async" "', argument " "8"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg8 = (TPM2B_ENCRYPTED_SECRET *)(argp8);
  result = (TSS2_RC)Esys_ActivateCredential_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_ID_OBJECT const *)arg7,(TPM2B_ENCRYPTED_SECRET const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ActivateCredential_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DIGEST **arg2 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ActivateCredential_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ActivateCredential_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ActivateCredential_Finish" "', argument " "2"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg2 = (TPM2B_DIGEST **)(argp2);
  result = (TSS2_RC)Esys_ActivateCredential_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_MakeCredential(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NAME *arg7 = (TPM2B_NAME *) 0 ;
  TPM2B_ID_OBJECT **arg8 = (TPM2B_ID_OBJECT **) 0 ;
  TPM2B_ENCRYPTED_SECRET **arg9 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_MakeCredential",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_MakeCredential" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_MakeCredential" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_MakeCredential" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_MakeCredential" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_MakeCredential" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_MakeCredential" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_MakeCredential" "', argument " "7"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg7 = (TPM2B_NAME *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_MakeCredential" "', argument " "8"" of type '" "TPM2B_ID_OBJECT **""'"); 
  }
  arg8 = (TPM2B_ID_OBJECT **)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_MakeCredential" "', argument " "9"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg9 = (TPM2B_ENCRYPTED_SECRET **)(argp9);
  result = (TSS2_RC)Esys_MakeCredential(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_NAME const *)arg7,arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_MakeCredential_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NAME *arg7 = (TPM2B_NAME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_MakeCredential_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_MakeCredential_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_MakeCredential_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_MakeCredential_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_MakeCredential_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_MakeCredential_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_MakeCredential_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_MakeCredential_Async" "', argument " "7"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg7 = (TPM2B_NAME *)(argp7);
  result = (TSS2_RC)Esys_MakeCredential_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_NAME const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_MakeCredential_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ID_OBJECT **arg2 = (TPM2B_ID_OBJECT **) 0 ;
  TPM2B_ENCRYPTED_SECRET **arg3 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_MakeCredential_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_MakeCredential_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ID_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_MakeCredential_Finish" "', argument " "2"" of type '" "TPM2B_ID_OBJECT **""'"); 
  }
  arg2 = (TPM2B_ID_OBJECT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_MakeCredential_Finish" "', argument " "3"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg3 = (TPM2B_ENCRYPTED_SECRET **)(argp3);
  result = (TSS2_RC)Esys_MakeCredential_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Unseal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_DATA **arg6 = (TPM2B_SENSITIVE_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_Unseal",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Unseal" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Unseal" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Unseal" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Unseal" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Unseal" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Unseal" "', argument " "6"" of type '" "TPM2B_SENSITIVE_DATA **""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_DATA **)(argp6);
  result = (TSS2_RC)Esys_Unseal(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Unseal_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Unseal_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Unseal_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Unseal_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Unseal_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Unseal_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Unseal_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_Unseal_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Unseal_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_SENSITIVE_DATA **arg2 = (TPM2B_SENSITIVE_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Unseal_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Unseal_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Unseal_Finish" "', argument " "2"" of type '" "TPM2B_SENSITIVE_DATA **""'"); 
  }
  arg2 = (TPM2B_SENSITIVE_DATA **)(argp2);
  result = (TSS2_RC)Esys_Unseal_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ObjectChangeAuth(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_AUTH *arg7 = (TPM2B_AUTH *) 0 ;
  TPM2B_PRIVATE **arg8 = (TPM2B_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_ObjectChangeAuth",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ObjectChangeAuth" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ObjectChangeAuth" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ObjectChangeAuth" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ObjectChangeAuth" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ObjectChangeAuth" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ObjectChangeAuth" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ObjectChangeAuth" "', argument " "7"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg7 = (TPM2B_AUTH *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_ObjectChangeAuth" "', argument " "8"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg8 = (TPM2B_PRIVATE **)(argp8);
  result = (TSS2_RC)Esys_ObjectChangeAuth(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ObjectChangeAuth_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_AUTH *arg7 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_ObjectChangeAuth_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ObjectChangeAuth_Async" "', argument " "7"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg7 = (TPM2B_AUTH *)(argp7);
  result = (TSS2_RC)Esys_ObjectChangeAuth_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ObjectChangeAuth_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PRIVATE **arg2 = (TPM2B_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ObjectChangeAuth_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ObjectChangeAuth_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ObjectChangeAuth_Finish" "', argument " "2"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg2 = (TPM2B_PRIVATE **)(argp2);
  result = (TSS2_RC)Esys_ObjectChangeAuth_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CreateLoaded(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_CREATE *arg6 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_TEMPLATE *arg7 = (TPM2B_TEMPLATE *) 0 ;
  unsigned int *arg8 = (unsigned int *) 0 ;
  TPM2B_PRIVATE **arg9 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_PUBLIC **arg10 = (TPM2B_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_CreateLoaded",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CreateLoaded" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_CreateLoaded" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_CreateLoaded" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_CreateLoaded" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_CreateLoaded" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_CreateLoaded" "', argument " "6"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_CREATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_CreateLoaded" "', argument " "7"" of type '" "TPM2B_TEMPLATE const *""'"); 
  }
  arg7 = (TPM2B_TEMPLATE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_CreateLoaded" "', argument " "8"" of type '" "unsigned int *""'"); 
  }
  arg8 = (unsigned int *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_CreateLoaded" "', argument " "9"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg9 = (TPM2B_PRIVATE **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_CreateLoaded" "', argument " "10"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg10 = (TPM2B_PUBLIC **)(argp10);
  result = (TSS2_RC)Esys_CreateLoaded(arg1,arg2,arg3,arg4,arg5,(TPM2B_SENSITIVE_CREATE const *)arg6,(TPM2B_TEMPLATE const *)arg7,arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CreateLoaded_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_CREATE *arg6 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_TEMPLATE *arg7 = (TPM2B_TEMPLATE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_CreateLoaded_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CreateLoaded_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_CreateLoaded_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_CreateLoaded_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_CreateLoaded_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_CreateLoaded_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_CreateLoaded_Async" "', argument " "6"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_CREATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_TEMPLATE, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_CreateLoaded_Async" "', argument " "7"" of type '" "TPM2B_TEMPLATE const *""'"); 
  }
  arg7 = (TPM2B_TEMPLATE *)(argp7);
  result = (TSS2_RC)Esys_CreateLoaded_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_SENSITIVE_CREATE const *)arg6,(TPM2B_TEMPLATE const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CreateLoaded_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  TPM2B_PRIVATE **arg3 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_PUBLIC **arg4 = (TPM2B_PUBLIC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_CreateLoaded_Finish",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CreateLoaded_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_CreateLoaded_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_CreateLoaded_Finish" "', argument " "3"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg3 = (TPM2B_PRIVATE **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_CreateLoaded_Finish" "', argument " "4"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg4 = (TPM2B_PUBLIC **)(argp4);
  result = (TSS2_RC)Esys_CreateLoaded_Finish(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Duplicate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg8 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  TPM2B_DATA **arg9 = (TPM2B_DATA **) 0 ;
  TPM2B_PRIVATE **arg10 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_ENCRYPTED_SECRET **arg11 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_Duplicate",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Duplicate" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Duplicate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Duplicate" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Duplicate" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Duplicate" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Duplicate" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Duplicate" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Duplicate" "', argument " "8"" of type '" "TPMT_SYM_DEF_OBJECT const *""'"); 
  }
  arg8 = (TPMT_SYM_DEF_OBJECT *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Duplicate" "', argument " "9"" of type '" "TPM2B_DATA **""'"); 
  }
  arg9 = (TPM2B_DATA **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Duplicate" "', argument " "10"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg10 = (TPM2B_PRIVATE **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_Duplicate" "', argument " "11"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg11 = (TPM2B_ENCRYPTED_SECRET **)(argp11);
  result = (TSS2_RC)Esys_Duplicate(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SYM_DEF_OBJECT const *)arg8,arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Duplicate_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg8 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_Duplicate_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Duplicate_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Duplicate_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Duplicate_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Duplicate_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Duplicate_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Duplicate_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Duplicate_Async" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Duplicate_Async" "', argument " "8"" of type '" "TPMT_SYM_DEF_OBJECT const *""'"); 
  }
  arg8 = (TPMT_SYM_DEF_OBJECT *)(argp8);
  result = (TSS2_RC)Esys_Duplicate_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SYM_DEF_OBJECT const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Duplicate_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DATA **arg2 = (TPM2B_DATA **) 0 ;
  TPM2B_PRIVATE **arg3 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_ENCRYPTED_SECRET **arg4 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_Duplicate_Finish",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Duplicate_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Duplicate_Finish" "', argument " "2"" of type '" "TPM2B_DATA **""'"); 
  }
  arg2 = (TPM2B_DATA **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Duplicate_Finish" "', argument " "3"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg3 = (TPM2B_PRIVATE **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_Duplicate_Finish" "', argument " "4"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg4 = (TPM2B_ENCRYPTED_SECRET **)(argp4);
  result = (TSS2_RC)Esys_Duplicate_Finish(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Rewrap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_PRIVATE *arg7 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_NAME *arg8 = (TPM2B_NAME *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg9 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  TPM2B_PRIVATE **arg10 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_ENCRYPTED_SECRET **arg11 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_Rewrap",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Rewrap" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Rewrap" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Rewrap" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Rewrap" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Rewrap" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Rewrap" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Rewrap" "', argument " "7"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg7 = (TPM2B_PRIVATE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Rewrap" "', argument " "8"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg8 = (TPM2B_NAME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Rewrap" "', argument " "9"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg9 = (TPM2B_ENCRYPTED_SECRET *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Rewrap" "', argument " "10"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg10 = (TPM2B_PRIVATE **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_Rewrap" "', argument " "11"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg11 = (TPM2B_ENCRYPTED_SECRET **)(argp11);
  result = (TSS2_RC)Esys_Rewrap(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_PRIVATE const *)arg7,(TPM2B_NAME const *)arg8,(TPM2B_ENCRYPTED_SECRET const *)arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Rewrap_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_PRIVATE *arg7 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_NAME *arg8 = (TPM2B_NAME *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg9 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_Rewrap_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Rewrap_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Rewrap_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Rewrap_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Rewrap_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Rewrap_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Rewrap_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Rewrap_Async" "', argument " "7"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg7 = (TPM2B_PRIVATE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Rewrap_Async" "', argument " "8"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg8 = (TPM2B_NAME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Rewrap_Async" "', argument " "9"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg9 = (TPM2B_ENCRYPTED_SECRET *)(argp9);
  result = (TSS2_RC)Esys_Rewrap_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_PRIVATE const *)arg7,(TPM2B_NAME const *)arg8,(TPM2B_ENCRYPTED_SECRET const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Rewrap_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PRIVATE **arg2 = (TPM2B_PRIVATE **) 0 ;
  TPM2B_ENCRYPTED_SECRET **arg3 = (TPM2B_ENCRYPTED_SECRET **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_Rewrap_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Rewrap_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Rewrap_Finish" "', argument " "2"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg2 = (TPM2B_PRIVATE **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Rewrap_Finish" "', argument " "3"" of type '" "TPM2B_ENCRYPTED_SECRET **""'"); 
  }
  arg3 = (TPM2B_ENCRYPTED_SECRET **)(argp3);
  result = (TSS2_RC)Esys_Rewrap_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Import(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DATA *arg6 = (TPM2B_DATA *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_PRIVATE *arg8 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg9 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg10 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  TPM2B_PRIVATE **arg11 = (TPM2B_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_Import",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Import" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Import" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Import" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Import" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Import" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Import" "', argument " "6"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg6 = (TPM2B_DATA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Import" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Import" "', argument " "8"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg8 = (TPM2B_PRIVATE *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Import" "', argument " "9"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg9 = (TPM2B_ENCRYPTED_SECRET *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Import" "', argument " "10"" of type '" "TPMT_SYM_DEF_OBJECT const *""'"); 
  }
  arg10 = (TPMT_SYM_DEF_OBJECT *)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_Import" "', argument " "11"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg11 = (TPM2B_PRIVATE **)(argp11);
  result = (TSS2_RC)Esys_Import(arg1,arg2,arg3,arg4,arg5,(TPM2B_DATA const *)arg6,(TPM2B_PUBLIC const *)arg7,(TPM2B_PRIVATE const *)arg8,(TPM2B_ENCRYPTED_SECRET const *)arg9,(TPMT_SYM_DEF_OBJECT const *)arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Import_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DATA *arg6 = (TPM2B_DATA *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_PRIVATE *arg8 = (TPM2B_PRIVATE *) 0 ;
  TPM2B_ENCRYPTED_SECRET *arg9 = (TPM2B_ENCRYPTED_SECRET *) 0 ;
  TPMT_SYM_DEF_OBJECT *arg10 = (TPMT_SYM_DEF_OBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_Import_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Import_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Import_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Import_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Import_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Import_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Import_Async" "', argument " "6"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg6 = (TPM2B_DATA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Import_Async" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Import_Async" "', argument " "8"" of type '" "TPM2B_PRIVATE const *""'"); 
  }
  arg8 = (TPM2B_PRIVATE *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_ENCRYPTED_SECRET, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Import_Async" "', argument " "9"" of type '" "TPM2B_ENCRYPTED_SECRET const *""'"); 
  }
  arg9 = (TPM2B_ENCRYPTED_SECRET *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_TPMT_SYM_DEF_OBJECT, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Import_Async" "', argument " "10"" of type '" "TPMT_SYM_DEF_OBJECT const *""'"); 
  }
  arg10 = (TPMT_SYM_DEF_OBJECT *)(argp10);
  result = (TSS2_RC)Esys_Import_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DATA const *)arg6,(TPM2B_PUBLIC const *)arg7,(TPM2B_PRIVATE const *)arg8,(TPM2B_ENCRYPTED_SECRET const *)arg9,(TPMT_SYM_DEF_OBJECT const *)arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Import_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PRIVATE **arg2 = (TPM2B_PRIVATE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Import_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Import_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PRIVATE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Import_Finish" "', argument " "2"" of type '" "TPM2B_PRIVATE **""'"); 
  }
  arg2 = (TPM2B_PRIVATE **)(argp2);
  result = (TSS2_RC)Esys_Import_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_RSA_Encrypt(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PUBLIC_KEY_RSA *arg6 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  TPMT_RSA_DECRYPT *arg7 = (TPMT_RSA_DECRYPT *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPM2B_PUBLIC_KEY_RSA **arg9 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_RSA_Encrypt",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_RSA_Encrypt" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_RSA_Encrypt" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_RSA_Encrypt" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_RSA_Encrypt" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_RSA_Encrypt" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_RSA_Encrypt" "', argument " "6"" of type '" "TPM2B_PUBLIC_KEY_RSA const *""'"); 
  }
  arg6 = (TPM2B_PUBLIC_KEY_RSA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_RSA_Encrypt" "', argument " "7"" of type '" "TPMT_RSA_DECRYPT const *""'"); 
  }
  arg7 = (TPMT_RSA_DECRYPT *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_RSA_Encrypt" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_RSA_Encrypt" "', argument " "9"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg9 = (TPM2B_PUBLIC_KEY_RSA **)(argp9);
  result = (TSS2_RC)Esys_RSA_Encrypt(arg1,arg2,arg3,arg4,arg5,(TPM2B_PUBLIC_KEY_RSA const *)arg6,(TPMT_RSA_DECRYPT const *)arg7,(TPM2B_DATA const *)arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_RSA_Encrypt_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PUBLIC_KEY_RSA *arg6 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  TPMT_RSA_DECRYPT *arg7 = (TPMT_RSA_DECRYPT *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_RSA_Encrypt_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "6"" of type '" "TPM2B_PUBLIC_KEY_RSA const *""'"); 
  }
  arg6 = (TPM2B_PUBLIC_KEY_RSA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "7"" of type '" "TPMT_RSA_DECRYPT const *""'"); 
  }
  arg7 = (TPMT_RSA_DECRYPT *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_RSA_Encrypt_Async" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  result = (TSS2_RC)Esys_RSA_Encrypt_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_PUBLIC_KEY_RSA const *)arg6,(TPMT_RSA_DECRYPT const *)arg7,(TPM2B_DATA const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_RSA_Encrypt_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PUBLIC_KEY_RSA **arg2 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_RSA_Encrypt_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_RSA_Encrypt_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_RSA_Encrypt_Finish" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg2 = (TPM2B_PUBLIC_KEY_RSA **)(argp2);
  result = (TSS2_RC)Esys_RSA_Encrypt_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_RSA_Decrypt(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PUBLIC_KEY_RSA *arg6 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  TPMT_RSA_DECRYPT *arg7 = (TPMT_RSA_DECRYPT *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPM2B_PUBLIC_KEY_RSA **arg9 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_RSA_Decrypt",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_RSA_Decrypt" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_RSA_Decrypt" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_RSA_Decrypt" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_RSA_Decrypt" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_RSA_Decrypt" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_RSA_Decrypt" "', argument " "6"" of type '" "TPM2B_PUBLIC_KEY_RSA const *""'"); 
  }
  arg6 = (TPM2B_PUBLIC_KEY_RSA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_RSA_Decrypt" "', argument " "7"" of type '" "TPMT_RSA_DECRYPT const *""'"); 
  }
  arg7 = (TPMT_RSA_DECRYPT *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_RSA_Decrypt" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_RSA_Decrypt" "', argument " "9"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg9 = (TPM2B_PUBLIC_KEY_RSA **)(argp9);
  result = (TSS2_RC)Esys_RSA_Decrypt(arg1,arg2,arg3,arg4,arg5,(TPM2B_PUBLIC_KEY_RSA const *)arg6,(TPMT_RSA_DECRYPT const *)arg7,(TPM2B_DATA const *)arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_RSA_Decrypt_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_PUBLIC_KEY_RSA *arg6 = (TPM2B_PUBLIC_KEY_RSA *) 0 ;
  TPMT_RSA_DECRYPT *arg7 = (TPMT_RSA_DECRYPT *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_RSA_Decrypt_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "6"" of type '" "TPM2B_PUBLIC_KEY_RSA const *""'"); 
  }
  arg6 = (TPM2B_PUBLIC_KEY_RSA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_RSA_DECRYPT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "7"" of type '" "TPMT_RSA_DECRYPT const *""'"); 
  }
  arg7 = (TPMT_RSA_DECRYPT *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_RSA_Decrypt_Async" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  result = (TSS2_RC)Esys_RSA_Decrypt_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_PUBLIC_KEY_RSA const *)arg6,(TPMT_RSA_DECRYPT const *)arg7,(TPM2B_DATA const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_RSA_Decrypt_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_PUBLIC_KEY_RSA **arg2 = (TPM2B_PUBLIC_KEY_RSA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_RSA_Decrypt_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_RSA_Decrypt_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_PUBLIC_KEY_RSA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_RSA_Decrypt_Finish" "', argument " "2"" of type '" "TPM2B_PUBLIC_KEY_RSA **""'"); 
  }
  arg2 = (TPM2B_PUBLIC_KEY_RSA **)(argp2);
  result = (TSS2_RC)Esys_RSA_Decrypt_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECDH_KeyGen(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT **arg6 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg7 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_ECDH_KeyGen",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECDH_KeyGen" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ECDH_KeyGen" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ECDH_KeyGen" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ECDH_KeyGen" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ECDH_KeyGen" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ECDH_KeyGen" "', argument " "6"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg6 = (TPM2B_ECC_POINT **)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ECDH_KeyGen" "', argument " "7"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg7 = (TPM2B_ECC_POINT **)(argp7);
  result = (TSS2_RC)Esys_ECDH_KeyGen(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECDH_KeyGen_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ECDH_KeyGen_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECDH_KeyGen_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ECDH_KeyGen_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ECDH_KeyGen_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ECDH_KeyGen_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ECDH_KeyGen_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_ECDH_KeyGen_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECDH_KeyGen_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ECC_POINT **arg2 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg3 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_ECDH_KeyGen_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECDH_KeyGen_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ECDH_KeyGen_Finish" "', argument " "2"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg2 = (TPM2B_ECC_POINT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_ECDH_KeyGen_Finish" "', argument " "3"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg3 = (TPM2B_ECC_POINT **)(argp3);
  result = (TSS2_RC)Esys_ECDH_KeyGen_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECDH_ZGen(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT *arg6 = (TPM2B_ECC_POINT *) 0 ;
  TPM2B_ECC_POINT **arg7 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_ECDH_ZGen",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECDH_ZGen" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ECDH_ZGen" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ECDH_ZGen" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ECDH_ZGen" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ECDH_ZGen" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ECDH_ZGen" "', argument " "6"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg6 = (TPM2B_ECC_POINT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ECDH_ZGen" "', argument " "7"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg7 = (TPM2B_ECC_POINT **)(argp7);
  result = (TSS2_RC)Esys_ECDH_ZGen(arg1,arg2,arg3,arg4,arg5,(TPM2B_ECC_POINT const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECDH_ZGen_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT *arg6 = (TPM2B_ECC_POINT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ECDH_ZGen_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECDH_ZGen_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ECDH_ZGen_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ECDH_ZGen_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ECDH_ZGen_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ECDH_ZGen_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ECDH_ZGen_Async" "', argument " "6"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg6 = (TPM2B_ECC_POINT *)(argp6);
  result = (TSS2_RC)Esys_ECDH_ZGen_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_ECC_POINT const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECDH_ZGen_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ECC_POINT **arg2 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ECDH_ZGen_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECDH_ZGen_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ECDH_ZGen_Finish" "', argument " "2"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg2 = (TPM2B_ECC_POINT **)(argp2);
  result = (TSS2_RC)Esys_ECDH_ZGen_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECC_Parameters(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMI_ECC_CURVE arg5 ;
  TPMS_ALGORITHM_DETAIL_ECC **arg6 = (TPMS_ALGORITHM_DETAIL_ECC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ECC_Parameters",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECC_Parameters" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ECC_Parameters" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ECC_Parameters" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ECC_Parameters" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ECC_Parameters" "', argument " "5"" of type '" "TPMI_ECC_CURVE""'");
  } 
  arg5 = (TPMI_ECC_CURVE)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ECC_Parameters" "', argument " "6"" of type '" "TPMS_ALGORITHM_DETAIL_ECC **""'"); 
  }
  arg6 = (TPMS_ALGORITHM_DETAIL_ECC **)(argp6);
  result = (TSS2_RC)Esys_ECC_Parameters(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECC_Parameters_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMI_ECC_CURVE arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ECC_Parameters_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECC_Parameters_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ECC_Parameters_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ECC_Parameters_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ECC_Parameters_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ECC_Parameters_Async" "', argument " "5"" of type '" "TPMI_ECC_CURVE""'");
  } 
  arg5 = (TPMI_ECC_CURVE)(val5);
  result = (TSS2_RC)Esys_ECC_Parameters_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ECC_Parameters_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMS_ALGORITHM_DETAIL_ECC **arg2 = (TPMS_ALGORITHM_DETAIL_ECC **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ECC_Parameters_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ECC_Parameters_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ECC_Parameters_Finish" "', argument " "2"" of type '" "TPMS_ALGORITHM_DETAIL_ECC **""'"); 
  }
  arg2 = (TPMS_ALGORITHM_DETAIL_ECC **)(argp2);
  result = (TSS2_RC)Esys_ECC_Parameters_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ZGen_2Phase(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT *arg6 = (TPM2B_ECC_POINT *) 0 ;
  TPM2B_ECC_POINT *arg7 = (TPM2B_ECC_POINT *) 0 ;
  TPMI_ECC_KEY_EXCHANGE arg8 ;
  UINT16 arg9 ;
  TPM2B_ECC_POINT **arg10 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg11 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  unsigned short val9 ;
  int ecode9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_ZGen_2Phase",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ZGen_2Phase" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ZGen_2Phase" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ZGen_2Phase" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ZGen_2Phase" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ZGen_2Phase" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ZGen_2Phase" "', argument " "6"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg6 = (TPM2B_ECC_POINT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ZGen_2Phase" "', argument " "7"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg7 = (TPM2B_ECC_POINT *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_ZGen_2Phase" "', argument " "8"" of type '" "TPMI_ECC_KEY_EXCHANGE""'");
  } 
  arg8 = (TPMI_ECC_KEY_EXCHANGE)(val8);
  ecode9 = SWIG_AsVal_unsigned_SS_short(obj8, &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "Esys_ZGen_2Phase" "', argument " "9"" of type '" "UINT16""'");
  } 
  arg9 = (UINT16)(val9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_ZGen_2Phase" "', argument " "10"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg10 = (TPM2B_ECC_POINT **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_ZGen_2Phase" "', argument " "11"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg11 = (TPM2B_ECC_POINT **)(argp11);
  result = (TSS2_RC)Esys_ZGen_2Phase(arg1,arg2,arg3,arg4,arg5,(TPM2B_ECC_POINT const *)arg6,(TPM2B_ECC_POINT const *)arg7,arg8,arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ZGen_2Phase_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT *arg6 = (TPM2B_ECC_POINT *) 0 ;
  TPM2B_ECC_POINT *arg7 = (TPM2B_ECC_POINT *) 0 ;
  TPMI_ECC_KEY_EXCHANGE arg8 ;
  UINT16 arg9 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  unsigned short val9 ;
  int ecode9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_ZGen_2Phase_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "6"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg6 = (TPM2B_ECC_POINT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "7"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg7 = (TPM2B_ECC_POINT *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "8"" of type '" "TPMI_ECC_KEY_EXCHANGE""'");
  } 
  arg8 = (TPMI_ECC_KEY_EXCHANGE)(val8);
  ecode9 = SWIG_AsVal_unsigned_SS_short(obj8, &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "Esys_ZGen_2Phase_Async" "', argument " "9"" of type '" "UINT16""'");
  } 
  arg9 = (UINT16)(val9);
  result = (TSS2_RC)Esys_ZGen_2Phase_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_ECC_POINT const *)arg6,(TPM2B_ECC_POINT const *)arg7,arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ZGen_2Phase_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ECC_POINT **arg2 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg3 = (TPM2B_ECC_POINT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_ZGen_2Phase_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ZGen_2Phase_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ZGen_2Phase_Finish" "', argument " "2"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg2 = (TPM2B_ECC_POINT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_ZGen_2Phase_Finish" "', argument " "3"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg3 = (TPM2B_ECC_POINT **)(argp3);
  result = (TSS2_RC)Esys_ZGen_2Phase_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EncryptDecrypt(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_YES_NO arg6 ;
  TPMI_ALG_SYM_MODE arg7 ;
  TPM2B_IV *arg8 = (TPM2B_IV *) 0 ;
  TPM2B_MAX_BUFFER *arg9 = (TPM2B_MAX_BUFFER *) 0 ;
  TPM2B_MAX_BUFFER **arg10 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2B_IV **arg11 = (TPM2B_IV **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_EncryptDecrypt",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EncryptDecrypt" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EncryptDecrypt" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EncryptDecrypt" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EncryptDecrypt" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EncryptDecrypt" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_EncryptDecrypt" "', argument " "6"" of type '" "TPMI_YES_NO""'");
  } 
  arg6 = (TPMI_YES_NO)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_EncryptDecrypt" "', argument " "7"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg7 = (TPMI_ALG_SYM_MODE)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_EncryptDecrypt" "', argument " "8"" of type '" "TPM2B_IV const *""'"); 
  }
  arg8 = (TPM2B_IV *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_EncryptDecrypt" "', argument " "9"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg9 = (TPM2B_MAX_BUFFER *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_EncryptDecrypt" "', argument " "10"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg10 = (TPM2B_MAX_BUFFER **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_EncryptDecrypt" "', argument " "11"" of type '" "TPM2B_IV **""'"); 
  }
  arg11 = (TPM2B_IV **)(argp11);
  result = (TSS2_RC)Esys_EncryptDecrypt(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_IV const *)arg8,(TPM2B_MAX_BUFFER const *)arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EncryptDecrypt_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_YES_NO arg6 ;
  TPMI_ALG_SYM_MODE arg7 ;
  TPM2B_IV *arg8 = (TPM2B_IV *) 0 ;
  TPM2B_MAX_BUFFER *arg9 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_EncryptDecrypt_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "6"" of type '" "TPMI_YES_NO""'");
  } 
  arg6 = (TPMI_YES_NO)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "7"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg7 = (TPMI_ALG_SYM_MODE)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "8"" of type '" "TPM2B_IV const *""'"); 
  }
  arg8 = (TPM2B_IV *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_EncryptDecrypt_Async" "', argument " "9"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg9 = (TPM2B_MAX_BUFFER *)(argp9);
  result = (TSS2_RC)Esys_EncryptDecrypt_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_IV const *)arg8,(TPM2B_MAX_BUFFER const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EncryptDecrypt_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_MAX_BUFFER **arg2 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2B_IV **arg3 = (TPM2B_IV **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_EncryptDecrypt_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EncryptDecrypt_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_EncryptDecrypt_Finish" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg2 = (TPM2B_MAX_BUFFER **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_EncryptDecrypt_Finish" "', argument " "3"" of type '" "TPM2B_IV **""'"); 
  }
  arg3 = (TPM2B_IV **)(argp3);
  result = (TSS2_RC)Esys_EncryptDecrypt_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EncryptDecrypt2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_YES_NO arg7 ;
  TPMI_ALG_SYM_MODE arg8 ;
  TPM2B_IV *arg9 = (TPM2B_IV *) 0 ;
  TPM2B_MAX_BUFFER **arg10 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2B_IV **arg11 = (TPM2B_IV **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned char val7 ;
  int ecode7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_EncryptDecrypt2",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EncryptDecrypt2" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EncryptDecrypt2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EncryptDecrypt2" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EncryptDecrypt2" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EncryptDecrypt2" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_EncryptDecrypt2" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_char(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_EncryptDecrypt2" "', argument " "7"" of type '" "TPMI_YES_NO""'");
  } 
  arg7 = (TPMI_YES_NO)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_EncryptDecrypt2" "', argument " "8"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg8 = (TPMI_ALG_SYM_MODE)(val8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_EncryptDecrypt2" "', argument " "9"" of type '" "TPM2B_IV const *""'"); 
  }
  arg9 = (TPM2B_IV *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_EncryptDecrypt2" "', argument " "10"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg10 = (TPM2B_MAX_BUFFER **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_EncryptDecrypt2" "', argument " "11"" of type '" "TPM2B_IV **""'"); 
  }
  arg11 = (TPM2B_IV **)(argp11);
  result = (TSS2_RC)Esys_EncryptDecrypt2(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6,arg7,arg8,(TPM2B_IV const *)arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EncryptDecrypt2_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_YES_NO arg7 ;
  TPMI_ALG_SYM_MODE arg8 ;
  TPM2B_IV *arg9 = (TPM2B_IV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned char val7 ;
  int ecode7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_EncryptDecrypt2_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_char(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "7"" of type '" "TPMI_YES_NO""'");
  } 
  arg7 = (TPMI_YES_NO)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "8"" of type '" "TPMI_ALG_SYM_MODE""'");
  } 
  arg8 = (TPMI_ALG_SYM_MODE)(val8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_EncryptDecrypt2_Async" "', argument " "9"" of type '" "TPM2B_IV const *""'"); 
  }
  arg9 = (TPM2B_IV *)(argp9);
  result = (TSS2_RC)Esys_EncryptDecrypt2_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6,arg7,arg8,(TPM2B_IV const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EncryptDecrypt2_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_MAX_BUFFER **arg2 = (TPM2B_MAX_BUFFER **) 0 ;
  TPM2B_IV **arg3 = (TPM2B_IV **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_EncryptDecrypt2_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EncryptDecrypt2_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_EncryptDecrypt2_Finish" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg2 = (TPM2B_MAX_BUFFER **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_IV, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_EncryptDecrypt2_Finish" "', argument " "3"" of type '" "TPM2B_IV **""'"); 
  }
  arg3 = (TPM2B_IV **)(argp3);
  result = (TSS2_RC)Esys_EncryptDecrypt2_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Hash(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_MAX_BUFFER *arg5 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_ALG_HASH arg6 ;
  TPMI_RH_HIERARCHY arg7 ;
  TPM2B_DIGEST **arg8 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_HASHCHECK **arg9 = (TPMT_TK_HASHCHECK **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned short val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_Hash",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Hash" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Hash" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Hash" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Hash" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_Hash" "', argument " "5"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg5 = (TPM2B_MAX_BUFFER *)(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_short(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Hash" "', argument " "6"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg6 = (TPMI_ALG_HASH)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_Hash" "', argument " "7"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg7 = (TPMI_RH_HIERARCHY)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Hash" "', argument " "8"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg8 = (TPM2B_DIGEST **)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Hash" "', argument " "9"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg9 = (TPMT_TK_HASHCHECK **)(argp9);
  result = (TSS2_RC)Esys_Hash(arg1,arg2,arg3,arg4,(TPM2B_MAX_BUFFER const *)arg5,arg6,arg7,arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Hash_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_MAX_BUFFER *arg5 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_ALG_HASH arg6 ;
  TPMI_RH_HIERARCHY arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned short val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_Hash_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Hash_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Hash_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Hash_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Hash_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_Hash_Async" "', argument " "5"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg5 = (TPM2B_MAX_BUFFER *)(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_short(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Hash_Async" "', argument " "6"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg6 = (TPMI_ALG_HASH)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_Hash_Async" "', argument " "7"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg7 = (TPMI_RH_HIERARCHY)(val7);
  result = (TSS2_RC)Esys_Hash_Async(arg1,arg2,arg3,arg4,(TPM2B_MAX_BUFFER const *)arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Hash_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DIGEST **arg2 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_HASHCHECK **arg3 = (TPMT_TK_HASHCHECK **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_Hash_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Hash_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Hash_Finish" "', argument " "2"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg2 = (TPM2B_DIGEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Hash_Finish" "', argument " "3"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg3 = (TPMT_TK_HASHCHECK **)(argp3);
  result = (TSS2_RC)Esys_Hash_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HMAC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_ALG_HASH arg7 ;
  TPM2B_DIGEST **arg8 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_HMAC",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HMAC" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HMAC" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HMAC" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HMAC" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HMAC" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_HMAC" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_HMAC" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_HMAC" "', argument " "8"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg8 = (TPM2B_DIGEST **)(argp8);
  result = (TSS2_RC)Esys_HMAC(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HMAC_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_ALG_HASH arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_HMAC_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HMAC_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HMAC_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HMAC_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HMAC_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HMAC_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_HMAC_Async" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_HMAC_Async" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  result = (TSS2_RC)Esys_HMAC_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HMAC_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DIGEST **arg2 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_HMAC_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HMAC_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_HMAC_Finish" "', argument " "2"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg2 = (TPM2B_DIGEST **)(argp2);
  result = (TSS2_RC)Esys_HMAC_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetRandom(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  UINT16 arg5 ;
  TPM2B_DIGEST **arg6 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_GetRandom",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetRandom" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetRandom" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetRandom" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetRandom" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetRandom" "', argument " "5"" of type '" "UINT16""'");
  } 
  arg5 = (UINT16)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_GetRandom" "', argument " "6"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg6 = (TPM2B_DIGEST **)(argp6);
  result = (TSS2_RC)Esys_GetRandom(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetRandom_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  UINT16 arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_GetRandom_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetRandom_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetRandom_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetRandom_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetRandom_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetRandom_Async" "', argument " "5"" of type '" "UINT16""'");
  } 
  arg5 = (UINT16)(val5);
  result = (TSS2_RC)Esys_GetRandom_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetRandom_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DIGEST **arg2 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_GetRandom_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetRandom_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetRandom_Finish" "', argument " "2"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg2 = (TPM2B_DIGEST **)(argp2);
  result = (TSS2_RC)Esys_GetRandom_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_StirRandom(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_SENSITIVE_DATA *arg5 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_StirRandom",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_StirRandom" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_StirRandom" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_StirRandom" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_StirRandom" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_StirRandom" "', argument " "5"" of type '" "TPM2B_SENSITIVE_DATA const *""'"); 
  }
  arg5 = (TPM2B_SENSITIVE_DATA *)(argp5);
  result = (TSS2_RC)Esys_StirRandom(arg1,arg2,arg3,arg4,(TPM2B_SENSITIVE_DATA const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_StirRandom_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_SENSITIVE_DATA *arg5 = (TPM2B_SENSITIVE_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_StirRandom_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_StirRandom_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_StirRandom_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_StirRandom_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_StirRandom_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_StirRandom_Async" "', argument " "5"" of type '" "TPM2B_SENSITIVE_DATA const *""'"); 
  }
  arg5 = (TPM2B_SENSITIVE_DATA *)(argp5);
  result = (TSS2_RC)Esys_StirRandom_Async(arg1,arg2,arg3,arg4,(TPM2B_SENSITIVE_DATA const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_StirRandom_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_StirRandom_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_StirRandom_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_StirRandom_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HMAC_Start(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  TPMI_ALG_HASH arg7 ;
  unsigned int *arg8 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_HMAC_Start",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HMAC_Start" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HMAC_Start" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HMAC_Start" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HMAC_Start" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HMAC_Start" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_HMAC_Start" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_HMAC_Start" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_HMAC_Start" "', argument " "8"" of type '" "unsigned int *""'"); 
  }
  arg8 = (unsigned int *)(argp8);
  result = (TSS2_RC)Esys_HMAC_Start(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HMAC_Start_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  TPMI_ALG_HASH arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_HMAC_Start_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HMAC_Start_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HMAC_Start_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HMAC_Start_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HMAC_Start_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HMAC_Start_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_HMAC_Start_Async" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_HMAC_Start_Async" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  result = (TSS2_RC)Esys_HMAC_Start_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HMAC_Start_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_HMAC_Start_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HMAC_Start_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_HMAC_Start_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_HMAC_Start_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HashSequenceStart(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_AUTH *arg5 = (TPM2B_AUTH *) 0 ;
  TPMI_ALG_HASH arg6 ;
  unsigned int *arg7 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned short val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_HashSequenceStart",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HashSequenceStart" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HashSequenceStart" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HashSequenceStart" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HashSequenceStart" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_HashSequenceStart" "', argument " "5"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg5 = (TPM2B_AUTH *)(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_short(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_HashSequenceStart" "', argument " "6"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg6 = (TPMI_ALG_HASH)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_HashSequenceStart" "', argument " "7"" of type '" "unsigned int *""'"); 
  }
  arg7 = (unsigned int *)(argp7);
  result = (TSS2_RC)Esys_HashSequenceStart(arg1,arg2,arg3,arg4,(TPM2B_DIGEST const *)arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HashSequenceStart_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_AUTH *arg5 = (TPM2B_AUTH *) 0 ;
  TPMI_ALG_HASH arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned short val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_HashSequenceStart_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HashSequenceStart_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HashSequenceStart_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HashSequenceStart_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HashSequenceStart_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_HashSequenceStart_Async" "', argument " "5"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg5 = (TPM2B_AUTH *)(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_short(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_HashSequenceStart_Async" "', argument " "6"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg6 = (TPMI_ALG_HASH)(val6);
  result = (TSS2_RC)Esys_HashSequenceStart_Async(arg1,arg2,arg3,arg4,(TPM2B_DIGEST const *)arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HashSequenceStart_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_HashSequenceStart_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HashSequenceStart_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_HashSequenceStart_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_HashSequenceStart_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SequenceUpdate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_SequenceUpdate",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SequenceUpdate" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SequenceUpdate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SequenceUpdate" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SequenceUpdate" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SequenceUpdate" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_SequenceUpdate" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  result = (TSS2_RC)Esys_SequenceUpdate(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SequenceUpdate_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_SequenceUpdate_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SequenceUpdate_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SequenceUpdate_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SequenceUpdate_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SequenceUpdate_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SequenceUpdate_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_SequenceUpdate_Async" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  result = (TSS2_RC)Esys_SequenceUpdate_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SequenceUpdate_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_SequenceUpdate_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SequenceUpdate_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_SequenceUpdate_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SequenceComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_RH_HIERARCHY arg7 ;
  TPM2B_DIGEST **arg8 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_HASHCHECK **arg9 = (TPMT_TK_HASHCHECK **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_SequenceComplete",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SequenceComplete" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SequenceComplete" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SequenceComplete" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SequenceComplete" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SequenceComplete" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_SequenceComplete" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_SequenceComplete" "', argument " "7"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg7 = (TPMI_RH_HIERARCHY)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_SequenceComplete" "', argument " "8"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg8 = (TPM2B_DIGEST **)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_SequenceComplete" "', argument " "9"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg9 = (TPMT_TK_HASHCHECK **)(argp9);
  result = (TSS2_RC)Esys_SequenceComplete(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6,arg7,arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SequenceComplete_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_MAX_BUFFER *arg6 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMI_RH_HIERARCHY arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_SequenceComplete_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SequenceComplete_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SequenceComplete_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SequenceComplete_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SequenceComplete_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SequenceComplete_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_SequenceComplete_Async" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_SequenceComplete_Async" "', argument " "7"" of type '" "TPMI_RH_HIERARCHY""'");
  } 
  arg7 = (TPMI_RH_HIERARCHY)(val7);
  result = (TSS2_RC)Esys_SequenceComplete_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_MAX_BUFFER const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SequenceComplete_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DIGEST **arg2 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_HASHCHECK **arg3 = (TPMT_TK_HASHCHECK **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_SequenceComplete_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SequenceComplete_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_SequenceComplete_Finish" "', argument " "2"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg2 = (TPM2B_DIGEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_SequenceComplete_Finish" "', argument " "3"" of type '" "TPMT_TK_HASHCHECK **""'"); 
  }
  arg3 = (TPMT_TK_HASHCHECK **)(argp3);
  result = (TSS2_RC)Esys_SequenceComplete_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EventSequenceComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_MAX_BUFFER *arg7 = (TPM2B_MAX_BUFFER *) 0 ;
  TPML_DIGEST_VALUES **arg8 = (TPML_DIGEST_VALUES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_EventSequenceComplete",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EventSequenceComplete" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EventSequenceComplete" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EventSequenceComplete" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EventSequenceComplete" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EventSequenceComplete" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_EventSequenceComplete" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_EventSequenceComplete" "', argument " "7"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg7 = (TPM2B_MAX_BUFFER *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_EventSequenceComplete" "', argument " "8"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg8 = (TPML_DIGEST_VALUES **)(argp8);
  result = (TSS2_RC)Esys_EventSequenceComplete(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_MAX_BUFFER const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EventSequenceComplete_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_MAX_BUFFER *arg7 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_EventSequenceComplete_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_EventSequenceComplete_Async" "', argument " "7"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg7 = (TPM2B_MAX_BUFFER *)(argp7);
  result = (TSS2_RC)Esys_EventSequenceComplete_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_MAX_BUFFER const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EventSequenceComplete_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPML_DIGEST_VALUES **arg2 = (TPML_DIGEST_VALUES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_EventSequenceComplete_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EventSequenceComplete_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_EventSequenceComplete_Finish" "', argument " "2"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg2 = (TPML_DIGEST_VALUES **)(argp2);
  result = (TSS2_RC)Esys_EventSequenceComplete_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Certify(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg8 = (TPMT_SIG_SCHEME *) 0 ;
  TPM2B_ATTEST **arg9 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg10 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_Certify",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Certify" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Certify" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Certify" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Certify" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Certify" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Certify" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Certify" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Certify" "', argument " "8"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg8 = (TPMT_SIG_SCHEME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Certify" "', argument " "9"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg9 = (TPM2B_ATTEST **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Certify" "', argument " "10"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg10 = (TPMT_SIGNATURE **)(argp10);
  result = (TSS2_RC)Esys_Certify(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SIG_SCHEME const *)arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Certify_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg8 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_Certify_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Certify_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Certify_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Certify_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Certify_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Certify_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_Certify_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Certify_Async" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Certify_Async" "', argument " "8"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg8 = (TPMT_SIG_SCHEME *)(argp8);
  result = (TSS2_RC)Esys_Certify_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SIG_SCHEME const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Certify_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_Certify_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Certify_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Certify_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Certify_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_Certify_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CertifyCreation(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPM2B_DIGEST *arg8 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIG_SCHEME *arg9 = (TPMT_SIG_SCHEME *) 0 ;
  TPMT_TK_CREATION *arg10 = (TPMT_TK_CREATION *) 0 ;
  TPM2B_ATTEST **arg11 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg12 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOO:Esys_CertifyCreation",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CertifyCreation" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_CertifyCreation" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_CertifyCreation" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_CertifyCreation" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_CertifyCreation" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_CertifyCreation" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_CertifyCreation" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_CertifyCreation" "', argument " "8"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg8 = (TPM2B_DIGEST *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_CertifyCreation" "', argument " "9"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg9 = (TPMT_SIG_SCHEME *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_CertifyCreation" "', argument " "10"" of type '" "TPMT_TK_CREATION const *""'"); 
  }
  arg10 = (TPMT_TK_CREATION *)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_CertifyCreation" "', argument " "11"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg11 = (TPM2B_ATTEST **)(argp11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_CertifyCreation" "', argument " "12"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg12 = (TPMT_SIGNATURE **)(argp12);
  result = (TSS2_RC)Esys_CertifyCreation(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPM2B_DIGEST const *)arg8,(TPMT_SIG_SCHEME const *)arg9,(TPMT_TK_CREATION const *)arg10,arg11,arg12);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CertifyCreation_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPM2B_DIGEST *arg8 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIG_SCHEME *arg9 = (TPMT_SIG_SCHEME *) 0 ;
  TPMT_TK_CREATION *arg10 = (TPMT_TK_CREATION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_CertifyCreation_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CertifyCreation_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_CertifyCreation_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_CertifyCreation_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_CertifyCreation_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_CertifyCreation_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_CertifyCreation_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_CertifyCreation_Async" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_CertifyCreation_Async" "', argument " "8"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg8 = (TPM2B_DIGEST *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_CertifyCreation_Async" "', argument " "9"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg9 = (TPMT_SIG_SCHEME *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_CertifyCreation_Async" "', argument " "10"" of type '" "TPMT_TK_CREATION const *""'"); 
  }
  arg10 = (TPMT_TK_CREATION *)(argp10);
  result = (TSS2_RC)Esys_CertifyCreation_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPM2B_DIGEST const *)arg8,(TPMT_SIG_SCHEME const *)arg9,(TPMT_TK_CREATION const *)arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CertifyCreation_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_CertifyCreation_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CertifyCreation_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_CertifyCreation_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_CertifyCreation_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_CertifyCreation_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Quote(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DATA *arg6 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg7 = (TPMT_SIG_SCHEME *) 0 ;
  TPML_PCR_SELECTION *arg8 = (TPML_PCR_SELECTION *) 0 ;
  TPM2B_ATTEST **arg9 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg10 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_Quote",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Quote" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Quote" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Quote" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Quote" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Quote" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Quote" "', argument " "6"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg6 = (TPM2B_DATA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Quote" "', argument " "7"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg7 = (TPMT_SIG_SCHEME *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Quote" "', argument " "8"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg8 = (TPML_PCR_SELECTION *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Quote" "', argument " "9"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg9 = (TPM2B_ATTEST **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Quote" "', argument " "10"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg10 = (TPMT_SIGNATURE **)(argp10);
  result = (TSS2_RC)Esys_Quote(arg1,arg2,arg3,arg4,arg5,(TPM2B_DATA const *)arg6,(TPMT_SIG_SCHEME const *)arg7,(TPML_PCR_SELECTION const *)arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Quote_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DATA *arg6 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg7 = (TPMT_SIG_SCHEME *) 0 ;
  TPML_PCR_SELECTION *arg8 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_Quote_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Quote_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Quote_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Quote_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Quote_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Quote_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Quote_Async" "', argument " "6"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg6 = (TPM2B_DATA *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Quote_Async" "', argument " "7"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg7 = (TPMT_SIG_SCHEME *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Quote_Async" "', argument " "8"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg8 = (TPML_PCR_SELECTION *)(argp8);
  result = (TSS2_RC)Esys_Quote_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DATA const *)arg6,(TPMT_SIG_SCHEME const *)arg7,(TPML_PCR_SELECTION const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Quote_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_Quote_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Quote_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Quote_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Quote_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_Quote_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetSessionAuditDigest(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg9 = (TPMT_SIG_SCHEME *) 0 ;
  TPM2B_ATTEST **arg10 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg11 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_GetSessionAuditDigest",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetSessionAuditDigest" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetSessionAuditDigest" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetSessionAuditDigest" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetSessionAuditDigest" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetSessionAuditDigest" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetSessionAuditDigest" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_GetSessionAuditDigest" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetSessionAuditDigest" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_GetSessionAuditDigest" "', argument " "9"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg9 = (TPMT_SIG_SCHEME *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_GetSessionAuditDigest" "', argument " "10"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg10 = (TPM2B_ATTEST **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_GetSessionAuditDigest" "', argument " "11"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg11 = (TPMT_SIGNATURE **)(argp11);
  result = (TSS2_RC)Esys_GetSessionAuditDigest(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_DATA const *)arg8,(TPMT_SIG_SCHEME const *)arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetSessionAuditDigest_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg9 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_GetSessionAuditDigest_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_GetSessionAuditDigest_Async" "', argument " "9"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg9 = (TPMT_SIG_SCHEME *)(argp9);
  result = (TSS2_RC)Esys_GetSessionAuditDigest_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_DATA const *)arg8,(TPMT_SIG_SCHEME const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetSessionAuditDigest_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_GetSessionAuditDigest_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetSessionAuditDigest_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetSessionAuditDigest_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_GetSessionAuditDigest_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_GetSessionAuditDigest_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetCommandAuditDigest(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg8 = (TPMT_SIG_SCHEME *) 0 ;
  TPM2B_ATTEST **arg9 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg10 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_GetCommandAuditDigest",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetCommandAuditDigest" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetCommandAuditDigest" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetCommandAuditDigest" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetCommandAuditDigest" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetCommandAuditDigest" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetCommandAuditDigest" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_GetCommandAuditDigest" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetCommandAuditDigest" "', argument " "8"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg8 = (TPMT_SIG_SCHEME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_GetCommandAuditDigest" "', argument " "9"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg9 = (TPM2B_ATTEST **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_GetCommandAuditDigest" "', argument " "10"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg10 = (TPMT_SIGNATURE **)(argp10);
  result = (TSS2_RC)Esys_GetCommandAuditDigest(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SIG_SCHEME const *)arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetCommandAuditDigest_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg8 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_GetCommandAuditDigest_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetCommandAuditDigest_Async" "', argument " "8"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg8 = (TPMT_SIG_SCHEME *)(argp8);
  result = (TSS2_RC)Esys_GetCommandAuditDigest_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SIG_SCHEME const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetCommandAuditDigest_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_GetCommandAuditDigest_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetCommandAuditDigest_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetCommandAuditDigest_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_GetCommandAuditDigest_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_GetCommandAuditDigest_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg8 = (TPMT_SIG_SCHEME *) 0 ;
  TPM2B_ATTEST **arg9 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg10 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_GetTime",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTime" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetTime" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetTime" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetTime" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetTime" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetTime" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_GetTime" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetTime" "', argument " "8"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg8 = (TPMT_SIG_SCHEME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_GetTime" "', argument " "9"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg9 = (TPM2B_ATTEST **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_GetTime" "', argument " "10"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg10 = (TPMT_SIGNATURE **)(argp10);
  result = (TSS2_RC)Esys_GetTime(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SIG_SCHEME const *)arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTime_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DATA *arg7 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg8 = (TPMT_SIG_SCHEME *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_GetTime_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTime_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetTime_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetTime_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetTime_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetTime_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetTime_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_GetTime_Async" "', argument " "7"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg7 = (TPM2B_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetTime_Async" "', argument " "8"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg8 = (TPMT_SIG_SCHEME *)(argp8);
  result = (TSS2_RC)Esys_GetTime_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DATA const *)arg7,(TPMT_SIG_SCHEME const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetTime_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_GetTime_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetTime_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetTime_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_GetTime_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_GetTime_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Commit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT *arg6 = (TPM2B_ECC_POINT *) 0 ;
  TPM2B_SENSITIVE_DATA *arg7 = (TPM2B_SENSITIVE_DATA *) 0 ;
  TPM2B_ECC_PARAMETER *arg8 = (TPM2B_ECC_PARAMETER *) 0 ;
  TPM2B_ECC_POINT **arg9 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg10 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg11 = (TPM2B_ECC_POINT **) 0 ;
  UINT16 *arg12 = (UINT16 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOO:Esys_Commit",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Commit" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Commit" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Commit" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Commit" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Commit" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Commit" "', argument " "6"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg6 = (TPM2B_ECC_POINT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Commit" "', argument " "7"" of type '" "TPM2B_SENSITIVE_DATA const *""'"); 
  }
  arg7 = (TPM2B_SENSITIVE_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Commit" "', argument " "8"" of type '" "TPM2B_ECC_PARAMETER const *""'"); 
  }
  arg8 = (TPM2B_ECC_PARAMETER *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Commit" "', argument " "9"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg9 = (TPM2B_ECC_POINT **)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_Commit" "', argument " "10"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg10 = (TPM2B_ECC_POINT **)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_Commit" "', argument " "11"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg11 = (TPM2B_ECC_POINT **)(argp11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_Commit" "', argument " "12"" of type '" "UINT16 *""'"); 
  }
  arg12 = (UINT16 *)(argp12);
  result = (TSS2_RC)Esys_Commit(arg1,arg2,arg3,arg4,arg5,(TPM2B_ECC_POINT const *)arg6,(TPM2B_SENSITIVE_DATA const *)arg7,(TPM2B_ECC_PARAMETER const *)arg8,arg9,arg10,arg11,arg12);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Commit_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_ECC_POINT *arg6 = (TPM2B_ECC_POINT *) 0 ;
  TPM2B_SENSITIVE_DATA *arg7 = (TPM2B_SENSITIVE_DATA *) 0 ;
  TPM2B_ECC_PARAMETER *arg8 = (TPM2B_ECC_PARAMETER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_Commit_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Commit_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Commit_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Commit_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Commit_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Commit_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Commit_Async" "', argument " "6"" of type '" "TPM2B_ECC_POINT const *""'"); 
  }
  arg6 = (TPM2B_ECC_POINT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_SENSITIVE_DATA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Commit_Async" "', argument " "7"" of type '" "TPM2B_SENSITIVE_DATA const *""'"); 
  }
  arg7 = (TPM2B_SENSITIVE_DATA *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_ECC_PARAMETER, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Commit_Async" "', argument " "8"" of type '" "TPM2B_ECC_PARAMETER const *""'"); 
  }
  arg8 = (TPM2B_ECC_PARAMETER *)(argp8);
  result = (TSS2_RC)Esys_Commit_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_ECC_POINT const *)arg6,(TPM2B_SENSITIVE_DATA const *)arg7,(TPM2B_ECC_PARAMETER const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Commit_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ECC_POINT **arg2 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg3 = (TPM2B_ECC_POINT **) 0 ;
  TPM2B_ECC_POINT **arg4 = (TPM2B_ECC_POINT **) 0 ;
  UINT16 *arg5 = (UINT16 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Commit_Finish",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Commit_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Commit_Finish" "', argument " "2"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg2 = (TPM2B_ECC_POINT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_Commit_Finish" "', argument " "3"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg3 = (TPM2B_ECC_POINT **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_Commit_Finish" "', argument " "4"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg4 = (TPM2B_ECC_POINT **)(argp4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_Commit_Finish" "', argument " "5"" of type '" "UINT16 *""'"); 
  }
  arg5 = (UINT16 *)(argp5);
  result = (TSS2_RC)Esys_Commit_Finish(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EC_Ephemeral(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMI_ECC_CURVE arg5 ;
  TPM2B_ECC_POINT **arg6 = (TPM2B_ECC_POINT **) 0 ;
  UINT16 *arg7 = (UINT16 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_EC_Ephemeral",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EC_Ephemeral" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EC_Ephemeral" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EC_Ephemeral" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EC_Ephemeral" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EC_Ephemeral" "', argument " "5"" of type '" "TPMI_ECC_CURVE""'");
  } 
  arg5 = (TPMI_ECC_CURVE)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_EC_Ephemeral" "', argument " "6"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg6 = (TPM2B_ECC_POINT **)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_EC_Ephemeral" "', argument " "7"" of type '" "UINT16 *""'"); 
  }
  arg7 = (UINT16 *)(argp7);
  result = (TSS2_RC)Esys_EC_Ephemeral(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EC_Ephemeral_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMI_ECC_CURVE arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned short val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_EC_Ephemeral_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EC_Ephemeral_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EC_Ephemeral_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EC_Ephemeral_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EC_Ephemeral_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_short(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EC_Ephemeral_Async" "', argument " "5"" of type '" "TPMI_ECC_CURVE""'");
  } 
  arg5 = (TPMI_ECC_CURVE)(val5);
  result = (TSS2_RC)Esys_EC_Ephemeral_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EC_Ephemeral_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ECC_POINT **arg2 = (TPM2B_ECC_POINT **) 0 ;
  UINT16 *arg3 = (UINT16 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_EC_Ephemeral_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EC_Ephemeral_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ECC_POINT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_EC_Ephemeral_Finish" "', argument " "2"" of type '" "TPM2B_ECC_POINT **""'"); 
  }
  arg2 = (TPM2B_ECC_POINT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_EC_Ephemeral_Finish" "', argument " "3"" of type '" "UINT16 *""'"); 
  }
  arg3 = (UINT16 *)(argp3);
  result = (TSS2_RC)Esys_EC_Ephemeral_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_VerifySignature(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIGNATURE *arg7 = (TPMT_SIGNATURE *) 0 ;
  TPMT_TK_VERIFIED **arg8 = (TPMT_TK_VERIFIED **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_VerifySignature",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_VerifySignature" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_VerifySignature" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_VerifySignature" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_VerifySignature" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_VerifySignature" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_VerifySignature" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_VerifySignature" "', argument " "7"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg7 = (TPMT_SIGNATURE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_VerifySignature" "', argument " "8"" of type '" "TPMT_TK_VERIFIED **""'"); 
  }
  arg8 = (TPMT_TK_VERIFIED **)(argp8);
  result = (TSS2_RC)Esys_VerifySignature(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPMT_SIGNATURE const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_VerifySignature_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIGNATURE *arg7 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_VerifySignature_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_VerifySignature_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_VerifySignature_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_VerifySignature_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_VerifySignature_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_VerifySignature_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_VerifySignature_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_VerifySignature_Async" "', argument " "7"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg7 = (TPMT_SIGNATURE *)(argp7);
  result = (TSS2_RC)Esys_VerifySignature_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPMT_SIGNATURE const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_VerifySignature_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMT_TK_VERIFIED **arg2 = (TPMT_TK_VERIFIED **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_VerifySignature_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_VerifySignature_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_VerifySignature_Finish" "', argument " "2"" of type '" "TPMT_TK_VERIFIED **""'"); 
  }
  arg2 = (TPMT_TK_VERIFIED **)(argp2);
  result = (TSS2_RC)Esys_VerifySignature_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Sign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIG_SCHEME *arg7 = (TPMT_SIG_SCHEME *) 0 ;
  TPMT_TK_HASHCHECK *arg8 = (TPMT_TK_HASHCHECK *) 0 ;
  TPMT_SIGNATURE **arg9 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_Sign",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Sign" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Sign" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Sign" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Sign" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Sign" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Sign" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Sign" "', argument " "7"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg7 = (TPMT_SIG_SCHEME *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Sign" "', argument " "8"" of type '" "TPMT_TK_HASHCHECK const *""'"); 
  }
  arg8 = (TPMT_TK_HASHCHECK *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_Sign" "', argument " "9"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg9 = (TPMT_SIGNATURE **)(argp9);
  result = (TSS2_RC)Esys_Sign(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPMT_SIG_SCHEME const *)arg7,(TPMT_TK_HASHCHECK const *)arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Sign_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIG_SCHEME *arg7 = (TPMT_SIG_SCHEME *) 0 ;
  TPMT_TK_HASHCHECK *arg8 = (TPMT_TK_HASHCHECK *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_Sign_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Sign_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Sign_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Sign_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Sign_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Sign_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Sign_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_Sign_Async" "', argument " "7"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg7 = (TPMT_SIG_SCHEME *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_TK_HASHCHECK, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_Sign_Async" "', argument " "8"" of type '" "TPMT_TK_HASHCHECK const *""'"); 
  }
  arg8 = (TPMT_TK_HASHCHECK *)(argp8);
  result = (TSS2_RC)Esys_Sign_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPMT_SIG_SCHEME const *)arg7,(TPMT_TK_HASHCHECK const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Sign_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMT_SIGNATURE **arg2 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Sign_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Sign_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Sign_Finish" "', argument " "2"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg2 = (TPMT_SIGNATURE **)(argp2);
  result = (TSS2_RC)Esys_Sign_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetCommandCodeAuditStatus(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_ALG_HASH arg6 ;
  TPML_CC *arg7 = (TPML_CC *) 0 ;
  TPML_CC *arg8 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned short val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_SetCommandCodeAuditStatus",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_short(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "6"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg6 = (TPMI_ALG_HASH)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "7"" of type '" "TPML_CC const *""'"); 
  }
  arg7 = (TPML_CC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_SetCommandCodeAuditStatus" "', argument " "8"" of type '" "TPML_CC const *""'"); 
  }
  arg8 = (TPML_CC *)(argp8);
  result = (TSS2_RC)Esys_SetCommandCodeAuditStatus(arg1,arg2,arg3,arg4,arg5,arg6,(TPML_CC const *)arg7,(TPML_CC const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetCommandCodeAuditStatus_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_ALG_HASH arg6 ;
  TPML_CC *arg7 = (TPML_CC *) 0 ;
  TPML_CC *arg8 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned short val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_SetCommandCodeAuditStatus_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_short(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "6"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg6 = (TPMI_ALG_HASH)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "7"" of type '" "TPML_CC const *""'"); 
  }
  arg7 = (TPML_CC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_SetCommandCodeAuditStatus_Async" "', argument " "8"" of type '" "TPML_CC const *""'"); 
  }
  arg8 = (TPML_CC *)(argp8);
  result = (TSS2_RC)Esys_SetCommandCodeAuditStatus_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPML_CC const *)arg7,(TPML_CC const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetCommandCodeAuditStatus_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_SetCommandCodeAuditStatus_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetCommandCodeAuditStatus_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_SetCommandCodeAuditStatus_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Extend(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_DIGEST_VALUES *arg6 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PCR_Extend",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Extend" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Extend" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Extend" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Extend" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Extend" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Extend" "', argument " "6"" of type '" "TPML_DIGEST_VALUES const *""'"); 
  }
  arg6 = (TPML_DIGEST_VALUES *)(argp6);
  result = (TSS2_RC)Esys_PCR_Extend(arg1,arg2,arg3,arg4,arg5,(TPML_DIGEST_VALUES const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Extend_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_DIGEST_VALUES *arg6 = (TPML_DIGEST_VALUES *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PCR_Extend_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Extend_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Extend_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Extend_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Extend_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Extend_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Extend_Async" "', argument " "6"" of type '" "TPML_DIGEST_VALUES const *""'"); 
  }
  arg6 = (TPML_DIGEST_VALUES *)(argp6);
  result = (TSS2_RC)Esys_PCR_Extend_Async(arg1,arg2,arg3,arg4,arg5,(TPML_DIGEST_VALUES const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Extend_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PCR_Extend_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Extend_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PCR_Extend_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Event(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_EVENT *arg6 = (TPM2B_EVENT *) 0 ;
  TPML_DIGEST_VALUES **arg7 = (TPML_DIGEST_VALUES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PCR_Event",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Event" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Event" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Event" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Event" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Event" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Event" "', argument " "6"" of type '" "TPM2B_EVENT const *""'"); 
  }
  arg6 = (TPM2B_EVENT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PCR_Event" "', argument " "7"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg7 = (TPML_DIGEST_VALUES **)(argp7);
  result = (TSS2_RC)Esys_PCR_Event(arg1,arg2,arg3,arg4,arg5,(TPM2B_EVENT const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Event_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_EVENT *arg6 = (TPM2B_EVENT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PCR_Event_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Event_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Event_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Event_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Event_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Event_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_EVENT, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Event_Async" "', argument " "6"" of type '" "TPM2B_EVENT const *""'"); 
  }
  arg6 = (TPM2B_EVENT *)(argp6);
  result = (TSS2_RC)Esys_PCR_Event_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_EVENT const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Event_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPML_DIGEST_VALUES **arg2 = (TPML_DIGEST_VALUES **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_PCR_Event_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Event_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPML_DIGEST_VALUES, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_PCR_Event_Finish" "', argument " "2"" of type '" "TPML_DIGEST_VALUES **""'"); 
  }
  arg2 = (TPML_DIGEST_VALUES **)(argp2);
  result = (TSS2_RC)Esys_PCR_Event_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Read(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPML_PCR_SELECTION *arg5 = (TPML_PCR_SELECTION *) 0 ;
  UINT32 *arg6 = (UINT32 *) 0 ;
  TPML_PCR_SELECTION **arg7 = (TPML_PCR_SELECTION **) 0 ;
  TPML_DIGEST **arg8 = (TPML_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PCR_Read",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Read" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Read" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Read" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Read" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_PCR_Read" "', argument " "5"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg5 = (TPML_PCR_SELECTION *)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Read" "', argument " "6"" of type '" "UINT32 *""'"); 
  }
  arg6 = (UINT32 *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PCR_Read" "', argument " "7"" of type '" "TPML_PCR_SELECTION **""'"); 
  }
  arg7 = (TPML_PCR_SELECTION **)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PCR_Read" "', argument " "8"" of type '" "TPML_DIGEST **""'"); 
  }
  arg8 = (TPML_DIGEST **)(argp8);
  result = (TSS2_RC)Esys_PCR_Read(arg1,arg2,arg3,arg4,(TPML_PCR_SELECTION const *)arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Read_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPML_PCR_SELECTION *arg5 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PCR_Read_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Read_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Read_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Read_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Read_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_PCR_Read_Async" "', argument " "5"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg5 = (TPML_PCR_SELECTION *)(argp5);
  result = (TSS2_RC)Esys_PCR_Read_Async(arg1,arg2,arg3,arg4,(TPML_PCR_SELECTION const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Read_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  UINT32 *arg2 = (UINT32 *) 0 ;
  TPML_PCR_SELECTION **arg3 = (TPML_PCR_SELECTION **) 0 ;
  TPML_DIGEST **arg4 = (TPML_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_PCR_Read_Finish",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Read_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_PCR_Read_Finish" "', argument " "2"" of type '" "UINT32 *""'"); 
  }
  arg2 = (UINT32 *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_PCR_Read_Finish" "', argument " "3"" of type '" "TPML_PCR_SELECTION **""'"); 
  }
  arg3 = (TPML_PCR_SELECTION **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_PCR_Read_Finish" "', argument " "4"" of type '" "TPML_DIGEST **""'"); 
  }
  arg4 = (TPML_DIGEST **)(argp4);
  result = (TSS2_RC)Esys_PCR_Read_Finish(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Allocate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_PCR_SELECTION *arg6 = (TPML_PCR_SELECTION *) 0 ;
  TPMI_YES_NO *arg7 = (TPMI_YES_NO *) 0 ;
  UINT32 *arg8 = (UINT32 *) 0 ;
  UINT32 *arg9 = (UINT32 *) 0 ;
  UINT32 *arg10 = (UINT32 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_PCR_Allocate",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Allocate" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Allocate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Allocate" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Allocate" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Allocate" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Allocate" "', argument " "6"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg6 = (TPML_PCR_SELECTION *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PCR_Allocate" "', argument " "7"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg7 = (TPMI_YES_NO *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PCR_Allocate" "', argument " "8"" of type '" "UINT32 *""'"); 
  }
  arg8 = (UINT32 *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PCR_Allocate" "', argument " "9"" of type '" "UINT32 *""'"); 
  }
  arg9 = (UINT32 *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_PCR_Allocate" "', argument " "10"" of type '" "UINT32 *""'"); 
  }
  arg10 = (UINT32 *)(argp10);
  result = (TSS2_RC)Esys_PCR_Allocate(arg1,arg2,arg3,arg4,arg5,(TPML_PCR_SELECTION const *)arg6,arg7,arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Allocate_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_PCR_SELECTION *arg6 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PCR_Allocate_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Allocate_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Allocate_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Allocate_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Allocate_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Allocate_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_Allocate_Async" "', argument " "6"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg6 = (TPML_PCR_SELECTION *)(argp6);
  result = (TSS2_RC)Esys_PCR_Allocate_Async(arg1,arg2,arg3,arg4,arg5,(TPML_PCR_SELECTION const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Allocate_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMI_YES_NO *arg2 = (TPMI_YES_NO *) 0 ;
  UINT32 *arg3 = (UINT32 *) 0 ;
  UINT32 *arg4 = (UINT32 *) 0 ;
  UINT32 *arg5 = (UINT32 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PCR_Allocate_Finish",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Allocate_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_PCR_Allocate_Finish" "', argument " "2"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg2 = (TPMI_YES_NO *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_PCR_Allocate_Finish" "', argument " "3"" of type '" "UINT32 *""'"); 
  }
  arg3 = (UINT32 *)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_PCR_Allocate_Finish" "', argument " "4"" of type '" "UINT32 *""'"); 
  }
  arg4 = (UINT32 *)(argp4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_PCR_Allocate_Finish" "', argument " "5"" of type '" "UINT32 *""'"); 
  }
  arg5 = (UINT32 *)(argp5);
  result = (TSS2_RC)Esys_PCR_Allocate_Finish(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_SetAuthPolicy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMI_ALG_HASH arg7 ;
  TPMI_DH_PCR arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PCR_SetAuthPolicy",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_PCR_SetAuthPolicy" "', argument " "8"" of type '" "TPMI_DH_PCR""'");
  } 
  arg8 = (TPMI_DH_PCR)(val8);
  result = (TSS2_RC)Esys_PCR_SetAuthPolicy(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_SetAuthPolicy_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMI_ALG_HASH arg7 ;
  TPMI_DH_PCR arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PCR_SetAuthPolicy_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_PCR_SetAuthPolicy_Async" "', argument " "8"" of type '" "TPMI_DH_PCR""'");
  } 
  arg8 = (TPMI_DH_PCR)(val8);
  result = (TSS2_RC)Esys_PCR_SetAuthPolicy_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_SetAuthPolicy_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PCR_SetAuthPolicy_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_SetAuthPolicy_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PCR_SetAuthPolicy_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_SetAuthValue(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PCR_SetAuthValue",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_SetAuthValue" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_SetAuthValue" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_SetAuthValue" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_SetAuthValue" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_SetAuthValue" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_SetAuthValue" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PCR_SetAuthValue(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_SetAuthValue_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PCR_SetAuthValue_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_SetAuthValue_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_SetAuthValue_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_SetAuthValue_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_SetAuthValue_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_SetAuthValue_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PCR_SetAuthValue_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PCR_SetAuthValue_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_SetAuthValue_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PCR_SetAuthValue_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_SetAuthValue_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PCR_SetAuthValue_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Reset(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PCR_Reset",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Reset" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Reset" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Reset" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Reset" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Reset" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PCR_Reset(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Reset_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PCR_Reset_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Reset_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PCR_Reset_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PCR_Reset_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PCR_Reset_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PCR_Reset_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PCR_Reset_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PCR_Reset_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PCR_Reset_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PCR_Reset_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PCR_Reset_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicySigned(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2B_DIGEST *arg8 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg9 = (TPM2B_NONCE *) 0 ;
  INT32 arg10 ;
  TPMT_SIGNATURE *arg11 = (TPMT_SIGNATURE *) 0 ;
  TPM2B_TIMEOUT **arg12 = (TPM2B_TIMEOUT **) 0 ;
  TPMT_TK_AUTH **arg13 = (TPMT_TK_AUTH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOOO:Esys_PolicySigned",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&obj12)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicySigned" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicySigned" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicySigned" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicySigned" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicySigned" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicySigned" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicySigned" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicySigned" "', argument " "8"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg8 = (TPM2B_DIGEST *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicySigned" "', argument " "9"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg9 = (TPM2B_NONCE *)(argp9);
  ecode10 = SWIG_AsVal_int(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_PolicySigned" "', argument " "10"" of type '" "INT32""'");
  } 
  arg10 = (INT32)(val10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_PolicySigned" "', argument " "11"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg11 = (TPMT_SIGNATURE *)(argp11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_PolicySigned" "', argument " "12"" of type '" "TPM2B_TIMEOUT **""'"); 
  }
  arg12 = (TPM2B_TIMEOUT **)(argp12);
  res13 = SWIG_ConvertPtr(obj12, &argp13,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), "in method '" "Esys_PolicySigned" "', argument " "13"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg13 = (TPMT_TK_AUTH **)(argp13);
  result = (TSS2_RC)Esys_PolicySigned(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_DIGEST const *)arg8,(TPM2B_DIGEST const *)arg9,arg10,(TPMT_SIGNATURE const *)arg11,arg12,arg13);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicySigned_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2B_DIGEST *arg8 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg9 = (TPM2B_NONCE *) 0 ;
  INT32 arg10 ;
  TPMT_SIGNATURE *arg11 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_PolicySigned_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicySigned_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicySigned_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicySigned_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicySigned_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicySigned_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicySigned_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicySigned_Async" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicySigned_Async" "', argument " "8"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg8 = (TPM2B_DIGEST *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicySigned_Async" "', argument " "9"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg9 = (TPM2B_NONCE *)(argp9);
  ecode10 = SWIG_AsVal_int(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_PolicySigned_Async" "', argument " "10"" of type '" "INT32""'");
  } 
  arg10 = (INT32)(val10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_PolicySigned_Async" "', argument " "11"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg11 = (TPMT_SIGNATURE *)(argp11);
  result = (TSS2_RC)Esys_PolicySigned_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_DIGEST const *)arg8,(TPM2B_DIGEST const *)arg9,arg10,(TPMT_SIGNATURE const *)arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicySigned_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_TIMEOUT **arg2 = (TPM2B_TIMEOUT **) 0 ;
  TPMT_TK_AUTH **arg3 = (TPMT_TK_AUTH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_PolicySigned_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicySigned_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_PolicySigned_Finish" "', argument " "2"" of type '" "TPM2B_TIMEOUT **""'"); 
  }
  arg2 = (TPM2B_TIMEOUT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_PolicySigned_Finish" "', argument " "3"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg3 = (TPMT_TK_AUTH **)(argp3);
  result = (TSS2_RC)Esys_PolicySigned_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicySecret(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2B_DIGEST *arg8 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg9 = (TPM2B_NONCE *) 0 ;
  INT32 arg10 ;
  TPM2B_TIMEOUT **arg11 = (TPM2B_TIMEOUT **) 0 ;
  TPMT_TK_AUTH **arg12 = (TPMT_TK_AUTH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOO:Esys_PolicySecret",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicySecret" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicySecret" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicySecret" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicySecret" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicySecret" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicySecret" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicySecret" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicySecret" "', argument " "8"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg8 = (TPM2B_DIGEST *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicySecret" "', argument " "9"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg9 = (TPM2B_NONCE *)(argp9);
  ecode10 = SWIG_AsVal_int(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_PolicySecret" "', argument " "10"" of type '" "INT32""'");
  } 
  arg10 = (INT32)(val10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_PolicySecret" "', argument " "11"" of type '" "TPM2B_TIMEOUT **""'"); 
  }
  arg11 = (TPM2B_TIMEOUT **)(argp11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_PolicySecret" "', argument " "12"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg12 = (TPMT_TK_AUTH **)(argp12);
  result = (TSS2_RC)Esys_PolicySecret(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_DIGEST const *)arg8,(TPM2B_DIGEST const *)arg9,arg10,arg11,arg12);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicySecret_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2B_DIGEST *arg8 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg9 = (TPM2B_NONCE *) 0 ;
  INT32 arg10 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_PolicySecret_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicySecret_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicySecret_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicySecret_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicySecret_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicySecret_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicySecret_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicySecret_Async" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicySecret_Async" "', argument " "8"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg8 = (TPM2B_DIGEST *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicySecret_Async" "', argument " "9"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg9 = (TPM2B_NONCE *)(argp9);
  ecode10 = SWIG_AsVal_int(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_PolicySecret_Async" "', argument " "10"" of type '" "INT32""'");
  } 
  arg10 = (INT32)(val10);
  result = (TSS2_RC)Esys_PolicySecret_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_DIGEST const *)arg8,(TPM2B_DIGEST const *)arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicySecret_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_TIMEOUT **arg2 = (TPM2B_TIMEOUT **) 0 ;
  TPMT_TK_AUTH **arg3 = (TPMT_TK_AUTH **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_PolicySecret_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicySecret_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_PolicySecret_Finish" "', argument " "2"" of type '" "TPM2B_TIMEOUT **""'"); 
  }
  arg2 = (TPM2B_TIMEOUT **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_PolicySecret_Finish" "', argument " "3"" of type '" "TPMT_TK_AUTH **""'"); 
  }
  arg3 = (TPMT_TK_AUTH **)(argp3);
  result = (TSS2_RC)Esys_PolicySecret_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyTicket(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_TIMEOUT *arg6 = (TPM2B_TIMEOUT *) 0 ;
  TPM2B_DIGEST *arg7 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg8 = (TPM2B_NONCE *) 0 ;
  TPM2B_NAME *arg9 = (TPM2B_NAME *) 0 ;
  TPMT_TK_AUTH *arg10 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_PolicyTicket",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyTicket" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyTicket" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyTicket" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyTicket" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyTicket" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyTicket" "', argument " "6"" of type '" "TPM2B_TIMEOUT const *""'"); 
  }
  arg6 = (TPM2B_TIMEOUT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyTicket" "', argument " "7"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg7 = (TPM2B_DIGEST *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicyTicket" "', argument " "8"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg8 = (TPM2B_NONCE *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicyTicket" "', argument " "9"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg9 = (TPM2B_NAME *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_PolicyTicket" "', argument " "10"" of type '" "TPMT_TK_AUTH const *""'"); 
  }
  arg10 = (TPMT_TK_AUTH *)(argp10);
  result = (TSS2_RC)Esys_PolicyTicket(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_DIGEST const *)arg8,(TPM2B_NAME const *)arg9,(TPMT_TK_AUTH const *)arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyTicket_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_TIMEOUT *arg6 = (TPM2B_TIMEOUT *) 0 ;
  TPM2B_DIGEST *arg7 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg8 = (TPM2B_NONCE *) 0 ;
  TPM2B_NAME *arg9 = (TPM2B_NAME *) 0 ;
  TPMT_TK_AUTH *arg10 = (TPMT_TK_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_PolicyTicket_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyTicket_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyTicket_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyTicket_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyTicket_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyTicket_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyTicket_Async" "', argument " "6"" of type '" "TPM2B_TIMEOUT const *""'"); 
  }
  arg6 = (TPM2B_TIMEOUT *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyTicket_Async" "', argument " "7"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg7 = (TPM2B_DIGEST *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicyTicket_Async" "', argument " "8"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg8 = (TPM2B_NONCE *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicyTicket_Async" "', argument " "9"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg9 = (TPM2B_NAME *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_TPMT_TK_AUTH, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_PolicyTicket_Async" "', argument " "10"" of type '" "TPMT_TK_AUTH const *""'"); 
  }
  arg10 = (TPMT_TK_AUTH *)(argp10);
  result = (TSS2_RC)Esys_PolicyTicket_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_DIGEST const *)arg8,(TPM2B_NAME const *)arg9,(TPMT_TK_AUTH const *)arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyTicket_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyTicket_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyTicket_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyTicket_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyOR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_DIGEST *arg6 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyOR",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyOR" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyOR" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyOR" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyOR" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyOR" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyOR" "', argument " "6"" of type '" "TPML_DIGEST const *""'"); 
  }
  arg6 = (TPML_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyOR(arg1,arg2,arg3,arg4,arg5,(TPML_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyOR_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_DIGEST *arg6 = (TPML_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyOR_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyOR_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyOR_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyOR_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyOR_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyOR_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyOR_Async" "', argument " "6"" of type '" "TPML_DIGEST const *""'"); 
  }
  arg6 = (TPML_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyOR_Async(arg1,arg2,arg3,arg4,arg5,(TPML_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyOR_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyOR_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyOR_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyOR_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPCR(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPML_PCR_SELECTION *arg7 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PolicyPCR",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPCR" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyPCR" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyPCR" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyPCR" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyPCR" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyPCR" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyPCR" "', argument " "7"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg7 = (TPML_PCR_SELECTION *)(argp7);
  result = (TSS2_RC)Esys_PolicyPCR(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPML_PCR_SELECTION const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPCR_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPML_PCR_SELECTION *arg7 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PolicyPCR_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPCR_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyPCR_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyPCR_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyPCR_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyPCR_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyPCR_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyPCR_Async" "', argument " "7"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg7 = (TPML_PCR_SELECTION *)(argp7);
  result = (TSS2_RC)Esys_PolicyPCR_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPML_PCR_SELECTION const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPCR_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyPCR_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPCR_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyPCR_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyLocality(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMA_LOCALITY arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyLocality",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyLocality" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyLocality" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyLocality" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyLocality" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyLocality" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyLocality" "', argument " "6"" of type '" "TPMA_LOCALITY""'");
  } 
  arg6 = (TPMA_LOCALITY)(val6);
  result = (TSS2_RC)Esys_PolicyLocality(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyLocality_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMA_LOCALITY arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyLocality_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyLocality_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyLocality_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyLocality_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyLocality_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyLocality_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyLocality_Async" "', argument " "6"" of type '" "TPMA_LOCALITY""'");
  } 
  arg6 = (TPMA_LOCALITY)(val6);
  result = (TSS2_RC)Esys_PolicyLocality_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyLocality_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyLocality_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyLocality_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyLocality_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNV(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  TPM2B_OPERAND *arg8 = (TPM2B_OPERAND *) 0 ;
  UINT16 arg9 ;
  TPM2_EO arg10 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  unsigned short val9 ;
  int ecode9 = 0 ;
  unsigned short val10 ;
  int ecode10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_PolicyNV",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNV" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyNV" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyNV" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyNV" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyNV" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyNV" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PolicyNV" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicyNV" "', argument " "8"" of type '" "TPM2B_OPERAND const *""'"); 
  }
  arg8 = (TPM2B_OPERAND *)(argp8);
  ecode9 = SWIG_AsVal_unsigned_SS_short(obj8, &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "Esys_PolicyNV" "', argument " "9"" of type '" "UINT16""'");
  } 
  arg9 = (UINT16)(val9);
  ecode10 = SWIG_AsVal_unsigned_SS_short(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_PolicyNV" "', argument " "10"" of type '" "TPM2_EO""'");
  } 
  arg10 = (TPM2_EO)(val10);
  result = (TSS2_RC)Esys_PolicyNV(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_DIGEST const *)arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNV_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  TPM2B_OPERAND *arg8 = (TPM2B_OPERAND *) 0 ;
  UINT16 arg9 ;
  TPM2_EO arg10 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  unsigned short val9 ;
  int ecode9 = 0 ;
  unsigned short val10 ;
  int ecode10 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:Esys_PolicyNV_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNV_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyNV_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyNV_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyNV_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyNV_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyNV_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PolicyNV_Async" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicyNV_Async" "', argument " "8"" of type '" "TPM2B_OPERAND const *""'"); 
  }
  arg8 = (TPM2B_OPERAND *)(argp8);
  ecode9 = SWIG_AsVal_unsigned_SS_short(obj8, &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "Esys_PolicyNV_Async" "', argument " "9"" of type '" "UINT16""'");
  } 
  arg9 = (UINT16)(val9);
  ecode10 = SWIG_AsVal_unsigned_SS_short(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_PolicyNV_Async" "', argument " "10"" of type '" "TPM2_EO""'");
  } 
  arg10 = (TPM2_EO)(val10);
  result = (TSS2_RC)Esys_PolicyNV_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_DIGEST const *)arg8,arg9,arg10);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNV_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyNV_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNV_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyNV_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCounterTimer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_OPERAND *arg6 = (TPM2B_OPERAND *) 0 ;
  UINT16 arg7 ;
  TPM2_EO arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PolicyCounterTimer",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCounterTimer" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyCounterTimer" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyCounterTimer" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyCounterTimer" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyCounterTimer" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyCounterTimer" "', argument " "6"" of type '" "TPM2B_OPERAND const *""'"); 
  }
  arg6 = (TPM2B_OPERAND *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PolicyCounterTimer" "', argument " "7"" of type '" "UINT16""'");
  } 
  arg7 = (UINT16)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_PolicyCounterTimer" "', argument " "8"" of type '" "TPM2_EO""'");
  } 
  arg8 = (TPM2_EO)(val8);
  result = (TSS2_RC)Esys_PolicyCounterTimer(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCounterTimer_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_OPERAND *arg6 = (TPM2B_OPERAND *) 0 ;
  UINT16 arg7 ;
  TPM2_EO arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PolicyCounterTimer_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "6"" of type '" "TPM2B_OPERAND const *""'"); 
  }
  arg6 = (TPM2B_OPERAND *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "7"" of type '" "UINT16""'");
  } 
  arg7 = (UINT16)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_PolicyCounterTimer_Async" "', argument " "8"" of type '" "TPM2_EO""'");
  } 
  arg8 = (TPM2_EO)(val8);
  result = (TSS2_RC)Esys_PolicyCounterTimer_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCounterTimer_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyCounterTimer_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCounterTimer_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyCounterTimer_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCommandCode(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2_CC arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyCommandCode",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCommandCode" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyCommandCode" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyCommandCode" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyCommandCode" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyCommandCode" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyCommandCode" "', argument " "6"" of type '" "TPM2_CC""'");
  } 
  arg6 = (TPM2_CC)(val6);
  result = (TSS2_RC)Esys_PolicyCommandCode(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCommandCode_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2_CC arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyCommandCode_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCommandCode_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyCommandCode_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyCommandCode_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyCommandCode_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyCommandCode_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyCommandCode_Async" "', argument " "6"" of type '" "TPM2_CC""'");
  } 
  arg6 = (TPM2_CC)(val6);
  result = (TSS2_RC)Esys_PolicyCommandCode_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCommandCode_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyCommandCode_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCommandCode_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyCommandCode_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPhysicalPresence(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyPhysicalPresence",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPhysicalPresence" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyPhysicalPresence" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyPhysicalPresence" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyPhysicalPresence" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyPhysicalPresence" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyPhysicalPresence(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPhysicalPresence_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyPhysicalPresence_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPhysicalPresence_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyPhysicalPresence_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyPhysicalPresence_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyPhysicalPresence_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyPhysicalPresence_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyPhysicalPresence_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPhysicalPresence_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyPhysicalPresence_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPhysicalPresence_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyPhysicalPresence_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCpHash(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyCpHash",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCpHash" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyCpHash" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyCpHash" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyCpHash" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyCpHash" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyCpHash" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyCpHash(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCpHash_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyCpHash_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCpHash_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyCpHash_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyCpHash_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyCpHash_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyCpHash_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyCpHash_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyCpHash_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyCpHash_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyCpHash_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyCpHash_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyCpHash_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNameHash(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyNameHash",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNameHash" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyNameHash" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyNameHash" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyNameHash" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyNameHash" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyNameHash" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyNameHash(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNameHash_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyNameHash_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNameHash_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyNameHash_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyNameHash_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyNameHash_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyNameHash_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyNameHash_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyNameHash_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNameHash_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyNameHash_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNameHash_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyNameHash_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyDuplicationSelect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_NAME *arg6 = (TPM2B_NAME *) 0 ;
  TPM2B_NAME *arg7 = (TPM2B_NAME *) 0 ;
  TPMI_YES_NO arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned char val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PolicyDuplicationSelect",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "6"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg6 = (TPM2B_NAME *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "7"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg7 = (TPM2B_NAME *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_char(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_PolicyDuplicationSelect" "', argument " "8"" of type '" "TPMI_YES_NO""'");
  } 
  arg8 = (TPMI_YES_NO)(val8);
  result = (TSS2_RC)Esys_PolicyDuplicationSelect(arg1,arg2,arg3,arg4,arg5,(TPM2B_NAME const *)arg6,(TPM2B_NAME const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyDuplicationSelect_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_NAME *arg6 = (TPM2B_NAME *) 0 ;
  TPM2B_NAME *arg7 = (TPM2B_NAME *) 0 ;
  TPMI_YES_NO arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned char val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_PolicyDuplicationSelect_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "6"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg6 = (TPM2B_NAME *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "7"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg7 = (TPM2B_NAME *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_char(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_PolicyDuplicationSelect_Async" "', argument " "8"" of type '" "TPMI_YES_NO""'");
  } 
  arg8 = (TPMI_YES_NO)(val8);
  result = (TSS2_RC)Esys_PolicyDuplicationSelect_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_NAME const *)arg6,(TPM2B_NAME const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyDuplicationSelect_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyDuplicationSelect_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyDuplicationSelect_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyDuplicationSelect_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthorize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2B_NAME *arg8 = (TPM2B_NAME *) 0 ;
  TPMT_TK_VERIFIED *arg9 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_PolicyAuthorize",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthorize" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyAuthorize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyAuthorize" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyAuthorize" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyAuthorize" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyAuthorize" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyAuthorize" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicyAuthorize" "', argument " "8"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg8 = (TPM2B_NAME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicyAuthorize" "', argument " "9"" of type '" "TPMT_TK_VERIFIED const *""'"); 
  }
  arg9 = (TPMT_TK_VERIFIED *)(argp9);
  result = (TSS2_RC)Esys_PolicyAuthorize(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_NAME const *)arg8,(TPMT_TK_VERIFIED const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthorize_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPM2B_NONCE *arg7 = (TPM2B_NONCE *) 0 ;
  TPM2B_NAME *arg8 = (TPM2B_NAME *) 0 ;
  TPMT_TK_VERIFIED *arg9 = (TPMT_TK_VERIFIED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_PolicyAuthorize_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "7"" of type '" "TPM2B_NONCE const *""'"); 
  }
  arg7 = (TPM2B_NONCE *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "8"" of type '" "TPM2B_NAME const *""'"); 
  }
  arg8 = (TPM2B_NAME *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_TK_VERIFIED, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_PolicyAuthorize_Async" "', argument " "9"" of type '" "TPMT_TK_VERIFIED const *""'"); 
  }
  arg9 = (TPMT_TK_VERIFIED *)(argp9);
  result = (TSS2_RC)Esys_PolicyAuthorize_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_DIGEST const *)arg7,(TPM2B_NAME const *)arg8,(TPMT_TK_VERIFIED const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthorize_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyAuthorize_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthorize_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyAuthorize_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthValue(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyAuthValue",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthValue" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyAuthValue" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyAuthValue" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyAuthValue" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyAuthValue" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyAuthValue(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthValue_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyAuthValue_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthValue_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyAuthValue_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyAuthValue_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyAuthValue_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyAuthValue_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyAuthValue_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthValue_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyAuthValue_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthValue_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyAuthValue_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPassword(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyPassword",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPassword" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyPassword" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyPassword" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyPassword" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyPassword" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyPassword(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPassword_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyPassword_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPassword_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyPassword_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyPassword_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyPassword_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyPassword_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyPassword_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyPassword_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyPassword_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyPassword_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyPassword_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyGetDigest(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST **arg6 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyGetDigest",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyGetDigest" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyGetDigest" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyGetDigest" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyGetDigest" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyGetDigest" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyGetDigest" "', argument " "6"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg6 = (TPM2B_DIGEST **)(argp6);
  result = (TSS2_RC)Esys_PolicyGetDigest(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyGetDigest_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_PolicyGetDigest_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyGetDigest_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyGetDigest_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyGetDigest_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyGetDigest_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyGetDigest_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_PolicyGetDigest_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyGetDigest_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DIGEST **arg2 = (TPM2B_DIGEST **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_PolicyGetDigest_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyGetDigest_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_PolicyGetDigest_Finish" "', argument " "2"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg2 = (TPM2B_DIGEST **)(argp2);
  result = (TSS2_RC)Esys_PolicyGetDigest_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNvWritten(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_YES_NO arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyNvWritten",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNvWritten" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyNvWritten" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyNvWritten" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyNvWritten" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyNvWritten" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyNvWritten" "', argument " "6"" of type '" "TPMI_YES_NO""'");
  } 
  arg6 = (TPMI_YES_NO)(val6);
  result = (TSS2_RC)Esys_PolicyNvWritten(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNvWritten_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_YES_NO arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyNvWritten_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNvWritten_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyNvWritten_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyNvWritten_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyNvWritten_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyNvWritten_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyNvWritten_Async" "', argument " "6"" of type '" "TPMI_YES_NO""'");
  } 
  arg6 = (TPMI_YES_NO)(val6);
  result = (TSS2_RC)Esys_PolicyNvWritten_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyNvWritten_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyNvWritten_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyNvWritten_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyNvWritten_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyTemplate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyTemplate",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyTemplate" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyTemplate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyTemplate" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyTemplate" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyTemplate" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyTemplate" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyTemplate(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyTemplate_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_PolicyTemplate_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyTemplate_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyTemplate_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyTemplate_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyTemplate_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyTemplate_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PolicyTemplate_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  result = (TSS2_RC)Esys_PolicyTemplate_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyTemplate_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyTemplate_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyTemplate_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyTemplate_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthorizeNV(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PolicyAuthorizeNV",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PolicyAuthorizeNV" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  result = (TSS2_RC)Esys_PolicyAuthorizeNV(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthorizeNV_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PolicyAuthorizeNV_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_PolicyAuthorizeNV_Async" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  result = (TSS2_RC)Esys_PolicyAuthorizeNV_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PolicyAuthorizeNV_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PolicyAuthorizeNV_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PolicyAuthorizeNV_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PolicyAuthorizeNV_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CreatePrimary(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_CREATE *arg6 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPML_PCR_SELECTION *arg9 = (TPML_PCR_SELECTION *) 0 ;
  unsigned int *arg10 = (unsigned int *) 0 ;
  TPM2B_PUBLIC **arg11 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_CREATION_DATA **arg12 = (TPM2B_CREATION_DATA **) 0 ;
  TPM2B_DIGEST **arg13 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_CREATION **arg14 = (TPMT_TK_CREATION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOOOO:Esys_CreatePrimary",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&obj12,&obj13)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CreatePrimary" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_CreatePrimary" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_CreatePrimary" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_CreatePrimary" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_CreatePrimary" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_CreatePrimary" "', argument " "6"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_CREATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_CreatePrimary" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_CreatePrimary" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_CreatePrimary" "', argument " "9"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg9 = (TPML_PCR_SELECTION *)(argp9);
  res10 = SWIG_ConvertPtr(obj9, &argp10,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "Esys_CreatePrimary" "', argument " "10"" of type '" "unsigned int *""'"); 
  }
  arg10 = (unsigned int *)(argp10);
  res11 = SWIG_ConvertPtr(obj10, &argp11,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), "in method '" "Esys_CreatePrimary" "', argument " "11"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg11 = (TPM2B_PUBLIC **)(argp11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_CreatePrimary" "', argument " "12"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg12 = (TPM2B_CREATION_DATA **)(argp12);
  res13 = SWIG_ConvertPtr(obj12, &argp13,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), "in method '" "Esys_CreatePrimary" "', argument " "13"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg13 = (TPM2B_DIGEST **)(argp13);
  res14 = SWIG_ConvertPtr(obj13, &argp14,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), "in method '" "Esys_CreatePrimary" "', argument " "14"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg14 = (TPMT_TK_CREATION **)(argp14);
  result = (TSS2_RC)Esys_CreatePrimary(arg1,arg2,arg3,arg4,arg5,(TPM2B_SENSITIVE_CREATE const *)arg6,(TPM2B_PUBLIC const *)arg7,(TPM2B_DATA const *)arg8,(TPML_PCR_SELECTION const *)arg9,arg10,arg11,arg12,arg13,arg14);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CreatePrimary_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_SENSITIVE_CREATE *arg6 = (TPM2B_SENSITIVE_CREATE *) 0 ;
  TPM2B_PUBLIC *arg7 = (TPM2B_PUBLIC *) 0 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPML_PCR_SELECTION *arg9 = (TPML_PCR_SELECTION *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_CreatePrimary_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CreatePrimary_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_CreatePrimary_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_CreatePrimary_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_CreatePrimary_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_CreatePrimary_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_SENSITIVE_CREATE, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_CreatePrimary_Async" "', argument " "6"" of type '" "TPM2B_SENSITIVE_CREATE const *""'"); 
  }
  arg6 = (TPM2B_SENSITIVE_CREATE *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_CreatePrimary_Async" "', argument " "7"" of type '" "TPM2B_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_CreatePrimary_Async" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPML_PCR_SELECTION, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_CreatePrimary_Async" "', argument " "9"" of type '" "TPML_PCR_SELECTION const *""'"); 
  }
  arg9 = (TPML_PCR_SELECTION *)(argp9);
  result = (TSS2_RC)Esys_CreatePrimary_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_SENSITIVE_CREATE const *)arg6,(TPM2B_PUBLIC const *)arg7,(TPM2B_DATA const *)arg8,(TPML_PCR_SELECTION const *)arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_CreatePrimary_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  TPM2B_PUBLIC **arg3 = (TPM2B_PUBLIC **) 0 ;
  TPM2B_CREATION_DATA **arg4 = (TPM2B_CREATION_DATA **) 0 ;
  TPM2B_DIGEST **arg5 = (TPM2B_DIGEST **) 0 ;
  TPMT_TK_CREATION **arg6 = (TPMT_TK_CREATION **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_CreatePrimary_Finish",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_CreatePrimary_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_CreatePrimary_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_CreatePrimary_Finish" "', argument " "3"" of type '" "TPM2B_PUBLIC **""'"); 
  }
  arg3 = (TPM2B_PUBLIC **)(argp3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_p_TPM2B_CREATION_DATA, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Esys_CreatePrimary_Finish" "', argument " "4"" of type '" "TPM2B_CREATION_DATA **""'"); 
  }
  arg4 = (TPM2B_CREATION_DATA **)(argp4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_CreatePrimary_Finish" "', argument " "5"" of type '" "TPM2B_DIGEST **""'"); 
  }
  arg5 = (TPM2B_DIGEST **)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPMT_TK_CREATION, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_CreatePrimary_Finish" "', argument " "6"" of type '" "TPMT_TK_CREATION **""'"); 
  }
  arg6 = (TPMT_TK_CREATION **)(argp6);
  result = (TSS2_RC)Esys_CreatePrimary_Finish(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HierarchyControl(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_RH_ENABLES arg6 ;
  TPMI_YES_NO arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned char val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_HierarchyControl",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HierarchyControl" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HierarchyControl" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HierarchyControl" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HierarchyControl" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HierarchyControl" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_HierarchyControl" "', argument " "6"" of type '" "TPMI_RH_ENABLES""'");
  } 
  arg6 = (TPMI_RH_ENABLES)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_char(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_HierarchyControl" "', argument " "7"" of type '" "TPMI_YES_NO""'");
  } 
  arg7 = (TPMI_YES_NO)(val7);
  result = (TSS2_RC)Esys_HierarchyControl(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HierarchyControl_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_RH_ENABLES arg6 ;
  TPMI_YES_NO arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned char val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_HierarchyControl_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HierarchyControl_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HierarchyControl_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HierarchyControl_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HierarchyControl_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HierarchyControl_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_HierarchyControl_Async" "', argument " "6"" of type '" "TPMI_RH_ENABLES""'");
  } 
  arg6 = (TPMI_RH_ENABLES)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_char(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_HierarchyControl_Async" "', argument " "7"" of type '" "TPMI_YES_NO""'");
  } 
  arg7 = (TPMI_YES_NO)(val7);
  result = (TSS2_RC)Esys_HierarchyControl_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HierarchyControl_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_HierarchyControl_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HierarchyControl_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_HierarchyControl_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetPrimaryPolicy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMI_ALG_HASH arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_SetPrimaryPolicy",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetPrimaryPolicy" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetPrimaryPolicy" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SetPrimaryPolicy" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SetPrimaryPolicy" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SetPrimaryPolicy" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_SetPrimaryPolicy" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_SetPrimaryPolicy" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  result = (TSS2_RC)Esys_SetPrimaryPolicy(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetPrimaryPolicy_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_DIGEST *arg6 = (TPM2B_DIGEST *) 0 ;
  TPMI_ALG_HASH arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_SetPrimaryPolicy_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "6"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg6 = (TPM2B_DIGEST *)(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_SetPrimaryPolicy_Async" "', argument " "7"" of type '" "TPMI_ALG_HASH""'");
  } 
  arg7 = (TPMI_ALG_HASH)(val7);
  result = (TSS2_RC)Esys_SetPrimaryPolicy_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetPrimaryPolicy_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_SetPrimaryPolicy_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetPrimaryPolicy_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_SetPrimaryPolicy_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ChangePPS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ChangePPS",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ChangePPS" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ChangePPS" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ChangePPS" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ChangePPS" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ChangePPS" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_ChangePPS(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ChangePPS_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ChangePPS_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ChangePPS_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ChangePPS_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ChangePPS_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ChangePPS_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ChangePPS_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_ChangePPS_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ChangePPS_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_ChangePPS_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ChangePPS_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_ChangePPS_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ChangeEPS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ChangeEPS",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ChangeEPS" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ChangeEPS" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ChangeEPS" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ChangeEPS" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ChangeEPS" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_ChangeEPS(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ChangeEPS_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ChangeEPS_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ChangeEPS_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ChangeEPS_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ChangeEPS_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ChangeEPS_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ChangeEPS_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_ChangeEPS_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ChangeEPS_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_ChangeEPS_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ChangeEPS_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_ChangeEPS_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Clear",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Clear" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Clear" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Clear" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Clear" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Clear" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_Clear(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Clear_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Clear_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Clear_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Clear_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Clear_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Clear_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_Clear_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_Clear_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Clear_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_Clear_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Clear_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_Clear_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClearControl(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_YES_NO arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ClearControl",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClearControl" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ClearControl" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ClearControl" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ClearControl" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ClearControl" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ClearControl" "', argument " "6"" of type '" "TPMI_YES_NO""'");
  } 
  arg6 = (TPMI_YES_NO)(val6);
  result = (TSS2_RC)Esys_ClearControl(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClearControl_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPMI_YES_NO arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ClearControl_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClearControl_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ClearControl_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ClearControl_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ClearControl_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ClearControl_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ClearControl_Async" "', argument " "6"" of type '" "TPMI_YES_NO""'");
  } 
  arg6 = (TPMI_YES_NO)(val6);
  result = (TSS2_RC)Esys_ClearControl_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClearControl_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_ClearControl_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClearControl_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_ClearControl_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HierarchyChangeAuth(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_HierarchyChangeAuth",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HierarchyChangeAuth" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HierarchyChangeAuth" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HierarchyChangeAuth" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HierarchyChangeAuth" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HierarchyChangeAuth" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_HierarchyChangeAuth" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  result = (TSS2_RC)Esys_HierarchyChangeAuth(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HierarchyChangeAuth_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_HierarchyChangeAuth_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HierarchyChangeAuth_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_HierarchyChangeAuth_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_HierarchyChangeAuth_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_HierarchyChangeAuth_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_HierarchyChangeAuth_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_HierarchyChangeAuth_Async" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  result = (TSS2_RC)Esys_HierarchyChangeAuth_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_HierarchyChangeAuth_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_HierarchyChangeAuth_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_HierarchyChangeAuth_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_HierarchyChangeAuth_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_DictionaryAttackLockReset(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_DictionaryAttackLockReset",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_DictionaryAttackLockReset" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_DictionaryAttackLockReset" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_DictionaryAttackLockReset" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_DictionaryAttackLockReset" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_DictionaryAttackLockReset" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_DictionaryAttackLockReset(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_DictionaryAttackLockReset_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_DictionaryAttackLockReset_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_DictionaryAttackLockReset_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_DictionaryAttackLockReset_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_DictionaryAttackLockReset_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_DictionaryAttackLockReset_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_DictionaryAttackLockReset_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_DictionaryAttackLockReset_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_DictionaryAttackLockReset_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_DictionaryAttackLockReset_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_DictionaryAttackLockReset_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_DictionaryAttackLockReset_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_DictionaryAttackParameters(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  UINT32 arg6 ;
  UINT32 arg7 ;
  UINT32 arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_DictionaryAttackParameters",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_DictionaryAttackParameters" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_DictionaryAttackParameters" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_DictionaryAttackParameters" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_DictionaryAttackParameters" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_DictionaryAttackParameters" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_DictionaryAttackParameters" "', argument " "6"" of type '" "UINT32""'");
  } 
  arg6 = (UINT32)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_DictionaryAttackParameters" "', argument " "7"" of type '" "UINT32""'");
  } 
  arg7 = (UINT32)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_DictionaryAttackParameters" "', argument " "8"" of type '" "UINT32""'");
  } 
  arg8 = (UINT32)(val8);
  result = (TSS2_RC)Esys_DictionaryAttackParameters(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_DictionaryAttackParameters_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  UINT32 arg6 ;
  UINT32 arg7 ;
  UINT32 arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_DictionaryAttackParameters_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "6"" of type '" "UINT32""'");
  } 
  arg6 = (UINT32)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "7"" of type '" "UINT32""'");
  } 
  arg7 = (UINT32)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_DictionaryAttackParameters_Async" "', argument " "8"" of type '" "UINT32""'");
  } 
  arg8 = (UINT32)(val8);
  result = (TSS2_RC)Esys_DictionaryAttackParameters_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_DictionaryAttackParameters_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_DictionaryAttackParameters_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_DictionaryAttackParameters_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_DictionaryAttackParameters_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PP_Commands(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_CC *arg6 = (TPML_CC *) 0 ;
  TPML_CC *arg7 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PP_Commands",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PP_Commands" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PP_Commands" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PP_Commands" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PP_Commands" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PP_Commands" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PP_Commands" "', argument " "6"" of type '" "TPML_CC const *""'"); 
  }
  arg6 = (TPML_CC *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PP_Commands" "', argument " "7"" of type '" "TPML_CC const *""'"); 
  }
  arg7 = (TPML_CC *)(argp7);
  result = (TSS2_RC)Esys_PP_Commands(arg1,arg2,arg3,arg4,arg5,(TPML_CC const *)arg6,(TPML_CC const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PP_Commands_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPML_CC *arg6 = (TPML_CC *) 0 ;
  TPML_CC *arg7 = (TPML_CC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_PP_Commands_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PP_Commands_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_PP_Commands_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_PP_Commands_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_PP_Commands_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_PP_Commands_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_PP_Commands_Async" "', argument " "6"" of type '" "TPML_CC const *""'"); 
  }
  arg6 = (TPML_CC *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPML_CC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_PP_Commands_Async" "', argument " "7"" of type '" "TPML_CC const *""'"); 
  }
  arg7 = (TPML_CC *)(argp7);
  result = (TSS2_RC)Esys_PP_Commands_Async(arg1,arg2,arg3,arg4,arg5,(TPML_CC const *)arg6,(TPML_CC const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_PP_Commands_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_PP_Commands_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_PP_Commands_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_PP_Commands_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetAlgorithmSet(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  UINT32 arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_SetAlgorithmSet",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetAlgorithmSet" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetAlgorithmSet" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SetAlgorithmSet" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SetAlgorithmSet" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SetAlgorithmSet" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_SetAlgorithmSet" "', argument " "6"" of type '" "UINT32""'");
  } 
  arg6 = (UINT32)(val6);
  result = (TSS2_RC)Esys_SetAlgorithmSet(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetAlgorithmSet_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  UINT32 arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_SetAlgorithmSet_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetAlgorithmSet_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_SetAlgorithmSet_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_SetAlgorithmSet_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_SetAlgorithmSet_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_SetAlgorithmSet_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_SetAlgorithmSet_Async" "', argument " "6"" of type '" "UINT32""'");
  } 
  arg6 = (UINT32)(val6);
  result = (TSS2_RC)Esys_SetAlgorithmSet_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_SetAlgorithmSet_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_SetAlgorithmSet_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_SetAlgorithmSet_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_SetAlgorithmSet_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FieldUpgradeStart(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DIGEST *arg7 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIGNATURE *arg8 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_FieldUpgradeStart",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FieldUpgradeStart" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FieldUpgradeStart" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_FieldUpgradeStart" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_FieldUpgradeStart" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_FieldUpgradeStart" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_FieldUpgradeStart" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_FieldUpgradeStart" "', argument " "7"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg7 = (TPM2B_DIGEST *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_FieldUpgradeStart" "', argument " "8"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg8 = (TPMT_SIGNATURE *)(argp8);
  result = (TSS2_RC)Esys_FieldUpgradeStart(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,(TPMT_SIGNATURE const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FieldUpgradeStart_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_DIGEST *arg7 = (TPM2B_DIGEST *) 0 ;
  TPMT_SIGNATURE *arg8 = (TPMT_SIGNATURE *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_FieldUpgradeStart_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "7"" of type '" "TPM2B_DIGEST const *""'"); 
  }
  arg7 = (TPM2B_DIGEST *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_FieldUpgradeStart_Async" "', argument " "8"" of type '" "TPMT_SIGNATURE const *""'"); 
  }
  arg8 = (TPMT_SIGNATURE *)(argp8);
  result = (TSS2_RC)Esys_FieldUpgradeStart_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_DIGEST const *)arg7,(TPMT_SIGNATURE const *)arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FieldUpgradeStart_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_FieldUpgradeStart_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FieldUpgradeStart_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_FieldUpgradeStart_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FieldUpgradeData(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_MAX_BUFFER *arg5 = (TPM2B_MAX_BUFFER *) 0 ;
  TPMT_HA **arg6 = (TPMT_HA **) 0 ;
  TPMT_HA **arg7 = (TPMT_HA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_FieldUpgradeData",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FieldUpgradeData" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FieldUpgradeData" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_FieldUpgradeData" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_FieldUpgradeData" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_FieldUpgradeData" "', argument " "5"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg5 = (TPM2B_MAX_BUFFER *)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_FieldUpgradeData" "', argument " "6"" of type '" "TPMT_HA **""'"); 
  }
  arg6 = (TPMT_HA **)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_FieldUpgradeData" "', argument " "7"" of type '" "TPMT_HA **""'"); 
  }
  arg7 = (TPMT_HA **)(argp7);
  result = (TSS2_RC)Esys_FieldUpgradeData(arg1,arg2,arg3,arg4,(TPM2B_MAX_BUFFER const *)arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FieldUpgradeData_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_MAX_BUFFER *arg5 = (TPM2B_MAX_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_FieldUpgradeData_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FieldUpgradeData_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FieldUpgradeData_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_FieldUpgradeData_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_FieldUpgradeData_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_FieldUpgradeData_Async" "', argument " "5"" of type '" "TPM2B_MAX_BUFFER const *""'"); 
  }
  arg5 = (TPM2B_MAX_BUFFER *)(argp5);
  result = (TSS2_RC)Esys_FieldUpgradeData_Async(arg1,arg2,arg3,arg4,(TPM2B_MAX_BUFFER const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FieldUpgradeData_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMT_HA **arg2 = (TPMT_HA **) 0 ;
  TPMT_HA **arg3 = (TPMT_HA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_FieldUpgradeData_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FieldUpgradeData_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_FieldUpgradeData_Finish" "', argument " "2"" of type '" "TPMT_HA **""'"); 
  }
  arg2 = (TPMT_HA **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_HA, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_FieldUpgradeData_Finish" "', argument " "3"" of type '" "TPMT_HA **""'"); 
  }
  arg3 = (TPMT_HA **)(argp3);
  result = (TSS2_RC)Esys_FieldUpgradeData_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FirmwareRead(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  UINT32 arg5 ;
  TPM2B_MAX_BUFFER **arg6 = (TPM2B_MAX_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_FirmwareRead",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FirmwareRead" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FirmwareRead" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_FirmwareRead" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_FirmwareRead" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_FirmwareRead" "', argument " "5"" of type '" "UINT32""'");
  } 
  arg5 = (UINT32)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_FirmwareRead" "', argument " "6"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg6 = (TPM2B_MAX_BUFFER **)(argp6);
  result = (TSS2_RC)Esys_FirmwareRead(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FirmwareRead_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  UINT32 arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_FirmwareRead_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FirmwareRead_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FirmwareRead_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_FirmwareRead_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_FirmwareRead_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_FirmwareRead_Async" "', argument " "5"" of type '" "UINT32""'");
  } 
  arg5 = (UINT32)(val5);
  result = (TSS2_RC)Esys_FirmwareRead_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FirmwareRead_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_MAX_BUFFER **arg2 = (TPM2B_MAX_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_FirmwareRead_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FirmwareRead_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_MAX_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_FirmwareRead_Finish" "', argument " "2"" of type '" "TPM2B_MAX_BUFFER **""'"); 
  }
  arg2 = (TPM2B_MAX_BUFFER **)(argp2);
  result = (TSS2_RC)Esys_FirmwareRead_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ContextSave(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  TPMS_CONTEXT **arg3 = (TPMS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_ContextSave",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ContextSave" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ContextSave" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_ContextSave" "', argument " "3"" of type '" "TPMS_CONTEXT **""'"); 
  }
  arg3 = (TPMS_CONTEXT **)(argp3);
  result = (TSS2_RC)Esys_ContextSave(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ContextSave_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ContextSave_Async",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ContextSave_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ContextSave_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  result = (TSS2_RC)Esys_ContextSave_Async(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ContextSave_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMS_CONTEXT **arg2 = (TPMS_CONTEXT **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ContextSave_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ContextSave_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ContextSave_Finish" "', argument " "2"" of type '" "TPMS_CONTEXT **""'"); 
  }
  arg2 = (TPMS_CONTEXT **)(argp2);
  result = (TSS2_RC)Esys_ContextSave_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ContextLoad(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMS_CONTEXT *arg2 = (TPMS_CONTEXT *) 0 ;
  unsigned int *arg3 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_ContextLoad",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ContextLoad" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ContextLoad" "', argument " "2"" of type '" "TPMS_CONTEXT const *""'"); 
  }
  arg2 = (TPMS_CONTEXT *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_ContextLoad" "', argument " "3"" of type '" "unsigned int *""'"); 
  }
  arg3 = (unsigned int *)(argp3);
  result = (TSS2_RC)Esys_ContextLoad(arg1,(TPMS_CONTEXT const *)arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ContextLoad_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMS_CONTEXT *arg2 = (TPMS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ContextLoad_Async",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ContextLoad_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_TPMS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ContextLoad_Async" "', argument " "2"" of type '" "TPMS_CONTEXT const *""'"); 
  }
  arg2 = (TPMS_CONTEXT *)(argp2);
  result = (TSS2_RC)Esys_ContextLoad_Async(arg1,(TPMS_CONTEXT const *)arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ContextLoad_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ContextLoad_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ContextLoad_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ContextLoad_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_ContextLoad_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FlushContext(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_FlushContext",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FlushContext" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FlushContext" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  result = (TSS2_RC)Esys_FlushContext(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FlushContext_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_FlushContext_Async",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FlushContext_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_FlushContext_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  result = (TSS2_RC)Esys_FlushContext_Async(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_FlushContext_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_FlushContext_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_FlushContext_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_FlushContext_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EvictControl(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPMI_DH_PERSISTENT arg7 ;
  unsigned int *arg8 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_EvictControl",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EvictControl" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EvictControl" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EvictControl" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EvictControl" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EvictControl" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_EvictControl" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_EvictControl" "', argument " "7"" of type '" "TPMI_DH_PERSISTENT""'");
  } 
  arg7 = (TPMI_DH_PERSISTENT)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_EvictControl" "', argument " "8"" of type '" "unsigned int *""'"); 
  }
  arg8 = (unsigned int *)(argp8);
  result = (TSS2_RC)Esys_EvictControl(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EvictControl_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPMI_DH_PERSISTENT arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_EvictControl_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EvictControl_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_EvictControl_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_EvictControl_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_EvictControl_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_EvictControl_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_EvictControl_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_EvictControl_Async" "', argument " "7"" of type '" "TPMI_DH_PERSISTENT""'");
  } 
  arg7 = (TPMI_DH_PERSISTENT)(val7);
  result = (TSS2_RC)Esys_EvictControl_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_EvictControl_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_EvictControl_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_EvictControl_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_EvictControl_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_EvictControl_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ReadClock(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMS_TIME_INFO **arg5 = (TPMS_TIME_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_ReadClock",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ReadClock" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ReadClock" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ReadClock" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ReadClock" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_ReadClock" "', argument " "5"" of type '" "TPMS_TIME_INFO **""'"); 
  }
  arg5 = (TPMS_TIME_INFO **)(argp5);
  result = (TSS2_RC)Esys_ReadClock(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ReadClock_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:Esys_ReadClock_Async",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ReadClock_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ReadClock_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ReadClock_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ReadClock_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  result = (TSS2_RC)Esys_ReadClock_Async(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ReadClock_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMS_TIME_INFO **arg2 = (TPMS_TIME_INFO **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_ReadClock_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ReadClock_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPMS_TIME_INFO, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_ReadClock_Finish" "', argument " "2"" of type '" "TPMS_TIME_INFO **""'"); 
  }
  arg2 = (TPMS_TIME_INFO **)(argp2);
  result = (TSS2_RC)Esys_ReadClock_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClockSet(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  UINT64 arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned long long val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ClockSet",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClockSet" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ClockSet" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ClockSet" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ClockSet" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ClockSet" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_long_SS_long(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ClockSet" "', argument " "6"" of type '" "UINT64""'");
  } 
  arg6 = (UINT64)(val6);
  result = (TSS2_RC)Esys_ClockSet(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClockSet_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  UINT64 arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned long long val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ClockSet_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClockSet_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ClockSet_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ClockSet_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ClockSet_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ClockSet_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_long_SS_long(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ClockSet_Async" "', argument " "6"" of type '" "UINT64""'");
  } 
  arg6 = (UINT64)(val6);
  result = (TSS2_RC)Esys_ClockSet_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClockSet_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_ClockSet_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClockSet_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_ClockSet_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClockRateAdjust(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2_CLOCK_ADJUST arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  signed char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ClockRateAdjust",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClockRateAdjust" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ClockRateAdjust" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ClockRateAdjust" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ClockRateAdjust" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ClockRateAdjust" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_signed_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ClockRateAdjust" "', argument " "6"" of type '" "TPM2_CLOCK_ADJUST""'");
  } 
  arg6 = (TPM2_CLOCK_ADJUST)(val6);
  result = (TSS2_RC)Esys_ClockRateAdjust(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClockRateAdjust_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2_CLOCK_ADJUST arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  signed char val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_ClockRateAdjust_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClockRateAdjust_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_ClockRateAdjust_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_ClockRateAdjust_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_ClockRateAdjust_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_ClockRateAdjust_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_signed_SS_char(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_ClockRateAdjust_Async" "', argument " "6"" of type '" "TPM2_CLOCK_ADJUST""'");
  } 
  arg6 = (TPM2_CLOCK_ADJUST)(val6);
  result = (TSS2_RC)Esys_ClockRateAdjust_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_ClockRateAdjust_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_ClockRateAdjust_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_ClockRateAdjust_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_ClockRateAdjust_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetCapability(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2_CAP arg5 ;
  UINT32 arg6 ;
  UINT32 arg7 ;
  TPMI_YES_NO *arg8 = (TPMI_YES_NO *) 0 ;
  TPMS_CAPABILITY_DATA **arg9 = (TPMS_CAPABILITY_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_GetCapability",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetCapability" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetCapability" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetCapability" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetCapability" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetCapability" "', argument " "5"" of type '" "TPM2_CAP""'");
  } 
  arg5 = (TPM2_CAP)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetCapability" "', argument " "6"" of type '" "UINT32""'");
  } 
  arg6 = (UINT32)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_GetCapability" "', argument " "7"" of type '" "UINT32""'");
  } 
  arg7 = (UINT32)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_GetCapability" "', argument " "8"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg8 = (TPMI_YES_NO *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_GetCapability" "', argument " "9"" of type '" "TPMS_CAPABILITY_DATA **""'"); 
  }
  arg9 = (TPMS_CAPABILITY_DATA **)(argp9);
  result = (TSS2_RC)Esys_GetCapability(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetCapability_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2_CAP arg5 ;
  UINT32 arg6 ;
  UINT32 arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_GetCapability_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetCapability_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_GetCapability_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_GetCapability_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_GetCapability_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_GetCapability_Async" "', argument " "5"" of type '" "TPM2_CAP""'");
  } 
  arg5 = (TPM2_CAP)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_GetCapability_Async" "', argument " "6"" of type '" "UINT32""'");
  } 
  arg6 = (UINT32)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_GetCapability_Async" "', argument " "7"" of type '" "UINT32""'");
  } 
  arg7 = (UINT32)(val7);
  result = (TSS2_RC)Esys_GetCapability_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_GetCapability_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPMI_YES_NO *arg2 = (TPMI_YES_NO *) 0 ;
  TPMS_CAPABILITY_DATA **arg3 = (TPMS_CAPABILITY_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_GetCapability_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_GetCapability_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_GetCapability_Finish" "', argument " "2"" of type '" "TPMI_YES_NO *""'"); 
  }
  arg2 = (TPMI_YES_NO *)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMS_CAPABILITY_DATA, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_GetCapability_Finish" "', argument " "3"" of type '" "TPMS_CAPABILITY_DATA **""'"); 
  }
  arg3 = (TPMS_CAPABILITY_DATA **)(argp3);
  result = (TSS2_RC)Esys_GetCapability_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TestParms(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMT_PUBLIC_PARMS *arg5 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_TestParms",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TestParms" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TestParms" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_TestParms" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_TestParms" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_TestParms" "', argument " "5"" of type '" "TPMT_PUBLIC_PARMS const *""'"); 
  }
  arg5 = (TPMT_PUBLIC_PARMS *)(argp5);
  result = (TSS2_RC)Esys_TestParms(arg1,arg2,arg3,arg4,(TPMT_PUBLIC_PARMS const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TestParms_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPMT_PUBLIC_PARMS *arg5 = (TPMT_PUBLIC_PARMS *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_TestParms_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TestParms_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_TestParms_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_TestParms_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_TestParms_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPMT_PUBLIC_PARMS, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_TestParms_Async" "', argument " "5"" of type '" "TPMT_PUBLIC_PARMS const *""'"); 
  }
  arg5 = (TPMT_PUBLIC_PARMS *)(argp5);
  result = (TSS2_RC)Esys_TestParms_Async(arg1,arg2,arg3,arg4,(TPMT_PUBLIC_PARMS const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_TestParms_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_TestParms_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_TestParms_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_TestParms_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_DefineSpace(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  TPM2B_NV_PUBLIC *arg7 = (TPM2B_NV_PUBLIC *) 0 ;
  unsigned int *arg8 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_NV_DefineSpace",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_DefineSpace" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_DefineSpace" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_DefineSpace" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_DefineSpace" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_DefineSpace" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_NV_DefineSpace" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_DefineSpace" "', argument " "7"" of type '" "TPM2B_NV_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_NV_PUBLIC *)(argp7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_NV_DefineSpace" "', argument " "8"" of type '" "unsigned int *""'"); 
  }
  arg8 = (unsigned int *)(argp8);
  result = (TSS2_RC)Esys_NV_DefineSpace(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_NV_PUBLIC const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_DefineSpace_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  TPM2B_NV_PUBLIC *arg7 = (TPM2B_NV_PUBLIC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_NV_DefineSpace_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_DefineSpace_Async" "', argument " "7"" of type '" "TPM2B_NV_PUBLIC const *""'"); 
  }
  arg7 = (TPM2B_NV_PUBLIC *)(argp7);
  result = (TSS2_RC)Esys_NV_DefineSpace_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6,(TPM2B_NV_PUBLIC const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_DefineSpace_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int *arg2 = (unsigned int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_NV_DefineSpace_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_DefineSpace_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_NV_DefineSpace_Finish" "', argument " "2"" of type '" "unsigned int *""'"); 
  }
  arg2 = (unsigned int *)(argp2);
  result = (TSS2_RC)Esys_NV_DefineSpace_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_UndefineSpace(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_UndefineSpace",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_UndefineSpace" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_UndefineSpace" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_UndefineSpace" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_UndefineSpace" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_UndefineSpace" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_UndefineSpace" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_UndefineSpace(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_UndefineSpace_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_UndefineSpace_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_UndefineSpace_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_UndefineSpace_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_UndefineSpace_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_UndefineSpace_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_UndefineSpace_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_UndefineSpace_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_UndefineSpace_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_UndefineSpace_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_UndefineSpace_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_UndefineSpace_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_UndefineSpace_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_UndefineSpaceSpecial(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_UndefineSpaceSpecial",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_UndefineSpaceSpecial" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_UndefineSpaceSpecial" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_UndefineSpaceSpecial" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_UndefineSpaceSpecial" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_UndefineSpaceSpecial" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_UndefineSpaceSpecial" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_UndefineSpaceSpecial(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_UndefineSpaceSpecial_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_UndefineSpaceSpecial_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_UndefineSpaceSpecial_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_UndefineSpaceSpecial_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_UndefineSpaceSpecial_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_UndefineSpaceSpecial_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_UndefineSpaceSpecial_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_UndefineSpaceSpecial_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_UndefineSpaceSpecial_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_UndefineSpaceSpecial_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_UndefineSpaceSpecial_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_UndefineSpaceSpecial_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_UndefineSpaceSpecial_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ReadPublic(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_NV_PUBLIC **arg6 = (TPM2B_NV_PUBLIC **) 0 ;
  TPM2B_NAME **arg7 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_NV_ReadPublic",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ReadPublic" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_ReadPublic" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_ReadPublic" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_ReadPublic" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_ReadPublic" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_NV_ReadPublic" "', argument " "6"" of type '" "TPM2B_NV_PUBLIC **""'"); 
  }
  arg6 = (TPM2B_NV_PUBLIC **)(argp6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_ReadPublic" "', argument " "7"" of type '" "TPM2B_NAME **""'"); 
  }
  arg7 = (TPM2B_NAME **)(argp7);
  result = (TSS2_RC)Esys_NV_ReadPublic(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ReadPublic_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_NV_ReadPublic_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ReadPublic_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_ReadPublic_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_ReadPublic_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_ReadPublic_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_ReadPublic_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_NV_ReadPublic_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ReadPublic_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_NV_PUBLIC **arg2 = (TPM2B_NV_PUBLIC **) 0 ;
  TPM2B_NAME **arg3 = (TPM2B_NAME **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_NV_ReadPublic_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ReadPublic_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_NV_PUBLIC, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_NV_ReadPublic_Finish" "', argument " "2"" of type '" "TPM2B_NV_PUBLIC **""'"); 
  }
  arg2 = (TPM2B_NV_PUBLIC **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPM2B_NAME, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_NV_ReadPublic_Finish" "', argument " "3"" of type '" "TPM2B_NAME **""'"); 
  }
  arg3 = (TPM2B_NAME **)(argp3);
  result = (TSS2_RC)Esys_NV_ReadPublic_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Write(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_MAX_NV_BUFFER *arg7 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  UINT16 arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_NV_Write",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Write" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Write" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Write" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Write" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Write" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Write" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_Write" "', argument " "7"" of type '" "TPM2B_MAX_NV_BUFFER const *""'"); 
  }
  arg7 = (TPM2B_MAX_NV_BUFFER *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_NV_Write" "', argument " "8"" of type '" "UINT16""'");
  } 
  arg8 = (UINT16)(val8);
  result = (TSS2_RC)Esys_NV_Write(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_MAX_NV_BUFFER const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Write_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_MAX_NV_BUFFER *arg7 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  UINT16 arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_NV_Write_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Write_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Write_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Write_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Write_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Write_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Write_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_Write_Async" "', argument " "7"" of type '" "TPM2B_MAX_NV_BUFFER const *""'"); 
  }
  arg7 = (TPM2B_MAX_NV_BUFFER *)(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_NV_Write_Async" "', argument " "8"" of type '" "UINT16""'");
  } 
  arg8 = (UINT16)(val8);
  result = (TSS2_RC)Esys_NV_Write_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_MAX_NV_BUFFER const *)arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Write_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_Write_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Write_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_Write_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Increment(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_Increment",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Increment" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Increment" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Increment" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Increment" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Increment" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Increment" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_Increment(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Increment_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_Increment_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Increment_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Increment_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Increment_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Increment_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Increment_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Increment_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_Increment_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Increment_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_Increment_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Increment_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_Increment_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Extend(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_MAX_NV_BUFFER *arg7 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_NV_Extend",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Extend" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Extend" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Extend" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Extend" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Extend" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Extend" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_Extend" "', argument " "7"" of type '" "TPM2B_MAX_NV_BUFFER const *""'"); 
  }
  arg7 = (TPM2B_MAX_NV_BUFFER *)(argp7);
  result = (TSS2_RC)Esys_NV_Extend(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_MAX_NV_BUFFER const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Extend_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  TPM2B_MAX_NV_BUFFER *arg7 = (TPM2B_MAX_NV_BUFFER *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_NV_Extend_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Extend_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Extend_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Extend_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Extend_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Extend_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Extend_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  res7 = SWIG_ConvertPtr(obj6, &argp7,SWIGTYPE_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Esys_NV_Extend_Async" "', argument " "7"" of type '" "TPM2B_MAX_NV_BUFFER const *""'"); 
  }
  arg7 = (TPM2B_MAX_NV_BUFFER *)(argp7);
  result = (TSS2_RC)Esys_NV_Extend_Async(arg1,arg2,arg3,arg4,arg5,arg6,(TPM2B_MAX_NV_BUFFER const *)arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Extend_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_Extend_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Extend_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_Extend_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_SetBits(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  UINT64 arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned long long val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_NV_SetBits",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_SetBits" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_SetBits" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_SetBits" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_SetBits" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_SetBits" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_SetBits" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_long_SS_long(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_NV_SetBits" "', argument " "7"" of type '" "UINT64""'");
  } 
  arg7 = (UINT64)(val7);
  result = (TSS2_RC)Esys_NV_SetBits(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_SetBits_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  UINT64 arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned long long val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOO:Esys_NV_SetBits_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_SetBits_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_SetBits_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_SetBits_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_SetBits_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_SetBits_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_SetBits_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_long_SS_long(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_NV_SetBits_Async" "', argument " "7"" of type '" "UINT64""'");
  } 
  arg7 = (UINT64)(val7);
  result = (TSS2_RC)Esys_NV_SetBits_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_SetBits_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_SetBits_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_SetBits_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_SetBits_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_WriteLock(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_WriteLock",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_WriteLock" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_WriteLock" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_WriteLock" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_WriteLock" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_WriteLock" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_WriteLock" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_WriteLock(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_WriteLock_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_WriteLock_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_WriteLock_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_WriteLock_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_WriteLock_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_WriteLock_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_WriteLock_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_WriteLock_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_WriteLock_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_WriteLock_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_WriteLock_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_WriteLock_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_WriteLock_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_GlobalWriteLock(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_NV_GlobalWriteLock",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_GlobalWriteLock" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_GlobalWriteLock" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_GlobalWriteLock" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_GlobalWriteLock" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_GlobalWriteLock" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_NV_GlobalWriteLock(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_GlobalWriteLock_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_NV_GlobalWriteLock_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_GlobalWriteLock_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_GlobalWriteLock_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_GlobalWriteLock_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_GlobalWriteLock_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_GlobalWriteLock_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  result = (TSS2_RC)Esys_NV_GlobalWriteLock_Async(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_GlobalWriteLock_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_GlobalWriteLock_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_GlobalWriteLock_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_GlobalWriteLock_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Read(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  UINT16 arg7 ;
  UINT16 arg8 ;
  TPM2B_MAX_NV_BUFFER **arg9 = (TPM2B_MAX_NV_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOO:Esys_NV_Read",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Read" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Read" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Read" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Read" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Read" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Read" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_NV_Read" "', argument " "7"" of type '" "UINT16""'");
  } 
  arg7 = (UINT16)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_NV_Read" "', argument " "8"" of type '" "UINT16""'");
  } 
  arg8 = (UINT16)(val8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_NV_Read" "', argument " "9"" of type '" "TPM2B_MAX_NV_BUFFER **""'"); 
  }
  arg9 = (TPM2B_MAX_NV_BUFFER **)(argp9);
  result = (TSS2_RC)Esys_NV_Read(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Read_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  UINT16 arg7 ;
  UINT16 arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned short val7 ;
  int ecode7 = 0 ;
  unsigned short val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOO:Esys_NV_Read_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Read_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Read_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Read_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Read_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Read_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Read_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_short(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_NV_Read_Async" "', argument " "7"" of type '" "UINT16""'");
  } 
  arg7 = (UINT16)(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_short(obj7, &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Esys_NV_Read_Async" "', argument " "8"" of type '" "UINT16""'");
  } 
  arg8 = (UINT16)(val8);
  result = (TSS2_RC)Esys_NV_Read_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Read_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_MAX_NV_BUFFER **arg2 = (TPM2B_MAX_NV_BUFFER **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_NV_Read_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Read_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_MAX_NV_BUFFER, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_NV_Read_Finish" "', argument " "2"" of type '" "TPM2B_MAX_NV_BUFFER **""'"); 
  }
  arg2 = (TPM2B_MAX_NV_BUFFER **)(argp2);
  result = (TSS2_RC)Esys_NV_Read_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ReadLock(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_ReadLock",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ReadLock" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_ReadLock" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_ReadLock" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_ReadLock" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_ReadLock" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_ReadLock" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_ReadLock(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ReadLock_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_ReadLock_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ReadLock_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_ReadLock_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_ReadLock_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_ReadLock_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_ReadLock_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_ReadLock_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  result = (TSS2_RC)Esys_NV_ReadLock_Async(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ReadLock_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_ReadLock_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ReadLock_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_ReadLock_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ChangeAuth(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_ChangeAuth",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ChangeAuth" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_ChangeAuth" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_ChangeAuth" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_ChangeAuth" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_ChangeAuth" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_NV_ChangeAuth" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  result = (TSS2_RC)Esys_NV_ChangeAuth(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ChangeAuth_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  TPM2B_AUTH *arg6 = (TPM2B_AUTH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_NV_ChangeAuth_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ChangeAuth_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_ChangeAuth_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_ChangeAuth_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_ChangeAuth_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_ChangeAuth_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_TPM2B_DIGEST, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_NV_ChangeAuth_Async" "', argument " "6"" of type '" "TPM2B_AUTH const *""'"); 
  }
  arg6 = (TPM2B_AUTH *)(argp6);
  result = (TSS2_RC)Esys_NV_ChangeAuth_Async(arg1,arg2,arg3,arg4,arg5,(TPM2B_DIGEST const *)arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_ChangeAuth_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_NV_ChangeAuth_Finish",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_ChangeAuth_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  result = (TSS2_RC)Esys_NV_ChangeAuth_Finish(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Certify(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg9 = (TPMT_SIG_SCHEME *) 0 ;
  UINT16 arg10 ;
  UINT16 arg11 ;
  TPM2B_ATTEST **arg12 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg13 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  unsigned short val10 ;
  int ecode10 = 0 ;
  unsigned short val11 ;
  int ecode11 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOOOO:Esys_NV_Certify",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10,&obj11,&obj12)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Certify" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Certify" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Certify" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Certify" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Certify" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Certify" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_NV_Certify" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_NV_Certify" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_NV_Certify" "', argument " "9"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg9 = (TPMT_SIG_SCHEME *)(argp9);
  ecode10 = SWIG_AsVal_unsigned_SS_short(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_NV_Certify" "', argument " "10"" of type '" "UINT16""'");
  } 
  arg10 = (UINT16)(val10);
  ecode11 = SWIG_AsVal_unsigned_SS_short(obj10, &val11);
  if (!SWIG_IsOK(ecode11)) {
    SWIG_exception_fail(SWIG_ArgError(ecode11), "in method '" "Esys_NV_Certify" "', argument " "11"" of type '" "UINT16""'");
  } 
  arg11 = (UINT16)(val11);
  res12 = SWIG_ConvertPtr(obj11, &argp12,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), "in method '" "Esys_NV_Certify" "', argument " "12"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg12 = (TPM2B_ATTEST **)(argp12);
  res13 = SWIG_ConvertPtr(obj12, &argp13,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), "in method '" "Esys_NV_Certify" "', argument " "13"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg13 = (TPMT_SIGNATURE **)(argp13);
  result = (TSS2_RC)Esys_NV_Certify(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_DATA const *)arg8,(TPMT_SIG_SCHEME const *)arg9,arg10,arg11,arg12,arg13);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Certify_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  unsigned int arg7 ;
  TPM2B_DATA *arg8 = (TPM2B_DATA *) 0 ;
  TPMT_SIG_SCHEME *arg9 = (TPMT_SIG_SCHEME *) 0 ;
  UINT16 arg10 ;
  UINT16 arg11 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  unsigned short val10 ;
  int ecode10 = 0 ;
  unsigned short val11 ;
  int ecode11 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOOO:Esys_NV_Certify_Async",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9,&obj10)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Certify_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_NV_Certify_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_NV_Certify_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_NV_Certify_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Esys_NV_Certify_Async" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = (unsigned int)(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Esys_NV_Certify_Async" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = (unsigned int)(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(obj6, &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "Esys_NV_Certify_Async" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = (unsigned int)(val7);
  res8 = SWIG_ConvertPtr(obj7, &argp8,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "Esys_NV_Certify_Async" "', argument " "8"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg8 = (TPM2B_DATA *)(argp8);
  res9 = SWIG_ConvertPtr(obj8, &argp9,SWIGTYPE_p_TPMT_SIG_SCHEME, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "Esys_NV_Certify_Async" "', argument " "9"" of type '" "TPMT_SIG_SCHEME const *""'"); 
  }
  arg9 = (TPMT_SIG_SCHEME *)(argp9);
  ecode10 = SWIG_AsVal_unsigned_SS_short(obj9, &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "Esys_NV_Certify_Async" "', argument " "10"" of type '" "UINT16""'");
  } 
  arg10 = (UINT16)(val10);
  ecode11 = SWIG_AsVal_unsigned_SS_short(obj10, &val11);
  if (!SWIG_IsOK(ecode11)) {
    SWIG_exception_fail(SWIG_ArgError(ecode11), "in method '" "Esys_NV_Certify_Async" "', argument " "11"" of type '" "UINT16""'");
  } 
  arg11 = (UINT16)(val11);
  result = (TSS2_RC)Esys_NV_Certify_Async(arg1,arg2,arg3,arg4,arg5,arg6,arg7,(TPM2B_DATA const *)arg8,(TPMT_SIG_SCHEME const *)arg9,arg10,arg11);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_NV_Certify_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_ATTEST **arg2 = (TPM2B_ATTEST **) 0 ;
  TPMT_SIGNATURE **arg3 = (TPMT_SIGNATURE **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:Esys_NV_Certify_Finish",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_NV_Certify_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_ATTEST, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_NV_Certify_Finish" "', argument " "2"" of type '" "TPM2B_ATTEST **""'"); 
  }
  arg2 = (TPM2B_ATTEST **)(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_p_TPMT_SIGNATURE, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Esys_NV_Certify_Finish" "', argument " "3"" of type '" "TPMT_SIGNATURE **""'"); 
  }
  arg3 = (TPMT_SIGNATURE **)(argp3);
  result = (TSS2_RC)Esys_NV_Certify_Finish(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Vendor_TCG_Test(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_DATA *arg5 = (TPM2B_DATA *) 0 ;
  TPM2B_DATA **arg6 = (TPM2B_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Esys_Vendor_TCG_Test",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Vendor_TCG_Test" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Vendor_TCG_Test" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Vendor_TCG_Test" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Vendor_TCG_Test" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_Vendor_TCG_Test" "', argument " "5"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg5 = (TPM2B_DATA *)(argp5);
  res6 = SWIG_ConvertPtr(obj5, &argp6,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Esys_Vendor_TCG_Test" "', argument " "6"" of type '" "TPM2B_DATA **""'"); 
  }
  arg6 = (TPM2B_DATA **)(argp6);
  result = (TSS2_RC)Esys_Vendor_TCG_Test(arg1,arg2,arg3,arg4,(TPM2B_DATA const *)arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Vendor_TCG_Test_Async(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  TPM2B_DATA *arg5 = (TPM2B_DATA *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Esys_Vendor_TCG_Test_Async",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Vendor_TCG_Test_Async" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Esys_Vendor_TCG_Test_Async" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = (unsigned int)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Esys_Vendor_TCG_Test_Async" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Esys_Vendor_TCG_Test_Async" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = (unsigned int)(val4);
  res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Esys_Vendor_TCG_Test_Async" "', argument " "5"" of type '" "TPM2B_DATA const *""'"); 
  }
  arg5 = (TPM2B_DATA *)(argp5);
  result = (TSS2_RC)Esys_Vendor_TCG_Test_Async(arg1,arg2,arg3,arg4,(TPM2B_DATA const *)arg5);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Vendor_TCG_Test_Finish(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  ESYS_CONTEXT *arg1 = (ESYS_CONTEXT *) 0 ;
  TPM2B_DATA **arg2 = (TPM2B_DATA **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  TSS2_RC result;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Esys_Vendor_TCG_Test_Finish",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_ESYS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Vendor_TCG_Test_Finish" "', argument " "1"" of type '" "ESYS_CONTEXT *""'"); 
  }
  arg1 = (ESYS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_p_TPM2B_DATA, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Esys_Vendor_TCG_Test_Finish" "', argument " "2"" of type '" "TPM2B_DATA **""'"); 
  }
  arg2 = (TPM2B_DATA **)(argp2);
  result = (TSS2_RC)Esys_Vendor_TCG_Test_Finish(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Esys_Free(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  void *arg1 = (void *) 0 ;
  int res1 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Esys_Free",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0,SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Esys_Free" "', argument " "1"" of type '" "void *""'"); 
  }
  Esys_Free(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


static PyMethodDef SwigMethods[] = {
	 { (char *)"SWIG_PyInstanceMethod_New", (PyCFunction)SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssCreator_set", _wrap_TSS2_ABI_VERSION_tssCreator_set, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssCreator_get", _wrap_TSS2_ABI_VERSION_tssCreator_get, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssFamily_set", _wrap_TSS2_ABI_VERSION_tssFamily_set, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssFamily_get", _wrap_TSS2_ABI_VERSION_tssFamily_get, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssLevel_set", _wrap_TSS2_ABI_VERSION_tssLevel_set, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssLevel_get", _wrap_TSS2_ABI_VERSION_tssLevel_get, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssVersion_set", _wrap_TSS2_ABI_VERSION_tssVersion_set, METH_VARARGS, NULL},
	 { (char *)"TSS2_ABI_VERSION_tssVersion_get", _wrap_TSS2_ABI_VERSION_tssVersion_get, METH_VARARGS, NULL},
	 { (char *)"new_TSS2_ABI_VERSION", _wrap_new_TSS2_ABI_VERSION, METH_VARARGS, (char *)"new_TSS2_ABI_VERSION() -> TSS2_ABI_VERSION"},
	 { (char *)"delete_TSS2_ABI_VERSION", _wrap_delete_TSS2_ABI_VERSION, METH_VARARGS, (char *)"delete_TSS2_ABI_VERSION(TSS2_ABI_VERSION self)"},
	 { (char *)"TSS2_ABI_VERSION_swigregister", TSS2_ABI_VERSION_swigregister, METH_VARARGS, NULL},
	 { (char *)"TSS2_RC_LAYER_SHIFT_swigconstant", TSS2_RC_LAYER_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_GENERAL_FAILURE_swigconstant", TSS2_BASE_RC_GENERAL_FAILURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NOT_IMPLEMENTED_swigconstant", TSS2_BASE_RC_NOT_IMPLEMENTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_CONTEXT_swigconstant", TSS2_BASE_RC_BAD_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_ABI_MISMATCH_swigconstant", TSS2_BASE_RC_ABI_MISMATCH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_REFERENCE_swigconstant", TSS2_BASE_RC_BAD_REFERENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_INSUFFICIENT_BUFFER_swigconstant", TSS2_BASE_RC_INSUFFICIENT_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_SEQUENCE_swigconstant", TSS2_BASE_RC_BAD_SEQUENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_CONNECTION_swigconstant", TSS2_BASE_RC_NO_CONNECTION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_TRY_AGAIN_swigconstant", TSS2_BASE_RC_TRY_AGAIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_IO_ERROR_swigconstant", TSS2_BASE_RC_IO_ERROR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_VALUE_swigconstant", TSS2_BASE_RC_BAD_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NOT_PERMITTED_swigconstant", TSS2_BASE_RC_NOT_PERMITTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_INVALID_SESSIONS_swigconstant", TSS2_BASE_RC_INVALID_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_DECRYPT_PARAM_swigconstant", TSS2_BASE_RC_NO_DECRYPT_PARAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_ENCRYPT_PARAM_swigconstant", TSS2_BASE_RC_NO_ENCRYPT_PARAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_SIZE_swigconstant", TSS2_BASE_RC_BAD_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_MALFORMED_RESPONSE_swigconstant", TSS2_BASE_RC_MALFORMED_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_INSUFFICIENT_CONTEXT_swigconstant", TSS2_BASE_RC_INSUFFICIENT_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_INSUFFICIENT_RESPONSE_swigconstant", TSS2_BASE_RC_INSUFFICIENT_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_INCOMPATIBLE_TCTI_swigconstant", TSS2_BASE_RC_INCOMPATIBLE_TCTI_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NOT_SUPPORTED_swigconstant", TSS2_BASE_RC_NOT_SUPPORTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_TCTI_STRUCTURE_swigconstant", TSS2_BASE_RC_BAD_TCTI_STRUCTURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_MEMORY_swigconstant", TSS2_BASE_RC_MEMORY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_TR_swigconstant", TSS2_BASE_RC_BAD_TR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_MULTIPLE_DECRYPT_SESSIONS_swigconstant", TSS2_BASE_RC_MULTIPLE_DECRYPT_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_MULTIPLE_ENCRYPT_SESSIONS_swigconstant", TSS2_BASE_RC_MULTIPLE_ENCRYPT_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_RSP_AUTH_FAILED_swigconstant", TSS2_BASE_RC_RSP_AUTH_FAILED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_CONFIG_swigconstant", TSS2_BASE_RC_NO_CONFIG_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_PATH_swigconstant", TSS2_BASE_RC_BAD_PATH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NOT_DELETABLE_swigconstant", TSS2_BASE_RC_NOT_DELETABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_PATH_ALREADY_EXISTS_swigconstant", TSS2_BASE_RC_PATH_ALREADY_EXISTS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_KEY_NOT_FOUND_swigconstant", TSS2_BASE_RC_KEY_NOT_FOUND_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_SIGNATURE_VERIFICATION_FAILED_swigconstant", TSS2_BASE_RC_SIGNATURE_VERIFICATION_FAILED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_HASH_MISMATCH_swigconstant", TSS2_BASE_RC_HASH_MISMATCH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_KEY_NOT_DUPLICABLE_swigconstant", TSS2_BASE_RC_KEY_NOT_DUPLICABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_PATH_NOT_FOUND_swigconstant", TSS2_BASE_RC_PATH_NOT_FOUND_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_CERT_swigconstant", TSS2_BASE_RC_NO_CERT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_PCR_swigconstant", TSS2_BASE_RC_NO_PCR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_PCR_NOT_RESETTABLE_swigconstant", TSS2_BASE_RC_PCR_NOT_RESETTABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_TEMPLATE_swigconstant", TSS2_BASE_RC_BAD_TEMPLATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_AUTHORIZATION_FAILED_swigconstant", TSS2_BASE_RC_AUTHORIZATION_FAILED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_AUTHORIZATION_UNKNOWN_swigconstant", TSS2_BASE_RC_AUTHORIZATION_UNKNOWN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NV_NOT_READABLE_swigconstant", TSS2_BASE_RC_NV_NOT_READABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NV_TOO_SMALL_swigconstant", TSS2_BASE_RC_NV_TOO_SMALL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NV_NOT_WRITEABLE_swigconstant", TSS2_BASE_RC_NV_NOT_WRITEABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_POLICY_UNKNOWN_swigconstant", TSS2_BASE_RC_POLICY_UNKNOWN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NV_WRONG_TYPE_swigconstant", TSS2_BASE_RC_NV_WRONG_TYPE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NAME_ALREADY_EXISTS_swigconstant", TSS2_BASE_RC_NAME_ALREADY_EXISTS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_TPM_swigconstant", TSS2_BASE_RC_NO_TPM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_BAD_KEY_swigconstant", TSS2_BASE_RC_BAD_KEY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_BASE_RC_NO_HANDLE_swigconstant", TSS2_BASE_RC_NO_HANDLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_LAYER_IMPLEMENTATION_SPECIFIC_OFFSET_swigconstant", TSS2_LAYER_IMPLEMENTATION_SPECIFIC_OFFSET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_LEVEL_IMPLEMENTATION_SPECIFIC_SHIFT_swigconstant", TSS2_LEVEL_IMPLEMENTATION_SPECIFIC_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_COMMAND_SIZE_swigconstant", TPM2_MAX_COMMAND_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_RESPONSE_SIZE_swigconstant", TPM2_MAX_RESPONSE_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NUM_PCR_BANKS_swigconstant", TPM2_NUM_PCR_BANKS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_DIGEST_BUFFER_swigconstant", TPM2_MAX_DIGEST_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_NV_BUFFER_SIZE_swigconstant", TPM2_MAX_NV_BUFFER_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_PCRS_swigconstant", TPM2_MAX_PCRS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_ALG_LIST_SIZE_swigconstant", TPM2_MAX_ALG_LIST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_CAP_CC_swigconstant", TPM2_MAX_CAP_CC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_CAP_BUFFER_swigconstant", TPM2_MAX_CAP_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_CONTEXT_SIZE_swigconstant", TPM2_MAX_CONTEXT_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SHA_DIGEST_SIZE_swigconstant", TPM2_SHA_DIGEST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SHA1_DIGEST_SIZE_swigconstant", TPM2_SHA1_DIGEST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SHA256_DIGEST_SIZE_swigconstant", TPM2_SHA256_DIGEST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SHA384_DIGEST_SIZE_swigconstant", TPM2_SHA384_DIGEST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SHA512_DIGEST_SIZE_swigconstant", TPM2_SHA512_DIGEST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SM3_256_DIGEST_SIZE_swigconstant", TPM2_SM3_256_DIGEST_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_SYM_BLOCK_SIZE_swigconstant", TPM2_MAX_SYM_BLOCK_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_SYM_DATA_swigconstant", TPM2_MAX_SYM_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_ECC_KEY_BYTES_swigconstant", TPM2_MAX_ECC_KEY_BYTES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_SYM_KEY_BYTES_swigconstant", TPM2_MAX_SYM_KEY_BYTES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_RSA_KEY_BYTES_swigconstant", TPM2_MAX_RSA_KEY_BYTES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_LABEL_MAX_BUFFER_swigconstant", TPM2_LABEL_MAX_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PCR_SELECT_MAX_swigconstant", TPM2_PCR_SELECT_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PRIVATE_VENDOR_SPECIFIC_BYTES_swigconstant", TPM2_PRIVATE_VENDOR_SPECIFIC_BYTES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_EXTENDED_SHIFT_swigconstant", TPMA_LOCALITY_EXTENDED_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_COMMANDINDEX_SHIFT_swigconstant", TPMA_CC_COMMANDINDEX_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_CHANDLES_SHIFT_swigconstant", TPMA_CC_CHANDLES_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_RES_SHIFT_swigconstant", TPMA_CC_RES_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NO_swigconstant", TPM2_NO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_YES_swigconstant", TPM2_YES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_empty_set", _wrap_TPMS_EMPTY_empty_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_empty_get", _wrap_TPMS_EMPTY_empty_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_EMPTY", _wrap_new_TPMS_EMPTY, METH_VARARGS, (char *)"new_TPMS_EMPTY() -> TPMS_EMPTY"},
	 { (char *)"delete_TPMS_EMPTY", _wrap_delete_TPMS_EMPTY, METH_VARARGS, (char *)"delete_TPMS_EMPTY(TPMS_EMPTY self)"},
	 { (char *)"TPMS_EMPTY_swigregister", TPMS_EMPTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_alg_set", _wrap_TPMS_ALGORITHM_DESCRIPTION_alg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_alg_get", _wrap_TPMS_ALGORITHM_DESCRIPTION_alg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_attributes_set", _wrap_TPMS_ALGORITHM_DESCRIPTION_attributes_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_attributes_get", _wrap_TPMS_ALGORITHM_DESCRIPTION_attributes_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DESCRIPTION", _wrap_new_TPMS_ALGORITHM_DESCRIPTION, METH_VARARGS, (char *)"new_TPMS_ALGORITHM_DESCRIPTION() -> TPMS_ALGORITHM_DESCRIPTION"},
	 { (char *)"delete_TPMS_ALGORITHM_DESCRIPTION", _wrap_delete_TPMS_ALGORITHM_DESCRIPTION, METH_VARARGS, (char *)"delete_TPMS_ALGORITHM_DESCRIPTION(TPMS_ALGORITHM_DESCRIPTION self)"},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_swigregister", TPMS_ALGORITHM_DESCRIPTION_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha1_set", _wrap_TPMU_HA_sha1_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha1_get", _wrap_TPMU_HA_sha1_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha256_set", _wrap_TPMU_HA_sha256_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha256_get", _wrap_TPMU_HA_sha256_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha384_set", _wrap_TPMU_HA_sha384_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha384_get", _wrap_TPMU_HA_sha384_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha512_set", _wrap_TPMU_HA_sha512_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sha512_get", _wrap_TPMU_HA_sha512_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sm3_256_set", _wrap_TPMU_HA_sm3_256_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_sm3_256_get", _wrap_TPMU_HA_sm3_256_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_HA", _wrap_new_TPMU_HA, METH_VARARGS, (char *)"new_TPMU_HA() -> TPMU_HA"},
	 { (char *)"delete_TPMU_HA", _wrap_delete_TPMU_HA, METH_VARARGS, (char *)"delete_TPMU_HA(TPMU_HA self)"},
	 { (char *)"TPMU_HA_swigregister", TPMU_HA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_hashAlg_set", _wrap_TPMT_HA_hashAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_hashAlg_get", _wrap_TPMT_HA_hashAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_digest_set", _wrap_TPMT_HA_digest_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_digest_get", _wrap_TPMT_HA_digest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_HA", _wrap_new_TPMT_HA, METH_VARARGS, (char *)"new_TPMT_HA() -> TPMT_HA"},
	 { (char *)"delete_TPMT_HA", _wrap_delete_TPMT_HA, METH_VARARGS, (char *)"delete_TPMT_HA(TPMT_HA self)"},
	 { (char *)"TPMT_HA_swigregister", TPMT_HA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_size_set", _wrap_TPM2B_DIGEST_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_size_get", _wrap_TPM2B_DIGEST_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_buffer_set", _wrap_TPM2B_DIGEST_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_buffer_get", _wrap_TPM2B_DIGEST_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DIGEST", _wrap_new_TPM2B_DIGEST, METH_VARARGS, (char *)"new_TPM2B_DIGEST() -> TPM2B_DIGEST"},
	 { (char *)"delete_TPM2B_DIGEST", _wrap_delete_TPM2B_DIGEST, METH_VARARGS, (char *)"delete_TPM2B_DIGEST(TPM2B_DIGEST self)"},
	 { (char *)"TPM2B_DIGEST_swigregister", TPM2B_DIGEST_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_size_set", _wrap_TPM2B_DATA_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_size_get", _wrap_TPM2B_DATA_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_buffer_set", _wrap_TPM2B_DATA_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_buffer_get", _wrap_TPM2B_DATA_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DATA", _wrap_new_TPM2B_DATA, METH_VARARGS, (char *)"new_TPM2B_DATA() -> TPM2B_DATA"},
	 { (char *)"delete_TPM2B_DATA", _wrap_delete_TPM2B_DATA, METH_VARARGS, (char *)"delete_TPM2B_DATA(TPM2B_DATA self)"},
	 { (char *)"TPM2B_DATA_swigregister", TPM2B_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_size_set", _wrap_TPM2B_EVENT_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_size_get", _wrap_TPM2B_EVENT_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_buffer_set", _wrap_TPM2B_EVENT_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_buffer_get", _wrap_TPM2B_EVENT_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_EVENT", _wrap_new_TPM2B_EVENT, METH_VARARGS, (char *)"new_TPM2B_EVENT() -> TPM2B_EVENT"},
	 { (char *)"delete_TPM2B_EVENT", _wrap_delete_TPM2B_EVENT, METH_VARARGS, (char *)"delete_TPM2B_EVENT(TPM2B_EVENT self)"},
	 { (char *)"TPM2B_EVENT_swigregister", TPM2B_EVENT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_size_set", _wrap_TPM2B_MAX_BUFFER_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_size_get", _wrap_TPM2B_MAX_BUFFER_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_buffer_set", _wrap_TPM2B_MAX_BUFFER_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_buffer_get", _wrap_TPM2B_MAX_BUFFER_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_BUFFER", _wrap_new_TPM2B_MAX_BUFFER, METH_VARARGS, (char *)"new_TPM2B_MAX_BUFFER() -> TPM2B_MAX_BUFFER"},
	 { (char *)"delete_TPM2B_MAX_BUFFER", _wrap_delete_TPM2B_MAX_BUFFER, METH_VARARGS, (char *)"delete_TPM2B_MAX_BUFFER(TPM2B_MAX_BUFFER self)"},
	 { (char *)"TPM2B_MAX_BUFFER_swigregister", TPM2B_MAX_BUFFER_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_size_set", _wrap_TPM2B_MAX_NV_BUFFER_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_size_get", _wrap_TPM2B_MAX_NV_BUFFER_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_buffer_set", _wrap_TPM2B_MAX_NV_BUFFER_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_buffer_get", _wrap_TPM2B_MAX_NV_BUFFER_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_NV_BUFFER", _wrap_new_TPM2B_MAX_NV_BUFFER, METH_VARARGS, (char *)"new_TPM2B_MAX_NV_BUFFER() -> TPM2B_MAX_NV_BUFFER"},
	 { (char *)"delete_TPM2B_MAX_NV_BUFFER", _wrap_delete_TPM2B_MAX_NV_BUFFER, METH_VARARGS, (char *)"delete_TPM2B_MAX_NV_BUFFER(TPM2B_MAX_NV_BUFFER self)"},
	 { (char *)"TPM2B_MAX_NV_BUFFER_swigregister", TPM2B_MAX_NV_BUFFER_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_size_set", _wrap_TPM2B_IV_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_size_get", _wrap_TPM2B_IV_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_buffer_set", _wrap_TPM2B_IV_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_buffer_get", _wrap_TPM2B_IV_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_IV", _wrap_new_TPM2B_IV, METH_VARARGS, (char *)"new_TPM2B_IV() -> TPM2B_IV"},
	 { (char *)"delete_TPM2B_IV", _wrap_delete_TPM2B_IV, METH_VARARGS, (char *)"delete_TPM2B_IV(TPM2B_IV self)"},
	 { (char *)"TPM2B_IV_swigregister", TPM2B_IV_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_digest_set", _wrap_TPMU_NAME_digest_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_digest_get", _wrap_TPMU_NAME_digest_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_handle_set", _wrap_TPMU_NAME_handle_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_handle_get", _wrap_TPMU_NAME_handle_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_NAME", _wrap_new_TPMU_NAME, METH_VARARGS, (char *)"new_TPMU_NAME() -> TPMU_NAME"},
	 { (char *)"delete_TPMU_NAME", _wrap_delete_TPMU_NAME, METH_VARARGS, (char *)"delete_TPMU_NAME(TPMU_NAME self)"},
	 { (char *)"TPMU_NAME_swigregister", TPMU_NAME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_size_set", _wrap_TPM2B_NAME_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_size_get", _wrap_TPM2B_NAME_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_name_set", _wrap_TPM2B_NAME_name_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_name_get", _wrap_TPM2B_NAME_name_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NAME", _wrap_new_TPM2B_NAME, METH_VARARGS, (char *)"new_TPM2B_NAME() -> TPM2B_NAME"},
	 { (char *)"delete_TPM2B_NAME", _wrap_delete_TPM2B_NAME, METH_VARARGS, (char *)"delete_TPM2B_NAME(TPM2B_NAME self)"},
	 { (char *)"TPM2B_NAME_swigregister", TPM2B_NAME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_sizeofSelect_set", _wrap_TPMS_PCR_SELECT_sizeofSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_sizeofSelect_get", _wrap_TPMS_PCR_SELECT_sizeofSelect_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_pcrSelect_set", _wrap_TPMS_PCR_SELECT_pcrSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_pcrSelect_get", _wrap_TPMS_PCR_SELECT_pcrSelect_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECT", _wrap_new_TPMS_PCR_SELECT, METH_VARARGS, (char *)"new_TPMS_PCR_SELECT() -> TPMS_PCR_SELECT"},
	 { (char *)"delete_TPMS_PCR_SELECT", _wrap_delete_TPMS_PCR_SELECT, METH_VARARGS, (char *)"delete_TPMS_PCR_SELECT(TPMS_PCR_SELECT self)"},
	 { (char *)"TPMS_PCR_SELECT_swigregister", TPMS_PCR_SELECT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_hash_set", _wrap_TPMS_PCR_SELECTION_hash_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_hash_get", _wrap_TPMS_PCR_SELECTION_hash_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_sizeofSelect_set", _wrap_TPMS_PCR_SELECTION_sizeofSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_sizeofSelect_get", _wrap_TPMS_PCR_SELECTION_sizeofSelect_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_pcrSelect_set", _wrap_TPMS_PCR_SELECTION_pcrSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_pcrSelect_get", _wrap_TPMS_PCR_SELECTION_pcrSelect_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECTION", _wrap_new_TPMS_PCR_SELECTION, METH_VARARGS, (char *)"new_TPMS_PCR_SELECTION() -> TPMS_PCR_SELECTION"},
	 { (char *)"delete_TPMS_PCR_SELECTION", _wrap_delete_TPMS_PCR_SELECTION, METH_VARARGS, (char *)"delete_TPMS_PCR_SELECTION(TPMS_PCR_SELECTION self)"},
	 { (char *)"TPMS_PCR_SELECTION_swigregister", TPMS_PCR_SELECTION_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_tag_set", _wrap_TPMT_TK_CREATION_tag_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_tag_get", _wrap_TPMT_TK_CREATION_tag_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_hierarchy_set", _wrap_TPMT_TK_CREATION_hierarchy_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_hierarchy_get", _wrap_TPMT_TK_CREATION_hierarchy_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_digest_set", _wrap_TPMT_TK_CREATION_digest_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_digest_get", _wrap_TPMT_TK_CREATION_digest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_CREATION", _wrap_new_TPMT_TK_CREATION, METH_VARARGS, (char *)"new_TPMT_TK_CREATION() -> TPMT_TK_CREATION"},
	 { (char *)"delete_TPMT_TK_CREATION", _wrap_delete_TPMT_TK_CREATION, METH_VARARGS, (char *)"delete_TPMT_TK_CREATION(TPMT_TK_CREATION self)"},
	 { (char *)"TPMT_TK_CREATION_swigregister", TPMT_TK_CREATION_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_tag_set", _wrap_TPMT_TK_VERIFIED_tag_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_tag_get", _wrap_TPMT_TK_VERIFIED_tag_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_hierarchy_set", _wrap_TPMT_TK_VERIFIED_hierarchy_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_hierarchy_get", _wrap_TPMT_TK_VERIFIED_hierarchy_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_digest_set", _wrap_TPMT_TK_VERIFIED_digest_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_digest_get", _wrap_TPMT_TK_VERIFIED_digest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_VERIFIED", _wrap_new_TPMT_TK_VERIFIED, METH_VARARGS, (char *)"new_TPMT_TK_VERIFIED() -> TPMT_TK_VERIFIED"},
	 { (char *)"delete_TPMT_TK_VERIFIED", _wrap_delete_TPMT_TK_VERIFIED, METH_VARARGS, (char *)"delete_TPMT_TK_VERIFIED(TPMT_TK_VERIFIED self)"},
	 { (char *)"TPMT_TK_VERIFIED_swigregister", TPMT_TK_VERIFIED_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_tag_set", _wrap_TPMT_TK_AUTH_tag_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_tag_get", _wrap_TPMT_TK_AUTH_tag_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_hierarchy_set", _wrap_TPMT_TK_AUTH_hierarchy_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_hierarchy_get", _wrap_TPMT_TK_AUTH_hierarchy_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_digest_set", _wrap_TPMT_TK_AUTH_digest_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_digest_get", _wrap_TPMT_TK_AUTH_digest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_AUTH", _wrap_new_TPMT_TK_AUTH, METH_VARARGS, (char *)"new_TPMT_TK_AUTH() -> TPMT_TK_AUTH"},
	 { (char *)"delete_TPMT_TK_AUTH", _wrap_delete_TPMT_TK_AUTH, METH_VARARGS, (char *)"delete_TPMT_TK_AUTH(TPMT_TK_AUTH self)"},
	 { (char *)"TPMT_TK_AUTH_swigregister", TPMT_TK_AUTH_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_tag_set", _wrap_TPMT_TK_HASHCHECK_tag_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_tag_get", _wrap_TPMT_TK_HASHCHECK_tag_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_hierarchy_set", _wrap_TPMT_TK_HASHCHECK_hierarchy_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_hierarchy_get", _wrap_TPMT_TK_HASHCHECK_hierarchy_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_digest_set", _wrap_TPMT_TK_HASHCHECK_digest_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_digest_get", _wrap_TPMT_TK_HASHCHECK_digest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_HASHCHECK", _wrap_new_TPMT_TK_HASHCHECK, METH_VARARGS, (char *)"new_TPMT_TK_HASHCHECK() -> TPMT_TK_HASHCHECK"},
	 { (char *)"delete_TPMT_TK_HASHCHECK", _wrap_delete_TPMT_TK_HASHCHECK, METH_VARARGS, (char *)"delete_TPMT_TK_HASHCHECK(TPMT_TK_HASHCHECK self)"},
	 { (char *)"TPMT_TK_HASHCHECK_swigregister", TPMT_TK_HASHCHECK_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_alg_set", _wrap_TPMS_ALG_PROPERTY_alg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_alg_get", _wrap_TPMS_ALG_PROPERTY_alg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_algProperties_set", _wrap_TPMS_ALG_PROPERTY_algProperties_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_algProperties_get", _wrap_TPMS_ALG_PROPERTY_algProperties_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALG_PROPERTY", _wrap_new_TPMS_ALG_PROPERTY, METH_VARARGS, (char *)"new_TPMS_ALG_PROPERTY() -> TPMS_ALG_PROPERTY"},
	 { (char *)"delete_TPMS_ALG_PROPERTY", _wrap_delete_TPMS_ALG_PROPERTY, METH_VARARGS, (char *)"delete_TPMS_ALG_PROPERTY(TPMS_ALG_PROPERTY self)"},
	 { (char *)"TPMS_ALG_PROPERTY_swigregister", TPMS_ALG_PROPERTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_property_set", _wrap_TPMS_TAGGED_PROPERTY_property_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_property_get", _wrap_TPMS_TAGGED_PROPERTY_property_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_value_set", _wrap_TPMS_TAGGED_PROPERTY_value_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_value_get", _wrap_TPMS_TAGGED_PROPERTY_value_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PROPERTY", _wrap_new_TPMS_TAGGED_PROPERTY, METH_VARARGS, (char *)"new_TPMS_TAGGED_PROPERTY() -> TPMS_TAGGED_PROPERTY"},
	 { (char *)"delete_TPMS_TAGGED_PROPERTY", _wrap_delete_TPMS_TAGGED_PROPERTY, METH_VARARGS, (char *)"delete_TPMS_TAGGED_PROPERTY(TPMS_TAGGED_PROPERTY self)"},
	 { (char *)"TPMS_TAGGED_PROPERTY_swigregister", TPMS_TAGGED_PROPERTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_tag_set", _wrap_TPMS_TAGGED_PCR_SELECT_tag_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_tag_get", _wrap_TPMS_TAGGED_PCR_SELECT_tag_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_sizeofSelect_set", _wrap_TPMS_TAGGED_PCR_SELECT_sizeofSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_sizeofSelect_get", _wrap_TPMS_TAGGED_PCR_SELECT_sizeofSelect_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_pcrSelect_set", _wrap_TPMS_TAGGED_PCR_SELECT_pcrSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_pcrSelect_get", _wrap_TPMS_TAGGED_PCR_SELECT_pcrSelect_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PCR_SELECT", _wrap_new_TPMS_TAGGED_PCR_SELECT, METH_VARARGS, (char *)"new_TPMS_TAGGED_PCR_SELECT() -> TPMS_TAGGED_PCR_SELECT"},
	 { (char *)"delete_TPMS_TAGGED_PCR_SELECT", _wrap_delete_TPMS_TAGGED_PCR_SELECT, METH_VARARGS, (char *)"delete_TPMS_TAGGED_PCR_SELECT(TPMS_TAGGED_PCR_SELECT self)"},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_swigregister", TPMS_TAGGED_PCR_SELECT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_handle_set", _wrap_TPMS_TAGGED_POLICY_handle_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_handle_get", _wrap_TPMS_TAGGED_POLICY_handle_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_policyHash_set", _wrap_TPMS_TAGGED_POLICY_policyHash_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_policyHash_get", _wrap_TPMS_TAGGED_POLICY_policyHash_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_POLICY", _wrap_new_TPMS_TAGGED_POLICY, METH_VARARGS, (char *)"new_TPMS_TAGGED_POLICY() -> TPMS_TAGGED_POLICY"},
	 { (char *)"delete_TPMS_TAGGED_POLICY", _wrap_delete_TPMS_TAGGED_POLICY, METH_VARARGS, (char *)"delete_TPMS_TAGGED_POLICY(TPMS_TAGGED_POLICY self)"},
	 { (char *)"TPMS_TAGGED_POLICY_swigregister", TPMS_TAGGED_POLICY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_count_set", _wrap_TPML_CC_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_count_get", _wrap_TPML_CC_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_commandCodes_set", _wrap_TPML_CC_commandCodes_set, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_commandCodes_get", _wrap_TPML_CC_commandCodes_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CC", _wrap_new_TPML_CC, METH_VARARGS, (char *)"new_TPML_CC() -> TPML_CC"},
	 { (char *)"delete_TPML_CC", _wrap_delete_TPML_CC, METH_VARARGS, (char *)"delete_TPML_CC(TPML_CC self)"},
	 { (char *)"TPML_CC_swigregister", TPML_CC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_count_set", _wrap_TPML_CCA_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_count_get", _wrap_TPML_CCA_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_commandAttributes_set", _wrap_TPML_CCA_commandAttributes_set, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_commandAttributes_get", _wrap_TPML_CCA_commandAttributes_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CCA", _wrap_new_TPML_CCA, METH_VARARGS, (char *)"new_TPML_CCA() -> TPML_CCA"},
	 { (char *)"delete_TPML_CCA", _wrap_delete_TPML_CCA, METH_VARARGS, (char *)"delete_TPML_CCA(TPML_CCA self)"},
	 { (char *)"TPML_CCA_swigregister", TPML_CCA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_count_set", _wrap_TPML_ALG_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_count_get", _wrap_TPML_ALG_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_algorithms_set", _wrap_TPML_ALG_algorithms_set, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_algorithms_get", _wrap_TPML_ALG_algorithms_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG", _wrap_new_TPML_ALG, METH_VARARGS, (char *)"new_TPML_ALG() -> TPML_ALG"},
	 { (char *)"delete_TPML_ALG", _wrap_delete_TPML_ALG, METH_VARARGS, (char *)"delete_TPML_ALG(TPML_ALG self)"},
	 { (char *)"TPML_ALG_swigregister", TPML_ALG_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_count_set", _wrap_TPML_HANDLE_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_count_get", _wrap_TPML_HANDLE_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_handle_set", _wrap_TPML_HANDLE_handle_set, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_handle_get", _wrap_TPML_HANDLE_handle_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_HANDLE", _wrap_new_TPML_HANDLE, METH_VARARGS, (char *)"new_TPML_HANDLE() -> TPML_HANDLE"},
	 { (char *)"delete_TPML_HANDLE", _wrap_delete_TPML_HANDLE, METH_VARARGS, (char *)"delete_TPML_HANDLE(TPML_HANDLE self)"},
	 { (char *)"TPML_HANDLE_swigregister", TPML_HANDLE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_count_set", _wrap_TPML_DIGEST_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_count_get", _wrap_TPML_DIGEST_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_digests_set", _wrap_TPML_DIGEST_digests_set, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_digests_get", _wrap_TPML_DIGEST_digests_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST", _wrap_new_TPML_DIGEST, METH_VARARGS, (char *)"new_TPML_DIGEST() -> TPML_DIGEST"},
	 { (char *)"delete_TPML_DIGEST", _wrap_delete_TPML_DIGEST, METH_VARARGS, (char *)"delete_TPML_DIGEST(TPML_DIGEST self)"},
	 { (char *)"TPML_DIGEST_swigregister", TPML_DIGEST_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_count_set", _wrap_TPML_DIGEST_VALUES_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_count_get", _wrap_TPML_DIGEST_VALUES_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_digests_set", _wrap_TPML_DIGEST_VALUES_digests_set, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_digests_get", _wrap_TPML_DIGEST_VALUES_digests_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_VALUES", _wrap_new_TPML_DIGEST_VALUES, METH_VARARGS, (char *)"new_TPML_DIGEST_VALUES() -> TPML_DIGEST_VALUES"},
	 { (char *)"delete_TPML_DIGEST_VALUES", _wrap_delete_TPML_DIGEST_VALUES, METH_VARARGS, (char *)"delete_TPML_DIGEST_VALUES(TPML_DIGEST_VALUES self)"},
	 { (char *)"TPML_DIGEST_VALUES_swigregister", TPML_DIGEST_VALUES_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_count_set", _wrap_TPML_PCR_SELECTION_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_count_get", _wrap_TPML_PCR_SELECTION_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_pcrSelections_set", _wrap_TPML_PCR_SELECTION_pcrSelections_set, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_pcrSelections_get", _wrap_TPML_PCR_SELECTION_pcrSelections_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_PCR_SELECTION", _wrap_new_TPML_PCR_SELECTION, METH_VARARGS, (char *)"new_TPML_PCR_SELECTION() -> TPML_PCR_SELECTION"},
	 { (char *)"delete_TPML_PCR_SELECTION", _wrap_delete_TPML_PCR_SELECTION, METH_VARARGS, (char *)"delete_TPML_PCR_SELECTION(TPML_PCR_SELECTION self)"},
	 { (char *)"TPML_PCR_SELECTION_swigregister", TPML_PCR_SELECTION_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_count_set", _wrap_TPML_ALG_PROPERTY_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_count_get", _wrap_TPML_ALG_PROPERTY_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_algProperties_set", _wrap_TPML_ALG_PROPERTY_algProperties_set, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_algProperties_get", _wrap_TPML_ALG_PROPERTY_algProperties_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_PROPERTY", _wrap_new_TPML_ALG_PROPERTY, METH_VARARGS, (char *)"new_TPML_ALG_PROPERTY() -> TPML_ALG_PROPERTY"},
	 { (char *)"delete_TPML_ALG_PROPERTY", _wrap_delete_TPML_ALG_PROPERTY, METH_VARARGS, (char *)"delete_TPML_ALG_PROPERTY(TPML_ALG_PROPERTY self)"},
	 { (char *)"TPML_ALG_PROPERTY_swigregister", TPML_ALG_PROPERTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_count_set", _wrap_TPML_TAGGED_TPM_PROPERTY_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_count_get", _wrap_TPML_TAGGED_TPM_PROPERTY_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_tpmProperty_set", _wrap_TPML_TAGGED_TPM_PROPERTY_tpmProperty_set, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_tpmProperty_get", _wrap_TPML_TAGGED_TPM_PROPERTY_tpmProperty_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_TPM_PROPERTY", _wrap_new_TPML_TAGGED_TPM_PROPERTY, METH_VARARGS, (char *)"new_TPML_TAGGED_TPM_PROPERTY() -> TPML_TAGGED_TPM_PROPERTY"},
	 { (char *)"delete_TPML_TAGGED_TPM_PROPERTY", _wrap_delete_TPML_TAGGED_TPM_PROPERTY, METH_VARARGS, (char *)"delete_TPML_TAGGED_TPM_PROPERTY(TPML_TAGGED_TPM_PROPERTY self)"},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_swigregister", TPML_TAGGED_TPM_PROPERTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_count_set", _wrap_TPML_TAGGED_PCR_PROPERTY_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_count_get", _wrap_TPML_TAGGED_PCR_PROPERTY_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_pcrProperty_set", _wrap_TPML_TAGGED_PCR_PROPERTY_pcrProperty_set, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_pcrProperty_get", _wrap_TPML_TAGGED_PCR_PROPERTY_pcrProperty_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_PCR_PROPERTY", _wrap_new_TPML_TAGGED_PCR_PROPERTY, METH_VARARGS, (char *)"new_TPML_TAGGED_PCR_PROPERTY() -> TPML_TAGGED_PCR_PROPERTY"},
	 { (char *)"delete_TPML_TAGGED_PCR_PROPERTY", _wrap_delete_TPML_TAGGED_PCR_PROPERTY, METH_VARARGS, (char *)"delete_TPML_TAGGED_PCR_PROPERTY(TPML_TAGGED_PCR_PROPERTY self)"},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_swigregister", TPML_TAGGED_PCR_PROPERTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_count_set", _wrap_TPML_ECC_CURVE_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_count_get", _wrap_TPML_ECC_CURVE_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_eccCurves_set", _wrap_TPML_ECC_CURVE_eccCurves_set, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_eccCurves_get", _wrap_TPML_ECC_CURVE_eccCurves_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ECC_CURVE", _wrap_new_TPML_ECC_CURVE, METH_VARARGS, (char *)"new_TPML_ECC_CURVE() -> TPML_ECC_CURVE"},
	 { (char *)"delete_TPML_ECC_CURVE", _wrap_delete_TPML_ECC_CURVE, METH_VARARGS, (char *)"delete_TPML_ECC_CURVE(TPML_ECC_CURVE self)"},
	 { (char *)"TPML_ECC_CURVE_swigregister", TPML_ECC_CURVE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_count_set", _wrap_TPML_INTEL_PTT_PROPERTY_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_count_get", _wrap_TPML_INTEL_PTT_PROPERTY_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_property_set", _wrap_TPML_INTEL_PTT_PROPERTY_property_set, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_property_get", _wrap_TPML_INTEL_PTT_PROPERTY_property_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_INTEL_PTT_PROPERTY", _wrap_new_TPML_INTEL_PTT_PROPERTY, METH_VARARGS, (char *)"new_TPML_INTEL_PTT_PROPERTY() -> TPML_INTEL_PTT_PROPERTY"},
	 { (char *)"delete_TPML_INTEL_PTT_PROPERTY", _wrap_delete_TPML_INTEL_PTT_PROPERTY, METH_VARARGS, (char *)"delete_TPML_INTEL_PTT_PROPERTY(TPML_INTEL_PTT_PROPERTY self)"},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_swigregister", TPML_INTEL_PTT_PROPERTY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_algorithms_set", _wrap_TPMU_CAPABILITIES_algorithms_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_algorithms_get", _wrap_TPMU_CAPABILITIES_algorithms_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_handles_set", _wrap_TPMU_CAPABILITIES_handles_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_handles_get", _wrap_TPMU_CAPABILITIES_handles_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_command_set", _wrap_TPMU_CAPABILITIES_command_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_command_get", _wrap_TPMU_CAPABILITIES_command_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ppCommands_set", _wrap_TPMU_CAPABILITIES_ppCommands_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ppCommands_get", _wrap_TPMU_CAPABILITIES_ppCommands_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_auditCommands_set", _wrap_TPMU_CAPABILITIES_auditCommands_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_auditCommands_get", _wrap_TPMU_CAPABILITIES_auditCommands_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_assignedPCR_set", _wrap_TPMU_CAPABILITIES_assignedPCR_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_assignedPCR_get", _wrap_TPMU_CAPABILITIES_assignedPCR_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_tpmProperties_set", _wrap_TPMU_CAPABILITIES_tpmProperties_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_tpmProperties_get", _wrap_TPMU_CAPABILITIES_tpmProperties_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_pcrProperties_set", _wrap_TPMU_CAPABILITIES_pcrProperties_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_pcrProperties_get", _wrap_TPMU_CAPABILITIES_pcrProperties_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_eccCurves_set", _wrap_TPMU_CAPABILITIES_eccCurves_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_eccCurves_get", _wrap_TPMU_CAPABILITIES_eccCurves_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_intelPttProperty_set", _wrap_TPMU_CAPABILITIES_intelPttProperty_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_intelPttProperty_get", _wrap_TPMU_CAPABILITIES_intelPttProperty_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_CAPABILITIES", _wrap_new_TPMU_CAPABILITIES, METH_VARARGS, (char *)"new_TPMU_CAPABILITIES() -> TPMU_CAPABILITIES"},
	 { (char *)"delete_TPMU_CAPABILITIES", _wrap_delete_TPMU_CAPABILITIES, METH_VARARGS, (char *)"delete_TPMU_CAPABILITIES(TPMU_CAPABILITIES self)"},
	 { (char *)"TPMU_CAPABILITIES_swigregister", TPMU_CAPABILITIES_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_capability_set", _wrap_TPMS_CAPABILITY_DATA_capability_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_capability_get", _wrap_TPMS_CAPABILITY_DATA_capability_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_data_set", _wrap_TPMS_CAPABILITY_DATA_data_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_data_get", _wrap_TPMS_CAPABILITY_DATA_data_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CAPABILITY_DATA", _wrap_new_TPMS_CAPABILITY_DATA, METH_VARARGS, (char *)"new_TPMS_CAPABILITY_DATA() -> TPMS_CAPABILITY_DATA"},
	 { (char *)"delete_TPMS_CAPABILITY_DATA", _wrap_delete_TPMS_CAPABILITY_DATA, METH_VARARGS, (char *)"delete_TPMS_CAPABILITY_DATA(TPMS_CAPABILITY_DATA self)"},
	 { (char *)"TPMS_CAPABILITY_DATA_swigregister", TPMS_CAPABILITY_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_clock_set", _wrap_TPMS_CLOCK_INFO_clock_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_clock_get", _wrap_TPMS_CLOCK_INFO_clock_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_resetCount_set", _wrap_TPMS_CLOCK_INFO_resetCount_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_resetCount_get", _wrap_TPMS_CLOCK_INFO_resetCount_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_restartCount_set", _wrap_TPMS_CLOCK_INFO_restartCount_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_restartCount_get", _wrap_TPMS_CLOCK_INFO_restartCount_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_safe_set", _wrap_TPMS_CLOCK_INFO_safe_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_safe_get", _wrap_TPMS_CLOCK_INFO_safe_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CLOCK_INFO", _wrap_new_TPMS_CLOCK_INFO, METH_VARARGS, (char *)"new_TPMS_CLOCK_INFO() -> TPMS_CLOCK_INFO"},
	 { (char *)"delete_TPMS_CLOCK_INFO", _wrap_delete_TPMS_CLOCK_INFO, METH_VARARGS, (char *)"delete_TPMS_CLOCK_INFO(TPMS_CLOCK_INFO self)"},
	 { (char *)"TPMS_CLOCK_INFO_swigregister", TPMS_CLOCK_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_time_set", _wrap_TPMS_TIME_INFO_time_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_time_get", _wrap_TPMS_TIME_INFO_time_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_clockInfo_set", _wrap_TPMS_TIME_INFO_clockInfo_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_clockInfo_get", _wrap_TPMS_TIME_INFO_clockInfo_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_INFO", _wrap_new_TPMS_TIME_INFO, METH_VARARGS, (char *)"new_TPMS_TIME_INFO() -> TPMS_TIME_INFO"},
	 { (char *)"delete_TPMS_TIME_INFO", _wrap_delete_TPMS_TIME_INFO, METH_VARARGS, (char *)"delete_TPMS_TIME_INFO(TPMS_TIME_INFO self)"},
	 { (char *)"TPMS_TIME_INFO_swigregister", TPMS_TIME_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_time_set", _wrap_TPMS_TIME_ATTEST_INFO_time_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_time_get", _wrap_TPMS_TIME_ATTEST_INFO_time_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_firmwareVersion_set", _wrap_TPMS_TIME_ATTEST_INFO_firmwareVersion_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_firmwareVersion_get", _wrap_TPMS_TIME_ATTEST_INFO_firmwareVersion_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_ATTEST_INFO", _wrap_new_TPMS_TIME_ATTEST_INFO, METH_VARARGS, (char *)"new_TPMS_TIME_ATTEST_INFO() -> TPMS_TIME_ATTEST_INFO"},
	 { (char *)"delete_TPMS_TIME_ATTEST_INFO", _wrap_delete_TPMS_TIME_ATTEST_INFO, METH_VARARGS, (char *)"delete_TPMS_TIME_ATTEST_INFO(TPMS_TIME_ATTEST_INFO self)"},
	 { (char *)"TPMS_TIME_ATTEST_INFO_swigregister", TPMS_TIME_ATTEST_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_name_set", _wrap_TPMS_CERTIFY_INFO_name_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_name_get", _wrap_TPMS_CERTIFY_INFO_name_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_qualifiedName_set", _wrap_TPMS_CERTIFY_INFO_qualifiedName_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_qualifiedName_get", _wrap_TPMS_CERTIFY_INFO_qualifiedName_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CERTIFY_INFO", _wrap_new_TPMS_CERTIFY_INFO, METH_VARARGS, (char *)"new_TPMS_CERTIFY_INFO() -> TPMS_CERTIFY_INFO"},
	 { (char *)"delete_TPMS_CERTIFY_INFO", _wrap_delete_TPMS_CERTIFY_INFO, METH_VARARGS, (char *)"delete_TPMS_CERTIFY_INFO(TPMS_CERTIFY_INFO self)"},
	 { (char *)"TPMS_CERTIFY_INFO_swigregister", TPMS_CERTIFY_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_pcrSelect_set", _wrap_TPMS_QUOTE_INFO_pcrSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_pcrSelect_get", _wrap_TPMS_QUOTE_INFO_pcrSelect_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_pcrDigest_set", _wrap_TPMS_QUOTE_INFO_pcrDigest_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_pcrDigest_get", _wrap_TPMS_QUOTE_INFO_pcrDigest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_QUOTE_INFO", _wrap_new_TPMS_QUOTE_INFO, METH_VARARGS, (char *)"new_TPMS_QUOTE_INFO() -> TPMS_QUOTE_INFO"},
	 { (char *)"delete_TPMS_QUOTE_INFO", _wrap_delete_TPMS_QUOTE_INFO, METH_VARARGS, (char *)"delete_TPMS_QUOTE_INFO(TPMS_QUOTE_INFO self)"},
	 { (char *)"TPMS_QUOTE_INFO_swigregister", TPMS_QUOTE_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_auditCounter_set", _wrap_TPMS_COMMAND_AUDIT_INFO_auditCounter_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_auditCounter_get", _wrap_TPMS_COMMAND_AUDIT_INFO_auditCounter_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_digestAlg_set", _wrap_TPMS_COMMAND_AUDIT_INFO_digestAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_digestAlg_get", _wrap_TPMS_COMMAND_AUDIT_INFO_digestAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_auditDigest_set", _wrap_TPMS_COMMAND_AUDIT_INFO_auditDigest_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_auditDigest_get", _wrap_TPMS_COMMAND_AUDIT_INFO_auditDigest_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_commandDigest_set", _wrap_TPMS_COMMAND_AUDIT_INFO_commandDigest_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_commandDigest_get", _wrap_TPMS_COMMAND_AUDIT_INFO_commandDigest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_COMMAND_AUDIT_INFO", _wrap_new_TPMS_COMMAND_AUDIT_INFO, METH_VARARGS, (char *)"new_TPMS_COMMAND_AUDIT_INFO() -> TPMS_COMMAND_AUDIT_INFO"},
	 { (char *)"delete_TPMS_COMMAND_AUDIT_INFO", _wrap_delete_TPMS_COMMAND_AUDIT_INFO, METH_VARARGS, (char *)"delete_TPMS_COMMAND_AUDIT_INFO(TPMS_COMMAND_AUDIT_INFO self)"},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_swigregister", TPMS_COMMAND_AUDIT_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_exclusiveSession_set", _wrap_TPMS_SESSION_AUDIT_INFO_exclusiveSession_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_exclusiveSession_get", _wrap_TPMS_SESSION_AUDIT_INFO_exclusiveSession_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_sessionDigest_set", _wrap_TPMS_SESSION_AUDIT_INFO_sessionDigest_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_sessionDigest_get", _wrap_TPMS_SESSION_AUDIT_INFO_sessionDigest_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SESSION_AUDIT_INFO", _wrap_new_TPMS_SESSION_AUDIT_INFO, METH_VARARGS, (char *)"new_TPMS_SESSION_AUDIT_INFO() -> TPMS_SESSION_AUDIT_INFO"},
	 { (char *)"delete_TPMS_SESSION_AUDIT_INFO", _wrap_delete_TPMS_SESSION_AUDIT_INFO, METH_VARARGS, (char *)"delete_TPMS_SESSION_AUDIT_INFO(TPMS_SESSION_AUDIT_INFO self)"},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_swigregister", TPMS_SESSION_AUDIT_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_objectName_set", _wrap_TPMS_CREATION_INFO_objectName_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_objectName_get", _wrap_TPMS_CREATION_INFO_objectName_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_creationHash_set", _wrap_TPMS_CREATION_INFO_creationHash_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_creationHash_get", _wrap_TPMS_CREATION_INFO_creationHash_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_INFO", _wrap_new_TPMS_CREATION_INFO, METH_VARARGS, (char *)"new_TPMS_CREATION_INFO() -> TPMS_CREATION_INFO"},
	 { (char *)"delete_TPMS_CREATION_INFO", _wrap_delete_TPMS_CREATION_INFO, METH_VARARGS, (char *)"delete_TPMS_CREATION_INFO(TPMS_CREATION_INFO self)"},
	 { (char *)"TPMS_CREATION_INFO_swigregister", TPMS_CREATION_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_indexName_set", _wrap_TPMS_NV_CERTIFY_INFO_indexName_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_indexName_get", _wrap_TPMS_NV_CERTIFY_INFO_indexName_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_offset_set", _wrap_TPMS_NV_CERTIFY_INFO_offset_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_offset_get", _wrap_TPMS_NV_CERTIFY_INFO_offset_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_nvContents_set", _wrap_TPMS_NV_CERTIFY_INFO_nvContents_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_nvContents_get", _wrap_TPMS_NV_CERTIFY_INFO_nvContents_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_CERTIFY_INFO", _wrap_new_TPMS_NV_CERTIFY_INFO, METH_VARARGS, (char *)"new_TPMS_NV_CERTIFY_INFO() -> TPMS_NV_CERTIFY_INFO"},
	 { (char *)"delete_TPMS_NV_CERTIFY_INFO", _wrap_delete_TPMS_NV_CERTIFY_INFO, METH_VARARGS, (char *)"delete_TPMS_NV_CERTIFY_INFO(TPMS_NV_CERTIFY_INFO self)"},
	 { (char *)"TPMS_NV_CERTIFY_INFO_swigregister", TPMS_NV_CERTIFY_INFO_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_certify_set", _wrap_TPMU_ATTEST_certify_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_certify_get", _wrap_TPMU_ATTEST_certify_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_creation_set", _wrap_TPMU_ATTEST_creation_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_creation_get", _wrap_TPMU_ATTEST_creation_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_quote_set", _wrap_TPMU_ATTEST_quote_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_quote_get", _wrap_TPMU_ATTEST_quote_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_commandAudit_set", _wrap_TPMU_ATTEST_commandAudit_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_commandAudit_get", _wrap_TPMU_ATTEST_commandAudit_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_sessionAudit_set", _wrap_TPMU_ATTEST_sessionAudit_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_sessionAudit_get", _wrap_TPMU_ATTEST_sessionAudit_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_time_set", _wrap_TPMU_ATTEST_time_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_time_get", _wrap_TPMU_ATTEST_time_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_nv_set", _wrap_TPMU_ATTEST_nv_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_nv_get", _wrap_TPMU_ATTEST_nv_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ATTEST", _wrap_new_TPMU_ATTEST, METH_VARARGS, (char *)"new_TPMU_ATTEST() -> TPMU_ATTEST"},
	 { (char *)"delete_TPMU_ATTEST", _wrap_delete_TPMU_ATTEST, METH_VARARGS, (char *)"delete_TPMU_ATTEST(TPMU_ATTEST self)"},
	 { (char *)"TPMU_ATTEST_swigregister", TPMU_ATTEST_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_magic_set", _wrap_TPMS_ATTEST_magic_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_magic_get", _wrap_TPMS_ATTEST_magic_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_type_set", _wrap_TPMS_ATTEST_type_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_type_get", _wrap_TPMS_ATTEST_type_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_qualifiedSigner_set", _wrap_TPMS_ATTEST_qualifiedSigner_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_qualifiedSigner_get", _wrap_TPMS_ATTEST_qualifiedSigner_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_extraData_set", _wrap_TPMS_ATTEST_extraData_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_extraData_get", _wrap_TPMS_ATTEST_extraData_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_clockInfo_set", _wrap_TPMS_ATTEST_clockInfo_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_clockInfo_get", _wrap_TPMS_ATTEST_clockInfo_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_firmwareVersion_set", _wrap_TPMS_ATTEST_firmwareVersion_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_firmwareVersion_get", _wrap_TPMS_ATTEST_firmwareVersion_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_attested_set", _wrap_TPMS_ATTEST_attested_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_attested_get", _wrap_TPMS_ATTEST_attested_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ATTEST", _wrap_new_TPMS_ATTEST, METH_VARARGS, (char *)"new_TPMS_ATTEST() -> TPMS_ATTEST"},
	 { (char *)"delete_TPMS_ATTEST", _wrap_delete_TPMS_ATTEST, METH_VARARGS, (char *)"delete_TPMS_ATTEST(TPMS_ATTEST self)"},
	 { (char *)"TPMS_ATTEST_swigregister", TPMS_ATTEST_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_size_set", _wrap_TPM2B_ATTEST_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_size_get", _wrap_TPM2B_ATTEST_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_attestationData_set", _wrap_TPM2B_ATTEST_attestationData_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_attestationData_get", _wrap_TPM2B_ATTEST_attestationData_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ATTEST", _wrap_new_TPM2B_ATTEST, METH_VARARGS, (char *)"new_TPM2B_ATTEST() -> TPM2B_ATTEST"},
	 { (char *)"delete_TPM2B_ATTEST", _wrap_delete_TPM2B_ATTEST, METH_VARARGS, (char *)"delete_TPM2B_ATTEST(TPM2B_ATTEST self)"},
	 { (char *)"TPM2B_ATTEST_swigregister", TPM2B_ATTEST_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_sessionHandle_set", _wrap_TPMS_AUTH_COMMAND_sessionHandle_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_sessionHandle_get", _wrap_TPMS_AUTH_COMMAND_sessionHandle_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_nonce_set", _wrap_TPMS_AUTH_COMMAND_nonce_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_nonce_get", _wrap_TPMS_AUTH_COMMAND_nonce_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_sessionAttributes_set", _wrap_TPMS_AUTH_COMMAND_sessionAttributes_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_sessionAttributes_get", _wrap_TPMS_AUTH_COMMAND_sessionAttributes_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_hmac_set", _wrap_TPMS_AUTH_COMMAND_hmac_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_hmac_get", _wrap_TPMS_AUTH_COMMAND_hmac_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_COMMAND", _wrap_new_TPMS_AUTH_COMMAND, METH_VARARGS, (char *)"new_TPMS_AUTH_COMMAND() -> TPMS_AUTH_COMMAND"},
	 { (char *)"delete_TPMS_AUTH_COMMAND", _wrap_delete_TPMS_AUTH_COMMAND, METH_VARARGS, (char *)"delete_TPMS_AUTH_COMMAND(TPMS_AUTH_COMMAND self)"},
	 { (char *)"TPMS_AUTH_COMMAND_swigregister", TPMS_AUTH_COMMAND_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_nonce_set", _wrap_TPMS_AUTH_RESPONSE_nonce_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_nonce_get", _wrap_TPMS_AUTH_RESPONSE_nonce_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_sessionAttributes_set", _wrap_TPMS_AUTH_RESPONSE_sessionAttributes_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_sessionAttributes_get", _wrap_TPMS_AUTH_RESPONSE_sessionAttributes_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_hmac_set", _wrap_TPMS_AUTH_RESPONSE_hmac_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_hmac_get", _wrap_TPMS_AUTH_RESPONSE_hmac_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_RESPONSE", _wrap_new_TPMS_AUTH_RESPONSE, METH_VARARGS, (char *)"new_TPMS_AUTH_RESPONSE() -> TPMS_AUTH_RESPONSE"},
	 { (char *)"delete_TPMS_AUTH_RESPONSE", _wrap_delete_TPMS_AUTH_RESPONSE, METH_VARARGS, (char *)"delete_TPMS_AUTH_RESPONSE(TPMS_AUTH_RESPONSE self)"},
	 { (char *)"TPMS_AUTH_RESPONSE_swigregister", TPMS_AUTH_RESPONSE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_aes_set", _wrap_TPMU_SYM_KEY_BITS_aes_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_aes_get", _wrap_TPMU_SYM_KEY_BITS_aes_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_sm4_set", _wrap_TPMU_SYM_KEY_BITS_sm4_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_sm4_get", _wrap_TPMU_SYM_KEY_BITS_sm4_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_camellia_set", _wrap_TPMU_SYM_KEY_BITS_camellia_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_camellia_get", _wrap_TPMU_SYM_KEY_BITS_camellia_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_sym_set", _wrap_TPMU_SYM_KEY_BITS_sym_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_sym_get", _wrap_TPMU_SYM_KEY_BITS_sym_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_exclusiveOr_set", _wrap_TPMU_SYM_KEY_BITS_exclusiveOr_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_exclusiveOr_get", _wrap_TPMU_SYM_KEY_BITS_exclusiveOr_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_KEY_BITS", _wrap_new_TPMU_SYM_KEY_BITS, METH_VARARGS, (char *)"new_TPMU_SYM_KEY_BITS() -> TPMU_SYM_KEY_BITS"},
	 { (char *)"delete_TPMU_SYM_KEY_BITS", _wrap_delete_TPMU_SYM_KEY_BITS, METH_VARARGS, (char *)"delete_TPMU_SYM_KEY_BITS(TPMU_SYM_KEY_BITS self)"},
	 { (char *)"TPMU_SYM_KEY_BITS_swigregister", TPMU_SYM_KEY_BITS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_aes_set", _wrap_TPMU_SYM_MODE_aes_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_aes_get", _wrap_TPMU_SYM_MODE_aes_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_sm4_set", _wrap_TPMU_SYM_MODE_sm4_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_sm4_get", _wrap_TPMU_SYM_MODE_sm4_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_camellia_set", _wrap_TPMU_SYM_MODE_camellia_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_camellia_get", _wrap_TPMU_SYM_MODE_camellia_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_sym_set", _wrap_TPMU_SYM_MODE_sym_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_sym_get", _wrap_TPMU_SYM_MODE_sym_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_MODE", _wrap_new_TPMU_SYM_MODE, METH_VARARGS, (char *)"new_TPMU_SYM_MODE() -> TPMU_SYM_MODE"},
	 { (char *)"delete_TPMU_SYM_MODE", _wrap_delete_TPMU_SYM_MODE, METH_VARARGS, (char *)"delete_TPMU_SYM_MODE(TPMU_SYM_MODE self)"},
	 { (char *)"TPMU_SYM_MODE_swigregister", TPMU_SYM_MODE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_algorithm_set", _wrap_TPMT_SYM_DEF_algorithm_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_algorithm_get", _wrap_TPMT_SYM_DEF_algorithm_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_keyBits_set", _wrap_TPMT_SYM_DEF_keyBits_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_keyBits_get", _wrap_TPMT_SYM_DEF_keyBits_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_mode_set", _wrap_TPMT_SYM_DEF_mode_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_mode_get", _wrap_TPMT_SYM_DEF_mode_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF", _wrap_new_TPMT_SYM_DEF, METH_VARARGS, (char *)"new_TPMT_SYM_DEF() -> TPMT_SYM_DEF"},
	 { (char *)"delete_TPMT_SYM_DEF", _wrap_delete_TPMT_SYM_DEF, METH_VARARGS, (char *)"delete_TPMT_SYM_DEF(TPMT_SYM_DEF self)"},
	 { (char *)"TPMT_SYM_DEF_swigregister", TPMT_SYM_DEF_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_algorithm_set", _wrap_TPMT_SYM_DEF_OBJECT_algorithm_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_algorithm_get", _wrap_TPMT_SYM_DEF_OBJECT_algorithm_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_keyBits_set", _wrap_TPMT_SYM_DEF_OBJECT_keyBits_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_keyBits_get", _wrap_TPMT_SYM_DEF_OBJECT_keyBits_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_mode_set", _wrap_TPMT_SYM_DEF_OBJECT_mode_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_mode_get", _wrap_TPMT_SYM_DEF_OBJECT_mode_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_OBJECT", _wrap_new_TPMT_SYM_DEF_OBJECT, METH_VARARGS, (char *)"new_TPMT_SYM_DEF_OBJECT() -> TPMT_SYM_DEF_OBJECT"},
	 { (char *)"delete_TPMT_SYM_DEF_OBJECT", _wrap_delete_TPMT_SYM_DEF_OBJECT, METH_VARARGS, (char *)"delete_TPMT_SYM_DEF_OBJECT(TPMT_SYM_DEF_OBJECT self)"},
	 { (char *)"TPMT_SYM_DEF_OBJECT_swigregister", TPMT_SYM_DEF_OBJECT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_size_set", _wrap_TPM2B_SYM_KEY_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_size_get", _wrap_TPM2B_SYM_KEY_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_buffer_set", _wrap_TPM2B_SYM_KEY_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_buffer_get", _wrap_TPM2B_SYM_KEY_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SYM_KEY", _wrap_new_TPM2B_SYM_KEY, METH_VARARGS, (char *)"new_TPM2B_SYM_KEY() -> TPM2B_SYM_KEY"},
	 { (char *)"delete_TPM2B_SYM_KEY", _wrap_delete_TPM2B_SYM_KEY, METH_VARARGS, (char *)"delete_TPM2B_SYM_KEY(TPM2B_SYM_KEY self)"},
	 { (char *)"TPM2B_SYM_KEY_swigregister", TPM2B_SYM_KEY_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_sym_set", _wrap_TPMS_SYMCIPHER_PARMS_sym_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_sym_get", _wrap_TPMS_SYMCIPHER_PARMS_sym_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SYMCIPHER_PARMS", _wrap_new_TPMS_SYMCIPHER_PARMS, METH_VARARGS, (char *)"new_TPMS_SYMCIPHER_PARMS() -> TPMS_SYMCIPHER_PARMS"},
	 { (char *)"delete_TPMS_SYMCIPHER_PARMS", _wrap_delete_TPMS_SYMCIPHER_PARMS, METH_VARARGS, (char *)"delete_TPMS_SYMCIPHER_PARMS(TPMS_SYMCIPHER_PARMS self)"},
	 { (char *)"TPMS_SYMCIPHER_PARMS_swigregister", TPMS_SYMCIPHER_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_size_set", _wrap_TPM2B_SENSITIVE_DATA_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_size_get", _wrap_TPM2B_SENSITIVE_DATA_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_buffer_set", _wrap_TPM2B_SENSITIVE_DATA_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_buffer_get", _wrap_TPM2B_SENSITIVE_DATA_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_DATA", _wrap_new_TPM2B_SENSITIVE_DATA, METH_VARARGS, (char *)"new_TPM2B_SENSITIVE_DATA() -> TPM2B_SENSITIVE_DATA"},
	 { (char *)"delete_TPM2B_SENSITIVE_DATA", _wrap_delete_TPM2B_SENSITIVE_DATA, METH_VARARGS, (char *)"delete_TPM2B_SENSITIVE_DATA(TPM2B_SENSITIVE_DATA self)"},
	 { (char *)"TPM2B_SENSITIVE_DATA_swigregister", TPM2B_SENSITIVE_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_userAuth_set", _wrap_TPMS_SENSITIVE_CREATE_userAuth_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_userAuth_get", _wrap_TPMS_SENSITIVE_CREATE_userAuth_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_data_set", _wrap_TPMS_SENSITIVE_CREATE_data_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_data_get", _wrap_TPMS_SENSITIVE_CREATE_data_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SENSITIVE_CREATE", _wrap_new_TPMS_SENSITIVE_CREATE, METH_VARARGS, (char *)"new_TPMS_SENSITIVE_CREATE() -> TPMS_SENSITIVE_CREATE"},
	 { (char *)"delete_TPMS_SENSITIVE_CREATE", _wrap_delete_TPMS_SENSITIVE_CREATE, METH_VARARGS, (char *)"delete_TPMS_SENSITIVE_CREATE(TPMS_SENSITIVE_CREATE self)"},
	 { (char *)"TPMS_SENSITIVE_CREATE_swigregister", TPMS_SENSITIVE_CREATE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_size_set", _wrap_TPM2B_SENSITIVE_CREATE_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_size_get", _wrap_TPM2B_SENSITIVE_CREATE_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_sensitive_set", _wrap_TPM2B_SENSITIVE_CREATE_sensitive_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_sensitive_get", _wrap_TPM2B_SENSITIVE_CREATE_sensitive_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_CREATE", _wrap_new_TPM2B_SENSITIVE_CREATE, METH_VARARGS, (char *)"new_TPM2B_SENSITIVE_CREATE() -> TPM2B_SENSITIVE_CREATE"},
	 { (char *)"delete_TPM2B_SENSITIVE_CREATE", _wrap_delete_TPM2B_SENSITIVE_CREATE, METH_VARARGS, (char *)"delete_TPM2B_SENSITIVE_CREATE(TPM2B_SENSITIVE_CREATE self)"},
	 { (char *)"TPM2B_SENSITIVE_CREATE_swigregister", TPM2B_SENSITIVE_CREATE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_hashAlg_set", _wrap_TPMS_SCHEME_HASH_hashAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_hashAlg_get", _wrap_TPMS_SCHEME_HASH_hashAlg_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_HASH", _wrap_new_TPMS_SCHEME_HASH, METH_VARARGS, (char *)"new_TPMS_SCHEME_HASH() -> TPMS_SCHEME_HASH"},
	 { (char *)"delete_TPMS_SCHEME_HASH", _wrap_delete_TPMS_SCHEME_HASH, METH_VARARGS, (char *)"delete_TPMS_SCHEME_HASH(TPMS_SCHEME_HASH self)"},
	 { (char *)"TPMS_SCHEME_HASH_swigregister", TPMS_SCHEME_HASH_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_hashAlg_set", _wrap_TPMS_SCHEME_ECDAA_hashAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_hashAlg_get", _wrap_TPMS_SCHEME_ECDAA_hashAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_count_set", _wrap_TPMS_SCHEME_ECDAA_count_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_count_get", _wrap_TPMS_SCHEME_ECDAA_count_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_ECDAA", _wrap_new_TPMS_SCHEME_ECDAA, METH_VARARGS, (char *)"new_TPMS_SCHEME_ECDAA() -> TPMS_SCHEME_ECDAA"},
	 { (char *)"delete_TPMS_SCHEME_ECDAA", _wrap_delete_TPMS_SCHEME_ECDAA, METH_VARARGS, (char *)"delete_TPMS_SCHEME_ECDAA(TPMS_SCHEME_ECDAA self)"},
	 { (char *)"TPMS_SCHEME_ECDAA_swigregister", TPMS_SCHEME_ECDAA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_hashAlg_set", _wrap_TPMS_SCHEME_XOR_hashAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_hashAlg_get", _wrap_TPMS_SCHEME_XOR_hashAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_kdf_set", _wrap_TPMS_SCHEME_XOR_kdf_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_kdf_get", _wrap_TPMS_SCHEME_XOR_kdf_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_XOR", _wrap_new_TPMS_SCHEME_XOR, METH_VARARGS, (char *)"new_TPMS_SCHEME_XOR() -> TPMS_SCHEME_XOR"},
	 { (char *)"delete_TPMS_SCHEME_XOR", _wrap_delete_TPMS_SCHEME_XOR, METH_VARARGS, (char *)"delete_TPMS_SCHEME_XOR(TPMS_SCHEME_XOR self)"},
	 { (char *)"TPMS_SCHEME_XOR_swigregister", TPMS_SCHEME_XOR_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_hmac_set", _wrap_TPMU_SCHEME_KEYEDHASH_hmac_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_hmac_get", _wrap_TPMU_SCHEME_KEYEDHASH_hmac_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_exclusiveOr_set", _wrap_TPMU_SCHEME_KEYEDHASH_exclusiveOr_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_exclusiveOr_get", _wrap_TPMU_SCHEME_KEYEDHASH_exclusiveOr_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SCHEME_KEYEDHASH", _wrap_new_TPMU_SCHEME_KEYEDHASH, METH_VARARGS, (char *)"new_TPMU_SCHEME_KEYEDHASH() -> TPMU_SCHEME_KEYEDHASH"},
	 { (char *)"delete_TPMU_SCHEME_KEYEDHASH", _wrap_delete_TPMU_SCHEME_KEYEDHASH, METH_VARARGS, (char *)"delete_TPMU_SCHEME_KEYEDHASH(TPMU_SCHEME_KEYEDHASH self)"},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_swigregister", TPMU_SCHEME_KEYEDHASH_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_scheme_set", _wrap_TPMT_KEYEDHASH_SCHEME_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_scheme_get", _wrap_TPMT_KEYEDHASH_SCHEME_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_details_set", _wrap_TPMT_KEYEDHASH_SCHEME_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_details_get", _wrap_TPMT_KEYEDHASH_SCHEME_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KEYEDHASH_SCHEME", _wrap_new_TPMT_KEYEDHASH_SCHEME, METH_VARARGS, (char *)"new_TPMT_KEYEDHASH_SCHEME() -> TPMT_KEYEDHASH_SCHEME"},
	 { (char *)"delete_TPMT_KEYEDHASH_SCHEME", _wrap_delete_TPMT_KEYEDHASH_SCHEME, METH_VARARGS, (char *)"delete_TPMT_KEYEDHASH_SCHEME(TPMT_KEYEDHASH_SCHEME self)"},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_swigregister", TPMT_KEYEDHASH_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_rsassa_set", _wrap_TPMU_SIG_SCHEME_rsassa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_rsassa_get", _wrap_TPMU_SIG_SCHEME_rsassa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_rsapss_set", _wrap_TPMU_SIG_SCHEME_rsapss_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_rsapss_get", _wrap_TPMU_SIG_SCHEME_rsapss_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ecdsa_set", _wrap_TPMU_SIG_SCHEME_ecdsa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ecdsa_get", _wrap_TPMU_SIG_SCHEME_ecdsa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ecdaa_set", _wrap_TPMU_SIG_SCHEME_ecdaa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ecdaa_get", _wrap_TPMU_SIG_SCHEME_ecdaa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_sm2_set", _wrap_TPMU_SIG_SCHEME_sm2_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_sm2_get", _wrap_TPMU_SIG_SCHEME_sm2_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ecschnorr_set", _wrap_TPMU_SIG_SCHEME_ecschnorr_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ecschnorr_get", _wrap_TPMU_SIG_SCHEME_ecschnorr_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_hmac_set", _wrap_TPMU_SIG_SCHEME_hmac_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_hmac_get", _wrap_TPMU_SIG_SCHEME_hmac_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_any_set", _wrap_TPMU_SIG_SCHEME_any_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_any_get", _wrap_TPMU_SIG_SCHEME_any_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIG_SCHEME", _wrap_new_TPMU_SIG_SCHEME, METH_VARARGS, (char *)"new_TPMU_SIG_SCHEME() -> TPMU_SIG_SCHEME"},
	 { (char *)"delete_TPMU_SIG_SCHEME", _wrap_delete_TPMU_SIG_SCHEME, METH_VARARGS, (char *)"delete_TPMU_SIG_SCHEME(TPMU_SIG_SCHEME self)"},
	 { (char *)"TPMU_SIG_SCHEME_swigregister", TPMU_SIG_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_scheme_set", _wrap_TPMT_SIG_SCHEME_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_scheme_get", _wrap_TPMT_SIG_SCHEME_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_details_set", _wrap_TPMT_SIG_SCHEME_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_details_get", _wrap_TPMT_SIG_SCHEME_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIG_SCHEME", _wrap_new_TPMT_SIG_SCHEME, METH_VARARGS, (char *)"new_TPMT_SIG_SCHEME() -> TPMT_SIG_SCHEME"},
	 { (char *)"delete_TPMT_SIG_SCHEME", _wrap_delete_TPMT_SIG_SCHEME, METH_VARARGS, (char *)"delete_TPMT_SIG_SCHEME(TPMT_SIG_SCHEME self)"},
	 { (char *)"TPMT_SIG_SCHEME_swigregister", TPMT_SIG_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_mgf1_set", _wrap_TPMU_KDF_SCHEME_mgf1_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_mgf1_get", _wrap_TPMU_KDF_SCHEME_mgf1_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_kdf1_sp800_56a_set", _wrap_TPMU_KDF_SCHEME_kdf1_sp800_56a_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_kdf1_sp800_56a_get", _wrap_TPMU_KDF_SCHEME_kdf1_sp800_56a_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_kdf2_set", _wrap_TPMU_KDF_SCHEME_kdf2_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_kdf2_get", _wrap_TPMU_KDF_SCHEME_kdf2_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_kdf1_sp800_108_set", _wrap_TPMU_KDF_SCHEME_kdf1_sp800_108_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_kdf1_sp800_108_get", _wrap_TPMU_KDF_SCHEME_kdf1_sp800_108_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_KDF_SCHEME", _wrap_new_TPMU_KDF_SCHEME, METH_VARARGS, (char *)"new_TPMU_KDF_SCHEME() -> TPMU_KDF_SCHEME"},
	 { (char *)"delete_TPMU_KDF_SCHEME", _wrap_delete_TPMU_KDF_SCHEME, METH_VARARGS, (char *)"delete_TPMU_KDF_SCHEME(TPMU_KDF_SCHEME self)"},
	 { (char *)"TPMU_KDF_SCHEME_swigregister", TPMU_KDF_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_scheme_set", _wrap_TPMT_KDF_SCHEME_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_scheme_get", _wrap_TPMT_KDF_SCHEME_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_details_set", _wrap_TPMT_KDF_SCHEME_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_details_get", _wrap_TPMT_KDF_SCHEME_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KDF_SCHEME", _wrap_new_TPMT_KDF_SCHEME, METH_VARARGS, (char *)"new_TPMT_KDF_SCHEME() -> TPMT_KDF_SCHEME"},
	 { (char *)"delete_TPMT_KDF_SCHEME", _wrap_delete_TPMT_KDF_SCHEME, METH_VARARGS, (char *)"delete_TPMT_KDF_SCHEME(TPMT_KDF_SCHEME self)"},
	 { (char *)"TPMT_KDF_SCHEME_swigregister", TPMT_KDF_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecdh_set", _wrap_TPMU_ASYM_SCHEME_ecdh_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecdh_get", _wrap_TPMU_ASYM_SCHEME_ecdh_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecmqv_set", _wrap_TPMU_ASYM_SCHEME_ecmqv_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecmqv_get", _wrap_TPMU_ASYM_SCHEME_ecmqv_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_rsassa_set", _wrap_TPMU_ASYM_SCHEME_rsassa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_rsassa_get", _wrap_TPMU_ASYM_SCHEME_rsassa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_rsapss_set", _wrap_TPMU_ASYM_SCHEME_rsapss_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_rsapss_get", _wrap_TPMU_ASYM_SCHEME_rsapss_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecdsa_set", _wrap_TPMU_ASYM_SCHEME_ecdsa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecdsa_get", _wrap_TPMU_ASYM_SCHEME_ecdsa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecdaa_set", _wrap_TPMU_ASYM_SCHEME_ecdaa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecdaa_get", _wrap_TPMU_ASYM_SCHEME_ecdaa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_sm2_set", _wrap_TPMU_ASYM_SCHEME_sm2_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_sm2_get", _wrap_TPMU_ASYM_SCHEME_sm2_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecschnorr_set", _wrap_TPMU_ASYM_SCHEME_ecschnorr_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ecschnorr_get", _wrap_TPMU_ASYM_SCHEME_ecschnorr_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_rsaes_set", _wrap_TPMU_ASYM_SCHEME_rsaes_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_rsaes_get", _wrap_TPMU_ASYM_SCHEME_rsaes_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_oaep_set", _wrap_TPMU_ASYM_SCHEME_oaep_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_oaep_get", _wrap_TPMU_ASYM_SCHEME_oaep_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_anySig_set", _wrap_TPMU_ASYM_SCHEME_anySig_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_anySig_get", _wrap_TPMU_ASYM_SCHEME_anySig_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ASYM_SCHEME", _wrap_new_TPMU_ASYM_SCHEME, METH_VARARGS, (char *)"new_TPMU_ASYM_SCHEME() -> TPMU_ASYM_SCHEME"},
	 { (char *)"delete_TPMU_ASYM_SCHEME", _wrap_delete_TPMU_ASYM_SCHEME, METH_VARARGS, (char *)"delete_TPMU_ASYM_SCHEME(TPMU_ASYM_SCHEME self)"},
	 { (char *)"TPMU_ASYM_SCHEME_swigregister", TPMU_ASYM_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_scheme_set", _wrap_TPMT_ASYM_SCHEME_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_scheme_get", _wrap_TPMT_ASYM_SCHEME_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_details_set", _wrap_TPMT_ASYM_SCHEME_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_details_get", _wrap_TPMT_ASYM_SCHEME_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ASYM_SCHEME", _wrap_new_TPMT_ASYM_SCHEME, METH_VARARGS, (char *)"new_TPMT_ASYM_SCHEME() -> TPMT_ASYM_SCHEME"},
	 { (char *)"delete_TPMT_ASYM_SCHEME", _wrap_delete_TPMT_ASYM_SCHEME, METH_VARARGS, (char *)"delete_TPMT_ASYM_SCHEME(TPMT_ASYM_SCHEME self)"},
	 { (char *)"TPMT_ASYM_SCHEME_swigregister", TPMT_ASYM_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_scheme_set", _wrap_TPMT_RSA_SCHEME_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_scheme_get", _wrap_TPMT_RSA_SCHEME_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_details_set", _wrap_TPMT_RSA_SCHEME_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_details_get", _wrap_TPMT_RSA_SCHEME_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_SCHEME", _wrap_new_TPMT_RSA_SCHEME, METH_VARARGS, (char *)"new_TPMT_RSA_SCHEME() -> TPMT_RSA_SCHEME"},
	 { (char *)"delete_TPMT_RSA_SCHEME", _wrap_delete_TPMT_RSA_SCHEME, METH_VARARGS, (char *)"delete_TPMT_RSA_SCHEME(TPMT_RSA_SCHEME self)"},
	 { (char *)"TPMT_RSA_SCHEME_swigregister", TPMT_RSA_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_scheme_set", _wrap_TPMT_RSA_DECRYPT_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_scheme_get", _wrap_TPMT_RSA_DECRYPT_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_details_set", _wrap_TPMT_RSA_DECRYPT_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_details_get", _wrap_TPMT_RSA_DECRYPT_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_DECRYPT", _wrap_new_TPMT_RSA_DECRYPT, METH_VARARGS, (char *)"new_TPMT_RSA_DECRYPT() -> TPMT_RSA_DECRYPT"},
	 { (char *)"delete_TPMT_RSA_DECRYPT", _wrap_delete_TPMT_RSA_DECRYPT, METH_VARARGS, (char *)"delete_TPMT_RSA_DECRYPT(TPMT_RSA_DECRYPT self)"},
	 { (char *)"TPMT_RSA_DECRYPT_swigregister", TPMT_RSA_DECRYPT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_size_set", _wrap_TPM2B_PUBLIC_KEY_RSA_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_size_get", _wrap_TPM2B_PUBLIC_KEY_RSA_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_buffer_set", _wrap_TPM2B_PUBLIC_KEY_RSA_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_buffer_get", _wrap_TPM2B_PUBLIC_KEY_RSA_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_KEY_RSA", _wrap_new_TPM2B_PUBLIC_KEY_RSA, METH_VARARGS, (char *)"new_TPM2B_PUBLIC_KEY_RSA() -> TPM2B_PUBLIC_KEY_RSA"},
	 { (char *)"delete_TPM2B_PUBLIC_KEY_RSA", _wrap_delete_TPM2B_PUBLIC_KEY_RSA, METH_VARARGS, (char *)"delete_TPM2B_PUBLIC_KEY_RSA(TPM2B_PUBLIC_KEY_RSA self)"},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_swigregister", TPM2B_PUBLIC_KEY_RSA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_size_set", _wrap_TPM2B_PRIVATE_KEY_RSA_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_size_get", _wrap_TPM2B_PRIVATE_KEY_RSA_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_buffer_set", _wrap_TPM2B_PRIVATE_KEY_RSA_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_buffer_get", _wrap_TPM2B_PRIVATE_KEY_RSA_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_KEY_RSA", _wrap_new_TPM2B_PRIVATE_KEY_RSA, METH_VARARGS, (char *)"new_TPM2B_PRIVATE_KEY_RSA() -> TPM2B_PRIVATE_KEY_RSA"},
	 { (char *)"delete_TPM2B_PRIVATE_KEY_RSA", _wrap_delete_TPM2B_PRIVATE_KEY_RSA, METH_VARARGS, (char *)"delete_TPM2B_PRIVATE_KEY_RSA(TPM2B_PRIVATE_KEY_RSA self)"},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_swigregister", TPM2B_PRIVATE_KEY_RSA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_size_set", _wrap_TPM2B_ECC_PARAMETER_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_size_get", _wrap_TPM2B_ECC_PARAMETER_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_buffer_set", _wrap_TPM2B_ECC_PARAMETER_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_buffer_get", _wrap_TPM2B_ECC_PARAMETER_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_PARAMETER", _wrap_new_TPM2B_ECC_PARAMETER, METH_VARARGS, (char *)"new_TPM2B_ECC_PARAMETER() -> TPM2B_ECC_PARAMETER"},
	 { (char *)"delete_TPM2B_ECC_PARAMETER", _wrap_delete_TPM2B_ECC_PARAMETER, METH_VARARGS, (char *)"delete_TPM2B_ECC_PARAMETER(TPM2B_ECC_PARAMETER self)"},
	 { (char *)"TPM2B_ECC_PARAMETER_swigregister", TPM2B_ECC_PARAMETER_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_x_set", _wrap_TPMS_ECC_POINT_x_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_x_get", _wrap_TPMS_ECC_POINT_x_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_y_set", _wrap_TPMS_ECC_POINT_y_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_y_get", _wrap_TPMS_ECC_POINT_y_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_POINT", _wrap_new_TPMS_ECC_POINT, METH_VARARGS, (char *)"new_TPMS_ECC_POINT() -> TPMS_ECC_POINT"},
	 { (char *)"delete_TPMS_ECC_POINT", _wrap_delete_TPMS_ECC_POINT, METH_VARARGS, (char *)"delete_TPMS_ECC_POINT(TPMS_ECC_POINT self)"},
	 { (char *)"TPMS_ECC_POINT_swigregister", TPMS_ECC_POINT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_size_set", _wrap_TPM2B_ECC_POINT_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_size_get", _wrap_TPM2B_ECC_POINT_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_point_set", _wrap_TPM2B_ECC_POINT_point_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_point_get", _wrap_TPM2B_ECC_POINT_point_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_POINT", _wrap_new_TPM2B_ECC_POINT, METH_VARARGS, (char *)"new_TPM2B_ECC_POINT() -> TPM2B_ECC_POINT"},
	 { (char *)"delete_TPM2B_ECC_POINT", _wrap_delete_TPM2B_ECC_POINT, METH_VARARGS, (char *)"delete_TPM2B_ECC_POINT(TPM2B_ECC_POINT self)"},
	 { (char *)"TPM2B_ECC_POINT_swigregister", TPM2B_ECC_POINT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_scheme_set", _wrap_TPMT_ECC_SCHEME_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_scheme_get", _wrap_TPMT_ECC_SCHEME_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_details_set", _wrap_TPMT_ECC_SCHEME_details_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_details_get", _wrap_TPMT_ECC_SCHEME_details_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ECC_SCHEME", _wrap_new_TPMT_ECC_SCHEME, METH_VARARGS, (char *)"new_TPMT_ECC_SCHEME() -> TPMT_ECC_SCHEME"},
	 { (char *)"delete_TPMT_ECC_SCHEME", _wrap_delete_TPMT_ECC_SCHEME, METH_VARARGS, (char *)"delete_TPMT_ECC_SCHEME(TPMT_ECC_SCHEME self)"},
	 { (char *)"TPMT_ECC_SCHEME_swigregister", TPMT_ECC_SCHEME_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_curveID_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_curveID_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_curveID_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_curveID_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_keySize_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_keySize_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_keySize_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_keySize_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_kdf_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_kdf_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_kdf_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_kdf_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_sign_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_sign_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_sign_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_sign_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_p_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_p_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_p_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_p_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_a_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_a_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_a_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_a_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_b_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_b_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_b_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_b_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_gX_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_gX_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_gX_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_gX_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_gY_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_gY_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_gY_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_gY_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_n_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_n_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_n_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_n_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_h_set", _wrap_TPMS_ALGORITHM_DETAIL_ECC_h_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_h_get", _wrap_TPMS_ALGORITHM_DETAIL_ECC_h_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DETAIL_ECC", _wrap_new_TPMS_ALGORITHM_DETAIL_ECC, METH_VARARGS, (char *)"new_TPMS_ALGORITHM_DETAIL_ECC() -> TPMS_ALGORITHM_DETAIL_ECC"},
	 { (char *)"delete_TPMS_ALGORITHM_DETAIL_ECC", _wrap_delete_TPMS_ALGORITHM_DETAIL_ECC, METH_VARARGS, (char *)"delete_TPMS_ALGORITHM_DETAIL_ECC(TPMS_ALGORITHM_DETAIL_ECC self)"},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_swigregister", TPMS_ALGORITHM_DETAIL_ECC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_hash_set", _wrap_TPMS_SIGNATURE_RSA_hash_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_hash_get", _wrap_TPMS_SIGNATURE_RSA_hash_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_sig_set", _wrap_TPMS_SIGNATURE_RSA_sig_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_sig_get", _wrap_TPMS_SIGNATURE_RSA_sig_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_RSA", _wrap_new_TPMS_SIGNATURE_RSA, METH_VARARGS, (char *)"new_TPMS_SIGNATURE_RSA() -> TPMS_SIGNATURE_RSA"},
	 { (char *)"delete_TPMS_SIGNATURE_RSA", _wrap_delete_TPMS_SIGNATURE_RSA, METH_VARARGS, (char *)"delete_TPMS_SIGNATURE_RSA(TPMS_SIGNATURE_RSA self)"},
	 { (char *)"TPMS_SIGNATURE_RSA_swigregister", TPMS_SIGNATURE_RSA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_hash_set", _wrap_TPMS_SIGNATURE_ECC_hash_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_hash_get", _wrap_TPMS_SIGNATURE_ECC_hash_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_signatureR_set", _wrap_TPMS_SIGNATURE_ECC_signatureR_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_signatureR_get", _wrap_TPMS_SIGNATURE_ECC_signatureR_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_signatureS_set", _wrap_TPMS_SIGNATURE_ECC_signatureS_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_signatureS_get", _wrap_TPMS_SIGNATURE_ECC_signatureS_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_ECC", _wrap_new_TPMS_SIGNATURE_ECC, METH_VARARGS, (char *)"new_TPMS_SIGNATURE_ECC() -> TPMS_SIGNATURE_ECC"},
	 { (char *)"delete_TPMS_SIGNATURE_ECC", _wrap_delete_TPMS_SIGNATURE_ECC, METH_VARARGS, (char *)"delete_TPMS_SIGNATURE_ECC(TPMS_SIGNATURE_ECC self)"},
	 { (char *)"TPMS_SIGNATURE_ECC_swigregister", TPMS_SIGNATURE_ECC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_rsassa_set", _wrap_TPMU_SIGNATURE_rsassa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_rsassa_get", _wrap_TPMU_SIGNATURE_rsassa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_rsapss_set", _wrap_TPMU_SIGNATURE_rsapss_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_rsapss_get", _wrap_TPMU_SIGNATURE_rsapss_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ecdsa_set", _wrap_TPMU_SIGNATURE_ecdsa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ecdsa_get", _wrap_TPMU_SIGNATURE_ecdsa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ecdaa_set", _wrap_TPMU_SIGNATURE_ecdaa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ecdaa_get", _wrap_TPMU_SIGNATURE_ecdaa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_sm2_set", _wrap_TPMU_SIGNATURE_sm2_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_sm2_get", _wrap_TPMU_SIGNATURE_sm2_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ecschnorr_set", _wrap_TPMU_SIGNATURE_ecschnorr_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ecschnorr_get", _wrap_TPMU_SIGNATURE_ecschnorr_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_hmac_set", _wrap_TPMU_SIGNATURE_hmac_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_hmac_get", _wrap_TPMU_SIGNATURE_hmac_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_any_set", _wrap_TPMU_SIGNATURE_any_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_any_get", _wrap_TPMU_SIGNATURE_any_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIGNATURE", _wrap_new_TPMU_SIGNATURE, METH_VARARGS, (char *)"new_TPMU_SIGNATURE() -> TPMU_SIGNATURE"},
	 { (char *)"delete_TPMU_SIGNATURE", _wrap_delete_TPMU_SIGNATURE, METH_VARARGS, (char *)"delete_TPMU_SIGNATURE(TPMU_SIGNATURE self)"},
	 { (char *)"TPMU_SIGNATURE_swigregister", TPMU_SIGNATURE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_sigAlg_set", _wrap_TPMT_SIGNATURE_sigAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_sigAlg_get", _wrap_TPMT_SIGNATURE_sigAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_signature_set", _wrap_TPMT_SIGNATURE_signature_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_signature_get", _wrap_TPMT_SIGNATURE_signature_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIGNATURE", _wrap_new_TPMT_SIGNATURE, METH_VARARGS, (char *)"new_TPMT_SIGNATURE() -> TPMT_SIGNATURE"},
	 { (char *)"delete_TPMT_SIGNATURE", _wrap_delete_TPMT_SIGNATURE, METH_VARARGS, (char *)"delete_TPMT_SIGNATURE(TPMT_SIGNATURE self)"},
	 { (char *)"TPMT_SIGNATURE_swigregister", TPMT_SIGNATURE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ecc_set", _wrap_TPMU_ENCRYPTED_SECRET_ecc_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ecc_get", _wrap_TPMU_ENCRYPTED_SECRET_ecc_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_rsa_set", _wrap_TPMU_ENCRYPTED_SECRET_rsa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_rsa_get", _wrap_TPMU_ENCRYPTED_SECRET_rsa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_symmetric_set", _wrap_TPMU_ENCRYPTED_SECRET_symmetric_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_symmetric_get", _wrap_TPMU_ENCRYPTED_SECRET_symmetric_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_keyedHash_set", _wrap_TPMU_ENCRYPTED_SECRET_keyedHash_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_keyedHash_get", _wrap_TPMU_ENCRYPTED_SECRET_keyedHash_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ENCRYPTED_SECRET", _wrap_new_TPMU_ENCRYPTED_SECRET, METH_VARARGS, (char *)"new_TPMU_ENCRYPTED_SECRET() -> TPMU_ENCRYPTED_SECRET"},
	 { (char *)"delete_TPMU_ENCRYPTED_SECRET", _wrap_delete_TPMU_ENCRYPTED_SECRET, METH_VARARGS, (char *)"delete_TPMU_ENCRYPTED_SECRET(TPMU_ENCRYPTED_SECRET self)"},
	 { (char *)"TPMU_ENCRYPTED_SECRET_swigregister", TPMU_ENCRYPTED_SECRET_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_size_set", _wrap_TPM2B_ENCRYPTED_SECRET_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_size_get", _wrap_TPM2B_ENCRYPTED_SECRET_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_secret_set", _wrap_TPM2B_ENCRYPTED_SECRET_secret_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_secret_get", _wrap_TPM2B_ENCRYPTED_SECRET_secret_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ENCRYPTED_SECRET", _wrap_new_TPM2B_ENCRYPTED_SECRET, METH_VARARGS, (char *)"new_TPM2B_ENCRYPTED_SECRET() -> TPM2B_ENCRYPTED_SECRET"},
	 { (char *)"delete_TPM2B_ENCRYPTED_SECRET", _wrap_delete_TPM2B_ENCRYPTED_SECRET, METH_VARARGS, (char *)"delete_TPM2B_ENCRYPTED_SECRET(TPM2B_ENCRYPTED_SECRET self)"},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_swigregister", TPM2B_ENCRYPTED_SECRET_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_keyedHash_set", _wrap_TPMU_PUBLIC_ID_keyedHash_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_keyedHash_get", _wrap_TPMU_PUBLIC_ID_keyedHash_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_sym_set", _wrap_TPMU_PUBLIC_ID_sym_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_sym_get", _wrap_TPMU_PUBLIC_ID_sym_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_rsa_set", _wrap_TPMU_PUBLIC_ID_rsa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_rsa_get", _wrap_TPMU_PUBLIC_ID_rsa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ecc_set", _wrap_TPMU_PUBLIC_ID_ecc_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ecc_get", _wrap_TPMU_PUBLIC_ID_ecc_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_ID", _wrap_new_TPMU_PUBLIC_ID, METH_VARARGS, (char *)"new_TPMU_PUBLIC_ID() -> TPMU_PUBLIC_ID"},
	 { (char *)"delete_TPMU_PUBLIC_ID", _wrap_delete_TPMU_PUBLIC_ID, METH_VARARGS, (char *)"delete_TPMU_PUBLIC_ID(TPMU_PUBLIC_ID self)"},
	 { (char *)"TPMU_PUBLIC_ID_swigregister", TPMU_PUBLIC_ID_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_scheme_set", _wrap_TPMS_KEYEDHASH_PARMS_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_scheme_get", _wrap_TPMS_KEYEDHASH_PARMS_scheme_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_KEYEDHASH_PARMS", _wrap_new_TPMS_KEYEDHASH_PARMS, METH_VARARGS, (char *)"new_TPMS_KEYEDHASH_PARMS() -> TPMS_KEYEDHASH_PARMS"},
	 { (char *)"delete_TPMS_KEYEDHASH_PARMS", _wrap_delete_TPMS_KEYEDHASH_PARMS, METH_VARARGS, (char *)"delete_TPMS_KEYEDHASH_PARMS(TPMS_KEYEDHASH_PARMS self)"},
	 { (char *)"TPMS_KEYEDHASH_PARMS_swigregister", TPMS_KEYEDHASH_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_symmetric_set", _wrap_TPMS_ASYM_PARMS_symmetric_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_symmetric_get", _wrap_TPMS_ASYM_PARMS_symmetric_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_scheme_set", _wrap_TPMS_ASYM_PARMS_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_scheme_get", _wrap_TPMS_ASYM_PARMS_scheme_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ASYM_PARMS", _wrap_new_TPMS_ASYM_PARMS, METH_VARARGS, (char *)"new_TPMS_ASYM_PARMS() -> TPMS_ASYM_PARMS"},
	 { (char *)"delete_TPMS_ASYM_PARMS", _wrap_delete_TPMS_ASYM_PARMS, METH_VARARGS, (char *)"delete_TPMS_ASYM_PARMS(TPMS_ASYM_PARMS self)"},
	 { (char *)"TPMS_ASYM_PARMS_swigregister", TPMS_ASYM_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_symmetric_set", _wrap_TPMS_RSA_PARMS_symmetric_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_symmetric_get", _wrap_TPMS_RSA_PARMS_symmetric_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_scheme_set", _wrap_TPMS_RSA_PARMS_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_scheme_get", _wrap_TPMS_RSA_PARMS_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_keyBits_set", _wrap_TPMS_RSA_PARMS_keyBits_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_keyBits_get", _wrap_TPMS_RSA_PARMS_keyBits_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_exponent_set", _wrap_TPMS_RSA_PARMS_exponent_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_exponent_get", _wrap_TPMS_RSA_PARMS_exponent_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_RSA_PARMS", _wrap_new_TPMS_RSA_PARMS, METH_VARARGS, (char *)"new_TPMS_RSA_PARMS() -> TPMS_RSA_PARMS"},
	 { (char *)"delete_TPMS_RSA_PARMS", _wrap_delete_TPMS_RSA_PARMS, METH_VARARGS, (char *)"delete_TPMS_RSA_PARMS(TPMS_RSA_PARMS self)"},
	 { (char *)"TPMS_RSA_PARMS_swigregister", TPMS_RSA_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_symmetric_set", _wrap_TPMS_ECC_PARMS_symmetric_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_symmetric_get", _wrap_TPMS_ECC_PARMS_symmetric_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_scheme_set", _wrap_TPMS_ECC_PARMS_scheme_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_scheme_get", _wrap_TPMS_ECC_PARMS_scheme_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_curveID_set", _wrap_TPMS_ECC_PARMS_curveID_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_curveID_get", _wrap_TPMS_ECC_PARMS_curveID_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_kdf_set", _wrap_TPMS_ECC_PARMS_kdf_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_kdf_get", _wrap_TPMS_ECC_PARMS_kdf_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_PARMS", _wrap_new_TPMS_ECC_PARMS, METH_VARARGS, (char *)"new_TPMS_ECC_PARMS() -> TPMS_ECC_PARMS"},
	 { (char *)"delete_TPMS_ECC_PARMS", _wrap_delete_TPMS_ECC_PARMS, METH_VARARGS, (char *)"delete_TPMS_ECC_PARMS(TPMS_ECC_PARMS self)"},
	 { (char *)"TPMS_ECC_PARMS_swigregister", TPMS_ECC_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_keyedHashDetail_set", _wrap_TPMU_PUBLIC_PARMS_keyedHashDetail_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_keyedHashDetail_get", _wrap_TPMU_PUBLIC_PARMS_keyedHashDetail_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_symDetail_set", _wrap_TPMU_PUBLIC_PARMS_symDetail_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_symDetail_get", _wrap_TPMU_PUBLIC_PARMS_symDetail_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_rsaDetail_set", _wrap_TPMU_PUBLIC_PARMS_rsaDetail_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_rsaDetail_get", _wrap_TPMU_PUBLIC_PARMS_rsaDetail_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_eccDetail_set", _wrap_TPMU_PUBLIC_PARMS_eccDetail_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_eccDetail_get", _wrap_TPMU_PUBLIC_PARMS_eccDetail_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_asymDetail_set", _wrap_TPMU_PUBLIC_PARMS_asymDetail_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_asymDetail_get", _wrap_TPMU_PUBLIC_PARMS_asymDetail_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_PARMS", _wrap_new_TPMU_PUBLIC_PARMS, METH_VARARGS, (char *)"new_TPMU_PUBLIC_PARMS() -> TPMU_PUBLIC_PARMS"},
	 { (char *)"delete_TPMU_PUBLIC_PARMS", _wrap_delete_TPMU_PUBLIC_PARMS, METH_VARARGS, (char *)"delete_TPMU_PUBLIC_PARMS(TPMU_PUBLIC_PARMS self)"},
	 { (char *)"TPMU_PUBLIC_PARMS_swigregister", TPMU_PUBLIC_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_type_set", _wrap_TPMT_PUBLIC_PARMS_type_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_type_get", _wrap_TPMT_PUBLIC_PARMS_type_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_parameters_set", _wrap_TPMT_PUBLIC_PARMS_parameters_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_parameters_get", _wrap_TPMT_PUBLIC_PARMS_parameters_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_PARMS", _wrap_new_TPMT_PUBLIC_PARMS, METH_VARARGS, (char *)"new_TPMT_PUBLIC_PARMS() -> TPMT_PUBLIC_PARMS"},
	 { (char *)"delete_TPMT_PUBLIC_PARMS", _wrap_delete_TPMT_PUBLIC_PARMS, METH_VARARGS, (char *)"delete_TPMT_PUBLIC_PARMS(TPMT_PUBLIC_PARMS self)"},
	 { (char *)"TPMT_PUBLIC_PARMS_swigregister", TPMT_PUBLIC_PARMS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_type_set", _wrap_TPMT_PUBLIC_type_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_type_get", _wrap_TPMT_PUBLIC_type_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_nameAlg_set", _wrap_TPMT_PUBLIC_nameAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_nameAlg_get", _wrap_TPMT_PUBLIC_nameAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_objectAttributes_set", _wrap_TPMT_PUBLIC_objectAttributes_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_objectAttributes_get", _wrap_TPMT_PUBLIC_objectAttributes_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_authPolicy_set", _wrap_TPMT_PUBLIC_authPolicy_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_authPolicy_get", _wrap_TPMT_PUBLIC_authPolicy_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_parameters_set", _wrap_TPMT_PUBLIC_parameters_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_parameters_get", _wrap_TPMT_PUBLIC_parameters_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_unique_set", _wrap_TPMT_PUBLIC_unique_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_unique_get", _wrap_TPMT_PUBLIC_unique_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC", _wrap_new_TPMT_PUBLIC, METH_VARARGS, (char *)"new_TPMT_PUBLIC() -> TPMT_PUBLIC"},
	 { (char *)"delete_TPMT_PUBLIC", _wrap_delete_TPMT_PUBLIC, METH_VARARGS, (char *)"delete_TPMT_PUBLIC(TPMT_PUBLIC self)"},
	 { (char *)"TPMT_PUBLIC_swigregister", TPMT_PUBLIC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_size_set", _wrap_TPM2B_PUBLIC_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_size_get", _wrap_TPM2B_PUBLIC_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_publicArea_set", _wrap_TPM2B_PUBLIC_publicArea_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_publicArea_get", _wrap_TPM2B_PUBLIC_publicArea_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC", _wrap_new_TPM2B_PUBLIC, METH_VARARGS, (char *)"new_TPM2B_PUBLIC() -> TPM2B_PUBLIC"},
	 { (char *)"delete_TPM2B_PUBLIC", _wrap_delete_TPM2B_PUBLIC, METH_VARARGS, (char *)"delete_TPM2B_PUBLIC(TPM2B_PUBLIC self)"},
	 { (char *)"TPM2B_PUBLIC_swigregister", TPM2B_PUBLIC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_size_set", _wrap_TPM2B_TEMPLATE_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_size_get", _wrap_TPM2B_TEMPLATE_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_buffer_set", _wrap_TPM2B_TEMPLATE_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_buffer_get", _wrap_TPM2B_TEMPLATE_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_TEMPLATE", _wrap_new_TPM2B_TEMPLATE, METH_VARARGS, (char *)"new_TPM2B_TEMPLATE() -> TPM2B_TEMPLATE"},
	 { (char *)"delete_TPM2B_TEMPLATE", _wrap_delete_TPM2B_TEMPLATE, METH_VARARGS, (char *)"delete_TPM2B_TEMPLATE(TPM2B_TEMPLATE self)"},
	 { (char *)"TPM2B_TEMPLATE_swigregister", TPM2B_TEMPLATE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_size_set", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_size_get", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_set", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_get", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_VENDOR_SPECIFIC", _wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC, METH_VARARGS, (char *)"new_TPM2B_PRIVATE_VENDOR_SPECIFIC() -> TPM2B_PRIVATE_VENDOR_SPECIFIC"},
	 { (char *)"delete_TPM2B_PRIVATE_VENDOR_SPECIFIC", _wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC, METH_VARARGS, (char *)"delete_TPM2B_PRIVATE_VENDOR_SPECIFIC(TPM2B_PRIVATE_VENDOR_SPECIFIC self)"},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_swigregister", TPM2B_PRIVATE_VENDOR_SPECIFIC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_rsa_set", _wrap_TPMU_SENSITIVE_COMPOSITE_rsa_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_rsa_get", _wrap_TPMU_SENSITIVE_COMPOSITE_rsa_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ecc_set", _wrap_TPMU_SENSITIVE_COMPOSITE_ecc_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ecc_get", _wrap_TPMU_SENSITIVE_COMPOSITE_ecc_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_bits_set", _wrap_TPMU_SENSITIVE_COMPOSITE_bits_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_bits_get", _wrap_TPMU_SENSITIVE_COMPOSITE_bits_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_sym_set", _wrap_TPMU_SENSITIVE_COMPOSITE_sym_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_sym_get", _wrap_TPMU_SENSITIVE_COMPOSITE_sym_get, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_any_set", _wrap_TPMU_SENSITIVE_COMPOSITE_any_set, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_any_get", _wrap_TPMU_SENSITIVE_COMPOSITE_any_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SENSITIVE_COMPOSITE", _wrap_new_TPMU_SENSITIVE_COMPOSITE, METH_VARARGS, (char *)"new_TPMU_SENSITIVE_COMPOSITE() -> TPMU_SENSITIVE_COMPOSITE"},
	 { (char *)"delete_TPMU_SENSITIVE_COMPOSITE", _wrap_delete_TPMU_SENSITIVE_COMPOSITE, METH_VARARGS, (char *)"delete_TPMU_SENSITIVE_COMPOSITE(TPMU_SENSITIVE_COMPOSITE self)"},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_swigregister", TPMU_SENSITIVE_COMPOSITE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_sensitiveType_set", _wrap_TPMT_SENSITIVE_sensitiveType_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_sensitiveType_get", _wrap_TPMT_SENSITIVE_sensitiveType_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_authValue_set", _wrap_TPMT_SENSITIVE_authValue_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_authValue_get", _wrap_TPMT_SENSITIVE_authValue_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_seedValue_set", _wrap_TPMT_SENSITIVE_seedValue_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_seedValue_get", _wrap_TPMT_SENSITIVE_seedValue_get, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_sensitive_set", _wrap_TPMT_SENSITIVE_sensitive_set, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_sensitive_get", _wrap_TPMT_SENSITIVE_sensitive_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SENSITIVE", _wrap_new_TPMT_SENSITIVE, METH_VARARGS, (char *)"new_TPMT_SENSITIVE() -> TPMT_SENSITIVE"},
	 { (char *)"delete_TPMT_SENSITIVE", _wrap_delete_TPMT_SENSITIVE, METH_VARARGS, (char *)"delete_TPMT_SENSITIVE(TPMT_SENSITIVE self)"},
	 { (char *)"TPMT_SENSITIVE_swigregister", TPMT_SENSITIVE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_size_set", _wrap_TPM2B_SENSITIVE_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_size_get", _wrap_TPM2B_SENSITIVE_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_sensitiveArea_set", _wrap_TPM2B_SENSITIVE_sensitiveArea_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_sensitiveArea_get", _wrap_TPM2B_SENSITIVE_sensitiveArea_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE", _wrap_new_TPM2B_SENSITIVE, METH_VARARGS, (char *)"new_TPM2B_SENSITIVE() -> TPM2B_SENSITIVE"},
	 { (char *)"delete_TPM2B_SENSITIVE", _wrap_delete_TPM2B_SENSITIVE, METH_VARARGS, (char *)"delete_TPM2B_SENSITIVE(TPM2B_SENSITIVE self)"},
	 { (char *)"TPM2B_SENSITIVE_swigregister", TPM2B_SENSITIVE_swigregister, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_integrityOuter_set", _wrap__PRIVATE_integrityOuter_set, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_integrityOuter_get", _wrap__PRIVATE_integrityOuter_get, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_integrityInner_set", _wrap__PRIVATE_integrityInner_set, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_integrityInner_get", _wrap__PRIVATE_integrityInner_get, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_sensitive_set", _wrap__PRIVATE_sensitive_set, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_sensitive_get", _wrap__PRIVATE_sensitive_get, METH_VARARGS, NULL},
	 { (char *)"new__PRIVATE", _wrap_new__PRIVATE, METH_VARARGS, (char *)"new__PRIVATE() -> _PRIVATE"},
	 { (char *)"delete__PRIVATE", _wrap_delete__PRIVATE, METH_VARARGS, (char *)"delete__PRIVATE(_PRIVATE self)"},
	 { (char *)"_PRIVATE_swigregister", _PRIVATE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_size_set", _wrap_TPM2B_PRIVATE_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_size_get", _wrap_TPM2B_PRIVATE_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_buffer_set", _wrap_TPM2B_PRIVATE_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_buffer_get", _wrap_TPM2B_PRIVATE_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE", _wrap_new_TPM2B_PRIVATE, METH_VARARGS, (char *)"new_TPM2B_PRIVATE() -> TPM2B_PRIVATE"},
	 { (char *)"delete_TPM2B_PRIVATE", _wrap_delete_TPM2B_PRIVATE, METH_VARARGS, (char *)"delete_TPM2B_PRIVATE(TPM2B_PRIVATE self)"},
	 { (char *)"TPM2B_PRIVATE_swigregister", TPM2B_PRIVATE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_integrityHMAC_set", _wrap_TPMS_ID_OBJECT_integrityHMAC_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_integrityHMAC_get", _wrap_TPMS_ID_OBJECT_integrityHMAC_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_encIdentity_set", _wrap_TPMS_ID_OBJECT_encIdentity_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_encIdentity_get", _wrap_TPMS_ID_OBJECT_encIdentity_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ID_OBJECT", _wrap_new_TPMS_ID_OBJECT, METH_VARARGS, (char *)"new_TPMS_ID_OBJECT() -> TPMS_ID_OBJECT"},
	 { (char *)"delete_TPMS_ID_OBJECT", _wrap_delete_TPMS_ID_OBJECT, METH_VARARGS, (char *)"delete_TPMS_ID_OBJECT(TPMS_ID_OBJECT self)"},
	 { (char *)"TPMS_ID_OBJECT_swigregister", TPMS_ID_OBJECT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_size_set", _wrap_TPM2B_ID_OBJECT_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_size_get", _wrap_TPM2B_ID_OBJECT_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_credential_set", _wrap_TPM2B_ID_OBJECT_credential_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_credential_get", _wrap_TPM2B_ID_OBJECT_credential_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ID_OBJECT", _wrap_new_TPM2B_ID_OBJECT, METH_VARARGS, (char *)"new_TPM2B_ID_OBJECT() -> TPM2B_ID_OBJECT"},
	 { (char *)"delete_TPM2B_ID_OBJECT", _wrap_delete_TPM2B_ID_OBJECT, METH_VARARGS, (char *)"delete_TPM2B_ID_OBJECT(TPM2B_ID_OBJECT self)"},
	 { (char *)"TPM2B_ID_OBJECT_swigregister", TPM2B_ID_OBJECT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2_NV_INDEX_INDEX_SHIFT_swigconstant", TPM2_NV_INDEX_INDEX_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NV_INDEX_RH_NV_SHIFT_swigconstant", TPM2_NV_INDEX_RH_NV_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_set", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_get", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinCount_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_set", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_get", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_pinLimit_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PIN_COUNTER_PARAMETERS", _wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS, METH_VARARGS, (char *)"new_TPMS_NV_PIN_COUNTER_PARAMETERS() -> TPMS_NV_PIN_COUNTER_PARAMETERS"},
	 { (char *)"delete_TPMS_NV_PIN_COUNTER_PARAMETERS", _wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS, METH_VARARGS, (char *)"delete_TPMS_NV_PIN_COUNTER_PARAMETERS(TPMS_NV_PIN_COUNTER_PARAMETERS self)"},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_swigregister", TPMS_NV_PIN_COUNTER_PARAMETERS_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_TPM2_NT_SHIFT_swigconstant", TPMA_NV_TPM2_NT_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_nvIndex_set", _wrap_TPMS_NV_PUBLIC_nvIndex_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_nvIndex_get", _wrap_TPMS_NV_PUBLIC_nvIndex_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_nameAlg_set", _wrap_TPMS_NV_PUBLIC_nameAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_nameAlg_get", _wrap_TPMS_NV_PUBLIC_nameAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_attributes_set", _wrap_TPMS_NV_PUBLIC_attributes_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_attributes_get", _wrap_TPMS_NV_PUBLIC_attributes_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_authPolicy_set", _wrap_TPMS_NV_PUBLIC_authPolicy_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_authPolicy_get", _wrap_TPMS_NV_PUBLIC_authPolicy_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_dataSize_set", _wrap_TPMS_NV_PUBLIC_dataSize_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_dataSize_get", _wrap_TPMS_NV_PUBLIC_dataSize_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PUBLIC", _wrap_new_TPMS_NV_PUBLIC, METH_VARARGS, (char *)"new_TPMS_NV_PUBLIC() -> TPMS_NV_PUBLIC"},
	 { (char *)"delete_TPMS_NV_PUBLIC", _wrap_delete_TPMS_NV_PUBLIC, METH_VARARGS, (char *)"delete_TPMS_NV_PUBLIC(TPMS_NV_PUBLIC self)"},
	 { (char *)"TPMS_NV_PUBLIC_swigregister", TPMS_NV_PUBLIC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_size_set", _wrap_TPM2B_NV_PUBLIC_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_size_get", _wrap_TPM2B_NV_PUBLIC_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_nvPublic_set", _wrap_TPM2B_NV_PUBLIC_nvPublic_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_nvPublic_get", _wrap_TPM2B_NV_PUBLIC_nvPublic_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NV_PUBLIC", _wrap_new_TPM2B_NV_PUBLIC, METH_VARARGS, (char *)"new_TPM2B_NV_PUBLIC() -> TPM2B_NV_PUBLIC"},
	 { (char *)"delete_TPM2B_NV_PUBLIC", _wrap_delete_TPM2B_NV_PUBLIC, METH_VARARGS, (char *)"delete_TPM2B_NV_PUBLIC(TPM2B_NV_PUBLIC self)"},
	 { (char *)"TPM2B_NV_PUBLIC_swigregister", TPM2B_NV_PUBLIC_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_size_set", _wrap_TPM2B_CONTEXT_SENSITIVE_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_size_get", _wrap_TPM2B_CONTEXT_SENSITIVE_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_buffer_set", _wrap_TPM2B_CONTEXT_SENSITIVE_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_buffer_get", _wrap_TPM2B_CONTEXT_SENSITIVE_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_SENSITIVE", _wrap_new_TPM2B_CONTEXT_SENSITIVE, METH_VARARGS, (char *)"new_TPM2B_CONTEXT_SENSITIVE() -> TPM2B_CONTEXT_SENSITIVE"},
	 { (char *)"delete_TPM2B_CONTEXT_SENSITIVE", _wrap_delete_TPM2B_CONTEXT_SENSITIVE, METH_VARARGS, (char *)"delete_TPM2B_CONTEXT_SENSITIVE(TPM2B_CONTEXT_SENSITIVE self)"},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_swigregister", TPM2B_CONTEXT_SENSITIVE_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_integrity_set", _wrap_TPMS_CONTEXT_DATA_integrity_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_integrity_get", _wrap_TPMS_CONTEXT_DATA_integrity_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_encrypted_set", _wrap_TPMS_CONTEXT_DATA_encrypted_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_encrypted_get", _wrap_TPMS_CONTEXT_DATA_encrypted_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_DATA", _wrap_new_TPMS_CONTEXT_DATA, METH_VARARGS, (char *)"new_TPMS_CONTEXT_DATA() -> TPMS_CONTEXT_DATA"},
	 { (char *)"delete_TPMS_CONTEXT_DATA", _wrap_delete_TPMS_CONTEXT_DATA, METH_VARARGS, (char *)"delete_TPMS_CONTEXT_DATA(TPMS_CONTEXT_DATA self)"},
	 { (char *)"TPMS_CONTEXT_DATA_swigregister", TPMS_CONTEXT_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_size_set", _wrap_TPM2B_CONTEXT_DATA_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_size_get", _wrap_TPM2B_CONTEXT_DATA_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_buffer_set", _wrap_TPM2B_CONTEXT_DATA_buffer_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_buffer_get", _wrap_TPM2B_CONTEXT_DATA_buffer_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_DATA", _wrap_new_TPM2B_CONTEXT_DATA, METH_VARARGS, (char *)"new_TPM2B_CONTEXT_DATA() -> TPM2B_CONTEXT_DATA"},
	 { (char *)"delete_TPM2B_CONTEXT_DATA", _wrap_delete_TPM2B_CONTEXT_DATA, METH_VARARGS, (char *)"delete_TPM2B_CONTEXT_DATA(TPM2B_CONTEXT_DATA self)"},
	 { (char *)"TPM2B_CONTEXT_DATA_swigregister", TPM2B_CONTEXT_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_sequence_set", _wrap_TPMS_CONTEXT_sequence_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_sequence_get", _wrap_TPMS_CONTEXT_sequence_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_savedHandle_set", _wrap_TPMS_CONTEXT_savedHandle_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_savedHandle_get", _wrap_TPMS_CONTEXT_savedHandle_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_hierarchy_set", _wrap_TPMS_CONTEXT_hierarchy_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_hierarchy_get", _wrap_TPMS_CONTEXT_hierarchy_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_contextBlob_set", _wrap_TPMS_CONTEXT_contextBlob_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_contextBlob_get", _wrap_TPMS_CONTEXT_contextBlob_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT", _wrap_new_TPMS_CONTEXT, METH_VARARGS, (char *)"new_TPMS_CONTEXT() -> TPMS_CONTEXT"},
	 { (char *)"delete_TPMS_CONTEXT", _wrap_delete_TPMS_CONTEXT, METH_VARARGS, (char *)"delete_TPMS_CONTEXT(TPMS_CONTEXT self)"},
	 { (char *)"TPMS_CONTEXT_swigregister", TPMS_CONTEXT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_pcrSelect_set", _wrap_TPMS_CREATION_DATA_pcrSelect_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_pcrSelect_get", _wrap_TPMS_CREATION_DATA_pcrSelect_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_pcrDigest_set", _wrap_TPMS_CREATION_DATA_pcrDigest_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_pcrDigest_get", _wrap_TPMS_CREATION_DATA_pcrDigest_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_locality_set", _wrap_TPMS_CREATION_DATA_locality_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_locality_get", _wrap_TPMS_CREATION_DATA_locality_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_parentNameAlg_set", _wrap_TPMS_CREATION_DATA_parentNameAlg_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_parentNameAlg_get", _wrap_TPMS_CREATION_DATA_parentNameAlg_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_parentName_set", _wrap_TPMS_CREATION_DATA_parentName_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_parentName_get", _wrap_TPMS_CREATION_DATA_parentName_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_parentQualifiedName_set", _wrap_TPMS_CREATION_DATA_parentQualifiedName_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_parentQualifiedName_get", _wrap_TPMS_CREATION_DATA_parentQualifiedName_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_outsideInfo_set", _wrap_TPMS_CREATION_DATA_outsideInfo_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_outsideInfo_get", _wrap_TPMS_CREATION_DATA_outsideInfo_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_DATA", _wrap_new_TPMS_CREATION_DATA, METH_VARARGS, (char *)"new_TPMS_CREATION_DATA() -> TPMS_CREATION_DATA"},
	 { (char *)"delete_TPMS_CREATION_DATA", _wrap_delete_TPMS_CREATION_DATA, METH_VARARGS, (char *)"delete_TPMS_CREATION_DATA(TPMS_CREATION_DATA self)"},
	 { (char *)"TPMS_CREATION_DATA_swigregister", TPMS_CREATION_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_size_set", _wrap_TPM2B_CREATION_DATA_size_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_size_get", _wrap_TPM2B_CREATION_DATA_size_get, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_creationData_set", _wrap_TPM2B_CREATION_DATA_creationData_set, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_creationData_get", _wrap_TPM2B_CREATION_DATA_creationData_get, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CREATION_DATA", _wrap_new_TPM2B_CREATION_DATA, METH_VARARGS, (char *)"new_TPM2B_CREATION_DATA() -> TPM2B_CREATION_DATA"},
	 { (char *)"delete_TPM2B_CREATION_DATA", _wrap_delete_TPM2B_CREATION_DATA, METH_VARARGS, (char *)"delete_TPM2B_CREATION_DATA(TPM2B_CREATION_DATA self)"},
	 { (char *)"TPM2B_CREATION_DATA_swigregister", TPM2B_CREATION_DATA_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_tag_set", _wrap_TPMS_AC_OUTPUT_tag_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_tag_get", _wrap_TPMS_AC_OUTPUT_tag_get, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_data_set", _wrap_TPMS_AC_OUTPUT_data_set, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_data_get", _wrap_TPMS_AC_OUTPUT_data_get, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AC_OUTPUT", _wrap_new_TPMS_AC_OUTPUT, METH_VARARGS, (char *)"new_TPMS_AC_OUTPUT() -> TPMS_AC_OUTPUT"},
	 { (char *)"delete_TPMS_AC_OUTPUT", _wrap_delete_TPMS_AC_OUTPUT, METH_VARARGS, (char *)"delete_TPMS_AC_OUTPUT(TPMS_AC_OUTPUT self)"},
	 { (char *)"TPMS_AC_OUTPUT_swigregister", TPMS_AC_OUTPUT_swigregister, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_count_set", _wrap_TPML_AC_CAPABILITIES_count_set, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_count_get", _wrap_TPML_AC_CAPABILITIES_count_get, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_acCapabilities_set", _wrap_TPML_AC_CAPABILITIES_acCapabilities_set, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_acCapabilities_get", _wrap_TPML_AC_CAPABILITIES_acCapabilities_get, METH_VARARGS, NULL},
	 { (char *)"new_TPML_AC_CAPABILITIES", _wrap_new_TPML_AC_CAPABILITIES, METH_VARARGS, (char *)"new_TPML_AC_CAPABILITIES() -> TPML_AC_CAPABILITIES"},
	 { (char *)"delete_TPML_AC_CAPABILITIES", _wrap_delete_TPML_AC_CAPABILITIES, METH_VARARGS, (char *)"delete_TPML_AC_CAPABILITIES(TPML_AC_CAPABILITIES self)"},
	 { (char *)"TPML_AC_CAPABILITIES_swigregister", TPML_AC_CAPABILITIES_swigregister, METH_VARARGS, NULL},
	 { (char *)"Tss2_RC_Decode", _wrap_Tss2_RC_Decode, METH_VARARGS, NULL},
	 { (char *)"new_TPMI_YES_NO_PTR", _wrap_new_TPMI_YES_NO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMI_YES_NO_PTR", _wrap_copy_TPMI_YES_NO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMI_YES_NO_PTR", _wrap_delete_TPMI_YES_NO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMI_YES_NO_PTR_assign", _wrap_TPMI_YES_NO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMI_YES_NO_PTR_value", _wrap_TPMI_YES_NO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_EMPTY_swigconstant", sizeof_TPMS_EMPTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_EMPTY_ARRAY", _wrap_new_TPMS_EMPTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_EMPTY_ARRAY", _wrap_delete_TPMS_EMPTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_ARRAY___getitem__", _wrap_TPMS_EMPTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_ARRAY___setitem__", _wrap_TPMS_EMPTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_ARRAY_cast", _wrap_TPMS_EMPTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_ARRAY_frompointer", _wrap_TPMS_EMPTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_ARRAY_swigregister", TPMS_EMPTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_EMPTY_PTR", _wrap_new_TPMS_EMPTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_EMPTY_PTR", _wrap_copy_TPMS_EMPTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_EMPTY_PTR", _wrap_delete_TPMS_EMPTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_PTR_assign", _wrap_TPMS_EMPTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_PTR_value", _wrap_TPMS_EMPTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_EMPTY_PTR_PTR", _wrap_new_TPMS_EMPTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_EMPTY_PTR_PTR", _wrap_copy_TPMS_EMPTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_EMPTY_PTR_PTR", _wrap_delete_TPMS_EMPTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_PTR_PTR_assign", _wrap_TPMS_EMPTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_EMPTY_PTR_PTR_value", _wrap_TPMS_EMPTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ALGORITHM_DESCRIPTION_swigconstant", sizeof_TPMS_ALGORITHM_DESCRIPTION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DESCRIPTION_ARRAY", _wrap_new_TPMS_ALGORITHM_DESCRIPTION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY", _wrap_delete_TPMS_ALGORITHM_DESCRIPTION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__", _wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__", _wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast", _wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer", _wrap_TPMS_ALGORITHM_DESCRIPTION_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_ARRAY_swigregister", TPMS_ALGORITHM_DESCRIPTION_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DESCRIPTION_PTR", _wrap_new_TPMS_ALGORITHM_DESCRIPTION_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ALGORITHM_DESCRIPTION_PTR", _wrap_copy_TPMS_ALGORITHM_DESCRIPTION_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALGORITHM_DESCRIPTION_PTR", _wrap_delete_TPMS_ALGORITHM_DESCRIPTION_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_PTR_assign", _wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_PTR_value", _wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR", _wrap_new_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR", _wrap_copy_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR", _wrap_delete_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign", _wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value", _wrap_TPMS_ALGORITHM_DESCRIPTION_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_HA_swigconstant", sizeof_TPMU_HA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_HA_ARRAY", _wrap_new_TPMU_HA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_HA_ARRAY", _wrap_delete_TPMU_HA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_ARRAY___getitem__", _wrap_TPMU_HA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_ARRAY___setitem__", _wrap_TPMU_HA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_ARRAY_cast", _wrap_TPMU_HA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_ARRAY_frompointer", _wrap_TPMU_HA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_ARRAY_swigregister", TPMU_HA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_HA_PTR", _wrap_new_TPMU_HA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_HA_PTR", _wrap_copy_TPMU_HA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_HA_PTR", _wrap_delete_TPMU_HA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_PTR_assign", _wrap_TPMU_HA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_PTR_value", _wrap_TPMU_HA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_HA_PTR_PTR", _wrap_new_TPMU_HA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_HA_PTR_PTR", _wrap_copy_TPMU_HA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_HA_PTR_PTR", _wrap_delete_TPMU_HA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_PTR_PTR_assign", _wrap_TPMU_HA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_HA_PTR_PTR_value", _wrap_TPMU_HA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_HA_swigconstant", sizeof_TPMT_HA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_HA_ARRAY", _wrap_new_TPMT_HA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_HA_ARRAY", _wrap_delete_TPMT_HA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_ARRAY___getitem__", _wrap_TPMT_HA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_ARRAY___setitem__", _wrap_TPMT_HA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_ARRAY_cast", _wrap_TPMT_HA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_ARRAY_frompointer", _wrap_TPMT_HA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_ARRAY_swigregister", TPMT_HA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_HA_PTR", _wrap_new_TPMT_HA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_HA_PTR", _wrap_copy_TPMT_HA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_HA_PTR", _wrap_delete_TPMT_HA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_PTR_assign", _wrap_TPMT_HA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_PTR_value", _wrap_TPMT_HA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_HA_PTR_PTR", _wrap_new_TPMT_HA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_HA_PTR_PTR", _wrap_copy_TPMT_HA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_HA_PTR_PTR", _wrap_delete_TPMT_HA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_PTR_PTR_assign", _wrap_TPMT_HA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_HA_PTR_PTR_value", _wrap_TPMT_HA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_DIGEST_swigconstant", sizeof_TPM2B_DIGEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DIGEST_ARRAY", _wrap_new_TPM2B_DIGEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_DIGEST_ARRAY", _wrap_delete_TPM2B_DIGEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_ARRAY___getitem__", _wrap_TPM2B_DIGEST_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_ARRAY___setitem__", _wrap_TPM2B_DIGEST_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_ARRAY_cast", _wrap_TPM2B_DIGEST_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_ARRAY_frompointer", _wrap_TPM2B_DIGEST_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_ARRAY_swigregister", TPM2B_DIGEST_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DIGEST_PTR", _wrap_new_TPM2B_DIGEST_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_DIGEST_PTR", _wrap_copy_TPM2B_DIGEST_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_DIGEST_PTR", _wrap_delete_TPM2B_DIGEST_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_PTR_assign", _wrap_TPM2B_DIGEST_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_PTR_value", _wrap_TPM2B_DIGEST_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DIGEST_PTR_PTR", _wrap_new_TPM2B_DIGEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_DIGEST_PTR_PTR", _wrap_copy_TPM2B_DIGEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_DIGEST_PTR_PTR", _wrap_delete_TPM2B_DIGEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_PTR_PTR_assign", _wrap_TPM2B_DIGEST_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DIGEST_PTR_PTR_value", _wrap_TPM2B_DIGEST_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_DATA_swigconstant", sizeof_TPM2B_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DATA_ARRAY", _wrap_new_TPM2B_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_DATA_ARRAY", _wrap_delete_TPM2B_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_ARRAY___getitem__", _wrap_TPM2B_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_ARRAY___setitem__", _wrap_TPM2B_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_ARRAY_cast", _wrap_TPM2B_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_ARRAY_frompointer", _wrap_TPM2B_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_ARRAY_swigregister", TPM2B_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DATA_PTR", _wrap_new_TPM2B_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_DATA_PTR", _wrap_copy_TPM2B_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_DATA_PTR", _wrap_delete_TPM2B_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_PTR_assign", _wrap_TPM2B_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_PTR_value", _wrap_TPM2B_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_DATA_PTR_PTR", _wrap_new_TPM2B_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_DATA_PTR_PTR", _wrap_copy_TPM2B_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_DATA_PTR_PTR", _wrap_delete_TPM2B_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_PTR_PTR_assign", _wrap_TPM2B_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_DATA_PTR_PTR_value", _wrap_TPM2B_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_EVENT_swigconstant", sizeof_TPM2B_EVENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_EVENT_ARRAY", _wrap_new_TPM2B_EVENT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_EVENT_ARRAY", _wrap_delete_TPM2B_EVENT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_ARRAY___getitem__", _wrap_TPM2B_EVENT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_ARRAY___setitem__", _wrap_TPM2B_EVENT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_ARRAY_cast", _wrap_TPM2B_EVENT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_ARRAY_frompointer", _wrap_TPM2B_EVENT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_ARRAY_swigregister", TPM2B_EVENT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_EVENT_PTR", _wrap_new_TPM2B_EVENT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_EVENT_PTR", _wrap_copy_TPM2B_EVENT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_EVENT_PTR", _wrap_delete_TPM2B_EVENT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_PTR_assign", _wrap_TPM2B_EVENT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_PTR_value", _wrap_TPM2B_EVENT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_EVENT_PTR_PTR", _wrap_new_TPM2B_EVENT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_EVENT_PTR_PTR", _wrap_copy_TPM2B_EVENT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_EVENT_PTR_PTR", _wrap_delete_TPM2B_EVENT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_PTR_PTR_assign", _wrap_TPM2B_EVENT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_EVENT_PTR_PTR_value", _wrap_TPM2B_EVENT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_MAX_BUFFER_swigconstant", sizeof_TPM2B_MAX_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_BUFFER_ARRAY", _wrap_new_TPM2B_MAX_BUFFER_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_MAX_BUFFER_ARRAY", _wrap_delete_TPM2B_MAX_BUFFER_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_ARRAY___getitem__", _wrap_TPM2B_MAX_BUFFER_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_ARRAY___setitem__", _wrap_TPM2B_MAX_BUFFER_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_ARRAY_cast", _wrap_TPM2B_MAX_BUFFER_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_ARRAY_frompointer", _wrap_TPM2B_MAX_BUFFER_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_ARRAY_swigregister", TPM2B_MAX_BUFFER_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_BUFFER_PTR", _wrap_new_TPM2B_MAX_BUFFER_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_MAX_BUFFER_PTR", _wrap_copy_TPM2B_MAX_BUFFER_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_MAX_BUFFER_PTR", _wrap_delete_TPM2B_MAX_BUFFER_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_PTR_assign", _wrap_TPM2B_MAX_BUFFER_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_PTR_value", _wrap_TPM2B_MAX_BUFFER_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_BUFFER_PTR_PTR", _wrap_new_TPM2B_MAX_BUFFER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_MAX_BUFFER_PTR_PTR", _wrap_copy_TPM2B_MAX_BUFFER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_MAX_BUFFER_PTR_PTR", _wrap_delete_TPM2B_MAX_BUFFER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_PTR_PTR_assign", _wrap_TPM2B_MAX_BUFFER_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_BUFFER_PTR_PTR_value", _wrap_TPM2B_MAX_BUFFER_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_MAX_NV_BUFFER_swigconstant", sizeof_TPM2B_MAX_NV_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_NV_BUFFER_ARRAY", _wrap_new_TPM2B_MAX_NV_BUFFER_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_MAX_NV_BUFFER_ARRAY", _wrap_delete_TPM2B_MAX_NV_BUFFER_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_ARRAY___getitem__", _wrap_TPM2B_MAX_NV_BUFFER_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_ARRAY___setitem__", _wrap_TPM2B_MAX_NV_BUFFER_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_ARRAY_cast", _wrap_TPM2B_MAX_NV_BUFFER_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_ARRAY_frompointer", _wrap_TPM2B_MAX_NV_BUFFER_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_ARRAY_swigregister", TPM2B_MAX_NV_BUFFER_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_NV_BUFFER_PTR", _wrap_new_TPM2B_MAX_NV_BUFFER_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_MAX_NV_BUFFER_PTR", _wrap_copy_TPM2B_MAX_NV_BUFFER_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_MAX_NV_BUFFER_PTR", _wrap_delete_TPM2B_MAX_NV_BUFFER_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_PTR_assign", _wrap_TPM2B_MAX_NV_BUFFER_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_PTR_value", _wrap_TPM2B_MAX_NV_BUFFER_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_MAX_NV_BUFFER_PTR_PTR", _wrap_new_TPM2B_MAX_NV_BUFFER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_MAX_NV_BUFFER_PTR_PTR", _wrap_copy_TPM2B_MAX_NV_BUFFER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_MAX_NV_BUFFER_PTR_PTR", _wrap_delete_TPM2B_MAX_NV_BUFFER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_PTR_PTR_assign", _wrap_TPM2B_MAX_NV_BUFFER_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_MAX_NV_BUFFER_PTR_PTR_value", _wrap_TPM2B_MAX_NV_BUFFER_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_IV_swigconstant", sizeof_TPM2B_IV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_IV_ARRAY", _wrap_new_TPM2B_IV_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_IV_ARRAY", _wrap_delete_TPM2B_IV_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_ARRAY___getitem__", _wrap_TPM2B_IV_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_ARRAY___setitem__", _wrap_TPM2B_IV_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_ARRAY_cast", _wrap_TPM2B_IV_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_ARRAY_frompointer", _wrap_TPM2B_IV_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_ARRAY_swigregister", TPM2B_IV_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_IV_PTR", _wrap_new_TPM2B_IV_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_IV_PTR", _wrap_copy_TPM2B_IV_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_IV_PTR", _wrap_delete_TPM2B_IV_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_PTR_assign", _wrap_TPM2B_IV_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_PTR_value", _wrap_TPM2B_IV_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_IV_PTR_PTR", _wrap_new_TPM2B_IV_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_IV_PTR_PTR", _wrap_copy_TPM2B_IV_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_IV_PTR_PTR", _wrap_delete_TPM2B_IV_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_PTR_PTR_assign", _wrap_TPM2B_IV_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_IV_PTR_PTR_value", _wrap_TPM2B_IV_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_NAME_swigconstant", sizeof_TPMU_NAME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_NAME_ARRAY", _wrap_new_TPMU_NAME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_NAME_ARRAY", _wrap_delete_TPMU_NAME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_ARRAY___getitem__", _wrap_TPMU_NAME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_ARRAY___setitem__", _wrap_TPMU_NAME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_ARRAY_cast", _wrap_TPMU_NAME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_ARRAY_frompointer", _wrap_TPMU_NAME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_ARRAY_swigregister", TPMU_NAME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_NAME_PTR", _wrap_new_TPMU_NAME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_NAME_PTR", _wrap_copy_TPMU_NAME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_NAME_PTR", _wrap_delete_TPMU_NAME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_PTR_assign", _wrap_TPMU_NAME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_PTR_value", _wrap_TPMU_NAME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_NAME_PTR_PTR", _wrap_new_TPMU_NAME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_NAME_PTR_PTR", _wrap_copy_TPMU_NAME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_NAME_PTR_PTR", _wrap_delete_TPMU_NAME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_PTR_PTR_assign", _wrap_TPMU_NAME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_NAME_PTR_PTR_value", _wrap_TPMU_NAME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_NAME_swigconstant", sizeof_TPM2B_NAME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NAME_ARRAY", _wrap_new_TPM2B_NAME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_NAME_ARRAY", _wrap_delete_TPM2B_NAME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_ARRAY___getitem__", _wrap_TPM2B_NAME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_ARRAY___setitem__", _wrap_TPM2B_NAME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_ARRAY_cast", _wrap_TPM2B_NAME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_ARRAY_frompointer", _wrap_TPM2B_NAME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_ARRAY_swigregister", TPM2B_NAME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NAME_PTR", _wrap_new_TPM2B_NAME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_NAME_PTR", _wrap_copy_TPM2B_NAME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_NAME_PTR", _wrap_delete_TPM2B_NAME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_PTR_assign", _wrap_TPM2B_NAME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_PTR_value", _wrap_TPM2B_NAME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NAME_PTR_PTR", _wrap_new_TPM2B_NAME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_NAME_PTR_PTR", _wrap_copy_TPM2B_NAME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_NAME_PTR_PTR", _wrap_delete_TPM2B_NAME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_PTR_PTR_assign", _wrap_TPM2B_NAME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NAME_PTR_PTR_value", _wrap_TPM2B_NAME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_PCR_SELECT_swigconstant", sizeof_TPMS_PCR_SELECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECT_ARRAY", _wrap_new_TPMS_PCR_SELECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_PCR_SELECT_ARRAY", _wrap_delete_TPMS_PCR_SELECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_ARRAY___getitem__", _wrap_TPMS_PCR_SELECT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_ARRAY___setitem__", _wrap_TPMS_PCR_SELECT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_ARRAY_cast", _wrap_TPMS_PCR_SELECT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_ARRAY_frompointer", _wrap_TPMS_PCR_SELECT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_ARRAY_swigregister", TPMS_PCR_SELECT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECT_PTR", _wrap_new_TPMS_PCR_SELECT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_PCR_SELECT_PTR", _wrap_copy_TPMS_PCR_SELECT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_PCR_SELECT_PTR", _wrap_delete_TPMS_PCR_SELECT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_PTR_assign", _wrap_TPMS_PCR_SELECT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_PTR_value", _wrap_TPMS_PCR_SELECT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECT_PTR_PTR", _wrap_new_TPMS_PCR_SELECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_PCR_SELECT_PTR_PTR", _wrap_copy_TPMS_PCR_SELECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_PCR_SELECT_PTR_PTR", _wrap_delete_TPMS_PCR_SELECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_PTR_PTR_assign", _wrap_TPMS_PCR_SELECT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECT_PTR_PTR_value", _wrap_TPMS_PCR_SELECT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_PCR_SELECTION_swigconstant", sizeof_TPMS_PCR_SELECTION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECTION_ARRAY", _wrap_new_TPMS_PCR_SELECTION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_PCR_SELECTION_ARRAY", _wrap_delete_TPMS_PCR_SELECTION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_ARRAY___getitem__", _wrap_TPMS_PCR_SELECTION_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_ARRAY___setitem__", _wrap_TPMS_PCR_SELECTION_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_ARRAY_cast", _wrap_TPMS_PCR_SELECTION_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_ARRAY_frompointer", _wrap_TPMS_PCR_SELECTION_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_ARRAY_swigregister", TPMS_PCR_SELECTION_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECTION_PTR", _wrap_new_TPMS_PCR_SELECTION_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_PCR_SELECTION_PTR", _wrap_copy_TPMS_PCR_SELECTION_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_PCR_SELECTION_PTR", _wrap_delete_TPMS_PCR_SELECTION_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_PTR_assign", _wrap_TPMS_PCR_SELECTION_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_PTR_value", _wrap_TPMS_PCR_SELECTION_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_PCR_SELECTION_PTR_PTR", _wrap_new_TPMS_PCR_SELECTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_PCR_SELECTION_PTR_PTR", _wrap_copy_TPMS_PCR_SELECTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_PCR_SELECTION_PTR_PTR", _wrap_delete_TPMS_PCR_SELECTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_PTR_PTR_assign", _wrap_TPMS_PCR_SELECTION_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_PCR_SELECTION_PTR_PTR_value", _wrap_TPMS_PCR_SELECTION_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_TK_CREATION_swigconstant", sizeof_TPMT_TK_CREATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_CREATION_ARRAY", _wrap_new_TPMT_TK_CREATION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_CREATION_ARRAY", _wrap_delete_TPMT_TK_CREATION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_ARRAY___getitem__", _wrap_TPMT_TK_CREATION_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_ARRAY___setitem__", _wrap_TPMT_TK_CREATION_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_ARRAY_cast", _wrap_TPMT_TK_CREATION_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_ARRAY_frompointer", _wrap_TPMT_TK_CREATION_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_ARRAY_swigregister", TPMT_TK_CREATION_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_CREATION_PTR", _wrap_new_TPMT_TK_CREATION_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_CREATION_PTR", _wrap_copy_TPMT_TK_CREATION_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_CREATION_PTR", _wrap_delete_TPMT_TK_CREATION_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_PTR_assign", _wrap_TPMT_TK_CREATION_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_PTR_value", _wrap_TPMT_TK_CREATION_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_CREATION_PTR_PTR", _wrap_new_TPMT_TK_CREATION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_CREATION_PTR_PTR", _wrap_copy_TPMT_TK_CREATION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_CREATION_PTR_PTR", _wrap_delete_TPMT_TK_CREATION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_PTR_PTR_assign", _wrap_TPMT_TK_CREATION_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_CREATION_PTR_PTR_value", _wrap_TPMT_TK_CREATION_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_TK_VERIFIED_swigconstant", sizeof_TPMT_TK_VERIFIED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_VERIFIED_ARRAY", _wrap_new_TPMT_TK_VERIFIED_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_VERIFIED_ARRAY", _wrap_delete_TPMT_TK_VERIFIED_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_ARRAY___getitem__", _wrap_TPMT_TK_VERIFIED_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_ARRAY___setitem__", _wrap_TPMT_TK_VERIFIED_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_ARRAY_cast", _wrap_TPMT_TK_VERIFIED_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_ARRAY_frompointer", _wrap_TPMT_TK_VERIFIED_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_ARRAY_swigregister", TPMT_TK_VERIFIED_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_VERIFIED_PTR", _wrap_new_TPMT_TK_VERIFIED_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_VERIFIED_PTR", _wrap_copy_TPMT_TK_VERIFIED_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_VERIFIED_PTR", _wrap_delete_TPMT_TK_VERIFIED_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_PTR_assign", _wrap_TPMT_TK_VERIFIED_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_PTR_value", _wrap_TPMT_TK_VERIFIED_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_VERIFIED_PTR_PTR", _wrap_new_TPMT_TK_VERIFIED_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_VERIFIED_PTR_PTR", _wrap_copy_TPMT_TK_VERIFIED_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_VERIFIED_PTR_PTR", _wrap_delete_TPMT_TK_VERIFIED_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_PTR_PTR_assign", _wrap_TPMT_TK_VERIFIED_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_VERIFIED_PTR_PTR_value", _wrap_TPMT_TK_VERIFIED_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_TK_AUTH_swigconstant", sizeof_TPMT_TK_AUTH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_AUTH_ARRAY", _wrap_new_TPMT_TK_AUTH_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_AUTH_ARRAY", _wrap_delete_TPMT_TK_AUTH_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_ARRAY___getitem__", _wrap_TPMT_TK_AUTH_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_ARRAY___setitem__", _wrap_TPMT_TK_AUTH_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_ARRAY_cast", _wrap_TPMT_TK_AUTH_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_ARRAY_frompointer", _wrap_TPMT_TK_AUTH_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_ARRAY_swigregister", TPMT_TK_AUTH_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_AUTH_PTR", _wrap_new_TPMT_TK_AUTH_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_AUTH_PTR", _wrap_copy_TPMT_TK_AUTH_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_AUTH_PTR", _wrap_delete_TPMT_TK_AUTH_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_PTR_assign", _wrap_TPMT_TK_AUTH_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_PTR_value", _wrap_TPMT_TK_AUTH_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_AUTH_PTR_PTR", _wrap_new_TPMT_TK_AUTH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_AUTH_PTR_PTR", _wrap_copy_TPMT_TK_AUTH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_AUTH_PTR_PTR", _wrap_delete_TPMT_TK_AUTH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_PTR_PTR_assign", _wrap_TPMT_TK_AUTH_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_AUTH_PTR_PTR_value", _wrap_TPMT_TK_AUTH_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_TK_HASHCHECK_swigconstant", sizeof_TPMT_TK_HASHCHECK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_HASHCHECK_ARRAY", _wrap_new_TPMT_TK_HASHCHECK_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_HASHCHECK_ARRAY", _wrap_delete_TPMT_TK_HASHCHECK_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_ARRAY___getitem__", _wrap_TPMT_TK_HASHCHECK_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_ARRAY___setitem__", _wrap_TPMT_TK_HASHCHECK_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_ARRAY_cast", _wrap_TPMT_TK_HASHCHECK_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_ARRAY_frompointer", _wrap_TPMT_TK_HASHCHECK_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_ARRAY_swigregister", TPMT_TK_HASHCHECK_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_HASHCHECK_PTR", _wrap_new_TPMT_TK_HASHCHECK_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_HASHCHECK_PTR", _wrap_copy_TPMT_TK_HASHCHECK_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_HASHCHECK_PTR", _wrap_delete_TPMT_TK_HASHCHECK_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_PTR_assign", _wrap_TPMT_TK_HASHCHECK_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_PTR_value", _wrap_TPMT_TK_HASHCHECK_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_TK_HASHCHECK_PTR_PTR", _wrap_new_TPMT_TK_HASHCHECK_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_TK_HASHCHECK_PTR_PTR", _wrap_copy_TPMT_TK_HASHCHECK_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_TK_HASHCHECK_PTR_PTR", _wrap_delete_TPMT_TK_HASHCHECK_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_PTR_PTR_assign", _wrap_TPMT_TK_HASHCHECK_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_TK_HASHCHECK_PTR_PTR_value", _wrap_TPMT_TK_HASHCHECK_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ALG_PROPERTY_swigconstant", sizeof_TPMS_ALG_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALG_PROPERTY_ARRAY", _wrap_new_TPMS_ALG_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALG_PROPERTY_ARRAY", _wrap_delete_TPMS_ALG_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_ARRAY___getitem__", _wrap_TPMS_ALG_PROPERTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_ARRAY___setitem__", _wrap_TPMS_ALG_PROPERTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_ARRAY_cast", _wrap_TPMS_ALG_PROPERTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_ARRAY_frompointer", _wrap_TPMS_ALG_PROPERTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_ARRAY_swigregister", TPMS_ALG_PROPERTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALG_PROPERTY_PTR", _wrap_new_TPMS_ALG_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ALG_PROPERTY_PTR", _wrap_copy_TPMS_ALG_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALG_PROPERTY_PTR", _wrap_delete_TPMS_ALG_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_PTR_assign", _wrap_TPMS_ALG_PROPERTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_PTR_value", _wrap_TPMS_ALG_PROPERTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALG_PROPERTY_PTR_PTR", _wrap_new_TPMS_ALG_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ALG_PROPERTY_PTR_PTR", _wrap_copy_TPMS_ALG_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALG_PROPERTY_PTR_PTR", _wrap_delete_TPMS_ALG_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_PTR_PTR_assign", _wrap_TPMS_ALG_PROPERTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALG_PROPERTY_PTR_PTR_value", _wrap_TPMS_ALG_PROPERTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_TAGGED_PROPERTY_swigconstant", sizeof_TPMS_TAGGED_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PROPERTY_ARRAY", _wrap_new_TPMS_TAGGED_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_PROPERTY_ARRAY", _wrap_delete_TPMS_TAGGED_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_ARRAY___getitem__", _wrap_TPMS_TAGGED_PROPERTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_ARRAY___setitem__", _wrap_TPMS_TAGGED_PROPERTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_ARRAY_cast", _wrap_TPMS_TAGGED_PROPERTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_ARRAY_frompointer", _wrap_TPMS_TAGGED_PROPERTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_ARRAY_swigregister", TPMS_TAGGED_PROPERTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PROPERTY_PTR", _wrap_new_TPMS_TAGGED_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TAGGED_PROPERTY_PTR", _wrap_copy_TPMS_TAGGED_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_PROPERTY_PTR", _wrap_delete_TPMS_TAGGED_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_PTR_assign", _wrap_TPMS_TAGGED_PROPERTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_PTR_value", _wrap_TPMS_TAGGED_PROPERTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PROPERTY_PTR_PTR", _wrap_new_TPMS_TAGGED_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TAGGED_PROPERTY_PTR_PTR", _wrap_copy_TPMS_TAGGED_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_PROPERTY_PTR_PTR", _wrap_delete_TPMS_TAGGED_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_PTR_PTR_assign", _wrap_TPMS_TAGGED_PROPERTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PROPERTY_PTR_PTR_value", _wrap_TPMS_TAGGED_PROPERTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_TAGGED_PCR_SELECT_swigconstant", sizeof_TPMS_TAGGED_PCR_SELECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PCR_SELECT_ARRAY", _wrap_new_TPMS_TAGGED_PCR_SELECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_PCR_SELECT_ARRAY", _wrap_delete_TPMS_TAGGED_PCR_SELECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__", _wrap_TPMS_TAGGED_PCR_SELECT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__", _wrap_TPMS_TAGGED_PCR_SELECT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_ARRAY_cast", _wrap_TPMS_TAGGED_PCR_SELECT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer", _wrap_TPMS_TAGGED_PCR_SELECT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_ARRAY_swigregister", TPMS_TAGGED_PCR_SELECT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PCR_SELECT_PTR", _wrap_new_TPMS_TAGGED_PCR_SELECT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TAGGED_PCR_SELECT_PTR", _wrap_copy_TPMS_TAGGED_PCR_SELECT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_PCR_SELECT_PTR", _wrap_delete_TPMS_TAGGED_PCR_SELECT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_PTR_assign", _wrap_TPMS_TAGGED_PCR_SELECT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_PTR_value", _wrap_TPMS_TAGGED_PCR_SELECT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_PCR_SELECT_PTR_PTR", _wrap_new_TPMS_TAGGED_PCR_SELECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR", _wrap_copy_TPMS_TAGGED_PCR_SELECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR", _wrap_delete_TPMS_TAGGED_PCR_SELECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign", _wrap_TPMS_TAGGED_PCR_SELECT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_PCR_SELECT_PTR_PTR_value", _wrap_TPMS_TAGGED_PCR_SELECT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_TAGGED_POLICY_swigconstant", sizeof_TPMS_TAGGED_POLICY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_POLICY_ARRAY", _wrap_new_TPMS_TAGGED_POLICY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_POLICY_ARRAY", _wrap_delete_TPMS_TAGGED_POLICY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_ARRAY___getitem__", _wrap_TPMS_TAGGED_POLICY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_ARRAY___setitem__", _wrap_TPMS_TAGGED_POLICY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_ARRAY_cast", _wrap_TPMS_TAGGED_POLICY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_ARRAY_frompointer", _wrap_TPMS_TAGGED_POLICY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_ARRAY_swigregister", TPMS_TAGGED_POLICY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_POLICY_PTR", _wrap_new_TPMS_TAGGED_POLICY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TAGGED_POLICY_PTR", _wrap_copy_TPMS_TAGGED_POLICY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_POLICY_PTR", _wrap_delete_TPMS_TAGGED_POLICY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_PTR_assign", _wrap_TPMS_TAGGED_POLICY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_PTR_value", _wrap_TPMS_TAGGED_POLICY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TAGGED_POLICY_PTR_PTR", _wrap_new_TPMS_TAGGED_POLICY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TAGGED_POLICY_PTR_PTR", _wrap_copy_TPMS_TAGGED_POLICY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TAGGED_POLICY_PTR_PTR", _wrap_delete_TPMS_TAGGED_POLICY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_PTR_PTR_assign", _wrap_TPMS_TAGGED_POLICY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TAGGED_POLICY_PTR_PTR_value", _wrap_TPMS_TAGGED_POLICY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_CC_swigconstant", sizeof_TPML_CC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CC_ARRAY", _wrap_new_TPML_CC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_CC_ARRAY", _wrap_delete_TPML_CC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_ARRAY___getitem__", _wrap_TPML_CC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_ARRAY___setitem__", _wrap_TPML_CC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_ARRAY_cast", _wrap_TPML_CC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_ARRAY_frompointer", _wrap_TPML_CC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_ARRAY_swigregister", TPML_CC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CC_PTR", _wrap_new_TPML_CC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_CC_PTR", _wrap_copy_TPML_CC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_CC_PTR", _wrap_delete_TPML_CC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_PTR_assign", _wrap_TPML_CC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_PTR_value", _wrap_TPML_CC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CC_PTR_PTR", _wrap_new_TPML_CC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_CC_PTR_PTR", _wrap_copy_TPML_CC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_CC_PTR_PTR", _wrap_delete_TPML_CC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_PTR_PTR_assign", _wrap_TPML_CC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_CC_PTR_PTR_value", _wrap_TPML_CC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_CCA_swigconstant", sizeof_TPML_CCA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CCA_ARRAY", _wrap_new_TPML_CCA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_CCA_ARRAY", _wrap_delete_TPML_CCA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_ARRAY___getitem__", _wrap_TPML_CCA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_ARRAY___setitem__", _wrap_TPML_CCA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_ARRAY_cast", _wrap_TPML_CCA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_ARRAY_frompointer", _wrap_TPML_CCA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_ARRAY_swigregister", TPML_CCA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CCA_PTR", _wrap_new_TPML_CCA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_CCA_PTR", _wrap_copy_TPML_CCA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_CCA_PTR", _wrap_delete_TPML_CCA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_PTR_assign", _wrap_TPML_CCA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_PTR_value", _wrap_TPML_CCA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_CCA_PTR_PTR", _wrap_new_TPML_CCA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_CCA_PTR_PTR", _wrap_copy_TPML_CCA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_CCA_PTR_PTR", _wrap_delete_TPML_CCA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_PTR_PTR_assign", _wrap_TPML_CCA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_CCA_PTR_PTR_value", _wrap_TPML_CCA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_ALG_swigconstant", sizeof_TPML_ALG_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_ARRAY", _wrap_new_TPML_ALG_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ALG_ARRAY", _wrap_delete_TPML_ALG_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_ARRAY___getitem__", _wrap_TPML_ALG_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_ARRAY___setitem__", _wrap_TPML_ALG_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_ARRAY_cast", _wrap_TPML_ALG_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_ARRAY_frompointer", _wrap_TPML_ALG_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_ARRAY_swigregister", TPML_ALG_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_PTR", _wrap_new_TPML_ALG_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_ALG_PTR", _wrap_copy_TPML_ALG_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ALG_PTR", _wrap_delete_TPML_ALG_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PTR_assign", _wrap_TPML_ALG_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PTR_value", _wrap_TPML_ALG_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_PTR_PTR", _wrap_new_TPML_ALG_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_ALG_PTR_PTR", _wrap_copy_TPML_ALG_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ALG_PTR_PTR", _wrap_delete_TPML_ALG_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PTR_PTR_assign", _wrap_TPML_ALG_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PTR_PTR_value", _wrap_TPML_ALG_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_HANDLE_swigconstant", sizeof_TPML_HANDLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_HANDLE_ARRAY", _wrap_new_TPML_HANDLE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_HANDLE_ARRAY", _wrap_delete_TPML_HANDLE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_ARRAY___getitem__", _wrap_TPML_HANDLE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_ARRAY___setitem__", _wrap_TPML_HANDLE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_ARRAY_cast", _wrap_TPML_HANDLE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_ARRAY_frompointer", _wrap_TPML_HANDLE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_ARRAY_swigregister", TPML_HANDLE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_HANDLE_PTR", _wrap_new_TPML_HANDLE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_HANDLE_PTR", _wrap_copy_TPML_HANDLE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_HANDLE_PTR", _wrap_delete_TPML_HANDLE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_PTR_assign", _wrap_TPML_HANDLE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_PTR_value", _wrap_TPML_HANDLE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_HANDLE_PTR_PTR", _wrap_new_TPML_HANDLE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_HANDLE_PTR_PTR", _wrap_copy_TPML_HANDLE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_HANDLE_PTR_PTR", _wrap_delete_TPML_HANDLE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_PTR_PTR_assign", _wrap_TPML_HANDLE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_HANDLE_PTR_PTR_value", _wrap_TPML_HANDLE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_DIGEST_swigconstant", sizeof_TPML_DIGEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_ARRAY", _wrap_new_TPML_DIGEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_DIGEST_ARRAY", _wrap_delete_TPML_DIGEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_ARRAY___getitem__", _wrap_TPML_DIGEST_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_ARRAY___setitem__", _wrap_TPML_DIGEST_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_ARRAY_cast", _wrap_TPML_DIGEST_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_ARRAY_frompointer", _wrap_TPML_DIGEST_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_ARRAY_swigregister", TPML_DIGEST_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_PTR", _wrap_new_TPML_DIGEST_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_DIGEST_PTR", _wrap_copy_TPML_DIGEST_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_DIGEST_PTR", _wrap_delete_TPML_DIGEST_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_PTR_assign", _wrap_TPML_DIGEST_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_PTR_value", _wrap_TPML_DIGEST_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_PTR_PTR", _wrap_new_TPML_DIGEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_DIGEST_PTR_PTR", _wrap_copy_TPML_DIGEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_DIGEST_PTR_PTR", _wrap_delete_TPML_DIGEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_PTR_PTR_assign", _wrap_TPML_DIGEST_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_PTR_PTR_value", _wrap_TPML_DIGEST_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_DIGEST_VALUES_swigconstant", sizeof_TPML_DIGEST_VALUES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_VALUES_ARRAY", _wrap_new_TPML_DIGEST_VALUES_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_DIGEST_VALUES_ARRAY", _wrap_delete_TPML_DIGEST_VALUES_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_ARRAY___getitem__", _wrap_TPML_DIGEST_VALUES_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_ARRAY___setitem__", _wrap_TPML_DIGEST_VALUES_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_ARRAY_cast", _wrap_TPML_DIGEST_VALUES_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_ARRAY_frompointer", _wrap_TPML_DIGEST_VALUES_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_ARRAY_swigregister", TPML_DIGEST_VALUES_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_VALUES_PTR", _wrap_new_TPML_DIGEST_VALUES_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_DIGEST_VALUES_PTR", _wrap_copy_TPML_DIGEST_VALUES_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_DIGEST_VALUES_PTR", _wrap_delete_TPML_DIGEST_VALUES_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_PTR_assign", _wrap_TPML_DIGEST_VALUES_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_PTR_value", _wrap_TPML_DIGEST_VALUES_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_DIGEST_VALUES_PTR_PTR", _wrap_new_TPML_DIGEST_VALUES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_DIGEST_VALUES_PTR_PTR", _wrap_copy_TPML_DIGEST_VALUES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_DIGEST_VALUES_PTR_PTR", _wrap_delete_TPML_DIGEST_VALUES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_PTR_PTR_assign", _wrap_TPML_DIGEST_VALUES_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_DIGEST_VALUES_PTR_PTR_value", _wrap_TPML_DIGEST_VALUES_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_PCR_SELECTION_swigconstant", sizeof_TPML_PCR_SELECTION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_PCR_SELECTION_ARRAY", _wrap_new_TPML_PCR_SELECTION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_PCR_SELECTION_ARRAY", _wrap_delete_TPML_PCR_SELECTION_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_ARRAY___getitem__", _wrap_TPML_PCR_SELECTION_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_ARRAY___setitem__", _wrap_TPML_PCR_SELECTION_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_ARRAY_cast", _wrap_TPML_PCR_SELECTION_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_ARRAY_frompointer", _wrap_TPML_PCR_SELECTION_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_ARRAY_swigregister", TPML_PCR_SELECTION_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_PCR_SELECTION_PTR", _wrap_new_TPML_PCR_SELECTION_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_PCR_SELECTION_PTR", _wrap_copy_TPML_PCR_SELECTION_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_PCR_SELECTION_PTR", _wrap_delete_TPML_PCR_SELECTION_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_PTR_assign", _wrap_TPML_PCR_SELECTION_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_PTR_value", _wrap_TPML_PCR_SELECTION_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_PCR_SELECTION_PTR_PTR", _wrap_new_TPML_PCR_SELECTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_PCR_SELECTION_PTR_PTR", _wrap_copy_TPML_PCR_SELECTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_PCR_SELECTION_PTR_PTR", _wrap_delete_TPML_PCR_SELECTION_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_PTR_PTR_assign", _wrap_TPML_PCR_SELECTION_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_PCR_SELECTION_PTR_PTR_value", _wrap_TPML_PCR_SELECTION_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_ALG_PROPERTY_swigconstant", sizeof_TPML_ALG_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_PROPERTY_ARRAY", _wrap_new_TPML_ALG_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ALG_PROPERTY_ARRAY", _wrap_delete_TPML_ALG_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_ARRAY___getitem__", _wrap_TPML_ALG_PROPERTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_ARRAY___setitem__", _wrap_TPML_ALG_PROPERTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_ARRAY_cast", _wrap_TPML_ALG_PROPERTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_ARRAY_frompointer", _wrap_TPML_ALG_PROPERTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_ARRAY_swigregister", TPML_ALG_PROPERTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_PROPERTY_PTR", _wrap_new_TPML_ALG_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_ALG_PROPERTY_PTR", _wrap_copy_TPML_ALG_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ALG_PROPERTY_PTR", _wrap_delete_TPML_ALG_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_PTR_assign", _wrap_TPML_ALG_PROPERTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_PTR_value", _wrap_TPML_ALG_PROPERTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ALG_PROPERTY_PTR_PTR", _wrap_new_TPML_ALG_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_ALG_PROPERTY_PTR_PTR", _wrap_copy_TPML_ALG_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ALG_PROPERTY_PTR_PTR", _wrap_delete_TPML_ALG_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_PTR_PTR_assign", _wrap_TPML_ALG_PROPERTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_ALG_PROPERTY_PTR_PTR_value", _wrap_TPML_ALG_PROPERTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_TAGGED_TPM_PROPERTY_swigconstant", sizeof_TPML_TAGGED_TPM_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_TPM_PROPERTY_ARRAY", _wrap_new_TPML_TAGGED_TPM_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_TAGGED_TPM_PROPERTY_ARRAY", _wrap_delete_TPML_TAGGED_TPM_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__", _wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__", _wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_ARRAY_cast", _wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer", _wrap_TPML_TAGGED_TPM_PROPERTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_ARRAY_swigregister", TPML_TAGGED_TPM_PROPERTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_TPM_PROPERTY_PTR", _wrap_new_TPML_TAGGED_TPM_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_TAGGED_TPM_PROPERTY_PTR", _wrap_copy_TPML_TAGGED_TPM_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_TAGGED_TPM_PROPERTY_PTR", _wrap_delete_TPML_TAGGED_TPM_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_PTR_assign", _wrap_TPML_TAGGED_TPM_PROPERTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_PTR_value", _wrap_TPML_TAGGED_TPM_PROPERTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_TPM_PROPERTY_PTR_PTR", _wrap_new_TPML_TAGGED_TPM_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR", _wrap_copy_TPML_TAGGED_TPM_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR", _wrap_delete_TPML_TAGGED_TPM_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign", _wrap_TPML_TAGGED_TPM_PROPERTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value", _wrap_TPML_TAGGED_TPM_PROPERTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_TAGGED_PCR_PROPERTY_swigconstant", sizeof_TPML_TAGGED_PCR_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_PCR_PROPERTY_ARRAY", _wrap_new_TPML_TAGGED_PCR_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_TAGGED_PCR_PROPERTY_ARRAY", _wrap_delete_TPML_TAGGED_PCR_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__", _wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__", _wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_ARRAY_cast", _wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer", _wrap_TPML_TAGGED_PCR_PROPERTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_ARRAY_swigregister", TPML_TAGGED_PCR_PROPERTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_PCR_PROPERTY_PTR", _wrap_new_TPML_TAGGED_PCR_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_TAGGED_PCR_PROPERTY_PTR", _wrap_copy_TPML_TAGGED_PCR_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_TAGGED_PCR_PROPERTY_PTR", _wrap_delete_TPML_TAGGED_PCR_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_PTR_assign", _wrap_TPML_TAGGED_PCR_PROPERTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_PTR_value", _wrap_TPML_TAGGED_PCR_PROPERTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_TAGGED_PCR_PROPERTY_PTR_PTR", _wrap_new_TPML_TAGGED_PCR_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR", _wrap_copy_TPML_TAGGED_PCR_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR", _wrap_delete_TPML_TAGGED_PCR_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign", _wrap_TPML_TAGGED_PCR_PROPERTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value", _wrap_TPML_TAGGED_PCR_PROPERTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_ECC_CURVE_swigconstant", sizeof_TPML_ECC_CURVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ECC_CURVE_ARRAY", _wrap_new_TPML_ECC_CURVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ECC_CURVE_ARRAY", _wrap_delete_TPML_ECC_CURVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_ARRAY___getitem__", _wrap_TPML_ECC_CURVE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_ARRAY___setitem__", _wrap_TPML_ECC_CURVE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_ARRAY_cast", _wrap_TPML_ECC_CURVE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_ARRAY_frompointer", _wrap_TPML_ECC_CURVE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_ARRAY_swigregister", TPML_ECC_CURVE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ECC_CURVE_PTR", _wrap_new_TPML_ECC_CURVE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_ECC_CURVE_PTR", _wrap_copy_TPML_ECC_CURVE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ECC_CURVE_PTR", _wrap_delete_TPML_ECC_CURVE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_PTR_assign", _wrap_TPML_ECC_CURVE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_PTR_value", _wrap_TPML_ECC_CURVE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_ECC_CURVE_PTR_PTR", _wrap_new_TPML_ECC_CURVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_ECC_CURVE_PTR_PTR", _wrap_copy_TPML_ECC_CURVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_ECC_CURVE_PTR_PTR", _wrap_delete_TPML_ECC_CURVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_PTR_PTR_assign", _wrap_TPML_ECC_CURVE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_ECC_CURVE_PTR_PTR_value", _wrap_TPML_ECC_CURVE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_INTEL_PTT_PROPERTY_swigconstant", sizeof_TPML_INTEL_PTT_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_INTEL_PTT_PROPERTY_ARRAY", _wrap_new_TPML_INTEL_PTT_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_INTEL_PTT_PROPERTY_ARRAY", _wrap_delete_TPML_INTEL_PTT_PROPERTY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__", _wrap_TPML_INTEL_PTT_PROPERTY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__", _wrap_TPML_INTEL_PTT_PROPERTY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_ARRAY_cast", _wrap_TPML_INTEL_PTT_PROPERTY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer", _wrap_TPML_INTEL_PTT_PROPERTY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_ARRAY_swigregister", TPML_INTEL_PTT_PROPERTY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_INTEL_PTT_PROPERTY_PTR", _wrap_new_TPML_INTEL_PTT_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_INTEL_PTT_PROPERTY_PTR", _wrap_copy_TPML_INTEL_PTT_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_INTEL_PTT_PROPERTY_PTR", _wrap_delete_TPML_INTEL_PTT_PROPERTY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_PTR_assign", _wrap_TPML_INTEL_PTT_PROPERTY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_PTR_value", _wrap_TPML_INTEL_PTT_PROPERTY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_INTEL_PTT_PROPERTY_PTR_PTR", _wrap_new_TPML_INTEL_PTT_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR", _wrap_copy_TPML_INTEL_PTT_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR", _wrap_delete_TPML_INTEL_PTT_PROPERTY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign", _wrap_TPML_INTEL_PTT_PROPERTY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_INTEL_PTT_PROPERTY_PTR_PTR_value", _wrap_TPML_INTEL_PTT_PROPERTY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_CAPABILITIES_swigconstant", sizeof_TPMU_CAPABILITIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_CAPABILITIES_ARRAY", _wrap_new_TPMU_CAPABILITIES_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_CAPABILITIES_ARRAY", _wrap_delete_TPMU_CAPABILITIES_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ARRAY___getitem__", _wrap_TPMU_CAPABILITIES_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ARRAY___setitem__", _wrap_TPMU_CAPABILITIES_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ARRAY_cast", _wrap_TPMU_CAPABILITIES_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ARRAY_frompointer", _wrap_TPMU_CAPABILITIES_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_ARRAY_swigregister", TPMU_CAPABILITIES_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_CAPABILITIES_PTR", _wrap_new_TPMU_CAPABILITIES_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_CAPABILITIES_PTR", _wrap_copy_TPMU_CAPABILITIES_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_CAPABILITIES_PTR", _wrap_delete_TPMU_CAPABILITIES_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_PTR_assign", _wrap_TPMU_CAPABILITIES_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_PTR_value", _wrap_TPMU_CAPABILITIES_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_CAPABILITIES_PTR_PTR", _wrap_new_TPMU_CAPABILITIES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_CAPABILITIES_PTR_PTR", _wrap_copy_TPMU_CAPABILITIES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_CAPABILITIES_PTR_PTR", _wrap_delete_TPMU_CAPABILITIES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_PTR_PTR_assign", _wrap_TPMU_CAPABILITIES_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_CAPABILITIES_PTR_PTR_value", _wrap_TPMU_CAPABILITIES_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CAPABILITY_DATA_swigconstant", sizeof_TPMS_CAPABILITY_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CAPABILITY_DATA_ARRAY", _wrap_new_TPMS_CAPABILITY_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CAPABILITY_DATA_ARRAY", _wrap_delete_TPMS_CAPABILITY_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_ARRAY___getitem__", _wrap_TPMS_CAPABILITY_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_ARRAY___setitem__", _wrap_TPMS_CAPABILITY_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_ARRAY_cast", _wrap_TPMS_CAPABILITY_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_ARRAY_frompointer", _wrap_TPMS_CAPABILITY_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_ARRAY_swigregister", TPMS_CAPABILITY_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CAPABILITY_DATA_PTR", _wrap_new_TPMS_CAPABILITY_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CAPABILITY_DATA_PTR", _wrap_copy_TPMS_CAPABILITY_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CAPABILITY_DATA_PTR", _wrap_delete_TPMS_CAPABILITY_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_PTR_assign", _wrap_TPMS_CAPABILITY_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_PTR_value", _wrap_TPMS_CAPABILITY_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CAPABILITY_DATA_PTR_PTR", _wrap_new_TPMS_CAPABILITY_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CAPABILITY_DATA_PTR_PTR", _wrap_copy_TPMS_CAPABILITY_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CAPABILITY_DATA_PTR_PTR", _wrap_delete_TPMS_CAPABILITY_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_PTR_PTR_assign", _wrap_TPMS_CAPABILITY_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CAPABILITY_DATA_PTR_PTR_value", _wrap_TPMS_CAPABILITY_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CLOCK_INFO_swigconstant", sizeof_TPMS_CLOCK_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CLOCK_INFO_ARRAY", _wrap_new_TPMS_CLOCK_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CLOCK_INFO_ARRAY", _wrap_delete_TPMS_CLOCK_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_ARRAY___getitem__", _wrap_TPMS_CLOCK_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_ARRAY___setitem__", _wrap_TPMS_CLOCK_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_ARRAY_cast", _wrap_TPMS_CLOCK_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_ARRAY_frompointer", _wrap_TPMS_CLOCK_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_ARRAY_swigregister", TPMS_CLOCK_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CLOCK_INFO_PTR", _wrap_new_TPMS_CLOCK_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CLOCK_INFO_PTR", _wrap_copy_TPMS_CLOCK_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CLOCK_INFO_PTR", _wrap_delete_TPMS_CLOCK_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_PTR_assign", _wrap_TPMS_CLOCK_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_PTR_value", _wrap_TPMS_CLOCK_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CLOCK_INFO_PTR_PTR", _wrap_new_TPMS_CLOCK_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CLOCK_INFO_PTR_PTR", _wrap_copy_TPMS_CLOCK_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CLOCK_INFO_PTR_PTR", _wrap_delete_TPMS_CLOCK_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_PTR_PTR_assign", _wrap_TPMS_CLOCK_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CLOCK_INFO_PTR_PTR_value", _wrap_TPMS_CLOCK_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_TIME_INFO_swigconstant", sizeof_TPMS_TIME_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_INFO_ARRAY", _wrap_new_TPMS_TIME_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TIME_INFO_ARRAY", _wrap_delete_TPMS_TIME_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_ARRAY___getitem__", _wrap_TPMS_TIME_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_ARRAY___setitem__", _wrap_TPMS_TIME_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_ARRAY_cast", _wrap_TPMS_TIME_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_ARRAY_frompointer", _wrap_TPMS_TIME_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_ARRAY_swigregister", TPMS_TIME_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_INFO_PTR", _wrap_new_TPMS_TIME_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TIME_INFO_PTR", _wrap_copy_TPMS_TIME_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TIME_INFO_PTR", _wrap_delete_TPMS_TIME_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_PTR_assign", _wrap_TPMS_TIME_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_PTR_value", _wrap_TPMS_TIME_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_INFO_PTR_PTR", _wrap_new_TPMS_TIME_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TIME_INFO_PTR_PTR", _wrap_copy_TPMS_TIME_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TIME_INFO_PTR_PTR", _wrap_delete_TPMS_TIME_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_PTR_PTR_assign", _wrap_TPMS_TIME_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_INFO_PTR_PTR_value", _wrap_TPMS_TIME_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_TIME_ATTEST_INFO_swigconstant", sizeof_TPMS_TIME_ATTEST_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_ATTEST_INFO_ARRAY", _wrap_new_TPMS_TIME_ATTEST_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TIME_ATTEST_INFO_ARRAY", _wrap_delete_TPMS_TIME_ATTEST_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_ARRAY___getitem__", _wrap_TPMS_TIME_ATTEST_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_ARRAY___setitem__", _wrap_TPMS_TIME_ATTEST_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_ARRAY_cast", _wrap_TPMS_TIME_ATTEST_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_ARRAY_frompointer", _wrap_TPMS_TIME_ATTEST_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_ARRAY_swigregister", TPMS_TIME_ATTEST_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_ATTEST_INFO_PTR", _wrap_new_TPMS_TIME_ATTEST_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TIME_ATTEST_INFO_PTR", _wrap_copy_TPMS_TIME_ATTEST_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TIME_ATTEST_INFO_PTR", _wrap_delete_TPMS_TIME_ATTEST_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_PTR_assign", _wrap_TPMS_TIME_ATTEST_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_PTR_value", _wrap_TPMS_TIME_ATTEST_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_TIME_ATTEST_INFO_PTR_PTR", _wrap_new_TPMS_TIME_ATTEST_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_TIME_ATTEST_INFO_PTR_PTR", _wrap_copy_TPMS_TIME_ATTEST_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_TIME_ATTEST_INFO_PTR_PTR", _wrap_delete_TPMS_TIME_ATTEST_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_PTR_PTR_assign", _wrap_TPMS_TIME_ATTEST_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_TIME_ATTEST_INFO_PTR_PTR_value", _wrap_TPMS_TIME_ATTEST_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CERTIFY_INFO_swigconstant", sizeof_TPMS_CERTIFY_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CERTIFY_INFO_ARRAY", _wrap_new_TPMS_CERTIFY_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CERTIFY_INFO_ARRAY", _wrap_delete_TPMS_CERTIFY_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_ARRAY___getitem__", _wrap_TPMS_CERTIFY_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_ARRAY___setitem__", _wrap_TPMS_CERTIFY_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_ARRAY_cast", _wrap_TPMS_CERTIFY_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_ARRAY_frompointer", _wrap_TPMS_CERTIFY_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_ARRAY_swigregister", TPMS_CERTIFY_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CERTIFY_INFO_PTR", _wrap_new_TPMS_CERTIFY_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CERTIFY_INFO_PTR", _wrap_copy_TPMS_CERTIFY_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CERTIFY_INFO_PTR", _wrap_delete_TPMS_CERTIFY_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_PTR_assign", _wrap_TPMS_CERTIFY_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_PTR_value", _wrap_TPMS_CERTIFY_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CERTIFY_INFO_PTR_PTR", _wrap_new_TPMS_CERTIFY_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CERTIFY_INFO_PTR_PTR", _wrap_copy_TPMS_CERTIFY_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CERTIFY_INFO_PTR_PTR", _wrap_delete_TPMS_CERTIFY_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_PTR_PTR_assign", _wrap_TPMS_CERTIFY_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CERTIFY_INFO_PTR_PTR_value", _wrap_TPMS_CERTIFY_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_QUOTE_INFO_swigconstant", sizeof_TPMS_QUOTE_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_QUOTE_INFO_ARRAY", _wrap_new_TPMS_QUOTE_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_QUOTE_INFO_ARRAY", _wrap_delete_TPMS_QUOTE_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_ARRAY___getitem__", _wrap_TPMS_QUOTE_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_ARRAY___setitem__", _wrap_TPMS_QUOTE_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_ARRAY_cast", _wrap_TPMS_QUOTE_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_ARRAY_frompointer", _wrap_TPMS_QUOTE_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_ARRAY_swigregister", TPMS_QUOTE_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_QUOTE_INFO_PTR", _wrap_new_TPMS_QUOTE_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_QUOTE_INFO_PTR", _wrap_copy_TPMS_QUOTE_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_QUOTE_INFO_PTR", _wrap_delete_TPMS_QUOTE_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_PTR_assign", _wrap_TPMS_QUOTE_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_PTR_value", _wrap_TPMS_QUOTE_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_QUOTE_INFO_PTR_PTR", _wrap_new_TPMS_QUOTE_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_QUOTE_INFO_PTR_PTR", _wrap_copy_TPMS_QUOTE_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_QUOTE_INFO_PTR_PTR", _wrap_delete_TPMS_QUOTE_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_PTR_PTR_assign", _wrap_TPMS_QUOTE_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_QUOTE_INFO_PTR_PTR_value", _wrap_TPMS_QUOTE_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_COMMAND_AUDIT_INFO_swigconstant", sizeof_TPMS_COMMAND_AUDIT_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_COMMAND_AUDIT_INFO_ARRAY", _wrap_new_TPMS_COMMAND_AUDIT_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_COMMAND_AUDIT_INFO_ARRAY", _wrap_delete_TPMS_COMMAND_AUDIT_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__", _wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__", _wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_ARRAY_cast", _wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer", _wrap_TPMS_COMMAND_AUDIT_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_ARRAY_swigregister", TPMS_COMMAND_AUDIT_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_COMMAND_AUDIT_INFO_PTR", _wrap_new_TPMS_COMMAND_AUDIT_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_COMMAND_AUDIT_INFO_PTR", _wrap_copy_TPMS_COMMAND_AUDIT_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_COMMAND_AUDIT_INFO_PTR", _wrap_delete_TPMS_COMMAND_AUDIT_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_PTR_assign", _wrap_TPMS_COMMAND_AUDIT_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_PTR_value", _wrap_TPMS_COMMAND_AUDIT_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_COMMAND_AUDIT_INFO_PTR_PTR", _wrap_new_TPMS_COMMAND_AUDIT_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR", _wrap_copy_TPMS_COMMAND_AUDIT_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR", _wrap_delete_TPMS_COMMAND_AUDIT_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign", _wrap_TPMS_COMMAND_AUDIT_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value", _wrap_TPMS_COMMAND_AUDIT_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SESSION_AUDIT_INFO_swigconstant", sizeof_TPMS_SESSION_AUDIT_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SESSION_AUDIT_INFO_ARRAY", _wrap_new_TPMS_SESSION_AUDIT_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SESSION_AUDIT_INFO_ARRAY", _wrap_delete_TPMS_SESSION_AUDIT_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__", _wrap_TPMS_SESSION_AUDIT_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__", _wrap_TPMS_SESSION_AUDIT_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_ARRAY_cast", _wrap_TPMS_SESSION_AUDIT_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer", _wrap_TPMS_SESSION_AUDIT_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_ARRAY_swigregister", TPMS_SESSION_AUDIT_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SESSION_AUDIT_INFO_PTR", _wrap_new_TPMS_SESSION_AUDIT_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SESSION_AUDIT_INFO_PTR", _wrap_copy_TPMS_SESSION_AUDIT_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SESSION_AUDIT_INFO_PTR", _wrap_delete_TPMS_SESSION_AUDIT_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_PTR_assign", _wrap_TPMS_SESSION_AUDIT_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_PTR_value", _wrap_TPMS_SESSION_AUDIT_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SESSION_AUDIT_INFO_PTR_PTR", _wrap_new_TPMS_SESSION_AUDIT_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR", _wrap_copy_TPMS_SESSION_AUDIT_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR", _wrap_delete_TPMS_SESSION_AUDIT_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign", _wrap_TPMS_SESSION_AUDIT_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SESSION_AUDIT_INFO_PTR_PTR_value", _wrap_TPMS_SESSION_AUDIT_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CREATION_INFO_swigconstant", sizeof_TPMS_CREATION_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_INFO_ARRAY", _wrap_new_TPMS_CREATION_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CREATION_INFO_ARRAY", _wrap_delete_TPMS_CREATION_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_ARRAY___getitem__", _wrap_TPMS_CREATION_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_ARRAY___setitem__", _wrap_TPMS_CREATION_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_ARRAY_cast", _wrap_TPMS_CREATION_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_ARRAY_frompointer", _wrap_TPMS_CREATION_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_ARRAY_swigregister", TPMS_CREATION_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_INFO_PTR", _wrap_new_TPMS_CREATION_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CREATION_INFO_PTR", _wrap_copy_TPMS_CREATION_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CREATION_INFO_PTR", _wrap_delete_TPMS_CREATION_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_PTR_assign", _wrap_TPMS_CREATION_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_PTR_value", _wrap_TPMS_CREATION_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_INFO_PTR_PTR", _wrap_new_TPMS_CREATION_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CREATION_INFO_PTR_PTR", _wrap_copy_TPMS_CREATION_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CREATION_INFO_PTR_PTR", _wrap_delete_TPMS_CREATION_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_PTR_PTR_assign", _wrap_TPMS_CREATION_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_INFO_PTR_PTR_value", _wrap_TPMS_CREATION_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_NV_CERTIFY_INFO_swigconstant", sizeof_TPMS_NV_CERTIFY_INFO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_CERTIFY_INFO_ARRAY", _wrap_new_TPMS_NV_CERTIFY_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_CERTIFY_INFO_ARRAY", _wrap_delete_TPMS_NV_CERTIFY_INFO_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_ARRAY___getitem__", _wrap_TPMS_NV_CERTIFY_INFO_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_ARRAY___setitem__", _wrap_TPMS_NV_CERTIFY_INFO_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_ARRAY_cast", _wrap_TPMS_NV_CERTIFY_INFO_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_ARRAY_frompointer", _wrap_TPMS_NV_CERTIFY_INFO_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_ARRAY_swigregister", TPMS_NV_CERTIFY_INFO_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_CERTIFY_INFO_PTR", _wrap_new_TPMS_NV_CERTIFY_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_NV_CERTIFY_INFO_PTR", _wrap_copy_TPMS_NV_CERTIFY_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_CERTIFY_INFO_PTR", _wrap_delete_TPMS_NV_CERTIFY_INFO_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_PTR_assign", _wrap_TPMS_NV_CERTIFY_INFO_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_PTR_value", _wrap_TPMS_NV_CERTIFY_INFO_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_CERTIFY_INFO_PTR_PTR", _wrap_new_TPMS_NV_CERTIFY_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_NV_CERTIFY_INFO_PTR_PTR", _wrap_copy_TPMS_NV_CERTIFY_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_CERTIFY_INFO_PTR_PTR", _wrap_delete_TPMS_NV_CERTIFY_INFO_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_PTR_PTR_assign", _wrap_TPMS_NV_CERTIFY_INFO_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_CERTIFY_INFO_PTR_PTR_value", _wrap_TPMS_NV_CERTIFY_INFO_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_ATTEST_swigconstant", sizeof_TPMU_ATTEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ATTEST_ARRAY", _wrap_new_TPMU_ATTEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ATTEST_ARRAY", _wrap_delete_TPMU_ATTEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_ARRAY___getitem__", _wrap_TPMU_ATTEST_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_ARRAY___setitem__", _wrap_TPMU_ATTEST_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_ARRAY_cast", _wrap_TPMU_ATTEST_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_ARRAY_frompointer", _wrap_TPMU_ATTEST_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_ARRAY_swigregister", TPMU_ATTEST_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ATTEST_PTR", _wrap_new_TPMU_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_ATTEST_PTR", _wrap_copy_TPMU_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ATTEST_PTR", _wrap_delete_TPMU_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_PTR_assign", _wrap_TPMU_ATTEST_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_PTR_value", _wrap_TPMU_ATTEST_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ATTEST_PTR_PTR", _wrap_new_TPMU_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_ATTEST_PTR_PTR", _wrap_copy_TPMU_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ATTEST_PTR_PTR", _wrap_delete_TPMU_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_PTR_PTR_assign", _wrap_TPMU_ATTEST_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_ATTEST_PTR_PTR_value", _wrap_TPMU_ATTEST_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ATTEST_swigconstant", sizeof_TPMS_ATTEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ATTEST_ARRAY", _wrap_new_TPMS_ATTEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ATTEST_ARRAY", _wrap_delete_TPMS_ATTEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_ARRAY___getitem__", _wrap_TPMS_ATTEST_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_ARRAY___setitem__", _wrap_TPMS_ATTEST_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_ARRAY_cast", _wrap_TPMS_ATTEST_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_ARRAY_frompointer", _wrap_TPMS_ATTEST_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_ARRAY_swigregister", TPMS_ATTEST_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ATTEST_PTR", _wrap_new_TPMS_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ATTEST_PTR", _wrap_copy_TPMS_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ATTEST_PTR", _wrap_delete_TPMS_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_PTR_assign", _wrap_TPMS_ATTEST_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_PTR_value", _wrap_TPMS_ATTEST_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ATTEST_PTR_PTR", _wrap_new_TPMS_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ATTEST_PTR_PTR", _wrap_copy_TPMS_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ATTEST_PTR_PTR", _wrap_delete_TPMS_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_PTR_PTR_assign", _wrap_TPMS_ATTEST_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ATTEST_PTR_PTR_value", _wrap_TPMS_ATTEST_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_ATTEST_swigconstant", sizeof_TPM2B_ATTEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ATTEST_ARRAY", _wrap_new_TPM2B_ATTEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ATTEST_ARRAY", _wrap_delete_TPM2B_ATTEST_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_ARRAY___getitem__", _wrap_TPM2B_ATTEST_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_ARRAY___setitem__", _wrap_TPM2B_ATTEST_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_ARRAY_cast", _wrap_TPM2B_ATTEST_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_ARRAY_frompointer", _wrap_TPM2B_ATTEST_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_ARRAY_swigregister", TPM2B_ATTEST_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ATTEST_PTR", _wrap_new_TPM2B_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ATTEST_PTR", _wrap_copy_TPM2B_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ATTEST_PTR", _wrap_delete_TPM2B_ATTEST_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_PTR_assign", _wrap_TPM2B_ATTEST_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_PTR_value", _wrap_TPM2B_ATTEST_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ATTEST_PTR_PTR", _wrap_new_TPM2B_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ATTEST_PTR_PTR", _wrap_copy_TPM2B_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ATTEST_PTR_PTR", _wrap_delete_TPM2B_ATTEST_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_PTR_PTR_assign", _wrap_TPM2B_ATTEST_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ATTEST_PTR_PTR_value", _wrap_TPM2B_ATTEST_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_AUTH_COMMAND_swigconstant", sizeof_TPMS_AUTH_COMMAND_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_COMMAND_ARRAY", _wrap_new_TPMS_AUTH_COMMAND_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AUTH_COMMAND_ARRAY", _wrap_delete_TPMS_AUTH_COMMAND_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_ARRAY___getitem__", _wrap_TPMS_AUTH_COMMAND_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_ARRAY___setitem__", _wrap_TPMS_AUTH_COMMAND_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_ARRAY_cast", _wrap_TPMS_AUTH_COMMAND_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_ARRAY_frompointer", _wrap_TPMS_AUTH_COMMAND_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_ARRAY_swigregister", TPMS_AUTH_COMMAND_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_COMMAND_PTR", _wrap_new_TPMS_AUTH_COMMAND_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_AUTH_COMMAND_PTR", _wrap_copy_TPMS_AUTH_COMMAND_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AUTH_COMMAND_PTR", _wrap_delete_TPMS_AUTH_COMMAND_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_PTR_assign", _wrap_TPMS_AUTH_COMMAND_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_PTR_value", _wrap_TPMS_AUTH_COMMAND_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_COMMAND_PTR_PTR", _wrap_new_TPMS_AUTH_COMMAND_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_AUTH_COMMAND_PTR_PTR", _wrap_copy_TPMS_AUTH_COMMAND_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AUTH_COMMAND_PTR_PTR", _wrap_delete_TPMS_AUTH_COMMAND_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_PTR_PTR_assign", _wrap_TPMS_AUTH_COMMAND_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_COMMAND_PTR_PTR_value", _wrap_TPMS_AUTH_COMMAND_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_AUTH_RESPONSE_swigconstant", sizeof_TPMS_AUTH_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_RESPONSE_ARRAY", _wrap_new_TPMS_AUTH_RESPONSE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AUTH_RESPONSE_ARRAY", _wrap_delete_TPMS_AUTH_RESPONSE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_ARRAY___getitem__", _wrap_TPMS_AUTH_RESPONSE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_ARRAY___setitem__", _wrap_TPMS_AUTH_RESPONSE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_ARRAY_cast", _wrap_TPMS_AUTH_RESPONSE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_ARRAY_frompointer", _wrap_TPMS_AUTH_RESPONSE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_ARRAY_swigregister", TPMS_AUTH_RESPONSE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_RESPONSE_PTR", _wrap_new_TPMS_AUTH_RESPONSE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_AUTH_RESPONSE_PTR", _wrap_copy_TPMS_AUTH_RESPONSE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AUTH_RESPONSE_PTR", _wrap_delete_TPMS_AUTH_RESPONSE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_PTR_assign", _wrap_TPMS_AUTH_RESPONSE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_PTR_value", _wrap_TPMS_AUTH_RESPONSE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AUTH_RESPONSE_PTR_PTR", _wrap_new_TPMS_AUTH_RESPONSE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_AUTH_RESPONSE_PTR_PTR", _wrap_copy_TPMS_AUTH_RESPONSE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AUTH_RESPONSE_PTR_PTR", _wrap_delete_TPMS_AUTH_RESPONSE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_PTR_PTR_assign", _wrap_TPMS_AUTH_RESPONSE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_AUTH_RESPONSE_PTR_PTR_value", _wrap_TPMS_AUTH_RESPONSE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_SYM_KEY_BITS_swigconstant", sizeof_TPMU_SYM_KEY_BITS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_KEY_BITS_ARRAY", _wrap_new_TPMU_SYM_KEY_BITS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SYM_KEY_BITS_ARRAY", _wrap_delete_TPMU_SYM_KEY_BITS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_ARRAY___getitem__", _wrap_TPMU_SYM_KEY_BITS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_ARRAY___setitem__", _wrap_TPMU_SYM_KEY_BITS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_ARRAY_cast", _wrap_TPMU_SYM_KEY_BITS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_ARRAY_frompointer", _wrap_TPMU_SYM_KEY_BITS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_ARRAY_swigregister", TPMU_SYM_KEY_BITS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_KEY_BITS_PTR", _wrap_new_TPMU_SYM_KEY_BITS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SYM_KEY_BITS_PTR", _wrap_copy_TPMU_SYM_KEY_BITS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SYM_KEY_BITS_PTR", _wrap_delete_TPMU_SYM_KEY_BITS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_PTR_assign", _wrap_TPMU_SYM_KEY_BITS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_PTR_value", _wrap_TPMU_SYM_KEY_BITS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_KEY_BITS_PTR_PTR", _wrap_new_TPMU_SYM_KEY_BITS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SYM_KEY_BITS_PTR_PTR", _wrap_copy_TPMU_SYM_KEY_BITS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SYM_KEY_BITS_PTR_PTR", _wrap_delete_TPMU_SYM_KEY_BITS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_PTR_PTR_assign", _wrap_TPMU_SYM_KEY_BITS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_KEY_BITS_PTR_PTR_value", _wrap_TPMU_SYM_KEY_BITS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_SYM_MODE_swigconstant", sizeof_TPMU_SYM_MODE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_MODE_ARRAY", _wrap_new_TPMU_SYM_MODE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SYM_MODE_ARRAY", _wrap_delete_TPMU_SYM_MODE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_ARRAY___getitem__", _wrap_TPMU_SYM_MODE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_ARRAY___setitem__", _wrap_TPMU_SYM_MODE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_ARRAY_cast", _wrap_TPMU_SYM_MODE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_ARRAY_frompointer", _wrap_TPMU_SYM_MODE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_ARRAY_swigregister", TPMU_SYM_MODE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_MODE_PTR", _wrap_new_TPMU_SYM_MODE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SYM_MODE_PTR", _wrap_copy_TPMU_SYM_MODE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SYM_MODE_PTR", _wrap_delete_TPMU_SYM_MODE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_PTR_assign", _wrap_TPMU_SYM_MODE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_PTR_value", _wrap_TPMU_SYM_MODE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SYM_MODE_PTR_PTR", _wrap_new_TPMU_SYM_MODE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SYM_MODE_PTR_PTR", _wrap_copy_TPMU_SYM_MODE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SYM_MODE_PTR_PTR", _wrap_delete_TPMU_SYM_MODE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_PTR_PTR_assign", _wrap_TPMU_SYM_MODE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SYM_MODE_PTR_PTR_value", _wrap_TPMU_SYM_MODE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_SYM_DEF_swigconstant", sizeof_TPMT_SYM_DEF_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_ARRAY", _wrap_new_TPMT_SYM_DEF_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SYM_DEF_ARRAY", _wrap_delete_TPMT_SYM_DEF_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_ARRAY___getitem__", _wrap_TPMT_SYM_DEF_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_ARRAY___setitem__", _wrap_TPMT_SYM_DEF_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_ARRAY_cast", _wrap_TPMT_SYM_DEF_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_ARRAY_frompointer", _wrap_TPMT_SYM_DEF_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_ARRAY_swigregister", TPMT_SYM_DEF_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_PTR", _wrap_new_TPMT_SYM_DEF_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SYM_DEF_PTR", _wrap_copy_TPMT_SYM_DEF_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SYM_DEF_PTR", _wrap_delete_TPMT_SYM_DEF_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_PTR_assign", _wrap_TPMT_SYM_DEF_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_PTR_value", _wrap_TPMT_SYM_DEF_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_PTR_PTR", _wrap_new_TPMT_SYM_DEF_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SYM_DEF_PTR_PTR", _wrap_copy_TPMT_SYM_DEF_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SYM_DEF_PTR_PTR", _wrap_delete_TPMT_SYM_DEF_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_PTR_PTR_assign", _wrap_TPMT_SYM_DEF_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_PTR_PTR_value", _wrap_TPMT_SYM_DEF_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_SYM_DEF_OBJECT_swigconstant", sizeof_TPMT_SYM_DEF_OBJECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_OBJECT_ARRAY", _wrap_new_TPMT_SYM_DEF_OBJECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SYM_DEF_OBJECT_ARRAY", _wrap_delete_TPMT_SYM_DEF_OBJECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_ARRAY___getitem__", _wrap_TPMT_SYM_DEF_OBJECT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_ARRAY___setitem__", _wrap_TPMT_SYM_DEF_OBJECT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_ARRAY_cast", _wrap_TPMT_SYM_DEF_OBJECT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_ARRAY_frompointer", _wrap_TPMT_SYM_DEF_OBJECT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_ARRAY_swigregister", TPMT_SYM_DEF_OBJECT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_OBJECT_PTR", _wrap_new_TPMT_SYM_DEF_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SYM_DEF_OBJECT_PTR", _wrap_copy_TPMT_SYM_DEF_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SYM_DEF_OBJECT_PTR", _wrap_delete_TPMT_SYM_DEF_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_PTR_assign", _wrap_TPMT_SYM_DEF_OBJECT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_PTR_value", _wrap_TPMT_SYM_DEF_OBJECT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SYM_DEF_OBJECT_PTR_PTR", _wrap_new_TPMT_SYM_DEF_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SYM_DEF_OBJECT_PTR_PTR", _wrap_copy_TPMT_SYM_DEF_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SYM_DEF_OBJECT_PTR_PTR", _wrap_delete_TPMT_SYM_DEF_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_PTR_PTR_assign", _wrap_TPMT_SYM_DEF_OBJECT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SYM_DEF_OBJECT_PTR_PTR_value", _wrap_TPMT_SYM_DEF_OBJECT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_SYM_KEY_swigconstant", sizeof_TPM2B_SYM_KEY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SYM_KEY_ARRAY", _wrap_new_TPM2B_SYM_KEY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SYM_KEY_ARRAY", _wrap_delete_TPM2B_SYM_KEY_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_ARRAY___getitem__", _wrap_TPM2B_SYM_KEY_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_ARRAY___setitem__", _wrap_TPM2B_SYM_KEY_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_ARRAY_cast", _wrap_TPM2B_SYM_KEY_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_ARRAY_frompointer", _wrap_TPM2B_SYM_KEY_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_ARRAY_swigregister", TPM2B_SYM_KEY_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SYM_KEY_PTR", _wrap_new_TPM2B_SYM_KEY_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SYM_KEY_PTR", _wrap_copy_TPM2B_SYM_KEY_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SYM_KEY_PTR", _wrap_delete_TPM2B_SYM_KEY_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_PTR_assign", _wrap_TPM2B_SYM_KEY_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_PTR_value", _wrap_TPM2B_SYM_KEY_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SYM_KEY_PTR_PTR", _wrap_new_TPM2B_SYM_KEY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SYM_KEY_PTR_PTR", _wrap_copy_TPM2B_SYM_KEY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SYM_KEY_PTR_PTR", _wrap_delete_TPM2B_SYM_KEY_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_PTR_PTR_assign", _wrap_TPM2B_SYM_KEY_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SYM_KEY_PTR_PTR_value", _wrap_TPM2B_SYM_KEY_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SYMCIPHER_PARMS_swigconstant", sizeof_TPMS_SYMCIPHER_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SYMCIPHER_PARMS_ARRAY", _wrap_new_TPMS_SYMCIPHER_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SYMCIPHER_PARMS_ARRAY", _wrap_delete_TPMS_SYMCIPHER_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_ARRAY___getitem__", _wrap_TPMS_SYMCIPHER_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_ARRAY___setitem__", _wrap_TPMS_SYMCIPHER_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_ARRAY_cast", _wrap_TPMS_SYMCIPHER_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_ARRAY_frompointer", _wrap_TPMS_SYMCIPHER_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_ARRAY_swigregister", TPMS_SYMCIPHER_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SYMCIPHER_PARMS_PTR", _wrap_new_TPMS_SYMCIPHER_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SYMCIPHER_PARMS_PTR", _wrap_copy_TPMS_SYMCIPHER_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SYMCIPHER_PARMS_PTR", _wrap_delete_TPMS_SYMCIPHER_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_PTR_assign", _wrap_TPMS_SYMCIPHER_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_PTR_value", _wrap_TPMS_SYMCIPHER_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SYMCIPHER_PARMS_PTR_PTR", _wrap_new_TPMS_SYMCIPHER_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SYMCIPHER_PARMS_PTR_PTR", _wrap_copy_TPMS_SYMCIPHER_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SYMCIPHER_PARMS_PTR_PTR", _wrap_delete_TPMS_SYMCIPHER_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_PTR_PTR_assign", _wrap_TPMS_SYMCIPHER_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SYMCIPHER_PARMS_PTR_PTR_value", _wrap_TPMS_SYMCIPHER_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_SENSITIVE_DATA_swigconstant", sizeof_TPM2B_SENSITIVE_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_DATA_ARRAY", _wrap_new_TPM2B_SENSITIVE_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_DATA_ARRAY", _wrap_delete_TPM2B_SENSITIVE_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_ARRAY___getitem__", _wrap_TPM2B_SENSITIVE_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_ARRAY___setitem__", _wrap_TPM2B_SENSITIVE_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_ARRAY_cast", _wrap_TPM2B_SENSITIVE_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_ARRAY_frompointer", _wrap_TPM2B_SENSITIVE_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_ARRAY_swigregister", TPM2B_SENSITIVE_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_DATA_PTR", _wrap_new_TPM2B_SENSITIVE_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SENSITIVE_DATA_PTR", _wrap_copy_TPM2B_SENSITIVE_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_DATA_PTR", _wrap_delete_TPM2B_SENSITIVE_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_PTR_assign", _wrap_TPM2B_SENSITIVE_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_PTR_value", _wrap_TPM2B_SENSITIVE_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_DATA_PTR_PTR", _wrap_new_TPM2B_SENSITIVE_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SENSITIVE_DATA_PTR_PTR", _wrap_copy_TPM2B_SENSITIVE_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_DATA_PTR_PTR", _wrap_delete_TPM2B_SENSITIVE_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_PTR_PTR_assign", _wrap_TPM2B_SENSITIVE_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_DATA_PTR_PTR_value", _wrap_TPM2B_SENSITIVE_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SENSITIVE_CREATE_swigconstant", sizeof_TPMS_SENSITIVE_CREATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SENSITIVE_CREATE_ARRAY", _wrap_new_TPMS_SENSITIVE_CREATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SENSITIVE_CREATE_ARRAY", _wrap_delete_TPMS_SENSITIVE_CREATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_ARRAY___getitem__", _wrap_TPMS_SENSITIVE_CREATE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_ARRAY___setitem__", _wrap_TPMS_SENSITIVE_CREATE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_ARRAY_cast", _wrap_TPMS_SENSITIVE_CREATE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_ARRAY_frompointer", _wrap_TPMS_SENSITIVE_CREATE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_ARRAY_swigregister", TPMS_SENSITIVE_CREATE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SENSITIVE_CREATE_PTR", _wrap_new_TPMS_SENSITIVE_CREATE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SENSITIVE_CREATE_PTR", _wrap_copy_TPMS_SENSITIVE_CREATE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SENSITIVE_CREATE_PTR", _wrap_delete_TPMS_SENSITIVE_CREATE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_PTR_assign", _wrap_TPMS_SENSITIVE_CREATE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_PTR_value", _wrap_TPMS_SENSITIVE_CREATE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SENSITIVE_CREATE_PTR_PTR", _wrap_new_TPMS_SENSITIVE_CREATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SENSITIVE_CREATE_PTR_PTR", _wrap_copy_TPMS_SENSITIVE_CREATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SENSITIVE_CREATE_PTR_PTR", _wrap_delete_TPMS_SENSITIVE_CREATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_PTR_PTR_assign", _wrap_TPMS_SENSITIVE_CREATE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SENSITIVE_CREATE_PTR_PTR_value", _wrap_TPMS_SENSITIVE_CREATE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_SENSITIVE_CREATE_swigconstant", sizeof_TPM2B_SENSITIVE_CREATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_CREATE_ARRAY", _wrap_new_TPM2B_SENSITIVE_CREATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_CREATE_ARRAY", _wrap_delete_TPM2B_SENSITIVE_CREATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_ARRAY___getitem__", _wrap_TPM2B_SENSITIVE_CREATE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_ARRAY___setitem__", _wrap_TPM2B_SENSITIVE_CREATE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_ARRAY_cast", _wrap_TPM2B_SENSITIVE_CREATE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_ARRAY_frompointer", _wrap_TPM2B_SENSITIVE_CREATE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_ARRAY_swigregister", TPM2B_SENSITIVE_CREATE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_CREATE_PTR", _wrap_new_TPM2B_SENSITIVE_CREATE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SENSITIVE_CREATE_PTR", _wrap_copy_TPM2B_SENSITIVE_CREATE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_CREATE_PTR", _wrap_delete_TPM2B_SENSITIVE_CREATE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_PTR_assign", _wrap_TPM2B_SENSITIVE_CREATE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_PTR_value", _wrap_TPM2B_SENSITIVE_CREATE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_CREATE_PTR_PTR", _wrap_new_TPM2B_SENSITIVE_CREATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SENSITIVE_CREATE_PTR_PTR", _wrap_copy_TPM2B_SENSITIVE_CREATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_CREATE_PTR_PTR", _wrap_delete_TPM2B_SENSITIVE_CREATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_PTR_PTR_assign", _wrap_TPM2B_SENSITIVE_CREATE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_CREATE_PTR_PTR_value", _wrap_TPM2B_SENSITIVE_CREATE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SCHEME_HASH_swigconstant", sizeof_TPMS_SCHEME_HASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_HASH_ARRAY", _wrap_new_TPMS_SCHEME_HASH_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_HASH_ARRAY", _wrap_delete_TPMS_SCHEME_HASH_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_ARRAY___getitem__", _wrap_TPMS_SCHEME_HASH_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_ARRAY___setitem__", _wrap_TPMS_SCHEME_HASH_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_ARRAY_cast", _wrap_TPMS_SCHEME_HASH_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_ARRAY_frompointer", _wrap_TPMS_SCHEME_HASH_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_ARRAY_swigregister", TPMS_SCHEME_HASH_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_HASH_PTR", _wrap_new_TPMS_SCHEME_HASH_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SCHEME_HASH_PTR", _wrap_copy_TPMS_SCHEME_HASH_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_HASH_PTR", _wrap_delete_TPMS_SCHEME_HASH_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_PTR_assign", _wrap_TPMS_SCHEME_HASH_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_PTR_value", _wrap_TPMS_SCHEME_HASH_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_HASH_PTR_PTR", _wrap_new_TPMS_SCHEME_HASH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SCHEME_HASH_PTR_PTR", _wrap_copy_TPMS_SCHEME_HASH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_HASH_PTR_PTR", _wrap_delete_TPMS_SCHEME_HASH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_PTR_PTR_assign", _wrap_TPMS_SCHEME_HASH_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_HASH_PTR_PTR_value", _wrap_TPMS_SCHEME_HASH_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SCHEME_ECDAA_swigconstant", sizeof_TPMS_SCHEME_ECDAA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_ECDAA_ARRAY", _wrap_new_TPMS_SCHEME_ECDAA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_ECDAA_ARRAY", _wrap_delete_TPMS_SCHEME_ECDAA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_ARRAY___getitem__", _wrap_TPMS_SCHEME_ECDAA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_ARRAY___setitem__", _wrap_TPMS_SCHEME_ECDAA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_ARRAY_cast", _wrap_TPMS_SCHEME_ECDAA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_ARRAY_frompointer", _wrap_TPMS_SCHEME_ECDAA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_ARRAY_swigregister", TPMS_SCHEME_ECDAA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_ECDAA_PTR", _wrap_new_TPMS_SCHEME_ECDAA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SCHEME_ECDAA_PTR", _wrap_copy_TPMS_SCHEME_ECDAA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_ECDAA_PTR", _wrap_delete_TPMS_SCHEME_ECDAA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_PTR_assign", _wrap_TPMS_SCHEME_ECDAA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_PTR_value", _wrap_TPMS_SCHEME_ECDAA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_ECDAA_PTR_PTR", _wrap_new_TPMS_SCHEME_ECDAA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SCHEME_ECDAA_PTR_PTR", _wrap_copy_TPMS_SCHEME_ECDAA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_ECDAA_PTR_PTR", _wrap_delete_TPMS_SCHEME_ECDAA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_PTR_PTR_assign", _wrap_TPMS_SCHEME_ECDAA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_ECDAA_PTR_PTR_value", _wrap_TPMS_SCHEME_ECDAA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SCHEME_XOR_swigconstant", sizeof_TPMS_SCHEME_XOR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_XOR_ARRAY", _wrap_new_TPMS_SCHEME_XOR_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_XOR_ARRAY", _wrap_delete_TPMS_SCHEME_XOR_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_ARRAY___getitem__", _wrap_TPMS_SCHEME_XOR_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_ARRAY___setitem__", _wrap_TPMS_SCHEME_XOR_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_ARRAY_cast", _wrap_TPMS_SCHEME_XOR_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_ARRAY_frompointer", _wrap_TPMS_SCHEME_XOR_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_ARRAY_swigregister", TPMS_SCHEME_XOR_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_XOR_PTR", _wrap_new_TPMS_SCHEME_XOR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SCHEME_XOR_PTR", _wrap_copy_TPMS_SCHEME_XOR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_XOR_PTR", _wrap_delete_TPMS_SCHEME_XOR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_PTR_assign", _wrap_TPMS_SCHEME_XOR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_PTR_value", _wrap_TPMS_SCHEME_XOR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SCHEME_XOR_PTR_PTR", _wrap_new_TPMS_SCHEME_XOR_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SCHEME_XOR_PTR_PTR", _wrap_copy_TPMS_SCHEME_XOR_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SCHEME_XOR_PTR_PTR", _wrap_delete_TPMS_SCHEME_XOR_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_PTR_PTR_assign", _wrap_TPMS_SCHEME_XOR_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SCHEME_XOR_PTR_PTR_value", _wrap_TPMS_SCHEME_XOR_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_SCHEME_KEYEDHASH_swigconstant", sizeof_TPMU_SCHEME_KEYEDHASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SCHEME_KEYEDHASH_ARRAY", _wrap_new_TPMU_SCHEME_KEYEDHASH_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SCHEME_KEYEDHASH_ARRAY", _wrap_delete_TPMU_SCHEME_KEYEDHASH_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__", _wrap_TPMU_SCHEME_KEYEDHASH_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__", _wrap_TPMU_SCHEME_KEYEDHASH_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_ARRAY_cast", _wrap_TPMU_SCHEME_KEYEDHASH_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer", _wrap_TPMU_SCHEME_KEYEDHASH_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_ARRAY_swigregister", TPMU_SCHEME_KEYEDHASH_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SCHEME_KEYEDHASH_PTR", _wrap_new_TPMU_SCHEME_KEYEDHASH_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SCHEME_KEYEDHASH_PTR", _wrap_copy_TPMU_SCHEME_KEYEDHASH_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SCHEME_KEYEDHASH_PTR", _wrap_delete_TPMU_SCHEME_KEYEDHASH_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_PTR_assign", _wrap_TPMU_SCHEME_KEYEDHASH_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_PTR_value", _wrap_TPMU_SCHEME_KEYEDHASH_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SCHEME_KEYEDHASH_PTR_PTR", _wrap_new_TPMU_SCHEME_KEYEDHASH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR", _wrap_copy_TPMU_SCHEME_KEYEDHASH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR", _wrap_delete_TPMU_SCHEME_KEYEDHASH_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign", _wrap_TPMU_SCHEME_KEYEDHASH_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SCHEME_KEYEDHASH_PTR_PTR_value", _wrap_TPMU_SCHEME_KEYEDHASH_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_KEYEDHASH_SCHEME_swigconstant", sizeof_TPMT_KEYEDHASH_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KEYEDHASH_SCHEME_ARRAY", _wrap_new_TPMT_KEYEDHASH_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_KEYEDHASH_SCHEME_ARRAY", _wrap_delete_TPMT_KEYEDHASH_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__", _wrap_TPMT_KEYEDHASH_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__", _wrap_TPMT_KEYEDHASH_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_ARRAY_cast", _wrap_TPMT_KEYEDHASH_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer", _wrap_TPMT_KEYEDHASH_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_ARRAY_swigregister", TPMT_KEYEDHASH_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KEYEDHASH_SCHEME_PTR", _wrap_new_TPMT_KEYEDHASH_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_KEYEDHASH_SCHEME_PTR", _wrap_copy_TPMT_KEYEDHASH_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_KEYEDHASH_SCHEME_PTR", _wrap_delete_TPMT_KEYEDHASH_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_PTR_assign", _wrap_TPMT_KEYEDHASH_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_PTR_value", _wrap_TPMT_KEYEDHASH_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KEYEDHASH_SCHEME_PTR_PTR", _wrap_new_TPMT_KEYEDHASH_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR", _wrap_copy_TPMT_KEYEDHASH_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR", _wrap_delete_TPMT_KEYEDHASH_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign", _wrap_TPMT_KEYEDHASH_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_KEYEDHASH_SCHEME_PTR_PTR_value", _wrap_TPMT_KEYEDHASH_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_SIG_SCHEME_swigconstant", sizeof_TPMU_SIG_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIG_SCHEME_ARRAY", _wrap_new_TPMU_SIG_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SIG_SCHEME_ARRAY", _wrap_delete_TPMU_SIG_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ARRAY___getitem__", _wrap_TPMU_SIG_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ARRAY___setitem__", _wrap_TPMU_SIG_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ARRAY_cast", _wrap_TPMU_SIG_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ARRAY_frompointer", _wrap_TPMU_SIG_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_ARRAY_swigregister", TPMU_SIG_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIG_SCHEME_PTR", _wrap_new_TPMU_SIG_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SIG_SCHEME_PTR", _wrap_copy_TPMU_SIG_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SIG_SCHEME_PTR", _wrap_delete_TPMU_SIG_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_PTR_assign", _wrap_TPMU_SIG_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_PTR_value", _wrap_TPMU_SIG_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIG_SCHEME_PTR_PTR", _wrap_new_TPMU_SIG_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SIG_SCHEME_PTR_PTR", _wrap_copy_TPMU_SIG_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SIG_SCHEME_PTR_PTR", _wrap_delete_TPMU_SIG_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_PTR_PTR_assign", _wrap_TPMU_SIG_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIG_SCHEME_PTR_PTR_value", _wrap_TPMU_SIG_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_SIG_SCHEME_swigconstant", sizeof_TPMT_SIG_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIG_SCHEME_ARRAY", _wrap_new_TPMT_SIG_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SIG_SCHEME_ARRAY", _wrap_delete_TPMT_SIG_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_ARRAY___getitem__", _wrap_TPMT_SIG_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_ARRAY___setitem__", _wrap_TPMT_SIG_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_ARRAY_cast", _wrap_TPMT_SIG_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_ARRAY_frompointer", _wrap_TPMT_SIG_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_ARRAY_swigregister", TPMT_SIG_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIG_SCHEME_PTR", _wrap_new_TPMT_SIG_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SIG_SCHEME_PTR", _wrap_copy_TPMT_SIG_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SIG_SCHEME_PTR", _wrap_delete_TPMT_SIG_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_PTR_assign", _wrap_TPMT_SIG_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_PTR_value", _wrap_TPMT_SIG_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIG_SCHEME_PTR_PTR", _wrap_new_TPMT_SIG_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SIG_SCHEME_PTR_PTR", _wrap_copy_TPMT_SIG_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SIG_SCHEME_PTR_PTR", _wrap_delete_TPMT_SIG_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_PTR_PTR_assign", _wrap_TPMT_SIG_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIG_SCHEME_PTR_PTR_value", _wrap_TPMT_SIG_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_KDF_SCHEME_swigconstant", sizeof_TPMU_KDF_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_KDF_SCHEME_ARRAY", _wrap_new_TPMU_KDF_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_KDF_SCHEME_ARRAY", _wrap_delete_TPMU_KDF_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_ARRAY___getitem__", _wrap_TPMU_KDF_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_ARRAY___setitem__", _wrap_TPMU_KDF_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_ARRAY_cast", _wrap_TPMU_KDF_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_ARRAY_frompointer", _wrap_TPMU_KDF_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_ARRAY_swigregister", TPMU_KDF_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_KDF_SCHEME_PTR", _wrap_new_TPMU_KDF_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_KDF_SCHEME_PTR", _wrap_copy_TPMU_KDF_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_KDF_SCHEME_PTR", _wrap_delete_TPMU_KDF_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_PTR_assign", _wrap_TPMU_KDF_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_PTR_value", _wrap_TPMU_KDF_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_KDF_SCHEME_PTR_PTR", _wrap_new_TPMU_KDF_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_KDF_SCHEME_PTR_PTR", _wrap_copy_TPMU_KDF_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_KDF_SCHEME_PTR_PTR", _wrap_delete_TPMU_KDF_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_PTR_PTR_assign", _wrap_TPMU_KDF_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_KDF_SCHEME_PTR_PTR_value", _wrap_TPMU_KDF_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_KDF_SCHEME_swigconstant", sizeof_TPMT_KDF_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KDF_SCHEME_ARRAY", _wrap_new_TPMT_KDF_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_KDF_SCHEME_ARRAY", _wrap_delete_TPMT_KDF_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_ARRAY___getitem__", _wrap_TPMT_KDF_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_ARRAY___setitem__", _wrap_TPMT_KDF_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_ARRAY_cast", _wrap_TPMT_KDF_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_ARRAY_frompointer", _wrap_TPMT_KDF_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_ARRAY_swigregister", TPMT_KDF_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KDF_SCHEME_PTR", _wrap_new_TPMT_KDF_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_KDF_SCHEME_PTR", _wrap_copy_TPMT_KDF_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_KDF_SCHEME_PTR", _wrap_delete_TPMT_KDF_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_PTR_assign", _wrap_TPMT_KDF_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_PTR_value", _wrap_TPMT_KDF_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_KDF_SCHEME_PTR_PTR", _wrap_new_TPMT_KDF_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_KDF_SCHEME_PTR_PTR", _wrap_copy_TPMT_KDF_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_KDF_SCHEME_PTR_PTR", _wrap_delete_TPMT_KDF_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_PTR_PTR_assign", _wrap_TPMT_KDF_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_KDF_SCHEME_PTR_PTR_value", _wrap_TPMT_KDF_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_ASYM_SCHEME_swigconstant", sizeof_TPMU_ASYM_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ASYM_SCHEME_ARRAY", _wrap_new_TPMU_ASYM_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ASYM_SCHEME_ARRAY", _wrap_delete_TPMU_ASYM_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ARRAY___getitem__", _wrap_TPMU_ASYM_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ARRAY___setitem__", _wrap_TPMU_ASYM_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ARRAY_cast", _wrap_TPMU_ASYM_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ARRAY_frompointer", _wrap_TPMU_ASYM_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_ARRAY_swigregister", TPMU_ASYM_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ASYM_SCHEME_PTR", _wrap_new_TPMU_ASYM_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_ASYM_SCHEME_PTR", _wrap_copy_TPMU_ASYM_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ASYM_SCHEME_PTR", _wrap_delete_TPMU_ASYM_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_PTR_assign", _wrap_TPMU_ASYM_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_PTR_value", _wrap_TPMU_ASYM_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ASYM_SCHEME_PTR_PTR", _wrap_new_TPMU_ASYM_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_ASYM_SCHEME_PTR_PTR", _wrap_copy_TPMU_ASYM_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ASYM_SCHEME_PTR_PTR", _wrap_delete_TPMU_ASYM_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_PTR_PTR_assign", _wrap_TPMU_ASYM_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_ASYM_SCHEME_PTR_PTR_value", _wrap_TPMU_ASYM_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_ASYM_SCHEME_swigconstant", sizeof_TPMT_ASYM_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ASYM_SCHEME_ARRAY", _wrap_new_TPMT_ASYM_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_ASYM_SCHEME_ARRAY", _wrap_delete_TPMT_ASYM_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_ARRAY___getitem__", _wrap_TPMT_ASYM_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_ARRAY___setitem__", _wrap_TPMT_ASYM_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_ARRAY_cast", _wrap_TPMT_ASYM_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_ARRAY_frompointer", _wrap_TPMT_ASYM_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_ARRAY_swigregister", TPMT_ASYM_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ASYM_SCHEME_PTR", _wrap_new_TPMT_ASYM_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_ASYM_SCHEME_PTR", _wrap_copy_TPMT_ASYM_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_ASYM_SCHEME_PTR", _wrap_delete_TPMT_ASYM_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_PTR_assign", _wrap_TPMT_ASYM_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_PTR_value", _wrap_TPMT_ASYM_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ASYM_SCHEME_PTR_PTR", _wrap_new_TPMT_ASYM_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_ASYM_SCHEME_PTR_PTR", _wrap_copy_TPMT_ASYM_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_ASYM_SCHEME_PTR_PTR", _wrap_delete_TPMT_ASYM_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_PTR_PTR_assign", _wrap_TPMT_ASYM_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_ASYM_SCHEME_PTR_PTR_value", _wrap_TPMT_ASYM_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_RSA_SCHEME_swigconstant", sizeof_TPMT_RSA_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_SCHEME_ARRAY", _wrap_new_TPMT_RSA_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_RSA_SCHEME_ARRAY", _wrap_delete_TPMT_RSA_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_ARRAY___getitem__", _wrap_TPMT_RSA_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_ARRAY___setitem__", _wrap_TPMT_RSA_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_ARRAY_cast", _wrap_TPMT_RSA_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_ARRAY_frompointer", _wrap_TPMT_RSA_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_ARRAY_swigregister", TPMT_RSA_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_SCHEME_PTR", _wrap_new_TPMT_RSA_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_RSA_SCHEME_PTR", _wrap_copy_TPMT_RSA_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_RSA_SCHEME_PTR", _wrap_delete_TPMT_RSA_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_PTR_assign", _wrap_TPMT_RSA_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_PTR_value", _wrap_TPMT_RSA_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_SCHEME_PTR_PTR", _wrap_new_TPMT_RSA_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_RSA_SCHEME_PTR_PTR", _wrap_copy_TPMT_RSA_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_RSA_SCHEME_PTR_PTR", _wrap_delete_TPMT_RSA_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_PTR_PTR_assign", _wrap_TPMT_RSA_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_SCHEME_PTR_PTR_value", _wrap_TPMT_RSA_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_RSA_DECRYPT_swigconstant", sizeof_TPMT_RSA_DECRYPT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_DECRYPT_ARRAY", _wrap_new_TPMT_RSA_DECRYPT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_RSA_DECRYPT_ARRAY", _wrap_delete_TPMT_RSA_DECRYPT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_ARRAY___getitem__", _wrap_TPMT_RSA_DECRYPT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_ARRAY___setitem__", _wrap_TPMT_RSA_DECRYPT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_ARRAY_cast", _wrap_TPMT_RSA_DECRYPT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_ARRAY_frompointer", _wrap_TPMT_RSA_DECRYPT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_ARRAY_swigregister", TPMT_RSA_DECRYPT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_DECRYPT_PTR", _wrap_new_TPMT_RSA_DECRYPT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_RSA_DECRYPT_PTR", _wrap_copy_TPMT_RSA_DECRYPT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_RSA_DECRYPT_PTR", _wrap_delete_TPMT_RSA_DECRYPT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_PTR_assign", _wrap_TPMT_RSA_DECRYPT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_PTR_value", _wrap_TPMT_RSA_DECRYPT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_RSA_DECRYPT_PTR_PTR", _wrap_new_TPMT_RSA_DECRYPT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_RSA_DECRYPT_PTR_PTR", _wrap_copy_TPMT_RSA_DECRYPT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_RSA_DECRYPT_PTR_PTR", _wrap_delete_TPMT_RSA_DECRYPT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_PTR_PTR_assign", _wrap_TPMT_RSA_DECRYPT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_RSA_DECRYPT_PTR_PTR_value", _wrap_TPMT_RSA_DECRYPT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_PUBLIC_KEY_RSA_swigconstant", sizeof_TPM2B_PUBLIC_KEY_RSA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_KEY_RSA_ARRAY", _wrap_new_TPM2B_PUBLIC_KEY_RSA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PUBLIC_KEY_RSA_ARRAY", _wrap_delete_TPM2B_PUBLIC_KEY_RSA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__", _wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__", _wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_ARRAY_cast", _wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer", _wrap_TPM2B_PUBLIC_KEY_RSA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_ARRAY_swigregister", TPM2B_PUBLIC_KEY_RSA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_KEY_RSA_PTR", _wrap_new_TPM2B_PUBLIC_KEY_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PUBLIC_KEY_RSA_PTR", _wrap_copy_TPM2B_PUBLIC_KEY_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PUBLIC_KEY_RSA_PTR", _wrap_delete_TPM2B_PUBLIC_KEY_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_PTR_assign", _wrap_TPM2B_PUBLIC_KEY_RSA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_PTR_value", _wrap_TPM2B_PUBLIC_KEY_RSA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_KEY_RSA_PTR_PTR", _wrap_new_TPM2B_PUBLIC_KEY_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR", _wrap_copy_TPM2B_PUBLIC_KEY_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR", _wrap_delete_TPM2B_PUBLIC_KEY_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign", _wrap_TPM2B_PUBLIC_KEY_RSA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value", _wrap_TPM2B_PUBLIC_KEY_RSA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_PRIVATE_KEY_RSA_swigconstant", sizeof_TPM2B_PRIVATE_KEY_RSA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_KEY_RSA_ARRAY", _wrap_new_TPM2B_PRIVATE_KEY_RSA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_KEY_RSA_ARRAY", _wrap_delete_TPM2B_PRIVATE_KEY_RSA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__", _wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__", _wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_ARRAY_cast", _wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer", _wrap_TPM2B_PRIVATE_KEY_RSA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_ARRAY_swigregister", TPM2B_PRIVATE_KEY_RSA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_KEY_RSA_PTR", _wrap_new_TPM2B_PRIVATE_KEY_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PRIVATE_KEY_RSA_PTR", _wrap_copy_TPM2B_PRIVATE_KEY_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_KEY_RSA_PTR", _wrap_delete_TPM2B_PRIVATE_KEY_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_PTR_assign", _wrap_TPM2B_PRIVATE_KEY_RSA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_PTR_value", _wrap_TPM2B_PRIVATE_KEY_RSA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_KEY_RSA_PTR_PTR", _wrap_new_TPM2B_PRIVATE_KEY_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR", _wrap_copy_TPM2B_PRIVATE_KEY_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR", _wrap_delete_TPM2B_PRIVATE_KEY_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign", _wrap_TPM2B_PRIVATE_KEY_RSA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value", _wrap_TPM2B_PRIVATE_KEY_RSA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_ECC_PARAMETER_swigconstant", sizeof_TPM2B_ECC_PARAMETER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_PARAMETER_ARRAY", _wrap_new_TPM2B_ECC_PARAMETER_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ECC_PARAMETER_ARRAY", _wrap_delete_TPM2B_ECC_PARAMETER_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_ARRAY___getitem__", _wrap_TPM2B_ECC_PARAMETER_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_ARRAY___setitem__", _wrap_TPM2B_ECC_PARAMETER_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_ARRAY_cast", _wrap_TPM2B_ECC_PARAMETER_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_ARRAY_frompointer", _wrap_TPM2B_ECC_PARAMETER_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_ARRAY_swigregister", TPM2B_ECC_PARAMETER_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_PARAMETER_PTR", _wrap_new_TPM2B_ECC_PARAMETER_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ECC_PARAMETER_PTR", _wrap_copy_TPM2B_ECC_PARAMETER_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ECC_PARAMETER_PTR", _wrap_delete_TPM2B_ECC_PARAMETER_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_PTR_assign", _wrap_TPM2B_ECC_PARAMETER_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_PTR_value", _wrap_TPM2B_ECC_PARAMETER_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_PARAMETER_PTR_PTR", _wrap_new_TPM2B_ECC_PARAMETER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ECC_PARAMETER_PTR_PTR", _wrap_copy_TPM2B_ECC_PARAMETER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ECC_PARAMETER_PTR_PTR", _wrap_delete_TPM2B_ECC_PARAMETER_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_PTR_PTR_assign", _wrap_TPM2B_ECC_PARAMETER_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_PARAMETER_PTR_PTR_value", _wrap_TPM2B_ECC_PARAMETER_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ECC_POINT_swigconstant", sizeof_TPMS_ECC_POINT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_POINT_ARRAY", _wrap_new_TPMS_ECC_POINT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ECC_POINT_ARRAY", _wrap_delete_TPMS_ECC_POINT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_ARRAY___getitem__", _wrap_TPMS_ECC_POINT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_ARRAY___setitem__", _wrap_TPMS_ECC_POINT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_ARRAY_cast", _wrap_TPMS_ECC_POINT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_ARRAY_frompointer", _wrap_TPMS_ECC_POINT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_ARRAY_swigregister", TPMS_ECC_POINT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_POINT_PTR", _wrap_new_TPMS_ECC_POINT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ECC_POINT_PTR", _wrap_copy_TPMS_ECC_POINT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ECC_POINT_PTR", _wrap_delete_TPMS_ECC_POINT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_PTR_assign", _wrap_TPMS_ECC_POINT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_PTR_value", _wrap_TPMS_ECC_POINT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_POINT_PTR_PTR", _wrap_new_TPMS_ECC_POINT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ECC_POINT_PTR_PTR", _wrap_copy_TPMS_ECC_POINT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ECC_POINT_PTR_PTR", _wrap_delete_TPMS_ECC_POINT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_PTR_PTR_assign", _wrap_TPMS_ECC_POINT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_POINT_PTR_PTR_value", _wrap_TPMS_ECC_POINT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_ECC_POINT_swigconstant", sizeof_TPM2B_ECC_POINT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_POINT_ARRAY", _wrap_new_TPM2B_ECC_POINT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ECC_POINT_ARRAY", _wrap_delete_TPM2B_ECC_POINT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_ARRAY___getitem__", _wrap_TPM2B_ECC_POINT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_ARRAY___setitem__", _wrap_TPM2B_ECC_POINT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_ARRAY_cast", _wrap_TPM2B_ECC_POINT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_ARRAY_frompointer", _wrap_TPM2B_ECC_POINT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_ARRAY_swigregister", TPM2B_ECC_POINT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_POINT_PTR", _wrap_new_TPM2B_ECC_POINT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ECC_POINT_PTR", _wrap_copy_TPM2B_ECC_POINT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ECC_POINT_PTR", _wrap_delete_TPM2B_ECC_POINT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_PTR_assign", _wrap_TPM2B_ECC_POINT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_PTR_value", _wrap_TPM2B_ECC_POINT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ECC_POINT_PTR_PTR", _wrap_new_TPM2B_ECC_POINT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ECC_POINT_PTR_PTR", _wrap_copy_TPM2B_ECC_POINT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ECC_POINT_PTR_PTR", _wrap_delete_TPM2B_ECC_POINT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_PTR_PTR_assign", _wrap_TPM2B_ECC_POINT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ECC_POINT_PTR_PTR_value", _wrap_TPM2B_ECC_POINT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_ECC_SCHEME_swigconstant", sizeof_TPMT_ECC_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ECC_SCHEME_ARRAY", _wrap_new_TPMT_ECC_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_ECC_SCHEME_ARRAY", _wrap_delete_TPMT_ECC_SCHEME_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_ARRAY___getitem__", _wrap_TPMT_ECC_SCHEME_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_ARRAY___setitem__", _wrap_TPMT_ECC_SCHEME_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_ARRAY_cast", _wrap_TPMT_ECC_SCHEME_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_ARRAY_frompointer", _wrap_TPMT_ECC_SCHEME_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_ARRAY_swigregister", TPMT_ECC_SCHEME_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ECC_SCHEME_PTR", _wrap_new_TPMT_ECC_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_ECC_SCHEME_PTR", _wrap_copy_TPMT_ECC_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_ECC_SCHEME_PTR", _wrap_delete_TPMT_ECC_SCHEME_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_PTR_assign", _wrap_TPMT_ECC_SCHEME_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_PTR_value", _wrap_TPMT_ECC_SCHEME_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_ECC_SCHEME_PTR_PTR", _wrap_new_TPMT_ECC_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_ECC_SCHEME_PTR_PTR", _wrap_copy_TPMT_ECC_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_ECC_SCHEME_PTR_PTR", _wrap_delete_TPMT_ECC_SCHEME_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_PTR_PTR_assign", _wrap_TPMT_ECC_SCHEME_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_ECC_SCHEME_PTR_PTR_value", _wrap_TPMT_ECC_SCHEME_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ALGORITHM_DETAIL_ECC_swigconstant", sizeof_TPMS_ALGORITHM_DETAIL_ECC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY", _wrap_new_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY", _wrap_delete_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__", _wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__", _wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast", _wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer", _wrap_TPMS_ALGORITHM_DETAIL_ECC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_ARRAY_swigregister", TPMS_ALGORITHM_DETAIL_ECC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DETAIL_ECC_PTR", _wrap_new_TPMS_ALGORITHM_DETAIL_ECC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ALGORITHM_DETAIL_ECC_PTR", _wrap_copy_TPMS_ALGORITHM_DETAIL_ECC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALGORITHM_DETAIL_ECC_PTR", _wrap_delete_TPMS_ALGORITHM_DETAIL_ECC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_PTR_assign", _wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_PTR_value", _wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR", _wrap_new_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR", _wrap_copy_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR", _wrap_delete_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign", _wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value", _wrap_TPMS_ALGORITHM_DETAIL_ECC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SIGNATURE_RSA_swigconstant", sizeof_TPMS_SIGNATURE_RSA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_RSA_ARRAY", _wrap_new_TPMS_SIGNATURE_RSA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SIGNATURE_RSA_ARRAY", _wrap_delete_TPMS_SIGNATURE_RSA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_ARRAY___getitem__", _wrap_TPMS_SIGNATURE_RSA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_ARRAY___setitem__", _wrap_TPMS_SIGNATURE_RSA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_ARRAY_cast", _wrap_TPMS_SIGNATURE_RSA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_ARRAY_frompointer", _wrap_TPMS_SIGNATURE_RSA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_ARRAY_swigregister", TPMS_SIGNATURE_RSA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_RSA_PTR", _wrap_new_TPMS_SIGNATURE_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SIGNATURE_RSA_PTR", _wrap_copy_TPMS_SIGNATURE_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SIGNATURE_RSA_PTR", _wrap_delete_TPMS_SIGNATURE_RSA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_PTR_assign", _wrap_TPMS_SIGNATURE_RSA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_PTR_value", _wrap_TPMS_SIGNATURE_RSA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_RSA_PTR_PTR", _wrap_new_TPMS_SIGNATURE_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SIGNATURE_RSA_PTR_PTR", _wrap_copy_TPMS_SIGNATURE_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SIGNATURE_RSA_PTR_PTR", _wrap_delete_TPMS_SIGNATURE_RSA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_PTR_PTR_assign", _wrap_TPMS_SIGNATURE_RSA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_RSA_PTR_PTR_value", _wrap_TPMS_SIGNATURE_RSA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_SIGNATURE_ECC_swigconstant", sizeof_TPMS_SIGNATURE_ECC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_ECC_ARRAY", _wrap_new_TPMS_SIGNATURE_ECC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SIGNATURE_ECC_ARRAY", _wrap_delete_TPMS_SIGNATURE_ECC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_ARRAY___getitem__", _wrap_TPMS_SIGNATURE_ECC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_ARRAY___setitem__", _wrap_TPMS_SIGNATURE_ECC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_ARRAY_cast", _wrap_TPMS_SIGNATURE_ECC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_ARRAY_frompointer", _wrap_TPMS_SIGNATURE_ECC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_ARRAY_swigregister", TPMS_SIGNATURE_ECC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_ECC_PTR", _wrap_new_TPMS_SIGNATURE_ECC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SIGNATURE_ECC_PTR", _wrap_copy_TPMS_SIGNATURE_ECC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SIGNATURE_ECC_PTR", _wrap_delete_TPMS_SIGNATURE_ECC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_PTR_assign", _wrap_TPMS_SIGNATURE_ECC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_PTR_value", _wrap_TPMS_SIGNATURE_ECC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_SIGNATURE_ECC_PTR_PTR", _wrap_new_TPMS_SIGNATURE_ECC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_SIGNATURE_ECC_PTR_PTR", _wrap_copy_TPMS_SIGNATURE_ECC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_SIGNATURE_ECC_PTR_PTR", _wrap_delete_TPMS_SIGNATURE_ECC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_PTR_PTR_assign", _wrap_TPMS_SIGNATURE_ECC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_SIGNATURE_ECC_PTR_PTR_value", _wrap_TPMS_SIGNATURE_ECC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_SIGNATURE_swigconstant", sizeof_TPMU_SIGNATURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIGNATURE_ARRAY", _wrap_new_TPMU_SIGNATURE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SIGNATURE_ARRAY", _wrap_delete_TPMU_SIGNATURE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ARRAY___getitem__", _wrap_TPMU_SIGNATURE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ARRAY___setitem__", _wrap_TPMU_SIGNATURE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ARRAY_cast", _wrap_TPMU_SIGNATURE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ARRAY_frompointer", _wrap_TPMU_SIGNATURE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_ARRAY_swigregister", TPMU_SIGNATURE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIGNATURE_PTR", _wrap_new_TPMU_SIGNATURE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SIGNATURE_PTR", _wrap_copy_TPMU_SIGNATURE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SIGNATURE_PTR", _wrap_delete_TPMU_SIGNATURE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_PTR_assign", _wrap_TPMU_SIGNATURE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_PTR_value", _wrap_TPMU_SIGNATURE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SIGNATURE_PTR_PTR", _wrap_new_TPMU_SIGNATURE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SIGNATURE_PTR_PTR", _wrap_copy_TPMU_SIGNATURE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SIGNATURE_PTR_PTR", _wrap_delete_TPMU_SIGNATURE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_PTR_PTR_assign", _wrap_TPMU_SIGNATURE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SIGNATURE_PTR_PTR_value", _wrap_TPMU_SIGNATURE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_SIGNATURE_swigconstant", sizeof_TPMT_SIGNATURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIGNATURE_ARRAY", _wrap_new_TPMT_SIGNATURE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SIGNATURE_ARRAY", _wrap_delete_TPMT_SIGNATURE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_ARRAY___getitem__", _wrap_TPMT_SIGNATURE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_ARRAY___setitem__", _wrap_TPMT_SIGNATURE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_ARRAY_cast", _wrap_TPMT_SIGNATURE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_ARRAY_frompointer", _wrap_TPMT_SIGNATURE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_ARRAY_swigregister", TPMT_SIGNATURE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIGNATURE_PTR", _wrap_new_TPMT_SIGNATURE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SIGNATURE_PTR", _wrap_copy_TPMT_SIGNATURE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SIGNATURE_PTR", _wrap_delete_TPMT_SIGNATURE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_PTR_assign", _wrap_TPMT_SIGNATURE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_PTR_value", _wrap_TPMT_SIGNATURE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SIGNATURE_PTR_PTR", _wrap_new_TPMT_SIGNATURE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SIGNATURE_PTR_PTR", _wrap_copy_TPMT_SIGNATURE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SIGNATURE_PTR_PTR", _wrap_delete_TPMT_SIGNATURE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_PTR_PTR_assign", _wrap_TPMT_SIGNATURE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SIGNATURE_PTR_PTR_value", _wrap_TPMT_SIGNATURE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_ENCRYPTED_SECRET_swigconstant", sizeof_TPMU_ENCRYPTED_SECRET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ENCRYPTED_SECRET_ARRAY", _wrap_new_TPMU_ENCRYPTED_SECRET_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ENCRYPTED_SECRET_ARRAY", _wrap_delete_TPMU_ENCRYPTED_SECRET_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ARRAY___getitem__", _wrap_TPMU_ENCRYPTED_SECRET_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ARRAY___setitem__", _wrap_TPMU_ENCRYPTED_SECRET_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ARRAY_cast", _wrap_TPMU_ENCRYPTED_SECRET_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ARRAY_frompointer", _wrap_TPMU_ENCRYPTED_SECRET_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_ARRAY_swigregister", TPMU_ENCRYPTED_SECRET_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ENCRYPTED_SECRET_PTR", _wrap_new_TPMU_ENCRYPTED_SECRET_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_ENCRYPTED_SECRET_PTR", _wrap_copy_TPMU_ENCRYPTED_SECRET_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ENCRYPTED_SECRET_PTR", _wrap_delete_TPMU_ENCRYPTED_SECRET_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_PTR_assign", _wrap_TPMU_ENCRYPTED_SECRET_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_PTR_value", _wrap_TPMU_ENCRYPTED_SECRET_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_ENCRYPTED_SECRET_PTR_PTR", _wrap_new_TPMU_ENCRYPTED_SECRET_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_ENCRYPTED_SECRET_PTR_PTR", _wrap_copy_TPMU_ENCRYPTED_SECRET_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_ENCRYPTED_SECRET_PTR_PTR", _wrap_delete_TPMU_ENCRYPTED_SECRET_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_PTR_PTR_assign", _wrap_TPMU_ENCRYPTED_SECRET_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_ENCRYPTED_SECRET_PTR_PTR_value", _wrap_TPMU_ENCRYPTED_SECRET_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_ENCRYPTED_SECRET_swigconstant", sizeof_TPM2B_ENCRYPTED_SECRET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ENCRYPTED_SECRET_ARRAY", _wrap_new_TPM2B_ENCRYPTED_SECRET_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ENCRYPTED_SECRET_ARRAY", _wrap_delete_TPM2B_ENCRYPTED_SECRET_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__", _wrap_TPM2B_ENCRYPTED_SECRET_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__", _wrap_TPM2B_ENCRYPTED_SECRET_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_ARRAY_cast", _wrap_TPM2B_ENCRYPTED_SECRET_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer", _wrap_TPM2B_ENCRYPTED_SECRET_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_ARRAY_swigregister", TPM2B_ENCRYPTED_SECRET_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ENCRYPTED_SECRET_PTR", _wrap_new_TPM2B_ENCRYPTED_SECRET_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ENCRYPTED_SECRET_PTR", _wrap_copy_TPM2B_ENCRYPTED_SECRET_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ENCRYPTED_SECRET_PTR", _wrap_delete_TPM2B_ENCRYPTED_SECRET_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_PTR_assign", _wrap_TPM2B_ENCRYPTED_SECRET_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_PTR_value", _wrap_TPM2B_ENCRYPTED_SECRET_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ENCRYPTED_SECRET_PTR_PTR", _wrap_new_TPM2B_ENCRYPTED_SECRET_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR", _wrap_copy_TPM2B_ENCRYPTED_SECRET_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR", _wrap_delete_TPM2B_ENCRYPTED_SECRET_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign", _wrap_TPM2B_ENCRYPTED_SECRET_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ENCRYPTED_SECRET_PTR_PTR_value", _wrap_TPM2B_ENCRYPTED_SECRET_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_PUBLIC_ID_swigconstant", sizeof_TPMU_PUBLIC_ID_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_ID_ARRAY", _wrap_new_TPMU_PUBLIC_ID_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_PUBLIC_ID_ARRAY", _wrap_delete_TPMU_PUBLIC_ID_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ARRAY___getitem__", _wrap_TPMU_PUBLIC_ID_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ARRAY___setitem__", _wrap_TPMU_PUBLIC_ID_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ARRAY_cast", _wrap_TPMU_PUBLIC_ID_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ARRAY_frompointer", _wrap_TPMU_PUBLIC_ID_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_ARRAY_swigregister", TPMU_PUBLIC_ID_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_ID_PTR", _wrap_new_TPMU_PUBLIC_ID_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_PUBLIC_ID_PTR", _wrap_copy_TPMU_PUBLIC_ID_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_PUBLIC_ID_PTR", _wrap_delete_TPMU_PUBLIC_ID_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_PTR_assign", _wrap_TPMU_PUBLIC_ID_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_PTR_value", _wrap_TPMU_PUBLIC_ID_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_ID_PTR_PTR", _wrap_new_TPMU_PUBLIC_ID_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_PUBLIC_ID_PTR_PTR", _wrap_copy_TPMU_PUBLIC_ID_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_PUBLIC_ID_PTR_PTR", _wrap_delete_TPMU_PUBLIC_ID_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_PTR_PTR_assign", _wrap_TPMU_PUBLIC_ID_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_ID_PTR_PTR_value", _wrap_TPMU_PUBLIC_ID_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_KEYEDHASH_PARMS_swigconstant", sizeof_TPMS_KEYEDHASH_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_KEYEDHASH_PARMS_ARRAY", _wrap_new_TPMS_KEYEDHASH_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_KEYEDHASH_PARMS_ARRAY", _wrap_delete_TPMS_KEYEDHASH_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_ARRAY___getitem__", _wrap_TPMS_KEYEDHASH_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_ARRAY___setitem__", _wrap_TPMS_KEYEDHASH_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_ARRAY_cast", _wrap_TPMS_KEYEDHASH_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_ARRAY_frompointer", _wrap_TPMS_KEYEDHASH_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_ARRAY_swigregister", TPMS_KEYEDHASH_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_KEYEDHASH_PARMS_PTR", _wrap_new_TPMS_KEYEDHASH_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_KEYEDHASH_PARMS_PTR", _wrap_copy_TPMS_KEYEDHASH_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_KEYEDHASH_PARMS_PTR", _wrap_delete_TPMS_KEYEDHASH_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_PTR_assign", _wrap_TPMS_KEYEDHASH_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_PTR_value", _wrap_TPMS_KEYEDHASH_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_KEYEDHASH_PARMS_PTR_PTR", _wrap_new_TPMS_KEYEDHASH_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_KEYEDHASH_PARMS_PTR_PTR", _wrap_copy_TPMS_KEYEDHASH_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_KEYEDHASH_PARMS_PTR_PTR", _wrap_delete_TPMS_KEYEDHASH_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_PTR_PTR_assign", _wrap_TPMS_KEYEDHASH_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_KEYEDHASH_PARMS_PTR_PTR_value", _wrap_TPMS_KEYEDHASH_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ASYM_PARMS_swigconstant", sizeof_TPMS_ASYM_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ASYM_PARMS_ARRAY", _wrap_new_TPMS_ASYM_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ASYM_PARMS_ARRAY", _wrap_delete_TPMS_ASYM_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_ARRAY___getitem__", _wrap_TPMS_ASYM_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_ARRAY___setitem__", _wrap_TPMS_ASYM_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_ARRAY_cast", _wrap_TPMS_ASYM_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_ARRAY_frompointer", _wrap_TPMS_ASYM_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_ARRAY_swigregister", TPMS_ASYM_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ASYM_PARMS_PTR", _wrap_new_TPMS_ASYM_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ASYM_PARMS_PTR", _wrap_copy_TPMS_ASYM_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ASYM_PARMS_PTR", _wrap_delete_TPMS_ASYM_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_PTR_assign", _wrap_TPMS_ASYM_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_PTR_value", _wrap_TPMS_ASYM_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ASYM_PARMS_PTR_PTR", _wrap_new_TPMS_ASYM_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ASYM_PARMS_PTR_PTR", _wrap_copy_TPMS_ASYM_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ASYM_PARMS_PTR_PTR", _wrap_delete_TPMS_ASYM_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_PTR_PTR_assign", _wrap_TPMS_ASYM_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ASYM_PARMS_PTR_PTR_value", _wrap_TPMS_ASYM_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_RSA_PARMS_swigconstant", sizeof_TPMS_RSA_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_RSA_PARMS_ARRAY", _wrap_new_TPMS_RSA_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_RSA_PARMS_ARRAY", _wrap_delete_TPMS_RSA_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_ARRAY___getitem__", _wrap_TPMS_RSA_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_ARRAY___setitem__", _wrap_TPMS_RSA_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_ARRAY_cast", _wrap_TPMS_RSA_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_ARRAY_frompointer", _wrap_TPMS_RSA_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_ARRAY_swigregister", TPMS_RSA_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_RSA_PARMS_PTR", _wrap_new_TPMS_RSA_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_RSA_PARMS_PTR", _wrap_copy_TPMS_RSA_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_RSA_PARMS_PTR", _wrap_delete_TPMS_RSA_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_PTR_assign", _wrap_TPMS_RSA_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_PTR_value", _wrap_TPMS_RSA_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_RSA_PARMS_PTR_PTR", _wrap_new_TPMS_RSA_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_RSA_PARMS_PTR_PTR", _wrap_copy_TPMS_RSA_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_RSA_PARMS_PTR_PTR", _wrap_delete_TPMS_RSA_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_PTR_PTR_assign", _wrap_TPMS_RSA_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_RSA_PARMS_PTR_PTR_value", _wrap_TPMS_RSA_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ECC_PARMS_swigconstant", sizeof_TPMS_ECC_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_PARMS_ARRAY", _wrap_new_TPMS_ECC_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ECC_PARMS_ARRAY", _wrap_delete_TPMS_ECC_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_ARRAY___getitem__", _wrap_TPMS_ECC_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_ARRAY___setitem__", _wrap_TPMS_ECC_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_ARRAY_cast", _wrap_TPMS_ECC_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_ARRAY_frompointer", _wrap_TPMS_ECC_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_ARRAY_swigregister", TPMS_ECC_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_PARMS_PTR", _wrap_new_TPMS_ECC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ECC_PARMS_PTR", _wrap_copy_TPMS_ECC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ECC_PARMS_PTR", _wrap_delete_TPMS_ECC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_PTR_assign", _wrap_TPMS_ECC_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_PTR_value", _wrap_TPMS_ECC_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ECC_PARMS_PTR_PTR", _wrap_new_TPMS_ECC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ECC_PARMS_PTR_PTR", _wrap_copy_TPMS_ECC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ECC_PARMS_PTR_PTR", _wrap_delete_TPMS_ECC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_PTR_PTR_assign", _wrap_TPMS_ECC_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ECC_PARMS_PTR_PTR_value", _wrap_TPMS_ECC_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_PUBLIC_PARMS_swigconstant", sizeof_TPMU_PUBLIC_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_PARMS_ARRAY", _wrap_new_TPMU_PUBLIC_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_PUBLIC_PARMS_ARRAY", _wrap_delete_TPMU_PUBLIC_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_ARRAY___getitem__", _wrap_TPMU_PUBLIC_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_ARRAY___setitem__", _wrap_TPMU_PUBLIC_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_ARRAY_cast", _wrap_TPMU_PUBLIC_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_ARRAY_frompointer", _wrap_TPMU_PUBLIC_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_ARRAY_swigregister", TPMU_PUBLIC_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_PARMS_PTR", _wrap_new_TPMU_PUBLIC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_PUBLIC_PARMS_PTR", _wrap_copy_TPMU_PUBLIC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_PUBLIC_PARMS_PTR", _wrap_delete_TPMU_PUBLIC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_PTR_assign", _wrap_TPMU_PUBLIC_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_PTR_value", _wrap_TPMU_PUBLIC_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_PUBLIC_PARMS_PTR_PTR", _wrap_new_TPMU_PUBLIC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_PUBLIC_PARMS_PTR_PTR", _wrap_copy_TPMU_PUBLIC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_PUBLIC_PARMS_PTR_PTR", _wrap_delete_TPMU_PUBLIC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_PTR_PTR_assign", _wrap_TPMU_PUBLIC_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_PUBLIC_PARMS_PTR_PTR_value", _wrap_TPMU_PUBLIC_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_PUBLIC_PARMS_swigconstant", sizeof_TPMT_PUBLIC_PARMS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_PARMS_ARRAY", _wrap_new_TPMT_PUBLIC_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_PUBLIC_PARMS_ARRAY", _wrap_delete_TPMT_PUBLIC_PARMS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_ARRAY___getitem__", _wrap_TPMT_PUBLIC_PARMS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_ARRAY___setitem__", _wrap_TPMT_PUBLIC_PARMS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_ARRAY_cast", _wrap_TPMT_PUBLIC_PARMS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_ARRAY_frompointer", _wrap_TPMT_PUBLIC_PARMS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_ARRAY_swigregister", TPMT_PUBLIC_PARMS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_PARMS_PTR", _wrap_new_TPMT_PUBLIC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_PUBLIC_PARMS_PTR", _wrap_copy_TPMT_PUBLIC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_PUBLIC_PARMS_PTR", _wrap_delete_TPMT_PUBLIC_PARMS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_PTR_assign", _wrap_TPMT_PUBLIC_PARMS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_PTR_value", _wrap_TPMT_PUBLIC_PARMS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_PARMS_PTR_PTR", _wrap_new_TPMT_PUBLIC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_PUBLIC_PARMS_PTR_PTR", _wrap_copy_TPMT_PUBLIC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_PUBLIC_PARMS_PTR_PTR", _wrap_delete_TPMT_PUBLIC_PARMS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_PTR_PTR_assign", _wrap_TPMT_PUBLIC_PARMS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PARMS_PTR_PTR_value", _wrap_TPMT_PUBLIC_PARMS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_PUBLIC_swigconstant", sizeof_TPMT_PUBLIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_ARRAY", _wrap_new_TPMT_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_PUBLIC_ARRAY", _wrap_delete_TPMT_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_ARRAY___getitem__", _wrap_TPMT_PUBLIC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_ARRAY___setitem__", _wrap_TPMT_PUBLIC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_ARRAY_cast", _wrap_TPMT_PUBLIC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_ARRAY_frompointer", _wrap_TPMT_PUBLIC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_ARRAY_swigregister", TPMT_PUBLIC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_PTR", _wrap_new_TPMT_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_PUBLIC_PTR", _wrap_copy_TPMT_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_PUBLIC_PTR", _wrap_delete_TPMT_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PTR_assign", _wrap_TPMT_PUBLIC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PTR_value", _wrap_TPMT_PUBLIC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_PUBLIC_PTR_PTR", _wrap_new_TPMT_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_PUBLIC_PTR_PTR", _wrap_copy_TPMT_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_PUBLIC_PTR_PTR", _wrap_delete_TPMT_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PTR_PTR_assign", _wrap_TPMT_PUBLIC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_PUBLIC_PTR_PTR_value", _wrap_TPMT_PUBLIC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_PUBLIC_swigconstant", sizeof_TPM2B_PUBLIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_ARRAY", _wrap_new_TPM2B_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PUBLIC_ARRAY", _wrap_delete_TPM2B_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_ARRAY___getitem__", _wrap_TPM2B_PUBLIC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_ARRAY___setitem__", _wrap_TPM2B_PUBLIC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_ARRAY_cast", _wrap_TPM2B_PUBLIC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_ARRAY_frompointer", _wrap_TPM2B_PUBLIC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_ARRAY_swigregister", TPM2B_PUBLIC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_PTR", _wrap_new_TPM2B_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PUBLIC_PTR", _wrap_copy_TPM2B_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PUBLIC_PTR", _wrap_delete_TPM2B_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_PTR_assign", _wrap_TPM2B_PUBLIC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_PTR_value", _wrap_TPM2B_PUBLIC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PUBLIC_PTR_PTR", _wrap_new_TPM2B_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PUBLIC_PTR_PTR", _wrap_copy_TPM2B_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PUBLIC_PTR_PTR", _wrap_delete_TPM2B_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_PTR_PTR_assign", _wrap_TPM2B_PUBLIC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PUBLIC_PTR_PTR_value", _wrap_TPM2B_PUBLIC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_TEMPLATE_swigconstant", sizeof_TPM2B_TEMPLATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_TEMPLATE_ARRAY", _wrap_new_TPM2B_TEMPLATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_TEMPLATE_ARRAY", _wrap_delete_TPM2B_TEMPLATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_ARRAY___getitem__", _wrap_TPM2B_TEMPLATE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_ARRAY___setitem__", _wrap_TPM2B_TEMPLATE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_ARRAY_cast", _wrap_TPM2B_TEMPLATE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_ARRAY_frompointer", _wrap_TPM2B_TEMPLATE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_ARRAY_swigregister", TPM2B_TEMPLATE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_TEMPLATE_PTR", _wrap_new_TPM2B_TEMPLATE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_TEMPLATE_PTR", _wrap_copy_TPM2B_TEMPLATE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_TEMPLATE_PTR", _wrap_delete_TPM2B_TEMPLATE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_PTR_assign", _wrap_TPM2B_TEMPLATE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_PTR_value", _wrap_TPM2B_TEMPLATE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_TEMPLATE_PTR_PTR", _wrap_new_TPM2B_TEMPLATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_TEMPLATE_PTR_PTR", _wrap_copy_TPM2B_TEMPLATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_TEMPLATE_PTR_PTR", _wrap_delete_TPM2B_TEMPLATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_PTR_PTR_assign", _wrap_TPM2B_TEMPLATE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_TEMPLATE_PTR_PTR_value", _wrap_TPM2B_TEMPLATE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_PRIVATE_VENDOR_SPECIFIC_swigconstant", sizeof_TPM2B_PRIVATE_VENDOR_SPECIFIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY", _wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY", _wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_swigregister", TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR", _wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR", _wrap_copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR", _wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR", _wrap_new_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR", _wrap_copy_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR", _wrap_delete_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value", _wrap_TPM2B_PRIVATE_VENDOR_SPECIFIC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMU_SENSITIVE_COMPOSITE_swigconstant", sizeof_TPMU_SENSITIVE_COMPOSITE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SENSITIVE_COMPOSITE_ARRAY", _wrap_new_TPMU_SENSITIVE_COMPOSITE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SENSITIVE_COMPOSITE_ARRAY", _wrap_delete_TPMU_SENSITIVE_COMPOSITE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__", _wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__", _wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ARRAY_cast", _wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer", _wrap_TPMU_SENSITIVE_COMPOSITE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_ARRAY_swigregister", TPMU_SENSITIVE_COMPOSITE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SENSITIVE_COMPOSITE_PTR", _wrap_new_TPMU_SENSITIVE_COMPOSITE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SENSITIVE_COMPOSITE_PTR", _wrap_copy_TPMU_SENSITIVE_COMPOSITE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SENSITIVE_COMPOSITE_PTR", _wrap_delete_TPMU_SENSITIVE_COMPOSITE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_PTR_assign", _wrap_TPMU_SENSITIVE_COMPOSITE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_PTR_value", _wrap_TPMU_SENSITIVE_COMPOSITE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMU_SENSITIVE_COMPOSITE_PTR_PTR", _wrap_new_TPMU_SENSITIVE_COMPOSITE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR", _wrap_copy_TPMU_SENSITIVE_COMPOSITE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR", _wrap_delete_TPMU_SENSITIVE_COMPOSITE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign", _wrap_TPMU_SENSITIVE_COMPOSITE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value", _wrap_TPMU_SENSITIVE_COMPOSITE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMT_SENSITIVE_swigconstant", sizeof_TPMT_SENSITIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SENSITIVE_ARRAY", _wrap_new_TPMT_SENSITIVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SENSITIVE_ARRAY", _wrap_delete_TPMT_SENSITIVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_ARRAY___getitem__", _wrap_TPMT_SENSITIVE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_ARRAY___setitem__", _wrap_TPMT_SENSITIVE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_ARRAY_cast", _wrap_TPMT_SENSITIVE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_ARRAY_frompointer", _wrap_TPMT_SENSITIVE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_ARRAY_swigregister", TPMT_SENSITIVE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SENSITIVE_PTR", _wrap_new_TPMT_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SENSITIVE_PTR", _wrap_copy_TPMT_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SENSITIVE_PTR", _wrap_delete_TPMT_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_PTR_assign", _wrap_TPMT_SENSITIVE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_PTR_value", _wrap_TPMT_SENSITIVE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMT_SENSITIVE_PTR_PTR", _wrap_new_TPMT_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMT_SENSITIVE_PTR_PTR", _wrap_copy_TPMT_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMT_SENSITIVE_PTR_PTR", _wrap_delete_TPMT_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_PTR_PTR_assign", _wrap_TPMT_SENSITIVE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMT_SENSITIVE_PTR_PTR_value", _wrap_TPMT_SENSITIVE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_SENSITIVE_swigconstant", sizeof_TPM2B_SENSITIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_ARRAY", _wrap_new_TPM2B_SENSITIVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_ARRAY", _wrap_delete_TPM2B_SENSITIVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_ARRAY___getitem__", _wrap_TPM2B_SENSITIVE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_ARRAY___setitem__", _wrap_TPM2B_SENSITIVE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_ARRAY_cast", _wrap_TPM2B_SENSITIVE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_ARRAY_frompointer", _wrap_TPM2B_SENSITIVE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_ARRAY_swigregister", TPM2B_SENSITIVE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_PTR", _wrap_new_TPM2B_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SENSITIVE_PTR", _wrap_copy_TPM2B_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_PTR", _wrap_delete_TPM2B_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_PTR_assign", _wrap_TPM2B_SENSITIVE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_PTR_value", _wrap_TPM2B_SENSITIVE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_SENSITIVE_PTR_PTR", _wrap_new_TPM2B_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_SENSITIVE_PTR_PTR", _wrap_copy_TPM2B_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_SENSITIVE_PTR_PTR", _wrap_delete_TPM2B_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_PTR_PTR_assign", _wrap_TPM2B_SENSITIVE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_SENSITIVE_PTR_PTR_value", _wrap_TPM2B_SENSITIVE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof__PRIVATE_swigconstant", sizeof__PRIVATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new__PRIVATE_ARRAY", _wrap_new__PRIVATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete__PRIVATE_ARRAY", _wrap_delete__PRIVATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_ARRAY___getitem__", _wrap__PRIVATE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_ARRAY___setitem__", _wrap__PRIVATE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_ARRAY_cast", _wrap__PRIVATE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_ARRAY_frompointer", _wrap__PRIVATE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_ARRAY_swigregister", _PRIVATE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new__PRIVATE_PTR", _wrap_new__PRIVATE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy__PRIVATE_PTR", _wrap_copy__PRIVATE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete__PRIVATE_PTR", _wrap_delete__PRIVATE_PTR, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_PTR_assign", _wrap__PRIVATE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_PTR_value", _wrap__PRIVATE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new__PRIVATE_PTR_PTR", _wrap_new__PRIVATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy__PRIVATE_PTR_PTR", _wrap_copy__PRIVATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete__PRIVATE_PTR_PTR", _wrap_delete__PRIVATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_PTR_PTR_assign", _wrap__PRIVATE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"_PRIVATE_PTR_PTR_value", _wrap__PRIVATE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_PRIVATE_swigconstant", sizeof_TPM2B_PRIVATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_ARRAY", _wrap_new_TPM2B_PRIVATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_ARRAY", _wrap_delete_TPM2B_PRIVATE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_ARRAY___getitem__", _wrap_TPM2B_PRIVATE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_ARRAY___setitem__", _wrap_TPM2B_PRIVATE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_ARRAY_cast", _wrap_TPM2B_PRIVATE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_ARRAY_frompointer", _wrap_TPM2B_PRIVATE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_ARRAY_swigregister", TPM2B_PRIVATE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_PTR", _wrap_new_TPM2B_PRIVATE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PRIVATE_PTR", _wrap_copy_TPM2B_PRIVATE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_PTR", _wrap_delete_TPM2B_PRIVATE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_PTR_assign", _wrap_TPM2B_PRIVATE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_PTR_value", _wrap_TPM2B_PRIVATE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_PRIVATE_PTR_PTR", _wrap_new_TPM2B_PRIVATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_PRIVATE_PTR_PTR", _wrap_copy_TPM2B_PRIVATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_PRIVATE_PTR_PTR", _wrap_delete_TPM2B_PRIVATE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_PTR_PTR_assign", _wrap_TPM2B_PRIVATE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_PRIVATE_PTR_PTR_value", _wrap_TPM2B_PRIVATE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_ID_OBJECT_swigconstant", sizeof_TPMS_ID_OBJECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ID_OBJECT_ARRAY", _wrap_new_TPMS_ID_OBJECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ID_OBJECT_ARRAY", _wrap_delete_TPMS_ID_OBJECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_ARRAY___getitem__", _wrap_TPMS_ID_OBJECT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_ARRAY___setitem__", _wrap_TPMS_ID_OBJECT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_ARRAY_cast", _wrap_TPMS_ID_OBJECT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_ARRAY_frompointer", _wrap_TPMS_ID_OBJECT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_ARRAY_swigregister", TPMS_ID_OBJECT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ID_OBJECT_PTR", _wrap_new_TPMS_ID_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ID_OBJECT_PTR", _wrap_copy_TPMS_ID_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ID_OBJECT_PTR", _wrap_delete_TPMS_ID_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_PTR_assign", _wrap_TPMS_ID_OBJECT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_PTR_value", _wrap_TPMS_ID_OBJECT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_ID_OBJECT_PTR_PTR", _wrap_new_TPMS_ID_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_ID_OBJECT_PTR_PTR", _wrap_copy_TPMS_ID_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_ID_OBJECT_PTR_PTR", _wrap_delete_TPMS_ID_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_PTR_PTR_assign", _wrap_TPMS_ID_OBJECT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_ID_OBJECT_PTR_PTR_value", _wrap_TPMS_ID_OBJECT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_ID_OBJECT_swigconstant", sizeof_TPM2B_ID_OBJECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ID_OBJECT_ARRAY", _wrap_new_TPM2B_ID_OBJECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ID_OBJECT_ARRAY", _wrap_delete_TPM2B_ID_OBJECT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_ARRAY___getitem__", _wrap_TPM2B_ID_OBJECT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_ARRAY___setitem__", _wrap_TPM2B_ID_OBJECT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_ARRAY_cast", _wrap_TPM2B_ID_OBJECT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_ARRAY_frompointer", _wrap_TPM2B_ID_OBJECT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_ARRAY_swigregister", TPM2B_ID_OBJECT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ID_OBJECT_PTR", _wrap_new_TPM2B_ID_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ID_OBJECT_PTR", _wrap_copy_TPM2B_ID_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ID_OBJECT_PTR", _wrap_delete_TPM2B_ID_OBJECT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_PTR_assign", _wrap_TPM2B_ID_OBJECT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_PTR_value", _wrap_TPM2B_ID_OBJECT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_ID_OBJECT_PTR_PTR", _wrap_new_TPM2B_ID_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_ID_OBJECT_PTR_PTR", _wrap_copy_TPM2B_ID_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_ID_OBJECT_PTR_PTR", _wrap_delete_TPM2B_ID_OBJECT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_PTR_PTR_assign", _wrap_TPM2B_ID_OBJECT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_ID_OBJECT_PTR_PTR_value", _wrap_TPM2B_ID_OBJECT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_NV_PIN_COUNTER_PARAMETERS_swigconstant", sizeof_TPMS_NV_PIN_COUNTER_PARAMETERS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY", _wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY", _wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_swigregister", TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR", _wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR", _wrap_copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR", _wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR", _wrap_new_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR", _wrap_copy_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR", _wrap_delete_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value", _wrap_TPMS_NV_PIN_COUNTER_PARAMETERS_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_NV_PUBLIC_swigconstant", sizeof_TPMS_NV_PUBLIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PUBLIC_ARRAY", _wrap_new_TPMS_NV_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_PUBLIC_ARRAY", _wrap_delete_TPMS_NV_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_ARRAY___getitem__", _wrap_TPMS_NV_PUBLIC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_ARRAY___setitem__", _wrap_TPMS_NV_PUBLIC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_ARRAY_cast", _wrap_TPMS_NV_PUBLIC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_ARRAY_frompointer", _wrap_TPMS_NV_PUBLIC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_ARRAY_swigregister", TPMS_NV_PUBLIC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PUBLIC_PTR", _wrap_new_TPMS_NV_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_NV_PUBLIC_PTR", _wrap_copy_TPMS_NV_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_PUBLIC_PTR", _wrap_delete_TPMS_NV_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_PTR_assign", _wrap_TPMS_NV_PUBLIC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_PTR_value", _wrap_TPMS_NV_PUBLIC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_NV_PUBLIC_PTR_PTR", _wrap_new_TPMS_NV_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_NV_PUBLIC_PTR_PTR", _wrap_copy_TPMS_NV_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_NV_PUBLIC_PTR_PTR", _wrap_delete_TPMS_NV_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_PTR_PTR_assign", _wrap_TPMS_NV_PUBLIC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_NV_PUBLIC_PTR_PTR_value", _wrap_TPMS_NV_PUBLIC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_NV_PUBLIC_swigconstant", sizeof_TPM2B_NV_PUBLIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NV_PUBLIC_ARRAY", _wrap_new_TPM2B_NV_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_NV_PUBLIC_ARRAY", _wrap_delete_TPM2B_NV_PUBLIC_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_ARRAY___getitem__", _wrap_TPM2B_NV_PUBLIC_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_ARRAY___setitem__", _wrap_TPM2B_NV_PUBLIC_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_ARRAY_cast", _wrap_TPM2B_NV_PUBLIC_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_ARRAY_frompointer", _wrap_TPM2B_NV_PUBLIC_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_ARRAY_swigregister", TPM2B_NV_PUBLIC_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NV_PUBLIC_PTR", _wrap_new_TPM2B_NV_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_NV_PUBLIC_PTR", _wrap_copy_TPM2B_NV_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_NV_PUBLIC_PTR", _wrap_delete_TPM2B_NV_PUBLIC_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_PTR_assign", _wrap_TPM2B_NV_PUBLIC_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_PTR_value", _wrap_TPM2B_NV_PUBLIC_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_NV_PUBLIC_PTR_PTR", _wrap_new_TPM2B_NV_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_NV_PUBLIC_PTR_PTR", _wrap_copy_TPM2B_NV_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_NV_PUBLIC_PTR_PTR", _wrap_delete_TPM2B_NV_PUBLIC_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_PTR_PTR_assign", _wrap_TPM2B_NV_PUBLIC_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_NV_PUBLIC_PTR_PTR_value", _wrap_TPM2B_NV_PUBLIC_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_CONTEXT_SENSITIVE_swigconstant", sizeof_TPM2B_CONTEXT_SENSITIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_SENSITIVE_ARRAY", _wrap_new_TPM2B_CONTEXT_SENSITIVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CONTEXT_SENSITIVE_ARRAY", _wrap_delete_TPM2B_CONTEXT_SENSITIVE_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__", _wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__", _wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_ARRAY_cast", _wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer", _wrap_TPM2B_CONTEXT_SENSITIVE_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_ARRAY_swigregister", TPM2B_CONTEXT_SENSITIVE_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_SENSITIVE_PTR", _wrap_new_TPM2B_CONTEXT_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_CONTEXT_SENSITIVE_PTR", _wrap_copy_TPM2B_CONTEXT_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CONTEXT_SENSITIVE_PTR", _wrap_delete_TPM2B_CONTEXT_SENSITIVE_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_PTR_assign", _wrap_TPM2B_CONTEXT_SENSITIVE_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_PTR_value", _wrap_TPM2B_CONTEXT_SENSITIVE_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_SENSITIVE_PTR_PTR", _wrap_new_TPM2B_CONTEXT_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR", _wrap_copy_TPM2B_CONTEXT_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR", _wrap_delete_TPM2B_CONTEXT_SENSITIVE_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign", _wrap_TPM2B_CONTEXT_SENSITIVE_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value", _wrap_TPM2B_CONTEXT_SENSITIVE_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CONTEXT_DATA_swigconstant", sizeof_TPMS_CONTEXT_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_DATA_ARRAY", _wrap_new_TPMS_CONTEXT_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CONTEXT_DATA_ARRAY", _wrap_delete_TPMS_CONTEXT_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_ARRAY___getitem__", _wrap_TPMS_CONTEXT_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_ARRAY___setitem__", _wrap_TPMS_CONTEXT_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_ARRAY_cast", _wrap_TPMS_CONTEXT_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_ARRAY_frompointer", _wrap_TPMS_CONTEXT_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_ARRAY_swigregister", TPMS_CONTEXT_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_DATA_PTR", _wrap_new_TPMS_CONTEXT_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CONTEXT_DATA_PTR", _wrap_copy_TPMS_CONTEXT_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CONTEXT_DATA_PTR", _wrap_delete_TPMS_CONTEXT_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_PTR_assign", _wrap_TPMS_CONTEXT_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_PTR_value", _wrap_TPMS_CONTEXT_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_DATA_PTR_PTR", _wrap_new_TPMS_CONTEXT_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CONTEXT_DATA_PTR_PTR", _wrap_copy_TPMS_CONTEXT_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CONTEXT_DATA_PTR_PTR", _wrap_delete_TPMS_CONTEXT_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_PTR_PTR_assign", _wrap_TPMS_CONTEXT_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_DATA_PTR_PTR_value", _wrap_TPMS_CONTEXT_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_CONTEXT_DATA_swigconstant", sizeof_TPM2B_CONTEXT_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_DATA_ARRAY", _wrap_new_TPM2B_CONTEXT_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CONTEXT_DATA_ARRAY", _wrap_delete_TPM2B_CONTEXT_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_ARRAY___getitem__", _wrap_TPM2B_CONTEXT_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_ARRAY___setitem__", _wrap_TPM2B_CONTEXT_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_ARRAY_cast", _wrap_TPM2B_CONTEXT_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_ARRAY_frompointer", _wrap_TPM2B_CONTEXT_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_ARRAY_swigregister", TPM2B_CONTEXT_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_DATA_PTR", _wrap_new_TPM2B_CONTEXT_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_CONTEXT_DATA_PTR", _wrap_copy_TPM2B_CONTEXT_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CONTEXT_DATA_PTR", _wrap_delete_TPM2B_CONTEXT_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_PTR_assign", _wrap_TPM2B_CONTEXT_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_PTR_value", _wrap_TPM2B_CONTEXT_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CONTEXT_DATA_PTR_PTR", _wrap_new_TPM2B_CONTEXT_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_CONTEXT_DATA_PTR_PTR", _wrap_copy_TPM2B_CONTEXT_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CONTEXT_DATA_PTR_PTR", _wrap_delete_TPM2B_CONTEXT_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_PTR_PTR_assign", _wrap_TPM2B_CONTEXT_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CONTEXT_DATA_PTR_PTR_value", _wrap_TPM2B_CONTEXT_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CONTEXT_swigconstant", sizeof_TPMS_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_ARRAY", _wrap_new_TPMS_CONTEXT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CONTEXT_ARRAY", _wrap_delete_TPMS_CONTEXT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_ARRAY___getitem__", _wrap_TPMS_CONTEXT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_ARRAY___setitem__", _wrap_TPMS_CONTEXT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_ARRAY_cast", _wrap_TPMS_CONTEXT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_ARRAY_frompointer", _wrap_TPMS_CONTEXT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_ARRAY_swigregister", TPMS_CONTEXT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_PTR", _wrap_new_TPMS_CONTEXT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CONTEXT_PTR", _wrap_copy_TPMS_CONTEXT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CONTEXT_PTR", _wrap_delete_TPMS_CONTEXT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_PTR_assign", _wrap_TPMS_CONTEXT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_PTR_value", _wrap_TPMS_CONTEXT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CONTEXT_PTR_PTR", _wrap_new_TPMS_CONTEXT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CONTEXT_PTR_PTR", _wrap_copy_TPMS_CONTEXT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CONTEXT_PTR_PTR", _wrap_delete_TPMS_CONTEXT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_PTR_PTR_assign", _wrap_TPMS_CONTEXT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CONTEXT_PTR_PTR_value", _wrap_TPMS_CONTEXT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_CREATION_DATA_swigconstant", sizeof_TPMS_CREATION_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_DATA_ARRAY", _wrap_new_TPMS_CREATION_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CREATION_DATA_ARRAY", _wrap_delete_TPMS_CREATION_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_ARRAY___getitem__", _wrap_TPMS_CREATION_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_ARRAY___setitem__", _wrap_TPMS_CREATION_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_ARRAY_cast", _wrap_TPMS_CREATION_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_ARRAY_frompointer", _wrap_TPMS_CREATION_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_ARRAY_swigregister", TPMS_CREATION_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_DATA_PTR", _wrap_new_TPMS_CREATION_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CREATION_DATA_PTR", _wrap_copy_TPMS_CREATION_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CREATION_DATA_PTR", _wrap_delete_TPMS_CREATION_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_PTR_assign", _wrap_TPMS_CREATION_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_PTR_value", _wrap_TPMS_CREATION_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_CREATION_DATA_PTR_PTR", _wrap_new_TPMS_CREATION_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_CREATION_DATA_PTR_PTR", _wrap_copy_TPMS_CREATION_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_CREATION_DATA_PTR_PTR", _wrap_delete_TPMS_CREATION_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_PTR_PTR_assign", _wrap_TPMS_CREATION_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_CREATION_DATA_PTR_PTR_value", _wrap_TPMS_CREATION_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPM2B_CREATION_DATA_swigconstant", sizeof_TPM2B_CREATION_DATA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CREATION_DATA_ARRAY", _wrap_new_TPM2B_CREATION_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CREATION_DATA_ARRAY", _wrap_delete_TPM2B_CREATION_DATA_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_ARRAY___getitem__", _wrap_TPM2B_CREATION_DATA_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_ARRAY___setitem__", _wrap_TPM2B_CREATION_DATA_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_ARRAY_cast", _wrap_TPM2B_CREATION_DATA_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_ARRAY_frompointer", _wrap_TPM2B_CREATION_DATA_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_ARRAY_swigregister", TPM2B_CREATION_DATA_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CREATION_DATA_PTR", _wrap_new_TPM2B_CREATION_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_CREATION_DATA_PTR", _wrap_copy_TPM2B_CREATION_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CREATION_DATA_PTR", _wrap_delete_TPM2B_CREATION_DATA_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_PTR_assign", _wrap_TPM2B_CREATION_DATA_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_PTR_value", _wrap_TPM2B_CREATION_DATA_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPM2B_CREATION_DATA_PTR_PTR", _wrap_new_TPM2B_CREATION_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPM2B_CREATION_DATA_PTR_PTR", _wrap_copy_TPM2B_CREATION_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPM2B_CREATION_DATA_PTR_PTR", _wrap_delete_TPM2B_CREATION_DATA_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_PTR_PTR_assign", _wrap_TPM2B_CREATION_DATA_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPM2B_CREATION_DATA_PTR_PTR_value", _wrap_TPM2B_CREATION_DATA_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPMS_AC_OUTPUT_swigconstant", sizeof_TPMS_AC_OUTPUT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AC_OUTPUT_ARRAY", _wrap_new_TPMS_AC_OUTPUT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AC_OUTPUT_ARRAY", _wrap_delete_TPMS_AC_OUTPUT_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_ARRAY___getitem__", _wrap_TPMS_AC_OUTPUT_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_ARRAY___setitem__", _wrap_TPMS_AC_OUTPUT_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_ARRAY_cast", _wrap_TPMS_AC_OUTPUT_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_ARRAY_frompointer", _wrap_TPMS_AC_OUTPUT_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_ARRAY_swigregister", TPMS_AC_OUTPUT_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AC_OUTPUT_PTR", _wrap_new_TPMS_AC_OUTPUT_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_AC_OUTPUT_PTR", _wrap_copy_TPMS_AC_OUTPUT_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AC_OUTPUT_PTR", _wrap_delete_TPMS_AC_OUTPUT_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_PTR_assign", _wrap_TPMS_AC_OUTPUT_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_PTR_value", _wrap_TPMS_AC_OUTPUT_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPMS_AC_OUTPUT_PTR_PTR", _wrap_new_TPMS_AC_OUTPUT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPMS_AC_OUTPUT_PTR_PTR", _wrap_copy_TPMS_AC_OUTPUT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPMS_AC_OUTPUT_PTR_PTR", _wrap_delete_TPMS_AC_OUTPUT_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_PTR_PTR_assign", _wrap_TPMS_AC_OUTPUT_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPMS_AC_OUTPUT_PTR_PTR_value", _wrap_TPMS_AC_OUTPUT_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"sizeof_TPML_AC_CAPABILITIES_swigconstant", sizeof_TPML_AC_CAPABILITIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"new_TPML_AC_CAPABILITIES_ARRAY", _wrap_new_TPML_AC_CAPABILITIES_ARRAY, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_AC_CAPABILITIES_ARRAY", _wrap_delete_TPML_AC_CAPABILITIES_ARRAY, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_ARRAY___getitem__", _wrap_TPML_AC_CAPABILITIES_ARRAY___getitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_ARRAY___setitem__", _wrap_TPML_AC_CAPABILITIES_ARRAY___setitem__, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_ARRAY_cast", _wrap_TPML_AC_CAPABILITIES_ARRAY_cast, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_ARRAY_frompointer", _wrap_TPML_AC_CAPABILITIES_ARRAY_frompointer, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_ARRAY_swigregister", TPML_AC_CAPABILITIES_ARRAY_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TPML_AC_CAPABILITIES_PTR", _wrap_new_TPML_AC_CAPABILITIES_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_AC_CAPABILITIES_PTR", _wrap_copy_TPML_AC_CAPABILITIES_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_AC_CAPABILITIES_PTR", _wrap_delete_TPML_AC_CAPABILITIES_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_PTR_assign", _wrap_TPML_AC_CAPABILITIES_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_PTR_value", _wrap_TPML_AC_CAPABILITIES_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_TPML_AC_CAPABILITIES_PTR_PTR", _wrap_new_TPML_AC_CAPABILITIES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_TPML_AC_CAPABILITIES_PTR_PTR", _wrap_copy_TPML_AC_CAPABILITIES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_TPML_AC_CAPABILITIES_PTR_PTR", _wrap_delete_TPML_AC_CAPABILITIES_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_PTR_PTR_assign", _wrap_TPML_AC_CAPABILITIES_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"TPML_AC_CAPABILITIES_PTR_PTR_value", _wrap_TPML_AC_CAPABILITIES_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_VOID_PTR_PTR", _wrap_new_VOID_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_VOID_PTR_PTR", _wrap_copy_VOID_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_VOID_PTR_PTR", _wrap_delete_VOID_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"VOID_PTR_PTR_assign", _wrap_VOID_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"VOID_PTR_PTR_value", _wrap_VOID_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_SIZE_T_PTR", _wrap_new_SIZE_T_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_SIZE_T_PTR", _wrap_copy_SIZE_T_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_SIZE_T_PTR", _wrap_delete_SIZE_T_PTR, METH_VARARGS, NULL},
	 { (char *)"SIZE_T_PTR_assign", _wrap_SIZE_T_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"SIZE_T_PTR_value", _wrap_SIZE_T_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_SIZE_T_PTR_PTR", _wrap_new_SIZE_T_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_SIZE_T_PTR_PTR", _wrap_copy_SIZE_T_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_SIZE_T_PTR_PTR", _wrap_delete_SIZE_T_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"SIZE_T_PTR_PTR_assign", _wrap_SIZE_T_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"SIZE_T_PTR_PTR_value", _wrap_SIZE_T_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_CHAR_PTR", _wrap_new_CHAR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_CHAR_PTR", _wrap_copy_CHAR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_CHAR_PTR", _wrap_delete_CHAR_PTR, METH_VARARGS, NULL},
	 { (char *)"CHAR_PTR_assign", _wrap_CHAR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"CHAR_PTR_value", _wrap_CHAR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_CHAR_PTR_PTR", _wrap_new_CHAR_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_CHAR_PTR_PTR", _wrap_copy_CHAR_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_CHAR_PTR_PTR", _wrap_delete_CHAR_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"CHAR_PTR_PTR_assign", _wrap_CHAR_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"CHAR_PTR_PTR_value", _wrap_CHAR_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_UINT8_PTR", _wrap_new_UINT8_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_UINT8_PTR", _wrap_copy_UINT8_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_UINT8_PTR", _wrap_delete_UINT8_PTR, METH_VARARGS, NULL},
	 { (char *)"UINT8_PTR_assign", _wrap_UINT8_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"UINT8_PTR_value", _wrap_UINT8_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_UINT8_PTR_PTR", _wrap_new_UINT8_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_UINT8_PTR_PTR", _wrap_copy_UINT8_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_UINT8_PTR_PTR", _wrap_delete_UINT8_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"UINT8_PTR_PTR_assign", _wrap_UINT8_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"UINT8_PTR_PTR_value", _wrap_UINT8_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_uint32_t_PTR", _wrap_new_uint32_t_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_uint32_t_PTR", _wrap_delete_uint32_t_PTR, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_assign", _wrap_uint32_t_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_value", _wrap_uint32_t_PTR_value, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_cast", _wrap_uint32_t_PTR_cast, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_frompointer", _wrap_uint32_t_PTR_frompointer, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_swigregister", uint32_t_PTR_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_uint32_t_PTR_PTR", _wrap_new_uint32_t_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_uint32_t_PTR_PTR", _wrap_copy_uint32_t_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_uint32_t_PTR_PTR", _wrap_delete_uint32_t_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_PTR_assign", _wrap_uint32_t_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"uint32_t_PTR_PTR_value", _wrap_uint32_t_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_UINT32_PTR", _wrap_new_UINT32_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_UINT32_PTR", _wrap_delete_UINT32_PTR, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_assign", _wrap_UINT32_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_value", _wrap_UINT32_PTR_value, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_cast", _wrap_UINT32_PTR_cast, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_frompointer", _wrap_UINT32_PTR_frompointer, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_swigregister", UINT32_PTR_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_UINT32_PTR_PTR", _wrap_new_UINT32_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_UINT32_PTR_PTR", _wrap_copy_UINT32_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_UINT32_PTR_PTR", _wrap_delete_UINT32_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_PTR_assign", _wrap_UINT32_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"UINT32_PTR_PTR_value", _wrap_UINT32_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_uint16_t_PTR", _wrap_new_uint16_t_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_uint16_t_PTR", _wrap_delete_uint16_t_PTR, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_assign", _wrap_uint16_t_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_value", _wrap_uint16_t_PTR_value, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_cast", _wrap_uint16_t_PTR_cast, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_frompointer", _wrap_uint16_t_PTR_frompointer, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_swigregister", uint16_t_PTR_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_uint16_t_PTR_PTR", _wrap_new_uint16_t_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_uint16_t_PTR_PTR", _wrap_copy_uint16_t_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_uint16_t_PTR_PTR", _wrap_delete_uint16_t_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_PTR_assign", _wrap_uint16_t_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"uint16_t_PTR_PTR_value", _wrap_uint16_t_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_UINT16_PTR", _wrap_new_UINT16_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_UINT16_PTR", _wrap_delete_UINT16_PTR, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_assign", _wrap_UINT16_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_value", _wrap_UINT16_PTR_value, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_cast", _wrap_UINT16_PTR_cast, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_frompointer", _wrap_UINT16_PTR_frompointer, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_swigregister", UINT16_PTR_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_UINT16_PTR_PTR", _wrap_new_UINT16_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"copy_UINT16_PTR_PTR", _wrap_copy_UINT16_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"delete_UINT16_PTR_PTR", _wrap_delete_UINT16_PTR_PTR, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_PTR_assign", _wrap_UINT16_PTR_PTR_assign, METH_VARARGS, NULL},
	 { (char *)"UINT16_PTR_PTR_value", _wrap_UINT16_PTR_PTR_value, METH_VARARGS, NULL},
	 { (char *)"new_ByteArray", _wrap_new_ByteArray, METH_VARARGS, NULL},
	 { (char *)"delete_ByteArray", _wrap_delete_ByteArray, METH_VARARGS, NULL},
	 { (char *)"ByteArray___getitem__", _wrap_ByteArray___getitem__, METH_VARARGS, NULL},
	 { (char *)"ByteArray___setitem__", _wrap_ByteArray___setitem__, METH_VARARGS, NULL},
	 { (char *)"ByteArray_cast", _wrap_ByteArray_cast, METH_VARARGS, NULL},
	 { (char *)"ByteArray_frompointer", _wrap_ByteArray_frompointer, METH_VARARGS, NULL},
	 { (char *)"ByteArray_swigregister", ByteArray_swigregister, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_TIMEOUT_BLOCK_swigconstant", TSS2_TCTI_TIMEOUT_BLOCK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_TIMEOUT_NONE_swigconstant", TSS2_TCTI_TIMEOUT_NONE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_NONE_swigconstant", ESYS_TR_NONE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PASSWORD_swigconstant", ESYS_TR_PASSWORD_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR0_swigconstant", ESYS_TR_PCR0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR1_swigconstant", ESYS_TR_PCR1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR2_swigconstant", ESYS_TR_PCR2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR3_swigconstant", ESYS_TR_PCR3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR4_swigconstant", ESYS_TR_PCR4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR5_swigconstant", ESYS_TR_PCR5_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR6_swigconstant", ESYS_TR_PCR6_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR7_swigconstant", ESYS_TR_PCR7_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR8_swigconstant", ESYS_TR_PCR8_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR9_swigconstant", ESYS_TR_PCR9_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR10_swigconstant", ESYS_TR_PCR10_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR11_swigconstant", ESYS_TR_PCR11_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR12_swigconstant", ESYS_TR_PCR12_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR13_swigconstant", ESYS_TR_PCR13_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR14_swigconstant", ESYS_TR_PCR14_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR15_swigconstant", ESYS_TR_PCR15_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR16_swigconstant", ESYS_TR_PCR16_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR17_swigconstant", ESYS_TR_PCR17_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR18_swigconstant", ESYS_TR_PCR18_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR19_swigconstant", ESYS_TR_PCR19_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR20_swigconstant", ESYS_TR_PCR20_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR21_swigconstant", ESYS_TR_PCR21_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR22_swigconstant", ESYS_TR_PCR22_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR23_swigconstant", ESYS_TR_PCR23_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR24_swigconstant", ESYS_TR_PCR24_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR25_swigconstant", ESYS_TR_PCR25_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR26_swigconstant", ESYS_TR_PCR26_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR27_swigconstant", ESYS_TR_PCR27_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR28_swigconstant", ESYS_TR_PCR28_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR29_swigconstant", ESYS_TR_PCR29_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR30_swigconstant", ESYS_TR_PCR30_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_PCR31_swigconstant", ESYS_TR_PCR31_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_OWNER_swigconstant", ESYS_TR_RH_OWNER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_NULL_swigconstant", ESYS_TR_RH_NULL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_LOCKOUT_swigconstant", ESYS_TR_RH_LOCKOUT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_ENDORSEMENT_swigconstant", ESYS_TR_RH_ENDORSEMENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_PLATFORM_swigconstant", ESYS_TR_RH_PLATFORM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_PLATFORM_NV_swigconstant", ESYS_TR_RH_PLATFORM_NV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_RH_AUTH_FIRST_swigconstant", ESYS_TR_RH_AUTH_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"ESYS_TR_MIN_OBJECT_swigconstant", ESYS_TR_MIN_OBJECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_CAP_HANDLES_swigconstant", TPM2_MAX_CAP_HANDLES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_CAP_ALGS_swigconstant", TPM2_MAX_CAP_ALGS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_TPM_PROPERTIES_swigconstant", TPM2_MAX_TPM_PROPERTIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_PCR_PROPERTIES_swigconstant", TPM2_MAX_PCR_PROPERTIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_ECC_CURVES_swigconstant", TPM2_MAX_ECC_CURVES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_TAGGED_POLICIES_swigconstant", TPM2_MAX_TAGGED_POLICIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_PTT_PROPERTIES_swigconstant", TPM2_MAX_PTT_PROPERTIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_MAX_AC_CAPABILITIES_swigconstant", TPM2_MAX_AC_CAPABILITIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ERROR_swigconstant", TPM2_ALG_ERROR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_RSA_swigconstant", TPM2_ALG_RSA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA_swigconstant", TPM2_ALG_SHA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA1_swigconstant", TPM2_ALG_SHA1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_HMAC_swigconstant", TPM2_ALG_HMAC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_AES_swigconstant", TPM2_ALG_AES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_MGF1_swigconstant", TPM2_ALG_MGF1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_KEYEDHASH_swigconstant", TPM2_ALG_KEYEDHASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_XOR_swigconstant", TPM2_ALG_XOR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA256_swigconstant", TPM2_ALG_SHA256_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA384_swigconstant", TPM2_ALG_SHA384_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA512_swigconstant", TPM2_ALG_SHA512_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_NULL_swigconstant", TPM2_ALG_NULL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SM3_256_swigconstant", TPM2_ALG_SM3_256_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SM4_swigconstant", TPM2_ALG_SM4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_RSASSA_swigconstant", TPM2_ALG_RSASSA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_RSAES_swigconstant", TPM2_ALG_RSAES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_RSAPSS_swigconstant", TPM2_ALG_RSAPSS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_OAEP_swigconstant", TPM2_ALG_OAEP_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECDSA_swigconstant", TPM2_ALG_ECDSA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECDH_swigconstant", TPM2_ALG_ECDH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECDAA_swigconstant", TPM2_ALG_ECDAA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SM2_swigconstant", TPM2_ALG_SM2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECSCHNORR_swigconstant", TPM2_ALG_ECSCHNORR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECMQV_swigconstant", TPM2_ALG_ECMQV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_KDF1_SP800_56A_swigconstant", TPM2_ALG_KDF1_SP800_56A_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_KDF2_swigconstant", TPM2_ALG_KDF2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_KDF1_SP800_108_swigconstant", TPM2_ALG_KDF1_SP800_108_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECC_swigconstant", TPM2_ALG_ECC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SYMCIPHER_swigconstant", TPM2_ALG_SYMCIPHER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_CAMELLIA_swigconstant", TPM2_ALG_CAMELLIA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_CMAC_swigconstant", TPM2_ALG_CMAC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_CTR_swigconstant", TPM2_ALG_CTR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA3_256_swigconstant", TPM2_ALG_SHA3_256_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA3_384_swigconstant", TPM2_ALG_SHA3_384_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_SHA3_512_swigconstant", TPM2_ALG_SHA3_512_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_OFB_swigconstant", TPM2_ALG_OFB_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_CBC_swigconstant", TPM2_ALG_CBC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_CFB_swigconstant", TPM2_ALG_CFB_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_ECB_swigconstant", TPM2_ALG_ECB_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_FIRST_swigconstant", TPM2_ALG_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ALG_LAST_swigconstant", TPM2_ALG_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_NONE_swigconstant", TPM2_ECC_NONE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_NIST_P192_swigconstant", TPM2_ECC_NIST_P192_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_NIST_P224_swigconstant", TPM2_ECC_NIST_P224_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_NIST_P256_swigconstant", TPM2_ECC_NIST_P256_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_NIST_P384_swigconstant", TPM2_ECC_NIST_P384_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_NIST_P521_swigconstant", TPM2_ECC_NIST_P521_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_BN_P256_swigconstant", TPM2_ECC_BN_P256_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_BN_P638_swigconstant", TPM2_ECC_BN_P638_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ECC_SM2_P256_swigconstant", TPM2_ECC_SM2_P256_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SPEC_FAMILY_swigconstant", TPM2_SPEC_FAMILY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SPEC_LEVEL_swigconstant", TPM2_SPEC_LEVEL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SPEC_VERSION_swigconstant", TPM2_SPEC_VERSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SPEC_YEAR_swigconstant", TPM2_SPEC_YEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SPEC_DAY_OF_YEAR_swigconstant", TPM2_SPEC_DAY_OF_YEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_GENERATED_VALUE_swigconstant", TPM2_GENERATED_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SUCCESS_swigconstant", TPM2_RC_SUCCESS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_BAD_TAG_swigconstant", TPM2_RC_BAD_TAG_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_VER1_swigconstant", TPM2_RC_VER1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_INITIALIZE_swigconstant", TPM2_RC_INITIALIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_FAILURE_swigconstant", TPM2_RC_FAILURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SEQUENCE_swigconstant", TPM2_RC_SEQUENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_PRIVATE_swigconstant", TPM2_RC_PRIVATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_HMAC_swigconstant", TPM2_RC_HMAC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_DISABLED_swigconstant", TPM2_RC_DISABLED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_EXCLUSIVE_swigconstant", TPM2_RC_EXCLUSIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_AUTH_TYPE_swigconstant", TPM2_RC_AUTH_TYPE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_AUTH_MISSING_swigconstant", TPM2_RC_AUTH_MISSING_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_POLICY_swigconstant", TPM2_RC_POLICY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_PCR_swigconstant", TPM2_RC_PCR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_PCR_CHANGED_swigconstant", TPM2_RC_PCR_CHANGED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_UPGRADE_swigconstant", TPM2_RC_UPGRADE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_TOO_MANY_CONTEXTS_swigconstant", TPM2_RC_TOO_MANY_CONTEXTS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_AUTH_UNAVAILABLE_swigconstant", TPM2_RC_AUTH_UNAVAILABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REBOOT_swigconstant", TPM2_RC_REBOOT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_UNBALANCED_swigconstant", TPM2_RC_UNBALANCED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_COMMAND_SIZE_swigconstant", TPM2_RC_COMMAND_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_COMMAND_CODE_swigconstant", TPM2_RC_COMMAND_CODE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_AUTHSIZE_swigconstant", TPM2_RC_AUTHSIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_AUTH_CONTEXT_swigconstant", TPM2_RC_AUTH_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_RANGE_swigconstant", TPM2_RC_NV_RANGE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_SIZE_swigconstant", TPM2_RC_NV_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_LOCKED_swigconstant", TPM2_RC_NV_LOCKED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_AUTHORIZATION_swigconstant", TPM2_RC_NV_AUTHORIZATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_UNINITIALIZED_swigconstant", TPM2_RC_NV_UNINITIALIZED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_SPACE_swigconstant", TPM2_RC_NV_SPACE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_DEFINED_swigconstant", TPM2_RC_NV_DEFINED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_BAD_CONTEXT_swigconstant", TPM2_RC_BAD_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_CPHASH_swigconstant", TPM2_RC_CPHASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_PARENT_swigconstant", TPM2_RC_PARENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NEEDS_TEST_swigconstant", TPM2_RC_NEEDS_TEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NO_RESULT_swigconstant", TPM2_RC_NO_RESULT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SENSITIVE_swigconstant", TPM2_RC_SENSITIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_MAX_FM0_swigconstant", TPM2_RC_MAX_FM0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_FMT1_swigconstant", TPM2_RC_FMT1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_ASYMMETRIC_swigconstant", TPM2_RC_ASYMMETRIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_ATTRIBUTES_swigconstant", TPM2_RC_ATTRIBUTES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_HASH_swigconstant", TPM2_RC_HASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_VALUE_swigconstant", TPM2_RC_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_HIERARCHY_swigconstant", TPM2_RC_HIERARCHY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_KEY_SIZE_swigconstant", TPM2_RC_KEY_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_MGF_swigconstant", TPM2_RC_MGF_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_MODE_swigconstant", TPM2_RC_MODE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_TYPE_swigconstant", TPM2_RC_TYPE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_HANDLE_swigconstant", TPM2_RC_HANDLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_KDF_swigconstant", TPM2_RC_KDF_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_RANGE_swigconstant", TPM2_RC_RANGE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_AUTH_FAIL_swigconstant", TPM2_RC_AUTH_FAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NONCE_swigconstant", TPM2_RC_NONCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_PP_swigconstant", TPM2_RC_PP_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SCHEME_swigconstant", TPM2_RC_SCHEME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SIZE_swigconstant", TPM2_RC_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SYMMETRIC_swigconstant", TPM2_RC_SYMMETRIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_TAG_swigconstant", TPM2_RC_TAG_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SELECTOR_swigconstant", TPM2_RC_SELECTOR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_INSUFFICIENT_swigconstant", TPM2_RC_INSUFFICIENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SIGNATURE_swigconstant", TPM2_RC_SIGNATURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_KEY_swigconstant", TPM2_RC_KEY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_POLICY_FAIL_swigconstant", TPM2_RC_POLICY_FAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_INTEGRITY_swigconstant", TPM2_RC_INTEGRITY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_TICKET_swigconstant", TPM2_RC_TICKET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_RESERVED_BITS_swigconstant", TPM2_RC_RESERVED_BITS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_BAD_AUTH_swigconstant", TPM2_RC_BAD_AUTH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_EXPIRED_swigconstant", TPM2_RC_EXPIRED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_POLICY_CC_swigconstant", TPM2_RC_POLICY_CC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_BINDING_swigconstant", TPM2_RC_BINDING_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_CURVE_swigconstant", TPM2_RC_CURVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_ECC_POINT_swigconstant", TPM2_RC_ECC_POINT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_WARN_swigconstant", TPM2_RC_WARN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_CONTEXT_GAP_swigconstant", TPM2_RC_CONTEXT_GAP_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_OBJECT_MEMORY_swigconstant", TPM2_RC_OBJECT_MEMORY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SESSION_MEMORY_swigconstant", TPM2_RC_SESSION_MEMORY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_MEMORY_swigconstant", TPM2_RC_MEMORY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_SESSION_HANDLES_swigconstant", TPM2_RC_SESSION_HANDLES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_OBJECT_HANDLES_swigconstant", TPM2_RC_OBJECT_HANDLES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_LOCALITY_swigconstant", TPM2_RC_LOCALITY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_YIELDED_swigconstant", TPM2_RC_YIELDED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_CANCELED_swigconstant", TPM2_RC_CANCELED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_TESTING_swigconstant", TPM2_RC_TESTING_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H0_swigconstant", TPM2_RC_REFERENCE_H0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H1_swigconstant", TPM2_RC_REFERENCE_H1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H2_swigconstant", TPM2_RC_REFERENCE_H2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H3_swigconstant", TPM2_RC_REFERENCE_H3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H4_swigconstant", TPM2_RC_REFERENCE_H4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H5_swigconstant", TPM2_RC_REFERENCE_H5_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_H6_swigconstant", TPM2_RC_REFERENCE_H6_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S0_swigconstant", TPM2_RC_REFERENCE_S0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S1_swigconstant", TPM2_RC_REFERENCE_S1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S2_swigconstant", TPM2_RC_REFERENCE_S2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S3_swigconstant", TPM2_RC_REFERENCE_S3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S4_swigconstant", TPM2_RC_REFERENCE_S4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S5_swigconstant", TPM2_RC_REFERENCE_S5_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_REFERENCE_S6_swigconstant", TPM2_RC_REFERENCE_S6_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_RATE_swigconstant", TPM2_RC_NV_RATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_LOCKOUT_swigconstant", TPM2_RC_LOCKOUT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_RETRY_swigconstant", TPM2_RC_RETRY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NV_UNAVAILABLE_swigconstant", TPM2_RC_NV_UNAVAILABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_NOT_USED_swigconstant", TPM2_RC_NOT_USED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_H_swigconstant", TPM2_RC_H_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_P_swigconstant", TPM2_RC_P_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_S_swigconstant", TPM2_RC_S_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_1_swigconstant", TPM2_RC_1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_2_swigconstant", TPM2_RC_2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_3_swigconstant", TPM2_RC_3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_4_swigconstant", TPM2_RC_4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_5_swigconstant", TPM2_RC_5_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_6_swigconstant", TPM2_RC_6_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_7_swigconstant", TPM2_RC_7_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_8_swigconstant", TPM2_RC_8_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_9_swigconstant", TPM2_RC_9_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_A_swigconstant", TPM2_RC_A_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_B_swigconstant", TPM2_RC_B_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_C_swigconstant", TPM2_RC_C_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_D_swigconstant", TPM2_RC_D_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_E_swigconstant", TPM2_RC_E_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_F_swigconstant", TPM2_RC_F_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RC_N_MASK_swigconstant", TPM2_RC_N_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_COARSE_SLOWER_swigconstant", TPM2_CLOCK_COARSE_SLOWER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_MEDIUM_SLOWER_swigconstant", TPM2_CLOCK_MEDIUM_SLOWER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_FINE_SLOWER_swigconstant", TPM2_CLOCK_FINE_SLOWER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_NO_CHANGE_swigconstant", TPM2_CLOCK_NO_CHANGE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_FINE_FASTER_swigconstant", TPM2_CLOCK_FINE_FASTER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_MEDIUM_FASTER_swigconstant", TPM2_CLOCK_MEDIUM_FASTER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CLOCK_COARSE_FASTER_swigconstant", TPM2_CLOCK_COARSE_FASTER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_EQ_swigconstant", TPM2_EO_EQ_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_NEQ_swigconstant", TPM2_EO_NEQ_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_SIGNED_GT_swigconstant", TPM2_EO_SIGNED_GT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_UNSIGNED_GT_swigconstant", TPM2_EO_UNSIGNED_GT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_SIGNED_LT_swigconstant", TPM2_EO_SIGNED_LT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_UNSIGNED_LT_swigconstant", TPM2_EO_UNSIGNED_LT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_SIGNED_GE_swigconstant", TPM2_EO_SIGNED_GE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_UNSIGNED_GE_swigconstant", TPM2_EO_UNSIGNED_GE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_SIGNED_LE_swigconstant", TPM2_EO_SIGNED_LE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_UNSIGNED_LE_swigconstant", TPM2_EO_UNSIGNED_LE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_BITSET_swigconstant", TPM2_EO_BITSET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_EO_BITCLEAR_swigconstant", TPM2_EO_BITCLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_RSP_COMMAND_swigconstant", TPM2_ST_RSP_COMMAND_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_NULL_swigconstant", TPM2_ST_NULL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_NO_SESSIONS_swigconstant", TPM2_ST_NO_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_SESSIONS_swigconstant", TPM2_ST_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_RESERVED1_swigconstant", TPM2_ST_RESERVED1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_RESERVED2_swigconstant", TPM2_ST_RESERVED2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_NV_swigconstant", TPM2_ST_ATTEST_NV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_COMMAND_AUDIT_swigconstant", TPM2_ST_ATTEST_COMMAND_AUDIT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_SESSION_AUDIT_swigconstant", TPM2_ST_ATTEST_SESSION_AUDIT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_CERTIFY_swigconstant", TPM2_ST_ATTEST_CERTIFY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_QUOTE_swigconstant", TPM2_ST_ATTEST_QUOTE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_TIME_swigconstant", TPM2_ST_ATTEST_TIME_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_ATTEST_CREATION_swigconstant", TPM2_ST_ATTEST_CREATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_RESERVED3_swigconstant", TPM2_ST_RESERVED3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_CREATION_swigconstant", TPM2_ST_CREATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_VERIFIED_swigconstant", TPM2_ST_VERIFIED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_AUTH_SECRET_swigconstant", TPM2_ST_AUTH_SECRET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_HASHCHECK_swigconstant", TPM2_ST_HASHCHECK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_AUTH_SIGNED_swigconstant", TPM2_ST_AUTH_SIGNED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ST_FU_MANIFEST_swigconstant", TPM2_ST_FU_MANIFEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SU_CLEAR_swigconstant", TPM2_SU_CLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SU_STATE_swigconstant", TPM2_SU_STATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SE_HMAC_swigconstant", TPM2_SE_HMAC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SE_POLICY_swigconstant", TPM2_SE_POLICY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_SE_TRIAL_swigconstant", TPM2_SE_TRIAL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_FIRST_swigconstant", TPM2_CAP_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_ALGS_swigconstant", TPM2_CAP_ALGS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_HANDLES_swigconstant", TPM2_CAP_HANDLES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_COMMANDS_swigconstant", TPM2_CAP_COMMANDS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_PP_COMMANDS_swigconstant", TPM2_CAP_PP_COMMANDS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_AUDIT_COMMANDS_swigconstant", TPM2_CAP_AUDIT_COMMANDS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_PCRS_swigconstant", TPM2_CAP_PCRS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_TPM_PROPERTIES_swigconstant", TPM2_CAP_TPM_PROPERTIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_PCR_PROPERTIES_swigconstant", TPM2_CAP_PCR_PROPERTIES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_ECC_CURVES_swigconstant", TPM2_CAP_ECC_CURVES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_LAST_swigconstant", TPM2_CAP_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_CAP_VENDOR_PROPERTY_swigconstant", TPM2_CAP_VENDOR_PROPERTY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NONE_swigconstant", TPM2_PT_NONE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_GROUP_swigconstant", TPM2_PT_GROUP_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_FIXED_swigconstant", TPM2_PT_FIXED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_FAMILY_INDICATOR_swigconstant", TPM2_PT_FAMILY_INDICATOR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_LEVEL_swigconstant", TPM2_PT_LEVEL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_REVISION_swigconstant", TPM2_PT_REVISION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_DAY_OF_YEAR_swigconstant", TPM2_PT_DAY_OF_YEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_YEAR_swigconstant", TPM2_PT_YEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MANUFACTURER_swigconstant", TPM2_PT_MANUFACTURER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VENDOR_STRING_1_swigconstant", TPM2_PT_VENDOR_STRING_1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VENDOR_STRING_2_swigconstant", TPM2_PT_VENDOR_STRING_2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VENDOR_STRING_3_swigconstant", TPM2_PT_VENDOR_STRING_3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VENDOR_STRING_4_swigconstant", TPM2_PT_VENDOR_STRING_4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VENDOR_TPM_TYPE_swigconstant", TPM2_PT_VENDOR_TPM_TYPE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_FIRMWARE_VERSION_1_swigconstant", TPM2_PT_FIRMWARE_VERSION_1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_FIRMWARE_VERSION_2_swigconstant", TPM2_PT_FIRMWARE_VERSION_2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_INPUT_BUFFER_swigconstant", TPM2_PT_INPUT_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_HR_TRANSIENT_MIN_swigconstant", TPM2_PT_TPM2_HR_TRANSIENT_MIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_HR_PERSISTENT_MIN_swigconstant", TPM2_PT_TPM2_HR_PERSISTENT_MIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_HR_LOADED_MIN_swigconstant", TPM2_PT_HR_LOADED_MIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_ACTIVE_SESSIONS_MAX_swigconstant", TPM2_PT_ACTIVE_SESSIONS_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_COUNT_swigconstant", TPM2_PT_PCR_COUNT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_SELECT_MIN_swigconstant", TPM2_PT_PCR_SELECT_MIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_CONTEXT_GAP_MAX_swigconstant", TPM2_PT_CONTEXT_GAP_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NV_COUNTERS_MAX_swigconstant", TPM2_PT_NV_COUNTERS_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NV_INDEX_MAX_swigconstant", TPM2_PT_NV_INDEX_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MEMORY_swigconstant", TPM2_PT_MEMORY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_CLOCK_UPDATE_swigconstant", TPM2_PT_CLOCK_UPDATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_CONTEXT_HASH_swigconstant", TPM2_PT_CONTEXT_HASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_CONTEXT_SYM_swigconstant", TPM2_PT_CONTEXT_SYM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_CONTEXT_SYM_SIZE_swigconstant", TPM2_PT_CONTEXT_SYM_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_ORDERLY_COUNT_swigconstant", TPM2_PT_ORDERLY_COUNT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MAX_COMMAND_SIZE_swigconstant", TPM2_PT_MAX_COMMAND_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MAX_RESPONSE_SIZE_swigconstant", TPM2_PT_MAX_RESPONSE_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MAX_DIGEST_swigconstant", TPM2_PT_MAX_DIGEST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MAX_OBJECT_CONTEXT_swigconstant", TPM2_PT_MAX_OBJECT_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MAX_SESSION_CONTEXT_swigconstant", TPM2_PT_MAX_SESSION_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PS_FAMILY_INDICATOR_swigconstant", TPM2_PT_PS_FAMILY_INDICATOR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PS_LEVEL_swigconstant", TPM2_PT_PS_LEVEL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PS_REVISION_swigconstant", TPM2_PT_PS_REVISION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PS_DAY_OF_YEAR_swigconstant", TPM2_PT_PS_DAY_OF_YEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PS_YEAR_swigconstant", TPM2_PT_PS_YEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_SPLIT_MAX_swigconstant", TPM2_PT_SPLIT_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TOTAL_COMMANDS_swigconstant", TPM2_PT_TOTAL_COMMANDS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_LIBRARY_COMMANDS_swigconstant", TPM2_PT_LIBRARY_COMMANDS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VENDOR_COMMANDS_swigconstant", TPM2_PT_VENDOR_COMMANDS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NV_BUFFER_MAX_swigconstant", TPM2_PT_NV_BUFFER_MAX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MODES_swigconstant", TPM2_PT_MODES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_VAR_swigconstant", TPM2_PT_VAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PERMANENT_swigconstant", TPM2_PT_PERMANENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_STARTUP_CLEAR_swigconstant", TPM2_PT_STARTUP_CLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_HR_NV_INDEX_swigconstant", TPM2_PT_TPM2_HR_NV_INDEX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_HR_LOADED_swigconstant", TPM2_PT_HR_LOADED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_HR_LOADED_AVAIL_swigconstant", TPM2_PT_HR_LOADED_AVAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_HR_ACTIVE_swigconstant", TPM2_PT_HR_ACTIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_HR_ACTIVE_AVAIL_swigconstant", TPM2_PT_HR_ACTIVE_AVAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_HR_TRANSIENT_AVAIL_swigconstant", TPM2_PT_TPM2_HR_TRANSIENT_AVAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_HR_PERSISTENT_swigconstant", TPM2_PT_TPM2_HR_PERSISTENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_HR_PERSISTENT_AVAIL_swigconstant", TPM2_PT_TPM2_HR_PERSISTENT_AVAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NV_COUNTERS_swigconstant", TPM2_PT_NV_COUNTERS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NV_COUNTERS_AVAIL_swigconstant", TPM2_PT_NV_COUNTERS_AVAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_ALGORITHM_SET_swigconstant", TPM2_PT_ALGORITHM_SET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_LOADED_CURVES_swigconstant", TPM2_PT_LOADED_CURVES_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_LOCKOUT_COUNTER_swigconstant", TPM2_PT_LOCKOUT_COUNTER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_MAX_AUTH_FAIL_swigconstant", TPM2_PT_MAX_AUTH_FAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_LOCKOUT_INTERVAL_swigconstant", TPM2_PT_LOCKOUT_INTERVAL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_LOCKOUT_RECOVERY_swigconstant", TPM2_PT_LOCKOUT_RECOVERY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_NV_WRITE_RECOVERY_swigconstant", TPM2_PT_NV_WRITE_RECOVERY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_AUDIT_COUNTER_0_swigconstant", TPM2_PT_AUDIT_COUNTER_0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_AUDIT_COUNTER_1_swigconstant", TPM2_PT_AUDIT_COUNTER_1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_PCR_FIRST_swigconstant", TPM2_PT_TPM2_PCR_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_SAVE_swigconstant", TPM2_PT_PCR_SAVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_EXTEND_L0_swigconstant", TPM2_PT_PCR_EXTEND_L0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_RESET_L0_swigconstant", TPM2_PT_PCR_RESET_L0_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_EXTEND_L1_swigconstant", TPM2_PT_PCR_EXTEND_L1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_RESET_L1_swigconstant", TPM2_PT_PCR_RESET_L1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_EXTEND_L2_swigconstant", TPM2_PT_PCR_EXTEND_L2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_RESET_L2_swigconstant", TPM2_PT_PCR_RESET_L2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_EXTEND_L3_swigconstant", TPM2_PT_PCR_EXTEND_L3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_RESET_L3_swigconstant", TPM2_PT_PCR_RESET_L3_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_EXTEND_L4_swigconstant", TPM2_PT_PCR_EXTEND_L4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_RESET_L4_swigconstant", TPM2_PT_PCR_RESET_L4_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_NO_INCREMENT_swigconstant", TPM2_PT_PCR_NO_INCREMENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_DRTM_RESET_swigconstant", TPM2_PT_PCR_DRTM_RESET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_POLICY_swigconstant", TPM2_PT_PCR_POLICY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_PCR_AUTH_swigconstant", TPM2_PT_PCR_AUTH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PT_TPM2_PCR_LAST_swigconstant", TPM2_PT_TPM2_PCR_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_MAIN_swigconstant", TPM2_PS_MAIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_PC_swigconstant", TPM2_PS_PC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_PDA_swigconstant", TPM2_PS_PDA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_CELL_PHONE_swigconstant", TPM2_PS_CELL_PHONE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_SERVER_swigconstant", TPM2_PS_SERVER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_PERIPHERAL_swigconstant", TPM2_PS_PERIPHERAL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_TSS_swigconstant", TPM2_PS_TSS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_STORAGE_swigconstant", TPM2_PS_STORAGE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_AUTHENTICATION_swigconstant", TPM2_PS_AUTHENTICATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_EMBEDDED_swigconstant", TPM2_PS_EMBEDDED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_HARDCOPY_swigconstant", TPM2_PS_HARDCOPY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_INFRASTRUCTURE_swigconstant", TPM2_PS_INFRASTRUCTURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_VIRTUALIZATION_swigconstant", TPM2_PS_VIRTUALIZATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_TNC_swigconstant", TPM2_PS_TNC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_MULTI_TENANT_swigconstant", TPM2_PS_MULTI_TENANT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PS_TC_swigconstant", TPM2_PS_TC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_PCR_swigconstant", TPM2_HT_PCR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_NV_INDEX_swigconstant", TPM2_HT_NV_INDEX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_HMAC_SESSION_swigconstant", TPM2_HT_HMAC_SESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_LOADED_SESSION_swigconstant", TPM2_HT_LOADED_SESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_POLICY_SESSION_swigconstant", TPM2_HT_POLICY_SESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_SAVED_SESSION_swigconstant", TPM2_HT_SAVED_SESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_PERMANENT_swigconstant", TPM2_HT_PERMANENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_TRANSIENT_swigconstant", TPM2_HT_TRANSIENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HT_PERSISTENT_swigconstant", TPM2_HT_PERSISTENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_FIRST_swigconstant", TPM2_RH_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_SRK_swigconstant", TPM2_RH_SRK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_OWNER_swigconstant", TPM2_RH_OWNER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_REVOKE_swigconstant", TPM2_RH_REVOKE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_TRANSPORT_swigconstant", TPM2_RH_TRANSPORT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_OPERATOR_swigconstant", TPM2_RH_OPERATOR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_ADMIN_swigconstant", TPM2_RH_ADMIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_EK_swigconstant", TPM2_RH_EK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_NULL_swigconstant", TPM2_RH_NULL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_UNASSIGNED_swigconstant", TPM2_RH_UNASSIGNED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RS_PW_swigconstant", TPM2_RS_PW_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_LOCKOUT_swigconstant", TPM2_RH_LOCKOUT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_ENDORSEMENT_swigconstant", TPM2_RH_ENDORSEMENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_PLATFORM_swigconstant", TPM2_RH_PLATFORM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_PLATFORM_NV_swigconstant", TPM2_RH_PLATFORM_NV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_AUTH_00_swigconstant", TPM2_RH_AUTH_00_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_AUTH_FF_swigconstant", TPM2_RH_AUTH_FF_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_RH_LAST_swigconstant", TPM2_RH_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_HANDLE_MASK_swigconstant", TPM2_HR_HANDLE_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_RANGE_MASK_swigconstant", TPM2_HR_RANGE_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_SHIFT_swigconstant", TPM2_HR_SHIFT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_PCR_swigconstant", TPM2_HR_PCR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_HMAC_SESSION_swigconstant", TPM2_HR_HMAC_SESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_POLICY_SESSION_swigconstant", TPM2_HR_POLICY_SESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_TRANSIENT_swigconstant", TPM2_HR_TRANSIENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_PERSISTENT_swigconstant", TPM2_HR_PERSISTENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_NV_INDEX_swigconstant", TPM2_HR_NV_INDEX_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HR_PERMANENT_swigconstant", TPM2_HR_PERMANENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PCR_FIRST_swigconstant", TPM2_PCR_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PCR_LAST_swigconstant", TPM2_PCR_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HMAC_SESSION_FIRST_swigconstant", TPM2_HMAC_SESSION_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_HMAC_SESSION_LAST_swigconstant", TPM2_HMAC_SESSION_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_LOADED_SESSION_FIRST_swigconstant", TPM2_LOADED_SESSION_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_LOADED_SESSION_LAST_swigconstant", TPM2_LOADED_SESSION_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_POLICY_SESSION_FIRST_swigconstant", TPM2_POLICY_SESSION_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_POLICY_SESSION_LAST_swigconstant", TPM2_POLICY_SESSION_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_TRANSIENT_FIRST_swigconstant", TPM2_TRANSIENT_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ACTIVE_SESSION_FIRST_swigconstant", TPM2_ACTIVE_SESSION_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_ACTIVE_SESSION_LAST_swigconstant", TPM2_ACTIVE_SESSION_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_TRANSIENT_LAST_swigconstant", TPM2_TRANSIENT_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PERSISTENT_FIRST_swigconstant", TPM2_PERSISTENT_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PERSISTENT_LAST_swigconstant", TPM2_PERSISTENT_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PLATFORM_PERSISTENT_swigconstant", TPM2_PLATFORM_PERSISTENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NV_INDEX_FIRST_swigconstant", TPM2_NV_INDEX_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NV_INDEX_LAST_swigconstant", TPM2_NV_INDEX_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PERMANENT_FIRST_swigconstant", TPM2_PERMANENT_FIRST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_PERMANENT_LAST_swigconstant", TPM2_PERMANENT_LAST_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_ASYMMETRIC_swigconstant", TPMA_ALGORITHM_ASYMMETRIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_SYMMETRIC_swigconstant", TPMA_ALGORITHM_SYMMETRIC_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_HASH_swigconstant", TPMA_ALGORITHM_HASH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_OBJECT_swigconstant", TPMA_ALGORITHM_OBJECT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_RESERVED1_MASK_swigconstant", TPMA_ALGORITHM_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_SIGNING_swigconstant", TPMA_ALGORITHM_SIGNING_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_ENCRYPTING_swigconstant", TPMA_ALGORITHM_ENCRYPTING_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_METHOD_swigconstant", TPMA_ALGORITHM_METHOD_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_ALGORITHM_RESERVED2_MASK_swigconstant", TPMA_ALGORITHM_RESERVED2_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_RESERVED1_MASK_swigconstant", TPMA_OBJECT_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_FIXEDTPM_swigconstant", TPMA_OBJECT_FIXEDTPM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_STCLEAR_swigconstant", TPMA_OBJECT_STCLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_RESERVED2_MASK_swigconstant", TPMA_OBJECT_RESERVED2_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_FIXEDPARENT_swigconstant", TPMA_OBJECT_FIXEDPARENT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_SENSITIVEDATAORIGIN_swigconstant", TPMA_OBJECT_SENSITIVEDATAORIGIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_USERWITHAUTH_swigconstant", TPMA_OBJECT_USERWITHAUTH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_ADMINWITHPOLICY_swigconstant", TPMA_OBJECT_ADMINWITHPOLICY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_RESERVED3_MASK_swigconstant", TPMA_OBJECT_RESERVED3_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_NODA_swigconstant", TPMA_OBJECT_NODA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_ENCRYPTEDDUPLICATION_swigconstant", TPMA_OBJECT_ENCRYPTEDDUPLICATION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_RESERVED4_MASK_swigconstant", TPMA_OBJECT_RESERVED4_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_RESTRICTED_swigconstant", TPMA_OBJECT_RESTRICTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_DECRYPT_swigconstant", TPMA_OBJECT_DECRYPT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_SIGN_ENCRYPT_swigconstant", TPMA_OBJECT_SIGN_ENCRYPT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_OBJECT_RESERVED5_MASK_swigconstant", TPMA_OBJECT_RESERVED5_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_CONTINUESESSION_swigconstant", TPMA_SESSION_CONTINUESESSION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_AUDITEXCLUSIVE_swigconstant", TPMA_SESSION_AUDITEXCLUSIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_AUDITRESET_swigconstant", TPMA_SESSION_AUDITRESET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_RESERVED1_MASK_swigconstant", TPMA_SESSION_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_DECRYPT_swigconstant", TPMA_SESSION_DECRYPT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_ENCRYPT_swigconstant", TPMA_SESSION_ENCRYPT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_SESSION_AUDIT_swigconstant", TPMA_SESSION_AUDIT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_TPM2_LOC_ZERO_swigconstant", TPMA_LOCALITY_TPM2_LOC_ZERO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_TPM2_LOC_ONE_swigconstant", TPMA_LOCALITY_TPM2_LOC_ONE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_TPM2_LOC_TWO_swigconstant", TPMA_LOCALITY_TPM2_LOC_TWO_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_TPM2_LOC_THREE_swigconstant", TPMA_LOCALITY_TPM2_LOC_THREE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_TPM2_LOC_FOUR_swigconstant", TPMA_LOCALITY_TPM2_LOC_FOUR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_LOCALITY_EXTENDED_MASK_swigconstant", TPMA_LOCALITY_EXTENDED_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_OWNERAUTHSET_swigconstant", TPMA_PERMANENT_OWNERAUTHSET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_ENDORSEMENTAUTHSET_swigconstant", TPMA_PERMANENT_ENDORSEMENTAUTHSET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_LOCKOUTAUTHSET_swigconstant", TPMA_PERMANENT_LOCKOUTAUTHSET_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_RESERVED1_MASK_swigconstant", TPMA_PERMANENT_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_DISABLECLEAR_swigconstant", TPMA_PERMANENT_DISABLECLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_INLOCKOUT_swigconstant", TPMA_PERMANENT_INLOCKOUT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_TPMGENERATEDEPS_swigconstant", TPMA_PERMANENT_TPMGENERATEDEPS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_PERMANENT_RESERVED2_MASK_swigconstant", TPMA_PERMANENT_RESERVED2_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_STARTUP_CLEAR_PHENABLE_swigconstant", TPMA_STARTUP_CLEAR_PHENABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_STARTUP_CLEAR_SHENABLE_swigconstant", TPMA_STARTUP_CLEAR_SHENABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_STARTUP_CLEAR_EHENABLE_swigconstant", TPMA_STARTUP_CLEAR_EHENABLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_STARTUP_CLEAR_PHENABLENV_swigconstant", TPMA_STARTUP_CLEAR_PHENABLENV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_STARTUP_CLEAR_RESERVED1_MASK_swigconstant", TPMA_STARTUP_CLEAR_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_STARTUP_CLEAR_ORDERLY_swigconstant", TPMA_STARTUP_CLEAR_ORDERLY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_MEMORY_SHAREDRAM_swigconstant", TPMA_MEMORY_SHAREDRAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_MEMORY_SHAREDNV_swigconstant", TPMA_MEMORY_SHAREDNV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_MEMORY_OBJECTCOPIEDTORAM_swigconstant", TPMA_MEMORY_OBJECTCOPIEDTORAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_MEMORY_RESERVED1_MASK_swigconstant", TPMA_MEMORY_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_COMMANDINDEX_MASK_swigconstant", TPMA_CC_COMMANDINDEX_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_RESERVED1_MASK_swigconstant", TPMA_CC_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_NV_swigconstant", TPMA_CC_NV_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_EXTENSIVE_swigconstant", TPMA_CC_EXTENSIVE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_FLUSHED_swigconstant", TPMA_CC_FLUSHED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_CHANDLES_MASK_swigconstant", TPMA_CC_CHANDLES_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_RHANDLE_swigconstant", TPMA_CC_RHANDLE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_V_swigconstant", TPMA_CC_V_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_CC_RES_MASK_swigconstant", TPMA_CC_RES_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_MODES_FIPS_140_2_swigconstant", TPMA_MODES_FIPS_140_2_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_MODES_RESERVED1_MASK_swigconstant", TPMA_MODES_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NV_INDEX_INDEX_MASK_swigconstant", TPM2_NV_INDEX_INDEX_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NV_INDEX_RH_NV_MASK_swigconstant", TPM2_NV_INDEX_RH_NV_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NT_ORDINARY_swigconstant", TPM2_NT_ORDINARY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NT_COUNTER_swigconstant", TPM2_NT_COUNTER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NT_BITS_swigconstant", TPM2_NT_BITS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NT_EXTEND_swigconstant", TPM2_NT_EXTEND_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NT_PIN_FAIL_swigconstant", TPM2_NT_PIN_FAIL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM2_NT_PIN_PASS_swigconstant", TPM2_NT_PIN_PASS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_PPWRITE_swigconstant", TPMA_NV_PPWRITE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_OWNERWRITE_swigconstant", TPMA_NV_OWNERWRITE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_AUTHWRITE_swigconstant", TPMA_NV_AUTHWRITE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_POLICYWRITE_swigconstant", TPMA_NV_POLICYWRITE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_TPM2_NT_MASK_swigconstant", TPMA_NV_TPM2_NT_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_RESERVED1_MASK_swigconstant", TPMA_NV_RESERVED1_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_POLICY_DELETE_swigconstant", TPMA_NV_POLICY_DELETE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_WRITELOCKED_swigconstant", TPMA_NV_WRITELOCKED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_WRITEALL_swigconstant", TPMA_NV_WRITEALL_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_WRITEDEFINE_swigconstant", TPMA_NV_WRITEDEFINE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_WRITE_STCLEAR_swigconstant", TPMA_NV_WRITE_STCLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_GLOBALLOCK_swigconstant", TPMA_NV_GLOBALLOCK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_PPREAD_swigconstant", TPMA_NV_PPREAD_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_OWNERREAD_swigconstant", TPMA_NV_OWNERREAD_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_AUTHREAD_swigconstant", TPMA_NV_AUTHREAD_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_POLICYREAD_swigconstant", TPMA_NV_POLICYREAD_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_RESERVED2_MASK_swigconstant", TPMA_NV_RESERVED2_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_NO_DA_swigconstant", TPMA_NV_NO_DA_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_ORDERLY_swigconstant", TPMA_NV_ORDERLY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_CLEAR_STCLEAR_swigconstant", TPMA_NV_CLEAR_STCLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_READLOCKED_swigconstant", TPMA_NV_READLOCKED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_WRITTEN_swigconstant", TPMA_NV_WRITTEN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_PLATFORMCREATE_swigconstant", TPMA_NV_PLATFORMCREATE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPMA_NV_READ_STCLEAR_swigconstant", TPMA_NV_READ_STCLEAR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM_AT_ANY_swigconstant", TPM_AT_ANY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM_AT_ERROR_swigconstant", TPM_AT_ERROR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TPM_AT_PV1_swigconstant", TPM_AT_PV1_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_RC_SUCCESS_swigconstant", TSS2_RC_SUCCESS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_RC_LAYER_MASK_swigconstant", TSS2_RC_LAYER_MASK_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TPM_RC_LAYER_swigconstant", TSS2_TPM_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_FEATURE_RC_LAYER_swigconstant", TSS2_FEATURE_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESAPI_RC_LAYER_swigconstant", TSS2_ESAPI_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_LAYER_swigconstant", TSS2_SYS_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_MU_RC_LAYER_swigconstant", TSS2_MU_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_LAYER_swigconstant", TSS2_TCTI_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_RESMGR_RC_LAYER_swigconstant", TSS2_RESMGR_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_RESMGR_TPM_RC_LAYER_swigconstant", TSS2_RESMGR_TPM_RC_LAYER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_GENERAL_FAILURE_swigconstant", TSS2_TCTI_RC_GENERAL_FAILURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_NOT_IMPLEMENTED_swigconstant", TSS2_TCTI_RC_NOT_IMPLEMENTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_BAD_CONTEXT_swigconstant", TSS2_TCTI_RC_BAD_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_ABI_MISMATCH_swigconstant", TSS2_TCTI_RC_ABI_MISMATCH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_BAD_REFERENCE_swigconstant", TSS2_TCTI_RC_BAD_REFERENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_INSUFFICIENT_BUFFER_swigconstant", TSS2_TCTI_RC_INSUFFICIENT_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_BAD_SEQUENCE_swigconstant", TSS2_TCTI_RC_BAD_SEQUENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_NO_CONNECTION_swigconstant", TSS2_TCTI_RC_NO_CONNECTION_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_IO_ERROR_swigconstant", TSS2_TCTI_RC_IO_ERROR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_BAD_VALUE_swigconstant", TSS2_TCTI_RC_BAD_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_NOT_PERMITTED_swigconstant", TSS2_TCTI_RC_NOT_PERMITTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_MALFORMED_RESPONSE_swigconstant", TSS2_TCTI_RC_MALFORMED_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_TCTI_RC_NOT_SUPPORTED_swigconstant", TSS2_TCTI_RC_NOT_SUPPORTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_GENERAL_FAILURE_swigconstant", TSS2_SYS_RC_GENERAL_FAILURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_ABI_MISMATCH_swigconstant", TSS2_SYS_RC_ABI_MISMATCH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_BAD_REFERENCE_swigconstant", TSS2_SYS_RC_BAD_REFERENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_INSUFFICIENT_BUFFER_swigconstant", TSS2_SYS_RC_INSUFFICIENT_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_BAD_SEQUENCE_swigconstant", TSS2_SYS_RC_BAD_SEQUENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_BAD_VALUE_swigconstant", TSS2_SYS_RC_BAD_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_INVALID_SESSIONS_swigconstant", TSS2_SYS_RC_INVALID_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_NO_DECRYPT_PARAM_swigconstant", TSS2_SYS_RC_NO_DECRYPT_PARAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_NO_ENCRYPT_PARAM_swigconstant", TSS2_SYS_RC_NO_ENCRYPT_PARAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_BAD_SIZE_swigconstant", TSS2_SYS_RC_BAD_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_MALFORMED_RESPONSE_swigconstant", TSS2_SYS_RC_MALFORMED_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_INSUFFICIENT_CONTEXT_swigconstant", TSS2_SYS_RC_INSUFFICIENT_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_INSUFFICIENT_RESPONSE_swigconstant", TSS2_SYS_RC_INSUFFICIENT_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_INCOMPATIBLE_TCTI_swigconstant", TSS2_SYS_RC_INCOMPATIBLE_TCTI_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_SYS_RC_BAD_TCTI_STRUCTURE_swigconstant", TSS2_SYS_RC_BAD_TCTI_STRUCTURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_MU_RC_GENERAL_FAILURE_swigconstant", TSS2_MU_RC_GENERAL_FAILURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_MU_RC_BAD_REFERENCE_swigconstant", TSS2_MU_RC_BAD_REFERENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_MU_RC_BAD_SIZE_swigconstant", TSS2_MU_RC_BAD_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_MU_RC_BAD_VALUE_swigconstant", TSS2_MU_RC_BAD_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_MU_RC_INSUFFICIENT_BUFFER_swigconstant", TSS2_MU_RC_INSUFFICIENT_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_GENERAL_FAILURE_swigconstant", TSS2_ESYS_RC_GENERAL_FAILURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_NOT_IMPLEMENTED_swigconstant", TSS2_ESYS_RC_NOT_IMPLEMENTED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_ABI_MISMATCH_swigconstant", TSS2_ESYS_RC_ABI_MISMATCH_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_BAD_REFERENCE_swigconstant", TSS2_ESYS_RC_BAD_REFERENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_INSUFFICIENT_BUFFER_swigconstant", TSS2_ESYS_RC_INSUFFICIENT_BUFFER_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_BAD_SEQUENCE_swigconstant", TSS2_ESYS_RC_BAD_SEQUENCE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_INVALID_SESSIONS_swigconstant", TSS2_ESYS_RC_INVALID_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_TRY_AGAIN_swigconstant", TSS2_ESYS_RC_TRY_AGAIN_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_IO_ERROR_swigconstant", TSS2_ESYS_RC_IO_ERROR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_BAD_VALUE_swigconstant", TSS2_ESYS_RC_BAD_VALUE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_NO_DECRYPT_PARAM_swigconstant", TSS2_ESYS_RC_NO_DECRYPT_PARAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_NO_ENCRYPT_PARAM_swigconstant", TSS2_ESYS_RC_NO_ENCRYPT_PARAM_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_BAD_SIZE_swigconstant", TSS2_ESYS_RC_BAD_SIZE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_MALFORMED_RESPONSE_swigconstant", TSS2_ESYS_RC_MALFORMED_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_INSUFFICIENT_CONTEXT_swigconstant", TSS2_ESYS_RC_INSUFFICIENT_CONTEXT_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_INSUFFICIENT_RESPONSE_swigconstant", TSS2_ESYS_RC_INSUFFICIENT_RESPONSE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_INCOMPATIBLE_TCTI_swigconstant", TSS2_ESYS_RC_INCOMPATIBLE_TCTI_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_BAD_TCTI_STRUCTURE_swigconstant", TSS2_ESYS_RC_BAD_TCTI_STRUCTURE_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_MEMORY_swigconstant", TSS2_ESYS_RC_MEMORY_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_BAD_TR_swigconstant", TSS2_ESYS_RC_BAD_TR_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS_swigconstant", TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS_swigconstant", TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS_swigconstant, METH_VARARGS, NULL},
	 { (char *)"TSS2_ESYS_RC_RSP_AUTH_FAILED_swigconstant", TSS2_ESYS_RC_RSP_AUTH_FAILED_swigconstant, METH_VARARGS, NULL},
	 { (char *)"Tss2_MU_BYTE_Marshal", _wrap_Tss2_MU_BYTE_Marshal, METH_VARARGS, (char *)"Tss2_MU_BYTE_Marshal(BYTE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_BYTE_Unmarshal", _wrap_Tss2_MU_BYTE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_BYTE_Unmarshal(uint8_t const [] buffer, size_t * offset, BYTE * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT8_Marshal", _wrap_Tss2_MU_INT8_Marshal, METH_VARARGS, (char *)"Tss2_MU_INT8_Marshal(INT8 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT8_Unmarshal", _wrap_Tss2_MU_INT8_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_INT8_Unmarshal(uint8_t const [] buffer, size_t * offset, INT8 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT16_Marshal", _wrap_Tss2_MU_INT16_Marshal, METH_VARARGS, (char *)"Tss2_MU_INT16_Marshal(INT16 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT16_Unmarshal", _wrap_Tss2_MU_INT16_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_INT16_Unmarshal(uint8_t const [] buffer, size_t * offset, INT16 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT32_Marshal", _wrap_Tss2_MU_INT32_Marshal, METH_VARARGS, (char *)"Tss2_MU_INT32_Marshal(INT32 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT32_Unmarshal", _wrap_Tss2_MU_INT32_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_INT32_Unmarshal(uint8_t const [] buffer, size_t * offset, INT32 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT64_Marshal", _wrap_Tss2_MU_INT64_Marshal, METH_VARARGS, (char *)"Tss2_MU_INT64_Marshal(INT64 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_INT64_Unmarshal", _wrap_Tss2_MU_INT64_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_INT64_Unmarshal(uint8_t const [] buffer, size_t * offset, INT64 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT8_Marshal", _wrap_Tss2_MU_UINT8_Marshal, METH_VARARGS, (char *)"Tss2_MU_UINT8_Marshal(UINT8 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT8_Unmarshal", _wrap_Tss2_MU_UINT8_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_UINT8_Unmarshal(uint8_t const [] buffer, size_t * offset, UINT8 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT16_Marshal", _wrap_Tss2_MU_UINT16_Marshal, METH_VARARGS, (char *)"Tss2_MU_UINT16_Marshal(UINT16 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT16_Unmarshal", _wrap_Tss2_MU_UINT16_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_UINT16_Unmarshal(uint8_t const [] buffer, size_t * offset, UINT16 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT32_Marshal", _wrap_Tss2_MU_UINT32_Marshal, METH_VARARGS, (char *)"Tss2_MU_UINT32_Marshal(UINT32 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT32_Unmarshal", _wrap_Tss2_MU_UINT32_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_UINT32_Unmarshal(uint8_t const [] buffer, size_t * offset, UINT32 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT64_Marshal", _wrap_Tss2_MU_UINT64_Marshal, METH_VARARGS, (char *)"Tss2_MU_UINT64_Marshal(UINT64 src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_UINT64_Unmarshal", _wrap_Tss2_MU_UINT64_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_UINT64_Unmarshal(uint8_t const [] buffer, size_t * offset, UINT64 * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_CC_Marshal", _wrap_Tss2_MU_TPM2_CC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_CC_Marshal(TPM2_CC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_CC_Unmarshal", _wrap_Tss2_MU_TPM2_CC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_CC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2_CC * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_ST_Marshal", _wrap_Tss2_MU_TPM2_ST_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_ST_Marshal(TPM2_ST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_ST_Unmarshal", _wrap_Tss2_MU_TPM2_ST_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_ST_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2_ST * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_ALGORITHM_Marshal", _wrap_Tss2_MU_TPMA_ALGORITHM_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_ALGORITHM_Marshal(TPMA_ALGORITHM src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_ALGORITHM_Unmarshal", _wrap_Tss2_MU_TPMA_ALGORITHM_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_ALGORITHM_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_ALGORITHM * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_CC_Marshal", _wrap_Tss2_MU_TPMA_CC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_CC_Marshal(TPMA_CC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_CC_Unmarshal", _wrap_Tss2_MU_TPMA_CC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_CC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_CC * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_LOCALITY_Marshal", _wrap_Tss2_MU_TPMA_LOCALITY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_LOCALITY_Marshal(TPMA_LOCALITY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_LOCALITY_Unmarshal", _wrap_Tss2_MU_TPMA_LOCALITY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_LOCALITY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_LOCALITY * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_NV_Marshal", _wrap_Tss2_MU_TPMA_NV_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_NV_Marshal(TPMA_NV src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_NV_Unmarshal", _wrap_Tss2_MU_TPMA_NV_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_NV_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_NV * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_OBJECT_Marshal", _wrap_Tss2_MU_TPMA_OBJECT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_OBJECT_Marshal(TPMA_OBJECT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_OBJECT_Unmarshal", _wrap_Tss2_MU_TPMA_OBJECT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_OBJECT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_OBJECT * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_PERMANENT_Marshal", _wrap_Tss2_MU_TPMA_PERMANENT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_PERMANENT_Marshal(TPMA_PERMANENT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_PERMANENT_Unmarshal", _wrap_Tss2_MU_TPMA_PERMANENT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_PERMANENT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_PERMANENT * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_SESSION_Marshal", _wrap_Tss2_MU_TPMA_SESSION_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_SESSION_Marshal(TPMA_SESSION src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_SESSION_Unmarshal", _wrap_Tss2_MU_TPMA_SESSION_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_SESSION_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_SESSION * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_STARTUP_CLEAR_Marshal", _wrap_Tss2_MU_TPMA_STARTUP_CLEAR_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_STARTUP_CLEAR_Marshal(TPMA_STARTUP_CLEAR src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal", _wrap_Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMA_STARTUP_CLEAR_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMA_STARTUP_CLEAR * dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_DIGEST_Marshal", _wrap_Tss2_MU_TPM2B_DIGEST_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_DIGEST_Marshal(TPM2B_DIGEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_DIGEST_Unmarshal", _wrap_Tss2_MU_TPM2B_DIGEST_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_DIGEST_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_DIGEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ATTEST_Marshal", _wrap_Tss2_MU_TPM2B_ATTEST_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ATTEST_Marshal(TPM2B_ATTEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ATTEST_Unmarshal", _wrap_Tss2_MU_TPM2B_ATTEST_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ATTEST_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_ATTEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_NAME_Marshal", _wrap_Tss2_MU_TPM2B_NAME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_NAME_Marshal(TPM2B_NAME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_NAME_Unmarshal", _wrap_Tss2_MU_TPM2B_NAME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_NAME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_NAME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal", _wrap_Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_MAX_NV_BUFFER_Marshal(TPM2B_MAX_NV_BUFFER src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal", _wrap_Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_MAX_NV_BUFFER_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_MAX_NV_BUFFER dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal", _wrap_Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SENSITIVE_DATA_Marshal(TPM2B_SENSITIVE_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal", _wrap_Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SENSITIVE_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_SENSITIVE_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ECC_PARAMETER_Marshal", _wrap_Tss2_MU_TPM2B_ECC_PARAMETER_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ECC_PARAMETER_Marshal(TPM2B_ECC_PARAMETER src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal", _wrap_Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ECC_PARAMETER_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_ECC_PARAMETER dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal", _wrap_Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Marshal(TPM2B_PUBLIC_KEY_RSA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal", _wrap_Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PUBLIC_KEY_RSA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_PUBLIC_KEY_RSA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal", _wrap_Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Marshal(TPM2B_PRIVATE_KEY_RSA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal", _wrap_Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PRIVATE_KEY_RSA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_PRIVATE_KEY_RSA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PRIVATE_Marshal", _wrap_Tss2_MU_TPM2B_PRIVATE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PRIVATE_Marshal(TPM2B_PRIVATE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PRIVATE_Unmarshal", _wrap_Tss2_MU_TPM2B_PRIVATE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PRIVATE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_PRIVATE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal", _wrap_Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Marshal(TPM2B_CONTEXT_SENSITIVE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal", _wrap_Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_CONTEXT_SENSITIVE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_CONTEXT_SENSITIVE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_CONTEXT_DATA_Marshal", _wrap_Tss2_MU_TPM2B_CONTEXT_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_CONTEXT_DATA_Marshal(TPM2B_CONTEXT_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal", _wrap_Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_CONTEXT_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_CONTEXT_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_DATA_Marshal", _wrap_Tss2_MU_TPM2B_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_DATA_Marshal(TPM2B_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_DATA_Unmarshal", _wrap_Tss2_MU_TPM2B_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SYM_KEY_Marshal", _wrap_Tss2_MU_TPM2B_SYM_KEY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SYM_KEY_Marshal(TPM2B_SYM_KEY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SYM_KEY_Unmarshal", _wrap_Tss2_MU_TPM2B_SYM_KEY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SYM_KEY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_SYM_KEY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ECC_POINT_Marshal", _wrap_Tss2_MU_TPM2B_ECC_POINT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ECC_POINT_Marshal(TPM2B_ECC_POINT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ECC_POINT_Unmarshal", _wrap_Tss2_MU_TPM2B_ECC_POINT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ECC_POINT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_ECC_POINT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_NV_PUBLIC_Marshal", _wrap_Tss2_MU_TPM2B_NV_PUBLIC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_NV_PUBLIC_Marshal(TPM2B_NV_PUBLIC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal", _wrap_Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_NV_PUBLIC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_NV_PUBLIC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SENSITIVE_Marshal", _wrap_Tss2_MU_TPM2B_SENSITIVE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SENSITIVE_Marshal(TPM2B_SENSITIVE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SENSITIVE_Unmarshal", _wrap_Tss2_MU_TPM2B_SENSITIVE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SENSITIVE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_SENSITIVE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal", _wrap_Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SENSITIVE_CREATE_Marshal(TPM2B_SENSITIVE_CREATE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal", _wrap_Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_SENSITIVE_CREATE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_SENSITIVE_CREATE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_CREATION_DATA_Marshal", _wrap_Tss2_MU_TPM2B_CREATION_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_CREATION_DATA_Marshal(TPM2B_CREATION_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_CREATION_DATA_Unmarshal", _wrap_Tss2_MU_TPM2B_CREATION_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_CREATION_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_CREATION_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PUBLIC_Marshal", _wrap_Tss2_MU_TPM2B_PUBLIC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PUBLIC_Marshal(TPM2B_PUBLIC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_PUBLIC_Unmarshal", _wrap_Tss2_MU_TPM2B_PUBLIC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_PUBLIC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_PUBLIC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal", _wrap_Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ENCRYPTED_SECRET_Marshal(TPM2B_ENCRYPTED_SECRET src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal", _wrap_Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ENCRYPTED_SECRET_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_ENCRYPTED_SECRET dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ID_OBJECT_Marshal", _wrap_Tss2_MU_TPM2B_ID_OBJECT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ID_OBJECT_Marshal(TPM2B_ID_OBJECT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_ID_OBJECT_Unmarshal", _wrap_Tss2_MU_TPM2B_ID_OBJECT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_ID_OBJECT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_ID_OBJECT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_IV_Marshal", _wrap_Tss2_MU_TPM2B_IV_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_IV_Marshal(TPM2B_IV src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_IV_Unmarshal", _wrap_Tss2_MU_TPM2B_IV_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_IV_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_IV dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_AUTH_Marshal", _wrap_Tss2_MU_TPM2B_AUTH_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_AUTH_Marshal(TPM2B_DIGEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_AUTH_Unmarshal", _wrap_Tss2_MU_TPM2B_AUTH_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_AUTH_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_DIGEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_EVENT_Marshal", _wrap_Tss2_MU_TPM2B_EVENT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_EVENT_Marshal(TPM2B_EVENT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_EVENT_Unmarshal", _wrap_Tss2_MU_TPM2B_EVENT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_EVENT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_EVENT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_MAX_BUFFER_Marshal", _wrap_Tss2_MU_TPM2B_MAX_BUFFER_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_MAX_BUFFER_Marshal(TPM2B_MAX_BUFFER src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal", _wrap_Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_MAX_BUFFER_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_MAX_BUFFER dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_NONCE_Marshal", _wrap_Tss2_MU_TPM2B_NONCE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_NONCE_Marshal(TPM2B_DIGEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_NONCE_Unmarshal", _wrap_Tss2_MU_TPM2B_NONCE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_NONCE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_DIGEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_OPERAND_Marshal", _wrap_Tss2_MU_TPM2B_OPERAND_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_OPERAND_Marshal(TPM2B_DIGEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_OPERAND_Unmarshal", _wrap_Tss2_MU_TPM2B_OPERAND_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_OPERAND_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_DIGEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_TIMEOUT_Marshal", _wrap_Tss2_MU_TPM2B_TIMEOUT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_TIMEOUT_Marshal(TPM2B_DIGEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_TIMEOUT_Unmarshal", _wrap_Tss2_MU_TPM2B_TIMEOUT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_TIMEOUT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_DIGEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_TEMPLATE_Marshal", _wrap_Tss2_MU_TPM2B_TEMPLATE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_TEMPLATE_Marshal(TPM2B_TEMPLATE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2B_TEMPLATE_Unmarshal", _wrap_Tss2_MU_TPM2B_TEMPLATE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2B_TEMPLATE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPM2B_TEMPLATE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CONTEXT_Marshal", _wrap_Tss2_MU_TPMS_CONTEXT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CONTEXT_Marshal(TPMS_CONTEXT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CONTEXT_Unmarshal", _wrap_Tss2_MU_TPMS_CONTEXT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CONTEXT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CONTEXT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TIME_INFO_Marshal", _wrap_Tss2_MU_TPMS_TIME_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TIME_INFO_Marshal(TPMS_TIME_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TIME_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_TIME_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TIME_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_TIME_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ECC_POINT_Marshal", _wrap_Tss2_MU_TPMS_ECC_POINT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ECC_POINT_Marshal(TPMS_ECC_POINT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ECC_POINT_Unmarshal", _wrap_Tss2_MU_TPMS_ECC_POINT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ECC_POINT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ECC_POINT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_NV_PUBLIC_Marshal", _wrap_Tss2_MU_TPMS_NV_PUBLIC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_NV_PUBLIC_Marshal(TPMS_NV_PUBLIC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_NV_PUBLIC_Unmarshal", _wrap_Tss2_MU_TPMS_NV_PUBLIC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_NV_PUBLIC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_NV_PUBLIC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ALG_PROPERTY_Marshal", _wrap_Tss2_MU_TPMS_ALG_PROPERTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ALG_PROPERTY_Marshal(TPMS_ALG_PROPERTY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal", _wrap_Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ALG_PROPERTY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ALG_PROPERTY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal", _wrap_Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal(TPMS_ALGORITHM_DESCRIPTION src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal", _wrap_Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ALGORITHM_DESCRIPTION dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal", _wrap_Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TAGGED_PROPERTY_Marshal(TPMS_TAGGED_PROPERTY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal", _wrap_Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TAGGED_PROPERTY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_TAGGED_PROPERTY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TAGGED_POLICY_Marshal", _wrap_Tss2_MU_TPMS_TAGGED_POLICY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TAGGED_POLICY_Marshal(TPMS_TAGGED_POLICY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal", _wrap_Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TAGGED_POLICY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_TAGGED_POLICY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CLOCK_INFO_Marshal", _wrap_Tss2_MU_TPMS_CLOCK_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CLOCK_INFO_Marshal(TPMS_CLOCK_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CLOCK_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_CLOCK_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CLOCK_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CLOCK_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal", _wrap_Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TIME_ATTEST_INFO_Marshal(TPMS_TIME_ATTEST_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TIME_ATTEST_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_TIME_ATTEST_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CERTIFY_INFO_Marshal", _wrap_Tss2_MU_TPMS_CERTIFY_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CERTIFY_INFO_Marshal(TPMS_CERTIFY_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CERTIFY_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CERTIFY_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal", _wrap_Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Marshal(TPMS_COMMAND_AUDIT_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_COMMAND_AUDIT_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_COMMAND_AUDIT_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal", _wrap_Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SESSION_AUDIT_INFO_Marshal(TPMS_SESSION_AUDIT_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SESSION_AUDIT_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SESSION_AUDIT_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CREATION_INFO_Marshal", _wrap_Tss2_MU_TPMS_CREATION_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CREATION_INFO_Marshal(TPMS_CREATION_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CREATION_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_CREATION_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CREATION_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CREATION_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal", _wrap_Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_NV_CERTIFY_INFO_Marshal(TPMS_NV_CERTIFY_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_NV_CERTIFY_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_NV_CERTIFY_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_AUTH_COMMAND_Marshal", _wrap_Tss2_MU_TPMS_AUTH_COMMAND_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_AUTH_COMMAND_Marshal(TPMS_AUTH_COMMAND src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal", _wrap_Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_AUTH_COMMAND_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_AUTH_COMMAND dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_AUTH_RESPONSE_Marshal", _wrap_Tss2_MU_TPMS_AUTH_RESPONSE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_AUTH_RESPONSE_Marshal(TPMS_AUTH_RESPONSE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal", _wrap_Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_AUTH_RESPONSE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_AUTH_RESPONSE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal", _wrap_Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SENSITIVE_CREATE_Marshal(TPMS_SENSITIVE_CREATE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal", _wrap_Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SENSITIVE_CREATE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SENSITIVE_CREATE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SCHEME_HASH_Marshal", _wrap_Tss2_MU_TPMS_SCHEME_HASH_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SCHEME_HASH_Marshal(TPMS_SCHEME_HASH src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SCHEME_HASH_Unmarshal", _wrap_Tss2_MU_TPMS_SCHEME_HASH_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SCHEME_HASH_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SCHEME_HASH dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SCHEME_ECDAA_Marshal", _wrap_Tss2_MU_TPMS_SCHEME_ECDAA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SCHEME_ECDAA_Marshal(TPMS_SCHEME_ECDAA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal", _wrap_Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SCHEME_ECDAA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SCHEME_ECDAA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SCHEME_XOR_Marshal", _wrap_Tss2_MU_TPMS_SCHEME_XOR_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SCHEME_XOR_Marshal(TPMS_SCHEME_XOR src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SCHEME_XOR_Unmarshal", _wrap_Tss2_MU_TPMS_SCHEME_XOR_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SCHEME_XOR_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SCHEME_XOR dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SIGNATURE_RSA_Marshal", _wrap_Tss2_MU_TPMS_SIGNATURE_RSA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SIGNATURE_RSA_Marshal(TPMS_SIGNATURE_RSA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal", _wrap_Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SIGNATURE_RSA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SIGNATURE_RSA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SIGNATURE_ECC_Marshal", _wrap_Tss2_MU_TPMS_SIGNATURE_ECC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SIGNATURE_ECC_Marshal(TPMS_SIGNATURE_ECC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal", _wrap_Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SIGNATURE_ECC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SIGNATURE_ECC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal", _wrap_Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Marshal(TPMS_NV_PIN_COUNTER_PARAMETERS src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal", _wrap_Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_NV_PIN_COUNTER_PARAMETERS_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_NV_PIN_COUNTER_PARAMETERS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CONTEXT_DATA_Marshal", _wrap_Tss2_MU_TPMS_CONTEXT_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CONTEXT_DATA_Marshal(TPMS_CONTEXT_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal", _wrap_Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CONTEXT_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CONTEXT_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_PCR_SELECT_Marshal", _wrap_Tss2_MU_TPMS_PCR_SELECT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_PCR_SELECT_Marshal(TPMS_PCR_SELECT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_PCR_SELECT_Unmarshal", _wrap_Tss2_MU_TPMS_PCR_SELECT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_PCR_SELECT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_PCR_SELECT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_PCR_SELECTION_Marshal", _wrap_Tss2_MU_TPMS_PCR_SELECTION_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_PCR_SELECTION_Marshal(TPMS_PCR_SELECTION src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_PCR_SELECTION_Unmarshal", _wrap_Tss2_MU_TPMS_PCR_SELECTION_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_PCR_SELECTION_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_PCR_SELECTION dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal", _wrap_Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TAGGED_PCR_SELECT_Marshal(TPMS_TAGGED_PCR_SELECT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal", _wrap_Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_TAGGED_PCR_SELECT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_TAGGED_PCR_SELECT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_QUOTE_INFO_Marshal", _wrap_Tss2_MU_TPMS_QUOTE_INFO_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_QUOTE_INFO_Marshal(TPMS_QUOTE_INFO src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_QUOTE_INFO_Unmarshal", _wrap_Tss2_MU_TPMS_QUOTE_INFO_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_QUOTE_INFO_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_QUOTE_INFO dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CREATION_DATA_Marshal", _wrap_Tss2_MU_TPMS_CREATION_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CREATION_DATA_Marshal(TPMS_CREATION_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CREATION_DATA_Unmarshal", _wrap_Tss2_MU_TPMS_CREATION_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CREATION_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CREATION_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ECC_PARMS_Marshal", _wrap_Tss2_MU_TPMS_ECC_PARMS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ECC_PARMS_Marshal(TPMS_ECC_PARMS src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ECC_PARMS_Unmarshal", _wrap_Tss2_MU_TPMS_ECC_PARMS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ECC_PARMS_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ECC_PARMS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ATTEST_Marshal", _wrap_Tss2_MU_TPMS_ATTEST_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ATTEST_Marshal(TPMS_ATTEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ATTEST_Unmarshal", _wrap_Tss2_MU_TPMS_ATTEST_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ATTEST_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ATTEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal", _wrap_Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Marshal(TPMS_ALGORITHM_DETAIL_ECC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal", _wrap_Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ALGORITHM_DETAIL_ECC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ALGORITHM_DETAIL_ECC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CAPABILITY_DATA_Marshal", _wrap_Tss2_MU_TPMS_CAPABILITY_DATA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CAPABILITY_DATA_Marshal(TPMS_CAPABILITY_DATA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal", _wrap_Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_CAPABILITY_DATA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_CAPABILITY_DATA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal", _wrap_Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_KEYEDHASH_PARMS_Marshal(TPMS_KEYEDHASH_PARMS src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal", _wrap_Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_KEYEDHASH_PARMS_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_KEYEDHASH_PARMS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_RSA_PARMS_Marshal", _wrap_Tss2_MU_TPMS_RSA_PARMS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_RSA_PARMS_Marshal(TPMS_RSA_PARMS src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_RSA_PARMS_Unmarshal", _wrap_Tss2_MU_TPMS_RSA_PARMS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_RSA_PARMS_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_RSA_PARMS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal", _wrap_Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SYMCIPHER_PARMS_Marshal(TPMS_SYMCIPHER_PARMS src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal", _wrap_Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_SYMCIPHER_PARMS_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_SYMCIPHER_PARMS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_AC_OUTPUT_Marshal", _wrap_Tss2_MU_TPMS_AC_OUTPUT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_AC_OUTPUT_Marshal(TPMS_AC_OUTPUT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_AC_OUTPUT_Unmarshal", _wrap_Tss2_MU_TPMS_AC_OUTPUT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_AC_OUTPUT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_AC_OUTPUT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ID_OBJECT_Marshal", _wrap_Tss2_MU_TPMS_ID_OBJECT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ID_OBJECT_Marshal(TPMS_ID_OBJECT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_ID_OBJECT_Unmarshal", _wrap_Tss2_MU_TPMS_ID_OBJECT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_ID_OBJECT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMS_ID_OBJECT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_CC_Marshal", _wrap_Tss2_MU_TPML_CC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_CC_Marshal(TPML_CC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_CC_Unmarshal", _wrap_Tss2_MU_TPML_CC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_CC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_CC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_CCA_Marshal", _wrap_Tss2_MU_TPML_CCA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_CCA_Marshal(TPML_CCA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_CCA_Unmarshal", _wrap_Tss2_MU_TPML_CCA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_CCA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_CCA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_ALG_Marshal", _wrap_Tss2_MU_TPML_ALG_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_ALG_Marshal(TPML_ALG src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_ALG_Unmarshal", _wrap_Tss2_MU_TPML_ALG_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_ALG_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_ALG dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_HANDLE_Marshal", _wrap_Tss2_MU_TPML_HANDLE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_HANDLE_Marshal(TPML_HANDLE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_HANDLE_Unmarshal", _wrap_Tss2_MU_TPML_HANDLE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_HANDLE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_HANDLE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_DIGEST_Marshal", _wrap_Tss2_MU_TPML_DIGEST_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_DIGEST_Marshal(TPML_DIGEST src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_DIGEST_Unmarshal", _wrap_Tss2_MU_TPML_DIGEST_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_DIGEST_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_DIGEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_DIGEST_VALUES_Marshal", _wrap_Tss2_MU_TPML_DIGEST_VALUES_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_DIGEST_VALUES_Marshal(TPML_DIGEST_VALUES src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_DIGEST_VALUES_Unmarshal", _wrap_Tss2_MU_TPML_DIGEST_VALUES_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_DIGEST_VALUES_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_DIGEST_VALUES dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_PCR_SELECTION_Marshal", _wrap_Tss2_MU_TPML_PCR_SELECTION_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_PCR_SELECTION_Marshal(TPML_PCR_SELECTION src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_PCR_SELECTION_Unmarshal", _wrap_Tss2_MU_TPML_PCR_SELECTION_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_PCR_SELECTION_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_PCR_SELECTION dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_ALG_PROPERTY_Marshal", _wrap_Tss2_MU_TPML_ALG_PROPERTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_ALG_PROPERTY_Marshal(TPML_ALG_PROPERTY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_ALG_PROPERTY_Unmarshal", _wrap_Tss2_MU_TPML_ALG_PROPERTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_ALG_PROPERTY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_ALG_PROPERTY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_ECC_CURVE_Marshal", _wrap_Tss2_MU_TPML_ECC_CURVE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_ECC_CURVE_Marshal(TPML_ECC_CURVE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_ECC_CURVE_Unmarshal", _wrap_Tss2_MU_TPML_ECC_CURVE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_ECC_CURVE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_ECC_CURVE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal", _wrap_Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Marshal(TPML_TAGGED_PCR_PROPERTY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal", _wrap_Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_TAGGED_PCR_PROPERTY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_TAGGED_PCR_PROPERTY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal", _wrap_Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Marshal(TPML_TAGGED_TPM_PROPERTY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal", _wrap_Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_TAGGED_TPM_PROPERTY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_TAGGED_TPM_PROPERTY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal", _wrap_Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_INTEL_PTT_PROPERTY_Marshal(TPML_INTEL_PTT_PROPERTY src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal", _wrap_Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_INTEL_PTT_PROPERTY_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_INTEL_PTT_PROPERTY dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_AC_CAPABILITIES_Marshal", _wrap_Tss2_MU_TPML_AC_CAPABILITIES_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPML_AC_CAPABILITIES_Marshal(TPML_AC_CAPABILITIES src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal", _wrap_Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPML_AC_CAPABILITIES_Unmarshal(uint8_t const [] buffer, size_t * offset, TPML_AC_CAPABILITIES dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_HA_Marshal", _wrap_Tss2_MU_TPMU_HA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_HA_Marshal(TPMU_HA src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_HA_Unmarshal", _wrap_Tss2_MU_TPMU_HA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_HA_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_HA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_CAPABILITIES_Marshal", _wrap_Tss2_MU_TPMU_CAPABILITIES_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_CAPABILITIES_Marshal(TPMU_CAPABILITIES src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_CAPABILITIES_Unmarshal", _wrap_Tss2_MU_TPMU_CAPABILITIES_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_CAPABILITIES_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_CAPABILITIES dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_ATTEST_Marshal", _wrap_Tss2_MU_TPMU_ATTEST_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_ATTEST_Marshal(TPMU_ATTEST src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_ATTEST_Unmarshal", _wrap_Tss2_MU_TPMU_ATTEST_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_ATTEST_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_ATTEST dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SYM_KEY_BITS_Marshal", _wrap_Tss2_MU_TPMU_SYM_KEY_BITS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SYM_KEY_BITS_Marshal(TPMU_SYM_KEY_BITS src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal", _wrap_Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SYM_KEY_BITS_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_SYM_KEY_BITS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SYM_MODE_Marshal", _wrap_Tss2_MU_TPMU_SYM_MODE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SYM_MODE_Marshal(TPMU_SYM_MODE src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SYM_MODE_Unmarshal", _wrap_Tss2_MU_TPMU_SYM_MODE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SYM_MODE_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_SYM_MODE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SIG_SCHEME_Marshal", _wrap_Tss2_MU_TPMU_SIG_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SIG_SCHEME_Marshal(TPMU_SIG_SCHEME src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SIG_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMU_SIG_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SIG_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_SIG_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_KDF_SCHEME_Marshal", _wrap_Tss2_MU_TPMU_KDF_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_KDF_SCHEME_Marshal(TPMU_KDF_SCHEME src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_KDF_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMU_KDF_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_KDF_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_KDF_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_ASYM_SCHEME_Marshal", _wrap_Tss2_MU_TPMU_ASYM_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_ASYM_SCHEME_Marshal(TPMU_ASYM_SCHEME src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_ASYM_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_ASYM_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal", _wrap_Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SCHEME_KEYEDHASH_Marshal(TPMU_SCHEME_KEYEDHASH src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal", _wrap_Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SCHEME_KEYEDHASH_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_SCHEME_KEYEDHASH dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SIGNATURE_Marshal", _wrap_Tss2_MU_TPMU_SIGNATURE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SIGNATURE_Marshal(TPMU_SIGNATURE src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SIGNATURE_Unmarshal", _wrap_Tss2_MU_TPMU_SIGNATURE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SIGNATURE_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_SIGNATURE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal", _wrap_Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Marshal(TPMU_SENSITIVE_COMPOSITE src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal", _wrap_Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_SENSITIVE_COMPOSITE_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_SENSITIVE_COMPOSITE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal", _wrap_Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_ENCRYPTED_SECRET_Marshal(TPMU_ENCRYPTED_SECRET src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal", _wrap_Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_ENCRYPTED_SECRET_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_ENCRYPTED_SECRET dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_PUBLIC_PARMS_Marshal", _wrap_Tss2_MU_TPMU_PUBLIC_PARMS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_PUBLIC_PARMS_Marshal(TPMU_PUBLIC_PARMS src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal", _wrap_Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_PUBLIC_PARMS_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_PUBLIC_PARMS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_PUBLIC_ID_Marshal", _wrap_Tss2_MU_TPMU_PUBLIC_ID_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_PUBLIC_ID_Marshal(TPMU_PUBLIC_ID src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_PUBLIC_ID_Unmarshal", _wrap_Tss2_MU_TPMU_PUBLIC_ID_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_PUBLIC_ID_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_PUBLIC_ID dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_NAME_Marshal", _wrap_Tss2_MU_TPMU_NAME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_NAME_Marshal(TPMU_NAME src, uint32_t selector_value, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMU_NAME_Unmarshal", _wrap_Tss2_MU_TPMU_NAME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMU_NAME_Unmarshal(uint8_t const [] buffer, size_t * offset, uint32_t selector_value, TPMU_NAME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_HA_Marshal", _wrap_Tss2_MU_TPMT_HA_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_HA_Marshal(TPMT_HA src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_HA_Unmarshal", _wrap_Tss2_MU_TPMT_HA_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_HA_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_HA dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SYM_DEF_Marshal", _wrap_Tss2_MU_TPMT_SYM_DEF_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SYM_DEF_Marshal(TPMT_SYM_DEF src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SYM_DEF_Unmarshal", _wrap_Tss2_MU_TPMT_SYM_DEF_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SYM_DEF_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_SYM_DEF dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal", _wrap_Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SYM_DEF_OBJECT_Marshal(TPMT_SYM_DEF_OBJECT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal", _wrap_Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SYM_DEF_OBJECT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_SYM_DEF_OBJECT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal", _wrap_Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_KEYEDHASH_SCHEME_Marshal(TPMT_KEYEDHASH_SCHEME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_KEYEDHASH_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_KEYEDHASH_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SIG_SCHEME_Marshal", _wrap_Tss2_MU_TPMT_SIG_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SIG_SCHEME_Marshal(TPMT_SIG_SCHEME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SIG_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMT_SIG_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SIG_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_SIG_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_KDF_SCHEME_Marshal", _wrap_Tss2_MU_TPMT_KDF_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_KDF_SCHEME_Marshal(TPMT_KDF_SCHEME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_KDF_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMT_KDF_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_KDF_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_KDF_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_ASYM_SCHEME_Marshal", _wrap_Tss2_MU_TPMT_ASYM_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_ASYM_SCHEME_Marshal(TPMT_ASYM_SCHEME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_ASYM_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_ASYM_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_RSA_SCHEME_Marshal", _wrap_Tss2_MU_TPMT_RSA_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_RSA_SCHEME_Marshal(TPMT_RSA_SCHEME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_RSA_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMT_RSA_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_RSA_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_RSA_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_RSA_DECRYPT_Marshal", _wrap_Tss2_MU_TPMT_RSA_DECRYPT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_RSA_DECRYPT_Marshal(TPMT_RSA_DECRYPT src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal", _wrap_Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_RSA_DECRYPT_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_RSA_DECRYPT dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_ECC_SCHEME_Marshal", _wrap_Tss2_MU_TPMT_ECC_SCHEME_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_ECC_SCHEME_Marshal(TPMT_ECC_SCHEME src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_ECC_SCHEME_Unmarshal", _wrap_Tss2_MU_TPMT_ECC_SCHEME_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_ECC_SCHEME_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_ECC_SCHEME dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SIGNATURE_Marshal", _wrap_Tss2_MU_TPMT_SIGNATURE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SIGNATURE_Marshal(TPMT_SIGNATURE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SIGNATURE_Unmarshal", _wrap_Tss2_MU_TPMT_SIGNATURE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SIGNATURE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_SIGNATURE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SENSITIVE_Marshal", _wrap_Tss2_MU_TPMT_SENSITIVE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SENSITIVE_Marshal(TPMT_SENSITIVE src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_SENSITIVE_Unmarshal", _wrap_Tss2_MU_TPMT_SENSITIVE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_SENSITIVE_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_SENSITIVE dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_PUBLIC_Marshal", _wrap_Tss2_MU_TPMT_PUBLIC_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_PUBLIC_Marshal(TPMT_PUBLIC src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_PUBLIC_Unmarshal", _wrap_Tss2_MU_TPMT_PUBLIC_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_PUBLIC_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_PUBLIC dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_PUBLIC_PARMS_Marshal", _wrap_Tss2_MU_TPMT_PUBLIC_PARMS_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_PUBLIC_PARMS_Marshal(TPMT_PUBLIC_PARMS src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal", _wrap_Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_PUBLIC_PARMS_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_PUBLIC_PARMS dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_CREATION_Marshal", _wrap_Tss2_MU_TPMT_TK_CREATION_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_CREATION_Marshal(TPMT_TK_CREATION src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_CREATION_Unmarshal", _wrap_Tss2_MU_TPMT_TK_CREATION_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_CREATION_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_TK_CREATION dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_VERIFIED_Marshal", _wrap_Tss2_MU_TPMT_TK_VERIFIED_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_VERIFIED_Marshal(TPMT_TK_VERIFIED src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_VERIFIED_Unmarshal", _wrap_Tss2_MU_TPMT_TK_VERIFIED_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_VERIFIED_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_TK_VERIFIED dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_AUTH_Marshal", _wrap_Tss2_MU_TPMT_TK_AUTH_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_AUTH_Marshal(TPMT_TK_AUTH src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_AUTH_Unmarshal", _wrap_Tss2_MU_TPMT_TK_AUTH_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_AUTH_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_TK_AUTH dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_HASHCHECK_Marshal", _wrap_Tss2_MU_TPMT_TK_HASHCHECK_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_HASHCHECK_Marshal(TPMT_TK_HASHCHECK src, uint8_t [] buffer, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal", _wrap_Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMT_TK_HASHCHECK_Unmarshal(uint8_t const [] buffer, size_t * offset, TPMT_TK_HASHCHECK dest) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_HANDLE_Marshal", _wrap_Tss2_MU_TPM2_HANDLE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_HANDLE_Marshal(TPM2_HANDLE arg1, uint8_t * buffer, size_t size, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_HANDLE_Unmarshal", _wrap_Tss2_MU_TPM2_HANDLE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_HANDLE_Unmarshal(uint8_t const [] buffer, size_t size, size_t * offset, TPM2_HANDLE * out) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMI_ALG_HASH_Marshal", _wrap_Tss2_MU_TPMI_ALG_HASH_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMI_ALG_HASH_Marshal(TPMI_ALG_HASH arg1, uint8_t * buffer, size_t size, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMI_ALG_HASH_Unmarshal", _wrap_Tss2_MU_TPMI_ALG_HASH_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMI_ALG_HASH_Unmarshal(uint8_t const [] buffer, size_t size, size_t * offset, TPMI_ALG_HASH * out) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_SE_Marshal", _wrap_Tss2_MU_TPM2_SE_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_SE_Marshal(TPM2_SE arg1, uint8_t * buffer, size_t size, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_SE_Unmarshal", _wrap_Tss2_MU_TPM2_SE_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_SE_Unmarshal(uint8_t const [] buffer, size_t size, size_t * offset, TPM2_SE * out) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_NT_Marshal", _wrap_Tss2_MU_TPM2_NT_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_NT_Marshal(TPM2_NT arg1, uint8_t * buffer, size_t size, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPM2_NT_Unmarshal", _wrap_Tss2_MU_TPM2_NT_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPM2_NT_Unmarshal(uint8_t const [] buffer, size_t size, size_t * offset, TPM2_NT * out) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_EMPTY_Marshal", _wrap_Tss2_MU_TPMS_EMPTY_Marshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_EMPTY_Marshal(TPMS_EMPTY arg1, uint8_t * buffer, size_t size, size_t * offset) -> TSS2_RC"},
	 { (char *)"Tss2_MU_TPMS_EMPTY_Unmarshal", _wrap_Tss2_MU_TPMS_EMPTY_Unmarshal, METH_VARARGS, (char *)"Tss2_MU_TPMS_EMPTY_Unmarshal(uint8_t const [] buffer, size_t size, size_t * offset, TPMS_EMPTY out) -> TSS2_RC"},
	 { (char *)"new_ESYS_TR_PTR", _wrap_new_ESYS_TR_PTR, METH_VARARGS, (char *)"new_ESYS_TR_PTR() -> ESYS_TR_PTR"},
	 { (char *)"delete_ESYS_TR_PTR", _wrap_delete_ESYS_TR_PTR, METH_VARARGS, (char *)"delete_ESYS_TR_PTR(ESYS_TR_PTR self)"},
	 { (char *)"ESYS_TR_PTR_assign", _wrap_ESYS_TR_PTR_assign, METH_VARARGS, (char *)"ESYS_TR_PTR_assign(ESYS_TR_PTR self, unsigned int value)"},
	 { (char *)"ESYS_TR_PTR_value", _wrap_ESYS_TR_PTR_value, METH_VARARGS, (char *)"ESYS_TR_PTR_value(ESYS_TR_PTR self) -> unsigned int"},
	 { (char *)"ESYS_TR_PTR_cast", _wrap_ESYS_TR_PTR_cast, METH_VARARGS, (char *)"ESYS_TR_PTR_cast(ESYS_TR_PTR self) -> unsigned int *"},
	 { (char *)"ESYS_TR_PTR_frompointer", _wrap_ESYS_TR_PTR_frompointer, METH_VARARGS, (char *)"ESYS_TR_PTR_frompointer(unsigned int * t) -> ESYS_TR_PTR"},
	 { (char *)"ESYS_TR_PTR_swigregister", ESYS_TR_PTR_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_ESYS_TR_PTR_PTR", _wrap_new_ESYS_TR_PTR_PTR, METH_VARARGS, (char *)"new_ESYS_TR_PTR_PTR() -> unsigned int **"},
	 { (char *)"copy_ESYS_TR_PTR_PTR", _wrap_copy_ESYS_TR_PTR_PTR, METH_VARARGS, (char *)"copy_ESYS_TR_PTR_PTR(unsigned int * value) -> unsigned int **"},
	 { (char *)"delete_ESYS_TR_PTR_PTR", _wrap_delete_ESYS_TR_PTR_PTR, METH_VARARGS, (char *)"delete_ESYS_TR_PTR_PTR(unsigned int ** obj)"},
	 { (char *)"ESYS_TR_PTR_PTR_assign", _wrap_ESYS_TR_PTR_PTR_assign, METH_VARARGS, (char *)"ESYS_TR_PTR_PTR_assign(unsigned int ** obj, unsigned int * value)"},
	 { (char *)"ESYS_TR_PTR_PTR_value", _wrap_ESYS_TR_PTR_PTR_value, METH_VARARGS, (char *)"ESYS_TR_PTR_PTR_value(unsigned int ** obj) -> unsigned int *"},
	 { (char *)"new_ctx_ptr", _wrap_new_ctx_ptr, METH_VARARGS, (char *)"new_ctx_ptr() -> struct ESYS_CONTEXT **"},
	 { (char *)"copy_ctx_ptr", _wrap_copy_ctx_ptr, METH_VARARGS, (char *)"copy_ctx_ptr(struct ESYS_CONTEXT * value) -> struct ESYS_CONTEXT **"},
	 { (char *)"delete_ctx_ptr", _wrap_delete_ctx_ptr, METH_VARARGS, (char *)"delete_ctx_ptr(struct ESYS_CONTEXT ** obj)"},
	 { (char *)"ctx_ptr_assign", _wrap_ctx_ptr_assign, METH_VARARGS, (char *)"ctx_ptr_assign(struct ESYS_CONTEXT ** obj, struct ESYS_CONTEXT * value)"},
	 { (char *)"ctx_ptr_value", _wrap_ctx_ptr_value, METH_VARARGS, (char *)"ctx_ptr_value(struct ESYS_CONTEXT ** obj) -> struct ESYS_CONTEXT *"},
	 { (char *)"new_tcti_ctx_ptr_ptr", _wrap_new_tcti_ctx_ptr_ptr, METH_VARARGS, (char *)"new_tcti_ctx_ptr_ptr() -> struct TSS2_TCTI_CONTEXT **"},
	 { (char *)"copy_tcti_ctx_ptr_ptr", _wrap_copy_tcti_ctx_ptr_ptr, METH_VARARGS, (char *)"copy_tcti_ctx_ptr_ptr(struct TSS2_TCTI_CONTEXT * value) -> struct TSS2_TCTI_CONTEXT **"},
	 { (char *)"delete_tcti_ctx_ptr_ptr", _wrap_delete_tcti_ctx_ptr_ptr, METH_VARARGS, (char *)"delete_tcti_ctx_ptr_ptr(struct TSS2_TCTI_CONTEXT ** obj)"},
	 { (char *)"tcti_ctx_ptr_ptr_assign", _wrap_tcti_ctx_ptr_ptr_assign, METH_VARARGS, (char *)"tcti_ctx_ptr_ptr_assign(struct TSS2_TCTI_CONTEXT ** obj, struct TSS2_TCTI_CONTEXT * value)"},
	 { (char *)"tcti_ctx_ptr_ptr_value", _wrap_tcti_ctx_ptr_ptr_value, METH_VARARGS, (char *)"tcti_ctx_ptr_ptr_value(struct TSS2_TCTI_CONTEXT ** obj) -> struct TSS2_TCTI_CONTEXT *"},
	 { (char *)"Tss2_TctiLdr_Initialize_Ex", _wrap_Tss2_TctiLdr_Initialize_Ex, METH_VARARGS, (char *)"Tss2_TctiLdr_Initialize_Ex(char const * name, char const * conf, TSS2_TCTI_CONTEXT ** context) -> TSS2_RC"},
	 { (char *)"Esys_Initialize", _wrap_Esys_Initialize, METH_VARARGS, (char *)"Esys_Initialize(ESYS_CONTEXT ** esys_context, TSS2_TCTI_CONTEXT * tcti, TSS2_ABI_VERSION abiVersion) -> TSS2_RC"},
	 { (char *)"Esys_Finalize", _wrap_Esys_Finalize, METH_VARARGS, (char *)"Esys_Finalize(ESYS_CONTEXT ** context)"},
	 { (char *)"Esys_GetTcti", _wrap_Esys_GetTcti, METH_VARARGS, (char *)"Esys_GetTcti(ESYS_CONTEXT * esys_context, TSS2_TCTI_CONTEXT ** tcti) -> TSS2_RC"},
	 { (char *)"Esys_GetPollHandles", _wrap_Esys_GetPollHandles, METH_VARARGS, (char *)"Esys_GetPollHandles(ESYS_CONTEXT * esys_context, TSS2_TCTI_POLL_HANDLE ** handles, size_t * count) -> TSS2_RC"},
	 { (char *)"Esys_SetTimeout", _wrap_Esys_SetTimeout, METH_VARARGS, (char *)"Esys_SetTimeout(ESYS_CONTEXT * esys_context, int32_t timeout) -> TSS2_RC"},
	 { (char *)"Esys_TR_Serialize", _wrap_Esys_TR_Serialize, METH_VARARGS, (char *)"Esys_TR_Serialize(ESYS_CONTEXT * esys_context, unsigned int object, uint8_t ** buffer, size_t * buffer_size) -> TSS2_RC"},
	 { (char *)"Esys_TR_Deserialize", _wrap_Esys_TR_Deserialize, METH_VARARGS, (char *)"Esys_TR_Deserialize(ESYS_CONTEXT * esys_context, uint8_t const * buffer, size_t buffer_size, unsigned int * esys_handle) -> TSS2_RC"},
	 { (char *)"Esys_TR_FromTPMPublic_Async", _wrap_Esys_TR_FromTPMPublic_Async, METH_VARARGS, (char *)"Esys_TR_FromTPMPublic_Async(ESYS_CONTEXT * esysContext, TPM2_HANDLE tpm_handle, unsigned int optionalSession1, unsigned int optionalSession2, unsigned int optionalSession3) -> TSS2_RC"},
	 { (char *)"Esys_TR_FromTPMPublic_Finish", _wrap_Esys_TR_FromTPMPublic_Finish, METH_VARARGS, (char *)"Esys_TR_FromTPMPublic_Finish(ESYS_CONTEXT * esysContext, unsigned int * object) -> TSS2_RC"},
	 { (char *)"Esys_TR_FromTPMPublic", _wrap_Esys_TR_FromTPMPublic, METH_VARARGS, (char *)"Esys_TR_FromTPMPublic(ESYS_CONTEXT * esysContext, TPM2_HANDLE tpm_handle, unsigned int optionalSession1, unsigned int optionalSession2, unsigned int optionalSession3, unsigned int * object) -> TSS2_RC"},
	 { (char *)"Esys_TR_Close", _wrap_Esys_TR_Close, METH_VARARGS, (char *)"Esys_TR_Close(ESYS_CONTEXT * esys_context, unsigned int * rsrc_handle) -> TSS2_RC"},
	 { (char *)"Esys_TR_SetAuth", _wrap_Esys_TR_SetAuth, METH_VARARGS, (char *)"Esys_TR_SetAuth(ESYS_CONTEXT * esysContext, unsigned int handle, TPM2B_DIGEST authValue) -> TSS2_RC"},
	 { (char *)"Esys_TR_GetName", _wrap_Esys_TR_GetName, METH_VARARGS, (char *)"Esys_TR_GetName(ESYS_CONTEXT * esysContext, unsigned int handle, TPM2B_NAME ** name) -> TSS2_RC"},
	 { (char *)"Esys_TRSess_GetAttributes", _wrap_Esys_TRSess_GetAttributes, METH_VARARGS, (char *)"Esys_TRSess_GetAttributes(ESYS_CONTEXT * esysContext, unsigned int session, TPMA_SESSION * flags) -> TSS2_RC"},
	 { (char *)"Esys_TRSess_SetAttributes", _wrap_Esys_TRSess_SetAttributes, METH_VARARGS, (char *)"Esys_TRSess_SetAttributes(ESYS_CONTEXT * esysContext, unsigned int session, TPMA_SESSION flags, TPMA_SESSION mask) -> TSS2_RC"},
	 { (char *)"Esys_TRSess_GetNonceTPM", _wrap_Esys_TRSess_GetNonceTPM, METH_VARARGS, (char *)"Esys_TRSess_GetNonceTPM(ESYS_CONTEXT * esysContext, unsigned int session, TPM2B_NONCE ** nonceTPM) -> TSS2_RC"},
	 { (char *)"Esys_Startup", _wrap_Esys_Startup, METH_VARARGS, (char *)"Esys_Startup(ESYS_CONTEXT * esysContext, TPM2_SU startupType) -> TSS2_RC"},
	 { (char *)"Esys_Startup_Async", _wrap_Esys_Startup_Async, METH_VARARGS, (char *)"Esys_Startup_Async(ESYS_CONTEXT * esysContext, TPM2_SU startupType) -> TSS2_RC"},
	 { (char *)"Esys_Startup_Finish", _wrap_Esys_Startup_Finish, METH_VARARGS, (char *)"Esys_Startup_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_Shutdown", _wrap_Esys_Shutdown, METH_VARARGS, (char *)"Esys_Shutdown(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_SU shutdownType) -> TSS2_RC"},
	 { (char *)"Esys_Shutdown_Async", _wrap_Esys_Shutdown_Async, METH_VARARGS, (char *)"Esys_Shutdown_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_SU shutdownType) -> TSS2_RC"},
	 { (char *)"Esys_Shutdown_Finish", _wrap_Esys_Shutdown_Finish, METH_VARARGS, (char *)"Esys_Shutdown_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_SelfTest", _wrap_Esys_SelfTest, METH_VARARGS, (char *)"Esys_SelfTest(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO fullTest) -> TSS2_RC"},
	 { (char *)"Esys_SelfTest_Async", _wrap_Esys_SelfTest_Async, METH_VARARGS, (char *)"Esys_SelfTest_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO fullTest) -> TSS2_RC"},
	 { (char *)"Esys_SelfTest_Finish", _wrap_Esys_SelfTest_Finish, METH_VARARGS, (char *)"Esys_SelfTest_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_IncrementalSelfTest", _wrap_Esys_IncrementalSelfTest, METH_VARARGS, (char *)"Esys_IncrementalSelfTest(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_ALG toTest, TPML_ALG ** toDoList) -> TSS2_RC"},
	 { (char *)"Esys_IncrementalSelfTest_Async", _wrap_Esys_IncrementalSelfTest_Async, METH_VARARGS, (char *)"Esys_IncrementalSelfTest_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_ALG toTest) -> TSS2_RC"},
	 { (char *)"Esys_IncrementalSelfTest_Finish", _wrap_Esys_IncrementalSelfTest_Finish, METH_VARARGS, (char *)"Esys_IncrementalSelfTest_Finish(ESYS_CONTEXT * esysContext, TPML_ALG ** toDoList) -> TSS2_RC"},
	 { (char *)"Esys_GetTestResult", _wrap_Esys_GetTestResult, METH_VARARGS, (char *)"Esys_GetTestResult(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER ** outData, TPM2_RC * testResult) -> TSS2_RC"},
	 { (char *)"Esys_GetTestResult_Async", _wrap_Esys_GetTestResult_Async, METH_VARARGS, (char *)"Esys_GetTestResult_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_GetTestResult_Finish", _wrap_Esys_GetTestResult_Finish, METH_VARARGS, (char *)"Esys_GetTestResult_Finish(ESYS_CONTEXT * esysContext, TPM2B_MAX_BUFFER ** outData, TPM2_RC * testResult) -> TSS2_RC"},
	 { (char *)"Esys_StartAuthSession", _wrap_Esys_StartAuthSession, METH_VARARGS, (char *)"Esys_StartAuthSession(ESYS_CONTEXT * esysContext, unsigned int tpmKey, unsigned int bind, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nonceCaller, TPM2_SE sessionType, TPMT_SYM_DEF symmetric, TPMI_ALG_HASH authHash, unsigned int * sessionHandle) -> TSS2_RC"},
	 { (char *)"Esys_StartAuthSession_Async", _wrap_Esys_StartAuthSession_Async, METH_VARARGS, (char *)"Esys_StartAuthSession_Async(ESYS_CONTEXT * esysContext, unsigned int tpmKey, unsigned int bind, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nonceCaller, TPM2_SE sessionType, TPMT_SYM_DEF symmetric, TPMI_ALG_HASH authHash) -> TSS2_RC"},
	 { (char *)"Esys_StartAuthSession_Finish", _wrap_Esys_StartAuthSession_Finish, METH_VARARGS, (char *)"Esys_StartAuthSession_Finish(ESYS_CONTEXT * esysContext, unsigned int * sessionHandle) -> TSS2_RC"},
	 { (char *)"Esys_PolicyRestart", _wrap_Esys_PolicyRestart, METH_VARARGS, (char *)"Esys_PolicyRestart(ESYS_CONTEXT * esysContext, unsigned int sessionHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyRestart_Async", _wrap_Esys_PolicyRestart_Async, METH_VARARGS, (char *)"Esys_PolicyRestart_Async(ESYS_CONTEXT * esysContext, unsigned int sessionHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyRestart_Finish", _wrap_Esys_PolicyRestart_Finish, METH_VARARGS, (char *)"Esys_PolicyRestart_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_Create", _wrap_Esys_Create, METH_VARARGS, (char *)"Esys_Create(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_CREATE inSensitive, TPM2B_PUBLIC inPublic, TPM2B_DATA outsideInfo, TPML_PCR_SELECTION creationPCR, TPM2B_PRIVATE ** outPrivate, TPM2B_PUBLIC ** outPublic, TPM2B_CREATION_DATA ** creationData, TPM2B_DIGEST ** creationHash, TPMT_TK_CREATION ** creationTicket) -> TSS2_RC"},
	 { (char *)"Esys_Create_Async", _wrap_Esys_Create_Async, METH_VARARGS, (char *)"Esys_Create_Async(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_CREATE inSensitive, TPM2B_PUBLIC inPublic, TPM2B_DATA outsideInfo, TPML_PCR_SELECTION creationPCR) -> TSS2_RC"},
	 { (char *)"Esys_Create_Finish", _wrap_Esys_Create_Finish, METH_VARARGS, (char *)"Esys_Create_Finish(ESYS_CONTEXT * esysContext, TPM2B_PRIVATE ** outPrivate, TPM2B_PUBLIC ** outPublic, TPM2B_CREATION_DATA ** creationData, TPM2B_DIGEST ** creationHash, TPMT_TK_CREATION ** creationTicket) -> TSS2_RC"},
	 { (char *)"Esys_Load", _wrap_Esys_Load, METH_VARARGS, (char *)"Esys_Load(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PRIVATE inPrivate, TPM2B_PUBLIC inPublic, unsigned int * objectHandle) -> TSS2_RC"},
	 { (char *)"Esys_Load_Async", _wrap_Esys_Load_Async, METH_VARARGS, (char *)"Esys_Load_Async(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PRIVATE inPrivate, TPM2B_PUBLIC inPublic) -> TSS2_RC"},
	 { (char *)"Esys_Load_Finish", _wrap_Esys_Load_Finish, METH_VARARGS, (char *)"Esys_Load_Finish(ESYS_CONTEXT * esysContext, unsigned int * objectHandle) -> TSS2_RC"},
	 { (char *)"Esys_LoadExternal", _wrap_Esys_LoadExternal, METH_VARARGS, (char *)"Esys_LoadExternal(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE inPrivate, TPM2B_PUBLIC inPublic, TPMI_RH_HIERARCHY hierarchy, unsigned int * objectHandle) -> TSS2_RC"},
	 { (char *)"Esys_LoadExternal_Async", _wrap_Esys_LoadExternal_Async, METH_VARARGS, (char *)"Esys_LoadExternal_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE inPrivate, TPM2B_PUBLIC inPublic, TPMI_RH_HIERARCHY hierarchy) -> TSS2_RC"},
	 { (char *)"Esys_LoadExternal_Finish", _wrap_Esys_LoadExternal_Finish, METH_VARARGS, (char *)"Esys_LoadExternal_Finish(ESYS_CONTEXT * esysContext, unsigned int * objectHandle) -> TSS2_RC"},
	 { (char *)"Esys_ReadPublic", _wrap_Esys_ReadPublic, METH_VARARGS, (char *)"Esys_ReadPublic(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PUBLIC ** outPublic, TPM2B_NAME ** name, TPM2B_NAME ** qualifiedName) -> TSS2_RC"},
	 { (char *)"Esys_ReadPublic_Async", _wrap_Esys_ReadPublic_Async, METH_VARARGS, (char *)"Esys_ReadPublic_Async(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ReadPublic_Finish", _wrap_Esys_ReadPublic_Finish, METH_VARARGS, (char *)"Esys_ReadPublic_Finish(ESYS_CONTEXT * esysContext, TPM2B_PUBLIC ** outPublic, TPM2B_NAME ** name, TPM2B_NAME ** qualifiedName) -> TSS2_RC"},
	 { (char *)"Esys_ActivateCredential", _wrap_Esys_ActivateCredential, METH_VARARGS, (char *)"Esys_ActivateCredential(ESYS_CONTEXT * esysContext, unsigned int activateHandle, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ID_OBJECT credentialBlob, TPM2B_ENCRYPTED_SECRET secret, TPM2B_DIGEST ** certInfo) -> TSS2_RC"},
	 { (char *)"Esys_ActivateCredential_Async", _wrap_Esys_ActivateCredential_Async, METH_VARARGS, (char *)"Esys_ActivateCredential_Async(ESYS_CONTEXT * esysContext, unsigned int activateHandle, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ID_OBJECT credentialBlob, TPM2B_ENCRYPTED_SECRET secret) -> TSS2_RC"},
	 { (char *)"Esys_ActivateCredential_Finish", _wrap_Esys_ActivateCredential_Finish, METH_VARARGS, (char *)"Esys_ActivateCredential_Finish(ESYS_CONTEXT * esysContext, TPM2B_DIGEST ** certInfo) -> TSS2_RC"},
	 { (char *)"Esys_MakeCredential", _wrap_Esys_MakeCredential, METH_VARARGS, (char *)"Esys_MakeCredential(ESYS_CONTEXT * esysContext, unsigned int handle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST credential, TPM2B_NAME objectName, TPM2B_ID_OBJECT ** credentialBlob, TPM2B_ENCRYPTED_SECRET ** secret) -> TSS2_RC"},
	 { (char *)"Esys_MakeCredential_Async", _wrap_Esys_MakeCredential_Async, METH_VARARGS, (char *)"Esys_MakeCredential_Async(ESYS_CONTEXT * esysContext, unsigned int handle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST credential, TPM2B_NAME objectName) -> TSS2_RC"},
	 { (char *)"Esys_MakeCredential_Finish", _wrap_Esys_MakeCredential_Finish, METH_VARARGS, (char *)"Esys_MakeCredential_Finish(ESYS_CONTEXT * esysContext, TPM2B_ID_OBJECT ** credentialBlob, TPM2B_ENCRYPTED_SECRET ** secret) -> TSS2_RC"},
	 { (char *)"Esys_Unseal", _wrap_Esys_Unseal, METH_VARARGS, (char *)"Esys_Unseal(ESYS_CONTEXT * esysContext, unsigned int itemHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_DATA ** outData) -> TSS2_RC"},
	 { (char *)"Esys_Unseal_Async", _wrap_Esys_Unseal_Async, METH_VARARGS, (char *)"Esys_Unseal_Async(ESYS_CONTEXT * esysContext, unsigned int itemHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_Unseal_Finish", _wrap_Esys_Unseal_Finish, METH_VARARGS, (char *)"Esys_Unseal_Finish(ESYS_CONTEXT * esysContext, TPM2B_SENSITIVE_DATA ** outData) -> TSS2_RC"},
	 { (char *)"Esys_ObjectChangeAuth", _wrap_Esys_ObjectChangeAuth, METH_VARARGS, (char *)"Esys_ObjectChangeAuth(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST newAuth, TPM2B_PRIVATE ** outPrivate) -> TSS2_RC"},
	 { (char *)"Esys_ObjectChangeAuth_Async", _wrap_Esys_ObjectChangeAuth_Async, METH_VARARGS, (char *)"Esys_ObjectChangeAuth_Async(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST newAuth) -> TSS2_RC"},
	 { (char *)"Esys_ObjectChangeAuth_Finish", _wrap_Esys_ObjectChangeAuth_Finish, METH_VARARGS, (char *)"Esys_ObjectChangeAuth_Finish(ESYS_CONTEXT * esysContext, TPM2B_PRIVATE ** outPrivate) -> TSS2_RC"},
	 { (char *)"Esys_CreateLoaded", _wrap_Esys_CreateLoaded, METH_VARARGS, (char *)"Esys_CreateLoaded(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_CREATE inSensitive, TPM2B_TEMPLATE inPublic, unsigned int * objectHandle, TPM2B_PRIVATE ** outPrivate, TPM2B_PUBLIC ** outPublic) -> TSS2_RC"},
	 { (char *)"Esys_CreateLoaded_Async", _wrap_Esys_CreateLoaded_Async, METH_VARARGS, (char *)"Esys_CreateLoaded_Async(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_CREATE inSensitive, TPM2B_TEMPLATE inPublic) -> TSS2_RC"},
	 { (char *)"Esys_CreateLoaded_Finish", _wrap_Esys_CreateLoaded_Finish, METH_VARARGS, (char *)"Esys_CreateLoaded_Finish(ESYS_CONTEXT * esysContext, unsigned int * objectHandle, TPM2B_PRIVATE ** outPrivate, TPM2B_PUBLIC ** outPublic) -> TSS2_RC"},
	 { (char *)"Esys_Duplicate", _wrap_Esys_Duplicate, METH_VARARGS, (char *)"Esys_Duplicate(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int newParentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA encryptionKeyIn, TPMT_SYM_DEF_OBJECT symmetricAlg, TPM2B_DATA ** encryptionKeyOut, TPM2B_PRIVATE ** duplicate, TPM2B_ENCRYPTED_SECRET ** outSymSeed) -> TSS2_RC"},
	 { (char *)"Esys_Duplicate_Async", _wrap_Esys_Duplicate_Async, METH_VARARGS, (char *)"Esys_Duplicate_Async(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int newParentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA encryptionKeyIn, TPMT_SYM_DEF_OBJECT symmetricAlg) -> TSS2_RC"},
	 { (char *)"Esys_Duplicate_Finish", _wrap_Esys_Duplicate_Finish, METH_VARARGS, (char *)"Esys_Duplicate_Finish(ESYS_CONTEXT * esysContext, TPM2B_DATA ** encryptionKeyOut, TPM2B_PRIVATE ** duplicate, TPM2B_ENCRYPTED_SECRET ** outSymSeed) -> TSS2_RC"},
	 { (char *)"Esys_Rewrap", _wrap_Esys_Rewrap, METH_VARARGS, (char *)"Esys_Rewrap(ESYS_CONTEXT * esysContext, unsigned int oldParent, unsigned int newParent, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PRIVATE inDuplicate, TPM2B_NAME name, TPM2B_ENCRYPTED_SECRET inSymSeed, TPM2B_PRIVATE ** outDuplicate, TPM2B_ENCRYPTED_SECRET ** outSymSeed) -> TSS2_RC"},
	 { (char *)"Esys_Rewrap_Async", _wrap_Esys_Rewrap_Async, METH_VARARGS, (char *)"Esys_Rewrap_Async(ESYS_CONTEXT * esysContext, unsigned int oldParent, unsigned int newParent, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PRIVATE inDuplicate, TPM2B_NAME name, TPM2B_ENCRYPTED_SECRET inSymSeed) -> TSS2_RC"},
	 { (char *)"Esys_Rewrap_Finish", _wrap_Esys_Rewrap_Finish, METH_VARARGS, (char *)"Esys_Rewrap_Finish(ESYS_CONTEXT * esysContext, TPM2B_PRIVATE ** outDuplicate, TPM2B_ENCRYPTED_SECRET ** outSymSeed) -> TSS2_RC"},
	 { (char *)"Esys_Import", _wrap_Esys_Import, METH_VARARGS, (char *)"Esys_Import(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA encryptionKey, TPM2B_PUBLIC objectPublic, TPM2B_PRIVATE duplicate, TPM2B_ENCRYPTED_SECRET inSymSeed, TPMT_SYM_DEF_OBJECT symmetricAlg, TPM2B_PRIVATE ** outPrivate) -> TSS2_RC"},
	 { (char *)"Esys_Import_Async", _wrap_Esys_Import_Async, METH_VARARGS, (char *)"Esys_Import_Async(ESYS_CONTEXT * esysContext, unsigned int parentHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA encryptionKey, TPM2B_PUBLIC objectPublic, TPM2B_PRIVATE duplicate, TPM2B_ENCRYPTED_SECRET inSymSeed, TPMT_SYM_DEF_OBJECT symmetricAlg) -> TSS2_RC"},
	 { (char *)"Esys_Import_Finish", _wrap_Esys_Import_Finish, METH_VARARGS, (char *)"Esys_Import_Finish(ESYS_CONTEXT * esysContext, TPM2B_PRIVATE ** outPrivate) -> TSS2_RC"},
	 { (char *)"Esys_RSA_Encrypt", _wrap_Esys_RSA_Encrypt, METH_VARARGS, (char *)"Esys_RSA_Encrypt(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PUBLIC_KEY_RSA message, TPMT_RSA_DECRYPT inScheme, TPM2B_DATA label, TPM2B_PUBLIC_KEY_RSA ** outData) -> TSS2_RC"},
	 { (char *)"Esys_RSA_Encrypt_Async", _wrap_Esys_RSA_Encrypt_Async, METH_VARARGS, (char *)"Esys_RSA_Encrypt_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PUBLIC_KEY_RSA message, TPMT_RSA_DECRYPT inScheme, TPM2B_DATA label) -> TSS2_RC"},
	 { (char *)"Esys_RSA_Encrypt_Finish", _wrap_Esys_RSA_Encrypt_Finish, METH_VARARGS, (char *)"Esys_RSA_Encrypt_Finish(ESYS_CONTEXT * esysContext, TPM2B_PUBLIC_KEY_RSA ** outData) -> TSS2_RC"},
	 { (char *)"Esys_RSA_Decrypt", _wrap_Esys_RSA_Decrypt, METH_VARARGS, (char *)"Esys_RSA_Decrypt(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PUBLIC_KEY_RSA cipherText, TPMT_RSA_DECRYPT inScheme, TPM2B_DATA label, TPM2B_PUBLIC_KEY_RSA ** message) -> TSS2_RC"},
	 { (char *)"Esys_RSA_Decrypt_Async", _wrap_Esys_RSA_Decrypt_Async, METH_VARARGS, (char *)"Esys_RSA_Decrypt_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_PUBLIC_KEY_RSA cipherText, TPMT_RSA_DECRYPT inScheme, TPM2B_DATA label) -> TSS2_RC"},
	 { (char *)"Esys_RSA_Decrypt_Finish", _wrap_Esys_RSA_Decrypt_Finish, METH_VARARGS, (char *)"Esys_RSA_Decrypt_Finish(ESYS_CONTEXT * esysContext, TPM2B_PUBLIC_KEY_RSA ** message) -> TSS2_RC"},
	 { (char *)"Esys_ECDH_KeyGen", _wrap_Esys_ECDH_KeyGen, METH_VARARGS, (char *)"Esys_ECDH_KeyGen(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT ** zPoint, TPM2B_ECC_POINT ** pubPoint) -> TSS2_RC"},
	 { (char *)"Esys_ECDH_KeyGen_Async", _wrap_Esys_ECDH_KeyGen_Async, METH_VARARGS, (char *)"Esys_ECDH_KeyGen_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ECDH_KeyGen_Finish", _wrap_Esys_ECDH_KeyGen_Finish, METH_VARARGS, (char *)"Esys_ECDH_KeyGen_Finish(ESYS_CONTEXT * esysContext, TPM2B_ECC_POINT ** zPoint, TPM2B_ECC_POINT ** pubPoint) -> TSS2_RC"},
	 { (char *)"Esys_ECDH_ZGen", _wrap_Esys_ECDH_ZGen, METH_VARARGS, (char *)"Esys_ECDH_ZGen(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT inPoint, TPM2B_ECC_POINT ** outPoint) -> TSS2_RC"},
	 { (char *)"Esys_ECDH_ZGen_Async", _wrap_Esys_ECDH_ZGen_Async, METH_VARARGS, (char *)"Esys_ECDH_ZGen_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT inPoint) -> TSS2_RC"},
	 { (char *)"Esys_ECDH_ZGen_Finish", _wrap_Esys_ECDH_ZGen_Finish, METH_VARARGS, (char *)"Esys_ECDH_ZGen_Finish(ESYS_CONTEXT * esysContext, TPM2B_ECC_POINT ** outPoint) -> TSS2_RC"},
	 { (char *)"Esys_ECC_Parameters", _wrap_Esys_ECC_Parameters, METH_VARARGS, (char *)"Esys_ECC_Parameters(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_ECC_CURVE curveID, TPMS_ALGORITHM_DETAIL_ECC ** parameters) -> TSS2_RC"},
	 { (char *)"Esys_ECC_Parameters_Async", _wrap_Esys_ECC_Parameters_Async, METH_VARARGS, (char *)"Esys_ECC_Parameters_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_ECC_CURVE curveID) -> TSS2_RC"},
	 { (char *)"Esys_ECC_Parameters_Finish", _wrap_Esys_ECC_Parameters_Finish, METH_VARARGS, (char *)"Esys_ECC_Parameters_Finish(ESYS_CONTEXT * esysContext, TPMS_ALGORITHM_DETAIL_ECC ** parameters) -> TSS2_RC"},
	 { (char *)"Esys_ZGen_2Phase", _wrap_Esys_ZGen_2Phase, METH_VARARGS, (char *)"Esys_ZGen_2Phase(ESYS_CONTEXT * esysContext, unsigned int keyA, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT inQsB, TPM2B_ECC_POINT inQeB, TPMI_ECC_KEY_EXCHANGE inScheme, UINT16 counter, TPM2B_ECC_POINT ** outZ1, TPM2B_ECC_POINT ** outZ2) -> TSS2_RC"},
	 { (char *)"Esys_ZGen_2Phase_Async", _wrap_Esys_ZGen_2Phase_Async, METH_VARARGS, (char *)"Esys_ZGen_2Phase_Async(ESYS_CONTEXT * esysContext, unsigned int keyA, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT inQsB, TPM2B_ECC_POINT inQeB, TPMI_ECC_KEY_EXCHANGE inScheme, UINT16 counter) -> TSS2_RC"},
	 { (char *)"Esys_ZGen_2Phase_Finish", _wrap_Esys_ZGen_2Phase_Finish, METH_VARARGS, (char *)"Esys_ZGen_2Phase_Finish(ESYS_CONTEXT * esysContext, TPM2B_ECC_POINT ** outZ1, TPM2B_ECC_POINT ** outZ2) -> TSS2_RC"},
	 { (char *)"Esys_EncryptDecrypt", _wrap_Esys_EncryptDecrypt, METH_VARARGS, (char *)"Esys_EncryptDecrypt(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO decrypt, TPMI_ALG_SYM_MODE mode, TPM2B_IV ivIn, TPM2B_MAX_BUFFER inData, TPM2B_MAX_BUFFER ** outData, TPM2B_IV ** ivOut) -> TSS2_RC"},
	 { (char *)"Esys_EncryptDecrypt_Async", _wrap_Esys_EncryptDecrypt_Async, METH_VARARGS, (char *)"Esys_EncryptDecrypt_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO decrypt, TPMI_ALG_SYM_MODE mode, TPM2B_IV ivIn, TPM2B_MAX_BUFFER inData) -> TSS2_RC"},
	 { (char *)"Esys_EncryptDecrypt_Finish", _wrap_Esys_EncryptDecrypt_Finish, METH_VARARGS, (char *)"Esys_EncryptDecrypt_Finish(ESYS_CONTEXT * esysContext, TPM2B_MAX_BUFFER ** outData, TPM2B_IV ** ivOut) -> TSS2_RC"},
	 { (char *)"Esys_EncryptDecrypt2", _wrap_Esys_EncryptDecrypt2, METH_VARARGS, (char *)"Esys_EncryptDecrypt2(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER inData, TPMI_YES_NO decrypt, TPMI_ALG_SYM_MODE mode, TPM2B_IV ivIn, TPM2B_MAX_BUFFER ** outData, TPM2B_IV ** ivOut) -> TSS2_RC"},
	 { (char *)"Esys_EncryptDecrypt2_Async", _wrap_Esys_EncryptDecrypt2_Async, METH_VARARGS, (char *)"Esys_EncryptDecrypt2_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER inData, TPMI_YES_NO decrypt, TPMI_ALG_SYM_MODE mode, TPM2B_IV ivIn) -> TSS2_RC"},
	 { (char *)"Esys_EncryptDecrypt2_Finish", _wrap_Esys_EncryptDecrypt2_Finish, METH_VARARGS, (char *)"Esys_EncryptDecrypt2_Finish(ESYS_CONTEXT * esysContext, TPM2B_MAX_BUFFER ** outData, TPM2B_IV ** ivOut) -> TSS2_RC"},
	 { (char *)"Esys_Hash", _wrap_Esys_Hash, METH_VARARGS, (char *)"Esys_Hash(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER data, TPMI_ALG_HASH hashAlg, TPMI_RH_HIERARCHY hierarchy, TPM2B_DIGEST ** outHash, TPMT_TK_HASHCHECK ** validation) -> TSS2_RC"},
	 { (char *)"Esys_Hash_Async", _wrap_Esys_Hash_Async, METH_VARARGS, (char *)"Esys_Hash_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER data, TPMI_ALG_HASH hashAlg, TPMI_RH_HIERARCHY hierarchy) -> TSS2_RC"},
	 { (char *)"Esys_Hash_Finish", _wrap_Esys_Hash_Finish, METH_VARARGS, (char *)"Esys_Hash_Finish(ESYS_CONTEXT * esysContext, TPM2B_DIGEST ** outHash, TPMT_TK_HASHCHECK ** validation) -> TSS2_RC"},
	 { (char *)"Esys_HMAC", _wrap_Esys_HMAC, METH_VARARGS, (char *)"Esys_HMAC(ESYS_CONTEXT * esysContext, unsigned int handle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer, TPMI_ALG_HASH hashAlg, TPM2B_DIGEST ** outHMAC) -> TSS2_RC"},
	 { (char *)"Esys_HMAC_Async", _wrap_Esys_HMAC_Async, METH_VARARGS, (char *)"Esys_HMAC_Async(ESYS_CONTEXT * esysContext, unsigned int handle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer, TPMI_ALG_HASH hashAlg) -> TSS2_RC"},
	 { (char *)"Esys_HMAC_Finish", _wrap_Esys_HMAC_Finish, METH_VARARGS, (char *)"Esys_HMAC_Finish(ESYS_CONTEXT * esysContext, TPM2B_DIGEST ** outHMAC) -> TSS2_RC"},
	 { (char *)"Esys_GetRandom", _wrap_Esys_GetRandom, METH_VARARGS, (char *)"Esys_GetRandom(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT16 bytesRequested, TPM2B_DIGEST ** randomBytes) -> TSS2_RC"},
	 { (char *)"Esys_GetRandom_Async", _wrap_Esys_GetRandom_Async, METH_VARARGS, (char *)"Esys_GetRandom_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT16 bytesRequested) -> TSS2_RC"},
	 { (char *)"Esys_GetRandom_Finish", _wrap_Esys_GetRandom_Finish, METH_VARARGS, (char *)"Esys_GetRandom_Finish(ESYS_CONTEXT * esysContext, TPM2B_DIGEST ** randomBytes) -> TSS2_RC"},
	 { (char *)"Esys_StirRandom", _wrap_Esys_StirRandom, METH_VARARGS, (char *)"Esys_StirRandom(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_DATA inData) -> TSS2_RC"},
	 { (char *)"Esys_StirRandom_Async", _wrap_Esys_StirRandom_Async, METH_VARARGS, (char *)"Esys_StirRandom_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_DATA inData) -> TSS2_RC"},
	 { (char *)"Esys_StirRandom_Finish", _wrap_Esys_StirRandom_Finish, METH_VARARGS, (char *)"Esys_StirRandom_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_HMAC_Start", _wrap_Esys_HMAC_Start, METH_VARARGS, (char *)"Esys_HMAC_Start(ESYS_CONTEXT * esysContext, unsigned int handle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth, TPMI_ALG_HASH hashAlg, unsigned int * sequenceHandle) -> TSS2_RC"},
	 { (char *)"Esys_HMAC_Start_Async", _wrap_Esys_HMAC_Start_Async, METH_VARARGS, (char *)"Esys_HMAC_Start_Async(ESYS_CONTEXT * esysContext, unsigned int handle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth, TPMI_ALG_HASH hashAlg) -> TSS2_RC"},
	 { (char *)"Esys_HMAC_Start_Finish", _wrap_Esys_HMAC_Start_Finish, METH_VARARGS, (char *)"Esys_HMAC_Start_Finish(ESYS_CONTEXT * esysContext, unsigned int * sequenceHandle) -> TSS2_RC"},
	 { (char *)"Esys_HashSequenceStart", _wrap_Esys_HashSequenceStart, METH_VARARGS, (char *)"Esys_HashSequenceStart(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth, TPMI_ALG_HASH hashAlg, unsigned int * sequenceHandle) -> TSS2_RC"},
	 { (char *)"Esys_HashSequenceStart_Async", _wrap_Esys_HashSequenceStart_Async, METH_VARARGS, (char *)"Esys_HashSequenceStart_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth, TPMI_ALG_HASH hashAlg) -> TSS2_RC"},
	 { (char *)"Esys_HashSequenceStart_Finish", _wrap_Esys_HashSequenceStart_Finish, METH_VARARGS, (char *)"Esys_HashSequenceStart_Finish(ESYS_CONTEXT * esysContext, unsigned int * sequenceHandle) -> TSS2_RC"},
	 { (char *)"Esys_SequenceUpdate", _wrap_Esys_SequenceUpdate, METH_VARARGS, (char *)"Esys_SequenceUpdate(ESYS_CONTEXT * esysContext, unsigned int sequenceHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer) -> TSS2_RC"},
	 { (char *)"Esys_SequenceUpdate_Async", _wrap_Esys_SequenceUpdate_Async, METH_VARARGS, (char *)"Esys_SequenceUpdate_Async(ESYS_CONTEXT * esysContext, unsigned int sequenceHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer) -> TSS2_RC"},
	 { (char *)"Esys_SequenceUpdate_Finish", _wrap_Esys_SequenceUpdate_Finish, METH_VARARGS, (char *)"Esys_SequenceUpdate_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_SequenceComplete", _wrap_Esys_SequenceComplete, METH_VARARGS, (char *)"Esys_SequenceComplete(ESYS_CONTEXT * esysContext, unsigned int sequenceHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer, TPMI_RH_HIERARCHY hierarchy, TPM2B_DIGEST ** result, TPMT_TK_HASHCHECK ** validation) -> TSS2_RC"},
	 { (char *)"Esys_SequenceComplete_Async", _wrap_Esys_SequenceComplete_Async, METH_VARARGS, (char *)"Esys_SequenceComplete_Async(ESYS_CONTEXT * esysContext, unsigned int sequenceHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer, TPMI_RH_HIERARCHY hierarchy) -> TSS2_RC"},
	 { (char *)"Esys_SequenceComplete_Finish", _wrap_Esys_SequenceComplete_Finish, METH_VARARGS, (char *)"Esys_SequenceComplete_Finish(ESYS_CONTEXT * esysContext, TPM2B_DIGEST ** result, TPMT_TK_HASHCHECK ** validation) -> TSS2_RC"},
	 { (char *)"Esys_EventSequenceComplete", _wrap_Esys_EventSequenceComplete, METH_VARARGS, (char *)"Esys_EventSequenceComplete(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int sequenceHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer, TPML_DIGEST_VALUES ** results) -> TSS2_RC"},
	 { (char *)"Esys_EventSequenceComplete_Async", _wrap_Esys_EventSequenceComplete_Async, METH_VARARGS, (char *)"Esys_EventSequenceComplete_Async(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int sequenceHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER buffer) -> TSS2_RC"},
	 { (char *)"Esys_EventSequenceComplete_Finish", _wrap_Esys_EventSequenceComplete_Finish, METH_VARARGS, (char *)"Esys_EventSequenceComplete_Finish(ESYS_CONTEXT * esysContext, TPML_DIGEST_VALUES ** results) -> TSS2_RC"},
	 { (char *)"Esys_Certify", _wrap_Esys_Certify, METH_VARARGS, (char *)"Esys_Certify(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, TPM2B_ATTEST ** certifyInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_Certify_Async", _wrap_Esys_Certify_Async, METH_VARARGS, (char *)"Esys_Certify_Async(ESYS_CONTEXT * esysContext, unsigned int objectHandle, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme) -> TSS2_RC"},
	 { (char *)"Esys_Certify_Finish", _wrap_Esys_Certify_Finish, METH_VARARGS, (char *)"Esys_Certify_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** certifyInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_CertifyCreation", _wrap_Esys_CertifyCreation, METH_VARARGS, (char *)"Esys_CertifyCreation(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int objectHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPM2B_DIGEST creationHash, TPMT_SIG_SCHEME inScheme, TPMT_TK_CREATION creationTicket, TPM2B_ATTEST ** certifyInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_CertifyCreation_Async", _wrap_Esys_CertifyCreation_Async, METH_VARARGS, (char *)"Esys_CertifyCreation_Async(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int objectHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPM2B_DIGEST creationHash, TPMT_SIG_SCHEME inScheme, TPMT_TK_CREATION creationTicket) -> TSS2_RC"},
	 { (char *)"Esys_CertifyCreation_Finish", _wrap_Esys_CertifyCreation_Finish, METH_VARARGS, (char *)"Esys_CertifyCreation_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** certifyInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_Quote", _wrap_Esys_Quote, METH_VARARGS, (char *)"Esys_Quote(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, TPML_PCR_SELECTION PCRselect, TPM2B_ATTEST ** quoted, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_Quote_Async", _wrap_Esys_Quote_Async, METH_VARARGS, (char *)"Esys_Quote_Async(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, TPML_PCR_SELECTION PCRselect) -> TSS2_RC"},
	 { (char *)"Esys_Quote_Finish", _wrap_Esys_Quote_Finish, METH_VARARGS, (char *)"Esys_Quote_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** quoted, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_GetSessionAuditDigest", _wrap_Esys_GetSessionAuditDigest, METH_VARARGS, (char *)"Esys_GetSessionAuditDigest(ESYS_CONTEXT * esysContext, unsigned int privacyAdminHandle, unsigned int signHandle, unsigned int sessionHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, TPM2B_ATTEST ** auditInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_GetSessionAuditDigest_Async", _wrap_Esys_GetSessionAuditDigest_Async, METH_VARARGS, (char *)"Esys_GetSessionAuditDigest_Async(ESYS_CONTEXT * esysContext, unsigned int privacyAdminHandle, unsigned int signHandle, unsigned int sessionHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme) -> TSS2_RC"},
	 { (char *)"Esys_GetSessionAuditDigest_Finish", _wrap_Esys_GetSessionAuditDigest_Finish, METH_VARARGS, (char *)"Esys_GetSessionAuditDigest_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** auditInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_GetCommandAuditDigest", _wrap_Esys_GetCommandAuditDigest, METH_VARARGS, (char *)"Esys_GetCommandAuditDigest(ESYS_CONTEXT * esysContext, unsigned int privacyHandle, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, TPM2B_ATTEST ** auditInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_GetCommandAuditDigest_Async", _wrap_Esys_GetCommandAuditDigest_Async, METH_VARARGS, (char *)"Esys_GetCommandAuditDigest_Async(ESYS_CONTEXT * esysContext, unsigned int privacyHandle, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme) -> TSS2_RC"},
	 { (char *)"Esys_GetCommandAuditDigest_Finish", _wrap_Esys_GetCommandAuditDigest_Finish, METH_VARARGS, (char *)"Esys_GetCommandAuditDigest_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** auditInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_GetTime", _wrap_Esys_GetTime, METH_VARARGS, (char *)"Esys_GetTime(ESYS_CONTEXT * esysContext, unsigned int privacyAdminHandle, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, TPM2B_ATTEST ** timeInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_GetTime_Async", _wrap_Esys_GetTime_Async, METH_VARARGS, (char *)"Esys_GetTime_Async(ESYS_CONTEXT * esysContext, unsigned int privacyAdminHandle, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme) -> TSS2_RC"},
	 { (char *)"Esys_GetTime_Finish", _wrap_Esys_GetTime_Finish, METH_VARARGS, (char *)"Esys_GetTime_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** timeInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_Commit", _wrap_Esys_Commit, METH_VARARGS, (char *)"Esys_Commit(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT P1, TPM2B_SENSITIVE_DATA s2, TPM2B_ECC_PARAMETER y2, TPM2B_ECC_POINT ** K, TPM2B_ECC_POINT ** L, TPM2B_ECC_POINT ** E, UINT16 * counter) -> TSS2_RC"},
	 { (char *)"Esys_Commit_Async", _wrap_Esys_Commit_Async, METH_VARARGS, (char *)"Esys_Commit_Async(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_ECC_POINT P1, TPM2B_SENSITIVE_DATA s2, TPM2B_ECC_PARAMETER y2) -> TSS2_RC"},
	 { (char *)"Esys_Commit_Finish", _wrap_Esys_Commit_Finish, METH_VARARGS, (char *)"Esys_Commit_Finish(ESYS_CONTEXT * esysContext, TPM2B_ECC_POINT ** K, TPM2B_ECC_POINT ** L, TPM2B_ECC_POINT ** E, UINT16 * counter) -> TSS2_RC"},
	 { (char *)"Esys_EC_Ephemeral", _wrap_Esys_EC_Ephemeral, METH_VARARGS, (char *)"Esys_EC_Ephemeral(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_ECC_CURVE curveID, TPM2B_ECC_POINT ** Q, UINT16 * counter) -> TSS2_RC"},
	 { (char *)"Esys_EC_Ephemeral_Async", _wrap_Esys_EC_Ephemeral_Async, METH_VARARGS, (char *)"Esys_EC_Ephemeral_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_ECC_CURVE curveID) -> TSS2_RC"},
	 { (char *)"Esys_EC_Ephemeral_Finish", _wrap_Esys_EC_Ephemeral_Finish, METH_VARARGS, (char *)"Esys_EC_Ephemeral_Finish(ESYS_CONTEXT * esysContext, TPM2B_ECC_POINT ** Q, UINT16 * counter) -> TSS2_RC"},
	 { (char *)"Esys_VerifySignature", _wrap_Esys_VerifySignature, METH_VARARGS, (char *)"Esys_VerifySignature(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST digest, TPMT_SIGNATURE signature, TPMT_TK_VERIFIED ** validation) -> TSS2_RC"},
	 { (char *)"Esys_VerifySignature_Async", _wrap_Esys_VerifySignature_Async, METH_VARARGS, (char *)"Esys_VerifySignature_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST digest, TPMT_SIGNATURE signature) -> TSS2_RC"},
	 { (char *)"Esys_VerifySignature_Finish", _wrap_Esys_VerifySignature_Finish, METH_VARARGS, (char *)"Esys_VerifySignature_Finish(ESYS_CONTEXT * esysContext, TPMT_TK_VERIFIED ** validation) -> TSS2_RC"},
	 { (char *)"Esys_Sign", _wrap_Esys_Sign, METH_VARARGS, (char *)"Esys_Sign(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST digest, TPMT_SIG_SCHEME inScheme, TPMT_TK_HASHCHECK validation, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_Sign_Async", _wrap_Esys_Sign_Async, METH_VARARGS, (char *)"Esys_Sign_Async(ESYS_CONTEXT * esysContext, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST digest, TPMT_SIG_SCHEME inScheme, TPMT_TK_HASHCHECK validation) -> TSS2_RC"},
	 { (char *)"Esys_Sign_Finish", _wrap_Esys_Sign_Finish, METH_VARARGS, (char *)"Esys_Sign_Finish(ESYS_CONTEXT * esysContext, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_SetCommandCodeAuditStatus", _wrap_Esys_SetCommandCodeAuditStatus, METH_VARARGS, (char *)"Esys_SetCommandCodeAuditStatus(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_ALG_HASH auditAlg, TPML_CC setList, TPML_CC clearList) -> TSS2_RC"},
	 { (char *)"Esys_SetCommandCodeAuditStatus_Async", _wrap_Esys_SetCommandCodeAuditStatus_Async, METH_VARARGS, (char *)"Esys_SetCommandCodeAuditStatus_Async(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_ALG_HASH auditAlg, TPML_CC setList, TPML_CC clearList) -> TSS2_RC"},
	 { (char *)"Esys_SetCommandCodeAuditStatus_Finish", _wrap_Esys_SetCommandCodeAuditStatus_Finish, METH_VARARGS, (char *)"Esys_SetCommandCodeAuditStatus_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Extend", _wrap_Esys_PCR_Extend, METH_VARARGS, (char *)"Esys_PCR_Extend(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_DIGEST_VALUES digests) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Extend_Async", _wrap_Esys_PCR_Extend_Async, METH_VARARGS, (char *)"Esys_PCR_Extend_Async(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_DIGEST_VALUES digests) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Extend_Finish", _wrap_Esys_PCR_Extend_Finish, METH_VARARGS, (char *)"Esys_PCR_Extend_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Event", _wrap_Esys_PCR_Event, METH_VARARGS, (char *)"Esys_PCR_Event(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_EVENT eventData, TPML_DIGEST_VALUES ** digests) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Event_Async", _wrap_Esys_PCR_Event_Async, METH_VARARGS, (char *)"Esys_PCR_Event_Async(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_EVENT eventData) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Event_Finish", _wrap_Esys_PCR_Event_Finish, METH_VARARGS, (char *)"Esys_PCR_Event_Finish(ESYS_CONTEXT * esysContext, TPML_DIGEST_VALUES ** digests) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Read", _wrap_Esys_PCR_Read, METH_VARARGS, (char *)"Esys_PCR_Read(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_PCR_SELECTION pcrSelectionIn, UINT32 * pcrUpdateCounter, TPML_PCR_SELECTION ** pcrSelectionOut, TPML_DIGEST ** pcrValues) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Read_Async", _wrap_Esys_PCR_Read_Async, METH_VARARGS, (char *)"Esys_PCR_Read_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_PCR_SELECTION pcrSelectionIn) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Read_Finish", _wrap_Esys_PCR_Read_Finish, METH_VARARGS, (char *)"Esys_PCR_Read_Finish(ESYS_CONTEXT * esysContext, UINT32 * pcrUpdateCounter, TPML_PCR_SELECTION ** pcrSelectionOut, TPML_DIGEST ** pcrValues) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Allocate", _wrap_Esys_PCR_Allocate, METH_VARARGS, (char *)"Esys_PCR_Allocate(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_PCR_SELECTION pcrAllocation, TPMI_YES_NO * allocationSuccess, UINT32 * maxPCR, UINT32 * sizeNeeded, UINT32 * sizeAvailable) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Allocate_Async", _wrap_Esys_PCR_Allocate_Async, METH_VARARGS, (char *)"Esys_PCR_Allocate_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_PCR_SELECTION pcrAllocation) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Allocate_Finish", _wrap_Esys_PCR_Allocate_Finish, METH_VARARGS, (char *)"Esys_PCR_Allocate_Finish(ESYS_CONTEXT * esysContext, TPMI_YES_NO * allocationSuccess, UINT32 * maxPCR, UINT32 * sizeNeeded, UINT32 * sizeAvailable) -> TSS2_RC"},
	 { (char *)"Esys_PCR_SetAuthPolicy", _wrap_Esys_PCR_SetAuthPolicy, METH_VARARGS, (char *)"Esys_PCR_SetAuthPolicy(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST authPolicy, TPMI_ALG_HASH hashAlg, TPMI_DH_PCR pcrNum) -> TSS2_RC"},
	 { (char *)"Esys_PCR_SetAuthPolicy_Async", _wrap_Esys_PCR_SetAuthPolicy_Async, METH_VARARGS, (char *)"Esys_PCR_SetAuthPolicy_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST authPolicy, TPMI_ALG_HASH hashAlg, TPMI_DH_PCR pcrNum) -> TSS2_RC"},
	 { (char *)"Esys_PCR_SetAuthPolicy_Finish", _wrap_Esys_PCR_SetAuthPolicy_Finish, METH_VARARGS, (char *)"Esys_PCR_SetAuthPolicy_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PCR_SetAuthValue", _wrap_Esys_PCR_SetAuthValue, METH_VARARGS, (char *)"Esys_PCR_SetAuthValue(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth) -> TSS2_RC"},
	 { (char *)"Esys_PCR_SetAuthValue_Async", _wrap_Esys_PCR_SetAuthValue_Async, METH_VARARGS, (char *)"Esys_PCR_SetAuthValue_Async(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth) -> TSS2_RC"},
	 { (char *)"Esys_PCR_SetAuthValue_Finish", _wrap_Esys_PCR_SetAuthValue_Finish, METH_VARARGS, (char *)"Esys_PCR_SetAuthValue_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Reset", _wrap_Esys_PCR_Reset, METH_VARARGS, (char *)"Esys_PCR_Reset(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Reset_Async", _wrap_Esys_PCR_Reset_Async, METH_VARARGS, (char *)"Esys_PCR_Reset_Async(ESYS_CONTEXT * esysContext, unsigned int pcrHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PCR_Reset_Finish", _wrap_Esys_PCR_Reset_Finish, METH_VARARGS, (char *)"Esys_PCR_Reset_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicySigned", _wrap_Esys_PolicySigned, METH_VARARGS, (char *)"Esys_PolicySigned(ESYS_CONTEXT * esysContext, unsigned int authObject, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nonceTPM, TPM2B_DIGEST cpHashA, TPM2B_DIGEST policyRef, INT32 expiration, TPMT_SIGNATURE auth, TPM2B_TIMEOUT ** timeout, TPMT_TK_AUTH ** policyTicket) -> TSS2_RC"},
	 { (char *)"Esys_PolicySigned_Async", _wrap_Esys_PolicySigned_Async, METH_VARARGS, (char *)"Esys_PolicySigned_Async(ESYS_CONTEXT * esysContext, unsigned int authObject, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nonceTPM, TPM2B_DIGEST cpHashA, TPM2B_DIGEST policyRef, INT32 expiration, TPMT_SIGNATURE auth) -> TSS2_RC"},
	 { (char *)"Esys_PolicySigned_Finish", _wrap_Esys_PolicySigned_Finish, METH_VARARGS, (char *)"Esys_PolicySigned_Finish(ESYS_CONTEXT * esysContext, TPM2B_TIMEOUT ** timeout, TPMT_TK_AUTH ** policyTicket) -> TSS2_RC"},
	 { (char *)"Esys_PolicySecret", _wrap_Esys_PolicySecret, METH_VARARGS, (char *)"Esys_PolicySecret(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nonceTPM, TPM2B_DIGEST cpHashA, TPM2B_DIGEST policyRef, INT32 expiration, TPM2B_TIMEOUT ** timeout, TPMT_TK_AUTH ** policyTicket) -> TSS2_RC"},
	 { (char *)"Esys_PolicySecret_Async", _wrap_Esys_PolicySecret_Async, METH_VARARGS, (char *)"Esys_PolicySecret_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nonceTPM, TPM2B_DIGEST cpHashA, TPM2B_DIGEST policyRef, INT32 expiration) -> TSS2_RC"},
	 { (char *)"Esys_PolicySecret_Finish", _wrap_Esys_PolicySecret_Finish, METH_VARARGS, (char *)"Esys_PolicySecret_Finish(ESYS_CONTEXT * esysContext, TPM2B_TIMEOUT ** timeout, TPMT_TK_AUTH ** policyTicket) -> TSS2_RC"},
	 { (char *)"Esys_PolicyTicket", _wrap_Esys_PolicyTicket, METH_VARARGS, (char *)"Esys_PolicyTicket(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST timeout, TPM2B_DIGEST cpHashA, TPM2B_DIGEST policyRef, TPM2B_NAME authName, TPMT_TK_AUTH ticket) -> TSS2_RC"},
	 { (char *)"Esys_PolicyTicket_Async", _wrap_Esys_PolicyTicket_Async, METH_VARARGS, (char *)"Esys_PolicyTicket_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST timeout, TPM2B_DIGEST cpHashA, TPM2B_DIGEST policyRef, TPM2B_NAME authName, TPMT_TK_AUTH ticket) -> TSS2_RC"},
	 { (char *)"Esys_PolicyTicket_Finish", _wrap_Esys_PolicyTicket_Finish, METH_VARARGS, (char *)"Esys_PolicyTicket_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyOR", _wrap_Esys_PolicyOR, METH_VARARGS, (char *)"Esys_PolicyOR(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_DIGEST pHashList) -> TSS2_RC"},
	 { (char *)"Esys_PolicyOR_Async", _wrap_Esys_PolicyOR_Async, METH_VARARGS, (char *)"Esys_PolicyOR_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_DIGEST pHashList) -> TSS2_RC"},
	 { (char *)"Esys_PolicyOR_Finish", _wrap_Esys_PolicyOR_Finish, METH_VARARGS, (char *)"Esys_PolicyOR_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPCR", _wrap_Esys_PolicyPCR, METH_VARARGS, (char *)"Esys_PolicyPCR(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST pcrDigest, TPML_PCR_SELECTION pcrs) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPCR_Async", _wrap_Esys_PolicyPCR_Async, METH_VARARGS, (char *)"Esys_PolicyPCR_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST pcrDigest, TPML_PCR_SELECTION pcrs) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPCR_Finish", _wrap_Esys_PolicyPCR_Finish, METH_VARARGS, (char *)"Esys_PolicyPCR_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyLocality", _wrap_Esys_PolicyLocality, METH_VARARGS, (char *)"Esys_PolicyLocality(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMA_LOCALITY locality) -> TSS2_RC"},
	 { (char *)"Esys_PolicyLocality_Async", _wrap_Esys_PolicyLocality_Async, METH_VARARGS, (char *)"Esys_PolicyLocality_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMA_LOCALITY locality) -> TSS2_RC"},
	 { (char *)"Esys_PolicyLocality_Finish", _wrap_Esys_PolicyLocality_Finish, METH_VARARGS, (char *)"Esys_PolicyLocality_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNV", _wrap_Esys_PolicyNV, METH_VARARGS, (char *)"Esys_PolicyNV(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST operandB, UINT16 offset, TPM2_EO operation) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNV_Async", _wrap_Esys_PolicyNV_Async, METH_VARARGS, (char *)"Esys_PolicyNV_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST operandB, UINT16 offset, TPM2_EO operation) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNV_Finish", _wrap_Esys_PolicyNV_Finish, METH_VARARGS, (char *)"Esys_PolicyNV_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCounterTimer", _wrap_Esys_PolicyCounterTimer, METH_VARARGS, (char *)"Esys_PolicyCounterTimer(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST operandB, UINT16 offset, TPM2_EO operation) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCounterTimer_Async", _wrap_Esys_PolicyCounterTimer_Async, METH_VARARGS, (char *)"Esys_PolicyCounterTimer_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST operandB, UINT16 offset, TPM2_EO operation) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCounterTimer_Finish", _wrap_Esys_PolicyCounterTimer_Finish, METH_VARARGS, (char *)"Esys_PolicyCounterTimer_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCommandCode", _wrap_Esys_PolicyCommandCode, METH_VARARGS, (char *)"Esys_PolicyCommandCode(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_CC code) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCommandCode_Async", _wrap_Esys_PolicyCommandCode_Async, METH_VARARGS, (char *)"Esys_PolicyCommandCode_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_CC code) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCommandCode_Finish", _wrap_Esys_PolicyCommandCode_Finish, METH_VARARGS, (char *)"Esys_PolicyCommandCode_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPhysicalPresence", _wrap_Esys_PolicyPhysicalPresence, METH_VARARGS, (char *)"Esys_PolicyPhysicalPresence(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPhysicalPresence_Async", _wrap_Esys_PolicyPhysicalPresence_Async, METH_VARARGS, (char *)"Esys_PolicyPhysicalPresence_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPhysicalPresence_Finish", _wrap_Esys_PolicyPhysicalPresence_Finish, METH_VARARGS, (char *)"Esys_PolicyPhysicalPresence_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCpHash", _wrap_Esys_PolicyCpHash, METH_VARARGS, (char *)"Esys_PolicyCpHash(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST cpHashA) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCpHash_Async", _wrap_Esys_PolicyCpHash_Async, METH_VARARGS, (char *)"Esys_PolicyCpHash_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST cpHashA) -> TSS2_RC"},
	 { (char *)"Esys_PolicyCpHash_Finish", _wrap_Esys_PolicyCpHash_Finish, METH_VARARGS, (char *)"Esys_PolicyCpHash_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNameHash", _wrap_Esys_PolicyNameHash, METH_VARARGS, (char *)"Esys_PolicyNameHash(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nameHash) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNameHash_Async", _wrap_Esys_PolicyNameHash_Async, METH_VARARGS, (char *)"Esys_PolicyNameHash_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST nameHash) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNameHash_Finish", _wrap_Esys_PolicyNameHash_Finish, METH_VARARGS, (char *)"Esys_PolicyNameHash_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyDuplicationSelect", _wrap_Esys_PolicyDuplicationSelect, METH_VARARGS, (char *)"Esys_PolicyDuplicationSelect(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_NAME objectName, TPM2B_NAME newParentName, TPMI_YES_NO includeObject) -> TSS2_RC"},
	 { (char *)"Esys_PolicyDuplicationSelect_Async", _wrap_Esys_PolicyDuplicationSelect_Async, METH_VARARGS, (char *)"Esys_PolicyDuplicationSelect_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_NAME objectName, TPM2B_NAME newParentName, TPMI_YES_NO includeObject) -> TSS2_RC"},
	 { (char *)"Esys_PolicyDuplicationSelect_Finish", _wrap_Esys_PolicyDuplicationSelect_Finish, METH_VARARGS, (char *)"Esys_PolicyDuplicationSelect_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthorize", _wrap_Esys_PolicyAuthorize, METH_VARARGS, (char *)"Esys_PolicyAuthorize(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST approvedPolicy, TPM2B_DIGEST policyRef, TPM2B_NAME keySign, TPMT_TK_VERIFIED checkTicket) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthorize_Async", _wrap_Esys_PolicyAuthorize_Async, METH_VARARGS, (char *)"Esys_PolicyAuthorize_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST approvedPolicy, TPM2B_DIGEST policyRef, TPM2B_NAME keySign, TPMT_TK_VERIFIED checkTicket) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthorize_Finish", _wrap_Esys_PolicyAuthorize_Finish, METH_VARARGS, (char *)"Esys_PolicyAuthorize_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthValue", _wrap_Esys_PolicyAuthValue, METH_VARARGS, (char *)"Esys_PolicyAuthValue(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthValue_Async", _wrap_Esys_PolicyAuthValue_Async, METH_VARARGS, (char *)"Esys_PolicyAuthValue_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthValue_Finish", _wrap_Esys_PolicyAuthValue_Finish, METH_VARARGS, (char *)"Esys_PolicyAuthValue_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPassword", _wrap_Esys_PolicyPassword, METH_VARARGS, (char *)"Esys_PolicyPassword(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPassword_Async", _wrap_Esys_PolicyPassword_Async, METH_VARARGS, (char *)"Esys_PolicyPassword_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyPassword_Finish", _wrap_Esys_PolicyPassword_Finish, METH_VARARGS, (char *)"Esys_PolicyPassword_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyGetDigest", _wrap_Esys_PolicyGetDigest, METH_VARARGS, (char *)"Esys_PolicyGetDigest(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST ** policyDigest) -> TSS2_RC"},
	 { (char *)"Esys_PolicyGetDigest_Async", _wrap_Esys_PolicyGetDigest_Async, METH_VARARGS, (char *)"Esys_PolicyGetDigest_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyGetDigest_Finish", _wrap_Esys_PolicyGetDigest_Finish, METH_VARARGS, (char *)"Esys_PolicyGetDigest_Finish(ESYS_CONTEXT * esysContext, TPM2B_DIGEST ** policyDigest) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNvWritten", _wrap_Esys_PolicyNvWritten, METH_VARARGS, (char *)"Esys_PolicyNvWritten(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO writtenSet) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNvWritten_Async", _wrap_Esys_PolicyNvWritten_Async, METH_VARARGS, (char *)"Esys_PolicyNvWritten_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO writtenSet) -> TSS2_RC"},
	 { (char *)"Esys_PolicyNvWritten_Finish", _wrap_Esys_PolicyNvWritten_Finish, METH_VARARGS, (char *)"Esys_PolicyNvWritten_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyTemplate", _wrap_Esys_PolicyTemplate, METH_VARARGS, (char *)"Esys_PolicyTemplate(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST templateHash) -> TSS2_RC"},
	 { (char *)"Esys_PolicyTemplate_Async", _wrap_Esys_PolicyTemplate_Async, METH_VARARGS, (char *)"Esys_PolicyTemplate_Async(ESYS_CONTEXT * esysContext, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST templateHash) -> TSS2_RC"},
	 { (char *)"Esys_PolicyTemplate_Finish", _wrap_Esys_PolicyTemplate_Finish, METH_VARARGS, (char *)"Esys_PolicyTemplate_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthorizeNV", _wrap_Esys_PolicyAuthorizeNV, METH_VARARGS, (char *)"Esys_PolicyAuthorizeNV(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthorizeNV_Async", _wrap_Esys_PolicyAuthorizeNV_Async, METH_VARARGS, (char *)"Esys_PolicyAuthorizeNV_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int policySession, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_PolicyAuthorizeNV_Finish", _wrap_Esys_PolicyAuthorizeNV_Finish, METH_VARARGS, (char *)"Esys_PolicyAuthorizeNV_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_CreatePrimary", _wrap_Esys_CreatePrimary, METH_VARARGS, (char *)"Esys_CreatePrimary(ESYS_CONTEXT * esysContext, unsigned int primaryHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_CREATE inSensitive, TPM2B_PUBLIC inPublic, TPM2B_DATA outsideInfo, TPML_PCR_SELECTION creationPCR, unsigned int * objectHandle, TPM2B_PUBLIC ** outPublic, TPM2B_CREATION_DATA ** creationData, TPM2B_DIGEST ** creationHash, TPMT_TK_CREATION ** creationTicket) -> TSS2_RC"},
	 { (char *)"Esys_CreatePrimary_Async", _wrap_Esys_CreatePrimary_Async, METH_VARARGS, (char *)"Esys_CreatePrimary_Async(ESYS_CONTEXT * esysContext, unsigned int primaryHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_SENSITIVE_CREATE inSensitive, TPM2B_PUBLIC inPublic, TPM2B_DATA outsideInfo, TPML_PCR_SELECTION creationPCR) -> TSS2_RC"},
	 { (char *)"Esys_CreatePrimary_Finish", _wrap_Esys_CreatePrimary_Finish, METH_VARARGS, (char *)"Esys_CreatePrimary_Finish(ESYS_CONTEXT * esysContext, unsigned int * objectHandle, TPM2B_PUBLIC ** outPublic, TPM2B_CREATION_DATA ** creationData, TPM2B_DIGEST ** creationHash, TPMT_TK_CREATION ** creationTicket) -> TSS2_RC"},
	 { (char *)"Esys_HierarchyControl", _wrap_Esys_HierarchyControl, METH_VARARGS, (char *)"Esys_HierarchyControl(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_RH_ENABLES enable, TPMI_YES_NO state) -> TSS2_RC"},
	 { (char *)"Esys_HierarchyControl_Async", _wrap_Esys_HierarchyControl_Async, METH_VARARGS, (char *)"Esys_HierarchyControl_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_RH_ENABLES enable, TPMI_YES_NO state) -> TSS2_RC"},
	 { (char *)"Esys_HierarchyControl_Finish", _wrap_Esys_HierarchyControl_Finish, METH_VARARGS, (char *)"Esys_HierarchyControl_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_SetPrimaryPolicy", _wrap_Esys_SetPrimaryPolicy, METH_VARARGS, (char *)"Esys_SetPrimaryPolicy(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST authPolicy, TPMI_ALG_HASH hashAlg) -> TSS2_RC"},
	 { (char *)"Esys_SetPrimaryPolicy_Async", _wrap_Esys_SetPrimaryPolicy_Async, METH_VARARGS, (char *)"Esys_SetPrimaryPolicy_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST authPolicy, TPMI_ALG_HASH hashAlg) -> TSS2_RC"},
	 { (char *)"Esys_SetPrimaryPolicy_Finish", _wrap_Esys_SetPrimaryPolicy_Finish, METH_VARARGS, (char *)"Esys_SetPrimaryPolicy_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_ChangePPS", _wrap_Esys_ChangePPS, METH_VARARGS, (char *)"Esys_ChangePPS(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ChangePPS_Async", _wrap_Esys_ChangePPS_Async, METH_VARARGS, (char *)"Esys_ChangePPS_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ChangePPS_Finish", _wrap_Esys_ChangePPS_Finish, METH_VARARGS, (char *)"Esys_ChangePPS_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_ChangeEPS", _wrap_Esys_ChangeEPS, METH_VARARGS, (char *)"Esys_ChangeEPS(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ChangeEPS_Async", _wrap_Esys_ChangeEPS_Async, METH_VARARGS, (char *)"Esys_ChangeEPS_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ChangeEPS_Finish", _wrap_Esys_ChangeEPS_Finish, METH_VARARGS, (char *)"Esys_ChangeEPS_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_Clear", _wrap_Esys_Clear, METH_VARARGS, (char *)"Esys_Clear(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_Clear_Async", _wrap_Esys_Clear_Async, METH_VARARGS, (char *)"Esys_Clear_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_Clear_Finish", _wrap_Esys_Clear_Finish, METH_VARARGS, (char *)"Esys_Clear_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_ClearControl", _wrap_Esys_ClearControl, METH_VARARGS, (char *)"Esys_ClearControl(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO disable) -> TSS2_RC"},
	 { (char *)"Esys_ClearControl_Async", _wrap_Esys_ClearControl_Async, METH_VARARGS, (char *)"Esys_ClearControl_Async(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_YES_NO disable) -> TSS2_RC"},
	 { (char *)"Esys_ClearControl_Finish", _wrap_Esys_ClearControl_Finish, METH_VARARGS, (char *)"Esys_ClearControl_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_HierarchyChangeAuth", _wrap_Esys_HierarchyChangeAuth, METH_VARARGS, (char *)"Esys_HierarchyChangeAuth(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST newAuth) -> TSS2_RC"},
	 { (char *)"Esys_HierarchyChangeAuth_Async", _wrap_Esys_HierarchyChangeAuth_Async, METH_VARARGS, (char *)"Esys_HierarchyChangeAuth_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST newAuth) -> TSS2_RC"},
	 { (char *)"Esys_HierarchyChangeAuth_Finish", _wrap_Esys_HierarchyChangeAuth_Finish, METH_VARARGS, (char *)"Esys_HierarchyChangeAuth_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_DictionaryAttackLockReset", _wrap_Esys_DictionaryAttackLockReset, METH_VARARGS, (char *)"Esys_DictionaryAttackLockReset(ESYS_CONTEXT * esysContext, unsigned int lockHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_DictionaryAttackLockReset_Async", _wrap_Esys_DictionaryAttackLockReset_Async, METH_VARARGS, (char *)"Esys_DictionaryAttackLockReset_Async(ESYS_CONTEXT * esysContext, unsigned int lockHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_DictionaryAttackLockReset_Finish", _wrap_Esys_DictionaryAttackLockReset_Finish, METH_VARARGS, (char *)"Esys_DictionaryAttackLockReset_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_DictionaryAttackParameters", _wrap_Esys_DictionaryAttackParameters, METH_VARARGS, (char *)"Esys_DictionaryAttackParameters(ESYS_CONTEXT * esysContext, unsigned int lockHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT32 newMaxTries, UINT32 newRecoveryTime, UINT32 lockoutRecovery) -> TSS2_RC"},
	 { (char *)"Esys_DictionaryAttackParameters_Async", _wrap_Esys_DictionaryAttackParameters_Async, METH_VARARGS, (char *)"Esys_DictionaryAttackParameters_Async(ESYS_CONTEXT * esysContext, unsigned int lockHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT32 newMaxTries, UINT32 newRecoveryTime, UINT32 lockoutRecovery) -> TSS2_RC"},
	 { (char *)"Esys_DictionaryAttackParameters_Finish", _wrap_Esys_DictionaryAttackParameters_Finish, METH_VARARGS, (char *)"Esys_DictionaryAttackParameters_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_PP_Commands", _wrap_Esys_PP_Commands, METH_VARARGS, (char *)"Esys_PP_Commands(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_CC setList, TPML_CC clearList) -> TSS2_RC"},
	 { (char *)"Esys_PP_Commands_Async", _wrap_Esys_PP_Commands_Async, METH_VARARGS, (char *)"Esys_PP_Commands_Async(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPML_CC setList, TPML_CC clearList) -> TSS2_RC"},
	 { (char *)"Esys_PP_Commands_Finish", _wrap_Esys_PP_Commands_Finish, METH_VARARGS, (char *)"Esys_PP_Commands_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_SetAlgorithmSet", _wrap_Esys_SetAlgorithmSet, METH_VARARGS, (char *)"Esys_SetAlgorithmSet(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT32 algorithmSet) -> TSS2_RC"},
	 { (char *)"Esys_SetAlgorithmSet_Async", _wrap_Esys_SetAlgorithmSet_Async, METH_VARARGS, (char *)"Esys_SetAlgorithmSet_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT32 algorithmSet) -> TSS2_RC"},
	 { (char *)"Esys_SetAlgorithmSet_Finish", _wrap_Esys_SetAlgorithmSet_Finish, METH_VARARGS, (char *)"Esys_SetAlgorithmSet_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_FieldUpgradeStart", _wrap_Esys_FieldUpgradeStart, METH_VARARGS, (char *)"Esys_FieldUpgradeStart(ESYS_CONTEXT * esysContext, unsigned int authorization, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST fuDigest, TPMT_SIGNATURE manifestSignature) -> TSS2_RC"},
	 { (char *)"Esys_FieldUpgradeStart_Async", _wrap_Esys_FieldUpgradeStart_Async, METH_VARARGS, (char *)"Esys_FieldUpgradeStart_Async(ESYS_CONTEXT * esysContext, unsigned int authorization, unsigned int keyHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST fuDigest, TPMT_SIGNATURE manifestSignature) -> TSS2_RC"},
	 { (char *)"Esys_FieldUpgradeStart_Finish", _wrap_Esys_FieldUpgradeStart_Finish, METH_VARARGS, (char *)"Esys_FieldUpgradeStart_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_FieldUpgradeData", _wrap_Esys_FieldUpgradeData, METH_VARARGS, (char *)"Esys_FieldUpgradeData(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER fuData, TPMT_HA ** nextDigest, TPMT_HA ** firstDigest) -> TSS2_RC"},
	 { (char *)"Esys_FieldUpgradeData_Async", _wrap_Esys_FieldUpgradeData_Async, METH_VARARGS, (char *)"Esys_FieldUpgradeData_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_BUFFER fuData) -> TSS2_RC"},
	 { (char *)"Esys_FieldUpgradeData_Finish", _wrap_Esys_FieldUpgradeData_Finish, METH_VARARGS, (char *)"Esys_FieldUpgradeData_Finish(ESYS_CONTEXT * esysContext, TPMT_HA ** nextDigest, TPMT_HA ** firstDigest) -> TSS2_RC"},
	 { (char *)"Esys_FirmwareRead", _wrap_Esys_FirmwareRead, METH_VARARGS, (char *)"Esys_FirmwareRead(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT32 sequenceNumber, TPM2B_MAX_BUFFER ** fuData) -> TSS2_RC"},
	 { (char *)"Esys_FirmwareRead_Async", _wrap_Esys_FirmwareRead_Async, METH_VARARGS, (char *)"Esys_FirmwareRead_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT32 sequenceNumber) -> TSS2_RC"},
	 { (char *)"Esys_FirmwareRead_Finish", _wrap_Esys_FirmwareRead_Finish, METH_VARARGS, (char *)"Esys_FirmwareRead_Finish(ESYS_CONTEXT * esysContext, TPM2B_MAX_BUFFER ** fuData) -> TSS2_RC"},
	 { (char *)"Esys_ContextSave", _wrap_Esys_ContextSave, METH_VARARGS, (char *)"Esys_ContextSave(ESYS_CONTEXT * esysContext, unsigned int saveHandle, TPMS_CONTEXT ** context) -> TSS2_RC"},
	 { (char *)"Esys_ContextSave_Async", _wrap_Esys_ContextSave_Async, METH_VARARGS, (char *)"Esys_ContextSave_Async(ESYS_CONTEXT * esysContext, unsigned int saveHandle) -> TSS2_RC"},
	 { (char *)"Esys_ContextSave_Finish", _wrap_Esys_ContextSave_Finish, METH_VARARGS, (char *)"Esys_ContextSave_Finish(ESYS_CONTEXT * esysContext, TPMS_CONTEXT ** context) -> TSS2_RC"},
	 { (char *)"Esys_ContextLoad", _wrap_Esys_ContextLoad, METH_VARARGS, (char *)"Esys_ContextLoad(ESYS_CONTEXT * esysContext, TPMS_CONTEXT context, unsigned int * loadedHandle) -> TSS2_RC"},
	 { (char *)"Esys_ContextLoad_Async", _wrap_Esys_ContextLoad_Async, METH_VARARGS, (char *)"Esys_ContextLoad_Async(ESYS_CONTEXT * esysContext, TPMS_CONTEXT context) -> TSS2_RC"},
	 { (char *)"Esys_ContextLoad_Finish", _wrap_Esys_ContextLoad_Finish, METH_VARARGS, (char *)"Esys_ContextLoad_Finish(ESYS_CONTEXT * esysContext, unsigned int * loadedHandle) -> TSS2_RC"},
	 { (char *)"Esys_FlushContext", _wrap_Esys_FlushContext, METH_VARARGS, (char *)"Esys_FlushContext(ESYS_CONTEXT * esysContext, unsigned int flushHandle) -> TSS2_RC"},
	 { (char *)"Esys_FlushContext_Async", _wrap_Esys_FlushContext_Async, METH_VARARGS, (char *)"Esys_FlushContext_Async(ESYS_CONTEXT * esysContext, unsigned int flushHandle) -> TSS2_RC"},
	 { (char *)"Esys_FlushContext_Finish", _wrap_Esys_FlushContext_Finish, METH_VARARGS, (char *)"Esys_FlushContext_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_EvictControl", _wrap_Esys_EvictControl, METH_VARARGS, (char *)"Esys_EvictControl(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int objectHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_DH_PERSISTENT persistentHandle, unsigned int * newObjectHandle) -> TSS2_RC"},
	 { (char *)"Esys_EvictControl_Async", _wrap_Esys_EvictControl_Async, METH_VARARGS, (char *)"Esys_EvictControl_Async(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int objectHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMI_DH_PERSISTENT persistentHandle) -> TSS2_RC"},
	 { (char *)"Esys_EvictControl_Finish", _wrap_Esys_EvictControl_Finish, METH_VARARGS, (char *)"Esys_EvictControl_Finish(ESYS_CONTEXT * esysContext, unsigned int * newObjectHandle) -> TSS2_RC"},
	 { (char *)"Esys_ReadClock", _wrap_Esys_ReadClock, METH_VARARGS, (char *)"Esys_ReadClock(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMS_TIME_INFO ** currentTime) -> TSS2_RC"},
	 { (char *)"Esys_ReadClock_Async", _wrap_Esys_ReadClock_Async, METH_VARARGS, (char *)"Esys_ReadClock_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_ReadClock_Finish", _wrap_Esys_ReadClock_Finish, METH_VARARGS, (char *)"Esys_ReadClock_Finish(ESYS_CONTEXT * esysContext, TPMS_TIME_INFO ** currentTime) -> TSS2_RC"},
	 { (char *)"Esys_ClockSet", _wrap_Esys_ClockSet, METH_VARARGS, (char *)"Esys_ClockSet(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT64 newTime) -> TSS2_RC"},
	 { (char *)"Esys_ClockSet_Async", _wrap_Esys_ClockSet_Async, METH_VARARGS, (char *)"Esys_ClockSet_Async(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT64 newTime) -> TSS2_RC"},
	 { (char *)"Esys_ClockSet_Finish", _wrap_Esys_ClockSet_Finish, METH_VARARGS, (char *)"Esys_ClockSet_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_ClockRateAdjust", _wrap_Esys_ClockRateAdjust, METH_VARARGS, (char *)"Esys_ClockRateAdjust(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_CLOCK_ADJUST rateAdjust) -> TSS2_RC"},
	 { (char *)"Esys_ClockRateAdjust_Async", _wrap_Esys_ClockRateAdjust_Async, METH_VARARGS, (char *)"Esys_ClockRateAdjust_Async(ESYS_CONTEXT * esysContext, unsigned int auth, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_CLOCK_ADJUST rateAdjust) -> TSS2_RC"},
	 { (char *)"Esys_ClockRateAdjust_Finish", _wrap_Esys_ClockRateAdjust_Finish, METH_VARARGS, (char *)"Esys_ClockRateAdjust_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_GetCapability", _wrap_Esys_GetCapability, METH_VARARGS, (char *)"Esys_GetCapability(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_CAP capability, UINT32 property, UINT32 propertyCount, TPMI_YES_NO * moreData, TPMS_CAPABILITY_DATA ** capabilityData) -> TSS2_RC"},
	 { (char *)"Esys_GetCapability_Async", _wrap_Esys_GetCapability_Async, METH_VARARGS, (char *)"Esys_GetCapability_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2_CAP capability, UINT32 property, UINT32 propertyCount) -> TSS2_RC"},
	 { (char *)"Esys_GetCapability_Finish", _wrap_Esys_GetCapability_Finish, METH_VARARGS, (char *)"Esys_GetCapability_Finish(ESYS_CONTEXT * esysContext, TPMI_YES_NO * moreData, TPMS_CAPABILITY_DATA ** capabilityData) -> TSS2_RC"},
	 { (char *)"Esys_TestParms", _wrap_Esys_TestParms, METH_VARARGS, (char *)"Esys_TestParms(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMT_PUBLIC_PARMS parameters) -> TSS2_RC"},
	 { (char *)"Esys_TestParms_Async", _wrap_Esys_TestParms_Async, METH_VARARGS, (char *)"Esys_TestParms_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPMT_PUBLIC_PARMS parameters) -> TSS2_RC"},
	 { (char *)"Esys_TestParms_Finish", _wrap_Esys_TestParms_Finish, METH_VARARGS, (char *)"Esys_TestParms_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_DefineSpace", _wrap_Esys_NV_DefineSpace, METH_VARARGS, (char *)"Esys_NV_DefineSpace(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth, TPM2B_NV_PUBLIC publicInfo, unsigned int * nvHandle) -> TSS2_RC"},
	 { (char *)"Esys_NV_DefineSpace_Async", _wrap_Esys_NV_DefineSpace_Async, METH_VARARGS, (char *)"Esys_NV_DefineSpace_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST auth, TPM2B_NV_PUBLIC publicInfo) -> TSS2_RC"},
	 { (char *)"Esys_NV_DefineSpace_Finish", _wrap_Esys_NV_DefineSpace_Finish, METH_VARARGS, (char *)"Esys_NV_DefineSpace_Finish(ESYS_CONTEXT * esysContext, unsigned int * nvHandle) -> TSS2_RC"},
	 { (char *)"Esys_NV_UndefineSpace", _wrap_Esys_NV_UndefineSpace, METH_VARARGS, (char *)"Esys_NV_UndefineSpace(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_UndefineSpace_Async", _wrap_Esys_NV_UndefineSpace_Async, METH_VARARGS, (char *)"Esys_NV_UndefineSpace_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_UndefineSpace_Finish", _wrap_Esys_NV_UndefineSpace_Finish, METH_VARARGS, (char *)"Esys_NV_UndefineSpace_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_UndefineSpaceSpecial", _wrap_Esys_NV_UndefineSpaceSpecial, METH_VARARGS, (char *)"Esys_NV_UndefineSpaceSpecial(ESYS_CONTEXT * esysContext, unsigned int nvIndex, unsigned int platform, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_UndefineSpaceSpecial_Async", _wrap_Esys_NV_UndefineSpaceSpecial_Async, METH_VARARGS, (char *)"Esys_NV_UndefineSpaceSpecial_Async(ESYS_CONTEXT * esysContext, unsigned int nvIndex, unsigned int platform, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_UndefineSpaceSpecial_Finish", _wrap_Esys_NV_UndefineSpaceSpecial_Finish, METH_VARARGS, (char *)"Esys_NV_UndefineSpaceSpecial_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_ReadPublic", _wrap_Esys_NV_ReadPublic, METH_VARARGS, (char *)"Esys_NV_ReadPublic(ESYS_CONTEXT * esysContext, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_NV_PUBLIC ** nvPublic, TPM2B_NAME ** nvName) -> TSS2_RC"},
	 { (char *)"Esys_NV_ReadPublic_Async", _wrap_Esys_NV_ReadPublic_Async, METH_VARARGS, (char *)"Esys_NV_ReadPublic_Async(ESYS_CONTEXT * esysContext, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_ReadPublic_Finish", _wrap_Esys_NV_ReadPublic_Finish, METH_VARARGS, (char *)"Esys_NV_ReadPublic_Finish(ESYS_CONTEXT * esysContext, TPM2B_NV_PUBLIC ** nvPublic, TPM2B_NAME ** nvName) -> TSS2_RC"},
	 { (char *)"Esys_NV_Write", _wrap_Esys_NV_Write, METH_VARARGS, (char *)"Esys_NV_Write(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_NV_BUFFER data, UINT16 offset) -> TSS2_RC"},
	 { (char *)"Esys_NV_Write_Async", _wrap_Esys_NV_Write_Async, METH_VARARGS, (char *)"Esys_NV_Write_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_NV_BUFFER data, UINT16 offset) -> TSS2_RC"},
	 { (char *)"Esys_NV_Write_Finish", _wrap_Esys_NV_Write_Finish, METH_VARARGS, (char *)"Esys_NV_Write_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_Increment", _wrap_Esys_NV_Increment, METH_VARARGS, (char *)"Esys_NV_Increment(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_Increment_Async", _wrap_Esys_NV_Increment_Async, METH_VARARGS, (char *)"Esys_NV_Increment_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_Increment_Finish", _wrap_Esys_NV_Increment_Finish, METH_VARARGS, (char *)"Esys_NV_Increment_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_Extend", _wrap_Esys_NV_Extend, METH_VARARGS, (char *)"Esys_NV_Extend(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_NV_BUFFER data) -> TSS2_RC"},
	 { (char *)"Esys_NV_Extend_Async", _wrap_Esys_NV_Extend_Async, METH_VARARGS, (char *)"Esys_NV_Extend_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_MAX_NV_BUFFER data) -> TSS2_RC"},
	 { (char *)"Esys_NV_Extend_Finish", _wrap_Esys_NV_Extend_Finish, METH_VARARGS, (char *)"Esys_NV_Extend_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_SetBits", _wrap_Esys_NV_SetBits, METH_VARARGS, (char *)"Esys_NV_SetBits(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT64 bits) -> TSS2_RC"},
	 { (char *)"Esys_NV_SetBits_Async", _wrap_Esys_NV_SetBits_Async, METH_VARARGS, (char *)"Esys_NV_SetBits_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT64 bits) -> TSS2_RC"},
	 { (char *)"Esys_NV_SetBits_Finish", _wrap_Esys_NV_SetBits_Finish, METH_VARARGS, (char *)"Esys_NV_SetBits_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_WriteLock", _wrap_Esys_NV_WriteLock, METH_VARARGS, (char *)"Esys_NV_WriteLock(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_WriteLock_Async", _wrap_Esys_NV_WriteLock_Async, METH_VARARGS, (char *)"Esys_NV_WriteLock_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_WriteLock_Finish", _wrap_Esys_NV_WriteLock_Finish, METH_VARARGS, (char *)"Esys_NV_WriteLock_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_GlobalWriteLock", _wrap_Esys_NV_GlobalWriteLock, METH_VARARGS, (char *)"Esys_NV_GlobalWriteLock(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_GlobalWriteLock_Async", _wrap_Esys_NV_GlobalWriteLock_Async, METH_VARARGS, (char *)"Esys_NV_GlobalWriteLock_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_GlobalWriteLock_Finish", _wrap_Esys_NV_GlobalWriteLock_Finish, METH_VARARGS, (char *)"Esys_NV_GlobalWriteLock_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_Read", _wrap_Esys_NV_Read, METH_VARARGS, (char *)"Esys_NV_Read(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT16 size, UINT16 offset, TPM2B_MAX_NV_BUFFER ** data) -> TSS2_RC"},
	 { (char *)"Esys_NV_Read_Async", _wrap_Esys_NV_Read_Async, METH_VARARGS, (char *)"Esys_NV_Read_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, UINT16 size, UINT16 offset) -> TSS2_RC"},
	 { (char *)"Esys_NV_Read_Finish", _wrap_Esys_NV_Read_Finish, METH_VARARGS, (char *)"Esys_NV_Read_Finish(ESYS_CONTEXT * esysContext, TPM2B_MAX_NV_BUFFER ** data) -> TSS2_RC"},
	 { (char *)"Esys_NV_ReadLock", _wrap_Esys_NV_ReadLock, METH_VARARGS, (char *)"Esys_NV_ReadLock(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_ReadLock_Async", _wrap_Esys_NV_ReadLock_Async, METH_VARARGS, (char *)"Esys_NV_ReadLock_Async(ESYS_CONTEXT * esysContext, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3) -> TSS2_RC"},
	 { (char *)"Esys_NV_ReadLock_Finish", _wrap_Esys_NV_ReadLock_Finish, METH_VARARGS, (char *)"Esys_NV_ReadLock_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_ChangeAuth", _wrap_Esys_NV_ChangeAuth, METH_VARARGS, (char *)"Esys_NV_ChangeAuth(ESYS_CONTEXT * esysContext, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST newAuth) -> TSS2_RC"},
	 { (char *)"Esys_NV_ChangeAuth_Async", _wrap_Esys_NV_ChangeAuth_Async, METH_VARARGS, (char *)"Esys_NV_ChangeAuth_Async(ESYS_CONTEXT * esysContext, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DIGEST newAuth) -> TSS2_RC"},
	 { (char *)"Esys_NV_ChangeAuth_Finish", _wrap_Esys_NV_ChangeAuth_Finish, METH_VARARGS, (char *)"Esys_NV_ChangeAuth_Finish(ESYS_CONTEXT * esysContext) -> TSS2_RC"},
	 { (char *)"Esys_NV_Certify", _wrap_Esys_NV_Certify, METH_VARARGS, (char *)"Esys_NV_Certify(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, UINT16 size, UINT16 offset, TPM2B_ATTEST ** certifyInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_NV_Certify_Async", _wrap_Esys_NV_Certify_Async, METH_VARARGS, (char *)"Esys_NV_Certify_Async(ESYS_CONTEXT * esysContext, unsigned int signHandle, unsigned int authHandle, unsigned int nvIndex, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA qualifyingData, TPMT_SIG_SCHEME inScheme, UINT16 size, UINT16 offset) -> TSS2_RC"},
	 { (char *)"Esys_NV_Certify_Finish", _wrap_Esys_NV_Certify_Finish, METH_VARARGS, (char *)"Esys_NV_Certify_Finish(ESYS_CONTEXT * esysContext, TPM2B_ATTEST ** certifyInfo, TPMT_SIGNATURE ** signature) -> TSS2_RC"},
	 { (char *)"Esys_Vendor_TCG_Test", _wrap_Esys_Vendor_TCG_Test, METH_VARARGS, (char *)"Esys_Vendor_TCG_Test(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA inputData, TPM2B_DATA ** outputData) -> TSS2_RC"},
	 { (char *)"Esys_Vendor_TCG_Test_Async", _wrap_Esys_Vendor_TCG_Test_Async, METH_VARARGS, (char *)"Esys_Vendor_TCG_Test_Async(ESYS_CONTEXT * esysContext, unsigned int shandle1, unsigned int shandle2, unsigned int shandle3, TPM2B_DATA inputData) -> TSS2_RC"},
	 { (char *)"Esys_Vendor_TCG_Test_Finish", _wrap_Esys_Vendor_TCG_Test_Finish, METH_VARARGS, (char *)"Esys_Vendor_TCG_Test_Finish(ESYS_CONTEXT * esysContext, TPM2B_DATA ** outputData) -> TSS2_RC"},
	 { (char *)"Esys_Free", _wrap_Esys_Free, METH_VARARGS, (char *)"Esys_Free(void * ptr)"},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_p_TPM2B_SENSITIVE_CREATE_ARRAYTo_p_p_TPM2B_SENSITIVE_CREATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SENSITIVE_CREATE **)  ((TPM2B_SENSITIVE_CREATE_ARRAY **) x));
}
static void *_p_p_TPMS_SENSITIVE_CREATE_ARRAYTo_p_p_TPMS_SENSITIVE_CREATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SENSITIVE_CREATE **)  ((TPMS_SENSITIVE_CREATE_ARRAY **) x));
}
static void *_p_TPMS_SENSITIVE_CREATE_ARRAYTo_p_TPMS_SENSITIVE_CREATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SENSITIVE_CREATE *)  ((TPMS_SENSITIVE_CREATE_ARRAY *) x));
}
static void *_p_TPM2B_SENSITIVE_CREATE_ARRAYTo_p_TPM2B_SENSITIVE_CREATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SENSITIVE_CREATE *)  ((TPM2B_SENSITIVE_CREATE_ARRAY *) x));
}
static void *_p_TPMS_CONTEXT_ARRAYTo_p_TPMS_CONTEXT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CONTEXT *)  ((TPMS_CONTEXT_ARRAY *) x));
}
static void *_p_p_TPMS_CONTEXT_ARRAYTo_p_p_TPMS_CONTEXT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CONTEXT **)  ((TPMS_CONTEXT_ARRAY **) x));
}
static void *_p_p_TPMS_SESSION_AUDIT_INFO_ARRAYTo_p_p_TPMS_SESSION_AUDIT_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SESSION_AUDIT_INFO **)  ((TPMS_SESSION_AUDIT_INFO_ARRAY **) x));
}
static void *_p_p_TPMS_COMMAND_AUDIT_INFO_ARRAYTo_p_p_TPMS_COMMAND_AUDIT_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_COMMAND_AUDIT_INFO **)  ((TPMS_COMMAND_AUDIT_INFO_ARRAY **) x));
}
static void *_p_p_TPMS_TIME_ATTEST_INFO_ARRAYTo_p_p_TPMS_TIME_ATTEST_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TIME_ATTEST_INFO **)  ((TPMS_TIME_ATTEST_INFO_ARRAY **) x));
}
static void *_p_TPMS_COMMAND_AUDIT_INFO_ARRAYTo_p_TPMS_COMMAND_AUDIT_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_COMMAND_AUDIT_INFO *)  ((TPMS_COMMAND_AUDIT_INFO_ARRAY *) x));
}
static void *_p_TPMS_TIME_ATTEST_INFO_ARRAYTo_p_TPMS_TIME_ATTEST_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TIME_ATTEST_INFO *)  ((TPMS_TIME_ATTEST_INFO_ARRAY *) x));
}
static void *_p_TPMS_SESSION_AUDIT_INFO_ARRAYTo_p_TPMS_SESSION_AUDIT_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SESSION_AUDIT_INFO *)  ((TPMS_SESSION_AUDIT_INFO_ARRAY *) x));
}
static void *_p_p_TPM2B_ECC_PARAMETER_ARRAYTo_p_p_TPM2B_ECC_PARAMETER(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ECC_PARAMETER **)  ((TPM2B_ECC_PARAMETER_ARRAY **) x));
}
static void *_p_TPM2B_ECC_PARAMETER_ARRAYTo_p_TPM2B_ECC_PARAMETER(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ECC_PARAMETER *)  ((TPM2B_ECC_PARAMETER_ARRAY *) x));
}
static void *_p_p_TPMS_SCHEME_ECDAA_ARRAYTo_p_p_TPMS_SCHEME_ECDAA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SCHEME_ECDAA **)  ((TPMS_SCHEME_ECDAA_ARRAY **) x));
}
static void *_p_TPMS_SCHEME_ECDAA_ARRAYTo_p_TPMS_SCHEME_ECDAA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SCHEME_ECDAA *)  ((TPMS_SCHEME_ECDAA_ARRAY *) x));
}
static void *_p_p_TPMS_EMPTY_ARRAYTo_p_p_TPMS_EMPTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_EMPTY **)  ((TPMS_EMPTY_ARRAY **) x));
}
static void *_p_TPMS_EMPTY_ARRAYTo_p_TPMS_EMPTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_EMPTY *)  ((TPMS_EMPTY_ARRAY *) x));
}
static void *_p_p_TPMU_SCHEME_KEYEDHASH_ARRAYTo_p_p_TPMU_SCHEME_KEYEDHASH(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SCHEME_KEYEDHASH **)  ((TPMU_SCHEME_KEYEDHASH_ARRAY **) x));
}
static void *_p_TPMU_SCHEME_KEYEDHASH_ARRAYTo_p_TPMU_SCHEME_KEYEDHASH(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SCHEME_KEYEDHASH *)  ((TPMU_SCHEME_KEYEDHASH_ARRAY *) x));
}
static void *_p_TPMS_AC_OUTPUT_ARRAYTo_p_TPMS_AC_OUTPUT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_AC_OUTPUT *)  ((TPMS_AC_OUTPUT_ARRAY *) x));
}
static void *_p_p_TPMS_AC_OUTPUT_ARRAYTo_p_p_TPMS_AC_OUTPUT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_AC_OUTPUT **)  ((TPMS_AC_OUTPUT_ARRAY **) x));
}
static void *_p_p_TPML_DIGEST_ARRAYTo_p_p_TPML_DIGEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_DIGEST **)  ((TPML_DIGEST_ARRAY **) x));
}
static void *_p_p_TPM2B_DIGEST_ARRAYTo_p_p_TPM2B_DIGEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_DIGEST **)  ((TPM2B_DIGEST_ARRAY **) x));
}
static void *_p_TPM2B_DIGEST_ARRAYTo_p_TPM2B_DIGEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_DIGEST *)  ((TPM2B_DIGEST_ARRAY *) x));
}
static void *_p_TPML_DIGEST_ARRAYTo_p_TPML_DIGEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_DIGEST *)  ((TPML_DIGEST_ARRAY *) x));
}
static void *_p_p_TPML_CCA_ARRAYTo_p_p_TPML_CCA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_CCA **)  ((TPML_CCA_ARRAY **) x));
}
static void *_p_TPML_CCA_ARRAYTo_p_TPML_CCA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_CCA *)  ((TPML_CCA_ARRAY *) x));
}
static void *_p_p_TPMT_RSA_DECRYPT_ARRAYTo_p_p_TPMT_RSA_DECRYPT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_RSA_DECRYPT **)  ((TPMT_RSA_DECRYPT_ARRAY **) x));
}
static void *_p_TPMT_RSA_DECRYPT_ARRAYTo_p_TPMT_RSA_DECRYPT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_RSA_DECRYPT *)  ((TPMT_RSA_DECRYPT_ARRAY *) x));
}
static void *_p_p_TPM2B_NAME_ARRAYTo_p_p_TPM2B_NAME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_NAME **)  ((TPM2B_NAME_ARRAY **) x));
}
static void *_p_TPM2B_NAME_ARRAYTo_p_TPM2B_NAME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_NAME *)  ((TPM2B_NAME_ARRAY *) x));
}
static void *_p_TPMS_TAGGED_PCR_SELECT_ARRAYTo_p_TPMS_TAGGED_PCR_SELECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TAGGED_PCR_SELECT *)  ((TPMS_TAGGED_PCR_SELECT_ARRAY *) x));
}
static void *_p_TPMS_TAGGED_PROPERTY_ARRAYTo_p_TPMS_TAGGED_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TAGGED_PROPERTY *)  ((TPMS_TAGGED_PROPERTY_ARRAY *) x));
}
static void *_p_TPMS_ALG_PROPERTY_ARRAYTo_p_TPMS_ALG_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ALG_PROPERTY *)  ((TPMS_ALG_PROPERTY_ARRAY *) x));
}
static void *_p_ESYS_TR_PTRTo_p_unsigned_int(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((unsigned int *)  ((ESYS_TR_PTR *) x));
}
static void *_p_TPMS_CONTEXT_DATA_ARRAYTo_p_TPMS_CONTEXT_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CONTEXT_DATA *)  ((TPMS_CONTEXT_DATA_ARRAY *) x));
}
static void *_p_TPM2B_CONTEXT_DATA_ARRAYTo_p_TPM2B_CONTEXT_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_CONTEXT_DATA *)  ((TPM2B_CONTEXT_DATA_ARRAY *) x));
}
static void *_p_p_TPMS_CONTEXT_DATA_ARRAYTo_p_p_TPMS_CONTEXT_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CONTEXT_DATA **)  ((TPMS_CONTEXT_DATA_ARRAY **) x));
}
static void *_p_p_TPM2B_CONTEXT_DATA_ARRAYTo_p_p_TPM2B_CONTEXT_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_CONTEXT_DATA **)  ((TPM2B_CONTEXT_DATA_ARRAY **) x));
}
static void *_p_p_TPMT_SYM_DEF_ARRAYTo_p_p_TPMT_SYM_DEF(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SYM_DEF **)  ((TPMT_SYM_DEF_ARRAY **) x));
}
static void *_p_TPMT_SYM_DEF_ARRAYTo_p_TPMT_SYM_DEF(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SYM_DEF *)  ((TPMT_SYM_DEF_ARRAY *) x));
}
static void *_p_p_TPMT_SYM_DEF_OBJECT_ARRAYTo_p_p_TPMT_SYM_DEF_OBJECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SYM_DEF_OBJECT **)  ((TPMT_SYM_DEF_OBJECT_ARRAY **) x));
}
static void *_p_TPMT_SYM_DEF_OBJECT_ARRAYTo_p_TPMT_SYM_DEF_OBJECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SYM_DEF_OBJECT *)  ((TPMT_SYM_DEF_OBJECT_ARRAY *) x));
}
static void *_p_TPMS_ID_OBJECT_ARRAYTo_p_TPMS_ID_OBJECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ID_OBJECT *)  ((TPMS_ID_OBJECT_ARRAY *) x));
}
static void *_p_TPM2B_ID_OBJECT_ARRAYTo_p_TPM2B_ID_OBJECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ID_OBJECT *)  ((TPM2B_ID_OBJECT_ARRAY *) x));
}
static void *_p_p_TPMS_ID_OBJECT_ARRAYTo_p_p_TPMS_ID_OBJECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ID_OBJECT **)  ((TPMS_ID_OBJECT_ARRAY **) x));
}
static void *_p_p_TPM2B_ID_OBJECT_ARRAYTo_p_p_TPM2B_ID_OBJECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ID_OBJECT **)  ((TPM2B_ID_OBJECT_ARRAY **) x));
}
static void *_p_p_TPMS_TAGGED_POLICY_ARRAYTo_p_p_TPMS_TAGGED_POLICY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TAGGED_POLICY **)  ((TPMS_TAGGED_POLICY_ARRAY **) x));
}
static void *_p_TPMS_TAGGED_POLICY_ARRAYTo_p_TPMS_TAGGED_POLICY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TAGGED_POLICY *)  ((TPMS_TAGGED_POLICY_ARRAY *) x));
}
static void *_p_TPM2B_TEMPLATE_ARRAYTo_p_TPM2B_TEMPLATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_TEMPLATE *)  ((TPM2B_TEMPLATE_ARRAY *) x));
}
static void *_p_p_TPM2B_TEMPLATE_ARRAYTo_p_p_TPM2B_TEMPLATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_TEMPLATE **)  ((TPM2B_TEMPLATE_ARRAY **) x));
}
static void *_p_p_TPMU_SIGNATURE_ARRAYTo_p_p_TPMU_SIGNATURE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SIGNATURE **)  ((TPMU_SIGNATURE_ARRAY **) x));
}
static void *_p_TPMU_SIGNATURE_ARRAYTo_p_TPMU_SIGNATURE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SIGNATURE *)  ((TPMU_SIGNATURE_ARRAY *) x));
}
static void *_p_TPMT_SIGNATURE_ARRAYTo_p_TPMT_SIGNATURE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SIGNATURE *)  ((TPMT_SIGNATURE_ARRAY *) x));
}
static void *_p_p_TPMT_SIGNATURE_ARRAYTo_p_p_TPMT_SIGNATURE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SIGNATURE **)  ((TPMT_SIGNATURE_ARRAY **) x));
}
static void *_p_p_TPMS_SIGNATURE_RSA_ARRAYTo_p_p_TPMS_SIGNATURE_RSA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SIGNATURE_RSA **)  ((TPMS_SIGNATURE_RSA_ARRAY **) x));
}
static void *_p_TPMS_SIGNATURE_RSA_ARRAYTo_p_TPMS_SIGNATURE_RSA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SIGNATURE_RSA *)  ((TPMS_SIGNATURE_RSA_ARRAY *) x));
}
static void *_p_TPMT_SENSITIVE_ARRAYTo_p_TPMT_SENSITIVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SENSITIVE *)  ((TPMT_SENSITIVE_ARRAY *) x));
}
static void *_p_TPM2B_SENSITIVE_ARRAYTo_p_TPM2B_SENSITIVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SENSITIVE *)  ((TPM2B_SENSITIVE_ARRAY *) x));
}
static void *_p_TPM2B_CONTEXT_SENSITIVE_ARRAYTo_p_TPM2B_CONTEXT_SENSITIVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_CONTEXT_SENSITIVE *)  ((TPM2B_CONTEXT_SENSITIVE_ARRAY *) x));
}
static void *_p_p_TPMT_SENSITIVE_ARRAYTo_p_p_TPMT_SENSITIVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SENSITIVE **)  ((TPMT_SENSITIVE_ARRAY **) x));
}
static void *_p_p_TPM2B_SENSITIVE_ARRAYTo_p_p_TPM2B_SENSITIVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SENSITIVE **)  ((TPM2B_SENSITIVE_ARRAY **) x));
}
static void *_p_p_TPM2B_CONTEXT_SENSITIVE_ARRAYTo_p_p_TPM2B_CONTEXT_SENSITIVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_CONTEXT_SENSITIVE **)  ((TPM2B_CONTEXT_SENSITIVE_ARRAY **) x));
}
static void *_p_p_TPM2B_ATTEST_ARRAYTo_p_p_TPM2B_ATTEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ATTEST **)  ((TPM2B_ATTEST_ARRAY **) x));
}
static void *_p_p_TPMS_ATTEST_ARRAYTo_p_p_TPMS_ATTEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ATTEST **)  ((TPMS_ATTEST_ARRAY **) x));
}
static void *_p_p_TPMU_ATTEST_ARRAYTo_p_p_TPMU_ATTEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_ATTEST **)  ((TPMU_ATTEST_ARRAY **) x));
}
static void *_p_TPMU_ATTEST_ARRAYTo_p_TPMU_ATTEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_ATTEST *)  ((TPMU_ATTEST_ARRAY *) x));
}
static void *_p_TPMS_ATTEST_ARRAYTo_p_TPMS_ATTEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ATTEST *)  ((TPMS_ATTEST_ARRAY *) x));
}
static void *_p_TPM2B_ATTEST_ARRAYTo_p_TPM2B_ATTEST(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ATTEST *)  ((TPM2B_ATTEST_ARRAY *) x));
}
static void *_p_p_TPML_DIGEST_VALUES_ARRAYTo_p_p_TPML_DIGEST_VALUES(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_DIGEST_VALUES **)  ((TPML_DIGEST_VALUES_ARRAY **) x));
}
static void *_p_TPML_DIGEST_VALUES_ARRAYTo_p_TPML_DIGEST_VALUES(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_DIGEST_VALUES *)  ((TPML_DIGEST_VALUES_ARRAY *) x));
}
static void *_p_p_TPMU_CAPABILITIES_ARRAYTo_p_p_TPMU_CAPABILITIES(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_CAPABILITIES **)  ((TPMU_CAPABILITIES_ARRAY **) x));
}
static void *_p_TPML_AC_CAPABILITIES_ARRAYTo_p_TPML_AC_CAPABILITIES(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_AC_CAPABILITIES *)  ((TPML_AC_CAPABILITIES_ARRAY *) x));
}
static void *_p_TPMU_CAPABILITIES_ARRAYTo_p_TPMU_CAPABILITIES(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_CAPABILITIES *)  ((TPMU_CAPABILITIES_ARRAY *) x));
}
static void *_p_p_TPML_AC_CAPABILITIES_ARRAYTo_p_p_TPML_AC_CAPABILITIES(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_AC_CAPABILITIES **)  ((TPML_AC_CAPABILITIES_ARRAY **) x));
}
static void *_p_p_TPMU_NAME_ARRAYTo_p_p_TPMU_NAME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_NAME **)  ((TPMU_NAME_ARRAY **) x));
}
static void *_p_TPMU_NAME_ARRAYTo_p_TPMU_NAME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_NAME *)  ((TPMU_NAME_ARRAY *) x));
}
static void *_p_p_TPML_CC_ARRAYTo_p_p_TPML_CC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_CC **)  ((TPML_CC_ARRAY **) x));
}
static void *_p_TPML_CC_ARRAYTo_p_TPML_CC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_CC *)  ((TPML_CC_ARRAY *) x));
}
static void *_p__PRIVATE_ARRAYTo_p__PRIVATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((_PRIVATE *)  ((_PRIVATE_ARRAY *) x));
}
static void *_p_TPM2B_PRIVATE_ARRAYTo_p_TPM2B_PRIVATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PRIVATE *)  ((TPM2B_PRIVATE_ARRAY *) x));
}
static void *_p_p__PRIVATE_ARRAYTo_p_p__PRIVATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((_PRIVATE **)  ((_PRIVATE_ARRAY **) x));
}
static void *_p_p_TPM2B_PRIVATE_ARRAYTo_p_p_TPM2B_PRIVATE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PRIVATE **)  ((TPM2B_PRIVATE_ARRAY **) x));
}
static void *_p_p_TPM2B_EVENT_ARRAYTo_p_p_TPM2B_EVENT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_EVENT **)  ((TPM2B_EVENT_ARRAY **) x));
}
static void *_p_TPM2B_EVENT_ARRAYTo_p_TPM2B_EVENT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_EVENT *)  ((TPM2B_EVENT_ARRAY *) x));
}
static void *_p_p_TPMS_QUOTE_INFO_ARRAYTo_p_p_TPMS_QUOTE_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_QUOTE_INFO **)  ((TPMS_QUOTE_INFO_ARRAY **) x));
}
static void *_p_TPMS_QUOTE_INFO_ARRAYTo_p_TPMS_QUOTE_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_QUOTE_INFO *)  ((TPMS_QUOTE_INFO_ARRAY *) x));
}
static void *_p_p_TPMS_TIME_INFO_ARRAYTo_p_p_TPMS_TIME_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TIME_INFO **)  ((TPMS_TIME_INFO_ARRAY **) x));
}
static void *_p_TPMS_TIME_INFO_ARRAYTo_p_TPMS_TIME_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TIME_INFO *)  ((TPMS_TIME_INFO_ARRAY *) x));
}
static void *_p_p_TPMS_SCHEME_HASH_ARRAYTo_p_p_TPMS_SCHEME_HASH(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SCHEME_HASH **)  ((TPMS_SCHEME_HASH_ARRAY **) x));
}
static void *_p_TPMS_SCHEME_HASH_ARRAYTo_p_TPMS_SCHEME_HASH(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SCHEME_HASH *)  ((TPMS_SCHEME_HASH_ARRAY *) x));
}
static void *_p_TPMT_PUBLIC_ARRAYTo_p_TPMT_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_PUBLIC *)  ((TPMT_PUBLIC_ARRAY *) x));
}
static void *_p_TPM2B_PUBLIC_ARRAYTo_p_TPM2B_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PUBLIC *)  ((TPM2B_PUBLIC_ARRAY *) x));
}
static void *_p_TPMS_NV_PUBLIC_ARRAYTo_p_TPMS_NV_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_NV_PUBLIC *)  ((TPMS_NV_PUBLIC_ARRAY *) x));
}
static void *_p_TPM2B_NV_PUBLIC_ARRAYTo_p_TPM2B_NV_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_NV_PUBLIC *)  ((TPM2B_NV_PUBLIC_ARRAY *) x));
}
static void *_p_p_TPMT_PUBLIC_ARRAYTo_p_p_TPMT_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_PUBLIC **)  ((TPMT_PUBLIC_ARRAY **) x));
}
static void *_p_p_TPM2B_PUBLIC_ARRAYTo_p_p_TPM2B_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PUBLIC **)  ((TPM2B_PUBLIC_ARRAY **) x));
}
static void *_p_p_TPMS_NV_PUBLIC_ARRAYTo_p_p_TPMS_NV_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_NV_PUBLIC **)  ((TPMS_NV_PUBLIC_ARRAY **) x));
}
static void *_p_p_TPM2B_NV_PUBLIC_ARRAYTo_p_p_TPM2B_NV_PUBLIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_NV_PUBLIC **)  ((TPM2B_NV_PUBLIC_ARRAY **) x));
}
static void *_p_p_TPMU_SYM_KEY_BITS_ARRAYTo_p_p_TPMU_SYM_KEY_BITS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SYM_KEY_BITS **)  ((TPMU_SYM_KEY_BITS_ARRAY **) x));
}
static void *_p_TPMU_SYM_KEY_BITS_ARRAYTo_p_TPMU_SYM_KEY_BITS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SYM_KEY_BITS *)  ((TPMU_SYM_KEY_BITS_ARRAY *) x));
}
static void *_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAYTo_p_TPMS_NV_PIN_COUNTER_PARAMETERS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_NV_PIN_COUNTER_PARAMETERS *)  ((TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *) x));
}
static void *_p_p_TPMS_SIGNATURE_ECC_ARRAYTo_p_p_TPMS_SIGNATURE_ECC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SIGNATURE_ECC **)  ((TPMS_SIGNATURE_ECC_ARRAY **) x));
}
static void *_p_TPMS_SIGNATURE_ECC_ARRAYTo_p_TPMS_SIGNATURE_ECC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SIGNATURE_ECC *)  ((TPMS_SIGNATURE_ECC_ARRAY *) x));
}
static void *_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAYTo_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_NV_PIN_COUNTER_PARAMETERS **)  ((TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY **) x));
}
static void *_p_p_TPML_PCR_SELECTION_ARRAYTo_p_p_TPML_PCR_SELECTION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_PCR_SELECTION **)  ((TPML_PCR_SELECTION_ARRAY **) x));
}
static void *_p_p_TPMS_PCR_SELECTION_ARRAYTo_p_p_TPMS_PCR_SELECTION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_PCR_SELECTION **)  ((TPMS_PCR_SELECTION_ARRAY **) x));
}
static void *_p_TPMS_PCR_SELECTION_ARRAYTo_p_TPMS_PCR_SELECTION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_PCR_SELECTION *)  ((TPMS_PCR_SELECTION_ARRAY *) x));
}
static void *_p_TPML_PCR_SELECTION_ARRAYTo_p_TPML_PCR_SELECTION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_PCR_SELECTION *)  ((TPML_PCR_SELECTION_ARRAY *) x));
}
static void *_p_p_TPMS_TAGGED_PCR_SELECT_ARRAYTo_p_p_TPMS_TAGGED_PCR_SELECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TAGGED_PCR_SELECT **)  ((TPMS_TAGGED_PCR_SELECT_ARRAY **) x));
}
static void *_p_p_TPMS_PCR_SELECT_ARRAYTo_p_p_TPMS_PCR_SELECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_PCR_SELECT **)  ((TPMS_PCR_SELECT_ARRAY **) x));
}
static void *_p_TPMS_PCR_SELECT_ARRAYTo_p_TPMS_PCR_SELECT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_PCR_SELECT *)  ((TPMS_PCR_SELECT_ARRAY *) x));
}
static void *_p_TPMS_ASYM_PARMS_ARRAYTo_p_TPMS_ASYM_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ASYM_PARMS *)  ((TPMS_ASYM_PARMS_ARRAY *) x));
}
static void *_p_TPMS_RSA_PARMS_ARRAYTo_p_TPMS_RSA_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_RSA_PARMS *)  ((TPMS_RSA_PARMS_ARRAY *) x));
}
static void *_p_TPMS_ECC_PARMS_ARRAYTo_p_TPMS_ECC_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ECC_PARMS *)  ((TPMS_ECC_PARMS_ARRAY *) x));
}
static void *_p_TPMU_PUBLIC_PARMS_ARRAYTo_p_TPMU_PUBLIC_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_PUBLIC_PARMS *)  ((TPMU_PUBLIC_PARMS_ARRAY *) x));
}
static void *_p_TPMT_PUBLIC_PARMS_ARRAYTo_p_TPMT_PUBLIC_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_PUBLIC_PARMS *)  ((TPMT_PUBLIC_PARMS_ARRAY *) x));
}
static void *_p_p_TPMS_ASYM_PARMS_ARRAYTo_p_p_TPMS_ASYM_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ASYM_PARMS **)  ((TPMS_ASYM_PARMS_ARRAY **) x));
}
static void *_p_p_TPMS_RSA_PARMS_ARRAYTo_p_p_TPMS_RSA_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_RSA_PARMS **)  ((TPMS_RSA_PARMS_ARRAY **) x));
}
static void *_p_p_TPMS_ECC_PARMS_ARRAYTo_p_p_TPMS_ECC_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ECC_PARMS **)  ((TPMS_ECC_PARMS_ARRAY **) x));
}
static void *_p_p_TPMU_PUBLIC_PARMS_ARRAYTo_p_p_TPMU_PUBLIC_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_PUBLIC_PARMS **)  ((TPMU_PUBLIC_PARMS_ARRAY **) x));
}
static void *_p_p_TPMT_PUBLIC_PARMS_ARRAYTo_p_p_TPMT_PUBLIC_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_PUBLIC_PARMS **)  ((TPMT_PUBLIC_PARMS_ARRAY **) x));
}
static void *_p_p_TPMS_SYMCIPHER_PARMS_ARRAYTo_p_p_TPMS_SYMCIPHER_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SYMCIPHER_PARMS **)  ((TPMS_SYMCIPHER_PARMS_ARRAY **) x));
}
static void *_p_TPMS_SYMCIPHER_PARMS_ARRAYTo_p_TPMS_SYMCIPHER_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SYMCIPHER_PARMS *)  ((TPMS_SYMCIPHER_PARMS_ARRAY *) x));
}
static void *_p_TPMS_KEYEDHASH_PARMS_ARRAYTo_p_TPMS_KEYEDHASH_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_KEYEDHASH_PARMS *)  ((TPMS_KEYEDHASH_PARMS_ARRAY *) x));
}
static void *_p_p_TPMS_KEYEDHASH_PARMS_ARRAYTo_p_p_TPMS_KEYEDHASH_PARMS(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_KEYEDHASH_PARMS **)  ((TPMS_KEYEDHASH_PARMS_ARRAY **) x));
}
static void *_p_p_TPM2B_PRIVATE_KEY_RSA_ARRAYTo_p_p_TPM2B_PRIVATE_KEY_RSA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PRIVATE_KEY_RSA **)  ((TPM2B_PRIVATE_KEY_RSA_ARRAY **) x));
}
static void *_p_p_TPM2B_PUBLIC_KEY_RSA_ARRAYTo_p_p_TPM2B_PUBLIC_KEY_RSA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PUBLIC_KEY_RSA **)  ((TPM2B_PUBLIC_KEY_RSA_ARRAY **) x));
}
static void *_p_TPM2B_PUBLIC_KEY_RSA_ARRAYTo_p_TPM2B_PUBLIC_KEY_RSA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PUBLIC_KEY_RSA *)  ((TPM2B_PUBLIC_KEY_RSA_ARRAY *) x));
}
static void *_p_TPM2B_PRIVATE_KEY_RSA_ARRAYTo_p_TPM2B_PRIVATE_KEY_RSA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PRIVATE_KEY_RSA *)  ((TPM2B_PRIVATE_KEY_RSA_ARRAY *) x));
}
static void *_p_p_TPM2B_SENSITIVE_DATA_ARRAYTo_p_p_TPM2B_SENSITIVE_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SENSITIVE_DATA **)  ((TPM2B_SENSITIVE_DATA_ARRAY **) x));
}
static void *_p_TPM2B_SENSITIVE_DATA_ARRAYTo_p_TPM2B_SENSITIVE_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SENSITIVE_DATA *)  ((TPM2B_SENSITIVE_DATA_ARRAY *) x));
}
static void *_p_p_TPMU_SYM_MODE_ARRAYTo_p_p_TPMU_SYM_MODE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SYM_MODE **)  ((TPMU_SYM_MODE_ARRAY **) x));
}
static void *_p_TPMU_SYM_MODE_ARRAYTo_p_TPMU_SYM_MODE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SYM_MODE *)  ((TPMU_SYM_MODE_ARRAY *) x));
}
static void *_p_p_TPML_ECC_CURVE_ARRAYTo_p_p_TPML_ECC_CURVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_ECC_CURVE **)  ((TPML_ECC_CURVE_ARRAY **) x));
}
static void *_p_TPML_ECC_CURVE_ARRAYTo_p_TPML_ECC_CURVE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_ECC_CURVE *)  ((TPML_ECC_CURVE_ARRAY *) x));
}
static void *_p_p_TPMS_NV_CERTIFY_INFO_ARRAYTo_p_p_TPMS_NV_CERTIFY_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_NV_CERTIFY_INFO **)  ((TPMS_NV_CERTIFY_INFO_ARRAY **) x));
}
static void *_p_p_TPMS_CERTIFY_INFO_ARRAYTo_p_p_TPMS_CERTIFY_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CERTIFY_INFO **)  ((TPMS_CERTIFY_INFO_ARRAY **) x));
}
static void *_p_TPMS_CERTIFY_INFO_ARRAYTo_p_TPMS_CERTIFY_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CERTIFY_INFO *)  ((TPMS_CERTIFY_INFO_ARRAY *) x));
}
static void *_p_TPMS_NV_CERTIFY_INFO_ARRAYTo_p_TPMS_NV_CERTIFY_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_NV_CERTIFY_INFO *)  ((TPMS_NV_CERTIFY_INFO_ARRAY *) x));
}
static void *_p_p_ESYS_TR_PTRTo_p_p_unsigned_int(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((unsigned int **)  ((ESYS_TR_PTR **) x));
}
static void *_p_p_TPMS_CREATION_INFO_ARRAYTo_p_p_TPMS_CREATION_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CREATION_INFO **)  ((TPMS_CREATION_INFO_ARRAY **) x));
}
static void *_p_TPMS_CREATION_INFO_ARRAYTo_p_TPMS_CREATION_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CREATION_INFO *)  ((TPMS_CREATION_INFO_ARRAY *) x));
}
static void *_p_p_TPMT_ECC_SCHEME_ARRAYTo_p_p_TPMT_ECC_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_ECC_SCHEME **)  ((TPMT_ECC_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMT_RSA_SCHEME_ARRAYTo_p_p_TPMT_RSA_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_RSA_SCHEME **)  ((TPMT_RSA_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMT_ASYM_SCHEME_ARRAYTo_p_p_TPMT_ASYM_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_ASYM_SCHEME **)  ((TPMT_ASYM_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMU_ASYM_SCHEME_ARRAYTo_p_p_TPMU_ASYM_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_ASYM_SCHEME **)  ((TPMU_ASYM_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMT_KDF_SCHEME_ARRAYTo_p_p_TPMT_KDF_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_KDF_SCHEME **)  ((TPMT_KDF_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMU_KDF_SCHEME_ARRAYTo_p_p_TPMU_KDF_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_KDF_SCHEME **)  ((TPMU_KDF_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMT_SIG_SCHEME_ARRAYTo_p_p_TPMT_SIG_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SIG_SCHEME **)  ((TPMT_SIG_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMU_SIG_SCHEME_ARRAYTo_p_p_TPMU_SIG_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SIG_SCHEME **)  ((TPMU_SIG_SCHEME_ARRAY **) x));
}
static void *_p_p_TPMT_KEYEDHASH_SCHEME_ARRAYTo_p_p_TPMT_KEYEDHASH_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_KEYEDHASH_SCHEME **)  ((TPMT_KEYEDHASH_SCHEME_ARRAY **) x));
}
static void *_p_TPMT_KEYEDHASH_SCHEME_ARRAYTo_p_TPMT_KEYEDHASH_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_KEYEDHASH_SCHEME *)  ((TPMT_KEYEDHASH_SCHEME_ARRAY *) x));
}
static void *_p_TPMU_SIG_SCHEME_ARRAYTo_p_TPMU_SIG_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SIG_SCHEME *)  ((TPMU_SIG_SCHEME_ARRAY *) x));
}
static void *_p_TPMT_SIG_SCHEME_ARRAYTo_p_TPMT_SIG_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_SIG_SCHEME *)  ((TPMT_SIG_SCHEME_ARRAY *) x));
}
static void *_p_TPMU_KDF_SCHEME_ARRAYTo_p_TPMU_KDF_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_KDF_SCHEME *)  ((TPMU_KDF_SCHEME_ARRAY *) x));
}
static void *_p_TPMT_KDF_SCHEME_ARRAYTo_p_TPMT_KDF_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_KDF_SCHEME *)  ((TPMT_KDF_SCHEME_ARRAY *) x));
}
static void *_p_TPMU_ASYM_SCHEME_ARRAYTo_p_TPMU_ASYM_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_ASYM_SCHEME *)  ((TPMU_ASYM_SCHEME_ARRAY *) x));
}
static void *_p_TPMT_ASYM_SCHEME_ARRAYTo_p_TPMT_ASYM_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_ASYM_SCHEME *)  ((TPMT_ASYM_SCHEME_ARRAY *) x));
}
static void *_p_TPMT_RSA_SCHEME_ARRAYTo_p_TPMT_RSA_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_RSA_SCHEME *)  ((TPMT_RSA_SCHEME_ARRAY *) x));
}
static void *_p_TPMT_ECC_SCHEME_ARRAYTo_p_TPMT_ECC_SCHEME(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_ECC_SCHEME *)  ((TPMT_ECC_SCHEME_ARRAY *) x));
}
static void *_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAYTo_p_TPM2B_PRIVATE_VENDOR_SPECIFIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PRIVATE_VENDOR_SPECIFIC *)  ((TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *) x));
}
static void *_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAYTo_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_PRIVATE_VENDOR_SPECIFIC **)  ((TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY **) x));
}
static void *_p_p_TPML_HANDLE_ARRAYTo_p_p_TPML_HANDLE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_HANDLE **)  ((TPML_HANDLE_ARRAY **) x));
}
static void *_p_TPML_HANDLE_ARRAYTo_p_TPML_HANDLE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_HANDLE *)  ((TPML_HANDLE_ARRAY *) x));
}
static void *_p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAYTo_p_p_TPMS_ALGORITHM_DETAIL_ECC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ALGORITHM_DETAIL_ECC **)  ((TPMS_ALGORITHM_DETAIL_ECC_ARRAY **) x));
}
static void *_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAYTo_p_TPMS_ALGORITHM_DETAIL_ECC(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ALGORITHM_DETAIL_ECC *)  ((TPMS_ALGORITHM_DETAIL_ECC_ARRAY *) x));
}
static void *_p_p_TPMS_CAPABILITY_DATA_ARRAYTo_p_p_TPMS_CAPABILITY_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CAPABILITY_DATA **)  ((TPMS_CAPABILITY_DATA_ARRAY **) x));
}
static void *_p_TPMS_CAPABILITY_DATA_ARRAYTo_p_TPMS_CAPABILITY_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CAPABILITY_DATA *)  ((TPMS_CAPABILITY_DATA_ARRAY *) x));
}
static void *_p_p_TPMT_TK_CREATION_ARRAYTo_p_p_TPMT_TK_CREATION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_CREATION **)  ((TPMT_TK_CREATION_ARRAY **) x));
}
static void *_p_TPMT_TK_CREATION_ARRAYTo_p_TPMT_TK_CREATION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_CREATION *)  ((TPMT_TK_CREATION_ARRAY *) x));
}
static void *_p_TPMS_CREATION_DATA_ARRAYTo_p_TPMS_CREATION_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CREATION_DATA *)  ((TPMS_CREATION_DATA_ARRAY *) x));
}
static void *_p_TPM2B_CREATION_DATA_ARRAYTo_p_TPM2B_CREATION_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_CREATION_DATA *)  ((TPM2B_CREATION_DATA_ARRAY *) x));
}
static void *_p_p_TPMS_CREATION_DATA_ARRAYTo_p_p_TPMS_CREATION_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CREATION_DATA **)  ((TPMS_CREATION_DATA_ARRAY **) x));
}
static void *_p_p_TPM2B_CREATION_DATA_ARRAYTo_p_p_TPM2B_CREATION_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_CREATION_DATA **)  ((TPM2B_CREATION_DATA_ARRAY **) x));
}
static void *_p_p_TPML_INTEL_PTT_PROPERTY_ARRAYTo_p_p_TPML_INTEL_PTT_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_INTEL_PTT_PROPERTY **)  ((TPML_INTEL_PTT_PROPERTY_ARRAY **) x));
}
static void *_p_p_TPML_TAGGED_PCR_PROPERTY_ARRAYTo_p_p_TPML_TAGGED_PCR_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_TAGGED_PCR_PROPERTY **)  ((TPML_TAGGED_PCR_PROPERTY_ARRAY **) x));
}
static void *_p_p_TPML_TAGGED_TPM_PROPERTY_ARRAYTo_p_p_TPML_TAGGED_TPM_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_TAGGED_TPM_PROPERTY **)  ((TPML_TAGGED_TPM_PROPERTY_ARRAY **) x));
}
static void *_p_p_TPML_ALG_PROPERTY_ARRAYTo_p_p_TPML_ALG_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_ALG_PROPERTY **)  ((TPML_ALG_PROPERTY_ARRAY **) x));
}
static void *_p_p_TPMS_TAGGED_PROPERTY_ARRAYTo_p_p_TPMS_TAGGED_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_TAGGED_PROPERTY **)  ((TPMS_TAGGED_PROPERTY_ARRAY **) x));
}
static void *_p_p_TPMS_ALG_PROPERTY_ARRAYTo_p_p_TPMS_ALG_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ALG_PROPERTY **)  ((TPMS_ALG_PROPERTY_ARRAY **) x));
}
static void *_p_TPML_INTEL_PTT_PROPERTY_ARRAYTo_p_TPML_INTEL_PTT_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_INTEL_PTT_PROPERTY *)  ((TPML_INTEL_PTT_PROPERTY_ARRAY *) x));
}
static void *_p_TPML_TAGGED_PCR_PROPERTY_ARRAYTo_p_TPML_TAGGED_PCR_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_TAGGED_PCR_PROPERTY *)  ((TPML_TAGGED_PCR_PROPERTY_ARRAY *) x));
}
static void *_p_TPML_TAGGED_TPM_PROPERTY_ARRAYTo_p_TPML_TAGGED_TPM_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_TAGGED_TPM_PROPERTY *)  ((TPML_TAGGED_TPM_PROPERTY_ARRAY *) x));
}
static void *_p_TPML_ALG_PROPERTY_ARRAYTo_p_TPML_ALG_PROPERTY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_ALG_PROPERTY *)  ((TPML_ALG_PROPERTY_ARRAY *) x));
}
static void *_p_p_TPMT_TK_HASHCHECK_ARRAYTo_p_p_TPMT_TK_HASHCHECK(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_HASHCHECK **)  ((TPMT_TK_HASHCHECK_ARRAY **) x));
}
static void *_p_TPMT_TK_HASHCHECK_ARRAYTo_p_TPMT_TK_HASHCHECK(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_HASHCHECK *)  ((TPMT_TK_HASHCHECK_ARRAY *) x));
}
static void *_p_p_TPM2B_ECC_POINT_ARRAYTo_p_p_TPM2B_ECC_POINT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ECC_POINT **)  ((TPM2B_ECC_POINT_ARRAY **) x));
}
static void *_p_p_TPMS_ECC_POINT_ARRAYTo_p_p_TPMS_ECC_POINT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ECC_POINT **)  ((TPMS_ECC_POINT_ARRAY **) x));
}
static void *_p_TPMS_ECC_POINT_ARRAYTo_p_TPMS_ECC_POINT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ECC_POINT *)  ((TPMS_ECC_POINT_ARRAY *) x));
}
static void *_p_TPM2B_ECC_POINT_ARRAYTo_p_TPM2B_ECC_POINT(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ECC_POINT *)  ((TPM2B_ECC_POINT_ARRAY *) x));
}
static void *_p_p_TPMS_AUTH_COMMAND_ARRAYTo_p_p_TPMS_AUTH_COMMAND(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_AUTH_COMMAND **)  ((TPMS_AUTH_COMMAND_ARRAY **) x));
}
static void *_p_TPMS_AUTH_COMMAND_ARRAYTo_p_TPMS_AUTH_COMMAND(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_AUTH_COMMAND *)  ((TPMS_AUTH_COMMAND_ARRAY *) x));
}
static void *_p_p_TPMT_TK_VERIFIED_ARRAYTo_p_p_TPMT_TK_VERIFIED(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_VERIFIED **)  ((TPMT_TK_VERIFIED_ARRAY **) x));
}
static void *_p_TPMT_TK_VERIFIED_ARRAYTo_p_TPMT_TK_VERIFIED(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_VERIFIED *)  ((TPMT_TK_VERIFIED_ARRAY *) x));
}
static void *_p_TPMU_PUBLIC_ID_ARRAYTo_p_TPMU_PUBLIC_ID(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_PUBLIC_ID *)  ((TPMU_PUBLIC_ID_ARRAY *) x));
}
static void *_p_p_TPMU_PUBLIC_ID_ARRAYTo_p_p_TPMU_PUBLIC_ID(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_PUBLIC_ID **)  ((TPMU_PUBLIC_ID_ARRAY **) x));
}
static void *_p_TPMU_SENSITIVE_COMPOSITE_ARRAYTo_p_TPMU_SENSITIVE_COMPOSITE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SENSITIVE_COMPOSITE *)  ((TPMU_SENSITIVE_COMPOSITE_ARRAY *) x));
}
static void *_p_p_TPMU_SENSITIVE_COMPOSITE_ARRAYTo_p_p_TPMU_SENSITIVE_COMPOSITE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_SENSITIVE_COMPOSITE **)  ((TPMU_SENSITIVE_COMPOSITE_ARRAY **) x));
}
static void *_p_p_TPMS_AUTH_RESPONSE_ARRAYTo_p_p_TPMS_AUTH_RESPONSE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_AUTH_RESPONSE **)  ((TPMS_AUTH_RESPONSE_ARRAY **) x));
}
static void *_p_TPMS_AUTH_RESPONSE_ARRAYTo_p_TPMS_AUTH_RESPONSE(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_AUTH_RESPONSE *)  ((TPMS_AUTH_RESPONSE_ARRAY *) x));
}
static void *_p_p_TPML_ALG_ARRAYTo_p_p_TPML_ALG(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_ALG **)  ((TPML_ALG_ARRAY **) x));
}
static void *_p_TPML_ALG_ARRAYTo_p_TPML_ALG(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPML_ALG *)  ((TPML_ALG_ARRAY *) x));
}
static void *_p_p_TPM2B_DATA_ARRAYTo_p_p_TPM2B_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_DATA **)  ((TPM2B_DATA_ARRAY **) x));
}
static void *_p_TPM2B_DATA_ARRAYTo_p_TPM2B_DATA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_DATA *)  ((TPM2B_DATA_ARRAY *) x));
}
static void *_p_p_TPMU_HA_ARRAYTo_p_p_TPMU_HA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_HA **)  ((TPMU_HA_ARRAY **) x));
}
static void *_p_TPMU_HA_ARRAYTo_p_TPMU_HA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_HA *)  ((TPMU_HA_ARRAY *) x));
}
static void *_p_p_TPMT_HA_ARRAYTo_p_p_TPMT_HA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_HA **)  ((TPMT_HA_ARRAY **) x));
}
static void *_p_TPMT_HA_ARRAYTo_p_TPMT_HA(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_HA *)  ((TPMT_HA_ARRAY *) x));
}
static void *_p_p_TPM2B_MAX_NV_BUFFER_ARRAYTo_p_p_TPM2B_MAX_NV_BUFFER(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_MAX_NV_BUFFER **)  ((TPM2B_MAX_NV_BUFFER_ARRAY **) x));
}
static void *_p_p_TPM2B_MAX_BUFFER_ARRAYTo_p_p_TPM2B_MAX_BUFFER(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_MAX_BUFFER **)  ((TPM2B_MAX_BUFFER_ARRAY **) x));
}
static void *_p_TPM2B_MAX_BUFFER_ARRAYTo_p_TPM2B_MAX_BUFFER(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_MAX_BUFFER *)  ((TPM2B_MAX_BUFFER_ARRAY *) x));
}
static void *_p_TPM2B_MAX_NV_BUFFER_ARRAYTo_p_TPM2B_MAX_NV_BUFFER(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_MAX_NV_BUFFER *)  ((TPM2B_MAX_NV_BUFFER_ARRAY *) x));
}
static void *_p_p_TPMS_CLOCK_INFO_ARRAYTo_p_p_TPMS_CLOCK_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CLOCK_INFO **)  ((TPMS_CLOCK_INFO_ARRAY **) x));
}
static void *_p_TPMS_CLOCK_INFO_ARRAYTo_p_TPMS_CLOCK_INFO(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_CLOCK_INFO *)  ((TPMS_CLOCK_INFO_ARRAY *) x));
}
static void *_p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAYTo_p_p_TPMS_ALGORITHM_DESCRIPTION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ALGORITHM_DESCRIPTION **)  ((TPMS_ALGORITHM_DESCRIPTION_ARRAY **) x));
}
static void *_p_TPMS_ALGORITHM_DESCRIPTION_ARRAYTo_p_TPMS_ALGORITHM_DESCRIPTION(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_ALGORITHM_DESCRIPTION *)  ((TPMS_ALGORITHM_DESCRIPTION_ARRAY *) x));
}
static void *_p_p_TPMT_TK_AUTH_ARRAYTo_p_p_TPMT_TK_AUTH(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_AUTH **)  ((TPMT_TK_AUTH_ARRAY **) x));
}
static void *_p_TPMT_TK_AUTH_ARRAYTo_p_TPMT_TK_AUTH(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMT_TK_AUTH *)  ((TPMT_TK_AUTH_ARRAY *) x));
}
static void *_p_p_TPM2B_SYM_KEY_ARRAYTo_p_p_TPM2B_SYM_KEY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SYM_KEY **)  ((TPM2B_SYM_KEY_ARRAY **) x));
}
static void *_p_TPM2B_SYM_KEY_ARRAYTo_p_TPM2B_SYM_KEY(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_SYM_KEY *)  ((TPM2B_SYM_KEY_ARRAY *) x));
}
static void *_p_p_TPMS_SCHEME_XOR_ARRAYTo_p_p_TPMS_SCHEME_XOR(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SCHEME_XOR **)  ((TPMS_SCHEME_XOR_ARRAY **) x));
}
static void *_p_TPMS_SCHEME_XOR_ARRAYTo_p_TPMS_SCHEME_XOR(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMS_SCHEME_XOR *)  ((TPMS_SCHEME_XOR_ARRAY *) x));
}
static void *_p_TPMU_ENCRYPTED_SECRET_ARRAYTo_p_TPMU_ENCRYPTED_SECRET(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_ENCRYPTED_SECRET *)  ((TPMU_ENCRYPTED_SECRET_ARRAY *) x));
}
static void *_p_TPM2B_ENCRYPTED_SECRET_ARRAYTo_p_TPM2B_ENCRYPTED_SECRET(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ENCRYPTED_SECRET *)  ((TPM2B_ENCRYPTED_SECRET_ARRAY *) x));
}
static void *_p_p_TPMU_ENCRYPTED_SECRET_ARRAYTo_p_p_TPMU_ENCRYPTED_SECRET(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPMU_ENCRYPTED_SECRET **)  ((TPMU_ENCRYPTED_SECRET_ARRAY **) x));
}
static void *_p_p_TPM2B_ENCRYPTED_SECRET_ARRAYTo_p_p_TPM2B_ENCRYPTED_SECRET(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_ENCRYPTED_SECRET **)  ((TPM2B_ENCRYPTED_SECRET_ARRAY **) x));
}
static void *_p_p_TPM2B_IV_ARRAYTo_p_p_TPM2B_IV(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_IV **)  ((TPM2B_IV_ARRAY **) x));
}
static void *_p_TPM2B_IV_ARRAYTo_p_TPM2B_IV(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TPM2B_IV *)  ((TPM2B_IV_ARRAY *) x));
}
static swig_type_info _swigt__p_ByteArray = {"_p_ByteArray", "ByteArray *|BYTE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ESYS_CONTEXT = {"_p_ESYS_CONTEXT", "struct ESYS_CONTEXT *|ESYS_CONTEXT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ESYS_TR_PTR = {"_p_ESYS_TR_PTR", "ESYS_TR_PTR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ATTEST = {"_p_TPM2B_ATTEST", "TPM2B_ATTEST *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ATTEST_ARRAY = {"_p_TPM2B_ATTEST_ARRAY", "TPM2B_ATTEST_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_CONTEXT_DATA = {"_p_TPM2B_CONTEXT_DATA", "TPM2B_CONTEXT_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_CONTEXT_DATA_ARRAY = {"_p_TPM2B_CONTEXT_DATA_ARRAY", "TPM2B_CONTEXT_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_CONTEXT_SENSITIVE = {"_p_TPM2B_CONTEXT_SENSITIVE", "TPM2B_CONTEXT_SENSITIVE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_CONTEXT_SENSITIVE_ARRAY = {"_p_TPM2B_CONTEXT_SENSITIVE_ARRAY", "TPM2B_CONTEXT_SENSITIVE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_CREATION_DATA = {"_p_TPM2B_CREATION_DATA", "TPM2B_CREATION_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_CREATION_DATA_ARRAY = {"_p_TPM2B_CREATION_DATA_ARRAY", "TPM2B_CREATION_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_DATA = {"_p_TPM2B_DATA", "TPM2B_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_DATA_ARRAY = {"_p_TPM2B_DATA_ARRAY", "TPM2B_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_DIGEST = {"_p_TPM2B_DIGEST", "TPM2B_TIMEOUT *|TPM2B_OPERAND *|TPM2B_NONCE *|TPM2B_DIGEST *|TPM2B_AUTH *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_DIGEST_ARRAY = {"_p_TPM2B_DIGEST_ARRAY", "TPM2B_DIGEST_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ECC_PARAMETER = {"_p_TPM2B_ECC_PARAMETER", "TPM2B_ECC_PARAMETER *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ECC_PARAMETER_ARRAY = {"_p_TPM2B_ECC_PARAMETER_ARRAY", "TPM2B_ECC_PARAMETER_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ECC_POINT = {"_p_TPM2B_ECC_POINT", "TPM2B_ECC_POINT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ECC_POINT_ARRAY = {"_p_TPM2B_ECC_POINT_ARRAY", "TPM2B_ECC_POINT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ENCRYPTED_SECRET = {"_p_TPM2B_ENCRYPTED_SECRET", "TPM2B_ENCRYPTED_SECRET *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ENCRYPTED_SECRET_ARRAY = {"_p_TPM2B_ENCRYPTED_SECRET_ARRAY", "TPM2B_ENCRYPTED_SECRET_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_EVENT = {"_p_TPM2B_EVENT", "TPM2B_EVENT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_EVENT_ARRAY = {"_p_TPM2B_EVENT_ARRAY", "TPM2B_EVENT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ID_OBJECT = {"_p_TPM2B_ID_OBJECT", "TPM2B_ID_OBJECT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_ID_OBJECT_ARRAY = {"_p_TPM2B_ID_OBJECT_ARRAY", "TPM2B_ID_OBJECT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_IV = {"_p_TPM2B_IV", "TPM2B_IV *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_IV_ARRAY = {"_p_TPM2B_IV_ARRAY", "TPM2B_IV_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_MAX_BUFFER = {"_p_TPM2B_MAX_BUFFER", "TPM2B_MAX_BUFFER *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_MAX_BUFFER_ARRAY = {"_p_TPM2B_MAX_BUFFER_ARRAY", "TPM2B_MAX_BUFFER_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_MAX_NV_BUFFER = {"_p_TPM2B_MAX_NV_BUFFER", "TPM2B_MAX_NV_BUFFER *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_MAX_NV_BUFFER_ARRAY = {"_p_TPM2B_MAX_NV_BUFFER_ARRAY", "TPM2B_MAX_NV_BUFFER_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_NAME = {"_p_TPM2B_NAME", "TPM2B_NAME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_NAME_ARRAY = {"_p_TPM2B_NAME_ARRAY", "TPM2B_NAME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_NV_PUBLIC = {"_p_TPM2B_NV_PUBLIC", "TPM2B_NV_PUBLIC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_NV_PUBLIC_ARRAY = {"_p_TPM2B_NV_PUBLIC_ARRAY", "TPM2B_NV_PUBLIC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PRIVATE = {"_p_TPM2B_PRIVATE", "TPM2B_PRIVATE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PRIVATE_ARRAY = {"_p_TPM2B_PRIVATE_ARRAY", "TPM2B_PRIVATE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PRIVATE_KEY_RSA = {"_p_TPM2B_PRIVATE_KEY_RSA", "TPM2B_PRIVATE_KEY_RSA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PRIVATE_KEY_RSA_ARRAY = {"_p_TPM2B_PRIVATE_KEY_RSA_ARRAY", "TPM2B_PRIVATE_KEY_RSA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC = {"_p_TPM2B_PRIVATE_VENDOR_SPECIFIC", "TPM2B_PRIVATE_VENDOR_SPECIFIC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY = {"_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY", "TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PUBLIC = {"_p_TPM2B_PUBLIC", "TPM2B_PUBLIC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PUBLIC_ARRAY = {"_p_TPM2B_PUBLIC_ARRAY", "TPM2B_PUBLIC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PUBLIC_KEY_RSA = {"_p_TPM2B_PUBLIC_KEY_RSA", "TPM2B_PUBLIC_KEY_RSA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_PUBLIC_KEY_RSA_ARRAY = {"_p_TPM2B_PUBLIC_KEY_RSA_ARRAY", "TPM2B_PUBLIC_KEY_RSA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SENSITIVE = {"_p_TPM2B_SENSITIVE", "TPM2B_SENSITIVE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SENSITIVE_ARRAY = {"_p_TPM2B_SENSITIVE_ARRAY", "TPM2B_SENSITIVE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SENSITIVE_CREATE = {"_p_TPM2B_SENSITIVE_CREATE", "TPM2B_SENSITIVE_CREATE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SENSITIVE_CREATE_ARRAY = {"_p_TPM2B_SENSITIVE_CREATE_ARRAY", "TPM2B_SENSITIVE_CREATE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SENSITIVE_DATA = {"_p_TPM2B_SENSITIVE_DATA", "TPM2B_SENSITIVE_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SENSITIVE_DATA_ARRAY = {"_p_TPM2B_SENSITIVE_DATA_ARRAY", "TPM2B_SENSITIVE_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SYM_KEY = {"_p_TPM2B_SYM_KEY", "TPM2B_SYM_KEY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_SYM_KEY_ARRAY = {"_p_TPM2B_SYM_KEY_ARRAY", "TPM2B_SYM_KEY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_TEMPLATE = {"_p_TPM2B_TEMPLATE", "TPM2B_TEMPLATE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPM2B_TEMPLATE_ARRAY = {"_p_TPM2B_TEMPLATE_ARRAY", "TPM2B_TEMPLATE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_AC_CAPABILITIES = {"_p_TPML_AC_CAPABILITIES", "TPML_AC_CAPABILITIES *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_AC_CAPABILITIES_ARRAY = {"_p_TPML_AC_CAPABILITIES_ARRAY", "TPML_AC_CAPABILITIES_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_ALG = {"_p_TPML_ALG", "TPML_ALG *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_ALG_ARRAY = {"_p_TPML_ALG_ARRAY", "TPML_ALG_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_ALG_PROPERTY = {"_p_TPML_ALG_PROPERTY", "TPML_ALG_PROPERTY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_ALG_PROPERTY_ARRAY = {"_p_TPML_ALG_PROPERTY_ARRAY", "TPML_ALG_PROPERTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_CC = {"_p_TPML_CC", "TPML_CC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_CCA = {"_p_TPML_CCA", "TPML_CCA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_CCA_ARRAY = {"_p_TPML_CCA_ARRAY", "TPML_CCA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_CC_ARRAY = {"_p_TPML_CC_ARRAY", "TPML_CC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_DIGEST = {"_p_TPML_DIGEST", "TPML_DIGEST *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_DIGEST_ARRAY = {"_p_TPML_DIGEST_ARRAY", "TPML_DIGEST_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_DIGEST_VALUES = {"_p_TPML_DIGEST_VALUES", "TPML_DIGEST_VALUES *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_DIGEST_VALUES_ARRAY = {"_p_TPML_DIGEST_VALUES_ARRAY", "TPML_DIGEST_VALUES_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_ECC_CURVE = {"_p_TPML_ECC_CURVE", "TPML_ECC_CURVE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_ECC_CURVE_ARRAY = {"_p_TPML_ECC_CURVE_ARRAY", "TPML_ECC_CURVE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_HANDLE = {"_p_TPML_HANDLE", "TPML_HANDLE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_HANDLE_ARRAY = {"_p_TPML_HANDLE_ARRAY", "TPML_HANDLE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_INTEL_PTT_PROPERTY = {"_p_TPML_INTEL_PTT_PROPERTY", "TPML_INTEL_PTT_PROPERTY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_INTEL_PTT_PROPERTY_ARRAY = {"_p_TPML_INTEL_PTT_PROPERTY_ARRAY", "TPML_INTEL_PTT_PROPERTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_PCR_SELECTION = {"_p_TPML_PCR_SELECTION", "TPML_PCR_SELECTION *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_PCR_SELECTION_ARRAY = {"_p_TPML_PCR_SELECTION_ARRAY", "TPML_PCR_SELECTION_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_TAGGED_PCR_PROPERTY = {"_p_TPML_TAGGED_PCR_PROPERTY", "TPML_TAGGED_PCR_PROPERTY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_TAGGED_PCR_PROPERTY_ARRAY = {"_p_TPML_TAGGED_PCR_PROPERTY_ARRAY", "TPML_TAGGED_PCR_PROPERTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_TAGGED_TPM_PROPERTY = {"_p_TPML_TAGGED_TPM_PROPERTY", "TPML_TAGGED_TPM_PROPERTY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPML_TAGGED_TPM_PROPERTY_ARRAY = {"_p_TPML_TAGGED_TPM_PROPERTY_ARRAY", "TPML_TAGGED_TPM_PROPERTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_AC_OUTPUT = {"_p_TPMS_AC_OUTPUT", "TPMS_AC_OUTPUT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_AC_OUTPUT_ARRAY = {"_p_TPMS_AC_OUTPUT_ARRAY", "TPMS_AC_OUTPUT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ALGORITHM_DESCRIPTION = {"_p_TPMS_ALGORITHM_DESCRIPTION", "TPMS_ALGORITHM_DESCRIPTION *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ALGORITHM_DESCRIPTION_ARRAY = {"_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY", "TPMS_ALGORITHM_DESCRIPTION_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ALGORITHM_DETAIL_ECC = {"_p_TPMS_ALGORITHM_DETAIL_ECC", "TPMS_ALGORITHM_DETAIL_ECC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY = {"_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY", "TPMS_ALGORITHM_DETAIL_ECC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ALG_PROPERTY = {"_p_TPMS_ALG_PROPERTY", "TPMS_ALG_PROPERTY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ALG_PROPERTY_ARRAY = {"_p_TPMS_ALG_PROPERTY_ARRAY", "TPMS_ALG_PROPERTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ASYM_PARMS = {"_p_TPMS_ASYM_PARMS", "TPMS_ASYM_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ASYM_PARMS_ARRAY = {"_p_TPMS_ASYM_PARMS_ARRAY", "TPMS_ASYM_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ATTEST = {"_p_TPMS_ATTEST", "TPMS_ATTEST *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ATTEST_ARRAY = {"_p_TPMS_ATTEST_ARRAY", "TPMS_ATTEST_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_AUTH_COMMAND = {"_p_TPMS_AUTH_COMMAND", "TPMS_AUTH_COMMAND *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_AUTH_COMMAND_ARRAY = {"_p_TPMS_AUTH_COMMAND_ARRAY", "TPMS_AUTH_COMMAND_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_AUTH_RESPONSE = {"_p_TPMS_AUTH_RESPONSE", "TPMS_AUTH_RESPONSE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_AUTH_RESPONSE_ARRAY = {"_p_TPMS_AUTH_RESPONSE_ARRAY", "TPMS_AUTH_RESPONSE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CAPABILITY_DATA = {"_p_TPMS_CAPABILITY_DATA", "TPMS_CAPABILITY_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CAPABILITY_DATA_ARRAY = {"_p_TPMS_CAPABILITY_DATA_ARRAY", "TPMS_CAPABILITY_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CERTIFY_INFO = {"_p_TPMS_CERTIFY_INFO", "TPMS_CERTIFY_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CERTIFY_INFO_ARRAY = {"_p_TPMS_CERTIFY_INFO_ARRAY", "TPMS_CERTIFY_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CLOCK_INFO = {"_p_TPMS_CLOCK_INFO", "TPMS_CLOCK_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CLOCK_INFO_ARRAY = {"_p_TPMS_CLOCK_INFO_ARRAY", "TPMS_CLOCK_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_COMMAND_AUDIT_INFO = {"_p_TPMS_COMMAND_AUDIT_INFO", "TPMS_COMMAND_AUDIT_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_COMMAND_AUDIT_INFO_ARRAY = {"_p_TPMS_COMMAND_AUDIT_INFO_ARRAY", "TPMS_COMMAND_AUDIT_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CONTEXT = {"_p_TPMS_CONTEXT", "TPMS_CONTEXT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CONTEXT_ARRAY = {"_p_TPMS_CONTEXT_ARRAY", "TPMS_CONTEXT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CONTEXT_DATA = {"_p_TPMS_CONTEXT_DATA", "TPMS_CONTEXT_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CONTEXT_DATA_ARRAY = {"_p_TPMS_CONTEXT_DATA_ARRAY", "TPMS_CONTEXT_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CREATION_DATA = {"_p_TPMS_CREATION_DATA", "TPMS_CREATION_DATA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CREATION_DATA_ARRAY = {"_p_TPMS_CREATION_DATA_ARRAY", "TPMS_CREATION_DATA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CREATION_INFO = {"_p_TPMS_CREATION_INFO", "TPMS_CREATION_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_CREATION_INFO_ARRAY = {"_p_TPMS_CREATION_INFO_ARRAY", "TPMS_CREATION_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ECC_PARMS = {"_p_TPMS_ECC_PARMS", "TPMS_ECC_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ECC_PARMS_ARRAY = {"_p_TPMS_ECC_PARMS_ARRAY", "TPMS_ECC_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ECC_POINT = {"_p_TPMS_ECC_POINT", "TPMS_ECC_POINT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ECC_POINT_ARRAY = {"_p_TPMS_ECC_POINT_ARRAY", "TPMS_ECC_POINT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_EMPTY = {"_p_TPMS_EMPTY", "TPMS_EMPTY *|TPMS_ENC_SCHEME_RSAES *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_EMPTY_ARRAY = {"_p_TPMS_EMPTY_ARRAY", "TPMS_EMPTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ID_OBJECT = {"_p_TPMS_ID_OBJECT", "TPMS_ID_OBJECT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_ID_OBJECT_ARRAY = {"_p_TPMS_ID_OBJECT_ARRAY", "TPMS_ID_OBJECT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_KEYEDHASH_PARMS = {"_p_TPMS_KEYEDHASH_PARMS", "TPMS_KEYEDHASH_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_KEYEDHASH_PARMS_ARRAY = {"_p_TPMS_KEYEDHASH_PARMS_ARRAY", "TPMS_KEYEDHASH_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_NV_CERTIFY_INFO = {"_p_TPMS_NV_CERTIFY_INFO", "TPMS_NV_CERTIFY_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_NV_CERTIFY_INFO_ARRAY = {"_p_TPMS_NV_CERTIFY_INFO_ARRAY", "TPMS_NV_CERTIFY_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS = {"_p_TPMS_NV_PIN_COUNTER_PARAMETERS", "TPMS_NV_PIN_COUNTER_PARAMETERS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY = {"_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY", "TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_NV_PUBLIC = {"_p_TPMS_NV_PUBLIC", "TPMS_NV_PUBLIC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_NV_PUBLIC_ARRAY = {"_p_TPMS_NV_PUBLIC_ARRAY", "TPMS_NV_PUBLIC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_PCR_SELECT = {"_p_TPMS_PCR_SELECT", "TPMS_PCR_SELECT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_PCR_SELECTION = {"_p_TPMS_PCR_SELECTION", "TPMS_PCR_SELECTION *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_PCR_SELECTION_ARRAY = {"_p_TPMS_PCR_SELECTION_ARRAY", "TPMS_PCR_SELECTION_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_PCR_SELECT_ARRAY = {"_p_TPMS_PCR_SELECT_ARRAY", "TPMS_PCR_SELECT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_QUOTE_INFO = {"_p_TPMS_QUOTE_INFO", "TPMS_QUOTE_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_QUOTE_INFO_ARRAY = {"_p_TPMS_QUOTE_INFO_ARRAY", "TPMS_QUOTE_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_RSA_PARMS = {"_p_TPMS_RSA_PARMS", "TPMS_RSA_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_RSA_PARMS_ARRAY = {"_p_TPMS_RSA_PARMS_ARRAY", "TPMS_RSA_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SCHEME_ECDAA = {"_p_TPMS_SCHEME_ECDAA", "TPMS_SCHEME_ECDAA *|TPMS_SIG_SCHEME_ECDAA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SCHEME_ECDAA_ARRAY = {"_p_TPMS_SCHEME_ECDAA_ARRAY", "TPMS_SCHEME_ECDAA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SCHEME_HASH = {"_p_TPMS_SCHEME_HASH", "TPMS_SIG_SCHEME_ECDSA *|TPMS_KEY_SCHEME_ECDH *|TPMS_SCHEME_HMAC *|TPMS_SCHEME_KDF1_SP800_108 *|TPMS_SCHEME_KDF1_SP800_56A *|TPMS_SIG_SCHEME_SM2 *|TPMS_SIG_SCHEME_RSAPSS *|TPMS_SCHEME_MGF1 *|TPMS_SIG_SCHEME_ECSCHNORR *|TPMS_KEY_SCHEME_ECMQV *|TPMS_SCHEME_HASH *|TPMS_SCHEME_KDF2 *|TPMS_ENC_SCHEME_OAEP *|TPMS_SIG_SCHEME_RSASSA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SCHEME_HASH_ARRAY = {"_p_TPMS_SCHEME_HASH_ARRAY", "TPMS_SCHEME_HASH_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SCHEME_XOR = {"_p_TPMS_SCHEME_XOR", "TPMS_SCHEME_XOR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SCHEME_XOR_ARRAY = {"_p_TPMS_SCHEME_XOR_ARRAY", "TPMS_SCHEME_XOR_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SENSITIVE_CREATE = {"_p_TPMS_SENSITIVE_CREATE", "TPMS_SENSITIVE_CREATE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SENSITIVE_CREATE_ARRAY = {"_p_TPMS_SENSITIVE_CREATE_ARRAY", "TPMS_SENSITIVE_CREATE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SESSION_AUDIT_INFO = {"_p_TPMS_SESSION_AUDIT_INFO", "TPMS_SESSION_AUDIT_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SESSION_AUDIT_INFO_ARRAY = {"_p_TPMS_SESSION_AUDIT_INFO_ARRAY", "TPMS_SESSION_AUDIT_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SIGNATURE_ECC = {"_p_TPMS_SIGNATURE_ECC", "TPMS_SIGNATURE_ECDAA *|TPMS_SIGNATURE_ECC *|TPMS_SIGNATURE_ECSCHNORR *|TPMS_SIGNATURE_ECDSA *|TPMS_SIGNATURE_SM2 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SIGNATURE_ECC_ARRAY = {"_p_TPMS_SIGNATURE_ECC_ARRAY", "TPMS_SIGNATURE_ECC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SIGNATURE_RSA = {"_p_TPMS_SIGNATURE_RSA", "TPMS_SIGNATURE_RSAPSS *|TPMS_SIGNATURE_RSA *|TPMS_SIGNATURE_RSASSA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SIGNATURE_RSA_ARRAY = {"_p_TPMS_SIGNATURE_RSA_ARRAY", "TPMS_SIGNATURE_RSA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SYMCIPHER_PARMS = {"_p_TPMS_SYMCIPHER_PARMS", "TPMS_SYMCIPHER_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_SYMCIPHER_PARMS_ARRAY = {"_p_TPMS_SYMCIPHER_PARMS_ARRAY", "TPMS_SYMCIPHER_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TAGGED_PCR_SELECT = {"_p_TPMS_TAGGED_PCR_SELECT", "TPMS_TAGGED_PCR_SELECT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TAGGED_PCR_SELECT_ARRAY = {"_p_TPMS_TAGGED_PCR_SELECT_ARRAY", "TPMS_TAGGED_PCR_SELECT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TAGGED_POLICY = {"_p_TPMS_TAGGED_POLICY", "TPMS_TAGGED_POLICY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TAGGED_POLICY_ARRAY = {"_p_TPMS_TAGGED_POLICY_ARRAY", "TPMS_TAGGED_POLICY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TAGGED_PROPERTY = {"_p_TPMS_TAGGED_PROPERTY", "TPMS_TAGGED_PROPERTY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TAGGED_PROPERTY_ARRAY = {"_p_TPMS_TAGGED_PROPERTY_ARRAY", "TPMS_TAGGED_PROPERTY_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TIME_ATTEST_INFO = {"_p_TPMS_TIME_ATTEST_INFO", "TPMS_TIME_ATTEST_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TIME_ATTEST_INFO_ARRAY = {"_p_TPMS_TIME_ATTEST_INFO_ARRAY", "TPMS_TIME_ATTEST_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TIME_INFO = {"_p_TPMS_TIME_INFO", "TPMS_TIME_INFO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMS_TIME_INFO_ARRAY = {"_p_TPMS_TIME_INFO_ARRAY", "TPMS_TIME_INFO_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_ASYM_SCHEME = {"_p_TPMT_ASYM_SCHEME", "TPMT_ASYM_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_ASYM_SCHEME_ARRAY = {"_p_TPMT_ASYM_SCHEME_ARRAY", "TPMT_ASYM_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_ECC_SCHEME = {"_p_TPMT_ECC_SCHEME", "TPMT_ECC_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_ECC_SCHEME_ARRAY = {"_p_TPMT_ECC_SCHEME_ARRAY", "TPMT_ECC_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_HA = {"_p_TPMT_HA", "TPMT_HA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_HA_ARRAY = {"_p_TPMT_HA_ARRAY", "TPMT_HA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_KDF_SCHEME = {"_p_TPMT_KDF_SCHEME", "TPMT_KDF_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_KDF_SCHEME_ARRAY = {"_p_TPMT_KDF_SCHEME_ARRAY", "TPMT_KDF_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_KEYEDHASH_SCHEME = {"_p_TPMT_KEYEDHASH_SCHEME", "TPMT_KEYEDHASH_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_KEYEDHASH_SCHEME_ARRAY = {"_p_TPMT_KEYEDHASH_SCHEME_ARRAY", "TPMT_KEYEDHASH_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_PUBLIC = {"_p_TPMT_PUBLIC", "TPMT_PUBLIC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_PUBLIC_ARRAY = {"_p_TPMT_PUBLIC_ARRAY", "TPMT_PUBLIC_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_PUBLIC_PARMS = {"_p_TPMT_PUBLIC_PARMS", "TPMT_PUBLIC_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_PUBLIC_PARMS_ARRAY = {"_p_TPMT_PUBLIC_PARMS_ARRAY", "TPMT_PUBLIC_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_RSA_DECRYPT = {"_p_TPMT_RSA_DECRYPT", "TPMT_RSA_DECRYPT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_RSA_DECRYPT_ARRAY = {"_p_TPMT_RSA_DECRYPT_ARRAY", "TPMT_RSA_DECRYPT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_RSA_SCHEME = {"_p_TPMT_RSA_SCHEME", "TPMT_RSA_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_RSA_SCHEME_ARRAY = {"_p_TPMT_RSA_SCHEME_ARRAY", "TPMT_RSA_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SENSITIVE = {"_p_TPMT_SENSITIVE", "TPMT_SENSITIVE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SENSITIVE_ARRAY = {"_p_TPMT_SENSITIVE_ARRAY", "TPMT_SENSITIVE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SIGNATURE = {"_p_TPMT_SIGNATURE", "TPMT_SIGNATURE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SIGNATURE_ARRAY = {"_p_TPMT_SIGNATURE_ARRAY", "TPMT_SIGNATURE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SIG_SCHEME = {"_p_TPMT_SIG_SCHEME", "TPMT_SIG_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SIG_SCHEME_ARRAY = {"_p_TPMT_SIG_SCHEME_ARRAY", "TPMT_SIG_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SYM_DEF = {"_p_TPMT_SYM_DEF", "TPMT_SYM_DEF *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SYM_DEF_ARRAY = {"_p_TPMT_SYM_DEF_ARRAY", "TPMT_SYM_DEF_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SYM_DEF_OBJECT = {"_p_TPMT_SYM_DEF_OBJECT", "TPMT_SYM_DEF_OBJECT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_SYM_DEF_OBJECT_ARRAY = {"_p_TPMT_SYM_DEF_OBJECT_ARRAY", "TPMT_SYM_DEF_OBJECT_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_AUTH = {"_p_TPMT_TK_AUTH", "TPMT_TK_AUTH *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_AUTH_ARRAY = {"_p_TPMT_TK_AUTH_ARRAY", "TPMT_TK_AUTH_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_CREATION = {"_p_TPMT_TK_CREATION", "TPMT_TK_CREATION *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_CREATION_ARRAY = {"_p_TPMT_TK_CREATION_ARRAY", "TPMT_TK_CREATION_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_HASHCHECK = {"_p_TPMT_TK_HASHCHECK", "TPMT_TK_HASHCHECK *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_HASHCHECK_ARRAY = {"_p_TPMT_TK_HASHCHECK_ARRAY", "TPMT_TK_HASHCHECK_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_VERIFIED = {"_p_TPMT_TK_VERIFIED", "TPMT_TK_VERIFIED *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMT_TK_VERIFIED_ARRAY = {"_p_TPMT_TK_VERIFIED_ARRAY", "TPMT_TK_VERIFIED_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_ASYM_SCHEME = {"_p_TPMU_ASYM_SCHEME", "TPMU_ASYM_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_ASYM_SCHEME_ARRAY = {"_p_TPMU_ASYM_SCHEME_ARRAY", "TPMU_ASYM_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_ATTEST = {"_p_TPMU_ATTEST", "TPMU_ATTEST *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_ATTEST_ARRAY = {"_p_TPMU_ATTEST_ARRAY", "TPMU_ATTEST_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_CAPABILITIES = {"_p_TPMU_CAPABILITIES", "TPMU_CAPABILITIES *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_CAPABILITIES_ARRAY = {"_p_TPMU_CAPABILITIES_ARRAY", "TPMU_CAPABILITIES_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_ENCRYPTED_SECRET = {"_p_TPMU_ENCRYPTED_SECRET", "TPMU_ENCRYPTED_SECRET *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_ENCRYPTED_SECRET_ARRAY = {"_p_TPMU_ENCRYPTED_SECRET_ARRAY", "TPMU_ENCRYPTED_SECRET_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_HA = {"_p_TPMU_HA", "TPMU_HA *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_HA_ARRAY = {"_p_TPMU_HA_ARRAY", "TPMU_HA_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_KDF_SCHEME = {"_p_TPMU_KDF_SCHEME", "TPMU_KDF_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_KDF_SCHEME_ARRAY = {"_p_TPMU_KDF_SCHEME_ARRAY", "TPMU_KDF_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_NAME = {"_p_TPMU_NAME", "TPMU_NAME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_NAME_ARRAY = {"_p_TPMU_NAME_ARRAY", "TPMU_NAME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_PUBLIC_ID = {"_p_TPMU_PUBLIC_ID", "TPMU_PUBLIC_ID *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_PUBLIC_ID_ARRAY = {"_p_TPMU_PUBLIC_ID_ARRAY", "TPMU_PUBLIC_ID_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_PUBLIC_PARMS = {"_p_TPMU_PUBLIC_PARMS", "TPMU_PUBLIC_PARMS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_PUBLIC_PARMS_ARRAY = {"_p_TPMU_PUBLIC_PARMS_ARRAY", "TPMU_PUBLIC_PARMS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SCHEME_KEYEDHASH = {"_p_TPMU_SCHEME_KEYEDHASH", "TPMU_SCHEME_KEYEDHASH *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SCHEME_KEYEDHASH_ARRAY = {"_p_TPMU_SCHEME_KEYEDHASH_ARRAY", "TPMU_SCHEME_KEYEDHASH_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SENSITIVE_COMPOSITE = {"_p_TPMU_SENSITIVE_COMPOSITE", "TPMU_SENSITIVE_COMPOSITE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SENSITIVE_COMPOSITE_ARRAY = {"_p_TPMU_SENSITIVE_COMPOSITE_ARRAY", "TPMU_SENSITIVE_COMPOSITE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SIGNATURE = {"_p_TPMU_SIGNATURE", "TPMU_SIGNATURE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SIGNATURE_ARRAY = {"_p_TPMU_SIGNATURE_ARRAY", "TPMU_SIGNATURE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SIG_SCHEME = {"_p_TPMU_SIG_SCHEME", "TPMU_SIG_SCHEME *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SIG_SCHEME_ARRAY = {"_p_TPMU_SIG_SCHEME_ARRAY", "TPMU_SIG_SCHEME_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SYM_KEY_BITS = {"_p_TPMU_SYM_KEY_BITS", "TPMU_SYM_KEY_BITS *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SYM_KEY_BITS_ARRAY = {"_p_TPMU_SYM_KEY_BITS_ARRAY", "TPMU_SYM_KEY_BITS_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SYM_MODE = {"_p_TPMU_SYM_MODE", "TPMU_SYM_MODE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TPMU_SYM_MODE_ARRAY = {"_p_TPMU_SYM_MODE_ARRAY", "TPMU_SYM_MODE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TSS2_ABI_VERSION = {"_p_TSS2_ABI_VERSION", "TSS2_ABI_VERSION *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TSS2_TCTI_CONTEXT = {"_p_TSS2_TCTI_CONTEXT", "struct TSS2_TCTI_CONTEXT *|TSS2_TCTI_CONTEXT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_UINT16_PTR = {"_p_UINT16_PTR", "UINT16_PTR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_UINT32_PTR = {"_p_UINT32_PTR", "UINT32_PTR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__PRIVATE = {"_p__PRIVATE", "_PRIVATE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__PRIVATE_ARRAY = {"_p__PRIVATE_ARRAY", "_PRIVATE_ARRAY *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "BOOL *|INT32 *|intptr_t *|int *|int_least32_t *|int_fast32_t *|int32_t *|int_fast16_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long_long = {"_p_long_long", "INT64 *|int_least64_t *|int_fast64_t *|int64_t *|long long *|intmax_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_ESYS_CONTEXT = {"_p_p_ESYS_CONTEXT", "struct ESYS_CONTEXT **|ESYS_CONTEXT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_ATTEST = {"_p_p_TPM2B_ATTEST", "TPM2B_ATTEST **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_ATTEST_ARRAY = {"_p_p_TPM2B_ATTEST_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_CONTEXT_DATA = {"_p_p_TPM2B_CONTEXT_DATA", "TPM2B_CONTEXT_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_CONTEXT_DATA_ARRAY = {"_p_p_TPM2B_CONTEXT_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_CONTEXT_SENSITIVE = {"_p_p_TPM2B_CONTEXT_SENSITIVE", "TPM2B_CONTEXT_SENSITIVE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY = {"_p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_CREATION_DATA = {"_p_p_TPM2B_CREATION_DATA", "TPM2B_CREATION_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_CREATION_DATA_ARRAY = {"_p_p_TPM2B_CREATION_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_DATA = {"_p_p_TPM2B_DATA", "TPM2B_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_DATA_ARRAY = {"_p_p_TPM2B_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_DIGEST = {"_p_p_TPM2B_DIGEST", "TPM2B_DIGEST **|TPM2B_TIMEOUT **|TPM2B_NONCE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_DIGEST_ARRAY = {"_p_p_TPM2B_DIGEST_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_ECC_PARAMETER = {"_p_p_TPM2B_ECC_PARAMETER", "TPM2B_ECC_PARAMETER **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_ECC_PARAMETER_ARRAY = {"_p_p_TPM2B_ECC_PARAMETER_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_ECC_POINT = {"_p_p_TPM2B_ECC_POINT", "TPM2B_ECC_POINT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_ECC_POINT_ARRAY = {"_p_p_TPM2B_ECC_POINT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_ENCRYPTED_SECRET = {"_p_p_TPM2B_ENCRYPTED_SECRET", "TPM2B_ENCRYPTED_SECRET **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_ENCRYPTED_SECRET_ARRAY = {"_p_p_TPM2B_ENCRYPTED_SECRET_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_EVENT = {"_p_p_TPM2B_EVENT", "TPM2B_EVENT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_EVENT_ARRAY = {"_p_p_TPM2B_EVENT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_ID_OBJECT = {"_p_p_TPM2B_ID_OBJECT", "TPM2B_ID_OBJECT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_ID_OBJECT_ARRAY = {"_p_p_TPM2B_ID_OBJECT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_IV = {"_p_p_TPM2B_IV", "TPM2B_IV **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_IV_ARRAY = {"_p_p_TPM2B_IV_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_MAX_BUFFER = {"_p_p_TPM2B_MAX_BUFFER", "TPM2B_MAX_BUFFER **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_MAX_BUFFER_ARRAY = {"_p_p_TPM2B_MAX_BUFFER_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_MAX_NV_BUFFER = {"_p_p_TPM2B_MAX_NV_BUFFER", "TPM2B_MAX_NV_BUFFER **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_MAX_NV_BUFFER_ARRAY = {"_p_p_TPM2B_MAX_NV_BUFFER_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_NAME = {"_p_p_TPM2B_NAME", "TPM2B_NAME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_NAME_ARRAY = {"_p_p_TPM2B_NAME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_NV_PUBLIC = {"_p_p_TPM2B_NV_PUBLIC", "TPM2B_NV_PUBLIC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_NV_PUBLIC_ARRAY = {"_p_p_TPM2B_NV_PUBLIC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_PRIVATE = {"_p_p_TPM2B_PRIVATE", "TPM2B_PRIVATE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_PRIVATE_ARRAY = {"_p_p_TPM2B_PRIVATE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_PRIVATE_KEY_RSA = {"_p_p_TPM2B_PRIVATE_KEY_RSA", "TPM2B_PRIVATE_KEY_RSA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY = {"_p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC = {"_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC", "TPM2B_PRIVATE_VENDOR_SPECIFIC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY = {"_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_PUBLIC = {"_p_p_TPM2B_PUBLIC", "TPM2B_PUBLIC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_PUBLIC_ARRAY = {"_p_p_TPM2B_PUBLIC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_PUBLIC_KEY_RSA = {"_p_p_TPM2B_PUBLIC_KEY_RSA", "TPM2B_PUBLIC_KEY_RSA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY = {"_p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_SENSITIVE = {"_p_p_TPM2B_SENSITIVE", "TPM2B_SENSITIVE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_SENSITIVE_ARRAY = {"_p_p_TPM2B_SENSITIVE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_SENSITIVE_CREATE = {"_p_p_TPM2B_SENSITIVE_CREATE", "TPM2B_SENSITIVE_CREATE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_SENSITIVE_CREATE_ARRAY = {"_p_p_TPM2B_SENSITIVE_CREATE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_SENSITIVE_DATA = {"_p_p_TPM2B_SENSITIVE_DATA", "TPM2B_SENSITIVE_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_SENSITIVE_DATA_ARRAY = {"_p_p_TPM2B_SENSITIVE_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_SYM_KEY = {"_p_p_TPM2B_SYM_KEY", "TPM2B_SYM_KEY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_SYM_KEY_ARRAY = {"_p_p_TPM2B_SYM_KEY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPM2B_TEMPLATE = {"_p_p_TPM2B_TEMPLATE", "TPM2B_TEMPLATE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPM2B_TEMPLATE_ARRAY = {"_p_p_TPM2B_TEMPLATE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_AC_CAPABILITIES = {"_p_p_TPML_AC_CAPABILITIES", "TPML_AC_CAPABILITIES **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_AC_CAPABILITIES_ARRAY = {"_p_p_TPML_AC_CAPABILITIES_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_ALG = {"_p_p_TPML_ALG", "TPML_ALG **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_ALG_ARRAY = {"_p_p_TPML_ALG_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_ALG_PROPERTY = {"_p_p_TPML_ALG_PROPERTY", "TPML_ALG_PROPERTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_ALG_PROPERTY_ARRAY = {"_p_p_TPML_ALG_PROPERTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_CC = {"_p_p_TPML_CC", "TPML_CC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_CC_ARRAY = {"_p_p_TPML_CC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_CCA = {"_p_p_TPML_CCA", "TPML_CCA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_CCA_ARRAY = {"_p_p_TPML_CCA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_DIGEST = {"_p_p_TPML_DIGEST", "TPML_DIGEST **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_DIGEST_ARRAY = {"_p_p_TPML_DIGEST_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_DIGEST_VALUES = {"_p_p_TPML_DIGEST_VALUES", "TPML_DIGEST_VALUES **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_DIGEST_VALUES_ARRAY = {"_p_p_TPML_DIGEST_VALUES_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_ECC_CURVE = {"_p_p_TPML_ECC_CURVE", "TPML_ECC_CURVE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_ECC_CURVE_ARRAY = {"_p_p_TPML_ECC_CURVE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_HANDLE = {"_p_p_TPML_HANDLE", "TPML_HANDLE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_HANDLE_ARRAY = {"_p_p_TPML_HANDLE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_INTEL_PTT_PROPERTY = {"_p_p_TPML_INTEL_PTT_PROPERTY", "TPML_INTEL_PTT_PROPERTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_INTEL_PTT_PROPERTY_ARRAY = {"_p_p_TPML_INTEL_PTT_PROPERTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_PCR_SELECTION = {"_p_p_TPML_PCR_SELECTION", "TPML_PCR_SELECTION **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_PCR_SELECTION_ARRAY = {"_p_p_TPML_PCR_SELECTION_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_TAGGED_PCR_PROPERTY = {"_p_p_TPML_TAGGED_PCR_PROPERTY", "TPML_TAGGED_PCR_PROPERTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY = {"_p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPML_TAGGED_TPM_PROPERTY = {"_p_p_TPML_TAGGED_TPM_PROPERTY", "TPML_TAGGED_TPM_PROPERTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY = {"_p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_AC_OUTPUT = {"_p_p_TPMS_AC_OUTPUT", "TPMS_AC_OUTPUT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_AC_OUTPUT_ARRAY = {"_p_p_TPMS_AC_OUTPUT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ALGORITHM_DESCRIPTION = {"_p_p_TPMS_ALGORITHM_DESCRIPTION", "TPMS_ALGORITHM_DESCRIPTION **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY = {"_p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC = {"_p_p_TPMS_ALGORITHM_DETAIL_ECC", "TPMS_ALGORITHM_DETAIL_ECC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY = {"_p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ALG_PROPERTY = {"_p_p_TPMS_ALG_PROPERTY", "TPMS_ALG_PROPERTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ALG_PROPERTY_ARRAY = {"_p_p_TPMS_ALG_PROPERTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ASYM_PARMS = {"_p_p_TPMS_ASYM_PARMS", "TPMS_ASYM_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ASYM_PARMS_ARRAY = {"_p_p_TPMS_ASYM_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ATTEST = {"_p_p_TPMS_ATTEST", "TPMS_ATTEST **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ATTEST_ARRAY = {"_p_p_TPMS_ATTEST_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_AUTH_COMMAND = {"_p_p_TPMS_AUTH_COMMAND", "TPMS_AUTH_COMMAND **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_AUTH_COMMAND_ARRAY = {"_p_p_TPMS_AUTH_COMMAND_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_AUTH_RESPONSE = {"_p_p_TPMS_AUTH_RESPONSE", "TPMS_AUTH_RESPONSE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_AUTH_RESPONSE_ARRAY = {"_p_p_TPMS_AUTH_RESPONSE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CAPABILITY_DATA = {"_p_p_TPMS_CAPABILITY_DATA", "TPMS_CAPABILITY_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CAPABILITY_DATA_ARRAY = {"_p_p_TPMS_CAPABILITY_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CERTIFY_INFO = {"_p_p_TPMS_CERTIFY_INFO", "TPMS_CERTIFY_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CERTIFY_INFO_ARRAY = {"_p_p_TPMS_CERTIFY_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CLOCK_INFO = {"_p_p_TPMS_CLOCK_INFO", "TPMS_CLOCK_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CLOCK_INFO_ARRAY = {"_p_p_TPMS_CLOCK_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_COMMAND_AUDIT_INFO = {"_p_p_TPMS_COMMAND_AUDIT_INFO", "TPMS_COMMAND_AUDIT_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY = {"_p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CONTEXT = {"_p_p_TPMS_CONTEXT", "TPMS_CONTEXT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CONTEXT_ARRAY = {"_p_p_TPMS_CONTEXT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CONTEXT_DATA = {"_p_p_TPMS_CONTEXT_DATA", "TPMS_CONTEXT_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CONTEXT_DATA_ARRAY = {"_p_p_TPMS_CONTEXT_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CREATION_DATA = {"_p_p_TPMS_CREATION_DATA", "TPMS_CREATION_DATA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CREATION_DATA_ARRAY = {"_p_p_TPMS_CREATION_DATA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_CREATION_INFO = {"_p_p_TPMS_CREATION_INFO", "TPMS_CREATION_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_CREATION_INFO_ARRAY = {"_p_p_TPMS_CREATION_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ECC_PARMS = {"_p_p_TPMS_ECC_PARMS", "TPMS_ECC_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ECC_PARMS_ARRAY = {"_p_p_TPMS_ECC_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ECC_POINT = {"_p_p_TPMS_ECC_POINT", "TPMS_ECC_POINT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ECC_POINT_ARRAY = {"_p_p_TPMS_ECC_POINT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_EMPTY = {"_p_p_TPMS_EMPTY", "TPMS_EMPTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_EMPTY_ARRAY = {"_p_p_TPMS_EMPTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_ID_OBJECT = {"_p_p_TPMS_ID_OBJECT", "TPMS_ID_OBJECT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_ID_OBJECT_ARRAY = {"_p_p_TPMS_ID_OBJECT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_KEYEDHASH_PARMS = {"_p_p_TPMS_KEYEDHASH_PARMS", "TPMS_KEYEDHASH_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_KEYEDHASH_PARMS_ARRAY = {"_p_p_TPMS_KEYEDHASH_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_NV_CERTIFY_INFO = {"_p_p_TPMS_NV_CERTIFY_INFO", "TPMS_NV_CERTIFY_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_NV_CERTIFY_INFO_ARRAY = {"_p_p_TPMS_NV_CERTIFY_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS = {"_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS", "TPMS_NV_PIN_COUNTER_PARAMETERS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY = {"_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_NV_PUBLIC = {"_p_p_TPMS_NV_PUBLIC", "TPMS_NV_PUBLIC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_NV_PUBLIC_ARRAY = {"_p_p_TPMS_NV_PUBLIC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_PCR_SELECT = {"_p_p_TPMS_PCR_SELECT", "TPMS_PCR_SELECT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_PCR_SELECT_ARRAY = {"_p_p_TPMS_PCR_SELECT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_PCR_SELECTION = {"_p_p_TPMS_PCR_SELECTION", "TPMS_PCR_SELECTION **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_PCR_SELECTION_ARRAY = {"_p_p_TPMS_PCR_SELECTION_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_QUOTE_INFO = {"_p_p_TPMS_QUOTE_INFO", "TPMS_QUOTE_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_QUOTE_INFO_ARRAY = {"_p_p_TPMS_QUOTE_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_RSA_PARMS = {"_p_p_TPMS_RSA_PARMS", "TPMS_RSA_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_RSA_PARMS_ARRAY = {"_p_p_TPMS_RSA_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SCHEME_ECDAA = {"_p_p_TPMS_SCHEME_ECDAA", "TPMS_SCHEME_ECDAA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SCHEME_ECDAA_ARRAY = {"_p_p_TPMS_SCHEME_ECDAA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SCHEME_HASH = {"_p_p_TPMS_SCHEME_HASH", "TPMS_SCHEME_HASH **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SCHEME_HASH_ARRAY = {"_p_p_TPMS_SCHEME_HASH_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SCHEME_XOR = {"_p_p_TPMS_SCHEME_XOR", "TPMS_SCHEME_XOR **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SCHEME_XOR_ARRAY = {"_p_p_TPMS_SCHEME_XOR_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SENSITIVE_CREATE = {"_p_p_TPMS_SENSITIVE_CREATE", "TPMS_SENSITIVE_CREATE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SENSITIVE_CREATE_ARRAY = {"_p_p_TPMS_SENSITIVE_CREATE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SESSION_AUDIT_INFO = {"_p_p_TPMS_SESSION_AUDIT_INFO", "TPMS_SESSION_AUDIT_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SESSION_AUDIT_INFO_ARRAY = {"_p_p_TPMS_SESSION_AUDIT_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SIGNATURE_ECC = {"_p_p_TPMS_SIGNATURE_ECC", "TPMS_SIGNATURE_ECC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SIGNATURE_ECC_ARRAY = {"_p_p_TPMS_SIGNATURE_ECC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SIGNATURE_RSA = {"_p_p_TPMS_SIGNATURE_RSA", "TPMS_SIGNATURE_RSA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SIGNATURE_RSA_ARRAY = {"_p_p_TPMS_SIGNATURE_RSA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_SYMCIPHER_PARMS = {"_p_p_TPMS_SYMCIPHER_PARMS", "TPMS_SYMCIPHER_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_SYMCIPHER_PARMS_ARRAY = {"_p_p_TPMS_SYMCIPHER_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_TAGGED_PCR_SELECT = {"_p_p_TPMS_TAGGED_PCR_SELECT", "TPMS_TAGGED_PCR_SELECT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_TAGGED_PCR_SELECT_ARRAY = {"_p_p_TPMS_TAGGED_PCR_SELECT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_TAGGED_POLICY = {"_p_p_TPMS_TAGGED_POLICY", "TPMS_TAGGED_POLICY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_TAGGED_POLICY_ARRAY = {"_p_p_TPMS_TAGGED_POLICY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_TAGGED_PROPERTY = {"_p_p_TPMS_TAGGED_PROPERTY", "TPMS_TAGGED_PROPERTY **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_TAGGED_PROPERTY_ARRAY = {"_p_p_TPMS_TAGGED_PROPERTY_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_TIME_ATTEST_INFO = {"_p_p_TPMS_TIME_ATTEST_INFO", "TPMS_TIME_ATTEST_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_TIME_ATTEST_INFO_ARRAY = {"_p_p_TPMS_TIME_ATTEST_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMS_TIME_INFO = {"_p_p_TPMS_TIME_INFO", "TPMS_TIME_INFO **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMS_TIME_INFO_ARRAY = {"_p_p_TPMS_TIME_INFO_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_ASYM_SCHEME = {"_p_p_TPMT_ASYM_SCHEME", "TPMT_ASYM_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_ASYM_SCHEME_ARRAY = {"_p_p_TPMT_ASYM_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_ECC_SCHEME = {"_p_p_TPMT_ECC_SCHEME", "TPMT_ECC_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_ECC_SCHEME_ARRAY = {"_p_p_TPMT_ECC_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_HA = {"_p_p_TPMT_HA", "TPMT_HA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_HA_ARRAY = {"_p_p_TPMT_HA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_KDF_SCHEME = {"_p_p_TPMT_KDF_SCHEME", "TPMT_KDF_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_KDF_SCHEME_ARRAY = {"_p_p_TPMT_KDF_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_KEYEDHASH_SCHEME = {"_p_p_TPMT_KEYEDHASH_SCHEME", "TPMT_KEYEDHASH_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_KEYEDHASH_SCHEME_ARRAY = {"_p_p_TPMT_KEYEDHASH_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_PUBLIC = {"_p_p_TPMT_PUBLIC", "TPMT_PUBLIC **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_PUBLIC_ARRAY = {"_p_p_TPMT_PUBLIC_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_PUBLIC_PARMS = {"_p_p_TPMT_PUBLIC_PARMS", "TPMT_PUBLIC_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_PUBLIC_PARMS_ARRAY = {"_p_p_TPMT_PUBLIC_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_RSA_DECRYPT = {"_p_p_TPMT_RSA_DECRYPT", "TPMT_RSA_DECRYPT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_RSA_DECRYPT_ARRAY = {"_p_p_TPMT_RSA_DECRYPT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_RSA_SCHEME = {"_p_p_TPMT_RSA_SCHEME", "TPMT_RSA_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_RSA_SCHEME_ARRAY = {"_p_p_TPMT_RSA_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_SENSITIVE = {"_p_p_TPMT_SENSITIVE", "TPMT_SENSITIVE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_SENSITIVE_ARRAY = {"_p_p_TPMT_SENSITIVE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_SIGNATURE = {"_p_p_TPMT_SIGNATURE", "TPMT_SIGNATURE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_SIGNATURE_ARRAY = {"_p_p_TPMT_SIGNATURE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_SIG_SCHEME = {"_p_p_TPMT_SIG_SCHEME", "TPMT_SIG_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_SIG_SCHEME_ARRAY = {"_p_p_TPMT_SIG_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_SYM_DEF = {"_p_p_TPMT_SYM_DEF", "TPMT_SYM_DEF **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_SYM_DEF_ARRAY = {"_p_p_TPMT_SYM_DEF_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_SYM_DEF_OBJECT = {"_p_p_TPMT_SYM_DEF_OBJECT", "TPMT_SYM_DEF_OBJECT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_SYM_DEF_OBJECT_ARRAY = {"_p_p_TPMT_SYM_DEF_OBJECT_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_AUTH = {"_p_p_TPMT_TK_AUTH", "TPMT_TK_AUTH **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_AUTH_ARRAY = {"_p_p_TPMT_TK_AUTH_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_CREATION = {"_p_p_TPMT_TK_CREATION", "TPMT_TK_CREATION **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_CREATION_ARRAY = {"_p_p_TPMT_TK_CREATION_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_HASHCHECK = {"_p_p_TPMT_TK_HASHCHECK", "TPMT_TK_HASHCHECK **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_HASHCHECK_ARRAY = {"_p_p_TPMT_TK_HASHCHECK_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_VERIFIED = {"_p_p_TPMT_TK_VERIFIED", "TPMT_TK_VERIFIED **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMT_TK_VERIFIED_ARRAY = {"_p_p_TPMT_TK_VERIFIED_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_ASYM_SCHEME = {"_p_p_TPMU_ASYM_SCHEME", "TPMU_ASYM_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_ASYM_SCHEME_ARRAY = {"_p_p_TPMU_ASYM_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_ATTEST = {"_p_p_TPMU_ATTEST", "TPMU_ATTEST **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_ATTEST_ARRAY = {"_p_p_TPMU_ATTEST_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_CAPABILITIES = {"_p_p_TPMU_CAPABILITIES", "TPMU_CAPABILITIES **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_CAPABILITIES_ARRAY = {"_p_p_TPMU_CAPABILITIES_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_ENCRYPTED_SECRET = {"_p_p_TPMU_ENCRYPTED_SECRET", "TPMU_ENCRYPTED_SECRET **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_ENCRYPTED_SECRET_ARRAY = {"_p_p_TPMU_ENCRYPTED_SECRET_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_HA = {"_p_p_TPMU_HA", "TPMU_HA **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_HA_ARRAY = {"_p_p_TPMU_HA_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_KDF_SCHEME = {"_p_p_TPMU_KDF_SCHEME", "TPMU_KDF_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_KDF_SCHEME_ARRAY = {"_p_p_TPMU_KDF_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_NAME = {"_p_p_TPMU_NAME", "TPMU_NAME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_NAME_ARRAY = {"_p_p_TPMU_NAME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_PUBLIC_ID = {"_p_p_TPMU_PUBLIC_ID", "TPMU_PUBLIC_ID **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_PUBLIC_ID_ARRAY = {"_p_p_TPMU_PUBLIC_ID_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_PUBLIC_PARMS = {"_p_p_TPMU_PUBLIC_PARMS", "TPMU_PUBLIC_PARMS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_PUBLIC_PARMS_ARRAY = {"_p_p_TPMU_PUBLIC_PARMS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_SCHEME_KEYEDHASH = {"_p_p_TPMU_SCHEME_KEYEDHASH", "TPMU_SCHEME_KEYEDHASH **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_SCHEME_KEYEDHASH_ARRAY = {"_p_p_TPMU_SCHEME_KEYEDHASH_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_SENSITIVE_COMPOSITE = {"_p_p_TPMU_SENSITIVE_COMPOSITE", "TPMU_SENSITIVE_COMPOSITE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY = {"_p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_SIGNATURE = {"_p_p_TPMU_SIGNATURE", "TPMU_SIGNATURE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_SIGNATURE_ARRAY = {"_p_p_TPMU_SIGNATURE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_SIG_SCHEME = {"_p_p_TPMU_SIG_SCHEME", "TPMU_SIG_SCHEME **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_SIG_SCHEME_ARRAY = {"_p_p_TPMU_SIG_SCHEME_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_SYM_KEY_BITS = {"_p_p_TPMU_SYM_KEY_BITS", "TPMU_SYM_KEY_BITS **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_SYM_KEY_BITS_ARRAY = {"_p_p_TPMU_SYM_KEY_BITS_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TPMU_SYM_MODE = {"_p_p_TPMU_SYM_MODE", "TPMU_SYM_MODE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TPMU_SYM_MODE_ARRAY = {"_p_p_TPMU_SYM_MODE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_TSS2_TCTI_CONTEXT = {"_p_p_TSS2_TCTI_CONTEXT", "struct TSS2_TCTI_CONTEXT **|TSS2_TCTI_CONTEXT **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_TSS2_TCTI_POLL_HANDLE = {"_p_p_TSS2_TCTI_POLL_HANDLE", "TSS2_TCTI_POLL_HANDLE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p__PRIVATE = {"_p_p__PRIVATE", "_PRIVATE **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p__PRIVATE_ARRAY = {"_p_p__PRIVATE_ARRAY", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_size_t = {"_p_p_size_t", "size_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_unsigned_char = {"_p_p_unsigned_char", "uint8_t **|unsigned char **|UINT8 **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_unsigned_int = {"_p_p_unsigned_int", "unsigned int **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_ESYS_TR_PTR = {"_p_p_ESYS_TR_PTR", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_short = {"_p_short", "short *|int_least16_t *|int16_t *|INT16 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_signed_char = {"_p_signed_char", "signed char *|INT8 *|TPM2_CLOCK_ADJUST *|int_least8_t *|int_fast8_t *|int8_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_size_t = {"_p_size_t", "size_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_uint16_t_PTR = {"_p_uint16_t_PTR", "uint16_t_PTR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_uint32_t_PTR = {"_p_uint32_t_PTR", "uint32_t_PTR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "TPM2_HT *|TPMA_LOCALITY *|unsigned char *|TPMA_SESSION *|TPM2_SE *|BYTE *|UINT8 *|uint_least8_t *|uint_fast8_t *|uint8_t *|TPM2_NT *|TPMI_YES_NO *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "uintptr_t *|TPM2_PARAMETER_SIZE *|TPM2_AUTHORIZATION_SIZE *|UINT32 *|TPM2_MODIFIER_INDICATOR *|TPMI_DH_ENTITY *|TPM2_PS *|TPM2_GENERATED *|TPMI_RH_LOCKOUT *|TPMI_DH_OBJECT *|TPMA_OBJECT *|TPM2_CC *|TPM2_HANDLE *|TPMI_SH_AUTH_SESSION *|TPMI_RH_PROVISION *|TPM2_HC *|TPMA_CC *|TPMI_RH_NV_AUTH *|TPMI_RH_HIERARCHY_AUTH *|TPMA_ALGORITHM *|TPM_EA *|uint_fast32_t *|uint_least32_t *|uint32_t *|TPMI_RH_OWNER *|TPMA_PERMANENT *|TPMI_SH_POLICY *|TPMI_RH_HIERARCHY *|unsigned int *|TPM2_RC *|TSS2_RC *|TPMI_RH_PLATFORM *|TPMI_RH_AC *|TPMI_DH_CONTEXT *|TPM2_NV_INDEX *|TPMI_RH_NV_INDEX *|TPMA_MEMORY *|TPMI_DH_PERSISTENT *|TPM_AT *|TPM2_PT *|TPM2_CAP *|TPMI_SH_HMAC *|TPMA_MODES *|TPM2_RH *|TPM2_ALGORITHM_ID *|TPM2_SPEC *|uint_fast16_t *|TPMI_RH_ENDORSEMENT *|TPMI_RH_ENABLES *|TPMA_STARTUP_CLEAR *|TPMI_RH_CLEAR *|TPM2_PT_PCR *|TPMA_NV *|TPMI_DH_PCR *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long_long = {"_p_unsigned_long_long", "UINT64 *|uint_least64_t *|uint_fast64_t *|uint64_t *|unsigned long long *|uintmax_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_short = {"_p_unsigned_short", "TPMI_ALG_SYM_OBJECT *|TPMI_ECC_KEY_EXCHANGE *|TPM2_SU *|TPM2_KEY_SIZE *|TPMI_ALG_SIG_SCHEME *|TPMI_ALG_KEYEDHASH_SCHEME *|TPMI_ALG_ASYM_SCHEME *|TPMI_ALG_RSA_SCHEME *|TPMI_ALG_ECC_SCHEME *|TPM2_EO *|TPMI_TPM2_CAMELLIA_KEY_BITS *|TPMI_RSA_KEY_BITS *|TPMI_CAMELLIA_KEY_BITS *|TPMI_SM4_KEY_BITS *|TPMI_AES_KEY_BITS *|TPM2_KEY_BITS *|TPMI_ALG_RSA_DECRYPT *|TPMI_ALG_HASH *|TPMI_ST_COMMAND_TAG *|TPM2_ALG_ID *|TPMI_ALG_SYM_MODE *|TPMI_ALG_PUBLIC *|TPMI_ALG_KDF *|TPMI_ST_ATTEST *|unsigned short *|UINT16 *|uint16_t *|TPMI_ECC_CURVE *|TPM2_ECC_CURVE *|uint_least16_t *|TPMI_ALG_ASYM *|TPMI_ALG_SYM *|TPM2_ST *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_ByteArray,
  &_swigt__p_ESYS_CONTEXT,
  &_swigt__p_ESYS_TR_PTR,
  &_swigt__p_TPM2B_ATTEST,
  &_swigt__p_TPM2B_ATTEST_ARRAY,
  &_swigt__p_TPM2B_CONTEXT_DATA,
  &_swigt__p_TPM2B_CONTEXT_DATA_ARRAY,
  &_swigt__p_TPM2B_CONTEXT_SENSITIVE,
  &_swigt__p_TPM2B_CONTEXT_SENSITIVE_ARRAY,
  &_swigt__p_TPM2B_CREATION_DATA,
  &_swigt__p_TPM2B_CREATION_DATA_ARRAY,
  &_swigt__p_TPM2B_DATA,
  &_swigt__p_TPM2B_DATA_ARRAY,
  &_swigt__p_TPM2B_DIGEST,
  &_swigt__p_TPM2B_DIGEST_ARRAY,
  &_swigt__p_TPM2B_ECC_PARAMETER,
  &_swigt__p_TPM2B_ECC_PARAMETER_ARRAY,
  &_swigt__p_TPM2B_ECC_POINT,
  &_swigt__p_TPM2B_ECC_POINT_ARRAY,
  &_swigt__p_TPM2B_ENCRYPTED_SECRET,
  &_swigt__p_TPM2B_ENCRYPTED_SECRET_ARRAY,
  &_swigt__p_TPM2B_EVENT,
  &_swigt__p_TPM2B_EVENT_ARRAY,
  &_swigt__p_TPM2B_ID_OBJECT,
  &_swigt__p_TPM2B_ID_OBJECT_ARRAY,
  &_swigt__p_TPM2B_IV,
  &_swigt__p_TPM2B_IV_ARRAY,
  &_swigt__p_TPM2B_MAX_BUFFER,
  &_swigt__p_TPM2B_MAX_BUFFER_ARRAY,
  &_swigt__p_TPM2B_MAX_NV_BUFFER,
  &_swigt__p_TPM2B_MAX_NV_BUFFER_ARRAY,
  &_swigt__p_TPM2B_NAME,
  &_swigt__p_TPM2B_NAME_ARRAY,
  &_swigt__p_TPM2B_NV_PUBLIC,
  &_swigt__p_TPM2B_NV_PUBLIC_ARRAY,
  &_swigt__p_TPM2B_PRIVATE,
  &_swigt__p_TPM2B_PRIVATE_ARRAY,
  &_swigt__p_TPM2B_PRIVATE_KEY_RSA,
  &_swigt__p_TPM2B_PRIVATE_KEY_RSA_ARRAY,
  &_swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC,
  &_swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY,
  &_swigt__p_TPM2B_PUBLIC,
  &_swigt__p_TPM2B_PUBLIC_ARRAY,
  &_swigt__p_TPM2B_PUBLIC_KEY_RSA,
  &_swigt__p_TPM2B_PUBLIC_KEY_RSA_ARRAY,
  &_swigt__p_TPM2B_SENSITIVE,
  &_swigt__p_TPM2B_SENSITIVE_ARRAY,
  &_swigt__p_TPM2B_SENSITIVE_CREATE,
  &_swigt__p_TPM2B_SENSITIVE_CREATE_ARRAY,
  &_swigt__p_TPM2B_SENSITIVE_DATA,
  &_swigt__p_TPM2B_SENSITIVE_DATA_ARRAY,
  &_swigt__p_TPM2B_SYM_KEY,
  &_swigt__p_TPM2B_SYM_KEY_ARRAY,
  &_swigt__p_TPM2B_TEMPLATE,
  &_swigt__p_TPM2B_TEMPLATE_ARRAY,
  &_swigt__p_TPML_AC_CAPABILITIES,
  &_swigt__p_TPML_AC_CAPABILITIES_ARRAY,
  &_swigt__p_TPML_ALG,
  &_swigt__p_TPML_ALG_ARRAY,
  &_swigt__p_TPML_ALG_PROPERTY,
  &_swigt__p_TPML_ALG_PROPERTY_ARRAY,
  &_swigt__p_TPML_CC,
  &_swigt__p_TPML_CCA,
  &_swigt__p_TPML_CCA_ARRAY,
  &_swigt__p_TPML_CC_ARRAY,
  &_swigt__p_TPML_DIGEST,
  &_swigt__p_TPML_DIGEST_ARRAY,
  &_swigt__p_TPML_DIGEST_VALUES,
  &_swigt__p_TPML_DIGEST_VALUES_ARRAY,
  &_swigt__p_TPML_ECC_CURVE,
  &_swigt__p_TPML_ECC_CURVE_ARRAY,
  &_swigt__p_TPML_HANDLE,
  &_swigt__p_TPML_HANDLE_ARRAY,
  &_swigt__p_TPML_INTEL_PTT_PROPERTY,
  &_swigt__p_TPML_INTEL_PTT_PROPERTY_ARRAY,
  &_swigt__p_TPML_PCR_SELECTION,
  &_swigt__p_TPML_PCR_SELECTION_ARRAY,
  &_swigt__p_TPML_TAGGED_PCR_PROPERTY,
  &_swigt__p_TPML_TAGGED_PCR_PROPERTY_ARRAY,
  &_swigt__p_TPML_TAGGED_TPM_PROPERTY,
  &_swigt__p_TPML_TAGGED_TPM_PROPERTY_ARRAY,
  &_swigt__p_TPMS_AC_OUTPUT,
  &_swigt__p_TPMS_AC_OUTPUT_ARRAY,
  &_swigt__p_TPMS_ALGORITHM_DESCRIPTION,
  &_swigt__p_TPMS_ALGORITHM_DESCRIPTION_ARRAY,
  &_swigt__p_TPMS_ALGORITHM_DETAIL_ECC,
  &_swigt__p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY,
  &_swigt__p_TPMS_ALG_PROPERTY,
  &_swigt__p_TPMS_ALG_PROPERTY_ARRAY,
  &_swigt__p_TPMS_ASYM_PARMS,
  &_swigt__p_TPMS_ASYM_PARMS_ARRAY,
  &_swigt__p_TPMS_ATTEST,
  &_swigt__p_TPMS_ATTEST_ARRAY,
  &_swigt__p_TPMS_AUTH_COMMAND,
  &_swigt__p_TPMS_AUTH_COMMAND_ARRAY,
  &_swigt__p_TPMS_AUTH_RESPONSE,
  &_swigt__p_TPMS_AUTH_RESPONSE_ARRAY,
  &_swigt__p_TPMS_CAPABILITY_DATA,
  &_swigt__p_TPMS_CAPABILITY_DATA_ARRAY,
  &_swigt__p_TPMS_CERTIFY_INFO,
  &_swigt__p_TPMS_CERTIFY_INFO_ARRAY,
  &_swigt__p_TPMS_CLOCK_INFO,
  &_swigt__p_TPMS_CLOCK_INFO_ARRAY,
  &_swigt__p_TPMS_COMMAND_AUDIT_INFO,
  &_swigt__p_TPMS_COMMAND_AUDIT_INFO_ARRAY,
  &_swigt__p_TPMS_CONTEXT,
  &_swigt__p_TPMS_CONTEXT_ARRAY,
  &_swigt__p_TPMS_CONTEXT_DATA,
  &_swigt__p_TPMS_CONTEXT_DATA_ARRAY,
  &_swigt__p_TPMS_CREATION_DATA,
  &_swigt__p_TPMS_CREATION_DATA_ARRAY,
  &_swigt__p_TPMS_CREATION_INFO,
  &_swigt__p_TPMS_CREATION_INFO_ARRAY,
  &_swigt__p_TPMS_ECC_PARMS,
  &_swigt__p_TPMS_ECC_PARMS_ARRAY,
  &_swigt__p_TPMS_ECC_POINT,
  &_swigt__p_TPMS_ECC_POINT_ARRAY,
  &_swigt__p_TPMS_EMPTY,
  &_swigt__p_TPMS_EMPTY_ARRAY,
  &_swigt__p_TPMS_ID_OBJECT,
  &_swigt__p_TPMS_ID_OBJECT_ARRAY,
  &_swigt__p_TPMS_KEYEDHASH_PARMS,
  &_swigt__p_TPMS_KEYEDHASH_PARMS_ARRAY,
  &_swigt__p_TPMS_NV_CERTIFY_INFO,
  &_swigt__p_TPMS_NV_CERTIFY_INFO_ARRAY,
  &_swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS,
  &_swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY,
  &_swigt__p_TPMS_NV_PUBLIC,
  &_swigt__p_TPMS_NV_PUBLIC_ARRAY,
  &_swigt__p_TPMS_PCR_SELECT,
  &_swigt__p_TPMS_PCR_SELECTION,
  &_swigt__p_TPMS_PCR_SELECTION_ARRAY,
  &_swigt__p_TPMS_PCR_SELECT_ARRAY,
  &_swigt__p_TPMS_QUOTE_INFO,
  &_swigt__p_TPMS_QUOTE_INFO_ARRAY,
  &_swigt__p_TPMS_RSA_PARMS,
  &_swigt__p_TPMS_RSA_PARMS_ARRAY,
  &_swigt__p_TPMS_SCHEME_ECDAA,
  &_swigt__p_TPMS_SCHEME_ECDAA_ARRAY,
  &_swigt__p_TPMS_SCHEME_HASH,
  &_swigt__p_TPMS_SCHEME_HASH_ARRAY,
  &_swigt__p_TPMS_SCHEME_XOR,
  &_swigt__p_TPMS_SCHEME_XOR_ARRAY,
  &_swigt__p_TPMS_SENSITIVE_CREATE,
  &_swigt__p_TPMS_SENSITIVE_CREATE_ARRAY,
  &_swigt__p_TPMS_SESSION_AUDIT_INFO,
  &_swigt__p_TPMS_SESSION_AUDIT_INFO_ARRAY,
  &_swigt__p_TPMS_SIGNATURE_ECC,
  &_swigt__p_TPMS_SIGNATURE_ECC_ARRAY,
  &_swigt__p_TPMS_SIGNATURE_RSA,
  &_swigt__p_TPMS_SIGNATURE_RSA_ARRAY,
  &_swigt__p_TPMS_SYMCIPHER_PARMS,
  &_swigt__p_TPMS_SYMCIPHER_PARMS_ARRAY,
  &_swigt__p_TPMS_TAGGED_PCR_SELECT,
  &_swigt__p_TPMS_TAGGED_PCR_SELECT_ARRAY,
  &_swigt__p_TPMS_TAGGED_POLICY,
  &_swigt__p_TPMS_TAGGED_POLICY_ARRAY,
  &_swigt__p_TPMS_TAGGED_PROPERTY,
  &_swigt__p_TPMS_TAGGED_PROPERTY_ARRAY,
  &_swigt__p_TPMS_TIME_ATTEST_INFO,
  &_swigt__p_TPMS_TIME_ATTEST_INFO_ARRAY,
  &_swigt__p_TPMS_TIME_INFO,
  &_swigt__p_TPMS_TIME_INFO_ARRAY,
  &_swigt__p_TPMT_ASYM_SCHEME,
  &_swigt__p_TPMT_ASYM_SCHEME_ARRAY,
  &_swigt__p_TPMT_ECC_SCHEME,
  &_swigt__p_TPMT_ECC_SCHEME_ARRAY,
  &_swigt__p_TPMT_HA,
  &_swigt__p_TPMT_HA_ARRAY,
  &_swigt__p_TPMT_KDF_SCHEME,
  &_swigt__p_TPMT_KDF_SCHEME_ARRAY,
  &_swigt__p_TPMT_KEYEDHASH_SCHEME,
  &_swigt__p_TPMT_KEYEDHASH_SCHEME_ARRAY,
  &_swigt__p_TPMT_PUBLIC,
  &_swigt__p_TPMT_PUBLIC_ARRAY,
  &_swigt__p_TPMT_PUBLIC_PARMS,
  &_swigt__p_TPMT_PUBLIC_PARMS_ARRAY,
  &_swigt__p_TPMT_RSA_DECRYPT,
  &_swigt__p_TPMT_RSA_DECRYPT_ARRAY,
  &_swigt__p_TPMT_RSA_SCHEME,
  &_swigt__p_TPMT_RSA_SCHEME_ARRAY,
  &_swigt__p_TPMT_SENSITIVE,
  &_swigt__p_TPMT_SENSITIVE_ARRAY,
  &_swigt__p_TPMT_SIGNATURE,
  &_swigt__p_TPMT_SIGNATURE_ARRAY,
  &_swigt__p_TPMT_SIG_SCHEME,
  &_swigt__p_TPMT_SIG_SCHEME_ARRAY,
  &_swigt__p_TPMT_SYM_DEF,
  &_swigt__p_TPMT_SYM_DEF_ARRAY,
  &_swigt__p_TPMT_SYM_DEF_OBJECT,
  &_swigt__p_TPMT_SYM_DEF_OBJECT_ARRAY,
  &_swigt__p_TPMT_TK_AUTH,
  &_swigt__p_TPMT_TK_AUTH_ARRAY,
  &_swigt__p_TPMT_TK_CREATION,
  &_swigt__p_TPMT_TK_CREATION_ARRAY,
  &_swigt__p_TPMT_TK_HASHCHECK,
  &_swigt__p_TPMT_TK_HASHCHECK_ARRAY,
  &_swigt__p_TPMT_TK_VERIFIED,
  &_swigt__p_TPMT_TK_VERIFIED_ARRAY,
  &_swigt__p_TPMU_ASYM_SCHEME,
  &_swigt__p_TPMU_ASYM_SCHEME_ARRAY,
  &_swigt__p_TPMU_ATTEST,
  &_swigt__p_TPMU_ATTEST_ARRAY,
  &_swigt__p_TPMU_CAPABILITIES,
  &_swigt__p_TPMU_CAPABILITIES_ARRAY,
  &_swigt__p_TPMU_ENCRYPTED_SECRET,
  &_swigt__p_TPMU_ENCRYPTED_SECRET_ARRAY,
  &_swigt__p_TPMU_HA,
  &_swigt__p_TPMU_HA_ARRAY,
  &_swigt__p_TPMU_KDF_SCHEME,
  &_swigt__p_TPMU_KDF_SCHEME_ARRAY,
  &_swigt__p_TPMU_NAME,
  &_swigt__p_TPMU_NAME_ARRAY,
  &_swigt__p_TPMU_PUBLIC_ID,
  &_swigt__p_TPMU_PUBLIC_ID_ARRAY,
  &_swigt__p_TPMU_PUBLIC_PARMS,
  &_swigt__p_TPMU_PUBLIC_PARMS_ARRAY,
  &_swigt__p_TPMU_SCHEME_KEYEDHASH,
  &_swigt__p_TPMU_SCHEME_KEYEDHASH_ARRAY,
  &_swigt__p_TPMU_SENSITIVE_COMPOSITE,
  &_swigt__p_TPMU_SENSITIVE_COMPOSITE_ARRAY,
  &_swigt__p_TPMU_SIGNATURE,
  &_swigt__p_TPMU_SIGNATURE_ARRAY,
  &_swigt__p_TPMU_SIG_SCHEME,
  &_swigt__p_TPMU_SIG_SCHEME_ARRAY,
  &_swigt__p_TPMU_SYM_KEY_BITS,
  &_swigt__p_TPMU_SYM_KEY_BITS_ARRAY,
  &_swigt__p_TPMU_SYM_MODE,
  &_swigt__p_TPMU_SYM_MODE_ARRAY,
  &_swigt__p_TSS2_ABI_VERSION,
  &_swigt__p_TSS2_TCTI_CONTEXT,
  &_swigt__p_UINT16_PTR,
  &_swigt__p_UINT32_PTR,
  &_swigt__p__PRIVATE,
  &_swigt__p__PRIVATE_ARRAY,
  &_swigt__p_char,
  &_swigt__p_int,
  &_swigt__p_long_long,
  &_swigt__p_p_ESYS_CONTEXT,
  &_swigt__p_p_ESYS_TR_PTR,
  &_swigt__p_p_TPM2B_ATTEST,
  &_swigt__p_p_TPM2B_ATTEST_ARRAY,
  &_swigt__p_p_TPM2B_CONTEXT_DATA,
  &_swigt__p_p_TPM2B_CONTEXT_DATA_ARRAY,
  &_swigt__p_p_TPM2B_CONTEXT_SENSITIVE,
  &_swigt__p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY,
  &_swigt__p_p_TPM2B_CREATION_DATA,
  &_swigt__p_p_TPM2B_CREATION_DATA_ARRAY,
  &_swigt__p_p_TPM2B_DATA,
  &_swigt__p_p_TPM2B_DATA_ARRAY,
  &_swigt__p_p_TPM2B_DIGEST,
  &_swigt__p_p_TPM2B_DIGEST_ARRAY,
  &_swigt__p_p_TPM2B_ECC_PARAMETER,
  &_swigt__p_p_TPM2B_ECC_PARAMETER_ARRAY,
  &_swigt__p_p_TPM2B_ECC_POINT,
  &_swigt__p_p_TPM2B_ECC_POINT_ARRAY,
  &_swigt__p_p_TPM2B_ENCRYPTED_SECRET,
  &_swigt__p_p_TPM2B_ENCRYPTED_SECRET_ARRAY,
  &_swigt__p_p_TPM2B_EVENT,
  &_swigt__p_p_TPM2B_EVENT_ARRAY,
  &_swigt__p_p_TPM2B_ID_OBJECT,
  &_swigt__p_p_TPM2B_ID_OBJECT_ARRAY,
  &_swigt__p_p_TPM2B_IV,
  &_swigt__p_p_TPM2B_IV_ARRAY,
  &_swigt__p_p_TPM2B_MAX_BUFFER,
  &_swigt__p_p_TPM2B_MAX_BUFFER_ARRAY,
  &_swigt__p_p_TPM2B_MAX_NV_BUFFER,
  &_swigt__p_p_TPM2B_MAX_NV_BUFFER_ARRAY,
  &_swigt__p_p_TPM2B_NAME,
  &_swigt__p_p_TPM2B_NAME_ARRAY,
  &_swigt__p_p_TPM2B_NV_PUBLIC,
  &_swigt__p_p_TPM2B_NV_PUBLIC_ARRAY,
  &_swigt__p_p_TPM2B_PRIVATE,
  &_swigt__p_p_TPM2B_PRIVATE_ARRAY,
  &_swigt__p_p_TPM2B_PRIVATE_KEY_RSA,
  &_swigt__p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY,
  &_swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC,
  &_swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY,
  &_swigt__p_p_TPM2B_PUBLIC,
  &_swigt__p_p_TPM2B_PUBLIC_ARRAY,
  &_swigt__p_p_TPM2B_PUBLIC_KEY_RSA,
  &_swigt__p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY,
  &_swigt__p_p_TPM2B_SENSITIVE,
  &_swigt__p_p_TPM2B_SENSITIVE_ARRAY,
  &_swigt__p_p_TPM2B_SENSITIVE_CREATE,
  &_swigt__p_p_TPM2B_SENSITIVE_CREATE_ARRAY,
  &_swigt__p_p_TPM2B_SENSITIVE_DATA,
  &_swigt__p_p_TPM2B_SENSITIVE_DATA_ARRAY,
  &_swigt__p_p_TPM2B_SYM_KEY,
  &_swigt__p_p_TPM2B_SYM_KEY_ARRAY,
  &_swigt__p_p_TPM2B_TEMPLATE,
  &_swigt__p_p_TPM2B_TEMPLATE_ARRAY,
  &_swigt__p_p_TPML_AC_CAPABILITIES,
  &_swigt__p_p_TPML_AC_CAPABILITIES_ARRAY,
  &_swigt__p_p_TPML_ALG,
  &_swigt__p_p_TPML_ALG_ARRAY,
  &_swigt__p_p_TPML_ALG_PROPERTY,
  &_swigt__p_p_TPML_ALG_PROPERTY_ARRAY,
  &_swigt__p_p_TPML_CC,
  &_swigt__p_p_TPML_CCA,
  &_swigt__p_p_TPML_CCA_ARRAY,
  &_swigt__p_p_TPML_CC_ARRAY,
  &_swigt__p_p_TPML_DIGEST,
  &_swigt__p_p_TPML_DIGEST_ARRAY,
  &_swigt__p_p_TPML_DIGEST_VALUES,
  &_swigt__p_p_TPML_DIGEST_VALUES_ARRAY,
  &_swigt__p_p_TPML_ECC_CURVE,
  &_swigt__p_p_TPML_ECC_CURVE_ARRAY,
  &_swigt__p_p_TPML_HANDLE,
  &_swigt__p_p_TPML_HANDLE_ARRAY,
  &_swigt__p_p_TPML_INTEL_PTT_PROPERTY,
  &_swigt__p_p_TPML_INTEL_PTT_PROPERTY_ARRAY,
  &_swigt__p_p_TPML_PCR_SELECTION,
  &_swigt__p_p_TPML_PCR_SELECTION_ARRAY,
  &_swigt__p_p_TPML_TAGGED_PCR_PROPERTY,
  &_swigt__p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY,
  &_swigt__p_p_TPML_TAGGED_TPM_PROPERTY,
  &_swigt__p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY,
  &_swigt__p_p_TPMS_AC_OUTPUT,
  &_swigt__p_p_TPMS_AC_OUTPUT_ARRAY,
  &_swigt__p_p_TPMS_ALGORITHM_DESCRIPTION,
  &_swigt__p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY,
  &_swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC,
  &_swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY,
  &_swigt__p_p_TPMS_ALG_PROPERTY,
  &_swigt__p_p_TPMS_ALG_PROPERTY_ARRAY,
  &_swigt__p_p_TPMS_ASYM_PARMS,
  &_swigt__p_p_TPMS_ASYM_PARMS_ARRAY,
  &_swigt__p_p_TPMS_ATTEST,
  &_swigt__p_p_TPMS_ATTEST_ARRAY,
  &_swigt__p_p_TPMS_AUTH_COMMAND,
  &_swigt__p_p_TPMS_AUTH_COMMAND_ARRAY,
  &_swigt__p_p_TPMS_AUTH_RESPONSE,
  &_swigt__p_p_TPMS_AUTH_RESPONSE_ARRAY,
  &_swigt__p_p_TPMS_CAPABILITY_DATA,
  &_swigt__p_p_TPMS_CAPABILITY_DATA_ARRAY,
  &_swigt__p_p_TPMS_CERTIFY_INFO,
  &_swigt__p_p_TPMS_CERTIFY_INFO_ARRAY,
  &_swigt__p_p_TPMS_CLOCK_INFO,
  &_swigt__p_p_TPMS_CLOCK_INFO_ARRAY,
  &_swigt__p_p_TPMS_COMMAND_AUDIT_INFO,
  &_swigt__p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY,
  &_swigt__p_p_TPMS_CONTEXT,
  &_swigt__p_p_TPMS_CONTEXT_ARRAY,
  &_swigt__p_p_TPMS_CONTEXT_DATA,
  &_swigt__p_p_TPMS_CONTEXT_DATA_ARRAY,
  &_swigt__p_p_TPMS_CREATION_DATA,
  &_swigt__p_p_TPMS_CREATION_DATA_ARRAY,
  &_swigt__p_p_TPMS_CREATION_INFO,
  &_swigt__p_p_TPMS_CREATION_INFO_ARRAY,
  &_swigt__p_p_TPMS_ECC_PARMS,
  &_swigt__p_p_TPMS_ECC_PARMS_ARRAY,
  &_swigt__p_p_TPMS_ECC_POINT,
  &_swigt__p_p_TPMS_ECC_POINT_ARRAY,
  &_swigt__p_p_TPMS_EMPTY,
  &_swigt__p_p_TPMS_EMPTY_ARRAY,
  &_swigt__p_p_TPMS_ID_OBJECT,
  &_swigt__p_p_TPMS_ID_OBJECT_ARRAY,
  &_swigt__p_p_TPMS_KEYEDHASH_PARMS,
  &_swigt__p_p_TPMS_KEYEDHASH_PARMS_ARRAY,
  &_swigt__p_p_TPMS_NV_CERTIFY_INFO,
  &_swigt__p_p_TPMS_NV_CERTIFY_INFO_ARRAY,
  &_swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS,
  &_swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY,
  &_swigt__p_p_TPMS_NV_PUBLIC,
  &_swigt__p_p_TPMS_NV_PUBLIC_ARRAY,
  &_swigt__p_p_TPMS_PCR_SELECT,
  &_swigt__p_p_TPMS_PCR_SELECTION,
  &_swigt__p_p_TPMS_PCR_SELECTION_ARRAY,
  &_swigt__p_p_TPMS_PCR_SELECT_ARRAY,
  &_swigt__p_p_TPMS_QUOTE_INFO,
  &_swigt__p_p_TPMS_QUOTE_INFO_ARRAY,
  &_swigt__p_p_TPMS_RSA_PARMS,
  &_swigt__p_p_TPMS_RSA_PARMS_ARRAY,
  &_swigt__p_p_TPMS_SCHEME_ECDAA,
  &_swigt__p_p_TPMS_SCHEME_ECDAA_ARRAY,
  &_swigt__p_p_TPMS_SCHEME_HASH,
  &_swigt__p_p_TPMS_SCHEME_HASH_ARRAY,
  &_swigt__p_p_TPMS_SCHEME_XOR,
  &_swigt__p_p_TPMS_SCHEME_XOR_ARRAY,
  &_swigt__p_p_TPMS_SENSITIVE_CREATE,
  &_swigt__p_p_TPMS_SENSITIVE_CREATE_ARRAY,
  &_swigt__p_p_TPMS_SESSION_AUDIT_INFO,
  &_swigt__p_p_TPMS_SESSION_AUDIT_INFO_ARRAY,
  &_swigt__p_p_TPMS_SIGNATURE_ECC,
  &_swigt__p_p_TPMS_SIGNATURE_ECC_ARRAY,
  &_swigt__p_p_TPMS_SIGNATURE_RSA,
  &_swigt__p_p_TPMS_SIGNATURE_RSA_ARRAY,
  &_swigt__p_p_TPMS_SYMCIPHER_PARMS,
  &_swigt__p_p_TPMS_SYMCIPHER_PARMS_ARRAY,
  &_swigt__p_p_TPMS_TAGGED_PCR_SELECT,
  &_swigt__p_p_TPMS_TAGGED_PCR_SELECT_ARRAY,
  &_swigt__p_p_TPMS_TAGGED_POLICY,
  &_swigt__p_p_TPMS_TAGGED_POLICY_ARRAY,
  &_swigt__p_p_TPMS_TAGGED_PROPERTY,
  &_swigt__p_p_TPMS_TAGGED_PROPERTY_ARRAY,
  &_swigt__p_p_TPMS_TIME_ATTEST_INFO,
  &_swigt__p_p_TPMS_TIME_ATTEST_INFO_ARRAY,
  &_swigt__p_p_TPMS_TIME_INFO,
  &_swigt__p_p_TPMS_TIME_INFO_ARRAY,
  &_swigt__p_p_TPMT_ASYM_SCHEME,
  &_swigt__p_p_TPMT_ASYM_SCHEME_ARRAY,
  &_swigt__p_p_TPMT_ECC_SCHEME,
  &_swigt__p_p_TPMT_ECC_SCHEME_ARRAY,
  &_swigt__p_p_TPMT_HA,
  &_swigt__p_p_TPMT_HA_ARRAY,
  &_swigt__p_p_TPMT_KDF_SCHEME,
  &_swigt__p_p_TPMT_KDF_SCHEME_ARRAY,
  &_swigt__p_p_TPMT_KEYEDHASH_SCHEME,
  &_swigt__p_p_TPMT_KEYEDHASH_SCHEME_ARRAY,
  &_swigt__p_p_TPMT_PUBLIC,
  &_swigt__p_p_TPMT_PUBLIC_ARRAY,
  &_swigt__p_p_TPMT_PUBLIC_PARMS,
  &_swigt__p_p_TPMT_PUBLIC_PARMS_ARRAY,
  &_swigt__p_p_TPMT_RSA_DECRYPT,
  &_swigt__p_p_TPMT_RSA_DECRYPT_ARRAY,
  &_swigt__p_p_TPMT_RSA_SCHEME,
  &_swigt__p_p_TPMT_RSA_SCHEME_ARRAY,
  &_swigt__p_p_TPMT_SENSITIVE,
  &_swigt__p_p_TPMT_SENSITIVE_ARRAY,
  &_swigt__p_p_TPMT_SIGNATURE,
  &_swigt__p_p_TPMT_SIGNATURE_ARRAY,
  &_swigt__p_p_TPMT_SIG_SCHEME,
  &_swigt__p_p_TPMT_SIG_SCHEME_ARRAY,
  &_swigt__p_p_TPMT_SYM_DEF,
  &_swigt__p_p_TPMT_SYM_DEF_ARRAY,
  &_swigt__p_p_TPMT_SYM_DEF_OBJECT,
  &_swigt__p_p_TPMT_SYM_DEF_OBJECT_ARRAY,
  &_swigt__p_p_TPMT_TK_AUTH,
  &_swigt__p_p_TPMT_TK_AUTH_ARRAY,
  &_swigt__p_p_TPMT_TK_CREATION,
  &_swigt__p_p_TPMT_TK_CREATION_ARRAY,
  &_swigt__p_p_TPMT_TK_HASHCHECK,
  &_swigt__p_p_TPMT_TK_HASHCHECK_ARRAY,
  &_swigt__p_p_TPMT_TK_VERIFIED,
  &_swigt__p_p_TPMT_TK_VERIFIED_ARRAY,
  &_swigt__p_p_TPMU_ASYM_SCHEME,
  &_swigt__p_p_TPMU_ASYM_SCHEME_ARRAY,
  &_swigt__p_p_TPMU_ATTEST,
  &_swigt__p_p_TPMU_ATTEST_ARRAY,
  &_swigt__p_p_TPMU_CAPABILITIES,
  &_swigt__p_p_TPMU_CAPABILITIES_ARRAY,
  &_swigt__p_p_TPMU_ENCRYPTED_SECRET,
  &_swigt__p_p_TPMU_ENCRYPTED_SECRET_ARRAY,
  &_swigt__p_p_TPMU_HA,
  &_swigt__p_p_TPMU_HA_ARRAY,
  &_swigt__p_p_TPMU_KDF_SCHEME,
  &_swigt__p_p_TPMU_KDF_SCHEME_ARRAY,
  &_swigt__p_p_TPMU_NAME,
  &_swigt__p_p_TPMU_NAME_ARRAY,
  &_swigt__p_p_TPMU_PUBLIC_ID,
  &_swigt__p_p_TPMU_PUBLIC_ID_ARRAY,
  &_swigt__p_p_TPMU_PUBLIC_PARMS,
  &_swigt__p_p_TPMU_PUBLIC_PARMS_ARRAY,
  &_swigt__p_p_TPMU_SCHEME_KEYEDHASH,
  &_swigt__p_p_TPMU_SCHEME_KEYEDHASH_ARRAY,
  &_swigt__p_p_TPMU_SENSITIVE_COMPOSITE,
  &_swigt__p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY,
  &_swigt__p_p_TPMU_SIGNATURE,
  &_swigt__p_p_TPMU_SIGNATURE_ARRAY,
  &_swigt__p_p_TPMU_SIG_SCHEME,
  &_swigt__p_p_TPMU_SIG_SCHEME_ARRAY,
  &_swigt__p_p_TPMU_SYM_KEY_BITS,
  &_swigt__p_p_TPMU_SYM_KEY_BITS_ARRAY,
  &_swigt__p_p_TPMU_SYM_MODE,
  &_swigt__p_p_TPMU_SYM_MODE_ARRAY,
  &_swigt__p_p_TSS2_TCTI_CONTEXT,
  &_swigt__p_p_TSS2_TCTI_POLL_HANDLE,
  &_swigt__p_p__PRIVATE,
  &_swigt__p_p__PRIVATE_ARRAY,
  &_swigt__p_p_char,
  &_swigt__p_p_size_t,
  &_swigt__p_p_unsigned_char,
  &_swigt__p_p_unsigned_int,
  &_swigt__p_p_void,
  &_swigt__p_short,
  &_swigt__p_signed_char,
  &_swigt__p_size_t,
  &_swigt__p_uint16_t_PTR,
  &_swigt__p_uint32_t_PTR,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
  &_swigt__p_unsigned_long_long,
  &_swigt__p_unsigned_short,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_ByteArray[] = {  {&_swigt__p_ByteArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ESYS_CONTEXT[] = {  {&_swigt__p_ESYS_CONTEXT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ESYS_TR_PTR[] = {  {&_swigt__p_ESYS_TR_PTR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ATTEST[] = {  {&_swigt__p_TPM2B_ATTEST_ARRAY, _p_TPM2B_ATTEST_ARRAYTo_p_TPM2B_ATTEST, 0, 0},  {&_swigt__p_TPM2B_ATTEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ATTEST_ARRAY[] = {  {&_swigt__p_TPM2B_ATTEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_CONTEXT_DATA[] = {  {&_swigt__p_TPM2B_CONTEXT_DATA, 0, 0, 0},  {&_swigt__p_TPM2B_CONTEXT_DATA_ARRAY, _p_TPM2B_CONTEXT_DATA_ARRAYTo_p_TPM2B_CONTEXT_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_CONTEXT_DATA_ARRAY[] = {  {&_swigt__p_TPM2B_CONTEXT_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_CONTEXT_SENSITIVE[] = {  {&_swigt__p_TPM2B_CONTEXT_SENSITIVE, 0, 0, 0},  {&_swigt__p_TPM2B_CONTEXT_SENSITIVE_ARRAY, _p_TPM2B_CONTEXT_SENSITIVE_ARRAYTo_p_TPM2B_CONTEXT_SENSITIVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_CONTEXT_SENSITIVE_ARRAY[] = {  {&_swigt__p_TPM2B_CONTEXT_SENSITIVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_CREATION_DATA[] = {  {&_swigt__p_TPM2B_CREATION_DATA, 0, 0, 0},  {&_swigt__p_TPM2B_CREATION_DATA_ARRAY, _p_TPM2B_CREATION_DATA_ARRAYTo_p_TPM2B_CREATION_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_CREATION_DATA_ARRAY[] = {  {&_swigt__p_TPM2B_CREATION_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_DATA[] = {  {&_swigt__p_TPM2B_DATA, 0, 0, 0},  {&_swigt__p_TPM2B_DATA_ARRAY, _p_TPM2B_DATA_ARRAYTo_p_TPM2B_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_DATA_ARRAY[] = {  {&_swigt__p_TPM2B_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_DIGEST[] = {  {&_swigt__p_TPM2B_DIGEST_ARRAY, _p_TPM2B_DIGEST_ARRAYTo_p_TPM2B_DIGEST, 0, 0},  {&_swigt__p_TPM2B_DIGEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_DIGEST_ARRAY[] = {  {&_swigt__p_TPM2B_DIGEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ECC_PARAMETER[] = {  {&_swigt__p_TPM2B_ECC_PARAMETER_ARRAY, _p_TPM2B_ECC_PARAMETER_ARRAYTo_p_TPM2B_ECC_PARAMETER, 0, 0},  {&_swigt__p_TPM2B_ECC_PARAMETER, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ECC_PARAMETER_ARRAY[] = {  {&_swigt__p_TPM2B_ECC_PARAMETER_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ECC_POINT[] = {  {&_swigt__p_TPM2B_ECC_POINT_ARRAY, _p_TPM2B_ECC_POINT_ARRAYTo_p_TPM2B_ECC_POINT, 0, 0},  {&_swigt__p_TPM2B_ECC_POINT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ECC_POINT_ARRAY[] = {  {&_swigt__p_TPM2B_ECC_POINT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ENCRYPTED_SECRET[] = {  {&_swigt__p_TPM2B_ENCRYPTED_SECRET_ARRAY, _p_TPM2B_ENCRYPTED_SECRET_ARRAYTo_p_TPM2B_ENCRYPTED_SECRET, 0, 0},  {&_swigt__p_TPM2B_ENCRYPTED_SECRET, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ENCRYPTED_SECRET_ARRAY[] = {  {&_swigt__p_TPM2B_ENCRYPTED_SECRET_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_EVENT[] = {  {&_swigt__p_TPM2B_EVENT_ARRAY, _p_TPM2B_EVENT_ARRAYTo_p_TPM2B_EVENT, 0, 0},  {&_swigt__p_TPM2B_EVENT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_EVENT_ARRAY[] = {  {&_swigt__p_TPM2B_EVENT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ID_OBJECT[] = {  {&_swigt__p_TPM2B_ID_OBJECT_ARRAY, _p_TPM2B_ID_OBJECT_ARRAYTo_p_TPM2B_ID_OBJECT, 0, 0},  {&_swigt__p_TPM2B_ID_OBJECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_ID_OBJECT_ARRAY[] = {  {&_swigt__p_TPM2B_ID_OBJECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_IV[] = {  {&_swigt__p_TPM2B_IV_ARRAY, _p_TPM2B_IV_ARRAYTo_p_TPM2B_IV, 0, 0},  {&_swigt__p_TPM2B_IV, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_IV_ARRAY[] = {  {&_swigt__p_TPM2B_IV_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_MAX_BUFFER[] = {  {&_swigt__p_TPM2B_MAX_BUFFER_ARRAY, _p_TPM2B_MAX_BUFFER_ARRAYTo_p_TPM2B_MAX_BUFFER, 0, 0},  {&_swigt__p_TPM2B_MAX_BUFFER, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_MAX_BUFFER_ARRAY[] = {  {&_swigt__p_TPM2B_MAX_BUFFER_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_MAX_NV_BUFFER[] = {  {&_swigt__p_TPM2B_MAX_NV_BUFFER_ARRAY, _p_TPM2B_MAX_NV_BUFFER_ARRAYTo_p_TPM2B_MAX_NV_BUFFER, 0, 0},  {&_swigt__p_TPM2B_MAX_NV_BUFFER, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_MAX_NV_BUFFER_ARRAY[] = {  {&_swigt__p_TPM2B_MAX_NV_BUFFER_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_NAME[] = {  {&_swigt__p_TPM2B_NAME, 0, 0, 0},  {&_swigt__p_TPM2B_NAME_ARRAY, _p_TPM2B_NAME_ARRAYTo_p_TPM2B_NAME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_NAME_ARRAY[] = {  {&_swigt__p_TPM2B_NAME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_NV_PUBLIC[] = {  {&_swigt__p_TPM2B_NV_PUBLIC, 0, 0, 0},  {&_swigt__p_TPM2B_NV_PUBLIC_ARRAY, _p_TPM2B_NV_PUBLIC_ARRAYTo_p_TPM2B_NV_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_NV_PUBLIC_ARRAY[] = {  {&_swigt__p_TPM2B_NV_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PRIVATE[] = {  {&_swigt__p_TPM2B_PRIVATE, 0, 0, 0},  {&_swigt__p_TPM2B_PRIVATE_ARRAY, _p_TPM2B_PRIVATE_ARRAYTo_p_TPM2B_PRIVATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PRIVATE_ARRAY[] = {  {&_swigt__p_TPM2B_PRIVATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PRIVATE_KEY_RSA[] = {  {&_swigt__p_TPM2B_PRIVATE_KEY_RSA, 0, 0, 0},  {&_swigt__p_TPM2B_PRIVATE_KEY_RSA_ARRAY, _p_TPM2B_PRIVATE_KEY_RSA_ARRAYTo_p_TPM2B_PRIVATE_KEY_RSA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PRIVATE_KEY_RSA_ARRAY[] = {  {&_swigt__p_TPM2B_PRIVATE_KEY_RSA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PRIVATE_VENDOR_SPECIFIC[] = {  {&_swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0, 0, 0},  {&_swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, _p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAYTo_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY[] = {  {&_swigt__p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PUBLIC[] = {  {&_swigt__p_TPM2B_PUBLIC, 0, 0, 0},  {&_swigt__p_TPM2B_PUBLIC_ARRAY, _p_TPM2B_PUBLIC_ARRAYTo_p_TPM2B_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PUBLIC_ARRAY[] = {  {&_swigt__p_TPM2B_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PUBLIC_KEY_RSA[] = {  {&_swigt__p_TPM2B_PUBLIC_KEY_RSA, 0, 0, 0},  {&_swigt__p_TPM2B_PUBLIC_KEY_RSA_ARRAY, _p_TPM2B_PUBLIC_KEY_RSA_ARRAYTo_p_TPM2B_PUBLIC_KEY_RSA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_PUBLIC_KEY_RSA_ARRAY[] = {  {&_swigt__p_TPM2B_PUBLIC_KEY_RSA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SENSITIVE[] = {  {&_swigt__p_TPM2B_SENSITIVE, 0, 0, 0},  {&_swigt__p_TPM2B_SENSITIVE_ARRAY, _p_TPM2B_SENSITIVE_ARRAYTo_p_TPM2B_SENSITIVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SENSITIVE_ARRAY[] = {  {&_swigt__p_TPM2B_SENSITIVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SENSITIVE_CREATE[] = {  {&_swigt__p_TPM2B_SENSITIVE_CREATE, 0, 0, 0},  {&_swigt__p_TPM2B_SENSITIVE_CREATE_ARRAY, _p_TPM2B_SENSITIVE_CREATE_ARRAYTo_p_TPM2B_SENSITIVE_CREATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SENSITIVE_CREATE_ARRAY[] = {  {&_swigt__p_TPM2B_SENSITIVE_CREATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SENSITIVE_DATA[] = {  {&_swigt__p_TPM2B_SENSITIVE_DATA, 0, 0, 0},  {&_swigt__p_TPM2B_SENSITIVE_DATA_ARRAY, _p_TPM2B_SENSITIVE_DATA_ARRAYTo_p_TPM2B_SENSITIVE_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SENSITIVE_DATA_ARRAY[] = {  {&_swigt__p_TPM2B_SENSITIVE_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SYM_KEY[] = {  {&_swigt__p_TPM2B_SYM_KEY, 0, 0, 0},  {&_swigt__p_TPM2B_SYM_KEY_ARRAY, _p_TPM2B_SYM_KEY_ARRAYTo_p_TPM2B_SYM_KEY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_SYM_KEY_ARRAY[] = {  {&_swigt__p_TPM2B_SYM_KEY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_TEMPLATE[] = {  {&_swigt__p_TPM2B_TEMPLATE, 0, 0, 0},  {&_swigt__p_TPM2B_TEMPLATE_ARRAY, _p_TPM2B_TEMPLATE_ARRAYTo_p_TPM2B_TEMPLATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPM2B_TEMPLATE_ARRAY[] = {  {&_swigt__p_TPM2B_TEMPLATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_AC_CAPABILITIES[] = {  {&_swigt__p_TPML_AC_CAPABILITIES, 0, 0, 0},  {&_swigt__p_TPML_AC_CAPABILITIES_ARRAY, _p_TPML_AC_CAPABILITIES_ARRAYTo_p_TPML_AC_CAPABILITIES, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_AC_CAPABILITIES_ARRAY[] = {  {&_swigt__p_TPML_AC_CAPABILITIES_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_ALG[] = {  {&_swigt__p_TPML_ALG, 0, 0, 0},  {&_swigt__p_TPML_ALG_ARRAY, _p_TPML_ALG_ARRAYTo_p_TPML_ALG, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_ALG_ARRAY[] = {  {&_swigt__p_TPML_ALG_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_ALG_PROPERTY[] = {  {&_swigt__p_TPML_ALG_PROPERTY, 0, 0, 0},  {&_swigt__p_TPML_ALG_PROPERTY_ARRAY, _p_TPML_ALG_PROPERTY_ARRAYTo_p_TPML_ALG_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_ALG_PROPERTY_ARRAY[] = {  {&_swigt__p_TPML_ALG_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_CC[] = {  {&_swigt__p_TPML_CC, 0, 0, 0},  {&_swigt__p_TPML_CC_ARRAY, _p_TPML_CC_ARRAYTo_p_TPML_CC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_CCA[] = {  {&_swigt__p_TPML_CCA, 0, 0, 0},  {&_swigt__p_TPML_CCA_ARRAY, _p_TPML_CCA_ARRAYTo_p_TPML_CCA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_CCA_ARRAY[] = {  {&_swigt__p_TPML_CCA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_CC_ARRAY[] = {  {&_swigt__p_TPML_CC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_DIGEST[] = {  {&_swigt__p_TPML_DIGEST_ARRAY, _p_TPML_DIGEST_ARRAYTo_p_TPML_DIGEST, 0, 0},  {&_swigt__p_TPML_DIGEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_DIGEST_ARRAY[] = {  {&_swigt__p_TPML_DIGEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_DIGEST_VALUES[] = {  {&_swigt__p_TPML_DIGEST_VALUES, 0, 0, 0},  {&_swigt__p_TPML_DIGEST_VALUES_ARRAY, _p_TPML_DIGEST_VALUES_ARRAYTo_p_TPML_DIGEST_VALUES, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_DIGEST_VALUES_ARRAY[] = {  {&_swigt__p_TPML_DIGEST_VALUES_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_ECC_CURVE[] = {  {&_swigt__p_TPML_ECC_CURVE, 0, 0, 0},  {&_swigt__p_TPML_ECC_CURVE_ARRAY, _p_TPML_ECC_CURVE_ARRAYTo_p_TPML_ECC_CURVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_ECC_CURVE_ARRAY[] = {  {&_swigt__p_TPML_ECC_CURVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_HANDLE[] = {  {&_swigt__p_TPML_HANDLE, 0, 0, 0},  {&_swigt__p_TPML_HANDLE_ARRAY, _p_TPML_HANDLE_ARRAYTo_p_TPML_HANDLE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_HANDLE_ARRAY[] = {  {&_swigt__p_TPML_HANDLE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_INTEL_PTT_PROPERTY[] = {  {&_swigt__p_TPML_INTEL_PTT_PROPERTY, 0, 0, 0},  {&_swigt__p_TPML_INTEL_PTT_PROPERTY_ARRAY, _p_TPML_INTEL_PTT_PROPERTY_ARRAYTo_p_TPML_INTEL_PTT_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_INTEL_PTT_PROPERTY_ARRAY[] = {  {&_swigt__p_TPML_INTEL_PTT_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_PCR_SELECTION[] = {  {&_swigt__p_TPML_PCR_SELECTION_ARRAY, _p_TPML_PCR_SELECTION_ARRAYTo_p_TPML_PCR_SELECTION, 0, 0},  {&_swigt__p_TPML_PCR_SELECTION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_PCR_SELECTION_ARRAY[] = {  {&_swigt__p_TPML_PCR_SELECTION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_TAGGED_PCR_PROPERTY[] = {  {&_swigt__p_TPML_TAGGED_PCR_PROPERTY, 0, 0, 0},  {&_swigt__p_TPML_TAGGED_PCR_PROPERTY_ARRAY, _p_TPML_TAGGED_PCR_PROPERTY_ARRAYTo_p_TPML_TAGGED_PCR_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_TAGGED_PCR_PROPERTY_ARRAY[] = {  {&_swigt__p_TPML_TAGGED_PCR_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_TAGGED_TPM_PROPERTY[] = {  {&_swigt__p_TPML_TAGGED_TPM_PROPERTY, 0, 0, 0},  {&_swigt__p_TPML_TAGGED_TPM_PROPERTY_ARRAY, _p_TPML_TAGGED_TPM_PROPERTY_ARRAYTo_p_TPML_TAGGED_TPM_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPML_TAGGED_TPM_PROPERTY_ARRAY[] = {  {&_swigt__p_TPML_TAGGED_TPM_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_AC_OUTPUT[] = {  {&_swigt__p_TPMS_AC_OUTPUT_ARRAY, _p_TPMS_AC_OUTPUT_ARRAYTo_p_TPMS_AC_OUTPUT, 0, 0},  {&_swigt__p_TPMS_AC_OUTPUT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_AC_OUTPUT_ARRAY[] = {  {&_swigt__p_TPMS_AC_OUTPUT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ALGORITHM_DESCRIPTION[] = {  {&_swigt__p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, _p_TPMS_ALGORITHM_DESCRIPTION_ARRAYTo_p_TPMS_ALGORITHM_DESCRIPTION, 0, 0},  {&_swigt__p_TPMS_ALGORITHM_DESCRIPTION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ALGORITHM_DESCRIPTION_ARRAY[] = {  {&_swigt__p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ALGORITHM_DETAIL_ECC[] = {  {&_swigt__p_TPMS_ALGORITHM_DETAIL_ECC, 0, 0, 0},  {&_swigt__p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, _p_TPMS_ALGORITHM_DETAIL_ECC_ARRAYTo_p_TPMS_ALGORITHM_DETAIL_ECC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY[] = {  {&_swigt__p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ALG_PROPERTY[] = {  {&_swigt__p_TPMS_ALG_PROPERTY, 0, 0, 0},  {&_swigt__p_TPMS_ALG_PROPERTY_ARRAY, _p_TPMS_ALG_PROPERTY_ARRAYTo_p_TPMS_ALG_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ALG_PROPERTY_ARRAY[] = {  {&_swigt__p_TPMS_ALG_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ASYM_PARMS[] = {  {&_swigt__p_TPMS_ASYM_PARMS, 0, 0, 0},  {&_swigt__p_TPMS_ASYM_PARMS_ARRAY, _p_TPMS_ASYM_PARMS_ARRAYTo_p_TPMS_ASYM_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ASYM_PARMS_ARRAY[] = {  {&_swigt__p_TPMS_ASYM_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ATTEST[] = {  {&_swigt__p_TPMS_ATTEST_ARRAY, _p_TPMS_ATTEST_ARRAYTo_p_TPMS_ATTEST, 0, 0},  {&_swigt__p_TPMS_ATTEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ATTEST_ARRAY[] = {  {&_swigt__p_TPMS_ATTEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_AUTH_COMMAND[] = {  {&_swigt__p_TPMS_AUTH_COMMAND_ARRAY, _p_TPMS_AUTH_COMMAND_ARRAYTo_p_TPMS_AUTH_COMMAND, 0, 0},  {&_swigt__p_TPMS_AUTH_COMMAND, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_AUTH_COMMAND_ARRAY[] = {  {&_swigt__p_TPMS_AUTH_COMMAND_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_AUTH_RESPONSE[] = {  {&_swigt__p_TPMS_AUTH_RESPONSE, 0, 0, 0},  {&_swigt__p_TPMS_AUTH_RESPONSE_ARRAY, _p_TPMS_AUTH_RESPONSE_ARRAYTo_p_TPMS_AUTH_RESPONSE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_AUTH_RESPONSE_ARRAY[] = {  {&_swigt__p_TPMS_AUTH_RESPONSE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CAPABILITY_DATA[] = {  {&_swigt__p_TPMS_CAPABILITY_DATA, 0, 0, 0},  {&_swigt__p_TPMS_CAPABILITY_DATA_ARRAY, _p_TPMS_CAPABILITY_DATA_ARRAYTo_p_TPMS_CAPABILITY_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CAPABILITY_DATA_ARRAY[] = {  {&_swigt__p_TPMS_CAPABILITY_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CERTIFY_INFO[] = {  {&_swigt__p_TPMS_CERTIFY_INFO, 0, 0, 0},  {&_swigt__p_TPMS_CERTIFY_INFO_ARRAY, _p_TPMS_CERTIFY_INFO_ARRAYTo_p_TPMS_CERTIFY_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CERTIFY_INFO_ARRAY[] = {  {&_swigt__p_TPMS_CERTIFY_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CLOCK_INFO[] = {  {&_swigt__p_TPMS_CLOCK_INFO, 0, 0, 0},  {&_swigt__p_TPMS_CLOCK_INFO_ARRAY, _p_TPMS_CLOCK_INFO_ARRAYTo_p_TPMS_CLOCK_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CLOCK_INFO_ARRAY[] = {  {&_swigt__p_TPMS_CLOCK_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_COMMAND_AUDIT_INFO[] = {  {&_swigt__p_TPMS_COMMAND_AUDIT_INFO, 0, 0, 0},  {&_swigt__p_TPMS_COMMAND_AUDIT_INFO_ARRAY, _p_TPMS_COMMAND_AUDIT_INFO_ARRAYTo_p_TPMS_COMMAND_AUDIT_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_COMMAND_AUDIT_INFO_ARRAY[] = {  {&_swigt__p_TPMS_COMMAND_AUDIT_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CONTEXT[] = {  {&_swigt__p_TPMS_CONTEXT, 0, 0, 0},  {&_swigt__p_TPMS_CONTEXT_ARRAY, _p_TPMS_CONTEXT_ARRAYTo_p_TPMS_CONTEXT, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CONTEXT_ARRAY[] = {  {&_swigt__p_TPMS_CONTEXT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CONTEXT_DATA[] = {  {&_swigt__p_TPMS_CONTEXT_DATA, 0, 0, 0},  {&_swigt__p_TPMS_CONTEXT_DATA_ARRAY, _p_TPMS_CONTEXT_DATA_ARRAYTo_p_TPMS_CONTEXT_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CONTEXT_DATA_ARRAY[] = {  {&_swigt__p_TPMS_CONTEXT_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CREATION_DATA[] = {  {&_swigt__p_TPMS_CREATION_DATA, 0, 0, 0},  {&_swigt__p_TPMS_CREATION_DATA_ARRAY, _p_TPMS_CREATION_DATA_ARRAYTo_p_TPMS_CREATION_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CREATION_DATA_ARRAY[] = {  {&_swigt__p_TPMS_CREATION_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CREATION_INFO[] = {  {&_swigt__p_TPMS_CREATION_INFO, 0, 0, 0},  {&_swigt__p_TPMS_CREATION_INFO_ARRAY, _p_TPMS_CREATION_INFO_ARRAYTo_p_TPMS_CREATION_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_CREATION_INFO_ARRAY[] = {  {&_swigt__p_TPMS_CREATION_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ECC_PARMS[] = {  {&_swigt__p_TPMS_ECC_PARMS, 0, 0, 0},  {&_swigt__p_TPMS_ECC_PARMS_ARRAY, _p_TPMS_ECC_PARMS_ARRAYTo_p_TPMS_ECC_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ECC_PARMS_ARRAY[] = {  {&_swigt__p_TPMS_ECC_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ECC_POINT[] = {  {&_swigt__p_TPMS_ECC_POINT_ARRAY, _p_TPMS_ECC_POINT_ARRAYTo_p_TPMS_ECC_POINT, 0, 0},  {&_swigt__p_TPMS_ECC_POINT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ECC_POINT_ARRAY[] = {  {&_swigt__p_TPMS_ECC_POINT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_EMPTY[] = {  {&_swigt__p_TPMS_EMPTY, 0, 0, 0},  {&_swigt__p_TPMS_EMPTY_ARRAY, _p_TPMS_EMPTY_ARRAYTo_p_TPMS_EMPTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_EMPTY_ARRAY[] = {  {&_swigt__p_TPMS_EMPTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ID_OBJECT[] = {  {&_swigt__p_TPMS_ID_OBJECT_ARRAY, _p_TPMS_ID_OBJECT_ARRAYTo_p_TPMS_ID_OBJECT, 0, 0},  {&_swigt__p_TPMS_ID_OBJECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_ID_OBJECT_ARRAY[] = {  {&_swigt__p_TPMS_ID_OBJECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_KEYEDHASH_PARMS[] = {  {&_swigt__p_TPMS_KEYEDHASH_PARMS, 0, 0, 0},  {&_swigt__p_TPMS_KEYEDHASH_PARMS_ARRAY, _p_TPMS_KEYEDHASH_PARMS_ARRAYTo_p_TPMS_KEYEDHASH_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_KEYEDHASH_PARMS_ARRAY[] = {  {&_swigt__p_TPMS_KEYEDHASH_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_NV_CERTIFY_INFO[] = {  {&_swigt__p_TPMS_NV_CERTIFY_INFO, 0, 0, 0},  {&_swigt__p_TPMS_NV_CERTIFY_INFO_ARRAY, _p_TPMS_NV_CERTIFY_INFO_ARRAYTo_p_TPMS_NV_CERTIFY_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_NV_CERTIFY_INFO_ARRAY[] = {  {&_swigt__p_TPMS_NV_CERTIFY_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_NV_PIN_COUNTER_PARAMETERS[] = {  {&_swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0, 0, 0},  {&_swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, _p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAYTo_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY[] = {  {&_swigt__p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_NV_PUBLIC[] = {  {&_swigt__p_TPMS_NV_PUBLIC, 0, 0, 0},  {&_swigt__p_TPMS_NV_PUBLIC_ARRAY, _p_TPMS_NV_PUBLIC_ARRAYTo_p_TPMS_NV_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_NV_PUBLIC_ARRAY[] = {  {&_swigt__p_TPMS_NV_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_PCR_SELECT[] = {  {&_swigt__p_TPMS_PCR_SELECT_ARRAY, _p_TPMS_PCR_SELECT_ARRAYTo_p_TPMS_PCR_SELECT, 0, 0},  {&_swigt__p_TPMS_PCR_SELECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_PCR_SELECTION[] = {  {&_swigt__p_TPMS_PCR_SELECTION_ARRAY, _p_TPMS_PCR_SELECTION_ARRAYTo_p_TPMS_PCR_SELECTION, 0, 0},  {&_swigt__p_TPMS_PCR_SELECTION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_PCR_SELECTION_ARRAY[] = {  {&_swigt__p_TPMS_PCR_SELECTION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_PCR_SELECT_ARRAY[] = {  {&_swigt__p_TPMS_PCR_SELECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_QUOTE_INFO[] = {  {&_swigt__p_TPMS_QUOTE_INFO, 0, 0, 0},  {&_swigt__p_TPMS_QUOTE_INFO_ARRAY, _p_TPMS_QUOTE_INFO_ARRAYTo_p_TPMS_QUOTE_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_QUOTE_INFO_ARRAY[] = {  {&_swigt__p_TPMS_QUOTE_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_RSA_PARMS[] = {  {&_swigt__p_TPMS_RSA_PARMS, 0, 0, 0},  {&_swigt__p_TPMS_RSA_PARMS_ARRAY, _p_TPMS_RSA_PARMS_ARRAYTo_p_TPMS_RSA_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_RSA_PARMS_ARRAY[] = {  {&_swigt__p_TPMS_RSA_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SCHEME_ECDAA[] = {  {&_swigt__p_TPMS_SCHEME_ECDAA, 0, 0, 0},  {&_swigt__p_TPMS_SCHEME_ECDAA_ARRAY, _p_TPMS_SCHEME_ECDAA_ARRAYTo_p_TPMS_SCHEME_ECDAA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SCHEME_ECDAA_ARRAY[] = {  {&_swigt__p_TPMS_SCHEME_ECDAA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SCHEME_HASH[] = {  {&_swigt__p_TPMS_SCHEME_HASH_ARRAY, _p_TPMS_SCHEME_HASH_ARRAYTo_p_TPMS_SCHEME_HASH, 0, 0},  {&_swigt__p_TPMS_SCHEME_HASH, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SCHEME_HASH_ARRAY[] = {  {&_swigt__p_TPMS_SCHEME_HASH_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SCHEME_XOR[] = {  {&_swigt__p_TPMS_SCHEME_XOR_ARRAY, _p_TPMS_SCHEME_XOR_ARRAYTo_p_TPMS_SCHEME_XOR, 0, 0},  {&_swigt__p_TPMS_SCHEME_XOR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SCHEME_XOR_ARRAY[] = {  {&_swigt__p_TPMS_SCHEME_XOR_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SENSITIVE_CREATE[] = {  {&_swigt__p_TPMS_SENSITIVE_CREATE, 0, 0, 0},  {&_swigt__p_TPMS_SENSITIVE_CREATE_ARRAY, _p_TPMS_SENSITIVE_CREATE_ARRAYTo_p_TPMS_SENSITIVE_CREATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SENSITIVE_CREATE_ARRAY[] = {  {&_swigt__p_TPMS_SENSITIVE_CREATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SESSION_AUDIT_INFO[] = {  {&_swigt__p_TPMS_SESSION_AUDIT_INFO, 0, 0, 0},  {&_swigt__p_TPMS_SESSION_AUDIT_INFO_ARRAY, _p_TPMS_SESSION_AUDIT_INFO_ARRAYTo_p_TPMS_SESSION_AUDIT_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SESSION_AUDIT_INFO_ARRAY[] = {  {&_swigt__p_TPMS_SESSION_AUDIT_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SIGNATURE_ECC[] = {  {&_swigt__p_TPMS_SIGNATURE_ECC, 0, 0, 0},  {&_swigt__p_TPMS_SIGNATURE_ECC_ARRAY, _p_TPMS_SIGNATURE_ECC_ARRAYTo_p_TPMS_SIGNATURE_ECC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SIGNATURE_ECC_ARRAY[] = {  {&_swigt__p_TPMS_SIGNATURE_ECC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SIGNATURE_RSA[] = {  {&_swigt__p_TPMS_SIGNATURE_RSA, 0, 0, 0},  {&_swigt__p_TPMS_SIGNATURE_RSA_ARRAY, _p_TPMS_SIGNATURE_RSA_ARRAYTo_p_TPMS_SIGNATURE_RSA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SIGNATURE_RSA_ARRAY[] = {  {&_swigt__p_TPMS_SIGNATURE_RSA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SYMCIPHER_PARMS[] = {  {&_swigt__p_TPMS_SYMCIPHER_PARMS, 0, 0, 0},  {&_swigt__p_TPMS_SYMCIPHER_PARMS_ARRAY, _p_TPMS_SYMCIPHER_PARMS_ARRAYTo_p_TPMS_SYMCIPHER_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_SYMCIPHER_PARMS_ARRAY[] = {  {&_swigt__p_TPMS_SYMCIPHER_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TAGGED_PCR_SELECT[] = {  {&_swigt__p_TPMS_TAGGED_PCR_SELECT_ARRAY, _p_TPMS_TAGGED_PCR_SELECT_ARRAYTo_p_TPMS_TAGGED_PCR_SELECT, 0, 0},  {&_swigt__p_TPMS_TAGGED_PCR_SELECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TAGGED_PCR_SELECT_ARRAY[] = {  {&_swigt__p_TPMS_TAGGED_PCR_SELECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TAGGED_POLICY[] = {  {&_swigt__p_TPMS_TAGGED_POLICY, 0, 0, 0},  {&_swigt__p_TPMS_TAGGED_POLICY_ARRAY, _p_TPMS_TAGGED_POLICY_ARRAYTo_p_TPMS_TAGGED_POLICY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TAGGED_POLICY_ARRAY[] = {  {&_swigt__p_TPMS_TAGGED_POLICY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TAGGED_PROPERTY[] = {  {&_swigt__p_TPMS_TAGGED_PROPERTY, 0, 0, 0},  {&_swigt__p_TPMS_TAGGED_PROPERTY_ARRAY, _p_TPMS_TAGGED_PROPERTY_ARRAYTo_p_TPMS_TAGGED_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TAGGED_PROPERTY_ARRAY[] = {  {&_swigt__p_TPMS_TAGGED_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TIME_ATTEST_INFO[] = {  {&_swigt__p_TPMS_TIME_ATTEST_INFO, 0, 0, 0},  {&_swigt__p_TPMS_TIME_ATTEST_INFO_ARRAY, _p_TPMS_TIME_ATTEST_INFO_ARRAYTo_p_TPMS_TIME_ATTEST_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TIME_ATTEST_INFO_ARRAY[] = {  {&_swigt__p_TPMS_TIME_ATTEST_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TIME_INFO[] = {  {&_swigt__p_TPMS_TIME_INFO, 0, 0, 0},  {&_swigt__p_TPMS_TIME_INFO_ARRAY, _p_TPMS_TIME_INFO_ARRAYTo_p_TPMS_TIME_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMS_TIME_INFO_ARRAY[] = {  {&_swigt__p_TPMS_TIME_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_ASYM_SCHEME[] = {  {&_swigt__p_TPMT_ASYM_SCHEME, 0, 0, 0},  {&_swigt__p_TPMT_ASYM_SCHEME_ARRAY, _p_TPMT_ASYM_SCHEME_ARRAYTo_p_TPMT_ASYM_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_ASYM_SCHEME_ARRAY[] = {  {&_swigt__p_TPMT_ASYM_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_ECC_SCHEME[] = {  {&_swigt__p_TPMT_ECC_SCHEME, 0, 0, 0},  {&_swigt__p_TPMT_ECC_SCHEME_ARRAY, _p_TPMT_ECC_SCHEME_ARRAYTo_p_TPMT_ECC_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_ECC_SCHEME_ARRAY[] = {  {&_swigt__p_TPMT_ECC_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_HA[] = {  {&_swigt__p_TPMT_HA, 0, 0, 0},  {&_swigt__p_TPMT_HA_ARRAY, _p_TPMT_HA_ARRAYTo_p_TPMT_HA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_HA_ARRAY[] = {  {&_swigt__p_TPMT_HA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_KDF_SCHEME[] = {  {&_swigt__p_TPMT_KDF_SCHEME, 0, 0, 0},  {&_swigt__p_TPMT_KDF_SCHEME_ARRAY, _p_TPMT_KDF_SCHEME_ARRAYTo_p_TPMT_KDF_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_KDF_SCHEME_ARRAY[] = {  {&_swigt__p_TPMT_KDF_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_KEYEDHASH_SCHEME[] = {  {&_swigt__p_TPMT_KEYEDHASH_SCHEME, 0, 0, 0},  {&_swigt__p_TPMT_KEYEDHASH_SCHEME_ARRAY, _p_TPMT_KEYEDHASH_SCHEME_ARRAYTo_p_TPMT_KEYEDHASH_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_KEYEDHASH_SCHEME_ARRAY[] = {  {&_swigt__p_TPMT_KEYEDHASH_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_PUBLIC[] = {  {&_swigt__p_TPMT_PUBLIC, 0, 0, 0},  {&_swigt__p_TPMT_PUBLIC_ARRAY, _p_TPMT_PUBLIC_ARRAYTo_p_TPMT_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_PUBLIC_ARRAY[] = {  {&_swigt__p_TPMT_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_PUBLIC_PARMS[] = {  {&_swigt__p_TPMT_PUBLIC_PARMS, 0, 0, 0},  {&_swigt__p_TPMT_PUBLIC_PARMS_ARRAY, _p_TPMT_PUBLIC_PARMS_ARRAYTo_p_TPMT_PUBLIC_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_PUBLIC_PARMS_ARRAY[] = {  {&_swigt__p_TPMT_PUBLIC_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_RSA_DECRYPT[] = {  {&_swigt__p_TPMT_RSA_DECRYPT_ARRAY, _p_TPMT_RSA_DECRYPT_ARRAYTo_p_TPMT_RSA_DECRYPT, 0, 0},  {&_swigt__p_TPMT_RSA_DECRYPT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_RSA_DECRYPT_ARRAY[] = {  {&_swigt__p_TPMT_RSA_DECRYPT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_RSA_SCHEME[] = {  {&_swigt__p_TPMT_RSA_SCHEME, 0, 0, 0},  {&_swigt__p_TPMT_RSA_SCHEME_ARRAY, _p_TPMT_RSA_SCHEME_ARRAYTo_p_TPMT_RSA_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_RSA_SCHEME_ARRAY[] = {  {&_swigt__p_TPMT_RSA_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SENSITIVE[] = {  {&_swigt__p_TPMT_SENSITIVE, 0, 0, 0},  {&_swigt__p_TPMT_SENSITIVE_ARRAY, _p_TPMT_SENSITIVE_ARRAYTo_p_TPMT_SENSITIVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SENSITIVE_ARRAY[] = {  {&_swigt__p_TPMT_SENSITIVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SIGNATURE[] = {  {&_swigt__p_TPMT_SIGNATURE, 0, 0, 0},  {&_swigt__p_TPMT_SIGNATURE_ARRAY, _p_TPMT_SIGNATURE_ARRAYTo_p_TPMT_SIGNATURE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SIGNATURE_ARRAY[] = {  {&_swigt__p_TPMT_SIGNATURE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SIG_SCHEME[] = {  {&_swigt__p_TPMT_SIG_SCHEME, 0, 0, 0},  {&_swigt__p_TPMT_SIG_SCHEME_ARRAY, _p_TPMT_SIG_SCHEME_ARRAYTo_p_TPMT_SIG_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SIG_SCHEME_ARRAY[] = {  {&_swigt__p_TPMT_SIG_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SYM_DEF[] = {  {&_swigt__p_TPMT_SYM_DEF_ARRAY, _p_TPMT_SYM_DEF_ARRAYTo_p_TPMT_SYM_DEF, 0, 0},  {&_swigt__p_TPMT_SYM_DEF, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SYM_DEF_ARRAY[] = {  {&_swigt__p_TPMT_SYM_DEF_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SYM_DEF_OBJECT[] = {  {&_swigt__p_TPMT_SYM_DEF_OBJECT_ARRAY, _p_TPMT_SYM_DEF_OBJECT_ARRAYTo_p_TPMT_SYM_DEF_OBJECT, 0, 0},  {&_swigt__p_TPMT_SYM_DEF_OBJECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_SYM_DEF_OBJECT_ARRAY[] = {  {&_swigt__p_TPMT_SYM_DEF_OBJECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_AUTH[] = {  {&_swigt__p_TPMT_TK_AUTH_ARRAY, _p_TPMT_TK_AUTH_ARRAYTo_p_TPMT_TK_AUTH, 0, 0},  {&_swigt__p_TPMT_TK_AUTH, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_AUTH_ARRAY[] = {  {&_swigt__p_TPMT_TK_AUTH_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_CREATION[] = {  {&_swigt__p_TPMT_TK_CREATION_ARRAY, _p_TPMT_TK_CREATION_ARRAYTo_p_TPMT_TK_CREATION, 0, 0},  {&_swigt__p_TPMT_TK_CREATION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_CREATION_ARRAY[] = {  {&_swigt__p_TPMT_TK_CREATION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_HASHCHECK[] = {  {&_swigt__p_TPMT_TK_HASHCHECK, 0, 0, 0},  {&_swigt__p_TPMT_TK_HASHCHECK_ARRAY, _p_TPMT_TK_HASHCHECK_ARRAYTo_p_TPMT_TK_HASHCHECK, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_HASHCHECK_ARRAY[] = {  {&_swigt__p_TPMT_TK_HASHCHECK_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_VERIFIED[] = {  {&_swigt__p_TPMT_TK_VERIFIED, 0, 0, 0},  {&_swigt__p_TPMT_TK_VERIFIED_ARRAY, _p_TPMT_TK_VERIFIED_ARRAYTo_p_TPMT_TK_VERIFIED, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMT_TK_VERIFIED_ARRAY[] = {  {&_swigt__p_TPMT_TK_VERIFIED_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_ASYM_SCHEME[] = {  {&_swigt__p_TPMU_ASYM_SCHEME, 0, 0, 0},  {&_swigt__p_TPMU_ASYM_SCHEME_ARRAY, _p_TPMU_ASYM_SCHEME_ARRAYTo_p_TPMU_ASYM_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_ASYM_SCHEME_ARRAY[] = {  {&_swigt__p_TPMU_ASYM_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_ATTEST[] = {  {&_swigt__p_TPMU_ATTEST_ARRAY, _p_TPMU_ATTEST_ARRAYTo_p_TPMU_ATTEST, 0, 0},  {&_swigt__p_TPMU_ATTEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_ATTEST_ARRAY[] = {  {&_swigt__p_TPMU_ATTEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_CAPABILITIES[] = {  {&_swigt__p_TPMU_CAPABILITIES, 0, 0, 0},  {&_swigt__p_TPMU_CAPABILITIES_ARRAY, _p_TPMU_CAPABILITIES_ARRAYTo_p_TPMU_CAPABILITIES, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_CAPABILITIES_ARRAY[] = {  {&_swigt__p_TPMU_CAPABILITIES_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_ENCRYPTED_SECRET[] = {  {&_swigt__p_TPMU_ENCRYPTED_SECRET_ARRAY, _p_TPMU_ENCRYPTED_SECRET_ARRAYTo_p_TPMU_ENCRYPTED_SECRET, 0, 0},  {&_swigt__p_TPMU_ENCRYPTED_SECRET, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_ENCRYPTED_SECRET_ARRAY[] = {  {&_swigt__p_TPMU_ENCRYPTED_SECRET_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_HA[] = {  {&_swigt__p_TPMU_HA, 0, 0, 0},  {&_swigt__p_TPMU_HA_ARRAY, _p_TPMU_HA_ARRAYTo_p_TPMU_HA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_HA_ARRAY[] = {  {&_swigt__p_TPMU_HA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_KDF_SCHEME[] = {  {&_swigt__p_TPMU_KDF_SCHEME, 0, 0, 0},  {&_swigt__p_TPMU_KDF_SCHEME_ARRAY, _p_TPMU_KDF_SCHEME_ARRAYTo_p_TPMU_KDF_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_KDF_SCHEME_ARRAY[] = {  {&_swigt__p_TPMU_KDF_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_NAME[] = {  {&_swigt__p_TPMU_NAME, 0, 0, 0},  {&_swigt__p_TPMU_NAME_ARRAY, _p_TPMU_NAME_ARRAYTo_p_TPMU_NAME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_NAME_ARRAY[] = {  {&_swigt__p_TPMU_NAME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_PUBLIC_ID[] = {  {&_swigt__p_TPMU_PUBLIC_ID_ARRAY, _p_TPMU_PUBLIC_ID_ARRAYTo_p_TPMU_PUBLIC_ID, 0, 0},  {&_swigt__p_TPMU_PUBLIC_ID, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_PUBLIC_ID_ARRAY[] = {  {&_swigt__p_TPMU_PUBLIC_ID_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_PUBLIC_PARMS[] = {  {&_swigt__p_TPMU_PUBLIC_PARMS, 0, 0, 0},  {&_swigt__p_TPMU_PUBLIC_PARMS_ARRAY, _p_TPMU_PUBLIC_PARMS_ARRAYTo_p_TPMU_PUBLIC_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_PUBLIC_PARMS_ARRAY[] = {  {&_swigt__p_TPMU_PUBLIC_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SCHEME_KEYEDHASH[] = {  {&_swigt__p_TPMU_SCHEME_KEYEDHASH, 0, 0, 0},  {&_swigt__p_TPMU_SCHEME_KEYEDHASH_ARRAY, _p_TPMU_SCHEME_KEYEDHASH_ARRAYTo_p_TPMU_SCHEME_KEYEDHASH, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SCHEME_KEYEDHASH_ARRAY[] = {  {&_swigt__p_TPMU_SCHEME_KEYEDHASH_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SENSITIVE_COMPOSITE[] = {  {&_swigt__p_TPMU_SENSITIVE_COMPOSITE, 0, 0, 0},  {&_swigt__p_TPMU_SENSITIVE_COMPOSITE_ARRAY, _p_TPMU_SENSITIVE_COMPOSITE_ARRAYTo_p_TPMU_SENSITIVE_COMPOSITE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SENSITIVE_COMPOSITE_ARRAY[] = {  {&_swigt__p_TPMU_SENSITIVE_COMPOSITE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SIGNATURE[] = {  {&_swigt__p_TPMU_SIGNATURE, 0, 0, 0},  {&_swigt__p_TPMU_SIGNATURE_ARRAY, _p_TPMU_SIGNATURE_ARRAYTo_p_TPMU_SIGNATURE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SIGNATURE_ARRAY[] = {  {&_swigt__p_TPMU_SIGNATURE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SIG_SCHEME[] = {  {&_swigt__p_TPMU_SIG_SCHEME, 0, 0, 0},  {&_swigt__p_TPMU_SIG_SCHEME_ARRAY, _p_TPMU_SIG_SCHEME_ARRAYTo_p_TPMU_SIG_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SIG_SCHEME_ARRAY[] = {  {&_swigt__p_TPMU_SIG_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SYM_KEY_BITS[] = {  {&_swigt__p_TPMU_SYM_KEY_BITS, 0, 0, 0},  {&_swigt__p_TPMU_SYM_KEY_BITS_ARRAY, _p_TPMU_SYM_KEY_BITS_ARRAYTo_p_TPMU_SYM_KEY_BITS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SYM_KEY_BITS_ARRAY[] = {  {&_swigt__p_TPMU_SYM_KEY_BITS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SYM_MODE[] = {  {&_swigt__p_TPMU_SYM_MODE, 0, 0, 0},  {&_swigt__p_TPMU_SYM_MODE_ARRAY, _p_TPMU_SYM_MODE_ARRAYTo_p_TPMU_SYM_MODE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TPMU_SYM_MODE_ARRAY[] = {  {&_swigt__p_TPMU_SYM_MODE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TSS2_ABI_VERSION[] = {  {&_swigt__p_TSS2_ABI_VERSION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TSS2_TCTI_CONTEXT[] = {  {&_swigt__p_TSS2_TCTI_CONTEXT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_UINT16_PTR[] = {  {&_swigt__p_UINT16_PTR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_UINT32_PTR[] = {  {&_swigt__p_UINT32_PTR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__PRIVATE[] = {  {&_swigt__p__PRIVATE, 0, 0, 0},  {&_swigt__p__PRIVATE_ARRAY, _p__PRIVATE_ARRAYTo_p__PRIVATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__PRIVATE_ARRAY[] = {  {&_swigt__p__PRIVATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long_long[] = {  {&_swigt__p_long_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_ESYS_CONTEXT[] = {  {&_swigt__p_p_ESYS_CONTEXT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ATTEST_ARRAY[] = {{&_swigt__p_p_TPM2B_ATTEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ATTEST[] = {  {&_swigt__p_p_TPM2B_ATTEST_ARRAY, _p_p_TPM2B_ATTEST_ARRAYTo_p_p_TPM2B_ATTEST, 0, 0},  {&_swigt__p_p_TPM2B_ATTEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_CONTEXT_DATA_ARRAY[] = {{&_swigt__p_p_TPM2B_CONTEXT_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_CONTEXT_DATA[] = {  {&_swigt__p_p_TPM2B_CONTEXT_DATA, 0, 0, 0},  {&_swigt__p_p_TPM2B_CONTEXT_DATA_ARRAY, _p_p_TPM2B_CONTEXT_DATA_ARRAYTo_p_p_TPM2B_CONTEXT_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY[] = {{&_swigt__p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_CONTEXT_SENSITIVE[] = {  {&_swigt__p_p_TPM2B_CONTEXT_SENSITIVE, 0, 0, 0},  {&_swigt__p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY, _p_p_TPM2B_CONTEXT_SENSITIVE_ARRAYTo_p_p_TPM2B_CONTEXT_SENSITIVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_CREATION_DATA_ARRAY[] = {{&_swigt__p_p_TPM2B_CREATION_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_CREATION_DATA[] = {  {&_swigt__p_p_TPM2B_CREATION_DATA, 0, 0, 0},  {&_swigt__p_p_TPM2B_CREATION_DATA_ARRAY, _p_p_TPM2B_CREATION_DATA_ARRAYTo_p_p_TPM2B_CREATION_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_DATA_ARRAY[] = {{&_swigt__p_p_TPM2B_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_DATA[] = {  {&_swigt__p_p_TPM2B_DATA, 0, 0, 0},  {&_swigt__p_p_TPM2B_DATA_ARRAY, _p_p_TPM2B_DATA_ARRAYTo_p_p_TPM2B_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_DIGEST_ARRAY[] = {{&_swigt__p_p_TPM2B_DIGEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_DIGEST[] = {  {&_swigt__p_p_TPM2B_DIGEST_ARRAY, _p_p_TPM2B_DIGEST_ARRAYTo_p_p_TPM2B_DIGEST, 0, 0},  {&_swigt__p_p_TPM2B_DIGEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ECC_PARAMETER_ARRAY[] = {{&_swigt__p_p_TPM2B_ECC_PARAMETER_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ECC_PARAMETER[] = {  {&_swigt__p_p_TPM2B_ECC_PARAMETER_ARRAY, _p_p_TPM2B_ECC_PARAMETER_ARRAYTo_p_p_TPM2B_ECC_PARAMETER, 0, 0},  {&_swigt__p_p_TPM2B_ECC_PARAMETER, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ECC_POINT_ARRAY[] = {{&_swigt__p_p_TPM2B_ECC_POINT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ECC_POINT[] = {  {&_swigt__p_p_TPM2B_ECC_POINT_ARRAY, _p_p_TPM2B_ECC_POINT_ARRAYTo_p_p_TPM2B_ECC_POINT, 0, 0},  {&_swigt__p_p_TPM2B_ECC_POINT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ENCRYPTED_SECRET_ARRAY[] = {{&_swigt__p_p_TPM2B_ENCRYPTED_SECRET_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ENCRYPTED_SECRET[] = {  {&_swigt__p_p_TPM2B_ENCRYPTED_SECRET_ARRAY, _p_p_TPM2B_ENCRYPTED_SECRET_ARRAYTo_p_p_TPM2B_ENCRYPTED_SECRET, 0, 0},  {&_swigt__p_p_TPM2B_ENCRYPTED_SECRET, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_EVENT_ARRAY[] = {{&_swigt__p_p_TPM2B_EVENT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_EVENT[] = {  {&_swigt__p_p_TPM2B_EVENT_ARRAY, _p_p_TPM2B_EVENT_ARRAYTo_p_p_TPM2B_EVENT, 0, 0},  {&_swigt__p_p_TPM2B_EVENT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ID_OBJECT_ARRAY[] = {{&_swigt__p_p_TPM2B_ID_OBJECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_ID_OBJECT[] = {  {&_swigt__p_p_TPM2B_ID_OBJECT_ARRAY, _p_p_TPM2B_ID_OBJECT_ARRAYTo_p_p_TPM2B_ID_OBJECT, 0, 0},  {&_swigt__p_p_TPM2B_ID_OBJECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_IV_ARRAY[] = {{&_swigt__p_p_TPM2B_IV_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_IV[] = {  {&_swigt__p_p_TPM2B_IV_ARRAY, _p_p_TPM2B_IV_ARRAYTo_p_p_TPM2B_IV, 0, 0},  {&_swigt__p_p_TPM2B_IV, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_MAX_BUFFER_ARRAY[] = {{&_swigt__p_p_TPM2B_MAX_BUFFER_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_MAX_BUFFER[] = {  {&_swigt__p_p_TPM2B_MAX_BUFFER_ARRAY, _p_p_TPM2B_MAX_BUFFER_ARRAYTo_p_p_TPM2B_MAX_BUFFER, 0, 0},  {&_swigt__p_p_TPM2B_MAX_BUFFER, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_MAX_NV_BUFFER_ARRAY[] = {{&_swigt__p_p_TPM2B_MAX_NV_BUFFER_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_MAX_NV_BUFFER[] = {  {&_swigt__p_p_TPM2B_MAX_NV_BUFFER_ARRAY, _p_p_TPM2B_MAX_NV_BUFFER_ARRAYTo_p_p_TPM2B_MAX_NV_BUFFER, 0, 0},  {&_swigt__p_p_TPM2B_MAX_NV_BUFFER, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_NAME_ARRAY[] = {{&_swigt__p_p_TPM2B_NAME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_NAME[] = {  {&_swigt__p_p_TPM2B_NAME, 0, 0, 0},  {&_swigt__p_p_TPM2B_NAME_ARRAY, _p_p_TPM2B_NAME_ARRAYTo_p_p_TPM2B_NAME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_NV_PUBLIC_ARRAY[] = {{&_swigt__p_p_TPM2B_NV_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_NV_PUBLIC[] = {  {&_swigt__p_p_TPM2B_NV_PUBLIC, 0, 0, 0},  {&_swigt__p_p_TPM2B_NV_PUBLIC_ARRAY, _p_p_TPM2B_NV_PUBLIC_ARRAYTo_p_p_TPM2B_NV_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PRIVATE_ARRAY[] = {{&_swigt__p_p_TPM2B_PRIVATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PRIVATE[] = {  {&_swigt__p_p_TPM2B_PRIVATE, 0, 0, 0},  {&_swigt__p_p_TPM2B_PRIVATE_ARRAY, _p_p_TPM2B_PRIVATE_ARRAYTo_p_p_TPM2B_PRIVATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY[] = {{&_swigt__p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PRIVATE_KEY_RSA[] = {  {&_swigt__p_p_TPM2B_PRIVATE_KEY_RSA, 0, 0, 0},  {&_swigt__p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY, _p_p_TPM2B_PRIVATE_KEY_RSA_ARRAYTo_p_p_TPM2B_PRIVATE_KEY_RSA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY[] = {{&_swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC[] = {  {&_swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0, 0, 0},  {&_swigt__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY, _p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAYTo_p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PUBLIC_ARRAY[] = {{&_swigt__p_p_TPM2B_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PUBLIC[] = {  {&_swigt__p_p_TPM2B_PUBLIC, 0, 0, 0},  {&_swigt__p_p_TPM2B_PUBLIC_ARRAY, _p_p_TPM2B_PUBLIC_ARRAYTo_p_p_TPM2B_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY[] = {{&_swigt__p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_PUBLIC_KEY_RSA[] = {  {&_swigt__p_p_TPM2B_PUBLIC_KEY_RSA, 0, 0, 0},  {&_swigt__p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY, _p_p_TPM2B_PUBLIC_KEY_RSA_ARRAYTo_p_p_TPM2B_PUBLIC_KEY_RSA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SENSITIVE_ARRAY[] = {{&_swigt__p_p_TPM2B_SENSITIVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SENSITIVE[] = {  {&_swigt__p_p_TPM2B_SENSITIVE, 0, 0, 0},  {&_swigt__p_p_TPM2B_SENSITIVE_ARRAY, _p_p_TPM2B_SENSITIVE_ARRAYTo_p_p_TPM2B_SENSITIVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SENSITIVE_CREATE_ARRAY[] = {{&_swigt__p_p_TPM2B_SENSITIVE_CREATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SENSITIVE_CREATE[] = {  {&_swigt__p_p_TPM2B_SENSITIVE_CREATE, 0, 0, 0},  {&_swigt__p_p_TPM2B_SENSITIVE_CREATE_ARRAY, _p_p_TPM2B_SENSITIVE_CREATE_ARRAYTo_p_p_TPM2B_SENSITIVE_CREATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SENSITIVE_DATA_ARRAY[] = {{&_swigt__p_p_TPM2B_SENSITIVE_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SENSITIVE_DATA[] = {  {&_swigt__p_p_TPM2B_SENSITIVE_DATA, 0, 0, 0},  {&_swigt__p_p_TPM2B_SENSITIVE_DATA_ARRAY, _p_p_TPM2B_SENSITIVE_DATA_ARRAYTo_p_p_TPM2B_SENSITIVE_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SYM_KEY_ARRAY[] = {{&_swigt__p_p_TPM2B_SYM_KEY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_SYM_KEY[] = {  {&_swigt__p_p_TPM2B_SYM_KEY, 0, 0, 0},  {&_swigt__p_p_TPM2B_SYM_KEY_ARRAY, _p_p_TPM2B_SYM_KEY_ARRAYTo_p_p_TPM2B_SYM_KEY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_TEMPLATE_ARRAY[] = {{&_swigt__p_p_TPM2B_TEMPLATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPM2B_TEMPLATE[] = {  {&_swigt__p_p_TPM2B_TEMPLATE, 0, 0, 0},  {&_swigt__p_p_TPM2B_TEMPLATE_ARRAY, _p_p_TPM2B_TEMPLATE_ARRAYTo_p_p_TPM2B_TEMPLATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_AC_CAPABILITIES_ARRAY[] = {{&_swigt__p_p_TPML_AC_CAPABILITIES_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_AC_CAPABILITIES[] = {  {&_swigt__p_p_TPML_AC_CAPABILITIES, 0, 0, 0},  {&_swigt__p_p_TPML_AC_CAPABILITIES_ARRAY, _p_p_TPML_AC_CAPABILITIES_ARRAYTo_p_p_TPML_AC_CAPABILITIES, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_ALG_ARRAY[] = {{&_swigt__p_p_TPML_ALG_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_ALG[] = {  {&_swigt__p_p_TPML_ALG, 0, 0, 0},  {&_swigt__p_p_TPML_ALG_ARRAY, _p_p_TPML_ALG_ARRAYTo_p_p_TPML_ALG, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_ALG_PROPERTY_ARRAY[] = {{&_swigt__p_p_TPML_ALG_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_ALG_PROPERTY[] = {  {&_swigt__p_p_TPML_ALG_PROPERTY, 0, 0, 0},  {&_swigt__p_p_TPML_ALG_PROPERTY_ARRAY, _p_p_TPML_ALG_PROPERTY_ARRAYTo_p_p_TPML_ALG_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_CC_ARRAY[] = {{&_swigt__p_p_TPML_CC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_CC[] = {  {&_swigt__p_p_TPML_CC, 0, 0, 0},  {&_swigt__p_p_TPML_CC_ARRAY, _p_p_TPML_CC_ARRAYTo_p_p_TPML_CC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_CCA_ARRAY[] = {{&_swigt__p_p_TPML_CCA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_CCA[] = {  {&_swigt__p_p_TPML_CCA, 0, 0, 0},  {&_swigt__p_p_TPML_CCA_ARRAY, _p_p_TPML_CCA_ARRAYTo_p_p_TPML_CCA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_DIGEST_ARRAY[] = {{&_swigt__p_p_TPML_DIGEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_DIGEST[] = {  {&_swigt__p_p_TPML_DIGEST_ARRAY, _p_p_TPML_DIGEST_ARRAYTo_p_p_TPML_DIGEST, 0, 0},  {&_swigt__p_p_TPML_DIGEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_DIGEST_VALUES_ARRAY[] = {{&_swigt__p_p_TPML_DIGEST_VALUES_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_DIGEST_VALUES[] = {  {&_swigt__p_p_TPML_DIGEST_VALUES, 0, 0, 0},  {&_swigt__p_p_TPML_DIGEST_VALUES_ARRAY, _p_p_TPML_DIGEST_VALUES_ARRAYTo_p_p_TPML_DIGEST_VALUES, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_ECC_CURVE_ARRAY[] = {{&_swigt__p_p_TPML_ECC_CURVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_ECC_CURVE[] = {  {&_swigt__p_p_TPML_ECC_CURVE, 0, 0, 0},  {&_swigt__p_p_TPML_ECC_CURVE_ARRAY, _p_p_TPML_ECC_CURVE_ARRAYTo_p_p_TPML_ECC_CURVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_HANDLE_ARRAY[] = {{&_swigt__p_p_TPML_HANDLE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_HANDLE[] = {  {&_swigt__p_p_TPML_HANDLE, 0, 0, 0},  {&_swigt__p_p_TPML_HANDLE_ARRAY, _p_p_TPML_HANDLE_ARRAYTo_p_p_TPML_HANDLE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_INTEL_PTT_PROPERTY_ARRAY[] = {{&_swigt__p_p_TPML_INTEL_PTT_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_INTEL_PTT_PROPERTY[] = {  {&_swigt__p_p_TPML_INTEL_PTT_PROPERTY, 0, 0, 0},  {&_swigt__p_p_TPML_INTEL_PTT_PROPERTY_ARRAY, _p_p_TPML_INTEL_PTT_PROPERTY_ARRAYTo_p_p_TPML_INTEL_PTT_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_PCR_SELECTION_ARRAY[] = {{&_swigt__p_p_TPML_PCR_SELECTION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_PCR_SELECTION[] = {  {&_swigt__p_p_TPML_PCR_SELECTION_ARRAY, _p_p_TPML_PCR_SELECTION_ARRAYTo_p_p_TPML_PCR_SELECTION, 0, 0},  {&_swigt__p_p_TPML_PCR_SELECTION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY[] = {{&_swigt__p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_TAGGED_PCR_PROPERTY[] = {  {&_swigt__p_p_TPML_TAGGED_PCR_PROPERTY, 0, 0, 0},  {&_swigt__p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY, _p_p_TPML_TAGGED_PCR_PROPERTY_ARRAYTo_p_p_TPML_TAGGED_PCR_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY[] = {{&_swigt__p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPML_TAGGED_TPM_PROPERTY[] = {  {&_swigt__p_p_TPML_TAGGED_TPM_PROPERTY, 0, 0, 0},  {&_swigt__p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY, _p_p_TPML_TAGGED_TPM_PROPERTY_ARRAYTo_p_p_TPML_TAGGED_TPM_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_AC_OUTPUT_ARRAY[] = {{&_swigt__p_p_TPMS_AC_OUTPUT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_AC_OUTPUT[] = {  {&_swigt__p_p_TPMS_AC_OUTPUT_ARRAY, _p_p_TPMS_AC_OUTPUT_ARRAYTo_p_p_TPMS_AC_OUTPUT, 0, 0},  {&_swigt__p_p_TPMS_AC_OUTPUT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY[] = {{&_swigt__p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ALGORITHM_DESCRIPTION[] = {  {&_swigt__p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY, _p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAYTo_p_p_TPMS_ALGORITHM_DESCRIPTION, 0, 0},  {&_swigt__p_p_TPMS_ALGORITHM_DESCRIPTION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY[] = {{&_swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ALGORITHM_DETAIL_ECC[] = {  {&_swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC, 0, 0, 0},  {&_swigt__p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY, _p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAYTo_p_p_TPMS_ALGORITHM_DETAIL_ECC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ALG_PROPERTY_ARRAY[] = {{&_swigt__p_p_TPMS_ALG_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ALG_PROPERTY[] = {  {&_swigt__p_p_TPMS_ALG_PROPERTY, 0, 0, 0},  {&_swigt__p_p_TPMS_ALG_PROPERTY_ARRAY, _p_p_TPMS_ALG_PROPERTY_ARRAYTo_p_p_TPMS_ALG_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ASYM_PARMS_ARRAY[] = {{&_swigt__p_p_TPMS_ASYM_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ASYM_PARMS[] = {  {&_swigt__p_p_TPMS_ASYM_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMS_ASYM_PARMS_ARRAY, _p_p_TPMS_ASYM_PARMS_ARRAYTo_p_p_TPMS_ASYM_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ATTEST_ARRAY[] = {{&_swigt__p_p_TPMS_ATTEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ATTEST[] = {  {&_swigt__p_p_TPMS_ATTEST_ARRAY, _p_p_TPMS_ATTEST_ARRAYTo_p_p_TPMS_ATTEST, 0, 0},  {&_swigt__p_p_TPMS_ATTEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_AUTH_COMMAND_ARRAY[] = {{&_swigt__p_p_TPMS_AUTH_COMMAND_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_AUTH_COMMAND[] = {  {&_swigt__p_p_TPMS_AUTH_COMMAND_ARRAY, _p_p_TPMS_AUTH_COMMAND_ARRAYTo_p_p_TPMS_AUTH_COMMAND, 0, 0},  {&_swigt__p_p_TPMS_AUTH_COMMAND, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_AUTH_RESPONSE_ARRAY[] = {{&_swigt__p_p_TPMS_AUTH_RESPONSE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_AUTH_RESPONSE[] = {  {&_swigt__p_p_TPMS_AUTH_RESPONSE, 0, 0, 0},  {&_swigt__p_p_TPMS_AUTH_RESPONSE_ARRAY, _p_p_TPMS_AUTH_RESPONSE_ARRAYTo_p_p_TPMS_AUTH_RESPONSE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CAPABILITY_DATA_ARRAY[] = {{&_swigt__p_p_TPMS_CAPABILITY_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CAPABILITY_DATA[] = {  {&_swigt__p_p_TPMS_CAPABILITY_DATA, 0, 0, 0},  {&_swigt__p_p_TPMS_CAPABILITY_DATA_ARRAY, _p_p_TPMS_CAPABILITY_DATA_ARRAYTo_p_p_TPMS_CAPABILITY_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CERTIFY_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_CERTIFY_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CERTIFY_INFO[] = {  {&_swigt__p_p_TPMS_CERTIFY_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_CERTIFY_INFO_ARRAY, _p_p_TPMS_CERTIFY_INFO_ARRAYTo_p_p_TPMS_CERTIFY_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CLOCK_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_CLOCK_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CLOCK_INFO[] = {  {&_swigt__p_p_TPMS_CLOCK_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_CLOCK_INFO_ARRAY, _p_p_TPMS_CLOCK_INFO_ARRAYTo_p_p_TPMS_CLOCK_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_COMMAND_AUDIT_INFO[] = {  {&_swigt__p_p_TPMS_COMMAND_AUDIT_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY, _p_p_TPMS_COMMAND_AUDIT_INFO_ARRAYTo_p_p_TPMS_COMMAND_AUDIT_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CONTEXT_ARRAY[] = {{&_swigt__p_p_TPMS_CONTEXT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CONTEXT[] = {  {&_swigt__p_p_TPMS_CONTEXT, 0, 0, 0},  {&_swigt__p_p_TPMS_CONTEXT_ARRAY, _p_p_TPMS_CONTEXT_ARRAYTo_p_p_TPMS_CONTEXT, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CONTEXT_DATA_ARRAY[] = {{&_swigt__p_p_TPMS_CONTEXT_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CONTEXT_DATA[] = {  {&_swigt__p_p_TPMS_CONTEXT_DATA, 0, 0, 0},  {&_swigt__p_p_TPMS_CONTEXT_DATA_ARRAY, _p_p_TPMS_CONTEXT_DATA_ARRAYTo_p_p_TPMS_CONTEXT_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CREATION_DATA_ARRAY[] = {{&_swigt__p_p_TPMS_CREATION_DATA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CREATION_DATA[] = {  {&_swigt__p_p_TPMS_CREATION_DATA, 0, 0, 0},  {&_swigt__p_p_TPMS_CREATION_DATA_ARRAY, _p_p_TPMS_CREATION_DATA_ARRAYTo_p_p_TPMS_CREATION_DATA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CREATION_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_CREATION_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_CREATION_INFO[] = {  {&_swigt__p_p_TPMS_CREATION_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_CREATION_INFO_ARRAY, _p_p_TPMS_CREATION_INFO_ARRAYTo_p_p_TPMS_CREATION_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ECC_PARMS_ARRAY[] = {{&_swigt__p_p_TPMS_ECC_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ECC_PARMS[] = {  {&_swigt__p_p_TPMS_ECC_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMS_ECC_PARMS_ARRAY, _p_p_TPMS_ECC_PARMS_ARRAYTo_p_p_TPMS_ECC_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ECC_POINT_ARRAY[] = {{&_swigt__p_p_TPMS_ECC_POINT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ECC_POINT[] = {  {&_swigt__p_p_TPMS_ECC_POINT_ARRAY, _p_p_TPMS_ECC_POINT_ARRAYTo_p_p_TPMS_ECC_POINT, 0, 0},  {&_swigt__p_p_TPMS_ECC_POINT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_EMPTY_ARRAY[] = {{&_swigt__p_p_TPMS_EMPTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_EMPTY[] = {  {&_swigt__p_p_TPMS_EMPTY, 0, 0, 0},  {&_swigt__p_p_TPMS_EMPTY_ARRAY, _p_p_TPMS_EMPTY_ARRAYTo_p_p_TPMS_EMPTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ID_OBJECT_ARRAY[] = {{&_swigt__p_p_TPMS_ID_OBJECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_ID_OBJECT[] = {  {&_swigt__p_p_TPMS_ID_OBJECT_ARRAY, _p_p_TPMS_ID_OBJECT_ARRAYTo_p_p_TPMS_ID_OBJECT, 0, 0},  {&_swigt__p_p_TPMS_ID_OBJECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_KEYEDHASH_PARMS_ARRAY[] = {{&_swigt__p_p_TPMS_KEYEDHASH_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_KEYEDHASH_PARMS[] = {  {&_swigt__p_p_TPMS_KEYEDHASH_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMS_KEYEDHASH_PARMS_ARRAY, _p_p_TPMS_KEYEDHASH_PARMS_ARRAYTo_p_p_TPMS_KEYEDHASH_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_NV_CERTIFY_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_NV_CERTIFY_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_NV_CERTIFY_INFO[] = {  {&_swigt__p_p_TPMS_NV_CERTIFY_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_NV_CERTIFY_INFO_ARRAY, _p_p_TPMS_NV_CERTIFY_INFO_ARRAYTo_p_p_TPMS_NV_CERTIFY_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY[] = {{&_swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS[] = {  {&_swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0, 0, 0},  {&_swigt__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY, _p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAYTo_p_p_TPMS_NV_PIN_COUNTER_PARAMETERS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_NV_PUBLIC_ARRAY[] = {{&_swigt__p_p_TPMS_NV_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_NV_PUBLIC[] = {  {&_swigt__p_p_TPMS_NV_PUBLIC, 0, 0, 0},  {&_swigt__p_p_TPMS_NV_PUBLIC_ARRAY, _p_p_TPMS_NV_PUBLIC_ARRAYTo_p_p_TPMS_NV_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_PCR_SELECT_ARRAY[] = {{&_swigt__p_p_TPMS_PCR_SELECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_PCR_SELECT[] = {  {&_swigt__p_p_TPMS_PCR_SELECT_ARRAY, _p_p_TPMS_PCR_SELECT_ARRAYTo_p_p_TPMS_PCR_SELECT, 0, 0},  {&_swigt__p_p_TPMS_PCR_SELECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_PCR_SELECTION_ARRAY[] = {{&_swigt__p_p_TPMS_PCR_SELECTION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_PCR_SELECTION[] = {  {&_swigt__p_p_TPMS_PCR_SELECTION_ARRAY, _p_p_TPMS_PCR_SELECTION_ARRAYTo_p_p_TPMS_PCR_SELECTION, 0, 0},  {&_swigt__p_p_TPMS_PCR_SELECTION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_QUOTE_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_QUOTE_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_QUOTE_INFO[] = {  {&_swigt__p_p_TPMS_QUOTE_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_QUOTE_INFO_ARRAY, _p_p_TPMS_QUOTE_INFO_ARRAYTo_p_p_TPMS_QUOTE_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_RSA_PARMS_ARRAY[] = {{&_swigt__p_p_TPMS_RSA_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_RSA_PARMS[] = {  {&_swigt__p_p_TPMS_RSA_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMS_RSA_PARMS_ARRAY, _p_p_TPMS_RSA_PARMS_ARRAYTo_p_p_TPMS_RSA_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SCHEME_ECDAA_ARRAY[] = {{&_swigt__p_p_TPMS_SCHEME_ECDAA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SCHEME_ECDAA[] = {  {&_swigt__p_p_TPMS_SCHEME_ECDAA, 0, 0, 0},  {&_swigt__p_p_TPMS_SCHEME_ECDAA_ARRAY, _p_p_TPMS_SCHEME_ECDAA_ARRAYTo_p_p_TPMS_SCHEME_ECDAA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SCHEME_HASH_ARRAY[] = {{&_swigt__p_p_TPMS_SCHEME_HASH_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SCHEME_HASH[] = {  {&_swigt__p_p_TPMS_SCHEME_HASH_ARRAY, _p_p_TPMS_SCHEME_HASH_ARRAYTo_p_p_TPMS_SCHEME_HASH, 0, 0},  {&_swigt__p_p_TPMS_SCHEME_HASH, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SCHEME_XOR_ARRAY[] = {{&_swigt__p_p_TPMS_SCHEME_XOR_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SCHEME_XOR[] = {  {&_swigt__p_p_TPMS_SCHEME_XOR_ARRAY, _p_p_TPMS_SCHEME_XOR_ARRAYTo_p_p_TPMS_SCHEME_XOR, 0, 0},  {&_swigt__p_p_TPMS_SCHEME_XOR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SENSITIVE_CREATE_ARRAY[] = {{&_swigt__p_p_TPMS_SENSITIVE_CREATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SENSITIVE_CREATE[] = {  {&_swigt__p_p_TPMS_SENSITIVE_CREATE, 0, 0, 0},  {&_swigt__p_p_TPMS_SENSITIVE_CREATE_ARRAY, _p_p_TPMS_SENSITIVE_CREATE_ARRAYTo_p_p_TPMS_SENSITIVE_CREATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SESSION_AUDIT_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_SESSION_AUDIT_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SESSION_AUDIT_INFO[] = {  {&_swigt__p_p_TPMS_SESSION_AUDIT_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_SESSION_AUDIT_INFO_ARRAY, _p_p_TPMS_SESSION_AUDIT_INFO_ARRAYTo_p_p_TPMS_SESSION_AUDIT_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SIGNATURE_ECC_ARRAY[] = {{&_swigt__p_p_TPMS_SIGNATURE_ECC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SIGNATURE_ECC[] = {  {&_swigt__p_p_TPMS_SIGNATURE_ECC, 0, 0, 0},  {&_swigt__p_p_TPMS_SIGNATURE_ECC_ARRAY, _p_p_TPMS_SIGNATURE_ECC_ARRAYTo_p_p_TPMS_SIGNATURE_ECC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SIGNATURE_RSA_ARRAY[] = {{&_swigt__p_p_TPMS_SIGNATURE_RSA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SIGNATURE_RSA[] = {  {&_swigt__p_p_TPMS_SIGNATURE_RSA, 0, 0, 0},  {&_swigt__p_p_TPMS_SIGNATURE_RSA_ARRAY, _p_p_TPMS_SIGNATURE_RSA_ARRAYTo_p_p_TPMS_SIGNATURE_RSA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SYMCIPHER_PARMS_ARRAY[] = {{&_swigt__p_p_TPMS_SYMCIPHER_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_SYMCIPHER_PARMS[] = {  {&_swigt__p_p_TPMS_SYMCIPHER_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMS_SYMCIPHER_PARMS_ARRAY, _p_p_TPMS_SYMCIPHER_PARMS_ARRAYTo_p_p_TPMS_SYMCIPHER_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TAGGED_PCR_SELECT_ARRAY[] = {{&_swigt__p_p_TPMS_TAGGED_PCR_SELECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TAGGED_PCR_SELECT[] = {  {&_swigt__p_p_TPMS_TAGGED_PCR_SELECT_ARRAY, _p_p_TPMS_TAGGED_PCR_SELECT_ARRAYTo_p_p_TPMS_TAGGED_PCR_SELECT, 0, 0},  {&_swigt__p_p_TPMS_TAGGED_PCR_SELECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TAGGED_POLICY_ARRAY[] = {{&_swigt__p_p_TPMS_TAGGED_POLICY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TAGGED_POLICY[] = {  {&_swigt__p_p_TPMS_TAGGED_POLICY, 0, 0, 0},  {&_swigt__p_p_TPMS_TAGGED_POLICY_ARRAY, _p_p_TPMS_TAGGED_POLICY_ARRAYTo_p_p_TPMS_TAGGED_POLICY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TAGGED_PROPERTY_ARRAY[] = {{&_swigt__p_p_TPMS_TAGGED_PROPERTY_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TAGGED_PROPERTY[] = {  {&_swigt__p_p_TPMS_TAGGED_PROPERTY, 0, 0, 0},  {&_swigt__p_p_TPMS_TAGGED_PROPERTY_ARRAY, _p_p_TPMS_TAGGED_PROPERTY_ARRAYTo_p_p_TPMS_TAGGED_PROPERTY, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TIME_ATTEST_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_TIME_ATTEST_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TIME_ATTEST_INFO[] = {  {&_swigt__p_p_TPMS_TIME_ATTEST_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_TIME_ATTEST_INFO_ARRAY, _p_p_TPMS_TIME_ATTEST_INFO_ARRAYTo_p_p_TPMS_TIME_ATTEST_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TIME_INFO_ARRAY[] = {{&_swigt__p_p_TPMS_TIME_INFO_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMS_TIME_INFO[] = {  {&_swigt__p_p_TPMS_TIME_INFO, 0, 0, 0},  {&_swigt__p_p_TPMS_TIME_INFO_ARRAY, _p_p_TPMS_TIME_INFO_ARRAYTo_p_p_TPMS_TIME_INFO, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_ASYM_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMT_ASYM_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_ASYM_SCHEME[] = {  {&_swigt__p_p_TPMT_ASYM_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMT_ASYM_SCHEME_ARRAY, _p_p_TPMT_ASYM_SCHEME_ARRAYTo_p_p_TPMT_ASYM_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_ECC_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMT_ECC_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_ECC_SCHEME[] = {  {&_swigt__p_p_TPMT_ECC_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMT_ECC_SCHEME_ARRAY, _p_p_TPMT_ECC_SCHEME_ARRAYTo_p_p_TPMT_ECC_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_HA_ARRAY[] = {{&_swigt__p_p_TPMT_HA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_HA[] = {  {&_swigt__p_p_TPMT_HA, 0, 0, 0},  {&_swigt__p_p_TPMT_HA_ARRAY, _p_p_TPMT_HA_ARRAYTo_p_p_TPMT_HA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_KDF_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMT_KDF_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_KDF_SCHEME[] = {  {&_swigt__p_p_TPMT_KDF_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMT_KDF_SCHEME_ARRAY, _p_p_TPMT_KDF_SCHEME_ARRAYTo_p_p_TPMT_KDF_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_KEYEDHASH_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMT_KEYEDHASH_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_KEYEDHASH_SCHEME[] = {  {&_swigt__p_p_TPMT_KEYEDHASH_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMT_KEYEDHASH_SCHEME_ARRAY, _p_p_TPMT_KEYEDHASH_SCHEME_ARRAYTo_p_p_TPMT_KEYEDHASH_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_PUBLIC_ARRAY[] = {{&_swigt__p_p_TPMT_PUBLIC_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_PUBLIC[] = {  {&_swigt__p_p_TPMT_PUBLIC, 0, 0, 0},  {&_swigt__p_p_TPMT_PUBLIC_ARRAY, _p_p_TPMT_PUBLIC_ARRAYTo_p_p_TPMT_PUBLIC, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_PUBLIC_PARMS_ARRAY[] = {{&_swigt__p_p_TPMT_PUBLIC_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_PUBLIC_PARMS[] = {  {&_swigt__p_p_TPMT_PUBLIC_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMT_PUBLIC_PARMS_ARRAY, _p_p_TPMT_PUBLIC_PARMS_ARRAYTo_p_p_TPMT_PUBLIC_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_RSA_DECRYPT_ARRAY[] = {{&_swigt__p_p_TPMT_RSA_DECRYPT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_RSA_DECRYPT[] = {  {&_swigt__p_p_TPMT_RSA_DECRYPT_ARRAY, _p_p_TPMT_RSA_DECRYPT_ARRAYTo_p_p_TPMT_RSA_DECRYPT, 0, 0},  {&_swigt__p_p_TPMT_RSA_DECRYPT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_RSA_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMT_RSA_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_RSA_SCHEME[] = {  {&_swigt__p_p_TPMT_RSA_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMT_RSA_SCHEME_ARRAY, _p_p_TPMT_RSA_SCHEME_ARRAYTo_p_p_TPMT_RSA_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SENSITIVE_ARRAY[] = {{&_swigt__p_p_TPMT_SENSITIVE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SENSITIVE[] = {  {&_swigt__p_p_TPMT_SENSITIVE, 0, 0, 0},  {&_swigt__p_p_TPMT_SENSITIVE_ARRAY, _p_p_TPMT_SENSITIVE_ARRAYTo_p_p_TPMT_SENSITIVE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SIGNATURE_ARRAY[] = {{&_swigt__p_p_TPMT_SIGNATURE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SIGNATURE[] = {  {&_swigt__p_p_TPMT_SIGNATURE, 0, 0, 0},  {&_swigt__p_p_TPMT_SIGNATURE_ARRAY, _p_p_TPMT_SIGNATURE_ARRAYTo_p_p_TPMT_SIGNATURE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SIG_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMT_SIG_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SIG_SCHEME[] = {  {&_swigt__p_p_TPMT_SIG_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMT_SIG_SCHEME_ARRAY, _p_p_TPMT_SIG_SCHEME_ARRAYTo_p_p_TPMT_SIG_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SYM_DEF_ARRAY[] = {{&_swigt__p_p_TPMT_SYM_DEF_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SYM_DEF[] = {  {&_swigt__p_p_TPMT_SYM_DEF_ARRAY, _p_p_TPMT_SYM_DEF_ARRAYTo_p_p_TPMT_SYM_DEF, 0, 0},  {&_swigt__p_p_TPMT_SYM_DEF, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SYM_DEF_OBJECT_ARRAY[] = {{&_swigt__p_p_TPMT_SYM_DEF_OBJECT_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_SYM_DEF_OBJECT[] = {  {&_swigt__p_p_TPMT_SYM_DEF_OBJECT_ARRAY, _p_p_TPMT_SYM_DEF_OBJECT_ARRAYTo_p_p_TPMT_SYM_DEF_OBJECT, 0, 0},  {&_swigt__p_p_TPMT_SYM_DEF_OBJECT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_AUTH_ARRAY[] = {{&_swigt__p_p_TPMT_TK_AUTH_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_AUTH[] = {  {&_swigt__p_p_TPMT_TK_AUTH_ARRAY, _p_p_TPMT_TK_AUTH_ARRAYTo_p_p_TPMT_TK_AUTH, 0, 0},  {&_swigt__p_p_TPMT_TK_AUTH, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_CREATION_ARRAY[] = {{&_swigt__p_p_TPMT_TK_CREATION_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_CREATION[] = {  {&_swigt__p_p_TPMT_TK_CREATION_ARRAY, _p_p_TPMT_TK_CREATION_ARRAYTo_p_p_TPMT_TK_CREATION, 0, 0},  {&_swigt__p_p_TPMT_TK_CREATION, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_HASHCHECK_ARRAY[] = {{&_swigt__p_p_TPMT_TK_HASHCHECK_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_HASHCHECK[] = {  {&_swigt__p_p_TPMT_TK_HASHCHECK, 0, 0, 0},  {&_swigt__p_p_TPMT_TK_HASHCHECK_ARRAY, _p_p_TPMT_TK_HASHCHECK_ARRAYTo_p_p_TPMT_TK_HASHCHECK, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_VERIFIED_ARRAY[] = {{&_swigt__p_p_TPMT_TK_VERIFIED_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMT_TK_VERIFIED[] = {  {&_swigt__p_p_TPMT_TK_VERIFIED, 0, 0, 0},  {&_swigt__p_p_TPMT_TK_VERIFIED_ARRAY, _p_p_TPMT_TK_VERIFIED_ARRAYTo_p_p_TPMT_TK_VERIFIED, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_ASYM_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMU_ASYM_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_ASYM_SCHEME[] = {  {&_swigt__p_p_TPMU_ASYM_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMU_ASYM_SCHEME_ARRAY, _p_p_TPMU_ASYM_SCHEME_ARRAYTo_p_p_TPMU_ASYM_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_ATTEST_ARRAY[] = {{&_swigt__p_p_TPMU_ATTEST_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_ATTEST[] = {  {&_swigt__p_p_TPMU_ATTEST_ARRAY, _p_p_TPMU_ATTEST_ARRAYTo_p_p_TPMU_ATTEST, 0, 0},  {&_swigt__p_p_TPMU_ATTEST, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_CAPABILITIES_ARRAY[] = {{&_swigt__p_p_TPMU_CAPABILITIES_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_CAPABILITIES[] = {  {&_swigt__p_p_TPMU_CAPABILITIES, 0, 0, 0},  {&_swigt__p_p_TPMU_CAPABILITIES_ARRAY, _p_p_TPMU_CAPABILITIES_ARRAYTo_p_p_TPMU_CAPABILITIES, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_ENCRYPTED_SECRET_ARRAY[] = {{&_swigt__p_p_TPMU_ENCRYPTED_SECRET_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_ENCRYPTED_SECRET[] = {  {&_swigt__p_p_TPMU_ENCRYPTED_SECRET_ARRAY, _p_p_TPMU_ENCRYPTED_SECRET_ARRAYTo_p_p_TPMU_ENCRYPTED_SECRET, 0, 0},  {&_swigt__p_p_TPMU_ENCRYPTED_SECRET, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_HA_ARRAY[] = {{&_swigt__p_p_TPMU_HA_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_HA[] = {  {&_swigt__p_p_TPMU_HA, 0, 0, 0},  {&_swigt__p_p_TPMU_HA_ARRAY, _p_p_TPMU_HA_ARRAYTo_p_p_TPMU_HA, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_KDF_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMU_KDF_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_KDF_SCHEME[] = {  {&_swigt__p_p_TPMU_KDF_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMU_KDF_SCHEME_ARRAY, _p_p_TPMU_KDF_SCHEME_ARRAYTo_p_p_TPMU_KDF_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_NAME_ARRAY[] = {{&_swigt__p_p_TPMU_NAME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_NAME[] = {  {&_swigt__p_p_TPMU_NAME, 0, 0, 0},  {&_swigt__p_p_TPMU_NAME_ARRAY, _p_p_TPMU_NAME_ARRAYTo_p_p_TPMU_NAME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_PUBLIC_ID_ARRAY[] = {{&_swigt__p_p_TPMU_PUBLIC_ID_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_PUBLIC_ID[] = {  {&_swigt__p_p_TPMU_PUBLIC_ID_ARRAY, _p_p_TPMU_PUBLIC_ID_ARRAYTo_p_p_TPMU_PUBLIC_ID, 0, 0},  {&_swigt__p_p_TPMU_PUBLIC_ID, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_PUBLIC_PARMS_ARRAY[] = {{&_swigt__p_p_TPMU_PUBLIC_PARMS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_PUBLIC_PARMS[] = {  {&_swigt__p_p_TPMU_PUBLIC_PARMS, 0, 0, 0},  {&_swigt__p_p_TPMU_PUBLIC_PARMS_ARRAY, _p_p_TPMU_PUBLIC_PARMS_ARRAYTo_p_p_TPMU_PUBLIC_PARMS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SCHEME_KEYEDHASH_ARRAY[] = {{&_swigt__p_p_TPMU_SCHEME_KEYEDHASH_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SCHEME_KEYEDHASH[] = {  {&_swigt__p_p_TPMU_SCHEME_KEYEDHASH, 0, 0, 0},  {&_swigt__p_p_TPMU_SCHEME_KEYEDHASH_ARRAY, _p_p_TPMU_SCHEME_KEYEDHASH_ARRAYTo_p_p_TPMU_SCHEME_KEYEDHASH, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY[] = {{&_swigt__p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SENSITIVE_COMPOSITE[] = {  {&_swigt__p_p_TPMU_SENSITIVE_COMPOSITE, 0, 0, 0},  {&_swigt__p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY, _p_p_TPMU_SENSITIVE_COMPOSITE_ARRAYTo_p_p_TPMU_SENSITIVE_COMPOSITE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SIGNATURE_ARRAY[] = {{&_swigt__p_p_TPMU_SIGNATURE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SIGNATURE[] = {  {&_swigt__p_p_TPMU_SIGNATURE, 0, 0, 0},  {&_swigt__p_p_TPMU_SIGNATURE_ARRAY, _p_p_TPMU_SIGNATURE_ARRAYTo_p_p_TPMU_SIGNATURE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SIG_SCHEME_ARRAY[] = {{&_swigt__p_p_TPMU_SIG_SCHEME_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SIG_SCHEME[] = {  {&_swigt__p_p_TPMU_SIG_SCHEME, 0, 0, 0},  {&_swigt__p_p_TPMU_SIG_SCHEME_ARRAY, _p_p_TPMU_SIG_SCHEME_ARRAYTo_p_p_TPMU_SIG_SCHEME, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SYM_KEY_BITS_ARRAY[] = {{&_swigt__p_p_TPMU_SYM_KEY_BITS_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SYM_KEY_BITS[] = {  {&_swigt__p_p_TPMU_SYM_KEY_BITS, 0, 0, 0},  {&_swigt__p_p_TPMU_SYM_KEY_BITS_ARRAY, _p_p_TPMU_SYM_KEY_BITS_ARRAYTo_p_p_TPMU_SYM_KEY_BITS, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SYM_MODE_ARRAY[] = {{&_swigt__p_p_TPMU_SYM_MODE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TPMU_SYM_MODE[] = {  {&_swigt__p_p_TPMU_SYM_MODE, 0, 0, 0},  {&_swigt__p_p_TPMU_SYM_MODE_ARRAY, _p_p_TPMU_SYM_MODE_ARRAYTo_p_p_TPMU_SYM_MODE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TSS2_TCTI_CONTEXT[] = {  {&_swigt__p_p_TSS2_TCTI_CONTEXT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_TSS2_TCTI_POLL_HANDLE[] = {  {&_swigt__p_p_TSS2_TCTI_POLL_HANDLE, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p__PRIVATE_ARRAY[] = {{&_swigt__p_p__PRIVATE_ARRAY, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p__PRIVATE[] = {  {&_swigt__p_p__PRIVATE, 0, 0, 0},  {&_swigt__p_p__PRIVATE_ARRAY, _p_p__PRIVATE_ARRAYTo_p_p__PRIVATE, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_size_t[] = {  {&_swigt__p_p_size_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_unsigned_char[] = {  {&_swigt__p_p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_ESYS_TR_PTR[] = {{&_swigt__p_p_ESYS_TR_PTR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_unsigned_int[] = {  {&_swigt__p_p_unsigned_int, 0, 0, 0},  {&_swigt__p_p_ESYS_TR_PTR, _p_p_ESYS_TR_PTRTo_p_p_unsigned_int, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_short[] = {  {&_swigt__p_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_signed_char[] = {  {&_swigt__p_signed_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_size_t[] = {  {&_swigt__p_size_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_uint16_t_PTR[] = {  {&_swigt__p_uint16_t_PTR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_uint32_t_PTR[] = {  {&_swigt__p_uint32_t_PTR, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},  {&_swigt__p_ESYS_TR_PTR, _p_ESYS_TR_PTRTo_p_unsigned_int, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long_long[] = {  {&_swigt__p_unsigned_long_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_short[] = {  {&_swigt__p_unsigned_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_ByteArray,
  _swigc__p_ESYS_CONTEXT,
  _swigc__p_ESYS_TR_PTR,
  _swigc__p_TPM2B_ATTEST,
  _swigc__p_TPM2B_ATTEST_ARRAY,
  _swigc__p_TPM2B_CONTEXT_DATA,
  _swigc__p_TPM2B_CONTEXT_DATA_ARRAY,
  _swigc__p_TPM2B_CONTEXT_SENSITIVE,
  _swigc__p_TPM2B_CONTEXT_SENSITIVE_ARRAY,
  _swigc__p_TPM2B_CREATION_DATA,
  _swigc__p_TPM2B_CREATION_DATA_ARRAY,
  _swigc__p_TPM2B_DATA,
  _swigc__p_TPM2B_DATA_ARRAY,
  _swigc__p_TPM2B_DIGEST,
  _swigc__p_TPM2B_DIGEST_ARRAY,
  _swigc__p_TPM2B_ECC_PARAMETER,
  _swigc__p_TPM2B_ECC_PARAMETER_ARRAY,
  _swigc__p_TPM2B_ECC_POINT,
  _swigc__p_TPM2B_ECC_POINT_ARRAY,
  _swigc__p_TPM2B_ENCRYPTED_SECRET,
  _swigc__p_TPM2B_ENCRYPTED_SECRET_ARRAY,
  _swigc__p_TPM2B_EVENT,
  _swigc__p_TPM2B_EVENT_ARRAY,
  _swigc__p_TPM2B_ID_OBJECT,
  _swigc__p_TPM2B_ID_OBJECT_ARRAY,
  _swigc__p_TPM2B_IV,
  _swigc__p_TPM2B_IV_ARRAY,
  _swigc__p_TPM2B_MAX_BUFFER,
  _swigc__p_TPM2B_MAX_BUFFER_ARRAY,
  _swigc__p_TPM2B_MAX_NV_BUFFER,
  _swigc__p_TPM2B_MAX_NV_BUFFER_ARRAY,
  _swigc__p_TPM2B_NAME,
  _swigc__p_TPM2B_NAME_ARRAY,
  _swigc__p_TPM2B_NV_PUBLIC,
  _swigc__p_TPM2B_NV_PUBLIC_ARRAY,
  _swigc__p_TPM2B_PRIVATE,
  _swigc__p_TPM2B_PRIVATE_ARRAY,
  _swigc__p_TPM2B_PRIVATE_KEY_RSA,
  _swigc__p_TPM2B_PRIVATE_KEY_RSA_ARRAY,
  _swigc__p_TPM2B_PRIVATE_VENDOR_SPECIFIC,
  _swigc__p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY,
  _swigc__p_TPM2B_PUBLIC,
  _swigc__p_TPM2B_PUBLIC_ARRAY,
  _swigc__p_TPM2B_PUBLIC_KEY_RSA,
  _swigc__p_TPM2B_PUBLIC_KEY_RSA_ARRAY,
  _swigc__p_TPM2B_SENSITIVE,
  _swigc__p_TPM2B_SENSITIVE_ARRAY,
  _swigc__p_TPM2B_SENSITIVE_CREATE,
  _swigc__p_TPM2B_SENSITIVE_CREATE_ARRAY,
  _swigc__p_TPM2B_SENSITIVE_DATA,
  _swigc__p_TPM2B_SENSITIVE_DATA_ARRAY,
  _swigc__p_TPM2B_SYM_KEY,
  _swigc__p_TPM2B_SYM_KEY_ARRAY,
  _swigc__p_TPM2B_TEMPLATE,
  _swigc__p_TPM2B_TEMPLATE_ARRAY,
  _swigc__p_TPML_AC_CAPABILITIES,
  _swigc__p_TPML_AC_CAPABILITIES_ARRAY,
  _swigc__p_TPML_ALG,
  _swigc__p_TPML_ALG_ARRAY,
  _swigc__p_TPML_ALG_PROPERTY,
  _swigc__p_TPML_ALG_PROPERTY_ARRAY,
  _swigc__p_TPML_CC,
  _swigc__p_TPML_CCA,
  _swigc__p_TPML_CCA_ARRAY,
  _swigc__p_TPML_CC_ARRAY,
  _swigc__p_TPML_DIGEST,
  _swigc__p_TPML_DIGEST_ARRAY,
  _swigc__p_TPML_DIGEST_VALUES,
  _swigc__p_TPML_DIGEST_VALUES_ARRAY,
  _swigc__p_TPML_ECC_CURVE,
  _swigc__p_TPML_ECC_CURVE_ARRAY,
  _swigc__p_TPML_HANDLE,
  _swigc__p_TPML_HANDLE_ARRAY,
  _swigc__p_TPML_INTEL_PTT_PROPERTY,
  _swigc__p_TPML_INTEL_PTT_PROPERTY_ARRAY,
  _swigc__p_TPML_PCR_SELECTION,
  _swigc__p_TPML_PCR_SELECTION_ARRAY,
  _swigc__p_TPML_TAGGED_PCR_PROPERTY,
  _swigc__p_TPML_TAGGED_PCR_PROPERTY_ARRAY,
  _swigc__p_TPML_TAGGED_TPM_PROPERTY,
  _swigc__p_TPML_TAGGED_TPM_PROPERTY_ARRAY,
  _swigc__p_TPMS_AC_OUTPUT,
  _swigc__p_TPMS_AC_OUTPUT_ARRAY,
  _swigc__p_TPMS_ALGORITHM_DESCRIPTION,
  _swigc__p_TPMS_ALGORITHM_DESCRIPTION_ARRAY,
  _swigc__p_TPMS_ALGORITHM_DETAIL_ECC,
  _swigc__p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY,
  _swigc__p_TPMS_ALG_PROPERTY,
  _swigc__p_TPMS_ALG_PROPERTY_ARRAY,
  _swigc__p_TPMS_ASYM_PARMS,
  _swigc__p_TPMS_ASYM_PARMS_ARRAY,
  _swigc__p_TPMS_ATTEST,
  _swigc__p_TPMS_ATTEST_ARRAY,
  _swigc__p_TPMS_AUTH_COMMAND,
  _swigc__p_TPMS_AUTH_COMMAND_ARRAY,
  _swigc__p_TPMS_AUTH_RESPONSE,
  _swigc__p_TPMS_AUTH_RESPONSE_ARRAY,
  _swigc__p_TPMS_CAPABILITY_DATA,
  _swigc__p_TPMS_CAPABILITY_DATA_ARRAY,
  _swigc__p_TPMS_CERTIFY_INFO,
  _swigc__p_TPMS_CERTIFY_INFO_ARRAY,
  _swigc__p_TPMS_CLOCK_INFO,
  _swigc__p_TPMS_CLOCK_INFO_ARRAY,
  _swigc__p_TPMS_COMMAND_AUDIT_INFO,
  _swigc__p_TPMS_COMMAND_AUDIT_INFO_ARRAY,
  _swigc__p_TPMS_CONTEXT,
  _swigc__p_TPMS_CONTEXT_ARRAY,
  _swigc__p_TPMS_CONTEXT_DATA,
  _swigc__p_TPMS_CONTEXT_DATA_ARRAY,
  _swigc__p_TPMS_CREATION_DATA,
  _swigc__p_TPMS_CREATION_DATA_ARRAY,
  _swigc__p_TPMS_CREATION_INFO,
  _swigc__p_TPMS_CREATION_INFO_ARRAY,
  _swigc__p_TPMS_ECC_PARMS,
  _swigc__p_TPMS_ECC_PARMS_ARRAY,
  _swigc__p_TPMS_ECC_POINT,
  _swigc__p_TPMS_ECC_POINT_ARRAY,
  _swigc__p_TPMS_EMPTY,
  _swigc__p_TPMS_EMPTY_ARRAY,
  _swigc__p_TPMS_ID_OBJECT,
  _swigc__p_TPMS_ID_OBJECT_ARRAY,
  _swigc__p_TPMS_KEYEDHASH_PARMS,
  _swigc__p_TPMS_KEYEDHASH_PARMS_ARRAY,
  _swigc__p_TPMS_NV_CERTIFY_INFO,
  _swigc__p_TPMS_NV_CERTIFY_INFO_ARRAY,
  _swigc__p_TPMS_NV_PIN_COUNTER_PARAMETERS,
  _swigc__p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY,
  _swigc__p_TPMS_NV_PUBLIC,
  _swigc__p_TPMS_NV_PUBLIC_ARRAY,
  _swigc__p_TPMS_PCR_SELECT,
  _swigc__p_TPMS_PCR_SELECTION,
  _swigc__p_TPMS_PCR_SELECTION_ARRAY,
  _swigc__p_TPMS_PCR_SELECT_ARRAY,
  _swigc__p_TPMS_QUOTE_INFO,
  _swigc__p_TPMS_QUOTE_INFO_ARRAY,
  _swigc__p_TPMS_RSA_PARMS,
  _swigc__p_TPMS_RSA_PARMS_ARRAY,
  _swigc__p_TPMS_SCHEME_ECDAA,
  _swigc__p_TPMS_SCHEME_ECDAA_ARRAY,
  _swigc__p_TPMS_SCHEME_HASH,
  _swigc__p_TPMS_SCHEME_HASH_ARRAY,
  _swigc__p_TPMS_SCHEME_XOR,
  _swigc__p_TPMS_SCHEME_XOR_ARRAY,
  _swigc__p_TPMS_SENSITIVE_CREATE,
  _swigc__p_TPMS_SENSITIVE_CREATE_ARRAY,
  _swigc__p_TPMS_SESSION_AUDIT_INFO,
  _swigc__p_TPMS_SESSION_AUDIT_INFO_ARRAY,
  _swigc__p_TPMS_SIGNATURE_ECC,
  _swigc__p_TPMS_SIGNATURE_ECC_ARRAY,
  _swigc__p_TPMS_SIGNATURE_RSA,
  _swigc__p_TPMS_SIGNATURE_RSA_ARRAY,
  _swigc__p_TPMS_SYMCIPHER_PARMS,
  _swigc__p_TPMS_SYMCIPHER_PARMS_ARRAY,
  _swigc__p_TPMS_TAGGED_PCR_SELECT,
  _swigc__p_TPMS_TAGGED_PCR_SELECT_ARRAY,
  _swigc__p_TPMS_TAGGED_POLICY,
  _swigc__p_TPMS_TAGGED_POLICY_ARRAY,
  _swigc__p_TPMS_TAGGED_PROPERTY,
  _swigc__p_TPMS_TAGGED_PROPERTY_ARRAY,
  _swigc__p_TPMS_TIME_ATTEST_INFO,
  _swigc__p_TPMS_TIME_ATTEST_INFO_ARRAY,
  _swigc__p_TPMS_TIME_INFO,
  _swigc__p_TPMS_TIME_INFO_ARRAY,
  _swigc__p_TPMT_ASYM_SCHEME,
  _swigc__p_TPMT_ASYM_SCHEME_ARRAY,
  _swigc__p_TPMT_ECC_SCHEME,
  _swigc__p_TPMT_ECC_SCHEME_ARRAY,
  _swigc__p_TPMT_HA,
  _swigc__p_TPMT_HA_ARRAY,
  _swigc__p_TPMT_KDF_SCHEME,
  _swigc__p_TPMT_KDF_SCHEME_ARRAY,
  _swigc__p_TPMT_KEYEDHASH_SCHEME,
  _swigc__p_TPMT_KEYEDHASH_SCHEME_ARRAY,
  _swigc__p_TPMT_PUBLIC,
  _swigc__p_TPMT_PUBLIC_ARRAY,
  _swigc__p_TPMT_PUBLIC_PARMS,
  _swigc__p_TPMT_PUBLIC_PARMS_ARRAY,
  _swigc__p_TPMT_RSA_DECRYPT,
  _swigc__p_TPMT_RSA_DECRYPT_ARRAY,
  _swigc__p_TPMT_RSA_SCHEME,
  _swigc__p_TPMT_RSA_SCHEME_ARRAY,
  _swigc__p_TPMT_SENSITIVE,
  _swigc__p_TPMT_SENSITIVE_ARRAY,
  _swigc__p_TPMT_SIGNATURE,
  _swigc__p_TPMT_SIGNATURE_ARRAY,
  _swigc__p_TPMT_SIG_SCHEME,
  _swigc__p_TPMT_SIG_SCHEME_ARRAY,
  _swigc__p_TPMT_SYM_DEF,
  _swigc__p_TPMT_SYM_DEF_ARRAY,
  _swigc__p_TPMT_SYM_DEF_OBJECT,
  _swigc__p_TPMT_SYM_DEF_OBJECT_ARRAY,
  _swigc__p_TPMT_TK_AUTH,
  _swigc__p_TPMT_TK_AUTH_ARRAY,
  _swigc__p_TPMT_TK_CREATION,
  _swigc__p_TPMT_TK_CREATION_ARRAY,
  _swigc__p_TPMT_TK_HASHCHECK,
  _swigc__p_TPMT_TK_HASHCHECK_ARRAY,
  _swigc__p_TPMT_TK_VERIFIED,
  _swigc__p_TPMT_TK_VERIFIED_ARRAY,
  _swigc__p_TPMU_ASYM_SCHEME,
  _swigc__p_TPMU_ASYM_SCHEME_ARRAY,
  _swigc__p_TPMU_ATTEST,
  _swigc__p_TPMU_ATTEST_ARRAY,
  _swigc__p_TPMU_CAPABILITIES,
  _swigc__p_TPMU_CAPABILITIES_ARRAY,
  _swigc__p_TPMU_ENCRYPTED_SECRET,
  _swigc__p_TPMU_ENCRYPTED_SECRET_ARRAY,
  _swigc__p_TPMU_HA,
  _swigc__p_TPMU_HA_ARRAY,
  _swigc__p_TPMU_KDF_SCHEME,
  _swigc__p_TPMU_KDF_SCHEME_ARRAY,
  _swigc__p_TPMU_NAME,
  _swigc__p_TPMU_NAME_ARRAY,
  _swigc__p_TPMU_PUBLIC_ID,
  _swigc__p_TPMU_PUBLIC_ID_ARRAY,
  _swigc__p_TPMU_PUBLIC_PARMS,
  _swigc__p_TPMU_PUBLIC_PARMS_ARRAY,
  _swigc__p_TPMU_SCHEME_KEYEDHASH,
  _swigc__p_TPMU_SCHEME_KEYEDHASH_ARRAY,
  _swigc__p_TPMU_SENSITIVE_COMPOSITE,
  _swigc__p_TPMU_SENSITIVE_COMPOSITE_ARRAY,
  _swigc__p_TPMU_SIGNATURE,
  _swigc__p_TPMU_SIGNATURE_ARRAY,
  _swigc__p_TPMU_SIG_SCHEME,
  _swigc__p_TPMU_SIG_SCHEME_ARRAY,
  _swigc__p_TPMU_SYM_KEY_BITS,
  _swigc__p_TPMU_SYM_KEY_BITS_ARRAY,
  _swigc__p_TPMU_SYM_MODE,
  _swigc__p_TPMU_SYM_MODE_ARRAY,
  _swigc__p_TSS2_ABI_VERSION,
  _swigc__p_TSS2_TCTI_CONTEXT,
  _swigc__p_UINT16_PTR,
  _swigc__p_UINT32_PTR,
  _swigc__p__PRIVATE,
  _swigc__p__PRIVATE_ARRAY,
  _swigc__p_char,
  _swigc__p_int,
  _swigc__p_long_long,
  _swigc__p_p_ESYS_CONTEXT,
  _swigc__p_p_ESYS_TR_PTR,
  _swigc__p_p_TPM2B_ATTEST,
  _swigc__p_p_TPM2B_ATTEST_ARRAY,
  _swigc__p_p_TPM2B_CONTEXT_DATA,
  _swigc__p_p_TPM2B_CONTEXT_DATA_ARRAY,
  _swigc__p_p_TPM2B_CONTEXT_SENSITIVE,
  _swigc__p_p_TPM2B_CONTEXT_SENSITIVE_ARRAY,
  _swigc__p_p_TPM2B_CREATION_DATA,
  _swigc__p_p_TPM2B_CREATION_DATA_ARRAY,
  _swigc__p_p_TPM2B_DATA,
  _swigc__p_p_TPM2B_DATA_ARRAY,
  _swigc__p_p_TPM2B_DIGEST,
  _swigc__p_p_TPM2B_DIGEST_ARRAY,
  _swigc__p_p_TPM2B_ECC_PARAMETER,
  _swigc__p_p_TPM2B_ECC_PARAMETER_ARRAY,
  _swigc__p_p_TPM2B_ECC_POINT,
  _swigc__p_p_TPM2B_ECC_POINT_ARRAY,
  _swigc__p_p_TPM2B_ENCRYPTED_SECRET,
  _swigc__p_p_TPM2B_ENCRYPTED_SECRET_ARRAY,
  _swigc__p_p_TPM2B_EVENT,
  _swigc__p_p_TPM2B_EVENT_ARRAY,
  _swigc__p_p_TPM2B_ID_OBJECT,
  _swigc__p_p_TPM2B_ID_OBJECT_ARRAY,
  _swigc__p_p_TPM2B_IV,
  _swigc__p_p_TPM2B_IV_ARRAY,
  _swigc__p_p_TPM2B_MAX_BUFFER,
  _swigc__p_p_TPM2B_MAX_BUFFER_ARRAY,
  _swigc__p_p_TPM2B_MAX_NV_BUFFER,
  _swigc__p_p_TPM2B_MAX_NV_BUFFER_ARRAY,
  _swigc__p_p_TPM2B_NAME,
  _swigc__p_p_TPM2B_NAME_ARRAY,
  _swigc__p_p_TPM2B_NV_PUBLIC,
  _swigc__p_p_TPM2B_NV_PUBLIC_ARRAY,
  _swigc__p_p_TPM2B_PRIVATE,
  _swigc__p_p_TPM2B_PRIVATE_ARRAY,
  _swigc__p_p_TPM2B_PRIVATE_KEY_RSA,
  _swigc__p_p_TPM2B_PRIVATE_KEY_RSA_ARRAY,
  _swigc__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC,
  _swigc__p_p_TPM2B_PRIVATE_VENDOR_SPECIFIC_ARRAY,
  _swigc__p_p_TPM2B_PUBLIC,
  _swigc__p_p_TPM2B_PUBLIC_ARRAY,
  _swigc__p_p_TPM2B_PUBLIC_KEY_RSA,
  _swigc__p_p_TPM2B_PUBLIC_KEY_RSA_ARRAY,
  _swigc__p_p_TPM2B_SENSITIVE,
  _swigc__p_p_TPM2B_SENSITIVE_ARRAY,
  _swigc__p_p_TPM2B_SENSITIVE_CREATE,
  _swigc__p_p_TPM2B_SENSITIVE_CREATE_ARRAY,
  _swigc__p_p_TPM2B_SENSITIVE_DATA,
  _swigc__p_p_TPM2B_SENSITIVE_DATA_ARRAY,
  _swigc__p_p_TPM2B_SYM_KEY,
  _swigc__p_p_TPM2B_SYM_KEY_ARRAY,
  _swigc__p_p_TPM2B_TEMPLATE,
  _swigc__p_p_TPM2B_TEMPLATE_ARRAY,
  _swigc__p_p_TPML_AC_CAPABILITIES,
  _swigc__p_p_TPML_AC_CAPABILITIES_ARRAY,
  _swigc__p_p_TPML_ALG,
  _swigc__p_p_TPML_ALG_ARRAY,
  _swigc__p_p_TPML_ALG_PROPERTY,
  _swigc__p_p_TPML_ALG_PROPERTY_ARRAY,
  _swigc__p_p_TPML_CC,
  _swigc__p_p_TPML_CCA,
  _swigc__p_p_TPML_CCA_ARRAY,
  _swigc__p_p_TPML_CC_ARRAY,
  _swigc__p_p_TPML_DIGEST,
  _swigc__p_p_TPML_DIGEST_ARRAY,
  _swigc__p_p_TPML_DIGEST_VALUES,
  _swigc__p_p_TPML_DIGEST_VALUES_ARRAY,
  _swigc__p_p_TPML_ECC_CURVE,
  _swigc__p_p_TPML_ECC_CURVE_ARRAY,
  _swigc__p_p_TPML_HANDLE,
  _swigc__p_p_TPML_HANDLE_ARRAY,
  _swigc__p_p_TPML_INTEL_PTT_PROPERTY,
  _swigc__p_p_TPML_INTEL_PTT_PROPERTY_ARRAY,
  _swigc__p_p_TPML_PCR_SELECTION,
  _swigc__p_p_TPML_PCR_SELECTION_ARRAY,
  _swigc__p_p_TPML_TAGGED_PCR_PROPERTY,
  _swigc__p_p_TPML_TAGGED_PCR_PROPERTY_ARRAY,
  _swigc__p_p_TPML_TAGGED_TPM_PROPERTY,
  _swigc__p_p_TPML_TAGGED_TPM_PROPERTY_ARRAY,
  _swigc__p_p_TPMS_AC_OUTPUT,
  _swigc__p_p_TPMS_AC_OUTPUT_ARRAY,
  _swigc__p_p_TPMS_ALGORITHM_DESCRIPTION,
  _swigc__p_p_TPMS_ALGORITHM_DESCRIPTION_ARRAY,
  _swigc__p_p_TPMS_ALGORITHM_DETAIL_ECC,
  _swigc__p_p_TPMS_ALGORITHM_DETAIL_ECC_ARRAY,
  _swigc__p_p_TPMS_ALG_PROPERTY,
  _swigc__p_p_TPMS_ALG_PROPERTY_ARRAY,
  _swigc__p_p_TPMS_ASYM_PARMS,
  _swigc__p_p_TPMS_ASYM_PARMS_ARRAY,
  _swigc__p_p_TPMS_ATTEST,
  _swigc__p_p_TPMS_ATTEST_ARRAY,
  _swigc__p_p_TPMS_AUTH_COMMAND,
  _swigc__p_p_TPMS_AUTH_COMMAND_ARRAY,
  _swigc__p_p_TPMS_AUTH_RESPONSE,
  _swigc__p_p_TPMS_AUTH_RESPONSE_ARRAY,
  _swigc__p_p_TPMS_CAPABILITY_DATA,
  _swigc__p_p_TPMS_CAPABILITY_DATA_ARRAY,
  _swigc__p_p_TPMS_CERTIFY_INFO,
  _swigc__p_p_TPMS_CERTIFY_INFO_ARRAY,
  _swigc__p_p_TPMS_CLOCK_INFO,
  _swigc__p_p_TPMS_CLOCK_INFO_ARRAY,
  _swigc__p_p_TPMS_COMMAND_AUDIT_INFO,
  _swigc__p_p_TPMS_COMMAND_AUDIT_INFO_ARRAY,
  _swigc__p_p_TPMS_CONTEXT,
  _swigc__p_p_TPMS_CONTEXT_ARRAY,
  _swigc__p_p_TPMS_CONTEXT_DATA,
  _swigc__p_p_TPMS_CONTEXT_DATA_ARRAY,
  _swigc__p_p_TPMS_CREATION_DATA,
  _swigc__p_p_TPMS_CREATION_DATA_ARRAY,
  _swigc__p_p_TPMS_CREATION_INFO,
  _swigc__p_p_TPMS_CREATION_INFO_ARRAY,
  _swigc__p_p_TPMS_ECC_PARMS,
  _swigc__p_p_TPMS_ECC_PARMS_ARRAY,
  _swigc__p_p_TPMS_ECC_POINT,
  _swigc__p_p_TPMS_ECC_POINT_ARRAY,
  _swigc__p_p_TPMS_EMPTY,
  _swigc__p_p_TPMS_EMPTY_ARRAY,
  _swigc__p_p_TPMS_ID_OBJECT,
  _swigc__p_p_TPMS_ID_OBJECT_ARRAY,
  _swigc__p_p_TPMS_KEYEDHASH_PARMS,
  _swigc__p_p_TPMS_KEYEDHASH_PARMS_ARRAY,
  _swigc__p_p_TPMS_NV_CERTIFY_INFO,
  _swigc__p_p_TPMS_NV_CERTIFY_INFO_ARRAY,
  _swigc__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS,
  _swigc__p_p_TPMS_NV_PIN_COUNTER_PARAMETERS_ARRAY,
  _swigc__p_p_TPMS_NV_PUBLIC,
  _swigc__p_p_TPMS_NV_PUBLIC_ARRAY,
  _swigc__p_p_TPMS_PCR_SELECT,
  _swigc__p_p_TPMS_PCR_SELECTION,
  _swigc__p_p_TPMS_PCR_SELECTION_ARRAY,
  _swigc__p_p_TPMS_PCR_SELECT_ARRAY,
  _swigc__p_p_TPMS_QUOTE_INFO,
  _swigc__p_p_TPMS_QUOTE_INFO_ARRAY,
  _swigc__p_p_TPMS_RSA_PARMS,
  _swigc__p_p_TPMS_RSA_PARMS_ARRAY,
  _swigc__p_p_TPMS_SCHEME_ECDAA,
  _swigc__p_p_TPMS_SCHEME_ECDAA_ARRAY,
  _swigc__p_p_TPMS_SCHEME_HASH,
  _swigc__p_p_TPMS_SCHEME_HASH_ARRAY,
  _swigc__p_p_TPMS_SCHEME_XOR,
  _swigc__p_p_TPMS_SCHEME_XOR_ARRAY,
  _swigc__p_p_TPMS_SENSITIVE_CREATE,
  _swigc__p_p_TPMS_SENSITIVE_CREATE_ARRAY,
  _swigc__p_p_TPMS_SESSION_AUDIT_INFO,
  _swigc__p_p_TPMS_SESSION_AUDIT_INFO_ARRAY,
  _swigc__p_p_TPMS_SIGNATURE_ECC,
  _swigc__p_p_TPMS_SIGNATURE_ECC_ARRAY,
  _swigc__p_p_TPMS_SIGNATURE_RSA,
  _swigc__p_p_TPMS_SIGNATURE_RSA_ARRAY,
  _swigc__p_p_TPMS_SYMCIPHER_PARMS,
  _swigc__p_p_TPMS_SYMCIPHER_PARMS_ARRAY,
  _swigc__p_p_TPMS_TAGGED_PCR_SELECT,
  _swigc__p_p_TPMS_TAGGED_PCR_SELECT_ARRAY,
  _swigc__p_p_TPMS_TAGGED_POLICY,
  _swigc__p_p_TPMS_TAGGED_POLICY_ARRAY,
  _swigc__p_p_TPMS_TAGGED_PROPERTY,
  _swigc__p_p_TPMS_TAGGED_PROPERTY_ARRAY,
  _swigc__p_p_TPMS_TIME_ATTEST_INFO,
  _swigc__p_p_TPMS_TIME_ATTEST_INFO_ARRAY,
  _swigc__p_p_TPMS_TIME_INFO,
  _swigc__p_p_TPMS_TIME_INFO_ARRAY,
  _swigc__p_p_TPMT_ASYM_SCHEME,
  _swigc__p_p_TPMT_ASYM_SCHEME_ARRAY,
  _swigc__p_p_TPMT_ECC_SCHEME,
  _swigc__p_p_TPMT_ECC_SCHEME_ARRAY,
  _swigc__p_p_TPMT_HA,
  _swigc__p_p_TPMT_HA_ARRAY,
  _swigc__p_p_TPMT_KDF_SCHEME,
  _swigc__p_p_TPMT_KDF_SCHEME_ARRAY,
  _swigc__p_p_TPMT_KEYEDHASH_SCHEME,
  _swigc__p_p_TPMT_KEYEDHASH_SCHEME_ARRAY,
  _swigc__p_p_TPMT_PUBLIC,
  _swigc__p_p_TPMT_PUBLIC_ARRAY,
  _swigc__p_p_TPMT_PUBLIC_PARMS,
  _swigc__p_p_TPMT_PUBLIC_PARMS_ARRAY,
  _swigc__p_p_TPMT_RSA_DECRYPT,
  _swigc__p_p_TPMT_RSA_DECRYPT_ARRAY,
  _swigc__p_p_TPMT_RSA_SCHEME,
  _swigc__p_p_TPMT_RSA_SCHEME_ARRAY,
  _swigc__p_p_TPMT_SENSITIVE,
  _swigc__p_p_TPMT_SENSITIVE_ARRAY,
  _swigc__p_p_TPMT_SIGNATURE,
  _swigc__p_p_TPMT_SIGNATURE_ARRAY,
  _swigc__p_p_TPMT_SIG_SCHEME,
  _swigc__p_p_TPMT_SIG_SCHEME_ARRAY,
  _swigc__p_p_TPMT_SYM_DEF,
  _swigc__p_p_TPMT_SYM_DEF_ARRAY,
  _swigc__p_p_TPMT_SYM_DEF_OBJECT,
  _swigc__p_p_TPMT_SYM_DEF_OBJECT_ARRAY,
  _swigc__p_p_TPMT_TK_AUTH,
  _swigc__p_p_TPMT_TK_AUTH_ARRAY,
  _swigc__p_p_TPMT_TK_CREATION,
  _swigc__p_p_TPMT_TK_CREATION_ARRAY,
  _swigc__p_p_TPMT_TK_HASHCHECK,
  _swigc__p_p_TPMT_TK_HASHCHECK_ARRAY,
  _swigc__p_p_TPMT_TK_VERIFIED,
  _swigc__p_p_TPMT_TK_VERIFIED_ARRAY,
  _swigc__p_p_TPMU_ASYM_SCHEME,
  _swigc__p_p_TPMU_ASYM_SCHEME_ARRAY,
  _swigc__p_p_TPMU_ATTEST,
  _swigc__p_p_TPMU_ATTEST_ARRAY,
  _swigc__p_p_TPMU_CAPABILITIES,
  _swigc__p_p_TPMU_CAPABILITIES_ARRAY,
  _swigc__p_p_TPMU_ENCRYPTED_SECRET,
  _swigc__p_p_TPMU_ENCRYPTED_SECRET_ARRAY,
  _swigc__p_p_TPMU_HA,
  _swigc__p_p_TPMU_HA_ARRAY,
  _swigc__p_p_TPMU_KDF_SCHEME,
  _swigc__p_p_TPMU_KDF_SCHEME_ARRAY,
  _swigc__p_p_TPMU_NAME,
  _swigc__p_p_TPMU_NAME_ARRAY,
  _swigc__p_p_TPMU_PUBLIC_ID,
  _swigc__p_p_TPMU_PUBLIC_ID_ARRAY,
  _swigc__p_p_TPMU_PUBLIC_PARMS,
  _swigc__p_p_TPMU_PUBLIC_PARMS_ARRAY,
  _swigc__p_p_TPMU_SCHEME_KEYEDHASH,
  _swigc__p_p_TPMU_SCHEME_KEYEDHASH_ARRAY,
  _swigc__p_p_TPMU_SENSITIVE_COMPOSITE,
  _swigc__p_p_TPMU_SENSITIVE_COMPOSITE_ARRAY,
  _swigc__p_p_TPMU_SIGNATURE,
  _swigc__p_p_TPMU_SIGNATURE_ARRAY,
  _swigc__p_p_TPMU_SIG_SCHEME,
  _swigc__p_p_TPMU_SIG_SCHEME_ARRAY,
  _swigc__p_p_TPMU_SYM_KEY_BITS,
  _swigc__p_p_TPMU_SYM_KEY_BITS_ARRAY,
  _swigc__p_p_TPMU_SYM_MODE,
  _swigc__p_p_TPMU_SYM_MODE_ARRAY,
  _swigc__p_p_TSS2_TCTI_CONTEXT,
  _swigc__p_p_TSS2_TCTI_POLL_HANDLE,
  _swigc__p_p__PRIVATE,
  _swigc__p_p__PRIVATE_ARRAY,
  _swigc__p_p_char,
  _swigc__p_p_size_t,
  _swigc__p_p_unsigned_char,
  _swigc__p_p_unsigned_int,
  _swigc__p_p_void,
  _swigc__p_short,
  _swigc__p_signed_char,
  _swigc__p_size_t,
  _swigc__p_uint16_t_PTR,
  _swigc__p_uint32_t_PTR,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
  _swigc__p_unsigned_long_long,
  _swigc__p_unsigned_short,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %d\n", swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
    return PyUnicode_InternFromString("<Swig global variables>");
#else
    return PyString_FromString("<Swig global variables>");
#endif
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *str = PyUnicode_InternFromString("(");
    PyObject *tail;
    PyObject *joined;
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      tail = PyUnicode_FromString(var->name);
      joined = PyUnicode_Concat(str, tail);
      Py_DecRef(str);
      Py_DecRef(tail);
      str = joined;
      if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
      }
    }
    tail = PyUnicode_InternFromString(")");
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
#else
    PyObject *str = PyString_FromString("(");
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
    return str;
  }
  
  SWIGINTERN int
  swig_varlink_print(swig_varlinkobject *v, FILE *fp, int SWIGUNUSEDPARM(flags)) {
    char *tmp;
    PyObject *str = swig_varlink_str(v);
    fprintf(fp,"Swig global variables ");
    fprintf(fp,"%s\n", tmp = SWIG_Python_str_AsChar(str));
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(str);
    return 0;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;
    if (!type_init) {
      const PyTypeObject tmp = {
        /* PyObject header changed in Python 3 */
#if PY_VERSION_HEX >= 0x03000000
        PyVarObject_HEAD_INIT(NULL, 0)
#else
        PyObject_HEAD_INIT(NULL)
        0,                                  /* ob_size */
#endif
        (char *)"swigvarlink",              /* tp_name */
        sizeof(swig_varlinkobject),         /* tp_basicsize */
        0,                                  /* tp_itemsize */
        (destructor) swig_varlink_dealloc,  /* tp_dealloc */
        (printfunc) swig_varlink_print,     /* tp_print */
        (getattrfunc) swig_varlink_getattr, /* tp_getattr */
        (setattrfunc) swig_varlink_setattr, /* tp_setattr */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc) swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
        0,                                  /* tp_del */
#endif
#if PY_VERSION_HEX >= 0x02060000
        0,                                  /* tp_version_tag */
#endif
#if PY_VERSION_HEX >= 0x03040000
        0,                                  /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
        0,                                  /* tp_allocs */
        0,                                  /* tp_frees */
        0,                                  /* tp_maxalloc */
#if PY_VERSION_HEX >= 0x02050000
        0,                                  /* tp_prev */
#endif
        0                                   /* tp_next */
#endif
      };
      varlink_type = tmp;
      type_init = 1;
#if PY_VERSION_HEX < 0x02020000
      varlink_type.ob_type = &PyType_Type;
#else
      if (PyType_Ready(&varlink_type) < 0)
      return NULL;
#endif
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        strncpy(gv->name,name,size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *_SWIG_globals = 0; 
    if (!_SWIG_globals) _SWIG_globals = SWIG_newvarlink();  
    return _SWIG_globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              strncpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              strncpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md;
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
# if PY_VERSION_HEX >= 0x03020000
    PyModuleDef_HEAD_INIT,
# else
    {
      PyObject_HEAD_INIT(NULL)
      NULL, /* m_init */
      0,    /* m_index */
      NULL, /* m_copy */
    },
# endif
    (char *) SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    (PyCFunction) SwigPyObject_own,
    (PyCFunction) SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyObject *metatype_args;
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* metatype is used to implement static member variables. */
  metatype_args = Py_BuildValue("(s(O){})", "SwigPyObjectType", &PyType_Type);
  assert(metatype_args);
  metatype = (PyTypeObject *) PyType_Type.tp_call((PyObject *) &PyType_Type, metatype_args, NULL);
  assert(metatype);
  Py_DECREF(metatype_args);
  metatype->tp_setattro = (setattrofunc) &SwigPyObjectType_setattro;
  assert(PyType_Ready(metatype) >= 0);
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule((char *) SWIG_name, SwigMethods);
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = SwigPyObject_TypeOnce();
  } else if (SwigPyObject_TypeOnce()->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

